/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#pragma once

#include <iostream>
#include "lib_everest/lib_everest.h"
#include "lib_utils/tools.h"

class SynthHost : public EVEREST::IHost
{
public:
  SynthHost(int iBufferSize)
  {
    SDL_AudioSpec spec;
    spec.freq = 44100;
    spec.format = AUDIO_S16;
    spec.channels = 2;
    spec.samples = iBufferSize;
    spec.callback = &SynthHost::MixAudio;
    spec.userdata = this;
    SDL_OpenAudio(&spec, &m_AudioSpec);

    std::cerr << "Audio: " << m_AudioSpec.freq << "Hz, latency=" << (iBufferSize * 1000 / m_AudioSpec.freq) << "ms" << std::endl;

    SDL_PauseAudio(0);
  }

  ~SynthHost()
  {
    SDL_PauseAudio(1);
    SDL_CloseAudio();
  }

  void SetSynth(EVEREST::IPlugin* pSynth)
  {
    pSynth->SetSampleRate(m_AudioSpec.freq);
    m_pSynth = pSynth;
  }

  void MixAudio(Uint8* stream, int len)
  {
    int iNumSamples = len / 4;
    m_LeftAudioBuffer.resize(iNumSamples);
    m_RightAudioBuffer.resize(iNumSamples);
    float* buffers[] =
    {
      &m_LeftAudioBuffer[0], &m_RightAudioBuffer[0]
    };

    if(m_pSynth)
      m_pSynth->ProcessAudio(NULL, buffers, iNumSamples);

    for(auto i = 0; i < iNumSamples; ++i)
    {
      auto iSampleL = int(m_LeftAudioBuffer[i] * m_Volume * 32767.0f);
      auto iSampleR = int(m_RightAudioBuffer[i] * m_Volume * 32767.0f);

      iSampleL = Clamp(iSampleL, -32768, 32767);
      iSampleR = Clamp(iSampleR, -32768, 32767);

      stream[i * 4 + 0] = (iSampleL >> 0) & 0xFF;
      stream[i * 4 + 1] = (iSampleL >> 8) & 0xFF;
      stream[i * 4 + 2] = (iSampleR >> 0) & 0xFF;
      stream[i * 4 + 3] = (iSampleR >> 8) & 0xFF;
    }
  }

  void OnParamChanged(int, float)
  {
  }

  static void MixAudio(void* userdata, Uint8* stream, int len)
  {
    SynthHost* pThis = reinterpret_cast<SynthHost*>(userdata);
    pThis->MixAudio(stream, len);
  }

private:
  std::vector<float> m_LeftAudioBuffer;
  std::vector<float> m_RightAudioBuffer;
  float m_Volume = 1.0f;
  EVEREST::IPlugin* m_pSynth = nullptr;
  SDL_AudioSpec m_AudioSpec;
};

