// A very simple test program for instanciating an Everest plugin.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include <iostream>
#include <algorithm>
#include <cmath>
#include <queue>
#include <string>
#include <sstream>
#include <fstream>
#include <stdexcept>

#define SDL_MAIN_HANDLED // prevent redefinition of 'main'
#include "SDL.h"
#include "SDL2_gfxPrimitives.h"

#include "lib_rtmidi/RtMidi.h"

#include "lib_utils/math_tools.h"
#include "lib_utils/tools.h"

#include "host.h"

using namespace std;

const int WIDTH = 1100;
const int HEIGHT = 600;

const float PI = 3.14159275f;
const int KNOB_HORZ_SPACING = 170;
const int KNOB_VERT_SPACING = 40;
const int KNOB_RADIUS = 16;
const Uint32 KNOB_COLOR = 0xFF888888;

class Renderer
{
public:
  Renderer(EVEREST::IPlugin* synth) :
    m_pSynth(synth)
  {
    m_DragParamIndex = -1;

    SDL_Init(SDL_INIT_EVERYTHING);
    m_Window = SDL_CreateWindow("Synth", 0, 0, WIDTH, HEIGHT, SDL_WINDOW_BORDERLESS);
    m_Screen = SDL_CreateRenderer(m_Window, -1, 0);

    // layout knobs
    for(int i = 0; i < m_pSynth->GetNumParams(); ++i)
    {
      Knob knob;
      knob.m_ParamIndex = i;
      knob.m_Name = m_pSynth->GetParamName(i);
      m_Knobs.push_back(knob);
    }

    sort(m_Knobs.begin(), m_Knobs.end());

    string prevCat;
    int row = 40;
    int col = 50;

    for(auto& knob : m_Knobs)
    {
      if(prevCat != "" && prevCat != getCategory(knob.m_Name))
      {
        if(row > 400)
        {
          row = 40;
          col += KNOB_HORZ_SPACING;
        }
        else
        {
          row += KNOB_VERT_SPACING;
        }
      }

      knob.m_Pos.x = col;
      knob.m_Pos.y = row;

      row += KNOB_VERT_SPACING;
      prevCat = getCategory(knob.m_Name);
    }
  }

  static string getCategory(string label)
  {
    auto i = label.rfind('.');

    if(i == string::npos)
      return "";

    return label.substr(0, i);
  }

  ~Renderer()
  {
    SDL_Quit();
  }

  void DrawKnob(int x, int y, float value, string const& sLabel, string const& sValue)
  {
    float angle = 2.0f * PI * 0.75f * value - 5.0f * PI / 4.0f;
    filledCircleColor(m_Screen, x, y, KNOB_RADIUS, KNOB_COLOR);
    filledCircleColor(m_Screen, x, y, int(KNOB_RADIUS * 0.85), 0xFFAAAAAA);
    int x2 = x + cos(angle) * KNOB_RADIUS;
    int y2 = y + sin(angle) * KNOB_RADIUS;
    lineColor(m_Screen, x, y, x2, y2, 0xff000000);
    stringColor(m_Screen, x + KNOB_RADIUS + 10, y - 10, sLabel.c_str(), 0xFFFFFFFF);
    stringColor(m_Screen, x + KNOB_RADIUS + 10, y, sValue.c_str(), KNOB_COLOR);
  }

  void DrawScreen()
  {
    setColor(0x33333333);
    SDL_RenderFillRect(m_Screen, 0);

    // draw knobs
    for(auto& knob : m_Knobs)
    {
      int i = knob.m_ParamIndex;
      auto value = m_pSynth->GetParam(i);
      auto sValue = m_pSynth->GetParamText(i);

      DrawKnob(knob.m_Pos.x, knob.m_Pos.y, value, knob.m_Name, sValue);
    }

    const auto border = 4;
    rectangleColor(m_Screen, border, border, WIDTH - border, HEIGHT - border, KNOB_COLOR);
    SDL_RenderPresent(m_Screen);
  }

  void OnLeftButtonDown(int x, int y)
  {
    m_DragParamIndex = -1;

    for(auto& knob : m_Knobs)
    {
      int dx = knob.m_Pos.x - x;
      int dy = knob.m_Pos.y - y;
      float squaredDist = dx * dx + dy * dy;

      if(squaredDist < KNOB_RADIUS * KNOB_RADIUS)
      {
        m_DragParamIndex = knob.m_ParamIndex;
        break;
      }
    }

    if(m_DragParamIndex == -1)
      return;

    m_fInitialParamValue = m_pSynth->GetParam(m_DragParamIndex);
    m_DragStartY = y;
  }

  void OnLeftButtonUp(int, int)
  {
    m_DragParamIndex = -1;
  }

  void OnMouseMove(int /*x*/, int y)
  {
    if(m_DragParamIndex == -1)
      return;

    float fParamValue = m_fInitialParamValue + -(y - m_DragStartY) * 0.01;
    fParamValue = Clamp(fParamValue, 0.0f, 1.0f);
    // cout << "SetParam('" << m_pSynth->GetParamName(m_DragParamIndex) << "', " << fParamValue << ")" << endl;
    m_pSynth->SetParam(m_DragParamIndex, fParamValue);
  }

private:
  class Knob
  {
public:
    int m_ParamIndex;
    string m_Name;
    SDL_Rect m_Pos;

    bool operator < (Knob const& other) const
    {
      return m_Name.compare(other.m_Name) < 0;
    }
  };

  void setColor(Uint32 color)
  {
    auto const r = (color >> 24) & 0xff;
    auto const g = (color >> 16) & 0xff;
    auto const b = (color >> 8) & 0xff;
    auto const a = (color >> 0) & 0xff;
    SDL_SetRenderDrawColor(m_Screen, r, g, b, a);
  }

  SDL_Renderer* m_Screen;
  SDL_Window* m_Window;
  vector<Knob> m_Knobs;
  EVEREST::IPlugin* const m_pSynth;

  // dragging info
  int m_DragParamIndex;
  int m_DragStartY;
  float m_fInitialParamValue;
};

void setParam(EVEREST::IPlugin* plugin, string const& sParamName, float fValue)
{
  for(int i = 0; i < plugin->GetNumParams(); ++i)
  {
    if(sParamName == plugin->GetParamName(i))
    {
      cout << "Set " << sParamName << " to " << fValue << endl;
      plugin->SetParam(i, fValue);
      break;
    }
  }
}

void SaveState(EVEREST::IPlugin* plugin)
{
  ofstream fp("saved_state.txt");

  for(int i = 0; i < plugin->GetNumParams(); ++i)
  {
    fp << plugin->GetParamName(i) << " " << plugin->GetParam(i) << endl;
  }
}

void LoadState(EVEREST::IPlugin* plugin)
{
  ifstream fp("saved_state.txt");

  while(fp.good())
  {
    string name;
    fp >> name;

    if(name == "")
      break;

    float fValue;
    fp >> fValue;
    setParam(plugin, name, fValue);
  }
}

void printMidiInputs(RtMidiIn* midiin)
{
  // Check inputs.
  unsigned int nPorts = midiin->getPortCount();

  for(unsigned int i = 0; i < nPorts; i++)
  {
    auto portName = midiin->getPortName(i);
    cout << "  Input Port #" << i << ": " << portName << '\n';
  }
}

void printParameters(EVEREST::IPlugin* plugin)
{
  for(int i = 0; i < plugin->GetNumParams(); ++i)
  {
    cout << "#" << i << " : '" << plugin->GetParamName(i) << "' : " << plugin->GetParam(i) << endl;
  }
}

void printUsage()
{
  cerr << "Usage: exe_synth [--port <port>] [--buffer <size>]" << endl;
}

ostream & operator << (ostream& o, vector<unsigned char> const& v)
{
  o << "[";

  for(size_t i = 0; i < v.size(); ++i)
  {
    if(i > 0)
      o << " ";

    char s[256];
    sprintf(s, "%2X", v[i]);
    o << s;
  }

  o << "]";

  return o;
}

enum
{
  EVENT_NOTE_ON,
  EVENT_NOTE_OFF,
  EVENT_PITCH_BEND,
  EVENT_CONTROL_CHANGE,
};

void midiCallback(double /*timeStamp*/, vector<unsigned char>* pMessage, void* /*userData*/)
{
  vector<unsigned char>& message = *pMessage;
  cout << "Message: " << message << endl;

  if(message.size() != 3)
    return;

  int iCommand = (message[0] & 0xF0) >> 4;
  int iChannel = (message[0] & 0x0F) >> 0;
  (void)iChannel;

  if(iCommand == 9)
  {
    int iNote = message[1];
    int iVel = message[2];

    SDL_Event e;
    e.type = SDL_USEREVENT;
    e.user.code = iVel > 0 ? EVENT_NOTE_ON : EVENT_NOTE_OFF;
    e.user.data1 = reinterpret_cast<void*>(iNote);
    e.user.data2 = reinterpret_cast<void*>(iVel);
    SDL_PushEvent(&e);
  }
  else if(iCommand == 8)
  {
    int iNote = message[1];
    int iVel = message[2];

    SDL_Event e;
    e.type = SDL_USEREVENT;
    e.user.code = EVENT_NOTE_OFF;
    e.user.data1 = reinterpret_cast<void*>(iNote);
    e.user.data2 = reinterpret_cast<void*>(iVel);
    SDL_PushEvent(&e);
  }
  else if(iCommand == 0xB)
  {
    int iControl = message[1];
    int iValue = message[2];

    SDL_Event e;
    e.type = SDL_USEREVENT;
    e.user.code = EVENT_CONTROL_CHANGE;
    e.user.data1 = reinterpret_cast<void*>(iControl);
    e.user.data2 = reinterpret_cast<void*>(iValue);
    SDL_PushEvent(&e);
  }
  else if(iCommand == 0xE)
  {
    int iBend = message[1] + message[2] * 256 - 16384;

    SDL_Event e;
    e.type = SDL_USEREVENT;
    e.user.code = EVENT_PITCH_BEND;
    e.user.data1 = reinterpret_cast<void*>(iBend);
    SDL_PushEvent(&e);
  }
  else
    cout << "Unknown command : " << iCommand << "(" << message << ")" << endl;
}

void stl_main(queue<string> args)
{
  cout << "exe_synth - Copyright (C) 2016 - Sebastien Alaiwan" << endl;

  int iPort = 4;
  int iBufferSize = 256;

  while(!args.empty())
  {
    auto arg = args.front();
    args.pop();

    if(arg == "--port" || arg == "-p")
    {
      if(args.empty())
        throw runtime_error("Command line error: excepted port number");

      auto sNum = args.front();
      args.pop();
      stringstream ss(sNum);
      ss >> iPort;
    }
    else if(arg == "--buffer" || arg == "-b")
    {
      if(args.empty())
        throw runtime_error("Command line error: excepted buffer size");

      auto sNum = args.front();
      args.pop();
      stringstream ss(sNum);
      ss >> iBufferSize;
    }
    else if(arg == "--help")
    {
      printUsage();
      return;
    }
  }

  SynthHost host(iBufferSize);
  auto synth = EVEREST::g_PlugDesc.create(&host);
  host.SetSynth(synth);

  RtMidiIn midiInput;
  printMidiInputs(&midiInput);

  if(iPort >= 0)
  {
    cout << "Opening port " << iPort << endl;
    midiInput.setCallback(&midiCallback, NULL);
    midiInput.openPort(iPort);
  }

  // Don't ignore sysex, timing. Discard active sensing messages.
  midiInput.ignoreTypes(false, false, true);

  // configure synthesizer
  setParam(synth, "Attack1", 0.01);
  setParam(synth, "Attack2", 0.01);
  printParameters(synth);

  Renderer renderer(synth);

  bool done = false;

  while(!done)
  {
    SDL_Event evt;

    if(!SDL_WaitEvent(&evt))
      continue;

    do
    {
      auto midi = dynamic_cast<EVEREST::IMidiListener*>(synth);
      switch(evt.type)
      {
      case SDL_QUIT:
        done = true;
        break;
      case SDL_KEYDOWN:

        if(evt.key.keysym.sym == SDLK_ESCAPE)
          done = true;
        else if(evt.key.keysym.sym == SDLK_F2)
        {
          SaveState(synth);
          cout << "Saved state" << endl;
        }
        else if(evt.key.keysym.sym == SDLK_F4)
        {
          LoadState(synth);
          cout << "Loaded state" << endl;
        }

        if(evt.key.keysym.sym == SDLK_SPACE)
          midi->controlChange(0, 1, 127);
        else if(evt.key.keysym.sym == SDLK_LCTRL)
          midi->noteOn(0, 60, 127);

        break;

      case SDL_KEYUP:

        if(evt.key.keysym.sym == SDLK_SPACE)
          midi->controlChange(0, 1, 0);
        else if(evt.key.keysym.sym == SDLK_LCTRL)
          midi->noteOff(0, 60, 127);

        break;
      case SDL_MOUSEBUTTONDOWN:
        renderer.OnLeftButtonDown(evt.button.x, evt.button.y);
        break;
      case SDL_MOUSEBUTTONUP:
        renderer.OnLeftButtonUp(evt.button.x, evt.button.y);
        break;
      case SDL_MOUSEMOTION:
        renderer.OnMouseMove(evt.motion.x, evt.motion.y);
        break;
      case SDL_USEREVENT:
        {
          int iNote = reinterpret_cast<long long>(evt.user.data1);
          int iVel = reinterpret_cast<long long>(evt.user.data2);
          switch(evt.user.code)
          {
          case EVENT_NOTE_ON:
            midi->noteOn(0, iNote, iVel);
            break;
          case EVENT_NOTE_OFF:
            midi->noteOff(0, iNote, iVel);
            break;
          case EVENT_PITCH_BEND:
            midi->pitchBend(0, float(iNote) / 16384.0f);
            break;
          case EVENT_CONTROL_CHANGE:
            midi->controlChange(0, iNote, iVel);
            break;
          }
        }
        break;
      }
    }
    while(SDL_PollEvent(&evt));

    renderer.DrawScreen();
  }
}

int main(int argc, char** argv)
{
  try
  {
    queue<string> args;

    for(int i = 1; i < argc; ++i)
      args.push(argv[i]);

    stl_main(args);

    return 0;
  }
  catch(runtime_error const& e)
  {
    cerr << "Runtime error : " << e.what() << endl;
    return 1;
  }
}

