/**
 * @file wav.d
 * @brief save WAV files.
 * @author Sebastien Alaiwan
 * @date 2012-08-02
 */

#include <cstdlib>
#include <cassert>
#include <iostream>
#include <fstream>
#include <vector>

void WriteValue(std::ostream& f, unsigned int iValue, int iNumBytes)
{
  for(int i = 0; i < iNumBytes; ++i)
  {
    auto const b = (unsigned char)((iValue >> (i * 8)) & 0xff);
    f.put(b);
  }
}

template<int N>
void WriteString(std::ostream& f, const char(&s)[N])
{
  assert(s[N - 1] == 0);

  for(size_t i = 0; i < N - 1; ++i) // -1 to skip the terminating zero
  {
    f.put(s[i]);
  }
}

void WriteWav(std::ostream& f, std::vector<short> const& samples, int sampleRate)
{
  auto const FORMAT_PCM = 1;
  auto const BytesPerSample = 2;
  auto const Channels = 1;
  auto const AudioSize = samples.size() * BytesPerSample;

  // RIFF chunk
  WriteString(f, "RIFF");
  WriteValue(f, AudioSize + 44 - 8, 4);

  {
    WriteString(f, "WAVE");

    // Format chunk
    WriteString(f, "fmt ");
    WriteValue(f, 16, 4); // Sub chunk size
    {
      WriteValue(f, FORMAT_PCM, 2);
      WriteValue(f, Channels, 2);
      WriteValue(f, sampleRate, 4);
      WriteValue(f, sampleRate * Channels * BytesPerSample, 4);
      WriteValue(f, 4, 2); // block align
      WriteValue(f, BytesPerSample * 8, 2);
    }

    // data chunk
    WriteString(f, "data");
    WriteValue(f, AudioSize, 4);
    {
      for(size_t i = 0; i < samples.size(); ++i)
      {
        WriteValue(f, samples[i], 2);
      }
    }
  }

  f.flush();
}

void WriteWav(std::string filename, std::vector<short> const& samples, int sampleRate)
{
  std::ofstream fp(filename.c_str(), std::ios::binary);
  WriteWav(fp, samples, sampleRate);
}

void WriteWav(std::string filename, std::vector<float> const& samples, int sampleRate)
{
  std::vector<short> output;
  output.reserve(samples.size() * 2); // output is 16 bit

  for(size_t i = 0; i < samples.size(); ++i)
  {
    assert(samples[i] >= -1.0f);
    assert(samples[i] <= +1.0f);
    auto const s = (short)(samples[i] * 32767.0);
    output.push_back(s);
  }

  WriteWav(filename, output, sampleRate);
}

