/**
 * @file main.cpp
 * @author Sebastien Alaiwan
 * @version 1.0
 * @date 2011-01-30
 */

/* Copyright (C)
 * 2011 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <cstdlib>
#include <ctime>
#include <memory>
#include <iostream>
#include "lib_echidna.h"

#include <sys/timeb.h>

struct DummyHost : EVEREST::IHost
{
  void OnParamChanged(int iIndex, float fValue) {}
};

class ProgressBar
{
public:
  ProgressBar() : lastShownPercent(-1)
  {
  }

  void update(int progress, int total)
  {
    int percent = (progress * 100) / total;

    if(percent != lastShownPercent)
    {
      std::cerr << "\r" << percent << "%";
      std::cerr.flush();
      lastShownPercent = percent;
    }
  }

  void finish()
  {
    update(1, 1);
    std::cerr << std::endl;
  }

private:
  int lastShownPercent;
};

int GetMilliCount()
{
  timeb tb;
  ftime(&tb);
  int nCount = tb.millitm + (tb.time & 0xfffff) * 1000;
  return nCount;
}

typedef EVEREST::IPlugin* PluginCreationFunction (EVEREST::IHost* pHost);

void ResetPluginParams(EVEREST::IPlugin* plugin)
{
  for(int i = 0; i < plugin->GetNumParams(); ++i)
  {
    plugin->SetParam(i, 0);
  }
}

void TestPluginQuality(EVEREST::IPlugin* plugin)
{
  plugin->SetSampleRate(44100.0f);

  static const unsigned int CHUNK_SIZE = 256;

  float input_chunk_left[CHUNK_SIZE];
  float input_chunk_right[CHUNK_SIZE];
  float output_chunk_left[CHUNK_SIZE];
  float output_chunk_right[CHUNK_SIZE];

  for(unsigned int i = 0; i < CHUNK_SIZE; ++i)
  {
    input_chunk_left[i] = 0;
    input_chunk_right[i] = 0;
    output_chunk_left[i] = 0;
    output_chunk_right[i] = 0;
  }

  float* inputs[] =
  {
    input_chunk_left, input_chunk_right
  };
  float* outputs[] =
  {
    output_chunk_left, output_chunk_right
  };

  for(int i = 0; i < 1024 * 100; ++i)
  {
    int iNote = i % 12 + 60;
    plugin->OnNoteOn(1, iNote, 100);
    plugin->OnTempoChange(150);
    plugin->ProcessAudio(inputs, outputs, CHUNK_SIZE);
    plugin->OnNoteOff(1, iNote, 0);
  }
}

void TestPluginPerformance(PluginCreationFunction plugCreation)
{
  DummyHost host;
  std::unique_ptr<EVEREST::IPlugin> plugin(plugCreation(&host));
  std::cout << "Testing plugin performance ... " << std::endl;
  plugin->SetSampleRate(44100.0f);
  static const unsigned int CHUNK_SIZE = 256;

  float input_chunk_left[CHUNK_SIZE];
  float input_chunk_right[CHUNK_SIZE];
  float output_chunk_left[CHUNK_SIZE];
  float output_chunk_right[CHUNK_SIZE];

  for(unsigned int i = 0; i < CHUNK_SIZE; ++i)
  {
    input_chunk_left[i] = 0;
    input_chunk_right[i] = 0;
    output_chunk_left[i] = 0;
    output_chunk_right[i] = 0;
  }

  float* inputs[] =
  {
    input_chunk_left, input_chunk_right
  };
  float* outputs[] =
  {
    output_chunk_left, output_chunk_right
  };

  int const NUM_CHUNKS = 1024 * 250;

  auto t1 = GetMilliCount();
  clock_t clock1 = clock();

  int lastShownPercent = -1;

  plugin->OnTempoChange(150);

  ProgressBar progressBar;

  for(int i = 0; i < NUM_CHUNKS; ++i)
  {
    int iNote = i % 12 + 60;

    plugin->OnNoteOn(1, iNote, 100);
    plugin->OnNoteOn(1, iNote + 1, 100);
    plugin->OnNoteOn(1, iNote + 2, 100);
    plugin->OnNoteOn(1, iNote + 3, 100);

    plugin->ProcessAudio(inputs, outputs, CHUNK_SIZE);

    plugin->OnNoteOff(1, iNote, 0);
    plugin->OnNoteOff(1, iNote + 1, 0);
    plugin->OnNoteOff(1, iNote + 2, 0);
    plugin->OnNoteOff(1, iNote + 3, 0);

    progressBar.update(i, NUM_CHUNKS);
  }

  progressBar.finish();

  auto t2 = GetMilliCount();
  clock_t clock2 = clock();

  int const NUM_SAMPLES = NUM_CHUNKS * CHUNK_SIZE;

  auto const dt = t2 - t1;
  auto const delta_clock = clock2 - clock1;
  std::cout << "Total time : " << dt << "ms" << std::endl;
  std::cout << "Total clocks : " << ((delta_clock * 1000) / CLOCKS_PER_SEC) << " ms" << std::endl;
  std::cout << "Mean processing frequency: " << NUM_SAMPLES / dt << " kHz (highest score: 1456 kHz)" << std::endl;
}

std::vector<float> GenerateSamplesWithRandomChunks(PluginCreationFunction plugCreation, int const iNumDesiredSamples, int seed)
{
  DummyHost host;
  std::unique_ptr<EVEREST::IPlugin> plugin(plugCreation(&host));
  std::vector<float> audio;

  srand(seed);
  plugin->SetSampleRate(44100.0f);
  plugin->OnTempoChange(150);
  plugin->OnNoteOn(1, 60, 100);
  int iTotalSamples = 0;

  while(iTotalSamples < iNumDesiredSamples)
  {
    int const iNumSamples = std::min(rand() % 2048 + 1, iNumDesiredSamples - iTotalSamples);

    std::vector<float> input_chunk_left(iNumSamples);
    std::vector<float> input_chunk_right(iNumSamples);
    std::vector<float> output_chunk_left(iNumSamples);
    std::vector<float> output_chunk_right(iNumSamples);

    float* inputs[] =
    {
      input_chunk_left.data(), input_chunk_right.data()
    };

    float* outputs[] =
    {
      output_chunk_left.data(), output_chunk_right.data()
    };

    plugin->ProcessAudio(inputs, outputs, iNumSamples);

    for(size_t i = 0; i < iNumSamples; ++i)
    {
      audio.push_back(output_chunk_left[i]);
    }

    iTotalSamples += iNumSamples;
  }

  plugin->OnNoteOff(1, 60, 100);

  return std::move(audio);
}

void TestPluginCorrectness(PluginCreationFunction plugCreation)
{
  ProgressBar progressBar;

  std::cout << "Testing plugin correctness ... " << std::endl;

  int const TEST_DURATION_IN_SAMPLES = 44100 * 10;

  auto const ref = GenerateSamplesWithRandomChunks(plugCreation, TEST_DURATION_IN_SAMPLES, 0);

  int const NUM_SEEDS = 10;

  for(int i = 0; i < NUM_SEEDS; ++i)
  {
    auto const test = GenerateSamplesWithRandomChunks(plugCreation, TEST_DURATION_IN_SAMPLES, i);

    if(ref != test)
    {
      std::cout << "Error with seed : " << i << std::endl;

      if(ref.size() != test.size())
      {
        std::cout << "ref.size() = " << ref.size() << std::endl;
        std::cout << "test.size() = " << test.size() << std::endl;
      }
    }

    progressBar.update(i, NUM_SEEDS);
  }

  progressBar.finish();
}

int main()
{
  TestPluginCorrectness(&LibEchidna::CreateInstance);
  TestPluginPerformance(&LibEchidna::CreateInstance);
  return 0;
}

