extern "C"
{
#include "jit/jit.h"
#include "frontend/ast.h"
}

#include <cassert>
#include <string>
#include <stdexcept>
#include <vector>
#include <map>
using namespace std;

#include "compiled_program.h"

namespace
{
struct Jitter
{
  jit_context_t context;
  vector<char> memory; // global variables

  jit_function_t process(jit_context_t context_, const Vector* constAst)
  {
    context = context_;

    jit_function_t mainFunc {};

    auto ast = const_cast<Vector*>(constAst);

    declareGlobals(ast);

    for(int i = 0; i < vec_len(ast); i++)
    {
      auto v = (Node*)vec_get(ast, i);

      if(v->kind == AST_FUNC)
      {
        auto func = onFunction(v);

        jit_function_compile(func);

        if(string(v->fname) == "main")
          mainFunc = func;
      }
      else if(v->kind == AST_DECL)
      {
      }
      else
      {
        assert(0);
      }
    }

    if(!mainFunc)
      throw runtime_error("no 'main' function");

    return mainFunc;
  }

  void declareGlobals(Vector* ast)
  {
    int globalSize = 0;

    for(int i = 0; i < vec_len(ast); i++)
    {
      auto v = (Node*)vec_get(ast, i);

      if(v->kind == AST_DECL)
        globalSize += v->declvar->ty->size;
    }

    memory.resize(globalSize);

    int offset = 0;

    for(int i = 0; i < vec_len(ast); i++)
    {
      auto v = (Node*)vec_get(ast, i);

      if(v->kind == AST_DECL)
      {
        v->declvar->user = memory.data() + offset;
        offset += v->declvar->ty->size;

        // initialize variable
        *((float*)v->declvar->user) = 0.0f;
      }
    }
  }

  void onCompound(Node* node)
  {
    for(int i = 0; i < vec_len(node->stmts); i++)
    {
      auto sub = (Node*)vec_get(node->stmts, i);
      onExpr(sub);
    }
  }

  jit_value_t onFloat32(float value)
  {
    jit_constant_t cons;
    cons.type = jit_type_float32;
    cons.un.float32_value = value;
    return jit_value_create_constant(F, &cons);
  }

  jit_value_t marker()
  {
    return onFloat32(3.1415f);
  }

  void onReturn(Node* node)
  {
    jit_value_t val = marker();

    if(node->retval)
      val = onExpr(node->retval);

    jit_insn_return(F, val);
  }

  jit_value_t onCall(Node* node)
  {
    vector<jit_value_t> argVal;

    for(int i = 0; i < vec_len(node->args); i++)
    {
      auto v = (Node*)vec_get(node->args, i);
      argVal.push_back(onExpr(v));
    }

    if(string(node->fname) == "sin")
    {
      assert(argVal.size() == 1);
      return jit_insn_sin(F, argVal[0]);
    }

    if(string(node->fname) == "round")
    {
      assert(argVal.size() == 1);
      return jit_insn_round(F, argVal[0]);
    }

    auto func = functions.find(node->fname);

    if(func == functions.end())
      throw runtime_error(string("no such function: '") + node->fname + "'");

    return jit_insn_call(F, nullptr,
                         func->second,
                         nullptr,
                         argVal.data(),
                         argVal.size(),
                         JIT_CALL_NOTHROW);
  }

  jit_value_t onLiteral(Node* node)
  {
    switch(node->ty->kind)
    {
    case KIND_FLOAT:
      return onFloat32(node->fval);
    case KIND_DOUBLE:
    case KIND_LDOUBLE:
    case KIND_ARRAY:
    case KIND_BOOL:
    case KIND_CHAR:
    case KIND_SHORT:
    case KIND_INT:
    case KIND_LONG:
    case KIND_LLONG:
      fprintf(stderr, "Unknown literal type %d\n", node->ty->kind);
      break;
    default:
      assert(0);
    }

    assert(0);
  }

  jit_value_t onLocalVariable(Node* node)
  {
    assert(node->user);
    return (jit_value_t)node->user;
  }

  jit_value_t onGlobalVariable(Node* node)
  {
    assert(node->user);
    void* address = node->user;

    jit_constant_t cons;
    cons.type = jit_type_void_ptr;
    cons.un.ptr_value = address;
    auto valueAddress = jit_value_create_constant(F, &cons);
    return jit_insn_load_relative(F, valueAddress, 0, jit_type_float32);
  }

  void onDeclaration(Node* node)
  {
    assert(!node->user);
    auto var = jit_value_create(F, jit_type_float32);

    if(node->declinit)
    {
      auto initNode = (Node*)vec_get(node->declinit, 0);
      auto initValue = onExpr(initNode->initval);
      jit_insn_store(F, var, initValue);
    }

    node->declvar->user = var;
  }

  void onIf(Node* node)
  {
    auto cond = onExpr(node->cond);

    jit_label_t label_exit = jit_label_undefined;
    jit_label_t label_false = jit_label_undefined;

    jit_insn_branch_if_not(F, cond, &label_false);

    assert(node->then);
    onExpr(node->then);
    jit_insn_branch(F, &label_exit);

    jit_insn_label(F, &label_false);

    if(node->els)
      onExpr(node->els);

    jit_insn_label(F, &label_exit);
  }

  jit_value_t onExpr(Node* node)
  {
    switch(node->kind)
    {
    // statements
    case AST_DECL:
      onDeclaration(node);
      return marker();
    case AST_RETURN:
      onReturn(node);
      return marker();
    case AST_COMPOUND_STMT:
      onCompound(node);
      return marker();
    case AST_IF:
      onIf(node);
      return marker();

    // expressions
    case AST_LITERAL:
      return onLiteral(node);
    case AST_LVAR:
      return onLocalVariable(node);
    case AST_GVAR:
      return onGlobalVariable(node);
    case AST_FUNCALL:
      return onCall(node);
    case AST_CONV:
      (void)node->operand->ty;
      (void)node->ty;
      return onExpr(node->operand);
    case '/':
      {
        auto leftVal = onExpr(node->left);
        auto rightVal = onExpr(node->right);
        return jit_insn_div(F, leftVal, rightVal);
      }
    case '-':
      {
        auto leftVal = onExpr(node->left);
        auto rightVal = onExpr(node->right);
        return jit_insn_sub(F, leftVal, rightVal);
      }
    case '+':
      {
        auto leftVal = onExpr(node->left);
        auto rightVal = onExpr(node->right);
        return jit_insn_add(F, leftVal, rightVal);
      }
    case '*':
      {
        auto leftVal = onExpr(node->left);
        auto rightVal = onExpr(node->right);
        return jit_insn_mul(F, leftVal, rightVal);
      }
    case '=':
      {
        auto rvalue = onExpr(node->right);
        storeTo(node->left, rvalue);
        return rvalue;
      }
    case OP_LE:
      {
        auto leftVal = onExpr(node->left);
        auto rightVal = onExpr(node->right);
        return jit_insn_le(F, leftVal, rightVal);
      }
    case '<':
      {
        auto leftVal = onExpr(node->left);
        auto rightVal = onExpr(node->right);
        return jit_insn_lt(F, leftVal, rightVal);
      }
    default:
      fprintf(stderr, "unknown kind: %d (%d)\n", node->kind, node->sourceLoc->line);
      assert(0);
      return marker();
    }
  }

  void storeTo(Node* dest, jit_value_t val)
  {
    if(dest->kind == AST_LVAR)
    {
      auto lvalue = (jit_value_t)dest->user;
      jit_insn_store(F, lvalue, val);
    }
    else if(dest->kind == AST_GVAR)
    {
      assert(dest->user);
      void* address = dest->user;

      jit_constant_t cons;
      cons.type = jit_type_void_ptr;
      cons.un.ptr_value = address;
      auto valueAddress = jit_value_create_constant(F, &cons);
      jit_insn_store_relative(F, valueAddress, 0, val);
    }
    else
    {
      fprintf(stderr, "unknown lvalue kind=%d\n", dest->kind);
      assert(0);
    }
  }

  jit_function_t onFunction(Node* func)
  {
    vector<jit_type_t> argTypes;

    for(int i = 0; i < vec_len(func->params); i++)
    {
      auto param = (Node*)vec_get(func->params, i);
      (void)param;
      argTypes.push_back(jit_type_float32); // param->ty
    }

    auto signature = jit_type_create_signature(
      jit_abi_cdecl,
      jit_type_float32,
      argTypes.data(),
      argTypes.size(),
      1);

    F = jit_function_create(context, signature);

    functions[func->fname] = F;

    for(int i = 0; i < (int)argTypes.size(); ++i)
    {
      auto param = (Node*)vec_get(func->params, i);
      param->user = (char*)jit_value_get_param(F, i);
    }

    onExpr(func->body);
    return F;
  }

private:
  map<string, jit_function_t> functions;
  jit_function_t F; // current function
};

struct CProgram : CompiledProgram
{
  CProgram(Vector* ast)
  {
    context = jit_context_create();

    if(0)
    {
      // dump ast
      for(int i = 0; i < vec_len(ast); i++)
      {
        auto v = (Node*)vec_get(ast, i);
        fprintf(stderr, "%s\n", node2s(v));
      }
    }

    jit_context_build_start(context);
    auto mainFunc = ctx.process(context, ast);
    jit_context_build_end(context);

    entryPoint = (FF)jit_function_to_closure(mainFunc);
  }

  ~CProgram()
  {
    jit_context_destroy(context);
  }

  float process(float input) override
  {
    return entryPoint(input);
  }

private:
  jit_context_t context;
  typedef float (* FF)(float);
  FF entryPoint;
  Jitter ctx;
};
}

unique_ptr<CompiledProgram> compileProgram(Vector* ast)
{
  return make_unique<CProgram>(ast);
}

