// Synthesizer engine.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */
#include <cassert>
#include "echidna_synth.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/fraction.h"
#include "lib_utils/fpu_util.h"
#include "lib_utils/tools.h"

Osc::SHAPE ParamToOscShape(float fValue)
{
  auto const fShapeRange = 1.1 / Osc::NUM_SHAPES;
  return static_cast<Osc::SHAPE>(fValue / fShapeRange);
}

Lfo::SHAPE ParamToShape(float fValue)
{
  auto const fShapeRange = 1.1 / Lfo::NUM_SHAPES;
  return static_cast<Lfo::SHAPE>(fValue / fShapeRange);
}

bool ParamToBool(float fValue)
{
  return fValue >= 0.5;
}

Fraction const g_Clocks[] =
{
  { 0, 0 },
  { 1, 64 },
  { 1, 32 },
  { 1, 24 },
  { 3, 64 },
  { 1, 16 },
  { 1, 12 },
  { 3, 32 },
  { 1, 8 },
  { 1, 6 },
  { 3, 16 },
  { 1, 4 },
  { 1, 3 },
  { 3, 8 },
  { 1, 2 },
  { 2, 3 },
  { 3, 4 },
// { 1, 1 },
// { 2, 1 },
// { 4, 1 },
// { 8, 1 },
// { 16, 1 },
};

auto const NUM_CLOCKS = ArrayLength(g_Clocks);

Fraction ToClock(float fValue)
{
  float const fStep = 1.01f / float(NUM_CLOCKS);
  int iClockIndex = int(fValue / fStep);
  assert(iClockIndex >= 0 && iClockIndex < NUM_CLOCKS);
  return g_Clocks[iClockIndex];
}

EchidnaSynth::EchidnaSynth()
  : m_Dispatcher(m_Voices, ArrayLength(m_Voices))
{
  m_fSampleRate = 44100.0;
  m_fTempo = 145.0f;
  m_iRemainingSamplesInChunk = 0;

  m_fModulationWheel = 0.0f;
}

void EchidnaSynth::SetSampleRate(float fSampleRate)
{
  m_fSampleRate = fSampleRate;

  for(auto& voice : m_Voices)
    voice.SetSampleRate(fSampleRate);

  for(auto& phaser : m_Phaser)
    phaser.SetSampleRate(fSampleRate);

  m_Delay[0].Initialize();
  m_Delay[1].Initialize();
}

int const SAMPLES_PER_CHUNK = 32;

void EchidnaSynth::NewChunk()
{
  float const osc1Pitch = floor(m_Values.osc1.Pitch);
  float const osc2Pitch = floor(m_Values.osc2.Pitch);

  // parameter acquisition
  for(auto& v : m_Voices)
  {
    v.SetADSR(0,
              m_Values.env1.Attack,
              m_Values.env1.Decay,
              m_Values.env1.Sustain,
              m_Values.env1.Release);

    v.SetADSR(1,
              m_Values.env2.Attack,
              m_Values.env2.Decay,
              m_Values.env2.Sustain,
              m_Values.env2.Release);

    v.SetFilterBalance(m_Values.m_FilterBalance);
    v.SetFilterPan(0, m_Values.filter1.Pan);
    v.SetFilterPan(1, m_Values.filter2.Pan);

    {
      // modulate filter cutoff
      float cutoff1 = m_Values.filter1.Cutoff * 100.0f;
      cutoff1 += m_Values.lfo1.ToFilter * m_Lfos[0].GetValue();
      cutoff1 += m_Values.lfo2.ToFilter * m_Lfos[1].GetValue();

      float cutoff2 = m_Values.filter2.Cutoff * 100.0f;
      cutoff2 += m_Values.lfo1.ToFilter * m_Lfos[0].GetValue();
      cutoff2 += m_Values.lfo2.ToFilter * m_Lfos[1].GetValue();

      v.SetBaseFilter(0, cutoff1, m_Values.filter1.Resonance * 100.0f);
      v.SetBaseFilter(1, cutoff2, m_Values.filter2.Resonance * 100.0f);
    }

    v.SetEnvToFilter(0, m_Values.env1.ToFilter);
    v.SetEnvToFilter(1, m_Values.env2.ToFilter);

    v.SetOscVol(m_Values.m_OscVolume);
    v.SetOscBalance(m_Values.m_OscBalance);
    v.SetOscRingMod(m_Values.m_OscRingMod);

    v.SetEnvToVol(0, m_Values.env1.ToVolume);
    v.SetEnvToVol(1, m_Values.env2.ToVolume);

    v.SetEnvToPitch(0, m_Values.env1.ToPitch);
    v.SetEnvToPitch(1, m_Values.env2.ToPitch);

    v.SetShape(0, ParamToOscShape(m_Values.osc1.Shape));
    v.SetShape(1, ParamToOscShape(m_Values.osc2.Shape));

    v.SetDetune(0, m_Values.osc1.Detune + osc1Pitch);
    v.SetDetune(1, m_Values.osc2.Detune + osc2Pitch);

    v.SetPulseWidth(0, m_Values.osc1.PulseWidth);
    v.SetPulseWidth(1, m_Values.osc2.PulseWidth);

    v.SetPortamento(exp(m_Values.m_Portamento) * SAMPLES_PER_CHUNK / m_fSampleRate);
  }

  // LFO processing
  UpdateLfos(0, m_Values.lfo1.Shape, m_Values.lfo1.Clock, m_Values.lfo1.Rate);
  UpdateLfos(1, m_Values.lfo2.Shape, m_Values.lfo2.Clock, m_Values.lfo2.Rate);

  UpdateDelay(0);
  UpdateDelay(1);

  // Routing destinations
// for(int i = 0; i < MAX_VOICES; ++i)
// {
// Voice& v = m_Voices[i];
// v.SetFrequency(220.0 * m_Lfos[0].GetValue() + 500.0);
// }

  float pitch[6];

  for(int i = 0; i < ArrayLength(pitch); ++i)
  {
    pitch[i] = m_Lfos[i].GetValue() * (m_Values.lfo1.ToPitch + m_fModulationWheel * m_Values.m_ModToLfo1ToPitch);
    pitch[i] += m_Values.m_ModToPitch * m_fModulationWheel;
  }

  for(auto& phaser : m_Phaser)
  {
    phaser.SetMix(m_Values.phaser.Send);
    phaser.SetFeedback(m_Values.phaser.Feedback);
    phaser.SetRate(m_Values.phaser.Freq);
    phaser.SetRange(800.0, 800.0 + m_Values.phaser.Depth);
  }

  for(auto& voice : m_Voices)
    voice.NewChunk(pitch);

  m_Dispatcher.SetMono(ParamToBool(m_Values.m_Mono));
}

void EchidnaSynth::UpdateDelay(int index)
{
  auto const delayClock = ToClock(m_Values.delay.clock);

  if(delayClock.num > 0)
  {
    // 1/4 means one beat.
    auto const delayFactor = ToFloat(delayClock) * 4.0;
    auto const beatSeconds = 60.0 / m_fTempo;
    auto const fDelaySeconds = beatSeconds * delayFactor;
    auto const delayTimeInSamples = int(m_fSampleRate * fDelaySeconds);
    m_Delay[index].SetTime(delayTimeInSamples);
    m_Delay[index].SetFeedback(m_Values.delay.feedback);
  }
  else
  {
    m_Delay[index].SetParams(m_Values.delay.time, m_Values.delay.feedback);
  }
}

void EchidnaSynth::UpdateLfos(int index, float lfoShape, float lfoClockParam, float lfoRate)
{
  m_Lfos[index].SetShape(ParamToShape(lfoShape));
  auto const lfoClock = ToClock(lfoClockParam);

  if(lfoClock.num > 0)
  {
    // 1/4 means one beat.
    auto const lfoFactor = ToFloat(lfoClock) * 4.0;
    auto const beatSeconds = 60.0 / m_fTempo;
    auto const lfoPeriodInSeconds = beatSeconds * lfoFactor;
    m_Lfos[index].SetRate(1.0f / lfoPeriodInSeconds);
  }
  else
  {
    m_Lfos[index].SetRate(lfoRate);
  }

  for(int i = 0; i < SAMPLES_PER_CHUNK; ++i)
  {
    m_Lfos[index].Work();
  }
}

static void silenceBuffer(AudioBuffer out)
{
  for(long i = 0; i < out.numSamples; ++i)
  {
    out.left[i] = 0;
    out.right[i] = 0;
  }
}

void EchidnaSynth::GenerateAudio(AudioBuffer out)
{
  silenceBuffer(out);

  for(auto& voice : m_Voices)
    voice.Mix(&out);

  // Apply delay.
  for(long i = 0; i < out.numSamples; ++i)
  {
    out.left[i] += m_Delay[0].Work(out.left[i] * m_Values.delay.send);
    out.right[i] += m_Delay[1].Work(out.right[i] * m_Values.delay.send);
  }

  // Apply phaser
  for(int i = 0; i < out.numSamples; ++i)
  {
    float phasedLeft = m_Phaser[0].Work(out.left[i]);
    out.left[i] = phasedLeft;

    float phasedRight = m_Phaser[1].Work(out.right[i]);
    out.right[i] = phasedRight;
  }

  // Apply volume
  for(long i = 0; i < out.numSamples; ++i)
  {
    out.left[i] *= m_Values.m_Volume;
    out.right[i] *= m_Values.m_Volume;
  }

  // Avoid clipping, slight distortion
  for(long i = 0; i < out.numSamples; ++i)
  {
    out.left[i] = atan(out.left[i]) / (MY_PI * 0.5f);
    out.right[i] = atan(out.right[i]) / (MY_PI * 0.5f);
  }
}

void EchidnaSynth::ProcessAudio(float** pOutput, long const iNumSamples)
{
  float* LeftOut = pOutput[0];
  float* RightOut = pOutput[1];

  NoDenormalScope flushDenormalsToZero;

  int iNumGeneratedSamples = 0;

  while(iNumGeneratedSamples < iNumSamples)
  {
    auto const iNumRemainingSamples = iNumSamples - iNumGeneratedSamples;

    if(m_iRemainingSamplesInChunk == 0)
    {
      NewChunk();
      m_iRemainingSamplesInChunk = SAMPLES_PER_CHUNK;
    }

    auto const n = std::min<long>(iNumRemainingSamples, m_iRemainingSamplesInChunk);
    AudioBuffer out;
    out.left = LeftOut;
    out.right = RightOut;
    out.numSamples = n;
    GenerateAudio(out);
    LeftOut += n;
    RightOut += n;
    m_iRemainingSamplesInChunk -= n;
    iNumGeneratedSamples += n;
  }
}

// midi events
void EchidnaSynth::OnTempoChange(float fTempo)
{
  m_fTempo = fTempo;
}

void EchidnaSynth::OnNoteOn(int iNote, int iVelocity)
{
  m_Dispatcher.NoteOn(iNote, iVelocity);
}

void EchidnaSynth::OnNoteOff(int iNote, int iVelocity)
{
  m_Dispatcher.NoteOff(iNote, iVelocity);
}

void EchidnaSynth::OnPitchBend(float fBend)
{
  for(auto& voice : m_Voices)
    voice.BendTo(fBend);
}

void EchidnaSynth::OnControlChange(int iControl, int iValue)
{
  if(iControl == 1)
    m_fModulationWheel = float(iValue) / 127.0f;
}

