// ADSR envelope class implementation. At the moment, the only difficulty
// here is to avoid changing brutaly the output level, and to correctly handle
// dynamic parameter changes.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */
#include "envelope.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/debug.h"

static const float SAMPLE_RATE = 44100.0f;

Envelope::Envelope()
{
  Initialize();
}

void Envelope::Initialize()
{
  m_fAttack = 0.01f;
  m_fAttackIncrement = 0;
  m_fDecay = 0.01f;
  m_fDecayIncrement = 0;
  m_fSustain = 0.3f;
  m_fRelease = 0.1f;
  m_fReleaseIncrement = 0;
  m_iState = DEAD;
  m_bHolding = false;
  m_fCurrLevel = 0.0f;
  m_fPos = 0.0f;
}

float Envelope::Work()
{
  float fTargetLevel = 0.0f;
  switch(m_iState)
  {
  case DEAD:
    {
      fTargetLevel = 0.0f;

      if(m_bHolding)
      {
        m_iState = ATTACKING;
        m_fPos = 0;

        // bypass smoothing.
        m_fCurrLevel = 0;
        return 0;
      }

      break;
    }
  case ATTACKING:
    {
      m_fPos += m_fAttackIncrement;
      fTargetLevel = m_fPos;

      if(m_fPos >= 1.0)
      {
        m_iState = DECAYING;
        m_fPos = 0;
        fTargetLevel = 1.0;
      }

      break;
    }
  case DECAYING:
    {
      m_fPos += m_fDecayIncrement;
      float fDecayLevel = 1.0f - m_fSustain;
      fTargetLevel = fDecayLevel * (1.0f - m_fPos) + m_fSustain;

      if(m_fPos >= 1.0f)
      {
        m_iState = SUSTAINING;
        m_fPos = 0;
        fTargetLevel = m_fSustain;
      }

      break;
    }
  case SUSTAINING:
    {
      fTargetLevel = m_fSustain;

      if(!m_bHolding)
      {
        m_iState = RELEASING;
        m_fPos = 0;
      }

      break;
    }
  case RELEASING:
    {
      m_fPos += m_fReleaseIncrement;
      fTargetLevel = (1.0f - m_fPos) * m_fSustain;

      if(m_fPos >= 1.0f)
      {
        m_iState = DEAD;
        m_fPos = 0;
        fTargetLevel = 0;
      }

      break;
    }
  }

  m_fCurrLevel = blend(fTargetLevel, m_fCurrLevel, 0.9f);
  float s = m_fCurrLevel;

  return s;
}

void Envelope::EnvAttack()
{
  m_iState = DEAD;
  m_bHolding = true;
}

void Envelope::EnvRelease()
{
  m_bHolding = false;
}

void Envelope::SetADSR(float fAttack, float fDecay, float fSustain, float fRelease)
{
  if(m_fAttack != fAttack)
  {
    m_fAttack = fAttack;
    m_fAttackIncrement = 1.0 / (fAttack * SAMPLE_RATE);
  }

  if(m_fDecay != fDecay)
  {
    m_fDecay = fDecay;
    m_fDecayIncrement = 1.0 / (fDecay * SAMPLE_RATE);
  }

  m_fSustain = fSustain;

  if(m_fRelease != fRelease)
  {
    m_fRelease = fRelease;
    m_fReleaseIncrement = 1.0 / (fRelease * SAMPLE_RATE);
  }
}

float Envelope::Value() const
{
  return m_fCurrLevel;
}

bool Envelope::IsDead() const
{
  return m_iState == DEAD && !m_bHolding;
}

