// Lowpass filter definition. This code is taken from LoopDub, thanks to its author!

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <math.h>
#include <stdio.h>
#include <string.h>
#include "lib_utils/math_tools.h"
#include "filter.h"

static const float PI = 3.141592654f;
inline bool FLTEQUAL(double a, double b)
{
  return fabs(a - b) <= 0.01;
}

Lowpass::Lowpass()
{
  Initialize();
}

void Lowpass::Initialize()
{
  // note: why does max cutoff of 22050 fall to zero db?

  cutoff = targetCutoff = 22000.0f;
  resonance = targetResonance = 1.0f;
  maxCutoff = 22000.0f;
  maxResonance = 40.0f; // 126.0f;
  minCutoff = 550.0f;
  minResonance = 0.1f;
  inertia = 0.2f;
  lastTargetC100 = 0;
  lastTargetR100 = 0;

  for(int i = 0; i < 5; ++i)
  {
    coef[i] = 0;
  }

  for(int i = 0; i < 4; ++i)
  {
    history[i] = 0;
  }

  m_bParamsChanged = false;

  cutoffPow = log10(maxCutoff - minCutoff);
  resonancePow = log10(maxResonance - minResonance);
}

void Lowpass::SetParams(float targetC100, float targetR100)
{
  targetC100 = Clamp(targetC100, 0.0f, 100.0f);
  targetR100 = Clamp(targetR100, 0.0f, 100.0f);

  if(!FLTEQUAL(lastTargetC100, targetC100) || !(FLTEQUAL(lastTargetR100, targetR100)))
  {
    targetCutoff = pow(targetC100 / 10.0f, cutoffPow) + minCutoff;
    targetResonance = pow(targetR100 / 10.0f, resonancePow) + minResonance;
    m_bParamsChanged = true;

    lastTargetC100 = targetC100;
    lastTargetR100 = targetR100;
  }
}

float Lowpass::Work(float sample)
{
  // go to target via inertia ratio
  const float C = blend(cutoff, targetCutoff, inertia);
  const float R = blend(resonance, targetResonance, inertia);

  if(m_bParamsChanged)
  {
    // Calculate filter coefficients (with interpolation)
    auto omega = 2.0f * PI * C / 44100.0f;
    auto sn = (float)sin(omega);
    auto cs = (float)cos(omega);
    auto alpha = sn / R;
    auto b0 = (1.0f - cs) / 2.0f;
    auto b1 = 1.0f - cs;
    auto b2 = (1.0f - cs) / 2.0f;
    auto a0 = 1.0f + alpha;
    auto a1 = -2.0f * cs;
    auto a2 = 1.0f - alpha;
    coef[0] = b0 / a0;
    coef[1] = b1 / a0;
    coef[2] = b2 / a0;
    coef[3] = -a1 / a0;
    coef[4] = -a2 / a0;
  }

  // perform filter function
  float temp_y = coef[0] * sample +
                 coef[1] * history[0] + // lx1 +
                 coef[2] * history[1] + // lx2 +
                 coef[3] * history[2] + // ly1 +
                 coef[4] * history[3];  // ly2;

  // shift buffer
  history[3] = history[2];
  history[2] = temp_y;
  history[1] = history[0];
  history[0] = sample;

  cutoff = C;
  resonance = R;

  if(FLTEQUAL(cutoff, targetCutoff) && FLTEQUAL(resonance, targetResonance))
    m_bParamsChanged = false;

  return Clamp(temp_y, -1.0f, +1.0f);
}

// ------------------------------------------------------------------------------

void Highpass::SetParams(float targetC100, float targetR100)
{
  m_Lowpass.SetParams(targetC100, targetR100);
}

float Highpass::Work(float sample)
{
  float lowPassed = m_Lowpass.Work(sample);
  return sample - lowPassed;
}

