// A moog filter LP implementation. Adapted from musicdsp.org, Many
// thanks to "toast" for his code and to Julius O. Smith III.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */
#include <cmath> // sin
#include "moogfilter.h"
#include "lib_utils/math_tools.h"

MoogFilter::MoogFilter()
{
  m_y1 = m_y2 = m_y3 = m_y4 = 0;
  oldy[0] = oldy[1] = oldy[2] = oldy[3] = 0;

  m_Freq = m_Res = 1;
  Update(0, 0);
}

void MoogFilter::Update(float freq, float res)
{
  if(freq == m_Freq && res == m_Res)
    return; // nothing to do

  m_Freq = freq;
  m_Res = res;

  m_p = m_Freq * (1.8f - 0.8f * m_Freq);
  m_k = 2.0f * sin(m_Freq * MY_PI * 0.5f) - 1.0f;

  auto const t1 = (1.f - m_p) * 1.386249f;
  auto const t2 = 12.f + t1 * t1;
  m_r = m_Res * (t2 + 6.f * t1) / (t2 - 6.f * t1);
}

float MoogFilter::Work(float input)
{
  // process input
  float y0 = input - m_r * m_y4;

  // Four cascaded onepole filters (bilinear transform)
  m_y1 = y0 * m_p + oldy[0] * m_p - m_k * m_y1;
  m_y2 = m_y1 * m_p + oldy[1] * m_p - m_k * m_y2;
  m_y3 = m_y2 * m_p + oldy[2] * m_p - m_k * m_y3;
  m_y4 = m_y3 * m_p + oldy[3] * m_p - m_k * m_y4;

  // Clipper band limited sigmoid
  m_y4 -= (m_y4 * m_y4 * m_y4) / 6.f;

  oldy[0] = y0;
  oldy[1] = m_y1;
  oldy[2] = m_y2;
  oldy[3] = m_y3;
  return m_y4;
}

void MoogFilter::setCutoff(float cutoff)
{
  auto const newFreq = Clamp(cutoff, 0.0f, 1.0f);
  Update(newFreq, m_Res);
}

void MoogFilter::setRes(float r)
{
  auto const newRes = Clamp(r, 0.0f, 1.0f);
  Update(m_Freq, newRes);
}

