// Phaser effect.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <cmath> // sin
#include "lib_utils/tools.h"
#include "lib_utils/math_tools.h"
#include "phaser.h"

AllpassDelay::AllpassDelay() :
  m_fDelay(0.f),
  m_fValue(0.f)
{
}

void AllpassDelay::SetDelay(float fDelay)
{
  m_fDelay = (1.f - fDelay) / (1.f + fDelay);
}

float AllpassDelay::Work(float input)
{
  float const y = Clamp(input * -m_fDelay + m_fValue, -1.0f, 1.0f);
  m_fValue = (y * m_fDelay + input);
  return y;
}

Phaser::Phaser()
{
  SetRange(440.f, 1600.f);
  SetRate(.5f);
}

void Phaser::SetSampleRate(float fSampleRate)
{
  m_fSampleRate = fSampleRate;
}

void Phaser::SetMix(float fMix)
{
  assert(!IsNan(fMix));
  m_Mix = fMix;
}

void Phaser::SetFeedback(float fb)
{
  assert(!IsNan(fb));
  m_Feedback = fb;
}

void Phaser::SetRange(float fMinHz, float fMaxHz)
{
  assert(!IsNan(fMinHz));
  assert(!IsNan(fMaxHz));
  m_fDelayMin = fMinHz / (m_fSampleRate / 2.f);
  m_fDelayMax = fMaxHz / (m_fSampleRate / 2.f);
}

void Phaser::SetRate(float rate)
{
  assert(!IsNan(rate));
  m_LfoInc = 2.f * MY_PI * (rate / m_fSampleRate);
}

float Phaser::UpdateLfo()
{
  auto const fLfoValue = (sin(m_LfoPhase) + 1.f) / 2.f;

  m_LfoPhase += m_LfoInc;

  if(m_LfoPhase >= MY_PI * 2.f)
    m_LfoPhase -= MY_PI * 2.f;

  return fLfoValue;
}

float Phaser::Work(float input)
{
  float const fLfoValue = UpdateLfo();

  // update filter coeffs
  float const fDelay = m_fDelayMin + (m_fDelayMax - m_fDelayMin) * fLfoValue;

  for(auto& delay : m_Delays)
    delay.SetDelay(fDelay);

  // calculate output
  float y = input + m_fValue * m_Feedback;

  for(auto& delay : m_Delays)
    y = delay.Work(y);

  m_fValue = y;

  return input + y * m_Mix;
}

