/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <cmath>
#include <stdio.h>
#include "lib_utils/math_tools.h"
#include "voice.h"
#include "lib_utils/debug.h"
#include "lib_utils/tools.h"

static double const LN_2 = std::log(2.0);

static float FrequencyToPitch(float fFrequency)
{
  return log(fFrequency) * 12.0 / LN_2;
}

static float PitchToFrequency(float fPitch)
{
  return exp(fPitch * LN_2 / 12.0);
}

Voice::Voice()
{
  m_bAlive = false;
  m_fSampleRate = 44100.0;

  for(auto& filter : m_Filter)
  {
    filter.setCutoff(1.0f);
    filter.setRes(0);
  }

  m_fBend = 0;
  m_fFilterBalance = 0.0f;
  m_fPortamentoPitch = 0;
}

void Voice::SetSampleRate(float fSampleRate)
{
  m_fSampleRate = fSampleRate;
}

bool Voice::IsAlive() const
{
  return m_bAlive;
}

void Voice::On(float fFrequency)
{
  auto const prevPitch = getCurrPitch();

  m_fPitch = FrequencyToPitch(fFrequency);

  if(m_bAlive)
    m_fPortamentoPitch = prevPitch - m_fPitch;
  else
    m_fPortamentoPitch = 0;

  m_Osc[0].SetFrequency(fFrequency / m_fSampleRate);
  m_Osc[1].SetFrequency(fFrequency / m_fSampleRate);

  if(!m_bAlive)
  {
    // reset the oscillators only if the voice was dead
    m_Osc[0].Sync();
    m_Osc[1].Sync();

    m_Envelope[0].EnvAttack();
    m_Envelope[1].EnvAttack();
  }

  m_bAlive = true;
}

void Voice::Off()
{
  m_Envelope[0].EnvRelease();
  m_Envelope[1].EnvRelease();
}

void Voice::SetPortamento(float fPortaTimeInChunks)
{
  m_fPortaTimeInChunks = fPortaTimeInChunks;
}

void Voice::SetADSR(int iEnvIdx, float A, float D, float S, float R)
{
  m_Envelope[iEnvIdx].SetADSR(A, D, S, R);
}

void Voice::SetShape(int iOscIdx, Osc::SHAPE eShape)
{
  m_Osc[iOscIdx].SetWaveform(eShape);
}

void Voice::SetEnvToPitch(int iEnvIdx, float fAmount)
{
  m_fEnvToPitch[iEnvIdx] = fAmount;
}

void Voice::SetEnvToFilter(int iEnvIdx, float fEnvToFilter)
{
  m_fEnvToFilter[iEnvIdx] = fEnvToFilter;
}

void Voice::SetEnvToVol(int iEnvIdx, float fEnvToVol)
{
  m_fEnvToVol[iEnvIdx] = fEnvToVol;
}

void Voice::SetFilterBalance(float fBalance)
{
  m_fFilterBalance = fBalance;
}

void Voice::SetBaseFilter(int iFilterIdx, float fCutoff, float fResonance)
{
  m_fCutoff[iFilterIdx] = fCutoff;
  m_fResonance[iFilterIdx] = fResonance;
}

void Voice::SetFilterPan(int iFilterIdx, float fPan)
{
  m_fFilterPan[iFilterIdx] = fPan;
}

void Voice::SetDetune(int iOscIdx, float fDetune)
{
  m_fOscDetune[iOscIdx] = fDetune;
}

void Voice::SetPulseWidth(int iOscIdx, float fPulseWidth)
{
  m_Osc[iOscIdx].SetPulseWidth(fPulseWidth);
}

void Voice::NewChunk(float const lfo[6])
{
  if(!m_bAlive)
    return;

  if(m_Envelope[0].IsDead() && m_Envelope[1].IsDead())
  {
    m_bAlive = false;
    return;
  }

  float fDeltaPitch1 = m_fOscDetune[0] + lfo[0] + m_fBend * 2;
  float fDeltaPitch2 = m_fOscDetune[1] + lfo[0] + m_fBend * 2;

  // modulation from enveloppes
  for(int i = 0; i < ArrayLength(m_Envelope); ++i)
  {
    fDeltaPitch1 += m_fEnvToPitch[i] * m_Envelope[i].Value();
    fDeltaPitch2 += m_fEnvToPitch[i] * m_Envelope[i].Value();
  }

  auto const pitch1 = getCurrPitch() + fDeltaPitch1;
  auto const pitch2 = getCurrPitch() + fDeltaPitch2;

  auto const freq1 = PitchToFrequency(pitch1);
  auto const freq2 = PitchToFrequency(pitch2);

  assert(freq1 == freq1);
  assert(freq2 == freq2);

  m_Osc[0].SetFrequency(freq1 / m_fSampleRate);
  m_Osc[1].SetFrequency(freq2 / m_fSampleRate);

  auto const alpha = exp(-m_fPortaTimeInChunks);
  m_fPortamentoPitch *= alpha;
}

// avoid clips in case of a computation error (NaN)
static float sanitizeValue(float val)
{
  if(IsNan(val) || val >= 4.0f || val <= -4.0f)
    return 0.0f;

  return val;
}

void Voice::Mix(AudioBuffer* out)
{
  if(!m_bAlive)
    return;

  for(long j = 0; j < out->numSamples; ++j)
  {
    float env[2];

    env[0] = m_Envelope[0].Work();
    env[1] = m_Envelope[1].Work();

    for(int i = 0; i < ArrayLength(m_Filter); ++i)
    {
      auto cutoff = m_fCutoff[i] + env[0] * m_fEnvToFilter[0] + env[1] * m_fEnvToFilter[1];
      m_Filter[i].setCutoff(cutoff * 0.5f / 100.0f);
      m_Filter[i].setRes(m_fResonance[i] / 100.0f);
    }

    float osc[2];
    osc[0] = m_Osc[0].Work();
    osc[1] = m_Osc[1].Work();

    auto oscOut = blend(osc[0], osc[1], m_fOscBalance) * m_fOscVol;

    if(m_fOscRingMod > 0)
      oscOut += osc[0] * osc[1];

    auto const f1 = m_Filter[0].Work(oscOut);
    auto const f2 = m_Filter[1].Work(oscOut);

    auto const R1 = (m_fFilterPan[0] + 1.0f) * 0.5f;
    auto const L1 = 1.0f - R1;
    auto const R2 = (m_fFilterPan[1] + 1.0f) * 0.5f;
    auto const L2 = 1.0f - R2;

    // Apply volume envelopes.
    auto const envVolume = (env[0] * m_fEnvToVol[0] + env[1] * m_fEnvToVol[1]);
    auto const left = (f1 * L1 + f2 * L2) * envVolume;
    auto const right = (f1 * R1 + f2 * R2) * envVolume;

    out->left[j] += sanitizeValue(left);
    out->right[j] += sanitizeValue(right);
  }
}

void Voice::BendTo(float fBend)
{
  if(fBend >= 0.95)
    fBend = 1.0f;

  if(fBend <= -0.95)
    fBend = -1.0f;

  m_fBend = fBend;
}

