/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */
#include <fstream>
#include <string>

#include <cassert>
#include <climits>
#include <cmath>

#include "waveform.h"
#include "lib_utils/math_tools.h"

WaveForm::WaveForm(SignalType type) :
  m_Type(type)
{
  // Check parameters
  static_assert(SINC_RESOL > 0, "");
  static_assert(SINC_RESOL < sizeof(long) * CHAR_BIT - 2, "");
  static_assert(SINC_ZEROS > 0, "");
  static_assert(static_cast<double>(SINC_ZEROS) * (1L << SINC_RESOL)
                < static_cast<double>(LONG_MAX), "");

  m_DutyCycle = 1.0 / 2.0;

  // ----------------------------------------------------------------------------
  // Sinc look-up table
  // ----------------------------------------------------------------------------

  for(long i = 0; i < SINC_TABLE_LEN; ++i)
  {
    if(i == SINC_ZERO_OFFSET)
    {
      sinc_table[i] = 1;
    }
    else
    {
      const double pos = (i - SINC_ZERO_OFFSET) * MY_PI;
      const double sinc_pos = pos * (1.0 / SINC_PTS_PER_SPL);
      const double win_pos = pos * (1.0 / SINC_TABLE_LEN);

      const double sinc = sin(sinc_pos) / sinc_pos;

      // Blackmann window
      const double win = 0.50 * cos(win_pos * 0.5)
                         + 0.08 * cos(win_pos)
                         + 0.42;

      sinc_table[i] = sinc * win;
    }
  }

  // ----------------------------------------------------------------------------
  // Signal generation
  // ----------------------------------------------------------------------------

  SetFrequency(0.01);

  // Signal parameters.

  // Buffer init
  for(auto& val : m_buffer)
    val = 0;

  // Initial values
  m_buf_pos = 0; // Current buffer positon
  m_cycle_cnt = 0; // Current cycle (for BP-BLITs)
  m_square = -m_DutyCycle;
  switch(m_Type)
  {
  case PULSE:
    m_result = -m_DcOffset;
    break;
  case SQUARE:
    m_result = -m_DutyCycle;
    break;
  default:
    m_result = -0.5; // Initial signal value
    break;
  }
}

void WaveForm::Sync()
{
  m_pos._int = 0;
  m_pos._frac = 0;
}

void WaveForm::SetFrequency(double f)
{
  m_Frequency = f;
  m_DcOffset = m_Frequency; // For standard BLIT
  m_Period = 1.0 / m_Frequency;
  m_PeriodFactor = m_Period / (m_Period - 1);
  m_Integrator = exp(log(0.8) / m_Period); // Leaky integrator

  if(m_Type == SQUARE || m_Type == TRIANGLE)
  {
    const double c_len_0 = m_Period * m_DutyCycle;
    const double c_len_1 = m_Period - c_len_0;

    m_per[0] = Cycle(c_len_0, +1);
    m_per[1] = Cycle(c_len_1, -1);
  }
  else
  {
    m_per[0] = Cycle(m_Period, +1);
    m_per[1] = Cycle(m_Period, +1);
  }

  // Check parameters
  assert(m_Frequency > 0);
  assert(m_Frequency <= 1.0 / 2);
  assert(m_DutyCycle >= m_Frequency / 1.0);
  assert(1 - m_DutyCycle >= m_Frequency / 1.0);
}

float WaveForm::Work()
{
  // New pulse
  if(m_pos._int <= 0)
    NewPulse();

  // Get current pulse value and clear buffer
  const double pulse = m_buffer[m_buf_pos];
  m_buffer[m_buf_pos] = 0;
  // Compute final signal
  switch(m_Type)
  {
  // BLIT based signals
  case PULSE:
    m_result = pulse - m_DcOffset;
    break;
  case SAWTOOTH:
    m_result = m_result * m_Integrator + (pulse - m_DcOffset) * m_PeriodFactor;
    break;

  // BP-BLIT based signals
  case SQUARE:
    m_result = m_result * m_Integrator + pulse;
    break;
  case TRIANGLE:
    m_square = m_square * m_Integrator + pulse;
    m_result = m_result * m_Integrator + m_square / (m_Period * m_DutyCycle * (1 - m_DutyCycle));
    break;
  }

  // Next sample
  --m_pos._int;
  m_buf_pos = (m_buf_pos + 1) % BUFFER_LEN;

  // check for NaN
  assert(m_result == m_result);

  m_result = std::min(1.0, m_result);
  m_result = std::max(-1.0, m_result);

  return m_result;
}

void WaveForm::SetDutyCycle(double fDutyCycle)
{
  m_DutyCycle = fDutyCycle;
}

void WaveForm::NewPulse()
{
  long offset = SINC_PTS_PER_SPL - long(m_pos._frac >> POS_TO_PHASE);
  long cur_pos = m_buf_pos;

  if(offset == SINC_PTS_PER_SPL)
  {
    offset = 0;
    --cur_pos;
  }

  // Generate 1 pulse and mix it with the buffer
  for(long i = 0; i < SINC_NUM_SAMPLES; ++i)
  {
    assert(offset >= 0);
    assert(offset < SINC_TABLE_LEN);

    ++cur_pos;
    m_buffer[cur_pos % BUFFER_LEN] += sinc_table[offset] * m_pos._sign;
    offset += SINC_PTS_PER_SPL;
  }

  // Next cycle (2 cycles = 1 period for BP-BLIT)
  m_pos += m_per[m_cycle_cnt];

  m_cycle_cnt ^= 1;
  m_pos._sign = m_per[m_cycle_cnt]._sign;
}

float SineWaveForm::Work()
{
  auto const PERIOD = 2.0 * MY_PI;
  m_fPhase += PERIOD * m_freq;

  while(m_fPhase >= PERIOD)
    m_fPhase -= PERIOD;

  return std::sin(m_fPhase);
}

void SineWaveForm::SetFrequency(double f)
{
  m_freq = f;
}

void SineWaveForm::Sync()
{
  m_fPhase = 0;
}

float NoiseWaveForm::Work()
{
  auto const scale = 2.0f / 0xffffffff;

  m_x1 ^= m_x2;
  auto const fSample = float(m_x2) * scale;
  m_x2 += m_x1;

  return fSample;
}

void NoiseWaveForm::SetFrequency(double /*freq*/)
{
}

void NoiseWaveForm::Sync()
{
  m_x1 = 0x01234567;
  m_x2 = 0x89ABCDEF;
}

float TriangleWaveForm::Work()
{
  auto const PERIOD = 1.0;
  m_fPhase += PERIOD * m_Frequency;

  while(m_fPhase >= PERIOD)
    m_fPhase -= PERIOD;

  // naive triangle implementation
  if(m_fPhase < PERIOD / 2)
  {
    return (m_fPhase / (PERIOD / 2)) * 2.0f - 1.0f;
  }
  else
  {
    float fSubPhase = m_fPhase - PERIOD / 2;
    return 1.0f - (fSubPhase / (PERIOD / 2)) * 2.0f;
  }
}

void TriangleWaveForm::SetFrequency(double f)
{
  m_Frequency = f;
}

void TriangleWaveForm::Sync()
{
  m_fPhase = 0;
}

