/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */
#pragma once

#include <cmath>
#include "limits.h"
#include "filter.h"

class IWaveForm
{
public:
  virtual void SetFrequency(double f) = 0;
  virtual void Sync() = 0;
  virtual float Work() = 0;
};

class WaveForm : public IWaveForm
{
public:
  enum SignalType
  {
    PULSE,
    SAWTOOTH,
    SQUARE,
    TRIANGLE // Not very stable in the current implementation
  };

  WaveForm(SignalType type = SQUARE);
  virtual float Work();

  void SetFrequency(double f);
  void Sync();

  void SetDutyCycle(double fDutyCycle);

private:
  long m_buf_pos;   // Current buffer positon
  int m_cycle_cnt;   // Current cycle (for BP-BLITs)
  double m_square;
  double m_result;

  // Look-up table parameters
  static const int SINC_RESOL = 4;   // Nbr of bits of the nbr of points/sample
  static const long SINC_ZEROS = 16;  // Nbr of zero-crossings.
  static const long SINC_NUM_SAMPLES = SINC_ZEROS * 2;
  static const int POS_TO_PHASE = sizeof(long) * CHAR_BIT - SINC_RESOL;

  static const long SINC_PTS_PER_SPL = 1L << SINC_RESOL;
  static const long SINC_TABLE_LEN = SINC_NUM_SAMPLES * SINC_PTS_PER_SPL;
  static const long SINC_ZERO_OFFSET = SINC_TABLE_LEN / 2;
  double sinc_table[SINC_TABLE_LEN]; // Prefer allocation on the heap for big values

  static const long BUFFER_LEN = SINC_NUM_SAMPLES << 1;  // Must be a power of 2, and greater than SINC_NBR_SPL

  double m_buffer[BUFFER_LEN];

  void NewPulse();

  struct Cycle
  {
    Cycle() :
      _sign(1), _int(0), _frac(0)
    {
    }

    Cycle(double value, int sign)
      : _sign(sign)
      , _int(static_cast<long>(floor(value)))
      , _frac(static_cast<unsigned long>
              ((value - floor(value)) * (static_cast<double>(ULONG_MAX) + 1)))
    {
    }

    Cycle const & operator += (Cycle const& other)
    {
      _int += other._int;
      _frac += other._frac;

      if(_frac < other._frac)
        ++_int;

      return *this;
    }

    // Pulse sign
    int _sign;

    // Fixed point value
    long _int;
    unsigned long _frac;
  };

  double m_Frequency; // Signal frequency
  double m_Period;
  double m_PeriodFactor; // precomputed: equal to m_Period / (m_Period - 1)
  double m_DcOffset;
  double m_Integrator;
  double m_DutyCycle; // Pulsewidth (for BP-BLIT based waveforms)

  Cycle m_pos;
  Cycle m_per[2];
  SignalType m_Type;
};

class SineWaveForm : public IWaveForm
{
public:
  virtual float Work();
  virtual void SetFrequency(double f);
  virtual void Sync();

private:
  double m_freq = 0; // periods/sample
  double m_fPhase = 0;
};

class NoiseWaveForm : public IWaveForm
{
public:
  virtual float Work();
  virtual void SetFrequency(double f);
  virtual void Sync();

private:
  unsigned int m_x1;
  unsigned int m_x2;
};

class TriangleWaveForm : public IWaveForm
{
public:
  virtual float Work();
  virtual void SetFrequency(double f);
  virtual void Sync();

private:
  double m_Frequency = 0;
  double m_fPhase = 0;
};

