// Copyright (C) 2018 - Sebastien Alaiwan
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Affero General Public License as
// published by the Free Software Foundation, either version 3 of the
// License, or (at your option) any later version.

// VST-adapter for Everest plugins

#if _WIN32
#include <windows.h>
#undef min
#elif __linux__
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#undef KeyPress
#elif EVEREST_MAC
#endif

#include <limits.h>
#include <memory.h>
#include <sstream>
#include <iomanip>
#include <algorithm>
#include "editor_wrapper.h"
#include "vst_adapter.h"
#include "lib_utils/string_format.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/debug.h"

#if EVEREST_ENABLE_GUI
static int activePlugins;

#if __linux__

namespace juce
{
extern Display* display;
}

class SharedMessageThread : public Thread
{
public:
  SharedMessageThread() : Thread("VstMessageThread")
  {
    startThread(7);

    while(!initialised)
    {
      sleep(1);
    }
  }

  ~SharedMessageThread()
  {
    signalThreadShouldExit();
    JUCEApplication::quit();
    waitForThreadToExit(5000);
    clearSingletonInstance();
  }

  void run()
  {
    initialiseJuce_GUI();
    initialised = true;

    MessageManager::getInstance()->setCurrentThreadAsMessageThread();

    while((!threadShouldExit()) && MessageManager::getInstance()->runDispatchLoopUntil(250))
    {
    }
  }

  juce_DeclareSingleton(SharedMessageThread, false);

private:
  bool initialised = false;
};

juce_ImplementSingleton(SharedMessageThread)

#endif
#endif

namespace EVEREST
{
VstAdapter::VstAdapter(HostControlFunction hostDo)
  : m_HostDo(hostDo)
{
}

void VstAdapter::Init(IPlugin* pPlugin)
{
  m_Plugin = pPlugin;

  assert(m_Plugin != nullptr);

  IVst2Plugin* vst = this;
  auto& m_Vst2Interface = *vst;

  memset(&m_Vst2Interface, 0, sizeof m_Vst2Interface);

  m_Vst2Interface.magic[0] = 'P';
  m_Vst2Interface.magic[1] = 't';
  m_Vst2Interface.magic[2] = 's';
  m_Vst2Interface.magic[3] = 'V';

  m_Vst2Interface.dispatcher = static_dispatcher;
  m_Vst2Interface.getParameter = static_getParameter;
  m_Vst2Interface.setParameter = static_setParameter;
  m_Vst2Interface.process = static_process;
  m_Vst2Interface.processReplacing = static_processReplacing;

  m_Vst2Interface.numPrograms = 1;
  m_Vst2Interface.numParams = m_Plugin->GetNumParams();
  m_Vst2Interface.numInputs = 2;
  m_Vst2Interface.numOutputs = 2;

  m_Vst2Interface.flags = VST_PLUGFLAG_CANDOREPLACING;
  m_Vst2Interface.flags |= VST_PLUGFLAG_SAVE_STATE;

  pPlugin->GetId(m_Vst2Interface.plugin_id);

  if(g_PlugDesc.isSynth)
    m_Vst2Interface.flags |= VST_PLUGFLAG_SYNTH;

  m_Vst2Interface.userdata = this;
  m_Vst2Interface.unkown_float = 1.0;

#if EVEREST_ENABLE_GUI

  if(g_PlugDesc.createEditor)
    m_Vst2Interface.flags |= VST_PLUGFLAG_HASEDITOR;

  m_pEditorComp = nullptr;
  ++activePlugins;
#endif
}

VstAdapter::~VstAdapter()
{
  delete m_Plugin;

#if EVEREST_ENABLE_GUI
  assert(activePlugins >= 1);
  --activePlugins;

  if(activePlugins == 0)
  {
#if JUCE_LINUX
    SharedMessageThread::deleteInstance();
#else
    shutdownJuce_GUI();
#endif
  }

#endif
}

void VstAdapter::OnParamChanged(int iIndex, float fValue)
{
  m_HostDo(this, 0, iIndex, 0, nullptr, fValue);
}

double VstAdapter::GetTempo()
{
  auto time = reinterpret_cast<VstTime*>(m_HostDo(this, 7, 1 << 10, 0, nullptr, 0));

  if(time == nullptr)
    return 145.0;

  return time->tempo;
}

#if EVEREST_ENABLE_GUI
void VstAdapter::CreateEditorComponent()
{
  if(m_pEditorComp == 0)
  {
    assert(g_PlugDesc.createEditor);
    Component* ed = g_PlugDesc.createEditor(m_Plugin, this);

    if(ed != 0)
    {
      ed->setOpaque(true);
      ed->setVisible(true);

      m_pEditorComp = new EditorWrapper(ed);
    }
  }
}

#endif

int VstAdapter::Dispatcher(VST_CMD eRequest, int iParam1, intptr_t iParam2, void* pData, float fParam)
{
  if(0)
  {
    std::stringstream s;
    s << "Dispatcher: " << ToString(eRequest) << " " << iParam1 << " " << iParam2 << std::hex << " 0x" << (uintptr_t)pData << " " << fParam << std::endl;

    if(eRequest != VST_CMD_PROCESSEVENTS && eRequest != 19)
      DebugTrace(s.str().c_str());
  }

  if(eRequest == VST_CMD_GETPLUGNAME || eRequest == VST_CMD_GETPRODUCTNAME)
  {
    std::string sName = g_PlugDesc.name;
#ifdef _DEBUG
    sName = "@" + sName;
#endif
    strncpy((char*)pData, sName.c_str(), 16);
    return 1;
  }
  else if(eRequest == VST_CMD_GETAUTHORNAME)
  {
    strncpy((char*)pData, g_PlugDesc.author, 16);
    return 1;
  }
  else if(eRequest == VST_CMD_GETPATCHNAME)
  {
    std::string sName = "- Init -";
    strncpy((char*)pData, sName.c_str(), 16);
    return 1;
  }
  else if(eRequest == VST_CMD_LOAD)
  {
    m_StateData.resize(iParam2);
    memcpy(&m_StateData[0], pData, iParam2);

    LoadPluginState();

    return 1;
  }
  else if(eRequest == VST_CMD_SAVE)
  {
    SavePluginState();

    *((void**)pData) = m_StateData.data();

    return m_StateData.size();
  }
  else if(eRequest == VST_CMD_CLOSE)
  {
    delete this;
    return 1;
  }

#if EVEREST_ENABLE_GUI
  else if(eRequest == VST_CMD_GETRECT)
  {
    VstRect** pRect = (VstRect**)pData;

    if(!g_PlugDesc.createEditor)
    {
      *pRect = nullptr;
      return 0;
    }

    m_Rect.x = m_Rect.y = 0;
    m_Rect.cx = g_PlugDesc.size.width;
    m_Rect.cy = g_PlugDesc.size.height;
    *pRect = &m_Rect;
    return 1;
  }
  else if(eRequest == VST_CMD_SHOW)
  {
    MessageManagerLock const mmLock;
    delete m_pEditorComp;
    m_pEditorComp = nullptr;

    CreateEditorComponent();

    if(m_pEditorComp)
    {
#if _WIN32
      m_pEditorComp->addToDesktop(0);
      HWND editorWnd = (HWND)m_pEditorComp->getWindowHandle();
      SetParent(editorWnd, (HWND)pData);

      // remove WS_POPUP style, add WS_CHILD style.
      DWORD val = GetWindowLong(editorWnd, GWL_STYLE);
      val = (val & ~WS_POPUP) | WS_CHILD;
      SetWindowLong(editorWnd, GWL_STYLE, val);

      SetWindowPos(editorWnd, 0, 0, 0, 0, 0, SWP_NOSIZE | SWP_NOOWNERZORDER | SWP_NOACTIVATE | SWP_SHOWWINDOW);
#elif __linux__
      m_pEditorComp->addToDesktop(0, pData);
      auto hostWindow = (Window)pData;
      auto editorWnd = (Window)m_pEditorComp->getWindowHandle();
      XReparentWindow(display, editorWnd, hostWindow, 0, 0);
#endif
      m_pEditorComp->setVisible(true);
    }

    return 1;
  }
  else if(eRequest == VST_CMD_HIDE)
  {
    MessageManagerLock const mmLock;
    delete m_pEditorComp;
    m_pEditorComp = nullptr;
    return 1;
  }
#endif
  else if(eRequest == VST_CMD_SETSAMPLERATE)
  {
    m_Plugin->SetSampleRate(fParam);
    return 1;
  }
  else if(eRequest == VST_CMD_PROCESSEVENTS)
  {
    VstEvents* events = reinterpret_cast<VstEvents*>(pData);

    for(int i = 0; i < events->numEvents; ++i)
    {
      VstEvent* e = events->events[i];
      switch(e->m_iType)
      {
      case VST_EVT_MIDI:
        // save MIDI events for use from Process (audio) call
        m_EventBuffer.push_back(events->events[i]->midi);
        break;
      default:
        break;
      }
    }

    return 1;
  }
  else if(eRequest == VST_CMD_ISAUTOMATABLE)
  {
    return 1;
  }
  else if(eRequest == VST_CMD_GETPARAMNAME)
  {
    std::string s;
    int iIndex = iParam1;

    if(iIndex >= 0 && iIndex < m_Plugin->GetNumParams())
      s = m_Plugin->GetParamName(iIndex);

    strncpy((char*)pData, s.c_str(), 16);
    return 1;
  }
  else if(eRequest == VST_CMD_GETPARAMSTRINGVALUE)
  {
    std::string s;
    int iIndex = iParam1;

    if(iIndex >= 0 && iIndex < m_Plugin->GetNumParams())
      s = m_Plugin->GetParamText(iIndex);

    strncpy((char*)pData, s.c_str(), 16);
    return 1;
  }
  else if(eRequest == VST_CMD_GETPARAMUNITS)
  {
    std::string s = "unit";
    strncpy((char*)pData, s.c_str(), 16);
    return 1;
  }
  else if(eRequest == VST_CMD_GETCAPABILITY)
  {
    std::string sCap = (char*)pData;

    if(sCap == "receiveVstMidiEvent")
      return 1;

    if(sCap == "sendVstMidiEvent")
      return 1;
  }
  else if(eRequest == VST_CMD_GETNUMMIDIINPUTS || eRequest == VST_CMD_GETNUMMIDIOUTPUTS)
  {
    return 16;
  }
  else if(eRequest == VST_CMD_SETTEMPO)
  {
    m_Plugin->OnTempoChange(fParam);
    return 1;
  }
  else if(eRequest == VST_CMD_OPEN)
  {
    // enable receiving of MIDI events.
    m_HostDo(this, 6, 0, 1, nullptr, 0);
  }
  else
  {
  }
  return 0;
}

void VstAdapter::LoadPluginState()
{
  if(auto serializable = dynamic_cast<ISerializable*>(m_Plugin))
  {
    serializable->loadState(m_StateData);
  }
  else
  {
    auto const iNumParams = m_Plugin->GetNumParams();

    for(int i = 0; i < iNumParams; ++i)
    {
      // if the buffer is not big enough, give up
      if(i * 4 + 4 > (int)m_StateData.size())
        break;

      unsigned int iValue = 0;

      for(int k = 3; k >= 0; --k)
      {
        iValue *= 256;
        iValue += m_StateData[i * 4 + k];
      }

      float fValue = float(iValue) / float(UINT_MAX);

      m_Plugin->SetParam(i, fValue);
    }
  }
}

void VstAdapter::SavePluginState()
{
  if(auto serializable = dynamic_cast<ISerializable*>(m_Plugin))
  {
    m_StateData = serializable->saveState();
  }
  else
  {
    auto const iNumParams = m_Plugin->GetNumParams();

    m_StateData.clear();

    for(int i = 0; i < iNumParams; ++i)
    {
      float fValue = std::max(0.0f, std::min(1.0f, m_Plugin->GetParam(i)));
      long long iValue = (long long)(UINT_MAX * fValue);
      iValue = std::min<long long>(iValue, UINT_MAX);
      iValue = std::max<long long>(iValue, 0);

      for(int k = 0; k < 4; ++k)
      {
        m_StateData.push_back(iValue & 0xFF);
        iValue /= 256;
      }
    }
  }
}

void VstAdapter::ProcessMidiEvent(VstMidiEvent* e)
{
  auto midi = dynamic_cast<EVEREST::IMidiListener*>(m_Plugin);

  if(!midi)
    return;

  int command = (e->midiData[0] & 0x70) >> 4;
  int channel = (e->midiData[0] & 0x0F) >> 0;

  if(command == 0x1) // note on
  {
    int note = (e->midiData[1] & 0x7F);
    int vel = (e->midiData[2] & 0x7F);

    if(vel > 0)
      // std::cout << "NoteOn" << std::endl;
      midi->noteOn(channel, note, vel);
    else
      midi->noteOff(channel, note, vel);
  }
  else if(command == 0x0) // note off
  {
    int note = (e->midiData[1] & 0x7F);
    int vel = (e->midiData[2] & 0x7F);
    midi->noteOff(channel, note, vel);
  }
  else if(command == 0x3)
  {
    int index = (e->midiData[1] & 0x7F);
    int value = (e->midiData[2] & 0x7F);
    midi->controlChange(channel, index, value);
  }
  else if(command == 0x6) // pitch-bend
  {
    int data1 = (e->midiData[1] & 0x7F);
    int data2 = (e->midiData[2] & 0x7F);
    int iValue = (data1 + data2 * 256) - 16384;
    midi->pitchBend(channel, float(iValue) / 16384.0f);
  }
}

void VstAdapter::Process(float** pInput, float** pOutput, int iNumSamples)
{
  double tempo = GetTempo();

  if(tempo != m_fTempo)
  {
    m_Plugin->OnTempoChange(tempo);
    m_fTempo = tempo;
  }

  auto pInputLeft = pInput[0];
  auto pInputRight = pInput[1];
  auto pOutputLeft = pOutput[0];
  auto pOutputRight = pOutput[1];

  int iTime = 0;
  int iNextTime = 0;
  size_t iEvtIdx = 0;

  while(iTime < iNumSamples)
  {
    VstMidiEvent* evt = nullptr;

    // find next event time
    if(iEvtIdx < m_EventBuffer.size())
    {
      iNextTime = m_EventBuffer[iEvtIdx].deltaFrames;
      evt = &m_EventBuffer[iEvtIdx];
      iEvtIdx++;
    }
    else
    {
      iNextTime = iNumSamples;
    }

    // generate audio until event time
    auto const iSamples = iNextTime - iTime;

    if(iSamples > 0)
    {
      float* pInputChannels[] = { pInputLeft, pInputRight };
      float* pOutputChannels[] = { pOutputLeft, pOutputRight };
      m_Plugin->ProcessAudio(pInputChannels, pOutputChannels, iSamples);

      pInputLeft += iSamples;
      pInputRight += iSamples;
      pOutputLeft += iSamples;
      pOutputRight += iSamples;

      iTime += iSamples;
    }

    if(evt)
    {
      ProcessMidiEvent(evt);
    }
  }

  m_EventBuffer.clear();
}
}

namespace EVEREST
{
IVst2Plugin* CreateVst2Plugin(HostControlFunction cb)
{
#if EVEREST_ENABLE_GUI

  if(activePlugins == 0)
  {
    // must be created before calling the plugin constructor.
#if __linux__
    SharedMessageThread::getInstance();
#else
    initialiseJuce_GUI();
#endif
  }

#endif

  auto* adapter = new VstAdapter(cb);
  auto pPlugin = g_PlugDesc.create(adapter);
  adapter->Init(pPlugin);

  return adapter;
}
}

