/*
	This file wraps together all the linux-specific code, so that
	we can include all the native headers just once, and compile all our
	platform-specific stuff in one big lump, keeping it out of the way of
	the rest of the codebase.
*/

#include "juce_TargetPlatform.h"
#include "juce_Config.h"
#include "juce_NativeIncludes.h"
#include "juce.h"

#if JUCE_LINUX

BEGIN_JUCE_NAMESPACE

#define JUCE_INCLUDED_FILE 1

// Now include the actual code files..


/*** Start of inlined file: juce_posix_SharedCode.h ***/
/*
	This file contains posix routines that are common to both the Linux and Mac builds.

	It gets included directly in the cpp files for these platforms.
*/

CriticalSection::CriticalSection() noexcept
{
	pthread_mutexattr_t atts;
	pthread_mutexattr_init (&atts);
	pthread_mutexattr_settype (&atts, PTHREAD_MUTEX_RECURSIVE);
   #if ! JUCE_ANDROID
	pthread_mutexattr_setprotocol (&atts, PTHREAD_PRIO_INHERIT);
   #endif
	pthread_mutex_init (&internal, &atts);
}

CriticalSection::~CriticalSection() noexcept
{
	pthread_mutex_destroy (&internal);
}

void CriticalSection::enter() const noexcept
{
	pthread_mutex_lock (&internal);
}

bool CriticalSection::tryEnter() const noexcept
{
	return pthread_mutex_trylock (&internal) == 0;
}

void CriticalSection::exit() const noexcept
{
	pthread_mutex_unlock (&internal);
}

class WaitableEventImpl
{
public:
	WaitableEventImpl (const bool manualReset_)
		: triggered (false),
		  manualReset (manualReset_)
	{
		pthread_cond_init (&condition, 0);

		pthread_mutexattr_t atts;
		pthread_mutexattr_init (&atts);
	   #if ! JUCE_ANDROID
		pthread_mutexattr_setprotocol (&atts, PTHREAD_PRIO_INHERIT);
	   #endif
		pthread_mutex_init (&mutex, &atts);
	}

	~WaitableEventImpl()
	{
		pthread_cond_destroy (&condition);
		pthread_mutex_destroy (&mutex);
	}

	bool wait (const int timeOutMillisecs) noexcept
	{
		pthread_mutex_lock (&mutex);

		if (! triggered)
		{
			if (timeOutMillisecs < 0)
			{
				do
				{
					pthread_cond_wait (&condition, &mutex);
				}
				while (! triggered);
			}
			else
			{
				struct timeval now;
				gettimeofday (&now, 0);

				struct timespec time;
				time.tv_sec  = now.tv_sec  + (timeOutMillisecs / 1000);
				time.tv_nsec = (now.tv_usec + ((timeOutMillisecs % 1000) * 1000)) * 1000;

				if (time.tv_nsec >= 1000000000)
				{
					time.tv_nsec -= 1000000000;
					time.tv_sec++;
				}

				do
				{
					if (pthread_cond_timedwait (&condition, &mutex, &time) == ETIMEDOUT)
					{
						pthread_mutex_unlock (&mutex);
						return false;
					}
				}
				while (! triggered);
			}
		}

		if (! manualReset)
			triggered = false;

		pthread_mutex_unlock (&mutex);
		return true;
	}

	void signal() noexcept
	{
		pthread_mutex_lock (&mutex);
		triggered = true;
		pthread_cond_broadcast (&condition);
		pthread_mutex_unlock (&mutex);
	}

	void reset() noexcept
	{
		pthread_mutex_lock (&mutex);
		triggered = false;
		pthread_mutex_unlock (&mutex);
	}

private:
	pthread_cond_t condition;
	pthread_mutex_t mutex;
	bool triggered;
	const bool manualReset;

	JUCE_DECLARE_NON_COPYABLE (WaitableEventImpl);
};

WaitableEvent::WaitableEvent (const bool manualReset) noexcept
	: internal (new WaitableEventImpl (manualReset))
{
}

WaitableEvent::~WaitableEvent() noexcept
{
	delete static_cast <WaitableEventImpl*> (internal);
}

bool WaitableEvent::wait (const int timeOutMillisecs) const noexcept
{
	return static_cast <WaitableEventImpl*> (internal)->wait (timeOutMillisecs);
}

void WaitableEvent::signal() const noexcept
{
	static_cast <WaitableEventImpl*> (internal)->signal();
}

void WaitableEvent::reset() const noexcept
{
	static_cast <WaitableEventImpl*> (internal)->reset();
}

void JUCE_CALLTYPE Thread::sleep (int millisecs)
{
	struct timespec time;
	time.tv_sec = millisecs / 1000;
	time.tv_nsec = (millisecs % 1000) * 1000000;
	nanosleep (&time, 0);
}

const juce_wchar File::separator = '/';
const String File::separatorString ("/");

File File::getCurrentWorkingDirectory()
{
	HeapBlock<char> heapBuffer;

	char localBuffer [1024];
	char* cwd = getcwd (localBuffer, sizeof (localBuffer) - 1);
	int bufferSize = 4096;

	while (cwd == nullptr && errno == ERANGE)
	{
		heapBuffer.malloc (bufferSize);
		cwd = getcwd (heapBuffer, bufferSize - 1);
		bufferSize += 1024;
	}

	return File (CharPointer_UTF8 (cwd));
}

bool File::setAsCurrentWorkingDirectory() const
{
	return chdir (getFullPathName().toUTF8()) == 0;
}

namespace
{
   #if JUCE_LINUX || (JUCE_IOS && ! __DARWIN_ONLY_64_BIT_INO_T) // (this iOS stuff is to avoid a simulator bug)
	typedef struct stat64 juce_statStruct;
	#define JUCE_STAT	 stat64
   #else
	typedef struct stat   juce_statStruct;
	#define JUCE_STAT	 stat
   #endif

	bool juce_stat (const String& fileName, juce_statStruct& info)
	{
		return fileName.isNotEmpty()
				 && JUCE_STAT (fileName.toUTF8(), &info) == 0;
	}

	// if this file doesn't exist, find a parent of it that does..
	bool juce_doStatFS (File f, struct statfs& result)
	{
		for (int i = 5; --i >= 0;)
		{
			if (f.exists())
				break;

			f = f.getParentDirectory();
		}

		return statfs (f.getFullPathName().toUTF8(), &result) == 0;
	}

	void updateStatInfoForFile (const String& path, bool* const isDir, int64* const fileSize,
								Time* const modTime, Time* const creationTime, bool* const isReadOnly)
	{
		if (isDir != nullptr || fileSize != nullptr || modTime != nullptr || creationTime != nullptr)
		{
			juce_statStruct info;
			const bool statOk = juce_stat (path, info);

			if (isDir != nullptr)	 *isDir	= statOk && ((info.st_mode & S_IFDIR) != 0);
			if (fileSize != nullptr)	  *fileSize	 = statOk ? info.st_size : 0;
			if (modTime != nullptr)	   *modTime	  = Time (statOk ? (int64) info.st_mtime * 1000 : 0);
			if (creationTime != nullptr)  *creationTime = Time (statOk ? (int64) info.st_ctime * 1000 : 0);
		}

		if (isReadOnly != nullptr)
			*isReadOnly = access (path.toUTF8(), W_OK) != 0;
	}

	Result getResultForErrno()
	{
		return Result::fail (String (strerror (errno)));
	}

	Result getResultForReturnValue (int value)
	{
		return value == -1 ? getResultForErrno() : Result::ok();
	}
}

bool File::isDirectory() const
{
	juce_statStruct info;

	return fullPath.isEmpty()
			|| (juce_stat (fullPath, info) && ((info.st_mode & S_IFDIR) != 0));
}

bool File::exists() const
{
	return fullPath.isNotEmpty()
			 && access (fullPath.toUTF8(), F_OK) == 0;
}

bool File::existsAsFile() const
{
	return exists() && ! isDirectory();
}

int64 File::getSize() const
{
	juce_statStruct info;
	return juce_stat (fullPath, info) ? info.st_size : 0;
}

bool File::hasWriteAccess() const
{
	if (exists())
		return access (fullPath.toUTF8(), W_OK) == 0;

	if ((! isDirectory()) && fullPath.containsChar (separator))
		return getParentDirectory().hasWriteAccess();

	return false;
}

bool File::setFileReadOnlyInternal (const bool shouldBeReadOnly) const
{
	juce_statStruct info;
	if (! juce_stat (fullPath, info))
		return false;

	info.st_mode &= 0777;   // Just permissions

	if (shouldBeReadOnly)
		info.st_mode &= ~(S_IWUSR | S_IWGRP | S_IWOTH);
	else
		// Give everybody write permission?
		info.st_mode |= S_IWUSR | S_IWGRP | S_IWOTH;

	return chmod (fullPath.toUTF8(), info.st_mode) == 0;
}

void File::getFileTimesInternal (int64& modificationTime, int64& accessTime, int64& creationTime) const
{
	modificationTime = 0;
	accessTime = 0;
	creationTime = 0;

	juce_statStruct info;
	if (juce_stat (fullPath, info))
	{
		modificationTime = (int64) info.st_mtime * 1000;
		accessTime = (int64) info.st_atime * 1000;
		creationTime = (int64) info.st_ctime * 1000;
	}
}

bool File::setFileTimesInternal (int64 modificationTime, int64 accessTime, int64 /*creationTime*/) const
{
	juce_statStruct info;

	if ((modificationTime != 0 || accessTime != 0) && juce_stat (fullPath, info))
	{
		struct utimbuf times;
		times.actime  = accessTime != 0	   ? (time_t) (accessTime / 1000)	   : info.st_atime;
		times.modtime = modificationTime != 0 ? (time_t) (modificationTime / 1000) : info.st_mtime;

		return utime (fullPath.toUTF8(), &times) == 0;
	}

	return false;
}

bool File::deleteFile() const
{
	if (! exists())
		return true;

	if (isDirectory())
		return rmdir (fullPath.toUTF8()) == 0;

	return remove (fullPath.toUTF8()) == 0;
}

bool File::moveInternal (const File& dest) const
{
	if (rename (fullPath.toUTF8(), dest.getFullPathName().toUTF8()) == 0)
		return true;

	if (hasWriteAccess() && copyInternal (dest))
	{
		if (deleteFile())
			return true;

		dest.deleteFile();
	}

	return false;
}

Result File::createDirectoryInternal (const String& fileName) const
{
	return getResultForReturnValue (mkdir (fileName.toUTF8(), 0777));
}

int64 juce_fileSetPosition (void* handle, int64 pos)
{
	if (handle != 0 && lseek ((int) (pointer_sized_int) handle, pos, SEEK_SET) == pos)
		return pos;

	return -1;
}

void FileInputStream::openHandle()
{
	totalSize = file.getSize();

	const int f = open (file.getFullPathName().toUTF8(), O_RDONLY, 00644);

	if (f != -1)
		fileHandle = (void*)(size_t)f;
	else
		status = getResultForErrno();
}

void FileInputStream::closeHandle()
{
	if (fileHandle != 0)
	{
		close ((int) (pointer_sized_int) fileHandle);
		fileHandle = 0;
	}
}

size_t FileInputStream::readInternal (void* const buffer, const size_t numBytes)
{
	ssize_t result = 0;

	if (fileHandle != 0)
	{
		result = ::read ((int) (pointer_sized_int) fileHandle, buffer, numBytes);

		if (result < 0)
		{
			status = getResultForErrno();
			result = 0;
		}
	}

	return (size_t) result;
}

void FileOutputStream::openHandle()
{
	if (file.exists())
	{
		const int f = open (file.getFullPathName().toUTF8(), O_RDWR, 00644);

		if (f != -1)
		{
			currentPosition = lseek (f, 0, SEEK_END);

			if (currentPosition >= 0)
			{
				fileHandle = (void*)(size_t)f;
			}
			else
			{
				status = getResultForErrno();
				close (f);
			}
		}
		else
		{
			status = getResultForErrno();
		}
	}
	else
	{
		const int f = open (file.getFullPathName().toUTF8(), O_RDWR + O_CREAT, 00644);

		if (f != -1)
			fileHandle = (void*)(size_t)f;
		else
			status = getResultForErrno();
	}
}

void FileOutputStream::closeHandle()
{
	if (fileHandle != 0)
	{
		close ((int) (pointer_sized_int) fileHandle);
		fileHandle = 0;
	}
}

int FileOutputStream::writeInternal (const void* const data, const int numBytes)
{
	ssize_t result = 0;

	if (fileHandle != 0)
	{
		result = ::write ((int) (pointer_sized_int) fileHandle, data, numBytes);

		if (result == -1)
			status = getResultForErrno();
	}

	return (int) result;
}

void FileOutputStream::flushInternal()
{
	if (fileHandle != 0)
		if (fsync ((int) (pointer_sized_int) fileHandle) == -1)
			status = getResultForErrno();
}

MemoryMappedFile::MemoryMappedFile (const File& file, MemoryMappedFile::AccessMode mode)
	: address (nullptr),
	  length (0),
	  fileHandle (0)
{
	jassert (mode == readOnly || mode == readWrite);

	fileHandle = open (file.getFullPathName().toUTF8(),
					   mode == readWrite ? (O_CREAT + O_RDWR) : O_RDONLY, 00644);

	if (fileHandle != -1)
	{
		const int64 fileSize = file.getSize();

		void* m = mmap (0, (size_t) fileSize,
						mode == readWrite ? (PROT_READ | PROT_WRITE) : PROT_READ,
						MAP_SHARED, fileHandle, 0);

		if (m != MAP_FAILED)
		{
			address = m;
			length = (size_t) fileSize;
		}
	}
}

MemoryMappedFile::~MemoryMappedFile()
{
	if (address != nullptr)
		munmap (address, length);

	if (fileHandle != 0)
		close (fileHandle);
}

File juce_getExecutableFile()
{
   #if JUCE_ANDROID
	return File (android.appFile);
   #else
	Dl_info exeInfo;
	dladdr ((void*) juce_getExecutableFile, &exeInfo);  // (can't be a const void* on android)
	return File::getCurrentWorkingDirectory().getChildFile (CharPointer_UTF8 (exeInfo.dli_fname));
   #endif
}

int64 File::getBytesFreeOnVolume() const
{
	struct statfs buf;
	if (juce_doStatFS (*this, buf))
		return (int64) buf.f_bsize * (int64) buf.f_bavail; // Note: this returns space available to non-super user

	return 0;
}

int64 File::getVolumeTotalSize() const
{
	struct statfs buf;
	if (juce_doStatFS (*this, buf))
		return (int64) buf.f_bsize * (int64) buf.f_blocks;

	return 0;
}

String File::getVolumeLabel() const
{
   #if JUCE_MAC
	struct VolAttrBuf
	{
		u_int32_t	   length;
		attrreference_t mountPointRef;
		char		mountPointSpace [MAXPATHLEN];
	} attrBuf;

	struct attrlist attrList = { 0 };
	attrList.bitmapcount = ATTR_BIT_MAP_COUNT;
	attrList.volattr = ATTR_VOL_INFO | ATTR_VOL_NAME;

	File f (*this);

	for (;;)
	{
		if (getattrlist (f.getFullPathName().toUTF8(), &attrList, &attrBuf, sizeof (attrBuf), 0) == 0)
			return String::fromUTF8 (((const char*) &attrBuf.mountPointRef) + attrBuf.mountPointRef.attr_dataoffset,
									 (int) attrBuf.mountPointRef.attr_length);

		const File parent (f.getParentDirectory());

		if (f == parent)
			break;

		f = parent;
	}
   #endif

	return String::empty;
}

int File::getVolumeSerialNumber() const
{
	int result = 0;
/*	int fd = open (getFullPathName().toUTF8(), O_RDONLY | O_NONBLOCK);

	char info [512];

	#ifndef HDIO_GET_IDENTITY
	 #define HDIO_GET_IDENTITY 0x030d
	#endif

	if (ioctl (fd, HDIO_GET_IDENTITY, info) == 0)
	{
		DBG (String (info + 20, 20));
		result = String (info + 20, 20).trim().getIntValue();
	}

	close (fd);*/
	return result;
}

class InterProcessLock::Pimpl
{
public:
	Pimpl (const String& name, const int timeOutMillisecs)
		: handle (0), refCount (1)
	{
		// Note that we can't get the normal temp folder here, as it might be different for each app.
		File tempFolder ("/var/tmp");

		if (! tempFolder.isDirectory())
			tempFolder = "/tmp";

		const File temp (tempFolder.getChildFile (name));

		temp.create();
		handle = open (temp.getFullPathName().toUTF8(), O_RDWR);

		if (handle != 0)
		{
			struct flock fl = { 0 };
			fl.l_whence = SEEK_SET;
			fl.l_type = F_WRLCK;

			const int64 endTime = Time::currentTimeMillis() + timeOutMillisecs;

			for (;;)
			{
				const int result = fcntl (handle, F_SETLK, &fl);

				if (result >= 0)
					return;

				if (errno != EINTR)
				{
					if (timeOutMillisecs == 0
						 || (timeOutMillisecs > 0 && Time::currentTimeMillis() >= endTime))
						break;

					Thread::sleep (10);
				}
			}
		}

		closeFile();
	}

	~Pimpl()
	{
		closeFile();
	}

	void closeFile()
	{
		if (handle != 0)
		{
			struct flock fl = { 0 };
			fl.l_whence = SEEK_SET;
			fl.l_type = F_UNLCK;

			while (! (fcntl (handle, F_SETLKW, &fl) >= 0 || errno != EINTR))
			{}

			close (handle);
			handle = 0;
		}
	}

	int handle, refCount;
};

InterProcessLock::InterProcessLock (const String& name_)
	: name (name_)
{
}

InterProcessLock::~InterProcessLock()
{
}

bool InterProcessLock::enter (const int timeOutMillisecs)
{
	const ScopedLock sl (lock);

	if (pimpl == nullptr)
	{
		pimpl = new Pimpl (name, timeOutMillisecs);

		if (pimpl->handle == 0)
			pimpl = 0;
	}
	else
	{
		pimpl->refCount++;
	}

	return pimpl != nullptr;
}

void InterProcessLock::exit()
{
	const ScopedLock sl (lock);

	// Trying to release the lock too many times!
	jassert (pimpl != nullptr);

	if (pimpl != nullptr && --(pimpl->refCount) == 0)
		pimpl = 0;
}

void JUCE_API juce_threadEntryPoint (void*);

void* threadEntryProc (void* userData)
{
   #if JUCE_ANDROID
	struct AndroidThreadScope
	{
		AndroidThreadScope()   { threadLocalJNIEnvHolder.attach(); }
		~AndroidThreadScope()  { threadLocalJNIEnvHolder.detach(); }
	};

	const AndroidThreadScope androidEnv;
   #endif

	juce_threadEntryPoint (userData);
	return nullptr;
}

void Thread::launchThread()
{
	threadHandle_ = 0;
	pthread_t handle = 0;

	if (pthread_create (&handle, 0, threadEntryProc, this) == 0)
	{
		pthread_detach (handle);
		threadHandle_ = (void*) handle;
		threadId_ = (ThreadID) threadHandle_;
	}
}

void Thread::closeThreadHandle()
{
	threadId_ = 0;
	threadHandle_ = 0;
}

void Thread::killThread()
{
	if (threadHandle_ != 0)
	{
	   #if JUCE_ANDROID
		jassertfalse; // pthread_cancel not available!
	   #else
		pthread_cancel ((pthread_t) threadHandle_);
	   #endif
	}
}

void Thread::setCurrentThreadName (const String& name)
{
   #if JUCE_MAC && defined (MAC_OS_X_VERSION_10_6) && MAC_OS_X_VERSION_MIN_REQUIRED >= MAC_OS_X_VERSION_10_6
	pthread_setname_np (name.toUTF8());
   #elif JUCE_LINUX
	prctl (PR_SET_NAME, name.toUTF8().getAddress(), 0, 0, 0);
   #endif
}

bool Thread::setThreadPriority (void* handle, int priority)
{
	struct sched_param param;
	int policy;
	priority = jlimit (0, 10, priority);

	if (handle == 0)
		handle = (void*) pthread_self();

	if (pthread_getschedparam ((pthread_t) handle, &policy, &param) != 0)
		return false;

	policy = priority == 0 ? SCHED_OTHER : SCHED_RR;

	const int minPriority = sched_get_priority_min (policy);
	const int maxPriority = sched_get_priority_max (policy);

	param.sched_priority = ((maxPriority - minPriority) * priority) / 10 + minPriority;
	return pthread_setschedparam ((pthread_t) handle, policy, &param) == 0;
}

Thread::ThreadID Thread::getCurrentThreadId()
{
	return (ThreadID) pthread_self();
}

void Thread::yield()
{
	sched_yield();
}

/* Remove this macro if you're having problems compiling the cpu affinity
   calls (the API for these has changed about quite a bit in various Linux
   versions, and a lot of distros seem to ship with obsolete versions)
*/
#if defined (CPU_ISSET) && ! defined (SUPPORT_AFFINITIES)
 #define SUPPORT_AFFINITIES 1
#endif

void Thread::setCurrentThreadAffinityMask (const uint32 affinityMask)
{
   #if SUPPORT_AFFINITIES
	cpu_set_t affinity;
	CPU_ZERO (&affinity);

	for (int i = 0; i < 32; ++i)
		if ((affinityMask & (1 << i)) != 0)
			CPU_SET (i, &affinity);

	/*
	   N.B. If this line causes a compile error, then you've probably not got the latest
	   version of glibc installed.

	   If you don't want to update your copy of glibc and don't care about cpu affinities,
	   then you can just disable all this stuff by setting the SUPPORT_AFFINITIES macro to 0.
	*/
	sched_setaffinity (getpid(), sizeof (cpu_set_t), &affinity);
	sched_yield();

   #else
	/* affinities aren't supported because either the appropriate header files weren't found,
	   or the SUPPORT_AFFINITIES macro was turned off
	*/
	jassertfalse;
	(void) affinityMask;
   #endif
}

bool DynamicLibrary::open (const String& name)
{
	close();
	handle = dlopen (name.toUTF8(), RTLD_LOCAL | RTLD_NOW);
	return handle != 0;
}

void DynamicLibrary::close()
{
	if (handle != nullptr)
	{
		dlclose (handle);
		handle = nullptr;
	}
}

void* DynamicLibrary::getFunction (const String& functionName) noexcept
{
	return handle != nullptr ? dlsym (handle, functionName.toUTF8()) : nullptr;
}

/*** End of inlined file: juce_posix_SharedCode.h ***/


/*** Start of inlined file: juce_linux_Files.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

enum
{
	U_ISOFS_SUPER_MAGIC = 0x9660,   // linux/iso_fs.h
	U_MSDOS_SUPER_MAGIC = 0x4d44,   // linux/msdos_fs.h
	U_NFS_SUPER_MAGIC = 0x6969,	 // linux/nfs_fs.h
	U_SMB_SUPER_MAGIC = 0x517B	  // linux/smb_fs.h
};

bool File::copyInternal (const File& dest) const
{
	FileInputStream in (*this);

	if (dest.deleteFile())
	{
		{
			FileOutputStream out (dest);

			if (out.failedToOpen())
				return false;

			if (out.writeFromInputStream (in, -1) == getSize())
				return true;
		}

		dest.deleteFile();
	}

	return false;
}

void File::findFileSystemRoots (Array<File>& destArray)
{
	destArray.add (File ("/"));
}

bool File::isOnCDRomDrive() const
{
	struct statfs buf;

	return statfs (getFullPathName().toUTF8(), &buf) == 0
			 && buf.f_type == (short) U_ISOFS_SUPER_MAGIC;
}

bool File::isOnHardDisk() const
{
	struct statfs buf;

	if (statfs (getFullPathName().toUTF8(), &buf) == 0)
	{
		switch (buf.f_type)
		{
			case U_ISOFS_SUPER_MAGIC:   // CD-ROM
			case U_MSDOS_SUPER_MAGIC:   // Probably floppy (but could be mounted FAT filesystem)
			case U_NFS_SUPER_MAGIC:	 // Network NFS
			case U_SMB_SUPER_MAGIC:	 // Network Samba
				return false;

			default:
				// Assume anything else is a hard-disk (but note it could
				// be a RAM disk.  There isn't a good way of determining
				// this for sure)
				return true;
		}
	}

	// Assume so if this fails for some reason
	return true;
}

bool File::isOnRemovableDrive() const
{
	jassertfalse; // xxx not implemented for linux!
	return false;
}

bool File::isHidden() const
{
	return getFileName().startsWithChar ('.');
}

namespace
{
	File juce_readlink (const String& file, const File& defaultFile)
	{
		const int size = 8192;
		HeapBlock<char> buffer;
		buffer.malloc (size + 4);

		const size_t numBytes = readlink (file.toUTF8(), buffer, size);

		if (numBytes > 0 && numBytes <= size)
			return File (file).getSiblingFile (String::fromUTF8 (buffer, (int) numBytes));

		return defaultFile;
	}
}

const char* juce_Argv0 = nullptr;  // referenced from juce_Application.cpp

File File::getSpecialLocation (const SpecialLocationType type)
{
	switch (type)
	{
	case userHomeDirectory:
	{
		const char* homeDir = getenv ("HOME");

		if (homeDir == nullptr)
		{
			struct passwd* const pw = getpwuid (getuid());
			if (pw != nullptr)
				homeDir = pw->pw_dir;
		}

		return File (CharPointer_UTF8 (homeDir));
	}

	case userDocumentsDirectory:
	case userMusicDirectory:
	case userMoviesDirectory:
	case userApplicationDataDirectory:
		return File ("~");

	case userDesktopDirectory:
		return File ("~/Desktop");

	case commonApplicationDataDirectory:
		return File ("/var");

	case globalApplicationsDirectory:
		return File ("/usr");

	case tempDirectory:
	{
		File tmp ("/var/tmp");

		if (! tmp.isDirectory())
		{
			tmp = "/tmp";

			if (! tmp.isDirectory())
				tmp = File::getCurrentWorkingDirectory();
		}

		return tmp;
	}

	case invokedExecutableFile:
		if (juce_Argv0 != nullptr)
			return File (CharPointer_UTF8 (juce_Argv0));
		// deliberate fall-through...

	case currentExecutableFile:
	case currentApplicationFile:
		return juce_getExecutableFile();

	case hostApplicationPath:
		return juce_readlink ("/proc/self/exe", juce_getExecutableFile());

	default:
		jassertfalse; // unknown type?
		break;
	}

	return File::nonexistent;
}

String File::getVersion() const
{
	return String::empty; // xxx not yet implemented
}

bool File::moveToTrash() const
{
	if (! exists())
		return true;

	File trashCan ("~/.Trash");

	if (! trashCan.isDirectory())
		trashCan = "~/.local/share/Trash/files";

	if (! trashCan.isDirectory())
		return false;

	return moveFileTo (trashCan.getNonexistentChildFile (getFileNameWithoutExtension(),
														 getFileExtension()));
}

class DirectoryIterator::NativeIterator::Pimpl
{
public:
	Pimpl (const File& directory, const String& wildCard_)
		: parentDir (File::addTrailingSeparator (directory.getFullPathName())),
		  wildCard (wildCard_),
		  dir (opendir (directory.getFullPathName().toUTF8()))
	{
	}

	~Pimpl()
	{
		if (dir != nullptr)
			closedir (dir);
	}

	bool next (String& filenameFound,
			   bool* const isDir, bool* const isHidden, int64* const fileSize,
			   Time* const modTime, Time* const creationTime, bool* const isReadOnly)
	{
		if (dir != nullptr)
		{
			const char* wildcardUTF8 = nullptr;

			for (;;)
			{
				struct dirent* const de = readdir (dir);

				if (de == nullptr)
					break;

				if (wildcardUTF8 == nullptr)
					wildcardUTF8 = wildCard.toUTF8();

				if (fnmatch (wildcardUTF8, de->d_name, FNM_CASEFOLD) == 0)
				{
					filenameFound = CharPointer_UTF8 (de->d_name);

					updateStatInfoForFile (parentDir + filenameFound, isDir, fileSize, modTime, creationTime, isReadOnly);

					if (isHidden != nullptr)
						*isHidden = filenameFound.startsWithChar ('.');

					return true;
				}
			}
		}

		return false;
	}

private:
	String parentDir, wildCard;
	DIR* dir;

	JUCE_DECLARE_NON_COPYABLE (Pimpl);
};

DirectoryIterator::NativeIterator::NativeIterator (const File& directory, const String& wildCard)
	: pimpl (new DirectoryIterator::NativeIterator::Pimpl (directory, wildCard))
{
}

DirectoryIterator::NativeIterator::~NativeIterator()
{
}

bool DirectoryIterator::NativeIterator::next (String& filenameFound,
											  bool* const isDir, bool* const isHidden, int64* const fileSize,
											  Time* const modTime, Time* const creationTime, bool* const isReadOnly)
{
	return pimpl->next (filenameFound, isDir, isHidden, fileSize, modTime, creationTime, isReadOnly);
}

bool Process::openDocument (const String& fileName, const String& parameters)
{
	String cmdString (fileName.replace (" ", "\\ ",false));
	cmdString << " " << parameters;

	if (URL::isProbablyAWebsiteURL (fileName)
		 || cmdString.startsWithIgnoreCase ("file:")
		 || URL::isProbablyAnEmailAddress (fileName))
	{
		// create a command that tries to launch a bunch of likely browsers
		const char* const browserNames[] = { "xdg-open", "/etc/alternatives/x-www-browser", "firefox", "mozilla", "konqueror", "opera" };

		StringArray cmdLines;

		for (int i = 0; i < numElementsInArray (browserNames); ++i)
			cmdLines.add (String (browserNames[i]) + " " + cmdString.trim().quoted());

		cmdString = cmdLines.joinIntoString (" || ");
	}

	const char* const argv[4] = { "/bin/sh", "-c", cmdString.toUTF8(), 0 };

	const int cpid = fork();

	if (cpid == 0)
	{
		setsid();

		// Child process
		execve (argv[0], (char**) argv, environ);
		exit (0);
	}

	return cpid >= 0;
}

void File::revealToUser() const
{
	if (isDirectory())
		startAsProcess();
	else if (getParentDirectory().exists())
		getParentDirectory().startAsProcess();
}

#endif

/*** End of inlined file: juce_linux_Files.cpp ***/


/*** Start of inlined file: juce_posix_NamedPipe.cpp ***/
// (This file gets included by juce_mac_NativeCode.mm, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

struct NamedPipeInternal
{
	String pipeInName, pipeOutName;
	int pipeIn, pipeOut;

	bool volatile createdPipe, blocked, stopReadOperation;

	static void signalHandler (int) {}
};

void NamedPipe::cancelPendingReads()
{
	while (internal != nullptr && static_cast <NamedPipeInternal*> (internal)->blocked)
	{
		NamedPipeInternal* const intern = static_cast <NamedPipeInternal*> (internal);

		intern->stopReadOperation = true;

		char buffer [1] = { 0 };
		int bytesWritten = (int) ::write (intern->pipeIn, buffer, 1);
		(void) bytesWritten;

		int timeout = 2000;
		while (intern->blocked && --timeout >= 0)
			Thread::sleep (2);

		intern->stopReadOperation = false;
	}
}

void NamedPipe::close()
{
	NamedPipeInternal* const intern = static_cast <NamedPipeInternal*> (internal);

	if (intern != nullptr)
	{
		internal = nullptr;

		if (intern->pipeIn != -1)
			::close (intern->pipeIn);

		if (intern->pipeOut != -1)
			::close (intern->pipeOut);

		if (intern->createdPipe)
		{
			unlink (intern->pipeInName.toUTF8());
			unlink (intern->pipeOutName.toUTF8());
		}

		delete intern;
	}
}

bool NamedPipe::openInternal (const String& pipeName, const bool createPipe)
{
	close();

	NamedPipeInternal* const intern = new NamedPipeInternal();
	internal = intern;
	intern->createdPipe = createPipe;
	intern->blocked = false;
	intern->stopReadOperation = false;

	signal (SIGPIPE, NamedPipeInternal::signalHandler);
	siginterrupt (SIGPIPE, 1);

	const String pipePath ("/tmp/" + File::createLegalFileName (pipeName));

	intern->pipeInName  = pipePath + "_in";
	intern->pipeOutName = pipePath + "_out";
	intern->pipeIn = -1;
	intern->pipeOut = -1;

	if (createPipe)
	{
		if ((mkfifo (intern->pipeInName.toUTF8(), 0666) && errno != EEXIST)
			|| (mkfifo (intern->pipeOutName.toUTF8(), 0666) && errno != EEXIST))
		{
			delete intern;
			internal = nullptr;

			return false;
		}
	}

	return true;
}

int NamedPipe::read (void* destBuffer, int maxBytesToRead, int /*timeOutMilliseconds*/)
{
	int bytesRead = -1;
	NamedPipeInternal* const intern = static_cast <NamedPipeInternal*> (internal);

	if (intern != nullptr)
	{
		intern->blocked = true;

		if (intern->pipeIn == -1)
		{
			if (intern->createdPipe)
				intern->pipeIn = ::open (intern->pipeInName.toUTF8(), O_RDWR);
			else
				intern->pipeIn = ::open (intern->pipeOutName.toUTF8(), O_RDWR);

			if (intern->pipeIn == -1)
			{
				intern->blocked = false;
				return -1;
			}
		}

		bytesRead = 0;

		char* p = static_cast<char*> (destBuffer);

		while (bytesRead < maxBytesToRead)
		{
			const int bytesThisTime = maxBytesToRead - bytesRead;
			const int numRead = (int) ::read (intern->pipeIn, p, bytesThisTime);

			if (numRead <= 0 || intern->stopReadOperation)
			{
				bytesRead = -1;
				break;
			}

			bytesRead += numRead;
			p += bytesRead;
		}

		intern->blocked = false;
	}

	return bytesRead;
}

int NamedPipe::write (const void* sourceBuffer, int numBytesToWrite, int timeOutMilliseconds)
{
	int bytesWritten = -1;
	NamedPipeInternal* const intern = static_cast <NamedPipeInternal*> (internal);

	if (intern != nullptr)
	{
		if (intern->pipeOut == -1)
		{
			if (intern->createdPipe)
				intern->pipeOut = ::open (intern->pipeOutName.toUTF8(), O_WRONLY);
			else
				intern->pipeOut = ::open (intern->pipeInName.toUTF8(), O_WRONLY);

			if (intern->pipeOut == -1)
			{
				return -1;
			}
		}

		const char* p = static_cast<const char*> (sourceBuffer);
		bytesWritten = 0;

		const uint32 timeOutTime = Time::getMillisecondCounter() + timeOutMilliseconds;

		while (bytesWritten < numBytesToWrite
			   && (timeOutMilliseconds < 0 || Time::getMillisecondCounter() < timeOutTime))
		{
			const int bytesThisTime = numBytesToWrite - bytesWritten;
			const int numWritten = (int) ::write (intern->pipeOut, p, bytesThisTime);

			if (numWritten <= 0)
			{
				bytesWritten = -1;
				break;
			}

			bytesWritten += numWritten;
			p += bytesWritten;
		}
	}

	return bytesWritten;
}

#endif

/*** End of inlined file: juce_posix_NamedPipe.cpp ***/


/*** Start of inlined file: juce_linux_Network.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

bool Process::openEmailWithAttachments (const String& targetEmailAddress,
										const String& emailSubject,
										const String& bodyText,
										const StringArray& filesToAttach)
{
	jassertfalse;	// xxx todo

	return false;
}

#endif

/*** End of inlined file: juce_linux_Network.cpp ***/


/*** Start of inlined file: juce_linux_SystemStats.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

void Logger::outputDebugString (const String& text)
{
	std::cerr << text << std::endl;
}

SystemStats::OperatingSystemType SystemStats::getOperatingSystemType()
{
	return Linux;
}

String SystemStats::getOperatingSystemName()
{
	return "Linux";
}

bool SystemStats::isOperatingSystem64Bit()
{
   #if JUCE_64BIT
	return true;
   #else
	//xxx not sure how to find this out?..
	return false;
   #endif
}

namespace LinuxStatsHelpers
{
	String getCpuInfo (const char* const key)
	{
		StringArray lines;
		File ("/proc/cpuinfo").readLines (lines);

		for (int i = lines.size(); --i >= 0;) // (NB - it's important that this runs in reverse order)
			if (lines[i].startsWithIgnoreCase (key))
				return lines[i].fromFirstOccurrenceOf (":", false, false).trim();

		return String::empty;
	}
}

String SystemStats::getCpuVendor()
{
	return LinuxStatsHelpers::getCpuInfo ("vendor_id");
}

int SystemStats::getCpuSpeedInMegaherz()
{
	return roundToInt (LinuxStatsHelpers::getCpuInfo ("cpu MHz").getFloatValue());
}

int SystemStats::getMemorySizeInMegabytes()
{
	struct sysinfo sysi;

	if (sysinfo (&sysi) == 0)
		return (sysi.totalram * sysi.mem_unit / (1024 * 1024));

	return 0;
}

int SystemStats::getPageSize()
{
	return sysconf (_SC_PAGESIZE);
}

String SystemStats::getLogonName()
{
	const char* user = getenv ("USER");

	if (user == nullptr)
	{
		struct passwd* const pw = getpwuid (getuid());
		if (pw != nullptr)
			user = pw->pw_name;
	}

	return CharPointer_UTF8 (user);
}

String SystemStats::getFullUserName()
{
	return getLogonName();
}

String SystemStats::getComputerName()
{
	char name [256] = { 0 };
	if (gethostname (name, sizeof (name) - 1) == 0)
		return name;

	return String::empty;
}

SystemStats::CPUFlags::CPUFlags()
{
	const String flags (LinuxStatsHelpers::getCpuInfo ("flags"));
	hasMMX   = flags.contains ("mmx");
	hasSSE   = flags.contains ("sse");
	hasSSE2  = flags.contains ("sse2");
	has3DNow = flags.contains ("3dnow");

	numCpus = LinuxStatsHelpers::getCpuInfo ("processor").getIntValue() + 1;
}

uint32 juce_millisecondsSinceStartup() noexcept
{
	timespec t;
	clock_gettime (CLOCK_MONOTONIC, &t);

	return t.tv_sec * 1000 + t.tv_nsec / 1000000;
}

int64 Time::getHighResolutionTicks() noexcept
{
	timespec t;
	clock_gettime (CLOCK_MONOTONIC, &t);

	return (t.tv_sec * (int64) 1000000) + (t.tv_nsec / 1000);
}

int64 Time::getHighResolutionTicksPerSecond() noexcept
{
	return 1000000;  // (microseconds)
}

double Time::getMillisecondCounterHiRes() noexcept
{
	return getHighResolutionTicks() * 0.001;
}

bool Time::setSystemTimeToThisTime() const
{
	timeval t;
	t.tv_sec = millisSinceEpoch / 1000;
	t.tv_usec = (millisSinceEpoch - t.tv_sec * 1000) * 1000;

	return settimeofday (&t, 0) == 0;
}

#endif

/*** End of inlined file: juce_linux_SystemStats.cpp ***/


/*** Start of inlined file: juce_linux_Threads.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

/*
	Note that a lot of methods that you'd expect to find in this file actually
	live in juce_posix_SharedCode.h!
*/

void Process::setPriority (const ProcessPriority prior)
{
	const int policy = (prior <= NormalPriority) ? SCHED_OTHER : SCHED_RR;
	const int minp = sched_get_priority_min (policy);
	const int maxp = sched_get_priority_max (policy);

	struct sched_param param;

	switch (prior)
	{
		case LowPriority:
		case NormalPriority:	param.sched_priority = 0; break;
		case HighPriority:	  param.sched_priority = minp + (maxp - minp) / 4; break;
		case RealtimePriority:  param.sched_priority = minp + (3 * (maxp - minp) / 4); break;
		default:		jassertfalse; break;
	}

	pthread_setschedparam (pthread_self(), policy, &param);
}

void Process::terminate()
{
	exit (0);
}

JUCE_API bool JUCE_CALLTYPE juce_isRunningUnderDebugger()
{
	static char testResult = 0;

	if (testResult == 0)
	{
		testResult = (char) ptrace (PT_TRACE_ME, 0, 0, 0);

		if (testResult >= 0)
		{
			ptrace (PT_DETACH, 0, (caddr_t) 1, 0);
			testResult = 1;
		}
	}

	return testResult < 0;
}

JUCE_API bool JUCE_CALLTYPE Process::isRunningUnderDebugger()
{
	return juce_isRunningUnderDebugger();
}

void Process::raisePrivilege()
{
	// If running suid root, change effective user to root
	if (geteuid() != 0 && getuid() == 0)
	{
		setreuid (geteuid(), getuid());
		setregid (getegid(), getgid());
	}
}

void Process::lowerPrivilege()
{
	// If runing suid root, change effective user back to real user
	if (geteuid() == 0 && getuid() != 0)
	{
		setreuid (geteuid(), getuid());
		setregid (getegid(), getgid());
	}
}

#endif

/*** End of inlined file: juce_linux_Threads.cpp ***/

#if ! JUCE_ONLY_BUILD_CORE_LIBRARY

/*** Start of inlined file: juce_linux_Clipboard.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

extern Display* display;
extern Window juce_messageWindowHandle;

namespace ClipboardHelpers
{
	static String localClipboardContent;
	static Atom   atom_UTF8_STRING;
	static Atom   atom_CLIPBOARD;
	static Atom   atom_TARGETS;

	static void initSelectionAtoms()
	{
		static bool isInitialised = false;
		if (! isInitialised)
		{
			atom_UTF8_STRING = XInternAtom (display, "UTF8_STRING", False);
			atom_CLIPBOARD   = XInternAtom (display, "CLIPBOARD", False);
			atom_TARGETS     = XInternAtom (display, "TARGETS", False);
		}
	}

	// Read the content of a window property as either a locale-dependent string or an utf8 string
	// works only for strings shorter than 1000000 bytes
	static String readWindowProperty (Window window, Atom prop, Atom fmt)
	{
		String returnData;
		char* clipData;
		Atom actualType;
		int  actualFormat;
		unsigned long numItems, bytesLeft;

		if (XGetWindowProperty (display, window, prop,
								0L /* offset */, 1000000 /* length (max) */, False,
								AnyPropertyType /* format */,
								&actualType, &actualFormat, &numItems, &bytesLeft,
								(unsigned char**) &clipData) == Success)
		{
			if (actualType == atom_UTF8_STRING && actualFormat == 8)
				returnData = String::fromUTF8 (clipData, numItems);
			else if (actualType == XA_STRING && actualFormat == 8)
				returnData = String (clipData, numItems);

			if (clipData != nullptr)
				XFree (clipData);

			jassert (bytesLeft == 0 || numItems == 1000000);
		}

		XDeleteProperty (display, window, prop);
		return returnData;
	}

	// Send a SelectionRequest to the window owning the selection and waits for its answer (with a timeout) */
	static bool requestSelectionContent (String& selectionContent, Atom selection, Atom requestedFormat)
	{
		Atom property_name = XInternAtom (display, "JUCE_SEL", false);

		// The selection owner will be asked to set the JUCE_SEL property on the
		// juce_messageWindowHandle with the selection content
		XConvertSelection (display, selection, requestedFormat, property_name,
						   juce_messageWindowHandle, CurrentTime);

		int count = 50; // will wait at most for 200 ms

		while (--count >= 0)
		{
			XEvent event;
			if (XCheckTypedWindowEvent (display, juce_messageWindowHandle, SelectionNotify, &event))
			{
				if (event.xselection.property == property_name)
				{
					jassert (event.xselection.requestor == juce_messageWindowHandle);

					selectionContent = readWindowProperty (event.xselection.requestor,
														   event.xselection.property,
														   requestedFormat);
					return true;
				}
				else
				{
					return false; // the format we asked for was denied.. (event.xselection.property == None)
				}
			}

			// not very elegant.. we could do a select() or something like that...
			// however clipboard content requesting is inherently slow on x11, it
			// often takes 50ms or more so...
			Thread::sleep (4);
		}

		return false;
	}
}

// Called from the event loop in juce_linux_Messaging in response to SelectionRequest events
void juce_handleSelectionRequest (XSelectionRequestEvent &evt)
{
	ClipboardHelpers::initSelectionAtoms();

	// the selection content is sent to the target window as a window property
	XSelectionEvent reply;
	reply.type = SelectionNotify;
	reply.display = evt.display;
	reply.requestor = evt.requestor;
	reply.selection = evt.selection;
	reply.target = evt.target;
	reply.property = None; // == "fail"
	reply.time = evt.time;

	HeapBlock <char> data;
	int propertyFormat = 0, numDataItems = 0;

	if (evt.selection == XA_PRIMARY || evt.selection == ClipboardHelpers::atom_CLIPBOARD)
	{
		if (evt.target == XA_STRING || evt.target == ClipboardHelpers::atom_UTF8_STRING)
		{
			// translate to utf8
			numDataItems = ClipboardHelpers::localClipboardContent.getNumBytesAsUTF8() + 1;
			data.calloc (numDataItems + 1);
			ClipboardHelpers::localClipboardContent.copyToUTF8 (data, numDataItems);
			propertyFormat = 8; // bits/item
		}
		else if (evt.target == ClipboardHelpers::atom_TARGETS)
		{
			// another application wants to know what we are able to send
			numDataItems = 2;
			propertyFormat = 32; // atoms are 32-bit
			data.calloc (numDataItems * 4);
			Atom* atoms = reinterpret_cast<Atom*> (data.getData());
			atoms[0] = ClipboardHelpers::atom_UTF8_STRING;
			atoms[1] = XA_STRING;
		}
	}
	else
	{
		DBG ("requested unsupported clipboard");
	}

	if (data != nullptr)
	{
		const int maxReasonableSelectionSize = 1000000;

		// for very big chunks of data, we should use the "INCR" protocol , which is a pain in the *ss
		if (evt.property != None && numDataItems < maxReasonableSelectionSize)
		{
			XChangeProperty (evt.display, evt.requestor,
							 evt.property, evt.target,
							 propertyFormat /* 8 or 32 */, PropModeReplace,
							 reinterpret_cast<const unsigned char*> (data.getData()), numDataItems);
			reply.property = evt.property; // " == success"
		}
	}

	XSendEvent (evt.display, evt.requestor, 0, NoEventMask, (XEvent*) &reply);
}

void SystemClipboard::copyTextToClipboard (const String& clipText)
{
	ClipboardHelpers::initSelectionAtoms();
	ClipboardHelpers::localClipboardContent = clipText;

	XSetSelectionOwner (display, XA_PRIMARY, juce_messageWindowHandle, CurrentTime);
	XSetSelectionOwner (display, ClipboardHelpers::atom_CLIPBOARD, juce_messageWindowHandle, CurrentTime);
}

String SystemClipboard::getTextFromClipboard()
{
	ClipboardHelpers::initSelectionAtoms();

	/* 1) try to read from the "CLIPBOARD" selection first (the "high
	   level" clipboard that is supposed to be filled by ctrl-C
	   etc). When a clipboard manager is running, the content of this
	   selection is preserved even when the original selection owner
	   exits.

	   2) and then try to read from "PRIMARY" selection (the "legacy" selection
	   filled by good old x11 apps such as xterm)
	*/
	String content;
	Atom selection = XA_PRIMARY;
	Window selectionOwner = None;

	if ((selectionOwner = XGetSelectionOwner (display, selection)) == None)
	{
		selection = ClipboardHelpers::atom_CLIPBOARD;
		selectionOwner = XGetSelectionOwner (display, selection);
	}

	if (selectionOwner != None)
	{
		if (selectionOwner == juce_messageWindowHandle)
		{
			content = ClipboardHelpers::localClipboardContent;
		}
		else
		{
			// first try: we want an utf8 string
			bool ok = ClipboardHelpers::requestSelectionContent (content, selection, ClipboardHelpers::atom_UTF8_STRING);

			if (! ok)
			{
				// second chance, ask for a good old locale-dependent string ..
				ok = ClipboardHelpers::requestSelectionContent (content, selection, XA_STRING);
			}
		}
	}

	return content;
}

#endif

/*** End of inlined file: juce_linux_Clipboard.cpp ***/



/*** Start of inlined file: juce_linux_Messaging.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

#if JUCE_DEBUG && ! defined (JUCE_DEBUG_XERRORS)
 #define JUCE_DEBUG_XERRORS 1
#endif

Display* display = nullptr;
Window juce_messageWindowHandle = None;
XContext windowHandleXContext;   // This is referenced from Windowing.cpp

extern void juce_windowMessageReceive (XEvent* event);  // Defined in Windowing.cpp
extern void juce_handleSelectionRequest (XSelectionRequestEvent &evt);  // Defined in Clipboard.cpp

ScopedXLock::ScopedXLock()	   { XLockDisplay (display); }
ScopedXLock::~ScopedXLock()	  { XUnlockDisplay (display); }

class InternalMessageQueue
{
public:
	InternalMessageQueue()
		: bytesInSocket (0),
		  totalEventCount (0)
	{
		int ret = ::socketpair (AF_LOCAL, SOCK_STREAM, 0, fd);
		(void) ret; jassert (ret == 0);

		//setNonBlocking (fd[0]);
		//setNonBlocking (fd[1]);
	}

	~InternalMessageQueue()
	{
		close (fd[0]);
		close (fd[1]);

		clearSingletonInstance();
	}

	void postMessage (Message* msg)
	{
		const int maxBytesInSocketQueue = 128;

		ScopedLock sl (lock);
		queue.add (msg);

		if (bytesInSocket < maxBytesInSocketQueue)
		{
			++bytesInSocket;

			ScopedUnlock ul (lock);
			const unsigned char x = 0xff;
			size_t bytesWritten = write (fd[0], &x, 1);
			(void) bytesWritten;
		}
	}

	bool isEmpty() const
	{
		ScopedLock sl (lock);
		return queue.size() == 0;
	}

	bool dispatchNextEvent()
	{
		// This alternates between giving priority to XEvents or internal messages,
		// to keep everything running smoothly..
		if ((++totalEventCount & 1) != 0)
			return dispatchNextXEvent() || dispatchNextInternalMessage();
		else
			return dispatchNextInternalMessage() || dispatchNextXEvent();
	}

	// Wait for an event (either XEvent, or an internal Message)
	bool sleepUntilEvent (const int timeoutMs)
	{
		if (! isEmpty())
			return true;

		if (display != 0)
		{
			ScopedXLock xlock;
			if (XPending (display))
				return true;
		}

		struct timeval tv;
		tv.tv_sec = 0;
		tv.tv_usec = timeoutMs * 1000;
		int fd0 = getWaitHandle();
		int fdmax = fd0;

		fd_set readset;
		FD_ZERO (&readset);
		FD_SET (fd0, &readset);

		if (display != 0)
		{
			ScopedXLock xlock;
			int fd1 = XConnectionNumber (display);
			FD_SET (fd1, &readset);
			fdmax = jmax (fd0, fd1);
		}

		const int ret = select (fdmax + 1, &readset, 0, 0, &tv);
		return (ret > 0); // ret <= 0 if error or timeout
	}

	struct MessageThreadFuncCall
	{
		enum { uniqueID = 0x73774623 };

		MessageCallbackFunction* func;
		void* parameter;
		void* result;
		CriticalSection lock;
		WaitableEvent event;
	};

	juce_DeclareSingleton_SingleThreaded_Minimal (InternalMessageQueue);

private:
	CriticalSection lock;
	ReferenceCountedArray <Message> queue;
	int fd[2];
	int bytesInSocket;
	int totalEventCount;

	int getWaitHandle() const noexcept	  { return fd[1]; }

	static bool setNonBlocking (int handle)
	{
		int socketFlags = fcntl (handle, F_GETFL, 0);
		if (socketFlags == -1)
			return false;

		socketFlags |= O_NONBLOCK;
		return fcntl (handle, F_SETFL, socketFlags) == 0;
	}

	static bool dispatchNextXEvent()
	{
		if (display == 0)
			return false;

		XEvent evt;

		{
			ScopedXLock xlock;
			if (! XPending (display))
				return false;

			XNextEvent (display, &evt);
		}

		if (evt.type == SelectionRequest && evt.xany.window == juce_messageWindowHandle)
			juce_handleSelectionRequest (evt.xselectionrequest);
		else if (evt.xany.window != juce_messageWindowHandle)
			juce_windowMessageReceive (&evt);

		return true;
	}

	Message::Ptr popNextMessage()
	{
		const ScopedLock sl (lock);

		if (bytesInSocket > 0)
		{
			--bytesInSocket;

			const ScopedUnlock ul (lock);
			unsigned char x;
			size_t numBytes = read (fd[1], &x, 1);
			(void) numBytes;
		}

		return queue.removeAndReturn (0);
	}

	bool dispatchNextInternalMessage()
	{
		const Message::Ptr msg (popNextMessage());

		if (msg == nullptr)
			return false;

		if (msg->intParameter1 == MessageThreadFuncCall::uniqueID)
		{
			// Handle callback message
			MessageThreadFuncCall* const call = (MessageThreadFuncCall*) msg->pointerParameter;

			call->result = (*(call->func)) (call->parameter);
			call->event.signal();
		}
		else
		{
			// Handle "normal" messages
			MessageManager::getInstance()->deliverMessage (msg);
		}

		return true;
	}
};

juce_ImplementSingleton_SingleThreaded (InternalMessageQueue);

namespace LinuxErrorHandling
{

	static bool errorOccurred = false;
	static bool keyboardBreakOccurred = false;
	static XErrorHandler oldErrorHandler = (XErrorHandler) 0;
	static XIOErrorHandler oldIOErrorHandler = (XIOErrorHandler) 0;

	// Usually happens when client-server connection is broken
	static int ioErrorHandler (Display* display)
	{
		DBG ("ERROR: connection to X server broken.. terminating.");

		if (JUCEApplication::isStandaloneApp())
			MessageManager::getInstance()->stopDispatchLoop();

		errorOccurred = true;
		return 0;
	}

	// A protocol error has occurred
	static int juce_XErrorHandler (Display* display, XErrorEvent* event)
	{
	   #if JUCE_DEBUG_XERRORS
		char errorStr[64] = { 0 };
		char requestStr[64] = { 0 };

		XGetErrorText (display, event->error_code, errorStr, 64);
		XGetErrorDatabaseText (display, "XRequest", String (event->request_code).toUTF8(), "Unknown", requestStr, 64);
		DBG ("ERROR: X returned " + String (errorStr) + " for operation " + String (requestStr));
	   #endif

		return 0;
	}

	static void installXErrorHandlers()
	{
		oldIOErrorHandler = XSetIOErrorHandler (ioErrorHandler);
		oldErrorHandler = XSetErrorHandler (juce_XErrorHandler);
	}

	static void removeXErrorHandlers()
	{
		if (JUCEApplication::isStandaloneApp())
		{
			XSetIOErrorHandler (oldIOErrorHandler);
			oldIOErrorHandler = 0;

			XSetErrorHandler (oldErrorHandler);
			oldErrorHandler = 0;
		}
	}

	static void keyboardBreakSignalHandler (int sig)
	{
		if (sig == SIGINT)
			keyboardBreakOccurred = true;
	}

	static void installKeyboardBreakHandler()
	{
		struct sigaction saction;
		sigset_t maskSet;
		sigemptyset (&maskSet);
		saction.sa_handler = keyboardBreakSignalHandler;
		saction.sa_mask = maskSet;
		saction.sa_flags = 0;
		sigaction (SIGINT, &saction, 0);
	}
}

void MessageManager::doPlatformSpecificInitialisation()
{
	if (JUCEApplication::isStandaloneApp())
	{
		// Initialise xlib for multiple thread support
		static bool initThreadCalled = false;

		if (! initThreadCalled)
		{
			if (! XInitThreads())
			{
				// This is fatal!  Print error and closedown
				Logger::outputDebugString ("Failed to initialise xlib thread support.");
				Process::terminate();
				return;
			}

			initThreadCalled = true;
		}

		LinuxErrorHandling::installXErrorHandlers();
		LinuxErrorHandling::installKeyboardBreakHandler();
	}

	// Create the internal message queue
	InternalMessageQueue::getInstance();

	// Try to connect to a display
	String displayName (getenv ("DISPLAY"));
	if (displayName.isEmpty())
		displayName = ":0.0";

	display = XOpenDisplay (displayName.toUTF8());

	if (display != 0)  // This is not fatal! we can run headless.
	{
		// Create a context to store user data associated with Windows we create in WindowDriver
		windowHandleXContext = XUniqueContext();

		// We're only interested in client messages for this window, which are always sent
		XSetWindowAttributes swa;
		swa.event_mask = NoEventMask;

		// Create our message window (this will never be mapped)
		const int screen = DefaultScreen (display);
		juce_messageWindowHandle = XCreateWindow (display, RootWindow (display, screen),
												  0, 0, 1, 1, 0, 0, InputOnly,
												  DefaultVisual (display, screen),
												  CWEventMask, &swa);
	}
}

void MessageManager::doPlatformSpecificShutdown()
{
	InternalMessageQueue::deleteInstance();

	if (display != 0 && ! LinuxErrorHandling::errorOccurred)
	{
		XDestroyWindow (display, juce_messageWindowHandle);
		XCloseDisplay (display);

		juce_messageWindowHandle = 0;
		display = nullptr;

		LinuxErrorHandling::removeXErrorHandlers();
	}
}

bool MessageManager::postMessageToSystemQueue (Message* message)
{
	if (LinuxErrorHandling::errorOccurred)
		return false;

	InternalMessageQueue::getInstanceWithoutCreating()->postMessage (message);
	return true;
}

void MessageManager::broadcastMessage (const String& value)
{
	/* TODO */
}

class AsyncFunctionCaller   : public AsyncUpdater
{
public:
	static void* call (MessageCallbackFunction* func_, void* parameter_)
	{
		if (MessageManager::getInstance()->isThisTheMessageThread())
			return func_ (parameter_);

		AsyncFunctionCaller caller (func_, parameter_);
		caller.triggerAsyncUpdate();
		caller.finished.wait();
		return caller.result;
	}

	void handleAsyncUpdate()
	{
		result = (*func) (parameter);
		finished.signal();
	}

private:
	WaitableEvent finished;
	MessageCallbackFunction* func;
	void* parameter;
	void* volatile result;

	AsyncFunctionCaller (MessageCallbackFunction* func_, void* parameter_)
		: result (0), func (func_), parameter (parameter_)
	{}

	JUCE_DECLARE_NON_COPYABLE (AsyncFunctionCaller);
};

void* MessageManager::callFunctionOnMessageThread (MessageCallbackFunction* func, void* parameter)
{
	if (LinuxErrorHandling::errorOccurred)
		return nullptr;

	return AsyncFunctionCaller::call (func, parameter);
}

// this function expects that it will NEVER be called simultaneously for two concurrent threads
bool MessageManager::dispatchNextMessageOnSystemQueue (bool returnIfNoPendingMessages)
{
	while (! LinuxErrorHandling::errorOccurred)
	{
		if (LinuxErrorHandling::keyboardBreakOccurred)
		{
			LinuxErrorHandling::errorOccurred = true;

			if (JUCEApplication::isStandaloneApp())
				Process::terminate();

			break;
		}

		if (InternalMessageQueue::getInstanceWithoutCreating()->dispatchNextEvent())
			return true;

		if (returnIfNoPendingMessages)
			break;

		InternalMessageQueue::getInstanceWithoutCreating()->sleepUntilEvent (2000);
	}

	return false;
}

#endif

/*** End of inlined file: juce_linux_Messaging.cpp ***/


/*** Start of inlined file: juce_linux_Fonts.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

class FreeTypeFontFace
{
public:

	enum FontStyle
	{
		Plain = 0,
		Bold = 1,
		Italic = 2
	};

	FreeTypeFontFace (const String& familyName)
	  : hasSerif (false),
		monospaced (false)
	{
		family = familyName;
	}

	void setFileName (const String& name, const int faceIndex, FontStyle style)
	{
		if (names [(int) style].fileName.isEmpty())
		{
			names [(int) style].fileName = name;
			names [(int) style].faceIndex = faceIndex;
		}
	}

	const String& getFamilyName() const noexcept	{ return family; }

	const String& getFileName (const int style, int& faceIndex) const noexcept
	{
		faceIndex = names[style].faceIndex;
		return names[style].fileName;
	}

	void setMonospaced (bool mono) noexcept	 { monospaced = mono; }
	bool getMonospaced() const noexcept	 { return monospaced; }

	void setSerif (const bool serif) noexcept   { hasSerif = serif; }
	bool getSerif() const noexcept		  { return hasSerif; }

private:

	String family;

	struct FontNameIndex
	{
		String fileName;
		int faceIndex;
	};

	FontNameIndex names[4];
	bool hasSerif, monospaced;
};

class LinuxFontFileIterator
{
public:
	LinuxFontFileIterator()
		: index (0)
	{
		fontDirs.addTokens (CharPointer_UTF8 (getenv ("JUCE_FONT_PATH")), ";,", String::empty);
		fontDirs.removeEmptyStrings (true);

		if (fontDirs.size() == 0)
		{
			const ScopedPointer<XmlElement> fontsInfo (XmlDocument::parse (File ("/etc/fonts/fonts.conf")));

			if (fontsInfo != nullptr)
			{
				forEachXmlChildElementWithTagName (*fontsInfo, e, "dir")
				{
					fontDirs.add (e->getAllSubText().trim());
				}
			}
		}

		if (fontDirs.size() == 0)
			fontDirs.add ("/usr/X11R6/lib/X11/fonts");

		fontDirs.removeEmptyStrings (true);
	}

	bool next()
	{
		if (iter != nullptr)
		{
			while (iter->next())
				if (getFile().hasFileExtension ("ttf;pfb;pcf"))
					return true;
		}

		if (index >= fontDirs.size())
			return false;

		iter = new DirectoryIterator (fontDirs [index++], true);
		return next();
	}

	File getFile() const	{ jassert (iter != nullptr); return iter->getFile(); }

private:
	StringArray fontDirs;
	int index;
	ScopedPointer<DirectoryIterator> iter;
};

class FreeTypeInterface  : public DeletedAtShutdown
{
public:

	FreeTypeInterface()
		: ftLib (0),
		  lastFace (0),
		  lastBold (false),
		  lastItalic (false)
	{
		if (FT_Init_FreeType (&ftLib) != 0)
		{
			ftLib = 0;
			DBG ("Failed to initialize FreeType");
		}

		LinuxFontFileIterator fontFileIterator;

		while (fontFileIterator.next())
		{
			FT_Face face;
			int faceIndex = 0;
			int numFaces = 0;

			do
			{
				if (FT_New_Face (ftLib, fontFileIterator.getFile().getFullPathName().toUTF8(),
								 faceIndex, &face) == 0)
				{
					if (faceIndex == 0)
						numFaces = face->num_faces;

					if ((face->face_flags & FT_FACE_FLAG_SCALABLE) != 0)
					{
						FreeTypeFontFace* const newFace = findOrCreate (face->family_name, true);
						int style = (int) FreeTypeFontFace::Plain;

						if ((face->style_flags & FT_STYLE_FLAG_BOLD) != 0)
							style |= (int) FreeTypeFontFace::Bold;

						if ((face->style_flags & FT_STYLE_FLAG_ITALIC) != 0)
							style |= (int) FreeTypeFontFace::Italic;

						newFace->setFileName (fontFileIterator.getFile().getFullPathName(),
											  faceIndex, (FreeTypeFontFace::FontStyle) style);
						newFace->setMonospaced ((face->face_flags & FT_FACE_FLAG_FIXED_WIDTH) != 0);

						// Surely there must be a better way to do this?
						const String name (face->family_name);
						newFace->setSerif (! (name.containsIgnoreCase ("Sans")
											   || name.containsIgnoreCase ("Verdana")
											   || name.containsIgnoreCase ("Arial")));

						//DBG (fontFileIterator.getFile().getFullPathName() << "  -  " << name);
					}

					FT_Done_Face (face);
				}

				++faceIndex;
			}
			while (faceIndex < numFaces);
		}
	}

	~FreeTypeInterface()
	{
		if (lastFace != 0)
			FT_Done_Face (lastFace);

		if (ftLib != 0)
			FT_Done_FreeType (ftLib);

		clearSingletonInstance();
	}

	FreeTypeFontFace* findOrCreate (const String& familyName, const bool create = false)
	{
		for (int i = 0; i < faces.size(); i++)
			if (faces[i]->getFamilyName() == familyName)
				return faces[i];

		if (! create)
			return nullptr;

		FreeTypeFontFace* newFace = new FreeTypeFontFace (familyName);
		faces.add (newFace);

		return newFace;
	}

	// Create a FreeType face object for a given font
	FT_Face createFT_Face (const String& fontName, const bool bold, const bool italic)
	{
		FT_Face face = 0;

		if (fontName == lastFontName && bold == lastBold && italic == lastItalic)
		{
			face = lastFace;
		}
		else
		{
			if (lastFace != 0)
			{
				FT_Done_Face (lastFace);
				lastFace = 0;
			}

			lastFontName = fontName;
			lastBold = bold;
			lastItalic = italic;

			FreeTypeFontFace* const ftFace = findOrCreate (fontName);

			if (ftFace != 0)
			{
				int style = (int) FreeTypeFontFace::Plain;

				if (bold)
					style |= (int) FreeTypeFontFace::Bold;

				if (italic)
					style |= (int) FreeTypeFontFace::Italic;

				int faceIndex;
				String fileName (ftFace->getFileName (style, faceIndex));

				if (fileName.isEmpty())
				{
					style ^= (int) FreeTypeFontFace::Bold;

					fileName = ftFace->getFileName (style, faceIndex);

					if (fileName.isEmpty())
					{
						style ^= (int) FreeTypeFontFace::Bold;
						style ^= (int) FreeTypeFontFace::Italic;

						fileName = ftFace->getFileName (style, faceIndex);

						if (! fileName.length())
						{
							style ^= (int) FreeTypeFontFace::Bold;
							fileName = ftFace->getFileName (style, faceIndex);
						}
					}
				}

				if (! FT_New_Face (ftLib, fileName.toUTF8(), faceIndex, &lastFace))
				{
					face = lastFace;

					// If there isn't a unicode charmap then select the first one.
					if (FT_Select_Charmap (face, ft_encoding_unicode))
						FT_Set_Charmap (face, face->charmaps[0]);
				}
			}
		}

		return face;
	}

	bool addGlyph (FT_Face face, CustomTypeface& dest, uint32 character)
	{
		const unsigned int glyphIndex = FT_Get_Char_Index (face, character);
		const float height = (float) (face->ascender - face->descender);
		const float scaleX = 1.0f / height;
		const float scaleY = -1.0f / height;
		Path destShape;

		if (FT_Load_Glyph (face, glyphIndex, FT_LOAD_NO_SCALE | FT_LOAD_NO_BITMAP | FT_LOAD_IGNORE_TRANSFORM) != 0
			 || face->glyph->format != ft_glyph_format_outline)
		{
			return false;
		}

		const FT_Outline* const outline = &face->glyph->outline;
		const short* const contours = outline->contours;
		const char* const tags = outline->tags;
		FT_Vector* const points = outline->points;

		for (int c = 0; c < outline->n_contours; c++)
		{
			const int startPoint = (c == 0) ? 0 : contours [c - 1] + 1;
			const int endPoint = contours[c];

			for (int p = startPoint; p <= endPoint; p++)
			{
				const float x = scaleX * points[p].x;
				const float y = scaleY * points[p].y;

				if (p == startPoint)
				{
					if (FT_CURVE_TAG (tags[p]) == FT_Curve_Tag_Conic)
					{
						float x2 = scaleX * points [endPoint].x;
						float y2 = scaleY * points [endPoint].y;

						if (FT_CURVE_TAG (tags[endPoint]) != FT_Curve_Tag_On)
						{
							x2 = (x + x2) * 0.5f;
							y2 = (y + y2) * 0.5f;
						}

						destShape.startNewSubPath (x2, y2);
					}
					else
					{
						destShape.startNewSubPath (x, y);
					}
				}

				if (FT_CURVE_TAG (tags[p]) == FT_Curve_Tag_On)
				{
					if (p != startPoint)
						destShape.lineTo (x, y);
				}
				else if (FT_CURVE_TAG (tags[p]) == FT_Curve_Tag_Conic)
				{
					const int nextIndex = (p == endPoint) ? startPoint : p + 1;
					float x2 = scaleX * points [nextIndex].x;
					float y2 = scaleY * points [nextIndex].y;

					if (FT_CURVE_TAG (tags [nextIndex]) == FT_Curve_Tag_Conic)
					{
						x2 = (x + x2) * 0.5f;
						y2 = (y + y2) * 0.5f;
					}
					else
					{
						++p;
					}

					destShape.quadraticTo (x, y, x2, y2);
				}
				else if (FT_CURVE_TAG (tags[p]) == FT_Curve_Tag_Cubic)
				{
					if (p >= endPoint)
						return false;

					const int next1 = p + 1;
					const int next2 = (p == (endPoint - 1)) ? startPoint : p + 2;

					const float x2 = scaleX * points [next1].x;
					const float y2 = scaleY * points [next1].y;
					const float x3 = scaleX * points [next2].x;
					const float y3 = scaleY * points [next2].y;

					if (FT_CURVE_TAG (tags[next1]) != FT_Curve_Tag_Cubic
						 || FT_CURVE_TAG (tags[next2]) != FT_Curve_Tag_On)
						return false;

					destShape.cubicTo (x, y, x2, y2, x3, y3);
					p += 2;
				}
			}

			destShape.closeSubPath();
		}

		dest.addGlyph (character, destShape, face->glyph->metrics.horiAdvance / height);

		if ((face->face_flags & FT_FACE_FLAG_KERNING) != 0)
			addKerning (face, dest, character, glyphIndex);

		return true;
	}

	void addKerning (FT_Face face, CustomTypeface& dest, const uint32 character, const uint32 glyphIndex)
	{
		const float height = (float) (face->ascender - face->descender);

		uint32 rightGlyphIndex;
		uint32 rightCharCode = FT_Get_First_Char (face, &rightGlyphIndex);

		while (rightGlyphIndex != 0)
		{
			FT_Vector kerning;

			if (FT_Get_Kerning (face, glyphIndex, rightGlyphIndex, ft_kerning_unscaled, &kerning) == 0)
			{
				if (kerning.x != 0)
					dest.addKerningPair (character, rightCharCode, kerning.x / height);
			}

			rightCharCode = FT_Get_Next_Char (face, rightCharCode, &rightGlyphIndex);
		}
	}

	// Add a glyph to a font
	bool addGlyphToFont (const uint32 character, const String& fontName,
						 bool bold, bool italic, CustomTypeface& dest)
	{
		FT_Face face = createFT_Face (fontName, bold, italic);

		return face != 0 && addGlyph (face, dest, character);
	}

	void getFamilyNames (StringArray& familyNames) const
	{
		for (int i = 0; i < faces.size(); i++)
			familyNames.add (faces[i]->getFamilyName());
	}

	void getMonospacedNames (StringArray& monoSpaced) const
	{
		for (int i = 0; i < faces.size(); i++)
			if (faces[i]->getMonospaced())
				monoSpaced.add (faces[i]->getFamilyName());
	}

	void getSerifNames (StringArray& serif) const
	{
		for (int i = 0; i < faces.size(); i++)
			if (faces[i]->getSerif())
				serif.add (faces[i]->getFamilyName());
	}

	void getSansSerifNames (StringArray& sansSerif) const
	{
		for (int i = 0; i < faces.size(); i++)
			if (! faces[i]->getSerif())
				sansSerif.add (faces[i]->getFamilyName());
	}

	juce_DeclareSingleton_SingleThreaded_Minimal (FreeTypeInterface);

private:

	FT_Library ftLib;
	FT_Face lastFace;
	String lastFontName;
	bool lastBold, lastItalic;
	OwnedArray<FreeTypeFontFace> faces;
};

juce_ImplementSingleton_SingleThreaded (FreeTypeInterface)

class FreetypeTypeface   : public CustomTypeface
{
public:
	FreetypeTypeface (const Font& font)
	{
		FT_Face face = FreeTypeInterface::getInstance()
							->createFT_Face (font.getTypefaceName(), font.isBold(), font.isItalic());

		if (face == 0)
		{
		   #if JUCE_DEBUG
			String msg ("Failed to create typeface: ");
			msg << font.getTypefaceName() << " " << (font.isBold() ? 'B' : ' ') << (font.isItalic() ? 'I' : ' ');
			DBG (msg);
		   #endif
		}
		else
		{
			setCharacteristics (font.getTypefaceName(),
								face->ascender / (float) (face->ascender - face->descender),
								font.isBold(), font.isItalic(),
								L' ');
		}
	}

	bool loadGlyphIfPossible (juce_wchar character)
	{
		return FreeTypeInterface::getInstance()
					->addGlyphToFont (character, name, isBold, isItalic, *this);
	}
};

Typeface::Ptr Typeface::createSystemTypefaceFor (const Font& font)
{
	return new FreetypeTypeface (font);
}

StringArray Font::findAllTypefaceNames()
{
	StringArray s;
	FreeTypeInterface::getInstance()->getFamilyNames (s);
	s.sort (true);
	return s;
}

namespace LinuxFontHelpers
{
	String pickBestFont (const StringArray& names,
						 const char* const* choicesString)
	{
		const StringArray choices (choicesString);

		int i, j;
		for (j = 0; j < choices.size(); ++j)
			if (names.contains (choices[j], true))
				return choices[j];

		for (j = 0; j < choices.size(); ++j)
			for (i = 0; i < names.size(); i++)
				if (names[i].startsWithIgnoreCase (choices[j]))
					return names[i];

		for (j = 0; j < choices.size(); ++j)
			for (i = 0; i < names.size(); i++)
				if (names[i].containsIgnoreCase (choices[j]))
					return names[i];

		return names[0];
	}

	String getDefaultSansSerifFontName()
	{
		StringArray allFonts;
		FreeTypeInterface::getInstance()->getSansSerifNames (allFonts);

		const char* targets[] = { "Verdana", "Bitstream Vera Sans", "Luxi Sans", "Sans", 0 };
		return pickBestFont (allFonts, targets);
	}

	String getDefaultSerifFontName()
	{
		StringArray allFonts;
		FreeTypeInterface::getInstance()->getSerifNames (allFonts);

		const char* targets[] = { "Bitstream Vera Serif", "Times", "Nimbus Roman", "Serif", 0 };
		return pickBestFont (allFonts, targets);
	}

	String getDefaultMonospacedFontName()
	{
		StringArray allFonts;
		FreeTypeInterface::getInstance()->getMonospacedNames (allFonts);

		const char* targets[] = { "Bitstream Vera Sans Mono", "Courier", "Sans Mono", "Mono", 0 };
		return pickBestFont (allFonts, targets);
	}
}

void Font::getPlatformDefaultFontNames (String& defaultSans, String& defaultSerif, String& defaultFixed, String& /*defaultFallback*/)
{
	defaultSans = LinuxFontHelpers::getDefaultSansSerifFontName();
	defaultSerif = LinuxFontHelpers::getDefaultSerifFontName();
	defaultFixed = LinuxFontHelpers::getDefaultMonospacedFontName();
}

#endif

/*** End of inlined file: juce_linux_Fonts.cpp ***/


/*** Start of inlined file: juce_linux_Windowing.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

// These are defined in juce_linux_Messaging.cpp
extern Display* display;
extern XContext windowHandleXContext;

namespace Atoms
{
	enum ProtocolItems
	{
		TAKE_FOCUS = 0,
		DELETE_WINDOW = 1,
		PING = 2
	};

	static Atom Protocols, ProtocolList[3], ChangeState, State,
				ActiveWin, Pid, WindowType, WindowState,
				XdndAware, XdndEnter, XdndLeave, XdndPosition, XdndStatus,
				XdndDrop, XdndFinished, XdndSelection, XdndTypeList, XdndActionList,
				XdndActionDescription, XdndActionCopy,
				allowedActions[5],
				allowedMimeTypes[2];

	const unsigned long DndVersion = 3;

	static void initialiseAtoms()
	{
		static bool atomsInitialised = false;

		if (! atomsInitialised)
		{
			atomsInitialised = true;

			Protocols                       = XInternAtom (display, "WM_PROTOCOLS", True);
			ProtocolList [TAKE_FOCUS]       = XInternAtom (display, "WM_TAKE_FOCUS", True);
			ProtocolList [DELETE_WINDOW]    = XInternAtom (display, "WM_DELETE_WINDOW", True);
			ProtocolList [PING]             = XInternAtom (display, "_NET_WM_PING", True);
			ChangeState                     = XInternAtom (display, "WM_CHANGE_STATE", True);
			State                           = XInternAtom (display, "WM_STATE", True);
			ActiveWin                       = XInternAtom (display, "_NET_ACTIVE_WINDOW", False);
			Pid                             = XInternAtom (display, "_NET_WM_PID", False);
			WindowType                      = XInternAtom (display, "_NET_WM_WINDOW_TYPE", True);
			WindowState                     = XInternAtom (display, "_NET_WM_STATE", True);

			XdndAware                       = XInternAtom (display, "XdndAware", False);
			XdndEnter                       = XInternAtom (display, "XdndEnter", False);
			XdndLeave                       = XInternAtom (display, "XdndLeave", False);
			XdndPosition                    = XInternAtom (display, "XdndPosition", False);
			XdndStatus                      = XInternAtom (display, "XdndStatus", False);
			XdndDrop                        = XInternAtom (display, "XdndDrop", False);
			XdndFinished                    = XInternAtom (display, "XdndFinished", False);
			XdndSelection                   = XInternAtom (display, "XdndSelection", False);

			XdndTypeList                    = XInternAtom (display, "XdndTypeList", False);
			XdndActionList                  = XInternAtom (display, "XdndActionList", False);
			XdndActionCopy                  = XInternAtom (display, "XdndActionCopy", False);
			XdndActionDescription           = XInternAtom (display, "XdndActionDescription", False);

			allowedMimeTypes[0]             = XInternAtom (display, "text/plain", False);
			allowedMimeTypes[1]             = XInternAtom (display, "text/uri-list", False);

			allowedActions[0]               = XInternAtom (display, "XdndActionMove", False);
			allowedActions[1]		   = XdndActionCopy;
			allowedActions[2]               = XInternAtom (display, "XdndActionLink", False);
			allowedActions[3]               = XInternAtom (display, "XdndActionAsk", False);
			allowedActions[4]               = XInternAtom (display, "XdndActionPrivate", False);
		}
	}
}

namespace Keys
{
	enum MouseButtons
	{
		NoButton = 0,
		LeftButton = 1,
		MiddleButton = 2,
		RightButton = 3,
		WheelUp = 4,
		WheelDown = 5
	};

	static int AltMask = 0;
	static int NumLockMask = 0;
	static bool numLock = false;
	static bool capsLock = false;
	static char keyStates [32];
	static const int extendedKeyModifier = 0x10000000;
}

bool KeyPress::isKeyCurrentlyDown (const int keyCode)
{
	int keysym;

	if (keyCode & Keys::extendedKeyModifier)
	{
		keysym = 0xff00 | (keyCode & 0xff);
	}
	else
	{
		keysym = keyCode;

		if (keysym == (XK_Tab & 0xff)
			|| keysym == (XK_Return & 0xff)
			|| keysym == (XK_Escape & 0xff)
			|| keysym == (XK_BackSpace & 0xff))
		{
			keysym |= 0xff00;
		}
	}

	ScopedXLock xlock;

	const int keycode = XKeysymToKeycode (display, keysym);

	const int keybyte = keycode >> 3;
	const int keybit = (1 << (keycode & 7));
	return (Keys::keyStates [keybyte] & keybit) != 0;
}

#if JUCE_USE_XSHM
namespace XSHMHelpers
{
	static int trappedErrorCode = 0;
	extern "C" int errorTrapHandler (Display*, XErrorEvent* err)
	{
		trappedErrorCode = err->error_code;
		return 0;
	}

	static bool isShmAvailable() noexcept
	{
		static bool isChecked = false;
		static bool isAvailable = false;

		if (! isChecked)
		{
			isChecked = true;
			int major, minor;
			Bool pixmaps;

			ScopedXLock xlock;

			if (XShmQueryVersion (display, &major, &minor, &pixmaps))
			{
				trappedErrorCode = 0;
				XErrorHandler oldHandler = XSetErrorHandler (errorTrapHandler);

				XShmSegmentInfo segmentInfo = { 0 };
				XImage* xImage = XShmCreateImage (display, DefaultVisual (display, DefaultScreen (display)),
												  24, ZPixmap, 0, &segmentInfo, 50, 50);

				if ((segmentInfo.shmid = shmget (IPC_PRIVATE,
												 xImage->bytes_per_line * xImage->height,
												 IPC_CREAT | 0777)) >= 0)
				{
					segmentInfo.shmaddr = (char*) shmat (segmentInfo.shmid, 0, 0);

					if (segmentInfo.shmaddr != (void*) -1)
					{
						segmentInfo.readOnly = False;
						xImage->data = segmentInfo.shmaddr;
						XSync (display, False);

						if (XShmAttach (display, &segmentInfo) != 0)
						{
							XSync (display, False);
							XShmDetach (display, &segmentInfo);

							isAvailable = true;
						}
					}

					XFlush (display);
					XDestroyImage (xImage);

					shmdt (segmentInfo.shmaddr);
				}

				shmctl (segmentInfo.shmid, IPC_RMID, 0);

				XSetErrorHandler (oldHandler);
				if (trappedErrorCode != 0)
					isAvailable = false;
			}
		}

		return isAvailable;
	}
}
#endif

#if JUCE_USE_XRENDER
namespace XRender
{
	typedef Status (*tXRenderQueryVersion) (Display*, int*, int*);
	typedef XRenderPictFormat* (*tXrenderFindStandardFormat) (Display*, int);
	typedef XRenderPictFormat* (*tXRenderFindFormat) (Display*, unsigned long, XRenderPictFormat*, int);
	typedef XRenderPictFormat* (*tXRenderFindVisualFormat) (Display*, Visual*);

	static tXRenderQueryVersion xRenderQueryVersion = 0;
	static tXrenderFindStandardFormat xRenderFindStandardFormat = 0;
	static tXRenderFindFormat xRenderFindFormat = 0;
	static tXRenderFindVisualFormat xRenderFindVisualFormat = 0;

	static bool isAvailable()
	{
		static bool hasLoaded = false;

		if (! hasLoaded)
		{
			ScopedXLock xlock;
			hasLoaded = true;

			void* h = dlopen ("libXrender.so", RTLD_GLOBAL | RTLD_NOW);

			if (h != 0)
			{
				xRenderQueryVersion         = (tXRenderQueryVersion)        dlsym (h, "XRenderQueryVersion");
				xRenderFindStandardFormat   = (tXrenderFindStandardFormat)  dlsym (h, "XrenderFindStandardFormat");
				xRenderFindFormat           = (tXRenderFindFormat)          dlsym (h, "XRenderFindFormat");
				xRenderFindVisualFormat     = (tXRenderFindVisualFormat)    dlsym (h, "XRenderFindVisualFormat");
			}

			if (xRenderQueryVersion != 0
				 && xRenderFindStandardFormat != 0
				 && xRenderFindFormat != 0
				 && xRenderFindVisualFormat != 0)
			{
				int major, minor;
				if (xRenderQueryVersion (display, &major, &minor))
					return true;
			}

			xRenderQueryVersion = 0;
		}

		return xRenderQueryVersion != 0;
	}

	static XRenderPictFormat* findPictureFormat()
	{
		ScopedXLock xlock;

		XRenderPictFormat* pictFormat = nullptr;

		if (isAvailable())
		{
			pictFormat = xRenderFindStandardFormat (display, PictStandardARGB32);

			if (pictFormat == 0)
			{
				XRenderPictFormat desiredFormat;
				desiredFormat.type = PictTypeDirect;
				desiredFormat.depth = 32;

				desiredFormat.direct.alphaMask = 0xff;
				desiredFormat.direct.redMask = 0xff;
				desiredFormat.direct.greenMask = 0xff;
				desiredFormat.direct.blueMask = 0xff;

				desiredFormat.direct.alpha = 24;
				desiredFormat.direct.red = 16;
				desiredFormat.direct.green = 8;
				desiredFormat.direct.blue = 0;

				pictFormat = xRenderFindFormat (display,
												PictFormatType | PictFormatDepth
												 | PictFormatRedMask | PictFormatRed
												 | PictFormatGreenMask | PictFormatGreen
												 | PictFormatBlueMask | PictFormatBlue
												 | PictFormatAlphaMask | PictFormatAlpha,
												&desiredFormat,
												0);
			}
		}

		return pictFormat;
	}
}
#endif

namespace Visuals
{
	static Visual* findVisualWithDepth (const int desiredDepth) noexcept
	{
		ScopedXLock xlock;

		Visual* visual = nullptr;
		int numVisuals = 0;
		long desiredMask = VisualNoMask;
		XVisualInfo desiredVisual;

		desiredVisual.screen = DefaultScreen (display);
		desiredVisual.depth = desiredDepth;

		desiredMask = VisualScreenMask | VisualDepthMask;

		if (desiredDepth == 32)
		{
			desiredVisual.c_class = TrueColor;
			desiredVisual.red_mask = 0x00FF0000;
			desiredVisual.green_mask = 0x0000FF00;
			desiredVisual.blue_mask = 0x000000FF;
			desiredVisual.bits_per_rgb = 8;

			desiredMask |= VisualClassMask;
			desiredMask |= VisualRedMaskMask;
			desiredMask |= VisualGreenMaskMask;
			desiredMask |= VisualBlueMaskMask;
			desiredMask |= VisualBitsPerRGBMask;
		}

		XVisualInfo* xvinfos = XGetVisualInfo (display,
											   desiredMask,
											   &desiredVisual,
											   &numVisuals);

		if (xvinfos != 0)
		{
			for (int i = 0; i < numVisuals; i++)
			{
				if (xvinfos[i].depth == desiredDepth)
				{
					visual = xvinfos[i].visual;
					break;
				}
			}

			XFree (xvinfos);
		}

		return visual;
	}

	static Visual* findVisualFormat (const int desiredDepth, int& matchedDepth) noexcept
	{
		Visual* visual = nullptr;

		if (desiredDepth == 32)
		{
		   #if JUCE_USE_XSHM
			if (XSHMHelpers::isShmAvailable())
			{
			   #if JUCE_USE_XRENDER
				if (XRender::isAvailable())
				{
					XRenderPictFormat* pictFormat = XRender::findPictureFormat();

					if (pictFormat != 0)
					{
						int numVisuals = 0;
						XVisualInfo desiredVisual;
						desiredVisual.screen = DefaultScreen (display);
						desiredVisual.depth = 32;
						desiredVisual.bits_per_rgb = 8;

						XVisualInfo* xvinfos = XGetVisualInfo (display,
															   VisualScreenMask | VisualDepthMask | VisualBitsPerRGBMask,
															   &desiredVisual, &numVisuals);
						if (xvinfos != 0)
						{
							for (int i = 0; i < numVisuals; ++i)
							{
								XRenderPictFormat* pictVisualFormat = XRender::xRenderFindVisualFormat (display, xvinfos[i].visual);

								if (pictVisualFormat != 0
									 && pictVisualFormat->type == PictTypeDirect
									 && pictVisualFormat->direct.alphaMask)
								{
									visual = xvinfos[i].visual;
									matchedDepth = 32;
									break;
								}
							}

							XFree (xvinfos);
						}
					}
				}
			   #endif
				if (visual == 0)
				{
					visual = findVisualWithDepth (32);
					if (visual != 0)
						matchedDepth = 32;
				}
			}
		   #endif
		}

		if (visual == 0 && desiredDepth >= 24)
		{
			visual = findVisualWithDepth (24);
			if (visual != 0)
				matchedDepth = 24;
		}

		if (visual == 0 && desiredDepth >= 16)
		{
			visual = findVisualWithDepth (16);
			if (visual != 0)
				matchedDepth = 16;
		}

		return visual;
	}
}

class XBitmapImage  : public Image::SharedImage
{
public:

	XBitmapImage (const Image::PixelFormat format_, const int w, const int h,
				  const bool clearImage, const int imageDepth_, Visual* visual)
		: Image::SharedImage (format_, w, h),
		  imageDepth (imageDepth_),
		  gc (None)
	{
		jassert (format_ == Image::RGB || format_ == Image::ARGB);

		pixelStride = (format_ == Image::RGB) ? 3 : 4;
		lineStride = ((w * pixelStride + 3) & ~3);

		ScopedXLock xlock;

	   #if JUCE_USE_XSHM
		usingXShm = false;

		if ((imageDepth > 16) && XSHMHelpers::isShmAvailable())
		{
			zerostruct (segmentInfo);

			segmentInfo.shmid = -1;
			segmentInfo.shmaddr = (char *) -1;
			segmentInfo.readOnly = False;

			xImage = XShmCreateImage (display, visual, imageDepth, ZPixmap, 0, &segmentInfo, w, h);

			if (xImage != 0)
			{
				if ((segmentInfo.shmid = shmget (IPC_PRIVATE,
												 xImage->bytes_per_line * xImage->height,
												 IPC_CREAT | 0777)) >= 0)
				{
					if (segmentInfo.shmid != -1)
					{
						segmentInfo.shmaddr = (char*) shmat (segmentInfo.shmid, 0, 0);

						if (segmentInfo.shmaddr != (void*) -1)
						{
							segmentInfo.readOnly = False;

							xImage->data = segmentInfo.shmaddr;
							imageData = (uint8*) segmentInfo.shmaddr;

							if (XShmAttach (display, &segmentInfo) != 0)
								usingXShm = true;
							else
								jassertfalse;
						}
						else
						{
							shmctl (segmentInfo.shmid, IPC_RMID, 0);
						}
					}
				}
			}
		}

		if (! usingXShm)
	   #endif
		{
			imageDataAllocated.allocate (lineStride * h, format_ == Image::ARGB && clearImage);
			imageData = imageDataAllocated;

			xImage = (XImage*) ::calloc (1, sizeof (XImage));

			xImage->width = w;
			xImage->height = h;
			xImage->xoffset = 0;
			xImage->format = ZPixmap;
			xImage->data = (char*) imageData;
			xImage->byte_order = ImageByteOrder (display);
			xImage->bitmap_unit = BitmapUnit (display);
			xImage->bitmap_bit_order = BitmapBitOrder (display);
			xImage->bitmap_pad = 32;
			xImage->depth = pixelStride * 8;
			xImage->bytes_per_line = lineStride;
			xImage->bits_per_pixel = pixelStride * 8;
			xImage->red_mask   = 0x00FF0000;
			xImage->green_mask = 0x0000FF00;
			xImage->blue_mask  = 0x000000FF;

			if (imageDepth == 16)
			{
				const int pixelStride = 2;
				const int lineStride = ((w * pixelStride + 3) & ~3);

				imageData16Bit.malloc (lineStride * h);
				xImage->data = imageData16Bit;
				xImage->bitmap_pad = 16;
				xImage->depth = pixelStride * 8;
				xImage->bytes_per_line = lineStride;
				xImage->bits_per_pixel = pixelStride * 8;
				xImage->red_mask   = visual->red_mask;
				xImage->green_mask = visual->green_mask;
				xImage->blue_mask  = visual->blue_mask;
			}

			if (! XInitImage (xImage))
				jassertfalse;
		}
	}

	~XBitmapImage()
	{
		ScopedXLock xlock;

		if (gc != None)
			XFreeGC (display, gc);

	   #if JUCE_USE_XSHM
		if (usingXShm)
		{
			XShmDetach (display, &segmentInfo);

			XFlush (display);
			XDestroyImage (xImage);

			shmdt (segmentInfo.shmaddr);
			shmctl (segmentInfo.shmid, IPC_RMID, 0);
		}
		else
	   #endif
		{
			xImage->data = nullptr;
			XDestroyImage (xImage);
		}
	}

	Image::ImageType getType() const	{ return Image::NativeImage; }

	LowLevelGraphicsContext* createLowLevelContext()
	{
		return new LowLevelGraphicsSoftwareRenderer (Image (this));
	}

	void initialiseBitmapData (Image::BitmapData& bitmap, int x, int y, Image::BitmapData::ReadWriteMode /*mode*/)
	{
		bitmap.data = imageData + x * pixelStride + y * lineStride;
		bitmap.pixelFormat = format;
		bitmap.lineStride = lineStride;
		bitmap.pixelStride = pixelStride;
	}

	SharedImage* clone()
	{
		jassertfalse;
		return nullptr;
	}

	void blitToWindow (Window window, int dx, int dy, int dw, int dh, int sx, int sy)
	{
		ScopedXLock xlock;

		if (gc == None)
		{
			XGCValues gcvalues;
			gcvalues.foreground = None;
			gcvalues.background = None;
			gcvalues.function = GXcopy;
			gcvalues.plane_mask = AllPlanes;
			gcvalues.clip_mask = None;
			gcvalues.graphics_exposures = False;

			gc = XCreateGC (display, window,
							GCBackground | GCForeground | GCFunction | GCPlaneMask | GCClipMask | GCGraphicsExposures,
							&gcvalues);
		}

		if (imageDepth == 16)
		{
			const uint32 rMask = xImage->red_mask;
			const uint32 rShiftL = jmax (0, getShiftNeeded (rMask));
			const uint32 rShiftR = jmax (0, -getShiftNeeded (rMask));
			const uint32 gMask = xImage->green_mask;
			const uint32 gShiftL = jmax (0, getShiftNeeded (gMask));
			const uint32 gShiftR = jmax (0, -getShiftNeeded (gMask));
			const uint32 bMask = xImage->blue_mask;
			const uint32 bShiftL = jmax (0, getShiftNeeded (bMask));
			const uint32 bShiftR = jmax (0, -getShiftNeeded (bMask));

			const Image::BitmapData srcData (Image (this), Image::BitmapData::readOnly);

			for (int y = sy; y < sy + dh; ++y)
			{
				const uint8* p = srcData.getPixelPointer (sx, y);

				for (int x = sx; x < sx + dw; ++x)
				{
					const PixelRGB* const pixel = (const PixelRGB*) p;
					p += srcData.pixelStride;

					XPutPixel (xImage, x, y,
							   (((((uint32) pixel->getRed()) << rShiftL) >> rShiftR) & rMask)
								 | (((((uint32) pixel->getGreen()) << gShiftL) >> gShiftR) & gMask)
								 | (((((uint32) pixel->getBlue()) << bShiftL) >> bShiftR) & bMask));
				}
			}
		}

		// blit results to screen.
	   #if JUCE_USE_XSHM
		if (usingXShm)
			XShmPutImage (display, (::Drawable) window, gc, xImage, sx, sy, dx, dy, dw, dh, True);
		else
	   #endif
			XPutImage (display, (::Drawable) window, gc, xImage, sx, sy, dx, dy, dw, dh);
	}

private:
	XImage* xImage;
	const int imageDepth;
	HeapBlock <uint8> imageDataAllocated;
	HeapBlock <char> imageData16Bit;
	int pixelStride, lineStride;
	uint8* imageData;
	GC gc;

   #if JUCE_USE_XSHM
	XShmSegmentInfo segmentInfo;
	bool usingXShm;
   #endif

	static int getShiftNeeded (const uint32 mask) noexcept
	{
		for (int i = 32; --i >= 0;)
			if (((mask >> i) & 1) != 0)
				return i - 7;

		jassertfalse;
		return 0;
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (XBitmapImage);
};

namespace PixmapHelpers
{
	Pixmap createColourPixmapFromImage (Display* display, const Image& image)
	{
		ScopedXLock xlock;

		const int width = image.getWidth();
		const int height = image.getHeight();
		HeapBlock <uint32> colour (width * height);
		int index = 0;

		for (int y = 0; y < height; ++y)
			for (int x = 0; x < width; ++x)
				colour[index++] = image.getPixelAt (x, y).getARGB();

		XImage* ximage = XCreateImage (display, CopyFromParent, 24, ZPixmap,
									   0, reinterpret_cast<char*> (colour.getData()),
									   width, height, 32, 0);

		Pixmap pixmap = XCreatePixmap (display, DefaultRootWindow (display),
									   width, height, 24);

		GC gc = XCreateGC (display, pixmap, 0, 0);
		XPutImage (display, pixmap, gc, ximage, 0, 0, 0, 0, width, height);
		XFreeGC (display, gc);

		return pixmap;
	}

	Pixmap createMaskPixmapFromImage (Display* display, const Image& image)
	{
		ScopedXLock xlock;

		const int width = image.getWidth();
		const int height = image.getHeight();
		const int stride = (width + 7) >> 3;
		HeapBlock <char> mask;
		mask.calloc (stride * height);
		const bool msbfirst = (BitmapBitOrder (display) == MSBFirst);

		for (int y = 0; y < height; ++y)
		{
			for (int x = 0; x < width; ++x)
			{
				const char bit = (char) (1 << (msbfirst ? (7 - (x & 7)) : (x & 7)));
				const int offset = y * stride + (x >> 3);

				if (image.getPixelAt (x, y).getAlpha() >= 128)
					mask[offset] |= bit;
			}
		}

		return XCreatePixmapFromBitmapData (display, DefaultRootWindow (display),
											mask.getData(), width, height, 1, 0, 1);
	}
}

class LinuxComponentPeer  : public ComponentPeer
{
public:

	LinuxComponentPeer (Component* const component, const int windowStyleFlags, Window parentToAddTo)
		: ComponentPeer (component, windowStyleFlags),
		  windowH (0), parentWindow (0),
		  wx (0), wy (0), ww (0), wh (0),
		  fullScreen (false), mapped (false),
		  visual (0), depth (0)
	{
		// it's dangerous to create a window on a thread other than the message thread..
		jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

		repainter = new LinuxRepaintManager (this);

		createWindow (parentToAddTo);

		setTitle (component->getName());
	}

	~LinuxComponentPeer()
	{
		// it's dangerous to delete a window on a thread other than the message thread..
		jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

		deleteIconPixmaps();

		destroyWindow();

		windowH = 0;
	}

	void* getNativeHandle() const
	{
		return (void*) windowH;
	}

	static LinuxComponentPeer* getPeerFor (Window windowHandle) noexcept
	{
		XPointer peer = 0;

		ScopedXLock xlock;
		if (! XFindContext (display, (XID) windowHandle, windowHandleXContext, &peer))
		{
			if (peer != nullptr && ! ComponentPeer::isValidPeer ((LinuxComponentPeer*) peer))
				peer = 0;
		}

		return (LinuxComponentPeer*) peer;
	}

	void setVisible (bool shouldBeVisible)
	{
		ScopedXLock xlock;
		if (shouldBeVisible)
			XMapWindow (display, windowH);
		else
			XUnmapWindow (display, windowH);
	}

	void setTitle (const String& title)
	{
		XTextProperty nameProperty;
		char* strings[] = { const_cast <char*> (title.toUTF8().getAddress()) };
		ScopedXLock xlock;

		if (XStringListToTextProperty (strings, 1, &nameProperty))
		{
			XSetWMName (display, windowH, &nameProperty);
			XSetWMIconName (display, windowH, &nameProperty);

			XFree (nameProperty.value);
		}
	}

	void setBounds (int x, int y, int w, int h, bool isNowFullScreen)
	{
		fullScreen = isNowFullScreen;

		if (windowH != 0)
		{
			WeakReference<Component> deletionChecker (component);

			wx = x;
			wy = y;
			ww = jmax (1, w);
			wh = jmax (1, h);

			ScopedXLock xlock;

			// Make sure the Window manager does what we want
			XSizeHints* hints = XAllocSizeHints();
			hints->flags = USSize | USPosition;
			hints->width = ww;
			hints->height = wh;
			hints->x = wx;
			hints->y = wy;

			if ((getStyleFlags() & (windowHasTitleBar | windowIsResizable)) == windowHasTitleBar)
			{
				hints->min_width  = hints->max_width  = hints->width;
				hints->min_height = hints->max_height = hints->height;
				hints->flags |= PMinSize | PMaxSize;
			}

			XSetWMNormalHints (display, windowH, hints);
			XFree (hints);

			XMoveResizeWindow (display, windowH,
							   wx - windowBorder.getLeft(),
							   wy - windowBorder.getTop(), ww, wh);

			if (deletionChecker != 0)
			{
				updateBorderSize();
				handleMovedOrResized();
			}
		}
	}

	void setPosition (int x, int y)		 { setBounds (x, y, ww, wh, false); }
	void setSize (int w, int h)			 { setBounds (wx, wy, w, h, false); }
	const Rectangle<int> getBounds() const	  { return Rectangle<int> (wx, wy, ww, wh); }
	const Point<int> getScreenPosition() const	  { return Point<int> (wx, wy); }

	const Point<int> localToGlobal (const Point<int>& relativePosition)
	{
		return relativePosition + getScreenPosition();
	}

	const Point<int> globalToLocal (const Point<int>& screenPosition)
	{
		return screenPosition - getScreenPosition();
	}

	void setAlpha (float newAlpha)
	{
		//xxx todo!
	}

	void setMinimised (bool shouldBeMinimised)
	{
		if (shouldBeMinimised)
		{
			Window root = RootWindow (display, DefaultScreen (display));

			XClientMessageEvent clientMsg;
			clientMsg.display = display;
			clientMsg.window = windowH;
			clientMsg.type = ClientMessage;
			clientMsg.format = 32;
			clientMsg.message_type = Atoms::ChangeState;
			clientMsg.data.l[0] = IconicState;

			ScopedXLock xlock;
			XSendEvent (display, root, false, SubstructureRedirectMask | SubstructureNotifyMask, (XEvent*) &clientMsg);
		}
		else
		{
			setVisible (true);
		}
	}

	bool isMinimised() const
	{
		bool minimised = false;

		unsigned char* stateProp;
		unsigned long nitems, bytesLeft;
		Atom actualType;
		int actualFormat;

		ScopedXLock xlock;
		if (XGetWindowProperty (display, windowH, Atoms::State, 0, 64, False,
								Atoms::State, &actualType, &actualFormat, &nitems, &bytesLeft,
								&stateProp) == Success
			&& actualType == Atoms::State
			&& actualFormat == 32
			&& nitems > 0)
		{
			if (((unsigned long*) stateProp)[0] == IconicState)
				minimised = true;

			XFree (stateProp);
		}

		return minimised;
	}

	void setFullScreen (const bool shouldBeFullScreen)
	{
		Rectangle<int> r (lastNonFullscreenBounds); // (get a copy of this before de-minimising)

		setMinimised (false);

		if (fullScreen != shouldBeFullScreen)
		{
			if (shouldBeFullScreen)
				r = Desktop::getInstance().getMainMonitorArea();

			if (! r.isEmpty())
				setBounds (r.getX(), r.getY(), r.getWidth(), r.getHeight(), shouldBeFullScreen);

			getComponent()->repaint();
		}
	}

	bool isFullScreen() const
	{
		return fullScreen;
	}

	bool isChildWindowOf (Window possibleParent) const
	{
		Window* windowList = nullptr;
		uint32 windowListSize = 0;
		Window parent, root;

		ScopedXLock xlock;
		if (XQueryTree (display, windowH, &root, &parent, &windowList, &windowListSize) != 0)
		{
			if (windowList != 0)
				XFree (windowList);

			return parent == possibleParent;
		}

		return false;
	}

	bool isFrontWindow() const
	{
		Window* windowList = nullptr;
		uint32 windowListSize = 0;
		bool result = false;

		ScopedXLock xlock;
		Window parent, root = RootWindow (display, DefaultScreen (display));

		if (XQueryTree (display, root, &root, &parent, &windowList, &windowListSize) != 0)
		{
			for (int i = windowListSize; --i >= 0;)
			{
				LinuxComponentPeer* const peer = LinuxComponentPeer::getPeerFor (windowList[i]);

				if (peer != 0)
				{
					result = (peer == this);
					break;
				}
			}
		}

		if (windowList != 0)
			XFree (windowList);

		return result;
	}

	bool contains (const Point<int>& position, bool trueIfInAChildWindow) const
	{
		if (! (isPositiveAndBelow (position.getX(), ww) && isPositiveAndBelow (position.getY(), wh)))
			return false;

		for (int i = Desktop::getInstance().getNumComponents(); --i >= 0;)
		{
			Component* const c = Desktop::getInstance().getComponent (i);

			if (c == getComponent())
				break;

			if (c->contains (position + Point<int> (wx, wy) - c->getScreenPosition()))
				return false;
		}

		if (trueIfInAChildWindow)
			return true;

		::Window root, child;
		unsigned int bw, depth;
		int wx, wy, w, h;

		ScopedXLock xlock;
		if (! XGetGeometry (display, (::Drawable) windowH, &root,
							&wx, &wy, (unsigned int*) &w, (unsigned int*) &h,
							&bw, &depth))
		{
			return false;
		}

		if (! XTranslateCoordinates (display, windowH, windowH, position.getX(), position.getY(), &wx, &wy, &child))
			return false;

		return child == None;
	}

	const BorderSize<int> getFrameSize() const
	{
		return BorderSize<int>();
	}

	bool setAlwaysOnTop (bool alwaysOnTop)
	{
		return false;
	}

	void toFront (bool makeActive)
	{
		if (makeActive)
		{
			setVisible (true);
			grabFocus();
		}

		XEvent ev;
		ev.xclient.type = ClientMessage;
		ev.xclient.serial = 0;
		ev.xclient.send_event = True;
		ev.xclient.message_type = Atoms::ActiveWin;
		ev.xclient.window = windowH;
		ev.xclient.format = 32;
		ev.xclient.data.l[0] = 2;
		ev.xclient.data.l[1] = CurrentTime;
		ev.xclient.data.l[2] = 0;
		ev.xclient.data.l[3] = 0;
		ev.xclient.data.l[4] = 0;

		{
			ScopedXLock xlock;
			XSendEvent (display, RootWindow (display, DefaultScreen (display)),
						False, SubstructureRedirectMask | SubstructureNotifyMask, &ev);

			XWindowAttributes attr;
			XGetWindowAttributes (display, windowH, &attr);

			if (component->isAlwaysOnTop())
				XRaiseWindow (display, windowH);

			XSync (display, False);
		}

		handleBroughtToFront();
	}

	void toBehind (ComponentPeer* other)
	{
		LinuxComponentPeer* const otherPeer = dynamic_cast <LinuxComponentPeer*> (other);
		jassert (otherPeer != nullptr); // wrong type of window?

		if (otherPeer != nullptr)
		{
			setMinimised (false);

			Window newStack[] = { otherPeer->windowH, windowH };

			ScopedXLock xlock;
			XRestackWindows (display, newStack, 2);
		}
	}

	bool isFocused() const
	{
		int revert = 0;
		Window focusedWindow = 0;
		ScopedXLock xlock;
		XGetInputFocus (display, &focusedWindow, &revert);

		return focusedWindow == windowH;
	}

	void grabFocus()
	{
		XWindowAttributes atts;
		ScopedXLock xlock;

		if (windowH != 0
			&& XGetWindowAttributes (display, windowH, &atts)
			&& atts.map_state == IsViewable
			&& ! isFocused())
		{
			XSetInputFocus (display, windowH, RevertToParent, CurrentTime);
			isActiveApplication = true;
		}
	}

	void textInputRequired (const Point<int>&)
	{
	}

	void repaint (const Rectangle<int>& area)
	{
		repainter->repaint (area.getIntersection (getComponent()->getLocalBounds()));
	}

	void performAnyPendingRepaintsNow()
	{
		repainter->performAnyPendingRepaintsNow();
	}

	void setIcon (const Image& newIcon)
	{
		const int dataSize = newIcon.getWidth() * newIcon.getHeight() + 2;
		HeapBlock <unsigned long> data (dataSize);

		int index = 0;
		data[index++] = newIcon.getWidth();
		data[index++] = newIcon.getHeight();

		for (int y = 0; y < newIcon.getHeight(); ++y)
			for (int x = 0; x < newIcon.getWidth(); ++x)
				data[index++] = newIcon.getPixelAt (x, y).getARGB();

		ScopedXLock xlock;
		XChangeProperty (display, windowH,
						 XInternAtom (display, "_NET_WM_ICON", False),
						 XA_CARDINAL, 32, PropModeReplace,
						 reinterpret_cast<unsigned char*> (data.getData()), dataSize);

		deleteIconPixmaps();

		XWMHints* wmHints = XGetWMHints (display, windowH);

		if (wmHints == 0)
			wmHints = XAllocWMHints();

		wmHints->flags |= IconPixmapHint | IconMaskHint;
		wmHints->icon_pixmap = PixmapHelpers::createColourPixmapFromImage (display, newIcon);
		wmHints->icon_mask = PixmapHelpers::createMaskPixmapFromImage (display, newIcon);

		XSetWMHints (display, windowH, wmHints);
		XFree (wmHints);

		XSync (display, False);
	}

	void deleteIconPixmaps()
	{
		ScopedXLock xlock;
		XWMHints* wmHints = XGetWMHints (display, windowH);

		if (wmHints != 0)
		{
			if ((wmHints->flags & IconPixmapHint) != 0)
			{
				wmHints->flags &= ~IconPixmapHint;
				XFreePixmap (display, wmHints->icon_pixmap);
			}

			if ((wmHints->flags & IconMaskHint) != 0)
			{
				wmHints->flags &= ~IconMaskHint;
				XFreePixmap (display, wmHints->icon_mask);
			}

			XSetWMHints (display, windowH, wmHints);
			XFree (wmHints);
		}
	}

	void handleWindowMessage (XEvent* event)
	{
		switch (event->xany.type)
		{
			case 2: /* KeyPress */	  handleKeyPressEvent ((XKeyEvent*) &event->xkey); break;
			case KeyRelease:		handleKeyReleaseEvent ((const XKeyEvent*) &event->xkey); break;
			case ButtonPress:	   handleButtonPressEvent ((const XButtonPressedEvent*) &event->xbutton); break;
			case ButtonRelease:	 handleButtonReleaseEvent ((const XButtonReleasedEvent*) &event->xbutton); break;
			case MotionNotify:	  handleMotionNotifyEvent ((const XPointerMovedEvent*) &event->xmotion); break;
			case EnterNotify:	   handleEnterNotifyEvent ((const XEnterWindowEvent*) &event->xcrossing); break;
			case LeaveNotify:	   handleLeaveNotifyEvent ((const XLeaveWindowEvent*) &event->xcrossing); break;
			case FocusIn:		   handleFocusInEvent(); break;
			case FocusOut:		  handleFocusOutEvent(); break;
			case Expose:		handleExposeEvent ((XExposeEvent*) &event->xexpose); break;
			case MappingNotify:	 handleMappingNotify ((XMappingEvent*) &event->xmapping); break;
			case ClientMessage:	 handleClientMessageEvent ((XClientMessageEvent*) &event->xclient, event); break;
			case SelectionNotify:	   handleDragAndDropSelection (event); break;
			case ConfigureNotify:	   handleConfigureNotifyEvent ((XConfigureEvent*) &event->xconfigure); break;
			case ReparentNotify:	handleReparentNotifyEvent(); break;
			case GravityNotify:	 handleGravityNotify(); break;

			case CirculateNotify:
			case CreateNotify:
			case DestroyNotify:
				// Think we can ignore these
				break;

			case MapNotify:
				mapped = true;
				handleBroughtToFront();
				break;

			case UnmapNotify:
				mapped = false;
				break;

			case SelectionClear:
			case SelectionRequest:
				break;

			default:
			   #if JUCE_USE_XSHM
				{
					ScopedXLock xlock;
					if (event->xany.type == XShmGetEventBase (display))
						repainter->notifyPaintCompleted();
				}
			   #endif
				break;
		}
	}

	void handleKeyPressEvent (XKeyEvent* const keyEvent)
	{
		char utf8 [64] = { 0 };
		juce_wchar unicodeChar = 0;
		int keyCode = 0;
		bool keyDownChange = false;
		KeySym sym;

		{
			ScopedXLock xlock;
			updateKeyStates (keyEvent->keycode, true);

			const char* oldLocale = ::setlocale (LC_ALL, 0);
			::setlocale (LC_ALL, "");
			XLookupString (keyEvent, utf8, sizeof (utf8), &sym, 0);
			::setlocale (LC_ALL, oldLocale);

			unicodeChar = String::fromUTF8 (utf8, sizeof (utf8) - 1) [0];
			keyCode = (int) unicodeChar;

			if (keyCode < 0x20)
				keyCode = XkbKeycodeToKeysym (display, keyEvent->keycode, currentModifiers.isShiftDown() ? 1 : 0, 0);

			keyDownChange = (sym != NoSymbol) && ! updateKeyModifiersFromSym (sym, true);
		}

		const ModifierKeys oldMods (currentModifiers);
		bool keyPressed = false;

		if ((sym & 0xff00) == 0xff00)
		{
			switch (sym)  // Translate keypad
			{
				case XK_KP_Divide:	  keyCode = XK_slash; break;
				case XK_KP_Multiply:	keyCode = XK_asterisk; break;
				case XK_KP_Subtract:	keyCode = XK_hyphen; break;
				case XK_KP_Add:	 keyCode = XK_plus; break;
				case XK_KP_Enter:	   keyCode = XK_Return; break;
				case XK_KP_Decimal:	 keyCode = Keys::numLock ? XK_period : XK_Delete; break;
				case XK_KP_0:	   keyCode = Keys::numLock ? XK_0 : XK_Insert; break;
				case XK_KP_1:	   keyCode = Keys::numLock ? XK_1 : XK_End; break;
				case XK_KP_2:	   keyCode = Keys::numLock ? XK_2 : XK_Down; break;
				case XK_KP_3:	   keyCode = Keys::numLock ? XK_3 : XK_Page_Down; break;
				case XK_KP_4:	   keyCode = Keys::numLock ? XK_4 : XK_Left; break;
				case XK_KP_5:	   keyCode = XK_5; break;
				case XK_KP_6:	   keyCode = Keys::numLock ? XK_6 : XK_Right; break;
				case XK_KP_7:	   keyCode = Keys::numLock ? XK_7 : XK_Home; break;
				case XK_KP_8:	   keyCode = Keys::numLock ? XK_8 : XK_Up; break;
				case XK_KP_9:	   keyCode = Keys::numLock ? XK_9 : XK_Page_Up; break;
				default:		break;
			}

			switch (sym)
			{
				case XK_Left:
				case XK_Right:
				case XK_Up:
				case XK_Down:
				case XK_Page_Up:
				case XK_Page_Down:
				case XK_End:
				case XK_Home:
				case XK_Delete:
				case XK_Insert:
					keyPressed = true;
					keyCode = (sym & 0xff) | Keys::extendedKeyModifier;
					break;

				case XK_Tab:
				case XK_Return:
				case XK_Escape:
				case XK_BackSpace:
					keyPressed = true;
					keyCode &= 0xff;
					break;

				default:
					if (sym >= XK_F1 && sym <= XK_F16)
					{
						keyPressed = true;
						keyCode = (sym & 0xff) | Keys::extendedKeyModifier;
					}
					break;
			}
		}

		if (utf8[0] != 0 || ((sym & 0xff00) == 0 && sym >= 8))
			keyPressed = true;

		if (oldMods != currentModifiers)
			handleModifierKeysChange();

		if (keyDownChange)
			handleKeyUpOrDown (true);

		if (keyPressed)
			handleKeyPress (keyCode, unicodeChar);
	}

	void handleKeyReleaseEvent (const XKeyEvent* const keyEvent)
	{
		updateKeyStates (keyEvent->keycode, false);
		KeySym sym;

		{
			ScopedXLock xlock;
			sym = XkbKeycodeToKeysym (display, keyEvent->keycode, 0, 0);
		}

		const ModifierKeys oldMods (currentModifiers);
		const bool keyDownChange = (sym != NoSymbol) && ! updateKeyModifiersFromSym (sym, false);

		if (oldMods != currentModifiers)
			handleModifierKeysChange();

		if (keyDownChange)
			handleKeyUpOrDown (false);
	}

	void handleButtonPressEvent (const XButtonPressedEvent* const buttonPressEvent)
	{
		updateKeyModifiers (buttonPressEvent->state);

		bool buttonMsg = false;
		const int map = pointerMap [buttonPressEvent->button - Button1];

		if (map == Keys::WheelUp || map == Keys::WheelDown)
		{
			handleMouseWheel (0, Point<int> (buttonPressEvent->x, buttonPressEvent->y),
							  getEventTime (buttonPressEvent->time), 0, map == Keys::WheelDown ? -84.0f : 84.0f);
		}
		if (map == Keys::LeftButton)
		{
			currentModifiers = currentModifiers.withFlags (ModifierKeys::leftButtonModifier);
			buttonMsg = true;
		}
		else if (map == Keys::RightButton)
		{
			currentModifiers = currentModifiers.withFlags (ModifierKeys::rightButtonModifier);
			buttonMsg = true;
		}
		else if (map == Keys::MiddleButton)
		{
			currentModifiers = currentModifiers.withFlags (ModifierKeys::middleButtonModifier);
			buttonMsg = true;
		}

		if (buttonMsg)
		{
			toFront (true);

			handleMouseEvent (0, Point<int> (buttonPressEvent->x, buttonPressEvent->y), currentModifiers,
							  getEventTime (buttonPressEvent->time));
		}

		clearLastMousePos();
	}

	void handleButtonReleaseEvent (const XButtonReleasedEvent* const buttonRelEvent)
	{
		updateKeyModifiers (buttonRelEvent->state);
		const int map = pointerMap [buttonRelEvent->button - Button1];

		if (map == Keys::LeftButton)	currentModifiers = currentModifiers.withoutFlags (ModifierKeys::leftButtonModifier);
		else if (map == Keys::RightButton)  currentModifiers = currentModifiers.withoutFlags (ModifierKeys::rightButtonModifier);
		else if (map == Keys::MiddleButton) currentModifiers = currentModifiers.withoutFlags (ModifierKeys::middleButtonModifier);

		handleMouseEvent (0, Point<int> (buttonRelEvent->x, buttonRelEvent->y), currentModifiers,
						  getEventTime (buttonRelEvent->time));

		clearLastMousePos();
	}

	void handleMotionNotifyEvent (const XPointerMovedEvent* const movedEvent)
	{
		updateKeyModifiers (movedEvent->state);
		const Point<int> mousePos (movedEvent->x_root, movedEvent->y_root);

		if (lastMousePos != mousePos)
		{
			lastMousePos = mousePos;

			if (parentWindow != 0 && (styleFlags & windowHasTitleBar) == 0)
			{
				Window wRoot = 0, wParent = 0;

				{
					ScopedXLock xlock;
					unsigned int numChildren;
					Window* wChild = nullptr;
					XQueryTree (display, windowH, &wRoot, &wParent, &wChild, &numChildren);
				}

				if (wParent != 0
					 && wParent != windowH
					 && wParent != wRoot)
				{
					parentWindow = wParent;
					updateBounds();
				}
				else
				{
					parentWindow = 0;
				}
			}

			handleMouseEvent (0, mousePos - getScreenPosition(), currentModifiers, getEventTime (movedEvent->time));
		}
	}

	void handleEnterNotifyEvent (const XEnterWindowEvent* const enterEvent)
	{
		clearLastMousePos();

		if (! currentModifiers.isAnyMouseButtonDown())
		{
			updateKeyModifiers (enterEvent->state);
			handleMouseEvent (0, Point<int> (enterEvent->x, enterEvent->y), currentModifiers, getEventTime (enterEvent->time));
		}
	}

	void handleLeaveNotifyEvent (const XLeaveWindowEvent* const leaveEvent)
	{
		// Suppress the normal leave if we've got a pointer grab, or if
		// it's a bogus one caused by clicking a mouse button when running
		// in a Window manager
		if (((! currentModifiers.isAnyMouseButtonDown()) && leaveEvent->mode == NotifyNormal)
			 || leaveEvent->mode == NotifyUngrab)
		{
			updateKeyModifiers (leaveEvent->state);
			handleMouseEvent (0, Point<int> (leaveEvent->x, leaveEvent->y), currentModifiers, getEventTime (leaveEvent->time));
		}
	}

	void handleFocusInEvent()
	{
		isActiveApplication = true;
		if (isFocused())
			handleFocusGain();
	}

	void handleFocusOutEvent()
	{
		isActiveApplication = false;
		if (! isFocused())
			handleFocusLoss();
	}

	void handleExposeEvent (XExposeEvent* exposeEvent)
	{
		// Batch together all pending expose events
		XEvent nextEvent;
		ScopedXLock xlock;

		if (exposeEvent->window != windowH)
		{
			Window child;
			XTranslateCoordinates (display, exposeEvent->window, windowH,
								   exposeEvent->x, exposeEvent->y, &exposeEvent->x, &exposeEvent->y,
								   &child);
		}

		repaint (Rectangle<int> (exposeEvent->x, exposeEvent->y,
								 exposeEvent->width, exposeEvent->height));

		while (XEventsQueued (display, QueuedAfterFlush) > 0)
		{
			XPeekEvent (display, (XEvent*) &nextEvent);
			if (nextEvent.type != Expose || nextEvent.xany.window != exposeEvent->window)
				break;

			XNextEvent (display, (XEvent*) &nextEvent);
			XExposeEvent* nextExposeEvent = (XExposeEvent*) &nextEvent.xexpose;
			repaint (Rectangle<int> (nextExposeEvent->x, nextExposeEvent->y,
									 nextExposeEvent->width, nextExposeEvent->height));
		}
	}

	void handleConfigureNotifyEvent (XConfigureEvent* const confEvent)
	{
		updateBounds();
		updateBorderSize();
		handleMovedOrResized();

		// if the native title bar is dragged, need to tell any active menus, etc.
		if ((styleFlags & windowHasTitleBar) != 0
			  && component->isCurrentlyBlockedByAnotherModalComponent())
		{
			Component* const currentModalComp = Component::getCurrentlyModalComponent();

			if (currentModalComp != 0)
				currentModalComp->inputAttemptWhenModal();
		}

		if (confEvent->window == windowH
			 && confEvent->above != 0
			 && isFrontWindow())
		{
			handleBroughtToFront();
		}
	}

	void handleReparentNotifyEvent()
	{
		parentWindow = 0;
		Window wRoot = 0;
		Window* wChild = nullptr;
		unsigned int numChildren;

		{
			ScopedXLock xlock;
			XQueryTree (display, windowH, &wRoot, &parentWindow, &wChild, &numChildren);
		}

		if (parentWindow == windowH || parentWindow == wRoot)
			parentWindow = 0;

		handleGravityNotify();
	}

	void handleGravityNotify()
	{
		updateBounds();
		updateBorderSize();
		handleMovedOrResized();
	}

	void handleMappingNotify (XMappingEvent* const mappingEvent)
	{
		if (mappingEvent->request != MappingPointer)
		{
			// Deal with modifier/keyboard mapping
			ScopedXLock xlock;
			XRefreshKeyboardMapping (mappingEvent);
			updateModifierMappings();
		}
	}

	void handleClientMessageEvent (XClientMessageEvent* const clientMsg, XEvent* event)
	{
		if (clientMsg->message_type == Atoms::Protocols && clientMsg->format == 32)
		{
			const Atom atom = (Atom) clientMsg->data.l[0];

			if (atom == Atoms::ProtocolList [Atoms::PING])
			{
				Window root = RootWindow (display, DefaultScreen (display));

				clientMsg->window = root;

				XSendEvent (display, root, False, NoEventMask, event);
				XFlush (display);
			}
			else if (atom == Atoms::ProtocolList [Atoms::TAKE_FOCUS])
			{
				XWindowAttributes atts;

				ScopedXLock xlock;
				if (clientMsg->window != 0
					 && XGetWindowAttributes (display, clientMsg->window, &atts))
				{
					if (atts.map_state == IsViewable)
						XSetInputFocus (display, clientMsg->window, RevertToParent, clientMsg->data.l[1]);
				}
			}
			else if (atom == Atoms::ProtocolList [Atoms::DELETE_WINDOW])
			{
				handleUserClosingWindow();
			}
		}
		else if (clientMsg->message_type == Atoms::XdndEnter)
		{
			handleDragAndDropEnter (clientMsg);
		}
		else if (clientMsg->message_type == Atoms::XdndLeave)
		{
			resetDragAndDrop();
		}
		else if (clientMsg->message_type == Atoms::XdndPosition)
		{
			handleDragAndDropPosition (clientMsg);
		}
		else if (clientMsg->message_type == Atoms::XdndDrop)
		{
			handleDragAndDropDrop (clientMsg);
		}
		else if (clientMsg->message_type == Atoms::XdndStatus)
		{
			handleDragAndDropStatus (clientMsg);
		}
		else if (clientMsg->message_type == Atoms::XdndFinished)
		{
			resetDragAndDrop();
		}
	}

	void showMouseCursor (Cursor cursor) noexcept
	{
		ScopedXLock xlock;
		XDefineCursor (display, windowH, cursor);
	}

	void setTaskBarIcon (const Image& image)
	{
		ScopedXLock xlock;
		taskbarImage = image;

		Screen* const screen = XDefaultScreenOfDisplay (display);
		const int screenNumber = XScreenNumberOfScreen (screen);

		String screenAtom ("_NET_SYSTEM_TRAY_S");
		screenAtom << screenNumber;
		Atom selectionAtom = XInternAtom (display, screenAtom.toUTF8(), false);

		XGrabServer (display);
		Window managerWin = XGetSelectionOwner (display, selectionAtom);

		if (managerWin != None)
			XSelectInput (display, managerWin, StructureNotifyMask);

		XUngrabServer (display);
		XFlush (display);

		if (managerWin != None)
		{
			XEvent ev = { 0 };
			ev.xclient.type = ClientMessage;
			ev.xclient.window = managerWin;
			ev.xclient.message_type = XInternAtom (display, "_NET_SYSTEM_TRAY_OPCODE", False);
			ev.xclient.format = 32;
			ev.xclient.data.l[0] = CurrentTime;
			ev.xclient.data.l[1] = 0 /*SYSTEM_TRAY_REQUEST_DOCK*/;
			ev.xclient.data.l[2] = windowH;
			ev.xclient.data.l[3] = 0;
			ev.xclient.data.l[4] = 0;

			XSendEvent (display, managerWin, False, NoEventMask, &ev);
			XSync (display, False);
		}

		// For older KDE's ...
		long atomData = 1;
		Atom trayAtom = XInternAtom (display, "KWM_DOCKWINDOW", false);
		XChangeProperty (display, windowH, trayAtom, trayAtom, 32, PropModeReplace, (unsigned char*) &atomData, 1);

		// For more recent KDE's...
		trayAtom = XInternAtom (display, "_KDE_NET_WM_SYSTEM_TRAY_WINDOW_FOR", false);
		XChangeProperty (display, windowH, trayAtom, XA_WINDOW, 32, PropModeReplace, (unsigned char*) &windowH, 1);

		// a minimum size must be specified for GNOME and Xfce, otherwise the icon is displayed with a width of 1
		XSizeHints* hints = XAllocSizeHints();
		hints->flags = PMinSize;
		hints->min_width = 22;
		hints->min_height = 22;
		XSetWMNormalHints (display, windowH, hints);
		XFree (hints);
	}

	const Image& getTaskbarIcon() const noexcept	  { return taskbarImage; }

	bool dontRepaint;

	static ModifierKeys currentModifiers;
	static bool isActiveApplication;

private:

	class LinuxRepaintManager : public Timer
	{
	public:
		LinuxRepaintManager (LinuxComponentPeer* const peer_)
			: peer (peer_),
			  lastTimeImageUsed (0)
		{
		   #if JUCE_USE_XSHM
			shmCompletedDrawing = true;

			useARGBImagesForRendering = XSHMHelpers::isShmAvailable();

			if (useARGBImagesForRendering)
			{
				ScopedXLock xlock;
				XShmSegmentInfo segmentinfo;

				XImage* const testImage
					= XShmCreateImage (display, DefaultVisual (display, DefaultScreen (display)),
									   24, ZPixmap, 0, &segmentinfo, 64, 64);

				useARGBImagesForRendering = (testImage->bits_per_pixel == 32);
				XDestroyImage (testImage);
			}
		   #endif
		}

		void timerCallback()
		{
		   #if JUCE_USE_XSHM
			if (! shmCompletedDrawing)
				return;
		   #endif
			if (! regionsNeedingRepaint.isEmpty())
			{
				stopTimer();
				performAnyPendingRepaintsNow();
			}
			else if (Time::getApproximateMillisecondCounter() > lastTimeImageUsed + 3000)
			{
				stopTimer();
				image = Image::null;
			}
		}

		void repaint (const Rectangle<int>& area)
		{
			if (! isTimerRunning())
				startTimer (repaintTimerPeriod);

			regionsNeedingRepaint.add (area);
		}

		void performAnyPendingRepaintsNow()
		{
		   #if JUCE_USE_XSHM
			if (! shmCompletedDrawing)
			{
				startTimer (repaintTimerPeriod);
				return;
			}
		   #endif

			peer->clearMaskedRegion();

			RectangleList originalRepaintRegion (regionsNeedingRepaint);
			regionsNeedingRepaint.clear();
			const Rectangle<int> totalArea (originalRepaintRegion.getBounds());

			if (! totalArea.isEmpty())
			{
				if (image.isNull() || image.getWidth() < totalArea.getWidth()
					 || image.getHeight() < totalArea.getHeight())
				{
				   #if JUCE_USE_XSHM
					image = Image (new XBitmapImage (useARGBImagesForRendering ? Image::ARGB
																			   : Image::RGB,
				   #else
					image = Image (new XBitmapImage (Image::RGB,
				   #endif
													 (totalArea.getWidth() + 31) & ~31,
													 (totalArea.getHeight() + 31) & ~31,
													 false, peer->depth, peer->visual));
				}

				startTimer (repaintTimerPeriod);

				RectangleList adjustedList (originalRepaintRegion);
				adjustedList.offsetAll (-totalArea.getX(), -totalArea.getY());
				LowLevelGraphicsSoftwareRenderer context (image, -totalArea.getX(), -totalArea.getY(), adjustedList);

				if (peer->depth == 32)
				{
					RectangleList::Iterator i (originalRepaintRegion);

					while (i.next())
						image.clear (*i.getRectangle() - totalArea.getPosition());
				}

				peer->handlePaint (context);

				if (! peer->maskedRegion.isEmpty())
					originalRepaintRegion.subtract (peer->maskedRegion);

				for (RectangleList::Iterator i (originalRepaintRegion); i.next();)
				{
				   #if JUCE_USE_XSHM
					shmCompletedDrawing = false;
				   #endif
					const Rectangle<int>& r = *i.getRectangle();

					static_cast<XBitmapImage*> (image.getSharedImage())
						->blitToWindow (peer->windowH,
										r.getX(), r.getY(), r.getWidth(), r.getHeight(),
										r.getX() - totalArea.getX(), r.getY() - totalArea.getY());
				}
			}

			lastTimeImageUsed = Time::getApproximateMillisecondCounter();
			startTimer (repaintTimerPeriod);
		}

	   #if JUCE_USE_XSHM
		void notifyPaintCompleted()		 { shmCompletedDrawing = true; }
	   #endif

	private:
		enum { repaintTimerPeriod = 1000 / 100 };

		LinuxComponentPeer* const peer;
		Image image;
		uint32 lastTimeImageUsed;
		RectangleList regionsNeedingRepaint;

	   #if JUCE_USE_XSHM
		bool useARGBImagesForRendering, shmCompletedDrawing;
	   #endif
		JUCE_DECLARE_NON_COPYABLE (LinuxRepaintManager);
	};

	ScopedPointer <LinuxRepaintManager> repainter;

	friend class LinuxRepaintManager;
	Window windowH, parentWindow;
	int wx, wy, ww, wh;
	Image taskbarImage;
	bool fullScreen, mapped;
	Visual* visual;
	int depth;
	BorderSize<int> windowBorder;

	struct MotifWmHints
	{
		unsigned long flags;
		unsigned long functions;
		unsigned long decorations;
		long input_mode;
		unsigned long status;
	};

	static void updateKeyStates (const int keycode, const bool press) noexcept
	{
		const int keybyte = keycode >> 3;
		const int keybit = (1 << (keycode & 7));

		if (press)
			Keys::keyStates [keybyte] |= keybit;
		else
			Keys::keyStates [keybyte] &= ~keybit;
	}

	static void updateKeyModifiers (const int status) noexcept
	{
		int keyMods = 0;

		if ((status & ShiftMask) != 0)	 keyMods |= ModifierKeys::shiftModifier;
		if ((status & ControlMask) != 0)   keyMods |= ModifierKeys::ctrlModifier;
		if ((status & Keys::AltMask) != 0) keyMods |= ModifierKeys::altModifier;

		currentModifiers = currentModifiers.withOnlyMouseButtons().withFlags (keyMods);

		Keys::numLock  = ((status & Keys::NumLockMask) != 0);
		Keys::capsLock = ((status & LockMask) != 0);
	}

	static bool updateKeyModifiersFromSym (KeySym sym, const bool press) noexcept
	{
		int modifier = 0;
		bool isModifier = true;

		switch (sym)
		{
			case XK_Shift_L:
			case XK_Shift_R:
				modifier = ModifierKeys::shiftModifier;
				break;

			case XK_Control_L:
			case XK_Control_R:
				modifier = ModifierKeys::ctrlModifier;
				break;

			case XK_Alt_L:
			case XK_Alt_R:
				modifier = ModifierKeys::altModifier;
				break;

			case XK_Num_Lock:
				if (press)
					Keys::numLock = ! Keys::numLock;

				break;

			case XK_Caps_Lock:
				if (press)
					Keys::capsLock = ! Keys::capsLock;

				break;

			case XK_Scroll_Lock:
				break;

			default:
				isModifier = false;
				break;
		}

		if (modifier != 0)
		{
			if (press)
				currentModifiers = currentModifiers.withFlags (modifier);
			else
				currentModifiers = currentModifiers.withoutFlags (modifier);
		}

		return isModifier;
	}

	// Alt and Num lock are not defined by standard X
	// modifier constants: check what they're mapped to
	static void updateModifierMappings() noexcept
	{
		ScopedXLock xlock;
		const int altLeftCode = XKeysymToKeycode (display, XK_Alt_L);
		const int numLockCode = XKeysymToKeycode (display, XK_Num_Lock);

		Keys::AltMask = 0;
		Keys::NumLockMask = 0;

		XModifierKeymap* mapping = XGetModifierMapping (display);

		if (mapping)
		{
			for (int i = 0; i < 8; i++)
			{
				if (mapping->modifiermap [i << 1] == altLeftCode)
					Keys::AltMask = 1 << i;
				else if (mapping->modifiermap [i << 1] == numLockCode)
					Keys::NumLockMask = 1 << i;
			}

			XFreeModifiermap (mapping);
		}
	}

	void removeWindowDecorations (Window wndH)
	{
		Atom hints = XInternAtom (display, "_MOTIF_WM_HINTS", True);

		if (hints != None)
		{
			MotifWmHints motifHints = { 0 };
			motifHints.flags = 2; /* MWM_HINTS_DECORATIONS */
			motifHints.decorations = 0;

			ScopedXLock xlock;
			XChangeProperty (display, wndH, hints, hints, 32, PropModeReplace,
							 (unsigned char*) &motifHints, 4);
		}

		hints = XInternAtom (display, "_WIN_HINTS", True);

		if (hints != None)
		{
			long gnomeHints = 0;

			ScopedXLock xlock;
			XChangeProperty (display, wndH, hints, hints, 32, PropModeReplace,
							 (unsigned char*) &gnomeHints, 1);
		}

		hints = XInternAtom (display, "KWM_WIN_DECORATION", True);

		if (hints != None)
		{
			long kwmHints = 2; /*KDE_tinyDecoration*/

			ScopedXLock xlock;
			XChangeProperty (display, wndH, hints, hints, 32, PropModeReplace,
							 (unsigned char*) &kwmHints, 1);
		}
	}

	void addWindowButtons (Window wndH)
	{
		ScopedXLock xlock;
		Atom hints = XInternAtom (display, "_MOTIF_WM_HINTS", True);

		if (hints != None)
		{
			MotifWmHints motifHints = { 0 };
			motifHints.flags = 1 | 2; /* MWM_HINTS_FUNCTIONS | MWM_HINTS_DECORATIONS */
			motifHints.decorations = 2 /* MWM_DECOR_BORDER */ | 8 /* MWM_DECOR_TITLE */ | 16; /* MWM_DECOR_MENU */

			motifHints.functions = 4 /* MWM_FUNC_MOVE */;

			if ((styleFlags & windowHasCloseButton) != 0)
				motifHints.functions |= 32; /* MWM_FUNC_CLOSE */

			if ((styleFlags & windowHasMinimiseButton) != 0)
			{
				motifHints.functions |= 8; /* MWM_FUNC_MINIMIZE */
				motifHints.decorations |= 0x20; /* MWM_DECOR_MINIMIZE */
			}

			if ((styleFlags & windowHasMaximiseButton) != 0)
			{
				motifHints.functions |= 0x10; /* MWM_FUNC_MAXIMIZE */
				motifHints.decorations |= 0x40; /* MWM_DECOR_MAXIMIZE */
			}

			if ((styleFlags & windowIsResizable) != 0)
			{
				motifHints.functions |= 2; /* MWM_FUNC_RESIZE */
				motifHints.decorations |= 0x4; /* MWM_DECOR_RESIZEH */
			}

			XChangeProperty (display, wndH, hints, hints, 32, 0, (unsigned char*) &motifHints, 5);
		}

		hints = XInternAtom (display, "_NET_WM_ALLOWED_ACTIONS", True);

		if (hints != None)
		{
			int netHints [6];
			int num = 0;

			if ((styleFlags & windowIsResizable) != 0)
				netHints [num++] = XInternAtom (display, "_NET_WM_ACTION_RESIZE", True);

			if ((styleFlags & windowHasMaximiseButton) != 0)
				netHints [num++] = XInternAtom (display, "_NET_WM_ACTION_FULLSCREEN", True);

			if ((styleFlags & windowHasMinimiseButton) != 0)
				netHints [num++] = XInternAtom (display, "_NET_WM_ACTION_MINIMIZE", True);

			if ((styleFlags & windowHasCloseButton) != 0)
				netHints [num++] = XInternAtom (display, "_NET_WM_ACTION_CLOSE", True);

			XChangeProperty (display, wndH, hints, XA_ATOM, 32, PropModeReplace, (unsigned char*) &netHints, num);
		}
	}

	void setWindowType()
	{
		int netHints [2];
		int numHints = 0;

		if ((styleFlags & windowIsTemporary) != 0
			 || ((styleFlags & windowHasDropShadow) == 0 && Desktop::canUseSemiTransparentWindows()))
			netHints [numHints++] = XInternAtom (display, "_NET_WM_WINDOW_TYPE_COMBO", True);
		else
			netHints [numHints++] = XInternAtom (display, "_NET_WM_WINDOW_TYPE_NORMAL", True);

		netHints[numHints++] = XInternAtom (display, "_KDE_NET_WM_WINDOW_TYPE_OVERRIDE", True);

		XChangeProperty (display, windowH, Atoms::WindowType, XA_ATOM, 32, PropModeReplace,
						 (unsigned char*) &netHints, numHints);

		numHints = 0;

		if ((styleFlags & windowAppearsOnTaskbar) == 0)
			netHints [numHints++] = XInternAtom (display, "_NET_WM_STATE_SKIP_TASKBAR", True);

		if (component->isAlwaysOnTop())
			netHints [numHints++] = XInternAtom (display, "_NET_WM_STATE_ABOVE", True);

		if (numHints > 0)
			XChangeProperty (display, windowH, Atoms::WindowState, XA_ATOM, 32, PropModeReplace,
							 (unsigned char*) &netHints, numHints);
	}

	void createWindow (Window parentToAddTo)
	{
		ScopedXLock xlock;
		Atoms::initialiseAtoms();
		resetDragAndDrop();

		// Get defaults for various properties
		const int screen = DefaultScreen (display);
		Window root = RootWindow (display, screen);

		// Try to obtain a 32-bit visual or fallback to 24 or 16
		visual = Visuals::findVisualFormat ((styleFlags & windowIsSemiTransparent) ? 32 : 24, depth);

		if (visual == 0)
		{
			Logger::outputDebugString ("ERROR: System doesn't support 32, 24 or 16 bit RGB display.\n");
			Process::terminate();
		}

		// Create and install a colormap suitable fr our visual
		Colormap colormap = XCreateColormap (display, root, visual, AllocNone);
		XInstallColormap (display, colormap);

		// Set up the window attributes
		XSetWindowAttributes swa;
		swa.border_pixel = 0;
		swa.background_pixmap = None;
		swa.colormap = colormap;
		swa.event_mask = getAllEventsMask();

		windowH = XCreateWindow (display, parentToAddTo != 0 ? parentToAddTo : root,
								 0, 0, 1, 1,
								 0, depth, InputOutput, visual,
								 CWBorderPixel | CWColormap | CWBackPixmap | CWEventMask,
								 &swa);

		XGrabButton (display, AnyButton, AnyModifier, windowH, False,
					 ButtonPressMask | ButtonReleaseMask | EnterWindowMask | LeaveWindowMask | PointerMotionMask,
					 GrabModeAsync, GrabModeAsync, None, None);

		// Set the window context to identify the window handle object
		if (XSaveContext (display, (XID) windowH, windowHandleXContext, (XPointer) this))
		{
			// Failed
			jassertfalse;
			Logger::outputDebugString ("Failed to create context information for window.\n");
			XDestroyWindow (display, windowH);
			windowH = 0;
			return;
		}

		// Set window manager hints
		XWMHints* wmHints = XAllocWMHints();
		wmHints->flags = InputHint | StateHint;
		wmHints->input = True;	  // Locally active input model
		wmHints->initial_state = NormalState;
		XSetWMHints (display, windowH, wmHints);
		XFree (wmHints);

		// Set the window type
		setWindowType();

		// Define decoration
		if ((styleFlags & windowHasTitleBar) == 0)
			removeWindowDecorations (windowH);
		else
			addWindowButtons (windowH);

		setTitle (getComponent()->getName());

		// Associate the PID, allowing to be shut down when something goes wrong
		unsigned long pid = getpid();
		XChangeProperty (display, windowH, Atoms::Pid, XA_CARDINAL, 32, PropModeReplace,
						 (unsigned char*) &pid, 1);

		// Set window manager protocols
		XChangeProperty (display, windowH, Atoms::Protocols, XA_ATOM, 32, PropModeReplace,
						 (unsigned char*) Atoms::ProtocolList, 2);

		// Set drag and drop flags
		XChangeProperty (display, windowH, Atoms::XdndTypeList, XA_ATOM, 32, PropModeReplace,
						 (const unsigned char*) Atoms::allowedMimeTypes, numElementsInArray (Atoms::allowedMimeTypes));

		XChangeProperty (display, windowH, Atoms::XdndActionList, XA_ATOM, 32, PropModeReplace,
						 (const unsigned char*) Atoms::allowedActions, numElementsInArray (Atoms::allowedActions));

		XChangeProperty (display, windowH, Atoms::XdndActionDescription, XA_STRING, 8, PropModeReplace,
						 (const unsigned char*) "", 0);

		unsigned long dndVersion = Atoms::DndVersion;
		XChangeProperty (display, windowH, Atoms::XdndAware, XA_ATOM, 32, PropModeReplace,
						 (const unsigned char*) &dndVersion, 1);

		// Initialise the pointer and keyboard mapping
		// This is not the same as the logical pointer mapping the X server uses:
		// we don't mess with this.
		bool mappingInitialised = false;

		if (! mappingInitialised)
		{
			mappingInitialised = true;

			const int numButtons = XGetPointerMapping (display, 0, 0);

			if (numButtons == 2)
			{
				pointerMap[0] = Keys::LeftButton;
				pointerMap[1] = Keys::RightButton;
				pointerMap[2] = pointerMap[3] = pointerMap[4] = Keys::NoButton;
			}
			else if (numButtons >= 3)
			{
				pointerMap[0] = Keys::LeftButton;
				pointerMap[1] = Keys::MiddleButton;
				pointerMap[2] = Keys::RightButton;

				if (numButtons >= 5)
				{
					pointerMap[3] = Keys::WheelUp;
					pointerMap[4] = Keys::WheelDown;
				}
			}

			updateModifierMappings();
		}
	}

	void destroyWindow()
	{
		ScopedXLock xlock;

		XPointer handlePointer;
		if (! XFindContext (display, (XID) windowH, windowHandleXContext, &handlePointer))
			XDeleteContext (display, (XID) windowH, windowHandleXContext);

		XDestroyWindow (display, windowH);

		// Wait for it to complete and then remove any events for this
		// window from the event queue.
		XSync (display, false);

		XEvent event;
		while (XCheckWindowEvent (display, windowH, getAllEventsMask(), &event) == True)
		{}
	}

	static int getAllEventsMask() noexcept
	{
		return NoEventMask | KeyPressMask | KeyReleaseMask | ButtonPressMask | ButtonReleaseMask
				 | EnterWindowMask | LeaveWindowMask | PointerMotionMask | KeymapStateMask
				 | ExposureMask | StructureNotifyMask | FocusChangeMask;
	}

	static int64 getEventTime (::Time t)
	{
		static int64 eventTimeOffset = 0x12345678;
		const int64 thisMessageTime = t;

		if (eventTimeOffset == 0x12345678)
			eventTimeOffset = Time::currentTimeMillis() - thisMessageTime;

		return eventTimeOffset + thisMessageTime;
	}

	void updateBorderSize()
	{
		if ((styleFlags & windowHasTitleBar) == 0)
		{
			windowBorder = BorderSize<int> (0);
		}
		else if (windowBorder.getTopAndBottom() == 0 && windowBorder.getLeftAndRight() == 0)
		{
			ScopedXLock xlock;
			Atom hints = XInternAtom (display, "_NET_FRAME_EXTENTS", True);

			if (hints != None)
			{
				unsigned char* data = nullptr;
				unsigned long nitems, bytesLeft;
				Atom actualType;
				int actualFormat;

				if (XGetWindowProperty (display, windowH, hints, 0, 4, False,
										XA_CARDINAL, &actualType, &actualFormat, &nitems, &bytesLeft,
										&data) == Success)
				{
					const unsigned long* const sizes = (const unsigned long*) data;

					if (actualFormat == 32)
						windowBorder = BorderSize<int> ((int) sizes[2], (int) sizes[0],
														(int) sizes[3], (int) sizes[1]);

					XFree (data);
				}
			}
		}
	}

	void updateBounds()
	{
		jassert (windowH != 0);
		if (windowH != 0)
		{
			Window root, child;
			unsigned int bw, depth;
			ScopedXLock xlock;

			if (! XGetGeometry (display, (::Drawable) windowH, &root,
								&wx, &wy, (unsigned int*) &ww, (unsigned int*) &wh,
								&bw, &depth))
			{
				wx = wy = ww = wh = 0;
			}
			else if (! XTranslateCoordinates (display, windowH, root, 0, 0, &wx, &wy, &child))
			{
				wx = wy = 0;
			}
		}
	}

	void resetDragAndDrop()
	{
		dragAndDropFiles.clear();
		lastDropPos = Point<int> (-1, -1);
		dragAndDropCurrentMimeType = 0;
		dragAndDropSourceWindow = 0;
		srcMimeTypeAtomList.clear();
	}

	void sendDragAndDropMessage (XClientMessageEvent& msg)
	{
		msg.type = ClientMessage;
		msg.display = display;
		msg.window = dragAndDropSourceWindow;
		msg.format = 32;
		msg.data.l[0] = windowH;

		ScopedXLock xlock;
		XSendEvent (display, dragAndDropSourceWindow, False, 0, (XEvent*) &msg);
	}

	void sendDragAndDropStatus (const bool acceptDrop, Atom dropAction)
	{
		XClientMessageEvent msg = { 0 };
		msg.message_type = Atoms::XdndStatus;
		msg.data.l[1] = (acceptDrop ? 1 : 0) | 2; // 2 indicates that we want to receive position messages
		msg.data.l[4] = dropAction;

		sendDragAndDropMessage (msg);
	}

	void sendDragAndDropLeave()
	{
		XClientMessageEvent msg = { 0 };
		msg.message_type = Atoms::XdndLeave;
		sendDragAndDropMessage (msg);
	}

	void sendDragAndDropFinish()
	{
		XClientMessageEvent msg = { 0 };
		msg.message_type = Atoms::XdndFinished;
		sendDragAndDropMessage (msg);
	}

	void handleDragAndDropStatus (const XClientMessageEvent* const clientMsg)
	{
		if ((clientMsg->data.l[1] & 1) == 0)
		{
			sendDragAndDropLeave();

			if (dragAndDropFiles.size() > 0)
				handleFileDragExit (dragAndDropFiles);

			dragAndDropFiles.clear();
		}
	}

	void handleDragAndDropPosition (const XClientMessageEvent* const clientMsg)
	{
		if (dragAndDropSourceWindow == 0)
			return;

		dragAndDropSourceWindow = clientMsg->data.l[0];

		Point<int> dropPos ((int) clientMsg->data.l[2] >> 16,
							(int) clientMsg->data.l[2] & 0xffff);
		dropPos -= getScreenPosition();

		if (lastDropPos != dropPos)
		{
			lastDropPos = dropPos;
			dragAndDropTimestamp = clientMsg->data.l[3];

			Atom targetAction = Atoms::XdndActionCopy;

			for (int i = numElementsInArray (Atoms::allowedActions); --i >= 0;)
			{
				if ((Atom) clientMsg->data.l[4] == Atoms::allowedActions[i])
				{
					targetAction = Atoms::allowedActions[i];
					break;
				}
			}

			sendDragAndDropStatus (true, targetAction);

			if (dragAndDropFiles.size() == 0)
				updateDraggedFileList (clientMsg);

			if (dragAndDropFiles.size() > 0)
				handleFileDragMove (dragAndDropFiles, dropPos);
		}
	}

	void handleDragAndDropDrop (const XClientMessageEvent* const clientMsg)
	{
		if (dragAndDropFiles.size() == 0)
			updateDraggedFileList (clientMsg);

		const StringArray files (dragAndDropFiles);
		const Point<int> lastPos (lastDropPos);

		sendDragAndDropFinish();
		resetDragAndDrop();

		if (files.size() > 0)
			handleFileDragDrop (files, lastPos);
	}

	void handleDragAndDropEnter (const XClientMessageEvent* const clientMsg)
	{
		dragAndDropFiles.clear();
		srcMimeTypeAtomList.clear();

		dragAndDropCurrentMimeType = 0;
		const unsigned long dndCurrentVersion = static_cast <unsigned long> (clientMsg->data.l[1] & 0xff000000) >> 24;

		if (dndCurrentVersion < 3 || dndCurrentVersion > Atoms::DndVersion)
		{
			dragAndDropSourceWindow = 0;
			return;
		}

		dragAndDropSourceWindow = clientMsg->data.l[0];

		if ((clientMsg->data.l[1] & 1) != 0)
		{
			Atom actual;
			int format;
			unsigned long count = 0, remaining = 0;
			unsigned char* data = 0;

			ScopedXLock xlock;
			XGetWindowProperty (display, dragAndDropSourceWindow, Atoms::XdndTypeList,
								0, 0x8000000L, False, XA_ATOM, &actual, &format,
								&count, &remaining, &data);

			if (data != 0)
			{
				if (actual == XA_ATOM && format == 32 && count != 0)
				{
					const unsigned long* const types = (const unsigned long*) data;

					for (unsigned int i = 0; i < count; ++i)
						if (types[i] != None)
							srcMimeTypeAtomList.add (types[i]);
				}

				XFree (data);
			}
		}

		if (srcMimeTypeAtomList.size() == 0)
		{
			for (int i = 2; i < 5; ++i)
				if (clientMsg->data.l[i] != None)
					srcMimeTypeAtomList.add (clientMsg->data.l[i]);

			if (srcMimeTypeAtomList.size() == 0)
			{
				dragAndDropSourceWindow = 0;
				return;
			}
		}

		for (int i = 0; i < srcMimeTypeAtomList.size() && dragAndDropCurrentMimeType == 0; ++i)
			for (int j = 0; j < numElementsInArray (Atoms::allowedMimeTypes); ++j)
				if (srcMimeTypeAtomList[i] == Atoms::allowedMimeTypes[j])
					dragAndDropCurrentMimeType = Atoms::allowedMimeTypes[j];

		handleDragAndDropPosition (clientMsg);
	}

	void handleDragAndDropSelection (const XEvent* const evt)
	{
		dragAndDropFiles.clear();

		if (evt->xselection.property != 0)
		{
			StringArray lines;

			{
				MemoryBlock dropData;

				for (;;)
				{
					Atom actual;
					uint8* data = 0;
					unsigned long count = 0, remaining = 0;
					int format = 0;
					ScopedXLock xlock;

					if (XGetWindowProperty (display, evt->xany.window, evt->xselection.property,
											dropData.getSize() / 4, 65536, 1, AnyPropertyType, &actual,
											&format, &count, &remaining, &data) == Success)
					{
						dropData.append (data, count * format / 8);
						XFree (data);

						if (remaining == 0)
							break;
					}
					else
					{
						XFree (data);
						break;
					}
				}

				lines.addLines (dropData.toString());
			}

			for (int i = 0; i < lines.size(); ++i)
				dragAndDropFiles.add (URL::removeEscapeChars (lines[i].fromFirstOccurrenceOf ("file://", false, true)));

			dragAndDropFiles.trim();
			dragAndDropFiles.removeEmptyStrings();
		}
	}

	void updateDraggedFileList (const XClientMessageEvent* const clientMsg)
	{
		dragAndDropFiles.clear();

		if (dragAndDropSourceWindow != None
			 && dragAndDropCurrentMimeType != 0)
		{
			dragAndDropTimestamp = clientMsg->data.l[2];

			ScopedXLock xlock;
			XConvertSelection (display,
							   Atoms::XdndSelection,
							   dragAndDropCurrentMimeType,
							   XInternAtom (display, "JXSelectionWindowProperty", 0),
							   windowH,
							   dragAndDropTimestamp);
		}
	}

	StringArray dragAndDropFiles;
	int dragAndDropTimestamp;
	Point<int> lastDropPos;

	Atom dragAndDropCurrentMimeType;
	Window dragAndDropSourceWindow;

	Array <Atom> srcMimeTypeAtomList;

	int pointerMap[5];
	static Point<int> lastMousePos;

	static void clearLastMousePos() noexcept
	{
		lastMousePos = Point<int> (0x100000, 0x100000);
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (LinuxComponentPeer);
};

ModifierKeys LinuxComponentPeer::currentModifiers;
bool LinuxComponentPeer::isActiveApplication = false;
Point<int> LinuxComponentPeer::lastMousePos;

bool Process::isForegroundProcess()
{
	return LinuxComponentPeer::isActiveApplication;
}

void ModifierKeys::updateCurrentModifiers() noexcept
{
	currentModifiers = LinuxComponentPeer::currentModifiers;
}

const ModifierKeys ModifierKeys::getCurrentModifiersRealtime() noexcept
{
	Window root, child;
	int x, y, winx, winy;
	unsigned int mask;
	int mouseMods = 0;

	ScopedXLock xlock;

	if (XQueryPointer (display, RootWindow (display, DefaultScreen (display)),
					   &root, &child, &x, &y, &winx, &winy, &mask) != False)
	{
		if ((mask & Button1Mask) != 0)  mouseMods |= ModifierKeys::leftButtonModifier;
		if ((mask & Button2Mask) != 0)  mouseMods |= ModifierKeys::middleButtonModifier;
		if ((mask & Button3Mask) != 0)  mouseMods |= ModifierKeys::rightButtonModifier;
	}

	LinuxComponentPeer::currentModifiers = LinuxComponentPeer::currentModifiers.withoutMouseButtons().withFlags (mouseMods);
	return LinuxComponentPeer::currentModifiers;
}

void Desktop::setKioskComponent (Component* kioskModeComponent, bool enableOrDisable, bool allowMenusAndBars)
{
	if (enableOrDisable)
		kioskModeComponent->setBounds (Desktop::getInstance().getMainMonitorArea (false));
}

ComponentPeer* Component::createNewPeer (int styleFlags, void* nativeWindowToAttachTo)
{
	return new LinuxComponentPeer (this, styleFlags, (Window) nativeWindowToAttachTo);
}

// (this callback is hooked up in the messaging code)
void juce_windowMessageReceive (XEvent* event)
{
	if (event->xany.window != None)
	{
		LinuxComponentPeer* const peer = LinuxComponentPeer::getPeerFor (event->xany.window);

		if (ComponentPeer::isValidPeer (peer))
			peer->handleWindowMessage (event);
	}
	else
	{
		switch (event->xany.type)
		{
			case KeymapNotify:
			{
				const XKeymapEvent* const keymapEvent = (const XKeymapEvent*) &event->xkeymap;
				memcpy (Keys::keyStates, keymapEvent->key_vector, 32);
				break;
			}

			default:
				break;
		}
	}
}

void Desktop::getCurrentMonitorPositions (Array <Rectangle<int> >& monitorCoords, const bool /*clipToWorkArea*/)
{
	if (display == 0)
		return;

  #if JUCE_USE_XINERAMA
	int major_opcode, first_event, first_error;

	ScopedXLock xlock;
	if (XQueryExtension (display, "XINERAMA", &major_opcode, &first_event, &first_error))
	{
		typedef Bool (*tXineramaIsActive) (Display*);
		typedef XineramaScreenInfo* (*tXineramaQueryScreens) (Display*, int*);

		static tXineramaIsActive xXineramaIsActive = 0;
		static tXineramaQueryScreens xXineramaQueryScreens = 0;

		if (xXineramaIsActive == 0 || xXineramaQueryScreens == 0)
		{
			void* h = dlopen ("libXinerama.so", RTLD_GLOBAL | RTLD_NOW);

			if (h == 0)
				h = dlopen ("libXinerama.so.1", RTLD_GLOBAL | RTLD_NOW);

			if (h != 0)
			{
				xXineramaIsActive = (tXineramaIsActive) dlsym (h, "XineramaIsActive");
				xXineramaQueryScreens = (tXineramaQueryScreens) dlsym (h, "XineramaQueryScreens");
			}
		}

		if (xXineramaIsActive != 0
			&& xXineramaQueryScreens != 0
			&& xXineramaIsActive (display))
		{
			int numMonitors = 0;
			XineramaScreenInfo* const screens = xXineramaQueryScreens (display, &numMonitors);

			if (screens != 0)
			{
				for (int i = numMonitors; --i >= 0;)
				{
					int index = screens[i].screen_number;

					if (index >= 0)
					{
						while (monitorCoords.size() < index)
							monitorCoords.add (Rectangle<int>());

						monitorCoords.set (index, Rectangle<int> (screens[i].x_org,
																  screens[i].y_org,
																  screens[i].width,
																  screens[i].height));
					}
				}

				XFree (screens);
			}
		}
	}

	if (monitorCoords.size() == 0)
  #endif
	{
		Atom hints = XInternAtom (display, "_NET_WORKAREA", True);

		if (hints != None)
		{
			const int numMonitors = ScreenCount (display);

			for (int i = 0; i < numMonitors; ++i)
			{
				Window root = RootWindow (display, i);

				unsigned long nitems, bytesLeft;
				Atom actualType;
				int actualFormat;
				unsigned char* data = nullptr;

				if (XGetWindowProperty (display, root, hints, 0, 4, False,
										XA_CARDINAL, &actualType, &actualFormat, &nitems, &bytesLeft,
										&data) == Success)
				{
					const long* const position = (const long*) data;

					if (actualType == XA_CARDINAL && actualFormat == 32 && nitems == 4)
						monitorCoords.add (Rectangle<int> (position[0], position[1],
														   position[2], position[3]));

					XFree (data);
				}
			}
		}

		if (monitorCoords.size() == 0)
		{
			monitorCoords.add (Rectangle<int> (DisplayWidth (display, DefaultScreen (display)),
											   DisplayHeight (display, DefaultScreen (display))));
		}
	}
}

void Desktop::createMouseInputSources()
{
	mouseSources.add (new MouseInputSource (0, true));
}

bool Desktop::canUseSemiTransparentWindows() noexcept
{
	int matchedDepth = 0;
	const int desiredDepth = 32;

	return Visuals::findVisualFormat (desiredDepth, matchedDepth) != 0
			 && (matchedDepth == desiredDepth);
}

const Point<int> MouseInputSource::getCurrentMousePosition()
{
	Window root, child;
	int x, y, winx, winy;
	unsigned int mask;

	ScopedXLock xlock;

	if (XQueryPointer (display,
					   RootWindow (display, DefaultScreen (display)),
					   &root, &child,
					   &x, &y, &winx, &winy, &mask) == False)
	{
		// Pointer not on the default screen
		x = y = -1;
	}

	return Point<int> (x, y);
}

void Desktop::setMousePosition (const Point<int>& newPosition)
{
	ScopedXLock xlock;
	Window root = RootWindow (display, DefaultScreen (display));
	XWarpPointer (display, None, root, 0, 0, 0, 0, newPosition.getX(), newPosition.getY());
}

Desktop::DisplayOrientation Desktop::getCurrentOrientation() const
{
	return upright;
}

static bool screenSaverAllowed = true;

void Desktop::setScreenSaverEnabled (const bool isEnabled)
{
	if (screenSaverAllowed != isEnabled)
	{
		screenSaverAllowed = isEnabled;

		typedef void (*tXScreenSaverSuspend) (Display*, Bool);
		static tXScreenSaverSuspend xScreenSaverSuspend = 0;

		if (xScreenSaverSuspend == 0)
		{
			void* h = dlopen ("libXss.so", RTLD_GLOBAL | RTLD_NOW);

			if (h != 0)
				xScreenSaverSuspend = (tXScreenSaverSuspend) dlsym (h, "XScreenSaverSuspend");
		}

		ScopedXLock xlock;
		if (xScreenSaverSuspend != 0)
			xScreenSaverSuspend (display, ! isEnabled);
	}
}

bool Desktop::isScreenSaverEnabled()
{
	return screenSaverAllowed;
}

void* MouseCursor::createMouseCursorFromImage (const Image& image, int hotspotX, int hotspotY)
{
	ScopedXLock xlock;
	const unsigned int imageW = image.getWidth();
	const unsigned int imageH = image.getHeight();

  #if JUCE_USE_XCURSOR
	{
		typedef XcursorBool (*tXcursorSupportsARGB) (Display*);
		typedef XcursorImage* (*tXcursorImageCreate) (int, int);
		typedef void (*tXcursorImageDestroy) (XcursorImage*);
		typedef Cursor (*tXcursorImageLoadCursor) (Display*, const XcursorImage*);

		static tXcursorSupportsARGB xXcursorSupportsARGB = 0;
		static tXcursorImageCreate xXcursorImageCreate = 0;
		static tXcursorImageDestroy xXcursorImageDestroy = 0;
		static tXcursorImageLoadCursor xXcursorImageLoadCursor = 0;
		static bool hasBeenLoaded = false;

		if (! hasBeenLoaded)
		{
			hasBeenLoaded = true;
			void* h = dlopen ("libXcursor.so", RTLD_GLOBAL | RTLD_NOW);

			if (h != 0)
			{
				xXcursorSupportsARGB    = (tXcursorSupportsARGB)    dlsym (h, "XcursorSupportsARGB");
				xXcursorImageCreate     = (tXcursorImageCreate)     dlsym (h, "XcursorImageCreate");
				xXcursorImageLoadCursor = (tXcursorImageLoadCursor) dlsym (h, "XcursorImageLoadCursor");
				xXcursorImageDestroy    = (tXcursorImageDestroy)    dlsym (h, "XcursorImageDestroy");

				if (xXcursorSupportsARGB == 0 || xXcursorImageCreate == 0
					  || xXcursorImageLoadCursor == 0 || xXcursorImageDestroy == 0
					  || ! xXcursorSupportsARGB (display))
					xXcursorSupportsARGB = 0;
			}
		}

		if (xXcursorSupportsARGB != 0)
		{
			XcursorImage* xcImage = xXcursorImageCreate (imageW, imageH);

			if (xcImage != 0)
			{
				xcImage->xhot = hotspotX;
				xcImage->yhot = hotspotY;
				XcursorPixel* dest = xcImage->pixels;

				for (int y = 0; y < (int) imageH; ++y)
					for (int x = 0; x < (int) imageW; ++x)
						*dest++ = image.getPixelAt (x, y).getARGB();

				void* result = (void*) xXcursorImageLoadCursor (display, xcImage);
				xXcursorImageDestroy (xcImage);

				if (result != 0)
					return result;
			}
		}
	}
  #endif

	Window root = RootWindow (display, DefaultScreen (display));
	unsigned int cursorW, cursorH;
	if (! XQueryBestCursor (display, root, imageW, imageH, &cursorW, &cursorH))
		return nullptr;

	Image im (Image::ARGB, cursorW, cursorH, true);

	{
		Graphics g (im);

		if (imageW > cursorW || imageH > cursorH)
		{
			hotspotX = (hotspotX * cursorW) / imageW;
			hotspotY = (hotspotY * cursorH) / imageH;

			g.drawImageWithin (image, 0, 0, imageW, imageH,
							   RectanglePlacement::xLeft | RectanglePlacement::yTop | RectanglePlacement::onlyReduceInSize,
							   false);
		}
		else
		{
			g.drawImageAt (image, 0, 0);
		}
	}

	const int stride = (cursorW + 7) >> 3;
	HeapBlock <char> maskPlane, sourcePlane;
	maskPlane.calloc (stride * cursorH);
	sourcePlane.calloc (stride * cursorH);

	const bool msbfirst = (BitmapBitOrder (display) == MSBFirst);

	for (int y = cursorH; --y >= 0;)
	{
		for (int x = cursorW; --x >= 0;)
		{
			const char mask = (char) (1 << (msbfirst ? (7 - (x & 7)) : (x & 7)));
			const int offset = y * stride + (x >> 3);

			const Colour c (im.getPixelAt (x, y));

			if (c.getAlpha() >= 128)
				maskPlane[offset] |= mask;

			if (c.getBrightness() >= 0.5f)
				sourcePlane[offset] |= mask;
		}
	}

	Pixmap sourcePixmap = XCreatePixmapFromBitmapData (display, root, sourcePlane.getData(), cursorW, cursorH, 0xffff, 0, 1);
	Pixmap maskPixmap = XCreatePixmapFromBitmapData (display, root, maskPlane.getData(), cursorW, cursorH, 0xffff, 0, 1);

	XColor white, black;
	black.red = black.green = black.blue = 0;
	white.red = white.green = white.blue = 0xffff;

	void* result = (void*) XCreatePixmapCursor (display, sourcePixmap, maskPixmap, &white, &black, hotspotX, hotspotY);

	XFreePixmap (display, sourcePixmap);
	XFreePixmap (display, maskPixmap);

	return result;
}

void MouseCursor::deleteMouseCursor (void* const cursorHandle, const bool)
{
	ScopedXLock xlock;
	if (cursorHandle != 0)
		XFreeCursor (display, (Cursor) cursorHandle);
}

void* MouseCursor::createStandardMouseCursor (MouseCursor::StandardCursorType type)
{
	unsigned int shape;

	switch (type)
	{
		case NormalCursor:		  return None; // Use parent cursor
		case NoCursor:			  return createMouseCursorFromImage (Image (Image::ARGB, 16, 16, true), 0, 0);

		case WaitCursor:			shape = XC_watch; break;
		case IBeamCursor:		   shape = XC_xterm; break;
		case PointingHandCursor:		shape = XC_hand2; break;
		case LeftRightResizeCursor:	 shape = XC_sb_h_double_arrow; break;
		case UpDownResizeCursor:		shape = XC_sb_v_double_arrow; break;
		case UpDownLeftRightResizeCursor:   shape = XC_fleur; break;
		case TopEdgeResizeCursor:	   shape = XC_top_side; break;
		case BottomEdgeResizeCursor:	shape = XC_bottom_side; break;
		case LeftEdgeResizeCursor:	  shape = XC_left_side; break;
		case RightEdgeResizeCursor:	 shape = XC_right_side; break;
		case TopLeftCornerResizeCursor:	 shape = XC_top_left_corner; break;
		case TopRightCornerResizeCursor:	shape = XC_top_right_corner; break;
		case BottomLeftCornerResizeCursor:  shape = XC_bottom_left_corner; break;
		case BottomRightCornerResizeCursor: shape = XC_bottom_right_corner; break;
		case CrosshairCursor:		   shape = XC_crosshair; break;

		case DraggingHandCursor:
		{
			static unsigned char dragHandData[] = { 71,73,70,56,57,97,16,0,16,0,145,2,0,0,0,0,255,255,255,0,
			  0,0,0,0,0,33,249,4,1,0,0,2,0,44,0,0,0,0,16,0, 16,0,0,2,52,148,47,0,200,185,16,130,90,12,74,139,107,84,123,39,
			  132,117,151,116,132,146,248,60,209,138,98,22,203,114,34,236,37,52,77,217, 247,154,191,119,110,240,193,128,193,95,163,56,60,234,98,135,2,0,59 };
			const int dragHandDataSize = 99;

			return createMouseCursorFromImage (ImageFileFormat::loadFrom (dragHandData, dragHandDataSize), 8, 7);
		}

		case CopyingCursor:
		{
			static unsigned char copyCursorData[] = { 71,73,70,56,57,97,21,0,21,0,145,0,0,0,0,0,255,255,255,0,
			  128,128,255,255,255,33,249,4,1,0,0,3,0,44,0,0,0,0,21,0, 21,0,0,2,72,4,134,169,171,16,199,98,11,79,90,71,161,93,56,111,
			  78,133,218,215,137,31,82,154,100,200,86,91,202,142,12,108,212,87,235,174, 15,54,214,126,237,226,37,96,59,141,16,37,18,201,142,157,230,204,51,112,
			  252,114,147,74,83,5,50,68,147,208,217,16,71,149,252,124,5,0,59,0,0 };
			const int copyCursorSize = 119;

			return createMouseCursorFromImage (ImageFileFormat::loadFrom (copyCursorData, copyCursorSize), 1, 3);
		}

		default:
			jassertfalse;
			return None;
	}

	ScopedXLock xlock;
	return (void*) XCreateFontCursor (display, shape);
}

void MouseCursor::showInWindow (ComponentPeer* peer) const
{
	LinuxComponentPeer* const lp = dynamic_cast <LinuxComponentPeer*> (peer);

	if (lp != 0)
		lp->showMouseCursor ((Cursor) getHandle());
}

void MouseCursor::showInAllWindows() const
{
	for (int i = ComponentPeer::getNumPeers(); --i >= 0;)
		showInWindow (ComponentPeer::getPeer (i));
}

Image juce_createIconForFile (const File& file)
{
	return Image::null;
}

Image::SharedImage* Image::SharedImage::createNativeImage (PixelFormat format, int width, int height, bool clearImage)
{
	return createSoftwareImage (format, width, height, clearImage);
}

bool DragAndDropContainer::performExternalDragDropOfFiles (const StringArray& files, const bool canMoveFiles)
{
	jassertfalse;	// not implemented!
	return false;
}

bool DragAndDropContainer::performExternalDragDropOfText (const String& text)
{
	jassertfalse;	// not implemented!
	return false;
}

void SystemTrayIconComponent::setIconImage (const Image& newImage)
{
	if (! isOnDesktop ())
		addToDesktop (0);

	LinuxComponentPeer* const wp = dynamic_cast <LinuxComponentPeer*> (getPeer());

	if (wp != 0)
	{
		wp->setTaskBarIcon (newImage);

		setVisible (true);
		toFront (false);
		repaint();
	}
}

void SystemTrayIconComponent::paint (Graphics& g)
{
	LinuxComponentPeer* const wp = dynamic_cast <LinuxComponentPeer*> (getPeer());

	if (wp != 0)
	{
		g.drawImageWithin (wp->getTaskbarIcon(), 0, 0, getWidth(), getHeight(),
						   RectanglePlacement::xLeft | RectanglePlacement::yTop | RectanglePlacement::onlyReduceInSize,
						   false);
	}
}

void SystemTrayIconComponent::setIconTooltip (const String& tooltip)
{
	// xxx not yet implemented!
}

void LookAndFeel::playAlertSound()
{
	std::cout << "\a" << std::flush;
}

void JUCE_CALLTYPE NativeMessageBox::showMessageBox (AlertWindow::AlertIconType iconType,
													 const String& title, const String& message,
													 Component* associatedComponent)
{
	AlertWindow::showMessageBox (AlertWindow::NoIcon, title, message);
}

void JUCE_CALLTYPE NativeMessageBox::showMessageBoxAsync (AlertWindow::AlertIconType iconType,
														  const String& title, const String& message,
														  Component* associatedComponent)
{
	AlertWindow::showMessageBoxAsync (AlertWindow::NoIcon, title, message);
}

bool JUCE_CALLTYPE NativeMessageBox::showOkCancelBox (AlertWindow::AlertIconType iconType,
													  const String& title, const String& message,
													  Component* associatedComponent,
													  ModalComponentManager::Callback* callback)
{
	return AlertWindow::showOkCancelBox (iconType, title, message, String::empty, String::empty,
										 associatedComponent, callback);
}

int JUCE_CALLTYPE NativeMessageBox::showYesNoCancelBox (AlertWindow::AlertIconType iconType,
														const String& title, const String& message,
														Component* associatedComponent,
														ModalComponentManager::Callback* callback)
{
	return AlertWindow::showYesNoCancelBox (iconType, title, message,
											String::empty, String::empty, String::empty,
											associatedComponent, callback);
}

const int KeyPress::spaceKey		= XK_space & 0xff;
const int KeyPress::returnKey		   = XK_Return & 0xff;
const int KeyPress::escapeKey		   = XK_Escape & 0xff;
const int KeyPress::backspaceKey		= XK_BackSpace & 0xff;
const int KeyPress::leftKey		 = (XK_Left & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::rightKey		= (XK_Right & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::upKey		   = (XK_Up & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::downKey		 = (XK_Down & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::pageUpKey		   = (XK_Page_Up & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::pageDownKey		 = (XK_Page_Down & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::endKey		  = (XK_End & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::homeKey		 = (XK_Home & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::insertKey		   = (XK_Insert & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::deleteKey		   = (XK_Delete & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::tabKey		  = XK_Tab & 0xff;
const int KeyPress::F1Key		   = (XK_F1 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F2Key		   = (XK_F2 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F3Key		   = (XK_F3 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F4Key		   = (XK_F4 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F5Key		   = (XK_F5 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F6Key		   = (XK_F6 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F7Key		   = (XK_F7 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F8Key		   = (XK_F8 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F9Key		   = (XK_F9 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F10Key		  = (XK_F10 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F11Key		  = (XK_F11 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F12Key		  = (XK_F12 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F13Key		  = (XK_F13 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F14Key		  = (XK_F14 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F15Key		  = (XK_F15 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::F16Key		  = (XK_F16 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad0		  = (XK_KP_0 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad1		  = (XK_KP_1 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad2		  = (XK_KP_2 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad3		  = (XK_KP_3 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad4		  = (XK_KP_4 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad5		  = (XK_KP_5 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad6		  = (XK_KP_6 & 0xff) | Keys::extendedKeyModifier;
const int KeyPress::numberPad7		  = (XK_KP_7 & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPad8		  = (XK_KP_8 & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPad9		  = (XK_KP_9 & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadAdd		= (XK_KP_Add & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadSubtract	   = (XK_KP_Subtract & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadMultiply	   = (XK_KP_Multiply & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadDivide	 = (XK_KP_Divide & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadSeparator	  = (XK_KP_Separator & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadDecimalPoint   = (XK_KP_Decimal & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadEquals	 = (XK_KP_Equal & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::numberPadDelete	 = (XK_KP_Delete & 0xff)| Keys::extendedKeyModifier;
const int KeyPress::playKey		 = (0xffeeff00) | Keys::extendedKeyModifier;
const int KeyPress::stopKey		 = (0xffeeff01) | Keys::extendedKeyModifier;
const int KeyPress::fastForwardKey	  = (0xffeeff02) | Keys::extendedKeyModifier;
const int KeyPress::rewindKey		   = (0xffeeff03) | Keys::extendedKeyModifier;

#endif

/*** End of inlined file: juce_linux_Windowing.cpp ***/



/*** Start of inlined file: juce_linux_FileChooser.cpp ***/
// (This file gets included by juce_linux_NativeCode.cpp, rather than being
// compiled on its own).
#if JUCE_INCLUDED_FILE

void FileChooser::showPlatformDialog (Array<File>& results,
									  const String& title,
									  const File& file,
									  const String& filters,
									  bool isDirectory,
									  bool selectsFiles,
									  bool isSave,
									  bool warnAboutOverwritingExistingFiles,
									  bool selectMultipleFiles,
									  FilePreviewComponent* previewComponent)
{
	const String separator (":");
	String command ("zenity --file-selection");

	if (title.isNotEmpty())
		command << " --title=\"" << title << "\"";

	if (file != File::nonexistent)
		command << " --filename=\"" << file.getFullPathName () << "\"";

	if (isDirectory)
		command << " --directory";

	if (isSave)
		command << " --save";

	if (selectMultipleFiles)
		command << " --multiple --separator=\"" << separator << "\"";

	command << " 2>&1";

	MemoryOutputStream result;
	int status = -1;
	FILE* stream = popen (command.toUTF8(), "r");

	if (stream != 0)
	{
		for (;;)
		{
			char buffer [1024];
			const int bytesRead = fread (buffer, 1, sizeof (buffer), stream);

			if (bytesRead <= 0)
				break;

			result.write (buffer, bytesRead);
		}

		status = pclose (stream);
	}

	if (status == 0)
	{
		StringArray tokens;

		if (selectMultipleFiles)
			tokens.addTokens (result.toUTF8(), separator, String::empty);
		else
			tokens.add (result.toUTF8());

		for (int i = 0; i < tokens.size(); i++)
			results.add (File (tokens[i]));

		return;
	}

	//xxx ain't got one!
	jassertfalse;
}

#endif

/*** End of inlined file: juce_linux_FileChooser.cpp ***/


#endif

END_JUCE_NAMESPACE

#endif

