/**********************************************************************/
/*! \class RtMidi
    \brief An abstract base class for realtime MIDI input/output.

    This class implements some common functionality for the realtime
    MIDI input/output subclasses RtMidiIn and RtMidiOut.

    RtMidi WWW site: http://music.mcgill.ca/~gary/rtmidi/

    RtMidi: realtime MIDI i/o C++ classes
    Copyright (c) 2003-2011 Gary P. Scavone

    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and associated documentation files
    (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    Any person wishing to distribute modifications to the Software is
    requested to send the modifications to the original developer so that
    they can be incorporated into the canonical version.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
    ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
    CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/**********************************************************************/

// RtMidi: Version 1.0.15

#include "RtMidi.h"

// *********************************************************************//
// Common RtMidi Definitions
// *********************************************************************//

RtMidi::RtMidi()
  :apiData_(0), connected_(false)
{
}

void RtMidi::error(RtError::Type type)
{
  if(type == RtError::WARNING)
  {
    std::cerr << '\n' << errorString_ << "\n\n";
  }
  else if(type == RtError::DEBUG_WARNING)
  {
  }
  else
  {
    std::cerr << '\n' << errorString_ << "\n\n";
    throw RtError(errorString_, type);
  }
}

bool RtMidi::isOpen() const
{
  return connected_;
}

// *********************************************************************//
// Common RtMidiIn Definitions
// *********************************************************************//

RtMidiIn::RtMidiIn(const std::string clientName, unsigned int queueSizeLimit) :
  RtMidi()
{
  this->initialize(clientName);

  // Allocate the MIDI queue.
  inputData_.queue.ringSize = queueSizeLimit;

  if(inputData_.queue.ringSize > 0)
  {
    inputData_.queue.ring = new MidiMessage[inputData_.queue.ringSize];
  }
}

void RtMidiIn::setCallback(RtMidiCallback callback, void* userData)
{
  if(inputData_.usingCallback)
  {
    errorString_ = "RtMidiIn::setCallback: a callback function is already set!";
    error(RtError::WARNING);
    return;
  }

  if(!callback)
  {
    errorString_ = "RtMidiIn::setCallback: callback function value is invalid!";
    error(RtError::WARNING);
    return;
  }

  inputData_.userCallback = (void*)callback;
  inputData_.userData = userData;
  inputData_.usingCallback = true;
}

void RtMidiIn::cancelCallback()
{
  if(!inputData_.usingCallback)
  {
    errorString_ = "RtMidiIn::cancelCallback: no callback function was set!";
    error(RtError::WARNING);
    return;
  }

  inputData_.userCallback = 0;
  inputData_.userData = 0;
  inputData_.usingCallback = false;
}

void RtMidiIn::ignoreTypes(bool midiSysex, bool midiTime, bool midiSense)
{
  inputData_.ignoreFlags = 0;

  if(midiSysex)
  {
    inputData_.ignoreFlags = 0x01;
  }

  if(midiTime)
  {
    inputData_.ignoreFlags |= 0x02;
  }

  if(midiSense)
  {
    inputData_.ignoreFlags |= 0x04;
  }
}

double RtMidiIn::getMessage(std::vector<unsigned char>* message)
{
  message->clear();

  if(inputData_.usingCallback)
  {
    errorString_ = "RtMidiIn::getNextMessage: a user callback is currently set for this port.";
    error(RtError::WARNING);
    return 0.0;
  }

  if(inputData_.queue.size == 0)
  {
    return 0.0;
  }

  // Copy queued message to the vector pointer argument and then "pop" it.
  std::vector<unsigned char>* bytes = &(inputData_.queue.ring[inputData_.queue.front].bytes);
  message->assign(bytes->begin(), bytes->end());
  double deltaTime = inputData_.queue.ring[inputData_.queue.front].timeStamp;
  inputData_.queue.size--;
  inputData_.queue.front++;

  if(inputData_.queue.front == inputData_.queue.ringSize)
  {
    inputData_.queue.front = 0;
  }

  return deltaTime;
}

// *********************************************************************//
// Common RtMidiOut Definitions
// *********************************************************************//

RtMidiOut::RtMidiOut(const std::string clientName) :
  RtMidi()
{
  this->initialize(clientName);
}







