// *********************************************************************//
// API: WINDOWS Kernel Streaming
//
// Written by Sebastien Alaiwan, 2012.
//
// *********************************************************************//

#include "RtMidi.h"

#if defined(__WINDOWS_KS__)

#include <memory>
#include "ksmidi/ksenum.h"
#include "ksmidi/ksfilter.h"
#include "ksmidi/ksmidifilter.h"
#include "ksmidi/comutils.h"

struct WindowsKsData
{
  WindowsKsData() : m_pPin(NULL), m_Buffer(1024), m_hInputThread(NULL)
  {
    memset(&overlapped, 0, sizeof(OVERLAPPED));
    m_hExitEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
    m_hMutex = ::CreateMutex(NULL, FALSE, TEXT("Global\\RtMidiKernelStreamingMutex"));
    overlapped.hEvent = ::CreateEvent(NULL, FALSE, FALSE, NULL);
    m_hInputThread = NULL;
  }

  ~WindowsKsData()
  {
    ::CloseHandle(overlapped.hEvent);
    ::CloseHandle(m_hExitEvent);
    ::CloseHandle(m_hMutex);
  }

  std::vector<unsigned char> m_Buffer;
  OVERLAPPED overlapped;
  CKsPin* m_pPin;
  std::auto_ptr<CKsEnumFilters<CKsMidiCapFilter> > m_pCaptureEnum;
  std::auto_ptr<CKsEnumFilters<CKsMidiRenFilter> > m_pRenderEnum;
  HANDLE m_hInputThread;
  HANDLE m_hExitEvent;
  HANDLE m_hMutex;
};

// *********************************************************************//
// API: WINDOWS Kernel Streaming
// Class Definitions: RtMidiIn
// *********************************************************************//

DWORD WINAPI midiKsInputThread(VOID* pUser)
{
  RtMidiIn::RtMidiInData* data = static_cast<RtMidiIn::RtMidiInData*>(pUser);
  WindowsKsData* apiData = static_cast<WindowsKsData*>(data->apiData);

  HANDLE hEvents[] = { apiData->overlapped.hEvent, apiData->m_hExitEvent };

  while(true)
  {
    // std::cerr << "KsMidiInput::Read begins" << std::endl;
    KSSTREAM_HEADER packet;
    Zero(packet);
    packet.Size = sizeof(KSSTREAM_HEADER);
    packet.PresentationTime.Time = 0;
    packet.PresentationTime.Numerator = 1;
    packet.PresentationTime.Denominator = 1;
    packet.Data = &apiData->m_Buffer[0];
    packet.DataUsed = 0;
    packet.FrameExtent = apiData->m_Buffer.size();
    apiData->m_pPin->ReadData(&packet, &apiData->overlapped);

    DWORD dwRet = ::WaitForMultipleObjects(2, hEvents, FALSE, INFINITE);

    if(dwRet == WAIT_OBJECT_0)
    {
      // parse packet
      unsigned char* pData = (unsigned char*)packet.Data;
      unsigned int iOffset = 0;

      while(iOffset < packet.DataUsed)
      {
        KSMUSICFORMAT* pMusic = (KSMUSICFORMAT*)&pData[iOffset];
        iOffset += sizeof(KSMUSICFORMAT);

        RtMidiIn::MidiMessage message;
        if(packet.OptionsFlags & KSSTREAM_HEADER_OPTIONSF_TIMEVALID)
          message.timeStamp = pMusic->TimeDeltaMs + (packet.PresentationTime.Time * packet.PresentationTime.Numerator / packet.PresentationTime.Denominator);
        else
          message.timeStamp = 0;
        for(size_t i=0;i < pMusic->ByteCount;++i)
          message.bytes.push_back(pData[iOffset+i]);

        if(data->usingCallback)
        {
          RtMidiIn::RtMidiCallback callback = (RtMidiIn::RtMidiCallback)data->userCallback;
          callback(message.timeStamp, &message.bytes, data->userData);
        }
        else
        {
          // As long as we haven't reached our queue size limit, push the message.
          if(data->queue.size < data->queue.ringSize)
          {
            data->queue.ring[data->queue.back++] = message;

            if(data->queue.back == data->queue.ringSize)
            {
              data->queue.back = 0;
            }

            data->queue.size++;
          }
          else
          {
            std::cerr << "\nRtMidiIn: message queue limit reached!!\n\n";
          }
        }

        iOffset += pMusic->ByteCount;

        // re-align on 32 bits
        if(iOffset % 4 != 0)
        {
          iOffset += (4 - iOffset % 4);
        }

        //assert(iOffset >= packet.DataUsed);
        //return pMusic->ByteCount; // FIXME: only return one MIDI message by call.
      }
    }
    else
    {
      break;
    }
  }
  return 0;
}


void RtMidiIn::initialize(const std::string& clientName)
{
  WindowsKsData* data = new WindowsKsData;
  apiData_ = (void*)data;
  inputData_.apiData = data;

  GUID const aguidEnumCats[] =
  {
    { STATIC_KSCATEGORY_AUDIO }, { STATIC_KSCATEGORY_CAPTURE }
  };
  data->m_pCaptureEnum.reset(new CKsEnumFilters<CKsMidiCapFilter> );
  data->m_pCaptureEnum->EnumFilters(aguidEnumCats, ArrayLength(aguidEnumCats), TRUE, FALSE, TRUE);
}

RtMidiIn::~RtMidiIn()
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);

  try
  {
    if(data->m_pPin)
    {
      closePort();
    }
  }
  catch(...)
  {
  }

  // Delete the MIDI queue.
  if(inputData_.queue.ringSize > 0)
  {
    delete[] inputData_.queue.ring;
  }

  delete data;
}

void RtMidiIn::openPort(unsigned int portNumber, const std::string portName)
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);

  if(portNumber < 0 || portNumber >= data->m_pCaptureEnum->m_listFilters.size())
  {
    std::stringstream ost;
    ost << "RtMidiIn::openPort: the 'portNumber' argument (" << portNumber << ") is invalid.";
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  CKsMidiCapFilter* pFilter = data->m_pCaptureEnum->m_listFilters[portNumber];
  //std::cout << "Opening : " << pFilter->GetName() << std::endl;
  data->m_pPin = pFilter->CreateCapturePin();

  if(data->m_pPin == NULL)
  {
    std::stringstream ost;
    ost << "RtMidiIn::openPort: KS error opening port (could not create pin)";
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  data->m_pPin->SetState(KSSTATE_RUN);

  DWORD threadId;
  data->m_hInputThread = ::CreateThread(NULL, 0, &midiKsInputThread, &inputData_, 0, &threadId);
  if(data->m_hInputThread == NULL)
  {
    delete data;
    std::stringstream ost;
    ost << "RtMidiIn::initialize: Could not create input thread : Windows error " << GetLastError() << std::endl;;
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  connected_ = true;
}

void RtMidiIn::openVirtualPort(const std::string portName)
{
  // This function cannot be implemented for the Windows KS MIDI API.
  errorString_ = "RtMidiIn::openVirtualPort: cannot be implemented in Windows KS MIDI API!";
  error(RtError::WARNING);
}

unsigned int RtMidiIn::getPortCount()
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);
  return (unsigned int)data->m_pCaptureEnum->m_listFilters.size();
}

std::string RtMidiIn::getPortName(unsigned int portNumber)
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);


  if(portNumber < 0 || portNumber >= data->m_pCaptureEnum->m_listFilters.size())
  {
    std::stringstream ost;
    ost << "RtMidiIn::getPortName: the 'portNumber' argument (" << portNumber << ") is invalid.";
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  CKsMidiCapFilter* pFilter = data->m_pCaptureEnum->m_listFilters[portNumber];
  return pFilter->GetFriendlyName();
}

void RtMidiIn::closePort()
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);

  connected_ = false;

  if(data->m_hInputThread)
  {
    ::SignalObjectAndWait(data->m_hExitEvent, data->m_hInputThread, INFINITE, FALSE);
    ::CloseHandle(data->m_hInputThread);
    data->m_hInputThread = NULL;
  }

  if(data->m_pPin)
  {
    data->m_pPin->SetState(KSSTATE_PAUSE);
    data->m_pPin->SetState(KSSTATE_STOP);
    data->m_pPin->ClosePin();
    data->m_pPin = NULL;
  }
}

// *********************************************************************//
// API: WINDOWS Kernel Streaming
// Class Definitions: RtMidiOut
// *********************************************************************//

void RtMidiOut::initialize(const std::string& clientName)
{
  WindowsKsData* data = new WindowsKsData;

  data->m_pPin = NULL;
  data->m_pRenderEnum.reset(new CKsEnumFilters<CKsMidiRenFilter> );
  GUID const aguidEnumCats[] =
  {
    { STATIC_KSCATEGORY_AUDIO }, { STATIC_KSCATEGORY_RENDER }
  };
  data->m_pRenderEnum->EnumFilters(aguidEnumCats, ArrayLength(aguidEnumCats), true, FALSE, true);

  apiData_ = (void*)data;
}

RtMidiOut::~RtMidiOut()
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);
  closePort();
}

void RtMidiOut::openPort(unsigned int portNumber, const std::string portName)
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);

  if(portNumber < 0 || portNumber >= data->m_pRenderEnum->m_listFilters.size())
  {
    std::stringstream ost;
    ost << "RtMidiOut::openPort: the 'portNumber' argument (" << portNumber << ") is invalid.";
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  CKsMidiRenFilter* pFilter = data->m_pRenderEnum->m_listFilters[portNumber];
  data->m_pPin = pFilter->CreateRenderPin();

  if(data->m_pPin == NULL)
  {
    std::stringstream ost;
    ost << "RtMidiOut::openPort: KS error opening port (could not create pin)";
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  data->m_pPin->SetState(KSSTATE_RUN);
  connected_ = true;
}

void RtMidiOut::openVirtualPort(const std::string portName)
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);

  // This function cannot be implemented for the Windows KS MIDI API.
  errorString_ = "RtMidiOut::openVirtualPort: cannot be implemented in Windows KS MIDI API!";
  error(RtError::WARNING);
}

unsigned int RtMidiOut::getPortCount()
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);

  return (unsigned int)data->m_pRenderEnum->m_listFilters.size();
}

std::string RtMidiOut::getPortName(unsigned int portNumber)
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);


  if(portNumber < 0 || portNumber >= data->m_pRenderEnum->m_listFilters.size())
  {
    std::stringstream ost;
    ost << "RtMidiOut::getPortName: the 'portNumber' argument (" << portNumber << ") is invalid.";
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  CKsMidiRenFilter* pFilter = data->m_pRenderEnum->m_listFilters[portNumber];
  return pFilter->GetFriendlyName();
}

void RtMidiOut::closePort()
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);
  connected_ = false;

  if(data->m_pPin == NULL)
    return;

  data->m_pPin->SetState(KSSTATE_PAUSE);
  data->m_pPin->SetState(KSSTATE_STOP);
  data->m_pPin->ClosePin();
  data->m_pPin = NULL;
}

void RtMidiOut::sendMessage(std::vector<unsigned char>* message)
{
  WindowsKsData* data = static_cast<WindowsKsData*>(apiData_);

  size_t iNumMidiBytes = message->size();

  size_t pos = 0;

  // write header
  KSMUSICFORMAT* pKsMusicFormat = reinterpret_cast<KSMUSICFORMAT*>(&data->m_Buffer[pos]);
  pKsMusicFormat->TimeDeltaMs = 0;
  pKsMusicFormat->ByteCount = iNumMidiBytes;
  pos += sizeof(KSMUSICFORMAT);

  // write MIDI bytes
  if(pos + iNumMidiBytes > data->m_Buffer.size())
  {
    std::stringstream ost;
    ost << "KsMidiInput::Write: MIDI buffer too small. Required " << pos + iNumMidiBytes << " bytes, only has " << data->m_Buffer.size();
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  if(data->m_pPin == NULL)
  {
    std::stringstream ost;
    ost << "RtMidiOut::sendMessage: port is not open";
    errorString_ = ost.str();
    error(RtError::WARNING);
  }

  memcpy(&data->m_Buffer[pos], &(*message)[0], iNumMidiBytes);
  pos += iNumMidiBytes;

  KSSTREAM_HEADER packet;
  Zero(packet);
  packet.Size = sizeof(packet);
  packet.PresentationTime.Time = 0;
  packet.PresentationTime.Numerator = 1;
  packet.PresentationTime.Denominator = 1;
  packet.Data = const_cast<unsigned char*>(&data->m_Buffer[0]);
  packet.DataUsed = ((pos+3)/4)*4;
  packet.FrameExtent = data->m_Buffer.size();

  struct scopeLock
  {
    HANDLE const _mutex;
    scopeLock(HANDLE hMutex) : _mutex(hMutex)
    {
      ::WaitForSingleObject(_mutex, INFINITE);
    }

    ~scopeLock()
    {
      ::ReleaseMutex(_mutex);
    }
  };

  {
    scopeLock lock(data->m_hMutex);
    data->m_pPin->WriteData(&packet, &data->overlapped);
  }
}

#endif  // __WINDOWS_KS__
