#pragma once

#include <memory>
#include <vector>
#include <stdexcept>
#include <windows.h>
#include <setupapi.h>
#include <mmsystem.h>

#include "util.h"

template<typename TFilterType>
class CKsEnumFilters
{
private:
  void DestroyLists()
  {
    clearVector(m_listFilters);
  }

public:
  // TODO: make this private.
  std::vector<TFilterType*> m_listFilters;

public: ~CKsEnumFilters()
  {
    DestroyLists();
  }

  void EnumFilters(GUID const* aguidCategories, ULONG cguidCategories, bool fNeedPins, bool fNeedNodes, bool fInstantiate)
  {
    DestroyLists();

    if(aguidCategories && !IsEqualGUIDAligned(GUID_NULL, aguidCategories[0]))
    {
      // Get a handle to the device set specified by the guid
      HDEVINFO hDevInfo = SetupDiGetClassDevs
                          (
        &aguidCategories[0],
        NULL,
        NULL,
        DIGCF_PRESENT | DIGCF_DEVICEINTERFACE
                          );

      if(!IsValidHandle(hDevInfo))
      {
        DebugPrintf(TRACE_NORMAL, TEXT("No devices found"));
        throw std::runtime_error("CKsEnumFilters: no devices found");
      }

      // Loop through members of the set and get details for each
      for(int iClassMember = 0;; iClassMember++)
      {
        SP_DEVICE_INTERFACE_DATA DID;
        DID.cbSize = sizeof(DID);
        DID.Reserved = 0;

        SP_DEVICE_INTERFACE_DATA DIDAlias;
        DIDAlias.cbSize = sizeof(DIDAlias);
        DIDAlias.Reserved = 0;

        SP_DEVINFO_DATA DevInfoData;
        DevInfoData.cbSize = sizeof(DevInfoData);
        DevInfoData.Reserved = 0;

        bool fRes = SetupDiEnumDeviceInterfaces(hDevInfo, NULL, &aguidCategories[0], iClassMember, &DID);

        if(!fRes)
        {
          // This just means that we've enumerate all the devices - it's not a real error
          break;
        }

        // Get details for the device registered in this class
        DWORD const cbItfDetails = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA) + MAX_PATH * sizeof(WCHAR);
        std::vector<BYTE> buffer(cbItfDetails);

        SP_DEVICE_INTERFACE_DETAIL_DATA* pDevInterfaceDetails = reinterpret_cast<SP_DEVICE_INTERFACE_DETAIL_DATA*>(&buffer[0]);
        pDevInterfaceDetails->cbSize = sizeof(*pDevInterfaceDetails);

        fRes = SetupDiGetDeviceInterfaceDetail(hDevInfo, &DID, pDevInterfaceDetails, cbItfDetails, NULL, &DevInfoData);

        if(fRes)
        {
          //
          // check additional category guids which may (or may not) have been supplied
          //
          for(ULONG nguidCategory = 1; nguidCategory < cguidCategories && fRes; nguidCategory++)
          {
            fRes = SetupDiGetDeviceInterfaceAlias
                   (
              hDevInfo,
              &DID,
              &aguidCategories[nguidCategory],
              &DIDAlias
                   );

            if(!fRes)
            {
              DebugPrintf(TRACE_LOW, TEXT("Failed to get requested DeviceInterfaceAlias"));
            }

            //
            // Check if the this interface alias is enabled.
            //
            if(fRes)
            {
              if(!DIDAlias.Flags || (DIDAlias.Flags & SPINT_REMOVED))
              {
                fRes = false;
                DebugPrintf(TRACE_LOW, TEXT("DeviceInterfaceAlias disabled."));
              }
            }
          }

          if(fRes)
          {
            try
            {
              std::string sFriendlyName;

              {
                HKEY hRegKey = SetupDiOpenDeviceInterfaceRegKey(hDevInfo, &DID, 0, KEY_READ);

                if(hRegKey != INVALID_HANDLE_VALUE)
                {
                  char friendlyName[256];
                  DWORD dwSize = sizeof friendlyName;
                  LONG lval = RegQueryValueEx(hRegKey, TEXT("FriendlyName"), NULL, NULL, (LPBYTE)friendlyName, &dwSize);
                  sFriendlyName = friendlyName;
                  RegCloseKey(hRegKey);
                }
              }

              std::auto_ptr<TFilterType> pFilter(new TFilterType(pDevInterfaceDetails->DevicePath, sFriendlyName));

              if(fNeedPins || fInstantiate)
              {
                pFilter->Instantiate();

                if(fNeedPins)
                {
                  pFilter->EnumeratePins();
                }
              }

              m_listFilters.push_back(pFilter.release());
            }
            catch(std::runtime_error const& err)
            {
              DebugPrintf(
                TRACE_ERROR,
                TEXT("Failed to create filter for %s : %s"),
                pDevInterfaceDetails->DevicePath,
                err.what());
            }
          }
        }
      }

      SetupDiDestroyDeviceInfoList(hDevInfo);
    }

    if(m_listFilters.empty())
    {
      throw std::runtime_error("CKsEnumFilters: no filter found");
    }
  }
};

