/**
 * @file ksfilter.cpp
 * @author Sebastien Alaiwan
 * @date 2011-11-20
 */

/*
 * Copyright (C) 2011 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ksfilter.h"
#include "kspin.h"
#include "util.h"
#include "comutils.h"
#include <stdexcept>
#include <memory>

CKsFilter::CKsFilter(tstring const& sName, std::string const& sFriendlyName) :
  CKsObject(INVALID_HANDLE_VALUE),
  m_sFriendlyName(sFriendlyName),
  m_sName(sName)
{
  if(sName.empty())
  {
    ThrowFatalError(std::runtime_error("CKsFilter::CKsFilter: name can't be empty"));
  }
}

CKsFilter::~CKsFilter()
{
  clearVector(m_listPins);
  SafeCloseHandle(m_handle);
}

void CKsFilter::Instantiate()
{
  m_handle = CreateFile(
    m_sName.c_str(),
    GENERIC_READ | GENERIC_WRITE,
    0,
    NULL,
    OPEN_EXISTING,
    FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
    NULL);

  if(!IsValidHandle(m_handle))
  {
    DWORD const dwError = GetLastError();
    ThrowFatalError(ComException("CKsFilter::Instantiate: can't open driver", HRESULT_FROM_WIN32(dwError)));
  }
}

void CKsFilter::EnumeratePins()
{
  HRESULT hr;

  ULONG numPins;
  hr = GetPinProperty(0, KSPROPSETID_Pin, KSPROPERTY_PIN_CTYPES, numPins);

  if(FAILED(hr))
  {
    ThrowFatalError(ComException("CKsFilter::EnumeratePins: Could not get pin count", hr));
  }

  for(ULONG i = 0; i < numPins; ++i)
  {
    try
    {
      std::auto_ptr<CKsPin> pNewPin(new CKsPin(this, i));
      m_listPins.push_back(pNewPin.release());
    }
    catch(std::runtime_error const& err)
    {
      DebugPrintf(TRACE_ERROR, TEXT("Failed to construct pin"));
      continue;
    }
  }

  ClassifyPins();

  DWORD fViableFilter = (
    m_listCaptureSinkPins.size() ||
    m_listCaptureSourcePins.size() ||
    m_listRenderSinkPins.size() ||
    m_listRenderSourcePins.size() ||
    m_listNoCommPins.size());

  if(!fViableFilter)
  {
    ThrowFatalError(std::runtime_error("Filter has no pins."));
  }
}

void CKsFilter::ClassifyPins()
{
  m_listRenderSinkPins.clear();
  m_listRenderSourcePins.clear();
  m_listCaptureSinkPins.clear();
  m_listCaptureSourcePins.clear();
  m_listNoCommPins.clear();

  for(size_t i = 0; i < m_listPins.size(); ++i)
  {
    CKsPin* const pPin = m_listPins[i];

    KSPIN_COMMUNICATION const Communication = pPin->GetCommunication();
    KSPIN_DATAFLOW const DataFlow = pPin->GetDataFlow();

    if(KSPIN_DATAFLOW_IN == DataFlow)
    {
      switch(Communication)
      {
      case KSPIN_COMMUNICATION_SINK:
        m_listRenderSinkPins.push_back(pPin);
        break;

      case KSPIN_COMMUNICATION_SOURCE:
        m_listCaptureSourcePins.push_back(pPin);
        break;

      case KSPIN_COMMUNICATION_BOTH:
        m_listRenderSinkPins.push_back(pPin);
        m_listCaptureSourcePins.push_back(pPin);
        break;

      case KSPIN_COMMUNICATION_NONE:
        m_listNoCommPins.push_back(pPin);
        break;

      default:
        DebugPrintf(TRACE_LOW, TEXT("Pin communication type not recognized"));
        break;
      }
    }
    else
    {
      switch(Communication)
      {
      case KSPIN_COMMUNICATION_SINK:
        m_listCaptureSinkPins.push_back(pPin);
        break;

      case KSPIN_COMMUNICATION_SOURCE:
        m_listRenderSourcePins.push_back(pPin);
        break;

      case KSPIN_COMMUNICATION_BOTH:
        m_listCaptureSinkPins.push_back(pPin);
        m_listRenderSourcePins.push_back(pPin);
        break;

      case KSPIN_COMMUNICATION_NONE:
        m_listNoCommPins.push_back(pPin);
        break;

      default:
        DebugPrintf(TRACE_LOW, TEXT("Pin communication type not recognized"));
        break;
      }
    }
  }
}

HRESULT CKsFilter::GetPinPropertyMulti(ULONG nPinId, REFGUID guidPropertySet, ULONG nProperty, PKSMULTIPLE_ITEM* ppKsMultipleItem)
{
  HRESULT hr = S_OK;

  KSP_PIN ksPProp;
  ksPProp.Property.Set = guidPropertySet;
  ksPProp.Property.Id = nProperty;
  ksPProp.Property.Flags = KSPROPERTY_TYPE_GET;
  ksPProp.PinId = nPinId;
  ksPProp.Reserved = 0;

  ULONG cbMultipleItem = 0;
  hr = DeviceIoControl(m_handle,
                       IOCTL_KS_PROPERTY,
                       &ksPProp.Property,
                       sizeof(KSP_PIN),
                       NULL,
                       0,
                       &cbMultipleItem,
                       NULL);

  if(FAILED(hr))
  {
    return hr;
  }

  *ppKsMultipleItem = (PKSMULTIPLE_ITEM) new BYTE[cbMultipleItem];

  ULONG ulReturned = 0;
  hr = DeviceIoControl(
    m_handle,
    IOCTL_KS_PROPERTY,
    &ksPProp,
    sizeof(KSP_PIN),
    (PVOID)*ppKsMultipleItem,
    cbMultipleItem,
    &ulReturned,
    NULL);

  if(FAILED(hr))
  {
    return hr;
  }

  return S_OK;
}

