/**
 * @file ksmidifilter.cpp
 * @author Sebastien Alaiwan
 * @date 2011-11-20
 */

/*
 * Copyright (C) 2011 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ksmidifilter.h"
#include "util.h"
#include "comutils.h"
#include <cassert>
#include <stdexcept>
#include <algorithm>
#include <memory>
#include <windows.h>
#include <mmsystem.h>
#include "ksmidipin.h"

CKsMidiFilter::CKsMidiFilter(tstring const& sPath, std::string const& sFriendlyName) :
  CKsFilter(sPath, sFriendlyName)
{
}

void CKsMidiFilter::EnumeratePins()
{
  HRESULT hr = S_OK;

  // Get pin count.
  ULONG numPins = 0;
  hr = GetPinProperty(0, KSPROPSETID_Pin, KSPROPERTY_PIN_CTYPES, numPins);

  if(FAILED(hr))
  {
    ThrowFatalError(ComException("CKsMidiFilter::EnumeratePins: failed to retrieve pin count", hr));
  }

  for(ULONG iPin = 0; iPin < numPins; ++iPin)
  {
    try
    {
      KSPIN_COMMUNICATION ksPinCommunication;
      hr = GetPinProperty(iPin, KSPROPSETID_Pin, KSPROPERTY_PIN_COMMUNICATION, ksPinCommunication);

      if(FAILED(hr))
      {
        ThrowFatalError(ComException("Failed to retrieve pin property KSPROPERTY_PIN_COMMUNICATION", hr));
      }

      if(ksPinCommunication != KSPIN_COMMUNICATION_SOURCE &&
         ksPinCommunication != KSPIN_COMMUNICATION_SINK &&
         ksPinCommunication != KSPIN_COMMUNICATION_BOTH)
      {
        ThrowFatalError(std::runtime_error("Pin communication value doesn't make sense"));
      }

      KSPIN_DATAFLOW ksPinDataflow;
      hr = GetPinProperty(iPin, KSPROPSETID_Pin, KSPROPERTY_PIN_DATAFLOW, ksPinDataflow);

      if(FAILED(hr))
      {
        ThrowFatalError(ComException("Failed to retrieve pin property KSPROPERTY_PIN_DATAFLOW", hr));
      }

      std::auto_ptr<CKsMidiPin> pNewPin;

      if(ksPinCommunication == KSPIN_COMMUNICATION_SINK)
      {
        switch(ksPinDataflow)
        {
        case KSPIN_DATAFLOW_IN:
          pNewPin.reset(new CKsMidiPin(this, iPin));
          break;
        case KSPIN_DATAFLOW_OUT:
          pNewPin.reset(new CKsMidiPin(this, iPin));
          break;
        default:
          ThrowFatalError(std::runtime_error("Pin dataflow value doesn't make sense"));
        }
      }
      else
      {
        pNewPin.reset(new CKsMidiPin(this, iPin));
      }

      m_listPins.push_back(pNewPin.release());
    }
    catch(std::runtime_error const& err)
    {
      // pin instanciation has failed, continue to the next pin.
    }
  }

  ClassifyPins();

  if(m_listRenderSinkPins.empty()
     && m_listRenderSourcePins.empty()
     && m_listCaptureSinkPins.empty()
     && m_listCaptureSourcePins.empty())
  {
    ThrowFatalError(std::runtime_error("No valid pins found on the filter."));
  }
}

CKsMidiRenFilter::CKsMidiRenFilter(tstring const& sPath, std::string const& sFriendlyName) :
  CKsMidiFilter(sPath, sFriendlyName)
{
}

CKsMidiPin* CKsMidiRenFilter::CreateRenderPin()
{
  if(m_listRenderSinkPins.empty())
  {
    ThrowFatalError(std::runtime_error("Could not find a MIDI render pin"));
  }

  CKsMidiPin* pPin = (CKsMidiPin*)m_listRenderSinkPins[0];
  assert(pPin != NULL);

  HRESULT hr = pPin->Instantiate(false);

  if(FAILED(hr))
  {
    ThrowFatalError(ComException("CKsMidiCapFilter::CreateRenderPin: Pin instanciation failed", hr));
  }

  return pPin;
}

CKsMidiCapFilter::CKsMidiCapFilter(tstring const& sPath, std::string const& sFriendlyName) :
  CKsMidiFilter(sPath, sFriendlyName)
{
}

CKsMidiPin* CKsMidiCapFilter::CreateCapturePin()
{
  if(m_listCaptureSinkPins.empty())
  {
    ThrowFatalError(std::runtime_error("Could not find a MIDI capture pin"));
  }

  CKsMidiPin* pPin = (CKsMidiPin*)m_listCaptureSinkPins[0];
  assert(pPin != NULL);

  HRESULT hr = pPin->Instantiate(false);

  if(FAILED(hr))
  {
    ThrowFatalError(ComException("CKsMidiCapFilter::CreateCapturePin: Pin instanciation failed", hr));
  }

  return pPin;
}

