/**
 * @file ksmidipin.cpp
 * @author Sebastien Alaiwan
 * @date 2011-11-20
 */

/*
 * Copyright (C) 2011 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "ksmidipin.h"
#include "util.h"
#include <stdexcept>

CKsMidiPin::CKsMidiPin(CKsFilter* pFilter, ULONG nId) :
  CKsPin(pFilter, nId)
{
  m_cbPinCreateSize = sizeof(KSPIN_CONNECT) + sizeof(KSDATAFORMAT);
  m_pksPinCreate = (PKSPIN_CONNECT) new BYTE[m_cbPinCreateSize];

  m_pksPinCreate->Interface.Set = KSINTERFACESETID_Standard;
  m_pksPinCreate->Interface.Id = KSINTERFACE_STANDARD_STREAMING;
  m_pksPinCreate->Interface.Flags = 0;
  m_pksPinCreate->Medium.Set = KSMEDIUMSETID_Standard;
  m_pksPinCreate->Medium.Id = KSMEDIUM_TYPE_ANYINSTANCE;
  m_pksPinCreate->Medium.Flags = 0;
  m_pksPinCreate->PinId = nId;
  m_pksPinCreate->PinToHandle = NULL;
  m_pksPinCreate->Priority.PriorityClass = KSPRIORITY_NORMAL;
  m_pksPinCreate->Priority.PrioritySubClass = 1;

  // point m_pDataFormat to just after the pConnect struct
  KSDATAFORMAT* m_pDataFormat = (KSDATAFORMAT*)(m_pksPinCreate + 1);
  m_pDataFormat->FormatSize = sizeof(KSDATAFORMAT);
  m_pDataFormat->Flags = 0;
  m_pDataFormat->SampleSize = 0;
  m_pDataFormat->Reserved = 0;
  m_pDataFormat->MajorFormat = GUID(KSDATAFORMAT_TYPE_MUSIC);
  m_pDataFormat->SubFormat = GUID(KSDATAFORMAT_SUBTYPE_MIDI);
  m_pDataFormat->Specifier = GUID(KSDATAFORMAT_SPECIFIER_NONE);

  HRESULT hr = CKsMidiPin::Init();

  if(FAILED(hr))
  {
    throw std::runtime_error("CKsMIdiPin::Init failed");
  }
}

HRESULT CKsMidiPin::Init()
{
  {
    // Make sure at least one interface is standard streaming
    bool hasStdStreamingInterface = false;

    for(ULONG i = 0; i < m_Descriptor.cInterfaces; i++)
    {
      if(!IsEqualGUIDAligned(m_Descriptor.pInterfaces[i].Set, KSINTERFACESETID_Standard))
      {
        continue;
      }

      if(m_Descriptor.pInterfaces[i].Id != KSINTERFACE_STANDARD_STREAMING)
      {
        continue;
      }

      hasStdStreamingInterface = true;
      break;
    }

    if(!hasStdStreamingInterface)
    {
      DebugPrintf(TRACE_ERROR, TEXT("No standard streaming interfaces on the pin"));
      return E_FAIL;
    }
  }

  {
    // Make sure at least one medium is standard streaming
    bool hasStdStreamingMedium = false;

    for(ULONG i = 0; i < m_Descriptor.cInterfaces; ++i)
    {
      if(!IsEqualGUIDAligned(m_Descriptor.pMediums[i].Set, KSMEDIUMSETID_Standard))
      {
        continue;
      }

      if(m_Descriptor.pMediums[i].Id != KSMEDIUM_STANDARD_DEVIO)
      {
        continue;
      }

      hasStdStreamingMedium = true;
      break;
    }

    if(!hasStdStreamingMedium)
    {
      DebugPrintf(TRACE_ERROR, TEXT("No standard streaming mediums on the pin"));
      return E_FAIL;
    }
  }

  {
    // Make sure at least one datarange supports MIDI
    bool hasMidiDataRange = false;
    BYTE const* pDataRangePtr = reinterpret_cast<BYTE const*>(m_Descriptor.pDataRanges);

    for(ULONG i = 0; i < m_Descriptor.cDataRanges; ++i)
    {
      KSDATARANGE const* pDataRange = reinterpret_cast<KSDATARANGE const*>(pDataRangePtr);

      if(IsEqualGUIDAligned(pDataRange->SubFormat, KSDATAFORMAT_SUBTYPE_MIDI))
      {
        hasMidiDataRange = true;
        break;
      }

      pDataRangePtr += pDataRange->FormatSize;
    }

    if(!hasMidiDataRange)
    {
      DebugPrintf(TRACE_ERROR, TEXT("No MIDI dataranges on the pin"));
      return E_FAIL;
    }
  }

  return S_OK;
}

