/**
 * @file kspin.cpp
 * @author Sebastien Alaiwan
 * @date 2011-11-20
 */

/*
 * Copyright (C) 2011 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "kspin.h"
#include "comutils.h"
#include "util.h"
#include <stdexcept>
#include <cassert>

CKsPin::CKsPin(CKsFilter* pFilter, ULONG PinId) :
  CKsObject(INVALID_HANDLE_VALUE),
  m_dwAlignment(0),
  m_pksPinCreate(NULL),
  m_cbPinCreateSize(0),
  m_nId(PinId),
  m_pFilter(pFilter),
  m_guidCategory(GUID_NULL),
  m_ksState(KSSTATE_STOP),
  m_fLooped(false),
  m_pLinkPin(NULL),
  m_fCompleteOnlyOnRunState(true)
{
  Zero(m_Descriptor);

  HRESULT hr = CKsPin::Init();

  if(FAILED(hr))
  {
    ThrowFatalError(ComException("CKsPin::CKsPin", hr));
  }
}

CKsPin::~CKsPin()
{
  ClosePin();

  delete[] (BYTE*)m_pksPinCreate;
  delete[] (BYTE*)m_Descriptor.pmiDataRanges;
  delete[] (BYTE*)m_Descriptor.pmiInterfaces;
  delete[] (BYTE*)m_Descriptor.pmiMediums;
}

void CKsPin::ClosePin()
{
  if(IsValidHandle(m_handle))
  {
    SetState(KSSTATE_STOP);
    SafeCloseHandle(m_handle);
  }
}

void CKsPin::SetState(KSSTATE ksState)
{
  HRESULT hr = SetProperty(
    KSPROPSETID_Connection,
    KSPROPERTY_CONNECTION_STATE,
    &ksState,
    sizeof(ksState));

  if(FAILED(hr))
  {
    ThrowFatalError(ComException("Can't set pin state", hr));
  }

  m_ksState = ksState;
}

HRESULT CKsPin::Instantiate(bool fLooped)
{
  if(!m_pksPinCreate)
  {
    return S_FALSE;
  }

  if(m_pLinkPin)
  {
    assert(KSPIN_COMMUNICATION_SINK == m_pLinkPin->m_Descriptor.Communication ||
           KSPIN_COMMUNICATION_BOTH == m_pLinkPin->m_Descriptor.Communication);

    m_pksPinCreate->PinToHandle = m_pLinkPin->m_handle;
  }

  m_fLooped = fLooped;
  m_pksPinCreate->Interface.Id =
    m_fLooped ? KSINTERFACE_STANDARD_LOOPED_STREAMING : KSINTERFACE_STANDARD_STREAMING;

  DWORD dwResult = KsCreatePin(m_pFilter->m_handle, m_pksPinCreate, GENERIC_WRITE | GENERIC_READ, &m_handle);

  if(ERROR_SUCCESS != dwResult)
  {
    HRESULT hr = HRESULT_FROM_WIN32(dwResult);

    if(SUCCEEDED(hr))
    {
      hr = E_FAIL;
    }

    DebugPrintf(TRACE_ERROR, TEXT("Failed to instantiate pin. hr=0x%08x"), hr);
    return hr;
  }

  return S_OK;
}

void CKsPin::GetState(KSSTATE& ksState)
{
  if(!IsValidHandle(m_handle))
  {
    throw std::runtime_error("CKsPin::GetState: invalid handle");
  }

  HRESULT hr = GetProperty(KSPROPSETID_Connection,
                           KSPROPERTY_CONNECTION_STATE,
                           &ksState,
                           sizeof(ksState));

  if(FAILED(hr))
  {
    ThrowFatalError(ComException("Can't get connection state", hr));
  }
}

// Submit some data to the pin, using the provided KSSTREAM_HEADER and OVERLAPPED structures.
// If the caller submits a valid event in the KSSTREAM_HEADER, the event must be unsignaled.
//
// Arguments:
// Pointer to the KSSTREAM_HEADER structure describing the data to send to the pin
// Pointer to the OVERLAPPED structure to use when doing the asynchronous I/O
void CKsPin::WriteData(KSSTREAM_HEADER* pKSSTREAM_HEADER, OVERLAPPED* pOVERLAPPED)
{
  if(pOVERLAPPED == NULL)
    ThrowFatalError(ComException("CKsPin::WriteData: pOVERLAPPED cannot be NULL", E_POINTER));
  DWORD cbReturned;
  BOOL fRes = DeviceIoControl(
    m_handle,
    IOCTL_KS_WRITE_STREAM,
    NULL,
    0,
    pKSSTREAM_HEADER,
    pKSSTREAM_HEADER->Size,
    NULL,
    pOVERLAPPED);

  if(!fRes)
  {
    DWORD const dwError = GetLastError();

    if(ERROR_IO_PENDING != dwError)
    {
      ThrowFatalError(ComException("CKsPin::WriteData: DeviceIoControl failed", HRESULT_FROM_WIN32(dwError)));
    }
  }

  // Wait for operation to complete.
  fRes = ::GetOverlappedResult(m_handle, pOVERLAPPED, &cbReturned, TRUE);
  if(!fRes)
  {
    DWORD const dwError = GetLastError();
    ThrowFatalError(ComException("CKsPin::WriteData: GetOverlappedResult failed", HRESULT_FROM_WIN32(dwError)));
  }
}

// Submit some memory for the pin to read into, using the provided KSSTREAM_HEADER and OVERLAPPED structures.
// If the caller submits a valid event in the KSSTREAM_HEADER, the event must be unsignaled.
//
// Arguments:
// Pointer to the KSSTREAM_HEADER structure describing where to store the data read from the pin
// Pointer to the OVERLAPPED structure to use when doing the asynchronous I/O
void CKsPin::ReadData(KSSTREAM_HEADER* pKSSTREAM_HEADER, OVERLAPPED* pOVERLAPPED)
{
  DWORD cbReturned;
  BOOL fRes = DeviceIoControl(
    m_handle,
    IOCTL_KS_READ_STREAM,
    NULL,
    0,
    pKSSTREAM_HEADER,
    pKSSTREAM_HEADER->Size,
    &cbReturned,
    pOVERLAPPED);

  if(!fRes)
  {
    DWORD const dwError = GetLastError();

    if(ERROR_IO_PENDING != dwError)
    {
      if(ERROR_IO_PENDING != dwError)
      {
        ThrowFatalError(ComException("CKsPin::ReadData: DeviceIoControl failed", HRESULT_FROM_WIN32(dwError)));
      }
    }
  }
}

HRESULT CKsPin::Init()
{
  HRESULT hr = S_OK;
  bool fFailure = false;

  // Get COMMUNICATION
  hr = m_pFilter->GetPinProperty(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_COMMUNICATION,
    m_Descriptor.Communication);

  if(FAILED(hr))
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_COMMUNICATION on PinID: %d"), m_nId);
    fFailure = true;
  }

  // get PKSPIN_INTERFACES
  hr = m_pFilter->GetPinPropertyMulti(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_INTERFACES,
    &m_Descriptor.pmiInterfaces);

  if(SUCCEEDED(hr))
  {
    m_Descriptor.cInterfaces = m_Descriptor.pmiInterfaces->Count;
    m_Descriptor.pInterfaces = (PKSPIN_INTERFACE)(m_Descriptor.pmiInterfaces + 1);
  }
  else
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPETY_PIN_INTERFACES on PinID: %d"), m_nId);
    fFailure = true;
  }

  // get PKSPIN_MEDIUMS
  hr = m_pFilter->GetPinPropertyMulti(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_MEDIUMS,
    &m_Descriptor.pmiMediums);

  if(SUCCEEDED(hr))
  {
    m_Descriptor.cMediums = m_Descriptor.pmiMediums->Count;
    m_Descriptor.pMediums = (PKSPIN_MEDIUM)(m_Descriptor.pmiMediums + 1);
  }
  else
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_MEDIUMS on PinID: %d"), m_nId);
    fFailure = true;
  }

  // get PKSPIN_DATARANGES
  hr = m_pFilter->GetPinPropertyMulti(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_DATARANGES,
    &m_Descriptor.pmiDataRanges);

  if(SUCCEEDED(hr))
  {
    m_Descriptor.cDataRanges = m_Descriptor.pmiDataRanges->Count;
    m_Descriptor.pDataRanges = (PKSDATARANGE)(m_Descriptor.pmiDataRanges + 1);
  }
  else
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_DATARANGES on PinID: %d"), m_nId);
    fFailure = true;
  }

  // get dataflow information
  hr = m_pFilter->GetPinProperty(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_DATAFLOW,
    m_Descriptor.DataFlow);

  if(FAILED(hr))
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_DATAFLOW on PinID: %d"), m_nId);
    fFailure = true;
  }

  // get instance information
  hr = m_pFilter->GetPinProperty(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_CINSTANCES,
    m_Descriptor.CInstances);

  if(FAILED(hr))
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_CINSTANCES on PinID: %d"), m_nId);
    fFailure = true;
  }

  // Get Global instance information
  hr = m_pFilter->GetPinProperty(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_GLOBALCINSTANCES,
    m_Descriptor.CInstancesGlobal);

  if(FAILED(hr))
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_GLOBALCINSTANCES on PinID: %d"), m_nId);
    fFailure = true;
  }

  // Get Pin Category Information
  hr = m_pFilter->GetPinProperty(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_CATEGORY,
    m_guidCategory);

  if(FAILED(hr))
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_CATEGORY on PinID: %d"), m_nId);
    fFailure = true;
  }

  // Get Pin Name
  TCHAR name[MAX_PATH];
  hr = m_pFilter->GetPinProperty(
    m_nId,
    KSPROPSETID_Pin,
    KSPROPERTY_PIN_NAME,
    name);
  m_szFriendlyName = name;

  if(FAILED(hr))
  {
    DebugPrintf(TRACE_HIGH, TEXT("Failed to retrieve pin property KSPROPERTY_PIN_NAME on PinID: %d"), m_nId);
    fFailure = true;
  }

  // if we experienced any failures, return S_FALSE, otherwise S_OK
  hr = fFailure ? S_FALSE : S_OK;

  return hr;
}

ULONG CKsPin::GetId() const
{
  return m_nId;
}

HRESULT CKsPin::SetPinConnect(KSPIN_CONNECT const* pksPinConnect, ULONG cbKsPinConnect)
{
  if(pksPinConnect == NULL)
  {
    return E_POINTER;
  }

  if(cbKsPinConnect < sizeof(KSPIN_CONNECT))
  {
    return E_INVALIDARG;
  }

  // Free the existing Pin Connect structure
  delete[] m_pksPinCreate;
  m_pksPinCreate = NULL;

  // Allocate memory for the new pin connect structure
  m_pksPinCreate = reinterpret_cast<PKSPIN_CONNECT>(new BYTE[cbKsPinConnect]);

  CopyMemory(m_pksPinCreate, pksPinConnect, cbKsPinConnect);

  return S_OK;
}

HRESULT CKsPin::SetFormat(KSDATAFORMAT* pFormat)
{
  HRESULT hr = S_OK;

  hr = SetProperty(
    KSPROPSETID_Connection,
    KSPROPERTY_CONNECTION_DATAFORMAT,
    pFormat,
    pFormat->FormatSize);

  return hr;
}

KSPIN_DATAFLOW CKsPin::GetDataFlow() const
{
  return m_Descriptor.DataFlow;
}

KSPIN_COMMUNICATION CKsPin::GetCommunication() const
{
  return m_Descriptor.Communication;
}

