// Definition of an object representing a Virus Synth.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <cassert>
#include <vector>
#include <memory.h>
#include <fstream>

#include "midivirus.h"

auto static const VIRUS_CMD_SINGLE_DUMP = 0x10;
auto static const VIRUS_CMD_MULTI_DUMP = 0x11;
auto static const VIRUS_CMD_SINGLE_REQ = 0x30;
auto static const VIRUS_CMD_MULTI_REQ = 0x31;
auto static const VIRUS_CMD_SINGLE_BANK_REQ = 0x32;
auto static const VIRUS_CMD_ARRANGEMENT_REQ = 0x34;
auto static const VIRUS_CMD_GLOBAL_REQ = 0x36;
auto static const VIRUS_CMD_TOTAL_REQ = 0x36;
auto static const VIRUS_CMD_CONTROLLER_DUMP_REQ = 0x37;
auto static const VIRUS_CMD_REMOTE_REQ = 0x38;
auto static const VIRUS_CMD_PAGE_C_CTRL_CHANGE = 0x6E;
auto static const VIRUS_CMD_PAGE_D_CTRL_CHANGE = 0x6F;
auto static const VIRUS_CMD_PAGE_A_CTRL_CHANGE = 0x70;
auto static const VIRUS_CMD_PAGE_B_CTRL_CHANGE = 0x71;
auto static const VIRUS_CMD_MULTI_CTRL_CHANGE = 0x72;
auto static const VIRUS_CMD_SYSTEM = 0x73;

static const uint8_t sysex_goto_seq_mode[] =
{
  0xF0, 0x00, 0x20, 0x33, 0x01, 0x10, 0x73, 0x00, 0x7A, 0x01, 0xF7
};

static const uint8_t sysex_req_dump[] =
{
  0xF0, 0x00, 0x20, 0x33, 0x01, 0x10, 0x34, 0x00, 0x00, 0xF7
};

static const uint8_t sysex_req_singledump[] =
{
  0xF0, 0x00, 0x20, 0x33, 0x01, 0x10, 0x30, 0x00, 0x00, 0xF7
};

CMidiVirus::CMidiVirus(IMidiSink* pSink)
{
  m_pOutput = pSink;
  m_pSource = nullptr;
  m_State = STATE_INIT;
}

void CMidiVirus::Init(IMidiSource* pSource, IMidiSink* pSink)
{
  m_pOutput = pSink;
  m_pSource = pSource;

  GoToSequencerMode();
}

void CMidiVirus::RequestSingleDump(int iPart)
{
  std::vector<uint8_t> msg(sizeof sysex_req_singledump);
  memcpy(&msg[0], sysex_req_singledump, sizeof sysex_req_singledump);
  msg[8] = iPart;
  SendSysex(&msg[0], msg.size());
}

bool CMidiVirus::DownloadPatch(int iPartNumber, VirusPatch& patch)
{
  RequestSingleDump(iPartNumber);
  m_pSource->Flush();
  std::vector<uint8_t> sysex;

  while(1)
  {
    if(!m_pSource->PullSysex(sysex))
      return false;

    // skip sysex that are not patches.
    if(sysex.size() > 100)
      break;
  }

  if(sysex.empty())
    return false;

  patch.Parse(&sysex[0], sysex.size());
  return true;
}

void CMidiVirus::GoToSequencerMode()
{
  SendSysex(sysex_goto_seq_mode, sizeof sysex_goto_seq_mode);
}

bool CMidiVirus::UploadPatch(int iPartNumber, VirusPatch const& patch)
{
  std::vector<uint8_t> buffer = patch.m_vSysex;

  // sanity check
  if(buffer.size() < 9)
    return false;

  buffer[8] = iPartNumber;
  SendSysex(&buffer[0], buffer.size());
  return true;
}

void CMidiVirus::SendSysex(uint8_t const* pSysex, int iLength)
{
  assert(m_pOutput);
  m_pOutput->WriteMidi(pSysex, iLength);
}

void CMidiVirus::ProcessMidi(uint8_t const* midiBuffer, size_t const iLength)
{
  assert(m_pOutput);

  for(size_t j = 0; j < iLength; ++j)
  {
    uint8_t midi_byte = midiBuffer[j];
    switch(m_State)
    {
    case STATE_INIT:
      {
        m_CurrentSysex.clear();
        m_CurrentData.clear();

        if(midi_byte == 0xF0)
        {
          m_State = STATE_SYSEX;
          m_CurrentSysex.push_back(midi_byte);
        }
        else if((midi_byte & 0xF0) == 0xB0) // control change
        {
          m_State = STATE_CC;
          m_StateChannel = midi_byte & 0x0F;
        }

        break;
      }
    case STATE_CC:
      {
        m_CurrentData.push_back(midi_byte);

        if(m_CurrentData.size() == 2)
        {
          m_State = STATE_INIT;
          int iOffset = m_CurrentData[0];
          int iValue = m_CurrentData[1];
          m_pOutput->OnVirusParamChanged(m_StateChannel, iOffset, iValue);
        }

        break;
      }
    case STATE_SYSEX:
      {
        m_CurrentSysex.push_back(midi_byte);

        if(midi_byte == 0xF7)
        {
          ProcessSysex(m_CurrentSysex);

          m_State = STATE_INIT;
          m_CurrentSysex.clear();
        }

        break;
      }
    }
  }
}

void CMidiVirus::ProcessSysex(std::vector<uint8_t> const& sysex)
{
  if(sysex.size() < 6)
    return;

  // check that the sysex begins with 'F0 00 20 33 01 10'
  if(sysex[0] != 0xF0)
    return;

  if(sysex[1] != 0x00)
    return;

  if(sysex[2] != 0x20)
    return;

  if(sysex[3] != 0x33)
    return;

  if(sysex[4] != 0x01)
    return;

  if(sysex[5] != 0x10)
    return;

  auto const cmd = sysex[6];

  if(cmd == VIRUS_CMD_SINGLE_DUMP)
  {
    int iPart = sysex[8];
    VirusPatch patch;
    patch.Parse(&sysex[0], sysex.size());
    m_pOutput->OnPatchReceived(iPart, patch);
  }
  else if(cmd == VIRUS_CMD_PAGE_A_CTRL_CHANGE)
  {
    int iPart = sysex[7];
    int iIndex = sysex[8];
    int iValue = sysex[9];
    m_pOutput->OnVirusParamChanged(iPart, iIndex, iValue);
  }
  else if(cmd == VIRUS_CMD_PAGE_B_CTRL_CHANGE)
  {
    int iPart = sysex[7];
    int iIndex = sysex[8];
    int iValue = sysex[9];
    m_pOutput->OnVirusParamChanged(iPart, 128 + iIndex, iValue);
  }
  else if(cmd == VIRUS_CMD_PAGE_C_CTRL_CHANGE)
  {
    int iPart = sysex[7];
    int iIndex = sysex[8];
    int iValue = sysex[9];
    m_pOutput->OnVirusParamChanged(iPart, 256 + iIndex, iValue);
  }
}

