/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */
#pragma once

#include <cmath> // floor
#include "math_tools.h"

template<typename ParamSetType>
struct OpaqueParamDesc
{
  virtual void SetDefault(ParamSetType& values) const = 0;
  virtual void Setf(ParamSetType& values, float fValue) const = 0;
  virtual float Getf(ParamSetType const& values) const = 0;
  virtual std::string GetName() const = 0;
  virtual std::string GetStringValue(ParamSetType const& values) const = 0;
};

template<typename ParamSetType, typename ValueType>
struct ParamDesc : public OpaqueParamDesc<ParamSetType>
{
  ParamDesc(char const* _sName, float _fMin, float _fMax, ValueType ParamSetType::* pValue)
    : sName(_sName),
    value(pValue),
    fMin(_fMin),
    fMax(_fMax)
  {
  }

  virtual std::string GetName() const
  {
    return sName;
  }

  virtual std::string GetStringValue(ParamSetType const& values) const
  {
    return internalToString(values.*value);
  }

  virtual void SetDefault(ParamSetType& values) const
  {
    values.*value = fMin;
  }

  virtual void Setf(ParamSetType& values, float fValue) const
  {
    internalFrom(values.*value, fValue);
  }

  virtual float Getf(ParamSetType const& values) const
  {
    float fValue;
    internalTo(values.*value, fValue);
    return fValue;
  }

private:
  char const* const sName;
  ValueType ParamSetType::* const value;
  float fMin, fMax;

  // float

  void internalFrom(float& realValue, float vstValue) const
  {
    float const fRange = fMax - fMin;
    realValue = vstValue * fRange + fMin;
    realValue = std::max(fMin, std::min(realValue, fMax));
  }

  void internalTo(float realValue, float& vstValue) const
  {
    float const fRange = fMax - fMin;
    realValue = std::max(fMin, std::min(realValue, fMax));
    vstValue = (realValue - fMin) / fRange;
  }

  static std::string internalToString(float fValue)
  {
    std::stringstream ss;
    ss << fValue;
    return ss.str();
  }
};

template<typename ParamSetType>
struct BoolParamDesc : public OpaqueParamDesc<ParamSetType>
{
  BoolParamDesc(char const* _sName, bool ParamSetType::* _pValue)
    : sName(_sName), pValue(_pValue)
  {
  }

  virtual std::string GetName() const
  {
    return sName;
  }

  virtual std::string GetStringValue(ParamSetType const& values) const
  {
    auto const currValue = values.*pValue;
    return currValue >= 0.5 ? "yes" : "no";
  }

  virtual void SetDefault(ParamSetType& values) const
  {
    values.*pValue = false;
  }

  virtual void Setf(ParamSetType& values, float fValue) const
  {
    if(fValue >= 0.5)
      values.*pValue = true;
    else
      values.*pValue = false;
  }

  virtual float Getf(ParamSetType const& values) const
  {
    auto const currValue = values.*pValue;
    return currValue >= 0.5 ? 1.0f : 0.0f;
  }

private:
  char const* const sName;
  bool ParamSetType::* const pValue;
};

template<typename ParamSetType, typename EnumType>
struct EnumParamDesc : public OpaqueParamDesc<ParamSetType>
{
  template<int N>
  EnumParamDesc(char const* _sName, EnumType const (&allowedValues)[N], EnumType ParamSetType::* pValue)
    : sName(_sName), pAllowedValues(allowedValues), iNumValues(N),
    value(pValue)
  {
  }

  virtual std::string GetName() const
  {
    return sName;
  }

  virtual std::string GetStringValue(ParamSetType const& values) const
  {
    return internalToString(values.*value);
  }

  virtual void SetDefault(ParamSetType& values) const
  {
    values.*value = pAllowedValues[0];
  }

  virtual void Setf(ParamSetType& values, float fValue) const
  {
    auto const index = int(floor(fValue * float(iNumValues + 0.5)));
    auto const clampedIndex = std::max(0, std::min(iNumValues - 1, index));
    values.*value = pAllowedValues[clampedIndex];
  }

  virtual float Getf(ParamSetType const& values) const
  {
    // search for the EnumValue index. Not very efficient...
    for(int i = 0; i < iNumValues; ++i)
    {
      if((values.*value) == pAllowedValues[i])
      {
        float vstValue = float(i) / float(iNumValues);
        return vstValue;
      }
    }

    std::cout << "Getf: VALUE NOT FOUND!" << std::endl;
    return -1.0f;
  }

private:
  char const* const sName;
  EnumType const* const pAllowedValues;
  int const iNumValues;
  EnumType ParamSetType::* const value;
};

inline static std::string internalToString(Fraction const& fraction)
{
  std::stringstream ss;
  ss << fraction.num << "/" << fraction.denom;
  return ss.str();
}

