/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#pragma once

#include <vector>
#include <string>
#include <sstream>
#include <memory>
#include <memory.h>
#include <iostream>
#include <stdexcept>

// -----------------------------------------------------------------------------
// Array operations
// -----------------------------------------------------------------------------

template<size_t N, typename T>
inline constexpr int ArrayLength(T const (&a)[N])
{
  (void)a;
  return (int)N;
}

template<typename TT, int NN>
inline void ArrayCopy(TT(&dstTab)[NN], TT const (&srcTab)[NN])
{
  for(size_t i = 0; i < NN; ++i)
    dstTab[i] = srcTab[i];
}

template<typename TT, int NN>
inline void ArraySet(TT(&dstTab)[NN], TT const& val)
{
  for(size_t i = 0; i < NN; ++i)
    dstTab[i] = val;
}

// -----------------------------------------------------------------------------
// Vector operations
// -----------------------------------------------------------------------------

template<typename T>
inline void AddToVector(std::vector<T>& v, const T& s)
{
  v.push_back(s);
}

template<typename T, typename... Arguments>
inline void AddToVector(std::vector<T>& v, const T& s, Arguments... values)
{
  v.push_back(s);
  AddToVector(v, values...);
}

template<typename T, typename... Arguments>
std::vector<T> MakeVector(const T& firstArg, Arguments... values)
{
  std::vector<T> r;
  AddToVector(r, firstArg, values...);
  return r;
}

inline std::string S(const char* s)
{
  return s;
}

template<typename T>
void ZeroStruct(T& s)
{
  memset(&s, 0, sizeof s);
}

// ------------------------------------------------------------------------------
// format implementation
// ------------------------------------------------------------------------------

template<typename T>
std::string ToString(T const& val)
{
  std::stringstream ss;
  ss << val;
  return ss.str();
}

template<>
inline std::string ToString(std::string const& val)
{
  std::stringstream ss;
  ss << "\"";
  ss << val;
  ss << "\"";
  return ss.str();
}

template<typename T>
std::string ToString(std::vector<T> const& v)
{
  std::stringstream ss;
  ss << "[";
  bool first = true;

  for(auto& val :  v)
  {
    if(!first)
      ss << ", ";

    first = false;

    ss << ToString(val);
  }

  ss << "]";
  return ss.str();
}

inline std::string format(const std::string& format)
{
  return format;
}

template<typename T, typename... Arguments>
std::string format(const std::string& fmt, const T& firstArg, Arguments... args)
{
  std::string r;
  size_t i = 0;

  while(i < fmt.size())
  {
    if(fmt[i] == '%')
    {
      ++i;

      if(i >= fmt.size())
        throw std::runtime_error("Invalid fmt specifier");

      if(fmt[i] == '%')
        r += '%';
      else if(fmt[i] == 's')
      {
        r += ToString(firstArg);
        return r + format(fmt.substr(i + 1), args...);
      }
    }
    else
    {
      r += fmt[i];
    }

    ++i;
  }

  return r;
}

// ------------------------------------------------------------------------------
// writefln implementation
// ------------------------------------------------------------------------------

template<typename... Arguments>
void writefln(const std::string& fmt, Arguments... args)
{
  std::cout << format(fmt, args...) << std::endl;
}

// ------------------------------------------------------------------------------
// Main wrapper
// ------------------------------------------------------------------------------

inline int runMain(int (* mainFunction)(int argc, char* argv[]), int argc, char* argv[])
{
  try
  {
    return mainFunction(argc, argv);
  }
  catch(const std::exception& e)
  {
    std::cerr << "Fatal error: " << e.what() << std::endl;
    return 255;
  }
}

