// Definition of an object representing a single patch.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <fstream>
#include <memory.h>
#include "viruspatch.h"

VirusPatch::VirusPatch()
{
  m_iPartNumber = -1;
}

void VirusPatch::Parse(std::vector<uint8_t> const& sysex)
{
  if(sysex.empty())
    return;

  Parse(&sysex[0], sysex.size());
}

void VirusPatch::Parse(uint8_t const* sSysex, unsigned int iLength)
{
  if(iLength == 0)
  {
    m_sName = "<none>";
    return;
  }

  m_vSysex.resize(iLength);

  memcpy(&m_vSysex[0], sSysex, iLength);

  if(iLength >= 524)
  {
    m_iPartNumber = sSysex[8];

    char name[16];
    memset(name, 0, 16);
    memcpy(name, &sSysex[0xF9], 10);
    m_sName = name;
  }
}

bool VirusPatch::Export(std::string const& sFilename) const
{
  std::ofstream fp(sFilename.c_str(), std::ios::binary);

  if(!fp.is_open())
    return false;

  fp.write((char*)&m_vSysex[0], m_vSysex.size());

  return true;
}

bool VirusPatch::ExportMidiFile(std::string const& sFilename, VirusPatch const* patches, int iNumPatches)
{
  std::ofstream fp(sFilename.c_str(), std::ios::binary);

  if(!fp.is_open())
    return false;

  // Write MIDI header.
  fp.write("MThd", 4);

  // Write chunk size.
  fp.put(0x00);
  fp.put(0x00);
  fp.put(0x00);
  fp.put(0x06);

  // Write file format : 0 for single track file format.
  fp.put(0x00);
  fp.put(0x00);

  // Write file format : 1 for one track chunk following the header chunk.
  uint8_t cNumTracks[2] =
  {
    0x00, 0x01
  };
  fp.write((char*)cNumTracks, sizeof cNumTracks);

  // Write file format : 96 divisions (0x0060). This is arbitrary.
  uint8_t cDivisions[2] =
  {
    0x00, 0x60
  };
  fp.write((char*)cDivisions, sizeof cDivisions);

  // Write track header.
  fp.write("MTrk", 4);

  int iLength = 0;

  for(int i = 0; i < iNumPatches; ++i)
  {
    if(patches[i].m_vSysex.empty())
      continue;

    iLength++; // delta time event.
    iLength += patches[i].m_vSysex.size() + 2;
  }

  iLength++; // delta time event.
  iLength += 3; // count for the end-of-track meta event.

  // Write track length.
  fp.put((iLength >> 24) & 0xFF);
  fp.put((iLength >> 16) & 0xFF);
  fp.put((iLength >> 8) & 0xFF);
  fp.put((iLength >> 0) & 0xFF);

  for(int i = 0; i < iNumPatches; ++i)
  {
    // skip empty patches.
    if(patches[i].m_vSysex.empty())
      continue;

    // 0 delta event.
    fp.put(0x70);

    // Write magic string.
    uint8_t cMagicString[] =
    {
      0xF0, 0x84, 0x0B
    };
    fp.write((char*)cMagicString, sizeof cMagicString);

    // Write sysex.
    fp.write((char*)&patches[i].m_vSysex[1], patches[i].m_vSysex.size() - 1);
  }

  // 0 delta event.
  fp.put(0x70);

  // Write end-of-track meta-event.
  uint8_t cEndOfTrack[3] =
  {
    0xFF, 0x2F, 0x00
  };
  fp.write((char*)cEndOfTrack, sizeof cEndOfTrack);

  return true;
}

bool VirusPatch::ExportMidiFile(std::string const& sFilename) const
{
  std::ofstream fp(sFilename.c_str(), std::ios::binary);

  if(!fp.is_open())
    return false;

  // Write MIDI header.
  fp.write("MThd", 4);

  // Write chunk size.
  fp.put(0x00);
  fp.put(0x00);
  fp.put(0x00);
  fp.put(0x06);

  // Write file format : 0 for single track file format.
  fp.put(0x00);
  fp.put(0x00);

  // Write file format : 1 for one track chunk following the header chunk.
  uint8_t cNumTracks[2] =
  {
    0x00, 0x01
  };
  fp.write((char*)cNumTracks, sizeof cNumTracks);

  // Write file format : 96 divisions (0x0060). This is arbitrary.
  uint8_t cDivisions[2] =
  {
    0x00, 0x60
  };
  fp.write((char*)cDivisions, sizeof cDivisions);

  // Write track header.
  fp.write("MTrk", 4);

  int iLength = 1 + m_vSysex.size() + 3 + 2;

  // Write track length.
  fp.put((iLength >> 24) & 0xFF);
  fp.put((iLength >> 16) & 0xFF);
  fp.put((iLength >> 8) & 0xFF);
  fp.put((iLength >> 0) & 0xFF);

  // 0 delta event.
  fp.put(0x78);

  // Write magic string.
  uint8_t cMagicString[] =
  {
    0xF0, 0x84, 0x0B
  };
  fp.write((char*)cMagicString, sizeof cMagicString);

  // Write sysex.
  fp.write((char*)&m_vSysex[1], m_vSysex.size() - 1);

  // Write end-of-track meta-event.
  uint8_t cEndOfTrack[3] =
  {
    0xFF, 0x2F, 0x00
  };
  fp.write((char*)cEndOfTrack, sizeof cEndOfTrack);

  return true;
}

bool VirusPatch::Import(std::string const& sFilename)
{
  std::ifstream fp(sFilename.c_str(), std::ios::binary);

  if(!fp.is_open())
    return false;

  std::vector<uint8_t> my_buff(32768);
  fp.read((char*)&my_buff[0], my_buff.size());
  std::streamsize length = fp.gcount();
  Parse(&my_buff[0], length);

  return true;
}

std::string VirusPatch::GetName() const
{
  return m_sName;
}

