/**
 * @file acecrusher.cpp
 * @brief Yet another bit crusher / rate reducer plugin.
 * @author Sebastien Alaiwan
 * @date 2010-12-01
 */

/* Copyright (C)
 * 2010 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#define USE_MATH_DEFINES
#include <iomanip>
#include <sstream>
#include <cmath>
#include <memory.h>

#include "lib_utils/math_tools.h"
#include "lib_utils/debug.h"
#include "lib_utils/string_format.h"
#include "lib_utils/tools.h"

#include "acecrusher.h"

struct PARAMETER_DESC
{
  int index;
  char const* sName;
  float fMin, fMax;
  string (* ToString)(float fValue);
};

static string ToStringNumber(float fValue)
{
  return format("%.4f", fValue);
}

static string ToStringNumberHeadroom(float fValue)
{
  return ToStringNumber(-30.0f * (1.0f - fValue));
}

PARAMETER_DESC const g_ParamDesc[] =
{
  { PARAM_CLIP, "Clip", 0.0f, 1.0f, &ToStringNumberHeadroom },
  { PARAM_BITS, "Quant", 0.0f, 1.0f, &ToStringNumber },
  { PARAM_RATE, "Rate", 0.235f, 0.5f, &ToStringNumber },
  { PARAM_POSTFILT, "PostFilt", 0.0f, 1.0f, &ToStringNumber },
  { PARAM_NONLIN, "Non-Lin", 0.0f, 1.0f, &ToStringNumber },
};

auto const NUM_PARAM_DESC = ArrayLength(g_ParamDesc);

PARAMETER_DESC const* GetParameterDesc(int index)
{
  if(index < 0 || index >= NUM_PARAM_DESC)
    return nullptr;

  return &g_ParamDesc[index];
}

AceCrusherPlugin::AceCrusherPlugin(EVEREST::IHost* pHost) :
  m_pHost(pHost)
{
  ArraySet(m_ParamValues, 0.0f);

  m_ParamValues[PARAM_CLIP] = 1.00f;
  m_ParamValues[PARAM_BITS] = 1.00f;
  m_ParamValues[PARAM_RATE] = 0.50f;
  m_ParamValues[PARAM_POSTFILT] = 0.9f;
  m_ParamValues[PARAM_NONLIN] = 0.58f;
}

string AceCrusherPlugin::GetParamName(int iParam) const
{
  PARAMETER_DESC const* pDesc = GetParameterDesc(iParam);

  if(!pDesc)
    return "";

  return pDesc->sName;
}

string AceCrusherPlugin::GetParamText(int iIndex) const
{
  PARAMETER_DESC const* pDesc = GetParameterDesc(iIndex);

  if(!pDesc)
    return "";

  return pDesc->ToString(m_ParamValues[iIndex]);
}

int AceCrusherPlugin::GetNumParams() const
{
  return NUM_PARAM_DESC;
}

float AceCrusherPlugin::GetParam(int iIndex) const
{
  float fRange = g_ParamDesc[iIndex].fMax - g_ParamDesc[iIndex].fMin;
  return (m_ParamValues[iIndex] - g_ParamDesc[iIndex].fMin) / fRange;
}

void AceCrusherPlugin::SetParam(int iIndex, float fValue)
{
  float fRange = g_ParamDesc[iIndex].fMax - g_ParamDesc[iIndex].fMin;
  m_ParamValues[iIndex] = fValue * fRange + g_ParamDesc[iIndex].fMin;

  m_dsp.m_fParamClip = m_ParamValues[PARAM_CLIP];
  m_dsp.m_fParamBits = m_ParamValues[PARAM_BITS];
  m_dsp.m_fParamRate = m_ParamValues[PARAM_RATE];
  m_dsp.m_fParamPostFilt = m_ParamValues[PARAM_POSTFILT];
  m_dsp.m_fParamNonLin = m_ParamValues[PARAM_NONLIN];

  m_dsp.m_bDirty = true;
}

// midi events
void AceCrusherPlugin::OnTempoChange(float /*fTempo*/)
{
}

void AceCrusherPlugin::SetSampleRate(float fSampleRate)
{
  m_dsp.m_fSampleRate = fSampleRate;
}

void AceCrusherPlugin::ProcessAudio(float** pInput, float** pOutput, long iNumSamples)
{
  m_dsp.ProcessAudio(pInput, pOutput, iNumSamples);
}

static EVEREST::IPlugin* CreatePlugin(EVEREST::IHost* pHost)
{
  return new AceCrusherPlugin(pHost);
}

#include "acecrusher_window.h"

#if EVEREST_ENABLE_GUI
static
juce::Component* CreateEditor(EVEREST::IPlugin* plugin, EVEREST::IHost* host)
{
  return new AceCrusherWindow(plugin, host);
}

#endif

namespace EVEREST
{
const PluginDescription g_PlugDesc =
{
  false, // isSynth
  { 232, 200 },
  "Ace",
  "AceCrusher",
  &CreatePlugin,
  &CreateEditor,
};
}

