// Delay effect plugin.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <algorithm>
#include <stdio.h>
#include <cassert>
#include "acedelay.h"
#include "acedelay_parameters.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/tools.h"

OpaqueParamDesc<DelayParamSet> const* const Descriptors[] =
{
  // add new parameters at the end, don't change the order of existing parameters.
  &DescSend,
  &DescClock,
  &DescTime,
  &DescFeedback,
  &DescSync,
};

int AceDelayEffect::GetNumParams() const
{
  return ArrayLength(Descriptors);
}

std::string AceDelayEffect::GetParamName(int iIndex) const
{
  return Descriptors[iIndex]->GetName();
}

float AceDelayEffect::GetParam(int iIndex) const
{
  auto& desc = Descriptors[iIndex];
  return desc->Getf(m_Values);
}

void AceDelayEffect::SetParam(int iIndex, float fValue)
{
  auto& desc = Descriptors[iIndex];
  return desc->Setf(m_Values, fValue);
}

std::string AceDelayEffect::GetParamText(int iIndex) const
{
  auto& desc = Descriptors[iIndex];
  return desc->GetStringValue(m_Values);
}

AceDelayEffect::AceDelayEffect()
{
  for(size_t i = 0; i < ArrayLength(Descriptors); ++i)
  {
    Descriptors[i]->SetDefault(m_Values);
  }

  SetSampleRate(44100.0);
}

void AceDelayEffect::SetSampleRate(float fSampleRate)
{
  m_fTempo = 145.0;
  m_fSampleRate = fSampleRate;
}

void AceDelayEffect::ProcessAudio(float** pInputs, float** pOutputs, long iNumSamples)
{
  auto const send = m_Values.delaySend;
  auto timeInSamples = m_fSampleRate * m_Values.delayTime;

  if(m_Values.delaySync)
  {
    // 1/4 means one beat.
    auto const delayFactor = ToFloat(m_Values.delayClock) * 4.0;
    auto const beatSeconds = 60.0 / m_fTempo;
    auto const fDelaySeconds = beatSeconds * delayFactor;
    timeInSamples = int(m_fSampleRate * fDelaySeconds);
  }

  for(int iChannel = 0; iChannel < 2; ++iChannel)
  {
    auto& delay = m_Delay[iChannel];
    delay.SetTime(timeInSamples);
    delay.SetFeedback(100.0f * m_Values.delayFeedback);

    for(long i = 0; i < iNumSamples; ++i)
    {
      float inputSample = pInputs[iChannel][i];
      pOutputs[iChannel][i] = blend(inputSample, delay.Work(inputSample), send);
    }
  }
}

void AceDelayEffect::OnTempoChange(float fTempo)
{
  m_fTempo = fTempo;
}

static EVEREST::IPlugin* CreatePlugin(EVEREST::IHost*)
{
  return new AceDelayEffect;
}

namespace EVEREST
{
const PluginDescription g_PlugDesc =
{
  false, // isSynth
  { 0, 0 },
  "Ace",
  "AceDelay",
  &CreatePlugin,
  nullptr,
};
}

