// Yet another software synthesizer plugin.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <cassert>
#include <iomanip>
#include <sstream>
#include <cmath>

#include "lib_utils/fpu_util.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/tools.h"

#include "plugin.h"

#include "lib_echidna/echidna_synth.h"
#include "echidna_params.h"
using namespace LibEchidna;

typedef float ZeroOneToValueFunction (float);

template<int MIN, int MAX, int SCALE = 1>
float AffineFunction(float fZeroOneValue)
{
  auto const fMin = float(MIN) / float(SCALE);
  auto const fMax = float(MAX) / float(SCALE);
  auto const fRange = fMax - fMin;
  auto const fValue = fMin + fRange * fZeroOneValue;
  return Clamp(fValue, fMin, fMax);
}

struct PARAMETER_DESC
{
  char const* sName;
  ZeroOneToValueFunction* pMappingFn;
  std::string (* ToString)(float fValue);
  size_t offset;
};

static std::string ToStringNumber(float fValue)
{
  std::stringstream s;
  s << std::setprecision(4) << fValue;
  return s.str();
}

static std::string ToStringSeconds(float fValue)
{
  std::stringstream s;

  if(-1.0 < fValue && fValue < 1.0)
    s << std::setprecision(4) << fValue * 1000.0f << "ms";
  else
    s << std::setprecision(4) << fValue << "s";

  return s.str();
}

static std::string ToStringBool(float fValue)
{
  return fValue >= 0.5 ? "yes" : "no";
}

static std::string ToStringClock(float fValue)
{
  auto const clock = ToClock(fValue);

  if(clock == 0)
    return "off";
  else
    return ToString(clock);
}

static std::string ToStringShape(float fValue)
{
  return ToString(ParamToOscShape(fValue));
}

static std::string ToStringLfoShape(float fValue)
{
  return ToString(ParamToShape(fValue));
}

auto const Identity = &AffineFunction<0, 1>;
auto const Linear100 = &AffineFunction<0, 100>;
auto const Pitch24 = &AffineFunction<-24, 24>;
auto const Pitch6 = &AffineFunction<-6, 6>;

float ExpAttack(float fValue)
{
  return (0.37f * exp(fValue * 128.0 * 0.0857f)) * 0.001;
}

#define Offset(member) \
  (size_t)(&((ParamValues*)nullptr)->member)

/**
 * @brief This array defines the order of the parameters as they are seen by the
 * client programs. This indirectly determines the file format of the patches.
 * Do not change the order of the elements! Add new parameters at the end of the
 * array, and do not remove old ones.
 */
PARAMETER_DESC const g_ParamDesc[] =
{
  { "Volume", Identity, &ToStringNumber, Offset(m_Volume) },
  { "Misc.Mono", Identity, &ToStringBool, Offset(m_Mono) },
  { "Porta", &AffineFunction<0, 10>, &ToStringNumber, Offset(m_Portamento) },

  { "Eg1.Attack", &ExpAttack, &ToStringSeconds, Offset(env1.Attack) },
  { "Eg1.Decay", &AffineFunction<1, 1000, 1000>, &ToStringSeconds, Offset(env1.Decay) },
  { "Eg1.Sustain", Identity, &ToStringNumber, Offset(env1.Sustain) },
  { "Eg1.Release", &ExpAttack, &ToStringSeconds, Offset(env1.Release) },
  { "Eg1.ToFlt", &AffineFunction<-100, 100>, &ToStringNumber, Offset(env1.ToFilter) },
  { "Eg1.ToVol", &AffineFunction<-1, 1>, &ToStringNumber, Offset(env1.ToVolume) },
  { "Eg1.ToPitch", &AffineFunction<-24, 24>, &ToStringNumber, Offset(env1.ToPitch) },

  { "Eg2.Attack", &ExpAttack, &ToStringSeconds, Offset(env2.Attack) },
  { "Eg2.Decay", &AffineFunction<1, 1000, 1000>, &ToStringSeconds, Offset(env2.Decay) },
  { "Eg2.Sustain", Identity, &ToStringNumber, Offset(env2.Sustain) },
  { "Eg2.Release", &ExpAttack, &ToStringSeconds, Offset(env2.Release) },
  { "Eg2.ToFlt", &AffineFunction<-100, 100>, &ToStringNumber, Offset(env2.ToFilter) },
  { "Eg2.ToVol", &AffineFunction<-1, 1>, &ToStringNumber, Offset(env2.ToVolume) },
  { "Eg2.ToPitch", &AffineFunction<-24, 24>, &ToStringNumber, Offset(env2.ToPitch) },

  { "Osc1.Shape", Identity, &ToStringShape, Offset(osc1.Shape) },
  { "Osc1.PWidth", Identity, &ToStringNumber, Offset(osc1.PulseWidth) },
  { "Osc1.Detune", &AffineFunction<-1, 1>, &ToStringNumber, Offset(osc1.Detune) },
  { "Osc1.Pitch", Pitch24, &ToStringNumber, Offset(osc1.Pitch) },

  { "Osc2.Shape", Identity, &ToStringShape, Offset(osc2.Shape) },
  { "Osc2.PWidth", Identity, &ToStringNumber, Offset(osc2.PulseWidth) },
  { "Osc2.Detune", &AffineFunction<-1, 1>, &ToStringNumber, Offset(osc2.Detune) },
  { "Osc2.Pitch", Pitch24, &ToStringNumber, Offset(osc2.Pitch) },

  { "Osc.Volume", Identity, &ToStringNumber, Offset(m_OscVolume) },
  { "Osc.Balance", Identity, &ToStringNumber, Offset(m_OscBalance) },
  { "Osc.RingMod", Identity, &ToStringNumber, Offset(m_OscRingMod) },

  { "Flt1.Cut", Identity, &ToStringNumber, Offset(filter1.Cutoff) },
  { "Flt1.Res", Identity, &ToStringNumber, Offset(filter1.Resonance) },
  { "Flt1.Pan", &AffineFunction<-1, 1>, &ToStringNumber, Offset(filter1.Pan) },
  { "Flt2.Cut", Identity, &ToStringNumber, Offset(filter2.Cutoff) },
  { "Flt2.Res", Identity, &ToStringNumber, Offset(filter2.Resonance) },
  { "Flt2.Pan", &AffineFunction<-1, 1>, &ToStringNumber, Offset(filter2.Pan) },
  { "FltBalance", &AffineFunction<-1, 1>, &ToStringNumber, Offset(m_FilterBalance) },
  { "FltGain", Identity, &ToStringNumber, Offset(m_FilterPregain) },

  { "Delay.Send", Identity, &ToStringNumber, Offset(delay.send) },
  { "Delay.Time", Identity, &ToStringNumber, Offset(delay.time) },
  { "Delay.Feedback", Identity, &ToStringNumber, Offset(delay.feedback) },
  { "Delay.Clock", Identity, &ToStringClock, Offset(delay.clock) },

  { "Lfo1.Shape", Identity, &ToStringLfoShape, Offset(lfo1.Shape) },
  { "Lfo1.Rate", &AffineFunction<1, 100, 10>, &ToStringNumber, Offset(lfo1.Rate) },
  { "Lfo1.Clock", Identity, &ToStringClock, Offset(lfo1.Clock) },
  { "Lfo1.ToFilter", Linear100, &ToStringNumber, Offset(lfo1.ToFilter) },
  { "Lfo1.ToVol", Identity, &ToStringNumber, Offset(lfo1.ToVolume) },
  { "Lfo1.ToPitch", Pitch6, &ToStringNumber, Offset(lfo1.ToPitch) },

  { "Lfo2.Shape", Identity, &ToStringLfoShape, Offset(lfo2.Shape) },
  { "Lfo2.Rate", &AffineFunction<1, 100, 10>, &ToStringNumber, Offset(lfo2.Rate) },
  { "Lfo2.Clock", Identity, &ToStringClock, Offset(lfo2.Clock) },
  { "Lfo2.ToFilter", Linear100, &ToStringNumber, Offset(lfo2.ToFilter) },
  { "Lfo2.ToVol", Identity, &ToStringNumber, Offset(lfo2.ToVolume) },
  { "Lfo2.ToPitch", Pitch6, &ToStringNumber, Offset(lfo2.ToPitch) },

  { "Vel.ToFilter", Linear100, &ToStringNumber, Offset(m_VelocityToFilter) },
  { "Vel.ToVol", Identity, &ToStringNumber, Offset(m_VelocityToVolume) },
  { "Vel.ToPitch", Pitch6, &ToStringNumber, Offset(m_VelocityToPitch) },

  { "Mod.ToFilter", Linear100, &ToStringNumber, Offset(m_ModToFilter) },
  { "Mod.ToVol", Identity, &ToStringNumber, Offset(m_ModToVolume) },
  { "Mod.ToPitch", Pitch6, &ToStringNumber, Offset(m_ModToPitch) },
  { "Mod.ToLfoToPitch", Identity, &ToStringNumber, Offset(m_ModToLfo1ToPitch) },

  { "Ph.Send", Identity, &ToStringNumber, Offset(phaser.Send) },
  { "Ph.Freq", Identity, &ToStringNumber, Offset(phaser.Freq) },
  { "Ph.Depth", &AffineFunction<220, 1600>, &ToStringNumber, Offset(phaser.Depth) },
  { "Ph.Feedback", Identity, &ToStringNumber, Offset(phaser.Feedback) },
};

const auto NUM_PARAM_DESC = ArrayLength(g_ParamDesc);
static_assert(NUM_PARAM_DESC == NUM_PARAMS, "g_ParamDesc and the parameter enumeration are out-of-sync");

PARAMETER_DESC const* GetParameterDesc(unsigned int index)
{
  if(index >= NUM_PARAM_DESC)
    return nullptr;

  return &g_ParamDesc[index];
}

EchidnaPlugin::EchidnaPlugin(EVEREST::IHost* pHost) :
  m_pHost(pHost)
{
  m_fVstValues.resize(NUM_PARAM_DESC);

  // set all parameters to their minimum value, just to avoid NANs.
  for(size_t i = 0; i < NUM_PARAM_DESC; ++i)
    SetParam(i, 0.0f);

  SetParam(PARAM_VOLUME, 0.5);

  SetParam(PARAM_ENV1_ATTACK, 0);
  SetParam(PARAM_ENV1_DECAY, 0.099);
  SetParam(PARAM_ENV1_SUSTAIN, 0.74);
  SetParam(PARAM_ENV1_RELEASE, 0);
  SetParam(PARAM_ENV1_TO_FILTER, 0.5);
  SetParam(PARAM_ENV1_TO_VOLUME, 1.0);
  SetParam(PARAM_ENV1_TO_PITCH, 0.5);

  SetParam(PARAM_ENV2_ATTACK, 0);
  SetParam(PARAM_ENV2_DECAY, 0.099);
  SetParam(PARAM_ENV2_SUSTAIN, 0.74);
  SetParam(PARAM_ENV2_RELEASE, 0);
  SetParam(PARAM_ENV2_TO_FILTER, 0.5);
  SetParam(PARAM_ENV2_TO_VOLUME, 0.5);
  SetParam(PARAM_ENV2_TO_PITCH, 0.5);

  SetParam(PARAM_OSC1_SHAPE, 0.5);
  SetParam(PARAM_OSC1_PULSEWIDTH, 0.75);
  SetParam(PARAM_OSC1_DETUNE, 0.5);
  SetParam(PARAM_OSC1_PITCH, 0.5);

  SetParam(PARAM_OSC2_SHAPE, 0.94);
  SetParam(PARAM_OSC2_PULSEWIDTH, 0.5);
  SetParam(PARAM_OSC2_DETUNE, 0.55);
  SetParam(PARAM_OSC2_PITCH, 0.5);

  SetParam(PARAM_OSC_VOLUME, 0.0);
  SetParam(PARAM_OSC_BALANCE, 0.5);
  SetParam(PARAM_OSC_RINGMOD, 0.0);

  SetParam(PARAM_FLT1_CUTOFF, 1);
  SetParam(PARAM_FLT1_RESO, 0);
  SetParam(PARAM_FLT1_PAN, 0.5);

  SetParam(PARAM_FLT2_CUTOFF, 1);
  SetParam(PARAM_FLT2_RESO, 0);
  SetParam(PARAM_FLT2_PAN, 0.5);

  SetParam(PARAM_DELAY_SEND, 0);
  SetParam(PARAM_DELAY_TIME, 0);
  SetParam(PARAM_DELAY_FEEDBACK, 0.44);
  SetParam(PARAM_DELAY_CLOCK, 0.67);

  SetParam(PARAM_MOD_TO_PITCH, 0.50);
  SetParam(PARAM_LFO1_TO_OSC1_PITCH, 0.50);
  SetParam(PARAM_MOD_TO_PITCH, 0.50);
  SetParam(PARAM_MOD_TO_LFO, 0.60);

  SetParam(PARAM_LFO1_RATE, 0.75);
  SetParam(PARAM_LFO2_RATE, 0.75);
  SetParam(PARAM_LFO2_TO_OSC1_PITCH, 0.50);

  SetParam(PARAM_PORTAMENTO, 1.00);
}

std::string EchidnaPlugin::GetParamName(int iParam) const
{
  auto const* pDesc = GetParameterDesc(iParam);

  if(!pDesc)
    return "";

  return pDesc->sName;
}

std::string EchidnaPlugin::GetParamText(int iIndex) const
{
  auto const* pDesc = GetParameterDesc(iIndex);

  if(!pDesc)
    return "";

  return pDesc->ToString(m_fVstValues[iIndex]);
}

int EchidnaPlugin::GetNumParams() const
{
  return NUM_PARAM_DESC;
}

float EchidnaPlugin::GetParam(int iIndex) const
{
  return m_fVstValues[iIndex];
}

void EchidnaPlugin::SetParam(int iIndex, float fValue)
{
  m_fVstValues[iIndex] = fValue; // save VST value.

  auto const& desc = g_ParamDesc[iIndex];
  auto fRealValue = desc.pMappingFn(fValue);

  auto value = (float*)(((char*)&m_Synth.m_Values) + desc.offset);
  *value = fRealValue;
}

void EchidnaPlugin::SetSampleRate(float fSampleRate)
{
  m_Synth.SetSampleRate(fSampleRate);
}

void EchidnaPlugin::ProcessAudio(float** /*pInput*/, float** pOutput, long const iNumSamples)
{
  m_Synth.ProcessAudio(pOutput, iNumSamples);
}

// midi events
void EchidnaPlugin::OnTempoChange(float fTempo)
{
  m_Synth.OnTempoChange(fTempo);
}

void EchidnaPlugin::noteOn(int /*channel*/, int iNote, int iVelocity)
{
  m_Synth.OnNoteOn(iNote, iVelocity);
}

void EchidnaPlugin::noteOff(int /*channel*/, int iNote, int iVelocity)
{
  m_Synth.OnNoteOff(iNote, iVelocity);
}

void EchidnaPlugin::pitchBend(int /*channel*/, float fBend)
{
  m_Synth.OnPitchBend(fBend);
}

void EchidnaPlugin::controlChange(int /*channel*/, int iControl, int iValue)
{
  m_Synth.OnControlChange(iControl, iValue);
}

#if EVEREST_ENABLE_GUI

#include "echidna_window.h"

static Component* CreateEditor(EVEREST::IPlugin* plugin, EVEREST::IHost* host)
{
  return new EchidnaWindow(plugin, host);
}

#endif

static EVEREST::IPlugin* CreatePlugin(EVEREST::IHost* pHost)
{
  return new EchidnaPlugin(pHost);
}

namespace EVEREST
{
const PluginDescription g_PlugDesc =
{
  true, // isSynth
  { 700, 580 },
  "Ace",
  "AceEchidna",
  &CreatePlugin,
#if EVEREST_ENABLE_GUI
  &CreateEditor,
#endif
};
}

