// Filter effect plugin.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <algorithm>
#include <stdio.h>
#include <cassert>
#include "acefilter.h"
#include "acefilter_parameters.h"
#include "acefilter_window.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/tools.h"

OpaqueParamDesc<FilterParamSet> const* const Descriptors[] =
{
  // add new parameters at the end, don't change the order of existing parameters.
  &DescType,
  &DescFreq,
  &DescReso,
};

int AceFilterEffect::GetNumParams() const
{
  return ArrayLength(Descriptors);
}

std::string AceFilterEffect::GetParamName(int iIndex) const
{
  return Descriptors[iIndex]->GetName();
}

float AceFilterEffect::GetParam(int iIndex) const
{
  auto& desc = Descriptors[iIndex];
  return desc->Getf(m_Values);
}

void AceFilterEffect::SetParam(int iIndex, float fValue)
{
  auto& desc = Descriptors[iIndex];
  desc->Setf(m_Values, fValue);
}

std::string AceFilterEffect::GetParamText(int iIndex) const
{
  auto& desc = Descriptors[iIndex];
  return desc->GetStringValue(m_Values);
}

EVEREST::IPlugin* AceFilterEffect::CreateInstance(EVEREST::IHost* pHost)
{
  return new AceFilterEffect(pHost);
}

AceFilterEffect::AceFilterEffect(EVEREST::IHost* pHost) :
  m_pHost(pHost)
{
  for(size_t i = 0; i < ArrayLength(Descriptors); ++i)
  {
    Descriptors[i]->SetDefault(m_Values);
  }

  SetSampleRate(44100.0);
}

void AceFilterEffect::SetSampleRate(float fSampleRate)
{
  m_fSampleRate = fSampleRate;
}

void AceFilterEffect::ProcessAudio(float** pInputs, float** pOutputs, long iNumSamples)
{
  for(int iChannel = 0; iChannel < 2; ++iChannel)
  {
    Filter* filter = nullptr;
    switch(m_Values.filterType)
    {
    case FilterTypeMoog:
      filter = &m_MoogFilter[iChannel];
      m_MoogFilter[iChannel].setCutoff((m_Values.filterFreq / 100.0f) * 10000.0f / m_fSampleRate);
      m_MoogFilter[iChannel].setRes(m_Values.filterReso / 100.0f);
      break;
    default:
    case FilterTypeLowpass:
      filter = &m_Filter[iChannel];
      m_Filter[iChannel].SetParams(m_Values.filterFreq, m_Values.filterReso);
      break;
    case FilterTypeHighpass:
      filter = &m_HighPassFilter[iChannel];
      m_HighPassFilter[iChannel].SetParams(m_Values.filterFreq, m_Values.filterReso);
      break;
    }

    for(long i = 0; i < iNumSamples; ++i)
    {
      float inputSample = pInputs[iChannel][i];
      float outputSample = filter->Work(inputSample);

      pOutputs[iChannel][i] = Clamp(outputSample, -1.0f, 1.0f);
    }
  }
}

#if EVEREST_ENABLE_GUI

static
juce::Component* CreateEditor(EVEREST::IPlugin* plugin, EVEREST::IHost* host)
{
  return new AceFilterWindow(plugin, host);
}

#endif

static EVEREST::IPlugin* CreatePlugin(EVEREST::IHost* pHost)
{
  return AceFilterEffect::CreateInstance(pHost);
}

namespace EVEREST
{
const PluginDescription g_PlugDesc =
{
  false, // isSynth
  { 232, 200 },
  "Ace",
  "AceFilter",
  &CreatePlugin,
  &CreateEditor,
};
}

