// SF2 based sampler

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <cassert>
#include <iomanip>
#include <sstream>
#include <cmath>

#define TSF_IMPLEMENTATION
#include "tsf.h"

using namespace std;

#include "lib_utils/debug.h"
#include "lib_utils/fpu_util.h"
#include "lib_utils/fraction.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/string_format.h"
#include "lib_utils/tools.h"

#include "plugin.h"

typedef float ZeroOneToValueFunction (float);

template<int MIN, int MAX, int SCALE = 1>
float AffineFunction(float fZeroOneValue)
{
  auto const fMin = float(MIN) / float(SCALE);
  auto const fMax = float(MAX) / float(SCALE);
  auto const fRange = fMax - fMin;
  auto const fValue = fMin + fRange * fZeroOneValue;
  return Clamp(fValue, fMin, fMax);
}

struct ParamValues
{
  float m_Volume;
  float m_Mono;
  float m_Portamento;
};

struct PARAMETER_DESC
{
  char const* sName;
  // float fMin, fMax;
  ZeroOneToValueFunction* pMappingFn;
  string (* ToString)(float fValue);
  float ParamValues::* value;
};

static
string ToStringNumber(float fValue)
{
  stringstream s;
  s << setprecision(4) << fValue;
  return s.str();
}

static
string ToStringBool(float fValue)
{
  return fValue >= 0.5 ? "yes" : "no";
}

typedef ParamValues PV;

auto const Identity = &AffineFunction<0, 1>;

/**
 * @brief This array defines the order of the parameters as they are seen by the
 * client programs. This indirectly determines the file format of the patches.
 * Do not change the order of the elements! Add new parameters at the end of the
 * array, and do not remove old ones.
 */
static
PARAMETER_DESC const g_ParamDesc[] =
{
  { "Volume", Identity, &ToStringNumber, &PV::m_Volume },
  { "Misc.Mono", Identity, &ToStringBool, &PV::m_Mono },
  { "Porta", &AffineFunction<0, 10>, &ToStringNumber, &PV::m_Portamento },
};

static unsigned int const NUM_PARAM_DESC = ArrayLength(g_ParamDesc);

PARAMETER_DESC const* GetParameterDesc(unsigned int index)
{
  if(index >= NUM_PARAM_DESC)
    return nullptr;

  return &g_ParamDesc[index];
}

AceFontPlugin::AceFontPlugin(EVEREST::IHost* /*pHost*/)
{
  m_fVstValues.resize(NUM_PARAM_DESC);

  // set all parameters to their minimum value, just to avoid NANs.
  for(size_t i = 0; i < NUM_PARAM_DESC; ++i)
    SetParam(i, 0.0f);

  SetParam(0, 1.0);
  SetParam(2, 1.00);

  m_sampler = tsf_load_filename("plug_acefont/default.sf2");
  assert(m_sampler);

  tsf_set_output(m_sampler, TSF_MONO, 48000, 0.0f);
}

AceFontPlugin::~AceFontPlugin()
{
  tsf_close(m_sampler);
}

string AceFontPlugin::GetParamName(int iParam) const
{
  auto const* pDesc = GetParameterDesc(iParam);

  if(!pDesc)
    return "";

  return pDesc->sName;
}

string AceFontPlugin::GetParamText(int iIndex) const
{
  auto const* pDesc = GetParameterDesc(iIndex);

  if(!pDesc)
    return "";

  return pDesc->ToString(m_fVstValues[iIndex]);
}

int AceFontPlugin::GetNumParams() const
{
  return NUM_PARAM_DESC;
}

float AceFontPlugin::GetParam(int iIndex) const
{
  return m_fVstValues[iIndex];
}

void AceFontPlugin::SetParam(int iIndex, float fValue)
{
  m_fVstValues[iIndex] = fValue; // save VST value.

  auto& desc = g_ParamDesc[iIndex];
  auto fRealValue = desc.pMappingFn(fValue);

  assert(desc.value != nullptr);

  (void)fRealValue;
  // m_Synth.m_Values.*desc.value = fRealValue;
}

void AceFontPlugin::SetSampleRate(float /*fSampleRate*/)
{
}

void AceFontPlugin::ProcessAudio(float** /*pInput*/, float** pOutput, long const iNumSamples)
{
  tsf_render_float(m_sampler, pOutput[0], iNumSamples, 0);

  for(int i = 0; i < iNumSamples; ++i)
    pOutput[1][i] = pOutput[0][i];
}

// midi events
void AceFontPlugin::OnTempoChange(float /*fTempo*/)
{
}

void AceFontPlugin::noteOn(int /*channel*/, int note, int vel)
{
  tsf_note_on(m_sampler, 0, note, vel / 128.0f);
}

void AceFontPlugin::noteOff(int /*channel*/, int note, int /*vel*/)
{
  tsf_note_off(m_sampler, 0, note);
}

void AceFontPlugin::pitchBend(int /*channel*/, float /*fBend*/)
{
}

void AceFontPlugin::controlChange(int /*channel*/, int /*control*/, int /*val*/)
{
}

static EVEREST::IPlugin* CreatePlugin(EVEREST::IHost* pHost)
{
  return new AceFontPlugin(pHost);
}

namespace EVEREST
{
const PluginDescription g_PlugDesc =
{
  true, // isSynth
  { 10, 10 },
  "Ace",
  "AceFont",
  &CreatePlugin,
#if EVEREST_ENABLE_GUI
  nullptr,
#endif
};
}

