/**
 * @file acevirus.cpp
 * @brief Implementation of the MIDI-only Virus VST plugin.
 * @author Sebastien Alaiwan
 * @version 1.0
 * @date 2010-09-15
 */

/* Copyright (C)
 * 2010 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <sstream>
#include "debug.h"
#include "virus_editor.h"
#include "acevirus.h"

#ifdef ENABLE_MIDI
int AceVirus::GetDeviceId(bool bInput)
{
  if(bInput)
  {
    int n = m_MidiInput.getPortCount();

    for(int i = 0; i < n; ++i)
    {
      std::string const sName = m_MidiOutput.getPortName(i);
      DebugTrace("[Ks-in][%02d] %s\n", i, sName.c_str());

      if(m_sInputDeviceName == sName)
        return i;
    }
  }
  else
  {
    int n = m_MidiOutput.getPortCount();

    for(int i = 0; i < n; ++i)
    {
      std::string const sName = m_MidiOutput.getPortName(i);
      DebugTrace("[Ks-out][%02d] %s\n", i, sName.c_str());

      if(m_sOutputDeviceName == sName)
        return i;
    }
  }

  return -1;
}

#endif

namespace EVEREST
{
IPlugin* CreatePlugin(IHost* pHost)
{
  return new AceVirus;
}
}

AceVirus::AceVirus() : m_MidiVirus(this)
{
  m_MidiInput.setCallback(&static_callback, this);
}

AceVirus::~AceVirus()
{
}

void AceVirus::Flush()
{
}

bool AceVirus::PullSysex(std::vector<uint8_t>& sysex_from_virus)
{
  return false;
}

void AceVirus::WriteMidi(uint8_t const* pData, int iLength)
{
  if(iLength == 0)
    return;

  /*
   * FL can't receive sysex.
   * eXT can't reveive sysex.
   * FL can send sysex (through FL API, not sure about how Fruity Wrapper handles sysex coming from a plugin).
   * eXT can't send sysex.
   *
   * My conclusion is: you have to completely bypass the host when you have to send/recv sysex.
   */

#ifdef ENABLE_MIDI

  if(pData[0] == 0xF0)
  {
    std::vector<uint8_t> msg(iLength);
    memcpy(msg.data(), pData, iLength);
    m_MidiOutput.sendMessage(&msg);
  }

#endif
}

bool AceVirus::DownloadPatch(int iPart)
{
  try
  {
    m_MidiVirus.RequestSingleDump(iPart);
    return true;
  }
  catch(std::runtime_error const& err)
  {
    std::basic_stringstream<char> ssError;
    ssError << "AceVirus: Could not download patch." << std::endl;
    ssError << "Error: " << err.what() << std::endl;
    AlertWindow::showMessageBox(AlertWindow::WarningIcon, T("Error"), ssError.str().c_str());
    return false;
  }
}

void AceVirus::UploadPatch(int iPart)
{
  try
  {
    m_MidiVirus.UploadPatch(iPart, m_Patches[iPart]);
  }
  catch(std::runtime_error const& err)
  {
    std::basic_stringstream<char> ssError;
    ssError << "AceVirus: Could not upload patch." << std::endl;
    ssError << "Error: " << err.what() << std::endl;
    AlertWindow::showMessageBox(AlertWindow::WarningIcon, T("Error"), ssError.str().c_str());
  }
}

VirusPatch AceVirus::GetPatch(int iPart) const
{
  return m_Patches[iPart];
}

void AceVirus::SetPatch(int iPart, VirusPatch const& p)
{
  m_Patches[iPart] = p;
}

void AceVirus::SetSampleRate(float fSampleRate)
{
}

void AceVirus::ProcessAudio(float** pInput, float** pOutput, long iNumSamples)
{
}

void AceVirus::OnTempoChange(float fTempo)
{
}

bool AceVirus::IsInputOK() const
{
#ifdef ENABLE_MIDI
  return m_MidiInput.isOpen();
#else
  return false;
#endif
}

bool AceVirus::IsOutputOK() const
{
#ifdef ENABLE_MIDI
  return m_MidiOutput.isOpen();
#else
  return false;
#endif
}

static void WriteDword(std::vector<uint8_t>& data, unsigned int iVal)
{
  data.push_back((iVal >> 24) & 0xFF);
  data.push_back((iVal >> 16) & 0xFF);
  data.push_back((iVal >> 8) & 0xFF);
  data.push_back((iVal >> 0) & 0xFF);
}

static bool ReadDword(std::vector<uint8_t> const& data, int& offset, unsigned int& iVal)
{
  if(offset + 4 > (int)data.size())
    return false;

  iVal = 0;

  for(int i = 0; i < 4; ++i)
  {
    iVal <<= 8;
    iVal |= data[offset++];
  }

  return true;
}

static void WriteString(std::vector<uint8_t>& data, std::string const& sVal)
{
  WriteDword(data, sVal.size());

  for(size_t i = 0; i < sVal.size(); ++i)
  {
    data.push_back(sVal[i]);
  }
}

static bool ReadString(std::vector<uint8_t> const& data, int& offset, std::string& sVal)
{
  unsigned int len;

  if(!ReadDword(data, offset, len))
    return false;

  sVal.clear();

  for(size_t i = 0; i < len; ++i)
  {
    sVal += data[offset + i];
  }

  offset += len;

  return true;
}

std::vector<uint8_t> AceVirus::SaveState() const
{
  std::vector<uint8_t> data;

  // save magic
  data.push_back(0xAC);
  data.push_back(0xE0);
  data.push_back(0x53);
  data.push_back(0x80);

  // save format version
  data.push_back(2);

  for(int i = 0; i < 16; ++i)
  {
    int iLength = (int)m_Patches[i].m_vSysex.size();

    // save patch length, big endian.
    data.push_back((iLength >> 24) & 0xFF);
    data.push_back((iLength >> 16) & 0xFF);
    data.push_back((iLength >> 8) & 0xFF);
    data.push_back((iLength >> 0) & 0xFF);

    // save raw sysex
    for(int j = 0; j < iLength; ++j)
    {
      data.push_back(m_Patches[i].m_vSysex[j]);
    }
  }

  WriteString(data, m_sInputDeviceName);
  WriteString(data, m_sOutputDeviceName);

  return data;
}

void AceVirus::LoadState(std::vector<uint8_t> const& data)
{
  DebugTrace("AceVirus::LoadState\n");

  uint8_t const* p = &data[0];

  // check magic
  if(data.size() < 4 || p[0] != 0xAC || p[1] != 0xE0 || p[2] != 0x53 || p[3] != 0x80)
    return;

  // check format version.
  if(data.size() < 5)
    return;

  int version = p[4];

  if(version != 1 && version != 2)
    return;

  int offset = 5;

  for(int i = 0; i < 16; ++i)
  {
    if(offset + 4 > (int)data.size())
      return;

    int iLength =
      (p[offset + 0] << 24)
      | (p[offset + 1] << 16)
      | (p[offset + 2] << 8)
      | (p[offset + 3] << 0);
    offset += 4;

    if(offset + iLength > (int)data.size())
      return;

    // load raw sysex
    m_Patches[i].Parse(&p[offset], iLength);
    offset += iLength;
  }

  if(version == 2)
  {
    ReadString(data, offset, m_sInputDeviceName);
    ReadString(data, offset, m_sOutputDeviceName);
  }

  ReconnectMidi();

  // Initialize Virus object. This will put the hardware synth in sequencer mode.
  //
  DebugTrace("Setting Virus to in sequencer mode\n");
  m_MidiVirus.Init(this, this);

  for(int i = 0; i < 16; ++i)
  {
    DebugTrace("Loading patch %d/16\n", i + 1);
    m_MidiVirus.UploadPatch(i, m_Patches[i]);
  }
}

void AceVirus::midiInputCallback(double timeStamp, std::vector<uint8_t> const& message)
{
  m_MidiVirus.ProcessMidi(&message[0], message.size());
}

void AceVirus::OnPatchReceived(int iPart, VirusPatch const& p)
{
  SetPatch(iPart, p);
  RefreshAllEditors();
}

void AceVirus::RefreshAllEditors()
{
  std::cout << "RefreshAllEditors" << std::endl;

  for(auto i_editor = m_Editors.begin(); i_editor != m_Editors.end(); ++i_editor)
  {
    (*i_editor)->ProtectedRefresh();
  }
}

void AceVirus::GetDimensions(int& cx, int& cy)
{
  cx = VirusEditor::WIDTH;
  cy = VirusEditor::HEIGHT;
}

juce::Component* AceVirus::CreateEditor()
{
  return new VirusEditor(this);
}

void AceVirus::RegisterEditor(VirusEditor* pEditor)
{
  m_Editors.insert(pEditor);
}

void AceVirus::UnregisterEditor(VirusEditor* pEditor)
{
  auto i_editor = m_Editors.find(pEditor);
  m_Editors.erase(i_editor);
}

std::vector<std::string> AceVirus::GetInputDeviceList() const
{
  std::vector<std::string> ret;

  RtMidiIn input;
  int const n = input.getPortCount();

  for(int i = 0; i < n; ++i)
  {
    ret.push_back(input.getPortName(i));
  }

  return ret;
}

std::vector<std::string> AceVirus::GetOutputDeviceList() const
{
  std::vector<std::string> ret;

  RtMidiOut output;
  int const n = output.getPortCount();

  for(int i = 0; i < n; ++i)
  {
    ret.push_back(output.getPortName(i));
  }

  return ret;
}

void AceVirus::ChangeMidiInputDevice(std::string const& sName)
{
  m_sInputDeviceName = sName;
  ReconnectMidi();
}

void AceVirus::ChangeMidiOutputDevice(std::string const& sName)
{
  m_sOutputDeviceName = sName;
  ReconnectMidi();
}

std::string AceVirus::GetMidiInputDeviceName() const
{
  return m_sInputDeviceName;
}

std::string AceVirus::GetMidiOutputDeviceName() const
{
  return m_sOutputDeviceName;
}

void AceVirus::ReconnectMidi()
{
#ifdef ENABLE_MIDI
  int err;

  std::basic_stringstream<char> ssError;

  try
  {
    m_MidiInput.closePort();
    int id_input = GetDeviceId(true);

    if(id_input != -1)
      m_MidiInput.openPort(id_input);
  }
  catch(std::runtime_error const& err)
  {
    ssError << "AceVirus: Could not open MIDI input device. Virus TI Synth must not be opened by the host! Patch receiving will be disabled." << std::endl;
    ssError << "Error: " << err.what() << std::endl;
    AlertWindow::showMessageBox(AlertWindow::WarningIcon, T("Error"), ssError.str().c_str());
  }

  try
  {
    m_MidiOutput.closePort();
    int id_output = GetDeviceId(false);

    if(id_output != -1)
      m_MidiOutput.openPort(id_output);
  }
  catch(std::runtime_error const& err)
  {
    ssError << "AceVirus: Could not open MIDI output device. Virus TI Synth must not be opened by the host! All communication with the Virus will be disabled." << std::endl;
    ssError << "Error: " << err.what() << std::endl;
  }

  try
  {
    if(m_MidiOutput.isOpen())
      m_MidiVirus.Init(this, this);
  }
  catch(std::runtime_error const& err)
  {
    std::basic_stringstream<char> ssError;
    ssError << "AceVirus: Error while setting the synth into sequencer mode." << std::endl;
    ssError << "Error: " << err.what() << std::endl;
  }

  if(!ssError.str().empty())
    AlertWindow::showMessageBox(AlertWindow::WarningIcon, T("Error"), ssError.str().c_str());

#endif
}

