/**
 * @file acevirus.h
 * @brief Declaration of the MIDI-only Virus VST plugin.
 * @author Sebastien Alaiwan
 * @version 1.0
 * @date 2010-09-15
 */

/* Copyright (C)
 * 2010 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#pragma once

#include <vector>
#include <set>
#include "lib_everest.h"
#ifdef ENABLE_MIDI
#include "RtMidi.h"
#endif
#include "viruspatch.h"

class VirusEditor;

class AceVirus : public EVEREST::IPlugin,
  public EVEREST::ISerializable,
  public IMidiSink,
  public IMidiSource
{
  VirusPatch m_Patches[16];
  CMidiVirus m_MidiVirus;
  std::vector<uint8_t> m_Chunk;
  std::string m_sInputDeviceName, m_sOutputDeviceName;

#ifdef ENABLE_MIDI
  RtMidiIn m_MidiInput;
  RtMidiOut m_MidiOutput;

  int GetDeviceId(bool bInput = true);
#endif

  void ReconnectMidi();
  void RefreshAllEditors();

  std::set<VirusEditor*> m_Editors;

  static void static_callback(double timeStamp, std::vector<uint8_t>* message, void* userData)
  {
    auto* pThis = reinterpret_cast<AceVirus*>(userData);
    pThis->midiInputCallback(timeStamp, *message);
  }

public:
  AceVirus();
  virtual ~AceVirus();

  void midiInputCallback(double timeStamp, std::vector<uint8_t> const& message);

  // implementation of IMidiSink
  virtual void WriteMidi(uint8_t const* pData, int iLength);
  virtual void Flush();
  virtual bool PullSysex(std::vector<uint8_t>& sysex_from_virus);
  virtual void OnPatchReceived(int iPart, VirusPatch const& p);

  bool DownloadPatch(int iPart);
  void UploadPatch(int iPart);
  VirusPatch GetPatch(int iPart) const;
  void SetPatch(int iPart, VirusPatch const& p);

  bool IsInputOK() const;
  bool IsOutputOK() const;

  std::vector<std::string> GetInputDeviceList() const;
  std::vector<std::string> GetOutputDeviceList() const;

  void ChangeMidiInputDevice(std::string const& sName);
  void ChangeMidiOutputDevice(std::string const& sName);
  std::string GetMidiInputDeviceName() const;
  std::string GetMidiOutputDeviceName() const;

  // EVEREST::IPlugin implementation
  virtual void SetSampleRate(float fSampleRate);

  virtual void ProcessAudio(float** pInput, float** pOutput, long iNumSamples);

  // midi events
  virtual void OnTempoChange(float fTempo);

  virtual void GetId(char* pId) const
  {
    pId[0] = '@';
    pId[1] = '7';
    pId[2] = '*';
    pId[3] = '0';
  }

  virtual void GetDimensions(int& cx, int& cy);
  virtual juce::Component* CreateEditor();

  virtual bool IsSynth() { return true; }

  virtual bool HasEditor() { return true; }

  virtual std::vector<uint8_t> SaveState() const override;
  virtual void LoadState(std::vector<uint8_t> const& data) override;

  void RegisterEditor(VirusEditor* pEditor);
  void UnregisterEditor(VirusEditor* pEditor);
};

