// Filter effect plugin.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */

#include <algorithm>
#include <stdio.h>
#include <cassert>
#include "aceversatile.h"
#include "aceversatile_parameters.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/tools.h"
#include "lib_utils/bytestream.h"
#include "lib_compiler/compiled_program.h"

OpaqueParamDesc<FilterParamSet> const* const Descriptors[] =
{
  // add new parameters at the end, don't change the order of existing parameters.
  &DescType,
  &DescFreq,
  &DescReso,
};

int AceVersatileEffect::GetNumParams() const
{
  return ArrayLength(Descriptors);
}

std::string AceVersatileEffect::GetParamName(int iIndex) const
{
  return Descriptors[iIndex]->GetName();
}

float AceVersatileEffect::GetParam(int iIndex) const
{
  auto& desc = Descriptors[iIndex];
  return desc->Getf(m_Values);
}

void AceVersatileEffect::SetParam(int iIndex, float fValue)
{
  auto& desc = Descriptors[iIndex];
  desc->Setf(m_Values, fValue);
}

std::string AceVersatileEffect::GetParamText(int iIndex) const
{
  auto& desc = Descriptors[iIndex];
  return desc->GetStringValue(m_Values);
}

EVEREST::IPlugin* AceVersatileEffect::CreateInstance(EVEREST::IHost* pHost)
{
  return new AceVersatileEffect(pHost);
}

AceVersatileEffect::AceVersatileEffect(EVEREST::IHost* pHost) :
  m_pHost(pHost)
{
  for(auto& desc : Descriptors)
    desc->SetDefault(m_Values);

  SetSampleRate(44100.0);

  m_sText = R"(
float main(float input)
{
  return 0.0f;
}
)";
}

void AceVersatileEffect::SetSampleRate(float fSampleRate)
{
  m_fSampleRate = fSampleRate;
}

void AceVersatileEffect::ProcessAudio(float** pInputs, float** pOutputs, long iNumSamples)
{
  for(int iChannel = 0; iChannel < 2; ++iChannel)
  {
    if(!m_program[iChannel])
      continue;

    for(long i = 0; i < iNumSamples; ++i)
    {
      float inputSample = pInputs[iChannel][i];
      float outputSample = m_program[iChannel]->process(inputSample);

      pOutputs[iChannel][i] = Clamp(outputSample, -1.0f, 1.0f);
    }
  }
}

unsigned int const MAGIC = 0xACE91337;

std::vector<uint8_t> AceVersatileEffect::saveState() const
{
  vector<uint8_t> data;
  obytestream bs(data);
  bs.write_uint32(MAGIC);

  bs.write_uint32(0);
  bs.write_uint32(0);
  bs.write_uint32(0);

  bs.write_string(m_sText);
  return data;
}

void AceVersatileEffect::loadState(std::vector<uint8_t> const& data)
{
  try
  {
    ibytestream bs(data);

    auto const magic = bs.read_uint32();

    if(magic != MAGIC)
      fprintf(stderr, "Invalid input stream\n");

    bs.read_uint32();
    bs.read_uint32();
    bs.read_uint32();

    setProgramSource(bs.read_string());
  }
  catch(bytestream::end_of_stream const &)
  {
    fprintf(stderr, "Unexpected end of stream\n");
  }
}

void AceVersatileEffect::setProgramSource(std::string text)
{
  m_sText = text;

  try
  {
    m_program[0] = compile(m_sText.c_str());
    m_program[1] = compile(m_sText.c_str());
  }
  catch(runtime_error const& e)
  {
    fprintf(stderr, "Compilation failed: %s\n", e.what());
  }
}

std::string AceVersatileEffect::getText() const
{
  return m_sText;
}

///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#include "window.h"
static
EVEREST::IPlugin* CreatePlugin(EVEREST::IHost* pHost)
{
  return AceVersatileEffect::CreateInstance(pHost);
}

#if EVEREST_ENABLE_GUI

static
juce::Component* CreateEditor(EVEREST::IPlugin* plug, EVEREST::IHost* host)
{
  return new AceVersatileWindow(dynamic_cast<AceVersatileEffect*>(plug), host);
}

#endif

namespace EVEREST
{
const PluginDescription g_PlugDesc =
{
  false, // isSynth
  { 600, 600 },
  "Ace",
  "AceVersatile",
  &CreatePlugin,
  &CreateEditor,
};
}

