// Implementation of the MIDI-only Virus VST plugin.

/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 */

#include <cassert>
#include <cmath> // round
#include <sstream>
using namespace std;

#include "acevirusti.h"
#include "virus_params.h"
#include "lib_utils/bytestream.h"
#include "lib_utils/tools.h"
#include "lib_utils/math_tools.h"
#include "lib_utils/debug.h"

map<int, int> CreateOffetToIndexMap()
{
  map<int, int> m;

  for(int i = 0; i < ArrayLength(g_PatchParameters); ++i)
  {
    int iBase = int(g_PatchParameters[i].m_Page - 'A') * 128;
    m[g_PatchParameters[i].m_iOffset + iBase] = i;
  }

  return m;
}

map<int, int> const g_OffsetToIndex = CreateOffetToIndexMap();

// AceVirusTi

AceVirusTi::AceVirusTi(EVEREST::IHost* pHost) :
  m_MidiVirus(this),
  m_pHost(pHost)
{
  ArraySet(m_fValues, 0.0f);

  m_MidiInput.setCallback(&static_callback, this);
  m_PartNumber = 0;
}

void AceVirusTi::SetSampleRate(float /*fSampleRate*/)
{
}

void AceVirusTi::ProcessAudio(float** /*pInput*/, float** /*pOutput*/, long /*iNumSamples*/)
{
}

void AceVirusTi::OnTempoChange(float /*fTempo*/)
{
}

void AceVirusTi::noteOn(int /*channel*/, int iNote, int vel)
{
  uint8_t midi_string[3];
  midi_string[0] = 0x90 | m_PartNumber;
  midi_string[1] = (uint8_t)iNote & 0x7F;
  midi_string[2] = (uint8_t)vel & 0x7F;
  WriteMidi(midi_string, ArrayLength(midi_string));
}

void AceVirusTi::noteOff(int /*channel*/, int iNote, int /*vel*/)
{
  uint8_t midi_string[3];
  midi_string[0] = 0x80 | m_PartNumber;
  midi_string[1] = (uint8_t)iNote & 0x7F;
  midi_string[2] = 0;
  WriteMidi(midi_string, ArrayLength(midi_string));
}

void AceVirusTi::pitchBend(int /*channel*/, float fBend)
{
  uint8_t midi_string[3];
  int bend = (int)(fBend * 16384.0f + 16384.0f);
  midi_string[0] = 0xE0 | m_PartNumber;
  midi_string[1] = (bend >> 0) & 0xFF;
  midi_string[2] = (bend >> 8) & 0xFF;
  WriteMidi(midi_string, ArrayLength(midi_string));
}

int AceVirusTi::GetNumParams() const
{
  return ArrayLength(g_PatchParameters);
}

string AceVirusTi::GetParamName(int iParam) const
{
  return g_PatchParameters[iParam].m_sName;
}

float AceVirusTi::GetParam(int iIndex) const
{
  return m_fValues[iIndex];
}

string AceVirusTi::GetParamText(int iIndex) const
{
  stringstream s;
  s << m_fValues[iIndex];
  return s.str();
}

void AceVirusTi::SetParam(int iIndex, float fValue)
{
  InternalSetParam(iIndex, fValue, true, false);
}

void AceVirusTi::InternalSetParam(int iIndex, float fValue, bool bSendMidi, bool bNotifyHost)
{
  // DebugTrace("%s := %.2f\n", GetParamName(iIndex).c_str(),  fValue);
  std::lock_guard<std::recursive_mutex> lock_it(m_PatchLock);
  float const fOldValue = m_fValues[iIndex];
  m_fValues[iIndex] = fValue;

  if(bSendMidi)
  {
    int iOldVal = int(round(127.0f * fOldValue));
    iOldVal = Clamp(iOldVal, 0, 127);
    uint8_t const iOldByteValue = (uint8_t)iOldVal & 0x7F;

    int iVal = int(round(127.0f * fValue));
    iVal = Clamp(iVal, 0, 127);
    uint8_t const iByteValue = (uint8_t)iVal & 0x7F;

    if(iByteValue != iOldByteValue)
    {
      int iNum = g_PatchParameters[iIndex].m_iOffset;
      assert(iNum >= 0 && iNum <= 127);

      uint8_t iPart = (uint8_t)(m_PartNumber & 0x0F);

      if(g_PatchParameters[iIndex].m_Page == 'A')
      {
        uint8_t midi_string[3];
        midi_string[0] = 0xB0 | iPart;
        midi_string[1] = (uint8_t)iNum & 0x7F;
        midi_string[2] = iByteValue;
        WriteMidi(midi_string, ArrayLength(midi_string));
      }
      else if(g_PatchParameters[iIndex].m_Page == 'B')
      {
        // F0 00 20 33 01 10 71 00 0F 02 F7
        uint8_t msg[11];
        msg[0] = 0xF0;
        msg[1] = 0x00;
        msg[2] = 0x20;
        msg[3] = 0x33;
        msg[4] = 0x01;
        msg[5] = 0x10;
        msg[6] = 0x71;
        msg[7] = iPart;
        msg[8] = (uint8_t)iNum & 0x7F;
        msg[9] = iByteValue;
        msg[10] = 0xF7;
        WriteMidi(msg, ArrayLength(msg));
      }
      else if(g_PatchParameters[iIndex].m_Page == 'C')
      {
        uint8_t msg[11];
        msg[0] = 0xF0;
        msg[1] = 0x00;
        msg[2] = 0x20;
        msg[3] = 0x33;
        msg[4] = 0x01;
        msg[5] = 0x10;
        msg[6] = 0x6E;
        msg[7] = iPart;
        msg[8] = (uint8_t)iNum & 0x7F;
        msg[9] = iByteValue;
        msg[10] = 0xF7;
        WriteMidi(msg, ArrayLength(msg));
      }
    }
  }

  if(bNotifyHost)
    m_pHost->OnParamChanged(iIndex, m_fValues[iIndex]);
}

vector<string> AceVirusTi::GetInputDeviceList() const
{
  RtMidiIn input;
  int const n = input.getPortCount();

  vector<string> ret;

  for(int i = 0; i < n; ++i)
    ret.push_back(input.getPortName(i));

  return ret;
}

vector<string> AceVirusTi::GetOutputDeviceList() const
{
  RtMidiOut output;
  int const n = output.getPortCount();

  vector<string> ret;

  for(int i = 0; i < n; ++i)
    ret.push_back(output.getPortName(i));

  return ret;
}

bool AceVirusTi::IsInputOK() const
{
  return m_MidiInput.isOpen();
}

bool AceVirusTi::IsOutputOK() const
{
  return m_MidiOutput.isOpen();
}

void AceVirusTi::ChangeMidiInputDevice(string const& sName)
{
  if(m_sInputDeviceName == sName)
    return;

  m_sInputDeviceName = sName;
  ReconnectMidi();
}

void AceVirusTi::ChangeMidiOutputDevice(string const& sName)
{
  if(m_sOutputDeviceName == sName)
    return;

  m_sOutputDeviceName = sName;
  ReconnectMidi();
}

void AceVirusTi::ReconnectMidi()
{
  std::lock_guard<std::recursive_mutex> lock_it(m_PatchLock);
  basic_stringstream<char> ssError;

  try
  {
    m_MidiInput.closePort();
    int id_input = GetDeviceId(true);

    if(id_input != -1)
    {
      DebugTrace("Opening input port %d\n", id_input);
      m_MidiInput.ignoreTypes(false, true, true); // ignore midi time and active sensing, but enable sysex.
      m_MidiInput.openPort(id_input);
    }
  }
  catch(exception const& err)
  {
    ssError << "AceVirusTi: Could not open MIDI input device. Patch receiving will be disabled." << endl;
    ssError << "Error: " << err.what() << endl;
  }

  try
  {
    m_MidiOutput.closePort();
    int id_output = GetDeviceId(false);

    if(id_output != -1)
    {
      DebugTrace("Opening output port %d\n", id_output);
      m_MidiOutput.openPort(id_output);
    }
  }
  catch(exception const& err)
  {
    ssError << "AceVirusTi: Could not open MIDI output device. All communication with the Virus will be disabled." << endl;
    ssError << "Error: " << err.what() << endl;
  }

  try
  {
    if(m_MidiOutput.isOpen())
      m_MidiVirus.Init(this, this);
  }
  catch(exception const& err)
  {
    basic_stringstream<char> ssError;
    ssError << "AceVirusTi: Error while setting the synth into sequencer mode." << endl;
    ssError << "Error: " << err.what() << endl;
  }

  if(!ssError.str().empty())
    DebugTrace("Error: %s", ssError.str().c_str());
}

int AceVirusTi::GetDeviceId(bool bInput)
{
  if(bInput)
  {
    int n = m_MidiInput.getPortCount();

    for(int i = 0; i < n; ++i)
    {
      string const sName = m_MidiOutput.getPortName(i);

      if(m_sInputDeviceName == sName)
        return i;
    }
  }
  else
  {
    int n = m_MidiOutput.getPortCount();

    for(int i = 0; i < n; ++i)
    {
      string const sName = m_MidiOutput.getPortName(i);

      if(m_sOutputDeviceName == sName)
        return i;
    }
  }

  return -1;
}

bool AceVirusTi::PullSysex(vector<uint8_t> & /*sysex_from_virus*/)
{
  return false;
}

void AceVirusTi::midiInputCallback(double /*timeStamp*/, vector<uint8_t> const& message)
{
  {
    stringstream ss;
    ss << "Received MIDI (" << message.size() << " bytes): ";

    for(int i = 0; i < (int)message.size(); ++i)
    {
      if(i % 16 == 0)
      {
        ss << endl;
        char buffer[16];
        sprintf(buffer, "%.8X: ", i);
        ss << buffer;
      }

      char buffer[16];
      sprintf(buffer, "%.2X ", message[i]);
      ss << buffer;
    }

    DebugTrace("%s\n", ss.str().c_str());
  }

  m_MidiVirus.ProcessMidi(&message[0], message.size());
}

void AceVirusTi::OnVirusParamChanged(int iPart, int iOffset, int iValue)
{
  auto i_index = g_OffsetToIndex.find(iOffset);

  if(i_index == g_OffsetToIndex.end())
  {
    DebugTrace("[Part %d] Virus sets unknown parameter (%d) to %d\n", iPart, iOffset, iValue);
    return;
  }

  auto const desc = g_PatchParameters[i_index->second];
  DebugTrace("[Part %d] Virus says %s = %d\n", iPart, desc.m_sName, iValue);
  InternalSetParam(i_index->second, iValue / 127.0f, false, true);
}

void AceVirusTi::WriteMidi(uint8_t const* pData, int iLength)
{
  if(iLength == 0)
    return;

  vector<uint8_t> msg(iLength);
  memcpy(&msg[0], pData, iLength);

  {
    stringstream ss;
    ss << "Write MIDI: ";

    for(size_t i = 0; i < msg.size(); ++i)
    {
      char buffer[16];
      sprintf(buffer, "%.2X ", msg[i]);
      ss << buffer;
    }

    DebugTrace("%s\n", ss.str().c_str());
  }
  try
  {
    std::lock_guard<std::recursive_mutex> lock_it(m_PatchLock);

    if(m_MidiOutput.isOpen())
      m_MidiOutput.sendMessage(&msg);
  }
  catch(exception const& err)
  {
    DebugTrace("Write MIDI failed : %s\n", err.what());
  }
}

void AceVirusTi::OnPatchReceived(int iPart, VirusPatch const& p)
{
  if(m_PartNumber != iPart)
    return;

  SetPatch(p);

  for(int i = 0; i < ArrayLength(g_PatchParameters); ++i)
  {
    int iBase = (g_PatchParameters[i].m_Page - 'A') * 128;
    int iOffset = g_PatchParameters[i].m_iOffset + iBase + 9;

    if(iOffset >= (int)p.m_vSysex.size())
    {
      DebugTrace("Skipping absent parameter value for %s\n", g_PatchParameters[i].m_sName);
      continue;
    }

    int iMidiValue = p.m_vSysex[iOffset];
    float fValue = float(iMidiValue) / 127.0f;
    DebugTrace("In this patch: %s=%.3f(%d)(@%.2X)\n", g_PatchParameters[i].m_sName, fValue, iMidiValue, iOffset);

    if(fValue != fValue)
      DebugTrace("NAN\n");

    InternalSetParam(i, fValue, false, true);
  }

  DebugTrace("[Part %d] Received patch : %s\n", iPart, p.GetName().c_str());
}

void AceVirusTi::Flush()
{
}

bool AceVirusTi::DownloadPatch()
{
  m_MidiVirus.RequestSingleDump(m_PartNumber);
  return true;
}

bool AceVirusTi::UploadPatch()
{
  return m_MidiVirus.UploadPatch(m_PartNumber, m_Patch);
}

unsigned int const MAGIC = 0xACE01337;

vector<uint8_t> AceVirusTi::saveState() const
{
  std::lock_guard<std::recursive_mutex> lock_it(m_PatchLock);
  vector<uint8_t> data;
  obytestream bs(data);
  bs.write_string(m_sInputDeviceName);
  bs.write_string(m_sOutputDeviceName);
  bs.write_uint32(m_PartNumber);
  bs.write_vector(m_Patch.m_vSysex);
  bs.write_uint32(MAGIC);
  return data;
}

void AceVirusTi::loadState(vector<uint8_t> const& data)
{
  std::lock_guard<std::recursive_mutex> lock_it(m_PatchLock);
  try
  {
    ibytestream bs(data);
    m_sInputDeviceName = bs.read_string();
    m_sOutputDeviceName = bs.read_string();
    m_PartNumber = bs.read_uint32();
    auto const sysex = bs.read_vector();

    if(!sysex.empty())
      m_Patch.Parse(&sysex[0], sysex.size());

    auto const magic = bs.read_uint32();

    if(magic != MAGIC)
      DebugTrace("Invalid input stream\n");

    DebugTrace("LoadState: patch='%s'\n", m_Patch.GetName().c_str());
  }
  catch(bytestream::end_of_stream const &)
  {
    DebugTrace("Unexpected end of stream\n");
  }
  ReconnectMidi();
  UploadPatch();
}

#include "acevirusti_editor.h"

static
juce::Component* CreateEditor(EVEREST::IPlugin* plugin, EVEREST::IHost* /*host*/)
{
  return new AceVirusTiEditor(dynamic_cast<AceVirusTi*>(plugin));
}

static
EVEREST::IPlugin* CreatePlugin(EVEREST::IHost* pHost)
{
  return new AceVirusTi(pHost);
}

namespace EVEREST
{
const PluginDescription g_PlugDesc =
{
  true, // isSynth
  { AceVirusTiEditor::WIDTH, AceVirusTiEditor::HEIGHT },
  "Ace",
  "AceVirusTi",
  &CreatePlugin,
  &CreateEditor,
};
}

