/*
 * Copyright (C) 2017 - Sebastien Alaiwan
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 */
#include <vector>
#include <string>
#include <sstream>
#include <cassert>
#include <windows.h>
#include <cstdlib>
#include "vst2_interface.h"
#include "debug.h"

template<typename T, int N>
std::string arrayToString(T(&tab)[N])
{
  std::stringstream ss;

  for(size_t i = 0; i < N; ++i)
  {
    if(i != 0)
      ss << ", ";

    ss << tab[i];
  }

  return ss.str();
}

template<int N>
std::string arrayToString(char(&tab)[N])
{
  std::stringstream ss;
  ss << "{ ";

  for(size_t i = 0; i < N; ++i)
  {
    if(i != 0)
      ss << ", ";

    if(isprint(tab[i]))
      ss << "'" << tab[i] << "'";
    else
      ss << std::hex << int(tab[i]);
  }

  ss << " }";
  return ss.str();
}

HostControlFunction m_pHostCtrlFunction;

class ApiMonitorPlugin
{
private:
  HMODULE m_hGoatModule;
  IVst2Plugin* m_pGoatInterface;

  static int VSTAPI hostControlFunction(IVst2Plugin* pThis, long int a, long int b, long int c, void* d, float e)
  {
    long int ret = m_pHostCtrlFunction(pThis, a, b, c, d, e);
    return ret;
  }

public:
  ApiMonitorPlugin(HostControlFunction ctrlFct)
  {
    m_pHostCtrlFunction = ctrlFct;
    // m_hGoatModule = ::LoadLibrary("Fruity Free Filter.dll");
    // m_hGoatModule = ::LoadLibrary("F:\\Program Files\\VstPlugins\\Sakura.dll");
    m_hGoatModule = ::LoadLibrary("F:\\Program Files\\VstPlugins\\Vanguard\\Vanguard.dll");
    // m_hGoatModule = ::LoadLibrary("H:\\Projects\\projects\\prog\\AudioPlugins\\acevirus_master_build\\bin\\Debug\\AceEchidna.dll");
    assert(m_hGoatModule != NULL);

    if(!m_hGoatModule)
      return;

    typedef IVst2Plugin* (VSTAPI * PluginCreationFunction)(HostControlFunction ctrlFct);

    PluginCreationFunction goatCreationFct = reinterpret_cast<PluginCreationFunction>(::GetProcAddress(m_hGoatModule, "VSTPluginMain"));

    if(!goatCreationFct)
      goatCreationFct = reinterpret_cast<PluginCreationFunction>(::GetProcAddress(m_hGoatModule, "main")); // "Fruity Free Filter" only exports "main".

    assert(goatCreationFct != NULL);

    if(!goatCreationFct)
    {
      CloseHandle(m_hGoatModule);
      return;
    }

    m_pGoatInterface = goatCreationFct(hostControlFunction);

    memcpy(&m_Vst2Interface, m_pGoatInterface, sizeof m_Vst2Interface);

    {
      std::stringstream s;
      s << "Magic: " << arrayToString(m_Vst2Interface.magic) << std::endl;
      s << "Flags: 0x" << std::hex << m_Vst2Interface.flags << std::endl;
      s << "numPrograms = " << m_Vst2Interface.numPrograms << std::endl;
      s << "numParams = " << m_Vst2Interface.numParams << std::endl;
      s << "numInputs = " << m_Vst2Interface.numInputs << std::endl;
      DebugTrace("%s", s.str().c_str());
    }

    {
      std::stringstream s;
      s << "numOutputs = " << m_Vst2Interface.numOutputs << std::endl;
      // s << "somedata = " << m_Vst2Interface.somedata << std::endl;
      DebugTrace("%s", s.str().c_str());
    }

    {
      std::stringstream s;
      s << "empty3=" << arrayToString(m_Vst2Interface.empty3) << std::endl;
      DebugTrace("%s", s.str().c_str());
    }

    {
      std::stringstream s;
      s << "unkown_float = " << m_Vst2Interface.unkown_float << std::endl;
      DebugTrace("%s", s.str().c_str());
    }

    {
      std::stringstream s;
      s << "empty3 = " << arrayToString(m_Vst2Interface.empty3) << std::endl;
      DebugTrace("%s", s.str().c_str());
    }

    {
      std::stringstream s;
      s << "userdata = " << m_Vst2Interface.userdata << std::endl;
      s << "plugin_id=" << arrayToString(m_Vst2Interface.plugin_id) << std::endl;
      s << "unknown1=" << arrayToString(m_Vst2Interface.unknown1) << std::endl;
      DebugTrace("%s", s.str().c_str());
    }

    m_Vst2Interface.dispatcher = static_dispatcher;
    m_Vst2Interface.getParameter = static_getParameter;
    m_Vst2Interface.setParameter = static_setParameter;
    m_Vst2Interface.process = static_process;
    m_Vst2Interface.processReplacing = static_processReplacing;

    m_Vst2Interface.userdata = this;
  }

  virtual ~ApiMonitorPlugin()
  {
    CloseHandle(m_hGoatModule);
  }

  static int static_dispatcher(IVst2Plugin* pVstInterface, int a, int b, int c, void* d, float e)
  {
    ApiMonitorPlugin* pWrapper = reinterpret_cast<ApiMonitorPlugin*>(pVstInterface->userdata);
    return pWrapper->Dispatcher(VST_CMD(a), b, c, d, e);
  }

  static void static_setParameter(IVst2Plugin* pVstInterface, int iParamId, float fValue)
  {
    ApiMonitorPlugin* pWrapper = reinterpret_cast<ApiMonitorPlugin*>(pVstInterface->userdata);
    pWrapper->SetParameter(iParamId, fValue);
  }

  static float static_getParameter(IVst2Plugin* pVstInterface, int iParamId)
  {
    ApiMonitorPlugin* pWrapper = reinterpret_cast<ApiMonitorPlugin*>(pVstInterface->userdata);
    return pWrapper->GetParameter(iParamId);
  }

  static void static_process(IVst2Plugin* pVstInterface, float** pInput, float** pOutput, int iNumSamples)
  {
    ApiMonitorPlugin* pWrapper = reinterpret_cast<ApiMonitorPlugin*>(pVstInterface->userdata);
    pWrapper->Process(pInput, pOutput, iNumSamples);
  }

  static void static_processReplacing(IVst2Plugin* pVstInterface, float** pInput, float** pOutput, int iNumSamples)
  {
    ApiMonitorPlugin* pWrapper = reinterpret_cast<ApiMonitorPlugin*>(pVstInterface->userdata);
    pWrapper->Process(pInput, pOutput, iNumSamples);
  }

  int Dispatcher(VST_CMD eRequest, int iParam1, int iParam2, void* pData, float fParam)
  {
    std::string sPointee;

    std::stringstream s;
    s << ToString(eRequest) << " " << iParam1 << " " << iParam2 << std::hex << " 0x" << (unsigned int)pData << " " << sPointee << " " << fParam;

    int ret = m_pGoatInterface->dispatcher(m_pGoatInterface, eRequest, iParam1, iParam2, pData, fParam);

    s << " -> " << ret << std::endl;

    if(eRequest != VST_CMD_PROCESSEVENTS && eRequest != 19)
      DebugTrace(s.str().c_str());

    return ret;
  }

  void SetParameter(int iParamIdx, float fValue)
  {
    std::stringstream s;
    s << "setParameter " << iParamIdx << " " << fValue << std::endl;
    DebugTrace(s.str().c_str());

    m_pGoatInterface->setParameter(m_pGoatInterface, iParamIdx, fValue);
  }

  float GetParameter(int iParamIdx)
  {
    float fValue = m_pGoatInterface->getParameter(m_pGoatInterface, iParamIdx);

    std::stringstream s;
    s << "getParameter " << iParamIdx << " " << fValue << std::endl;
    DebugTrace(s.str().c_str());

    return fValue;
  }

  void Process(float** pInput, float** pOutput, int iNumSamples)
  {
    m_pGoatInterface->processReplacing(m_pGoatInterface, pInput, pOutput, iNumSamples);
  }

  IVst2Plugin m_Vst2Interface;
};

extern "C"
void* VSTPluginMain(HostControlFunction ctrlFct)
{
  ApiMonitorPlugin* pWrapper = new ApiMonitorPlugin(ctrlFct);
  return &pWrapper->m_Vst2Interface;
}

