/**
 * @file virus_editor.h
 * @brief Definition of the GUI part of the plugin.
 * @author Sebastien Alaiwan
 * @version 1.0
 * @date 2010-09-15
 */

/* Copyright (C)
 * 2010 - Sebastien Alaiwan
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <cassert>
#include "virus_editor.h"
#include "acevirus.h"
#include "math_tools.h"
#include "resource.h"
#include "../revision.h"
#include "lib_utils/tools.h"

VirusEditor::VirusEditor(AceVirus* pEffect) :
  m_pVirusPlugin(pEffect)
{
  for(int i = 0; i < 16; ++i)
  {
    m_bPatchValid[i] = true;
  }

  setBounds(0, 0, WIDTH, HEIGHT);
  // setSize(WIDTH, HEIGHT);

  int line = 10;

  for(int i = 0; i < 16; ++i)
  {
    int col = 10;

    juce::TextEditor* pEdit = new juce::TextEditor;
    pEdit->setReadOnly(true);
    pEdit->setBounds(col, line, 90, 20);
    addAndMakeVisible(pEdit);
    m_pEdit[i] = pEdit;
    col += 100;

    juce::Button* pButton = new juce::TextButton("yo");
    pButton->setButtonText("Recv");
    pButton->setBounds(col, line, 50, 20);
    AddButton(pButton, IDC_BUTTON_PATCH1_SAVE + i);
    col += 60;

    juce::Button* pButton2 = new juce::TextButton("yo2");
    pButton2->setTopLeftPosition(10, 10);
    pButton2->setButtonText("Send");
    pButton2->setBounds(col, line, 50, 20);
    AddButton(pButton2, IDC_BUTTON_PATCH1_LOAD + i);
    col += 60;

    juce::Button* pExportButton = new juce::TextButton("yo2");
    pExportButton->setTopLeftPosition(10, 10);
    pExportButton->setButtonText("Export");
    pExportButton->setBounds(col, line, 50, 20);
    AddButton(pExportButton, IDC_BUTTON_PATCH1_EXPORT + i);
    col += 60;

    juce::Button* pImportButton = new juce::TextButton("yo2");
    pImportButton->setTopLeftPosition(10, 10);
    pImportButton->setButtonText("Import");
    pImportButton->setBounds(col, line, 50, 20);
    AddButton(pImportButton, IDC_BUTTON_PATCH1_IMPORT + i);
    col += 60;

    line += 24;
  }

  {
    juce::DrawableText* pText = new juce::DrawableText;
    pText->setBoundingBox
    (
      RelativeParallelogram
      (
        RelativePoint(10.0f, (float)line),
        RelativePoint(115.0f, (float)line),
        RelativePoint(10.0f, (float)line + 25.0f)
      )
    );
    pText->setFont(Font(16.0f), true);
    pText->setText("Virus MIDI input:");
    addAndMakeVisible(pText);

    std::vector<std::string> input_devices = m_pVirusPlugin->GetInputDeviceList();

    juce::ComboBox* pCombo = new juce::ComboBox("yo3");

    for(size_t i = 0; i < input_devices.size(); ++i)
    {
      pCombo->addItem(juce::String::fromUTF8(input_devices[i].c_str()), i + 1);
    }

    pCombo->setTopLeftPosition(10, 10);
    pCombo->setBounds(120, line, 200, 20);
    pCombo->addListener(this);
    addAndMakeVisible(pCombo);
    m_pCombo[0] = pCombo;
  }

  line += 24;

  {
    juce::DrawableText* pText = new juce::DrawableText;
    pText->setBoundingBox
    (
      RelativeParallelogram
      (
        RelativePoint(10.0f, (float)line),
        RelativePoint(115.0f, (float)line),
        RelativePoint(10.0f, (float)line + 25.0f)
      )
    );
    pText->setFont(Font(16.0f), true);
    pText->setText("Virus MIDI output:");
    addAndMakeVisible(pText);

    std::vector<std::string> devices = m_pVirusPlugin->GetOutputDeviceList();

    juce::ComboBox* pCombo = new juce::ComboBox("yo3");

    for(size_t i = 0; i < devices.size(); ++i)
    {
      pCombo->addItem(juce::String::fromUTF8(devices[i].c_str()), i + 1);
    }

    pCombo->setTopLeftPosition(10, 10);
    pCombo->setBounds(120, line, 200, 20);
    pCombo->addListener(this);
    addAndMakeVisible(pCombo);
    m_pCombo[1] = pCombo;
  }

  line += 24;

  {
    juce::DrawableText* pText = new juce::DrawableText;
    pText->setBoundingBox
    (
      RelativeParallelogram
      (
        RelativePoint(10.0f, (float)line),
        RelativePoint(150.0f, (float)line),
        RelativePoint(10.0f, (float)line + 25.0f)
      )
    );
    pText->setFont(Font(16.0f), true);
#define QUOTEME_(x) # x
#define QUOTEME(x) QUOTEME_(x)
    pText->setText("Ace Midivirus r" QUOTEME(REVISION));
    addAndMakeVisible(pText);
  }

  m_pVirusPlugin->RegisterEditor(this);
}

void VirusEditor::buttonClicked(juce::Button* button)
{
  int iId = m_ButtonToId[button];
  OnPressButton(iId);
}

void VirusEditor::comboBoxChanged(juce::ComboBox* cb)
{
  int idx = cb->getSelectedItemIndex();
  std::string s(cb->getItemText(idx).getCharPointer());

  if(cb == m_pCombo[0])
  {
    m_pVirusPlugin->ChangeMidiInputDevice(s);
    ProtectedRefresh();
  }
  else if(cb == m_pCombo[1])
  {
    m_pVirusPlugin->ChangeMidiOutputDevice(s);
    ProtectedRefresh();
  }
}

void VirusEditor::AddButton(juce::Button* pButton, int iID)
{
  m_ButtonToId[pButton] = iID;
  m_IdToButton[iID] = pButton;
  pButton->addListener(this);
  addAndMakeVisible(pButton);
}

VirusEditor::~VirusEditor()
{
  m_pVirusPlugin->UnregisterEditor(this);
  this->deleteAllChildren();
}

void VirusEditor::paint(juce::Graphics& g)
{
  g.fillAll(juce::Colour(0xff303040));
  ProtectedRefresh();
/*
   g.setFont(10.0);
   g.setColour(juce::Colour(255, 255, 255));
   g.drawText("AceVirus", 10, 120, 200, 40, juce::Justification::centred, false);
   g.setColour(juce::Colour(255, 0, 0));
   g.drawText("AceVirus", 10+1, 120+1, 200, 40, juce::Justification::centred, false);
 */
};

static std::string DoFileDialog(const char* sFilter, bool bSave, bool& bIsMidi)
{
  unsigned int flags = juce::FileBrowserComponent::canSelectFiles;

  if(bSave)
    flags |= juce::FileBrowserComponent::saveMode;
  else
    flags |= juce::FileBrowserComponent::openMode;

  juce::WildcardFileFilter filter(sFilter, "", "");
  juce::FileBrowserComponent browser(flags, juce::File::nonexistent, &filter, nullptr);

  FileChooserDialogBox dialogBox(T("Export patch"), T("Choose a file name for your patch."), browser, true, Colour(0xffb8b8b8));

  if(dialogBox.show())
  {
    juce::File f = browser.getSelectedFile(0);
    bIsMidi = f.getFileExtension().compareIgnoreCase(".mid") == 0;
    std::string s(f.getFullPathName().getCharPointer());
    return s;
  }

  return "";
}

void VirusEditor::SavePatch(int iPartNumber)
{
  m_pEdit[iPartNumber]->setText("< receiving ... >");

  bool bOk = m_pVirusPlugin->DownloadPatch(iPartNumber);

  if(bOk)
    m_bPatchValid[iPartNumber] = true;
  else
    m_bPatchValid[iPartNumber] = false;

  ProtectedRefresh();
}

void VirusEditor::LoadPatch(int iPartNumber)
{
  m_pVirusPlugin->UploadPatch(iPartNumber);
}

void VirusEditor::ExportPatch(int iPartNumber)
{
  bool bIsMidi;
  std::string sPath = DoFileDialog("Raw sysex (*.viruspatch)\0*.viruspatch\0Midi (*.mid)\0*.mid\0\0", true, bIsMidi);

  if(sPath.empty())
    return;

  VirusPatch p = m_pVirusPlugin->GetPatch(iPartNumber);

  if(bIsMidi)
  {
    if(!p.ExportMidiFile(sPath))
    {
      AlertWindow::showMessageBox(AlertWindow::WarningIcon, T("Error"), T("Failed saving file."));
      return;
    }
  }
  else if(!p.Export(sPath))
  {
    AlertWindow::showMessageBox(AlertWindow::WarningIcon, T("Error"), T("Failed saving file."));
    return;
  }
}

void VirusEditor::ImportPatch(int iPartNumber)
{
  bool bIsMidi;
  std::string sPath = DoFileDialog("Raw sysex (*.viruspatch)\0*.viruspatch", false, bIsMidi);

  if(sPath.empty())
    return;

  VirusPatch p;

  if(!p.Import(sPath))
  {
    AlertWindow::showMessageBox(AlertWindow::WarningIcon, T("Error"), T("Failed importing file."));
    return;
  }

  m_pVirusPlugin->SetPatch(iPartNumber, p);

  ProtectedRefresh();
}

void VirusEditor::OnPressButton(int id)
{
  if(id >= IDC_BUTTON_PATCH1_SAVE && id <= IDC_BUTTON_PATCH16_SAVE)
  {
    int iPartNumber = id - IDC_BUTTON_PATCH1_SAVE;
    SavePatch(iPartNumber);
  }
  else if(id >= IDC_BUTTON_PATCH1_LOAD && id <= IDC_BUTTON_PATCH16_LOAD)
  {
    int iPartNumber = id - IDC_BUTTON_PATCH1_LOAD;
    LoadPatch(iPartNumber);
  }
  else if(id >= IDC_BUTTON_PATCH1_EXPORT && id <= IDC_BUTTON_PATCH16_EXPORT)
  {
    int iPartNumber = id - IDC_BUTTON_PATCH1_EXPORT;
    ExportPatch(iPartNumber);
  }
  else if(id >= IDC_BUTTON_PATCH1_IMPORT && id <= IDC_BUTTON_PATCH16_IMPORT)
  {
    int iPartNumber = id - IDC_BUTTON_PATCH1_IMPORT;
    ImportPatch(iPartNumber);
  }
  else if(id == IDC_EXPORT_AS_BANK)
  {
    int iPartNumber = id - IDC_BUTTON_PATCH1_EXPORT;

    bool bIsMidi;
    std::string sPath = DoFileDialog("Midi Bank (*.mid)\0*.mid\0\0", true, bIsMidi);

    if(sPath.empty())
      return;

    VirusPatch patches[16];

    for(int i = 0; i < 16; ++i)
    {
      patches[i] = m_pVirusPlugin->GetPatch(i);
    }

    VirusPatch::ExportMidiFile(sPath, patches, 16);
  }
}

static void ComboSelectItemByText(juce::ComboBox* p, std::string const& sText)
{
  int const n = p->getNumItems();

  for(int i = 0; i < n; ++i)
  {
    std::string s(p->getItemText(i).getCharPointer());

    if(s == sText)
    {
      p->setSelectedItemIndex(i);
      return;
    }
  }
}

void VirusEditor::ProtectedRefresh()
{
  for(int i = 0; i < 16; ++i)
  {
    if(m_bPatchValid[i])
    {
      VirusPatch p = m_pVirusPlugin->GetPatch(i);
      m_pEdit[i]->setText(p.GetName().c_str());
    }
    else
      m_pEdit[i]->setText("< failed >");
  }

  // without input, we can still send patches to the virus
  int iButtonLoadIds[] =
  {
    IDC_BUTTON_PATCH1_LOAD,
    IDC_BUTTON_PATCH2_LOAD,
    IDC_BUTTON_PATCH3_LOAD,
    IDC_BUTTON_PATCH4_LOAD,
    IDC_BUTTON_PATCH5_LOAD,
    IDC_BUTTON_PATCH6_LOAD,
    IDC_BUTTON_PATCH7_LOAD,
    IDC_BUTTON_PATCH8_LOAD,
    IDC_BUTTON_PATCH9_LOAD,
    IDC_BUTTON_PATCH10_LOAD,
    IDC_BUTTON_PATCH11_LOAD,
    IDC_BUTTON_PATCH12_LOAD,
    IDC_BUTTON_PATCH13_LOAD,
    IDC_BUTTON_PATCH14_LOAD,
    IDC_BUTTON_PATCH15_LOAD,
    IDC_BUTTON_PATCH16_LOAD,
  };

  for(size_t i = 0; i < ArrayLength(iButtonLoadIds); ++i)
  {
    juce::Button* pButton = m_IdToButton[iButtonLoadIds[i]];
    assert(pButton);
    pButton->setEnabled(m_pVirusPlugin->IsOutputOK());
  }

  // without output, we can't communicate with the virus.

  int iButtonSaveIds[] =
  {
    IDC_BUTTON_PATCH1_SAVE,
    IDC_BUTTON_PATCH2_SAVE,
    IDC_BUTTON_PATCH3_SAVE,
    IDC_BUTTON_PATCH4_SAVE,
    IDC_BUTTON_PATCH5_SAVE,
    IDC_BUTTON_PATCH6_SAVE,
    IDC_BUTTON_PATCH7_SAVE,
    IDC_BUTTON_PATCH8_SAVE,
    IDC_BUTTON_PATCH9_SAVE,
    IDC_BUTTON_PATCH10_SAVE,
    IDC_BUTTON_PATCH11_SAVE,
    IDC_BUTTON_PATCH12_SAVE,
    IDC_BUTTON_PATCH13_SAVE,
    IDC_BUTTON_PATCH14_SAVE,
    IDC_BUTTON_PATCH15_SAVE,
    IDC_BUTTON_PATCH16_SAVE,
  };

  for(size_t i = 0; i < ArrayLength(iButtonSaveIds); ++i)
  {
    juce::Button* pButton = m_IdToButton[iButtonSaveIds[i]];
    assert(pButton);
    pButton->setEnabled(m_pVirusPlugin->IsInputOK() && m_pVirusPlugin->IsOutputOK());
  }

  ComboSelectItemByText(m_pCombo[0], m_pVirusPlugin->GetMidiInputDeviceName());
  ComboSelectItemByText(m_pCombo[1], m_pVirusPlugin->GetMidiOutputDeviceName());
}

