/*
  ==============================================================================

   This file is part of the JUCE library - "Jules' Utility Class Extensions"
   Copyright 2004-9 by Raw Material Software Ltd.

  ------------------------------------------------------------------------------

   JUCE can be redistributed and/or modified under the terms of the GNU General
   Public License (Version 2), as published by the Free Software Foundation.
   A copy of the license is included in the JUCE distribution, or can be found
   online at www.gnu.org/licenses.

   JUCE is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  ------------------------------------------------------------------------------

   To release a closed-source product which uses JUCE, commercial licenses are
   available: visit www.rawmaterialsoftware.com/juce for more information.

  ==============================================================================
*/

#ifdef __JUCE_JUCEHEADER__
 /* When you add the amalgamated cpp file to your project, you mustn't include it in
	a file where you've already included juce.h - just put it inside a file on its own,
	possibly with your config flags preceding it, but don't include anything else. */
 #error
#endif


#include "juce_TargetPlatform.h"

 // FORCE_AMALGAMATOR_INCLUDE

#include "juce_Config.h"

  // FORCE_AMALGAMATOR_INCLUDE

#ifndef JUCE_BUILD_CORE
 #define JUCE_BUILD_CORE 1
#endif
#ifndef JUCE_BUILD_MISC
 #define JUCE_BUILD_MISC 1
#endif
#ifndef JUCE_BUILD_GUI
 #define JUCE_BUILD_GUI 1
#endif
#ifndef JUCE_BUILD_NATIVE
 #define JUCE_BUILD_NATIVE 1
#endif

#if JUCE_ONLY_BUILD_CORE_LIBRARY
 #undef JUCE_BUILD_MISC
 #undef JUCE_BUILD_GUI
#endif

#ifndef JUCE_NAMESPACE
 #define JUCE_NAMESPACE juce
#endif

//==============================================================================
#include "juce_NativeIncludes.h"
//==============================================================================
#define DONT_SET_USING_JUCE_NAMESPACE 1
#undef max
#undef min
#define NO_DUMMY_DECL

#define JUCE_AMALGAMATED_TEMPLATE 1

#if JUCE_BUILD_NATIVE
 #include "juce_amalgamated.h"  // FORCE_AMALGAMATOR_INCLUDE
#endif

#if (defined(_MSC_VER) && (_MSC_VER <= 1200))
 #pragma warning (disable: 4309 4305)
#endif


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////









//==============================================================================
#if JUCE_BUILD_CORE






/*** Start of inlined file: juce_Random.cpp ***/
BEGIN_JUCE_NAMESPACE

Random::Random (const int64 seedValue) noexcept
	: seed (seedValue)
{
}

Random::Random()
	: seed (1)
{
	setSeedRandomly();
}

Random::~Random() noexcept
{
}

void Random::setSeed (const int64 newSeed) noexcept
{
	seed = newSeed;
}

void Random::combineSeed (const int64 seedValue) noexcept
{
	seed ^= nextInt64() ^ seedValue;
}

void Random::setSeedRandomly()
{
	combineSeed ((int64) (pointer_sized_int) this);
	combineSeed (Time::getMillisecondCounter());
	combineSeed (Time::getHighResolutionTicks());
	combineSeed (Time::getHighResolutionTicksPerSecond());
	combineSeed (Time::currentTimeMillis());
}

Random& Random::getSystemRandom() noexcept
{
	static Random sysRand;
	return sysRand;
}

int Random::nextInt() noexcept
{
	seed = (seed * literal64bit (0x5deece66d) + 11) & literal64bit (0xffffffffffff);

	return (int) (seed >> 16);
}

int Random::nextInt (const int maxValue) noexcept
{
	jassert (maxValue > 0);
	return (nextInt() & 0x7fffffff) % maxValue;
}

int64 Random::nextInt64() noexcept
{
	return (((int64) nextInt()) << 32) | (int64) (uint64) (uint32) nextInt();
}

bool Random::nextBool() noexcept
{
	return (nextInt() & 0x80000000) != 0;
}

float Random::nextFloat() noexcept
{
	return static_cast <uint32> (nextInt()) / (float) 0xffffffff;
}

double Random::nextDouble() noexcept
{
	return static_cast <uint32> (nextInt()) / (double) 0xffffffff;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Random.cpp ***/


/*** Start of inlined file: juce_RelativeTime.cpp ***/
BEGIN_JUCE_NAMESPACE

RelativeTime::RelativeTime (const double seconds_) noexcept
	: seconds (seconds_)
{
}

RelativeTime::RelativeTime (const RelativeTime& other) noexcept
	: seconds (other.seconds)
{
}

RelativeTime::~RelativeTime() noexcept
{
}

const RelativeTime RelativeTime::milliseconds (const int milliseconds) noexcept   { return RelativeTime (milliseconds * 0.001); }
const RelativeTime RelativeTime::milliseconds (const int64 milliseconds) noexcept { return RelativeTime (milliseconds * 0.001); }
const RelativeTime RelativeTime::minutes (const double numberOfMinutes) noexcept  { return RelativeTime (numberOfMinutes * 60.0); }
const RelativeTime RelativeTime::hours (const double numberOfHours) noexcept	  { return RelativeTime (numberOfHours * (60.0 * 60.0)); }
const RelativeTime RelativeTime::days (const double numberOfDays) noexcept	{ return RelativeTime (numberOfDays  * (60.0 * 60.0 * 24.0)); }
const RelativeTime RelativeTime::weeks (const double numberOfWeeks) noexcept	  { return RelativeTime (numberOfWeeks * (60.0 * 60.0 * 24.0 * 7.0)); }

int64 RelativeTime::inMilliseconds() const noexcept { return (int64) (seconds * 1000.0); }
double RelativeTime::inMinutes() const noexcept	 { return seconds / 60.0; }
double RelativeTime::inHours() const noexcept	   { return seconds / (60.0 * 60.0); }
double RelativeTime::inDays() const noexcept	{ return seconds / (60.0 * 60.0 * 24.0); }
double RelativeTime::inWeeks() const noexcept	   { return seconds / (60.0 * 60.0 * 24.0 * 7.0); }

String RelativeTime::getDescription (const String& returnValueForZeroTime) const
{
	if (seconds < 0.001 && seconds > -0.001)
		return returnValueForZeroTime;

	String result;
	result.preallocateBytes (32);

	if (seconds < 0)
		result << '-';

	int fieldsShown = 0;
	int n = std::abs ((int) inWeeks());
	if (n > 0)
	{
		result << n << (n == 1 ? TRANS(" week ")
							   : TRANS(" weeks "));
		++fieldsShown;
	}

	n = std::abs ((int) inDays()) % 7;
	if (n > 0)
	{
		result << n << (n == 1 ? TRANS(" day ")
							   : TRANS(" days "));
		++fieldsShown;
	}

	if (fieldsShown < 2)
	{
		n = std::abs ((int) inHours()) % 24;
		if (n > 0)
		{
			result << n << (n == 1 ? TRANS(" hr ")
								   : TRANS(" hrs "));
			++fieldsShown;
		}

		if (fieldsShown < 2)
		{
			n = std::abs ((int) inMinutes()) % 60;
			if (n > 0)
			{
				result << n << (n == 1 ? TRANS(" min ")
									   : TRANS(" mins "));
				++fieldsShown;
			}

			if (fieldsShown < 2)
			{
				n = std::abs ((int) inSeconds()) % 60;
				if (n > 0)
				{
					result << n << (n == 1 ? TRANS(" sec ")
										   : TRANS(" secs "));
					++fieldsShown;
				}

				if (fieldsShown == 0)
				{
					n = std::abs ((int) inMilliseconds()) % 1000;
					if (n > 0)
						result << n << TRANS(" ms");
				}
			}
		}
	}

	return result.trimEnd();
}

RelativeTime& RelativeTime::operator= (const RelativeTime& other) noexcept
{
	seconds = other.seconds;
	return *this;
}

const RelativeTime& RelativeTime::operator+= (const RelativeTime& timeToAdd) noexcept
{
	seconds += timeToAdd.seconds;
	return *this;
}

const RelativeTime& RelativeTime::operator-= (const RelativeTime& timeToSubtract) noexcept
{
	seconds -= timeToSubtract.seconds;
	return *this;
}

const RelativeTime& RelativeTime::operator+= (const double secondsToAdd) noexcept
{
	seconds += secondsToAdd;
	return *this;
}

const RelativeTime& RelativeTime::operator-= (const double secondsToSubtract) noexcept
{
	seconds -= secondsToSubtract;
	return *this;
}

bool operator== (const RelativeTime& t1, const RelativeTime& t2) noexcept { return t1.inSeconds() == t2.inSeconds(); }
bool operator!= (const RelativeTime& t1, const RelativeTime& t2) noexcept { return t1.inSeconds() != t2.inSeconds(); }
bool operator>  (const RelativeTime& t1, const RelativeTime& t2) noexcept { return t1.inSeconds() >  t2.inSeconds(); }
bool operator<  (const RelativeTime& t1, const RelativeTime& t2) noexcept { return t1.inSeconds() <  t2.inSeconds(); }
bool operator>= (const RelativeTime& t1, const RelativeTime& t2) noexcept { return t1.inSeconds() >= t2.inSeconds(); }
bool operator<= (const RelativeTime& t1, const RelativeTime& t2) noexcept { return t1.inSeconds() <= t2.inSeconds(); }

RelativeTime operator+ (const RelativeTime&  t1, const RelativeTime& t2) noexcept   { RelativeTime t (t1); return t += t2; }
RelativeTime operator- (const RelativeTime&  t1, const RelativeTime& t2) noexcept   { RelativeTime t (t1); return t -= t2; }

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RelativeTime.cpp ***/


/*** Start of inlined file: juce_SystemStats.cpp ***/
BEGIN_JUCE_NAMESPACE

const SystemStats::CPUFlags& SystemStats::getCPUFlags()
{
	static CPUFlags cpuFlags;
	return cpuFlags;
}

String SystemStats::getJUCEVersion()
{
	// Some basic tests, to keep an eye on things and make sure these types work ok
	// on all platforms. Let me know if any of these assertions fail on your system!
	static_assert (sizeof (pointer_sized_int) == sizeof (void*), "");
	static_assert (sizeof (int8) == 1, "");
	static_assert (sizeof (uint8) == 1, "");
	static_assert (sizeof (int16) == 2, "");
	static_assert (sizeof (uint16) == 2, "");
	static_assert (sizeof (int32) == 4, "");
	static_assert (sizeof (uint32) == 4, "");
	static_assert (sizeof (int64) == 8, "");
	static_assert (sizeof (uint64) == 8, "");

	// (these confusing macros convert numbers into a single string literal)
	#define JUCE_STRINGIFYVERSION2(a) #a
	#define JUCE_STRINGIFYVERSION(a) JUCE_STRINGIFYVERSION2(a)

	return "JUCE v" JUCE_STRINGIFYVERSION(JUCE_MAJOR_VERSION)
				"." JUCE_STRINGIFYVERSION(JUCE_MINOR_VERSION)
				"." JUCE_STRINGIFYVERSION(JUCE_BUILDNUMBER);

	#undef JUCE_STRINGIFYVERSION
	#undef JUCE_STRINGIFYVERSION2
}

#if JUCE_DEBUG && ! JUCE_ANDROID
 struct JuceVersionPrinter
 {
	 JuceVersionPrinter()
	 {
		 DBG (SystemStats::getJUCEVersion());
	 }
 };

 static JuceVersionPrinter juceVersionPrinter;
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_SystemStats.cpp ***/


/*** Start of inlined file: juce_Result.cpp ***/
BEGIN_JUCE_NAMESPACE

Result::Result (const String& message) noexcept
	: errorMessage (message)
{
}

Result::Result (const Result& other)
	: errorMessage (other.errorMessage)
{
}

Result& Result::operator= (const Result& other)
{
	errorMessage = other.errorMessage;
	return *this;
}

bool Result::operator== (const Result& other) const noexcept
{
	return errorMessage == other.errorMessage;
}

bool Result::operator!= (const Result& other) const noexcept
{
	return errorMessage != other.errorMessage;
}

Result Result::ok() noexcept
{
	return Result (String::empty);
}

Result Result::fail (const String& errorMessage) noexcept
{
	return Result (errorMessage.isEmpty() ? "Unknown Error" : errorMessage);
}

const String& Result::getErrorMessage() const noexcept
{
	return errorMessage;
}

bool Result::wasOk() const noexcept	 { return errorMessage.isEmpty(); }
Result::operator bool() const noexcept	  { return errorMessage.isEmpty(); }
bool Result::failed() const noexcept	{ return errorMessage.isNotEmpty(); }
bool Result::operator!() const noexcept	 { return errorMessage.isNotEmpty(); }

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Result.cpp ***/


/*** Start of inlined file: juce_Time.cpp ***/
#if JUCE_MSVC
  #pragma warning (push)
  #pragma warning (disable: 4514)
#endif

#ifndef JUCE_WINDOWS
  #include <sys/time.h>
#else
  #include <ctime>
#endif

#include <sys/timeb.h>

#if JUCE_MSVC
  #pragma warning (pop)

  #ifdef _INC_TIME_INL
	#define USE_NEW_SECURE_TIME_FNS
  #endif
#endif

BEGIN_JUCE_NAMESPACE

namespace TimeHelpers
{
	struct tm  millisToLocal (const int64 millis) noexcept
	{
		struct tm result;
		const int64 seconds = millis / 1000;

		if (seconds < literal64bit (86400) || seconds >= literal64bit (2145916800))
		{
			// use extended maths for dates beyond 1970 to 2037..
			const int timeZoneAdjustment = 31536000 - (int) (Time (1971, 0, 1, 0, 0).toMilliseconds() / 1000);
			const int64 jdm = seconds + timeZoneAdjustment + literal64bit (210866803200);

			const int days = (int) (jdm / literal64bit (86400));
			const int a = 32044 + days;
			const int b = (4 * a + 3) / 146097;
			const int c = a - (b * 146097) / 4;
			const int d = (4 * c + 3) / 1461;
			const int e = c - (d * 1461) / 4;
			const int m = (5 * e + 2) / 153;

			result.tm_mday  = e - (153 * m + 2) / 5 + 1;
			result.tm_mon   = m + 2 - 12 * (m / 10);
			result.tm_year  = b * 100 + d - 6700 + (m / 10);
			result.tm_wday  = (days + 1) % 7;
			result.tm_yday  = -1;

			int t = (int) (jdm % literal64bit (86400));
			result.tm_hour  = t / 3600;
			t %= 3600;
			result.tm_min   = t / 60;
			result.tm_sec   = t % 60;
			result.tm_isdst = -1;
		}
		else
		{
			time_t now = static_cast <time_t> (seconds);

		  #if JUCE_WINDOWS
		   #ifdef USE_NEW_SECURE_TIME_FNS
			if (now >= 0 && now <= 0x793406fff)
				localtime_s (&result, &now);
			else
				zerostruct (result);
		   #else
			result = *localtime (&now);
		   #endif
		  #else

			localtime_r (&now, &result); // more thread-safe
		  #endif
		}

		return result;
	}

	int extendedModulo (const int64 value, const int modulo) noexcept
	{
		return (int) (value >= 0 ? (value % modulo)
								 : (value - ((value / modulo) + 1) * modulo));
	}

	int doFTime (CharPointer_UTF32 dest, const int maxChars, const String& format, const struct tm* const tm) noexcept
	{
	   #if JUCE_ANDROID
		HeapBlock <char> tempDest;
		tempDest.calloc (maxChars + 2);
		const int result = (int) strftime (tempDest, maxChars, format.toUTF8(), tm);
		if (result > 0)
			dest.writeAll (CharPointer_UTF8 (tempDest.getData()));
		return result;
	   #elif JUCE_WINDOWS
		HeapBlock <wchar_t> tempDest;
		tempDest.calloc (maxChars + 2);
		const int result = (int) wcsftime (tempDest, maxChars, format.toWideCharPointer(), tm);
		if (result > 0)
			dest.writeAll (CharPointer_UTF16 (tempDest.getData()));
		return result;
	   #else
		return (int) wcsftime (dest.getAddress(), maxChars, format.toUTF32(), tm);
	   #endif
	}

	static uint32 lastMSCounterValue = 0;
}

Time::Time() noexcept
	: millisSinceEpoch (0)
{
}

Time::Time (const Time& other) noexcept
	: millisSinceEpoch (other.millisSinceEpoch)
{
}

Time::Time (const int64 ms) noexcept
	: millisSinceEpoch (ms)
{
}

Time::Time (const int year,
			const int month,
			const int day,
			const int hours,
			const int minutes,
			const int seconds,
			const int milliseconds,
			const bool useLocalTime) noexcept
{
	jassert (year > 100); // year must be a 4-digit version

	if (year < 1971 || year >= 2038 || ! useLocalTime)
	{
		// use extended maths for dates beyond 1970 to 2037..
		const int timeZoneAdjustment = useLocalTime ? (31536000 - (int) (Time (1971, 0, 1, 0, 0).toMilliseconds() / 1000))
													: 0;
		const int a = (13 - month) / 12;
		const int y = year + 4800 - a;
		const int jd = day + (153 * (month + 12 * a - 2) + 2) / 5
						   + (y * 365) + (y /  4) - (y / 100) + (y / 400)
						   - 32045;

		const int64 s = ((int64) jd) * literal64bit (86400) - literal64bit (210866803200);

		millisSinceEpoch = 1000 * (s + (hours * 3600 + minutes * 60 + seconds - timeZoneAdjustment))
							 + milliseconds;
	}
	else
	{
		struct tm t;
		t.tm_year   = year - 1900;
		t.tm_mon	= month;
		t.tm_mday   = day;
		t.tm_hour   = hours;
		t.tm_min	= minutes;
		t.tm_sec	= seconds;
		t.tm_isdst  = -1;

		millisSinceEpoch = 1000 * (int64) mktime (&t);

		if (millisSinceEpoch < 0)
			millisSinceEpoch = 0;
		else
			millisSinceEpoch += milliseconds;
	}
}

Time::~Time() noexcept
{
}

Time& Time::operator= (const Time& other) noexcept
{
	millisSinceEpoch = other.millisSinceEpoch;
	return *this;
}

int64 Time::currentTimeMillis() noexcept
{
	static uint32 lastCounterResult = 0xffffffff;
	static int64 correction = 0;

	const uint32 now = getMillisecondCounter();

	// check the counter hasn't wrapped (also triggered the first time this function is called)
	if (now < lastCounterResult)
	{
		// double-check it's actually wrapped, in case multi-cpu machines have timers that drift a bit.
		if (lastCounterResult == 0xffffffff || now < lastCounterResult - 10)
		{
			// get the time once using normal library calls, and store the difference needed to
			// turn the millisecond counter into a real time.
		  #if JUCE_WINDOWS
			struct _timeb t;
		   #ifdef USE_NEW_SECURE_TIME_FNS
			_ftime_s (&t);
		   #else
			_ftime (&t);
		   #endif
			correction = (((int64) t.time) * 1000 + t.millitm) - now;
		  #else
			struct timeval tv;
			struct timezone tz;
			gettimeofday (&tv, &tz);
			correction = (((int64) tv.tv_sec) * 1000 + tv.tv_usec / 1000) - now;
		  #endif
		}
	}

	lastCounterResult = now;

	return correction + now;
}

uint32 juce_millisecondsSinceStartup() noexcept;

uint32 Time::getMillisecondCounter() noexcept
{
	const uint32 now = juce_millisecondsSinceStartup();

	if (now < TimeHelpers::lastMSCounterValue)
	{
		// in multi-threaded apps this might be called concurrently, so
		// make sure that our last counter value only increases and doesn't
		// go backwards..
		if (now < TimeHelpers::lastMSCounterValue - 1000)
			TimeHelpers::lastMSCounterValue = now;
	}
	else
	{
		TimeHelpers::lastMSCounterValue = now;
	}

	return now;
}

uint32 Time::getApproximateMillisecondCounter() noexcept
{
	if (TimeHelpers::lastMSCounterValue == 0)
		getMillisecondCounter();

	return TimeHelpers::lastMSCounterValue;
}

void Time::waitForMillisecondCounter (const uint32 targetTime) noexcept
{
	for (;;)
	{
		const uint32 now = getMillisecondCounter();

		if (now >= targetTime)
			break;

		const int toWait = targetTime - now;

		if (toWait > 2)
		{
			Thread::sleep (jmin (20, toWait >> 1));
		}
		else
		{
			// xxx should consider using mutex_pause on the mac as it apparently
			// makes it seem less like a spinlock and avoids lowering the thread pri.
			for (int i = 10; --i >= 0;)
				Thread::yield();
		}
	}
}

double Time::highResolutionTicksToSeconds (const int64 ticks) noexcept
{
	return ticks / (double) getHighResolutionTicksPerSecond();
}

int64 Time::secondsToHighResolutionTicks (const double seconds) noexcept
{
	return (int64) (seconds * (double) getHighResolutionTicksPerSecond());
}

Time JUCE_CALLTYPE Time::getCurrentTime() noexcept
{
	return Time (currentTimeMillis());
}

String Time::toString (const bool includeDate,
					   const bool includeTime,
					   const bool includeSeconds,
					   const bool use24HourClock) const noexcept
{
	String result;

	if (includeDate)
	{
		result << getDayOfMonth() << ' '
			   << getMonthName (true) << ' '
			   << getYear();

		if (includeTime)
			result << ' ';
	}

	if (includeTime)
	{
		const int mins = getMinutes();

		result << (use24HourClock ? getHours() : getHoursInAmPmFormat())
			   << (mins < 10 ? ":0" : ":") << mins;

		if (includeSeconds)
		{
			const int secs = getSeconds();
			result << (secs < 10 ? ":0" : ":") << secs;
		}

		if (! use24HourClock)
			result << (isAfternoon() ? "pm" : "am");
	}

	return result.trimEnd();
}

String Time::formatted (const String& format) const
{
	int bufferSize = 128;
	HeapBlock<juce_wchar> buffer (128);

	struct tm t (TimeHelpers::millisToLocal (millisSinceEpoch));

	while (TimeHelpers::doFTime (CharPointer_UTF32 (buffer.getData()), bufferSize, format, &t) <= 0)
	{
		bufferSize += 128;
		buffer.malloc (bufferSize);
	}

	return CharPointer_UTF32 (buffer.getData());
}

int Time::getYear() const noexcept	  { return TimeHelpers::millisToLocal (millisSinceEpoch).tm_year + 1900; }
int Time::getMonth() const noexcept	 { return TimeHelpers::millisToLocal (millisSinceEpoch).tm_mon; }
int Time::getDayOfMonth() const noexcept	{ return TimeHelpers::millisToLocal (millisSinceEpoch).tm_mday; }
int Time::getDayOfWeek() const noexcept	 { return TimeHelpers::millisToLocal (millisSinceEpoch).tm_wday; }
int Time::getHours() const noexcept	 { return TimeHelpers::millisToLocal (millisSinceEpoch).tm_hour; }
int Time::getMinutes() const noexcept	   { return TimeHelpers::millisToLocal (millisSinceEpoch).tm_min; }
int Time::getSeconds() const noexcept	   { return TimeHelpers::extendedModulo (millisSinceEpoch / 1000, 60); }
int Time::getMilliseconds() const noexcept  { return TimeHelpers::extendedModulo (millisSinceEpoch, 1000); }

int Time::getHoursInAmPmFormat() const noexcept
{
	const int hours = getHours();

	if (hours == 0)
		return 12;
	else if (hours <= 12)
		return hours;
	else
		return hours - 12;
}

bool Time::isAfternoon() const noexcept
{
	return getHours() >= 12;
}

bool Time::isDaylightSavingTime() const noexcept
{
	return TimeHelpers::millisToLocal (millisSinceEpoch).tm_isdst != 0;
}

String Time::getMonthName (const bool threeLetterVersion) const
{
	return getMonthName (getMonth(), threeLetterVersion);
}

String Time::getWeekdayName (const bool threeLetterVersion) const
{
	return getWeekdayName (getDayOfWeek(), threeLetterVersion);
}

String Time::getMonthName (int monthNumber, const bool threeLetterVersion)
{
	const char* const shortMonthNames[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
	const char* const longMonthNames[]  = { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" };

	monthNumber %= 12;

	return TRANS (threeLetterVersion ? shortMonthNames [monthNumber]
									 : longMonthNames [monthNumber]);
}

String Time::getWeekdayName (int day, const bool threeLetterVersion)
{
	const char* const shortDayNames[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
	const char* const longDayNames[]  = { "Sunday", "Monday", "Tuesday", "Wednesday", "Thursday", "Friday", "Saturday" };

	day %= 7;

	return TRANS (threeLetterVersion ? shortDayNames [day]
									 : longDayNames [day]);
}

Time& Time::operator+= (const RelativeTime& delta)	{ millisSinceEpoch += delta.inMilliseconds(); return *this; }
Time& Time::operator-= (const RelativeTime& delta)	{ millisSinceEpoch -= delta.inMilliseconds(); return *this; }

Time operator+ (const Time& time, const RelativeTime& delta)	{ Time t (time); return t += delta; }
Time operator- (const Time& time, const RelativeTime& delta)	{ Time t (time); return t -= delta; }
Time operator+ (const RelativeTime& delta, const Time& time)	{ Time t (time); return t += delta; }
const RelativeTime operator- (const Time& time1, const Time& time2) { return RelativeTime::milliseconds (time1.toMilliseconds() - time2.toMilliseconds()); }

bool operator== (const Time& time1, const Time& time2)	  { return time1.toMilliseconds() == time2.toMilliseconds(); }
bool operator!= (const Time& time1, const Time& time2)	  { return time1.toMilliseconds() != time2.toMilliseconds(); }
bool operator<  (const Time& time1, const Time& time2)	  { return time1.toMilliseconds() <  time2.toMilliseconds(); }
bool operator>  (const Time& time1, const Time& time2)	  { return time1.toMilliseconds() >  time2.toMilliseconds(); }
bool operator<= (const Time& time1, const Time& time2)	  { return time1.toMilliseconds() <= time2.toMilliseconds(); }
bool operator>= (const Time& time1, const Time& time2)	  { return time1.toMilliseconds() >= time2.toMilliseconds(); }

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Time.cpp ***/


/*** Start of inlined file: juce_AbstractFifo.cpp ***/
BEGIN_JUCE_NAMESPACE

AbstractFifo::AbstractFifo (const int capacity) noexcept
	: bufferSize (capacity)
{
	jassert (bufferSize > 0);
}

AbstractFifo::~AbstractFifo() {}

int AbstractFifo::getTotalSize() const noexcept	   { return bufferSize; }
int AbstractFifo::getFreeSpace() const noexcept	   { return bufferSize - getNumReady(); }

int AbstractFifo::getNumReady() const noexcept
{
	const int vs = validStart.get();
	const int ve = validEnd.get();
	return ve >= vs ? (ve - vs) : (bufferSize - (vs - ve));
}

void AbstractFifo::reset() noexcept
{
	validEnd = 0;
	validStart = 0;
}

void AbstractFifo::setTotalSize (int newSize) noexcept
{
	jassert (newSize > 0);
	reset();
	bufferSize = newSize;
}

void AbstractFifo::prepareToWrite (int numToWrite, int& startIndex1, int& blockSize1, int& startIndex2, int& blockSize2) const noexcept
{
	const int vs = validStart.get();
	const int ve = validEnd.value;

	const int freeSpace = ve >= vs ? (bufferSize - (ve - vs)) : (vs - ve);
	numToWrite = jmin (numToWrite, freeSpace - 1);

	if (numToWrite <= 0)
	{
		startIndex1 = 0;
		startIndex2 = 0;
		blockSize1 = 0;
		blockSize2 = 0;
	}
	else
	{
		startIndex1 = ve;
		startIndex2 = 0;
		blockSize1 = jmin (bufferSize - ve, numToWrite);
		numToWrite -= blockSize1;
		blockSize2 = numToWrite <= 0 ? 0 : jmin (numToWrite, vs);
	}
}

void AbstractFifo::finishedWrite (int numWritten) noexcept
{
	jassert (numWritten >= 0 && numWritten < bufferSize);
	int newEnd = validEnd.value + numWritten;
	if (newEnd >= bufferSize)
		newEnd -= bufferSize;

	validEnd = newEnd;
}

void AbstractFifo::prepareToRead (int numWanted, int& startIndex1, int& blockSize1, int& startIndex2, int& blockSize2) const noexcept
{
	const int vs = validStart.value;
	const int ve = validEnd.get();

	const int numReady = ve >= vs ? (ve - vs) : (bufferSize - (vs - ve));
	numWanted = jmin (numWanted, numReady);

	if (numWanted <= 0)
	{
		startIndex1 = 0;
		startIndex2 = 0;
		blockSize1 = 0;
		blockSize2 = 0;
	}
	else
	{
		startIndex1 = vs;
		startIndex2 = 0;
		blockSize1 = jmin (bufferSize - vs, numWanted);
		numWanted -= blockSize1;
		blockSize2 = numWanted <= 0 ? 0 : jmin (numWanted, ve);
	}
}

void AbstractFifo::finishedRead (int numRead) noexcept
{
	jassert (numRead >= 0 && numRead <= bufferSize);

	int newStart = validStart.value + numRead;
	if (newStart >= bufferSize)
		newStart -= bufferSize;

	validStart = newStart;
}


END_JUCE_NAMESPACE

/*** End of inlined file: juce_AbstractFifo.cpp ***/




/*** Start of inlined file: juce_MemoryBlock.cpp ***/
BEGIN_JUCE_NAMESPACE

MemoryBlock::MemoryBlock() noexcept
	: size (0)
{
}

MemoryBlock::MemoryBlock (const size_t initialSize, const bool initialiseToZero)
{
	if (initialSize > 0)
	{
		size = initialSize;
		data.allocate (initialSize, initialiseToZero);
	}
	else
	{
		size = 0;
	}
}

MemoryBlock::MemoryBlock (const MemoryBlock& other)
	: size (other.size)
{
	if (size > 0)
	{
		jassert (other.data != nullptr);
		data.malloc (size);
		memcpy (data, other.data, size);
	}
}

MemoryBlock::MemoryBlock (const void* const dataToInitialiseFrom, const size_t sizeInBytes)
	: size (jmax ((size_t) 0, sizeInBytes))
{
	jassert (sizeInBytes >= 0);

	if (size > 0)
	{
		jassert (dataToInitialiseFrom != nullptr); // non-zero size, but a zero pointer passed-in?

		data.malloc (size);

		if (dataToInitialiseFrom != nullptr)
			memcpy (data, dataToInitialiseFrom, size);
	}
}

MemoryBlock::~MemoryBlock() noexcept
{
	jassert (size >= 0);	// should never happen
	jassert (size == 0 || data != nullptr); // non-zero size but no data allocated?
}

MemoryBlock& MemoryBlock::operator= (const MemoryBlock& other)
{
	if (this != &other)
	{
		setSize (other.size, false);
		memcpy (data, other.data, size);
	}

	return *this;
}

bool MemoryBlock::operator== (const MemoryBlock& other) const noexcept
{
	return matches (other.data, other.size);
}

bool MemoryBlock::operator!= (const MemoryBlock& other) const noexcept
{
	return ! operator== (other);
}

bool MemoryBlock::matches (const void* dataToCompare, size_t dataSize) const noexcept
{
	return size == dataSize
			&& memcmp (data, dataToCompare, size) == 0;
}

// this will resize the block to this size
void MemoryBlock::setSize (const size_t newSize, const bool initialiseToZero)
{
	if (size != newSize)
	{
		if (newSize <= 0)
		{
			data.free();
			size = 0;
		}
		else
		{
			if (data != nullptr)
			{
				data.realloc (newSize);

				if (initialiseToZero && (newSize > size))
					zeromem (data + size, newSize - size);
			}
			else
			{
				data.allocate (newSize, initialiseToZero);
			}

			size = newSize;
		}
	}
}

void MemoryBlock::ensureSize (const size_t minimumSize, const bool initialiseToZero)
{
	if (size < minimumSize)
		setSize (minimumSize, initialiseToZero);
}

void MemoryBlock::swapWith (MemoryBlock& other) noexcept
{
	std::swap (size, other.size);
	data.swapWith (other.data);
}

void MemoryBlock::fillWith (const uint8 value) noexcept
{
	memset (data, (int) value, size);
}

void MemoryBlock::append (const void* const srcData, const size_t numBytes)
{
	if (numBytes > 0)
	{
		const size_t oldSize = size;
		setSize (size + numBytes);
		memcpy (data + oldSize, srcData, numBytes);
	}
}

void MemoryBlock::copyFrom (const void* const src, int offset, size_t num) noexcept
{
	const char* d = static_cast<const char*> (src);

	if (offset < 0)
	{
		d -= offset;
		num -= offset;
		offset = 0;
	}

	if (offset + num > size)
		num = size - offset;

	if (num > 0)
		memcpy (data + offset, d, num);
}

void MemoryBlock::copyTo (void* const dst, int offset, size_t num) const noexcept
{
	char* d = static_cast<char*> (dst);

	if (offset < 0)
	{
		zeromem (d, -offset);
		d -= offset;

		num += offset;
		offset = 0;
	}

	if (offset + num > size)
	{
		const size_t newNum = size - offset;
		zeromem (d + newNum, num - newNum);
		num = newNum;
	}

	if (num > 0)
		memcpy (d, data + offset, num);
}

void MemoryBlock::removeSection (size_t startByte, size_t numBytesToRemove)
{
	if (startByte + numBytesToRemove >= size)
	{
		setSize (startByte);
	}
	else if (numBytesToRemove > 0)
	{
		memmove (data + startByte,
				 data + startByte + numBytesToRemove,
				 size - (startByte + numBytesToRemove));

		setSize (size - numBytesToRemove);
	}
}

String MemoryBlock::toString() const
{
	return String (CharPointer_UTF8 (data), size);
}

int MemoryBlock::getBitRange (const size_t bitRangeStart, size_t numBits) const noexcept
{
	int res = 0;

	size_t byte = bitRangeStart >> 3;
	int offsetInByte = (int) bitRangeStart & 7;
	size_t bitsSoFar = 0;

	while (numBits > 0 && (size_t) byte < size)
	{
		const int bitsThisTime = jmin ((int) numBits, 8 - offsetInByte);
		const int mask = (0xff >> (8 - bitsThisTime)) << offsetInByte;

		res |= (((data[byte] & mask) >> offsetInByte) << bitsSoFar);

		bitsSoFar += bitsThisTime;
		numBits -= bitsThisTime;
		++byte;
		offsetInByte = 0;
	}

	return res;
}

void MemoryBlock::setBitRange (const size_t bitRangeStart, size_t numBits, int bitsToSet) noexcept
{
	size_t byte = bitRangeStart >> 3;
	int offsetInByte = (int) bitRangeStart & 7;
	unsigned int mask = ~((((unsigned int) 0xffffffff) << (32 - numBits)) >> (32 - numBits));

	while (numBits > 0 && (size_t) byte < size)
	{
		const int bitsThisTime = jmin ((int) numBits, 8 - offsetInByte);

		const unsigned int tempMask = (mask << offsetInByte) | ~((((unsigned int) 0xffffffff) >> offsetInByte) << offsetInByte);
		const unsigned int tempBits = bitsToSet << offsetInByte;

		data[byte] = (char) ((data[byte] & tempMask) | tempBits);

		++byte;
		numBits -= bitsThisTime;
		bitsToSet >>= bitsThisTime;
		mask >>= bitsThisTime;
		offsetInByte = 0;
	}
}

void MemoryBlock::loadFromHexString (const String& hex)
{
	ensureSize (hex.length() >> 1);
	char* dest = data;
	String::CharPointerType t (hex.getCharPointer());

	for (;;)
	{
		int byte = 0;

		for (int loop = 2; --loop >= 0;)
		{
			byte <<= 4;

			for (;;)
			{
				const juce_wchar c = t.getAndAdvance();

				if (c >= '0' && c <= '9')
				{
					byte |= c - '0';
					break;
				}
				else if (c >= 'a' && c <= 'z')
				{
					byte |= c - ('a' - 10);
					break;
				}
				else if (c >= 'A' && c <= 'Z')
				{
					byte |= c - ('A' - 10);
					break;
				}
				else if (c == 0)
				{
					setSize (static_cast <size_t> (dest - data));
					return;
				}
			}
		}

		*dest++ = (char) byte;
	}
}

const char* const MemoryBlock::encodingTable = ".ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+";

String MemoryBlock::toBase64Encoding() const
{
	const size_t numChars = ((size << 3) + 5) / 6;

	String destString ((unsigned int) size); // store the length, followed by a '.', and then the data.
	const int initialLen = destString.length();
	destString.preallocateBytes (sizeof (String::CharPointerType::CharType) * (initialLen + 2 + numChars));

	String::CharPointerType d (destString.getCharPointer());
	d += initialLen;
	d.write ('.');

	for (size_t i = 0; i < numChars; ++i)
		d.write (encodingTable [getBitRange (i * 6, 6)]);

	d.writeNull();
	return destString;
}

bool MemoryBlock::fromBase64Encoding (const String& s)
{
	const int startPos = s.indexOfChar ('.') + 1;

	if (startPos <= 0)
		return false;

	const int numBytesNeeded = s.substring (0, startPos - 1).getIntValue();

	setSize (numBytesNeeded, true);

	const int numChars = s.length() - startPos;

	String::CharPointerType srcChars (s.getCharPointer());
	srcChars += startPos;
	int pos = 0;

	for (int i = 0; i < numChars; ++i)
	{
		const char c = (char) srcChars.getAndAdvance();

		for (int j = 0; j < 64; ++j)
		{
			if (encodingTable[j] == c)
			{
				setBitRange (pos, 6, j);
				pos += 6;
				break;
			}
		}
	}

	return true;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MemoryBlock.cpp ***/


/*** Start of inlined file: juce_PropertySet.cpp ***/
BEGIN_JUCE_NAMESPACE

PropertySet::PropertySet (const bool ignoreCaseOfKeyNames)
	: properties (ignoreCaseOfKeyNames),
	  fallbackProperties (nullptr),
	  ignoreCaseOfKeys (ignoreCaseOfKeyNames)
{
}

PropertySet::PropertySet (const PropertySet& other)
	: properties (other.properties),
	  fallbackProperties (other.fallbackProperties),
	  ignoreCaseOfKeys (other.ignoreCaseOfKeys)
{
}

PropertySet& PropertySet::operator= (const PropertySet& other)
{
	properties = other.properties;
	fallbackProperties = other.fallbackProperties;
	ignoreCaseOfKeys = other.ignoreCaseOfKeys;

	propertyChanged();
	return *this;
}

PropertySet::~PropertySet()
{
}

void PropertySet::clear()
{
	const ScopedLock sl (lock);

	if (properties.size() > 0)
	{
		properties.clear();
		propertyChanged();
	}
}

String PropertySet::getValue (const String& keyName,
							  const String& defaultValue) const noexcept
{
	const ScopedLock sl (lock);

	const int index = properties.getAllKeys().indexOf (keyName, ignoreCaseOfKeys);

	if (index >= 0)
		return properties.getAllValues() [index];

	return fallbackProperties != nullptr ? fallbackProperties->getValue (keyName, defaultValue)
										 : defaultValue;
}

int PropertySet::getIntValue (const String& keyName,
							  const int defaultValue) const noexcept
{
	const ScopedLock sl (lock);
	const int index = properties.getAllKeys().indexOf (keyName, ignoreCaseOfKeys);

	if (index >= 0)
		return properties.getAllValues() [index].getIntValue();

	return fallbackProperties != nullptr ? fallbackProperties->getIntValue (keyName, defaultValue)
										 : defaultValue;
}

double PropertySet::getDoubleValue (const String& keyName,
									const double defaultValue) const noexcept
{
	const ScopedLock sl (lock);
	const int index = properties.getAllKeys().indexOf (keyName, ignoreCaseOfKeys);

	if (index >= 0)
		return properties.getAllValues()[index].getDoubleValue();

	return fallbackProperties != nullptr ? fallbackProperties->getDoubleValue (keyName, defaultValue)
										 : defaultValue;
}

bool PropertySet::getBoolValue (const String& keyName,
								const bool defaultValue) const noexcept
{
	const ScopedLock sl (lock);
	const int index = properties.getAllKeys().indexOf (keyName, ignoreCaseOfKeys);

	if (index >= 0)
		return properties.getAllValues() [index].getIntValue() != 0;

	return fallbackProperties != nullptr ? fallbackProperties->getBoolValue (keyName, defaultValue)
										 : defaultValue;
}

XmlElement* PropertySet::getXmlValue (const String& keyName) const
{
	return XmlDocument::parse (getValue (keyName));
}

void PropertySet::setValue (const String& keyName, const var& v)
{
	jassert (keyName.isNotEmpty()); // shouldn't use an empty key name!

	if (keyName.isNotEmpty())
	{
		const String value (v.toString());
		const ScopedLock sl (lock);

		const int index = properties.getAllKeys().indexOf (keyName, ignoreCaseOfKeys);

		if (index < 0 || properties.getAllValues() [index] != value)
		{
			properties.set (keyName, value);
			propertyChanged();
		}
	}
}

void PropertySet::removeValue (const String& keyName)
{
	if (keyName.isNotEmpty())
	{
		const ScopedLock sl (lock);
		const int index = properties.getAllKeys().indexOf (keyName, ignoreCaseOfKeys);

		if (index >= 0)
		{
			properties.remove (keyName);
			propertyChanged();
		}
	}
}

void PropertySet::setValue (const String& keyName, const XmlElement* const xml)
{
	setValue (keyName, xml == nullptr ? var::null
									  : var (xml->createDocument (String::empty, true)));
}

bool PropertySet::containsKey (const String& keyName) const noexcept
{
	const ScopedLock sl (lock);
	return properties.getAllKeys().contains (keyName, ignoreCaseOfKeys);
}

void PropertySet::addAllPropertiesFrom (const PropertySet& source)
{
	const ScopedLock sl (source.getLock());

	for (int i = 0; i < source.properties.size(); ++i)
		setValue (source.properties.getAllKeys() [i],
				  source.properties.getAllValues() [i]);
}

void PropertySet::setFallbackPropertySet (PropertySet* fallbackProperties_) noexcept
{
	const ScopedLock sl (lock);
	fallbackProperties = fallbackProperties_;
}

XmlElement* PropertySet::createXml (const String& nodeName) const
{
	const ScopedLock sl (lock);
	XmlElement* const xml = new XmlElement (nodeName);

	for (int i = 0; i < properties.getAllKeys().size(); ++i)
	{
		XmlElement* const e = xml->createNewChildElement ("VALUE");
		e->setAttribute ("name", properties.getAllKeys()[i]);
		e->setAttribute ("val", properties.getAllValues()[i]);
	}

	return xml;
}

void PropertySet::restoreFromXml (const XmlElement& xml)
{
	const ScopedLock sl (lock);
	clear();

	forEachXmlChildElementWithTagName (xml, e, "VALUE")
	{
		if (e->hasAttribute ("name")
			 && e->hasAttribute ("val"))
		{
			properties.set (e->getStringAttribute ("name"),
							e->getStringAttribute ("val"));
		}
	}

	if (properties.size() > 0)
		propertyChanged();
}

void PropertySet::propertyChanged()
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_PropertySet.cpp ***/


/*** Start of inlined file: juce_Identifier.cpp ***/
BEGIN_JUCE_NAMESPACE

StringPool& Identifier::getPool()
{
	static StringPool pool;
	return pool;
}

Identifier::Identifier() noexcept
	: name (nullptr)
{
}

Identifier::Identifier (const Identifier& other) noexcept
	: name (other.name)
{
}

Identifier& Identifier::operator= (const Identifier& other) noexcept
{
	name = other.name;
	return *this;
}

Identifier::Identifier (const String& name_)
	: name (Identifier::getPool().getPooledString (name_))
{
	/* An Identifier string must be suitable for use as a script variable or XML
	   attribute, so it can only contain this limited set of characters.. */
	jassert (isValidIdentifier (name_));
}

Identifier::Identifier (const char* const name_)
	: name (Identifier::getPool().getPooledString (name_))
{
	/* An Identifier string must be suitable for use as a script variable or XML
	   attribute, so it can only contain this limited set of characters.. */
	jassert (isValidIdentifier (toString()));
}

Identifier::~Identifier()
{
}

bool Identifier::isValidIdentifier (const String& possibleIdentifier) noexcept
{
	return possibleIdentifier.isNotEmpty()
			&& possibleIdentifier.containsOnly ("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789_-:");
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Identifier.cpp ***/


/*** Start of inlined file: juce_Variant.cpp ***/
BEGIN_JUCE_NAMESPACE

enum VariantStreamMarkers
{
	varMarker_Int	   = 1,
	varMarker_BoolTrue  = 2,
	varMarker_BoolFalse = 3,
	varMarker_Double	= 4,
	varMarker_String	= 5,
	varMarker_Int64	 = 6,
	varMarker_Array	 = 7
};

class var::VariantType
{
public:
	VariantType() noexcept {}
	virtual ~VariantType() noexcept {}

	virtual int toInt (const ValueUnion&) const noexcept			{ return 0; }
	virtual int64 toInt64 (const ValueUnion&) const noexcept			{ return 0; }
	virtual double toDouble (const ValueUnion&) const noexcept		  { return 0; }
	virtual String toString (const ValueUnion&) const			   { return String::empty; }
	virtual bool toBool (const ValueUnion&) const noexcept			  { return false; }
	virtual ReferenceCountedObject* toObject (const ValueUnion&) const noexcept { return nullptr; }
	virtual Array<var>* toArray (const ValueUnion&) const noexcept		  { return 0; }

	virtual bool isVoid() const noexcept	  { return false; }
	virtual bool isInt() const noexcept	   { return false; }
	virtual bool isInt64() const noexcept	 { return false; }
	virtual bool isBool() const noexcept	  { return false; }
	virtual bool isDouble() const noexcept	{ return false; }
	virtual bool isString() const noexcept	{ return false; }
	virtual bool isObject() const noexcept	{ return false; }
	virtual bool isArray() const noexcept	 { return false; }
	virtual bool isMethod() const noexcept	{ return false; }

	virtual void cleanUp (ValueUnion&) const noexcept {}
	virtual void createCopy (ValueUnion& dest, const ValueUnion& source) const	  { dest = source; }
	virtual bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept = 0;
	virtual void writeToStream (const ValueUnion& data, OutputStream& output) const = 0;
};

class var::VariantType_Void  : public var::VariantType
{
public:
	VariantType_Void() noexcept {}
	static const VariantType_Void instance;

	bool isVoid() const noexcept	{ return true; }
	bool equals (const ValueUnion&, const ValueUnion&, const VariantType& otherType) const noexcept { return otherType.isVoid(); }
	void writeToStream (const ValueUnion&, OutputStream& output) const   { output.writeCompressedInt (0); }
};

class var::VariantType_Int  : public var::VariantType
{
public:
	VariantType_Int() noexcept {}
	static const VariantType_Int instance;

	int toInt (const ValueUnion& data) const noexcept	   { return data.intValue; };
	int64 toInt64 (const ValueUnion& data) const noexcept   { return (int64) data.intValue; };
	double toDouble (const ValueUnion& data) const noexcept { return (double) data.intValue; }
	String toString (const ValueUnion& data) const	  { return String (data.intValue); }
	bool toBool (const ValueUnion& data) const noexcept	 { return data.intValue != 0; }
	bool isInt() const noexcept				 { return true; }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		return otherType.toInt (otherData) == data.intValue;
	}

	void writeToStream (const ValueUnion& data, OutputStream& output) const
	{
		output.writeCompressedInt (5);
		output.writeByte (varMarker_Int);
		output.writeInt (data.intValue);
	}
};

class var::VariantType_Int64  : public var::VariantType
{
public:
	VariantType_Int64() noexcept {}
	static const VariantType_Int64 instance;

	int toInt (const ValueUnion& data) const noexcept	   { return (int) data.int64Value; };
	int64 toInt64 (const ValueUnion& data) const noexcept   { return data.int64Value; };
	double toDouble (const ValueUnion& data) const noexcept { return (double) data.int64Value; }
	String toString (const ValueUnion& data) const	  { return String (data.int64Value); }
	bool toBool (const ValueUnion& data) const noexcept	 { return data.int64Value != 0; }
	bool isInt64() const noexcept			   { return true; }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		return otherType.toInt64 (otherData) == data.int64Value;
	}

	void writeToStream (const ValueUnion& data, OutputStream& output) const
	{
		output.writeCompressedInt (9);
		output.writeByte (varMarker_Int64);
		output.writeInt64 (data.int64Value);
	}
};

class var::VariantType_Double   : public var::VariantType
{
public:
	VariantType_Double() noexcept {}
	static const VariantType_Double instance;

	int toInt (const ValueUnion& data) const noexcept	   { return (int) data.doubleValue; };
	int64 toInt64 (const ValueUnion& data) const noexcept   { return (int64) data.doubleValue; };
	double toDouble (const ValueUnion& data) const noexcept { return data.doubleValue; }
	String toString (const ValueUnion& data) const	  { return String (data.doubleValue); }
	bool toBool (const ValueUnion& data) const noexcept	 { return data.doubleValue != 0; }
	bool isDouble() const noexcept			  { return true; }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		return otherType.toDouble (otherData) == data.doubleValue;
	}

	void writeToStream (const ValueUnion& data, OutputStream& output) const
	{
		output.writeCompressedInt (9);
		output.writeByte (varMarker_Double);
		output.writeDouble (data.doubleValue);
	}
};

class var::VariantType_Bool   : public var::VariantType
{
public:
	VariantType_Bool() noexcept {}
	static const VariantType_Bool instance;

	int toInt (const ValueUnion& data) const noexcept	   { return data.boolValue ? 1 : 0; };
	int64 toInt64 (const ValueUnion& data) const noexcept   { return data.boolValue ? 1 : 0; };
	double toDouble (const ValueUnion& data) const noexcept { return data.boolValue ? 1.0 : 0.0; }
	String toString (const ValueUnion& data) const	  { return String::charToString (data.boolValue ? '1' : '0'); }
	bool toBool (const ValueUnion& data) const noexcept	 { return data.boolValue; }
	bool isBool() const noexcept				{ return true; }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		return otherType.toBool (otherData) == data.boolValue;
	}

	void writeToStream (const ValueUnion& data, OutputStream& output) const
	{
		output.writeCompressedInt (1);
		output.writeByte (data.boolValue ? (char) varMarker_BoolTrue : (char) varMarker_BoolFalse);
	}
};

class var::VariantType_String   : public var::VariantType
{
public:
	VariantType_String() noexcept {}
	static const VariantType_String instance;

	void cleanUp (ValueUnion& data) const noexcept			   { getString (data)-> ~String(); }
	void createCopy (ValueUnion& dest, const ValueUnion& source) const   { new (dest.stringValue) String (*getString (source)); }

	bool isString() const noexcept			  { return true; }
	int toInt (const ValueUnion& data) const noexcept	   { return getString (data)->getIntValue(); };
	int64 toInt64 (const ValueUnion& data) const noexcept   { return getString (data)->getLargeIntValue(); };
	double toDouble (const ValueUnion& data) const noexcept { return getString (data)->getDoubleValue(); }
	String toString (const ValueUnion& data) const	  { return *getString (data); }
	bool toBool (const ValueUnion& data) const noexcept	 { return getString (data)->getIntValue() != 0
																	  || getString (data)->trim().equalsIgnoreCase ("true")
																	  || getString (data)->trim().equalsIgnoreCase ("yes"); }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		return otherType.toString (otherData) == *getString (data);
	}

	void writeToStream (const ValueUnion& data, OutputStream& output) const
	{
		const String* const s = getString (data);
		const int len = s->getNumBytesAsUTF8() + 1;
		HeapBlock<char> temp (len);
		s->copyToUTF8 (temp, len);
		output.writeCompressedInt (len + 1);
		output.writeByte (varMarker_String);
		output.write (temp, len);
	}

private:
	static inline const String* getString (const ValueUnion& data) noexcept { return reinterpret_cast <const String*> (data.stringValue); }
	static inline String* getString (ValueUnion& data) noexcept		 { return reinterpret_cast <String*> (data.stringValue); }
};

class var::VariantType_Object   : public var::VariantType
{
public:
	VariantType_Object() noexcept {}
	static const VariantType_Object instance;

	void cleanUp (ValueUnion& data) const noexcept			  { if (data.objectValue != nullptr) data.objectValue->decReferenceCount(); }

	void createCopy (ValueUnion& dest, const ValueUnion& source) const
	{
		dest.objectValue = source.objectValue;
		if (dest.objectValue != nullptr)
			dest.objectValue->incReferenceCount();
	}

	String toString (const ValueUnion& data) const                            { return "Object 0x" + String::toHexString ((int) (pointer_sized_int) data.objectValue); }
	bool toBool (const ValueUnion& data) const noexcept			   { return data.objectValue != 0; }
	ReferenceCountedObject* toObject (const ValueUnion& data) const noexcept  { return data.objectValue; }
	bool isObject() const noexcept						{ return true; }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		return otherType.toObject (otherData) == data.objectValue;
	}

	void writeToStream (const ValueUnion&, OutputStream& output) const
	{
		jassertfalse; // Can't write an object to a stream!
		output.writeCompressedInt (0);
	}
};

class var::VariantType_Array   : public var::VariantType
{
public:
	VariantType_Array() noexcept {}
	static const VariantType_Array instance;

	void cleanUp (ValueUnion& data) const noexcept			  { delete data.arrayValue; }
	void createCopy (ValueUnion& dest, const ValueUnion& source) const  { dest.arrayValue = new Array<var> (*(source.arrayValue)); }

	String toString (const ValueUnion&) const                           { return "[Array]"; }
	bool isArray() const noexcept					   { return true; }
	Array<var>* toArray (const ValueUnion& data) const noexcept	 { return data.arrayValue; }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		const Array<var>* const otherArray = otherType.toArray (otherData);
		return otherArray != nullptr && *otherArray == *(data.arrayValue);
	}

	void writeToStream (const ValueUnion& data, OutputStream& output) const
	{
		MemoryOutputStream buffer (512);
		const int numItems = data.arrayValue->size();
		buffer.writeCompressedInt (numItems);

		for (int i = 0; i < numItems; ++i)
			data.arrayValue->getReference(i).writeToStream (buffer);

		output.writeCompressedInt (1 + (int) buffer.getDataSize());
		output.writeByte (varMarker_Array);
		output << buffer;
	}
};

class var::VariantType_Method   : public var::VariantType
{
public:
	VariantType_Method() noexcept {}
	static const VariantType_Method instance;

	String toString (const ValueUnion&) const               { return "Method"; }
	bool toBool (const ValueUnion& data) const noexcept	 { return data.methodValue != 0; }
	bool isMethod() const noexcept			  { return true; }

	bool equals (const ValueUnion& data, const ValueUnion& otherData, const VariantType& otherType) const noexcept
	{
		return otherType.isMethod() && otherData.methodValue == data.methodValue;
	}

	void writeToStream (const ValueUnion&, OutputStream& output) const
	{
		jassertfalse; // Can't write a method to a stream!
		output.writeCompressedInt (0);
	}
};

const var::VariantType_Void	var::VariantType_Void::instance;
const var::VariantType_Int	 var::VariantType_Int::instance;
const var::VariantType_Int64   var::VariantType_Int64::instance;
const var::VariantType_Bool	var::VariantType_Bool::instance;
const var::VariantType_Double  var::VariantType_Double::instance;
const var::VariantType_String  var::VariantType_String::instance;
const var::VariantType_Object  var::VariantType_Object::instance;
const var::VariantType_Array   var::VariantType_Array::instance;
const var::VariantType_Method  var::VariantType_Method::instance;

var::var() noexcept : type (&VariantType_Void::instance)
{
}

var::~var() noexcept
{
	type->cleanUp (value);
}

const var var::null;

var::var (const var& valueToCopy)  : type (valueToCopy.type)
{
	type->createCopy (value, valueToCopy.value);
}

var::var (const int value_) noexcept	   : type (&VariantType_Int::instance)	{ value.intValue = value_; }
var::var (const int64 value_) noexcept	 : type (&VariantType_Int64::instance)  { value.int64Value = value_; }
var::var (const bool value_) noexcept	  : type (&VariantType_Bool::instance)   { value.boolValue = value_; }
var::var (const double value_) noexcept	: type (&VariantType_Double::instance) { value.doubleValue = value_; }
var::var (MethodFunction method_) noexcept : type (&VariantType_Method::instance) { value.methodValue = method_; }

var::var (const String& value_)	: type (&VariantType_String::instance) { new (value.stringValue) String (value_); }
var::var (const char* const value_)	: type (&VariantType_String::instance) { new (value.stringValue) String (value_); }
var::var (const wchar_t* const value_) : type (&VariantType_String::instance) { new (value.stringValue) String (value_); }

var::var (const Array<var>& value_)  : type (&VariantType_Array::instance)
{
	value.arrayValue = new Array<var> (value_);
}

var::var (ReferenceCountedObject* const object)  : type (&VariantType_Object::instance)
{
	value.objectValue = object;

	if (object != nullptr)
		object->incReferenceCount();
}

bool var::isVoid() const noexcept	 { return type->isVoid(); }
bool var::isInt() const noexcept	  { return type->isInt(); }
bool var::isInt64() const noexcept	{ return type->isInt64(); }
bool var::isBool() const noexcept	 { return type->isBool(); }
bool var::isDouble() const noexcept   { return type->isDouble(); }
bool var::isString() const noexcept   { return type->isString(); }
bool var::isObject() const noexcept   { return type->isObject(); }
bool var::isArray() const noexcept	{ return type->isArray(); }
bool var::isMethod() const noexcept   { return type->isMethod(); }

var::operator int() const noexcept			  { return type->toInt (value); }
var::operator int64() const noexcept			{ return type->toInt64 (value); }
var::operator bool() const noexcept			 { return type->toBool (value); }
var::operator float() const noexcept			{ return (float) type->toDouble (value); }
var::operator double() const noexcept		   { return type->toDouble (value); }
String var::toString() const				{ return type->toString (value); }
var::operator String() const				{ return type->toString (value); }
ReferenceCountedObject* var::getObject() const noexcept { return type->toObject (value); }
Array<var>* var::getArray() const noexcept		  { return type->toArray (value); }
DynamicObject* var::getDynamicObject() const noexcept   { return dynamic_cast <DynamicObject*> (getObject()); }

void var::swapWith (var& other) noexcept
{
	std::swap (type, other.type);
	std::swap (value, other.value);
}

const var& var::operator= (const var& v)		   { type->cleanUp (value); type = v.type; type->createCopy (value, v.value); return *this; }
const var& var::operator= (const int v)		{ type->cleanUp (value); type = &VariantType_Int::instance; value.intValue = v; return *this; }
const var& var::operator= (const int64 v)		  { type->cleanUp (value); type = &VariantType_Int64::instance; value.int64Value = v; return *this; }
const var& var::operator= (const bool v)		   { type->cleanUp (value); type = &VariantType_Bool::instance; value.boolValue = v; return *this; }
const var& var::operator= (const double v)		 { type->cleanUp (value); type = &VariantType_Double::instance; value.doubleValue = v; return *this; }
const var& var::operator= (const char* const v)	{ type->cleanUp (value); type = &VariantType_String::instance; new (value.stringValue) String (v); return *this; }
const var& var::operator= (const wchar_t* const v)	 { type->cleanUp (value); type = &VariantType_String::instance; new (value.stringValue) String (v); return *this; }
const var& var::operator= (const String& v)		{ type->cleanUp (value); type = &VariantType_String::instance; new (value.stringValue) String (v); return *this; }
const var& var::operator= (const Array<var>& v)	{ var v2 (v); swapWith (v2); return *this; }
const var& var::operator= (ReferenceCountedObject* v)  { var v2 (v); swapWith (v2); return *this; }
const var& var::operator= (MethodFunction v)	   { var v2 (v); swapWith (v2); return *this; }

bool var::equals (const var& other) const noexcept
{
	return type->equals (value, other.value, *other.type);
}

bool var::equalsWithSameType (const var& other) const noexcept
{
	return type == other.type && equals (other);
}

bool operator== (const var& v1, const var& v2) noexcept	 { return v1.equals (v2); }
bool operator!= (const var& v1, const var& v2) noexcept	 { return ! v1.equals (v2); }
bool operator== (const var& v1, const String& v2)	   { return v1.toString() == v2; }
bool operator!= (const var& v1, const String& v2)	   { return v1.toString() != v2; }
bool operator== (const var& v1, const char* const v2)	   { return v1.toString() == v2; }
bool operator!= (const var& v1, const char* const v2)	   { return v1.toString() != v2; }

var var::operator[] (const Identifier& propertyName) const
{
	DynamicObject* const o = getDynamicObject();
	return o != nullptr ? o->getProperty (propertyName) : var::null;
}

var var::operator[] (const char* const propertyName) const
{
	return operator[] (Identifier (propertyName));
}

var var::getProperty (const Identifier& propertyName, const var& defaultReturnValue) const
{
	DynamicObject* const o = getDynamicObject();
	return o != nullptr ? o->getProperties().getWithDefault (propertyName, defaultReturnValue) : defaultReturnValue;
}

var var::invoke (const Identifier& method, const var* arguments, int numArguments) const
{
	DynamicObject* const o = getDynamicObject();
	return o != nullptr ? o->invokeMethod (method, arguments, numArguments) : var::null;
}

var var::invokeMethod (DynamicObject* const target, const var* const arguments, const int numArguments) const
{
	jassert (target != nullptr);

	if (isMethod())
		return (target->*(value.methodValue)) (arguments, numArguments);

	return var::null;
}

var var::call (const Identifier& method) const
{
	return invoke (method, nullptr, 0);
}

var var::call (const Identifier& method, const var& arg1) const
{
	return invoke (method, &arg1, 1);
}

var var::call (const Identifier& method, const var& arg1, const var& arg2) const
{
	var args[] = { arg1, arg2 };
	return invoke (method, args, 2);
}

var var::call (const Identifier& method, const var& arg1, const var& arg2, const var& arg3)
{
	var args[] = { arg1, arg2, arg3 };
	return invoke (method, args, 3);
}

var var::call (const Identifier& method, const var& arg1, const var& arg2, const var& arg3, const var& arg4) const
{
	var args[] = { arg1, arg2, arg3, arg4 };
	return invoke (method, args, 4);
}

var var::call (const Identifier& method, const var& arg1, const var& arg2, const var& arg3, const var& arg4, const var& arg5) const
{
	var args[] = { arg1, arg2, arg3, arg4, arg5 };
	return invoke (method, args, 5);
}

int var::size() const
{
	const Array<var>* const array = getArray();
	return array != nullptr ? array->size() : 0;
}

const var& var::operator[] (int arrayIndex) const
{
	const Array<var>* const array = getArray();

	// When using this method, the var must actually be an array, and the index
	// must be in-range!
	jassert (array != nullptr && isPositiveAndBelow (arrayIndex, array->size()));

	return array->getReference (arrayIndex);
}

var& var::operator[] (int arrayIndex)
{
	const Array<var>* const array = getArray();

	// When using this method, the var must actually be an array, and the index
	// must be in-range!
	jassert (array != nullptr && isPositiveAndBelow (arrayIndex, array->size()));

	return array->getReference (arrayIndex);
}

Array<var>* var::convertToArray()
{
	Array<var>* array = getArray();

	if (array == nullptr)
	{
		const Array<var> tempVar;
		var v (tempVar);
		array = v.value.arrayValue;

		if (! isVoid())
			array->add (*this);

		swapWith (v);
	}

	return array;
}

void var::append (const var& n)
{
	convertToArray()->add (n);
}

void var::remove (const int index)
{
	Array<var>* const array = getArray();

	if (array != nullptr)
		array->remove (index);
}

void var::insert (const int index, const var& n)
{
	convertToArray()->insert (index, n);
}

void var::resize (const int numArrayElementsWanted)
{
	convertToArray()->resize (numArrayElementsWanted);
}

int var::indexOf (const var& n) const
{
	const Array<var>* const array = getArray();
	return array != nullptr ? array->indexOf (n) : -1;
}

void var::writeToStream (OutputStream& output) const
{
	type->writeToStream (value, output);
}

var var::readFromStream (InputStream& input)
{
	const int numBytes = input.readCompressedInt();

	if (numBytes > 0)
	{
		switch (input.readByte())
		{
			case varMarker_Int:	 return var (input.readInt());
			case varMarker_Int64:	   return var (input.readInt64());
			case varMarker_BoolTrue:	return var (true);
			case varMarker_BoolFalse:   return var (false);
			case varMarker_Double:	  return var (input.readDouble());
			case varMarker_String:
			{
				MemoryOutputStream mo;
				mo.writeFromInputStream (input, numBytes - 1);
				return var (mo.toUTF8());
			}

			case varMarker_Array:
			{
				var v;
				Array<var>* const destArray = v.convertToArray();

				for (int i = input.readCompressedInt(); --i >= 0;)
					destArray->add (readFromStream (input));

				return v;
			}

			default:
				input.skipNextBytes (numBytes - 1); break;
		}
	}

	return var::null;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Variant.cpp ***/


/*** Start of inlined file: juce_NamedValueSet.cpp ***/
BEGIN_JUCE_NAMESPACE

NamedValueSet::NamedValue::NamedValue() noexcept
{
}

inline NamedValueSet::NamedValue::NamedValue (const Identifier& name_, const var& value_)
	: name (name_), value (value_)
{
}

NamedValueSet::NamedValue::NamedValue (const NamedValue& other)
	: name (other.name), value (other.value)
{
}

NamedValueSet::NamedValue& NamedValueSet::NamedValue::operator= (const NamedValueSet::NamedValue& other)
{
	name = other.name;
	value = other.value;
	return *this;
}

bool NamedValueSet::NamedValue::operator== (const NamedValueSet::NamedValue& other) const noexcept
{
	return name == other.name && value == other.value;
}

NamedValueSet::NamedValueSet() noexcept
{
}

NamedValueSet::NamedValueSet (const NamedValueSet& other)
{
	values.addCopyOfList (other.values);
}

NamedValueSet& NamedValueSet::operator= (const NamedValueSet& other)
{
	clear();
	values.addCopyOfList (other.values);
	return *this;
}

NamedValueSet::~NamedValueSet()
{
	clear();
}

void NamedValueSet::clear()
{
	values.deleteAll();
}

bool NamedValueSet::operator== (const NamedValueSet& other) const
{
	const NamedValue* i1 = values;
	const NamedValue* i2 = other.values;

	while (i1 != nullptr && i2 != nullptr)
	{
		if (! (*i1 == *i2))
			return false;

		i1 = i1->nextListItem;
		i2 = i2->nextListItem;
	}

	return true;
}

bool NamedValueSet::operator!= (const NamedValueSet& other) const
{
	return ! operator== (other);
}

int NamedValueSet::size() const noexcept
{
	return values.size();
}

const var& NamedValueSet::operator[] (const Identifier& name) const
{
	for (NamedValue* i = values; i != nullptr; i = i->nextListItem)
		if (i->name == name)
			return i->value;

	return var::null;
}

var NamedValueSet::getWithDefault (const Identifier& name, const var& defaultReturnValue) const
{
	const var* const v = getVarPointer (name);
	return v != nullptr ? *v : defaultReturnValue;
}

var* NamedValueSet::getVarPointer (const Identifier& name) const noexcept
{
	for (NamedValue* i = values; i != nullptr; i = i->nextListItem)
		if (i->name == name)
			return &(i->value);

	return nullptr;
}

bool NamedValueSet::set (const Identifier& name, const var& newValue)
{
	LinkedListPointer<NamedValue>* i = &values;

	while (i->get() != nullptr)
	{
		NamedValue* const v = i->get();

		if (v->name == name)
		{
			if (v->value.equalsWithSameType (newValue))
				return false;

			v->value = newValue;
			return true;
		}

		i = &(v->nextListItem);
	}

	i->insertNext (new NamedValue (name, newValue));
	return true;
}

bool NamedValueSet::contains (const Identifier& name) const
{
	return getVarPointer (name) != nullptr;
}

bool NamedValueSet::remove (const Identifier& name)
{
	LinkedListPointer<NamedValue>* i = &values;

	for (;;)
	{
		NamedValue* const v = i->get();

		if (v == nullptr)
			break;

		if (v->name == name)
		{
			delete i->removeNext();
			return true;
		}

		i = &(v->nextListItem);
	}

	return false;
}

const Identifier NamedValueSet::getName (const int index) const
{
	const NamedValue* const v = values[index];
	jassert (v != nullptr);
	return v->name;
}

const var& NamedValueSet::getValueAt (const int index) const
{
	const NamedValue* const v = values[index];
	jassert (v != nullptr);
	return v->value;
}

void NamedValueSet::setFromXmlAttributes (const XmlElement& xml)
{
	clear();
	LinkedListPointer<NamedValue>::Appender appender (values);

	const int numAtts = xml.getNumAttributes(); // xxx inefficient - should write an att iterator..

	for (int i = 0; i < numAtts; ++i)
		appender.append (new NamedValue (xml.getAttributeName (i), var (xml.getAttributeValue (i))));
}

void NamedValueSet::copyToXmlAttributes (XmlElement& xml) const
{
	for (NamedValue* i = values; i != nullptr; i = i->nextListItem)
	{
		jassert (! i->value.isObject()); // DynamicObjects can't be stored as XML!

		xml.setAttribute (i->name.toString(),
						  i->value.toString());
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_NamedValueSet.cpp ***/


/*** Start of inlined file: juce_DynamicObject.cpp ***/
BEGIN_JUCE_NAMESPACE

DynamicObject::DynamicObject()
{
}

DynamicObject::~DynamicObject()
{
}

bool DynamicObject::hasProperty (const Identifier& propertyName) const
{
	const var* const v = properties.getVarPointer (propertyName);
	return v != nullptr && ! v->isMethod();
}

var DynamicObject::getProperty (const Identifier& propertyName) const
{
	return properties [propertyName];
}

void DynamicObject::setProperty (const Identifier& propertyName, const var& newValue)
{
	properties.set (propertyName, newValue);
}

void DynamicObject::removeProperty (const Identifier& propertyName)
{
	properties.remove (propertyName);
}

bool DynamicObject::hasMethod (const Identifier& methodName) const
{
	return getProperty (methodName).isMethod();
}

var DynamicObject::invokeMethod (const Identifier& methodName,
								 const var* parameters,
								 int numParameters)
{
	return properties [methodName].invokeMethod (this, parameters, numParameters);
}

void DynamicObject::setMethod (const Identifier& name,
							   var::MethodFunction methodFunction)
{
	properties.set (name, var (methodFunction));
}

void DynamicObject::clear()
{
	properties.clear();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DynamicObject.cpp ***/


/*** Start of inlined file: juce_Expression.cpp ***/
BEGIN_JUCE_NAMESPACE

class Expression::Term  : public SingleThreadedReferenceCountedObject
{
public:
	Term() {}
	virtual ~Term() {}

	virtual Type getType() const noexcept = 0;
	virtual Term* clone() const = 0;
	virtual ReferenceCountedObjectPtr<Term> resolve (const Scope&, int recursionDepth) = 0;
	virtual String toString() const = 0;
	virtual double toDouble() const					  { return 0; }
	virtual int getInputIndexFor (const Term*) const			 { return -1; }
	virtual int getOperatorPrecedence() const				{ return 0; }
	virtual int getNumInputs() const					 { return 0; }
	virtual Term* getInput (int) const					   { return nullptr; }
	virtual ReferenceCountedObjectPtr<Term> negated();

	virtual ReferenceCountedObjectPtr<Term> createTermToEvaluateInput (const Scope&, const Term* /*inputTerm*/,
																			 double /*overallTarget*/, Term* /*topLevelTerm*/) const
	{
		jassertfalse;
		return nullptr;
	}

	virtual String getName() const
	{
		jassertfalse; // You shouldn't call this for an expression that's not actually a function!
		return String::empty;
	}

	virtual void renameSymbol (const Symbol& oldSymbol, const String& newName, const Scope& scope, int recursionDepth)
	{
		for (int i = getNumInputs(); --i >= 0;)
			getInput (i)->renameSymbol (oldSymbol, newName, scope, recursionDepth);
	}

	class SymbolVisitor
	{
	public:
		virtual ~SymbolVisitor() {}
		virtual void useSymbol (const Symbol&) = 0;
	};

	virtual void visitAllSymbols (SymbolVisitor& visitor, const Scope& scope, int recursionDepth)
	{
		for (int i = getNumInputs(); --i >= 0;)
			getInput(i)->visitAllSymbols (visitor, scope, recursionDepth);
	}

private:
	JUCE_DECLARE_NON_COPYABLE (Term);
};

class Expression::Helpers
{
public:
	typedef ReferenceCountedObjectPtr<Term> TermPtr;

	// This helper function is needed to work around VC6 scoping bugs
	static inline const TermPtr& getTermFor (const Expression& exp) noexcept	  { return exp.term; }

	static void checkRecursionDepth (const int depth)
	{
		if (depth > 256)
			throw EvaluationError ("Recursive symbol references");
	}

	friend class Expression::Term; // (also only needed as a VC6 workaround)

	/** An exception that can be thrown by Expression::evaluate(). */
	class EvaluationError  : public std::exception
	{
	public:
		EvaluationError (const String& description_)
			: description (description_)
		{
			DBG ("Expression::EvaluationError: " + description);
		}

		String description;
	};

	class Constant  : public Term
	{
	public:
		Constant (const double value_, const bool isResolutionTarget_)
			: value (value_), isResolutionTarget (isResolutionTarget_) {}

		Type getType() const noexcept		{ return constantType; }
		Term* clone() const			  { return new Constant (value, isResolutionTarget); }
		TermPtr resolve (const Scope&, int)	  { return this; }
		double toDouble() const			  { return value; }
		TermPtr negated()				{ return new Constant (-value, isResolutionTarget); }

		String toString() const
		{
			String s (value);
			if (isResolutionTarget)
				s = "@" + s;

			return s;
		}

		double value;
		bool isResolutionTarget;
	};

	class BinaryTerm  : public Term
	{
	public:
		BinaryTerm (Term* const left_, Term* const right_) : left (left_), right (right_)
		{
			jassert (left_ != nullptr && right_ != nullptr);
		}

		int getInputIndexFor (const Term* possibleInput) const
		{
			return possibleInput == left ? 0 : (possibleInput == right ? 1 : -1);
		}

		Type getType() const noexcept	   { return operatorType; }
		int getNumInputs() const		{ return 2; }
		Term* getInput (int index) const	{ return index == 0 ? left.getObject() : (index == 1 ? right.getObject() : 0); }

		virtual double performFunction (double left, double right) const = 0;
		virtual void writeOperator (String& dest) const = 0;

		TermPtr resolve (const Scope& scope, int recursionDepth)
		{
			return new Constant (performFunction (left->resolve (scope, recursionDepth)->toDouble(),
												  right->resolve (scope, recursionDepth)->toDouble()), false);
		}

		String toString() const
		{
			String s;

			const int ourPrecendence = getOperatorPrecedence();
			if (left->getOperatorPrecedence() > ourPrecendence)
				s << '(' << left->toString() << ')';
			else
				s = left->toString();

			writeOperator (s);

			if (right->getOperatorPrecedence() >= ourPrecendence)
				s << '(' << right->toString() << ')';
			else
				s << right->toString();

			return s;
		}

	protected:
		const TermPtr left, right;

		TermPtr createDestinationTerm (const Scope& scope, const Term* input, double overallTarget, Term* topLevelTerm) const
		{
			jassert (input == left || input == right);
			if (input != left && input != right)
				return nullptr;

			const Term* const dest = findDestinationFor (topLevelTerm, this);

			if (dest == nullptr)
				return new Constant (overallTarget, false);

			return dest->createTermToEvaluateInput (scope, this, overallTarget, topLevelTerm);
		}
	};

	class SymbolTerm  : public Term
	{
	public:
		explicit SymbolTerm (const String& symbol_) : symbol (symbol_) {}

		TermPtr resolve (const Scope& scope, int recursionDepth)
		{
			checkRecursionDepth (recursionDepth);
			return getTermFor (scope.getSymbolValue (symbol))->resolve (scope, recursionDepth + 1);
		}

		Type getType() const noexcept   { return symbolType; }
		Term* clone() const		 { return new SymbolTerm (symbol); }
		String toString() const	 { return symbol; }
		String getName() const	  { return symbol; }

		void visitAllSymbols (SymbolVisitor& visitor, const Scope& scope, int recursionDepth)
		{
			checkRecursionDepth (recursionDepth);
			visitor.useSymbol (Symbol (scope.getScopeUID(), symbol));
			getTermFor (scope.getSymbolValue (symbol))->visitAllSymbols (visitor, scope, recursionDepth + 1);
		}

		void renameSymbol (const Symbol& oldSymbol, const String& newName, const Scope& scope, int /*recursionDepth*/)
		{
			if (oldSymbol.symbolName == symbol && scope.getScopeUID() == oldSymbol.scopeUID)
				symbol = newName;
		}

		String symbol;
	};

	class Function  : public Term
	{
	public:
		explicit Function (const String& functionName_)  : functionName (functionName_) {}

		Function (const String& functionName_, const Array<Expression>& parameters_)
			: functionName (functionName_), parameters (parameters_)
		{}

		Type getType() const noexcept   { return functionType; }
		Term* clone() const		 { return new Function (functionName, parameters); }
		int getNumInputs() const	{ return parameters.size(); }
		Term* getInput (int i) const	{ return getTermFor (parameters [i]); }
		String getName() const	  { return functionName; }

		TermPtr resolve (const Scope& scope, int recursionDepth)
		{
			checkRecursionDepth (recursionDepth);
			double result = 0;
			const int numParams = parameters.size();
			if (numParams > 0)
			{
				HeapBlock<double> params (numParams);
				for (int i = 0; i < numParams; ++i)
					params[i] = getTermFor (parameters.getReference(i))->resolve (scope, recursionDepth + 1)->toDouble();

				result = scope.evaluateFunction (functionName, params, numParams);
			}
			else
			{
				result = scope.evaluateFunction (functionName, nullptr, 0);
			}

			return new Constant (result, false);
		}

		int getInputIndexFor (const Term* possibleInput) const
		{
			for (int i = 0; i < parameters.size(); ++i)
				if (getTermFor (parameters.getReference(i)) == possibleInput)
					return i;

			return -1;
		}

		String toString() const
		{
			if (parameters.size() == 0)
				return functionName + "()";

			String s (functionName + " (");

			for (int i = 0; i < parameters.size(); ++i)
			{
				s << getTermFor (parameters.getReference(i))->toString();

				if (i < parameters.size() - 1)
					s << ", ";
			}

			s << ')';
			return s;
		}

		const String functionName;
		Array<Expression> parameters;
	};

	class DotOperator  : public BinaryTerm
	{
	public:
		DotOperator (SymbolTerm* const left_, Term* const right_)  : BinaryTerm (left_, right_) {}

		TermPtr resolve (const Scope& scope, int recursionDepth)
		{
			checkRecursionDepth (recursionDepth);

			EvaluationVisitor visitor (right, recursionDepth + 1);
			scope.visitRelativeScope (getSymbol()->symbol, visitor);
			return visitor.output;
		}

		Term* clone() const				 { return new DotOperator (getSymbol(), right); }
		String getName() const                          { return "."; }
		int getOperatorPrecedence() const		   { return 1; }
		void writeOperator (String& dest) const	 { dest << '.'; }
		double performFunction (double, double) const   { return 0.0; }

		void visitAllSymbols (SymbolVisitor& visitor, const Scope& scope, int recursionDepth)
		{
			checkRecursionDepth (recursionDepth);
			visitor.useSymbol (Symbol (scope.getScopeUID(), getSymbol()->symbol));

			SymbolVisitingVisitor v (right, visitor, recursionDepth + 1);

			try
			{
				scope.visitRelativeScope (getSymbol()->symbol, v);
			}
			catch (...) {}
		}

		void renameSymbol (const Symbol& oldSymbol, const String& newName, const Scope& scope, int recursionDepth)
		{
			checkRecursionDepth (recursionDepth);
			getSymbol()->renameSymbol (oldSymbol, newName, scope, recursionDepth);

			SymbolRenamingVisitor visitor (right, oldSymbol, newName, recursionDepth + 1);

			try
			{
				scope.visitRelativeScope (getSymbol()->symbol, visitor);
			}
			catch (...) {}
		}

	private:

		class EvaluationVisitor  : public Scope::Visitor
		{
		public:
			EvaluationVisitor (const TermPtr& input_, const int recursionCount_)
				: input (input_), output (input_), recursionCount (recursionCount_) {}

			void visit (const Scope& scope)   { output = input->resolve (scope, recursionCount); }

			const TermPtr input;
			TermPtr output;
			const int recursionCount;

		private:
			JUCE_DECLARE_NON_COPYABLE (EvaluationVisitor);
		};

		class SymbolVisitingVisitor  : public Scope::Visitor
		{
		public:
			SymbolVisitingVisitor (const TermPtr& input_, SymbolVisitor& visitor_, const int recursionCount_)
				: input (input_), visitor (visitor_), recursionCount (recursionCount_) {}

			void visit (const Scope& scope)   { input->visitAllSymbols (visitor, scope, recursionCount); }

		private:
			const TermPtr input;
			SymbolVisitor& visitor;
			const int recursionCount;

			JUCE_DECLARE_NON_COPYABLE (SymbolVisitingVisitor);
		};

		class SymbolRenamingVisitor   : public Scope::Visitor
		{
		public:
			SymbolRenamingVisitor (const TermPtr& input_, const Expression::Symbol& symbol_, const String& newName_, const int recursionCount_)
				: input (input_), symbol (symbol_), newName (newName_), recursionCount (recursionCount_)  {}

			void visit (const Scope& scope)   { input->renameSymbol (symbol, newName, scope, recursionCount); }

		private:
			const TermPtr input;
			const Symbol& symbol;
			const String newName;
			const int recursionCount;

			JUCE_DECLARE_NON_COPYABLE (SymbolRenamingVisitor);
		};

		SymbolTerm* getSymbol() const  { return static_cast <SymbolTerm*> (left.getObject()); }

		JUCE_DECLARE_NON_COPYABLE (DotOperator);
	};

	class Negate  : public Term
	{
	public:
		explicit Negate (const TermPtr& input_) : input (input_)
		{
			jassert (input_ != nullptr);
		}

		Type getType() const noexcept			   { return operatorType; }
		int getInputIndexFor (const Term* possibleInput) const  { return possibleInput == input ? 0 : -1; }
		int getNumInputs() const				{ return 1; }
		Term* getInput (int index) const			{ return index == 0 ? input.getObject() : nullptr; }
		Term* clone() const					 { return new Negate (input->clone()); }

		TermPtr resolve (const Scope& scope, int recursionDepth)
		{
			return new Constant (-input->resolve (scope, recursionDepth)->toDouble(), false);
		}

		String getName() const          { return "-"; }
		TermPtr negated()		   { return input; }

		TermPtr createTermToEvaluateInput (const Scope& scope, const Term* input_, double overallTarget, Term* topLevelTerm) const
		{
			(void) input_;
			jassert (input_ == input);

			const Term* const dest = findDestinationFor (topLevelTerm, this);

			return new Negate (dest == nullptr ? new Constant (overallTarget, false)
											   : dest->createTermToEvaluateInput (scope, this, overallTarget, topLevelTerm));
		}

		String toString() const
		{
			if (input->getOperatorPrecedence() > 0)
				return "-(" + input->toString() + ")";
			else
				return "-" + input->toString();
		}

	private:
		const TermPtr input;
	};

	class Add  : public BinaryTerm
	{
	public:
		Add (Term* const left_, Term* const right_) : BinaryTerm (left_, right_) {}

		Term* clone() const			 { return new Add (left->clone(), right->clone()); }
		double performFunction (double lhs, double rhs) const	{ return lhs + rhs; }
		int getOperatorPrecedence() const	   { return 3; }
		String getName() const                  { return "+"; }
		void writeOperator (String& dest) const { dest << " + "; }

		TermPtr createTermToEvaluateInput (const Scope& scope, const Term* input, double overallTarget, Term* topLevelTerm) const
		{
			const TermPtr newDest (createDestinationTerm (scope, input, overallTarget, topLevelTerm));
			if (newDest == nullptr)
				return nullptr;

			return new Subtract (newDest, (input == left ? right : left)->clone());
		}

	private:
		JUCE_DECLARE_NON_COPYABLE (Add);
	};

	class Subtract  : public BinaryTerm
	{
	public:
		Subtract (Term* const left_, Term* const right_) : BinaryTerm (left_, right_) {}

		Term* clone() const			 { return new Subtract (left->clone(), right->clone()); }
		double performFunction (double lhs, double rhs) const	{ return lhs - rhs; }
		int getOperatorPrecedence() const	   { return 3; }
		String getName() const                  { return "-"; }
		void writeOperator (String& dest) const { dest << " - "; }

		TermPtr createTermToEvaluateInput (const Scope& scope, const Term* input, double overallTarget, Term* topLevelTerm) const
		{
			const TermPtr newDest (createDestinationTerm (scope, input, overallTarget, topLevelTerm));
			if (newDest == nullptr)
				return nullptr;

			if (input == left)
				return new Add (newDest, right->clone());
			else
				return new Subtract (left->clone(), newDest);
		}

	private:
		JUCE_DECLARE_NON_COPYABLE (Subtract);
	};

	class Multiply  : public BinaryTerm
	{
	public:
		Multiply (Term* const left_, Term* const right_) : BinaryTerm (left_, right_) {}

		Term* clone() const			 { return new Multiply (left->clone(), right->clone()); }
		double performFunction (double lhs, double rhs) const	{ return lhs * rhs; }
		String getName() const                  { return "*"; }
		void writeOperator (String& dest) const { dest << " * "; }
		int getOperatorPrecedence() const	   { return 2; }

		TermPtr createTermToEvaluateInput (const Scope& scope, const Term* input, double overallTarget, Term* topLevelTerm) const
		{
			const TermPtr newDest (createDestinationTerm (scope, input, overallTarget, topLevelTerm));
			if (newDest == nullptr)
				return nullptr;

			return new Divide (newDest, (input == left ? right : left)->clone());
		}

	private:
		JUCE_DECLARE_NON_COPYABLE (Multiply);
	};

	class Divide  : public BinaryTerm
	{
	public:
		Divide (Term* const left_, Term* const right_) : BinaryTerm (left_, right_) {}

		Term* clone() const			 { return new Divide (left->clone(), right->clone()); }
		double performFunction (double lhs, double rhs) const	{ return lhs / rhs; }
		String getName() const                  { return "/"; }
		void writeOperator (String& dest) const { dest << " / "; }
		int getOperatorPrecedence() const	   { return 2; }

		TermPtr createTermToEvaluateInput (const Scope& scope, const Term* input, double overallTarget, Term* topLevelTerm) const
		{
			const TermPtr newDest (createDestinationTerm (scope, input, overallTarget, topLevelTerm));
			if (newDest == nullptr)
				return nullptr;

			if (input == left)
				return new Multiply (newDest, right->clone());
			else
				return new Divide (left->clone(), newDest);
		}

	private:
		JUCE_DECLARE_NON_COPYABLE (Divide);
	};

	static Term* findDestinationFor (Term* const topLevel, const Term* const inputTerm)
	{
		const int inputIndex = topLevel->getInputIndexFor (inputTerm);
		if (inputIndex >= 0)
			return topLevel;

		for (int i = topLevel->getNumInputs(); --i >= 0;)
		{
			Term* const t = findDestinationFor (topLevel->getInput (i), inputTerm);

			if (t != nullptr)
				return t;
		}

		return nullptr;
	}

	static Constant* findTermToAdjust (Term* const term, const bool mustBeFlagged)
	{
		{
			Constant* const c = dynamic_cast<Constant*> (term);
			if (c != nullptr && (c->isResolutionTarget || ! mustBeFlagged))
				return c;
		}

		if (dynamic_cast<Function*> (term) != nullptr)
			return nullptr;

		int i;
		const int numIns = term->getNumInputs();
		for (i = 0; i < numIns; ++i)
		{
			Constant* const c = dynamic_cast<Constant*> (term->getInput (i));
			if (c != nullptr && (c->isResolutionTarget || ! mustBeFlagged))
				return c;
		}

		for (i = 0; i < numIns; ++i)
		{
			Constant* const c = findTermToAdjust (term->getInput (i), mustBeFlagged);
			if (c != nullptr)
				return c;
		}

		return nullptr;
	}

	static bool containsAnySymbols (const Term* const t)
	{
		if (t->getType() == Expression::symbolType)
			return true;

		for (int i = t->getNumInputs(); --i >= 0;)
			if (containsAnySymbols (t->getInput (i)))
				return true;

		return false;
	}

	class SymbolCheckVisitor  : public Term::SymbolVisitor
	{
	public:
		SymbolCheckVisitor (const Symbol& symbol_) : wasFound (false), symbol (symbol_) {}
		void useSymbol (const Symbol& s)	{ wasFound = wasFound || s == symbol; }

		bool wasFound;

	private:
		const Symbol& symbol;

		JUCE_DECLARE_NON_COPYABLE (SymbolCheckVisitor);
	};

	class SymbolListVisitor  : public Term::SymbolVisitor
	{
	public:
		SymbolListVisitor (Array<Symbol>& list_) : list (list_) {}
		void useSymbol (const Symbol& s)	{ list.addIfNotAlreadyThere (s); }

	private:
		Array<Symbol>& list;

		JUCE_DECLARE_NON_COPYABLE (SymbolListVisitor);
	};

	class Parser
	{
	public:

		Parser (String::CharPointerType& stringToParse)
			: text (stringToParse)
		{
		}

		TermPtr readUpToComma()
		{
			if (text.isEmpty())
				return new Constant (0.0, false);

			const TermPtr e (readExpression());

			if (e == nullptr || ((! readOperator (",")) && ! text.isEmpty()))
				throw ParseError ("Syntax error: \"" + String (text) + "\"");

			return e;
		}

	private:
		String::CharPointerType& text;

		static inline bool isDecimalDigit (const juce_wchar c) noexcept
		{
			return c >= '0' && c <= '9';
		}

		bool readChar (const juce_wchar required) noexcept
		{
			if (*text == required)
			{
				++text;
				return true;
			}

			return false;
		}

		bool readOperator (const char* ops, char* const opType = nullptr) noexcept
		{
			text = text.findEndOfWhitespace();

			while (*ops != 0)
			{
				if (readChar (*ops))
				{
					if (opType != nullptr)
						*opType = *ops;

					return true;
				}

				++ops;
			}

			return false;
		}

		bool readIdentifier (String& identifier) noexcept
		{
			text = text.findEndOfWhitespace();
			String::CharPointerType t (text);
			int numChars = 0;

			if (t.isLetter() || *t == '_')
			{
				++t;
				++numChars;

				while (t.isLetterOrDigit() || *t == '_')
				{
					++t;
					++numChars;
				}
			}

			if (numChars > 0)
			{
				identifier = String (text, numChars);
				text = t;
				return true;
			}

			return false;
		}

		Term* readNumber() noexcept
		{
			text = text.findEndOfWhitespace();
			String::CharPointerType t (text);

			const bool isResolutionTarget = (*t == '@');
			if (isResolutionTarget)
			{
				++t;
				t = t.findEndOfWhitespace();
				text = t;
			}

			if (*t == '-')
			{
				++t;
				t = t.findEndOfWhitespace();
			}

			if (isDecimalDigit (*t) || (*t == '.' && isDecimalDigit (t[1])))
				return new Constant (CharacterFunctions::readDoubleValue (text), isResolutionTarget);

			return nullptr;
		}

		TermPtr readExpression()
		{
			TermPtr lhs (readMultiplyOrDivideExpression());

			char opType;
			while (lhs != nullptr && readOperator ("+-", &opType))
			{
				TermPtr rhs (readMultiplyOrDivideExpression());

				if (rhs == nullptr)
					throw ParseError ("Expected expression after \"" + String::charToString (opType) + "\"");

				if (opType == '+')
					lhs = new Add (lhs, rhs);
				else
					lhs = new Subtract (lhs, rhs);
			}

			return lhs;
		}

		TermPtr readMultiplyOrDivideExpression()
		{
			TermPtr lhs (readUnaryExpression());

			char opType;
			while (lhs != nullptr && readOperator ("*/", &opType))
			{
				TermPtr rhs (readUnaryExpression());

				if (rhs == nullptr)
					throw ParseError ("Expected expression after \"" + String::charToString (opType) + "\"");

				if (opType == '*')
					lhs = new Multiply (lhs, rhs);
				else
					lhs = new Divide (lhs, rhs);
			}

			return lhs;
		}

		TermPtr readUnaryExpression()
		{
			char opType;
			if (readOperator ("+-", &opType))
			{
				TermPtr term (readUnaryExpression());

				if (term == nullptr)
					throw ParseError ("Expected expression after \"" + String::charToString (opType) + "\"");

				if (opType == '-')
					term = term->negated();

				return term;
			}

			return readPrimaryExpression();
		}

		TermPtr readPrimaryExpression()
		{
			TermPtr e (readParenthesisedExpression());
			if (e != nullptr)
				return e;

			e = readNumber();
			if (e != nullptr)
				return e;

			return readSymbolOrFunction();
		}

		TermPtr readSymbolOrFunction()
		{
			String identifier;
			if (readIdentifier (identifier))
			{
				if (readOperator ("(")) // method call...
				{
					Function* const f = new Function (identifier);
					ScopedPointer<Term> func (f);  // (can't use ScopedPointer<Function> in MSVC)

					TermPtr param (readExpression());

					if (param == nullptr)
					{
						if (readOperator (")"))
							return func.release();

						throw ParseError ("Expected parameters after \"" + identifier + " (\"");
					}

					f->parameters.add (Expression (param));

					while (readOperator (","))
					{
						param = readExpression();

						if (param == nullptr)
							throw ParseError ("Expected expression after \",\"");

						f->parameters.add (Expression (param));
					}

					if (readOperator (")"))
						return func.release();

					throw ParseError ("Expected \")\"");
				}
				else if (readOperator ("."))
				{
					TermPtr rhs (readSymbolOrFunction());

					if (rhs == nullptr)
						throw ParseError ("Expected symbol or function after \".\"");

					if (identifier == "this")
						return rhs;

					return new DotOperator (new SymbolTerm (identifier), rhs);
				}
				else // just a symbol..
				{
					jassert (identifier.trim() == identifier);
					return new SymbolTerm (identifier);
				}
			}

			return nullptr;
		}

		TermPtr readParenthesisedExpression()
		{
			if (! readOperator ("("))
				return nullptr;

			const TermPtr e (readExpression());
			if (e == nullptr || ! readOperator (")"))
				return nullptr;

			return e;
		}

		JUCE_DECLARE_NON_COPYABLE (Parser);
	};
};

Expression::Expression()
	: term (new Expression::Helpers::Constant (0, false))
{
}

Expression::~Expression()
{
}

Expression::Expression (Term* const term_)
	: term (term_)
{
	jassert (term != nullptr);
}

Expression::Expression (const double constant)
	: term (new Expression::Helpers::Constant (constant, false))
{
}

Expression::Expression (const Expression& other)
	: term (other.term)
{
}

Expression& Expression::operator= (const Expression& other)
{
	term = other.term;
	return *this;
}

Expression::Expression (const String& stringToParse)
{
	String::CharPointerType text (stringToParse.getCharPointer());
	Helpers::Parser parser (text);
	term = parser.readUpToComma();
}

Expression Expression::parse (String::CharPointerType& stringToParse)
{
	Helpers::Parser parser (stringToParse);
	return Expression (parser.readUpToComma());
}

double Expression::evaluate() const
{
	return evaluate (Expression::Scope());
}

double Expression::evaluate (const Expression::Scope& scope) const
{
	try
	{
		return term->resolve (scope, 0)->toDouble();
	}
	catch (Helpers::EvaluationError&)
	{}

	return 0;
}

double Expression::evaluate (const Scope& scope, String& evaluationError) const
{
	try
	{
		return term->resolve (scope, 0)->toDouble();
	}
	catch (Helpers::EvaluationError& e)
	{
		evaluationError = e.description;
	}

	return 0;
}

Expression Expression::operator+ (const Expression& other) const  { return Expression (new Helpers::Add (term, other.term)); }
Expression Expression::operator- (const Expression& other) const  { return Expression (new Helpers::Subtract (term, other.term)); }
Expression Expression::operator* (const Expression& other) const  { return Expression (new Helpers::Multiply (term, other.term)); }
Expression Expression::operator/ (const Expression& other) const  { return Expression (new Helpers::Divide (term, other.term)); }
Expression Expression::operator-() const			  { return Expression (term->negated()); }
Expression Expression::symbol (const String& symbol)		  { return Expression (new Helpers::SymbolTerm (symbol)); }

Expression Expression::function (const String& functionName, const Array<Expression>& parameters)
{
	return Expression (new Helpers::Function (functionName, parameters));
}

Expression Expression::adjustedToGiveNewResult (const double targetValue, const Expression::Scope& scope) const
{
	ScopedPointer<Term> newTerm (term->clone());

	Helpers::Constant* termToAdjust = Helpers::findTermToAdjust (newTerm, true);

	if (termToAdjust == nullptr)
		termToAdjust = Helpers::findTermToAdjust (newTerm, false);

	if (termToAdjust == nullptr)
	{
		newTerm = new Helpers::Add (newTerm.release(), new Helpers::Constant (0, false));
		termToAdjust = Helpers::findTermToAdjust (newTerm, false);
	}

	jassert (termToAdjust != nullptr);

	const Term* const parent = Helpers::findDestinationFor (newTerm, termToAdjust);

	if (parent == nullptr)
	{
		termToAdjust->value = targetValue;
	}
	else
	{
		const Helpers::TermPtr reverseTerm (parent->createTermToEvaluateInput (scope, termToAdjust, targetValue, newTerm));

		if (reverseTerm == nullptr)
			return Expression (targetValue);

		termToAdjust->value = reverseTerm->resolve (scope, 0)->toDouble();
	}

	return Expression (newTerm.release());
}

Expression Expression::withRenamedSymbol (const Expression::Symbol& oldSymbol, const String& newName, const Scope& scope) const
{
	jassert (newName.toLowerCase().containsOnly ("abcdefghijklmnopqrstuvwxyz0123456789_"));

	if (oldSymbol.symbolName == newName)
		return *this;

	Expression e (term->clone());
	e.term->renameSymbol (oldSymbol, newName, scope, 0);
	return e;
}

bool Expression::referencesSymbol (const Expression::Symbol& symbolToCheck, const Scope& scope) const
{
	Helpers::SymbolCheckVisitor visitor (symbolToCheck);

	try
	{
		term->visitAllSymbols (visitor, scope, 0);
	}
	catch (Helpers::EvaluationError&)
	{}

	return visitor.wasFound;
}

void Expression::findReferencedSymbols (Array<Symbol>& results, const Scope& scope) const
{
	try
	{
		Helpers::SymbolListVisitor visitor (results);
		term->visitAllSymbols (visitor, scope, 0);
	}
	catch (Helpers::EvaluationError&)
	{}
}

String Expression::toString() const			 { return term->toString(); }
bool Expression::usesAnySymbols() const		 { return Helpers::containsAnySymbols (term); }
Expression::Type Expression::getType() const noexcept   { return term->getType(); }
String Expression::getSymbolOrFunction() const	  { return term->getName(); }
int Expression::getNumInputs() const			{ return term->getNumInputs(); }
Expression Expression::getInput (int index) const	   { return Expression (term->getInput (index)); }

ReferenceCountedObjectPtr<Expression::Term> Expression::Term::negated()
{
	return new Helpers::Negate (this);
}

Expression::ParseError::ParseError (const String& message)
	: description (message)
{
	DBG ("Expression::ParseError: " + message);
}

Expression::Symbol::Symbol (const String& scopeUID_, const String& symbolName_)
	: scopeUID (scopeUID_), symbolName (symbolName_)
{
}

bool Expression::Symbol::operator== (const Symbol& other) const noexcept
{
	return symbolName == other.symbolName && scopeUID == other.scopeUID;
}

bool Expression::Symbol::operator!= (const Symbol& other) const noexcept
{
	return ! operator== (other);
}

Expression::Scope::Scope()  {}
Expression::Scope::~Scope() {}

Expression Expression::Scope::getSymbolValue (const String& symbol) const
{
	throw Helpers::EvaluationError ("Unknown symbol: " + symbol);
}

double Expression::Scope::evaluateFunction (const String& functionName, const double* parameters, int numParams) const
{
	if (numParams > 0)
	{
		if (functionName == "min")
		{
			double v = parameters[0];
			for (int i = 1; i < numParams; ++i)
				v = jmin (v, parameters[i]);

			return v;
		}
		else if (functionName == "max")
		{
			double v = parameters[0];
			for (int i = 1; i < numParams; ++i)
				v = jmax (v, parameters[i]);

			return v;
		}
		else if (numParams == 1)
		{
			if      (functionName == "sin")     return sin (parameters[0]);
			else if (functionName == "cos")     return cos (parameters[0]);
			else if (functionName == "tan")     return tan (parameters[0]);
			else if (functionName == "abs")     return std::abs (parameters[0]);
		}
	}

	throw Helpers::EvaluationError ("Unknown function: \"" + functionName + "\"");
}

void Expression::Scope::visitRelativeScope (const String& scopeName, Visitor&) const
{
	throw Helpers::EvaluationError ("Unknown symbol: " + scopeName);
}

String Expression::Scope::getScopeUID() const
{
	return String::empty;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Expression.cpp ***/




/*** Start of inlined file: juce_InputStream.cpp ***/
BEGIN_JUCE_NAMESPACE

char InputStream::readByte()
{
	char temp = 0;
	read (&temp, 1);
	return temp;
}

bool InputStream::readBool()
{
	return readByte() != 0;
}

short InputStream::readShort()
{
	char temp[2];

	if (read (temp, 2) == 2)
		return (short) ByteOrder::littleEndianShort (temp);

	return 0;
}

short InputStream::readShortBigEndian()
{
	char temp[2];

	if (read (temp, 2) == 2)
		return (short) ByteOrder::bigEndianShort (temp);

	return 0;
}

int InputStream::readInt()
{
	char temp[4];

	if (read (temp, 4) == 4)
		return (int) ByteOrder::littleEndianInt (temp);

	return 0;
}

int InputStream::readIntBigEndian()
{
	char temp[4];

	if (read (temp, 4) == 4)
		return (int) ByteOrder::bigEndianInt (temp);

	return 0;
}

int InputStream::readCompressedInt()
{
	const unsigned char sizeByte = readByte();
	if (sizeByte == 0)
		return 0;

	const int numBytes = (sizeByte & 0x7f);
	if (numBytes > 4)
	{
		jassertfalse;	// trying to read corrupt data - this method must only be used
					   // to read data that was written by OutputStream::writeCompressedInt()
		return 0;
	}

	char bytes[4] = { 0, 0, 0, 0 };
	if (read (bytes, numBytes) != numBytes)
		return 0;

	const int num = (int) ByteOrder::littleEndianInt (bytes);
	return (sizeByte >> 7) ? -num : num;
}

int64 InputStream::readInt64()
{
	union { uint8 asBytes[8]; uint64 asInt64; } n;

	if (read (n.asBytes, 8) == 8)
		return (int64) ByteOrder::swapIfBigEndian (n.asInt64);

	return 0;
}

int64 InputStream::readInt64BigEndian()
{
	union { uint8 asBytes[8]; uint64 asInt64; } n;

	if (read (n.asBytes, 8) == 8)
		return (int64) ByteOrder::swapIfLittleEndian (n.asInt64);

	return 0;
}

float InputStream::readFloat()
{
	// the union below relies on these types being the same size...
	static_assert (sizeof (int32) == sizeof (float), "");
	union { int32 asInt; float asFloat; } n;
	n.asInt = (int32) readInt();
	return n.asFloat;
}

float InputStream::readFloatBigEndian()
{
	union { int32 asInt; float asFloat; } n;
	n.asInt = (int32) readIntBigEndian();
	return n.asFloat;
}

double InputStream::readDouble()
{
	union { int64 asInt; double asDouble; } n;
	n.asInt = readInt64();
	return n.asDouble;
}

double InputStream::readDoubleBigEndian()
{
	union { int64 asInt; double asDouble; } n;
	n.asInt = readInt64BigEndian();
	return n.asDouble;
}

String InputStream::readString()
{
	MemoryBlock buffer (256);
	char* data = static_cast<char*> (buffer.getData());
	size_t i = 0;

	while ((data[i] = readByte()) != 0)
	{
		if (++i >= buffer.getSize())
		{
			buffer.setSize (buffer.getSize() + 512);
			data = static_cast<char*> (buffer.getData());
		}
	}

	return String::fromUTF8 (data, (int) i);
}

String InputStream::readNextLine()
{
	MemoryBlock buffer (256);
	char* data = static_cast<char*> (buffer.getData());
	size_t i = 0;

	while ((data[i] = readByte()) != 0)
	{
		if (data[i] == '\n')
			break;

		if (data[i] == '\r')
		{
			const int64 lastPos = getPosition();

			if (readByte() != '\n')
				setPosition (lastPos);

			break;
		}

		if (++i >= buffer.getSize())
		{
			buffer.setSize (buffer.getSize() + 512);
			data = static_cast<char*> (buffer.getData());
		}
	}

	return String::fromUTF8 (data, (int) i);
}

int InputStream::readIntoMemoryBlock (MemoryBlock& block, int numBytes)
{
	MemoryOutputStream mo (block, true);
	return mo.writeFromInputStream (*this, numBytes);
}

String InputStream::readEntireStreamAsString()
{
	MemoryOutputStream mo;
	mo.writeFromInputStream (*this, -1);
	return mo.toString();
}

void InputStream::skipNextBytes (int64 numBytesToSkip)
{
	if (numBytesToSkip > 0)
	{
		const int skipBufferSize = (int) jmin (numBytesToSkip, (int64) 16384);
		HeapBlock<char> temp (skipBufferSize);

		while (numBytesToSkip > 0 && ! isExhausted())
			numBytesToSkip -= read (temp, (int) jmin (numBytesToSkip, (int64) skipBufferSize));
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_InputStream.cpp ***/


/*** Start of inlined file: juce_OutputStream.cpp ***/
BEGIN_JUCE_NAMESPACE

#if JUCE_DEBUG

struct DanglingStreamChecker
{
	DanglingStreamChecker() {}

	~DanglingStreamChecker()
	{
		/*
			It's always a bad idea to leak any object, but if you're leaking output
			streams, then there's a good chance that you're failing to flush a file
			to disk properly, which could result in corrupted data and other similar
			nastiness..
		*/
		jassert (activeStreams.size() == 0);
	}

	Array<void*, CriticalSection> activeStreams;
};

static DanglingStreamChecker danglingStreamChecker;
#endif

OutputStream::OutputStream()
	: newLineString (NewLine::getDefault())
{
   #if JUCE_DEBUG
	danglingStreamChecker.activeStreams.add (this);
   #endif
}

OutputStream::~OutputStream()
{
   #if JUCE_DEBUG
	danglingStreamChecker.activeStreams.removeValue (this);
   #endif
}

void OutputStream::writeBool (const bool b)
{
	writeByte (b ? (char) 1
				 : (char) 0);
}

void OutputStream::writeByte (char byte)
{
	write (&byte, 1);
}

void OutputStream::writeRepeatedByte (uint8 byte, int numTimesToRepeat)
{
	while (--numTimesToRepeat >= 0)
		writeByte (byte);
}

void OutputStream::writeShort (short value)
{
	const unsigned short v = ByteOrder::swapIfBigEndian ((unsigned short) value);
	write (&v, 2);
}

void OutputStream::writeShortBigEndian (short value)
{
	const unsigned short v = ByteOrder::swapIfLittleEndian ((unsigned short) value);
	write (&v, 2);
}

void OutputStream::writeInt (int value)
{
	const unsigned int v = ByteOrder::swapIfBigEndian ((unsigned int) value);
	write (&v, 4);
}

void OutputStream::writeIntBigEndian (int value)
{
	const unsigned int v = ByteOrder::swapIfLittleEndian ((unsigned int) value);
	write (&v, 4);
}

void OutputStream::writeCompressedInt (int value)
{
	unsigned int un = (value < 0) ? (unsigned int) -value
								  : (unsigned int) value;

	uint8 data[5];
	int num = 0;

	while (un > 0)
	{
		data[++num] = (uint8) un;
		un >>= 8;
	}

	data[0] = (uint8) num;

	if (value < 0)
		data[0] |= 0x80;

	write (data, num + 1);
}

void OutputStream::writeInt64 (int64 value)
{
	const uint64 v = ByteOrder::swapIfBigEndian ((uint64) value);
	write (&v, 8);
}

void OutputStream::writeInt64BigEndian (int64 value)
{
	const uint64 v = ByteOrder::swapIfLittleEndian ((uint64) value);
	write (&v, 8);
}

void OutputStream::writeFloat (float value)
{
	union { int asInt; float asFloat; } n;
	n.asFloat = value;
	writeInt (n.asInt);
}

void OutputStream::writeFloatBigEndian (float value)
{
	union { int asInt; float asFloat; } n;
	n.asFloat = value;
	writeIntBigEndian (n.asInt);
}

void OutputStream::writeDouble (double value)
{
	union { int64 asInt; double asDouble; } n;
	n.asDouble = value;
	writeInt64 (n.asInt);
}

void OutputStream::writeDoubleBigEndian (double value)
{
	union { int64 asInt; double asDouble; } n;
	n.asDouble = value;
	writeInt64BigEndian (n.asInt);
}

void OutputStream::writeString (const String& text)
{
	// (This avoids using toUTF8() to prevent the memory bloat that it would leave behind
	// if lots of large, persistent strings were to be written to streams).
	const int numBytes = text.getNumBytesAsUTF8() + 1;
	HeapBlock<char> temp (numBytes);
	text.copyToUTF8 (temp, numBytes);
	write (temp, numBytes);
}

void OutputStream::writeText (const String& text, const bool asUTF16,
							  const bool writeUTF16ByteOrderMark)
{
	if (asUTF16)
	{
		if (writeUTF16ByteOrderMark)
			write ("\x0ff\x0fe", 2);

		String::CharPointerType src (text.getCharPointer());
		bool lastCharWasReturn = false;

		for (;;)
		{
			const juce_wchar c = src.getAndAdvance();

			if (c == 0)
				break;

			if (c == '\n' && ! lastCharWasReturn)
				writeShort ((short) '\r');

			lastCharWasReturn = (c == L'\r');
			writeShort ((short) c);
		}
	}
	else
	{
		const char* src = text.toUTF8();
		const char* t = src;

		for (;;)
		{
			if (*t == '\n')
			{
				if (t > src)
					write (src, (int) (t - src));

				write ("\r\n", 2);
				src = t + 1;
			}
			else if (*t == '\r')
			{
				if (t[1] == '\n')
					++t;
			}
			else if (*t == 0)
			{
				if (t > src)
					write (src, (int) (t - src));

				break;
			}

			++t;
		}
	}
}

int OutputStream::writeFromInputStream (InputStream& source, int64 numBytesToWrite)
{
	if (numBytesToWrite < 0)
		numBytesToWrite = std::numeric_limits<int64>::max();

	int numWritten = 0;

	while (numBytesToWrite > 0 && ! source.isExhausted())
	{
		char buffer [8192];
		const int num = source.read (buffer, (int) jmin (numBytesToWrite, (int64) sizeof (buffer)));

		if (num <= 0)
			break;

		write (buffer, num);

		numBytesToWrite -= num;
		numWritten += num;
	}

	return numWritten;
}

void OutputStream::setNewLineString (const String& newLineString_)
{
	newLineString = newLineString_;
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const int number)
{
	return stream << String (number);
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const double number)
{
	return stream << String (number);
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const char character)
{
	stream.writeByte (character);
	return stream;
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const char* const text)
{
	stream.write (text, (int) strlen (text));
	return stream;
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const MemoryBlock& data)
{
	stream.write (data.getData(), (int) data.getSize());
	return stream;
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const File& fileToRead)
{
	const ScopedPointer<FileInputStream> in (fileToRead.createInputStream());

	if (in != nullptr)
		stream.writeFromInputStream (*in, -1);

	return stream;
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const NewLine&)
{
	return stream << stream.getNewLineString();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_OutputStream.cpp ***/


/*** Start of inlined file: juce_DirectoryIterator.cpp ***/
BEGIN_JUCE_NAMESPACE

DirectoryIterator::DirectoryIterator (const File& directory,
									  bool isRecursive_,
									  const String& wildCard_,
									  const int whatToLookFor_)
  : fileFinder (directory, isRecursive_ ? "*" : wildCard_),
	wildCard (wildCard_),
	path (File::addTrailingSeparator (directory.getFullPathName())),
	index (-1),
	totalNumFiles (-1),
	whatToLookFor (whatToLookFor_),
	isRecursive (isRecursive_),
	hasBeenAdvanced (false)
{
	// you have to specify the type of files you're looking for!
	jassert ((whatToLookFor_ & (File::findFiles | File::findDirectories)) != 0);
	jassert (whatToLookFor_ > 0 && whatToLookFor_ <= 7);
}

DirectoryIterator::~DirectoryIterator()
{
}

bool DirectoryIterator::next()
{
	return next (nullptr, nullptr, nullptr, nullptr, nullptr, nullptr);
}

bool DirectoryIterator::next (bool* const isDirResult, bool* const isHiddenResult, int64* const fileSize,
							  Time* const modTime, Time* const creationTime, bool* const isReadOnly)
{
	hasBeenAdvanced = true;

	if (subIterator != nullptr)
	{
		if (subIterator->next (isDirResult, isHiddenResult, fileSize, modTime, creationTime, isReadOnly))
			return true;

		subIterator = nullptr;
	}

	String filename;
	bool isDirectory, isHidden;
	while (fileFinder.next (filename, &isDirectory, &isHidden, fileSize, modTime, creationTime, isReadOnly))
	{
		++index;

		if (! filename.containsOnly ("."))
		{
			const File fileFound (path + filename, 0);
			bool matches = false;

			if (isDirectory)
			{
				if (isRecursive && ((whatToLookFor & File::ignoreHiddenFiles) == 0 || ! isHidden))
					subIterator = new DirectoryIterator (fileFound, true, wildCard, whatToLookFor);

				matches = (whatToLookFor & File::findDirectories) != 0;
			}
			else
			{
				matches = (whatToLookFor & File::findFiles) != 0;
			}

			// if recursive, we're not relying on the OS iterator to do the wildcard match, so do it now..
			if (matches && isRecursive)
				matches = filename.matchesWildcard (wildCard, ! File::areFileNamesCaseSensitive());

			if (matches && (whatToLookFor & File::ignoreHiddenFiles) != 0)
				matches = ! isHidden;

			if (matches)
			{
				currentFile = fileFound;
				if (isHiddenResult != nullptr)	 *isHiddenResult = isHidden;
				if (isDirResult != nullptr)	*isDirResult = isDirectory;

				return true;
			}
			else if (subIterator != nullptr)
			{
				return next();
			}
		}
	}

	return false;
}

const File& DirectoryIterator::getFile() const
{
	if (subIterator != nullptr && subIterator->hasBeenAdvanced)
		return subIterator->getFile();

	// You need to call DirectoryIterator::next() before asking it for the file that it found!
	jassert (hasBeenAdvanced);

	return currentFile;
}

float DirectoryIterator::getEstimatedProgress() const
{
	if (totalNumFiles < 0)
		totalNumFiles = File (path).getNumberOfChildFiles (File::findFilesAndDirectories);

	if (totalNumFiles <= 0)
		return 0.0f;

	const float detailedIndex = (subIterator != nullptr) ? index + subIterator->getEstimatedProgress()
														 : (float) index;

	return detailedIndex / totalNumFiles;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DirectoryIterator.cpp ***/


/*** Start of inlined file: juce_File.cpp ***/
#if ! JUCE_WINDOWS
  #include <pwd.h>
#endif

BEGIN_JUCE_NAMESPACE

File::File (const String& fullPathName)
	: fullPath (parseAbsolutePath (fullPathName))
{
}

File::File (const String& path, int)
	: fullPath (path)
{
}

File File::createFileWithoutCheckingPath (const String& path)
{
	return File (path, 0);
}

File::File (const File& other)
	: fullPath (other.fullPath)
{
}

File& File::operator= (const String& newPath)
{
	fullPath = parseAbsolutePath (newPath);
	return *this;
}

File& File::operator= (const File& other)
{
	fullPath = other.fullPath;
	return *this;
}

const File File::nonexistent;

String File::parseAbsolutePath (const String& p)
{
	if (p.isEmpty())
		return String::empty;

#if JUCE_WINDOWS
	// Windows..
	String path (p.replaceCharacter ('/', '\\'));

	if (path.startsWithChar (File::separator))
	{
		if (path[1] != File::separator)
		{
			/*  When you supply a raw string to the File object constructor, it must be an absolute path.
				If you're trying to parse a string that may be either a relative path or an absolute path,
				you MUST provide a context against which the partial path can be evaluated - you can do
				this by simply using File::getChildFile() instead of the File constructor. E.g. saying
				"File::getCurrentWorkingDirectory().getChildFile (myUnknownPath)" would return an absolute
				path if that's what was supplied, or would evaluate a partial path relative to the CWD.
			*/
			jassertfalse;

			path = File::getCurrentWorkingDirectory().getFullPathName().substring (0, 2) + path;
		}
	}
	else if (! path.containsChar (':'))
	{
		/*  When you supply a raw string to the File object constructor, it must be an absolute path.
			If you're trying to parse a string that may be either a relative path or an absolute path,
			you MUST provide a context against which the partial path can be evaluated - you can do
			this by simply using File::getChildFile() instead of the File constructor. E.g. saying
			"File::getCurrentWorkingDirectory().getChildFile (myUnknownPath)" would return an absolute
			path if that's what was supplied, or would evaluate a partial path relative to the CWD.
		*/
		jassertfalse;

		return File::getCurrentWorkingDirectory().getChildFile (path).getFullPathName();
	}
#else
	// Mac or Linux..

	// Yes, I know it's legal for a unix pathname to contain a backslash, but this assertion is here
	// to catch anyone who's trying to run code that was written on Windows with hard-coded path names.
	// If that's why you've ended up here, use File::getChildFile() to build your paths instead.
	jassert ((! p.containsChar ('\\')) || (p.indexOfChar ('/') >= 0 && p.indexOfChar ('/') < p.indexOfChar ('\\')));

	String path (p);

	if (path.startsWithChar ('~'))
	{
		if (path[1] == File::separator || path[1] == 0)
		{
			// expand a name of the form "~/abc"
			path = File::getSpecialLocation (File::userHomeDirectory).getFullPathName()
					+ path.substring (1);
		}
		else
		{
			// expand a name of type "~dave/abc"
			const String userName (path.substring (1).upToFirstOccurrenceOf ("/", false, false));

			struct passwd* const pw = getpwnam (userName.toUTF8());
			if (pw != nullptr)
				path = addTrailingSeparator (pw->pw_dir) + path.fromFirstOccurrenceOf ("/", false, false);
		}
	}
	else if (! path.startsWithChar (File::separator))
	{
		/*  When you supply a raw string to the File object constructor, it must be an absolute path.
			If you're trying to parse a string that may be either a relative path or an absolute path,
			you MUST provide a context against which the partial path can be evaluated - you can do
			this by simply using File::getChildFile() instead of the File constructor. E.g. saying
			"File::getCurrentWorkingDirectory().getChildFile (myUnknownPath)" would return an absolute
			path if that's what was supplied, or would evaluate a partial path relative to the CWD.
		*/
		jassert (path.startsWith ("./") || path.startsWith ("../")); // (assume that a path "./xyz" is deliberately intended to be relative to the CWD)

		return File::getCurrentWorkingDirectory().getChildFile (path).getFullPathName();
	}
#endif

	while (path.endsWithChar (separator) && path != separatorString) // careful not to turn a single "/" into an empty string.
		path = path.dropLastCharacters (1);

	return path;
}

String File::addTrailingSeparator (const String& path)
{
	return path.endsWithChar (File::separator) ? path
											   : path + File::separator;
}

#if JUCE_LINUX
 #define NAMES_ARE_CASE_SENSITIVE 1
#endif

bool File::areFileNamesCaseSensitive()
{
   #if NAMES_ARE_CASE_SENSITIVE
	return true;
   #else
	return false;
   #endif
}

bool File::operator== (const File& other) const
{
   #if NAMES_ARE_CASE_SENSITIVE
	return fullPath == other.fullPath;
   #else
	return fullPath.equalsIgnoreCase (other.fullPath);
   #endif
}

bool File::operator!= (const File& other) const
{
	return ! operator== (other);
}

bool File::operator< (const File& other) const
{
   #if NAMES_ARE_CASE_SENSITIVE
	return fullPath < other.fullPath;
   #else
	return fullPath.compareIgnoreCase (other.fullPath) < 0;
   #endif
}

bool File::operator> (const File& other) const
{
   #if NAMES_ARE_CASE_SENSITIVE
	return fullPath > other.fullPath;
   #else
	return fullPath.compareIgnoreCase (other.fullPath) > 0;
   #endif
}

bool File::setReadOnly (const bool shouldBeReadOnly,
						const bool applyRecursively) const
{
	bool worked = true;

	if (applyRecursively && isDirectory())
	{
		Array <File> subFiles;
		findChildFiles (subFiles, File::findFilesAndDirectories, false);

		for (int i = subFiles.size(); --i >= 0;)
			worked = subFiles.getReference(i).setReadOnly (shouldBeReadOnly, true) && worked;
	}

	return setFileReadOnlyInternal (shouldBeReadOnly) && worked;
}

bool File::deleteRecursively() const
{
	bool worked = true;

	if (isDirectory())
	{
		Array<File> subFiles;
		findChildFiles (subFiles, File::findFilesAndDirectories, false);

		for (int i = subFiles.size(); --i >= 0;)
			worked = subFiles.getReference(i).deleteRecursively() && worked;
	}

	return deleteFile() && worked;
}

bool File::moveFileTo (const File& newFile) const
{
	if (newFile.fullPath == fullPath)
		return true;

   #if ! NAMES_ARE_CASE_SENSITIVE
	if (*this != newFile)
   #endif
		if (! newFile.deleteFile())
			return false;

	return moveInternal (newFile);
}

bool File::copyFileTo (const File& newFile) const
{
	return (*this == newFile)
			|| (exists() && newFile.deleteFile() && copyInternal (newFile));
}

bool File::copyDirectoryTo (const File& newDirectory) const
{
	if (isDirectory() && newDirectory.createDirectory())
	{
		Array<File> subFiles;
		findChildFiles (subFiles, File::findFiles, false);

		int i;
		for (i = 0; i < subFiles.size(); ++i)
			if (! subFiles.getReference(i).copyFileTo (newDirectory.getChildFile (subFiles.getReference(i).getFileName())))
				return false;

		subFiles.clear();
		findChildFiles (subFiles, File::findDirectories, false);

		for (i = 0; i < subFiles.size(); ++i)
			if (! subFiles.getReference(i).copyDirectoryTo (newDirectory.getChildFile (subFiles.getReference(i).getFileName())))
				return false;

		return true;
	}

	return false;
}

String File::getPathUpToLastSlash() const
{
	const int lastSlash = fullPath.lastIndexOfChar (separator);

	if (lastSlash > 0)
		return fullPath.substring (0, lastSlash);
	else if (lastSlash == 0)
		return separatorString;
	else
		return fullPath;
}

File File::getParentDirectory() const
{
	return File (getPathUpToLastSlash(), (int) 0);
}

String File::getFileName() const
{
	return fullPath.substring (fullPath.lastIndexOfChar (separator) + 1);
}

int File::hashCode() const
{
	return fullPath.hashCode();
}

int64 File::hashCode64() const
{
	return fullPath.hashCode64();
}

String File::getFileNameWithoutExtension() const
{
	const int lastSlash = fullPath.lastIndexOfChar (separator) + 1;
	const int lastDot = fullPath.lastIndexOfChar ('.');

	if (lastDot > lastSlash)
		return fullPath.substring (lastSlash, lastDot);
	else
		return fullPath.substring (lastSlash);
}

bool File::isAChildOf (const File& potentialParent) const
{
	if (potentialParent == File::nonexistent)
		return false;

	const String ourPath (getPathUpToLastSlash());

   #if NAMES_ARE_CASE_SENSITIVE
	if (potentialParent.fullPath == ourPath)
   #else
	if (potentialParent.fullPath.equalsIgnoreCase (ourPath))
   #endif
	{
		return true;
	}
	else if (potentialParent.fullPath.length() >= ourPath.length())
	{
		return false;
	}
	else
	{
		return getParentDirectory().isAChildOf (potentialParent);
	}
}

bool File::isAbsolutePath (const String& path)
{
	return path.startsWithChar (separator)
		   #if JUCE_WINDOWS
			|| (path.isNotEmpty() && path[1] == ':');
		   #else
			|| path.startsWithChar ('~');
		   #endif
}

File File::getChildFile (String relativePath) const
{
	if (isAbsolutePath (relativePath))
		return File (relativePath);

	String path (fullPath);

	// It's relative, so remove any ../ or ./ bits at the start..
	if (relativePath[0] == '.')
	{
	   #if JUCE_WINDOWS
		relativePath = relativePath.replaceCharacter ('/', '\\').trimStart();
	   #else
		relativePath = relativePath.trimStart();
	   #endif

		while (relativePath[0] == '.')
		{
			if (relativePath[1] == '.')
			{
				if (relativePath [2] == 0 || relativePath[2] == separator)
				{
					const int lastSlash = path.lastIndexOfChar (separator);
					if (lastSlash >= 0)
						path = path.substring (0, lastSlash);

					relativePath = relativePath.substring (3);
				}
				else
				{
					break;
				}
			}
			else if (relativePath[1] == separator)
			{
				relativePath = relativePath.substring (2);
			}
			else
			{
				break;
			}
		}
	}

	return File (addTrailingSeparator (path) + relativePath);
}

File File::getSiblingFile (const String& fileName) const
{
	return getParentDirectory().getChildFile (fileName);
}

String File::descriptionOfSizeInBytes (const int64 bytes)
{
	if (bytes == 1)                       return "1 byte";
	else if (bytes < 1024)                return String (bytes) + " bytes";
	else if (bytes < 1024 * 1024)         return String (bytes / 1024.0, 1) + " KB";
	else if (bytes < 1024 * 1024 * 1024)  return String (bytes / (1024.0 * 1024.0), 1) + " MB";
	else                                  return String (bytes / (1024.0 * 1024.0 * 1024.0), 1) + " GB";
}

Result File::create() const
{
	if (exists())
		return Result::ok();

	const File parentDir (getParentDirectory());

	if (parentDir == *this)
		return Result::fail ("Cannot create parent directory");

	Result r (parentDir.createDirectory());

	if (r.wasOk())
	{
		FileOutputStream fo (*this, 8);

		r = fo.getStatus();
	}

	return r;
}

Result File::createDirectory() const
{
	if (isDirectory())
		return Result::ok();

	const File parentDir (getParentDirectory());

	if (parentDir == *this)
		return Result::fail ("Cannot create parent directory");

	Result r (parentDir.createDirectory());

	if (r.wasOk())
		r = createDirectoryInternal (fullPath.trimCharactersAtEnd (separatorString));

	return r;
}

Time File::getLastModificationTime() const		  { int64 m, a, c; getFileTimesInternal (m, a, c); return Time (m); }
Time File::getLastAccessTime() const			{ int64 m, a, c; getFileTimesInternal (m, a, c); return Time (a); }
Time File::getCreationTime() const			  { int64 m, a, c; getFileTimesInternal (m, a, c); return Time (c); }

bool File::setLastModificationTime (const Time& t) const	{ return setFileTimesInternal (t.toMilliseconds(), 0, 0); }
bool File::setLastAccessTime (const Time& t) const	  { return setFileTimesInternal (0, t.toMilliseconds(), 0); }
bool File::setCreationTime (const Time& t) const		{ return setFileTimesInternal (0, 0, t.toMilliseconds()); }

bool File::loadFileAsData (MemoryBlock& destBlock) const
{
	if (! existsAsFile())
		return false;

	FileInputStream in (*this);
	return getSize() == in.readIntoMemoryBlock (destBlock);
}

String File::loadFileAsString() const
{
	if (! existsAsFile())
		return String::empty;

	FileInputStream in (*this);
	return in.readEntireStreamAsString();
}

void File::readLines (StringArray& destLines) const
{
	destLines.addLines (loadFileAsString());
}

int File::findChildFiles (Array<File>& results,
						  const int whatToLookFor,
						  const bool searchRecursively,
						  const String& wildCardPattern) const
{
	DirectoryIterator di (*this, searchRecursively, wildCardPattern, whatToLookFor);

	int total = 0;
	while (di.next())
	{
		results.add (di.getFile());
		++total;
	}

	return total;
}

int File::getNumberOfChildFiles (const int whatToLookFor, const String& wildCardPattern) const
{
	DirectoryIterator di (*this, false, wildCardPattern, whatToLookFor);

	int total = 0;
	while (di.next())
		++total;

	return total;
}

bool File::containsSubDirectories() const
{
	if (isDirectory())
	{
		DirectoryIterator di (*this, false, "*", findDirectories);
		return di.next();
	}

	return false;
}

File File::getNonexistentChildFile (const String& prefix_,
									const String& suffix,
									bool putNumbersInBrackets) const
{
	File f (getChildFile (prefix_ + suffix));

	if (f.exists())
	{
		int num = 2;
		String prefix (prefix_);

		// remove any bracketed numbers that may already be on the end..
		if (prefix.trim().endsWithChar (')'))
		{
			putNumbersInBrackets = true;

			const int openBracks  = prefix.lastIndexOfChar ('(');
			const int closeBracks = prefix.lastIndexOfChar (')');

			if (openBracks > 0
				 && closeBracks > openBracks
				 && prefix.substring (openBracks + 1, closeBracks).containsOnly ("0123456789"))
			{
				num = prefix.substring (openBracks + 1, closeBracks).getIntValue() + 1;
				prefix = prefix.substring (0, openBracks);
			}
		}

		// also use brackets if it ends in a digit.
		putNumbersInBrackets = putNumbersInBrackets
								|| CharacterFunctions::isDigit (prefix.getLastCharacter());

		do
		{
			String newName (prefix);

			if (putNumbersInBrackets)
				newName << '(' << num++ << ')';
			else
				newName << num++;

			f = getChildFile (newName + suffix);

		} while (f.exists());
	}

	return f;
}

File File::getNonexistentSibling (const bool putNumbersInBrackets) const
{
	if (exists())
		return getParentDirectory()
				.getNonexistentChildFile (getFileNameWithoutExtension(),
										  getFileExtension(),
										  putNumbersInBrackets);
	return *this;
}

String File::getFileExtension() const
{
	const int indexOfDot = fullPath.lastIndexOfChar ('.');

	if (indexOfDot > fullPath.lastIndexOfChar (separator))
		return fullPath.substring (indexOfDot);

	return String::empty;
}

bool File::hasFileExtension (const String& possibleSuffix) const
{
	if (possibleSuffix.isEmpty())
		return fullPath.lastIndexOfChar ('.') <= fullPath.lastIndexOfChar (separator);

	const int semicolon = possibleSuffix.indexOfChar (0, ';');

	if (semicolon >= 0)
	{
		return hasFileExtension (possibleSuffix.substring (0, semicolon).trimEnd())
				|| hasFileExtension (possibleSuffix.substring (semicolon + 1).trimStart());
	}
	else
	{
		if (fullPath.endsWithIgnoreCase (possibleSuffix))
		{
			if (possibleSuffix.startsWithChar ('.'))
				return true;

			const int dotPos = fullPath.length() - possibleSuffix.length() - 1;

			if (dotPos >= 0)
				return fullPath [dotPos] == '.';
		}
	}

	return false;
}

File File::withFileExtension (const String& newExtension) const
{
	if (fullPath.isEmpty())
		return File::nonexistent;

	String filePart (getFileName());

	int i = filePart.lastIndexOfChar ('.');
	if (i >= 0)
		filePart = filePart.substring (0, i);

	if (newExtension.isNotEmpty() && ! newExtension.startsWithChar ('.'))
		filePart << '.';

	return getSiblingFile (filePart + newExtension);
}

bool File::startAsProcess (const String& parameters) const
{
	return exists() && Process::openDocument (fullPath, parameters);
}

FileInputStream* File::createInputStream() const
{
	if (existsAsFile())
		return new FileInputStream (*this);

	return nullptr;
}

FileOutputStream* File::createOutputStream (const int bufferSize) const
{
	ScopedPointer <FileOutputStream> out (new FileOutputStream (*this, bufferSize));

	if (out->failedToOpen())
		return nullptr;

	return out.release();
}

bool File::appendData (const void* const dataToAppend,
					   const int numberOfBytes) const
{
	if (numberOfBytes > 0)
	{
		const ScopedPointer <FileOutputStream> out (createOutputStream());

		if (out == 0)
			return false;

		out->write (dataToAppend, numberOfBytes);
	}

	return true;
}

bool File::replaceWithData (const void* const dataToWrite,
							const int numberOfBytes) const
{
	jassert (numberOfBytes >= 0); // a negative number of bytes??

	if (numberOfBytes <= 0)
		return deleteFile();

	TemporaryFile tempFile (*this, TemporaryFile::useHiddenFile);
	tempFile.getFile().appendData (dataToWrite, numberOfBytes);
	return tempFile.overwriteTargetFileWithTemporary();
}

bool File::appendText (const String& text,
					   const bool asUnicode,
					   const bool writeUnicodeHeaderBytes) const
{
	const ScopedPointer <FileOutputStream> out (createOutputStream());

	if (out != nullptr)
	{
		out->writeText (text, asUnicode, writeUnicodeHeaderBytes);
		return true;
	}

	return false;
}

bool File::replaceWithText (const String& textToWrite,
							const bool asUnicode,
							const bool writeUnicodeHeaderBytes) const
{
	TemporaryFile tempFile (*this, TemporaryFile::useHiddenFile);
	tempFile.getFile().appendText (textToWrite, asUnicode, writeUnicodeHeaderBytes);
	return tempFile.overwriteTargetFileWithTemporary();
}

bool File::hasIdenticalContentTo (const File& other) const
{
	if (other == *this)
		return true;

	if (getSize() == other.getSize() && existsAsFile() && other.existsAsFile())
	{
		FileInputStream in1 (*this), in2 (other);

		const int bufferSize = 4096;
		HeapBlock <char> buffer1 (bufferSize), buffer2 (bufferSize);

		for (;;)
		{
			const int num1 = in1.read (buffer1, bufferSize);
			const int num2 = in2.read (buffer2, bufferSize);

			if (num1 != num2)
				break;

			if (num1 <= 0)
				return true;

			if (memcmp (buffer1, buffer2, num1) != 0)
				break;
		}
	}

	return false;
}

String File::createLegalPathName (const String& original)
{
	String s (original);
	String start;

	if (s[1] == ':')
	{
		start = s.substring (0, 2);
		s = s.substring (2);
	}

	return start + s.removeCharacters ("\"#@,;:<>*^|?")
					.substring (0, 1024);
}

String File::createLegalFileName (const String& original)
{
	String s (original.removeCharacters ("\"#@,;:<>*^|?\\/"));

	const int maxLength = 128; // only the length of the filename, not the whole path
	const int len = s.length();

	if (len > maxLength)
	{
		const int lastDot = s.lastIndexOfChar ('.');

		if (lastDot > jmax (0, len - 12))
		{
			s = s.substring (0, maxLength - (len - lastDot))
				 + s.substring (lastDot);
		}
		else
		{
			s = s.substring (0, maxLength);
		}
	}

	return s;
}

String File::getRelativePathFrom (const File& dir)  const
{
	String thisPath (fullPath);

	while (thisPath.endsWithChar (separator))
		thisPath = thisPath.dropLastCharacters (1);

	String dirPath (addTrailingSeparator (dir.existsAsFile() ? dir.getParentDirectory().getFullPathName()
															 : dir.fullPath));

	const int len = jmin (thisPath.length(), dirPath.length());
	int commonBitLength = 0;

	{
		String::CharPointerType thisPathIter (thisPath.getCharPointer());
		String::CharPointerType dirPathIter (dirPath.getCharPointer());

		for (int i = 0; i < len; ++i)
		{
			const juce_wchar c1 = thisPathIter.getAndAdvance();
			const juce_wchar c2 = dirPathIter.getAndAdvance();

		   #if NAMES_ARE_CASE_SENSITIVE
			if (c1 != c2)
		   #else
			if (c1 != c2 && CharacterFunctions::toLowerCase (c1) != CharacterFunctions::toLowerCase (c2))
		   #endif
				break;

			++commonBitLength;
		}
	}

	while (commonBitLength > 0 && thisPath [commonBitLength - 1] != File::separator)
		--commonBitLength;

	// if the only common bit is the root, then just return the full path..
	if (commonBitLength <= 0
		 || (commonBitLength == 1 && thisPath [1] == File::separator))
		return fullPath;

	thisPath = thisPath.substring (commonBitLength);
	dirPath  = dirPath.substring (commonBitLength);

	while (dirPath.isNotEmpty())
	{
	   #if JUCE_WINDOWS
		thisPath = "..\\" + thisPath;
	   #else
		thisPath = "../" + thisPath;
	   #endif

		const int sep = dirPath.indexOfChar (separator);

		if (sep >= 0)
			dirPath = dirPath.substring (sep + 1);
		else
			dirPath = String::empty;
	}

	return thisPath;
}

File File::createTempFile (const String& fileNameEnding)
{
	const File tempFile (getSpecialLocation (tempDirectory)
							.getChildFile ("temp_" + String (Random::getSystemRandom().nextInt()))
							.withFileExtension (fileNameEnding));

	if (tempFile.exists())
		return createTempFile (fileNameEnding);

	return tempFile;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_File.cpp ***/


/*** Start of inlined file: juce_FileInputStream.cpp ***/
BEGIN_JUCE_NAMESPACE

int64 juce_fileSetPosition (void* handle, int64 pos);

FileInputStream::FileInputStream (const File& f)
	: file (f),
	  fileHandle (nullptr),
	  currentPosition (0),
	  totalSize (0),
	  status (Result::ok()),
	  needToSeek (true)
{
	openHandle();
}

FileInputStream::~FileInputStream()
{
	closeHandle();
}

int64 FileInputStream::getTotalLength()
{
	return totalSize;
}

int FileInputStream::read (void* buffer, int bytesToRead)
{
	if (needToSeek)
	{
		if (juce_fileSetPosition (fileHandle, currentPosition) < 0)
			return 0;

		needToSeek = false;
	}

	const size_t num = readInternal (buffer, bytesToRead);
	currentPosition += num;

	return (int) num;
}

bool FileInputStream::isExhausted()
{
	return currentPosition >= totalSize;
}

int64 FileInputStream::getPosition()
{
	return currentPosition;
}

bool FileInputStream::setPosition (int64 pos)
{
	pos = jlimit ((int64) 0, totalSize, pos);

	needToSeek |= (currentPosition != pos);
	currentPosition = pos;

	return true;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileInputStream.cpp ***/


/*** Start of inlined file: juce_FileOutputStream.cpp ***/
BEGIN_JUCE_NAMESPACE

int64 juce_fileSetPosition (void* handle, int64 pos);

FileOutputStream::FileOutputStream (const File& f, const int bufferSize_)
	: file (f),
	  fileHandle (nullptr),
	  status (Result::ok()),
	  currentPosition (0),
	  bufferSize (bufferSize_),
	  bytesInBuffer (0),
	  buffer (jmax (bufferSize_, 16))
{
	openHandle();
}

FileOutputStream::~FileOutputStream()
{
	flushBuffer();
	flushInternal();
	closeHandle();
}

int64 FileOutputStream::getPosition()
{
	return currentPosition;
}

bool FileOutputStream::setPosition (int64 newPosition)
{
	if (newPosition != currentPosition)
	{
		flushBuffer();
		currentPosition = juce_fileSetPosition (fileHandle, newPosition);
	}

	return newPosition == currentPosition;
}

bool FileOutputStream::flushBuffer()
{
	bool ok = true;

	if (bytesInBuffer > 0)
	{
		ok = (writeInternal (buffer, bytesInBuffer) == bytesInBuffer);
		bytesInBuffer = 0;
	}

	return ok;
}

void FileOutputStream::flush()
{
	flushBuffer();
	flushInternal();
}

bool FileOutputStream::write (const void* const src, const int numBytes)
{
	if (bytesInBuffer + numBytes < bufferSize)
	{
		memcpy (buffer + bytesInBuffer, src, numBytes);
		bytesInBuffer += numBytes;
		currentPosition += numBytes;
	}
	else
	{
		if (! flushBuffer())
			return false;

		if (numBytes < bufferSize)
		{
			memcpy (buffer + bytesInBuffer, src, numBytes);
			bytesInBuffer += numBytes;
			currentPosition += numBytes;
		}
		else
		{
			const int bytesWritten = writeInternal (src, numBytes);

			if (bytesWritten < 0)
				return false;

			currentPosition += bytesWritten;
			return bytesWritten == numBytes;
		}
	}

	return true;
}

void FileOutputStream::writeRepeatedByte (uint8 byte, int numBytes)
{
	if (bytesInBuffer + numBytes < bufferSize)
	{
		memset (buffer + bytesInBuffer, byte, numBytes);
		bytesInBuffer += numBytes;
		currentPosition += numBytes;
	}
	else
	{
		OutputStream::writeRepeatedByte (byte, numBytes);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileOutputStream.cpp ***/


/*** Start of inlined file: juce_FileSearchPath.cpp ***/
BEGIN_JUCE_NAMESPACE

FileSearchPath::FileSearchPath()
{
}

FileSearchPath::FileSearchPath (const String& path)
{
	init (path);
}

FileSearchPath::FileSearchPath (const FileSearchPath& other)
  : directories (other.directories)
{
}

FileSearchPath::~FileSearchPath()
{
}

FileSearchPath& FileSearchPath::operator= (const String& path)
{
	init (path);
	return *this;
}

void FileSearchPath::init (const String& path)
{
	directories.clear();
	directories.addTokens (path, ";", "\"");
	directories.trim();
	directories.removeEmptyStrings();

	for (int i = directories.size(); --i >= 0;)
		directories.set (i, directories[i].unquoted());
}

int FileSearchPath::getNumPaths() const
{
	return directories.size();
}

File FileSearchPath::operator[] (const int index) const
{
	return File (directories [index]);
}

String FileSearchPath::toString() const
{
	StringArray directories2 (directories);
	for (int i = directories2.size(); --i >= 0;)
		if (directories2[i].containsChar (';'))
			directories2.set (i, directories2[i].quoted());

	return directories2.joinIntoString (";");
}

void FileSearchPath::add (const File& dir, const int insertIndex)
{
	directories.insert (insertIndex, dir.getFullPathName());
}

void FileSearchPath::addIfNotAlreadyThere (const File& dir)
{
	for (int i = 0; i < directories.size(); ++i)
		if (File (directories[i]) == dir)
			return;

	add (dir);
}

void FileSearchPath::remove (const int index)
{
	directories.remove (index);
}

void FileSearchPath::addPath (const FileSearchPath& other)
{
	for (int i = 0; i < other.getNumPaths(); ++i)
		addIfNotAlreadyThere (other[i]);
}

void FileSearchPath::removeRedundantPaths()
{
	for (int i = directories.size(); --i >= 0;)
	{
		const File d1 (directories[i]);

		for (int j = directories.size(); --j >= 0;)
		{
			const File d2 (directories[j]);

			if ((i != j) && (d1.isAChildOf (d2) || d1 == d2))
			{
				directories.remove (i);
				break;
			}
		}
	}
}

void FileSearchPath::removeNonExistentPaths()
{
	for (int i = directories.size(); --i >= 0;)
		if (! File (directories[i]).isDirectory())
			directories.remove (i);
}

int FileSearchPath::findChildFiles (Array<File>& results,
									const int whatToLookFor,
									const bool searchRecursively,
									const String& wildCardPattern) const
{
	int total = 0;

	for (int i = 0; i < directories.size(); ++i)
		total += operator[] (i).findChildFiles (results,
												whatToLookFor,
												searchRecursively,
												wildCardPattern);

	return total;
}

bool FileSearchPath::isFileInPath (const File& fileToCheck,
								   const bool checkRecursively) const
{
	for (int i = directories.size(); --i >= 0;)
	{
		const File d (directories[i]);

		if (checkRecursively)
		{
			if (fileToCheck.isAChildOf (d))
				return true;
		}
		else
		{
			if (fileToCheck.getParentDirectory() == d)
				return true;
		}
	}

	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileSearchPath.cpp ***/


/*** Start of inlined file: juce_NamedPipe.cpp ***/
BEGIN_JUCE_NAMESPACE

NamedPipe::NamedPipe()
	: internal (nullptr)
{
}

NamedPipe::~NamedPipe()
{
	close();
}

bool NamedPipe::openExisting (const String& pipeName)
{
	currentPipeName = pipeName;
	return openInternal (pipeName, false);
}

bool NamedPipe::createNewPipe (const String& pipeName)
{
	currentPipeName = pipeName;
	return openInternal (pipeName, true);
}

bool NamedPipe::isOpen() const
{
	return internal != nullptr;
}

String NamedPipe::getName() const
{
	return currentPipeName;
}

// other methods for this class are implemented in the platform-specific files

END_JUCE_NAMESPACE

/*** End of inlined file: juce_NamedPipe.cpp ***/


/*** Start of inlined file: juce_TemporaryFile.cpp ***/
BEGIN_JUCE_NAMESPACE

TemporaryFile::TemporaryFile (const String& suffix, const int optionFlags)
{
	createTempFile (File::getSpecialLocation (File::tempDirectory),
					"temp_" + String (Random::getSystemRandom().nextInt()),
					suffix,
					optionFlags);
}

TemporaryFile::TemporaryFile (const File& targetFile_, const int optionFlags)
	: targetFile (targetFile_)
{
	// If you use this constructor, you need to give it a valid target file!
	jassert (targetFile != File::nonexistent);

	createTempFile (targetFile.getParentDirectory(),
					targetFile.getFileNameWithoutExtension() + "_temp" + String (Random::getSystemRandom().nextInt()),
					targetFile.getFileExtension(),
					optionFlags);
}

void TemporaryFile::createTempFile (const File& parentDirectory, String name,
									const String& suffix, const int optionFlags)
{
	if ((optionFlags & useHiddenFile) != 0)
		name = "." + name;

	temporaryFile = parentDirectory.getNonexistentChildFile (name, suffix, (optionFlags & putNumbersInBrackets) != 0);
}

TemporaryFile::~TemporaryFile()
{
	if (! deleteTemporaryFile())
	{
		/* Failed to delete our temporary file! The most likely reason for this would be
		   that you've not closed an output stream that was being used to write to file.

		   If you find that something beyond your control is changing permissions on
		   your temporary files and preventing them from being deleted, you may want to
		   call TemporaryFile::deleteTemporaryFile() to detect those error cases and
		   handle them appropriately.
		*/
		jassertfalse;
	}
}

bool TemporaryFile::overwriteTargetFileWithTemporary() const
{
	// This method only works if you created this object with the constructor
	// that takes a target file!
	jassert (targetFile != File::nonexistent);

	if (temporaryFile.exists())
	{
		// Have a few attempts at overwriting the file before giving up..
		for (int i = 5; --i >= 0;)
		{
			if (temporaryFile.moveFileTo (targetFile))
				return true;

			Thread::sleep (100);
		}
	}
	else
	{
		// There's no temporary file to use. If your write failed, you should
		// probably check, and not bother calling this method.
		jassertfalse;
	}

	return false;
}

bool TemporaryFile::deleteTemporaryFile() const
{
	// Have a few attempts at deleting the file before giving up..
	for (int i = 5; --i >= 0;)
	{
		if (temporaryFile.deleteFile())
			return true;

		Thread::sleep (50);
	}

	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TemporaryFile.cpp ***/


/*** Start of inlined file: juce_Socket.cpp ***/
#if JUCE_WINDOWS
 #include <winsock2.h>
 #include <imm.h>
 #include <ws2tcpip.h>

 #if JUCE_MSVC
  #pragma warning (push)
  #pragma warning (disable : 4127 4389 4018)
 #endif

#else
 #if JUCE_LINUX || JUCE_ANDROID
  #include <sys/types.h>
  #include <sys/socket.h>
  #include <sys/errno.h>
  #include <unistd.h>
  #include <netinet/in.h>
 #endif

 #include <fcntl.h>
 #include <netdb.h>
 #include <arpa/inet.h>
 #include <netinet/tcp.h>
#endif

#ifndef AI_NUMERICSERV  // (missing in older Mac SDKs)
 #define AI_NUMERICSERV 0x1000
#endif

BEGIN_JUCE_NAMESPACE

#if JUCE_WINDOWS
 typedef int	   juce_socklen_t;
#else
 typedef socklen_t juce_socklen_t;
#endif

namespace SocketHelpers
{
	void initSockets()
	{
	   #if JUCE_WINDOWS
		static bool socketsStarted = false;

		if (! socketsStarted)
		{
			socketsStarted = true;

			WSADATA wsaData;
			const WORD wVersionRequested = MAKEWORD (1, 1);
			WSAStartup (wVersionRequested, &wsaData);
		}
	   #endif
	}

	bool resetSocketOptions (const int handle, const bool isDatagram, const bool allowBroadcast) noexcept
	{
		const int sndBufSize = 65536;
		const int rcvBufSize = 65536;
		const int one = 1;

		return handle > 0
				&& setsockopt (handle, SOL_SOCKET, SO_RCVBUF, (const char*) &rcvBufSize, sizeof (rcvBufSize)) == 0
				&& setsockopt (handle, SOL_SOCKET, SO_SNDBUF, (const char*) &sndBufSize, sizeof (sndBufSize)) == 0
				&& (isDatagram ? ((! allowBroadcast) || setsockopt (handle, SOL_SOCKET, SO_BROADCAST, (const char*) &one, sizeof (one)) == 0)
							   : (setsockopt (handle, IPPROTO_TCP, TCP_NODELAY, (const char*) &one, sizeof (one)) == 0));
	}

	bool bindSocketToPort (const int handle, const int port) noexcept
	{
		if (handle <= 0 || port <= 0)
			return false;

		struct sockaddr_in servTmpAddr = { 0 };
		servTmpAddr.sin_family = PF_INET;
		servTmpAddr.sin_addr.s_addr = htonl (INADDR_ANY);
		servTmpAddr.sin_port = htons ((uint16) port);

		return bind (handle, (struct sockaddr*) &servTmpAddr, sizeof (struct sockaddr_in)) >= 0;
	}

	int readSocket (const int handle,
					void* const destBuffer, const int maxBytesToRead,
					bool volatile& connected,
					const bool blockUntilSpecifiedAmountHasArrived) noexcept
	{
		int bytesRead = 0;

		while (bytesRead < maxBytesToRead)
		{
			int bytesThisTime;

		   #if JUCE_WINDOWS
			bytesThisTime = recv (handle, static_cast<char*> (destBuffer) + bytesRead, maxBytesToRead - bytesRead, 0);
		   #else
			while ((bytesThisTime = (int) ::read (handle, addBytesToPointer (destBuffer, bytesRead), maxBytesToRead - bytesRead)) < 0
					 && errno == EINTR
					 && connected)
			{
			}
		   #endif

			if (bytesThisTime <= 0 || ! connected)
			{
				if (bytesRead == 0)
					bytesRead = -1;

				break;
			}

			bytesRead += bytesThisTime;

			if (! blockUntilSpecifiedAmountHasArrived)
				break;
		}

		return bytesRead;
	}

	int waitForReadiness (const int handle, const bool forReading, const int timeoutMsecs) noexcept
	{
		struct timeval timeout;
		struct timeval* timeoutp;

		if (timeoutMsecs >= 0)
		{
			timeout.tv_sec = timeoutMsecs / 1000;
			timeout.tv_usec = (timeoutMsecs % 1000) * 1000;
			timeoutp = &timeout;
		}
		else
		{
			timeoutp = 0;
		}

		fd_set rset, wset;
		FD_ZERO (&rset);
		FD_SET (handle, &rset);
		FD_ZERO (&wset);
		FD_SET (handle, &wset);

		fd_set* const prset = forReading ? &rset : nullptr;
		fd_set* const pwset = forReading ? nullptr : &wset;

	   #if JUCE_WINDOWS
		if (select (handle + 1, prset, pwset, 0, timeoutp) < 0)
			return -1;
	   #else
		{
			int result;
			while ((result = select (handle + 1, prset, pwset, 0, timeoutp)) < 0
					&& errno == EINTR)
			{
			}

			if (result < 0)
				return -1;
		}
	   #endif

		{
			int opt;
			juce_socklen_t len = sizeof (opt);

			if (getsockopt (handle, SOL_SOCKET, SO_ERROR, (char*) &opt, &len) < 0
				 || opt != 0)
				return -1;
		}

		return FD_ISSET (handle, forReading ? &rset : &wset) ? 1 : 0;
	}

	bool setSocketBlockingState (const int handle, const bool shouldBlock) noexcept
	{
	   #if JUCE_WINDOWS
		u_long nonBlocking = shouldBlock ? 0 : 1;
		return ioctlsocket (handle, FIONBIO, &nonBlocking) == 0;
	   #else
		int socketFlags = fcntl (handle, F_GETFL, 0);

		if (socketFlags == -1)
			return false;

		if (shouldBlock)
			socketFlags &= ~O_NONBLOCK;
		else
			socketFlags |= O_NONBLOCK;

		return fcntl (handle, F_SETFL, socketFlags) == 0;
	   #endif
	}

	bool connectSocket (int volatile& handle,
						const bool isDatagram,
						void** serverAddress,
						const String& hostName,
						const int portNumber,
						const int timeOutMillisecs) noexcept
	{
		struct addrinfo hints = { 0 };
		hints.ai_family = AF_UNSPEC;
		hints.ai_socktype = isDatagram ? SOCK_DGRAM : SOCK_STREAM;
		hints.ai_flags = AI_NUMERICSERV;

		struct addrinfo* info = nullptr;
		if (getaddrinfo (hostName.toUTF8(), String (portNumber).toUTF8(), &hints, &info) != 0 || info == 0)
			return false;

		if (handle < 0)
			handle = (int) socket (info->ai_family, info->ai_socktype, 0);

		if (handle < 0)
		{
			freeaddrinfo (info);
			return false;
		}

		if (isDatagram)
		{
			struct sockaddr* s = new struct sockaddr();
			*s = *(info->ai_addr);
			*serverAddress = s;

			freeaddrinfo (info);
			return true;
		}

		setSocketBlockingState (handle, false);
		const int result = ::connect (handle, info->ai_addr, (int) info->ai_addrlen);
		freeaddrinfo (info);

		if (result < 0)
		{
		   #if JUCE_WINDOWS
			if (result == SOCKET_ERROR && WSAGetLastError() == WSAEWOULDBLOCK)
		   #else
			if (errno == EINPROGRESS)
		   #endif
			{
				if (waitForReadiness (handle, false, timeOutMillisecs) != 1)
				{
					setSocketBlockingState (handle, true);
					return false;
				}
			}
		}

		setSocketBlockingState (handle, true);
		resetSocketOptions (handle, false, false);

		return true;
	}
}

StreamingSocket::StreamingSocket()
	: portNumber (0),
	  handle (-1),
	  connected (false),
	  isListener (false)
{
	SocketHelpers::initSockets();
}

StreamingSocket::StreamingSocket (const String& hostName_,
								  const int portNumber_,
								  const int handle_)
	: hostName (hostName_),
	  portNumber (portNumber_),
	  handle (handle_),
	  connected (true),
	  isListener (false)
{
	SocketHelpers::initSockets();
	SocketHelpers::resetSocketOptions (handle_, false, false);
}

StreamingSocket::~StreamingSocket()
{
	close();
}

int StreamingSocket::read (void* destBuffer, const int maxBytesToRead, const bool blockUntilSpecifiedAmountHasArrived)
{
	return (connected && ! isListener) ? SocketHelpers::readSocket (handle, destBuffer, maxBytesToRead, connected, blockUntilSpecifiedAmountHasArrived)
									   : -1;
}

int StreamingSocket::write (const void* sourceBuffer, const int numBytesToWrite)
{
	if (isListener || ! connected)
		return -1;

   #if JUCE_WINDOWS
	return send (handle, (const char*) sourceBuffer, numBytesToWrite, 0);
   #else
	int result;

	while ((result = (int) ::write (handle, sourceBuffer, numBytesToWrite)) < 0
			&& errno == EINTR)
	{
	}

	return result;
   #endif
}

int StreamingSocket::waitUntilReady (const bool readyForReading,
									 const int timeoutMsecs) const
{
	return connected ? SocketHelpers::waitForReadiness (handle, readyForReading, timeoutMsecs)
					 : -1;
}

bool StreamingSocket::bindToPort (const int port)
{
	return SocketHelpers::bindSocketToPort (handle, port);
}

bool StreamingSocket::connect (const String& remoteHostName,
							   const int remotePortNumber,
							   const int timeOutMillisecs)
{
	if (isListener)
	{
		jassertfalse;	// a listener socket can't connect to another one!
		return false;
	}

	if (connected)
		close();

	hostName = remoteHostName;
	portNumber = remotePortNumber;
	isListener = false;

	connected = SocketHelpers::connectSocket (handle, false, 0, remoteHostName,
											  remotePortNumber, timeOutMillisecs);

	if (! (connected && SocketHelpers::resetSocketOptions (handle, false, false)))
	{
		close();
		return false;
	}

	return true;
}

void StreamingSocket::close()
{
   #if JUCE_WINDOWS
	if (handle != SOCKET_ERROR || connected)
		closesocket (handle);

	connected = false;
   #else
	if (connected)
	{
		connected = false;

		if (isListener)
		{
			// need to do this to interrupt the accept() function..
			StreamingSocket temp;
			temp.connect ("localhost", portNumber, 1000);
		}
	}

	if (handle != -1)
		::close (handle);
   #endif

	hostName = String::empty;
	portNumber = 0;
	handle = -1;
	isListener = false;
}

bool StreamingSocket::createListener (const int newPortNumber, const String& localHostName)
{
	if (connected)
		close();

	hostName = "listener";
	portNumber = newPortNumber;
	isListener = true;

	struct sockaddr_in servTmpAddr = { 0 };
	servTmpAddr.sin_family = PF_INET;
	servTmpAddr.sin_addr.s_addr = htonl (INADDR_ANY);

	if (localHostName.isNotEmpty())
		servTmpAddr.sin_addr.s_addr = ::inet_addr (localHostName.toUTF8());

	servTmpAddr.sin_port = htons ((uint16) portNumber);

	handle = (int) socket (AF_INET, SOCK_STREAM, 0);

	if (handle < 0)
		return false;

	const int reuse = 1;
	setsockopt (handle, SOL_SOCKET, SO_REUSEADDR, (const char*) &reuse, sizeof (reuse));

	if (bind (handle, (struct sockaddr*) &servTmpAddr, sizeof (struct sockaddr_in)) < 0
		 || listen (handle, SOMAXCONN) < 0)
	{
		close();
		return false;
	}

	connected = true;
	return true;
}

StreamingSocket* StreamingSocket::waitForNextConnection() const
{
	jassert (isListener || ! connected); // to call this method, you first have to use createListener() to
										 // prepare this socket as a listener.

	if (connected && isListener)
	{
		struct sockaddr address;
		juce_socklen_t len = sizeof (sockaddr);
		const int newSocket = (int) accept (handle, &address, &len);

		if (newSocket >= 0 && connected)
			return new StreamingSocket (inet_ntoa (((struct sockaddr_in*) &address)->sin_addr),
										portNumber, newSocket);
	}

	return nullptr;
}

bool StreamingSocket::isLocal() const noexcept
{
	return hostName == "127.0.0.1";
}

DatagramSocket::DatagramSocket (const int localPortNumber, const bool allowBroadcast_)
	: portNumber (0),
	  handle (-1),
	  connected (true),
	  allowBroadcast (allowBroadcast_),
	  serverAddress (0)
{
	SocketHelpers::initSockets();

	handle = (int) socket (AF_INET, SOCK_DGRAM, 0);
	bindToPort (localPortNumber);
}

DatagramSocket::DatagramSocket (const String& hostName_, const int portNumber_,
								const int handle_, const int localPortNumber)
	: hostName (hostName_),
	  portNumber (portNumber_),
	  handle (handle_),
	  connected (true),
	  allowBroadcast (false),
	  serverAddress (0)
{
	SocketHelpers::initSockets();

	SocketHelpers::resetSocketOptions (handle_, true, allowBroadcast);
	bindToPort (localPortNumber);
}

DatagramSocket::~DatagramSocket()
{
	close();

	delete static_cast <struct sockaddr*> (serverAddress);
	serverAddress = 0;
}

void DatagramSocket::close()
{
   #if JUCE_WINDOWS
	closesocket (handle);
	connected = false;
   #else
	connected = false;
	::close (handle);
   #endif

	hostName = String::empty;
	portNumber = 0;
	handle = -1;
}

bool DatagramSocket::bindToPort (const int port)
{
	return SocketHelpers::bindSocketToPort (handle, port);
}

bool DatagramSocket::connect (const String& remoteHostName,
							  const int remotePortNumber,
							  const int timeOutMillisecs)
{
	if (connected)
		close();

	hostName = remoteHostName;
	portNumber = remotePortNumber;

	connected = SocketHelpers::connectSocket (handle, true, &serverAddress,
											  remoteHostName, remotePortNumber,
											  timeOutMillisecs);

	if (! (connected && SocketHelpers::resetSocketOptions (handle, true, allowBroadcast)))
	{
		close();
		return false;
	}

	return true;
}

DatagramSocket* DatagramSocket::waitForNextConnection() const
{
	struct sockaddr address;
	juce_socklen_t len = sizeof (sockaddr);

	while (waitUntilReady (true, -1) == 1)
	{
		char buf[1];

		if (recvfrom (handle, buf, 0, 0, &address, &len) > 0)
		{
			return new DatagramSocket (inet_ntoa (((struct sockaddr_in*) &address)->sin_addr),
									   ntohs (((struct sockaddr_in*) &address)->sin_port),
									   -1, -1);
		}
	}

	return nullptr;
}

int DatagramSocket::waitUntilReady (const bool readyForReading,
									const int timeoutMsecs) const
{
	return connected ? SocketHelpers::waitForReadiness (handle, readyForReading, timeoutMsecs)
					 : -1;
}

int DatagramSocket::read (void* destBuffer, const int maxBytesToRead, const bool blockUntilSpecifiedAmountHasArrived)
{
	return connected ? SocketHelpers::readSocket (handle, destBuffer, maxBytesToRead, connected, blockUntilSpecifiedAmountHasArrived)
					 : -1;
}

int DatagramSocket::write (const void* sourceBuffer, const int numBytesToWrite)
{
	// You need to call connect() first to set the server address..
	jassert (serverAddress != 0 && connected);

	return connected ? (int) sendto (handle, (const char*) sourceBuffer,
									 numBytesToWrite, 0,
									 (const struct sockaddr*) serverAddress,
									 sizeof (struct sockaddr_in))
					 : -1;
}

bool DatagramSocket::isLocal() const noexcept
{
	return hostName == "127.0.0.1";
}

#if JUCE_MSVC
 #pragma warning (pop)
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Socket.cpp ***/


/*** Start of inlined file: juce_URL.cpp ***/
BEGIN_JUCE_NAMESPACE

URL::URL()
{
}

URL::URL (const String& url_)
	: url (url_)
{
	int i = url.indexOfChar ('?');

	if (i >= 0)
	{
		do
		{
			const int nextAmp   = url.indexOfChar (i + 1, '&');
			const int equalsPos = url.indexOfChar (i + 1, '=');

			if (equalsPos > i + 1)
			{
				if (nextAmp < 0)
				{
					parameters.set (removeEscapeChars (url.substring (i + 1, equalsPos)),
									removeEscapeChars (url.substring (equalsPos + 1)));
				}
				else if (nextAmp > 0 && equalsPos < nextAmp)
				{
					parameters.set (removeEscapeChars (url.substring (i + 1, equalsPos)),
									removeEscapeChars (url.substring (equalsPos + 1, nextAmp)));
				}
			}

			i = nextAmp;
		}
		while (i >= 0);

		url = url.upToFirstOccurrenceOf ("?", false, false);
	}
}

URL::URL (const URL& other)
	: url (other.url),
	  postData (other.postData),
	  parameters (other.parameters),
	  filesToUpload (other.filesToUpload),
	  mimeTypes (other.mimeTypes)
{
}

URL& URL::operator= (const URL& other)
{
	url = other.url;
	postData = other.postData;
	parameters = other.parameters;
	filesToUpload = other.filesToUpload;
	mimeTypes = other.mimeTypes;

	return *this;
}

URL::~URL()
{
}

namespace URLHelpers
{
	String getMangledParameters (const StringPairArray& parameters)
	{
		String p;

		for (int i = 0; i < parameters.size(); ++i)
		{
			if (i > 0)
				p << '&';

			p << URL::addEscapeChars (parameters.getAllKeys() [i], true)
			  << '='
			  << URL::addEscapeChars (parameters.getAllValues() [i], true);
		}

		return p;
	}

	int findStartOfDomain (const String& url)
	{
		int i = 0;

		while (CharacterFunctions::isLetterOrDigit (url[i])
				|| url[i] == '+' || url[i] == '-' || url[i] == '.')
			++i;

		return url[i] == ':' ? i + 1 : 0;
	}

	void createHeadersAndPostData (const URL& url, String& headers, MemoryBlock& postData)
	{
		MemoryOutputStream data (postData, false);

		if (url.getFilesToUpload().size() > 0)
		{
			// need to upload some files, so do it as multi-part...
			const String boundary (String::toHexString (Random::getSystemRandom().nextInt64()));

			headers << "Content-Type: multipart/form-data; boundary=" << boundary << "\r\n";

			data << "--" << boundary;

			int i;
			for (i = 0; i < url.getParameters().size(); ++i)
			{
				data << "\r\nContent-Disposition: form-data; name=\""
					 << url.getParameters().getAllKeys() [i]
					 << "\"\r\n\r\n"
					 << url.getParameters().getAllValues() [i]
					 << "\r\n--"
					 << boundary;
			}

			for (i = 0; i < url.getFilesToUpload().size(); ++i)
			{
				const File file (url.getFilesToUpload().getAllValues() [i]);
				const String paramName (url.getFilesToUpload().getAllKeys() [i]);

				data << "\r\nContent-Disposition: form-data; name=\"" << paramName
					 << "\"; filename=\"" << file.getFileName() << "\"\r\n";

				const String mimeType (url.getMimeTypesOfUploadFiles()
										  .getValue (paramName, String::empty));

				if (mimeType.isNotEmpty())
					data << "Content-Type: " << mimeType << "\r\n";

				data << "Content-Transfer-Encoding: binary\r\n\r\n"
					 << file << "\r\n--" << boundary;
			}

			data << "--\r\n";
			data.flush();
		}
		else
		{
			data << getMangledParameters (url.getParameters()) << url.getPostData();
			data.flush();

			// just a short text attachment, so use simple url encoding..
			headers << "Content-Type: application/x-www-form-urlencoded\r\nContent-length: "
					<< (int) postData.getSize() << "\r\n";
		}
	}
}

String URL::toString (const bool includeGetParameters) const
{
	if (includeGetParameters && parameters.size() > 0)
		return url + "?" + URLHelpers::getMangledParameters (parameters);
	else
		return url;
}

bool URL::isWellFormed() const
{
	//xxx TODO
	return url.isNotEmpty();
}

String URL::getDomain() const
{
	int start = URLHelpers::findStartOfDomain (url);
	while (url[start] == '/')
		++start;

	const int end1 = url.indexOfChar (start, '/');
	const int end2 = url.indexOfChar (start, ':');

	const int end = (end1 < 0 || end2 < 0) ? jmax (end1, end2)
										   : jmin (end1, end2);

	return url.substring (start, end);
}

String URL::getSubPath() const
{
	int start = URLHelpers::findStartOfDomain (url);
	while (url[start] == '/')
		++start;

	const int startOfPath = url.indexOfChar (start, '/') + 1;

	return startOfPath <= 0 ? String::empty
							: url.substring (startOfPath);
}

String URL::getScheme() const
{
	return url.substring (0, URLHelpers::findStartOfDomain (url) - 1);
}

const URL URL::withNewSubPath (const String& newPath) const
{
	int start = URLHelpers::findStartOfDomain (url);
	while (url[start] == '/')
		++start;

	const int startOfPath = url.indexOfChar (start, '/') + 1;

	URL u (*this);

	if (startOfPath > 0)
		u.url = url.substring (0, startOfPath);

	if (! u.url.endsWithChar ('/'))
		u.url << '/';

	if (newPath.startsWithChar ('/'))
		u.url << newPath.substring (1);
	else
		u.url << newPath;

	return u;
}

bool URL::isProbablyAWebsiteURL (const String& possibleURL)
{
	const char* validProtocols[] = { "http:", "ftp:", "https:" };

	for (int i = 0; i < numElementsInArray (validProtocols); ++i)
		if (possibleURL.startsWithIgnoreCase (validProtocols[i]))
			return true;

	if (possibleURL.containsChar ('@')
		 || possibleURL.containsChar (' '))
		return false;

	const String topLevelDomain (possibleURL.upToFirstOccurrenceOf ("/", false, false)
											.fromLastOccurrenceOf (".", false, false));

	return topLevelDomain.isNotEmpty() && topLevelDomain.length() <= 3;
}

bool URL::isProbablyAnEmailAddress (const String& possibleEmailAddress)
{
	const int atSign = possibleEmailAddress.indexOfChar ('@');

	return atSign > 0
			&& possibleEmailAddress.lastIndexOfChar ('.') > (atSign + 1)
			&& (! possibleEmailAddress.endsWithChar ('.'));
}

const URL URL::withParameter (const String& parameterName,
							  const String& parameterValue) const
{
	URL u (*this);
	u.parameters.set (parameterName, parameterValue);
	return u;
}

const URL URL::withFileToUpload (const String& parameterName,
								 const File& fileToUpload,
								 const String& mimeType) const
{
	jassert (mimeType.isNotEmpty()); // You need to supply a mime type!

	URL u (*this);
	u.filesToUpload.set (parameterName, fileToUpload.getFullPathName());
	u.mimeTypes.set (parameterName, mimeType);
	return u;
}

const URL URL::withPOSTData (const String& postData_) const
{
	URL u (*this);
	u.postData = postData_;
	return u;
}

const StringPairArray& URL::getParameters() const
{
	return parameters;
}

const StringPairArray& URL::getFilesToUpload() const
{
	return filesToUpload;
}

const StringPairArray& URL::getMimeTypesOfUploadFiles() const
{
	return mimeTypes;
}

String URL::removeEscapeChars (const String& s)
{
	String result (s.replaceCharacter ('+', ' '));

	if (! result.containsChar ('%'))
		return result;

	// We need to operate on the string as raw UTF8 chars, and then recombine them into unicode
	// after all the replacements have been made, so that multi-byte chars are handled.
	Array<char> utf8 (result.toUTF8().getAddress(), result.getNumBytesAsUTF8());

	for (int i = 0; i < utf8.size(); ++i)
	{
		if (utf8.getUnchecked(i) == '%')
		{
			const int hexDigit1 = CharacterFunctions::getHexDigitValue (utf8 [i + 1]);
			const int hexDigit2 = CharacterFunctions::getHexDigitValue (utf8 [i + 2]);

			if (hexDigit1 >= 0 && hexDigit2 >= 0)
			{
				utf8.set (i, (char) ((hexDigit1 << 4) + hexDigit2));
				utf8.removeRange (i + 1, 2);
			}
		}
	}

	return String::fromUTF8 (utf8.getRawDataPointer(), utf8.size());
}

String URL::addEscapeChars (const String& s, const bool isParameter)
{
	const CharPointer_UTF8 legalChars (isParameter ? "_-.*!'()"
												   : ",$_-.*!'()");

	Array<char> utf8 (s.toUTF8().getAddress(), s.getNumBytesAsUTF8());

	for (int i = 0; i < utf8.size(); ++i)
	{
		const char c = utf8.getUnchecked(i);

		if (! (CharacterFunctions::isLetterOrDigit (c)
				 || legalChars.indexOf ((juce_wchar) c) >= 0))
		{
			if (c == ' ')
			{
				utf8.set (i, '+');
			}
			else
			{
				static const char* const hexDigits = "0123456789abcdef";

				utf8.set (i, '%');
				utf8.insert (++i, hexDigits [((uint8) c) >> 4]);
				utf8.insert (++i, hexDigits [c & 15]);
			}
		}
	}

	return String::fromUTF8 (utf8.getRawDataPointer(), utf8.size());
}

bool URL::launchInDefaultBrowser() const
{
	String u (toString (true));

	if (u.containsChar ('@') && ! u.containsChar (':'))
		u = "mailto:" + u;

	return Process::openDocument (u, String::empty);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_URL.cpp ***/


/*** Start of inlined file: juce_BufferedInputStream.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace
{
	int calcBufferStreamBufferSize (int requestedSize, InputStream* const source) noexcept
	{
		// You need to supply a real stream when creating a BufferedInputStream
		jassert (source != nullptr);

		requestedSize = jmax (256, requestedSize);

		const int64 sourceSize = source->getTotalLength();
		if (sourceSize >= 0 && sourceSize < requestedSize)
			requestedSize = jmax (32, (int) sourceSize);

		return requestedSize;
	}
}

BufferedInputStream::BufferedInputStream (InputStream* const sourceStream, const int bufferSize_,
										  const bool deleteSourceWhenDestroyed)
   : source (sourceStream, deleteSourceWhenDestroyed),
	 bufferSize (calcBufferStreamBufferSize (bufferSize_, sourceStream)),
	 position (sourceStream->getPosition()),
	 lastReadPos (0),
	 bufferStart (position),
	 bufferOverlap (128)
{
	buffer.malloc (bufferSize);
}

BufferedInputStream::BufferedInputStream (InputStream& sourceStream, const int bufferSize_)
   : source (&sourceStream, false),
	 bufferSize (calcBufferStreamBufferSize (bufferSize_, &sourceStream)),
	 position (sourceStream.getPosition()),
	 lastReadPos (0),
	 bufferStart (position),
	 bufferOverlap (128)
{
	buffer.malloc (bufferSize);
}

BufferedInputStream::~BufferedInputStream()
{
}

int64 BufferedInputStream::getTotalLength()
{
	return source->getTotalLength();
}

int64 BufferedInputStream::getPosition()
{
	return position;
}

bool BufferedInputStream::setPosition (int64 newPosition)
{
	position = jmax ((int64) 0, newPosition);
	return true;
}

bool BufferedInputStream::isExhausted()
{
	return (position >= lastReadPos)
			 && source->isExhausted();
}

void BufferedInputStream::ensureBuffered()
{
	const int64 bufferEndOverlap = lastReadPos - bufferOverlap;

	if (position < bufferStart || position >= bufferEndOverlap)
	{
		int bytesRead;

		if (position < lastReadPos
			 && position >= bufferEndOverlap
			 && position >= bufferStart)
		{
			const int bytesToKeep = (int) (lastReadPos - position);
			memmove (buffer, buffer + (int) (position - bufferStart), bytesToKeep);

			bufferStart = position;

			bytesRead = source->read (buffer + bytesToKeep,
									  bufferSize - bytesToKeep);

			lastReadPos += bytesRead;
			bytesRead += bytesToKeep;
		}
		else
		{
			bufferStart = position;
			source->setPosition (bufferStart);
			bytesRead = source->read (buffer, bufferSize);
			lastReadPos = bufferStart + bytesRead;
		}

		while (bytesRead < bufferSize)
			buffer [bytesRead++] = 0;
	}
}

int BufferedInputStream::read (void* destBuffer, int maxBytesToRead)
{
	if (position >= bufferStart
		 && position + maxBytesToRead <= lastReadPos)
	{
		memcpy (destBuffer, buffer + (int) (position - bufferStart), maxBytesToRead);
		position += maxBytesToRead;

		return maxBytesToRead;
	}
	else
	{
		if (position < bufferStart || position >= lastReadPos)
			ensureBuffered();

		int bytesRead = 0;

		while (maxBytesToRead > 0)
		{
			const int bytesAvailable = jmin (maxBytesToRead, (int) (lastReadPos - position));

			if (bytesAvailable > 0)
			{
				memcpy (destBuffer, buffer + (int) (position - bufferStart), bytesAvailable);
				maxBytesToRead -= bytesAvailable;
				bytesRead += bytesAvailable;
				position += bytesAvailable;
				destBuffer = static_cast <char*> (destBuffer) + bytesAvailable;
			}

			const int64 oldLastReadPos = lastReadPos;
			ensureBuffered();

			if (oldLastReadPos == lastReadPos)
				break; // if ensureBuffered() failed to read any more data, bail out

			if (isExhausted())
				break;
		}

		return bytesRead;
	}
}

String BufferedInputStream::readString()
{
	if (position >= bufferStart
		 && position < lastReadPos)
	{
		const int maxChars = (int) (lastReadPos - position);

		const char* const src = buffer + (int) (position - bufferStart);

		for (int i = 0; i < maxChars; ++i)
		{
			if (src[i] == 0)
			{
				position += i + 1;
				return String::fromUTF8 (src, i);
			}
		}
	}

	return InputStream::readString();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_BufferedInputStream.cpp ***/


/*** Start of inlined file: juce_FileInputSource.cpp ***/
BEGIN_JUCE_NAMESPACE

FileInputSource::FileInputSource (const File& file_, bool useFileTimeInHashGeneration_)
	: file (file_), useFileTimeInHashGeneration (useFileTimeInHashGeneration_)
{
}

FileInputSource::~FileInputSource()
{
}

InputStream* FileInputSource::createInputStream()
{
	return file.createInputStream();
}

InputStream* FileInputSource::createInputStreamFor (const String& relatedItemPath)
{
	return file.getSiblingFile (relatedItemPath).createInputStream();
}

int64 FileInputSource::hashCode() const
{
	int64 h = file.hashCode();

	if (useFileTimeInHashGeneration)
		h ^= file.getLastModificationTime().toMilliseconds();

	return h;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileInputSource.cpp ***/


/*** Start of inlined file: juce_MemoryInputStream.cpp ***/
BEGIN_JUCE_NAMESPACE

MemoryInputStream::MemoryInputStream (const void* const sourceData,
									  const size_t sourceDataSize,
									  const bool keepInternalCopy)
	: data (static_cast <const char*> (sourceData)),
	  dataSize (sourceDataSize),
	  position (0)
{
	if (keepInternalCopy)
		createInternalCopy();
}

MemoryInputStream::MemoryInputStream (const MemoryBlock& sourceData,
									  const bool keepInternalCopy)
	: data (static_cast <const char*> (sourceData.getData())),
	  dataSize (sourceData.getSize()),
	  position (0)
{
	if (keepInternalCopy)
		createInternalCopy();
}

void MemoryInputStream::createInternalCopy()
{
	internalCopy.malloc (dataSize);
	memcpy (internalCopy, data, dataSize);
	data = internalCopy;
}

MemoryInputStream::~MemoryInputStream()
{
}

int64 MemoryInputStream::getTotalLength()
{
	return dataSize;
}

int MemoryInputStream::read (void* const buffer, const int howMany)
{
	jassert (howMany >= 0);
	const int num = jmin (howMany, (int) (dataSize - position));
	if (num <= 0)
		return 0;

	memcpy (buffer, data + position, num);
	position += num;
	return (int) num;
}

bool MemoryInputStream::isExhausted()
{
	return position >= dataSize;
}

bool MemoryInputStream::setPosition (const int64 pos)
{
	position = (int) jlimit ((int64) 0, (int64) dataSize, pos);
	return true;
}

int64 MemoryInputStream::getPosition()
{
	return position;
}


END_JUCE_NAMESPACE

/*** End of inlined file: juce_MemoryInputStream.cpp ***/


/*** Start of inlined file: juce_MemoryOutputStream.cpp ***/
BEGIN_JUCE_NAMESPACE

MemoryOutputStream::MemoryOutputStream (const size_t initialSize)
  : data (internalBlock),
	position (0),
	size (0)
{
	internalBlock.setSize (initialSize, false);
}

MemoryOutputStream::MemoryOutputStream (MemoryBlock& memoryBlockToWriteTo,
										const bool appendToExistingBlockContent)
  : data (memoryBlockToWriteTo),
	position (0),
	size (0)
{
	if (appendToExistingBlockContent)
		position = size = memoryBlockToWriteTo.getSize();
}

MemoryOutputStream::~MemoryOutputStream()
{
	trimExternalBlockSize();
}

void MemoryOutputStream::flush()
{
	trimExternalBlockSize();
}

void MemoryOutputStream::trimExternalBlockSize()
{
	if (&data != &internalBlock)
		data.setSize (size, false);
}

void MemoryOutputStream::preallocate (const size_t bytesToPreallocate)
{
	data.ensureSize (bytesToPreallocate + 1);
}

void MemoryOutputStream::reset() noexcept
{
	position = 0;
	size = 0;
}

void MemoryOutputStream::prepareToWrite (int numBytes)
{
	const size_t storageNeeded = position + numBytes;

	if (storageNeeded >= data.getSize())
		data.ensureSize ((storageNeeded + jmin ((int) (storageNeeded / 2), 1024 * 1024) + 32) & ~31);
}

bool MemoryOutputStream::write (const void* const buffer, int howMany)
{
	if (howMany > 0)
	{
		prepareToWrite (howMany);
		memcpy (static_cast<char*> (data.getData()) + position, buffer, howMany);
		position += howMany;
		size = jmax (size, position);
	}

	return true;
}

void MemoryOutputStream::writeRepeatedByte (uint8 byte, int howMany)
{
	if (howMany > 0)
	{
		prepareToWrite (howMany);
		memset (static_cast<char*> (data.getData()) + position, byte, howMany);
		position += howMany;
		size = jmax (size, position);
	}
}

const void* MemoryOutputStream::getData() const noexcept
{
	void* const d = data.getData();

	if (data.getSize() > size)
		static_cast <char*> (d) [size] = 0;

	return d;
}

bool MemoryOutputStream::setPosition (int64 newPosition)
{
	if (newPosition <= (int64) size)
	{
		// ok to seek backwards
		position = jlimit ((size_t) 0, size, (size_t) newPosition);
		return true;
	}
	else
	{
		// trying to make it bigger isn't a good thing to do..
		return false;
	}
}

int MemoryOutputStream::writeFromInputStream (InputStream& source, int64 maxNumBytesToWrite)
{
	// before writing from an input, see if we can preallocate to make it more efficient..
	int64 availableData = source.getTotalLength() - source.getPosition();

	if (availableData > 0)
	{
		if (maxNumBytesToWrite > 0 && maxNumBytesToWrite < availableData)
			availableData = maxNumBytesToWrite;

		preallocate (data.getSize() + (size_t) maxNumBytesToWrite);
	}

	return OutputStream::writeFromInputStream (source, maxNumBytesToWrite);
}

String MemoryOutputStream::toUTF8() const
{
	const char* const d = static_cast <const char*> (getData());
	return String (CharPointer_UTF8 (d), CharPointer_UTF8 (d + getDataSize()));
}

String MemoryOutputStream::toString() const
{
	return String::createStringFromData (getData(), (int) getDataSize());
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const MemoryOutputStream& streamToRead)
{
	stream.write (streamToRead.getData(), (int) streamToRead.getDataSize());
	return stream;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MemoryOutputStream.cpp ***/


/*** Start of inlined file: juce_SubregionStream.cpp ***/
BEGIN_JUCE_NAMESPACE

SubregionStream::SubregionStream (InputStream* const sourceStream,
								  const int64 startPositionInSourceStream_,
								  const int64 lengthOfSourceStream_,
								  const bool deleteSourceWhenDestroyed)
  : source (sourceStream, deleteSourceWhenDestroyed),
	startPositionInSourceStream (startPositionInSourceStream_),
	lengthOfSourceStream (lengthOfSourceStream_)
{
	SubregionStream::setPosition (0);
}

SubregionStream::~SubregionStream()
{
}

int64 SubregionStream::getTotalLength()
{
	const int64 srcLen = source->getTotalLength() - startPositionInSourceStream;

	return (lengthOfSourceStream >= 0) ? jmin (lengthOfSourceStream, srcLen)
									   : srcLen;
}

int64 SubregionStream::getPosition()
{
	return source->getPosition() - startPositionInSourceStream;
}

bool SubregionStream::setPosition (int64 newPosition)
{
	return source->setPosition (jmax ((int64) 0, newPosition + startPositionInSourceStream));
}

int SubregionStream::read (void* destBuffer, int maxBytesToRead)
{
	if (lengthOfSourceStream < 0)
	{
		return source->read (destBuffer, maxBytesToRead);
	}
	else
	{
		maxBytesToRead = (int) jmin ((int64) maxBytesToRead, lengthOfSourceStream - getPosition());

		if (maxBytesToRead <= 0)
			return 0;

		return source->read (destBuffer, maxBytesToRead);
	}
}

bool SubregionStream::isExhausted()
{
	if (lengthOfSourceStream >= 0)
		return (getPosition() >= lengthOfSourceStream) || source->isExhausted();
	else
		return source->isExhausted();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_SubregionStream.cpp ***/


/*** Start of inlined file: juce_CharacterFunctions.cpp ***/
#if JUCE_MSVC
  #pragma warning (push)
  #pragma warning (disable: 4514 4996)
#endif

#if ! JUCE_ANDROID
 #include <cwctype>
#endif

#include <cctype>

BEGIN_JUCE_NAMESPACE

juce_wchar CharacterFunctions::toUpperCase (const juce_wchar character) noexcept
{
	return towupper ((wchar_t) character);
}

juce_wchar CharacterFunctions::toLowerCase (const juce_wchar character) noexcept
{
	return towlower ((wchar_t) character);
}

bool CharacterFunctions::isUpperCase (const juce_wchar character) noexcept
{
   #if JUCE_WINDOWS
	return iswupper ((wchar_t) character) != 0;
   #else
	return toLowerCase (character) != character;
   #endif
}

bool CharacterFunctions::isLowerCase (const juce_wchar character) noexcept
{
   #if JUCE_WINDOWS
	return iswlower ((wchar_t) character) != 0;
   #else
	return toUpperCase (character) != character;
   #endif
}

#if JUCE_MSVC
  #pragma warning (pop)
#endif

bool CharacterFunctions::isWhitespace (const char character) noexcept
{
	return character == ' ' || (character <= 13 && character >= 9);
}

bool CharacterFunctions::isWhitespace (const juce_wchar character) noexcept
{
	return iswspace ((wchar_t) character) != 0;
}

bool CharacterFunctions::isDigit (const char character) noexcept
{
	return (character >= '0' && character <= '9');
}

bool CharacterFunctions::isDigit (const juce_wchar character) noexcept
{
	return iswdigit ((wchar_t) character) != 0;
}

bool CharacterFunctions::isLetter (const char character) noexcept
{
	return (character >= 'a' && character <= 'z')
		|| (character >= 'A' && character <= 'Z');
}

bool CharacterFunctions::isLetter (const juce_wchar character) noexcept
{
	return iswalpha ((wchar_t) character) != 0;
}

bool CharacterFunctions::isLetterOrDigit (const char character) noexcept
{
	return (character >= 'a' && character <= 'z')
		|| (character >= 'A' && character <= 'Z')
		|| (character >= '0' && character <= '9');
}

bool CharacterFunctions::isLetterOrDigit (const juce_wchar character) noexcept
{
	return iswalnum ((wchar_t) character) != 0;
}

int CharacterFunctions::getHexDigitValue (const juce_wchar digit) noexcept
{
	unsigned int d = digit - '0';
	if (d < (unsigned int) 10)
		return (int) d;

	d += (unsigned int) ('0' - 'a');
	if (d < (unsigned int) 6)
		return (int) d + 10;

	d += (unsigned int) ('a' - 'A');
	if (d < (unsigned int) 6)
		return (int) d + 10;

	return -1;
}

double CharacterFunctions::mulexp10 (const double value, int exponent) noexcept
{
	if (exponent == 0)
		return value;

	if (value == 0)
		return 0;

	const bool negative = (exponent < 0);
	if (negative)
		exponent = -exponent;

	double result = 1.0, power = 10.0;
	for (int bit = 1; exponent != 0; bit <<= 1)
	{
		if ((exponent & bit) != 0)
		{
			exponent ^= bit;
			result *= power;
			if (exponent == 0)
				break;
		}
		power *= power;
	}

	return negative ? (value / result) : (value * result);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_CharacterFunctions.cpp ***/


/*** Start of inlined file: juce_LocalisedStrings.cpp ***/
BEGIN_JUCE_NAMESPACE

LocalisedStrings::LocalisedStrings (const String& fileContents)
{
	loadFromText (fileContents);
}

LocalisedStrings::LocalisedStrings (const File& fileToLoad)
{
	loadFromText (fileToLoad.loadFileAsString());
}

LocalisedStrings::~LocalisedStrings()
{
}

String LocalisedStrings::translate (const String& text) const
{
	return translations.getValue (text, text);
}

namespace
{
   #if JUCE_CHECK_MEMORY_LEAKS
	// By using this object to force a LocalisedStrings object to be created
	// before the currentMappings object, we can force the static order-of-destruction to
	// delete the currentMappings object first, which avoids a bogus leak warning.
	// (Oddly, just creating a LocalisedStrings on the stack doesn't work in gcc, it
	// has to be created with 'new' for this to work..)
	struct LeakAvoidanceTrick
	{
		LeakAvoidanceTrick()
		{
			const ScopedPointer<LocalisedStrings> dummy (new LocalisedStrings (String()));
		}
	};

	LeakAvoidanceTrick leakAvoidanceTrick;
   #endif

	SpinLock currentMappingsLock;
	ScopedPointer<LocalisedStrings> currentMappings;

	int findCloseQuote (const String& text, int startPos)
	{
		juce_wchar lastChar = 0;
		String::CharPointerType t (text.getCharPointer() + startPos);

		for (;;)
		{
			const juce_wchar c = t.getAndAdvance();

			if (c == 0 || (c == '"' && lastChar != '\\'))
				break;

			lastChar = c;
			++startPos;
		}

		return startPos;
	}

	String unescapeString (const String& s)
	{
		return s.replace ("\\\"", "\"")
				.replace ("\\\'", "\'")
				.replace ("\\t", "\t")
				.replace ("\\r", "\r")
				.replace ("\\n", "\n");
	}
}

void LocalisedStrings::loadFromText (const String& fileContents)
{
	StringArray lines;
	lines.addLines (fileContents);

	for (int i = 0; i < lines.size(); ++i)
	{
		String line (lines[i].trim());

		if (line.startsWithChar ('"'))
		{
			int closeQuote = findCloseQuote (line, 1);

			const String originalText (unescapeString (line.substring (1, closeQuote)));

			if (originalText.isNotEmpty())
			{
				const int openingQuote = findCloseQuote (line, closeQuote + 1);
				closeQuote = findCloseQuote (line, openingQuote + 1);

				const String newText (unescapeString (line.substring (openingQuote + 1, closeQuote)));

				if (newText.isNotEmpty())
					translations.set (originalText, newText);
			}
		}
		else if (line.startsWithIgnoreCase ("language:"))
		{
			languageName = line.substring (9).trim();
		}
		else if (line.startsWithIgnoreCase ("countries:"))
		{
			countryCodes.addTokens (line.substring (10).trim(), true);
			countryCodes.trim();
			countryCodes.removeEmptyStrings();
		}
	}
}

void LocalisedStrings::setIgnoresCase (const bool shouldIgnoreCase)
{
	translations.setIgnoresCase (shouldIgnoreCase);
}

void LocalisedStrings::setCurrentMappings (LocalisedStrings* newTranslations)
{
	const SpinLock::ScopedLockType sl (currentMappingsLock);
	currentMappings = newTranslations;
}

LocalisedStrings* LocalisedStrings::getCurrentMappings()
{
	return currentMappings;
}

String LocalisedStrings::translateWithCurrentMappings (const String& text)
{
	const SpinLock::ScopedLockType sl (currentMappingsLock);

	if (currentMappings != nullptr)
		return currentMappings->translate (text);

	return text;
}

String LocalisedStrings::translateWithCurrentMappings (const char* text)
{
	return translateWithCurrentMappings (String (text));
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_LocalisedStrings.cpp ***/


/*** Start of inlined file: juce_String.cpp ***/
#if JUCE_MSVC
  #pragma warning (push)
  #pragma warning (disable: 4514 4996)
#endif

#include <locale>

BEGIN_JUCE_NAMESPACE

NewLine newLine;

#if defined (JUCE_STRINGS_ARE_UNICODE) && ! JUCE_STRINGS_ARE_UNICODE
 #error "JUCE_STRINGS_ARE_UNICODE is deprecated! All strings are now unicode by default."
#endif

#if JUCE_NATIVE_WCHAR_IS_UTF8
 typedef CharPointer_UTF8	  CharPointer_wchar_t;
#elif JUCE_NATIVE_WCHAR_IS_UTF16
 typedef CharPointer_UTF16	 CharPointer_wchar_t;
#else
 typedef CharPointer_UTF32	 CharPointer_wchar_t;
#endif

static inline CharPointer_wchar_t castToCharPointer_wchar_t (const void* t) noexcept
{
	return CharPointer_wchar_t (static_cast <const CharPointer_wchar_t::CharType*> (t));
}

class StringHolder
{
public:
	StringHolder()
		: refCount (0x3fffffff), allocatedNumBytes (sizeof (*text))
	{
		text[0] = 0;
	}

	typedef String::CharPointerType CharPointerType;
	typedef String::CharPointerType::CharType CharType;

	static const CharPointerType createUninitialisedBytes (const size_t numBytes)
	{
		StringHolder* const s = reinterpret_cast <StringHolder*> (new char [sizeof (StringHolder) - sizeof (CharType) + numBytes]);
		s->refCount.value = 0;
		s->allocatedNumBytes = numBytes;
		return CharPointerType (s->text);
	}

	template <class CharPointer>
	static const CharPointerType createFromCharPointer (const CharPointer& text)
	{
		if (text.getAddress() == nullptr || text.isEmpty())
			return getEmpty();

		CharPointer t (text);
		size_t bytesNeeded = sizeof (CharType);

		while (! t.isEmpty())
			bytesNeeded += CharPointerType::getBytesRequiredFor (t.getAndAdvance());

		const CharPointerType dest (createUninitialisedBytes (bytesNeeded));
		CharPointerType (dest).writeAll (text);
		return dest;
	}

	template <class CharPointer>
	static const CharPointerType createFromCharPointer (const CharPointer& text, size_t maxChars)
	{
		if (text.getAddress() == nullptr || text.isEmpty() || maxChars == 0)
			return getEmpty();

		CharPointer end (text);
		size_t numChars = 0;
		size_t bytesNeeded = sizeof (CharType);

		while (numChars < maxChars && ! end.isEmpty())
		{
			bytesNeeded += CharPointerType::getBytesRequiredFor (end.getAndAdvance());
			++numChars;
		}

		const CharPointerType dest (createUninitialisedBytes (bytesNeeded));
		CharPointerType (dest).writeWithCharLimit (text, (int) numChars + 1);
		return dest;
	}

	template <class CharPointer>
	static const CharPointerType createFromCharPointer (const CharPointer& start, const CharPointer& end)
	{
		if (start.getAddress() == nullptr || start.isEmpty())
			return getEmpty();

		CharPointer e (start);
		int numChars = 0;
		size_t bytesNeeded = sizeof (CharType);

		while (e < end && ! e.isEmpty())
		{
			bytesNeeded += CharPointerType::getBytesRequiredFor (e.getAndAdvance());
			++numChars;
		}

		const CharPointerType dest (createUninitialisedBytes (bytesNeeded));
		CharPointerType (dest).writeWithCharLimit (start, numChars + 1);
		return dest;
	}

	static const CharPointerType createFromFixedLength (const char* const src, const size_t numChars)
	{
		const CharPointerType dest (createUninitialisedBytes (numChars * sizeof (CharType) + sizeof (CharType)));
		CharPointerType (dest).writeWithCharLimit (CharPointer_UTF8 (src), (int) (numChars + 1));
		return dest;
	}

	static inline const CharPointerType getEmpty() noexcept
	{
		return CharPointerType (empty.text);
	}

	static void retain (const CharPointerType& text) noexcept
	{
		++(bufferFromText (text)->refCount);
	}

	static inline void release (StringHolder* const b) noexcept
	{
		if (--(b->refCount) == -1 && b != &empty)
			delete[] reinterpret_cast <char*> (b);
	}

	static void release (const CharPointerType& text) noexcept
	{
		release (bufferFromText (text));
	}

	static CharPointerType makeUnique (const CharPointerType& text)
	{
		StringHolder* const b = bufferFromText (text);

		if (b->refCount.get() <= 0)
			return text;

		CharPointerType newText (createUninitialisedBytes (b->allocatedNumBytes));
		memcpy (newText.getAddress(), text.getAddress(), b->allocatedNumBytes);
		release (b);

		return newText;
	}

	static CharPointerType makeUniqueWithByteSize (const CharPointerType& text, size_t numBytes)
	{
		StringHolder* const b = bufferFromText (text);

		if (b->refCount.get() <= 0 && b->allocatedNumBytes >= numBytes)
			return text;

		CharPointerType newText (createUninitialisedBytes (jmax (b->allocatedNumBytes, numBytes)));
		memcpy (newText.getAddress(), text.getAddress(), b->allocatedNumBytes);
		release (b);

		return newText;
	}

	static size_t getAllocatedNumBytes (const CharPointerType& text) noexcept
	{
		return bufferFromText (text)->allocatedNumBytes;
	}

	Atomic<int> refCount;
	size_t allocatedNumBytes;
	CharType text[1];

	static StringHolder empty;

private:
	static inline StringHolder* bufferFromText (const CharPointerType& text) noexcept
	{
		// (Can't use offsetof() here because of warnings about this not being a POD)
		return reinterpret_cast <StringHolder*> (reinterpret_cast <char*> (text.getAddress())
					- (reinterpret_cast <size_t> (reinterpret_cast <StringHolder*> (1)->text) - 1));
	}

	void compileTimeChecks()
	{
		// Let me know if any of these assertions fail on your system!
	   #if JUCE_NATIVE_WCHAR_IS_UTF8
		static_assert (sizeof (wchar_t) == 1, "");
	   #elif JUCE_NATIVE_WCHAR_IS_UTF16
		static_assert (sizeof (wchar_t) == 2, "");
	   #elif JUCE_NATIVE_WCHAR_IS_UTF32
		static_assert (sizeof (wchar_t) == 4, "");
	   #else
		#error "native wchar_t size is unknown"
	   #endif
	}
};

StringHolder StringHolder::empty;
const String String::empty;

void String::preallocateBytes (const size_t numBytesNeeded)
{
	text = StringHolder::makeUniqueWithByteSize (text, numBytesNeeded + sizeof (CharPointerType::CharType));
}

String::String() noexcept
	: text (StringHolder::getEmpty())
{
}

String::~String() noexcept
{
	StringHolder::release (text);
}

String::String (const String& other) noexcept
	: text (other.text)
{
	StringHolder::retain (text);
}

void String::swapWith (String& other) noexcept
{
	std::swap (text, other.text);
}

String& String::operator= (const String& other) noexcept
{
	StringHolder::retain (other.text);
	StringHolder::release (text.atomicSwap (other.text));
	return *this;
}

inline String::PreallocationBytes::PreallocationBytes (const size_t numBytes_) : numBytes (numBytes_) {}

String::String (const PreallocationBytes& preallocationSize)
	: text (StringHolder::createUninitialisedBytes (preallocationSize.numBytes + sizeof (CharPointerType::CharType)))
{
}

String::String (const char* const t)
	: text (StringHolder::createFromCharPointer (CharPointer_ASCII (t)))
{
	/*  If you get an assertion here, then you're trying to create a string from 8-bit data
		that contains values greater than 127. These can NOT be correctly converted to unicode
		because there's no way for the String class to know what encoding was used to
		create them. The source data could be UTF-8, ASCII or one of many local code-pages.

		To get around this problem, you must be more explicit when you pass an ambiguous 8-bit
		string to the String class - so for example if your source data is actually UTF-8,
		you'd call String (CharPointer_UTF8 ("my utf8 string..")), and it would be able to
		correctly convert the multi-byte characters to unicode. It's *highly* recommended that
		you use UTF-8 with escape characters in your source code to represent extended characters,
		because there's no other way to represent these strings in a way that isn't dependent on
		the compiler, source code editor and platform.
	*/
	jassert (t == nullptr || CharPointer_ASCII::isValidString (t, std::numeric_limits<int>::max()));
}

String::String (const char* const t, const size_t maxChars)
	: text (StringHolder::createFromCharPointer (CharPointer_ASCII (t), maxChars))
{
	/*  If you get an assertion here, then you're trying to create a string from 8-bit data
		that contains values greater than 127. These can NOT be correctly converted to unicode
		because there's no way for the String class to know what encoding was used to
		create them. The source data could be UTF-8, ASCII or one of many local code-pages.

		To get around this problem, you must be more explicit when you pass an ambiguous 8-bit
		string to the String class - so for example if your source data is actually UTF-8,
		you'd call String (CharPointer_UTF8 ("my utf8 string..")), and it would be able to
		correctly convert the multi-byte characters to unicode. It's *highly* recommended that
		you use UTF-8 with escape characters in your source code to represent extended characters,
		because there's no other way to represent these strings in a way that isn't dependent on
		the compiler, source code editor and platform.
	*/
	jassert (t == nullptr || CharPointer_ASCII::isValidString (t, (int) maxChars));
}

String::String (const wchar_t* const t)	  : text (StringHolder::createFromCharPointer (castToCharPointer_wchar_t (t))) {}
String::String (const CharPointer_UTF8& t)   : text (StringHolder::createFromCharPointer (t)) {}
String::String (const CharPointer_UTF16& t)  : text (StringHolder::createFromCharPointer (t)) {}
String::String (const CharPointer_UTF32& t)  : text (StringHolder::createFromCharPointer (t)) {}
String::String (const CharPointer_ASCII& t)  : text (StringHolder::createFromCharPointer (t)) {}

String::String (const CharPointer_UTF8& t, const size_t maxChars)   : text (StringHolder::createFromCharPointer (t, maxChars)) {}
String::String (const CharPointer_UTF16& t, const size_t maxChars)  : text (StringHolder::createFromCharPointer (t, maxChars)) {}
String::String (const CharPointer_UTF32& t, const size_t maxChars)  : text (StringHolder::createFromCharPointer (t, maxChars)) {}
String::String (const wchar_t* const t, size_t maxChars)		: text (StringHolder::createFromCharPointer (castToCharPointer_wchar_t (t), maxChars)) {}

String::String (const CharPointer_UTF8& start, const CharPointer_UTF8& end)   : text (StringHolder::createFromCharPointer (start, end)) {}
String::String (const CharPointer_UTF16& start, const CharPointer_UTF16& end) : text (StringHolder::createFromCharPointer (start, end)) {}
String::String (const CharPointer_UTF32& start, const CharPointer_UTF32& end) : text (StringHolder::createFromCharPointer (start, end)) {}

String String::charToString (const juce_wchar character)
{
	String result (PreallocationBytes (CharPointerType::getBytesRequiredFor (character)));
	CharPointerType t (result.text);
	t.write (character);
	t.writeNull();
	return result;
}

namespace NumberToStringConverters
{
	// pass in a pointer to the END of a buffer..
	char* numberToString (char* t, const int64 n) noexcept
	{
		*--t = 0;
		int64 v = (n >= 0) ? n : -n;

		do
		{
			*--t = (char) ('0' + (int) (v % 10));
			v /= 10;

		} while (v > 0);

		if (n < 0)
			*--t = '-';

		return t;
	}

	char* numberToString (char* t, uint64 v) noexcept
	{
		*--t = 0;

		do
		{
			*--t = (char) ('0' + (int) (v % 10));
			v /= 10;

		} while (v > 0);

		return t;
	}

	char* numberToString (char* t, const int n) noexcept
	{
		if (n == (int) 0x80000000) // (would cause an overflow)
			return numberToString (t, (int64) n);

		*--t = 0;
		int v = abs (n);

		do
		{
			*--t = (char) ('0' + (v % 10));
			v /= 10;

		} while (v > 0);

		if (n < 0)
			*--t = '-';

		return t;
	}

	char* numberToString (char* t, unsigned int v) noexcept
	{
		*--t = 0;

		do
		{
			*--t = (char) ('0' + (v % 10));
			v /= 10;

		} while (v > 0);

		return t;
	}

	char getDecimalPoint()
	{
		static char dp = (char)
		   #if JUCE_VC7_OR_EARLIER
			std::_USE (std::locale(), std::numpunct <char>).decimal_point();
		   #else
			std::use_facet <std::numpunct <char> > (std::locale()).decimal_point();
		   #endif

		return dp;
	}

	char* doubleToString (char* buffer, int numChars, double n, int numDecPlaces, size_t& len) noexcept
	{
		if (numDecPlaces > 0 && n > -1.0e20 && n < 1.0e20)
		{
			char* const end = buffer + numChars;
			char* t = end;
			int64 v = (int64) (pow (10.0, numDecPlaces) * std::abs (n) + 0.5);
			*--t = (char) 0;

			while (numDecPlaces >= 0 || v > 0)
			{
				if (numDecPlaces == 0)
					*--t = (char) getDecimalPoint();

				*--t = (char) ('0' + (v % 10));

				v /= 10;
				--numDecPlaces;
			}

			if (n < 0)
				*--t = '-';

			len = end - t - 1;
			return t;
		}
		else
		{
			len = sprintf (buffer, "%.9g", n);
			return buffer;
		}
	}

	template <typename IntegerType>
	const String::CharPointerType createFromInteger (const IntegerType number)
	{
		char buffer [32];
		char* const end = buffer + numElementsInArray (buffer);
		char* const start = numberToString (end, number);

		return StringHolder::createFromFixedLength (start, end - start - 1);
	}

	const String::CharPointerType createFromDouble (const double number, const int numberOfDecimalPlaces)
	{
		char buffer [48];
		size_t len;
		char* const start = doubleToString (buffer, numElementsInArray (buffer), (double) number, numberOfDecimalPlaces, len);
		return StringHolder::createFromFixedLength (start, len);
	}
}

String::String (const int number)		: text (NumberToStringConverters::createFromInteger (number)) {}
String::String (const unsigned int number)   : text (NumberToStringConverters::createFromInteger (number)) {}
String::String (const short number)	  : text (NumberToStringConverters::createFromInteger ((int) number)) {}
String::String (const unsigned short number) : text (NumberToStringConverters::createFromInteger ((unsigned int) number)) {}
String::String (const int64 number)	  : text (NumberToStringConverters::createFromInteger (number)) {}
String::String (const uint64 number)	 : text (NumberToStringConverters::createFromInteger (number)) {}

String::String (const float number, const int numberOfDecimalPlaces)   : text (NumberToStringConverters::createFromDouble ((double) number, numberOfDecimalPlaces)) {}
String::String (const double number, const int numberOfDecimalPlaces)  : text (NumberToStringConverters::createFromDouble (number, numberOfDecimalPlaces)) {}

int String::length() const noexcept
{
	return (int) text.length();
}

size_t String::getByteOffsetOfEnd() const noexcept
{
	return ((char*) text.findTerminatingNull().getAddress()) - (char*) text.getAddress();
}

juce_wchar String::operator[] (int index) const noexcept
{
	jassert (index == 0 || (index > 0 && index <= (int) text.lengthUpTo (index + 1)));
	return text [index];
}

int String::hashCode() const noexcept
{
	CharPointerType t (text);
	int result = 0;

	while (! t.isEmpty())
		result = 31 * result + t.getAndAdvance();

	return result;
}

int64 String::hashCode64() const noexcept
{
	CharPointerType t (text);
	int64 result = 0;

	while (! t.isEmpty())
		result = 101 * result + t.getAndAdvance();

	return result;
}

JUCE_API bool JUCE_CALLTYPE operator== (const String& s1, const String& s2) noexcept		{ return s1.compare (s2) == 0; }
JUCE_API bool JUCE_CALLTYPE operator== (const String& s1, const char* const s2) noexcept	{ return s1.compare (s2) == 0; }
JUCE_API bool JUCE_CALLTYPE operator== (const String& s1, const wchar_t* const s2) noexcept	 { return s1.compare (s2) == 0; }
JUCE_API bool JUCE_CALLTYPE operator== (const String& s1, const CharPointer_UTF8& s2) noexcept  { return s1.getCharPointer().compare (s2) == 0; }
JUCE_API bool JUCE_CALLTYPE operator== (const String& s1, const CharPointer_UTF16& s2) noexcept { return s1.getCharPointer().compare (s2) == 0; }
JUCE_API bool JUCE_CALLTYPE operator== (const String& s1, const CharPointer_UTF32& s2) noexcept { return s1.getCharPointer().compare (s2) == 0; }
JUCE_API bool JUCE_CALLTYPE operator!= (const String& s1, const String& s2) noexcept		{ return s1.compare (s2) != 0; }
JUCE_API bool JUCE_CALLTYPE operator!= (const String& s1, const char* const s2) noexcept	{ return s1.compare (s2) != 0; }
JUCE_API bool JUCE_CALLTYPE operator!= (const String& s1, const wchar_t* const s2) noexcept	 { return s1.compare (s2) != 0; }
JUCE_API bool JUCE_CALLTYPE operator!= (const String& s1, const CharPointer_UTF8& s2) noexcept  { return s1.getCharPointer().compare (s2) != 0; }
JUCE_API bool JUCE_CALLTYPE operator!= (const String& s1, const CharPointer_UTF16& s2) noexcept { return s1.getCharPointer().compare (s2) != 0; }
JUCE_API bool JUCE_CALLTYPE operator!= (const String& s1, const CharPointer_UTF32& s2) noexcept { return s1.getCharPointer().compare (s2) != 0; }
JUCE_API bool JUCE_CALLTYPE operator>  (const String& s1, const String& s2) noexcept		{ return s1.compare (s2) > 0; }
JUCE_API bool JUCE_CALLTYPE operator<  (const String& s1, const String& s2) noexcept		{ return s1.compare (s2) < 0; }
JUCE_API bool JUCE_CALLTYPE operator>= (const String& s1, const String& s2) noexcept		{ return s1.compare (s2) >= 0; }
JUCE_API bool JUCE_CALLTYPE operator<= (const String& s1, const String& s2) noexcept		{ return s1.compare (s2) <= 0; }

bool String::equalsIgnoreCase (const wchar_t* const t) const noexcept
{
	return t != nullptr ? text.compareIgnoreCase (castToCharPointer_wchar_t (t)) == 0
						: isEmpty();
}

bool String::equalsIgnoreCase (const char* const t) const noexcept
{
	return t != nullptr ? text.compareIgnoreCase (CharPointer_UTF8 (t)) == 0
						: isEmpty();
}

bool String::equalsIgnoreCase (const String& other) const noexcept
{
	return text == other.text
			|| text.compareIgnoreCase (other.text) == 0;
}

int String::compare (const String& other) const noexcept	   { return (text == other.text) ? 0 : text.compare (other.text); }
int String::compare (const char* const other) const noexcept	   { return text.compare (CharPointer_UTF8 (other)); }
int String::compare (const wchar_t* const other) const noexcept	{ return text.compare (castToCharPointer_wchar_t (other)); }
int String::compareIgnoreCase (const String& other) const noexcept { return (text == other.text) ? 0 : text.compareIgnoreCase (other.text); }

int String::compareLexicographically (const String& other) const noexcept
{
	CharPointerType s1 (text);

	while (! (s1.isEmpty() || s1.isLetterOrDigit()))
		++s1;

	CharPointerType s2 (other.text);

	while (! (s2.isEmpty() || s2.isLetterOrDigit()))
		++s2;

	return s1.compareIgnoreCase (s2);
}

void String::append (const String& textToAppend, size_t maxCharsToTake)
{
	appendCharPointer (textToAppend.text, maxCharsToTake);
}

String& String::operator+= (const wchar_t* const t)
{
	appendCharPointer (castToCharPointer_wchar_t (t));
	return *this;
}

String& String::operator+= (const char* const t)
{
	/*  If you get an assertion here, then you're trying to create a string from 8-bit data
		that contains values greater than 127. These can NOT be correctly converted to unicode
		because there's no way for the String class to know what encoding was used to
		create them. The source data could be UTF-8, ASCII or one of many local code-pages.

		To get around this problem, you must be more explicit when you pass an ambiguous 8-bit
		string to the String class - so for example if your source data is actually UTF-8,
		you'd call String (CharPointer_UTF8 ("my utf8 string..")), and it would be able to
		correctly convert the multi-byte characters to unicode. It's *highly* recommended that
		you use UTF-8 with escape characters in your source code to represent extended characters,
		because there's no other way to represent these strings in a way that isn't dependent on
		the compiler, source code editor and platform.
	*/
	jassert (t == nullptr || CharPointer_ASCII::isValidString (t, std::numeric_limits<int>::max()));

	appendCharPointer (CharPointer_ASCII (t));
	return *this;
}

String& String::operator+= (const String& other)
{
	if (isEmpty())
		return operator= (other);

	appendCharPointer (other.text);
	return *this;
}

String& String::operator+= (const char ch)
{
	const char asString[] = { ch, 0 };
	return operator+= (asString);
}

String& String::operator+= (const wchar_t ch)
{
	const wchar_t asString[] = { ch, 0 };
	return operator+= (asString);
}

#if ! JUCE_NATIVE_WCHAR_IS_UTF32
String& String::operator+= (const juce_wchar ch)
{
	const juce_wchar asString[] = { ch, 0 };
	appendCharPointer (CharPointer_UTF32 (asString));
	return *this;
}
#endif

String& String::operator+= (const int number)
{
	char buffer [16];
	char* const end = buffer + numElementsInArray (buffer);
	char* const start = NumberToStringConverters::numberToString (end, number);

	const int numExtraChars = (int) (end - start);

	if (numExtraChars > 0)
	{
		const size_t byteOffsetOfNull = getByteOffsetOfEnd();
		const size_t newBytesNeeded = sizeof (CharPointerType::CharType) + byteOffsetOfNull
										+ sizeof (CharPointerType::CharType) * numExtraChars;

		text = StringHolder::makeUniqueWithByteSize (text, newBytesNeeded);

		CharPointerType newEnd (addBytesToPointer (text.getAddress(), (int) byteOffsetOfNull));
		newEnd.writeWithCharLimit (CharPointer_ASCII (start), numExtraChars);
	}

	return *this;
}

JUCE_API String JUCE_CALLTYPE operator+ (const char* const string1, const String& string2)
{
	String s (string1);
	return s += string2;
}

JUCE_API String JUCE_CALLTYPE operator+ (const wchar_t* const string1, const String& string2)
{
	String s (string1);
	return s += string2;
}

JUCE_API String JUCE_CALLTYPE operator+ (const char s1, const String& s2)	   { return String::charToString (s1) + s2; }
JUCE_API String JUCE_CALLTYPE operator+ (const wchar_t s1, const String& s2)	{ return String::charToString (s1) + s2; }
#if ! JUCE_NATIVE_WCHAR_IS_UTF32
JUCE_API String JUCE_CALLTYPE operator+ (const juce_wchar s1, const String& s2) { return String::charToString (s1) + s2; }
#endif

JUCE_API String JUCE_CALLTYPE operator+ (String s1, const String& s2)	   { return s1 += s2; }
JUCE_API String JUCE_CALLTYPE operator+ (String s1, const char* const s2)   { return s1 += s2; }
JUCE_API String JUCE_CALLTYPE operator+ (String s1, const wchar_t* s2)	  { return s1 += s2; }

JUCE_API String JUCE_CALLTYPE operator+ (String s1, const char s2)	  { return s1 += s2; }
JUCE_API String JUCE_CALLTYPE operator+ (String s1, const wchar_t s2)	   { return s1 += s2; }
#if ! JUCE_NATIVE_WCHAR_IS_UTF32
JUCE_API String JUCE_CALLTYPE operator+ (String s1, const juce_wchar s2)	{ return s1 += s2; }
#endif

JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const char s2)		 { return s1 += s2; }
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const wchar_t s2)	  { return s1 += s2; }
#if ! JUCE_NATIVE_WCHAR_IS_UTF32
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const juce_wchar s2)	   { return s1 += s2; }
#endif

JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const char* const s2)	  { return s1 += s2; }
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const wchar_t* const s2)   { return s1 += s2; }
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const String& s2)	  { return s1 += s2; }

JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const short number)	{ return s1 += (int) number; }
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const int number)	  { return s1 += number; }
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const long number)	 { return s1 += (int) number; }
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const float number)	{ return s1 += String (number); }
JUCE_API String& JUCE_CALLTYPE operator<< (String& s1, const double number)	   { return s1 += String (number); }

JUCE_API OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const String& text)
{
	const int numBytes = text.getNumBytesAsUTF8();

   #if (JUCE_STRING_UTF_TYPE == 8)
	stream.write (text.getCharPointer().getAddress(), numBytes);
   #else
	// (This avoids using toUTF8() to prevent the memory bloat that it would leave behind
	// if lots of large, persistent strings were to be written to streams).
	HeapBlock<char> temp (numBytes + 1);
	CharPointer_UTF8 (temp).writeAll (text.getCharPointer());
	stream.write (temp, numBytes);
   #endif

	return stream;
}

JUCE_API String& JUCE_CALLTYPE operator<< (String& string1, const NewLine&)
{
	return string1 += NewLine::getDefault();
}

int String::indexOfChar (const juce_wchar character) const noexcept
{
	return text.indexOf (character);
}

int String::indexOfChar (const int startIndex, const juce_wchar character) const noexcept
{
	CharPointerType t (text);

	for (int i = 0; ! t.isEmpty(); ++i)
	{
		if (i >= startIndex)
		{
			if (t.getAndAdvance() == character)
				return i;
		}
		else
		{
			++t;
		}
	}

	return -1;
}

int String::lastIndexOfChar (const juce_wchar character) const noexcept
{
	CharPointerType t (text);
	int last = -1;

	for (int i = 0; ! t.isEmpty(); ++i)
		if (t.getAndAdvance() == character)
			last = i;

	return last;
}

int String::indexOfAnyOf (const String& charactersToLookFor, const int startIndex, const bool ignoreCase) const noexcept
{
	CharPointerType t (text);

	for (int i = 0; ! t.isEmpty(); ++i)
	{
		if (i >= startIndex)
		{
			if (charactersToLookFor.text.indexOf (t.getAndAdvance(), ignoreCase) >= 0)
				return i;
		}
		else
		{
			++t;
		}
	}

	return -1;
}

int String::indexOf (const String& other) const noexcept
{
	return other.isEmpty() ? 0 : text.indexOf (other.text);
}

int String::indexOfIgnoreCase (const String& other) const noexcept
{
	return other.isEmpty() ? 0 : CharacterFunctions::indexOfIgnoreCase (text, other.text);
}

int String::indexOf (const int startIndex, const String& other) const noexcept
{
	if (other.isEmpty())
		return -1;

	CharPointerType t (text);

	for (int i = startIndex; --i >= 0;)
	{
		if (t.isEmpty())
			return -1;

		++t;
	}

	int found = t.indexOf (other.text);
	if (found >= 0)
		found += startIndex;
	return found;
}

int String::indexOfIgnoreCase (const int startIndex, const String& other) const noexcept
{
	if (other.isEmpty())
		return -1;

	CharPointerType t (text);

	for (int i = startIndex; --i >= 0;)
	{
		if (t.isEmpty())
			return -1;

		++t;
	}

	int found = CharacterFunctions::indexOfIgnoreCase (t, other.text);
	if (found >= 0)
		found += startIndex;
	return found;
}

int String::lastIndexOf (const String& other) const noexcept
{
	if (other.isNotEmpty())
	{
		const int len = other.length();
		int i = length() - len;

		if (i >= 0)
		{
			CharPointerType n (text + i);

			while (i >= 0)
			{
				if (n.compareUpTo (other.text, len) == 0)
					return i;

				--n;
				--i;
			}
		}
	}

	return -1;
}

int String::lastIndexOfIgnoreCase (const String& other) const noexcept
{
	if (other.isNotEmpty())
	{
		const int len = other.length();
		int i = length() - len;

		if (i >= 0)
		{
			CharPointerType n (text + i);

			while (i >= 0)
			{
				if (n.compareIgnoreCaseUpTo (other.text, len) == 0)
					return i;

				--n;
				--i;
			}
		}
	}

	return -1;
}

int String::lastIndexOfAnyOf (const String& charactersToLookFor, const bool ignoreCase) const noexcept
{
	CharPointerType t (text);
	int last = -1;

	for (int i = 0; ! t.isEmpty(); ++i)
		if (charactersToLookFor.text.indexOf (t.getAndAdvance(), ignoreCase) >= 0)
			last = i;

	return last;
}

bool String::contains (const String& other) const noexcept
{
	return indexOf (other) >= 0;
}

bool String::containsChar (const juce_wchar character) const noexcept
{
	return text.indexOf (character) >= 0;
}

bool String::containsIgnoreCase (const String& t) const noexcept
{
	return indexOfIgnoreCase (t) >= 0;
}

int String::indexOfWholeWord (const String& word) const noexcept
{
	if (word.isNotEmpty())
	{
		CharPointerType t (text);
		const int wordLen = word.length();
		const int end = (int) t.length() - wordLen;

		for (int i = 0; i <= end; ++i)
		{
			if (t.compareUpTo (word.text, wordLen) == 0
				  && (i == 0 || ! (t - 1).isLetterOrDigit())
				  && ! (t + wordLen).isLetterOrDigit())
				return i;

			++t;
		}
	}

	return -1;
}

int String::indexOfWholeWordIgnoreCase (const String& word) const noexcept
{
	if (word.isNotEmpty())
	{
		CharPointerType t (text);
		const int wordLen = word.length();
		const int end = (int) t.length() - wordLen;

		for (int i = 0; i <= end; ++i)
		{
			if (t.compareIgnoreCaseUpTo (word.text, wordLen) == 0
				  && (i == 0 || ! (t - 1).isLetterOrDigit())
				  && ! (t + wordLen).isLetterOrDigit())
				return i;

			++t;
		}
	}

	return -1;
}

bool String::containsWholeWord (const String& wordToLookFor) const noexcept
{
	return indexOfWholeWord (wordToLookFor) >= 0;
}

bool String::containsWholeWordIgnoreCase (const String& wordToLookFor) const noexcept
{
	return indexOfWholeWordIgnoreCase (wordToLookFor) >= 0;
}

namespace WildCardHelpers
{
	int indexOfMatch (const String::CharPointerType& wildcard,
					  String::CharPointerType test,
					  const bool ignoreCase) noexcept
	{
		int start = 0;

		while (! test.isEmpty())
		{
			String::CharPointerType t (test);
			String::CharPointerType w (wildcard);

			for (;;)
			{
				const juce_wchar wc = *w;
				const juce_wchar tc = *t;

				if (wc == tc
					 || (ignoreCase && w.toLowerCase() == t.toLowerCase())
					 || (wc == '?' && tc != 0))
				{
					if (wc == 0)
						return start;

					++t;
					++w;
				}
				else
				{
					if (wc == '*' && (w[1] == 0 || indexOfMatch (w + 1, t, ignoreCase) >= 0))
						return start;

					break;
				}
			}

			++start;
			++test;
		}

		return -1;
	}
}

bool String::matchesWildcard (const String& wildcard, const bool ignoreCase) const noexcept
{
	CharPointerType w (wildcard.text);
	CharPointerType t (text);

	for (;;)
	{
		const juce_wchar wc = *w;
		const juce_wchar tc = *t;

		if (wc == tc
			 || (ignoreCase && w.toLowerCase() == t.toLowerCase())
			 || (wc == '?' && tc != 0))
		{
			if (wc == 0)
				return true;

			++w;
			++t;
		}
		else
		{
			return wc == '*' && (w[1] == 0 || WildCardHelpers::indexOfMatch (w + 1, t, ignoreCase) >= 0);
		}
	}
}

String String::repeatedString (const String& stringToRepeat, int numberOfTimesToRepeat)
{
	if (numberOfTimesToRepeat <= 0)
		return String::empty;

	String result (PreallocationBytes (stringToRepeat.getByteOffsetOfEnd() * numberOfTimesToRepeat));
	CharPointerType n (result.text);

	while (--numberOfTimesToRepeat >= 0)
		n.writeAll (stringToRepeat.text);

	return result;
}

String String::paddedLeft (const juce_wchar padCharacter, int minimumLength) const
{
	jassert (padCharacter != 0);

	int extraChars = minimumLength;
	CharPointerType end (text);

	while (! end.isEmpty())
	{
		--extraChars;
		++end;
	}

	if (extraChars <= 0 || padCharacter == 0)
		return *this;

	const size_t currentByteSize = ((char*) end.getAddress()) - (char*) text.getAddress();
	String result (PreallocationBytes (currentByteSize + extraChars * CharPointerType::getBytesRequiredFor (padCharacter)));
	CharPointerType n (result.text);

	while (--extraChars >= 0)
		n.write (padCharacter);

	n.writeAll (text);
	return result;
}

String String::paddedRight (const juce_wchar padCharacter, int minimumLength) const
{
	jassert (padCharacter != 0);

	int extraChars = minimumLength;
	CharPointerType end (text);

	while (! end.isEmpty())
	{
		--extraChars;
		++end;
	}

	if (extraChars <= 0 || padCharacter == 0)
		return *this;

	const size_t currentByteSize = ((char*) end.getAddress()) - (char*) text.getAddress();
	String result (PreallocationBytes (currentByteSize + extraChars * CharPointerType::getBytesRequiredFor (padCharacter)));
	CharPointerType n (result.text);

	n.writeAll (text);

	while (--extraChars >= 0)
		n.write (padCharacter);

	n.writeNull();
	return result;
}

String String::replaceSection (int index, int numCharsToReplace, const String& stringToInsert) const
{
	if (index < 0)
	{
		// a negative index to replace from?
		jassertfalse;
		index = 0;
	}

	if (numCharsToReplace < 0)
	{
		// replacing a negative number of characters?
		numCharsToReplace = 0;
		jassertfalse;
	}

	int i = 0;
	CharPointerType insertPoint (text);

	while (i < index)
	{
		if (insertPoint.isEmpty())
		{
			// replacing beyond the end of the string?
			jassertfalse;
			return *this + stringToInsert;
		}

		++insertPoint;
		++i;
	}

	CharPointerType startOfRemainder (insertPoint);

	i = 0;
	while (i < numCharsToReplace && ! startOfRemainder.isEmpty())
	{
		++startOfRemainder;
		++i;
	}

	if (insertPoint == text && startOfRemainder.isEmpty())
		return stringToInsert;

	const size_t initialBytes = ((char*) insertPoint.getAddress()) - (char*) text.getAddress();
	const size_t newStringBytes = stringToInsert.getByteOffsetOfEnd();
	const size_t remainderBytes = ((char*) startOfRemainder.findTerminatingNull().getAddress()) - (char*) startOfRemainder.getAddress();

	const size_t newTotalBytes = initialBytes + newStringBytes + remainderBytes;
	if (newTotalBytes <= 0)
		return String::empty;

	String result (PreallocationBytes ((size_t) newTotalBytes));

	char* dest = (char*) result.text.getAddress();
	memcpy (dest, text.getAddress(), initialBytes);
	dest += initialBytes;
	memcpy (dest, stringToInsert.text.getAddress(), newStringBytes);
	dest += newStringBytes;
	memcpy (dest, startOfRemainder.getAddress(), remainderBytes);
	dest += remainderBytes;
	CharPointerType ((CharPointerType::CharType*) dest).writeNull();

	return result;
}

String String::replace (const String& stringToReplace, const String& stringToInsert, const bool ignoreCase) const
{
	const int stringToReplaceLen = stringToReplace.length();
	const int stringToInsertLen = stringToInsert.length();

	int i = 0;
	String result (*this);

	while ((i = (ignoreCase ? result.indexOfIgnoreCase (i, stringToReplace)
							: result.indexOf (i, stringToReplace))) >= 0)
	{
		result = result.replaceSection (i, stringToReplaceLen, stringToInsert);
		i += stringToInsertLen;
	}

	return result;
}

class StringCreationHelper
{
public:
	StringCreationHelper (const size_t initialBytes)
		: source (nullptr), dest (nullptr), allocatedBytes (initialBytes), bytesWritten (0)
	{
		result.preallocateBytes (allocatedBytes);
		dest = result.getCharPointer();
	}

	StringCreationHelper (const String::CharPointerType& source_)
		: source (source_), dest (nullptr), allocatedBytes (StringHolder::getAllocatedNumBytes (source)), bytesWritten (0)
	{
		result.preallocateBytes (allocatedBytes);
		dest = result.getCharPointer();
	}

	void write (juce_wchar c)
	{
		bytesWritten += String::CharPointerType::getBytesRequiredFor (c);

		if (bytesWritten > allocatedBytes)
		{
			allocatedBytes += jmax ((size_t) 8, allocatedBytes / 16);
			const size_t destOffset = ((char*) dest.getAddress()) - (char*) result.getCharPointer().getAddress();
			result.preallocateBytes (allocatedBytes);
			dest = addBytesToPointer (result.getCharPointer().getAddress(), (int) destOffset);
		}

		dest.write (c);
	}

	String result;
	String::CharPointerType source;

private:
	String::CharPointerType dest;
	size_t allocatedBytes, bytesWritten;
};

String String::replaceCharacter (const juce_wchar charToReplace, const juce_wchar charToInsert) const
{
	if (! containsChar (charToReplace))
		return *this;

	StringCreationHelper builder (text);

	for (;;)
	{
		juce_wchar c = builder.source.getAndAdvance();

		if (c == charToReplace)
			c = charToInsert;

		builder.write (c);

		if (c == 0)
			break;
	}

	return builder.result;
}

String String::replaceCharacters (const String& charactersToReplace, const String& charactersToInsertInstead) const
{
	StringCreationHelper builder (text);

	for (;;)
	{
		juce_wchar c = builder.source.getAndAdvance();

		const int index = charactersToReplace.indexOfChar (c);
		if (index >= 0)
			c = charactersToInsertInstead [index];

		builder.write (c);

		if (c == 0)
			break;
	}

	return builder.result;
}

bool String::startsWith (const String& other) const noexcept
{
	return text.compareUpTo (other.text, other.length()) == 0;
}

bool String::startsWithIgnoreCase (const String& other) const noexcept
{
	return text.compareIgnoreCaseUpTo (other.text, other.length()) == 0;
}

bool String::startsWithChar (const juce_wchar character) const noexcept
{
	jassert (character != 0); // strings can't contain a null character!

	return *text == character;
}

bool String::endsWithChar (const juce_wchar character) const noexcept
{
	jassert (character != 0); // strings can't contain a null character!

	if (text.isEmpty())
		return false;

	CharPointerType t (text.findTerminatingNull());
	return *--t == character;
}

bool String::endsWith (const String& other) const noexcept
{
	CharPointerType end (text.findTerminatingNull());
	CharPointerType otherEnd (other.text.findTerminatingNull());

	while (end > text && otherEnd > other.text)
	{
		--end;
		--otherEnd;

		if (*end != *otherEnd)
			return false;
	}

	return otherEnd == other.text;
}

bool String::endsWithIgnoreCase (const String& other) const noexcept
{
	CharPointerType end (text.findTerminatingNull());
	CharPointerType otherEnd (other.text.findTerminatingNull());

	while (end > text && otherEnd > other.text)
	{
		--end;
		--otherEnd;

		if (end.toLowerCase() != otherEnd.toLowerCase())
			return false;
	}

	return otherEnd == other.text;
}

String String::toUpperCase() const
{
	StringCreationHelper builder (text);

	for (;;)
	{
		const juce_wchar c = builder.source.toUpperCase();
		++(builder.source);
		builder.write (c);

		if (c == 0)
			break;
	}

	return builder.result;
}

String String::toLowerCase() const
{
	StringCreationHelper builder (text);

	for (;;)
	{
		const juce_wchar c = builder.source.toLowerCase();
		++(builder.source);
		builder.write (c);

		if (c == 0)
			break;
	}

	return builder.result;
}

juce_wchar String::getLastCharacter() const noexcept
{
	return isEmpty() ? juce_wchar() : text [length() - 1];
}

String String::substring (int start, const int end) const
{
	if (start < 0)
		start = 0;

	if (end <= start)
		return empty;

	int i = 0;
	CharPointerType t1 (text);

	while (i < start)
	{
		if (t1.isEmpty())
			return empty;

		++i;
		++t1;
	}

	CharPointerType t2 (t1);
	while (i < end)
	{
		if (t2.isEmpty())
		{
			if (start == 0)
				return *this;

			break;
		}

		++i;
		++t2;
	}

	return String (t1, t2);
}

String String::substring (int start) const
{
	if (start <= 0)
		return *this;

	CharPointerType t (text);

	while (--start >= 0)
	{
		if (t.isEmpty())
			return empty;

		++t;
	}

	return String (t);
}

String String::dropLastCharacters (const int numberToDrop) const
{
	return String (text, jmax (0, length() - numberToDrop));
}

String String::getLastCharacters (const int numCharacters) const
{
	return String (text + jmax (0, length() - jmax (0, numCharacters)));
}

String String::fromFirstOccurrenceOf (const String& sub,
									  const bool includeSubString,
									  const bool ignoreCase) const
{
	const int i = ignoreCase ? indexOfIgnoreCase (sub)
							 : indexOf (sub);
	if (i < 0)
		return empty;

	return substring (includeSubString ? i : i + sub.length());
}

String String::fromLastOccurrenceOf (const String& sub,
									 const bool includeSubString,
									 const bool ignoreCase) const
{
	const int i = ignoreCase ? lastIndexOfIgnoreCase (sub)
							 : lastIndexOf (sub);
	if (i < 0)
		return *this;

	return substring (includeSubString ? i : i + sub.length());
}

String String::upToFirstOccurrenceOf (const String& sub,
									  const bool includeSubString,
									  const bool ignoreCase) const
{
	const int i = ignoreCase ? indexOfIgnoreCase (sub)
							 : indexOf (sub);
	if (i < 0)
		return *this;

	return substring (0, includeSubString ? i + sub.length() : i);
}

String String::upToLastOccurrenceOf (const String& sub,
									 const bool includeSubString,
									 const bool ignoreCase) const
{
	const int i = ignoreCase ? lastIndexOfIgnoreCase (sub)
							 : lastIndexOf (sub);
	if (i < 0)
		return *this;

	return substring (0, includeSubString ? i + sub.length() : i);
}

bool String::isQuotedString() const
{
	const String trimmed (trimStart());

	return trimmed[0] == '"'
		|| trimmed[0] == '\'';
}

String String::unquoted() const
{
	const int len = length();

	if (len == 0)
		return empty;

	const juce_wchar lastChar = text [len - 1];
	const int dropAtStart = (*text == '"' || *text == '\'') ? 1 : 0;
	const int dropAtEnd = (lastChar == '"' || lastChar == '\'') ? 1 : 0;

	return substring (dropAtStart, len - dropAtEnd);
}

String String::quoted (const juce_wchar quoteCharacter) const
{
	if (isEmpty())
		return charToString (quoteCharacter) + quoteCharacter;

	String t (*this);

	if (! t.startsWithChar (quoteCharacter))
		t = charToString (quoteCharacter) + t;

	if (! t.endsWithChar (quoteCharacter))
		t += quoteCharacter;

	return t;
}

static String::CharPointerType findTrimmedEnd (const String::CharPointerType& start, String::CharPointerType end)
{
	while (end > start)
	{
		if (! (--end).isWhitespace())
		{
			++end;
			break;
		}
	}

	return end;
}

String String::trim() const
{
	if (isNotEmpty())
	{
		CharPointerType start (text.findEndOfWhitespace());

		const CharPointerType end (start.findTerminatingNull());
		CharPointerType trimmedEnd (findTrimmedEnd (start, end));

		if (trimmedEnd <= start)
			return empty;
		else if (text < start || trimmedEnd < end)
			return String (start, trimmedEnd);
	}

	return *this;
}

String String::trimStart() const
{
	if (isNotEmpty())
	{
		const CharPointerType t (text.findEndOfWhitespace());

		if (t != text)
			return String (t);
	}

	return *this;
}

String String::trimEnd() const
{
	if (isNotEmpty())
	{
		const CharPointerType end (text.findTerminatingNull());
		CharPointerType trimmedEnd (findTrimmedEnd (text, end));

		if (trimmedEnd < end)
			return String (text, trimmedEnd);
	}

	return *this;
}

String String::trimCharactersAtStart (const String& charactersToTrim) const
{
	CharPointerType t (text);

	while (charactersToTrim.containsChar (*t))
		++t;

	return t == text ? *this : String (t);
}

String String::trimCharactersAtEnd (const String& charactersToTrim) const
{
	if (isNotEmpty())
	{
		const CharPointerType end (text.findTerminatingNull());
		CharPointerType trimmedEnd (end);

		while (trimmedEnd > text)
		{
			if (! charactersToTrim.containsChar (*--trimmedEnd))
			{
				++trimmedEnd;
				break;
			}
		}

		if (trimmedEnd < end)
			return String (text, trimmedEnd);
	}

	return *this;
}

String String::retainCharacters (const String& charactersToRetain) const
{
	if (isEmpty())
		return empty;

	StringCreationHelper builder (text);

	for (;;)
	{
		juce_wchar c = builder.source.getAndAdvance();

		if (charactersToRetain.containsChar (c))
			builder.write (c);

		if (c == 0)
			break;
	}

	builder.write (0);
	return builder.result;
}

String String::removeCharacters (const String& charactersToRemove) const
{
	if (isEmpty())
		return empty;

	StringCreationHelper builder (text);

	for (;;)
	{
		juce_wchar c = builder.source.getAndAdvance();

		if (! charactersToRemove.containsChar (c))
			builder.write (c);

		if (c == 0)
			break;
	}

	return builder.result;
}

String String::initialSectionContainingOnly (const String& permittedCharacters) const
{
	CharPointerType t (text);

	while (! t.isEmpty())
	{
		if (! permittedCharacters.containsChar (*t))
			return String (text, t);

		++t;
	}

	return *this;
}

String String::initialSectionNotContaining (const String& charactersToStopAt) const
{
	CharPointerType t (text);

	while (! t.isEmpty())
	{
		if (charactersToStopAt.containsChar (*t))
			return String (text, t);

		++t;
	}

	return *this;
}

bool String::containsOnly (const String& chars) const noexcept
{
	CharPointerType t (text);

	while (! t.isEmpty())
		if (! chars.containsChar (t.getAndAdvance()))
			return false;

	return true;
}

bool String::containsAnyOf (const String& chars) const noexcept
{
	CharPointerType t (text);

	while (! t.isEmpty())
		if (chars.containsChar (t.getAndAdvance()))
			return true;

	return false;
}

bool String::containsNonWhitespaceChars() const noexcept
{
	CharPointerType t (text);

	while (! t.isEmpty())
	{
		if (! t.isWhitespace())
			return true;

		++t;
	}

	return false;
}

// Note! The format parameter here MUST NOT be a reference, otherwise MS's va_start macro fails to work (but still compiles).
String String::formatted (const String pf, ... )
{
	size_t bufferSize = 256;

	for (;;)
	{
		va_list args;
		va_start (args, pf);

	   #if JUCE_WINDOWS
		HeapBlock <wchar_t> temp (bufferSize);
		const int num = (int) _vsnwprintf (temp.getData(), bufferSize - 1, pf.toWideCharPointer(), args);
	   #elif JUCE_ANDROID
		HeapBlock <char> temp (bufferSize);
		const int num = (int) vsnprintf (temp.getData(), bufferSize - 1, pf.toUTF8(), args);
	   #else
		HeapBlock <wchar_t> temp (bufferSize);
		const int num = (int) vswprintf (temp.getData(), bufferSize - 1, pf.toWideCharPointer(), args);
	   #endif

		va_end (args);

		if (num > 0)
			return String (temp);

		bufferSize += 256;

		if (num == 0 || bufferSize > 65536) // the upper limit is a sanity check to avoid situations where vprintf repeatedly
			break;			  // returns -1 because of an error rather than because it needs more space.
	}

	return empty;
}

int String::getIntValue() const noexcept
{
	return text.getIntValue32();
}

int String::getTrailingIntValue() const noexcept
{
	int n = 0;
	int mult = 1;
	CharPointerType t (text.findTerminatingNull());

	while (--t >= text)
	{
		if (! t.isDigit())
		{
			if (*t == '-')
				n = -n;

			break;
		}

		n += mult * (*t - '0');
		mult *= 10;
	}

	return n;
}

int64 String::getLargeIntValue() const noexcept
{
	return text.getIntValue64();
}

float String::getFloatValue() const noexcept
{
	return (float) getDoubleValue();
}

double String::getDoubleValue() const noexcept
{
	return text.getDoubleValue();
}

static const char hexDigits[] = "0123456789abcdef";

template <typename Type>
struct HexConverter
{
	static String hexToString (Type v)
	{
		char buffer[32];
		char* const end = buffer + 32;
		char* t = end;
		*--t = 0;

		do
		{
			*--t = hexDigits [(int) (v & 15)];
			v >>= 4;

		} while (v != 0);

		return String (t, (int) (end - t) - 1);
	}

	static Type stringToHex (String::CharPointerType t) noexcept
	{
		Type result = 0;

		while (! t.isEmpty())
		{
			const int hexValue = CharacterFunctions::getHexDigitValue (t.getAndAdvance());

			if (hexValue >= 0)
				result = (result << 4) | hexValue;
		}

		return result;
	}
};

String String::toHexString (const int number)
{
	return HexConverter <unsigned int>::hexToString ((unsigned int) number);
}

String String::toHexString (const int64 number)
{
	return HexConverter <uint64>::hexToString ((uint64) number);
}

String String::toHexString (const short number)
{
	return toHexString ((int) (unsigned short) number);
}

String String::toHexString (const void* const d, const int size, const int groupSize)
{
	if (size <= 0)
		return empty;

	int numChars = (size * 2) + 2;
	if (groupSize > 0)
		numChars += size / groupSize;

	String s (PreallocationBytes (sizeof (CharPointerType::CharType) * (size_t) numChars));

	const unsigned char* data = static_cast <const unsigned char*> (d);
	CharPointerType dest (s.text);

	for (int i = 0; i < size; ++i)
	{
		const unsigned char nextByte = *data++;
		dest.write ((juce_wchar) hexDigits [nextByte >> 4]);
		dest.write ((juce_wchar) hexDigits [nextByte & 0xf]);

		if (groupSize > 0 && (i % groupSize) == (groupSize - 1) && i < (size - 1))
			dest.write ((juce_wchar) ' ');
	}

	dest.writeNull();
	return s;
}

int String::getHexValue32() const noexcept
{
	return HexConverter<int>::stringToHex (text);
}

int64 String::getHexValue64() const noexcept
{
	return HexConverter<int64>::stringToHex (text);
}

String String::createStringFromData (const void* const data_, const int size)
{
	const uint8* const data = static_cast <const uint8*> (data_);

	if (size <= 0 || data == nullptr)
	{
		return empty;
	}
	else if (size == 1)
	{
		return charToString ((char) data[0]);
	}
	else if ((data[0] == (uint8) CharPointer_UTF16::byteOrderMarkBE1 && data[1] == (uint8) CharPointer_UTF16::byteOrderMarkBE2)
		  || (data[0] == (uint8) CharPointer_UTF16::byteOrderMarkLE1 && data[1] == (uint8) CharPointer_UTF16::byteOrderMarkLE2))
	{
		const bool bigEndian = (data[0] == (uint8) CharPointer_UTF16::byteOrderMarkBE1);
		const int numChars = size / 2 - 1;

		StringCreationHelper builder (numChars);

		const uint16* const src = (const uint16*) (data + 2);

		if (bigEndian)
		{
			for (int i = 0; i < numChars; ++i)
				builder.write ((juce_wchar) ByteOrder::swapIfLittleEndian (src[i]));
		}
		else
		{
			for (int i = 0; i < numChars; ++i)
				builder.write ((juce_wchar) ByteOrder::swapIfBigEndian (src[i]));
		}

		builder.write (0);
		return builder.result;
	}
	else
	{
		const uint8* start = data;
		const uint8* end = data + size;

		if (size >= 3
			  && data[0] == (uint8) CharPointer_UTF8::byteOrderMark1
			  && data[1] == (uint8) CharPointer_UTF8::byteOrderMark2
			  && data[2] == (uint8) CharPointer_UTF8::byteOrderMark3)
			start += 3;

		return String (CharPointer_UTF8 ((const char*) start),
					   CharPointer_UTF8 ((const char*) end));
	}
}

static juce_wchar emptyChar = 0;

template <class CharPointerType_Src, class CharPointerType_Dest>
struct StringEncodingConverter
{
	static const CharPointerType_Dest convert (const String& s)
	{
		String& source = const_cast <String&> (s);

		typedef typename CharPointerType_Dest::CharType DestChar;

		if (source.isEmpty())
			return CharPointerType_Dest (reinterpret_cast <const DestChar*> (&emptyChar));

		CharPointerType_Src text (source.getCharPointer());
		const size_t extraBytesNeeded = CharPointerType_Dest::getBytesRequiredFor (text);
		const size_t endOffset = (text.sizeInBytes() + 3) & ~3; // the new string must be word-aligned or many win32
																// functions will fail to read it correctly!
		source.preallocateBytes (endOffset + extraBytesNeeded);
		text = source.getCharPointer();

		void* newSpace = addBytesToPointer (text.getAddress(), (int) endOffset);
		const CharPointerType_Dest extraSpace (static_cast <DestChar*> (newSpace));

	   #if JUCE_DEBUG // (This just avoids spurious warnings from valgrind about the uninitialised bytes at the end of the buffer..)
		const int bytesToClear = jmin ((int) extraBytesNeeded, 4);
		zeromem (addBytesToPointer (newSpace, (int) (extraBytesNeeded - bytesToClear)), bytesToClear);
	   #endif

		CharPointerType_Dest (extraSpace).writeAll (text);
		return extraSpace;
	}
};

template <>
struct StringEncodingConverter <CharPointer_UTF8, CharPointer_UTF8>
{
	static const CharPointer_UTF8 convert (const String& source) noexcept   { return CharPointer_UTF8 ((CharPointer_UTF8::CharType*) source.getCharPointer().getAddress()); }
};

template <>
struct StringEncodingConverter <CharPointer_UTF16, CharPointer_UTF16>
{
	static const CharPointer_UTF16 convert (const String& source) noexcept  { return CharPointer_UTF16 ((CharPointer_UTF16::CharType*) source.getCharPointer().getAddress()); }
};

template <>
struct StringEncodingConverter <CharPointer_UTF32, CharPointer_UTF32>
{
	static const CharPointer_UTF32 convert (const String& source) noexcept  { return CharPointer_UTF32 ((CharPointer_UTF32::CharType*) source.getCharPointer().getAddress()); }
};

const CharPointer_UTF8 String::toUTF8() const
{
	return StringEncodingConverter <CharPointerType, CharPointer_UTF8>::convert (*this);
}

CharPointer_UTF16 String::toUTF16() const
{
	return StringEncodingConverter <CharPointerType, CharPointer_UTF16>::convert (*this);
}

CharPointer_UTF32 String::toUTF32() const
{
	return StringEncodingConverter <CharPointerType, CharPointer_UTF32>::convert (*this);
}

const wchar_t* String::toWideCharPointer() const
{
	return (const wchar_t*) StringEncodingConverter <CharPointerType, CharPointer_wchar_t>::convert (*this).getAddress();
}

template <class CharPointerType_Src, class CharPointerType_Dest>
struct StringCopier
{
	static int copyToBuffer (const CharPointerType_Src& source, typename CharPointerType_Dest::CharType* const buffer, const int maxBufferSizeBytes)
	{
		jassert (maxBufferSizeBytes >= 0); // keep this value positive, or no characters will be copied!

		if (buffer == nullptr)
			return (int) (CharPointerType_Dest::getBytesRequiredFor (source) + sizeof (typename CharPointerType_Dest::CharType));

		return CharPointerType_Dest (buffer).writeWithDestByteLimit (source, maxBufferSizeBytes);
	}
};

int String::copyToUTF8 (CharPointer_UTF8::CharType* const buffer, const int maxBufferSizeBytes) const noexcept
{
	return StringCopier <CharPointerType, CharPointer_UTF8>::copyToBuffer (text, buffer, maxBufferSizeBytes);
}

int String::copyToUTF16 (CharPointer_UTF16::CharType* const buffer, int maxBufferSizeBytes) const noexcept
{
	return StringCopier <CharPointerType, CharPointer_UTF16>::copyToBuffer (text, buffer, maxBufferSizeBytes);
}

int String::copyToUTF32 (CharPointer_UTF32::CharType* const buffer, int maxBufferSizeBytes) const noexcept
{
	return StringCopier <CharPointerType, CharPointer_UTF32>::copyToBuffer (text, buffer, maxBufferSizeBytes);
}

int String::getNumBytesAsUTF8() const noexcept
{
	return (int) CharPointer_UTF8::getBytesRequiredFor (text);
}

String String::fromUTF8 (const char* const buffer, int bufferSizeBytes)
{
	if (buffer != nullptr)
	{
		if (bufferSizeBytes < 0)
			return String (CharPointer_UTF8 (buffer));
		else if (bufferSizeBytes > 0)
			return String (CharPointer_UTF8 (buffer), CharPointer_UTF8 (buffer + bufferSizeBytes));
	}

	return String::empty;
}

#if JUCE_MSVC
  #pragma warning (pop)
#endif

String::Concatenator::Concatenator (String& stringToAppendTo)
	: result (stringToAppendTo),
	  nextIndex (stringToAppendTo.length())
{
}

String::Concatenator::~Concatenator()
{
}

void String::Concatenator::append (const String& s)
{
	const int len = s.length();

	if (len > 0)
	{
		size_t extraBytes = s.getCharPointer().sizeInBytes();
		result.preallocateBytes (nextIndex + extraBytes);
		CharPointerType (result.text + nextIndex).writeAll (s.text);
		nextIndex += len;
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_String.cpp ***/


/*** Start of inlined file: juce_StringArray.cpp ***/
BEGIN_JUCE_NAMESPACE

StringArray::StringArray() noexcept
{
}

StringArray::StringArray (const StringArray& other)
	: strings (other.strings)
{
}

StringArray::StringArray (const String& firstValue)
{
	strings.add (firstValue);
}

namespace StringArrayHelpers
{
	template <typename CharType>
	void addArray (Array<String>& dest, const CharType* const* strings)
	{
		if (strings != nullptr)
			while (*strings != nullptr)
				dest.add (*strings++);
	}

	template <typename CharType>
	void addArray (Array<String>& dest, const CharType* const* const strings, const int numberOfStrings)
	{
		for (int i = 0; i < numberOfStrings; ++i)
			dest.add (strings [i]);
	}
}

StringArray::StringArray (const char* const* const initialStrings)
{
	StringArrayHelpers::addArray (strings, initialStrings);
}

StringArray::StringArray (const char* const* const initialStrings, const int numberOfStrings)
{
	StringArrayHelpers::addArray (strings, initialStrings, numberOfStrings);
}

StringArray::StringArray (const wchar_t* const* const initialStrings)
{
	StringArrayHelpers::addArray (strings, initialStrings);
}

StringArray::StringArray (const wchar_t* const* const initialStrings, const int numberOfStrings)
{
	StringArrayHelpers::addArray (strings, initialStrings, numberOfStrings);
}

StringArray& StringArray::operator= (const StringArray& other)
{
	strings = other.strings;
	return *this;
}

StringArray::~StringArray()
{
}

bool StringArray::operator== (const StringArray& other) const noexcept
{
	if (other.size() != size())
		return false;

	for (int i = size(); --i >= 0;)
		if (other.strings.getReference(i) != strings.getReference(i))
			return false;

	return true;
}

bool StringArray::operator!= (const StringArray& other) const noexcept
{
	return ! operator== (other);
}

void StringArray::clear()
{
	strings.clear();
}

const String& StringArray::operator[] (const int index) const noexcept
{
	if (isPositiveAndBelow (index, strings.size()))
		return strings.getReference (index);

	return String::empty;
}

String& StringArray::getReference (const int index) noexcept
{
	jassert (isPositiveAndBelow (index, strings.size()));
	return strings.getReference (index);
}

void StringArray::add (const String& newString)
{
	strings.add (newString);
}

void StringArray::insert (const int index, const String& newString)
{
	strings.insert (index, newString);
}

void StringArray::addIfNotAlreadyThere (const String& newString, const bool ignoreCase)
{
	if (! contains (newString, ignoreCase))
		add (newString);
}

void StringArray::addArray (const StringArray& otherArray, int startIndex, int numElementsToAdd)
{
	if (startIndex < 0)
	{
		jassertfalse;
		startIndex = 0;
	}

	if (numElementsToAdd < 0 || startIndex + numElementsToAdd > otherArray.size())
		numElementsToAdd = otherArray.size() - startIndex;

	while (--numElementsToAdd >= 0)
		strings.add (otherArray.strings.getReference (startIndex++));
}

void StringArray::set (const int index, const String& newString)
{
	strings.set (index, newString);
}

bool StringArray::contains (const String& stringToLookFor, const bool ignoreCase) const
{
	if (ignoreCase)
	{
		for (int i = size(); --i >= 0;)
			if (strings.getReference(i).equalsIgnoreCase (stringToLookFor))
				return true;
	}
	else
	{
		for (int i = size(); --i >= 0;)
			if (stringToLookFor == strings.getReference(i))
				return true;
	}

	return false;
}

int StringArray::indexOf (const String& stringToLookFor, const bool ignoreCase, int i) const
{
	if (i < 0)
		i = 0;

	const int numElements = size();

	if (ignoreCase)
	{
		while (i < numElements)
		{
			if (strings.getReference(i).equalsIgnoreCase (stringToLookFor))
				return i;

			++i;
		}
	}
	else
	{
		while (i < numElements)
		{
			if (stringToLookFor == strings.getReference (i))
				return i;

			++i;
		}
	}

	return -1;
}

void StringArray::remove (const int index)
{
	strings.remove (index);
}

void StringArray::removeString (const String& stringToRemove,
								const bool ignoreCase)
{
	if (ignoreCase)
	{
		for (int i = size(); --i >= 0;)
			if (strings.getReference(i).equalsIgnoreCase (stringToRemove))
				strings.remove (i);
	}
	else
	{
		for (int i = size(); --i >= 0;)
			if (stringToRemove == strings.getReference (i))
				strings.remove (i);
	}
}

void StringArray::removeRange (int startIndex, int numberToRemove)
{
	strings.removeRange (startIndex, numberToRemove);
}

void StringArray::removeEmptyStrings (const bool removeWhitespaceStrings)
{
	if (removeWhitespaceStrings)
	{
		for (int i = size(); --i >= 0;)
			if (! strings.getReference(i).containsNonWhitespaceChars())
				strings.remove (i);
	}
	else
	{
		for (int i = size(); --i >= 0;)
			if (strings.getReference(i).isEmpty())
				strings.remove (i);
	}
}

void StringArray::trim()
{
	for (int i = size(); --i >= 0;)
	{
		String& s = strings.getReference(i);
		s = s.trim();
	}
}

class InternalStringArrayComparator_CaseSensitive
{
public:
	static int compareElements (String& first, String& second)	  { return first.compare (second); }
};

class InternalStringArrayComparator_CaseInsensitive
{
public:
	static int compareElements (String& first, String& second)	  { return first.compareIgnoreCase (second); }
};

void StringArray::sort (const bool ignoreCase)
{
	if (ignoreCase)
	{
		InternalStringArrayComparator_CaseInsensitive comp;
		strings.sort (comp);
	}
	else
	{
		InternalStringArrayComparator_CaseSensitive comp;
		strings.sort (comp);
	}
}

void StringArray::move (const int currentIndex, int newIndex) noexcept
{
	strings.move (currentIndex, newIndex);
}

String StringArray::joinIntoString (const String& separator, int start, int numberToJoin) const
{
	const int last = (numberToJoin < 0) ? size()
										: jmin (size(), start + numberToJoin);

	if (start < 0)
		start = 0;

	if (start >= last)
		return String::empty;

	if (start == last - 1)
		return strings.getReference (start);

	const size_t separatorBytes = separator.getCharPointer().sizeInBytes() - sizeof (String::CharPointerType::CharType);
	size_t bytesNeeded = separatorBytes * (last - start - 1);

	for (int i = start; i < last; ++i)
		bytesNeeded += strings.getReference(i).getCharPointer().sizeInBytes() - sizeof (String::CharPointerType::CharType);

	String result;
	result.preallocateBytes (bytesNeeded);

	String::CharPointerType dest (result.getCharPointer());

	while (start < last)
	{
		const String& s = strings.getReference (start);

		if (! s.isEmpty())
			dest.writeAll (s.getCharPointer());

		if (++start < last && separatorBytes > 0)
			dest.writeAll (separator.getCharPointer());
	}

	dest.writeNull();

	return result;
}

int StringArray::addTokens (const String& text, const bool preserveQuotedStrings)
{
	return addTokens (text, " \n\r\t", preserveQuotedStrings ? "\"" : "");
}

int StringArray::addTokens (const String& text, const String& breakCharacters, const String& quoteCharacters)
{
	int num = 0;
	String::CharPointerType t (text.getCharPointer());

	while (! t.isEmpty())
	{
		String::CharPointerType tokenEnd (CharacterFunctions::findEndOfToken (t,
																			  breakCharacters.getCharPointer(),
																			  quoteCharacters.getCharPointer()));
		add (String (t, tokenEnd));
		++num;

		if (tokenEnd.isEmpty())
			break;

		t = ++tokenEnd;
	}

	return num;
}

int StringArray::addLines (const String& sourceText)
{
	int numLines = 0;
	String::CharPointerType text (sourceText.getCharPointer());
	bool finished = text.isEmpty();

	while (! finished)
	{
		String::CharPointerType startOfLine (text);
		int numChars = 0;

		for (;;)
		{
			const juce_wchar c = text.getAndAdvance();

			if (c == 0)
			{
				finished = true;
				break;
			}

			if (c == '\n')
				break;

			if (c == '\r')
			{
				if (*text == '\n')
					++text;

				break;
			}

			++numChars;
		}

		add (String (startOfLine, numChars));
		++numLines;
	}

	return numLines;
}

void StringArray::removeDuplicates (const bool ignoreCase)
{
	for (int i = 0; i < size() - 1; ++i)
	{
		const String s (strings.getReference(i));

		int nextIndex = i + 1;

		for (;;)
		{
			nextIndex = indexOf (s, ignoreCase, nextIndex);

			if (nextIndex < 0)
				break;

			strings.remove (nextIndex);
		}
	}
}

void StringArray::appendNumbersToDuplicates (const bool ignoreCase,
											 const bool appendNumberToFirstInstance,
											 CharPointer_UTF8 preNumberString,
											 CharPointer_UTF8 postNumberString)
{
	CharPointer_UTF8 defaultPre (" ("), defaultPost (")");

	if (preNumberString.getAddress() == nullptr)
		preNumberString = defaultPre;

	if (postNumberString.getAddress() == nullptr)
		postNumberString = defaultPost;

	for (int i = 0; i < size() - 1; ++i)
	{
		String& s = strings.getReference(i);

		int nextIndex = indexOf (s, ignoreCase, i + 1);

		if (nextIndex >= 0)
		{
			const String original (s);

			int number = 0;

			if (appendNumberToFirstInstance)
				s = original + String (preNumberString) + String (++number) + String (postNumberString);
			else
				++number;

			while (nextIndex >= 0)
			{
				set (nextIndex, (*this)[nextIndex] + String (preNumberString) + String (++number) + String (postNumberString));
				nextIndex = indexOf (original, ignoreCase, nextIndex + 1);
			}
		}
	}
}

void StringArray::minimiseStorageOverheads()
{
	strings.minimiseStorageOverheads();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_StringArray.cpp ***/


/*** Start of inlined file: juce_StringPairArray.cpp ***/
BEGIN_JUCE_NAMESPACE

StringPairArray::StringPairArray (const bool ignoreCase_)
	: ignoreCase (ignoreCase_)
{
}

StringPairArray::StringPairArray (const StringPairArray& other)
	: keys (other.keys),
	  values (other.values),
	  ignoreCase (other.ignoreCase)
{
}

StringPairArray::~StringPairArray()
{
}

StringPairArray& StringPairArray::operator= (const StringPairArray& other)
{
	keys = other.keys;
	values = other.values;
	return *this;
}

bool StringPairArray::operator== (const StringPairArray& other) const
{
	for (int i = keys.size(); --i >= 0;)
		if (other [keys[i]] != values[i])
			return false;

	return true;
}

bool StringPairArray::operator!= (const StringPairArray& other) const
{
	return ! operator== (other);
}

const String& StringPairArray::operator[] (const String& key) const
{
	return values [keys.indexOf (key, ignoreCase)];
}

String StringPairArray::getValue (const String& key, const String& defaultReturnValue) const
{
	const int i = keys.indexOf (key, ignoreCase);

	if (i >= 0)
		return values[i];

	return defaultReturnValue;
}

void StringPairArray::set (const String& key, const String& value)
{
	const int i = keys.indexOf (key, ignoreCase);

	if (i >= 0)
	{
		values.set (i, value);
	}
	else
	{
		keys.add (key);
		values.add (value);
	}
}

void StringPairArray::addArray (const StringPairArray& other)
{
	for (int i = 0; i < other.size(); ++i)
		set (other.keys[i], other.values[i]);
}

void StringPairArray::clear()
{
	keys.clear();
	values.clear();
}

void StringPairArray::remove (const String& key)
{
	remove (keys.indexOf (key, ignoreCase));
}

void StringPairArray::remove (const int index)
{
	keys.remove (index);
	values.remove (index);
}

void StringPairArray::setIgnoresCase (const bool shouldIgnoreCase)
{
	ignoreCase = shouldIgnoreCase;
}

String StringPairArray::getDescription() const
{
	String s;

	for (int i = 0; i < keys.size(); ++i)
	{
		s << keys[i] << " = " << values[i];
		if (i < keys.size())
			s << ", ";
	}

	return s;
}

void StringPairArray::minimiseStorageOverheads()
{
	keys.minimiseStorageOverheads();
	values.minimiseStorageOverheads();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_StringPairArray.cpp ***/


/*** Start of inlined file: juce_StringPool.cpp ***/
BEGIN_JUCE_NAMESPACE

StringPool::StringPool() noexcept   {}
StringPool::~StringPool()	   {}

namespace StringPoolHelpers
{
	template <class StringType>
	const String::CharPointerType getPooledStringFromArray (Array<String>& strings, StringType newString)
	{
		int start = 0;
		int end = strings.size();

		for (;;)
		{
			if (start >= end)
			{
				jassert (start <= end);
				strings.insert (start, newString);
				return strings.getReference (start).getCharPointer();
			}
			else
			{
				const String& startString = strings.getReference (start);

				if (startString == newString)
					return startString.getCharPointer();

				const int halfway = (start + end) >> 1;

				if (halfway == start)
				{
					if (startString.compare (newString) < 0)
						++start;

					strings.insert (start, newString);
					return strings.getReference (start).getCharPointer();
				}

				const int comp = strings.getReference (halfway).compare (newString);

				if (comp == 0)
					return strings.getReference (halfway).getCharPointer();
				else if (comp < 0)
					start = halfway;
				else
					end = halfway;
			}
		}
	}
}

const String::CharPointerType StringPool::getPooledString (const String& s)
{
	if (s.isEmpty())
		return String::empty.getCharPointer();

	return StringPoolHelpers::getPooledStringFromArray (strings, s);
}

const String::CharPointerType StringPool::getPooledString (const char* const s)
{
	if (s == nullptr || *s == 0)
		return String::empty.getCharPointer();

	return StringPoolHelpers::getPooledStringFromArray (strings, s);
}

const String::CharPointerType StringPool::getPooledString (const wchar_t* const s)
{
	if (s == nullptr || *s == 0)
		return String::empty.getCharPointer();

	return StringPoolHelpers::getPooledStringFromArray (strings, s);
}

int StringPool::size() const noexcept
{
	return strings.size();
}

const String::CharPointerType StringPool::operator[] (const int index) const noexcept
{
	return strings [index].getCharPointer();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_StringPool.cpp ***/


/*** Start of inlined file: juce_XmlDocument.cpp ***/
BEGIN_JUCE_NAMESPACE

XmlDocument::XmlDocument (const String& documentText)
	: originalText (documentText),
	  input (nullptr),
	  ignoreEmptyTextElements (true)
{
}

XmlDocument::XmlDocument (const File& file)
	: input (nullptr),
	  ignoreEmptyTextElements (true),
	  inputSource (new FileInputSource (file))
{
}

XmlDocument::~XmlDocument()
{
}

XmlElement* XmlDocument::parse (const File& file)
{
	XmlDocument doc (file);
	return doc.getDocumentElement();
}

XmlElement* XmlDocument::parse (const String& xmlData)
{
	XmlDocument doc (xmlData);
	return doc.getDocumentElement();
}

void XmlDocument::setInputSource (InputSource* const newSource) noexcept
{
	inputSource = newSource;
}

void XmlDocument::setEmptyTextElementsIgnored (const bool shouldBeIgnored) noexcept
{
	ignoreEmptyTextElements = shouldBeIgnored;
}

namespace XmlIdentifierChars
{
	bool isIdentifierCharSlow (const juce_wchar c) noexcept
	{
		return CharacterFunctions::isLetterOrDigit (c)
				 || c == '_' || c == '-' || c == ':' || c == '.';
	}

	bool isIdentifierChar (const juce_wchar c) noexcept
	{
		static const uint32 legalChars[] = { 0, 0x7ff6000, 0x87fffffe, 0x7fffffe, 0 };

		return ((int) c < (int) numElementsInArray (legalChars) * 32) ? ((legalChars [c >> 5] & (1 << (c & 31))) != 0)
																	  : isIdentifierCharSlow (c);
	}

	/*static void generateIdentifierCharConstants()
	{
		uint32 n[8] = { 0 };
		for (int i = 0; i < 256; ++i)
			if (isIdentifierCharSlow (i))
				n[i >> 5] |= (1 << (i & 31));

		String s;
		for (int i = 0; i < 8; ++i)
			s << "0x" << String::toHexString ((int) n[i]) << ", ";

		DBG (s);
	}*/
}

XmlElement* XmlDocument::getDocumentElement (const bool onlyReadOuterDocumentElement)
{
	String textToParse (originalText);

	if (textToParse.isEmpty() && inputSource != nullptr)
	{
		ScopedPointer <InputStream> in (inputSource->createInputStream());

		if (in != nullptr)
		{
			MemoryOutputStream data;
			data.writeFromInputStream (*in, onlyReadOuterDocumentElement ? 8192 : -1);
			textToParse = data.toString();

			if (! onlyReadOuterDocumentElement)
				originalText = textToParse;
		}
	}

	input = textToParse.getCharPointer();
	lastError = String::empty;
	errorOccurred = false;
	outOfData = false;
	needToLoadDTD = true;

	if (textToParse.isEmpty())
	{
		lastError = "not enough input";
	}
	else
	{
		skipHeader();

		if (input.getAddress() != nullptr)
		{
			ScopedPointer <XmlElement> result (readNextElement (! onlyReadOuterDocumentElement));

			if (! errorOccurred)
				return result.release();
		}
		else
		{
			lastError = "incorrect xml header";
		}
	}

	return nullptr;
}

const String& XmlDocument::getLastParseError() const noexcept
{
	return lastError;
}

void XmlDocument::setLastError (const String& desc, const bool carryOn)
{
	lastError = desc;
	errorOccurred = ! carryOn;
}

String XmlDocument::getFileContents (const String& filename) const
{
	if (inputSource != nullptr)
	{
		const ScopedPointer <InputStream> in (inputSource->createInputStreamFor (filename.trim().unquoted()));

		if (in != nullptr)
			return in->readEntireStreamAsString();
	}

	return String::empty;
}

juce_wchar XmlDocument::readNextChar() noexcept
{
	const juce_wchar c = input.getAndAdvance();

	if (c == 0)
	{
		outOfData = true;
		--input;
	}

	return c;
}

int XmlDocument::findNextTokenLength() noexcept
{
	int len = 0;
	juce_wchar c = *input;

	while (XmlIdentifierChars::isIdentifierChar (c))
		c = input [++len];

	return len;
}

void XmlDocument::skipHeader()
{
	const int headerStart = input.indexOf (CharPointer_UTF8 ("<?xml"));

	if (headerStart >= 0)
	{
		const int headerEnd = (input + headerStart).indexOf (CharPointer_UTF8 ("?>"));
		if (headerEnd < 0)
			return;

	   #if JUCE_DEBUG
		const String header (input + headerStart, headerEnd - headerStart);
		const String encoding (header.fromFirstOccurrenceOf ("encoding", false, true)
									 .fromFirstOccurrenceOf ("=", false, false)
									 .fromFirstOccurrenceOf ("\"", false, false)
									 .upToFirstOccurrenceOf ("\"", false, false).trim());

		/* If you load an XML document with a non-UTF encoding type, it may have been
		   loaded wrongly.. Since all the files are read via the normal juce file streams,
		   they're treated as UTF-8, so by the time it gets to the parser, the encoding will
		   have been lost. Best plan is to stick to utf-8 or if you have specific files to
		   read, use your own code to convert them to a unicode String, and pass that to the
		   XML parser.
		*/
		jassert (encoding.isEmpty() || encoding.startsWithIgnoreCase ("utf-"));
	   #endif

		input += headerEnd + 2;
	}

	skipNextWhiteSpace();

	const int docTypeIndex = input.indexOf (CharPointer_UTF8 ("<!DOCTYPE"));
	if (docTypeIndex < 0)
		return;

	input += docTypeIndex + 9;
	const String::CharPointerType docType (input);

	int n = 1;

	while (n > 0)
	{
		const juce_wchar c = readNextChar();

		if (outOfData)
			return;

		if (c == '<')
			++n;
		else if (c == '>')
			--n;
	}

	dtdText = String (docType, (int) (input.getAddress() - (docType.getAddress() + 1))).trim();
}

void XmlDocument::skipNextWhiteSpace()
{
	for (;;)
	{
		juce_wchar c = *input;

		while (CharacterFunctions::isWhitespace (c))
			c = *++input;

		if (c == 0)
		{
			outOfData = true;
			break;
		}
		else if (c == '<')
		{
			if (input[1] == '!'
				 && input[2] == '-'
				 && input[3] == '-')
			{
				const int closeComment = input.indexOf (CharPointer_UTF8 ("-->"));

				if (closeComment < 0)
				{
					outOfData = true;
					break;
				}

				input += closeComment + 3;
				continue;
			}
			else if (input[1] == '?')
			{
				const int closeBracket = input.indexOf (CharPointer_UTF8 ("?>"));

				if (closeBracket < 0)
				{
					outOfData = true;
					break;
				}

				input += closeBracket + 2;
				continue;
			}
		}

		break;
	}
}

void XmlDocument::readQuotedString (String& result)
{
	const juce_wchar quote = readNextChar();

	while (! outOfData)
	{
		const juce_wchar c = readNextChar();

		if (c == quote)
			break;

		--input;

		if (c == '&')
		{
			readEntity (result);
		}
		else
		{
			const String::CharPointerType start (input);
			int numChars = 0;

			for (;;)
			{
				const juce_wchar character = *input;

				if (character == quote)
				{
					result.appendCharPointer (start, numChars);
					++input;
					return;
				}
				else if (character == '&')
				{
					result.appendCharPointer (start, numChars);
					break;
				}
				else if (character == 0)
				{
					outOfData = true;
					setLastError ("unmatched quotes", false);
					break;
				}

				++input;
				++numChars;
			}
		}
	}
}

XmlElement* XmlDocument::readNextElement (const bool alsoParseSubElements)
{
	XmlElement* node = nullptr;

	skipNextWhiteSpace();
	if (outOfData)
		return nullptr;

	const int openBracket = input.indexOf ((juce_wchar) '<');

	if (openBracket >= 0)
	{
		input += openBracket + 1;
		int tagLen = findNextTokenLength();

		if (tagLen == 0)
		{
			// no tag name - but allow for a gap after the '<' before giving an error
			skipNextWhiteSpace();
			tagLen = findNextTokenLength();

			if (tagLen == 0)
			{
				setLastError ("tag name missing", false);
				return node;
			}
		}

		node = new XmlElement (String (input, tagLen));
		input += tagLen;
		LinkedListPointer<XmlElement::XmlAttributeNode>::Appender attributeAppender (node->attributes);

		// look for attributes
		for (;;)
		{
			skipNextWhiteSpace();

			const juce_wchar c = *input;

			// empty tag..
			if (c == '/' && input[1] == '>')
			{
				input += 2;
				break;
			}

			// parse the guts of the element..
			if (c == '>')
			{
				++input;

				if (alsoParseSubElements)
					readChildElements (node);

				break;
			}

			// get an attribute..
			if (XmlIdentifierChars::isIdentifierChar (c))
			{
				const int attNameLen = findNextTokenLength();

				if (attNameLen > 0)
				{
					const String::CharPointerType attNameStart (input);
					input += attNameLen;

					skipNextWhiteSpace();

					if (readNextChar() == '=')
					{
						skipNextWhiteSpace();

						const juce_wchar nextChar = *input;

						if (nextChar == '"' || nextChar == '\'')
						{
							XmlElement::XmlAttributeNode* const newAtt
								= new XmlElement::XmlAttributeNode (String (attNameStart, attNameLen),
																	String::empty);

							readQuotedString (newAtt->value);
							attributeAppender.append (newAtt);
							continue;
						}
					}
				}
			}
			else
			{
				if (! outOfData)
					setLastError ("illegal character found in " + node->getTagName() + ": '" + c + "'", false);
			}

			break;
		}
	}

	return node;
}

void XmlDocument::readChildElements (XmlElement* parent)
{
	LinkedListPointer<XmlElement>::Appender childAppender (parent->firstChildElement);

	for (;;)
	{
		const String::CharPointerType preWhitespaceInput (input);
		skipNextWhiteSpace();

		if (outOfData)
		{
			setLastError ("unmatched tags", false);
			break;
		}

		if (*input == '<')
		{
			if (input[1] == '/')
			{
				// our close tag..
				const int closeTag = input.indexOf ((juce_wchar) '>');

				if (closeTag >= 0)
					input += closeTag + 1;

				break;
			}
			else if (input[1] == '!'
				  && input[2] == '['
				  && input[3] == 'C'
				  && input[4] == 'D'
				  && input[5] == 'A'
				  && input[6] == 'T'
				  && input[7] == 'A'
				  && input[8] == '[')
			{
				input += 9;
				const String::CharPointerType inputStart (input);

				int len = 0;

				for (;;)
				{
					if (*input == 0)
					{
						setLastError ("unterminated CDATA section", false);
						outOfData = true;
						break;
					}
					else if (input[0] == ']'
							  && input[1] == ']'
							  && input[2] == '>')
					{
						input += 3;
						break;
					}

					++input;
					++len;
				}

				childAppender.append (XmlElement::createTextElement (String (inputStart, len)));
			}
			else
			{
				// this is some other element, so parse and add it..
				XmlElement* const n = readNextElement (true);

				if (n != nullptr)
					childAppender.append (n);
				else
					break;
			}
		}
		else  // must be a character block
		{
			input = preWhitespaceInput; // roll back to include the leading whitespace
			String textElementContent;

			for (;;)
			{
				const juce_wchar c = *input;

				if (c == '<')
					break;

				if (c == 0)
				{
					setLastError ("unmatched tags", false);
					outOfData = true;
					return;
				}

				if (c == '&')
				{
					String entity;
					readEntity (entity);

					if (entity.startsWithChar ('<') && entity [1] != 0)
					{
						const String::CharPointerType oldInput (input);
						const bool oldOutOfData = outOfData;

						input = entity.getCharPointer();
						outOfData = false;

						for (;;)
						{
							XmlElement* const n = readNextElement (true);

							if (n == nullptr)
								break;

							childAppender.append (n);
						}

						input = oldInput;
						outOfData = oldOutOfData;
					}
					else
					{
						textElementContent += entity;
					}
				}
				else
				{
					const String::CharPointerType start (input);
					int len = 0;

					for (;;)
					{
						const juce_wchar nextChar = *input;

						if (nextChar == '<' || nextChar == '&')
						{
							break;
						}
						else if (nextChar == 0)
						{
							setLastError ("unmatched tags", false);
							outOfData = true;
							return;
						}

						++input;
						++len;
					}

					textElementContent.appendCharPointer (start, len);
				}
			}

			if ((! ignoreEmptyTextElements) || textElementContent.containsNonWhitespaceChars())
			{
				childAppender.append (XmlElement::createTextElement (textElementContent));
			}
		}
	}
}

void XmlDocument::readEntity (String& result)
{
	// skip over the ampersand
	++input;

	if (input.compareIgnoreCaseUpTo (CharPointer_UTF8 ("amp;"), 4) == 0)
	{
		input += 4;
		result += '&';
	}
	else if (input.compareIgnoreCaseUpTo (CharPointer_UTF8 ("quot;"), 5) == 0)
	{
		input += 5;
		result += '"';
	}
	else if (input.compareIgnoreCaseUpTo (CharPointer_UTF8 ("apos;"), 5) == 0)
	{
		input += 5;
		result += '\'';
	}
	else if (input.compareIgnoreCaseUpTo (CharPointer_UTF8 ("lt;"), 3) == 0)
	{
		input += 3;
		result += '<';
	}
	else if (input.compareIgnoreCaseUpTo (CharPointer_UTF8 ("gt;"), 3) == 0)
	{
		input += 3;
		result += '>';
	}
	else if (*input == '#')
	{
		int charCode = 0;
		++input;

		if (*input == 'x' || *input == 'X')
		{
			++input;
			int numChars = 0;

			while (input[0] != ';')
			{
				const int hexValue = CharacterFunctions::getHexDigitValue (input[0]);

				if (hexValue < 0 || ++numChars > 8)
				{
					setLastError ("illegal escape sequence", true);
					break;
				}

				charCode = (charCode << 4) | hexValue;
				++input;
			}

			++input;
		}
		else if (input[0] >= '0' && input[0] <= '9')
		{
			int numChars = 0;

			while (input[0] != ';')
			{
				if (++numChars > 12)
				{
					setLastError ("illegal escape sequence", true);
					break;
				}

				charCode = charCode * 10 + (input[0] - '0');
				++input;
			}

			++input;
		}
		else
		{
			setLastError ("illegal escape sequence", true);
			result += '&';
			return;
		}

		result << (juce_wchar) charCode;
	}
	else
	{
		const String::CharPointerType entityNameStart (input);
		const int closingSemiColon = input.indexOf ((juce_wchar) ';');

		if (closingSemiColon < 0)
		{
			outOfData = true;
			result += '&';
		}
		else
		{
			input += closingSemiColon + 1;

			result += expandExternalEntity (String (entityNameStart, closingSemiColon));
		}
	}
}

String XmlDocument::expandEntity (const String& ent)
{
	if (ent.equalsIgnoreCase ("amp"))   return String::charToString ('&');
	if (ent.equalsIgnoreCase ("quot"))  return String::charToString ('"');
	if (ent.equalsIgnoreCase ("apos"))  return String::charToString ('\'');
	if (ent.equalsIgnoreCase ("lt"))    return String::charToString ('<');
	if (ent.equalsIgnoreCase ("gt"))    return String::charToString ('>');

	if (ent[0] == '#')
	{
		const juce_wchar char1 = ent[1];

		if (char1 == 'x' || char1 == 'X')
			return String::charToString (static_cast <juce_wchar> (ent.substring (2).getHexValue32()));

		if (char1 >= '0' && char1 <= '9')
			return String::charToString (static_cast <juce_wchar> (ent.substring (1).getIntValue()));

		setLastError ("illegal escape sequence", false);
		return String::charToString ('&');
	}

	return expandExternalEntity (ent);
}

String XmlDocument::expandExternalEntity (const String& entity)
{
	if (needToLoadDTD)
	{
		if (dtdText.isNotEmpty())
		{
			dtdText = dtdText.trimCharactersAtEnd (">");
			tokenisedDTD.addTokens (dtdText, true);

			if (tokenisedDTD [tokenisedDTD.size() - 2].equalsIgnoreCase ("system")
				 && tokenisedDTD [tokenisedDTD.size() - 1].isQuotedString())
			{
				const String fn (tokenisedDTD [tokenisedDTD.size() - 1]);

				tokenisedDTD.clear();
				tokenisedDTD.addTokens (getFileContents (fn), true);
			}
			else
			{
				tokenisedDTD.clear();
				const int openBracket = dtdText.indexOfChar ('[');

				if (openBracket > 0)
				{
					const int closeBracket = dtdText.lastIndexOfChar (']');

					if (closeBracket > openBracket)
						tokenisedDTD.addTokens (dtdText.substring (openBracket + 1,
																   closeBracket), true);
				}
			}

			for (int i = tokenisedDTD.size(); --i >= 0;)
			{
				if (tokenisedDTD[i].startsWithChar ('%')
					 && tokenisedDTD[i].endsWithChar (';'))
				{
					const String parsed (getParameterEntity (tokenisedDTD[i].substring (1, tokenisedDTD[i].length() - 1)));
					StringArray newToks;
					newToks.addTokens (parsed, true);

					tokenisedDTD.remove (i);

					for (int j = newToks.size(); --j >= 0;)
						tokenisedDTD.insert (i, newToks[j]);
				}
			}
		}

		needToLoadDTD = false;
	}

	for (int i = 0; i < tokenisedDTD.size(); ++i)
	{
		if (tokenisedDTD[i] == entity)
		{
			if (tokenisedDTD[i - 1].equalsIgnoreCase ("<!entity"))
			{
				String ent (tokenisedDTD [i + 1].trimCharactersAtEnd (">").trim().unquoted());

				// check for sub-entities..
				int ampersand = ent.indexOfChar ('&');

				while (ampersand >= 0)
				{
					const int semiColon = ent.indexOf (i + 1, ";");

					if (semiColon < 0)
					{
						setLastError ("entity without terminating semi-colon", false);
						break;
					}

					const String resolved (expandEntity (ent.substring (i + 1, semiColon)));

					ent = ent.substring (0, ampersand)
						   + resolved
						   + ent.substring (semiColon + 1);

					ampersand = ent.indexOfChar (semiColon + 1, '&');
				}

				return ent;
			}
		}
	}

	setLastError ("unknown entity", true);

	return entity;
}

String XmlDocument::getParameterEntity (const String& entity)
{
	for (int i = 0; i < tokenisedDTD.size(); ++i)
	{
		if (tokenisedDTD[i] == entity
			 && tokenisedDTD [i - 1] == "%"
			 && tokenisedDTD [i - 2].equalsIgnoreCase ("<!entity"))
		{
			const String ent (tokenisedDTD [i + 1].trimCharactersAtEnd (">"));

			if (ent.equalsIgnoreCase ("system"))
				return getFileContents (tokenisedDTD [i + 2].trimCharactersAtEnd (">"));
			else
				return ent.trim().unquoted();
		}
	}

	return entity;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_XmlDocument.cpp ***/


/*** Start of inlined file: juce_XmlElement.cpp ***/
BEGIN_JUCE_NAMESPACE

XmlElement::XmlAttributeNode::XmlAttributeNode (const XmlAttributeNode& other) noexcept
	: name (other.name),
	  value (other.value)
{
}

XmlElement::XmlAttributeNode::XmlAttributeNode (const String& name_, const String& value_) noexcept
	: name (name_),
	  value (value_)
{
   #if JUCE_DEBUG
	// this checks whether the attribute name string contains any illegal characters..
	for (String::CharPointerType t (name.getCharPointer()); ! t.isEmpty(); ++t)
		jassert (t.isLetterOrDigit() || *t == '_' || *t == '-' || *t == ':');
   #endif
}

inline bool XmlElement::XmlAttributeNode::hasName (const String& nameToMatch) const noexcept
{
	return name.equalsIgnoreCase (nameToMatch);
}

XmlElement::XmlElement (const String& tagName_) noexcept
	: tagName (tagName_)
{
	// the tag name mustn't be empty, or it'll look like a text element!
	jassert (tagName_.containsNonWhitespaceChars())

	// The tag can't contain spaces or other characters that would create invalid XML!
	jassert (! tagName_.containsAnyOf (" <>/&"));
}

XmlElement::XmlElement (int /*dummy*/) noexcept
{
}

XmlElement::XmlElement (const XmlElement& other)
	: tagName (other.tagName)
{
	copyChildrenAndAttributesFrom (other);
}

XmlElement& XmlElement::operator= (const XmlElement& other)
{
	if (this != &other)
	{
		removeAllAttributes();
		deleteAllChildElements();

		tagName = other.tagName;

		copyChildrenAndAttributesFrom (other);
	}

	return *this;
}

void XmlElement::copyChildrenAndAttributesFrom (const XmlElement& other)
{
	jassert (firstChildElement.get() == nullptr);
	firstChildElement.addCopyOfList (other.firstChildElement);

	jassert (attributes.get() == nullptr);
	attributes.addCopyOfList (other.attributes);
}

XmlElement::~XmlElement() noexcept
{
	firstChildElement.deleteAll();
	attributes.deleteAll();
}

namespace XmlOutputFunctions
{
	/*bool isLegalXmlCharSlow (const juce_wchar character) noexcept
	{
		if ((character >= 'a' && character <= 'z')
			 || (character >= 'A' && character <= 'Z')
				|| (character >= '0' && character <= '9'))
			return true;

		const char* t = " .,;:-()_+=?!'#@[]/\\*%~{}$|";

		do
		{
			if (((juce_wchar) (uint8) *t) == character)
				return true;
		}
		while (*++t != 0);

		return false;
	}

	void generateLegalCharConstants()
	{
		uint8 n[32] = { 0 };
		for (int i = 0; i < 256; ++i)
			if (isLegalXmlCharSlow (i))
				n[i >> 3] |= (1 << (i & 7));

		String s;
		for (int i = 0; i < 32; ++i)
			s << (int) n[i] << ", ";

		DBG (s);
	}*/

	bool isLegalXmlChar (const uint32 c) noexcept
	{
		static const unsigned char legalChars[] = { 0, 0, 0, 0, 187, 255, 255, 175, 255, 255, 255, 191, 254, 255, 255, 127 };

		return c < sizeof (legalChars) * 8
				 && (legalChars [c >> 3] & (1 << (c & 7))) != 0;
	}

	void escapeIllegalXmlChars (OutputStream& outputStream, const String& text, const bool changeNewLines)
	{
		String::CharPointerType t (text.getCharPointer());

		for (;;)
		{
			const uint32 character = (uint32) t.getAndAdvance();

			if (character == 0)
				break;

			if (isLegalXmlChar (character))
			{
				outputStream << (char) character;
			}
			else
			{
				switch (character)
				{
				case '&':   outputStream << "&amp;"; break;
				case '"':   outputStream << "&quot;"; break;
				case '>':   outputStream << "&gt;"; break;
				case '<':   outputStream << "&lt;"; break;

				case '\n':
				case '\r':
					if (! changeNewLines)
					{
						outputStream << (char) character;
						break;
					}
					// Note: deliberate fall-through here!
				default:
					outputStream << "&#" << ((int) character) << ';';
					break;
				}
			}
		}
	}

	void writeSpaces (OutputStream& out, const int numSpaces)
	{
		out.writeRepeatedByte (' ', numSpaces);
	}
}

void XmlElement::writeElementAsText (OutputStream& outputStream,
									 const int indentationLevel,
									 const int lineWrapLength) const
{
	using namespace XmlOutputFunctions;
	writeSpaces (outputStream, indentationLevel);

	if (! isTextElement())
	{
		outputStream.writeByte ('<');
		outputStream << tagName;

		{
			const int attIndent = indentationLevel + tagName.length() + 1;
			int lineLen = 0;

			for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
			{
				if (lineLen > lineWrapLength && indentationLevel >= 0)
				{
					outputStream << newLine;
					writeSpaces (outputStream, attIndent);
					lineLen = 0;
				}

				const int64 startPos = outputStream.getPosition();
				outputStream.writeByte (' ');
				outputStream << att->name;
				outputStream.write ("=\"", 2);
				escapeIllegalXmlChars (outputStream, att->value, true);
				outputStream.writeByte ('"');
				lineLen += (int) (outputStream.getPosition() - startPos);
			}
		}

		if (firstChildElement != nullptr)
		{
			outputStream.writeByte ('>');

			XmlElement* child = firstChildElement;
			bool lastWasTextNode = false;

			while (child != nullptr)
			{
				if (child->isTextElement())
				{
					escapeIllegalXmlChars (outputStream, child->getText(), false);
					lastWasTextNode = true;
				}
				else
				{
					if (indentationLevel >= 0 && ! lastWasTextNode)
						outputStream << newLine;

					child->writeElementAsText (outputStream,
											   lastWasTextNode ? 0 : (indentationLevel + (indentationLevel >= 0 ? 2 : 0)), lineWrapLength);
					lastWasTextNode = false;
				}

				child = child->getNextElement();
			}

			if (indentationLevel >= 0 && ! lastWasTextNode)
			{
				outputStream << newLine;
				writeSpaces (outputStream, indentationLevel);
			}

			outputStream.write ("</", 2);
			outputStream << tagName;
			outputStream.writeByte ('>');
		}
		else
		{
			outputStream.write ("/>", 2);
		}
	}
	else
	{
		escapeIllegalXmlChars (outputStream, getText(), false);
	}
}

String XmlElement::createDocument (const String& dtdToUse,
								   const bool allOnOneLine,
								   const bool includeXmlHeader,
								   const String& encodingType,
								   const int lineWrapLength) const
{
	MemoryOutputStream mem (2048);
	writeToStream (mem, dtdToUse, allOnOneLine, includeXmlHeader, encodingType, lineWrapLength);

	return mem.toUTF8();
}

void XmlElement::writeToStream (OutputStream& output,
								const String& dtdToUse,
								const bool allOnOneLine,
								const bool includeXmlHeader,
								const String& encodingType,
								const int lineWrapLength) const
{
	using namespace XmlOutputFunctions;

	if (includeXmlHeader)
	{
		output << "<?xml version=\"1.0\" encoding=\"" << encodingType << "\"?>";

		if (allOnOneLine)
			output.writeByte (' ');
		else
			output << newLine << newLine;
	}

	if (dtdToUse.isNotEmpty())
	{
		output << dtdToUse;

		if (allOnOneLine)
			output.writeByte (' ');
		else
			output << newLine;
	}

	writeElementAsText (output, allOnOneLine ? -1 : 0, lineWrapLength);

	if (! allOnOneLine)
		output << newLine;
}

bool XmlElement::writeToFile (const File& file,
							  const String& dtdToUse,
							  const String& encodingType,
							  const int lineWrapLength) const
{
	if (file.hasWriteAccess())
	{
		TemporaryFile tempFile (file);
		ScopedPointer <FileOutputStream> out (tempFile.getFile().createOutputStream());

		if (out != nullptr)
		{
			writeToStream (*out, dtdToUse, false, true, encodingType, lineWrapLength);
			out = nullptr;

			return tempFile.overwriteTargetFileWithTemporary();
		}
	}

	return false;
}

bool XmlElement::hasTagName (const String& tagNameWanted) const noexcept
{
   #if JUCE_DEBUG
	// if debugging, check that the case is actually the same, because
	// valid xml is case-sensitive, and although this lets it pass, it's
	// better not to..
	if (tagName.equalsIgnoreCase (tagNameWanted))
	{
		jassert (tagName == tagNameWanted);
		return true;
	}
	else
	{
		return false;
	}
   #else
	return tagName.equalsIgnoreCase (tagNameWanted);
   #endif
}

XmlElement* XmlElement::getNextElementWithTagName (const String& requiredTagName) const
{
	XmlElement* e = nextListItem;

	while (e != nullptr && ! e->hasTagName (requiredTagName))
		e = e->nextListItem;

	return e;
}

int XmlElement::getNumAttributes() const noexcept
{
	return attributes.size();
}

const String& XmlElement::getAttributeName (const int index) const noexcept
{
	const XmlAttributeNode* const att = attributes [index];
	return att != nullptr ? att->name : String::empty;
}

const String& XmlElement::getAttributeValue (const int index) const noexcept
{
	const XmlAttributeNode* const att = attributes [index];
	return att != nullptr ? att->value : String::empty;
}

bool XmlElement::hasAttribute (const String& attributeName) const noexcept
{
	for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
		if (att->hasName (attributeName))
			return true;

	return false;
}

const String& XmlElement::getStringAttribute (const String& attributeName) const noexcept
{
	for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
		if (att->hasName (attributeName))
			return att->value;

	return String::empty;
}

String XmlElement::getStringAttribute (const String& attributeName, const String& defaultReturnValue) const
{
	for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
		if (att->hasName (attributeName))
			return att->value;

	return defaultReturnValue;
}

int XmlElement::getIntAttribute (const String& attributeName, const int defaultReturnValue) const
{
	for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
		if (att->hasName (attributeName))
			return att->value.getIntValue();

	return defaultReturnValue;
}

double XmlElement::getDoubleAttribute (const String& attributeName, const double defaultReturnValue) const
{
	for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
		if (att->hasName (attributeName))
			return att->value.getDoubleValue();

	return defaultReturnValue;
}

bool XmlElement::getBoolAttribute (const String& attributeName, const bool defaultReturnValue) const
{
	for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
	{
		if (att->hasName (attributeName))
		{
			juce_wchar firstChar = att->value[0];

			if (CharacterFunctions::isWhitespace (firstChar))
				firstChar = att->value.trimStart() [0];

			return firstChar == '1'
				|| firstChar == 't'
				|| firstChar == 'y'
				|| firstChar == 'T'
				|| firstChar == 'Y';
		}
	}

	return defaultReturnValue;
}

bool XmlElement::compareAttribute (const String& attributeName,
								   const String& stringToCompareAgainst,
								   const bool ignoreCase) const noexcept
{
	for (const XmlAttributeNode* att = attributes; att != nullptr; att = att->nextListItem)
		if (att->hasName (attributeName))
			return ignoreCase ? att->value.equalsIgnoreCase (stringToCompareAgainst)
							  : att->value == stringToCompareAgainst;

	return false;
}

void XmlElement::setAttribute (const String& attributeName, const String& value)
{
	if (attributes == nullptr)
	{
		attributes = new XmlAttributeNode (attributeName, value);
	}
	else
	{
		XmlAttributeNode* att = attributes;

		for (;;)
		{
			if (att->hasName (attributeName))
			{
				att->value = value;
				break;
			}
			else if (att->nextListItem == nullptr)
			{
				att->nextListItem = new XmlAttributeNode (attributeName, value);
				break;
			}

			att = att->nextListItem;
		}
	}
}

void XmlElement::setAttribute (const String& attributeName, const int number)
{
	setAttribute (attributeName, String (number));
}

void XmlElement::setAttribute (const String& attributeName, const double number)
{
	setAttribute (attributeName, String (number));
}

void XmlElement::removeAttribute (const String& attributeName) noexcept
{
	LinkedListPointer<XmlAttributeNode>* att = &attributes;

	while (att->get() != nullptr)
	{
		if (att->get()->hasName (attributeName))
		{
			delete att->removeNext();
			break;
		}

		att = &(att->get()->nextListItem);
	}
}

void XmlElement::removeAllAttributes() noexcept
{
	attributes.deleteAll();
}

int XmlElement::getNumChildElements() const noexcept
{
	return firstChildElement.size();
}

XmlElement* XmlElement::getChildElement (const int index) const noexcept
{
	return firstChildElement [index].get();
}

XmlElement* XmlElement::getChildByName (const String& childName) const noexcept
{
	XmlElement* child = firstChildElement;

	while (child != nullptr)
	{
		if (child->hasTagName (childName))
			break;

		child = child->nextListItem;
	}

	return child;
}

void XmlElement::addChildElement (XmlElement* const newNode) noexcept
{
	if (newNode != nullptr)
		firstChildElement.append (newNode);
}

void XmlElement::insertChildElement (XmlElement* const newNode,
									 int indexToInsertAt) noexcept
{
	if (newNode != nullptr)
	{
		removeChildElement (newNode, false);
		firstChildElement.insertAtIndex (indexToInsertAt, newNode);
	}
}

XmlElement* XmlElement::createNewChildElement (const String& childTagName)
{
	XmlElement* const newElement = new XmlElement (childTagName);
	addChildElement (newElement);
	return newElement;
}

bool XmlElement::replaceChildElement (XmlElement* const currentChildElement,
									  XmlElement* const newNode) noexcept
{
	if (newNode != nullptr)
	{
		LinkedListPointer<XmlElement>* const p = firstChildElement.findPointerTo (currentChildElement);

		if (p != nullptr)
		{
			if (currentChildElement != newNode)
				delete p->replaceNext (newNode);

			return true;
		}
	}

	return false;
}

void XmlElement::removeChildElement (XmlElement* const childToRemove,
									 const bool shouldDeleteTheChild) noexcept
{
	if (childToRemove != nullptr)
	{
		firstChildElement.remove (childToRemove);

		if (shouldDeleteTheChild)
			delete childToRemove;
	}
}

bool XmlElement::isEquivalentTo (const XmlElement* const other,
								 const bool ignoreOrderOfAttributes) const noexcept
{
	if (this != other)
	{
		if (other == nullptr || tagName != other->tagName)
			return false;

		if (ignoreOrderOfAttributes)
		{
			int totalAtts = 0;
			const XmlAttributeNode* att = attributes;

			while (att != nullptr)
			{
				if (! other->compareAttribute (att->name, att->value))
					return false;

				att = att->nextListItem;
				++totalAtts;
			}

			if (totalAtts != other->getNumAttributes())
				return false;
		}
		else
		{
			const XmlAttributeNode* thisAtt = attributes;
			const XmlAttributeNode* otherAtt = other->attributes;

			for (;;)
			{
				if (thisAtt == nullptr || otherAtt == nullptr)
				{
					if (thisAtt == otherAtt) // both 0, so it's a match
						break;

					return false;
				}

				if (thisAtt->name != otherAtt->name
					 || thisAtt->value != otherAtt->value)
				{
					return false;
				}

				thisAtt = thisAtt->nextListItem;
				otherAtt = otherAtt->nextListItem;
			}
		}

		const XmlElement* thisChild = firstChildElement;
		const XmlElement* otherChild = other->firstChildElement;

		for (;;)
		{
			if (thisChild == nullptr || otherChild == nullptr)
			{
				if (thisChild == otherChild) // both 0, so it's a match
					break;

				return false;
			}

			if (! thisChild->isEquivalentTo (otherChild, ignoreOrderOfAttributes))
				return false;

			thisChild = thisChild->nextListItem;
			otherChild = otherChild->nextListItem;
		}
	}

	return true;
}

void XmlElement::deleteAllChildElements() noexcept
{
	firstChildElement.deleteAll();
}

void XmlElement::deleteAllChildElementsWithTagName (const String& name) noexcept
{
	XmlElement* child = firstChildElement;

	while (child != nullptr)
	{
		XmlElement* const nextChild = child->nextListItem;

		if (child->hasTagName (name))
			removeChildElement (child, true);

		child = nextChild;
	}
}

bool XmlElement::containsChildElement (const XmlElement* const possibleChild) const noexcept
{
	return firstChildElement.contains (possibleChild);
}

XmlElement* XmlElement::findParentElementOf (const XmlElement* const elementToLookFor) noexcept
{
	if (this == elementToLookFor || elementToLookFor == nullptr)
		return nullptr;

	XmlElement* child = firstChildElement;

	while (child != nullptr)
	{
		if (elementToLookFor == child)
			return this;

		XmlElement* const found = child->findParentElementOf (elementToLookFor);

		if (found != nullptr)
			return found;

		child = child->nextListItem;
	}

	return nullptr;
}

void XmlElement::getChildElementsAsArray (XmlElement** elems) const noexcept
{
	firstChildElement.copyToArray (elems);
}

void XmlElement::reorderChildElements (XmlElement** const elems, const int num) noexcept
{
	XmlElement* e = firstChildElement = elems[0];

	for (int i = 1; i < num; ++i)
	{
		e->nextListItem = elems[i];
		e = e->nextListItem;
	}

	e->nextListItem = nullptr;
}

bool XmlElement::isTextElement() const noexcept
{
	return tagName.isEmpty();
}

static const String juce_xmltextContentAttributeName ("text");

const String& XmlElement::getText() const noexcept
{
	jassert (isTextElement());  // you're trying to get the text from an element that
								// isn't actually a text element.. If this contains text sub-nodes, you
								// probably want to use getAllSubText instead.

	return getStringAttribute (juce_xmltextContentAttributeName);
}

void XmlElement::setText (const String& newText)
{
	if (isTextElement())
		setAttribute (juce_xmltextContentAttributeName, newText);
	else
		jassertfalse; // you can only change the text in a text element, not a normal one.
}

String XmlElement::getAllSubText() const
{
	if (isTextElement())
		return getText();

	String result;
	String::Concatenator concatenator (result);
	const XmlElement* child = firstChildElement;

	while (child != nullptr)
	{
		concatenator.append (child->getAllSubText());
		child = child->nextListItem;
	}

	return result;
}

String XmlElement::getChildElementAllSubText (const String& childTagName,
											  const String& defaultReturnValue) const
{
	const XmlElement* const child = getChildByName (childTagName);

	if (child != nullptr)
		return child->getAllSubText();

	return defaultReturnValue;
}

XmlElement* XmlElement::createTextElement (const String& text)
{
	XmlElement* const e = new XmlElement ((int) 0);
	e->setAttribute (juce_xmltextContentAttributeName, text);
	return e;
}

void XmlElement::addTextElement (const String& text)
{
	addChildElement (createTextElement (text));
}

void XmlElement::deleteAllTextElements() noexcept
{
	XmlElement* child = firstChildElement;

	while (child != nullptr)
	{
		XmlElement* const next = child->nextListItem;

		if (child->isTextElement())
			removeChildElement (child, true);

		child = next;
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_XmlElement.cpp ***/


/*** Start of inlined file: juce_JSON.cpp ***/
BEGIN_JUCE_NAMESPACE

class JSONParser
{
public:
	static Result parseAny (String::CharPointerType& t, var& result)
	{
		t = t.findEndOfWhitespace();
		String::CharPointerType t2 (t);

		switch (t2.getAndAdvance())
		{
			case '{':	t = t2; return parseObject (t, result);
			case '[':	t = t2; return parseArray (t, result);
			case '"':    t = t2; return parseString (t, result);

			case '-':
				t2 = t2.findEndOfWhitespace();
				if (! CharacterFunctions::isDigit (*t2))
					break;

				t = t2;
				return parseNumber (t, result, true);

			case '0': case '1': case '2': case '3': case '4':
			case '5': case '6': case '7': case '8': case '9':
				return parseNumber (t, result, false);

			case 't':   // "true"
				if (t2.getAndAdvance() == 'r' && t2.getAndAdvance() == 'u' && t2.getAndAdvance() == 'e')
				{
					t = t2;
					result = var (true);
					return Result::ok();
				}
				break;

			case 'f':   // "false"
				if (t2.getAndAdvance() == 'a' && t2.getAndAdvance() == 'l'
					  && t2.getAndAdvance() == 's' && t2.getAndAdvance() == 'e')
				{
					t = t2;
					result = var (false);
					return Result::ok();
				}
				break;

			case 'n':   // "null"
				if (t2.getAndAdvance() == 'u' && t2.getAndAdvance() == 'l' && t2.getAndAdvance() == 'l')
				{
					t = t2;
					result = var::null;
					return Result::ok();
				}
				break;

			default:
				break;
		}

		return createFail ("Syntax error", &t);
	}

private:
	static Result createFail (const char* const message, const String::CharPointerType* location = nullptr)
	{
		String m (message);
		if (location != nullptr)
			m << ": \"" << String (*location, 20) << '"';

		return Result::fail (m);
	}

	static Result parseNumber (String::CharPointerType& t, var& result, const bool isNegative)
	{
		String::CharPointerType oldT (t);

		int64 intValue = t.getAndAdvance() - '0';
		jassert (intValue >= 0 && intValue < 10);

		for (;;)
		{
			String::CharPointerType previousChar (t);
			const juce_wchar c = t.getAndAdvance();
			const int digit = ((int) c) - '0';

			if (isPositiveAndBelow (digit, 10))
			{
				intValue = intValue * 10 + digit;
				continue;
			}

			if (c == 'e' || c == 'E' || c == '.')
			{
				t = oldT;
				const double asDouble = CharacterFunctions::readDoubleValue (t);
				result = isNegative ? -asDouble : asDouble;
				return Result::ok();
			}

			if (CharacterFunctions::isWhitespace (c)
				 || c == ',' || c == '}' || c == ']' || c == 0)
			{
				t = previousChar;
				break;
			}

			return createFail ("Syntax error in number", &oldT);
		}

		const int64 correctedValue = isNegative ? -intValue : intValue;

		if ((intValue >> 31) != 0)
			result = correctedValue;
		else
			result = (int) correctedValue;

		return Result::ok();
	}

	static Result parseObject (String::CharPointerType& t, var& result)
	{
		DynamicObject* const resultObject = new DynamicObject();
		result = resultObject;
		NamedValueSet& resultProperties = resultObject->getProperties();

		for (;;)
		{
			t = t.findEndOfWhitespace();

			String::CharPointerType oldT (t);
			const juce_wchar c = t.getAndAdvance();

			if (c == '}')
				break;

			if (c == 0)
				return createFail ("Unexpected end-of-input in object declaration");

			if (c == '"')
			{
				var propertyNameVar;
				Result r (parseString (t, propertyNameVar));

				if (r.failed())
					return r;

				const String propertyName (propertyNameVar.toString());

				if (propertyName.isNotEmpty())
				{
					t = t.findEndOfWhitespace();
					oldT = t;

					const juce_wchar c2 = t.getAndAdvance();
					if (c2 != ':')
						return createFail ("Expected ':', but found", &oldT);

					resultProperties.set (propertyName, var::null);
					var* propertyValue = resultProperties.getVarPointer (propertyName);

					Result r2 (parseAny (t, *propertyValue));

					if (r2.failed())
						return r2;

					t = t.findEndOfWhitespace();
					oldT = t;

					const juce_wchar nextChar = t.getAndAdvance();

					if (nextChar == ',')
						continue;
					else if (nextChar == '}')
						break;
				}
			}

			return createFail ("Expected object member declaration, but found", &oldT);
		}

		return Result::ok();
	}

	static Result parseArray (String::CharPointerType& t, var& result)
	{
		result = var (Array<var>());
		Array<var>* const destArray = result.getArray();

		for (;;)
		{
			t = t.findEndOfWhitespace();

			String::CharPointerType oldT (t);
			const juce_wchar c = t.getAndAdvance();

			if (c == ']')
				break;

			if (c == 0)
				return createFail ("Unexpected end-of-input in array declaration");

			t = oldT;
			destArray->add (var::null);
			Result r (parseAny (t, destArray->getReference (destArray->size() - 1)));

			if (r.failed())
				return r;

			t = t.findEndOfWhitespace();
			oldT = t;

			const juce_wchar nextChar = t.getAndAdvance();

			if (nextChar == ',')
				continue;
			else if (nextChar == ']')
				break;

			return createFail ("Expected object array item, but found", &oldT);
		}

		return Result::ok();
	}

	static Result parseString (String::CharPointerType& t, var& result)
	{
		Array<juce_wchar> buffer;
		buffer.ensureStorageAllocated (256);

		for (;;)
		{
			juce_wchar c = t.getAndAdvance();

			if (c == '"')
				break;

			if (c == '\\')
			{
				c = t.getAndAdvance();

				switch (c)
				{
					case '"':
					case '\\':
					case '/':  break;

					case 'b':  c = '\b'; break;
					case 'f':  c = '\f'; break;
					case 'n':  c = '\n'; break;
					case 'r':  c = '\r'; break;
					case 't':  c = '\t'; break;

					case 'u':
					{
						c = 0;

						for (int i = 4; --i >= 0;)
						{
							const int digitValue = CharacterFunctions::getHexDigitValue (t.getAndAdvance());
							if (digitValue < 0)
								return createFail ("Syntax error in unicode escape sequence");

							c = (juce_wchar) ((c << 4) + digitValue);
						}

						break;
					}
				}
			}

			if (c == 0)
				return createFail ("Unexpected end-of-input in string constant");

			buffer.add (c);
		}

		buffer.add (0);
		result = String (CharPointer_UTF32 (buffer.getRawDataPointer()));
		return Result::ok();
	}
};

class JSONFormatter
{
public:
	static void write (OutputStream& out, const var& v,
					   const int indentLevel, const bool allOnOneLine)
	{
		if (v.isString())
		{
			writeString (out, v.toString().getCharPointer());
		}
		else if (v.isVoid())
		{
			out << "null";
		}
		else if (v.isBool())
		{
			out << (static_cast<bool> (v) ? "true" : "false");
		}
		else if (v.isArray())
		{
			writeArray (out, *v.getArray(), indentLevel, allOnOneLine);
		}
		else if (v.isObject())
		{
			DynamicObject* object = dynamic_cast<DynamicObject*> (v.getObject());

			jassert (object != nullptr); // Only DynamicObjects can be converted to JSON!

			writeObject (out, *object, indentLevel, allOnOneLine);
		}
		else
		{
			jassert (! v.isMethod()); // Can't convert an object with methods to JSON!

			out << v.toString();
		}
	}

private:
	enum { indentSize = 2 };

	static void writeEscapedChar (OutputStream& out, const unsigned short value)
	{
		out << "\\u" << String::toHexString ((int) value).paddedLeft ('0', 4);
	}

	static void writeString (OutputStream& out, String::CharPointerType t)
	{
		out << '"';

		for (;;)
		{
			const juce_wchar c (t.getAndAdvance());

			switch (c)
			{
				case 0:  out << '"'; return;

				case '\"':  out << "\\\""; break;
				case '\\':  out << "\\\\"; break;
				case '\b':  out << "\\b";  break;
				case '\f':  out << "\\f";  break;
				case '\t':  out << "\\t";  break;
				case '\r':  out << "\\r";  break;
				case '\n':  out << "\\n";  break;

				default:
					if (c >= 32 && c < 127)
					{
						out << (char) c;
					}
					else
					{
						if (CharPointer_UTF16::getBytesRequiredFor (c) > 2)
						{
							CharPointer_UTF16::CharType chars[2];
							CharPointer_UTF16 utf16 (chars);
							utf16.write (c);

							for (int i = 0; i < 2; ++i)
								writeEscapedChar (out, (unsigned short) chars[i]);
						}
						else
						{
							writeEscapedChar (out, (unsigned short) c);
						}
					}

					break;
			}
		}
	}

	static void writeSpaces (OutputStream& out, int numSpaces)
	{
		out.writeRepeatedByte (' ', numSpaces);
	}

	static void writeArray (OutputStream& out, const Array<var>& array,
							const int indentLevel, const bool allOnOneLine)
	{
		out << '[';
		if (! allOnOneLine)
			out << newLine;

		for (int i = 0; i < array.size(); ++i)
		{
			if (! allOnOneLine)
				writeSpaces (out, indentLevel + indentSize);

			write (out, array.getReference(i), indentLevel + indentSize, allOnOneLine);

			if (i < array.size() - 1)
			{
				if (allOnOneLine)
					out << ", ";
				else
					out << ',' << newLine;
			}
			else if (! allOnOneLine)
				out << newLine;
		}

		if (! allOnOneLine)
			writeSpaces (out, indentLevel);

		out << ']';
	}

	static void writeObject (OutputStream& out, DynamicObject& object,
							 const int indentLevel, const bool allOnOneLine)
	{
		NamedValueSet& props = object.getProperties();

		out << '{';
		if (! allOnOneLine)
			out << newLine;

		LinkedListPointer<NamedValueSet::NamedValue>* i = &(props.values);

		for (;;)
		{
			NamedValueSet::NamedValue* const v = i->get();

			if (v == nullptr)
				break;

			if (! allOnOneLine)
				writeSpaces (out, indentLevel + indentSize);

			writeString (out, v->name);
			out << ": ";
			write (out, v->value, indentLevel + indentSize, allOnOneLine);

			if (v->nextListItem.get() != nullptr)
			{
				if (allOnOneLine)
					out << ", ";
				else
					out << ',' << newLine;
			}
			else if (! allOnOneLine)
				out << newLine;

			i = &(v->nextListItem);
		}

		if (! allOnOneLine)
			writeSpaces (out, indentLevel);

		out << '}';
	}
};

var JSON::parse (const String& text)
{
	var result;
	String::CharPointerType t (text.getCharPointer());
	if (! JSONParser::parseAny (t, result))
		result = var::null;

	return result;
}

var JSON::parse (InputStream& input)
{
	return parse (input.readEntireStreamAsString());
}

var JSON::parse (const File& file)
{
	return parse (file.loadFileAsString());
}

Result JSON::parse (const String& text, var& result)
{
	String::CharPointerType t (text.getCharPointer());
	return JSONParser::parseAny (t, result);
}

String JSON::toString (const var& data, const bool allOnOneLine)
{
	MemoryOutputStream mo (1024);
	JSONFormatter::write (mo, data, 0, allOnOneLine);
	return mo.toString();
}

void JSON::writeToStream (OutputStream& output, const var& data, const bool allOnOneLine)
{
	JSONFormatter::write (output, data, 0, allOnOneLine);
}


END_JUCE_NAMESPACE

/*** End of inlined file: juce_JSON.cpp ***/




/*** Start of inlined file: juce_Thread.cpp ***/
BEGIN_JUCE_NAMESPACE

class RunningThreadsList
{
public:
	RunningThreadsList()
	{
	}

	~RunningThreadsList()
	{
		// Some threads are still running! Make sure you stop all your
		// threads cleanly before your app quits!
		jassert (threads.size() == 0);
	}

	void add (Thread* const thread)
	{
		const SpinLock::ScopedLockType sl (lock);
		jassert (! threads.contains (thread));
		threads.add (thread);
	}

	void remove (Thread* const thread)
	{
		const SpinLock::ScopedLockType sl (lock);
		jassert (threads.contains (thread));
		threads.removeValue (thread);
	}

	int size() const noexcept
	{
		return threads.size();
	}

	Thread* getThreadWithID (const Thread::ThreadID targetID) const noexcept
	{
		const SpinLock::ScopedLockType sl (lock);

		for (int i = threads.size(); --i >= 0;)
		{
			Thread* const t = threads.getUnchecked(i);

			if (t->getThreadId() == targetID)
				return t;
		}

		return nullptr;
	}

	void stopAll (const int timeOutMilliseconds)
	{
		signalAllThreadsToStop();

		for (;;)
		{
			Thread* firstThread = getFirstThread();

			if (firstThread != nullptr)
				firstThread->stopThread (timeOutMilliseconds);
			else
				break;
		}
	}

	static RunningThreadsList& getInstance()
	{
		static RunningThreadsList runningThreads;
		return runningThreads;
	}

private:
	Array<Thread*> threads;
	SpinLock lock;

	void signalAllThreadsToStop()
	{
		const SpinLock::ScopedLockType sl (lock);

		for (int i = threads.size(); --i >= 0;)
			threads.getUnchecked(i)->signalThreadShouldExit();
	}

	Thread* getFirstThread() const
	{
		const SpinLock::ScopedLockType sl (lock);
		return threads.getFirst();
	}
};

void Thread::threadEntryPoint()
{
	RunningThreadsList::getInstance().add (this);

	JUCE_TRY
	{
		if (threadName_.isNotEmpty())
			setCurrentThreadName (threadName_);

		if (startSuspensionEvent_.wait (10000))
		{
			jassert (getCurrentThreadId() == threadId_);

			if (affinityMask_ != 0)
				setCurrentThreadAffinityMask (affinityMask_);

			run();
		}
	}
	JUCE_CATCH_ALL_ASSERT

	RunningThreadsList::getInstance().remove (this);
	closeThreadHandle();
}

// used to wrap the incoming call from the platform-specific code
void JUCE_API juce_threadEntryPoint (void* userData)
{
	static_cast <Thread*> (userData)->threadEntryPoint();
}

Thread::Thread (const String& threadName)
	: threadName_ (threadName),
	  threadHandle_ (nullptr),
	  threadId_ (0),
	  threadPriority_ (5),
	  affinityMask_ (0),
	  threadShouldExit_ (false)
{
}

Thread::~Thread()
{
	/* If your thread class's destructor has been called without first stopping the thread, that
	   means that this partially destructed object is still performing some work - and that's
	   probably a Bad Thing!

	   To avoid this type of nastiness, always make sure you call stopThread() before or during
	   your subclass's destructor.
	*/
	jassert (! isThreadRunning());

	stopThread (100);
}

void Thread::startThread()
{
	const ScopedLock sl (startStopLock);

	threadShouldExit_ = false;

	if (threadHandle_ == nullptr)
	{
		launchThread();
		setThreadPriority (threadHandle_, threadPriority_);
		startSuspensionEvent_.signal();
	}
}

void Thread::startThread (const int priority)
{
	const ScopedLock sl (startStopLock);

	if (threadHandle_ == nullptr)
	{
		threadPriority_ = priority;
		startThread();
	}
	else
	{
		setPriority (priority);
	}
}

bool Thread::isThreadRunning() const
{
	return threadHandle_ != nullptr;
}

void Thread::signalThreadShouldExit()
{
	threadShouldExit_ = true;
}

bool Thread::waitForThreadToExit (const int timeOutMilliseconds) const
{
	// Doh! So how exactly do you expect this thread to wait for itself to stop??
	jassert (getThreadId() != getCurrentThreadId() || getCurrentThreadId() == 0);

	const int sleepMsPerIteration = 5;
	int count = timeOutMilliseconds / sleepMsPerIteration;

	while (isThreadRunning())
	{
		if (timeOutMilliseconds > 0 && --count < 0)
			return false;

		sleep (sleepMsPerIteration);
	}

	return true;
}

void Thread::stopThread (const int timeOutMilliseconds)
{
	// agh! You can't stop the thread that's calling this method! How on earth
	// would that work??
	jassert (getCurrentThreadId() != getThreadId());

	const ScopedLock sl (startStopLock);

	if (isThreadRunning())
	{
		signalThreadShouldExit();
		notify();

		if (timeOutMilliseconds != 0)
			waitForThreadToExit (timeOutMilliseconds);

		if (isThreadRunning())
		{
			// very bad karma if this point is reached, as there are bound to be
			// locks and events left in silly states when a thread is killed by force..
			jassertfalse;

			killThread();

			RunningThreadsList::getInstance().remove (this);
			threadHandle_ = nullptr;
			threadId_ = 0;
		}
	}
}

bool Thread::setPriority (const int priority)
{
	const ScopedLock sl (startStopLock);

	if (setThreadPriority (threadHandle_, priority))
	{
		threadPriority_ = priority;
		return true;
	}

	return false;
}

bool Thread::setCurrentThreadPriority (const int priority)
{
	return setThreadPriority (0, priority);
}

void Thread::setAffinityMask (const uint32 affinityMask)
{
	affinityMask_ = affinityMask;
}

bool Thread::wait (const int timeOutMilliseconds) const
{
	return defaultEvent_.wait (timeOutMilliseconds);
}

void Thread::notify() const
{
	defaultEvent_.signal();
}

int Thread::getNumRunningThreads()
{
	return RunningThreadsList::getInstance().size();
}

Thread* Thread::getCurrentThread()
{
	return RunningThreadsList::getInstance().getThreadWithID (getCurrentThreadId());
}

void Thread::stopAllThreads (const int timeOutMilliseconds)
{
	RunningThreadsList::getInstance().stopAll (timeOutMilliseconds);
}

void SpinLock::enter() const noexcept
{
	if (! tryEnter())
	{
		for (int i = 20; --i >= 0;)
			if (tryEnter())
				return;

		while (! tryEnter())
			Thread::yield();
	}
}


END_JUCE_NAMESPACE

/*** End of inlined file: juce_Thread.cpp ***/


/*** Start of inlined file: juce_ThreadPool.cpp ***/
BEGIN_JUCE_NAMESPACE

ThreadPoolJob::ThreadPoolJob (const String& name)
	: jobName (name),
	  pool (nullptr),
	  shouldStop (false),
	  isActive (false),
	  shouldBeDeleted (false)
{
}

ThreadPoolJob::~ThreadPoolJob()
{
	// you mustn't delete a job while it's still in a pool! Use ThreadPool::removeJob()
	// to remove it first!
	jassert (pool == nullptr || ! pool->contains (this));
}

String ThreadPoolJob::getJobName() const
{
	return jobName;
}

void ThreadPoolJob::setJobName (const String& newName)
{
	jobName = newName;
}

void ThreadPoolJob::signalJobShouldExit()
{
	shouldStop = true;
}

class ThreadPool::ThreadPoolThread  : public Thread
{
public:
	ThreadPoolThread (ThreadPool& pool_)
		: Thread ("Pool"),
		  pool (pool_),
		  busy (false)
	{
	}

	void run()
	{
		while (! threadShouldExit())
		{
			if (! pool.runNextJob())
				wait (500);
		}
	}

private:
	ThreadPool& pool;
	bool volatile busy;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ThreadPoolThread);
};

ThreadPool::ThreadPool (const int numThreads,
						const bool startThreadsOnlyWhenNeeded,
						const int stopThreadsWhenNotUsedTimeoutMs)
	: threadStopTimeout (stopThreadsWhenNotUsedTimeoutMs),
	  priority (5)
{
	jassert (numThreads > 0); // not much point having one of these with no threads in it.

	for (int i = jmax (1, numThreads); --i >= 0;)
		threads.add (new ThreadPoolThread (*this));

	if (! startThreadsOnlyWhenNeeded)
		for (int i = threads.size(); --i >= 0;)
			threads.getUnchecked(i)->startThread (priority);
}

ThreadPool::~ThreadPool()
{
	removeAllJobs (true, 4000);

	int i;
	for (i = threads.size(); --i >= 0;)
		threads.getUnchecked(i)->signalThreadShouldExit();

	for (i = threads.size(); --i >= 0;)
		threads.getUnchecked(i)->stopThread (500);
}

void ThreadPool::addJob (ThreadPoolJob* const job)
{
	jassert (job != nullptr);
	jassert (job->pool == nullptr);

	if (job->pool == nullptr)
	{
		job->pool = this;
		job->shouldStop = false;
		job->isActive = false;

		{
			const ScopedLock sl (lock);
			jobs.add (job);

			int numRunning = 0;

			for (int i = threads.size(); --i >= 0;)
				if (threads.getUnchecked(i)->isThreadRunning() && ! threads.getUnchecked(i)->threadShouldExit())
					++numRunning;

			if (numRunning < threads.size())
			{
				bool startedOne = false;
				int n = 1000;

				while (--n >= 0 && ! startedOne)
				{
					for (int i = threads.size(); --i >= 0;)
					{
						if (! threads.getUnchecked(i)->isThreadRunning())
						{
							threads.getUnchecked(i)->startThread (priority);
							startedOne = true;
							break;
						}
					}

					if (! startedOne)
						Thread::sleep (2);
				}
			}
		}

		for (int i = threads.size(); --i >= 0;)
			threads.getUnchecked(i)->notify();
	}
}

int ThreadPool::getNumJobs() const
{
	return jobs.size();
}

ThreadPoolJob* ThreadPool::getJob (const int index) const
{
	const ScopedLock sl (lock);
	return jobs [index];
}

bool ThreadPool::contains (const ThreadPoolJob* const job) const
{
	const ScopedLock sl (lock);
	return jobs.contains (const_cast <ThreadPoolJob*> (job));
}

bool ThreadPool::isJobRunning (const ThreadPoolJob* const job) const
{
	const ScopedLock sl (lock);
	return jobs.contains (const_cast <ThreadPoolJob*> (job)) && job->isActive;
}

bool ThreadPool::waitForJobToFinish (const ThreadPoolJob* const job,
									 const int timeOutMs) const
{
	if (job != nullptr)
	{
		const uint32 start = Time::getMillisecondCounter();

		while (contains (job))
		{
			if (timeOutMs >= 0 && Time::getMillisecondCounter() >= start + timeOutMs)
				return false;

			jobFinishedSignal.wait (2);
		}
	}

	return true;
}

bool ThreadPool::removeJob (ThreadPoolJob* const job,
							const bool interruptIfRunning,
							const int timeOutMs)
{
	bool dontWait = true;

	if (job != nullptr)
	{
		const ScopedLock sl (lock);

		if (jobs.contains (job))
		{
			if (job->isActive)
			{
				if (interruptIfRunning)
					job->signalJobShouldExit();

				dontWait = false;
			}
			else
			{
				jobs.removeValue (job);
				job->pool = nullptr;
			}
		}
	}

	return dontWait || waitForJobToFinish (job, timeOutMs);
}

bool ThreadPool::removeAllJobs (const bool interruptRunningJobs,
								const int timeOutMs,
								const bool deleteInactiveJobs,
								ThreadPool::JobSelector* selectedJobsToRemove)
{
	Array <ThreadPoolJob*> jobsToWaitFor;

	{
		const ScopedLock sl (lock);

		for (int i = jobs.size(); --i >= 0;)
		{
			ThreadPoolJob* const job = jobs.getUnchecked(i);

			if (selectedJobsToRemove == nullptr || selectedJobsToRemove->isJobSuitable (job))
			{
				if (job->isActive)
				{
					jobsToWaitFor.add (job);

					if (interruptRunningJobs)
						job->signalJobShouldExit();
				}
				else
				{
					jobs.remove (i);

					if (deleteInactiveJobs)
						delete job;
					else
						job->pool = nullptr;
				}
			}
		}
	}

	const uint32 start = Time::getMillisecondCounter();

	for (;;)
	{
		for (int i = jobsToWaitFor.size(); --i >= 0;)
			if (! isJobRunning (jobsToWaitFor.getUnchecked (i)))
				jobsToWaitFor.remove (i);

		if (jobsToWaitFor.size() == 0)
			break;

		if (timeOutMs >= 0 && Time::getMillisecondCounter() >= start + timeOutMs)
			return false;

		jobFinishedSignal.wait (20);
	}

	return true;
}

StringArray ThreadPool::getNamesOfAllJobs (const bool onlyReturnActiveJobs) const
{
	StringArray s;
	const ScopedLock sl (lock);

	for (int i = 0; i < jobs.size(); ++i)
	{
		const ThreadPoolJob* const job = jobs.getUnchecked(i);
		if (job->isActive || ! onlyReturnActiveJobs)
			s.add (job->getJobName());
	}

	return s;
}

bool ThreadPool::setThreadPriorities (const int newPriority)
{
	bool ok = true;

	if (priority != newPriority)
	{
		priority = newPriority;

		for (int i = threads.size(); --i >= 0;)
			if (! threads.getUnchecked(i)->setPriority (newPriority))
				ok = false;
	}

	return ok;
}

bool ThreadPool::runNextJob()
{
	ThreadPoolJob* job = nullptr;

	{
		const ScopedLock sl (lock);

		for (int i = 0; i < jobs.size(); ++i)
		{
			job = jobs[i];

			if (job != nullptr && ! (job->isActive || job->shouldStop))
				break;

			job = nullptr;
		}

		if (job != nullptr)
			job->isActive = true;

	}

	if (job != nullptr)
	{
		JUCE_TRY
		{
			ThreadPoolJob::JobStatus result = job->runJob();

			lastJobEndTime = Time::getApproximateMillisecondCounter();

			const ScopedLock sl (lock);

			if (jobs.contains (job))
			{
				job->isActive = false;

				if (result != ThreadPoolJob::jobNeedsRunningAgain || job->shouldStop)
				{
					job->pool = nullptr;
					job->shouldStop = true;
					jobs.removeValue (job);

					if (result == ThreadPoolJob::jobHasFinishedAndShouldBeDeleted)
						delete job;

					jobFinishedSignal.signal();
				}
				else
				{
					// move the job to the end of the queue if it wants another go
					jobs.move (jobs.indexOf (job), -1);
				}
			}
		}
#if JUCE_CATCH_UNHANDLED_EXCEPTIONS
		catch (...)
		{
			const ScopedLock sl (lock);
			jobs.removeValue (job);
		}
#endif
	}
	else
	{
		if (threadStopTimeout > 0
			 && Time::getApproximateMillisecondCounter() > lastJobEndTime + threadStopTimeout)
		{
			const ScopedLock sl (lock);

			if (jobs.size() == 0)
				for (int i = threads.size(); --i >= 0;)
					threads.getUnchecked(i)->signalThreadShouldExit();
		}
		else
		{
			return false;
		}
	}

	return true;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ThreadPool.cpp ***/


/*** Start of inlined file: juce_TimeSliceThread.cpp ***/
BEGIN_JUCE_NAMESPACE

TimeSliceThread::TimeSliceThread (const String& threadName)
	: Thread (threadName),
	  clientBeingCalled (nullptr)
{
}

TimeSliceThread::~TimeSliceThread()
{
	stopThread (2000);
}

void TimeSliceThread::addTimeSliceClient (TimeSliceClient* const client, int millisecondsBeforeStarting)
{
	if (client != nullptr)
	{
		const ScopedLock sl (listLock);
		client->nextCallTime = Time::getCurrentTime() + RelativeTime::milliseconds (millisecondsBeforeStarting);
		clients.addIfNotAlreadyThere (client);
		notify();
	}
}

void TimeSliceThread::removeTimeSliceClient (TimeSliceClient* const client)
{
	const ScopedLock sl1 (listLock);

	// if there's a chance we're in the middle of calling this client, we need to
	// also lock the outer lock..
	if (clientBeingCalled == client)
	{
		const ScopedUnlock ul (listLock); // unlock first to get the order right..

		const ScopedLock sl2 (callbackLock);
		const ScopedLock sl3 (listLock);

		clients.removeValue (client);
	}
	else
	{
		clients.removeValue (client);
	}
}

int TimeSliceThread::getNumClients() const
{
	return clients.size();
}

TimeSliceClient* TimeSliceThread::getClient (const int i) const
{
	const ScopedLock sl (listLock);
	return clients [i];
}

TimeSliceClient* TimeSliceThread::getNextClient (int index) const
{
	Time soonest;
	TimeSliceClient* client = nullptr;

	for (int i = clients.size(); --i >= 0;)
	{
		TimeSliceClient* const c = clients.getUnchecked ((i + index) % clients.size());

		if (client == nullptr || c->nextCallTime < soonest)
		{
			client = c;
			soonest = c->nextCallTime;
		}
	}

	return client;
}

void TimeSliceThread::run()
{
	int index = 0;

	while (! threadShouldExit())
	{
		int timeToWait = 500;

		{
			Time nextClientTime;

			{
				const ScopedLock sl2 (listLock);

				index = clients.size() > 0 ? ((index + 1) % clients.size()) : 0;

				TimeSliceClient* const firstClient = getNextClient (index);
				if (firstClient != nullptr)
					nextClientTime = firstClient->nextCallTime;
			}

			const Time now (Time::getCurrentTime());

			if (nextClientTime > now)
			{
				timeToWait = (int) jmin ((int64) 500, (nextClientTime - now).inMilliseconds());
			}
			else
			{
				timeToWait = index == 0 ? 1 : 0;

				const ScopedLock sl (callbackLock);

				{
					const ScopedLock sl2 (listLock);
					clientBeingCalled = getNextClient (index);
				}

				if (clientBeingCalled != nullptr)
				{
					const int msUntilNextCall = clientBeingCalled->useTimeSlice();

					const ScopedLock sl2 (listLock);

					if (msUntilNextCall >= 0)
						clientBeingCalled->nextCallTime += RelativeTime::milliseconds (msUntilNextCall);
					else
						clients.removeValue (clientBeingCalled);

					clientBeingCalled = nullptr;
				}
			}
		}

		if (timeToWait > 0)
			wait (timeToWait);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TimeSliceThread.cpp ***/

#endif

#if JUCE_BUILD_MISC

/*** Start of inlined file: juce_ValueTree.cpp ***/
BEGIN_JUCE_NAMESPACE

class ValueTree::SetPropertyAction  : public UndoableAction
{
public:
	SetPropertyAction (const SharedObjectPtr& target_, const Identifier& name_,
					   const var& newValue_, const var& oldValue_,
					   const bool isAddingNewProperty_, const bool isDeletingProperty_)
		: target (target_), name (name_), newValue (newValue_), oldValue (oldValue_),
		  isAddingNewProperty (isAddingNewProperty_), isDeletingProperty (isDeletingProperty_)
	{
	}

	bool perform()
	{
		jassert (! (isAddingNewProperty && target->hasProperty (name)));

		if (isDeletingProperty)
			target->removeProperty (name, nullptr);
		else
			target->setProperty (name, newValue, nullptr);

		return true;
	}

	bool undo()
	{
		if (isAddingNewProperty)
			target->removeProperty (name, nullptr);
		else
			target->setProperty (name, oldValue, nullptr);

		return true;
	}

	int getSizeInUnits()
	{
		return (int) sizeof (*this); //xxx should be more accurate
	}

	UndoableAction* createCoalescedAction (UndoableAction* nextAction)
	{
		if (! (isAddingNewProperty || isDeletingProperty))
		{
			SetPropertyAction* next = dynamic_cast <SetPropertyAction*> (nextAction);

			if (next != nullptr && next->target == target && next->name == name
				 && ! (next->isAddingNewProperty || next->isDeletingProperty))
			{
				return new SetPropertyAction (target, name, next->newValue, oldValue, false, false);
			}
		}

		return nullptr;
	}

private:
	const SharedObjectPtr target;
	const Identifier name;
	const var newValue;
	var oldValue;
	const bool isAddingNewProperty : 1, isDeletingProperty : 1;

	JUCE_DECLARE_NON_COPYABLE (SetPropertyAction);
};

class ValueTree::AddOrRemoveChildAction  : public UndoableAction
{
public:
	AddOrRemoveChildAction (const SharedObjectPtr& target_, const int childIndex_,
							const SharedObjectPtr& newChild_)
		: target (target_),
		  child (newChild_ != nullptr ? newChild_ : target_->children [childIndex_]),
		  childIndex (childIndex_),
		  isDeleting (newChild_ == nullptr)
	{
		jassert (child != nullptr);
	}

	bool perform()
	{
		if (isDeleting)
			target->removeChild (childIndex, nullptr);
		else
			target->addChild (child, childIndex, nullptr);

		return true;
	}

	bool undo()
	{
		if (isDeleting)
		{
			target->addChild (child, childIndex, nullptr);
		}
		else
		{
			// If you hit this, it seems that your object's state is getting confused - probably
			// because you've interleaved some undoable and non-undoable operations?
			jassert (childIndex < target->children.size());
			target->removeChild (childIndex, nullptr);
		}

		return true;
	}

	int getSizeInUnits()
	{
		return (int) sizeof (*this); //xxx should be more accurate
	}

private:
	const SharedObjectPtr target, child;
	const int childIndex;
	const bool isDeleting;

	JUCE_DECLARE_NON_COPYABLE (AddOrRemoveChildAction);
};

class ValueTree::MoveChildAction  : public UndoableAction
{
public:
	MoveChildAction (const SharedObjectPtr& parent_,
					 const int startIndex_, const int endIndex_)
		: parent (parent_),
		  startIndex (startIndex_),
		  endIndex (endIndex_)
	{
	}

	bool perform()
	{
		parent->moveChild (startIndex, endIndex, nullptr);
		return true;
	}

	bool undo()
	{
		parent->moveChild (endIndex, startIndex, nullptr);
		return true;
	}

	int getSizeInUnits()
	{
		return (int) sizeof (*this); //xxx should be more accurate
	}

	UndoableAction* createCoalescedAction (UndoableAction* nextAction)
	{
		MoveChildAction* next = dynamic_cast <MoveChildAction*> (nextAction);

		if (next != nullptr && next->parent == parent && next->startIndex == endIndex)
			return new MoveChildAction (parent, startIndex, next->endIndex);

		return nullptr;
	}

private:
	const SharedObjectPtr parent;
	const int startIndex, endIndex;

	JUCE_DECLARE_NON_COPYABLE (MoveChildAction);
};

ValueTree::SharedObject::SharedObject (const Identifier& type_)
	: type (type_), parent (nullptr)
{
}

ValueTree::SharedObject::SharedObject (const SharedObject& other)
	: type (other.type), properties (other.properties), parent (nullptr)
{
	for (int i = 0; i < other.children.size(); ++i)
	{
		SharedObject* const child = new SharedObject (*other.children.getUnchecked(i));
		child->parent = this;
		children.add (child);
	}
}

ValueTree::SharedObject::~SharedObject()
{
	jassert (parent == nullptr); // this should never happen unless something isn't obeying the ref-counting!

	for (int i = children.size(); --i >= 0;)
	{
		const SharedObjectPtr c (children.getUnchecked(i));
		c->parent = nullptr;
		children.remove (i);
		c->sendParentChangeMessage();
	}
}

void ValueTree::SharedObject::sendPropertyChangeMessage (ValueTree& tree, const Identifier& property)
{
	for (int i = valueTreesWithListeners.size(); --i >= 0;)
	{
		ValueTree* const v = valueTreesWithListeners[i];
		if (v != nullptr)
			v->listeners.call (&ValueTree::Listener::valueTreePropertyChanged, tree, property);
	}
}

void ValueTree::SharedObject::sendPropertyChangeMessage (const Identifier& property)
{
	ValueTree tree (this);

	for (ValueTree::SharedObject* t = this; t != nullptr; t = t->parent)
		t->sendPropertyChangeMessage (tree, property);
}

void ValueTree::SharedObject::sendChildAddedMessage (ValueTree& tree, ValueTree& child)
{
	for (int i = valueTreesWithListeners.size(); --i >= 0;)
	{
		ValueTree* const v = valueTreesWithListeners[i];
		if (v != nullptr)
			v->listeners.call (&ValueTree::Listener::valueTreeChildAdded, tree, child);
	}
}

void ValueTree::SharedObject::sendChildAddedMessage (ValueTree child)
{
	ValueTree tree (this);

	for (ValueTree::SharedObject* t = this; t != nullptr; t = t->parent)
		t->sendChildAddedMessage (tree, child);
}

void ValueTree::SharedObject::sendChildRemovedMessage (ValueTree& tree, ValueTree& child)
{
	for (int i = valueTreesWithListeners.size(); --i >= 0;)
	{
		ValueTree* const v = valueTreesWithListeners[i];
		if (v != nullptr)
			v->listeners.call (&ValueTree::Listener::valueTreeChildRemoved, tree, child);
	}
}

void ValueTree::SharedObject::sendChildRemovedMessage (ValueTree child)
{
	ValueTree tree (this);

	for (ValueTree::SharedObject* t = this; t != nullptr; t = t->parent)
		t->sendChildRemovedMessage (tree, child);
}

void ValueTree::SharedObject::sendChildOrderChangedMessage (ValueTree& tree)
{
	for (int i = valueTreesWithListeners.size(); --i >= 0;)
	{
		ValueTree* const v = valueTreesWithListeners[i];
		if (v != nullptr)
			v->listeners.call (&ValueTree::Listener::valueTreeChildOrderChanged, tree);
	}
}

void ValueTree::SharedObject::sendChildOrderChangedMessage()
{
	ValueTree tree (this);

	for (ValueTree::SharedObject* t = this; t != nullptr; t = t->parent)
		t->sendChildOrderChangedMessage (tree);
}

void ValueTree::SharedObject::sendParentChangeMessage()
{
	ValueTree tree (this);

	int i;
	for (i = children.size(); --i >= 0;)
	{
		SharedObject* const t = children[i];
		if (t != nullptr)
			t->sendParentChangeMessage();
	}

	for (i = valueTreesWithListeners.size(); --i >= 0;)
	{
		ValueTree* const v = valueTreesWithListeners[i];
		if (v != nullptr)
			v->listeners.call (&ValueTree::Listener::valueTreeParentChanged, tree);
	}
}

const var& ValueTree::SharedObject::getProperty (const Identifier& name) const
{
	return properties [name];
}

var ValueTree::SharedObject::getProperty (const Identifier& name, const var& defaultReturnValue) const
{
	return properties.getWithDefault (name, defaultReturnValue);
}

void ValueTree::SharedObject::setProperty (const Identifier& name, const var& newValue, UndoManager* const undoManager)
{
	if (undoManager == nullptr)
	{
		if (properties.set (name, newValue))
			sendPropertyChangeMessage (name);
	}
	else
	{
		const var* const existingValue = properties.getVarPointer (name);

		if (existingValue != nullptr)
		{
			if (*existingValue != newValue)
				undoManager->perform (new SetPropertyAction (this, name, newValue, *existingValue, false, false));
		}
		else
		{
			undoManager->perform (new SetPropertyAction (this, name, newValue, var::null, true, false));
		}
	}
}

bool ValueTree::SharedObject::hasProperty (const Identifier& name) const
{
	return properties.contains (name);
}

void ValueTree::SharedObject::removeProperty (const Identifier& name, UndoManager* const undoManager)
{
	if (undoManager == nullptr)
	{
		if (properties.remove (name))
			sendPropertyChangeMessage (name);
	}
	else
	{
		if (properties.contains (name))
			undoManager->perform (new SetPropertyAction (this, name, var::null, properties [name], false, true));
	}
}

void ValueTree::SharedObject::removeAllProperties (UndoManager* const undoManager)
{
	if (undoManager == nullptr)
	{
		while (properties.size() > 0)
		{
			const Identifier name (properties.getName (properties.size() - 1));
			properties.remove (name);
			sendPropertyChangeMessage (name);
		}
	}
	else
	{
		for (int i = properties.size(); --i >= 0;)
			undoManager->perform (new SetPropertyAction (this, properties.getName(i), var::null, properties.getValueAt(i), false, true));
	}
}

ValueTree ValueTree::SharedObject::getChildWithName (const Identifier& typeToMatch) const
{
	for (int i = 0; i < children.size(); ++i)
		if (children.getUnchecked(i)->type == typeToMatch)
			return ValueTree (children.getUnchecked(i).getObject());

	return ValueTree::invalid;
}

ValueTree ValueTree::SharedObject::getOrCreateChildWithName (const Identifier& typeToMatch, UndoManager* undoManager)
{
	for (int i = 0; i < children.size(); ++i)
		if (children.getUnchecked(i)->type == typeToMatch)
			return ValueTree (children.getUnchecked(i).getObject());

	SharedObject* const newObject = new SharedObject (typeToMatch);
	addChild (newObject, -1, undoManager);
	return ValueTree (newObject);

}

ValueTree ValueTree::SharedObject::getChildWithProperty (const Identifier& propertyName, const var& propertyValue) const
{
	for (int i = 0; i < children.size(); ++i)
		if (children.getUnchecked(i)->getProperty (propertyName) == propertyValue)
			return ValueTree (children.getUnchecked(i).getObject());

	return ValueTree::invalid;
}

bool ValueTree::SharedObject::isAChildOf (const SharedObject* const possibleParent) const
{
	const SharedObject* p = parent;

	while (p != nullptr)
	{
		if (p == possibleParent)
			return true;

		p = p->parent;
	}

	return false;
}

int ValueTree::SharedObject::indexOf (const ValueTree& child) const
{
	return children.indexOf (child.object);
}

void ValueTree::SharedObject::addChild (SharedObject* child, int index, UndoManager* const undoManager)
{
	if (child != nullptr && child->parent != this)
	{
		if (child != this && ! isAChildOf (child))
		{
			// You should always make sure that a child is removed from its previous parent before
			// adding it somewhere else - otherwise, it's ambiguous as to whether a different
			// undomanager should be used when removing it from its current parent..
			jassert (child->parent == nullptr);

			if (child->parent != nullptr)
			{
				jassert (child->parent->children.indexOf (child) >= 0);
				child->parent->removeChild (child->parent->children.indexOf (child), undoManager);
			}

			if (undoManager == nullptr)
			{
				children.insert (index, child);
				child->parent = this;
				sendChildAddedMessage (ValueTree (child));
				child->sendParentChangeMessage();
			}
			else
			{
				if (index < 0)
					index = children.size();

				undoManager->perform (new AddOrRemoveChildAction (this, index, child));
			}
		}
		else
		{
			// You're attempting to create a recursive loop! A node
			// can't be a child of one of its own children!
			jassertfalse;
		}
	}
}

void ValueTree::SharedObject::removeChild (const int childIndex, UndoManager* const undoManager)
{
	const SharedObjectPtr child (children [childIndex]);

	if (child != nullptr)
	{
		if (undoManager == nullptr)
		{
			children.remove (childIndex);
			child->parent = nullptr;
			sendChildRemovedMessage (ValueTree (child));
			child->sendParentChangeMessage();
		}
		else
		{
			undoManager->perform (new AddOrRemoveChildAction (this, childIndex, nullptr));
		}
	}
}

void ValueTree::SharedObject::removeAllChildren (UndoManager* const undoManager)
{
	while (children.size() > 0)
		removeChild (children.size() - 1, undoManager);
}

void ValueTree::SharedObject::moveChild (int currentIndex, int newIndex, UndoManager* undoManager)
{
	// The source index must be a valid index!
	jassert (isPositiveAndBelow (currentIndex, children.size()));

	if (currentIndex != newIndex
		 && isPositiveAndBelow (currentIndex, children.size()))
	{
		if (undoManager == nullptr)
		{
			children.move (currentIndex, newIndex);
			sendChildOrderChangedMessage();
		}
		else
		{
			if (! isPositiveAndBelow (newIndex, children.size()))
				newIndex = children.size() - 1;

			undoManager->perform (new MoveChildAction (this, currentIndex, newIndex));
		}
	}
}

void ValueTree::SharedObject::reorderChildren (const ReferenceCountedArray <SharedObject>& newOrder, UndoManager* undoManager)
{
	jassert (newOrder.size() == children.size());

	if (undoManager == nullptr)
	{
		children = newOrder;
		sendChildOrderChangedMessage();
	}
	else
	{
		for (int i = 0; i < children.size(); ++i)
		{
			const SharedObjectPtr child (newOrder.getUnchecked(i));

			if (children.getUnchecked(i) != child)
			{
				const int oldIndex = children.indexOf (child);
				jassert (oldIndex >= 0);
				moveChild (oldIndex, i, undoManager);
			}
		}
	}
}

bool ValueTree::SharedObject::isEquivalentTo (const SharedObject& other) const
{
	if (type != other.type
		 || properties.size() != other.properties.size()
		 || children.size() != other.children.size()
		 || properties != other.properties)
		return false;

	for (int i = 0; i < children.size(); ++i)
		if (! children.getUnchecked(i)->isEquivalentTo (*other.children.getUnchecked(i)))
			return false;

	return true;
}

ValueTree::ValueTree() noexcept
{
}

const ValueTree ValueTree::invalid;

ValueTree::ValueTree (const Identifier& type_)
	: object (new ValueTree::SharedObject (type_))
{
	jassert (type_.toString().isNotEmpty()); // All objects should be given a sensible type name!
}

ValueTree::ValueTree (SharedObject* const object_)
	: object (object_)
{
}

ValueTree::ValueTree (const ValueTree& other)
	: object (other.object)
{
}

ValueTree& ValueTree::operator= (const ValueTree& other)
{
	if (listeners.size() > 0)
	{
		if (object != nullptr)
			object->valueTreesWithListeners.removeValue (this);

		if (other.object != nullptr)
			other.object->valueTreesWithListeners.add (this);
	}

	object = other.object;

	return *this;
}

ValueTree::~ValueTree()
{
	if (listeners.size() > 0 && object != nullptr)
		object->valueTreesWithListeners.removeValue (this);
}

bool ValueTree::operator== (const ValueTree& other) const noexcept
{
	return object == other.object;
}

bool ValueTree::operator!= (const ValueTree& other) const noexcept
{
	return object != other.object;
}

bool ValueTree::isEquivalentTo (const ValueTree& other) const
{
	return object == other.object
			|| (object != nullptr && other.object != nullptr && object->isEquivalentTo (*other.object));
}

ValueTree ValueTree::createCopy() const
{
	return ValueTree (object != nullptr ? new SharedObject (*object) : nullptr);
}

bool ValueTree::hasType (const Identifier& typeName) const
{
	return object != nullptr && object->type == typeName;
}

Identifier ValueTree::getType() const
{
	return object != nullptr ? object->type : Identifier();
}

ValueTree ValueTree::getParent() const
{
	return ValueTree (object != nullptr ? object->parent : (SharedObject*) nullptr);
}

ValueTree ValueTree::getSibling (const int delta) const
{
	if (object == nullptr || object->parent == nullptr)
		return invalid;

	const int index = object->parent->indexOf (*this) + delta;
	return ValueTree (object->parent->children [index].getObject());
}

const var& ValueTree::operator[] (const Identifier& name) const
{
	return object == nullptr ? var::null : object->getProperty (name);
}

const var& ValueTree::getProperty (const Identifier& name) const
{
	return object == nullptr ? var::null : object->getProperty (name);
}

var ValueTree::getProperty (const Identifier& name, const var& defaultReturnValue) const
{
	return object == nullptr ? defaultReturnValue : object->getProperty (name, defaultReturnValue);
}

void ValueTree::setProperty (const Identifier& name, const var& newValue, UndoManager* const undoManager)
{
	jassert (name.toString().isNotEmpty());

	if (object != nullptr && name.toString().isNotEmpty())
		object->setProperty (name, newValue, undoManager);
}

bool ValueTree::hasProperty (const Identifier& name) const
{
	return object != nullptr && object->hasProperty (name);
}

void ValueTree::removeProperty (const Identifier& name, UndoManager* const undoManager)
{
	if (object != nullptr)
		object->removeProperty (name, undoManager);
}

void ValueTree::removeAllProperties (UndoManager* const undoManager)
{
	if (object != nullptr)
		object->removeAllProperties (undoManager);
}

int ValueTree::getNumProperties() const
{
	return object == nullptr ? 0 : object->properties.size();
}

Identifier ValueTree::getPropertyName (const int index) const
{
	return object == nullptr ? Identifier()
							 : object->properties.getName (index);
}

class ValueTreePropertyValueSource  : public Value::ValueSource,
									  public ValueTree::Listener
{
public:
	ValueTreePropertyValueSource (const ValueTree& tree_,
								  const Identifier& property_,
								  UndoManager* const undoManager_)
		: tree (tree_),
		  property (property_),
		  undoManager (undoManager_)
	{
		tree.addListener (this);
	}

	~ValueTreePropertyValueSource()
	{
		tree.removeListener (this);
	}

	var getValue() const
	{
		return tree [property];
	}

	void setValue (const var& newValue)
	{
		tree.setProperty (property, newValue, undoManager);
	}

	void valueTreePropertyChanged (ValueTree& treeWhosePropertyHasChanged, const Identifier& changedProperty)
	{
		if (tree == treeWhosePropertyHasChanged && property == changedProperty)
			sendChangeMessage (false);
	}

	void valueTreeChildAdded (ValueTree&, ValueTree&) {}
	void valueTreeChildRemoved (ValueTree&, ValueTree&) {}
	void valueTreeChildOrderChanged (ValueTree&) {}
	void valueTreeParentChanged (ValueTree&) {}

private:
	ValueTree tree;
	const Identifier property;
	UndoManager* const undoManager;

	ValueTreePropertyValueSource& operator= (const ValueTreePropertyValueSource&);
};

Value ValueTree::getPropertyAsValue (const Identifier& name, UndoManager* const undoManager) const
{
	return Value (new ValueTreePropertyValueSource (*this, name, undoManager));
}

int ValueTree::getNumChildren() const
{
	return object == nullptr ? 0 : object->children.size();
}

ValueTree ValueTree::getChild (int index) const
{
	return ValueTree (object != nullptr ? (SharedObject*) object->children [index] : (SharedObject*) nullptr);
}

ValueTree ValueTree::getChildWithName (const Identifier& type) const
{
	return object != nullptr ? object->getChildWithName (type) : ValueTree::invalid;
}

ValueTree ValueTree::getOrCreateChildWithName (const Identifier& type, UndoManager* undoManager)
{
	return object != nullptr ? object->getOrCreateChildWithName (type, undoManager) : ValueTree::invalid;
}

ValueTree ValueTree::getChildWithProperty (const Identifier& propertyName, const var& propertyValue) const
{
	return object != nullptr ? object->getChildWithProperty (propertyName, propertyValue) : ValueTree::invalid;
}

bool ValueTree::isAChildOf (const ValueTree& possibleParent) const
{
	return object != nullptr && object->isAChildOf (possibleParent.object);
}

int ValueTree::indexOf (const ValueTree& child) const
{
	return object != nullptr ? object->indexOf (child) : -1;
}

void ValueTree::addChild (const ValueTree& child, int index, UndoManager* const undoManager)
{
	if (object != nullptr)
		object->addChild (child.object, index, undoManager);
}

void ValueTree::removeChild (const int childIndex, UndoManager* const undoManager)
{
	if (object != nullptr)
		object->removeChild (childIndex, undoManager);
}

void ValueTree::removeChild (const ValueTree& child, UndoManager* const undoManager)
{
	if (object != nullptr)
		object->removeChild (object->children.indexOf (child.object), undoManager);
}

void ValueTree::removeAllChildren (UndoManager* const undoManager)
{
	if (object != nullptr)
		object->removeAllChildren (undoManager);
}

void ValueTree::moveChild (int currentIndex, int newIndex, UndoManager* undoManager)
{
	if (object != nullptr)
		object->moveChild (currentIndex, newIndex, undoManager);
}

void ValueTree::addListener (Listener* listener)
{
	if (listener != nullptr)
	{
		if (listeners.size() == 0 && object != nullptr)
			object->valueTreesWithListeners.add (this);

		listeners.add (listener);
	}
}

void ValueTree::removeListener (Listener* listener)
{
	listeners.remove (listener);

	if (listeners.size() == 0 && object != nullptr)
		object->valueTreesWithListeners.removeValue (this);
}

XmlElement* ValueTree::SharedObject::createXml() const
{
	XmlElement* const xml = new XmlElement (type.toString());
	properties.copyToXmlAttributes (*xml);

	for (int i = 0; i < children.size(); ++i)
		xml->addChildElement (children.getUnchecked(i)->createXml());

	return xml;
}

XmlElement* ValueTree::createXml() const
{
	return object != nullptr ? object->createXml() : nullptr;
}

ValueTree ValueTree::fromXml (const XmlElement& xml)
{
	ValueTree v (xml.getTagName());
	v.object->properties.setFromXmlAttributes (xml);

	forEachXmlChildElement (xml, e)
		v.addChild (fromXml (*e), -1, nullptr);

	return v;
}

void ValueTree::writeToStream (OutputStream& output)
{
	output.writeString (getType().toString());

	const int numProps = getNumProperties();
	output.writeCompressedInt (numProps);

	int i;
	for (i = 0; i < numProps; ++i)
	{
		const Identifier name (getPropertyName(i));
		output.writeString (name.toString());
		getProperty(name).writeToStream (output);
	}

	const int numChildren = getNumChildren();
	output.writeCompressedInt (numChildren);

	for (i = 0; i < numChildren; ++i)
		getChild (i).writeToStream (output);
}

ValueTree ValueTree::readFromStream (InputStream& input)
{
	const String type (input.readString());

	if (type.isEmpty())
		return ValueTree::invalid;

	ValueTree v (type);

	const int numProps = input.readCompressedInt();

	if (numProps < 0)
	{
		jassertfalse;  // trying to read corrupted data!
		return v;
	}

	int i;
	for (i = 0; i < numProps; ++i)
	{
		const String name (input.readString());
		jassert (name.isNotEmpty());
		const var value (var::readFromStream (input));
		v.object->properties.set (name, value);
	}

	const int numChildren = input.readCompressedInt();

	for (i = 0; i < numChildren; ++i)
	{
		ValueTree child (readFromStream (input));

		v.object->children.add (child.object);
		child.object->parent = v.object;
	}

	return v;
}

ValueTree ValueTree::readFromData (const void* const data, const size_t numBytes)
{
	MemoryInputStream in (data, numBytes, false);
	return readFromStream (in);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ValueTree.cpp ***/



/*** Start of inlined file: juce_Value.cpp ***/
BEGIN_JUCE_NAMESPACE

Value::ValueSource::ValueSource()
{
}

Value::ValueSource::~ValueSource()
{
}

void Value::ValueSource::sendChangeMessage (const bool synchronous)
{
	if (synchronous)
	{
		// (hold a local reference to this object in case it's freed during the callbacks)
		const ReferenceCountedObjectPtr<ValueSource> localRef (this);

		for (int i = valuesWithListeners.size(); --i >= 0;)
		{
			Value* const v = valuesWithListeners[i];

			if (v != nullptr)
				v->callListeners();
		}
	}
	else
	{
		triggerAsyncUpdate();
	}
}

void Value::ValueSource::handleAsyncUpdate()
{
	sendChangeMessage (true);
}

class SimpleValueSource  : public Value::ValueSource
{
public:
	SimpleValueSource()
	{
	}

	SimpleValueSource (const var& initialValue)
		: value (initialValue)
	{
	}

	var getValue() const
	{
		return value;
	}

	void setValue (const var& newValue)
	{
		if (! newValue.equalsWithSameType (value))
		{
			value = newValue;
			sendChangeMessage (false);
		}
	}

private:
	var value;

	JUCE_DECLARE_NON_COPYABLE (SimpleValueSource);
};

Value::Value()
	: value (new SimpleValueSource())
{
}

Value::Value (ValueSource* const value_)
	: value (value_)
{
	jassert (value_ != nullptr);
}

Value::Value (const var& initialValue)
	: value (new SimpleValueSource (initialValue))
{
}

Value::Value (const Value& other)
	: value (other.value)
{
}

Value& Value::operator= (const Value& other)
{
	value = other.value;
	return *this;
}

Value::~Value()
{
	if (listeners.size() > 0)
		value->valuesWithListeners.removeValue (this);
}

var Value::getValue() const
{
	return value->getValue();
}

Value::operator var() const
{
	return getValue();
}

void Value::setValue (const var& newValue)
{
	value->setValue (newValue);
}

String Value::toString() const
{
	return value->getValue().toString();
}

Value& Value::operator= (const var& newValue)
{
	value->setValue (newValue);
	return *this;
}

void Value::referTo (const Value& valueToReferTo)
{
	if (valueToReferTo.value != value)
	{
		if (listeners.size() > 0)
		{
			value->valuesWithListeners.removeValue (this);
			valueToReferTo.value->valuesWithListeners.add (this);
		}

		value = valueToReferTo.value;
		callListeners();
	}
}

bool Value::refersToSameSourceAs (const Value& other) const
{
	return value == other.value;
}

bool Value::operator== (const Value& other) const
{
	return value == other.value || value->getValue() == other.getValue();
}

bool Value::operator!= (const Value& other) const
{
	return value != other.value && value->getValue() != other.getValue();
}

void Value::addListener (ValueListener* const listener)
{
	if (listener != nullptr)
	{
		if (listeners.size() == 0)
			value->valuesWithListeners.add (this);

		listeners.add (listener);
	}
}

void Value::removeListener (ValueListener* const listener)
{
	listeners.remove (listener);

	if (listeners.size() == 0)
		value->valuesWithListeners.removeValue (this);
}

void Value::callListeners()
{
	Value v (*this); // (create a copy in case this gets deleted by a callback)
	listeners.call (&ValueListener::valueChanged, v);
}

OutputStream& JUCE_CALLTYPE operator<< (OutputStream& stream, const Value& value)
{
	return stream << value.toString();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Value.cpp ***/


/*** Start of inlined file: juce_Application.cpp ***/
BEGIN_JUCE_NAMESPACE

#if JUCE_MAC
 extern void juce_initialiseMacMainMenu();
#endif

class AppBroadcastCallback  : public ActionListener
{
public:
	AppBroadcastCallback()	{ MessageManager::getInstance()->registerBroadcastListener (this); }
	~AppBroadcastCallback()   { MessageManager::getInstance()->deregisterBroadcastListener (this); }

	void actionListenerCallback (const String& message)
	{
		JUCEApplication* const app = JUCEApplication::getInstance();

		if (app != 0 && message.startsWith (app->getApplicationName() + "/"))
			app->anotherInstanceStarted (message.substring (app->getApplicationName().length() + 1));
	}
};

JUCEApplication::JUCEApplication()
	: appReturnValue (0),
	  stillInitialising (true)
{
	jassert (isStandaloneApp() && appInstance == nullptr);
	appInstance = this;
}

JUCEApplication::~JUCEApplication()
{
	if (appLock != nullptr)
	{
		appLock->exit();
		appLock = nullptr;
	}

	jassert (appInstance == this);
	appInstance = nullptr;
}

JUCEApplication::CreateInstanceFunction JUCEApplication::createInstance = 0;
JUCEApplication* JUCEApplication::appInstance = nullptr;

bool JUCEApplication::moreThanOneInstanceAllowed()
{
	return true;
}

void JUCEApplication::anotherInstanceStarted (const String&)
{
}

void JUCEApplication::systemRequestedQuit()
{
	quit();
}

void JUCEApplication::quit()
{
	MessageManager::getInstance()->stopDispatchLoop();
}

void JUCEApplication::setApplicationReturnValue (const int newReturnValue) noexcept
{
	appReturnValue = newReturnValue;
}

void JUCEApplication::unhandledException (const std::exception*,
										  const String&,
										  const int)
{
	jassertfalse;
}

void JUCEApplication::sendUnhandledException (const std::exception* const e,
											  const char* const sourceFile,
											  const int lineNumber)
{
	if (appInstance != nullptr)
		appInstance->unhandledException (e, sourceFile, lineNumber);
}

ApplicationCommandTarget* JUCEApplication::getNextCommandTarget()
{
	return nullptr;
}

void JUCEApplication::getAllCommands (Array <CommandID>& commands)
{
	commands.add (StandardApplicationCommandIDs::quit);
}

void JUCEApplication::getCommandInfo (const CommandID commandID, ApplicationCommandInfo& result)
{
	if (commandID == StandardApplicationCommandIDs::quit)
	{
		result.setInfo (TRANS("Quit"),
						TRANS("Quits the application"),
						"Application",
						0);

		result.defaultKeypresses.add (KeyPress ('q', ModifierKeys::commandModifier, 0));
	}
}

bool JUCEApplication::perform (const InvocationInfo& info)
{
	if (info.commandID == StandardApplicationCommandIDs::quit)
	{
		systemRequestedQuit();
		return true;
	}

	return false;
}

bool JUCEApplication::initialiseApp (const String& commandLine)
{
	commandLineParameters = commandLine.trim();

   #if ! JUCE_IOS
	jassert (appLock == nullptr); // initialiseApp must only be called once!

	if (! moreThanOneInstanceAllowed())
	{
		appLock = new InterProcessLock ("juceAppLock_" + getApplicationName());

		if (! appLock->enter(0))
		{
			appLock = nullptr;
			MessageManager::broadcastMessage (getApplicationName() + "/" + commandLineParameters);

			DBG ("Another instance is running - quitting...");
			return false;
		}
	}
   #endif

	// let the app do its setting-up..
	initialise (commandLineParameters);

   #if JUCE_MAC
	juce_initialiseMacMainMenu(); // needs to be called after the app object has created, to get its name
   #endif

   #if ! JUCE_IOS
	broadcastCallback = new AppBroadcastCallback();
   #endif

	stillInitialising = false;
	return true;
}

int JUCEApplication::shutdownApp()
{
	jassert (appInstance == this);

	broadcastCallback = nullptr;

	JUCE_TRY
	{
		// give the app a chance to clean up..
		shutdown();
	}
	JUCE_CATCH_EXCEPTION

	return getApplicationReturnValue();
}

// This is called on the Mac and iOS where the OS doesn't allow the stack to unwind on shutdown..
void JUCEApplication::appWillTerminateByForce()
{
	{
		const ScopedPointer<JUCEApplication> app (JUCEApplication::getInstance());

		if (app != nullptr)
			app->shutdownApp();
	}

	shutdownJuce_GUI();
}

#if ! JUCE_ANDROID
int JUCEApplication::main (const String& commandLine)
{
	ScopedJuceInitialiser_GUI libraryInitialiser;
	jassert (createInstance != nullptr);
	int returnCode = 0;

	{
		const ScopedPointer<JUCEApplication> app (createInstance());

		if (! app->initialiseApp (commandLine))
			return 0;

		JUCE_TRY
		{
			// loop until a quit message is received..
			MessageManager::getInstance()->runDispatchLoop();
		}
		JUCE_CATCH_EXCEPTION

		returnCode = app->shutdownApp();
	}

	return returnCode;
}

#if JUCE_IOS
 extern int juce_iOSMain (int argc, const char* argv[]);
#endif

#if ! JUCE_WINDOWS
 extern const char* juce_Argv0;
#endif

#if JUCE_MAC
 extern void initialiseNSApplication();
#endif

int JUCEApplication::main (int argc, const char* argv[])
{
   #if JUCE_MAC
	initialiseNSApplication();
   #endif

   #if ! JUCE_WINDOWS
	jassert (createInstance != nullptr);
	juce_Argv0 = argv[0];
   #endif

   #if JUCE_IOS
	return juce_iOSMain (argc, argv);
   #else
	String cmd;
	for (int i = 1; i < argc; ++i)
		cmd << argv[i] << ' ';

	return JUCEApplication::main (cmd);
   #endif
}
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Application.cpp ***/


/*** Start of inlined file: juce_ApplicationCommandInfo.cpp ***/
BEGIN_JUCE_NAMESPACE

ApplicationCommandInfo::ApplicationCommandInfo (const CommandID commandID_) noexcept
	: commandID (commandID_),
	  flags (0)
{
}

void ApplicationCommandInfo::setInfo (const String& shortName_,
									  const String& description_,
									  const String& categoryName_,
									  const int flags_) noexcept
{
	shortName = shortName_;
	description = description_;
	categoryName = categoryName_;
	flags = flags_;
}

void ApplicationCommandInfo::setActive (const bool b) noexcept
{
	if (b)
		flags &= ~isDisabled;
	else
		flags |= isDisabled;
}

void ApplicationCommandInfo::setTicked (const bool b) noexcept
{
	if (b)
		flags |= isTicked;
	else
		flags &= ~isTicked;
}

void ApplicationCommandInfo::addDefaultKeypress (const int keyCode, const ModifierKeys& modifiers) noexcept
{
	defaultKeypresses.add (KeyPress (keyCode, modifiers, 0));
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ApplicationCommandInfo.cpp ***/


/*** Start of inlined file: juce_ApplicationCommandManager.cpp ***/
BEGIN_JUCE_NAMESPACE

ApplicationCommandManager::ApplicationCommandManager()
	: firstTarget (nullptr)
{
	keyMappings = new KeyPressMappingSet (this);

	Desktop::getInstance().addFocusChangeListener (this);
}

ApplicationCommandManager::~ApplicationCommandManager()
{
	Desktop::getInstance().removeFocusChangeListener (this);
	keyMappings = nullptr;
}

void ApplicationCommandManager::clearCommands()
{
	commands.clear();
	keyMappings->clearAllKeyPresses();
	triggerAsyncUpdate();
}

void ApplicationCommandManager::registerCommand (const ApplicationCommandInfo& newCommand)
{
	// zero isn't a valid command ID!
	jassert (newCommand.commandID != 0);

	// the name isn't optional!
	jassert (newCommand.shortName.isNotEmpty());

	if (getCommandForID (newCommand.commandID) == 0)
	{
		ApplicationCommandInfo* const newInfo = new ApplicationCommandInfo (newCommand);
		newInfo->flags &= ~ApplicationCommandInfo::isTicked;
		commands.add (newInfo);

		keyMappings->resetToDefaultMapping (newCommand.commandID);

		triggerAsyncUpdate();
	}
	else
	{
		// trying to re-register the same command with different parameters?
		jassert (newCommand.shortName == getCommandForID (newCommand.commandID)->shortName
				  && (newCommand.description == getCommandForID (newCommand.commandID)->description || newCommand.description.isEmpty())
				  && newCommand.categoryName == getCommandForID (newCommand.commandID)->categoryName
				  && newCommand.defaultKeypresses == getCommandForID (newCommand.commandID)->defaultKeypresses
				  && (newCommand.flags & (ApplicationCommandInfo::wantsKeyUpDownCallbacks | ApplicationCommandInfo::hiddenFromKeyEditor | ApplicationCommandInfo::readOnlyInKeyEditor))
					   == (getCommandForID (newCommand.commandID)->flags & (ApplicationCommandInfo::wantsKeyUpDownCallbacks | ApplicationCommandInfo::hiddenFromKeyEditor | ApplicationCommandInfo::readOnlyInKeyEditor)));
	}
}

void ApplicationCommandManager::registerAllCommandsForTarget (ApplicationCommandTarget* target)
{
	if (target != nullptr)
	{
		Array <CommandID> commandIDs;
		target->getAllCommands (commandIDs);

		for (int i = 0; i < commandIDs.size(); ++i)
		{
			ApplicationCommandInfo info (commandIDs.getUnchecked(i));
			target->getCommandInfo (info.commandID, info);

			registerCommand (info);
		}
	}
}

void ApplicationCommandManager::removeCommand (const CommandID commandID)
{
	for (int i = commands.size(); --i >= 0;)
	{
		if (commands.getUnchecked (i)->commandID == commandID)
		{
			commands.remove (i);
			triggerAsyncUpdate();

			const Array <KeyPress> keys (keyMappings->getKeyPressesAssignedToCommand (commandID));

			for (int j = keys.size(); --j >= 0;)
				keyMappings->removeKeyPress (keys.getReference (j));
		}
	}
}

void ApplicationCommandManager::commandStatusChanged()
{
	triggerAsyncUpdate();
}

const ApplicationCommandInfo* ApplicationCommandManager::getCommandForID (const CommandID commandID) const noexcept
{
	for (int i = commands.size(); --i >= 0;)
		if (commands.getUnchecked(i)->commandID == commandID)
			return commands.getUnchecked(i);

	return nullptr;
}

String ApplicationCommandManager::getNameOfCommand (const CommandID commandID) const noexcept
{
	const ApplicationCommandInfo* const ci = getCommandForID (commandID);

	return ci != nullptr ? ci->shortName : String::empty;
}

String ApplicationCommandManager::getDescriptionOfCommand (const CommandID commandID) const noexcept
{
	const ApplicationCommandInfo* const ci = getCommandForID (commandID);

	return ci != nullptr ? (ci->description.isNotEmpty() ? ci->description : ci->shortName)
						 : String::empty;
}

StringArray ApplicationCommandManager::getCommandCategories() const
{
	StringArray s;

	for (int i = 0; i < commands.size(); ++i)
		s.addIfNotAlreadyThere (commands.getUnchecked(i)->categoryName, false);

	return s;
}

Array<CommandID> ApplicationCommandManager::getCommandsInCategory (const String& categoryName) const
{
	Array <CommandID> results;

	for (int i = 0; i < commands.size(); ++i)
		if (commands.getUnchecked(i)->categoryName == categoryName)
			results.add (commands.getUnchecked(i)->commandID);

	return results;
}

bool ApplicationCommandManager::invokeDirectly (const CommandID commandID, const bool asynchronously)
{
	ApplicationCommandTarget::InvocationInfo info (commandID);
	info.invocationMethod = ApplicationCommandTarget::InvocationInfo::direct;

	return invoke (info, asynchronously);
}

bool ApplicationCommandManager::invoke (const ApplicationCommandTarget::InvocationInfo& info_, const bool asynchronously)
{
	// This call isn't thread-safe for use from a non-UI thread without locking the message
	// manager first..
	jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

	ApplicationCommandInfo commandInfo (0);
	ApplicationCommandTarget* const target = getTargetForCommand (info_.commandID, commandInfo);

	if (target == nullptr)
		return false;

	ApplicationCommandTarget::InvocationInfo info (info_);
	info.commandFlags = commandInfo.flags;

	sendListenerInvokeCallback (info);

	const bool ok = target->invoke (info, asynchronously);

	commandStatusChanged();

	return ok;
}

ApplicationCommandTarget* ApplicationCommandManager::getFirstCommandTarget (const CommandID)
{
	return firstTarget != nullptr ? firstTarget
								  : findDefaultComponentTarget();
}

void ApplicationCommandManager::setFirstCommandTarget (ApplicationCommandTarget* const newTarget) noexcept
{
	firstTarget = newTarget;
}

ApplicationCommandTarget* ApplicationCommandManager::getTargetForCommand (const CommandID commandID,
																		  ApplicationCommandInfo& upToDateInfo)
{
	ApplicationCommandTarget* target = getFirstCommandTarget (commandID);

	if (target == nullptr)
		target = JUCEApplication::getInstance();

	if (target != nullptr)
		target = target->getTargetForCommand (commandID);

	if (target != nullptr)
		target->getCommandInfo (commandID, upToDateInfo);

	return target;
}

ApplicationCommandTarget* ApplicationCommandManager::findTargetForComponent (Component* c)
{
	ApplicationCommandTarget* target = dynamic_cast <ApplicationCommandTarget*> (c);

	if (target == nullptr && c != nullptr)
		// (unable to use the syntax findParentComponentOfClass <ApplicationCommandTarget> () because of a VC6 compiler bug)
		target = c->findParentComponentOfClass ((ApplicationCommandTarget*) nullptr);

	return target;
}

ApplicationCommandTarget* ApplicationCommandManager::findDefaultComponentTarget()
{
	Component* c = Component::getCurrentlyFocusedComponent();

	if (c == nullptr)
	{
		TopLevelWindow* const activeWindow = TopLevelWindow::getActiveTopLevelWindow();

		if (activeWindow != nullptr)
		{
			c = activeWindow->getPeer()->getLastFocusedSubcomponent();

			if (c == nullptr)
				c = activeWindow;
		}
	}

	if (c == nullptr && Process::isForegroundProcess())
	{
		// getting a bit desperate now - try all desktop comps..
		for (int i = Desktop::getInstance().getNumComponents(); --i >= 0;)
		{
			ApplicationCommandTarget* const target
				= findTargetForComponent (Desktop::getInstance().getComponent (i)
											  ->getPeer()->getLastFocusedSubcomponent());

			if (target != nullptr)
				return target;
		}
	}

	if (c != nullptr)
	{
		ResizableWindow* const resizableWindow = dynamic_cast <ResizableWindow*> (c);

		// if we're focused on a ResizableWindow, chances are that it's the content
		// component that really should get the event. And if not, the event will
		// still be passed up to the top level window anyway, so let's send it to the
		// content comp.
		if (resizableWindow != nullptr && resizableWindow->getContentComponent() != nullptr)
			c = resizableWindow->getContentComponent();

		ApplicationCommandTarget* const target = findTargetForComponent (c);

		if (target != nullptr)
			return target;
	}

	return JUCEApplication::getInstance();
}

void ApplicationCommandManager::addListener (ApplicationCommandManagerListener* const listener)
{
	listeners.add (listener);
}

void ApplicationCommandManager::removeListener (ApplicationCommandManagerListener* const listener)
{
	listeners.remove (listener);
}

void ApplicationCommandManager::sendListenerInvokeCallback (const ApplicationCommandTarget::InvocationInfo& info)
{
	listeners.call (&ApplicationCommandManagerListener::applicationCommandInvoked, info);
}

void ApplicationCommandManager::handleAsyncUpdate()
{
	listeners.call (&ApplicationCommandManagerListener::applicationCommandListChanged);
}

void ApplicationCommandManager::globalFocusChanged (Component*)
{
	commandStatusChanged();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ApplicationCommandManager.cpp ***/


/*** Start of inlined file: juce_ApplicationCommandTarget.cpp ***/
BEGIN_JUCE_NAMESPACE

ApplicationCommandTarget::ApplicationCommandTarget()
{
}

ApplicationCommandTarget::~ApplicationCommandTarget()
{
	messageInvoker = nullptr;
}

bool ApplicationCommandTarget::tryToInvoke (const InvocationInfo& info, const bool async)
{
	if (isCommandActive (info.commandID))
	{
		if (async)
		{
			if (messageInvoker == nullptr)
				messageInvoker = new CommandTargetMessageInvoker (this);

			messageInvoker->postMessage (new Message (0, 0, 0, new ApplicationCommandTarget::InvocationInfo (info)));
			return true;
		}
		else
		{
			const bool success = perform (info);

			jassert (success);  // hmm - your target should have been able to perform this command. If it can't
								// do it at the moment for some reason, it should clear the 'isActive' flag when it
								// returns the command's info.
			return success;
		}
	}

	return false;
}

ApplicationCommandTarget* ApplicationCommandTarget::findFirstTargetParentComponent()
{
	Component* c = dynamic_cast <Component*> (this);

	if (c != nullptr)
		// (unable to use the syntax findParentComponentOfClass <ApplicationCommandTarget> () because of a VC6 compiler bug)
		return c->findParentComponentOfClass ((ApplicationCommandTarget*) nullptr);

	return nullptr;
}

ApplicationCommandTarget* ApplicationCommandTarget::getTargetForCommand (const CommandID commandID)
{
	ApplicationCommandTarget* target = this;
	int depth = 0;

	while (target != nullptr)
	{
		Array <CommandID> commandIDs;
		target->getAllCommands (commandIDs);

		if (commandIDs.contains (commandID))
			return target;

		target = target->getNextCommandTarget();

		++depth;
		jassert (depth < 100); // could be a recursive command chain??
		jassert (target != this); // definitely a recursive command chain!

		if (depth > 100 || target == this)
			break;
	}

	if (target == nullptr)
	{
		target = JUCEApplication::getInstance();

		if (target != nullptr)
		{
			Array <CommandID> commandIDs;
			target->getAllCommands (commandIDs);

			if (commandIDs.contains (commandID))
				return target;
		}
	}

	return nullptr;
}

bool ApplicationCommandTarget::isCommandActive (const CommandID commandID)
{
	ApplicationCommandInfo info (commandID);
	info.flags = ApplicationCommandInfo::isDisabled;

	getCommandInfo (commandID, info);

	return (info.flags & ApplicationCommandInfo::isDisabled) == 0;
}

bool ApplicationCommandTarget::invoke (const InvocationInfo& info, const bool async)
{
	ApplicationCommandTarget* target = this;
	int depth = 0;

	while (target != nullptr)
	{
		if (target->tryToInvoke (info, async))
			return true;

		target = target->getNextCommandTarget();

		++depth;
		jassert (depth < 100); // could be a recursive command chain??
		jassert (target != this); // definitely a recursive command chain!

		if (depth > 100 || target == this)
			break;
	}

	if (target == nullptr)
	{
		target = JUCEApplication::getInstance();

		if (target != nullptr)
			return target->tryToInvoke (info, async);
	}

	return false;
}

bool ApplicationCommandTarget::invokeDirectly (const CommandID commandID, const bool asynchronously)
{
	ApplicationCommandTarget::InvocationInfo info (commandID);
	info.invocationMethod = ApplicationCommandTarget::InvocationInfo::direct;

	return invoke (info, asynchronously);
}

ApplicationCommandTarget::InvocationInfo::InvocationInfo (const CommandID commandID_)
	: commandID (commandID_),
	  commandFlags (0),
	  invocationMethod (direct),
	  originatingComponent (nullptr),
	  isKeyDown (false),
	  millisecsSinceKeyPressed (0)
{
}

ApplicationCommandTarget::CommandTargetMessageInvoker::CommandTargetMessageInvoker (ApplicationCommandTarget* const owner_)
	: owner (owner_)
{
}

ApplicationCommandTarget::CommandTargetMessageInvoker::~CommandTargetMessageInvoker()
{
}

void ApplicationCommandTarget::CommandTargetMessageInvoker::handleMessage (const Message& message)
{
	const ScopedPointer <InvocationInfo> info (static_cast <InvocationInfo*> (message.pointerParameter));
	owner->tryToInvoke (*info, false);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ApplicationCommandTarget.cpp ***/




/*** Start of inlined file: juce_RecentlyOpenedFilesList.cpp ***/
BEGIN_JUCE_NAMESPACE

RecentlyOpenedFilesList::RecentlyOpenedFilesList()
	: maxNumberOfItems (10)
{
}

RecentlyOpenedFilesList::~RecentlyOpenedFilesList()
{
}

void RecentlyOpenedFilesList::setMaxNumberOfItems (const int newMaxNumber)
{
	maxNumberOfItems = jmax (1, newMaxNumber);

	files.removeRange (maxNumberOfItems, getNumFiles());
}

int RecentlyOpenedFilesList::getNumFiles() const
{
	return files.size();
}

File RecentlyOpenedFilesList::getFile (const int index) const
{
	return File (files [index]);
}

void RecentlyOpenedFilesList::clear()
{
	files.clear();
}

void RecentlyOpenedFilesList::addFile (const File& file)
{
	removeFile (file);
	files.insert (0, file.getFullPathName());

	setMaxNumberOfItems (maxNumberOfItems);
}

void RecentlyOpenedFilesList::removeFile (const File& file)
{
	files.removeString (file.getFullPathName());
}

void RecentlyOpenedFilesList::removeNonExistentFiles()
{
	for (int i = getNumFiles(); --i >= 0;)
		if (! getFile(i).exists())
			files.remove (i);
}

int RecentlyOpenedFilesList::createPopupMenuItems (PopupMenu& menuToAddTo,
												   const int baseItemId,
												   const bool showFullPaths,
												   const bool dontAddNonExistentFiles,
												   const File** filesToAvoid)
{
	int num = 0;

	for (int i = 0; i < getNumFiles(); ++i)
	{
		const File f (getFile(i));

		if ((! dontAddNonExistentFiles) || f.exists())
		{
			bool needsAvoiding = false;

			if (filesToAvoid != nullptr)
			{
				for (const File** avoid = filesToAvoid; *avoid != nullptr; ++avoid)
				{
					if (f == **avoid)
					{
						needsAvoiding = true;
						break;
					}
				}
			}

			if (! needsAvoiding)
			{
				menuToAddTo.addItem (baseItemId + i,
									 showFullPaths ? f.getFullPathName()
												   : f.getFileName());
				++num;
			}
		}
	}

	return num;
}

String RecentlyOpenedFilesList::toString() const
{
	return files.joinIntoString ("\n");
}

void RecentlyOpenedFilesList::restoreFromString (const String& stringifiedVersion)
{
	clear();
	files.addLines (stringifiedVersion);

	setMaxNumberOfItems (maxNumberOfItems);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RecentlyOpenedFilesList.cpp ***/


/*** Start of inlined file: juce_UndoManager.cpp ***/
BEGIN_JUCE_NAMESPACE

UndoManager::UndoManager (const int maxNumberOfUnitsToKeep,
						  const int minimumTransactions)
   : totalUnitsStored (0),
	 nextIndex (0),
	 newTransaction (true),
	 reentrancyCheck (false)
{
	setMaxNumberOfStoredUnits (maxNumberOfUnitsToKeep,
							   minimumTransactions);
}

UndoManager::~UndoManager()
{
	clearUndoHistory();
}

void UndoManager::clearUndoHistory()
{
	transactions.clear();
	transactionNames.clear();
	totalUnitsStored = 0;
	nextIndex = 0;
	sendChangeMessage();
}

int UndoManager::getNumberOfUnitsTakenUpByStoredCommands() const
{
	return totalUnitsStored;
}

void UndoManager::setMaxNumberOfStoredUnits (const int maxNumberOfUnitsToKeep,
											 const int minimumTransactions)
{
	maxNumUnitsToKeep	  = jmax (1, maxNumberOfUnitsToKeep);
	minimumTransactionsToKeep  = jmax (1, minimumTransactions);
}

bool UndoManager::perform (UndoableAction* const command_, const String& actionName)
{
	if (command_ != nullptr)
	{
		ScopedPointer<UndoableAction> command (command_);

		if (actionName.isNotEmpty())
			currentTransactionName = actionName;

		if (reentrancyCheck)
		{
			jassertfalse;	// don't call perform() recursively from the UndoableAction::perform() or
							 // undo() methods, or else these actions won't actually get done.

			return false;
		}
		else if (command->perform())
		{
			OwnedArray<UndoableAction>* commandSet = transactions [nextIndex - 1];

			if (commandSet != nullptr && ! newTransaction)
			{
				UndoableAction* lastAction = commandSet->getLast();

				if (lastAction != nullptr)
				{
					UndoableAction* coalescedAction = lastAction->createCoalescedAction (command);

					if (coalescedAction != nullptr)
					{
						command = coalescedAction;
						totalUnitsStored -= lastAction->getSizeInUnits();
						commandSet->removeLast();
					}
				}
			}
			else
			{
				commandSet = new OwnedArray<UndoableAction>();
				transactions.insert (nextIndex, commandSet);
				transactionNames.insert (nextIndex, currentTransactionName);
				++nextIndex;
			}

			totalUnitsStored += command->getSizeInUnits();
			commandSet->add (command.release());
			newTransaction = false;

			while (nextIndex < transactions.size())
			{
				const OwnedArray <UndoableAction>* const lastSet = transactions.getLast();

				for (int i = lastSet->size(); --i >= 0;)
					totalUnitsStored -= lastSet->getUnchecked (i)->getSizeInUnits();

				transactions.removeLast();
				transactionNames.remove (transactionNames.size() - 1);
			}

			while (nextIndex > 0
				   && totalUnitsStored > maxNumUnitsToKeep
				   && transactions.size() > minimumTransactionsToKeep)
			{
				const OwnedArray <UndoableAction>* const firstSet = transactions.getFirst();

				for (int i = firstSet->size(); --i >= 0;)
					totalUnitsStored -= firstSet->getUnchecked (i)->getSizeInUnits();

				jassert (totalUnitsStored >= 0); // something fishy going on if this fails!

				transactions.remove (0);
				transactionNames.remove (0);
				--nextIndex;
			}

			sendChangeMessage();

			return true;
		}
	}

	return false;
}

void UndoManager::beginNewTransaction (const String& actionName)
{
	newTransaction = true;
	currentTransactionName = actionName;
}

void UndoManager::setCurrentTransactionName (const String& newName)
{
	currentTransactionName = newName;
}

bool UndoManager::canUndo() const
{
	return nextIndex > 0;
}

bool UndoManager::canRedo() const
{
	return nextIndex < transactions.size();
}

String UndoManager::getUndoDescription() const
{
	return transactionNames [nextIndex - 1];
}

String UndoManager::getRedoDescription() const
{
	return transactionNames [nextIndex];
}

bool UndoManager::undo()
{
	const OwnedArray<UndoableAction>* const commandSet = transactions [nextIndex - 1];

	if (commandSet == nullptr)
		return false;

	bool failed = false;

	{
		const ScopedValueSetter<bool> setter (reentrancyCheck, true);

		for (int i = commandSet->size(); --i >= 0;)
		{
			if (! commandSet->getUnchecked(i)->undo())
			{
				jassertfalse;
				failed = true;
				break;
			}
		}
	}

	if (failed)
		clearUndoHistory();
	else
		--nextIndex;

	beginNewTransaction();

	sendChangeMessage();
	return true;
}

bool UndoManager::redo()
{
	const OwnedArray<UndoableAction>* const commandSet = transactions [nextIndex];

	if (commandSet == nullptr)
		return false;

	bool failed = false;

	{
		const ScopedValueSetter<bool> setter (reentrancyCheck, true);

		for (int i = 0; i < commandSet->size(); ++i)
		{
			if (! commandSet->getUnchecked(i)->perform())
			{
				jassertfalse;
				failed = true;
				break;
			}
		}
	}

	if (failed)
		clearUndoHistory();
	else
		++nextIndex;

	beginNewTransaction();

	sendChangeMessage();
	return true;
}

bool UndoManager::undoCurrentTransactionOnly()
{
	return newTransaction ? false : undo();
}

void UndoManager::getActionsInCurrentTransaction (Array <const UndoableAction*>& actionsFound) const
{
	const OwnedArray <UndoableAction>* const commandSet = transactions [nextIndex - 1];

	if (commandSet != nullptr && ! newTransaction)
	{
		for (int i = 0; i < commandSet->size(); ++i)
			actionsFound.add (commandSet->getUnchecked(i));
	}
}

int UndoManager::getNumActionsInCurrentTransaction() const
{
	const OwnedArray <UndoableAction>* const commandSet = transactions [nextIndex - 1];

	if (commandSet != nullptr && ! newTransaction)
		return commandSet->size();

	return 0;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_UndoManager.cpp ***/



/*** Start of inlined file: juce_DeletedAtShutdown.cpp ***/
BEGIN_JUCE_NAMESPACE

static SpinLock deletedAtShutdownLock;

DeletedAtShutdown::DeletedAtShutdown()
{
	const SpinLock::ScopedLockType sl (deletedAtShutdownLock);
	getObjects().add (this);
}

DeletedAtShutdown::~DeletedAtShutdown()
{
	const SpinLock::ScopedLockType sl (deletedAtShutdownLock);
	getObjects().removeValue (this);
}

void DeletedAtShutdown::deleteAll()
{
	// make a local copy of the array, so it can't get into a loop if something
	// creates another DeletedAtShutdown object during its destructor.
	Array <DeletedAtShutdown*> localCopy;

	{
		const SpinLock::ScopedLockType sl (deletedAtShutdownLock);
		localCopy = getObjects();
	}

	for (int i = localCopy.size(); --i >= 0;)
	{
		JUCE_TRY
		{
			DeletedAtShutdown* deletee = localCopy.getUnchecked(i);

			// double-check that it's not already been deleted during another object's destructor.
			{
				const SpinLock::ScopedLockType sl (deletedAtShutdownLock);
				if (! getObjects().contains (deletee))
					deletee = nullptr;
			}

			delete deletee;
		}
		JUCE_CATCH_EXCEPTION
	}

	// if no objects got re-created during shutdown, this should have been emptied by their
	// destructors
	jassert (getObjects().size() == 0);

	getObjects().clear(); // just to make sure the array doesn't have any memory still allocated
}

Array <DeletedAtShutdown*>& DeletedAtShutdown::getObjects()
{
	static Array <DeletedAtShutdown*> objects;
	return objects;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DeletedAtShutdown.cpp ***/



/*** Start of inlined file: juce_AudioFormatReader.cpp ***/
BEGIN_JUCE_NAMESPACE

AudioFormatReader::AudioFormatReader (InputStream* const in,
									  const String& formatName_)
	: sampleRate (0),
	  bitsPerSample (0),
	  lengthInSamples (0),
	  numChannels (0),
	  usesFloatingPointData (false),
	  input (in),
	  formatName (formatName_)
{
}

AudioFormatReader::~AudioFormatReader()
{
	delete input;
}

bool AudioFormatReader::read (int* const* destSamples,
							  int numDestChannels,
							  int64 startSampleInSource,
							  int numSamplesToRead,
							  const bool fillLeftoverChannelsWithCopies)
{
	jassert (numDestChannels > 0); // you have to actually give this some channels to work with!

	int startOffsetInDestBuffer = 0;

	if (startSampleInSource < 0)
	{
		const int silence = (int) jmin (-startSampleInSource, (int64) numSamplesToRead);

		for (int i = numDestChannels; --i >= 0;)
			if (destSamples[i] != nullptr)
				zeromem (destSamples[i], sizeof (int) * silence);

		startOffsetInDestBuffer += silence;
		numSamplesToRead -= silence;
		startSampleInSource = 0;
	}

	if (numSamplesToRead <= 0)
		return true;

	if (! readSamples (const_cast<int**> (destSamples),
					   jmin ((int) numChannels, numDestChannels), startOffsetInDestBuffer,
					   startSampleInSource, numSamplesToRead))
		return false;

	if (numDestChannels > (int) numChannels)
	{
		if (fillLeftoverChannelsWithCopies)
		{
			int* lastFullChannel = destSamples[0];

			for (int i = (int) numChannels; --i > 0;)
			{
				if (destSamples[i] != nullptr)
				{
					lastFullChannel = destSamples[i];
					break;
				}
			}

			if (lastFullChannel != nullptr)
				for (int i = numChannels; i < numDestChannels; ++i)
					if (destSamples[i] != nullptr)
						memcpy (destSamples[i], lastFullChannel, sizeof (int) * numSamplesToRead);
		}
		else
		{
			for (int i = numChannels; i < numDestChannels; ++i)
				if (destSamples[i] != nullptr)
					zeromem (destSamples[i], sizeof (int) * numSamplesToRead);
		}
	}

	return true;
}

void AudioFormatReader::readMaxLevels (int64 startSampleInFile,
									   int64 numSamples,
									   float& lowestLeft, float& highestLeft,
									   float& lowestRight, float& highestRight)
{
	if (numSamples <= 0)
	{
		lowestLeft = 0;
		lowestRight = 0;
		highestLeft = 0;
		highestRight = 0;
		return;
	}

	const int bufferSize = (int) jmin (numSamples, (int64) 4096);
	HeapBlock<int> tempSpace (bufferSize * 2 + 64);

	int* tempBuffer[3];
	tempBuffer[0] = tempSpace.getData();
	tempBuffer[1] = tempSpace.getData() + bufferSize;
	tempBuffer[2] = 0;

	if (usesFloatingPointData)
	{
		float lmin = 1.0e6f;
		float lmax = -lmin;
		float rmin = lmin;
		float rmax = lmax;

		while (numSamples > 0)
		{
			const int numToDo = (int) jmin (numSamples, (int64) bufferSize);
			read (tempBuffer, 2, startSampleInFile, numToDo, false);

			numSamples -= numToDo;
			startSampleInFile += numToDo;

			float bufMin, bufMax;
			findMinAndMax (reinterpret_cast<float*> (tempBuffer[0]), numToDo, bufMin, bufMax);
			lmin = jmin (lmin, bufMin);
			lmax = jmax (lmax, bufMax);

			if (numChannels > 1)
			{
				findMinAndMax (reinterpret_cast<float*> (tempBuffer[1]), numToDo, bufMin, bufMax);
				rmin = jmin (rmin, bufMin);
				rmax = jmax (rmax, bufMax);
			}
		}

		if (numChannels <= 1)
		{
			rmax = lmax;
			rmin = lmin;
		}

		lowestLeft = lmin;
		highestLeft = lmax;
		lowestRight = rmin;
		highestRight = rmax;
	}
	else
	{
		int lmax = std::numeric_limits<int>::min();
		int lmin = std::numeric_limits<int>::max();
		int rmax = std::numeric_limits<int>::min();
		int rmin = std::numeric_limits<int>::max();

		while (numSamples > 0)
		{
			const int numToDo = (int) jmin (numSamples, (int64) bufferSize);
			if (! read (tempBuffer, 2, startSampleInFile, numToDo, false))
				break;

			numSamples -= numToDo;
			startSampleInFile += numToDo;

			for (int j = numChannels; --j >= 0;)
			{
				int bufMin, bufMax;
				findMinAndMax (tempBuffer[j], numToDo, bufMin, bufMax);

				if (j == 0)
				{
					lmax = jmax (lmax, bufMax);
					lmin = jmin (lmin, bufMin);
				}
				else
				{
					rmax = jmax (rmax, bufMax);
					rmin = jmin (rmin, bufMin);
				}
			}
		}

		if (numChannels <= 1)
		{
			rmax = lmax;
			rmin = lmin;
		}

		lowestLeft = lmin / (float) std::numeric_limits<int>::max();
		highestLeft = lmax / (float) std::numeric_limits<int>::max();
		lowestRight = rmin / (float) std::numeric_limits<int>::max();
		highestRight = rmax / (float) std::numeric_limits<int>::max();
	}
}

int64 AudioFormatReader::searchForLevel (int64 startSample,
										 int64 numSamplesToSearch,
										 const double magnitudeRangeMinimum,
										 const double magnitudeRangeMaximum,
										 const int minimumConsecutiveSamples)
{
	if (numSamplesToSearch == 0)
		return -1;

	const int bufferSize = 4096;
	HeapBlock<int> tempSpace (bufferSize * 2 + 64);

	int* tempBuffer[3];
	tempBuffer[0] = tempSpace.getData();
	tempBuffer[1] = tempSpace.getData() + bufferSize;
	tempBuffer[2] = 0;

	int consecutive = 0;
	int64 firstMatchPos = -1;

	jassert (magnitudeRangeMaximum > magnitudeRangeMinimum);

	const double doubleMin = jlimit (0.0, (double) std::numeric_limits<int>::max(), magnitudeRangeMinimum * std::numeric_limits<int>::max());
	const double doubleMax = jlimit (doubleMin, (double) std::numeric_limits<int>::max(), magnitudeRangeMaximum * std::numeric_limits<int>::max());
	const int intMagnitudeRangeMinimum = roundToInt (doubleMin);
	const int intMagnitudeRangeMaximum = roundToInt (doubleMax);

	while (numSamplesToSearch != 0)
	{
		const int numThisTime = (int) jmin (abs64 (numSamplesToSearch), (int64) bufferSize);
		int64 bufferStart = startSample;

		if (numSamplesToSearch < 0)
			bufferStart -= numThisTime;

		if (bufferStart >= (int) lengthInSamples)
			break;

		read (tempBuffer, 2, bufferStart, numThisTime, false);

		int num = numThisTime;
		while (--num >= 0)
		{
			if (numSamplesToSearch < 0)
				--startSample;

			bool matches = false;
			const int index = (int) (startSample - bufferStart);

			if (usesFloatingPointData)
			{
				const float sample1 = std::abs (((float*) tempBuffer[0]) [index]);

				if (sample1 >= magnitudeRangeMinimum
					 && sample1 <= magnitudeRangeMaximum)
				{
					matches = true;
				}
				else if (numChannels > 1)
				{
					const float sample2 = std::abs (((float*) tempBuffer[1]) [index]);

					matches = (sample2 >= magnitudeRangeMinimum
								 && sample2 <= magnitudeRangeMaximum);
				}
			}
			else
			{
				const int sample1 = abs (tempBuffer[0] [index]);

				if (sample1 >= intMagnitudeRangeMinimum
					 && sample1 <= intMagnitudeRangeMaximum)
				{
					matches = true;
				}
				else if (numChannels > 1)
				{
					const int sample2 = abs (tempBuffer[1][index]);

					matches = (sample2 >= intMagnitudeRangeMinimum
								 && sample2 <= intMagnitudeRangeMaximum);
				}
			}

			if (matches)
			{
				if (firstMatchPos < 0)
					firstMatchPos = startSample;

				if (++consecutive >= minimumConsecutiveSamples)
				{
					if (firstMatchPos < 0 || firstMatchPos >= lengthInSamples)
						return -1;

					return firstMatchPos;
				}
			}
			else
			{
				consecutive = 0;
				firstMatchPos = -1;
			}

			if (numSamplesToSearch > 0)
				++startSample;
		}

		if (numSamplesToSearch > 0)
			numSamplesToSearch -= numThisTime;
		else
			numSamplesToSearch += numThisTime;
	}

	return -1;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_AudioFormatReader.cpp ***/


/*** Start of inlined file: juce_AudioSampleBuffer.cpp ***/
BEGIN_JUCE_NAMESPACE

AudioSampleBuffer::AudioSampleBuffer (const int numChannels_,
									  const int numSamples) noexcept
  : numChannels (numChannels_),
	size (numSamples)
{
	jassert (numSamples >= 0);
	jassert (numChannels_ > 0);

	allocateData();
}

AudioSampleBuffer::AudioSampleBuffer (const AudioSampleBuffer& other) noexcept
  : numChannels (other.numChannels),
	size (other.size)
{
	allocateData();
	const size_t numBytes = size * sizeof (float);

	for (int i = 0; i < numChannels; ++i)
		memcpy (channels[i], other.channels[i], numBytes);
}

void AudioSampleBuffer::allocateData()
{
	const size_t channelListSize = (numChannels + 1) * sizeof (float*);
	allocatedBytes = (int) (numChannels * size * sizeof (float) + channelListSize + 32);
	allocatedData.malloc (allocatedBytes);
	channels = reinterpret_cast <float**> (allocatedData.getData());

	float* chan = (float*) (allocatedData + channelListSize);
	for (int i = 0; i < numChannels; ++i)
	{
		channels[i] = chan;
		chan += size;
	}

	channels [numChannels] = 0;
}

AudioSampleBuffer::AudioSampleBuffer (float** dataToReferTo,
									  const int numChannels_,
									  const int numSamples) noexcept
	: numChannels (numChannels_),
	  size (numSamples),
	  allocatedBytes (0)
{
	jassert (numChannels_ > 0);
	allocateChannels (dataToReferTo, 0);
}

AudioSampleBuffer::AudioSampleBuffer (float** dataToReferTo,
									  const int numChannels_,
									  const int startSample,
									  const int numSamples) noexcept
	: numChannels (numChannels_),
	  size (numSamples),
	  allocatedBytes (0)
{
	jassert (numChannels_ > 0);
	allocateChannels (dataToReferTo, startSample);
}

void AudioSampleBuffer::setDataToReferTo (float** dataToReferTo,
										  const int newNumChannels,
										  const int newNumSamples) noexcept
{
	jassert (newNumChannels > 0);

	allocatedBytes = 0;
	allocatedData.free();

	numChannels = newNumChannels;
	size = newNumSamples;

	allocateChannels (dataToReferTo, 0);
}

void AudioSampleBuffer::allocateChannels (float** const dataToReferTo, int offset)
{
	// (try to avoid doing a malloc here, as that'll blow up things like Pro-Tools)
	if (numChannels < numElementsInArray (preallocatedChannelSpace))
	{
		channels = static_cast <float**> (preallocatedChannelSpace);
	}
	else
	{
		allocatedData.malloc (numChannels + 1, sizeof (float*));
		channels = reinterpret_cast <float**> (allocatedData.getData());
	}

	for (int i = 0; i < numChannels; ++i)
	{
		// you have to pass in the same number of valid pointers as numChannels
		jassert (dataToReferTo[i] != nullptr);

		channels[i] = dataToReferTo[i] + offset;
	}

	channels [numChannels] = 0;
}

AudioSampleBuffer& AudioSampleBuffer::operator= (const AudioSampleBuffer& other) noexcept
{
	if (this != &other)
	{
		setSize (other.getNumChannels(), other.getNumSamples(), false, false, false);

		const size_t numBytes = size * sizeof (float);

		for (int i = 0; i < numChannels; ++i)
			memcpy (channels[i], other.channels[i], numBytes);
	}

	return *this;
}

AudioSampleBuffer::~AudioSampleBuffer() noexcept
{
}

void AudioSampleBuffer::setSize (const int newNumChannels,
								 const int newNumSamples,
								 const bool keepExistingContent,
								 const bool clearExtraSpace,
								 const bool avoidReallocating) noexcept
{
	jassert (newNumChannels > 0);

	if (newNumSamples != size || newNumChannels != numChannels)
	{
		const size_t channelListSize = (newNumChannels + 1) * sizeof (float*);
		const size_t newTotalBytes = (newNumChannels * newNumSamples * sizeof (float)) + channelListSize + 32;

		if (keepExistingContent)
		{
			HeapBlock <char> newData;
			newData.allocate (newTotalBytes, clearExtraSpace);

			const int numChansToCopy = jmin (numChannels, newNumChannels);
			const size_t numBytesToCopy = sizeof (float) * jmin (newNumSamples, size);

			float** const newChannels = reinterpret_cast <float**> (newData.getData());
			float* newChan = reinterpret_cast <float*> (newData + channelListSize);
			for (int i = 0; i < numChansToCopy; ++i)
			{
				memcpy (newChan, channels[i], numBytesToCopy);
				newChannels[i] = newChan;
				newChan += newNumSamples;
			}

			allocatedData.swapWith (newData);
			allocatedBytes = (int) newTotalBytes;
			channels = newChannels;
		}
		else
		{
			if (avoidReallocating && allocatedBytes >= newTotalBytes)
			{
				if (clearExtraSpace)
					allocatedData.clear (newTotalBytes);
			}
			else
			{
				allocatedBytes = newTotalBytes;
				allocatedData.allocate (newTotalBytes, clearExtraSpace);
				channels = reinterpret_cast <float**> (allocatedData.getData());
			}

			float* chan = reinterpret_cast <float*> (allocatedData + channelListSize);
			for (int i = 0; i < newNumChannels; ++i)
			{
				channels[i] = chan;
				chan += newNumSamples;
			}
		}

		channels [newNumChannels] = 0;
		size = newNumSamples;
		numChannels = newNumChannels;
	}
}

void AudioSampleBuffer::clear() noexcept
{
	for (int i = 0; i < numChannels; ++i)
		zeromem (channels[i], size * sizeof (float));
}

void AudioSampleBuffer::clear (const int startSample,
							   const int numSamples) noexcept
{
	jassert (startSample >= 0 && startSample + numSamples <= size);

	for (int i = 0; i < numChannels; ++i)
		zeromem (channels [i] + startSample, numSamples * sizeof (float));
}

void AudioSampleBuffer::clear (const int channel,
							   const int startSample,
							   const int numSamples) noexcept
{
	jassert (isPositiveAndBelow (channel, numChannels));
	jassert (startSample >= 0 && startSample + numSamples <= size);

	zeromem (channels [channel] + startSample, numSamples * sizeof (float));
}

void AudioSampleBuffer::applyGain (const int channel,
								   const int startSample,
								   int numSamples,
								   const float gain) noexcept
{
	jassert (isPositiveAndBelow (channel, numChannels));
	jassert (startSample >= 0 && startSample + numSamples <= size);

	if (gain != 1.0f)
	{
		float* d = channels [channel] + startSample;

		if (gain == 0.0f)
		{
			zeromem (d, sizeof (float) * numSamples);
		}
		else
		{
			while (--numSamples >= 0)
				*d++ *= gain;
		}
	}
}

void AudioSampleBuffer::applyGainRamp (const int channel,
									   const int startSample,
									   int numSamples,
									   float startGain,
									   float endGain) noexcept
{
	if (startGain == endGain)
	{
		applyGain (channel, startSample, numSamples, startGain);
	}
	else
	{
		jassert (isPositiveAndBelow (channel, numChannels));
		jassert (startSample >= 0 && startSample + numSamples <= size);

		const float increment = (endGain - startGain) / numSamples;
		float* d = channels [channel] + startSample;

		while (--numSamples >= 0)
		{
			*d++ *= startGain;
			startGain += increment;
		}
	}
}

void AudioSampleBuffer::applyGain (const int startSample,
								   const int numSamples,
								   const float gain) noexcept
{
	for (int i = 0; i < numChannels; ++i)
		applyGain (i, startSample, numSamples, gain);
}

void AudioSampleBuffer::addFrom (const int destChannel,
								 const int destStartSample,
								 const AudioSampleBuffer& source,
								 const int sourceChannel,
								 const int sourceStartSample,
								 int numSamples,
								 const float gain) noexcept
{
	jassert (&source != this || sourceChannel != destChannel);
	jassert (isPositiveAndBelow (destChannel, numChannels));
	jassert (destStartSample >= 0 && destStartSample + numSamples <= size);
	jassert (isPositiveAndBelow (sourceChannel, source.numChannels));
	jassert (sourceStartSample >= 0 && sourceStartSample + numSamples <= source.size);

	if (gain != 0.0f && numSamples > 0)
	{
		float* d = channels [destChannel] + destStartSample;
		const float* s  = source.channels [sourceChannel] + sourceStartSample;

		if (gain != 1.0f)
		{
			while (--numSamples >= 0)
				*d++ += gain * *s++;
		}
		else
		{
			while (--numSamples >= 0)
				*d++ += *s++;
		}
	}
}

void AudioSampleBuffer::addFrom (const int destChannel,
								 const int destStartSample,
								 const float* source,
								 int numSamples,
								 const float gain) noexcept
{
	jassert (isPositiveAndBelow (destChannel, numChannels));
	jassert (destStartSample >= 0 && destStartSample + numSamples <= size);
	jassert (source != nullptr);

	if (gain != 0.0f && numSamples > 0)
	{
		float* d = channels [destChannel] + destStartSample;

		if (gain != 1.0f)
		{
			while (--numSamples >= 0)
				*d++ += gain * *source++;
		}
		else
		{
			while (--numSamples >= 0)
				*d++ += *source++;
		}
	}
}

void AudioSampleBuffer::addFromWithRamp (const int destChannel,
										 const int destStartSample,
										 const float* source,
										 int numSamples,
										 float startGain,
										 const float endGain) noexcept
{
	jassert (isPositiveAndBelow (destChannel, numChannels));
	jassert (destStartSample >= 0 && destStartSample + numSamples <= size);
	jassert (source != nullptr);

	if (startGain == endGain)
	{
		addFrom (destChannel,
				 destStartSample,
				 source,
				 numSamples,
				 startGain);
	}
	else
	{
		if (numSamples > 0 && (startGain != 0.0f || endGain != 0.0f))
		{
			const float increment = (endGain - startGain) / numSamples;
			float* d = channels [destChannel] + destStartSample;

			while (--numSamples >= 0)
			{
				*d++ += startGain * *source++;
				startGain += increment;
			}
		}
	}
}

void AudioSampleBuffer::copyFrom (const int destChannel,
								  const int destStartSample,
								  const AudioSampleBuffer& source,
								  const int sourceChannel,
								  const int sourceStartSample,
								  int numSamples) noexcept
{
	jassert (&source != this || sourceChannel != destChannel);
	jassert (isPositiveAndBelow (destChannel, numChannels));
	jassert (destStartSample >= 0 && destStartSample + numSamples <= size);
	jassert (isPositiveAndBelow (sourceChannel, source.numChannels));
	jassert (sourceStartSample >= 0 && sourceStartSample + numSamples <= source.size);

	if (numSamples > 0)
	{
		memcpy (channels [destChannel] + destStartSample,
				source.channels [sourceChannel] + sourceStartSample,
				sizeof (float) * numSamples);
	}
}

void AudioSampleBuffer::copyFrom (const int destChannel,
								  const int destStartSample,
								  const float* source,
								  int numSamples) noexcept
{
	jassert (isPositiveAndBelow (destChannel, numChannels));
	jassert (destStartSample >= 0 && destStartSample + numSamples <= size);
	jassert (source != nullptr);

	if (numSamples > 0)
	{
		memcpy (channels [destChannel] + destStartSample,
				source,
				sizeof (float) * numSamples);
	}
}

void AudioSampleBuffer::copyFrom (const int destChannel,
								  const int destStartSample,
								  const float* source,
								  int numSamples,
								  const float gain) noexcept
{
	jassert (isPositiveAndBelow (destChannel, numChannels));
	jassert (destStartSample >= 0 && destStartSample + numSamples <= size);
	jassert (source != nullptr);

	if (numSamples > 0)
	{
		float* d = channels [destChannel] + destStartSample;

		if (gain != 1.0f)
		{
			if (gain == 0)
			{
				zeromem (d, sizeof (float) * numSamples);
			}
			else
			{
				while (--numSamples >= 0)
					*d++ = gain * *source++;
			}
		}
		else
		{
			memcpy (d, source, sizeof (float) * numSamples);
		}
	}
}

void AudioSampleBuffer::copyFromWithRamp (const int destChannel,
										  const int destStartSample,
										  const float* source,
										  int numSamples,
										  float startGain,
										  float endGain) noexcept
{
	jassert (isPositiveAndBelow (destChannel, numChannels));
	jassert (destStartSample >= 0 && destStartSample + numSamples <= size);
	jassert (source != nullptr);

	if (startGain == endGain)
	{
		copyFrom (destChannel,
				  destStartSample,
				  source,
				  numSamples,
				  startGain);
	}
	else
	{
		if (numSamples > 0 && (startGain != 0.0f || endGain != 0.0f))
		{
			const float increment = (endGain - startGain) / numSamples;
			float* d = channels [destChannel] + destStartSample;

			while (--numSamples >= 0)
			{
				*d++ = startGain * *source++;
				startGain += increment;
			}
		}
	}
}

void AudioSampleBuffer::findMinMax (const int channel,
									const int startSample,
									int numSamples,
									float& minVal,
									float& maxVal) const noexcept
{
	jassert (isPositiveAndBelow (channel, numChannels));
	jassert (startSample >= 0 && startSample + numSamples <= size);

	findMinAndMax (channels [channel] + startSample, numSamples, minVal, maxVal);
}

float AudioSampleBuffer::getMagnitude (const int channel,
									   const int startSample,
									   const int numSamples) const noexcept
{
	jassert (isPositiveAndBelow (channel, numChannels));
	jassert (startSample >= 0 && startSample + numSamples <= size);

	float mn, mx;
	findMinMax (channel, startSample, numSamples, mn, mx);

	return jmax (mn, -mn, mx, -mx);
}

float AudioSampleBuffer::getMagnitude (const int startSample,
									   const int numSamples) const noexcept
{
	float mag = 0.0f;

	for (int i = 0; i < numChannels; ++i)
		mag = jmax (mag, getMagnitude (i, startSample, numSamples));

	return mag;
}

float AudioSampleBuffer::getRMSLevel (const int channel,
									  const int startSample,
									  const int numSamples) const noexcept
{
	jassert (isPositiveAndBelow (channel, numChannels));
	jassert (startSample >= 0 && startSample + numSamples <= size);

	if (numSamples <= 0 || channel < 0 || channel >= numChannels)
		return 0.0f;

	const float* const data = channels [channel] + startSample;
	double sum = 0.0;

	for (int i = 0; i < numSamples; ++i)
	{
		const float sample = data [i];
		sum += sample * sample;
	}

	return (float) std::sqrt (sum / numSamples);
}

void AudioSampleBuffer::readFromAudioReader (AudioFormatReader* reader,
											 const int startSample,
											 const int numSamples,
											 const int64 readerStartSample,
											 const bool useLeftChan,
											 const bool useRightChan)
{
	jassert (reader != nullptr);
	jassert (startSample >= 0 && startSample + numSamples <= size);

	if (numSamples > 0)
	{
		int* chans[3];

		if (useLeftChan == useRightChan)
		{
			chans[0] = reinterpret_cast<int*> (getSampleData (0, startSample));
			chans[1] = (reader->numChannels > 1 && getNumChannels() > 1) ? reinterpret_cast<int*> (getSampleData (1, startSample)) : 0;
		}
		else if (useLeftChan || (reader->numChannels == 1))
		{
			chans[0] = reinterpret_cast<int*> (getSampleData (0, startSample));
			chans[1] = nullptr;
		}
		else if (useRightChan)
		{
			chans[0] = nullptr;
			chans[1] = reinterpret_cast<int*> (getSampleData (0, startSample));
		}

		chans[2] = nullptr;

		reader->read (chans, 2, readerStartSample, numSamples, true);

		if (! reader->usesFloatingPointData)
		{
			for (int j = 0; j < 2; ++j)
			{
				float* const d = reinterpret_cast <float*> (chans[j]);

				if (d != nullptr)
				{
					const float multiplier = 1.0f / 0x7fffffff;

					for (int i = 0; i < numSamples; ++i)
						d[i] = *reinterpret_cast<int*> (d + i) * multiplier;
				}
			}
		}

		if (numChannels > 1 && (chans[0] == nullptr || chans[1] == nullptr))
		{
			// if this is a stereo buffer and the source was mono, dupe the first channel..
			memcpy (getSampleData (1, startSample),
					getSampleData (0, startSample),
					sizeof (float) * numSamples);
		}
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_AudioSampleBuffer.cpp ***/






/*** Start of inlined file: juce_ActionBroadcaster.cpp ***/
BEGIN_JUCE_NAMESPACE

// special message of our own with a string in it
class ActionMessage  : public Message
{
public:
	ActionMessage (const String& messageText, ActionListener* const listener_) noexcept
		: message (messageText)
	{
		pointerParameter = listener_;
	}

	const String message;

private:
	JUCE_DECLARE_NON_COPYABLE (ActionMessage);
};

ActionBroadcaster::CallbackReceiver::CallbackReceiver() {}

void ActionBroadcaster::CallbackReceiver::handleMessage (const Message& message)
{
	const ActionMessage& am = static_cast <const ActionMessage&> (message);
	ActionListener* const target = static_cast <ActionListener*> (am.pointerParameter);

	if (owner->actionListeners.contains (target))
		target->actionListenerCallback (am.message);
}

ActionBroadcaster::ActionBroadcaster()
{
	// are you trying to create this object before or after juce has been intialised??
	jassert (MessageManager::instance != nullptr);

	callback.owner = this;
}

ActionBroadcaster::~ActionBroadcaster()
{
	// all event-based objects must be deleted BEFORE juce is shut down!
	jassert (MessageManager::instance != nullptr);
}

void ActionBroadcaster::addActionListener (ActionListener* const listener)
{
	const ScopedLock sl (actionListenerLock);

	if (listener != nullptr)
		actionListeners.add (listener);
}

void ActionBroadcaster::removeActionListener (ActionListener* const listener)
{
	const ScopedLock sl (actionListenerLock);
	actionListeners.removeValue (listener);
}

void ActionBroadcaster::removeAllActionListeners()
{
	const ScopedLock sl (actionListenerLock);
	actionListeners.clear();
}

void ActionBroadcaster::sendActionMessage (const String& message) const
{
	const ScopedLock sl (actionListenerLock);

	for (int i = actionListeners.size(); --i >= 0;)
		callback.postMessage (new ActionMessage (message, actionListeners.getUnchecked(i)));
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ActionBroadcaster.cpp ***/


/*** Start of inlined file: juce_AsyncUpdater.cpp ***/
BEGIN_JUCE_NAMESPACE

class AsyncUpdaterMessage  : public CallbackMessage
{
public:
	AsyncUpdaterMessage (AsyncUpdater& owner_)
		: owner (owner_)
	{
	}

	void messageCallback()
	{
		if (shouldDeliver.compareAndSetBool (0, 1))
			owner.handleAsyncUpdate();
	}

	Atomic<int> shouldDeliver;

private:
	AsyncUpdater& owner;
};

AsyncUpdater::AsyncUpdater()
{
	message = new AsyncUpdaterMessage (*this);
}

inline Atomic<int>& AsyncUpdater::getDeliveryFlag() const noexcept
{
	return static_cast <AsyncUpdaterMessage*> (message.getObject())->shouldDeliver;
}

AsyncUpdater::~AsyncUpdater()
{
	// You're deleting this object with a background thread while there's an update
	// pending on the main event thread - that's pretty dodgy threading, as the callback could
	// happen after this destructor has finished. You should either use a MessageManagerLock while
	// deleting this object, or find some other way to avoid such a race condition.
	jassert ((! isUpdatePending()) || MessageManager::getInstance()->currentThreadHasLockedMessageManager());

	getDeliveryFlag().set (0);
}

void AsyncUpdater::triggerAsyncUpdate()
{
	if (getDeliveryFlag().compareAndSetBool (1, 0))
		message->post();
}

void AsyncUpdater::cancelPendingUpdate() noexcept
{
	getDeliveryFlag().set (0);
}

void AsyncUpdater::handleUpdateNowIfNeeded()
{
	// This can only be called by the event thread.
	jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

	if (getDeliveryFlag().exchange (0) != 0)
		handleAsyncUpdate();
}

bool AsyncUpdater::isUpdatePending() const noexcept
{
	return getDeliveryFlag().value != 0;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_AsyncUpdater.cpp ***/


/*** Start of inlined file: juce_ChangeBroadcaster.cpp ***/
BEGIN_JUCE_NAMESPACE

ChangeBroadcaster::ChangeBroadcaster() noexcept
{
	// are you trying to create this object before or after juce has been intialised??
	jassert (MessageManager::instance != nullptr);

	callback.owner = this;
}

ChangeBroadcaster::~ChangeBroadcaster()
{
	// all event-based objects must be deleted BEFORE juce is shut down!
	jassert (MessageManager::instance != nullptr);
}

void ChangeBroadcaster::addChangeListener (ChangeListener* const listener)
{
	// Listeners can only be safely added when the event thread is locked
	// You can  use a MessageManagerLock if you need to call this from another thread.
	jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

	changeListeners.add (listener);
}

void ChangeBroadcaster::removeChangeListener (ChangeListener* const listener)
{
	// Listeners can only be safely added when the event thread is locked
	// You can  use a MessageManagerLock if you need to call this from another thread.
	jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

	changeListeners.remove (listener);
}

void ChangeBroadcaster::removeAllChangeListeners()
{
	// Listeners can only be safely added when the event thread is locked
	// You can  use a MessageManagerLock if you need to call this from another thread.
	jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

	changeListeners.clear();
}

void ChangeBroadcaster::sendChangeMessage()
{
	if (changeListeners.size() > 0)
		callback.triggerAsyncUpdate();
}

void ChangeBroadcaster::sendSynchronousChangeMessage()
{
	// This can only be called by the event thread.
	jassert (MessageManager::getInstance()->isThisTheMessageThread());

	callback.cancelPendingUpdate();
	callListeners();
}

void ChangeBroadcaster::dispatchPendingMessages()
{
	callback.handleUpdateNowIfNeeded();
}

void ChangeBroadcaster::callListeners()
{
	changeListeners.call (&ChangeListener::changeListenerCallback, this);
}

ChangeBroadcaster::ChangeBroadcasterCallback::ChangeBroadcasterCallback()
	: owner (nullptr)
{
}

void ChangeBroadcaster::ChangeBroadcasterCallback::handleAsyncUpdate()
{
	jassert (owner != nullptr);
	owner->callListeners();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ChangeBroadcaster.cpp ***/


/*** Start of inlined file: juce_Message.cpp ***/
BEGIN_JUCE_NAMESPACE

Message::Message() noexcept
	: intParameter1 (0),
	  intParameter2 (0),
	  intParameter3 (0),
	  pointerParameter (nullptr),
	  messageRecipient (nullptr)
{
}

Message::Message (const int intParameter1_,
				  const int intParameter2_,
				  const int intParameter3_,
				  void* const pointerParameter_) noexcept
	: intParameter1 (intParameter1_),
	  intParameter2 (intParameter2_),
	  intParameter3 (intParameter3_),
	  pointerParameter (pointerParameter_),
	  messageRecipient (nullptr)
{
}

Message::~Message()
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Message.cpp ***/


/*** Start of inlined file: juce_MessageListener.cpp ***/
BEGIN_JUCE_NAMESPACE

MessageListener::MessageListener() noexcept
{
	// are you trying to create a messagelistener before or after juce has been intialised??
	jassert (MessageManager::instance != nullptr);

	if (MessageManager::instance != nullptr)
		MessageManager::instance->messageListeners.add (this);
}

MessageListener::~MessageListener()
{
	if (MessageManager::instance != nullptr)
		MessageManager::instance->messageListeners.removeValue (this);
}

void MessageListener::postMessage (Message* const message) const
{
	message->messageRecipient = const_cast <MessageListener*> (this);

	if (MessageManager::instance == nullptr)
		MessageManager::getInstance();

	MessageManager::instance->postMessageToQueue (message);
}

bool MessageListener::isValidMessageListener() const noexcept
{
	return (MessageManager::instance != nullptr)
			 && MessageManager::instance->messageListeners.contains (this);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MessageListener.cpp ***/


/*** Start of inlined file: juce_MessageManager.cpp ***/
BEGIN_JUCE_NAMESPACE

MessageManager* MessageManager::instance = nullptr;

enum { quitMessageId = 0xfffff321 };

MessageManager::MessageManager() noexcept
  : quitMessagePosted (false),
	quitMessageReceived (false),
	messageThreadId (Thread::getCurrentThreadId()),
	threadWithLock (0)
{
	if (JUCEApplication::isStandaloneApp())
		Thread::setCurrentThreadName ("Juce Message Thread");
}

MessageManager::~MessageManager() noexcept
{
	broadcaster = nullptr;

	doPlatformSpecificShutdown();

	// If you hit this assertion, then you've probably leaked some kind of MessageListener object..
	jassert (messageListeners.size() == 0);

	jassert (instance == this);
	instance = nullptr;  // do this last in case this instance is still needed by doPlatformSpecificShutdown()
}

MessageManager* MessageManager::getInstance()
{
	if (instance == nullptr)
	{
		instance = new MessageManager();
		doPlatformSpecificInitialisation();
	}

	return instance;
}

void MessageManager::deleteInstance()
{
	deleteAndZero (instance);
}

void MessageManager::postMessageToQueue (Message* const message)
{
	if (quitMessagePosted || ! postMessageToSystemQueue (message))
		Message::Ptr deleter (message); // (this will delete messages that were just created with a 0 ref count)
}

CallbackMessage::CallbackMessage() noexcept {}
CallbackMessage::~CallbackMessage() {}

void CallbackMessage::post()
{
	if (MessageManager::instance != nullptr)
		MessageManager::instance->postMessageToQueue (this);
}

// not for public use..
void MessageManager::deliverMessage (Message* const message)
{
	JUCE_TRY
	{
		MessageListener* const recipient = message->messageRecipient;

		if (recipient == nullptr)
		{
			CallbackMessage* const callbackMessage = dynamic_cast <CallbackMessage*> (message);

			if (callbackMessage != nullptr)
			{
				callbackMessage->messageCallback();
			}
			else if (message->intParameter1 == (int) quitMessageId)
			{
				quitMessageReceived = true;
			}
		}
		else if (messageListeners.contains (recipient))
		{
			recipient->handleMessage (*message);
		}
	}
	JUCE_CATCH_EXCEPTION
}

#if JUCE_MODAL_LOOPS_PERMITTED && ! (JUCE_MAC || JUCE_IOS)
void MessageManager::runDispatchLoop()
{
	jassert (isThisTheMessageThread()); // must only be called by the message thread

	runDispatchLoopUntil (-1);
}

void MessageManager::stopDispatchLoop()
{
	postMessageToQueue (new Message ((int) quitMessageId, 0, 0, nullptr));
	quitMessagePosted = true;
}

bool MessageManager::runDispatchLoopUntil (int millisecondsToRunFor)
{
	jassert (isThisTheMessageThread()); // must only be called by the message thread

	const int64 endTime = Time::currentTimeMillis() + millisecondsToRunFor;

	while ((millisecondsToRunFor < 0 || endTime > Time::currentTimeMillis())
			&& ! quitMessageReceived)
	{
		JUCE_TRY
		{
			if (! dispatchNextMessageOnSystemQueue (millisecondsToRunFor >= 0))
			{
				const int msToWait = (int) (endTime - Time::currentTimeMillis());

				if (msToWait > 0)
					Thread::sleep (jmin (5, msToWait));
			}
		}
		JUCE_CATCH_EXCEPTION
	}

	return ! quitMessageReceived;
}

#endif

void MessageManager::deliverBroadcastMessage (const String& value)
{
	if (broadcaster != nullptr)
		broadcaster->sendActionMessage (value);
}

void MessageManager::registerBroadcastListener (ActionListener* const listener)
{
	if (broadcaster == nullptr)
		broadcaster = new ActionBroadcaster();

	broadcaster->addActionListener (listener);
}

void MessageManager::deregisterBroadcastListener (ActionListener* const listener)
{
	if (broadcaster != nullptr)
		broadcaster->removeActionListener (listener);
}

bool MessageManager::isThisTheMessageThread() const noexcept
{
	return Thread::getCurrentThreadId() == messageThreadId;
}

void MessageManager::setCurrentThreadAsMessageThread()
{
	const Thread::ThreadID thisThread = Thread::getCurrentThreadId();

	if (messageThreadId != thisThread)
	{
		messageThreadId = thisThread;

		// This is needed on windows to make sure the message window is created by this thread
		doPlatformSpecificShutdown();
		doPlatformSpecificInitialisation();
	}
}

bool MessageManager::currentThreadHasLockedMessageManager() const noexcept
{
	const Thread::ThreadID thisThread = Thread::getCurrentThreadId();
	return thisThread == messageThreadId || thisThread == threadWithLock;
}

/*  The only safe way to lock the message thread while another thread does
	some work is by posting a special message, whose purpose is to tie up the event
	loop until the other thread has finished its business.

	Any other approach can get horribly deadlocked if the OS uses its own hidden locks which
	get locked before making an event callback, because if the same OS lock gets indirectly
	accessed from another thread inside a MM lock, you're screwed. (this is exactly what happens
	in Cocoa).
*/
class MessageManagerLock::BlockingMessage   : public CallbackMessage
{
public:
	BlockingMessage() {}

	void messageCallback()
	{
		lockedEvent.signal();
		releaseEvent.wait();
	}

	WaitableEvent lockedEvent, releaseEvent;

private:
	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (BlockingMessage);
};

MessageManagerLock::MessageManagerLock (Thread* const threadToCheck)
	: locked (false)
{
	init (threadToCheck, nullptr);
}

MessageManagerLock::MessageManagerLock (ThreadPoolJob* const jobToCheckForExitSignal)
	: locked (false)
{
	init (nullptr, jobToCheckForExitSignal);
}

void MessageManagerLock::init (Thread* const threadToCheck, ThreadPoolJob* const job)
{
	if (MessageManager::instance != nullptr)
	{
		if (MessageManager::instance->currentThreadHasLockedMessageManager())
		{
			locked = true;   // either we're on the message thread, or this is a re-entrant call.
		}
		else
		{
			if (threadToCheck == nullptr && job == nullptr)
			{
				MessageManager::instance->lockingLock.enter();
			}
			else
			{
				while (! MessageManager::instance->lockingLock.tryEnter())
				{
					if ((threadToCheck != nullptr && threadToCheck->threadShouldExit())
						  || (job != nullptr && job->shouldExit()))
						return;

					Thread::sleep (1);
				}
			}

			blockingMessage = new BlockingMessage();
			blockingMessage->post();

			while (! blockingMessage->lockedEvent.wait (20))
			{
				if ((threadToCheck != nullptr && threadToCheck->threadShouldExit())
					  || (job != nullptr && job->shouldExit()))
				{
					blockingMessage->releaseEvent.signal();
					blockingMessage = nullptr;
					MessageManager::instance->lockingLock.exit();
					return;
				}
			}

			jassert (MessageManager::instance->threadWithLock == 0);

			MessageManager::instance->threadWithLock = Thread::getCurrentThreadId();
			locked = true;
		}
	}
}

MessageManagerLock::~MessageManagerLock() noexcept
{
	if (blockingMessage != nullptr)
	{
		jassert (MessageManager::instance == nullptr || MessageManager::instance->currentThreadHasLockedMessageManager());

		blockingMessage->releaseEvent.signal();
		blockingMessage = nullptr;

		if (MessageManager::instance != nullptr)
		{
			MessageManager::instance->threadWithLock = 0;
			MessageManager::instance->lockingLock.exit();
		}
	}
}

JUCE_API void JUCE_CALLTYPE initialiseJuce_GUI()
{
	MessageManager::getInstance();
}

JUCE_API void JUCE_CALLTYPE shutdownJuce_GUI()
{
	DeletedAtShutdown::deleteAll();
	MessageManager::deleteInstance();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MessageManager.cpp ***/


/*** Start of inlined file: juce_MultiTimer.cpp ***/
BEGIN_JUCE_NAMESPACE

class MultiTimer::MultiTimerCallback	: public Timer
{
public:
	MultiTimerCallback (const int timerId_, MultiTimer& owner_)
		: timerId (timerId_),
		  owner (owner_)
	{
	}

	void timerCallback()
	{
		owner.timerCallback (timerId);
	}

	const int timerId;

private:
	MultiTimer& owner;
};

MultiTimer::MultiTimer() noexcept
{
}

MultiTimer::MultiTimer (const MultiTimer&) noexcept
{
}

MultiTimer::~MultiTimer()
{
	const SpinLock::ScopedLockType sl (timerListLock);
	timers.clear();
}

void MultiTimer::startTimer (const int timerId, const int intervalInMilliseconds) noexcept
{
	const SpinLock::ScopedLockType sl (timerListLock);

	for (int i = timers.size(); --i >= 0;)
	{
		MultiTimerCallback* const t = timers.getUnchecked(i);

		if (t->timerId == timerId)
		{
			t->startTimer (intervalInMilliseconds);
			return;
		}
	}

	MultiTimerCallback* const newTimer = new MultiTimerCallback (timerId, *this);
	timers.add (newTimer);
	newTimer->startTimer (intervalInMilliseconds);
}

void MultiTimer::stopTimer (const int timerId) noexcept
{
	const SpinLock::ScopedLockType sl (timerListLock);

	for (int i = timers.size(); --i >= 0;)
	{
		MultiTimerCallback* const t = timers.getUnchecked(i);

		if (t->timerId == timerId)
			t->stopTimer();
	}
}

bool MultiTimer::isTimerRunning (const int timerId) const noexcept
{
	const SpinLock::ScopedLockType sl (timerListLock);

	for (int i = timers.size(); --i >= 0;)
	{
		const MultiTimerCallback* const t = timers.getUnchecked(i);
		if (t->timerId == timerId)
			return t->isTimerRunning();
	}

	return false;
}

int MultiTimer::getTimerInterval (const int timerId) const noexcept
{
	const SpinLock::ScopedLockType sl (timerListLock);

	for (int i = timers.size(); --i >= 0;)
	{
		const MultiTimerCallback* const t = timers.getUnchecked(i);
		if (t->timerId == timerId)
			return t->getTimerInterval();
	}

	return 0;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MultiTimer.cpp ***/


/*** Start of inlined file: juce_Timer.cpp ***/
BEGIN_JUCE_NAMESPACE

class InternalTimerThread  : private Thread,
							 private MessageListener,
							 private DeletedAtShutdown,
							 private AsyncUpdater
{
public:
	typedef CriticalSection LockType; // (mysteriously, using a SpinLock here causes problems on some XP machines..)

	InternalTimerThread()
		: Thread ("Juce Timer"),
		  firstTimer (nullptr),
		  callbackNeeded (0)
	{
		triggerAsyncUpdate();
	}

	~InternalTimerThread() noexcept
	{
		stopThread (4000);

		jassert (instance == this || instance == nullptr);
		if (instance == this)
			instance = nullptr;
	}

	void run()
	{
		uint32 lastTime = Time::getMillisecondCounter();
		Message::Ptr messageToSend (new Message());

		while (! threadShouldExit())
		{
			const uint32 now = Time::getMillisecondCounter();

			if (now == lastTime)
			{
				wait (1);
				continue;
			}

			const int elapsed = now >= lastTime ? (now - lastTime)
												: (std::numeric_limits<uint32>::max() - (lastTime - now));
			lastTime = now;

			const int timeUntilFirstTimer = getTimeUntilFirstTimer (elapsed);

			if (timeUntilFirstTimer <= 0)
			{
				/* If we managed to set the atomic boolean to true then send a message, this is needed
				   as a memory barrier so the message won't be sent before callbackNeeded is set to true,
				   but if it fails it means the message-thread changed the value from under us so at least
				   some processing is happenening and we can just loop around and try again
				*/
				if (callbackNeeded.compareAndSetBool (1, 0))
				{
					postMessage (messageToSend);

					/* Sometimes our message can get discarded by the OS (e.g. when running as an RTAS
					   when the app has a modal loop), so this is how long to wait before assuming the
					   message has been lost and trying again.
					*/
					const uint32 messageDeliveryTimeout = now + 2000;

					while (callbackNeeded.get() != 0)
					{
						wait (4);

						if (threadShouldExit())
							return;

						if (Time::getMillisecondCounter() > messageDeliveryTimeout)
							break;
					}
				}
			}
			else
			{
				// don't wait for too long because running this loop also helps keep the
				// Time::getApproximateMillisecondTimer value stay up-to-date
				wait (jlimit (1, 50, timeUntilFirstTimer));
			}
		}
	}

	void callTimers()
	{
		const LockType::ScopedLockType sl (lock);

		while (firstTimer != nullptr && firstTimer->countdownMs <= 0)
		{
			Timer* const t = firstTimer;
			t->countdownMs = t->periodMs;

			removeTimer (t);
			addTimer (t);

			const LockType::ScopedUnlockType ul (lock);

			JUCE_TRY
			{
				t->timerCallback();
			}
			JUCE_CATCH_EXCEPTION
		}

		/* This is needed as a memory barrier to make sure all processing of current timers is done
		   before the boolean is set. This set should never fail since if it was false in the first place,
		   we wouldn't get a message (so it can't be changed from false to true from under us), and if we
		   get a message then the value is true and the other thread can only set it to true again and
		   we will get another callback to set it to false.
		*/
		callbackNeeded.set (0);
	}

	void handleMessage (const Message&)
	{
		callTimers();
	}

	void callTimersSynchronously()
	{
		if (! isThreadRunning())
		{
			// (This is relied on by some plugins in cases where the MM has
			// had to restart and the async callback never started)
			cancelPendingUpdate();
			triggerAsyncUpdate();
		}

		callTimers();
	}

	static void callAnyTimersSynchronously()
	{
		if (InternalTimerThread::instance != nullptr)
			InternalTimerThread::instance->callTimersSynchronously();
	}

	static inline void add (Timer* const tim) noexcept
	{
		if (instance == nullptr)
			instance = new InternalTimerThread();

		instance->addTimer (tim);
	}

	static inline void remove (Timer* const tim) noexcept
	{
		if (instance != nullptr)
			instance->removeTimer (tim);
	}

	static inline void resetCounter (Timer* const tim, const int newCounter) noexcept
	{
		if (instance != nullptr)
		{
			tim->countdownMs = newCounter;
			tim->periodMs = newCounter;

			if ((tim->next != nullptr && tim->next->countdownMs < tim->countdownMs)
				 || (tim->previous != nullptr && tim->previous->countdownMs > tim->countdownMs))
			{
				instance->removeTimer (tim);
				instance->addTimer (tim);
			}
		}
	}

private:
	friend class Timer;
	static InternalTimerThread* instance;
	static LockType lock;
	Timer* volatile firstTimer;
	Atomic <int> callbackNeeded;

	void addTimer (Timer* const t) noexcept
	{
	   #if JUCE_DEBUG
		Timer* tt = firstTimer;

		while (tt != nullptr)
		{
			// trying to add a timer that's already here - shouldn't get to this point,
			// so if you get this assertion, let me know!
			jassert (tt != t);

			tt = tt->next;
		}

		jassert (t->previous == nullptr && t->next == nullptr);
	   #endif

		Timer* i = firstTimer;

		if (i == nullptr || i->countdownMs > t->countdownMs)
		{
			t->next = firstTimer;
			firstTimer = t;
		}
		else
		{
			while (i->next != nullptr && i->next->countdownMs <= t->countdownMs)
				i = i->next;

			jassert (i != nullptr);

			t->next = i->next;
			t->previous = i;
			i->next = t;
		}

		if (t->next != nullptr)
			t->next->previous = t;

		jassert ((t->next == nullptr || t->next->countdownMs >= t->countdownMs)
				  && (t->previous == nullptr || t->previous->countdownMs <= t->countdownMs));

		notify();
	}

	void removeTimer (Timer* const t) noexcept
	{
	   #if JUCE_DEBUG
		Timer* tt = firstTimer;
		bool found = false;

		while (tt != nullptr)
		{
			if (tt == t)
			{
				found = true;
				break;
			}

			tt = tt->next;
		}

		// trying to remove a timer that's not here - shouldn't get to this point,
		// so if you get this assertion, let me know!
		jassert (found);
	   #endif

		if (t->previous != nullptr)
		{
			jassert (firstTimer != t);
			t->previous->next = t->next;
		}
		else
		{
			jassert (firstTimer == t);
			firstTimer = t->next;
		}

		if (t->next != nullptr)
			t->next->previous = t->previous;

		t->next = nullptr;
		t->previous = nullptr;
	}

	int getTimeUntilFirstTimer (const int numMillisecsElapsed) const
	{
		const LockType::ScopedLockType sl (lock);

		for (Timer* t = firstTimer; t != nullptr; t = t->next)
			t->countdownMs -= numMillisecsElapsed;

		return firstTimer != nullptr ? firstTimer->countdownMs : 1000;
	}

	void handleAsyncUpdate()
	{
		startThread (7);
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (InternalTimerThread);
};

InternalTimerThread* InternalTimerThread::instance = nullptr;
InternalTimerThread::LockType InternalTimerThread::lock;

#if JUCE_DEBUG
static SortedSet <Timer*> activeTimers;
#endif

Timer::Timer() noexcept
   : countdownMs (0),
	 periodMs (0),
	 previous (nullptr),
	 next (nullptr)
{
   #if JUCE_DEBUG
	const InternalTimerThread::LockType::ScopedLockType sl (InternalTimerThread::lock);
	activeTimers.add (this);
   #endif
}

Timer::Timer (const Timer&) noexcept
   : countdownMs (0),
	 periodMs (0),
	 previous (nullptr),
	 next (nullptr)
{
   #if JUCE_DEBUG
	const InternalTimerThread::LockType::ScopedLockType sl (InternalTimerThread::lock);
	activeTimers.add (this);
   #endif
}

Timer::~Timer()
{
	stopTimer();

   #if JUCE_DEBUG
	activeTimers.removeValue (this);
   #endif
}

void Timer::startTimer (const int interval) noexcept
{
	const InternalTimerThread::LockType::ScopedLockType sl (InternalTimerThread::lock);

   #if JUCE_DEBUG
	// this isn't a valid object! Your timer might be a dangling pointer or something..
	jassert (activeTimers.contains (this));
   #endif

	if (periodMs == 0)
	{
		countdownMs = interval;
		periodMs = jmax (1, interval);
		InternalTimerThread::add (this);
	}
	else
	{
		InternalTimerThread::resetCounter (this, interval);
	}
}

void Timer::stopTimer() noexcept
{
	const InternalTimerThread::LockType::ScopedLockType sl (InternalTimerThread::lock);

   #if JUCE_DEBUG
	// this isn't a valid object! Your timer might be a dangling pointer or something..
	jassert (activeTimers.contains (this));
   #endif

	if (periodMs > 0)
	{
		InternalTimerThread::remove (this);
		periodMs = 0;
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Timer.cpp ***/

#endif

