/*
  ==============================================================================

   This file is part of the JUCE library - "Jules' Utility Class Extensions"
   Copyright 2004-9 by Raw Material Software Ltd.

  ------------------------------------------------------------------------------

   JUCE can be redistributed and/or modified under the terms of the GNU General
   Public License (Version 2), as published by the Free Software Foundation.
   A copy of the license is included in the JUCE distribution, or can be found
   online at www.gnu.org/licenses.

   JUCE is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  ------------------------------------------------------------------------------

   To release a closed-source product which uses JUCE, commercial licenses are
   available: visit www.rawmaterialsoftware.com/juce for more information.

  ==============================================================================
*/

#ifdef __JUCE_JUCEHEADER__
 /* When you add the amalgamated cpp file to your project, you mustn't include it in
	a file where you've already included juce.h - just put it inside a file on its own,
	possibly with your config flags preceding it, but don't include anything else. */
 #error
#endif


#include "juce_TargetPlatform.h"

 // FORCE_AMALGAMATOR_INCLUDE

#include "juce_Config.h"

  // FORCE_AMALGAMATOR_INCLUDE

#ifndef JUCE_BUILD_CORE
 #define JUCE_BUILD_CORE 1
#endif
#ifndef JUCE_BUILD_MISC
 #define JUCE_BUILD_MISC 1
#endif
#ifndef JUCE_BUILD_GUI
 #define JUCE_BUILD_GUI 1
#endif
#ifndef JUCE_BUILD_NATIVE
 #define JUCE_BUILD_NATIVE 1
#endif

#if JUCE_ONLY_BUILD_CORE_LIBRARY
 #undef JUCE_BUILD_MISC
 #undef JUCE_BUILD_GUI
#endif

#ifndef JUCE_NAMESPACE
 #define JUCE_NAMESPACE juce
#endif

//==============================================================================
#include "juce_NativeIncludes.h"
//==============================================================================
#define DONT_SET_USING_JUCE_NAMESPACE 1
#undef max
#undef min
#define NO_DUMMY_DECL

#define JUCE_AMALGAMATED_TEMPLATE 1

#if JUCE_BUILD_NATIVE
 #include "juce_amalgamated.h"  // FORCE_AMALGAMATOR_INCLUDE
#endif

#if (defined(_MSC_VER) && (_MSC_VER <= 1200))
 #pragma warning (disable: 4309 4305)
#endif


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
  
  
  
  
  
  
  
  
#if JUCE_BUILD_GUI

/*** Start of inlined file: juce_Component.cpp ***/
BEGIN_JUCE_NAMESPACE

#define CHECK_MESSAGE_MANAGER_IS_LOCKED	 jassert (MessageManager::getInstance()->currentThreadHasLockedMessageManager());

Component* Component::currentlyFocusedComponent = nullptr;

class Component::MouseListenerList
{
public:
	MouseListenerList()
		: numDeepMouseListeners (0)
	{
	}

	void addListener (MouseListener* const newListener, const bool wantsEventsForAllNestedChildComponents)
	{
		if (! listeners.contains (newListener))
		{
			if (wantsEventsForAllNestedChildComponents)
			{
				listeners.insert (0, newListener);
				++numDeepMouseListeners;
			}
			else
			{
				listeners.add (newListener);
			}
		}
	}

	void removeListener (MouseListener* const listenerToRemove)
	{
		const int index = listeners.indexOf (listenerToRemove);

		if (index >= 0)
		{
			if (index < numDeepMouseListeners)
				--numDeepMouseListeners;

			listeners.remove (index);
		}
	}

	static void sendMouseEvent (Component& comp, BailOutChecker& checker,
								void (MouseListener::*eventMethod) (const MouseEvent&), const MouseEvent& e)
	{
		if (checker.shouldBailOut())
			return;

		{
			MouseListenerList* const list = comp.mouseListeners;

			if (list != nullptr)
			{
				for (int i = list->listeners.size(); --i >= 0;)
				{
					(list->listeners.getUnchecked(i)->*eventMethod) (e);

					if (checker.shouldBailOut())
						return;

					i = jmin (i, list->listeners.size());
				}
			}
		}

		Component* p = comp.parentComponent;

		while (p != nullptr)
		{
			MouseListenerList* const list = p->mouseListeners;

			if (list != nullptr && list->numDeepMouseListeners > 0)
			{
				BailOutChecker2 checker2 (checker, p);

				for (int i = list->numDeepMouseListeners; --i >= 0;)
				{
					(list->listeners.getUnchecked(i)->*eventMethod) (e);

					if (checker2.shouldBailOut())
						return;

					i = jmin (i, list->numDeepMouseListeners);
				}
			}

			p = p->parentComponent;
		}
	}

	static void sendWheelEvent (Component& comp, BailOutChecker& checker, const MouseEvent& e,
								const float wheelIncrementX, const float wheelIncrementY)
	{
		{
			MouseListenerList* const list = comp.mouseListeners;

			if (list != nullptr)
			{
				for (int i = list->listeners.size(); --i >= 0;)
				{
					list->listeners.getUnchecked(i)->mouseWheelMove (e, wheelIncrementX, wheelIncrementY);

					if (checker.shouldBailOut())
						return;

					i = jmin (i, list->listeners.size());
				}
			}
		}

		Component* p = comp.parentComponent;

		while (p != nullptr)
		{
			MouseListenerList* const list = p->mouseListeners;

			if (list != nullptr && list->numDeepMouseListeners > 0)
			{
				BailOutChecker2 checker2 (checker, p);

				for (int i = list->numDeepMouseListeners; --i >= 0;)
				{
					list->listeners.getUnchecked(i)->mouseWheelMove (e, wheelIncrementX, wheelIncrementY);

					if (checker2.shouldBailOut())
						return;

					i = jmin (i, list->numDeepMouseListeners);
				}
			}

			p = p->parentComponent;
		}
	}

private:
	Array <MouseListener*> listeners;
	int numDeepMouseListeners;

	class BailOutChecker2
	{
	public:
		BailOutChecker2 (BailOutChecker& checker_, Component* const component)
			: checker (checker_), safePointer (component)
		{
		}

		bool shouldBailOut() const noexcept
		{
			return checker.shouldBailOut() || safePointer == 0;
		}

	private:
		BailOutChecker& checker;
		const WeakReference<Component> safePointer;

		JUCE_DECLARE_NON_COPYABLE (BailOutChecker2);
	};

	JUCE_DECLARE_NON_COPYABLE (MouseListenerList);
};

class Component::ComponentHelpers
{
public:

   #if JUCE_MODAL_LOOPS_PERMITTED
	static void* runModalLoopCallback (void* userData)
	{
		return (void*) (pointer_sized_int) static_cast <Component*> (userData)->runModalLoop();
	}
   #endif

	static const Identifier getColourPropertyId (const int colourId)
	{
		String s;
		s.preallocateBytes (32);
		s << "jcclr_" << String::toHexString (colourId);
		return s;
	}

	static inline bool hitTest (Component& comp, const Point<int>& localPoint)
	{
		return isPositiveAndBelow (localPoint.getX(), comp.getWidth())
				 && isPositiveAndBelow (localPoint.getY(), comp.getHeight())
				 && comp.hitTest (localPoint.getX(), localPoint.getY());
	}

	static const Point<int> convertFromParentSpace (const Component& comp, const Point<int>& pointInParentSpace)
	{
		if (comp.affineTransform == nullptr)
			return pointInParentSpace - comp.getPosition();

		return pointInParentSpace.toFloat().transformedBy (comp.affineTransform->inverted()).toInt() - comp.getPosition();
	}

	static const Rectangle<int> convertFromParentSpace (const Component& comp, const Rectangle<int>& areaInParentSpace)
	{
		if (comp.affineTransform == nullptr)
			return areaInParentSpace - comp.getPosition();

		return areaInParentSpace.toFloat().transformed (comp.affineTransform->inverted()).getSmallestIntegerContainer() - comp.getPosition();
	}

	static const Point<int> convertToParentSpace (const Component& comp, const Point<int>& pointInLocalSpace)
	{
		if (comp.affineTransform == nullptr)
			return pointInLocalSpace + comp.getPosition();

		return (pointInLocalSpace + comp.getPosition()).toFloat().transformedBy (*comp.affineTransform).toInt();
	}

	static const Rectangle<int> convertToParentSpace (const Component& comp, const Rectangle<int>& areaInLocalSpace)
	{
		if (comp.affineTransform == nullptr)
			return areaInLocalSpace + comp.getPosition();

		return (areaInLocalSpace + comp.getPosition()).toFloat().transformed (*comp.affineTransform).getSmallestIntegerContainer();
	}

	template <typename Type>
	static const Type convertFromDistantParentSpace (const Component* parent, const Component& target, Type coordInParent)
	{
		const Component* const directParent = target.getParentComponent();
		jassert (directParent != nullptr);

		if (directParent == parent)
			return convertFromParentSpace (target, coordInParent);

		return convertFromParentSpace (target, convertFromDistantParentSpace (parent, *directParent, coordInParent));
	}

	template <typename Type>
	static const Type convertCoordinate (const Component* target, const Component* source, Type p)
	{
		while (source != nullptr)
		{
			if (source == target)
				return p;

			if (source->isParentOf (target))
				return convertFromDistantParentSpace (source, *target, p);

			if (source->isOnDesktop())
			{
				p = source->getPeer()->localToGlobal (p);
				source = nullptr;
			}
			else
			{
				p = convertToParentSpace (*source, p);
				source = source->getParentComponent();
			}
		}

		jassert (source == nullptr);
		if (target == nullptr)
			return p;

		const Component* const topLevelComp = target->getTopLevelComponent();

		if (topLevelComp->isOnDesktop())
			p = topLevelComp->getPeer()->globalToLocal (p);
		else
			p = convertFromParentSpace (*topLevelComp, p);

		if (topLevelComp == target)
			return p;

		return convertFromDistantParentSpace (topLevelComp, *target, p);
	}

	static const Rectangle<int> getUnclippedArea (const Component& comp)
	{
		Rectangle<int> r (comp.getLocalBounds());

		Component* const p = comp.getParentComponent();

		if (p != nullptr)
			r = r.getIntersection (convertFromParentSpace (comp, getUnclippedArea (*p)));

		return r;
	}

	static void clipObscuredRegions (const Component& comp, Graphics& g, const Rectangle<int>& clipRect, const Point<int>& delta)
	{
		for (int i = comp.childComponentList.size(); --i >= 0;)
		{
			const Component& child = *comp.childComponentList.getUnchecked(i);

			if (child.isVisible() && ! child.isTransformed())
			{
				const Rectangle<int> newClip (clipRect.getIntersection (child.bounds));

				if (! newClip.isEmpty())
				{
					if (child.isOpaque())
					{
						g.excludeClipRegion (newClip + delta);
					}
					else
					{
						const Point<int> childPos (child.getPosition());
						clipObscuredRegions (child, g, newClip - childPos, childPos + delta);
					}
				}
			}
		}
	}

	static void subtractObscuredRegions (const Component& comp, RectangleList& result,
										 const Point<int>& delta,
										 const Rectangle<int>& clipRect,
										 const Component* const compToAvoid)
	{
		for (int i = comp.childComponentList.size(); --i >= 0;)
		{
			const Component* const c = comp.childComponentList.getUnchecked(i);

			if (c != compToAvoid && c->isVisible())
			{
				if (c->isOpaque())
				{
					Rectangle<int> childBounds (c->bounds.getIntersection (clipRect));
					childBounds.translate (delta.getX(), delta.getY());

					result.subtract (childBounds);
				}
				else
				{
					Rectangle<int> newClip (clipRect.getIntersection (c->bounds));
					newClip.translate (-c->getX(), -c->getY());

					subtractObscuredRegions (*c, result, c->getPosition() + delta,
											 newClip, compToAvoid);
				}
			}
		}
	}

	static const Rectangle<int> getParentOrMainMonitorBounds (const Component& comp)
	{
		return comp.getParentComponent() != nullptr ? comp.getParentComponent()->getLocalBounds()
													: Desktop::getInstance().getMainMonitorArea();
	}
};

Component::Component()
  : parentComponent (nullptr),
	lookAndFeel (nullptr),
	effect (nullptr),
	componentFlags (0),
	componentTransparency (0)
{
}

Component::Component (const String& name)
  : componentName (name),
	parentComponent (nullptr),
	lookAndFeel (nullptr),
	effect (nullptr),
	componentFlags (0),
	componentTransparency (0)
{
}

Component::~Component()
{
   #if ! JUCE_VC6  // (access to private union not allowed in VC6)
	static_assert (sizeof (flags) <= sizeof (componentFlags), "");
   #endif

	componentListeners.call (&ComponentListener::componentBeingDeleted, *this);

	weakReferenceMaster.clear();

	while (childComponentList.size() > 0)
		removeChildComponent (childComponentList.size() - 1, false, true);

	if (parentComponent != nullptr)
		parentComponent->removeChildComponent (parentComponent->childComponentList.indexOf (this), true, false);
	else if (currentlyFocusedComponent == this || isParentOf (currentlyFocusedComponent))
		giveAwayFocus (currentlyFocusedComponent != this);

	if (flags.hasHeavyweightPeerFlag)
		removeFromDesktop();

	// Something has added some children to this component during its destructor! Not a smart idea!
	jassert (childComponentList.size() == 0);
}

const WeakReference<Component>::SharedRef& Component::getWeakReference()
{
	return weakReferenceMaster (this);
}

void Component::setName (const String& name)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (componentName != name)
	{
		componentName = name;

		if (flags.hasHeavyweightPeerFlag)
		{
			ComponentPeer* const peer = getPeer();

			jassert (peer != nullptr);
			if (peer != nullptr)
				peer->setTitle (name);
		}

		BailOutChecker checker (this);
		componentListeners.callChecked (checker, &ComponentListener::componentNameChanged, *this);
	}
}

void Component::setComponentID (const String& newID)
{
	componentID = newID;
}

void Component::setVisible (bool shouldBeVisible)
{
	if (flags.visibleFlag != shouldBeVisible)
	{
		// if component methods are being called from threads other than the message
		// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
		CHECK_MESSAGE_MANAGER_IS_LOCKED

		WeakReference<Component> safePointer (this);

		flags.visibleFlag = shouldBeVisible;

		internalRepaint (0, 0, getWidth(), getHeight());

		sendFakeMouseMove();

		if (! shouldBeVisible)
		{
			if (currentlyFocusedComponent == this || isParentOf (currentlyFocusedComponent))
			{
				if (parentComponent != nullptr)
					parentComponent->grabKeyboardFocus();
				else
					giveAwayFocus (true);
			}
		}

		if (safePointer != nullptr)
		{
			sendVisibilityChangeMessage();

			if (safePointer != nullptr && flags.hasHeavyweightPeerFlag)
			{
				ComponentPeer* const peer = getPeer();

				jassert (peer != nullptr);
				if (peer != nullptr)
				{
					peer->setVisible (shouldBeVisible);
					internalHierarchyChanged();
				}
			}
		}
	}
}

void Component::visibilityChanged()
{
}

void Component::sendVisibilityChangeMessage()
{
	BailOutChecker checker (this);

	visibilityChanged();

	if (! checker.shouldBailOut())
		componentListeners.callChecked (checker, &ComponentListener::componentVisibilityChanged, *this);
}

bool Component::isShowing() const
{
	if (flags.visibleFlag)
	{
		if (parentComponent != nullptr)
		{
			return parentComponent->isShowing();
		}
		else
		{
			const ComponentPeer* const peer = getPeer();

			return peer != nullptr && ! peer->isMinimised();
		}
	}

	return false;
}

void* Component::getWindowHandle() const
{
	const ComponentPeer* const peer = getPeer();

	if (peer != nullptr)
		return peer->getNativeHandle();

	return nullptr;
}

void Component::addToDesktop (int styleWanted, void* nativeWindowToAttachTo)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (isOpaque())
		styleWanted &= ~ComponentPeer::windowIsSemiTransparent;
	else
		styleWanted |= ComponentPeer::windowIsSemiTransparent;

	int currentStyleFlags = 0;

	// don't use getPeer(), so that we only get the peer that's specifically
	// for this comp, and not for one of its parents.
	ComponentPeer* peer = ComponentPeer::getPeerFor (this);

	if (peer != nullptr)
		currentStyleFlags = peer->getStyleFlags();

	if (styleWanted != currentStyleFlags || ! flags.hasHeavyweightPeerFlag)
	{
		WeakReference<Component> safePointer (this);

	   #if JUCE_LINUX
		// it's wise to give the component a non-zero size before
		// putting it on the desktop, as X windows get confused by this, and
		// a (1, 1) minimum size is enforced here.
		setSize (jmax (1, getWidth()),
				 jmax (1, getHeight()));
	   #endif

		const Point<int> topLeft (getScreenPosition());

		bool wasFullscreen = false;
		bool wasMinimised = false;
		ComponentBoundsConstrainer* currentConstainer = nullptr;
		Rectangle<int> oldNonFullScreenBounds;

		if (peer != nullptr)
		{
			wasFullscreen = peer->isFullScreen();
			wasMinimised = peer->isMinimised();
			currentConstainer = peer->getConstrainer();
			oldNonFullScreenBounds = peer->getNonFullScreenBounds();

			removeFromDesktop();

			setTopLeftPosition (topLeft.getX(), topLeft.getY());
		}

		if (parentComponent != nullptr)
			parentComponent->removeChildComponent (this);

		if (safePointer != nullptr)
		{
			flags.hasHeavyweightPeerFlag = true;

			peer = createNewPeer (styleWanted, nativeWindowToAttachTo);

			Desktop::getInstance().addDesktopComponent (this);

			bounds.setPosition (topLeft);
			peer->setBounds (topLeft.getX(), topLeft.getY(), getWidth(), getHeight(), false);

			peer->setVisible (isVisible());

			if (wasFullscreen)
			{
				peer->setFullScreen (true);
				peer->setNonFullScreenBounds (oldNonFullScreenBounds);
			}

			if (wasMinimised)
				peer->setMinimised (true);

			if (isAlwaysOnTop())
				peer->setAlwaysOnTop (true);

			peer->setConstrainer (currentConstainer);

			repaint();
		}

		internalHierarchyChanged();
	}
}

void Component::removeFromDesktop()
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (flags.hasHeavyweightPeerFlag)
	{
		ComponentPeer* const peer = ComponentPeer::getPeerFor (this);

		flags.hasHeavyweightPeerFlag = false;

		jassert (peer != nullptr);
		delete peer;

		Desktop::getInstance().removeDesktopComponent (this);
	}
}

bool Component::isOnDesktop() const noexcept
{
	return flags.hasHeavyweightPeerFlag;
}

void Component::userTriedToCloseWindow()
{
	/* This means that the user's trying to get rid of your window with the 'close window' system
	   menu option (on windows) or possibly the task manager - you should really handle this
	   and delete or hide your component in an appropriate way.

	   If you want to ignore the event and don't want to trigger this assertion, just override
	   this method and do nothing.
	*/
	jassertfalse;
}

void Component::minimisationStateChanged (bool)
{
}

void Component::setOpaque (const bool shouldBeOpaque)
{
	if (shouldBeOpaque != flags.opaqueFlag)
	{
		flags.opaqueFlag = shouldBeOpaque;

		if (flags.hasHeavyweightPeerFlag)
		{
			const ComponentPeer* const peer = ComponentPeer::getPeerFor (this);

			if (peer != nullptr)
			{
				// to make it recreate the heavyweight window
				addToDesktop (peer->getStyleFlags());
			}
		}

		repaint();
	}
}

bool Component::isOpaque() const noexcept
{
	return flags.opaqueFlag;
}

void Component::setBufferedToImage (const bool shouldBeBuffered)
{
	if (shouldBeBuffered != flags.bufferToImageFlag)
	{
		bufferedImage = Image::null;
		flags.bufferToImageFlag = shouldBeBuffered;
	}
}

void Component::moveChildInternal (const int sourceIndex, const int destIndex)
{
	if (sourceIndex != destIndex)
	{
		Component* const c = childComponentList.getUnchecked (sourceIndex);
		jassert (c != nullptr);
		c->repaintParent();

		childComponentList.move (sourceIndex, destIndex);

		sendFakeMouseMove();
		internalChildrenChanged();
	}
}

void Component::toFront (const bool setAsForeground)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (flags.hasHeavyweightPeerFlag)
	{
		ComponentPeer* const peer = getPeer();

		if (peer != nullptr)
		{
			peer->toFront (setAsForeground);

			if (setAsForeground && ! hasKeyboardFocus (true))
				grabKeyboardFocus();
		}
	}
	else if (parentComponent != nullptr)
	{
		const Array<Component*>& childList = parentComponent->childComponentList;

		if (childList.getLast() != this)
		{
			const int index = childList.indexOf (this);

			if (index >= 0)
			{
				int insertIndex = -1;

				if (! flags.alwaysOnTopFlag)
				{
					insertIndex = childList.size() - 1;

					while (insertIndex > 0 && childList.getUnchecked (insertIndex)->isAlwaysOnTop())
						--insertIndex;
				}

				parentComponent->moveChildInternal (index, insertIndex);
			}
		}

		if (setAsForeground)
		{
			internalBroughtToFront();
			grabKeyboardFocus();
		}
	}
}

void Component::toBehind (Component* const other)
{
	if (other != nullptr && other != this)
	{
		// the two components must belong to the same parent..
		jassert (parentComponent == other->parentComponent);

		if (parentComponent != nullptr)
		{
			const Array<Component*>& childList = parentComponent->childComponentList;
			const int index = childList.indexOf (this);

			if (index >= 0 && childList [index + 1] != other)
			{
				int otherIndex = childList.indexOf (other);

				if (otherIndex >= 0)
				{
					if (index < otherIndex)
						--otherIndex;

					parentComponent->moveChildInternal (index, otherIndex);
				}
			}
		}
		else if (isOnDesktop())
		{
			jassert (other->isOnDesktop());

			if (other->isOnDesktop())
			{
				ComponentPeer* const us = getPeer();
				ComponentPeer* const them = other->getPeer();

				jassert (us != nullptr && them != nullptr);
				if (us != nullptr && them != nullptr)
					us->toBehind (them);
			}
		}
	}
}

void Component::toBack()
{
	if (isOnDesktop())
	{
		jassertfalse; //xxx need to add this to native window
	}
	else if (parentComponent != nullptr)
	{
		const Array<Component*>& childList = parentComponent->childComponentList;

		if (childList.getFirst() != this)
		{
			const int index = childList.indexOf (this);

			if (index > 0)
			{
				int insertIndex = 0;

				if (flags.alwaysOnTopFlag)
					while (insertIndex < childList.size() && ! childList.getUnchecked (insertIndex)->isAlwaysOnTop())
						++insertIndex;

				parentComponent->moveChildInternal (index, insertIndex);
			}
		}
	}
}

void Component::setAlwaysOnTop (const bool shouldStayOnTop)
{
	if (shouldStayOnTop != flags.alwaysOnTopFlag)
	{
		BailOutChecker checker (this);

		flags.alwaysOnTopFlag = shouldStayOnTop;

		if (isOnDesktop())
		{
			ComponentPeer* const peer = getPeer();

			jassert (peer != nullptr);
			if (peer != nullptr)
			{
				if (! peer->setAlwaysOnTop (shouldStayOnTop))
				{
					// some kinds of peer can't change their always-on-top status, so
					// for these, we'll need to create a new window
					const int oldFlags = peer->getStyleFlags();
					removeFromDesktop();
					addToDesktop (oldFlags);
				}
			}
		}

		if (shouldStayOnTop && ! checker.shouldBailOut())
			toFront (false);

		if (! checker.shouldBailOut())
			internalHierarchyChanged();
	}
}

bool Component::isAlwaysOnTop() const noexcept
{
	return flags.alwaysOnTopFlag;
}

int Component::proportionOfWidth (const float proportion) const noexcept
{
	return roundToInt (proportion * bounds.getWidth());
}

int Component::proportionOfHeight (const float proportion) const noexcept
{
	return roundToInt (proportion * bounds.getHeight());
}

int Component::getParentWidth() const noexcept
{
	return parentComponent != nullptr ? parentComponent->getWidth()
									  : getParentMonitorArea().getWidth();
}

int Component::getParentHeight() const noexcept
{
	return parentComponent != nullptr ? parentComponent->getHeight()
									  : getParentMonitorArea().getHeight();
}

int Component::getScreenX() const   { return getScreenPosition().getX(); }
int Component::getScreenY() const   { return getScreenPosition().getY(); }

Point<int> Component::getScreenPosition() const	   { return localPointToGlobal (Point<int>()); }
Rectangle<int> Component::getScreenBounds() const	 { return localAreaToGlobal (getLocalBounds()); }

Point<int> Component::getLocalPoint (const Component* source, const Point<int>& point) const
{
	return ComponentHelpers::convertCoordinate (this, source, point);
}

Rectangle<int> Component::getLocalArea (const Component* source, const Rectangle<int>& area) const
{
	return ComponentHelpers::convertCoordinate (this, source, area);
}

Point<int> Component::localPointToGlobal (const Point<int>& point) const
{
	return ComponentHelpers::convertCoordinate (nullptr, this, point);
}

Rectangle<int> Component::localAreaToGlobal (const Rectangle<int>& area) const
{
	return ComponentHelpers::convertCoordinate (nullptr, this, area);
}

/* Deprecated methods... */
const Point<int> Component::relativePositionToGlobal (const Point<int>& relativePosition) const
{
	return localPointToGlobal (relativePosition);
}

const Point<int> Component::globalPositionToRelative (const Point<int>& screenPosition) const
{
	return getLocalPoint (nullptr, screenPosition);
}

const Point<int> Component::relativePositionToOtherComponent (const Component* const targetComponent, const Point<int>& positionRelativeToThis) const
{
	return targetComponent == nullptr ? localPointToGlobal (positionRelativeToThis)
									  : targetComponent->getLocalPoint (this, positionRelativeToThis);
}

void Component::setBounds (const int x, const int y, int w, int h)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (w < 0) w = 0;
	if (h < 0) h = 0;

	const bool wasResized  = (getWidth() != w || getHeight() != h);
	const bool wasMoved	= (getX() != x || getY() != y);

   #if JUCE_DEBUG
	// It's a very bad idea to try to resize a window during its paint() method!
	jassert (! (flags.isInsidePaintCall && wasResized && isOnDesktop()));
   #endif

	if (wasMoved || wasResized)
	{
		const bool showing = isShowing();
		if (showing)
		{
			// send a fake mouse move to trigger enter/exit messages if needed..
			sendFakeMouseMove();

			if (! flags.hasHeavyweightPeerFlag)
				repaintParent();
		}

		bounds.setBounds (x, y, w, h);

		if (showing)
		{
			if (wasResized)
				repaint();
			else if (! flags.hasHeavyweightPeerFlag)
				repaintParent();
		}
		else
		{
			bufferedImage = Image::null;
		}

		if (flags.hasHeavyweightPeerFlag)
		{
			ComponentPeer* const peer = getPeer();

			if (peer != nullptr)
			{
				if (wasMoved && wasResized)
					peer->setBounds (getX(), getY(), getWidth(), getHeight(), false);
				else if (wasMoved)
					peer->setPosition (getX(), getY());
				else if (wasResized)
					peer->setSize (getWidth(), getHeight());
			}
		}

		sendMovedResizedMessages (wasMoved, wasResized);
	}
}

void Component::sendMovedResizedMessages (const bool wasMoved, const bool wasResized)
{
	BailOutChecker checker (this);

	if (wasMoved)
	{
		moved();

		if (checker.shouldBailOut())
			return;
	}

	if (wasResized)
	{
		resized();

		if (checker.shouldBailOut())
			return;

		for (int i = childComponentList.size(); --i >= 0;)
		{
			childComponentList.getUnchecked(i)->parentSizeChanged();

			if (checker.shouldBailOut())
				return;

			i = jmin (i, childComponentList.size());
		}
	}

	if (parentComponent != nullptr)
		parentComponent->childBoundsChanged (this);

	if (! checker.shouldBailOut())
		componentListeners.callChecked (checker, &ComponentListener::componentMovedOrResized,
										*this, wasMoved, wasResized);
}

void Component::setSize (const int w, const int h)
{
	setBounds (getX(), getY(), w, h);
}

void Component::setTopLeftPosition (const int x, const int y)
{
	setBounds (x, y, getWidth(), getHeight());
}

void Component::setTopRightPosition (const int x, const int y)
{
	setTopLeftPosition (x - getWidth(), y);
}

void Component::setBounds (const Rectangle<int>& r)
{
	setBounds (r.getX(), r.getY(), r.getWidth(), r.getHeight());
}

void Component::setBounds (const RelativeRectangle& newBounds)
{
	newBounds.applyToComponent (*this);
}

void Component::setBounds (const String& newBoundsExpression)
{
	setBounds (RelativeRectangle (newBoundsExpression));
}

void Component::setBoundsRelative (const float x, const float y,
								   const float w, const float h)
{
	const int pw = getParentWidth();
	const int ph = getParentHeight();

	setBounds (roundToInt (x * pw),
			   roundToInt (y * ph),
			   roundToInt (w * pw),
			   roundToInt (h * ph));
}

void Component::setCentrePosition (const int x, const int y)
{
	setTopLeftPosition (x - getWidth() / 2,
						y - getHeight() / 2);
}

void Component::setCentreRelative (const float x, const float y)
{
	setCentrePosition (roundToInt (getParentWidth() * x),
					   roundToInt (getParentHeight() * y));
}

void Component::centreWithSize (const int width, const int height)
{
	const Rectangle<int> parentArea (ComponentHelpers::getParentOrMainMonitorBounds (*this));

	setBounds (parentArea.getCentreX() - width / 2,
			   parentArea.getCentreY() - height / 2,
			   width, height);
}

void Component::setBoundsInset (const BorderSize<int>& borders)
{
	setBounds (borders.subtractedFrom (ComponentHelpers::getParentOrMainMonitorBounds (*this)));
}

void Component::setBoundsToFit (int x, int y, int width, int height,
								const Justification& justification,
								const bool onlyReduceInSize)
{
	// it's no good calling this method unless both the component and
	// target rectangle have a finite size.
	jassert (getWidth() > 0 && getHeight() > 0 && width > 0 && height > 0);

	if (getWidth() > 0 && getHeight() > 0
		 && width > 0 && height > 0)
	{
		int newW, newH;

		if (onlyReduceInSize && getWidth() <= width && getHeight() <= height)
		{
			newW = getWidth();
			newH = getHeight();
		}
		else
		{
			const double imageRatio = getHeight() / (double) getWidth();
			const double targetRatio = height / (double) width;

			if (imageRatio <= targetRatio)
			{
				newW = width;
				newH = jmin (height, roundToInt (newW * imageRatio));
			}
			else
			{
				newH = height;
				newW = jmin (width, roundToInt (newH / imageRatio));
			}
		}

		if (newW > 0 && newH > 0)
			setBounds (justification.appliedToRectangle (Rectangle<int> (0, 0, newW, newH),
														 Rectangle<int> (x, y, width, height)));
	}
}

bool Component::isTransformed() const noexcept
{
	return affineTransform != nullptr;
}

void Component::setTransform (const AffineTransform& newTransform)
{
	// If you pass in a transform with no inverse, the component will have no dimensions,
	// and there will be all sorts of maths errors when converting coordinates.
	jassert (! newTransform.isSingularity());

	if (newTransform.isIdentity())
	{
		if (affineTransform != nullptr)
		{
			repaint();
			affineTransform = nullptr;
			repaint();

			sendMovedResizedMessages (false, false);
		}
	}
	else if (affineTransform == nullptr)
	{
		repaint();
		affineTransform = new AffineTransform (newTransform);
		repaint();
		sendMovedResizedMessages (false, false);
	}
	else if (*affineTransform != newTransform)
	{
		repaint();
		*affineTransform = newTransform;
		repaint();
		sendMovedResizedMessages (false, false);
	}
}

AffineTransform Component::getTransform() const
{
	return affineTransform != nullptr ? *affineTransform : AffineTransform::identity;
}

bool Component::hitTest (int x, int y)
{
	if (! flags.ignoresMouseClicksFlag)
		return true;

	if (flags.allowChildMouseClicksFlag)
	{
		for (int i = getNumChildComponents(); --i >= 0;)
		{
			Component& child = *getChildComponent (i);

			if (child.isVisible()
				 && ComponentHelpers::hitTest (child, ComponentHelpers::convertFromParentSpace (child, Point<int> (x, y))))
				return true;
		}
	}

	return false;
}

void Component::setInterceptsMouseClicks (const bool allowClicks,
										  const bool allowClicksOnChildComponents) noexcept
{
	flags.ignoresMouseClicksFlag = ! allowClicks;
	flags.allowChildMouseClicksFlag = allowClicksOnChildComponents;
}

void Component::getInterceptsMouseClicks (bool& allowsClicksOnThisComponent,
										  bool& allowsClicksOnChildComponents) const noexcept
{
	allowsClicksOnThisComponent = ! flags.ignoresMouseClicksFlag;
	allowsClicksOnChildComponents = flags.allowChildMouseClicksFlag;
}

bool Component::contains (const Point<int>& point)
{
	if (ComponentHelpers::hitTest (*this, point))
	{
		if (parentComponent != nullptr)
		{
			return parentComponent->contains (ComponentHelpers::convertToParentSpace (*this, point));
		}
		else if (flags.hasHeavyweightPeerFlag)
		{
			const ComponentPeer* const peer = getPeer();

			if (peer != nullptr)
				return peer->contains (point, true);
		}
	}

	return false;
}

bool Component::reallyContains (const Point<int>& point, const bool returnTrueIfWithinAChild)
{
	if (! contains (point))
		return false;

	Component* const top = getTopLevelComponent();
	const Component* const compAtPosition = top->getComponentAt (top->getLocalPoint (this, point));

	return (compAtPosition == this) || (returnTrueIfWithinAChild && isParentOf (compAtPosition));
}

Component* Component::getComponentAt (const Point<int>& position)
{
	if (flags.visibleFlag && ComponentHelpers::hitTest (*this, position))
	{
		for (int i = childComponentList.size(); --i >= 0;)
		{
			Component* child = childComponentList.getUnchecked(i);
			child = child->getComponentAt (ComponentHelpers::convertFromParentSpace (*child, position));

			if (child != nullptr)
				return child;
		}

		return this;
	}

	return nullptr;
}

Component* Component::getComponentAt (const int x, const int y)
{
	return getComponentAt (Point<int> (x, y));
}

void Component::addChildComponent (Component* const child, int zOrder)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (child != nullptr && child->parentComponent != this)
	{
		if (child->parentComponent != nullptr)
			child->parentComponent->removeChildComponent (child);
		else
			child->removeFromDesktop();

		child->parentComponent = this;

		if (child->isVisible())
			child->repaintParent();

		if (! child->isAlwaysOnTop())
		{
			if (zOrder < 0 || zOrder > childComponentList.size())
				zOrder = childComponentList.size();

			while (zOrder > 0)
			{
				if (! childComponentList.getUnchecked (zOrder - 1)->isAlwaysOnTop())
					break;

				--zOrder;
			}
		}

		childComponentList.insert (zOrder, child);

		child->internalHierarchyChanged();
		internalChildrenChanged();
	}
}

void Component::addAndMakeVisible (Component* const child, int zOrder)
{
	if (child != nullptr)
	{
		child->setVisible (true);
		addChildComponent (child, zOrder);
	}
}

void Component::removeChildComponent (Component* const child)
{
	removeChildComponent (childComponentList.indexOf (child), true, true);
}

Component* Component::removeChildComponent (const int index)
{
	return removeChildComponent (index, true, true);
}

Component* Component::removeChildComponent (const int index, bool sendParentEvents, const bool sendChildEvents)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	Component* const child = childComponentList [index];

	if (child != nullptr)
	{
		sendParentEvents = sendParentEvents && child->isShowing();

		if (sendParentEvents)
		{
			sendFakeMouseMove();
			child->repaintParent();
		}

		childComponentList.remove (index);
		child->parentComponent = nullptr;

		// (NB: there are obscure situations where child->isShowing() = false, but it still has the focus)
		if (currentlyFocusedComponent == child || child->isParentOf (currentlyFocusedComponent))
		{
			if (sendParentEvents)
			{
				const WeakReference<Component> thisPointer (this);

				giveAwayFocus (sendChildEvents || currentlyFocusedComponent != child);

				if (thisPointer == nullptr)
					return child;

				grabKeyboardFocus();
			}
			else
			{
				giveAwayFocus (sendChildEvents || currentlyFocusedComponent != child);
			}
		}

		if (sendChildEvents)
			child->internalHierarchyChanged();

		if (sendParentEvents)
			internalChildrenChanged();
	}

	return child;
}

void Component::removeAllChildren()
{
	while (childComponentList.size() > 0)
		removeChildComponent (childComponentList.size() - 1);
}

void Component::deleteAllChildren()
{
	while (childComponentList.size() > 0)
		delete (removeChildComponent (childComponentList.size() - 1));
}

int Component::getNumChildComponents() const noexcept
{
	return childComponentList.size();
}

Component* Component::getChildComponent (const int index) const noexcept
{
	return childComponentList [index];
}

int Component::getIndexOfChildComponent (const Component* const child) const noexcept
{
	return childComponentList.indexOf (const_cast <Component*> (child));
}

Component* Component::getTopLevelComponent() const noexcept
{
	const Component* comp = this;

	while (comp->parentComponent != nullptr)
		comp = comp->parentComponent;

	return const_cast <Component*> (comp);
}

bool Component::isParentOf (const Component* possibleChild) const noexcept
{
	while (possibleChild != nullptr)
	{
		possibleChild = possibleChild->parentComponent;

		if (possibleChild == this)
			return true;
	}

	return false;
}

void Component::parentHierarchyChanged()
{
}

void Component::childrenChanged()
{
}

void Component::internalChildrenChanged()
{
	if (componentListeners.isEmpty())
	{
		childrenChanged();
	}
	else
	{
		BailOutChecker checker (this);

		childrenChanged();

		if (! checker.shouldBailOut())
			componentListeners.callChecked (checker, &ComponentListener::componentChildrenChanged, *this);
	}
}

void Component::internalHierarchyChanged()
{
	BailOutChecker checker (this);

	parentHierarchyChanged();

	if (checker.shouldBailOut())
		return;

	componentListeners.callChecked (checker, &ComponentListener::componentParentHierarchyChanged, *this);

	if (checker.shouldBailOut())
		return;

	for (int i = childComponentList.size(); --i >= 0;)
	{
		childComponentList.getUnchecked (i)->internalHierarchyChanged();

		if (checker.shouldBailOut())
		{
			// you really shouldn't delete the parent component during a callback telling you
			// that it's changed..
			jassertfalse;
			return;
		}

		i = jmin (i, childComponentList.size());
	}
}

#if JUCE_MODAL_LOOPS_PERMITTED
int Component::runModalLoop()
{
	if (! MessageManager::getInstance()->isThisTheMessageThread())
	{
		// use a callback so this can be called from non-gui threads
		return (int) (pointer_sized_int) MessageManager::getInstance()
										   ->callFunctionOnMessageThread (&ComponentHelpers::runModalLoopCallback, this);
	}

	if (! isCurrentlyModal())
		enterModalState (true);

	return ModalComponentManager::getInstance()->runEventLoopForCurrentComponent();
}
#endif

class ModalAutoDeleteCallback   : public ModalComponentManager::Callback
{
public:
	ModalAutoDeleteCallback (Component* const comp_)
		: comp (comp_)
	{}

	void modalStateFinished (int)
	{
		delete comp.get();
	}

private:
	WeakReference<Component> comp;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ModalAutoDeleteCallback);
};

void Component::enterModalState (const bool shouldTakeKeyboardFocus,
								 ModalComponentManager::Callback* callback,
								 const bool deleteWhenDismissed)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	// Check for an attempt to make a component modal when it already is!
	// This can cause nasty problems..
	jassert (! flags.currentlyModalFlag);

	if (! isCurrentlyModal())
	{
		ModalComponentManager* const mcm = ModalComponentManager::getInstance();
		mcm->startModal (this);
		mcm->attachCallback (this, callback);

		if (deleteWhenDismissed)
			mcm->attachCallback (this, new ModalAutoDeleteCallback (this));

		flags.currentlyModalFlag = true;
		setVisible (true);

		if (shouldTakeKeyboardFocus)
			grabKeyboardFocus();
	}
}

void Component::exitModalState (const int returnValue)
{
	if (flags.currentlyModalFlag)
	{
		if (MessageManager::getInstance()->isThisTheMessageThread())
		{
			ModalComponentManager::getInstance()->endModal (this, returnValue);
			flags.currentlyModalFlag = false;

			ModalComponentManager::getInstance()->bringModalComponentsToFront();
		}
		else
		{
			class ExitModalStateMessage   : public CallbackMessage
			{
			public:
				ExitModalStateMessage (Component* const target_, const int result_)
					: target (target_), result (result_)   {}

				void messageCallback()
				{
					if (target.get() != nullptr) // (get() required for VS2003 bug)
						target->exitModalState (result);
				}

			private:
				WeakReference<Component> target;
				int result;
			};

			(new ExitModalStateMessage (this, returnValue))->post();
		}
	}
}

bool Component::isCurrentlyModal() const noexcept
{
	return flags.currentlyModalFlag
			&& getCurrentlyModalComponent() == this;
}

bool Component::isCurrentlyBlockedByAnotherModalComponent() const
{
	Component* const mc = getCurrentlyModalComponent();

	return ! (mc == nullptr || mc == this || mc->isParentOf (this)
			   || mc->canModalEventBeSentToComponent (this));
}

int JUCE_CALLTYPE Component::getNumCurrentlyModalComponents() noexcept
{
	return ModalComponentManager::getInstance()->getNumModalComponents();
}

Component* JUCE_CALLTYPE Component::getCurrentlyModalComponent (int index) noexcept
{
	return ModalComponentManager::getInstance()->getModalComponent (index);
}

void Component::setBroughtToFrontOnMouseClick (const bool shouldBeBroughtToFront) noexcept
{
	flags.bringToFrontOnClickFlag = shouldBeBroughtToFront;
}

bool Component::isBroughtToFrontOnMouseClick() const noexcept
{
	return flags.bringToFrontOnClickFlag;
}

void Component::setMouseCursor (const MouseCursor& newCursor)
{
	if (cursor != newCursor)
	{
		cursor = newCursor;

		if (flags.visibleFlag)
			updateMouseCursor();
	}
}

const MouseCursor Component::getMouseCursor()
{
	return cursor;
}

void Component::updateMouseCursor() const
{
	Desktop::getInstance().getMainMouseSource().forceMouseCursorUpdate();
}

void Component::setRepaintsOnMouseActivity (const bool shouldRepaint) noexcept
{
	flags.repaintOnMouseActivityFlag = shouldRepaint;
}

void Component::setAlpha (const float newAlpha)
{
	const uint8 newIntAlpha = (uint8) (255 - jlimit (0, 255, roundToInt (newAlpha * 255.0)));

	if (componentTransparency != newIntAlpha)
	{
		componentTransparency = newIntAlpha;

		if (flags.hasHeavyweightPeerFlag)
		{
			ComponentPeer* const peer = getPeer();

			if (peer != nullptr)
				peer->setAlpha (newAlpha);
		}
		else
		{
			repaint();
		}
	}
}

float Component::getAlpha() const
{
	return (255 - componentTransparency) / 255.0f;
}

void Component::repaintParent()
{
	if (flags.visibleFlag)
		internalRepaint (0, 0, getWidth(), getHeight());
}

void Component::repaint()
{
	repaint (0, 0, getWidth(), getHeight());
}

void Component::repaint (const int x, const int y,
						 const int w, const int h)
{
	bufferedImage = Image::null;

	if (flags.visibleFlag)
		internalRepaint (x, y, w, h);
}

void Component::repaint (const Rectangle<int>& area)
{
	repaint (area.getX(), area.getY(), area.getWidth(), area.getHeight());
}

void Component::internalRepaint (int x, int y, int w, int h)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (x < 0)
	{
		w += x;
		x = 0;
	}

	if (x + w > getWidth())
		w = getWidth() - x;

	if (w > 0)
	{
		if (y < 0)
		{
			h += y;
			y = 0;
		}

		if (y + h > getHeight())
			h = getHeight() - y;

		if (h > 0)
		{
			if (parentComponent != nullptr)
			{
				if (parentComponent->flags.visibleFlag)
				{
					if (affineTransform == nullptr)
					{
						parentComponent->internalRepaint (x + getX(), y + getY(), w, h);
					}
					else
					{
						const Rectangle<int> r (ComponentHelpers::convertToParentSpace (*this, Rectangle<int> (x, y, w, h)));
						parentComponent->internalRepaint (r.getX(), r.getY(), r.getWidth(), r.getHeight());
					}
				}
			}
			else if (flags.hasHeavyweightPeerFlag)
			{
				ComponentPeer* const peer = getPeer();

				if (peer != nullptr)
					peer->repaint (Rectangle<int> (x, y, w, h));
			}
		}
	}
}

void Component::paintComponent (Graphics& g)
{
	if (flags.bufferToImageFlag)
	{
		if (bufferedImage.isNull())
		{
			bufferedImage = Image (flags.opaqueFlag ? Image::RGB : Image::ARGB,
								   getWidth(), getHeight(), ! flags.opaqueFlag, Image::NativeImage);

			Graphics imG (bufferedImage);
			paint (imG);
		}

		g.setColour (Colours::black);
		g.drawImageAt (bufferedImage, 0, 0);
	}
	else
	{
		paint (g);
	}
}

void Component::paintWithinParentContext (Graphics& g)
{
	g.setOrigin (getX(), getY());
	paintEntireComponent (g, false);
}

void Component::paintComponentAndChildren (Graphics& g)
{
	const Rectangle<int> clipBounds (g.getClipBounds());

	if (flags.dontClipGraphicsFlag)
	{
		paintComponent (g);
	}
	else
	{
		g.saveState();
		ComponentHelpers::clipObscuredRegions (*this, g, clipBounds, Point<int>());

		if (! g.isClipEmpty())
			paintComponent (g);

		g.restoreState();
	}

	for (int i = 0; i < childComponentList.size(); ++i)
	{
		Component& child = *childComponentList.getUnchecked (i);

		if (child.isVisible())
		{
			if (child.affineTransform != nullptr)
			{
				g.saveState();
				g.addTransform (*child.affineTransform);

				if ((child.flags.dontClipGraphicsFlag && ! g.isClipEmpty()) || g.reduceClipRegion (child.getBounds()))
					child.paintWithinParentContext (g);

				g.restoreState();
			}
			else if (clipBounds.intersects (child.getBounds()))
			{
				g.saveState();

				if (child.flags.dontClipGraphicsFlag)
				{
					child.paintWithinParentContext (g);
				}
				else if (g.reduceClipRegion (child.getBounds()))
				{
					bool nothingClipped = true;

					for (int j = i + 1; j < childComponentList.size(); ++j)
					{
						const Component& sibling = *childComponentList.getUnchecked (j);

						if (sibling.flags.opaqueFlag && sibling.isVisible() && sibling.affineTransform == nullptr)
						{
							nothingClipped = false;
							g.excludeClipRegion (sibling.getBounds());
						}
					}

					if (nothingClipped || ! g.isClipEmpty())
						child.paintWithinParentContext (g);
				}

				g.restoreState();
			}
		}
	}

	g.saveState();
	paintOverChildren (g);
	g.restoreState();
}

void Component::paintEntireComponent (Graphics& g, const bool ignoreAlphaLevel)
{
	jassert (! g.isClipEmpty());

   #if JUCE_DEBUG
	flags.isInsidePaintCall = true;
   #endif

	if (effect != nullptr)
	{
		Image effectImage (flags.opaqueFlag ? Image::RGB : Image::ARGB,
						   getWidth(), getHeight(), ! flags.opaqueFlag, Image::NativeImage);
		{
			Graphics g2 (effectImage);
			paintComponentAndChildren (g2);
		}

		effect->applyEffect (effectImage, g, ignoreAlphaLevel ? 1.0f : getAlpha());
	}
	else if (componentTransparency > 0 && ! ignoreAlphaLevel)
	{
		if (componentTransparency < 255)
		{
			g.beginTransparencyLayer (getAlpha());
			paintComponentAndChildren (g);
			g.endTransparencyLayer();
		}
	}
	else
	{
		paintComponentAndChildren (g);
	}

   #if JUCE_DEBUG
	flags.isInsidePaintCall = false;
   #endif
}

void Component::setPaintingIsUnclipped (const bool shouldPaintWithoutClipping) noexcept
{
	flags.dontClipGraphicsFlag = shouldPaintWithoutClipping;
}

Image Component::createComponentSnapshot (const Rectangle<int>& areaToGrab,
										  const bool clipImageToComponentBounds)
{
	Rectangle<int> r (areaToGrab);

	if (clipImageToComponentBounds)
		r = r.getIntersection (getLocalBounds());

	Image componentImage (flags.opaqueFlag ? Image::RGB : Image::ARGB,
						  jmax (1, r.getWidth()),
						  jmax (1, r.getHeight()),
						  true);

	Graphics imageContext (componentImage);
	imageContext.setOrigin (-r.getX(), -r.getY());
	paintEntireComponent (imageContext, true);

	return componentImage;
}

void Component::setComponentEffect (ImageEffectFilter* const newEffect)
{
	if (effect != newEffect)
	{
		effect = newEffect;
		repaint();
	}
}

LookAndFeel& Component::getLookAndFeel() const noexcept
{
	const Component* c = this;

	do
	{
		if (c->lookAndFeel != nullptr)
			return *(c->lookAndFeel);

		c = c->parentComponent;
	}
	while (c != nullptr);

	return LookAndFeel::getDefaultLookAndFeel();
}

void Component::setLookAndFeel (LookAndFeel* const newLookAndFeel)
{
	if (lookAndFeel != newLookAndFeel)
	{
		lookAndFeel = newLookAndFeel;

		sendLookAndFeelChange();
	}
}

void Component::lookAndFeelChanged()
{
}

void Component::sendLookAndFeelChange()
{
	repaint();

	WeakReference<Component> safePointer (this);

	lookAndFeelChanged();

	if (safePointer != nullptr)
	{
		for (int i = childComponentList.size(); --i >= 0;)
		{
			childComponentList.getUnchecked (i)->sendLookAndFeelChange();

			if (safePointer == nullptr)
				return;

			i = jmin (i, childComponentList.size());
		}
	}
}

const Colour Component::findColour (const int colourId, const bool inheritFromParent) const
{
	const var* const v = properties.getVarPointer (ComponentHelpers::getColourPropertyId (colourId));

	if (v != nullptr)
		return Colour ((int) *v);

	if (inheritFromParent && parentComponent != nullptr
		 && (lookAndFeel == nullptr || ! lookAndFeel->isColourSpecified (colourId)))
		return parentComponent->findColour (colourId, true);

	return getLookAndFeel().findColour (colourId);
}

bool Component::isColourSpecified (const int colourId) const
{
	return properties.contains (ComponentHelpers::getColourPropertyId (colourId));
}

void Component::removeColour (const int colourId)
{
	if (properties.remove (ComponentHelpers::getColourPropertyId (colourId)))
		colourChanged();
}

void Component::setColour (const int colourId, const Colour& colour)
{
	if (properties.set (ComponentHelpers::getColourPropertyId (colourId), (int) colour.getARGB()))
		colourChanged();
}

void Component::copyAllExplicitColoursTo (Component& target) const
{
	bool changed = false;

	for (int i = properties.size(); --i >= 0;)
	{
		const Identifier name (properties.getName(i));

		if (name.toString().startsWith ("jcclr_"))
			if (target.properties.set (name, properties [name]))
				changed = true;
	}

	if (changed)
		target.colourChanged();
}

void Component::colourChanged()
{
}

MarkerList* Component::getMarkers (bool /*xAxis*/)
{
	return nullptr;
}

Component::Positioner::Positioner (Component& component_) noexcept
	: component (component_)
{
}

Component::Positioner* Component::getPositioner() const noexcept
{
	return positioner;
}

void Component::setPositioner (Positioner* newPositioner)
{
	// You can only assign a positioner to the component that it was created for!
	jassert (newPositioner == nullptr || this == &(newPositioner->getComponent()));
	positioner = newPositioner;
}

Rectangle<int> Component::getLocalBounds() const noexcept
{
	return Rectangle<int> (getWidth(), getHeight());
}

Rectangle<int> Component::getBoundsInParent() const noexcept
{
	return affineTransform == nullptr ? bounds
									  : bounds.toFloat().transformed (*affineTransform).getSmallestIntegerContainer();
}

void Component::getVisibleArea (RectangleList& result, const bool includeSiblings) const
{
	result.clear();
	const Rectangle<int> unclipped (ComponentHelpers::getUnclippedArea (*this));

	if (! unclipped.isEmpty())
	{
		result.add (unclipped);

		if (includeSiblings)
		{
			const Component* const c = getTopLevelComponent();

			ComponentHelpers::subtractObscuredRegions (*c, result, getLocalPoint (c, Point<int>()),
													   c->getLocalBounds(), this);
		}

		ComponentHelpers::subtractObscuredRegions (*this, result, Point<int>(), unclipped, nullptr);
		result.consolidate();
	}
}

void Component::mouseEnter (const MouseEvent&)
{
	// base class does nothing
}

void Component::mouseExit (const MouseEvent&)
{
	// base class does nothing
}

void Component::mouseDown (const MouseEvent&)
{
	// base class does nothing
}

void Component::mouseUp (const MouseEvent&)
{
	// base class does nothing
}

void Component::mouseDrag (const MouseEvent&)
{
	// base class does nothing
}

void Component::mouseMove (const MouseEvent&)
{
	// base class does nothing
}

void Component::mouseDoubleClick (const MouseEvent&)
{
	// base class does nothing
}

void Component::mouseWheelMove (const MouseEvent& e, float wheelIncrementX, float wheelIncrementY)
{
	// the base class just passes this event up to its parent..

	if (parentComponent != nullptr)
		parentComponent->mouseWheelMove (e.getEventRelativeTo (parentComponent),
										 wheelIncrementX, wheelIncrementY);
}

void Component::resized()
{
	// base class does nothing
}

void Component::moved()
{
	// base class does nothing
}

void Component::childBoundsChanged (Component*)
{
	// base class does nothing
}

void Component::parentSizeChanged()
{
	// base class does nothing
}

void Component::addComponentListener (ComponentListener* const newListener)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	componentListeners.add (newListener);
}

void Component::removeComponentListener (ComponentListener* const listenerToRemove)
{
	componentListeners.remove (listenerToRemove);
}

void Component::inputAttemptWhenModal()
{
	ModalComponentManager::getInstance()->bringModalComponentsToFront();
	getLookAndFeel().playAlertSound();
}

bool Component::canModalEventBeSentToComponent (const Component*)
{
	return false;
}

void Component::internalModalInputAttempt()
{
	Component* const current = getCurrentlyModalComponent();

	if (current != nullptr)
		current->inputAttemptWhenModal();
}

void Component::paint (Graphics&)
{
	// all painting is done in the subclasses

	jassert (! isOpaque()); // if your component's opaque, you've gotta paint it!
}

void Component::paintOverChildren (Graphics&)
{
	// all painting is done in the subclasses
}

void Component::postCommandMessage (const int commandId)
{
	class CustomCommandMessage   : public CallbackMessage
	{
	public:
		CustomCommandMessage (Component* const target_, const int commandId_)
			: target (target_), commandId (commandId_) {}

		void messageCallback()
		{
			if (target.get() != nullptr)  // (get() required for VS2003 bug)
				target->handleCommandMessage (commandId);
		}

	private:
		WeakReference<Component> target;
		int commandId;
	};

	(new CustomCommandMessage (this, commandId))->post();
}

void Component::handleCommandMessage (int)
{
	// used by subclasses
}

void Component::addMouseListener (MouseListener* const newListener,
								  const bool wantsEventsForAllNestedChildComponents)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	// If you register a component as a mouselistener for itself, it'll receive all the events
	// twice - once via the direct callback that all components get anyway, and then again as a listener!
	jassert ((newListener != this) || wantsEventsForAllNestedChildComponents);

	if (mouseListeners == nullptr)
		mouseListeners = new MouseListenerList();

	mouseListeners->addListener (newListener, wantsEventsForAllNestedChildComponents);
}

void Component::removeMouseListener (MouseListener* const listenerToRemove)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (mouseListeners != nullptr)
		mouseListeners->removeListener (listenerToRemove);
}

void Component::internalMouseEnter (MouseInputSource& source, const Point<int>& relativePos, const Time& time)
{
	if (isCurrentlyBlockedByAnotherModalComponent())
	{
		// if something else is modal, always just show a normal mouse cursor
		source.showMouseCursor (MouseCursor::NormalCursor);
		return;
	}

	if (! flags.mouseInsideFlag)
	{
		flags.mouseInsideFlag = true;
		flags.mouseOverFlag = true;
		flags.mouseDownFlag = false;

		BailOutChecker checker (this);

		if (flags.repaintOnMouseActivityFlag)
			repaint();

		const MouseEvent me (source, relativePos, source.getCurrentModifiers(),
							 this, this, time, relativePos, time, 0, false);
		mouseEnter (me);

		if (checker.shouldBailOut())
			return;

		Desktop& desktop = Desktop::getInstance();
		desktop.resetTimer();
		desktop.mouseListeners.callChecked (checker, &MouseListener::mouseEnter, me);

		MouseListenerList::sendMouseEvent (*this, checker, &MouseListener::mouseEnter, me);
	}
}

void Component::internalMouseExit (MouseInputSource& source, const Point<int>& relativePos, const Time& time)
{
	BailOutChecker checker (this);

	if (flags.mouseDownFlag)
	{
		internalMouseUp (source, relativePos, time, source.getCurrentModifiers().getRawFlags());

		if (checker.shouldBailOut())
			return;
	}

	if (flags.mouseInsideFlag || flags.mouseOverFlag)
	{
		flags.mouseInsideFlag = false;
		flags.mouseOverFlag = false;
		flags.mouseDownFlag = false;

		if (flags.repaintOnMouseActivityFlag)
			repaint();

		const MouseEvent me (source, relativePos, source.getCurrentModifiers(),
							 this, this, time, relativePos, time, 0, false);
		mouseExit (me);

		if (checker.shouldBailOut())
			return;

		Desktop& desktop = Desktop::getInstance();
		desktop.resetTimer();
		desktop.mouseListeners.callChecked (checker, &MouseListener::mouseExit, me);

		MouseListenerList::sendMouseEvent (*this, checker, &MouseListener::mouseExit, me);
	}
}

void Component::internalMouseDown (MouseInputSource& source, const Point<int>& relativePos, const Time& time)
{
	Desktop& desktop = Desktop::getInstance();

	BailOutChecker checker (this);

	if (isCurrentlyBlockedByAnotherModalComponent())
	{
		internalModalInputAttempt();

		if (checker.shouldBailOut())
			return;

		// If processing the input attempt has exited the modal loop, we'll allow the event
		// to be delivered..
		if (isCurrentlyBlockedByAnotherModalComponent())
		{
			// allow blocked mouse-events to go to global listeners..
			const MouseEvent me (source, relativePos, source.getCurrentModifiers(),
								 this, this, time, relativePos, time,
								 source.getNumberOfMultipleClicks(), false);

			desktop.resetTimer();
			desktop.mouseListeners.callChecked (checker, &MouseListener::mouseDown, me);
			return;
		}
	}

	{
		Component* c = this;

		while (c != nullptr)
		{
			if (c->isBroughtToFrontOnMouseClick())
			{
				c->toFront (true);

				if (checker.shouldBailOut())
					return;
			}

			c = c->parentComponent;
		}
	}

	if (! flags.dontFocusOnMouseClickFlag)
	{
		grabFocusInternal (focusChangedByMouseClick);

		if (checker.shouldBailOut())
			return;
	}

	flags.mouseDownFlag = true;
	flags.mouseOverFlag = true;

	if (flags.repaintOnMouseActivityFlag)
		repaint();

	const MouseEvent me (source, relativePos, source.getCurrentModifiers(),
						 this, this, time, relativePos, time,
						 source.getNumberOfMultipleClicks(), false);
	mouseDown (me);

	if (checker.shouldBailOut())
		return;

	desktop.resetTimer();
	desktop.mouseListeners.callChecked (checker, &MouseListener::mouseDown, me);

	MouseListenerList::sendMouseEvent (*this, checker, &MouseListener::mouseDown, me);
}

void Component::internalMouseUp (MouseInputSource& source, const Point<int>& relativePos, const Time& time, const ModifierKeys& oldModifiers)
{
	if (flags.mouseDownFlag)
	{
		flags.mouseDownFlag = false;

		BailOutChecker checker (this);

		if (flags.repaintOnMouseActivityFlag)
			repaint();

		const MouseEvent me (source, relativePos,
							 oldModifiers, this, this, time,
							 getLocalPoint (nullptr, source.getLastMouseDownPosition()),
							 source.getLastMouseDownTime(),
							 source.getNumberOfMultipleClicks(),
							 source.hasMouseMovedSignificantlySincePressed());

		mouseUp (me);

		if (checker.shouldBailOut())
			return;

		Desktop& desktop = Desktop::getInstance();
		desktop.resetTimer();
		desktop.mouseListeners.callChecked (checker, &MouseListener::mouseUp, me);

		MouseListenerList::sendMouseEvent (*this, checker, &MouseListener::mouseUp, me);

		if (checker.shouldBailOut())
			return;

		// check for double-click
		if (me.getNumberOfClicks() >= 2)
		{
			mouseDoubleClick (me);

			if (checker.shouldBailOut())
				return;

			desktop.mouseListeners.callChecked (checker, &MouseListener::mouseDoubleClick, me);
			MouseListenerList::sendMouseEvent (*this, checker, &MouseListener::mouseDoubleClick, me);
		}
	}
}

void Component::internalMouseDrag (MouseInputSource& source, const Point<int>& relativePos, const Time& time)
{
	if (flags.mouseDownFlag)
	{
		flags.mouseOverFlag = reallyContains (relativePos, false);

		BailOutChecker checker (this);

		const MouseEvent me (source, relativePos,
							 source.getCurrentModifiers(), this, this, time,
							 getLocalPoint (nullptr, source.getLastMouseDownPosition()),
							 source.getLastMouseDownTime(),
							 source.getNumberOfMultipleClicks(),
							 source.hasMouseMovedSignificantlySincePressed());

		mouseDrag (me);

		if (checker.shouldBailOut())
			return;

		Desktop& desktop = Desktop::getInstance();
		desktop.resetTimer();
		desktop.mouseListeners.callChecked (checker, &MouseListener::mouseDrag, me);

		MouseListenerList::sendMouseEvent (*this, checker, &MouseListener::mouseDrag, me);
	}
}

void Component::internalMouseMove (MouseInputSource& source, const Point<int>& relativePos, const Time& time)
{
	Desktop& desktop = Desktop::getInstance();
	BailOutChecker checker (this);

	const MouseEvent me (source, relativePos, source.getCurrentModifiers(),
						 this, this, time, relativePos, time, 0, false);

	if (isCurrentlyBlockedByAnotherModalComponent())
	{
		// allow blocked mouse-events to go to global listeners..
		desktop.sendMouseMove();
	}
	else
	{
		flags.mouseOverFlag = true;

		mouseMove (me);

		if (checker.shouldBailOut())
			return;

		desktop.resetTimer();
		desktop.mouseListeners.callChecked (checker, &MouseListener::mouseMove, me);

		MouseListenerList::sendMouseEvent (*this, checker, &MouseListener::mouseMove, me);
	}
}

void Component::internalMouseWheel (MouseInputSource& source, const Point<int>& relativePos,
									const Time& time, const float amountX, const float amountY)
{
	Desktop& desktop = Desktop::getInstance();
	BailOutChecker checker (this);

	const float wheelIncrementX = amountX / 256.0f;
	const float wheelIncrementY = amountY / 256.0f;

	const MouseEvent me (source, relativePos, source.getCurrentModifiers(),
						 this, this, time, relativePos, time, 0, false);

	if (isCurrentlyBlockedByAnotherModalComponent())
	{
		// allow blocked mouse-events to go to global listeners..
		desktop.mouseListeners.callChecked (checker, &MouseListener::mouseWheelMove, me, wheelIncrementX, wheelIncrementY);
	}
	else
	{
		mouseWheelMove (me, wheelIncrementX, wheelIncrementY);

		if (checker.shouldBailOut())
			return;

		desktop.mouseListeners.callChecked (checker, &MouseListener::mouseWheelMove, me, wheelIncrementX, wheelIncrementY);

		if (! checker.shouldBailOut())
			MouseListenerList::sendWheelEvent (*this, checker, me, wheelIncrementX, wheelIncrementY);
	}
}

void Component::sendFakeMouseMove() const
{
	MouseInputSource& mainMouse = Desktop::getInstance().getMainMouseSource();

	if (! mainMouse.isDragging())
		mainMouse.triggerFakeMove();
}

void Component::beginDragAutoRepeat (const int interval)
{
	Desktop::getInstance().beginDragAutoRepeat (interval);
}

void Component::broughtToFront()
{
}

void Component::internalBroughtToFront()
{
	if (flags.hasHeavyweightPeerFlag)
		Desktop::getInstance().componentBroughtToFront (this);

	BailOutChecker checker (this);
	broughtToFront();

	if (checker.shouldBailOut())
		return;

	componentListeners.callChecked (checker, &ComponentListener::componentBroughtToFront, *this);

	if (checker.shouldBailOut())
		return;

	// When brought to the front and there's a modal component blocking this one,
	// we need to bring the modal one to the front instead..
	Component* const cm = getCurrentlyModalComponent();

	if (cm != nullptr && cm->getTopLevelComponent() != getTopLevelComponent())
		ModalComponentManager::getInstance()->bringModalComponentsToFront (false); // very important that this is false, otherwise in win32,
																				   // non-front components can't get focus when another modal comp is
																				   // active, and therefore can't receive mouse-clicks
}

void Component::focusGained (FocusChangeType)
{
	// base class does nothing
}

void Component::internalFocusGain (const FocusChangeType cause)
{
	internalFocusGain (cause, WeakReference<Component> (this));
}

void Component::internalFocusGain (const FocusChangeType cause, const WeakReference<Component>& safePointer)
{
	focusGained (cause);

	if (safePointer != nullptr)
		internalChildFocusChange (cause, safePointer);
}

void Component::focusLost (FocusChangeType)
{
	// base class does nothing
}

void Component::internalFocusLoss (const FocusChangeType cause)
{
	WeakReference<Component> safePointer (this);

	focusLost (focusChangedDirectly);

	if (safePointer != nullptr)
		internalChildFocusChange (cause, safePointer);
}

void Component::focusOfChildComponentChanged (FocusChangeType /*cause*/)
{
	// base class does nothing
}

void Component::internalChildFocusChange (FocusChangeType cause, const WeakReference<Component>& safePointer)
{
	const bool childIsNowFocused = hasKeyboardFocus (true);

	if (flags.childCompFocusedFlag != childIsNowFocused)
	{
		flags.childCompFocusedFlag = childIsNowFocused;

		focusOfChildComponentChanged (cause);

		if (safePointer == nullptr)
			return;
	}

	if (parentComponent != nullptr)
		parentComponent->internalChildFocusChange (cause, WeakReference<Component> (parentComponent));
}

bool Component::isEnabled() const noexcept
{
	return (! flags.isDisabledFlag)
			&& (parentComponent == nullptr || parentComponent->isEnabled());
}

void Component::setEnabled (const bool shouldBeEnabled)
{
	if (flags.isDisabledFlag == shouldBeEnabled)
	{
		flags.isDisabledFlag = ! shouldBeEnabled;

		// if any parent components are disabled, setting our flag won't make a difference,
		// so no need to send a change message
		if (parentComponent == nullptr || parentComponent->isEnabled())
			sendEnablementChangeMessage();
	}
}

void Component::sendEnablementChangeMessage()
{
	WeakReference<Component> safePointer (this);

	enablementChanged();

	if (safePointer == nullptr)
		return;

	for (int i = getNumChildComponents(); --i >= 0;)
	{
		Component* const c = getChildComponent (i);

		if (c != nullptr)
		{
			c->sendEnablementChangeMessage();

			if (safePointer == nullptr)
				return;
		}
	}
}

void Component::enablementChanged()
{
}

void Component::setWantsKeyboardFocus (const bool wantsFocus) noexcept
{
	flags.wantsFocusFlag = wantsFocus;
}

void Component::setMouseClickGrabsKeyboardFocus (const bool shouldGrabFocus)
{
	flags.dontFocusOnMouseClickFlag = ! shouldGrabFocus;
}

bool Component::getMouseClickGrabsKeyboardFocus() const noexcept
{
	return ! flags.dontFocusOnMouseClickFlag;
}

bool Component::getWantsKeyboardFocus() const noexcept
{
	return flags.wantsFocusFlag && ! flags.isDisabledFlag;
}

void Component::setFocusContainer (const bool shouldBeFocusContainer) noexcept
{
	flags.isFocusContainerFlag = shouldBeFocusContainer;
}

bool Component::isFocusContainer() const noexcept
{
	return flags.isFocusContainerFlag;
}

static const Identifier juce_explicitFocusOrderId ("_jexfo");

int Component::getExplicitFocusOrder() const
{
	return properties [juce_explicitFocusOrderId];
}

void Component::setExplicitFocusOrder (const int newFocusOrderIndex)
{
	properties.set (juce_explicitFocusOrderId, newFocusOrderIndex);
}

KeyboardFocusTraverser* Component::createFocusTraverser()
{
	if (flags.isFocusContainerFlag || parentComponent == nullptr)
		return new KeyboardFocusTraverser();

	return parentComponent->createFocusTraverser();
}

void Component::takeKeyboardFocus (const FocusChangeType cause)
{
	// give the focus to this component
	if (currentlyFocusedComponent != this)
	{
		// get the focus onto our desktop window
		ComponentPeer* const peer = getPeer();

		if (peer != nullptr)
		{
			WeakReference<Component> safePointer (this);

			peer->grabFocus();

			if (peer->isFocused() && currentlyFocusedComponent != this)
			{
				WeakReference<Component> componentLosingFocus (currentlyFocusedComponent);

				currentlyFocusedComponent = this;

				Desktop::getInstance().triggerFocusCallback();

				// call this after setting currentlyFocusedComponent so that the one that's
				// losing it has a chance to see where focus is going
				if (componentLosingFocus != nullptr)
					componentLosingFocus->internalFocusLoss (cause);

				if (currentlyFocusedComponent == this)
					internalFocusGain (cause, safePointer);
			}
		}
	}
}

void Component::grabFocusInternal (const FocusChangeType cause, const bool canTryParent)
{
	if (isShowing())
	{
		if (flags.wantsFocusFlag && (isEnabled() || parentComponent == nullptr))
		{
			takeKeyboardFocus (cause);
		}
		else
		{
			if (isParentOf (currentlyFocusedComponent)
				 && currentlyFocusedComponent->isShowing())
			{
				// do nothing if the focused component is actually a child of ours..
			}
			else
			{
				// find the default child component..
				ScopedPointer <KeyboardFocusTraverser> traverser (createFocusTraverser());

				if (traverser != nullptr)
				{
					Component* const defaultComp = traverser->getDefaultComponent (this);
					traverser = nullptr;

					if (defaultComp != nullptr)
					{
						defaultComp->grabFocusInternal (cause, false);
						return;
					}
				}

				if (canTryParent && parentComponent != nullptr)
				{
					// if no children want it and we're allowed to try our parent comp,
					// then pass up to parent, which will try our siblings.
					parentComponent->grabFocusInternal (cause, true);
				}
			}
		}
	}
}

void Component::grabKeyboardFocus()
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	grabFocusInternal (focusChangedDirectly);
}

void Component::moveKeyboardFocusToSibling (const bool moveToNext)
{
	// if component methods are being called from threads other than the message
	// thread, you'll need to use a MessageManagerLock object to make sure it's thread-safe.
	CHECK_MESSAGE_MANAGER_IS_LOCKED

	if (parentComponent != nullptr)
	{
		ScopedPointer <KeyboardFocusTraverser> traverser (createFocusTraverser());

		if (traverser != nullptr)
		{
			Component* const nextComp = moveToNext ? traverser->getNextComponent (this)
												   : traverser->getPreviousComponent (this);
			traverser = nullptr;

			if (nextComp != nullptr)
			{
				if (nextComp->isCurrentlyBlockedByAnotherModalComponent())
				{
					WeakReference<Component> nextCompPointer (nextComp);
					internalModalInputAttempt();

					if (nextCompPointer == nullptr || nextComp->isCurrentlyBlockedByAnotherModalComponent())
						return;
				}

				nextComp->grabFocusInternal (focusChangedByTabKey);
				return;
			}
		}

		parentComponent->moveKeyboardFocusToSibling (moveToNext);
	}
}

bool Component::hasKeyboardFocus (const bool trueIfChildIsFocused) const
{
	return (currentlyFocusedComponent == this)
			|| (trueIfChildIsFocused && isParentOf (currentlyFocusedComponent));
}

Component* JUCE_CALLTYPE Component::getCurrentlyFocusedComponent() noexcept
{
	return currentlyFocusedComponent;
}

void Component::giveAwayFocus (const bool sendFocusLossEvent)
{
	Component* const componentLosingFocus = currentlyFocusedComponent;
	currentlyFocusedComponent = nullptr;

	if (sendFocusLossEvent && componentLosingFocus != nullptr)
		componentLosingFocus->internalFocusLoss (focusChangedDirectly);

	Desktop::getInstance().triggerFocusCallback();
}

bool Component::isMouseOver (const bool includeChildren) const
{
	if (flags.mouseOverFlag)
		return true;

	if (includeChildren)
	{
		Desktop& desktop = Desktop::getInstance();

		for (int i = desktop.getNumMouseSources(); --i >= 0;)
		{
			Component* const c = desktop.getMouseSource(i)->getComponentUnderMouse();

			if (isParentOf (c) && c->flags.mouseOverFlag) // (mouseOverFlag checked in case it's being dragged outside the comp)
				return true;
		}
	}

	return false;
}

bool Component::isMouseButtonDown() const noexcept	  { return flags.mouseDownFlag; }
bool Component::isMouseOverOrDragging() const noexcept  { return flags.mouseOverFlag || flags.mouseDownFlag; }

bool JUCE_CALLTYPE Component::isMouseButtonDownAnywhere() noexcept
{
	return ModifierKeys::getCurrentModifiers().isAnyMouseButtonDown();
}

Point<int> Component::getMouseXYRelative() const
{
	return getLocalPoint (nullptr, Desktop::getMousePosition());
}

Rectangle<int> Component::getParentMonitorArea() const
{
	return Desktop::getInstance().getMonitorAreaContaining (getScreenBounds().getCentre());
}

void Component::addKeyListener (KeyListener* const newListener)
{
	if (keyListeners == nullptr)
		keyListeners = new Array <KeyListener*>();

	keyListeners->addIfNotAlreadyThere (newListener);
}

void Component::removeKeyListener (KeyListener* const listenerToRemove)
{
	if (keyListeners != nullptr)
		keyListeners->removeValue (listenerToRemove);
}

bool Component::keyPressed (const KeyPress&)
{
	return false;
}

bool Component::keyStateChanged (const bool /*isKeyDown*/)
{
	return false;
}

void Component::modifierKeysChanged (const ModifierKeys& modifiers)
{
	if (parentComponent != nullptr)
		parentComponent->modifierKeysChanged (modifiers);
}

void Component::internalModifierKeysChanged()
{
	sendFakeMouseMove();

	modifierKeysChanged (ModifierKeys::getCurrentModifiers());
}

ComponentPeer* Component::getPeer() const
{
	if (flags.hasHeavyweightPeerFlag)
		return ComponentPeer::getPeerFor (this);
	else if (parentComponent == nullptr)
		return nullptr;

	return parentComponent->getPeer();
}

Component::BailOutChecker::BailOutChecker (Component* const component)
	: safePointer (component)
{
	jassert (component != nullptr);
}

bool Component::BailOutChecker::shouldBailOut() const noexcept
{
	return safePointer == nullptr;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Component.cpp ***/



/*** Start of inlined file: juce_ComponentListener.cpp ***/
BEGIN_JUCE_NAMESPACE

void ComponentListener::componentMovedOrResized (Component&, bool, bool)	{}
void ComponentListener::componentBroughtToFront (Component&)		{}
void ComponentListener::componentVisibilityChanged (Component&)		 {}
void ComponentListener::componentChildrenChanged (Component&)		   {}
void ComponentListener::componentParentHierarchyChanged (Component&)	{}
void ComponentListener::componentNameChanged (Component&)		   {}
void ComponentListener::componentBeingDeleted (Component&)		  {}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComponentListener.cpp ***/


/*** Start of inlined file: juce_Desktop.cpp ***/
BEGIN_JUCE_NAMESPACE

Desktop::Desktop()
	: mouseClickCounter (0),
	  kioskModeComponent (nullptr),
	  allowedOrientations (allOrientations)
{
	createMouseInputSources();
	refreshMonitorSizes();
}

Desktop::~Desktop()
{
	setScreenSaverEnabled (true);

	jassert (instance == this);
	instance = nullptr;

	// doh! If you don't delete all your windows before exiting, you're going to
	// be leaking memory!
	jassert (desktopComponents.size() == 0);
}

Desktop& JUCE_CALLTYPE Desktop::getInstance()
{
	if (instance == nullptr)
		instance = new Desktop();

	return *instance;
}

Desktop* Desktop::instance = nullptr;

void Desktop::refreshMonitorSizes()
{
	Array <Rectangle<int> > oldClipped, oldUnclipped;
	oldClipped.swapWithArray (monitorCoordsClipped);
	oldUnclipped.swapWithArray (monitorCoordsUnclipped);

	getCurrentMonitorPositions (monitorCoordsClipped, true);
	getCurrentMonitorPositions (monitorCoordsUnclipped, false);
	jassert (monitorCoordsClipped.size() == monitorCoordsUnclipped.size());

	if (oldClipped != monitorCoordsClipped
		 || oldUnclipped != monitorCoordsUnclipped)
	{
		for (int i = ComponentPeer::getNumPeers(); --i >= 0;)
		{
			ComponentPeer* const p = ComponentPeer::getPeer (i);
			if (p != nullptr)
				p->handleScreenSizeChange();
		}
	}
}

int Desktop::getNumDisplayMonitors() const noexcept
{
	return monitorCoordsClipped.size();
}

const Rectangle<int> Desktop::getDisplayMonitorCoordinates (const int index, const bool clippedToWorkArea) const noexcept
{
	return clippedToWorkArea ? monitorCoordsClipped [index]
							 : monitorCoordsUnclipped [index];
}

const RectangleList Desktop::getAllMonitorDisplayAreas (const bool clippedToWorkArea) const
{
	RectangleList rl;

	for (int i = 0; i < getNumDisplayMonitors(); ++i)
		rl.addWithoutMerging (getDisplayMonitorCoordinates (i, clippedToWorkArea));

	return rl;
}

const Rectangle<int> Desktop::getMainMonitorArea (const bool clippedToWorkArea) const noexcept
{
	return getDisplayMonitorCoordinates (0, clippedToWorkArea);
}

const Rectangle<int> Desktop::getMonitorAreaContaining (const Point<int>& position, const bool clippedToWorkArea) const
{
	Rectangle<int> best (getMainMonitorArea (clippedToWorkArea));
	double bestDistance = 1.0e10;

	for (int i = getNumDisplayMonitors(); --i >= 0;)
	{
		const Rectangle<int> rect (getDisplayMonitorCoordinates (i, clippedToWorkArea));

		if (rect.contains (position))
			return rect;

		const double distance = rect.getCentre().getDistanceFrom (position);

		if (distance < bestDistance)
		{
			bestDistance = distance;
			best = rect;
		}
	}

	return best;
}

int Desktop::getNumComponents() const noexcept
{
	return desktopComponents.size();
}

Component* Desktop::getComponent (const int index) const noexcept
{
	return desktopComponents [index];
}

Component* Desktop::findComponentAt (const Point<int>& screenPosition) const
{
	for (int i = desktopComponents.size(); --i >= 0;)
	{
		Component* const c = desktopComponents.getUnchecked(i);

		if (c->isVisible())
		{
			const Point<int> relative (c->getLocalPoint (nullptr, screenPosition));

			if (c->contains (relative))
				return c->getComponentAt (relative);
		}
	}

	return nullptr;
}

LookAndFeel& Desktop::getDefaultLookAndFeel() noexcept
{
	if (currentLookAndFeel == nullptr)
	{
		if (defaultLookAndFeel == nullptr)
			defaultLookAndFeel = new LookAndFeel();

		currentLookAndFeel = defaultLookAndFeel;
	}

	return *currentLookAndFeel;
}

void Desktop::setDefaultLookAndFeel (LookAndFeel* newDefaultLookAndFeel)
{
	currentLookAndFeel = newDefaultLookAndFeel;

	for (int i = getNumComponents(); --i >= 0;)
	{
		Component* const c = getComponent (i);

		if (c != nullptr)
			c->sendLookAndFeelChange();
	}
}

void Desktop::addDesktopComponent (Component* const c)
{
	jassert (c != nullptr);
	jassert (! desktopComponents.contains (c));
	desktopComponents.addIfNotAlreadyThere (c);
}

void Desktop::removeDesktopComponent (Component* const c)
{
	desktopComponents.removeValue (c);
}

void Desktop::componentBroughtToFront (Component* const c)
{
	const int index = desktopComponents.indexOf (c);
	jassert (index >= 0);

	if (index >= 0)
	{
		int newIndex = -1;

		if (! c->isAlwaysOnTop())
		{
			newIndex = desktopComponents.size();

			while (newIndex > 0 && desktopComponents.getUnchecked (newIndex - 1)->isAlwaysOnTop())
				--newIndex;

			--newIndex;
		}

		desktopComponents.move (index, newIndex);
	}
}

const Point<int> Desktop::getMousePosition()
{
	return getInstance().getMainMouseSource().getScreenPosition();
}

const Point<int> Desktop::getLastMouseDownPosition()
{
	return getInstance().getMainMouseSource().getLastMouseDownPosition();
}

int Desktop::getMouseButtonClickCounter()
{
	return getInstance().mouseClickCounter;
}

void Desktop::incrementMouseClickCounter() noexcept
{
	++mouseClickCounter;
}

int Desktop::getNumDraggingMouseSources() const noexcept
{
	int num = 0;
	for (int i = mouseSources.size(); --i >= 0;)
		if (mouseSources.getUnchecked(i)->isDragging())
			++num;

	return num;
}

MouseInputSource* Desktop::getDraggingMouseSource (int index) const noexcept
{
	int num = 0;
	for (int i = mouseSources.size(); --i >= 0;)
	{
		MouseInputSource* const mi = mouseSources.getUnchecked(i);

		if (mi->isDragging())
		{
			if (index == num)
				return mi;

			++num;
		}
	}

	return nullptr;
}

class MouseDragAutoRepeater  : public Timer
{
public:
	MouseDragAutoRepeater() {}

	void timerCallback()
	{
		Desktop& desktop = Desktop::getInstance();
		int numMiceDown = 0;

		for (int i = desktop.getNumMouseSources(); --i >= 0;)
		{
			MouseInputSource* const source = desktop.getMouseSource(i);
			if (source->isDragging())
			{
				source->triggerFakeMove();
				++numMiceDown;
			}
		}

		if (numMiceDown == 0)
			desktop.beginDragAutoRepeat (0);
	}

private:
	JUCE_DECLARE_NON_COPYABLE (MouseDragAutoRepeater);
};

void Desktop::beginDragAutoRepeat (const int interval)
{
	if (interval > 0)
	{
		if (dragRepeater == nullptr)
			dragRepeater = new MouseDragAutoRepeater();

		if (dragRepeater->getTimerInterval() != interval)
			dragRepeater->startTimer (interval);
	}
	else
	{
		dragRepeater = nullptr;
	}
}

void Desktop::addFocusChangeListener (FocusChangeListener* const listener)
{
	focusListeners.add (listener);
}

void Desktop::removeFocusChangeListener (FocusChangeListener* const listener)
{
	focusListeners.remove (listener);
}

void Desktop::triggerFocusCallback()
{
	triggerAsyncUpdate();
}

void Desktop::handleAsyncUpdate()
{
	// The component may be deleted during this operation, but we'll use a SafePointer rather than a
	// BailOutChecker so that any remaining listeners will still get a callback (with a null pointer).
	WeakReference<Component> currentFocus (Component::getCurrentlyFocusedComponent());
	focusListeners.call (&FocusChangeListener::globalFocusChanged, currentFocus);
}

void Desktop::addGlobalMouseListener (MouseListener* const listener)
{
	mouseListeners.add (listener);
	resetTimer();
}

void Desktop::removeGlobalMouseListener (MouseListener* const listener)
{
	mouseListeners.remove (listener);
	resetTimer();
}

void Desktop::timerCallback()
{
	if (lastFakeMouseMove != getMousePosition())
		sendMouseMove();
}

void Desktop::sendMouseMove()
{
	if (! mouseListeners.isEmpty())
	{
		startTimer (20);

		lastFakeMouseMove = getMousePosition();

		Component* const target = findComponentAt (lastFakeMouseMove);

		if (target != nullptr)
		{
			Component::BailOutChecker checker (target);
			const Point<int> pos (target->getLocalPoint (nullptr, lastFakeMouseMove));
			const Time now (Time::getCurrentTime());

			const MouseEvent me (getMainMouseSource(), pos, ModifierKeys::getCurrentModifiers(),
								 target, target, now, pos, now, 0, false);

			if (me.mods.isAnyMouseButtonDown())
				mouseListeners.callChecked (checker, &MouseListener::mouseDrag, me);
			else
				mouseListeners.callChecked (checker, &MouseListener::mouseMove, me);
		}
	}
}

void Desktop::resetTimer()
{
	if (mouseListeners.size() == 0)
		stopTimer();
	else
		startTimer (100);

	lastFakeMouseMove = getMousePosition();
}

void Desktop::setKioskModeComponent (Component* componentToUse, const bool allowMenusAndBars)
{
	if (kioskModeComponent != componentToUse)
	{
		// agh! Don't delete or remove a component from the desktop while it's still the kiosk component!
		jassert (kioskModeComponent == nullptr || ComponentPeer::getPeerFor (kioskModeComponent) != nullptr);

		if (kioskModeComponent != nullptr)
		{
			setKioskComponent (kioskModeComponent, false, allowMenusAndBars);

			kioskModeComponent->setBounds (kioskComponentOriginalBounds);
		}

		kioskModeComponent = componentToUse;

		if (kioskModeComponent != nullptr)
		{
			// Only components that are already on the desktop can be put into kiosk mode!
			jassert (ComponentPeer::getPeerFor (kioskModeComponent) != nullptr);

			kioskComponentOriginalBounds = kioskModeComponent->getBounds();

			setKioskComponent (kioskModeComponent, true, allowMenusAndBars);
		}
	}
}

void Desktop::setOrientationsEnabled (const int newOrientations)
{
	// Dodgy set of flags being passed here! Make sure you specify at least one permitted orientation.
	jassert (newOrientations != 0 && (newOrientations & ~allOrientations) == 0);

	allowedOrientations = newOrientations;
}

bool Desktop::isOrientationEnabled (const DisplayOrientation orientation) const noexcept
{
	// Make sure you only pass one valid flag in here...
	jassert (orientation == upright || orientation == upsideDown || orientation == rotatedClockwise || orientation ==  rotatedAntiClockwise);

	return (allowedOrientations & orientation) != 0;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Desktop.cpp ***/


/*** Start of inlined file: juce_ModalComponentManager.cpp ***/
BEGIN_JUCE_NAMESPACE

class ModalComponentManager::ModalItem  : public ComponentMovementWatcher
{
public:
	ModalItem (Component* const comp)
		: ComponentMovementWatcher (comp),
		  component (comp), returnValue (0), isActive (true)
	{
		jassert (comp != nullptr);
	}

	void componentMovedOrResized (bool, bool) {}

	void componentPeerChanged()
	{
		if (! component->isShowing())
			cancel();
	}

	void componentVisibilityChanged()
	{
		if (! component->isShowing())
			cancel();
	}

	void componentBeingDeleted (Component& comp)
	{
		ComponentMovementWatcher::componentBeingDeleted (comp);

		if (component == &comp || comp.isParentOf (component))
			cancel();
	}

	void cancel()
	{
		if (isActive)
		{
			isActive = false;
			ModalComponentManager::getInstance()->triggerAsyncUpdate();
		}
	}

	Component* component;
	OwnedArray<Callback> callbacks;
	int returnValue;
	bool isActive;

private:
	JUCE_DECLARE_NON_COPYABLE (ModalItem);
};

ModalComponentManager::ModalComponentManager()
{
}

ModalComponentManager::~ModalComponentManager()
{
	clearSingletonInstance();
}

juce_ImplementSingleton_SingleThreaded (ModalComponentManager);

void ModalComponentManager::startModal (Component* component)
{
	if (component != nullptr)
		stack.add (new ModalItem (component));
}

void ModalComponentManager::attachCallback (Component* component, Callback* callback)
{
	if (callback != nullptr)
	{
		ScopedPointer<Callback> callbackDeleter (callback);

		for (int i = stack.size(); --i >= 0;)
		{
			ModalItem* const item = stack.getUnchecked(i);

			if (item->component == component)
			{
				item->callbacks.add (callback);
				callbackDeleter.release();
				break;
			}
		}
	}
}

void ModalComponentManager::endModal (Component* component)
{
	for (int i = stack.size(); --i >= 0;)
	{
		ModalItem* const item = stack.getUnchecked(i);

		if (item->component == component)
			item->cancel();
	}
}

void ModalComponentManager::endModal (Component* component, int returnValue)
{
	for (int i = stack.size(); --i >= 0;)
	{
		ModalItem* const item = stack.getUnchecked(i);

		if (item->component == component)
		{
			item->returnValue = returnValue;
			item->cancel();
		}
	}
}

int ModalComponentManager::getNumModalComponents() const
{
	int n = 0;
	for (int i = 0; i < stack.size(); ++i)
		if (stack.getUnchecked(i)->isActive)
			++n;

	return n;
}

Component* ModalComponentManager::getModalComponent (const int index) const
{
	int n = 0;
	for (int i = stack.size(); --i >= 0;)
	{
		const ModalItem* const item = stack.getUnchecked(i);
		if (item->isActive)
			if (n++ == index)
				return item->component;
	}

	return nullptr;
}

bool ModalComponentManager::isModal (Component* const comp) const
{
	for (int i = stack.size(); --i >= 0;)
	{
		const ModalItem* const item = stack.getUnchecked(i);
		if (item->isActive && item->component == comp)
			return true;
	}

	return false;
}

bool ModalComponentManager::isFrontModalComponent (Component* const comp) const
{
	return comp == getModalComponent (0);
}

void ModalComponentManager::handleAsyncUpdate()
{
	for (int i = stack.size(); --i >= 0;)
	{
		const ModalItem* const item = stack.getUnchecked(i);

		if (! item->isActive)
		{
			ScopedPointer<ModalItem> deleter (stack.removeAndReturn (i));

			for (int j = item->callbacks.size(); --j >= 0;)
				item->callbacks.getUnchecked(j)->modalStateFinished (item->returnValue);
		}
	}
}

void ModalComponentManager::bringModalComponentsToFront (bool topOneShouldGrabFocus)
{
	ComponentPeer* lastOne = nullptr;

	for (int i = 0; i < getNumModalComponents(); ++i)
	{
		Component* const c = getModalComponent (i);

		if (c == nullptr)
			break;

		ComponentPeer* peer = c->getPeer();

		if (peer != nullptr && peer != lastOne)
		{
			if (lastOne == nullptr)
			{
				peer->toFront (topOneShouldGrabFocus);

				if (topOneShouldGrabFocus)
					peer->grabFocus();
			}
			else
				peer->toBehind (lastOne);

			lastOne = peer;
		}
	}
}

#if JUCE_MODAL_LOOPS_PERMITTED
class ModalComponentManager::ReturnValueRetriever	 : public ModalComponentManager::Callback
{
public:
	ReturnValueRetriever (int& value_, bool& finished_) : value (value_), finished (finished_) {}

	void modalStateFinished (int returnValue)
	{
		finished = true;
		value = returnValue;
	}

private:
	int& value;
	bool& finished;

	JUCE_DECLARE_NON_COPYABLE (ReturnValueRetriever);
};

int ModalComponentManager::runEventLoopForCurrentComponent()
{
	// This can only be run from the message thread!
	jassert (MessageManager::getInstance()->isThisTheMessageThread());

	Component* currentlyModal = getModalComponent (0);

	if (currentlyModal == nullptr)
		return 0;

	WeakReference<Component> prevFocused (Component::getCurrentlyFocusedComponent());

	int returnValue = 0;
	bool finished = false;
	attachCallback (currentlyModal, new ReturnValueRetriever (returnValue, finished));

	JUCE_TRY
	{
		while (! finished)
		{
			if  (! MessageManager::getInstance()->runDispatchLoopUntil (20))
				break;
		}
	}
	JUCE_CATCH_EXCEPTION

	if (prevFocused != nullptr)
		prevFocused->grabKeyboardFocus();

	return returnValue;
}
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ModalComponentManager.cpp ***/


/*** Start of inlined file: juce_ArrowButton.cpp ***/
BEGIN_JUCE_NAMESPACE

ArrowButton::ArrowButton (const String& name,
						  float arrowDirectionInRadians,
						  const Colour& arrowColour)
   : Button (name),
	 colour (arrowColour)
{
	path.lineTo (0.0f, 1.0f);
	path.lineTo (1.0f, 0.5f);
	path.closeSubPath();

	path.applyTransform (AffineTransform::rotation (float_Pi * 2.0f * arrowDirectionInRadians,
													0.5f, 0.5f));

	setComponentEffect (&shadow);
	updateShadowAndOffset();
}

ArrowButton::~ArrowButton()
{
}

void ArrowButton::paintButton (Graphics& g,
							   bool /*isMouseOverButton*/,
							   bool /*isButtonDown*/)
{
	g.setColour (colour);

	g.fillPath (path, path.getTransformToScaleToFit ((float) offset,
													 (float) offset,
													 (float) (getWidth() - 3),
													 (float) (getHeight() - 3),
													 false));
}

void ArrowButton::buttonStateChanged()
{
	updateShadowAndOffset();
}

void ArrowButton::updateShadowAndOffset()
{
	offset = (isDown()) ? 1 : 0;

	shadow.setShadowProperties ((isDown()) ? 1.2f : 3.0f,
								0.3f, -1, 0);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ArrowButton.cpp ***/


/*** Start of inlined file: juce_Button.cpp ***/
BEGIN_JUCE_NAMESPACE

class Button::RepeatTimer  : public Timer
{
public:
	RepeatTimer (Button& owner_) : owner (owner_)   {}
	void timerCallback()	{ owner.repeatTimerCallback(); }

private:
	Button& owner;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (RepeatTimer);
};

Button::Button (const String& name)
  : Component (name),
	text (name),
	buttonPressTime (0),
	lastRepeatTime (0),
	commandManagerToUse (nullptr),
	autoRepeatDelay (-1),
	autoRepeatSpeed (0),
	autoRepeatMinimumDelay (-1),
	radioGroupId (0),
	commandID (0),
	connectedEdgeFlags (0),
	buttonState (buttonNormal),
	lastToggleState (false),
	clickTogglesState (false),
	needsToRelease (false),
	needsRepainting (false),
	isKeyDown (false),
	triggerOnMouseDown (false),
	generateTooltip (false)
{
	setWantsKeyboardFocus (true);
	isOn.addListener (this);
}

Button::~Button()
{
	isOn.removeListener (this);

	if (commandManagerToUse != nullptr)
		commandManagerToUse->removeListener (this);

	repeatTimer = nullptr;
	clearShortcuts();
}

void Button::setButtonText (const String& newText)
{
	if (text != newText)
	{
		text = newText;
		repaint();
	}
}

void Button::setTooltip (const String& newTooltip)
{
	SettableTooltipClient::setTooltip (newTooltip);
	generateTooltip = false;
}

const String Button::getTooltip()
{
	if (generateTooltip && commandManagerToUse != nullptr && commandID != 0)
	{
		String tt (commandManagerToUse->getDescriptionOfCommand (commandID));

		Array <KeyPress> keyPresses (commandManagerToUse->getKeyMappings()->getKeyPressesAssignedToCommand (commandID));

		for (int i = 0; i < keyPresses.size(); ++i)
		{
			const String key (keyPresses.getReference(i).getTextDescription());

			tt << " [";

			if (key.length() == 1)
				tt << TRANS("shortcut") << ": '" << key << "']";
			else
				tt << key << ']';
		}

		return tt;
	}

	return SettableTooltipClient::getTooltip();
}

void Button::setConnectedEdges (const int connectedEdgeFlags_)
{
	if (connectedEdgeFlags != connectedEdgeFlags_)
	{
		connectedEdgeFlags = connectedEdgeFlags_;
		repaint();
	}
}

void Button::setToggleState (const bool shouldBeOn,
							 const bool sendChangeNotification)
{
	if (shouldBeOn != lastToggleState)
	{
		if (isOn != shouldBeOn)  // this test means that if the value is void rather than explicitly set to
			isOn = shouldBeOn;   // false, it won't be changed unless the required value is true.

		lastToggleState = shouldBeOn;
		repaint();

		WeakReference<Component> deletionWatcher (this);

		if (sendChangeNotification)
		{
			sendClickMessage (ModifierKeys());

			if (deletionWatcher == nullptr)
				return;
		}

		if (lastToggleState)
		{
			turnOffOtherButtonsInGroup (sendChangeNotification);

			if (deletionWatcher == nullptr)
				return;
		}

		sendStateMessage();
	}
}

void Button::setClickingTogglesState (const bool shouldToggle) noexcept
{
	clickTogglesState = shouldToggle;

	// if you've got clickTogglesState turned on, you shouldn't also connect the button
	// up to be a command invoker. Instead, your command handler must flip the state of whatever
	// it is that this button represents, and the button will update its state to reflect this
	// in the applicationCommandListChanged() method.
	jassert (commandManagerToUse == nullptr || ! clickTogglesState);
}

bool Button::getClickingTogglesState() const noexcept
{
	return clickTogglesState;
}

void Button::valueChanged (Value& value)
{
	if (value.refersToSameSourceAs (isOn))
		setToggleState (isOn.getValue(), true);
}

void Button::setRadioGroupId (const int newGroupId)
{
	if (radioGroupId != newGroupId)
	{
		radioGroupId = newGroupId;

		if (lastToggleState)
			turnOffOtherButtonsInGroup (true);
	}
}

void Button::turnOffOtherButtonsInGroup (const bool sendChangeNotification)
{
	Component* const p = getParentComponent();

	if (p != nullptr && radioGroupId != 0)
	{
		WeakReference<Component> deletionWatcher (this);

		for (int i = p->getNumChildComponents(); --i >= 0;)
		{
			Component* const c = p->getChildComponent (i);

			if (c != this)
			{
				Button* const b = dynamic_cast <Button*> (c);

				if (b != nullptr && b->getRadioGroupId() == radioGroupId)
				{
					b->setToggleState (false, sendChangeNotification);

					if (deletionWatcher == nullptr)
						return;
				}
			}
		}
	}
}

void Button::enablementChanged()
{
	updateState();
	repaint();
}

Button::ButtonState Button::updateState()
{
	return updateState (isMouseOver (true), isMouseButtonDown());
}

Button::ButtonState Button::updateState (const bool over, const bool down)
{
	ButtonState newState = buttonNormal;

	if (isEnabled() && isVisible() && ! isCurrentlyBlockedByAnotherModalComponent())
	{
		if ((down && (over || (triggerOnMouseDown && buttonState == buttonDown))) || isKeyDown)
			newState = buttonDown;
		else if (over)
			newState = buttonOver;
	}

	setState (newState);
	return newState;
}

void Button::setState (const ButtonState newState)
{
	if (buttonState != newState)
	{
		buttonState = newState;
		repaint();

		if (buttonState == buttonDown)
		{
			buttonPressTime = Time::getApproximateMillisecondCounter();
			lastRepeatTime = 0;
		}

		sendStateMessage();
	}
}

bool Button::isDown() const noexcept
{
	return buttonState == buttonDown;
}

bool Button::isOver() const noexcept
{
	return buttonState != buttonNormal;
}

void Button::buttonStateChanged()
{
}

uint32 Button::getMillisecondsSinceButtonDown() const noexcept
{
	const uint32 now = Time::getApproximateMillisecondCounter();
	return now > buttonPressTime ? now - buttonPressTime : 0;
}

void Button::setTriggeredOnMouseDown (const bool isTriggeredOnMouseDown) noexcept
{
	triggerOnMouseDown = isTriggeredOnMouseDown;
}

void Button::clicked()
{
}

void Button::clicked (const ModifierKeys& /*modifiers*/)
{
	clicked();
}

static const int clickMessageId = 0x2f3f4f99;

void Button::triggerClick()
{
	postCommandMessage (clickMessageId);
}

void Button::internalClickCallback (const ModifierKeys& modifiers)
{
	if (clickTogglesState)
		setToggleState ((radioGroupId != 0) || ! lastToggleState, false);

	sendClickMessage (modifiers);
}

void Button::flashButtonState()
{
	if (isEnabled())
	{
		needsToRelease = true;
		setState (buttonDown);
		getRepeatTimer().startTimer (100);
	}
}

void Button::handleCommandMessage (int commandId)
{
	if (commandId == clickMessageId)
	{
		if (isEnabled())
		{
			flashButtonState();
			internalClickCallback (ModifierKeys::getCurrentModifiers());
		}
	}
	else
	{
		Component::handleCommandMessage (commandId);
	}
}

void Button::addListener (ButtonListener* const newListener)
{
	buttonListeners.add (newListener);
}

void Button::removeListener (ButtonListener* const listener)
{
	buttonListeners.remove (listener);
}

void Button::addButtonListener (ButtonListener* l)	  { addListener (l); }
void Button::removeButtonListener (ButtonListener* l)   { removeListener (l); }

void Button::sendClickMessage (const ModifierKeys& modifiers)
{
	Component::BailOutChecker checker (this);

	if (commandManagerToUse != nullptr && commandID != 0)
	{
		ApplicationCommandTarget::InvocationInfo info (commandID);
		info.invocationMethod = ApplicationCommandTarget::InvocationInfo::fromButton;
		info.originatingComponent = this;

		commandManagerToUse->invoke (info, true);
	}

	clicked (modifiers);

	if (! checker.shouldBailOut())
		buttonListeners.callChecked (checker, &ButtonListener::buttonClicked, this);  // (can't use Button::Listener due to idiotic VC2005 bug)
}

void Button::sendStateMessage()
{
	Component::BailOutChecker checker (this);

	buttonStateChanged();

	if (! checker.shouldBailOut())
		buttonListeners.callChecked (checker, &ButtonListener::buttonStateChanged, this);
}

void Button::paint (Graphics& g)
{
	if (needsToRelease && isEnabled())
	{
		needsToRelease = false;
		needsRepainting = true;
	}

	paintButton (g, isOver(), isDown());
}

void Button::mouseEnter (const MouseEvent&)
{
	updateState (true, false);
}

void Button::mouseExit (const MouseEvent&)
{
	updateState (false, false);
}

void Button::mouseDown (const MouseEvent& e)
{
	updateState (true, true);

	if (isDown())
	{
		if (autoRepeatDelay >= 0)
			getRepeatTimer().startTimer (autoRepeatDelay);

		if (triggerOnMouseDown)
			internalClickCallback (e.mods);
	}
}

void Button::mouseUp (const MouseEvent& e)
{
	const bool wasDown = isDown();
	updateState (isMouseOver(), false);

	if (wasDown && isOver() && ! triggerOnMouseDown)
		internalClickCallback (e.mods);
}

void Button::mouseDrag (const MouseEvent&)
{
	const ButtonState oldState = buttonState;
	updateState (isMouseOver(), true);

	if (autoRepeatDelay >= 0 && buttonState != oldState && isDown())
		getRepeatTimer().startTimer (autoRepeatSpeed);
}

void Button::focusGained (FocusChangeType)
{
	updateState();
	repaint();
}

void Button::focusLost (FocusChangeType)
{
	updateState();
	repaint();
}

void Button::visibilityChanged()
{
	needsToRelease = false;
	updateState();
}

void Button::parentHierarchyChanged()
{
	Component* const newKeySource = (shortcuts.size() == 0) ? nullptr : getTopLevelComponent();

	if (newKeySource != keySource.get())
	{
		if (keySource != nullptr)
			keySource->removeKeyListener (this);

		keySource = newKeySource;

		if (keySource != nullptr)
			keySource->addKeyListener (this);
	}
}

void Button::setCommandToTrigger (ApplicationCommandManager* const commandManagerToUse_,
								  const int commandID_,
								  const bool generateTooltip_)
{
	commandID = commandID_;
	generateTooltip = generateTooltip_;

	if (commandManagerToUse != commandManagerToUse_)
	{
		if (commandManagerToUse != nullptr)
			commandManagerToUse->removeListener (this);

		commandManagerToUse = commandManagerToUse_;

		if (commandManagerToUse != nullptr)
			commandManagerToUse->addListener (this);

		// if you've got clickTogglesState turned on, you shouldn't also connect the button
		// up to be a command invoker. Instead, your command handler must flip the state of whatever
		// it is that this button represents, and the button will update its state to reflect this
		// in the applicationCommandListChanged() method.
		jassert (commandManagerToUse == nullptr || ! clickTogglesState);
	}

	if (commandManagerToUse != nullptr)
		applicationCommandListChanged();
	else
		setEnabled (true);
}

void Button::applicationCommandInvoked (const ApplicationCommandTarget::InvocationInfo& info)
{
	if (info.commandID == commandID
		 && (info.commandFlags & ApplicationCommandInfo::dontTriggerVisualFeedback) == 0)
	{
		flashButtonState();
	}
}

void Button::applicationCommandListChanged()
{
	if (commandManagerToUse != nullptr)
	{
		ApplicationCommandInfo info (0);

		ApplicationCommandTarget* const target = commandManagerToUse->getTargetForCommand (commandID, info);

		setEnabled (target != nullptr && (info.flags & ApplicationCommandInfo::isDisabled) == 0);

		if (target != nullptr)
			setToggleState ((info.flags & ApplicationCommandInfo::isTicked) != 0, false);
	}
}

void Button::addShortcut (const KeyPress& key)
{
	if (key.isValid())
	{
		jassert (! isRegisteredForShortcut (key));  // already registered!

		shortcuts.add (key);
		parentHierarchyChanged();
	}
}

void Button::clearShortcuts()
{
	shortcuts.clear();

	parentHierarchyChanged();
}

bool Button::isShortcutPressed() const
{
	if (! isCurrentlyBlockedByAnotherModalComponent())
	{
		for (int i = shortcuts.size(); --i >= 0;)
			if (shortcuts.getReference(i).isCurrentlyDown())
				return true;
	}

	return false;
}

bool Button::isRegisteredForShortcut (const KeyPress& key) const
{
	for (int i = shortcuts.size(); --i >= 0;)
		if (key == shortcuts.getReference(i))
			return true;

	return false;
}

bool Button::keyStateChanged (const bool, Component*)
{
	if (! isEnabled())
		return false;

	const bool wasDown = isKeyDown;
	isKeyDown = isShortcutPressed();

	if (autoRepeatDelay >= 0 && (isKeyDown && ! wasDown))
		getRepeatTimer().startTimer (autoRepeatDelay);

	updateState();

	if (isEnabled() && wasDown && ! isKeyDown)
	{
		internalClickCallback (ModifierKeys::getCurrentModifiers());

		// (return immediately - this button may now have been deleted)
		return true;
	}

	return wasDown || isKeyDown;
}

bool Button::keyPressed (const KeyPress&, Component*)
{
	// returning true will avoid forwarding events for keys that we're using as shortcuts
	return isShortcutPressed();
}

bool Button::keyPressed (const KeyPress& key)
{
	if (isEnabled() && key.isKeyCode (KeyPress::returnKey))
	{
		triggerClick();
		return true;
	}

	return false;
}

void Button::setRepeatSpeed (const int initialDelayMillisecs,
							 const int repeatMillisecs,
							 const int minimumDelayInMillisecs) noexcept
{
	autoRepeatDelay = initialDelayMillisecs;
	autoRepeatSpeed = repeatMillisecs;
	autoRepeatMinimumDelay = jmin (autoRepeatSpeed, minimumDelayInMillisecs);
}

void Button::repeatTimerCallback()
{
	if (needsRepainting)
	{
		getRepeatTimer().stopTimer();
		updateState();
		needsRepainting = false;
	}
	else if (autoRepeatSpeed > 0 && (isKeyDown || (updateState() == buttonDown)))
	{
		int repeatSpeed = autoRepeatSpeed;

		if (autoRepeatMinimumDelay >= 0)
		{
			double timeHeldDown = jmin (1.0, getMillisecondsSinceButtonDown() / 4000.0);
			timeHeldDown *= timeHeldDown;

			repeatSpeed = repeatSpeed + (int) (timeHeldDown * (autoRepeatMinimumDelay - repeatSpeed));
		}

		repeatSpeed = jmax (1, repeatSpeed);

		const uint32 now = Time::getMillisecondCounter();

		// if we've been blocked from repeating often enough, speed up the repeat timer to compensate..
		if (lastRepeatTime != 0 && (int) (now - lastRepeatTime) > repeatSpeed * 2)
			repeatSpeed = jmax (1, repeatSpeed / 2);

		lastRepeatTime = now;
		getRepeatTimer().startTimer (repeatSpeed);

		internalClickCallback (ModifierKeys::getCurrentModifiers());
	}
	else if (! needsToRelease)
	{
		getRepeatTimer().stopTimer();
	}
}

Button::RepeatTimer& Button::getRepeatTimer()
{
	if (repeatTimer == nullptr)
		repeatTimer = new RepeatTimer (*this);

	return *repeatTimer;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Button.cpp ***/


/*** Start of inlined file: juce_DrawableButton.cpp ***/
BEGIN_JUCE_NAMESPACE

DrawableButton::DrawableButton (const String& name,
								const DrawableButton::ButtonStyle buttonStyle)
	: Button (name),
	  style (buttonStyle),
	  currentImage (nullptr),
	  edgeIndent (3)
{
	if (buttonStyle == ImageOnButtonBackground)
	{
		backgroundOff = Colour (0xffbbbbff);
		backgroundOn = Colour (0xff3333ff);
	}
	else
	{
		backgroundOff = Colours::transparentBlack;
		backgroundOn = Colour (0xaabbbbff);
	}
}

DrawableButton::~DrawableButton()
{
}

void DrawableButton::setImages (const Drawable* normal,
								const Drawable* over,
								const Drawable* down,
								const Drawable* disabled,
								const Drawable* normalOn,
								const Drawable* overOn,
								const Drawable* downOn,
								const Drawable* disabledOn)
{
	jassert (normal != nullptr); // you really need to give it at least a normal image..

	if (normal != nullptr)	normalImage = normal->createCopy();
	if (over != nullptr)	  overImage = over->createCopy();
	if (down != nullptr)	  downImage = down->createCopy();
	if (disabled != nullptr)	  disabledImage = disabled->createCopy();
	if (normalOn != nullptr)	  normalImageOn = normalOn->createCopy();
	if (overOn != nullptr)	overImageOn = overOn->createCopy();
	if (downOn != nullptr)	downImageOn = downOn->createCopy();
	if (disabledOn != nullptr)	disabledImageOn = disabledOn->createCopy();

	buttonStateChanged();
}

void DrawableButton::setButtonStyle (const DrawableButton::ButtonStyle newStyle)
{
	if (style != newStyle)
	{
		style = newStyle;
		buttonStateChanged();
	}
}

void DrawableButton::setBackgroundColours (const Colour& toggledOffColour,
										   const Colour& toggledOnColour)
{
	if (backgroundOff != toggledOffColour
		 || backgroundOn != toggledOnColour)
	{
		backgroundOff = toggledOffColour;
		backgroundOn = toggledOnColour;

		repaint();
	}
}

const Colour& DrawableButton::getBackgroundColour() const noexcept
{
	return getToggleState() ? backgroundOn
							: backgroundOff;
}

void DrawableButton::setEdgeIndent (const int numPixelsIndent)
{
	edgeIndent = numPixelsIndent;
	repaint();
	resized();
}

void DrawableButton::resized()
{
	Button::resized();

	if (currentImage != nullptr)
	{
		if (style == ImageRaw)
		{
			currentImage->setOriginWithOriginalSize (Point<float>());
		}
		else
		{
			Rectangle<int> imageSpace;

			if (style == ImageOnButtonBackground)
			{
				imageSpace = getLocalBounds().reduced (getWidth() / 4, getHeight() / 4);
			}
			else
			{
				const int textH = (style == ImageAboveTextLabel) ? jmin (16, proportionOfHeight (0.25f)) : 0;

				const int indentX = jmin (edgeIndent, proportionOfWidth (0.3f));
				const int indentY = jmin (edgeIndent, proportionOfHeight (0.3f));

				imageSpace.setBounds (indentX, indentY,
									  getWidth() - indentX * 2,
									  getHeight() - indentY * 2 - textH);
			}

			currentImage->setTransformToFit (imageSpace.toFloat(), RectanglePlacement::centred);
		}
	}
}

void DrawableButton::buttonStateChanged()
{
	repaint();

	Drawable* imageToDraw = nullptr;
	float opacity = 1.0f;

	if (isEnabled())
	{
		imageToDraw = getCurrentImage();
	}
	else
	{
		imageToDraw = getToggleState() ? disabledImageOn
									   : disabledImage;

		if (imageToDraw == nullptr)
		{
			opacity = 0.4f;
			imageToDraw = getNormalImage();
		}
	}

	if (imageToDraw != currentImage)
	{
		removeChildComponent (currentImage);
		currentImage = imageToDraw;

		if (currentImage != nullptr)
		{
			currentImage->setInterceptsMouseClicks (false, false);
			addAndMakeVisible (currentImage);
			DrawableButton::resized();
		}
	}

	if (currentImage != nullptr)
		currentImage->setAlpha (opacity);
}

void DrawableButton::paintButton (Graphics& g,
								  bool isMouseOverButton,
								  bool isButtonDown)
{
	if (style == ImageOnButtonBackground)
	{
		getLookAndFeel().drawButtonBackground (g, *this,
											   getBackgroundColour(),
											   isMouseOverButton,
											   isButtonDown);
	}
	else
	{
		g.fillAll (getBackgroundColour());

		const int textH = (style == ImageAboveTextLabel)
							? jmin (16, proportionOfHeight (0.25f))
							: 0;

		if (textH > 0)
		{
			g.setFont ((float) textH);

			g.setColour (findColour (DrawableButton::textColourId)
							.withMultipliedAlpha (isEnabled() ? 1.0f : 0.4f));

			g.drawFittedText (getButtonText(),
							  2, getHeight() - textH - 1,
							  getWidth() - 4, textH,
							  Justification::centred, 1);
		}
	}
}

Drawable* DrawableButton::getCurrentImage() const noexcept
{
	if (isDown())
		return getDownImage();

	if (isOver())
		return getOverImage();

	return getNormalImage();
}

Drawable* DrawableButton::getNormalImage() const noexcept
{
	return (getToggleState() && normalImageOn != nullptr) ? normalImageOn
														  : normalImage;
}

Drawable* DrawableButton::getOverImage() const noexcept
{
	Drawable* d = normalImage;

	if (getToggleState())
	{
		if (overImageOn != nullptr)
			d = overImageOn;
		else if (normalImageOn != nullptr)
			d = normalImageOn;
		else if (overImage != nullptr)
			d = overImage;
	}
	else
	{
		if (overImage != nullptr)
			d = overImage;
	}

	return d;
}

Drawable* DrawableButton::getDownImage() const noexcept
{
	Drawable* d = normalImage;

	if (getToggleState())
	{
		if (downImageOn != nullptr)
			d = downImageOn;
		else if (overImageOn != nullptr)
			d = overImageOn;
		else if (normalImageOn != nullptr)
			d = normalImageOn;
		else if (downImage != nullptr)
			d = downImage;
		else
			d = getOverImage();
	}
	else
	{
		if (downImage != nullptr)
			d = downImage;
		else
			d = getOverImage();
	}

	return d;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DrawableButton.cpp ***/


/*** Start of inlined file: juce_HyperlinkButton.cpp ***/
BEGIN_JUCE_NAMESPACE

HyperlinkButton::HyperlinkButton (const String& linkText,
								  const URL& linkURL)
   : Button (linkText),
	 url (linkURL),
	 font (14.0f, Font::underlined),
	 resizeFont (true),
	 justification (Justification::centred)
{
	setMouseCursor (MouseCursor::PointingHandCursor);
	setTooltip (linkURL.toString (false));
}

HyperlinkButton::~HyperlinkButton()
{
}

void HyperlinkButton::setFont (const Font& newFont,
							   const bool resizeToMatchComponentHeight,
							   const Justification& justificationType)
{
	font = newFont;
	resizeFont = resizeToMatchComponentHeight;
	justification = justificationType;
	repaint();
}

void HyperlinkButton::setURL (const URL& newURL) noexcept
{
	url = newURL;
	setTooltip (newURL.toString (false));
}

Font HyperlinkButton::getFontToUse() const
{
	Font f (font);

	if (resizeFont)
		f.setHeight (getHeight() * 0.7f);

	return f;
}

void HyperlinkButton::changeWidthToFitText()
{
	setSize (getFontToUse().getStringWidth (getName()) + 6, getHeight());
}

void HyperlinkButton::colourChanged()
{
	repaint();
}

void HyperlinkButton::clicked()
{
	if (url.isWellFormed())
		url.launchInDefaultBrowser();
}

void HyperlinkButton::paintButton (Graphics& g,
								   bool isMouseOverButton,
								   bool isButtonDown)
{
	const Colour textColour (findColour (textColourId));

	if (isEnabled())
		g.setColour ((isMouseOverButton) ? textColour.darker ((isButtonDown) ? 1.3f : 0.4f)
										 : textColour);
	else
		g.setColour (textColour.withMultipliedAlpha (0.4f));

	g.setFont (getFontToUse());

	g.drawText (getButtonText(),
				2, 0, getWidth() - 2, getHeight(),
				justification.getOnlyHorizontalFlags() | Justification::verticallyCentred,
				true);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_HyperlinkButton.cpp ***/


/*** Start of inlined file: juce_ImageButton.cpp ***/
BEGIN_JUCE_NAMESPACE

ImageButton::ImageButton (const String& text_)
	: Button (text_),
	  scaleImageToFit (true),
	  preserveProportions (true),
	  alphaThreshold (0),
	  imageX (0),
	  imageY (0),
	  imageW (0),
	  imageH (0)
{
}

ImageButton::~ImageButton()
{
}

void ImageButton::setImages (const bool resizeButtonNowToFitThisImage,
							 const bool rescaleImagesWhenButtonSizeChanges,
							 const bool preserveImageProportions,
							 const Image& normalImage_,
							 const float imageOpacityWhenNormal,
							 const Colour& overlayColourWhenNormal,
							 const Image& overImage_,
							 const float imageOpacityWhenOver,
							 const Colour& overlayColourWhenOver,
							 const Image& downImage_,
							 const float imageOpacityWhenDown,
							 const Colour& overlayColourWhenDown,
							 const float hitTestAlphaThreshold)
{
	normalImage = normalImage_;
	overImage = overImage_;
	downImage = downImage_;

	if (resizeButtonNowToFitThisImage && normalImage.isValid())
	{
		imageW = normalImage.getWidth();
		imageH = normalImage.getHeight();

		setSize (imageW, imageH);
	}

	scaleImageToFit = rescaleImagesWhenButtonSizeChanges;
	preserveProportions = preserveImageProportions;

	normalOpacity = imageOpacityWhenNormal;
	normalOverlay = overlayColourWhenNormal;
	overOpacity   = imageOpacityWhenOver;
	overOverlay   = overlayColourWhenOver;
	downOpacity   = imageOpacityWhenDown;
	downOverlay   = overlayColourWhenDown;

	alphaThreshold = (unsigned char) jlimit (0, 0xff, roundToInt (255.0f * hitTestAlphaThreshold));

	repaint();
}

Image ImageButton::getCurrentImage() const
{
	if (isDown() || getToggleState())
		return getDownImage();

	if (isOver())
		return getOverImage();

	return getNormalImage();
}

Image ImageButton::getNormalImage() const
{
	return normalImage;
}

Image ImageButton::getOverImage() const
{
	return overImage.isValid() ? overImage
							   : normalImage;
}

Image ImageButton::getDownImage() const
{
	return downImage.isValid() ? downImage
							   : getOverImage();
}

void ImageButton::paintButton (Graphics& g,
							   bool isMouseOverButton,
							   bool isButtonDown)
{
	if (! isEnabled())
	{
		isMouseOverButton = false;
		isButtonDown = false;
	}

	Image im (getCurrentImage());

	if (im.isValid())
	{
		const int iw = im.getWidth();
		const int ih = im.getHeight();
		imageW = getWidth();
		imageH = getHeight();
		imageX = (imageW - iw) >> 1;
		imageY = (imageH - ih) >> 1;

		if (scaleImageToFit)
		{
			if (preserveProportions)
			{
				int newW, newH;
				const float imRatio = ih / (float)iw;
				const float destRatio = imageH / (float)imageW;

				if (imRatio > destRatio)
				{
					newW = roundToInt (imageH / imRatio);
					newH = imageH;
				}
				else
				{
					newW = imageW;
					newH = roundToInt (imageW * imRatio);
				}

				imageX = (imageW - newW) / 2;
				imageY = (imageH - newH) / 2;
				imageW = newW;
				imageH = newH;
			}
			else
			{
				imageX = 0;
				imageY = 0;
			}
		}

		if (! scaleImageToFit)
		{
			imageW = iw;
			imageH = ih;
		}

		getLookAndFeel().drawImageButton (g, &im, imageX, imageY, imageW, imageH,
										  isButtonDown ? downOverlay
													   : (isMouseOverButton ? overOverlay
																			: normalOverlay),
										  isButtonDown ? downOpacity
													   : (isMouseOverButton ? overOpacity
																			: normalOpacity),
										  *this);
	}
}

bool ImageButton::hitTest (int x, int y)
{
	if (alphaThreshold == 0)
		return true;

	Image im (getCurrentImage());

	return im.isNull() || (imageW > 0 && imageH > 0
							&& alphaThreshold < im.getPixelAt (((x - imageX) * im.getWidth()) / imageW,
															   ((y - imageY) * im.getHeight()) / imageH).getAlpha());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ImageButton.cpp ***/


/*** Start of inlined file: juce_ShapeButton.cpp ***/
BEGIN_JUCE_NAMESPACE

ShapeButton::ShapeButton (const String& text_,
						  const Colour& normalColour_,
						  const Colour& overColour_,
						  const Colour& downColour_)
  : Button (text_),
	normalColour (normalColour_),
	overColour (overColour_),
	downColour (downColour_),
	maintainShapeProportions (false),
	outlineWidth (0.0f)
{
}

ShapeButton::~ShapeButton()
{
}

void ShapeButton::setColours (const Colour& newNormalColour,
							  const Colour& newOverColour,
							  const Colour& newDownColour)
{
	normalColour = newNormalColour;
	overColour = newOverColour;
	downColour = newDownColour;
}

void ShapeButton::setOutline (const Colour& newOutlineColour,
							  const float newOutlineWidth)
{
	outlineColour = newOutlineColour;
	outlineWidth = newOutlineWidth;
}

void ShapeButton::setShape (const Path& newShape,
							const bool resizeNowToFitThisShape,
							const bool maintainShapeProportions_,
							const bool hasShadow)
{
	shape = newShape;
	maintainShapeProportions = maintainShapeProportions_;

	shadow.setShadowProperties (3.0f, 0.5f, 0, 0);
	setComponentEffect ((hasShadow) ? &shadow : 0);

	if (resizeNowToFitThisShape)
	{
		Rectangle<float> newBounds (shape.getBounds());

		if (hasShadow)
			newBounds.expand (4.0f, 4.0f);

		shape.applyTransform (AffineTransform::translation (-newBounds.getX(), -newBounds.getY()));

		setSize (1 + (int) (newBounds.getWidth() + outlineWidth),
				 1 + (int) (newBounds.getHeight() + outlineWidth));
	}
}

void ShapeButton::paintButton (Graphics& g, bool isMouseOverButton, bool isButtonDown)
{
	if (! isEnabled())
	{
		isMouseOverButton = false;
		isButtonDown = false;
	}

	g.setColour ((isButtonDown) ? downColour
								: (isMouseOverButton) ? overColour
													  : normalColour);

	int w = getWidth();
	int h = getHeight();

	if (getComponentEffect() != nullptr)
	{
		w -= 4;
		h -= 4;
	}

	const float offset = (outlineWidth * 0.5f) + (isButtonDown ? 1.5f : 0.0f);

	const AffineTransform trans (shape.getTransformToScaleToFit (offset, offset,
																 w - offset - outlineWidth,
																 h - offset - outlineWidth,
																 maintainShapeProportions));
	g.fillPath (shape, trans);

	if (outlineWidth > 0.0f)
	{
		g.setColour (outlineColour);
		g.strokePath (shape, PathStrokeType (outlineWidth), trans);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ShapeButton.cpp ***/


/*** Start of inlined file: juce_TextButton.cpp ***/
BEGIN_JUCE_NAMESPACE

TextButton::TextButton (const String& name,
						const String& toolTip)
	: Button (name)
{
	setTooltip (toolTip);
}

TextButton::~TextButton()
{
}

void TextButton::paintButton (Graphics& g,
							  bool isMouseOverButton,
							  bool isButtonDown)
{
	getLookAndFeel().drawButtonBackground (g, *this,
										   findColour (getToggleState() ? buttonOnColourId
																		: buttonColourId),
										   isMouseOverButton,
										   isButtonDown);

	getLookAndFeel().drawButtonText (g, *this,
									 isMouseOverButton,
									 isButtonDown);
}

void TextButton::colourChanged()
{
	repaint();
}

const Font TextButton::getFont()
{
	return Font (jmin (15.0f, getHeight() * 0.6f));
}

void TextButton::changeWidthToFitText (const int newHeight)
{
	if (newHeight >= 0)
		setSize (jmax (1, getWidth()), newHeight);

	setSize (getFont().getStringWidth (getButtonText()) + getHeight(),
			 getHeight());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TextButton.cpp ***/


/*** Start of inlined file: juce_ToggleButton.cpp ***/
BEGIN_JUCE_NAMESPACE

ToggleButton::ToggleButton (const String& buttonText)
	: Button (buttonText)
{
	setClickingTogglesState (true);
}

ToggleButton::~ToggleButton()
{
}

void ToggleButton::paintButton (Graphics& g,
								bool isMouseOverButton,
								bool isButtonDown)
{
	getLookAndFeel().drawToggleButton (g, *this,
									   isMouseOverButton,
									   isButtonDown);
}

void ToggleButton::changeWidthToFitText()
{
	getLookAndFeel().changeToggleButtonWidthToFitText (*this);
}

void ToggleButton::colourChanged()
{
	repaint();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ToggleButton.cpp ***/


/*** Start of inlined file: juce_ToolbarButton.cpp ***/
BEGIN_JUCE_NAMESPACE

ToolbarButton::ToolbarButton (const int itemId_, const String& buttonText,
							  Drawable* const normalImage_, Drawable* const toggledOnImage_)
   : ToolbarItemComponent (itemId_, buttonText, true),
	 normalImage (normalImage_),
	 toggledOnImage (toggledOnImage_),
	 currentImage (nullptr)
{
	jassert (normalImage_ != nullptr);
}

ToolbarButton::~ToolbarButton()
{
}

bool ToolbarButton::getToolbarItemSizes (int toolbarDepth, bool /*isToolbarVertical*/, int& preferredSize, int& minSize, int& maxSize)
{
	preferredSize = minSize = maxSize = toolbarDepth;
	return true;
}

void ToolbarButton::paintButtonArea (Graphics&, int /*width*/, int /*height*/, bool /*isMouseOver*/, bool /*isMouseDown*/)
{
}

void ToolbarButton::contentAreaChanged (const Rectangle<int>&)
{
	buttonStateChanged();
}

void ToolbarButton::updateDrawable()
{
	if (currentImage != nullptr)
	{
		currentImage->setTransformToFit (getContentArea().toFloat(), RectanglePlacement::centred);
		currentImage->setAlpha (isEnabled() ? 1.0f : 0.5f);
	}
}

void ToolbarButton::resized()
{
	ToolbarItemComponent::resized();
	updateDrawable();
}

void ToolbarButton::enablementChanged()
{
	ToolbarItemComponent::enablementChanged();
	updateDrawable();
}

void ToolbarButton::buttonStateChanged()
{
	Drawable* d = normalImage;

	if (getToggleState() && toggledOnImage != nullptr)
		d = toggledOnImage;

	if (d != currentImage)
	{
		removeChildComponent (currentImage);
		currentImage = d;

		if (d != nullptr)
		{
			enablementChanged();
			addAndMakeVisible (d);
			updateDrawable();
		}
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ToolbarButton.cpp ***/


/*** Start of inlined file: juce_CodeDocument.cpp ***/
BEGIN_JUCE_NAMESPACE

class CodeDocumentLine
{
public:
	CodeDocumentLine (const String::CharPointerType& line_,
					  const int lineLength_,
					  const int numNewLineChars,
					  const int lineStartInFile_)
		: line (line_, lineLength_),
		  lineStartInFile (lineStartInFile_),
		  lineLength (lineLength_),
		  lineLengthWithoutNewLines (lineLength_ - numNewLineChars)
	{
	}

	static void createLines (Array <CodeDocumentLine*>& newLines, const String& text)
	{
		String::CharPointerType t (text.getCharPointer());
		int charNumInFile = 0;
		bool finished = false;

		while (! (finished || t.isEmpty()))
		{
			String::CharPointerType startOfLine (t);
			int startOfLineInFile = charNumInFile;
			int lineLength = 0;
			int numNewLineChars = 0;

			for (;;)
			{
				const juce_wchar c = t.getAndAdvance();

				if (c == 0)
				{
					finished = true;
					break;
				}

				++charNumInFile;
				++lineLength;

				if (c == '\r')
				{
					++numNewLineChars;

					if (*t == '\n')
					{
						++t;
						++charNumInFile;
						++lineLength;
						++numNewLineChars;
					}

					break;
				}

				if (c == '\n')
				{
					++numNewLineChars;
					break;
				}
			}

			newLines.add (new CodeDocumentLine (startOfLine, lineLength,
												numNewLineChars, startOfLineInFile));
		}

		jassert (charNumInFile == text.length());
	}

	bool endsWithLineBreak() const noexcept
	{
		return lineLengthWithoutNewLines != lineLength;
	}

	void updateLength() noexcept
	{
		lineLength = 0;
		lineLengthWithoutNewLines = 0;

		String::CharPointerType t (line.getCharPointer());

		for (;;)
		{
			const juce_wchar c = t.getAndAdvance();

			if (c == 0)
				break;

			++lineLength;

			if (c != '\n' && c != '\r')
				lineLengthWithoutNewLines = lineLength;
		}
	}

	String line;
	int lineStartInFile, lineLength, lineLengthWithoutNewLines;
};

CodeDocument::Iterator::Iterator (CodeDocument* const document_)
	: document (document_),
	  charPointer (nullptr),
	  line (0),
	  position (0)
{
}

CodeDocument::Iterator::Iterator (const CodeDocument::Iterator& other)
	: document (other.document),
	  charPointer (other.charPointer),
	  line (other.line),
	  position (other.position)
{
}

CodeDocument::Iterator& CodeDocument::Iterator::operator= (const CodeDocument::Iterator& other) noexcept
{
	document = other.document;
	charPointer = other.charPointer;
	line = other.line;
	position = other.position;

	return *this;
}

CodeDocument::Iterator::~Iterator() noexcept
{
}

juce_wchar CodeDocument::Iterator::nextChar()
{
	for (;;)
	{
		if (charPointer.getAddress() == nullptr)
		{
			CodeDocumentLine* const l = document->lines[line];

			if (l == nullptr)
				return 0;

			charPointer = l->line.getCharPointer();
		}

		const juce_wchar result = charPointer.getAndAdvance();

		if (result == 0)
		{
			++line;
			charPointer = nullptr;
		}
		else
		{
			++position;
			return result;
		}
	}
}

void CodeDocument::Iterator::skip()
{
	nextChar();
}

void CodeDocument::Iterator::skipToEndOfLine()
{
	if (charPointer.getAddress() == nullptr)
	{
		CodeDocumentLine* const l = document->lines[line];

		if (l == nullptr)
			return;

		charPointer = l->line.getCharPointer();
	}

	position += (int) charPointer.length();
	++line;
	charPointer = nullptr;
}

juce_wchar CodeDocument::Iterator::peekNextChar() const
{
	if (charPointer.getAddress() == nullptr)
	{
		CodeDocumentLine* const l = document->lines[line];

		if (l == nullptr)
			return 0;

		charPointer = l->line.getCharPointer();
	}

	const juce_wchar c = *charPointer;

	if (c != 0)
		return c;

	CodeDocumentLine* const l = document->lines [line + 1];
	return l == nullptr ? 0 : l->line[0];
}

void CodeDocument::Iterator::skipWhitespace()
{
	while (CharacterFunctions::isWhitespace (peekNextChar()))
		skip();
}

bool CodeDocument::Iterator::isEOF() const noexcept
{
	return charPointer.getAddress() == nullptr && line >= document->lines.size();
}

CodeDocument::Position::Position() noexcept
	: owner (0), characterPos (0), line (0),
	  indexInLine (0), positionMaintained (false)
{
}

CodeDocument::Position::Position (const CodeDocument* const ownerDocument,
								  const int line_, const int indexInLine_) noexcept
	: owner (const_cast <CodeDocument*> (ownerDocument)),
	  characterPos (0), line (line_),
	  indexInLine (indexInLine_), positionMaintained (false)
{
	setLineAndIndex (line_, indexInLine_);
}

CodeDocument::Position::Position (const CodeDocument* const ownerDocument,
								  const int characterPos_) noexcept
	: owner (const_cast <CodeDocument*> (ownerDocument)),
	  positionMaintained (false)
{
	setPosition (characterPos_);
}

CodeDocument::Position::Position (const Position& other) noexcept
	: owner (other.owner), characterPos (other.characterPos), line (other.line),
	  indexInLine (other.indexInLine), positionMaintained (false)
{
	jassert (*this == other);
}

CodeDocument::Position::~Position()
{
	setPositionMaintained (false);
}

CodeDocument::Position& CodeDocument::Position::operator= (const Position& other)
{
	if (this != &other)
	{
		const bool wasPositionMaintained = positionMaintained;
		if (owner != other.owner)
			setPositionMaintained (false);

		owner = other.owner;
		line = other.line;
		indexInLine = other.indexInLine;
		characterPos = other.characterPos;
		setPositionMaintained (wasPositionMaintained);

		jassert (*this == other);
	}

	return *this;
}

bool CodeDocument::Position::operator== (const Position& other) const noexcept
{
	jassert ((characterPos == other.characterPos)
			   == (line == other.line && indexInLine == other.indexInLine));

	return characterPos == other.characterPos
			&& line == other.line
			&& indexInLine == other.indexInLine
			&& owner == other.owner;
}

bool CodeDocument::Position::operator!= (const Position& other) const noexcept
{
	return ! operator== (other);
}

void CodeDocument::Position::setLineAndIndex (const int newLineNum, const int newIndexInLine)
{
	jassert (owner != nullptr);

	if (owner->lines.size() == 0)
	{
		line = 0;
		indexInLine = 0;
		characterPos = 0;
	}
	else
	{
		if (newLineNum >= owner->lines.size())
		{
			line = owner->lines.size() - 1;

			CodeDocumentLine* const l = owner->lines.getUnchecked (line);
			jassert (l != nullptr);

			indexInLine = l->lineLengthWithoutNewLines;
			characterPos = l->lineStartInFile + indexInLine;
		}
		else
		{
			line = jmax (0, newLineNum);

			CodeDocumentLine* const l = owner->lines.getUnchecked (line);
			jassert (l != nullptr);

			if (l->lineLengthWithoutNewLines > 0)
				indexInLine = jlimit (0, l->lineLengthWithoutNewLines, newIndexInLine);
			else
				indexInLine = 0;

			characterPos = l->lineStartInFile + indexInLine;
		}
	}
}

void CodeDocument::Position::setPosition (const int newPosition)
{
	jassert (owner != nullptr);

	line = 0;
	indexInLine = 0;
	characterPos = 0;

	if (newPosition > 0)
	{
		int lineStart = 0;
		int lineEnd = owner->lines.size();

		for (;;)
		{
			if (lineEnd - lineStart < 4)
			{
				for (int i = lineStart; i < lineEnd; ++i)
				{
					CodeDocumentLine* const l = owner->lines.getUnchecked (i);

					int index = newPosition - l->lineStartInFile;

					if (index >= 0 && (index < l->lineLength || i == lineEnd - 1))
					{
						line = i;
						indexInLine = jmin (l->lineLengthWithoutNewLines, index);
						characterPos = l->lineStartInFile + indexInLine;
					}
				}

				break;
			}
			else
			{
				const int midIndex = (lineStart + lineEnd + 1) / 2;
				CodeDocumentLine* const mid = owner->lines.getUnchecked (midIndex);

				if (newPosition >= mid->lineStartInFile)
					lineStart = midIndex;
				else
					lineEnd = midIndex;
			}
		}
	}
}

void CodeDocument::Position::moveBy (int characterDelta)
{
	jassert (owner != nullptr);

	if (characterDelta == 1)
	{
		setPosition (getPosition());

		// If moving right, make sure we don't get stuck between the \r and \n characters..
		if (line < owner->lines.size())
		{
			CodeDocumentLine* const l = owner->lines.getUnchecked (line);
			if (indexInLine + characterDelta < l->lineLength
				 && indexInLine + characterDelta >= l->lineLengthWithoutNewLines + 1)
				++characterDelta;
		}
	}

	setPosition (characterPos + characterDelta);
}

CodeDocument::Position CodeDocument::Position::movedBy (const int characterDelta) const
{
	CodeDocument::Position p (*this);
	p.moveBy (characterDelta);
	return p;
}

CodeDocument::Position CodeDocument::Position::movedByLines (const int deltaLines) const
{
	CodeDocument::Position p (*this);
	p.setLineAndIndex (getLineNumber() + deltaLines, getIndexInLine());
	return p;
}

juce_wchar CodeDocument::Position::getCharacter() const
{
	const CodeDocumentLine* const l = owner->lines [line];
	return l == nullptr ? 0 : l->line [getIndexInLine()];
}

String CodeDocument::Position::getLineText() const
{
	const CodeDocumentLine* const l = owner->lines [line];
	return l == nullptr ? String::empty : l->line;
}

void CodeDocument::Position::setPositionMaintained (const bool isMaintained)
{
	if (isMaintained != positionMaintained)
	{
		positionMaintained = isMaintained;

		if (owner != nullptr)
		{
			if (isMaintained)
			{
				jassert (! owner->positionsToMaintain.contains (this));
				owner->positionsToMaintain.add (this);
			}
			else
			{
				// If this happens, you may have deleted the document while there are Position objects that are still using it...
				jassert (owner->positionsToMaintain.contains (this));
				owner->positionsToMaintain.removeValue (this);
			}
		}
	}
}

CodeDocument::CodeDocument()
	: undoManager (std::numeric_limits<int>::max(), 10000),
	  currentActionIndex (0),
	  indexOfSavedState (-1),
	  maximumLineLength (-1),
	  newLineChars ("\r\n")
{
}

CodeDocument::~CodeDocument()
{
}

String CodeDocument::getAllContent() const
{
	return getTextBetween (Position (this, 0),
						   Position (this, lines.size(), 0));
}

String CodeDocument::getTextBetween (const Position& start, const Position& end) const
{
	if (end.getPosition() <= start.getPosition())
		return String::empty;

	const int startLine = start.getLineNumber();
	const int endLine = end.getLineNumber();

	if (startLine == endLine)
	{
		CodeDocumentLine* const line = lines [startLine];
		return (line == nullptr) ? String::empty : line->line.substring (start.getIndexInLine(), end.getIndexInLine());
	}

	MemoryOutputStream mo;
	mo.preallocate (end.getPosition() - start.getPosition() + 4);

	const int maxLine = jmin (lines.size() - 1, endLine);

	for (int i = jmax (0, startLine); i <= maxLine; ++i)
	{
		const CodeDocumentLine* line = lines.getUnchecked(i);
		int len = line->lineLength;

		if (i == startLine)
		{
			const int index = start.getIndexInLine();
			mo << line->line.substring (index, len);
		}
		else if (i == endLine)
		{
			len = end.getIndexInLine();
			mo << line->line.substring (0, len);
		}
		else
		{
			mo << line->line;
		}
	}

	return mo.toString();
}

int CodeDocument::getNumCharacters() const noexcept
{
	const CodeDocumentLine* const lastLine = lines.getLast();
	return (lastLine == nullptr) ? 0 : lastLine->lineStartInFile + lastLine->lineLength;
}

String CodeDocument::getLine (const int lineIndex) const noexcept
{
	const CodeDocumentLine* const line = lines [lineIndex];
	return (line == nullptr) ? String::empty : line->line;
}

int CodeDocument::getMaximumLineLength() noexcept
{
	if (maximumLineLength < 0)
	{
		maximumLineLength = 0;

		for (int i = lines.size(); --i >= 0;)
			maximumLineLength = jmax (maximumLineLength, lines.getUnchecked(i)->lineLength);
	}

	return maximumLineLength;
}

void CodeDocument::deleteSection (const Position& startPosition, const Position& endPosition)
{
	remove (startPosition.getPosition(), endPosition.getPosition(), true);
}

void CodeDocument::insertText (const Position& position, const String& text)
{
	insert (text, position.getPosition(), true);
}

void CodeDocument::replaceAllContent (const String& newContent)
{
	remove (0, getNumCharacters(), true);
	insert (newContent, 0, true);
}

bool CodeDocument::loadFromStream (InputStream& stream)
{
	remove (0, getNumCharacters(), false);
	insert (stream.readEntireStreamAsString(), 0, false);
	setSavePoint();
	clearUndoHistory();
	return true;
}

bool CodeDocument::writeToStream (OutputStream& stream)
{
	for (int i = 0; i < lines.size(); ++i)
	{
		String temp (lines.getUnchecked(i)->line); // use a copy to avoid bloating the memory footprint of the stored string.
		const char* utf8 = temp.toUTF8();

		if (! stream.write (utf8, (int) strlen (utf8)))
			return false;
	}

	return true;
}

void CodeDocument::setNewLineCharacters (const String& newLineChars_) noexcept
{
	jassert (newLineChars_ == "\r\n" || newLineChars_ == "\n" || newLineChars_ == "\r");
	newLineChars = newLineChars_;
}

void CodeDocument::newTransaction()
{
	undoManager.beginNewTransaction (String::empty);
}

void CodeDocument::undo()
{
	newTransaction();
	undoManager.undo();
}

void CodeDocument::redo()
{
	undoManager.redo();
}

void CodeDocument::clearUndoHistory()
{
	undoManager.clearUndoHistory();
}

void CodeDocument::setSavePoint() noexcept
{
	indexOfSavedState = currentActionIndex;
}

bool CodeDocument::hasChangedSinceSavePoint() const noexcept
{
	return currentActionIndex != indexOfSavedState;
}

namespace CodeDocumentHelpers
{
	int getCharacterType (const juce_wchar character) noexcept
	{
		return (CharacterFunctions::isLetterOrDigit (character) || character == '_')
					? 2 : (CharacterFunctions::isWhitespace (character) ? 0 : 1);
	}
}

const CodeDocument::Position CodeDocument::findWordBreakAfter (const Position& position) const noexcept
{
	Position p (position);
	const int maxDistance = 256;
	int i = 0;

	while (i < maxDistance
			&& CharacterFunctions::isWhitespace (p.getCharacter())
			&& (i == 0 || (p.getCharacter() != '\n'
							&& p.getCharacter() != '\r')))
	{
		++i;
		p.moveBy (1);
	}

	if (i == 0)
	{
		const int type = CodeDocumentHelpers::getCharacterType (p.getCharacter());

		while (i < maxDistance && type == CodeDocumentHelpers::getCharacterType (p.getCharacter()))
		{
			++i;
			p.moveBy (1);
		}

		while (i < maxDistance
				&& CharacterFunctions::isWhitespace (p.getCharacter())
				&& (i == 0 || (p.getCharacter() != '\n'
								&& p.getCharacter() != '\r')))
		{
			++i;
			p.moveBy (1);
		}
	}

	return p;
}

const CodeDocument::Position CodeDocument::findWordBreakBefore (const Position& position) const noexcept
{
	Position p (position);
	const int maxDistance = 256;
	int i = 0;
	bool stoppedAtLineStart = false;

	while (i < maxDistance)
	{
		const juce_wchar c = p.movedBy (-1).getCharacter();

		if (c == '\r' || c == '\n')
		{
			stoppedAtLineStart = true;

			if (i > 0)
				break;
		}

		if (! CharacterFunctions::isWhitespace (c))
			break;

		p.moveBy (-1);
		++i;
	}

	if (i < maxDistance && ! stoppedAtLineStart)
	{
		const int type = CodeDocumentHelpers::getCharacterType (p.movedBy (-1).getCharacter());

		while (i < maxDistance && type == CodeDocumentHelpers::getCharacterType (p.movedBy (-1).getCharacter()))
		{
			p.moveBy (-1);
			++i;
		}
	}

	return p;
}

void CodeDocument::checkLastLineStatus()
{
	while (lines.size() > 0
			&& lines.getLast()->lineLength == 0
			&& (lines.size() == 1 || ! lines.getUnchecked (lines.size() - 2)->endsWithLineBreak()))
	{
		// remove any empty lines at the end if the preceding line doesn't end in a newline.
		lines.removeLast();
	}

	const CodeDocumentLine* const lastLine = lines.getLast();

	if (lastLine != nullptr && lastLine->endsWithLineBreak())
	{
		// check that there's an empty line at the end if the preceding one ends in a newline..
		lines.add (new CodeDocumentLine (String::empty.getCharPointer(), 0, 0, lastLine->lineStartInFile + lastLine->lineLength));
	}
}

void CodeDocument::addListener (CodeDocument::Listener* const listener) noexcept
{
	listeners.add (listener);
}

void CodeDocument::removeListener (CodeDocument::Listener* const listener) noexcept
{
	listeners.remove (listener);
}

void CodeDocument::sendListenerChangeMessage (const int startLine, const int endLine)
{
	Position startPos (this, startLine, 0);
	Position endPos (this, endLine, 0);

	listeners.call (&CodeDocument::Listener::codeDocumentChanged, startPos, endPos);
}

class CodeDocumentInsertAction   : public UndoableAction
{
public:
	CodeDocumentInsertAction (CodeDocument& owner_, const String& text_, const int insertPos_) noexcept
		: owner (owner_),
		  text (text_),
		  insertPos (insertPos_)
	{
	}

	bool perform()
	{
		owner.currentActionIndex++;
		owner.insert (text, insertPos, false);
		return true;
	}

	bool undo()
	{
		owner.currentActionIndex--;
		owner.remove (insertPos, insertPos + text.length(), false);
		return true;
	}

	int getSizeInUnits()	{ return text.length() + 32; }

private:
	CodeDocument& owner;
	const String text;
	int insertPos;

	JUCE_DECLARE_NON_COPYABLE (CodeDocumentInsertAction);
};

void CodeDocument::insert (const String& text, const int insertPos, const bool undoable)
{
	if (text.isEmpty())
		return;

	if (undoable)
	{
		undoManager.perform (new CodeDocumentInsertAction (*this, text, insertPos));
	}
	else
	{
		Position pos (this, insertPos);
		const int firstAffectedLine = pos.getLineNumber();
		int lastAffectedLine = firstAffectedLine + 1;

		CodeDocumentLine* const firstLine = lines [firstAffectedLine];
		String textInsideOriginalLine (text);

		if (firstLine != nullptr)
		{
			const int index = pos.getIndexInLine();
			textInsideOriginalLine = firstLine->line.substring (0, index)
									 + textInsideOriginalLine
									 + firstLine->line.substring (index);
		}

		maximumLineLength = -1;
		Array <CodeDocumentLine*> newLines;
		CodeDocumentLine::createLines (newLines, textInsideOriginalLine);
		jassert (newLines.size() > 0);

		CodeDocumentLine* const newFirstLine = newLines.getUnchecked (0);
		newFirstLine->lineStartInFile = firstLine != nullptr ? firstLine->lineStartInFile : 0;
		lines.set (firstAffectedLine, newFirstLine);

		if (newLines.size() > 1)
		{
			for (int i = 1; i < newLines.size(); ++i)
			{
				CodeDocumentLine* const l = newLines.getUnchecked (i);
				lines.insert (firstAffectedLine + i, l);
			}

			lastAffectedLine = lines.size();
		}

		int i, lineStart = newFirstLine->lineStartInFile;
		for (i = firstAffectedLine; i < lines.size(); ++i)
		{
			CodeDocumentLine* const l = lines.getUnchecked (i);
			l->lineStartInFile = lineStart;
			lineStart += l->lineLength;
		}

		checkLastLineStatus();

		const int newTextLength = text.length();
		for (i = 0; i < positionsToMaintain.size(); ++i)
		{
			CodeDocument::Position* const p = positionsToMaintain.getUnchecked(i);

			if (p->getPosition() >= insertPos)
				p->setPosition (p->getPosition() + newTextLength);
		}

		sendListenerChangeMessage (firstAffectedLine, lastAffectedLine);
	}
}

class CodeDocumentDeleteAction  : public UndoableAction
{
public:
	CodeDocumentDeleteAction (CodeDocument& owner_, const int startPos_, const int endPos_) noexcept
		: owner (owner_),
		  startPos (startPos_),
		  endPos (endPos_)
	{
		removedText = owner.getTextBetween (CodeDocument::Position (&owner, startPos),
											CodeDocument::Position (&owner, endPos));
	}

	bool perform()
	{
		owner.currentActionIndex++;
		owner.remove (startPos, endPos, false);
		return true;
	}

	bool undo()
	{
		owner.currentActionIndex--;
		owner.insert (removedText, startPos, false);
		return true;
	}

	int getSizeInUnits()	{ return removedText.length() + 32; }

private:
	CodeDocument& owner;
	int startPos, endPos;
	String removedText;

	JUCE_DECLARE_NON_COPYABLE (CodeDocumentDeleteAction);
};

void CodeDocument::remove (const int startPos, const int endPos, const bool undoable)
{
	if (endPos <= startPos)
		return;

	if (undoable)
	{
		undoManager.perform (new CodeDocumentDeleteAction (*this, startPos, endPos));
	}
	else
	{
		Position startPosition (this, startPos);
		Position endPosition (this, endPos);

		maximumLineLength = -1;
		const int firstAffectedLine = startPosition.getLineNumber();
		const int endLine = endPosition.getLineNumber();
		int lastAffectedLine = firstAffectedLine + 1;
		CodeDocumentLine* const firstLine = lines.getUnchecked (firstAffectedLine);

		if (firstAffectedLine == endLine)
		{
			firstLine->line = firstLine->line.substring (0, startPosition.getIndexInLine())
							+ firstLine->line.substring (endPosition.getIndexInLine());
			firstLine->updateLength();
		}
		else
		{
			lastAffectedLine = lines.size();

			CodeDocumentLine* const lastLine = lines.getUnchecked (endLine);
			jassert (lastLine != nullptr);

			firstLine->line = firstLine->line.substring (0, startPosition.getIndexInLine())
								+ lastLine->line.substring (endPosition.getIndexInLine());
			firstLine->updateLength();

			int numLinesToRemove = endLine - firstAffectedLine;
			lines.removeRange (firstAffectedLine + 1, numLinesToRemove);
		}

		int i;
		for (i = firstAffectedLine + 1; i < lines.size(); ++i)
		{
			CodeDocumentLine* const l = lines.getUnchecked (i);
			const CodeDocumentLine* const previousLine = lines.getUnchecked (i - 1);
			l->lineStartInFile = previousLine->lineStartInFile + previousLine->lineLength;
		}

		checkLastLineStatus();

		const int totalChars = getNumCharacters();

		for (i = 0; i < positionsToMaintain.size(); ++i)
		{
			CodeDocument::Position* p = positionsToMaintain.getUnchecked(i);

			if (p->getPosition() > startPosition.getPosition())
				p->setPosition (jmax (startPos, p->getPosition() + startPos - endPos));

			if (p->getPosition() > totalChars)
				p->setPosition (totalChars);
		}

		sendListenerChangeMessage (firstAffectedLine, lastAffectedLine);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_CodeDocument.cpp ***/


/*** Start of inlined file: juce_CodeEditorComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

class CodeEditorComponent::CodeEditorLine
{
public:
	CodeEditorLine() noexcept
	   : highlightColumnStart (0), highlightColumnEnd (0)
	{
	}

	bool update (CodeDocument& document, int lineNum,
				 CodeDocument::Iterator& source,
				 CodeTokeniser* analyser, const int spacesPerTab,
				 const CodeDocument::Position& selectionStart,
				 const CodeDocument::Position& selectionEnd)
	{
		Array <SyntaxToken> newTokens;
		newTokens.ensureStorageAllocated (8);

		if (analyser == nullptr)
		{
			newTokens.add (SyntaxToken (document.getLine (lineNum), -1));
		}
		else if (lineNum < document.getNumLines())
		{
			const CodeDocument::Position pos (&document, lineNum, 0);
			createTokens (pos.getPosition(), pos.getLineText(),
						  source, analyser, newTokens);
		}

		replaceTabsWithSpaces (newTokens, spacesPerTab);

		int newHighlightStart = 0;
		int newHighlightEnd = 0;

		if (selectionStart.getLineNumber() <= lineNum && selectionEnd.getLineNumber() >= lineNum)
		{
			const String line (document.getLine (lineNum));

			CodeDocument::Position lineStart (&document, lineNum, 0), lineEnd (&document, lineNum + 1, 0);
			newHighlightStart = indexToColumn (jmax (0, selectionStart.getPosition() - lineStart.getPosition()),
											   line, spacesPerTab);
			newHighlightEnd = indexToColumn (jmin (lineEnd.getPosition() - lineStart.getPosition(), selectionEnd.getPosition() - lineStart.getPosition()),
											 line, spacesPerTab);
		}

		if (newHighlightStart != highlightColumnStart || newHighlightEnd != highlightColumnEnd)
		{
			highlightColumnStart = newHighlightStart;
			highlightColumnEnd = newHighlightEnd;
		}
		else
		{
			if (tokens.size() == newTokens.size())
			{
				bool allTheSame = true;

				for (int i = newTokens.size(); --i >= 0;)
				{
					if (tokens.getReference(i) != newTokens.getReference(i))
					{
						allTheSame = false;
						break;
					}
				}

				if (allTheSame)
					return false;
			}
		}

		tokens.swapWithArray (newTokens);
		return true;
	}

	void draw (CodeEditorComponent& owner, Graphics& g, const Font& font,
			   float x, const int y, const int baselineOffset, const int lineHeight,
			   const Colour& highlightColour) const
	{
		if (highlightColumnStart < highlightColumnEnd)
		{
			g.setColour (highlightColour);
			g.fillRect (roundToInt (x + highlightColumnStart * owner.getCharWidth()), y,
						roundToInt ((highlightColumnEnd - highlightColumnStart) * owner.getCharWidth()), lineHeight);
		}

		int lastType = std::numeric_limits<int>::min();

		for (int i = 0; i < tokens.size(); ++i)
		{
			SyntaxToken& token = tokens.getReference(i);

			if (lastType != token.tokenType)
			{
				lastType = token.tokenType;
				g.setColour (owner.getColourForTokenType (lastType));
			}

			g.drawSingleLineText (token.text, roundToInt (x), y + baselineOffset);

			if (i < tokens.size() - 1)
			{
				if (token.width < 0)
					token.width = font.getStringWidthFloat (token.text);

				x += token.width;
			}
		}
	}

private:
	struct SyntaxToken
	{
		SyntaxToken (const String& text_, const int type) noexcept
			: text (text_), tokenType (type), width (-1.0f)
		{
		}

		bool operator!= (const SyntaxToken& other) const noexcept
		{
			return text != other.text || tokenType != other.tokenType;
		}

		String text;
		int tokenType;
		float width;
	};

	Array <SyntaxToken> tokens;
	int highlightColumnStart, highlightColumnEnd;

	static void createTokens (int startPosition, const String& lineText,
							  CodeDocument::Iterator& source,
							  CodeTokeniser* analyser,
							  Array <SyntaxToken>& newTokens)
	{
		CodeDocument::Iterator lastIterator (source);
		const int lineLength = lineText.length();

		for (;;)
		{
			int tokenType = analyser->readNextToken (source);
			int tokenStart = lastIterator.getPosition();
			int tokenEnd = source.getPosition();

			if (tokenEnd <= tokenStart)
				break;

			tokenEnd -= startPosition;

			if (tokenEnd > 0)
			{
				tokenStart -= startPosition;
				newTokens.add (SyntaxToken (lineText.substring (jmax (0, tokenStart), tokenEnd),
											tokenType));

				if (tokenEnd >= lineLength)
					break;
			}

			lastIterator = source;
		}

		source = lastIterator;
	}

	static void replaceTabsWithSpaces (Array <SyntaxToken>& tokens, const int spacesPerTab)
	{
		int x = 0;
		for (int i = 0; i < tokens.size(); ++i)
		{
			SyntaxToken& t = tokens.getReference(i);

			for (;;)
			{
				int tabPos = t.text.indexOfChar ('\t');
				if (tabPos < 0)
					break;

				const int spacesNeeded = spacesPerTab - ((tabPos + x) % spacesPerTab);
				t.text = t.text.replaceSection (tabPos, 1, String::repeatedString (" ", spacesNeeded));
			}

			x += t.text.length();
		}
	}

	int indexToColumn (int index, const String& line, int spacesPerTab) const noexcept
	{
		jassert (index <= line.length());

		String::CharPointerType t (line.getCharPointer());
		int col = 0;
		for (int i = 0; i < index; ++i)
		{
			if (t.getAndAdvance() != '\t')
				++col;
			else
				col += spacesPerTab - (col % spacesPerTab);
		}

		return col;
	}
};

CodeEditorComponent::CodeEditorComponent (CodeDocument& document_,
										  CodeTokeniser* const codeTokeniser_)
	: document (document_),
	  firstLineOnScreen (0),
	  gutter (5),
	  spacesPerTab (4),
	  lineHeight (0),
	  linesOnScreen (0),
	  columnsOnScreen (0),
	  scrollbarThickness (16),
	  columnToTryToMaintain (-1),
	  useSpacesForTabs (false),
	  xOffset (0),
	  verticalScrollBar (true),
	  horizontalScrollBar (false),
	  codeTokeniser (codeTokeniser_)
{
	caretPos = CodeDocument::Position (&document_, 0, 0);
	caretPos.setPositionMaintained (true);

	selectionStart = CodeDocument::Position (&document_, 0, 0);
	selectionStart.setPositionMaintained (true);

	selectionEnd = CodeDocument::Position (&document_, 0, 0);
	selectionEnd.setPositionMaintained (true);

	setOpaque (true);
	setMouseCursor (MouseCursor (MouseCursor::IBeamCursor));
	setWantsKeyboardFocus (true);

	addAndMakeVisible (&verticalScrollBar);
	verticalScrollBar.setSingleStepSize (1.0);

	addAndMakeVisible (&horizontalScrollBar);
	horizontalScrollBar.setSingleStepSize (1.0);

	addAndMakeVisible (caret = getLookAndFeel().createCaretComponent (this));

	Font f (12.0f);
	f.setTypefaceName (Font::getDefaultMonospacedFontName());
	setFont (f);

	resetToDefaultColours();

	verticalScrollBar.addListener (this);
	horizontalScrollBar.addListener (this);
	document.addListener (this);
}

CodeEditorComponent::~CodeEditorComponent()
{
	document.removeListener (this);
}

void CodeEditorComponent::loadContent (const String& newContent)
{
	clearCachedIterators (0);
	document.replaceAllContent (newContent);
	document.clearUndoHistory();
	document.setSavePoint();
	caretPos.setPosition (0);
	selectionStart.setPosition (0);
	selectionEnd.setPosition (0);
	scrollToLine (0);
}

bool CodeEditorComponent::isTextInputActive() const
{
	return true;
}

void CodeEditorComponent::setTemporaryUnderlining (const Array <Range<int> >&)
{
	jassertfalse; // TODO Windows IME not yet supported for this comp..
}

const Rectangle<int> CodeEditorComponent::getCaretRectangle()
{
	return getLocalArea (caret, caret->getLocalBounds());
}

void CodeEditorComponent::codeDocumentChanged (const CodeDocument::Position& affectedTextStart,
											   const CodeDocument::Position& affectedTextEnd)
{
	clearCachedIterators (affectedTextStart.getLineNumber());

	triggerAsyncUpdate();

	updateCaretPosition();
	columnToTryToMaintain = -1;

	if (affectedTextEnd.getPosition() >= selectionStart.getPosition()
		 && affectedTextStart.getPosition() <= selectionEnd.getPosition())
		deselectAll();

	if (caretPos.getPosition() > affectedTextEnd.getPosition()
		 || caretPos.getPosition() < affectedTextStart.getPosition())
		moveCaretTo (affectedTextStart, false);

	updateScrollBars();
}

void CodeEditorComponent::resized()
{
	linesOnScreen = (getHeight() - scrollbarThickness) / lineHeight;
	columnsOnScreen = (int) ((getWidth() - scrollbarThickness) / charWidth);
	lines.clear();
	rebuildLineTokens();
	updateCaretPosition();

	verticalScrollBar.setBounds (getWidth() - scrollbarThickness, 0, scrollbarThickness, getHeight() - scrollbarThickness);
	horizontalScrollBar.setBounds (gutter, getHeight() - scrollbarThickness, getWidth() - scrollbarThickness - gutter, scrollbarThickness);
	updateScrollBars();
}

void CodeEditorComponent::paint (Graphics& g)
{
	handleUpdateNowIfNeeded();

	g.fillAll (findColour (CodeEditorComponent::backgroundColourId));

	g.reduceClipRegion (gutter, 0, verticalScrollBar.getX() - gutter, horizontalScrollBar.getY());

	g.setFont (font);
	const int baselineOffset = (int) font.getAscent();
	const Colour defaultColour (findColour (CodeEditorComponent::defaultTextColourId));
	const Colour highlightColour (findColour (CodeEditorComponent::highlightColourId));

	const Rectangle<int> clip (g.getClipBounds());
	const int firstLineToDraw = jmax (0, clip.getY() / lineHeight);
	const int lastLineToDraw = jmin (lines.size(), clip.getBottom() / lineHeight + 1);

	for (int j = firstLineToDraw; j < lastLineToDraw; ++j)
	{
		lines.getUnchecked(j)->draw (*this, g, font,
									 (float) (gutter - xOffset * charWidth),
									 lineHeight * j, baselineOffset, lineHeight,
									 highlightColour);
	}
}

void CodeEditorComponent::setScrollbarThickness (const int thickness)
{
	if (scrollbarThickness != thickness)
	{
		scrollbarThickness = thickness;
		resized();
	}
}

void CodeEditorComponent::handleAsyncUpdate()
{
	rebuildLineTokens();
}

void CodeEditorComponent::rebuildLineTokens()
{
	cancelPendingUpdate();

	const int numNeeded = linesOnScreen + 1;

	int minLineToRepaint = numNeeded;
	int maxLineToRepaint = 0;

	if (numNeeded != lines.size())
	{
		lines.clear();

		for (int i = numNeeded; --i >= 0;)
			lines.add (new CodeEditorLine());

		minLineToRepaint = 0;
		maxLineToRepaint = numNeeded;
	}

	jassert (numNeeded == lines.size());

	CodeDocument::Iterator source (&document);
	getIteratorForPosition (CodeDocument::Position (&document, firstLineOnScreen, 0).getPosition(), source);

	for (int i = 0; i < numNeeded; ++i)
	{
		CodeEditorLine* const line = lines.getUnchecked(i);

		if (line->update (document, firstLineOnScreen + i, source, codeTokeniser, spacesPerTab,
						  selectionStart, selectionEnd))
		{
			minLineToRepaint = jmin (minLineToRepaint, i);
			maxLineToRepaint = jmax (maxLineToRepaint, i);
		}
	}

	if (minLineToRepaint <= maxLineToRepaint)
	{
		repaint (gutter, lineHeight * minLineToRepaint - 1,
				 verticalScrollBar.getX() - gutter,
				 lineHeight * (1 + maxLineToRepaint - minLineToRepaint) + 2);
	}
}

void CodeEditorComponent::updateCaretPosition()
{
	caret->setCaretPosition (getCharacterBounds (getCaretPos()));
}

void CodeEditorComponent::moveCaretTo (const CodeDocument::Position& newPos, const bool highlighting)
{
	caretPos = newPos;
	columnToTryToMaintain = -1;

	if (highlighting)
	{
		if (dragType == notDragging)
		{
			if (abs (caretPos.getPosition() - selectionStart.getPosition())
				  < abs (caretPos.getPosition() - selectionEnd.getPosition()))
				dragType = draggingSelectionStart;
			else
				dragType = draggingSelectionEnd;
		}

		if (dragType == draggingSelectionStart)
		{
			selectionStart = caretPos;

			if (selectionEnd.getPosition() < selectionStart.getPosition())
			{
				const CodeDocument::Position temp (selectionStart);
				selectionStart = selectionEnd;
				selectionEnd = temp;

				dragType = draggingSelectionEnd;
			}
		}
		else
		{
			selectionEnd = caretPos;

			if (selectionEnd.getPosition() < selectionStart.getPosition())
			{
				const CodeDocument::Position temp (selectionStart);
				selectionStart = selectionEnd;
				selectionEnd = temp;

				dragType = draggingSelectionStart;
			}
		}

		triggerAsyncUpdate();
	}
	else
	{
		deselectAll();
	}

	updateCaretPosition();
	scrollToKeepCaretOnScreen();
	updateScrollBars();
}

void CodeEditorComponent::deselectAll()
{
	if (selectionStart != selectionEnd)
		triggerAsyncUpdate();

	selectionStart = caretPos;
	selectionEnd = caretPos;
}

void CodeEditorComponent::updateScrollBars()
{
	verticalScrollBar.setRangeLimits (0, jmax (document.getNumLines(), firstLineOnScreen + linesOnScreen));
	verticalScrollBar.setCurrentRange (firstLineOnScreen, linesOnScreen);

	horizontalScrollBar.setRangeLimits (0, jmax ((double) document.getMaximumLineLength(), xOffset + columnsOnScreen));
	horizontalScrollBar.setCurrentRange (xOffset, columnsOnScreen);
}

void CodeEditorComponent::scrollToLineInternal (int newFirstLineOnScreen)
{
	newFirstLineOnScreen = jlimit (0, jmax (0, document.getNumLines() - 1),
								   newFirstLineOnScreen);

	if (newFirstLineOnScreen != firstLineOnScreen)
	{
		firstLineOnScreen = newFirstLineOnScreen;
		updateCaretPosition();

		updateCachedIterators (firstLineOnScreen);
		triggerAsyncUpdate();
	}
}

void CodeEditorComponent::scrollToColumnInternal (double column)
{
	const double newOffset = jlimit (0.0, document.getMaximumLineLength() + 3.0, column);

	if (xOffset != newOffset)
	{
		xOffset = newOffset;
		updateCaretPosition();
		repaint();
	}
}

void CodeEditorComponent::scrollToLine (int newFirstLineOnScreen)
{
	scrollToLineInternal (newFirstLineOnScreen);
	updateScrollBars();
}

void CodeEditorComponent::scrollToColumn (int newFirstColumnOnScreen)
{
	scrollToColumnInternal (newFirstColumnOnScreen);
	updateScrollBars();
}

void CodeEditorComponent::scrollBy (int deltaLines)
{
	scrollToLine (firstLineOnScreen + deltaLines);
}

void CodeEditorComponent::scrollToKeepCaretOnScreen()
{
	if (caretPos.getLineNumber() < firstLineOnScreen)
		scrollBy (caretPos.getLineNumber() - firstLineOnScreen);
	else if (caretPos.getLineNumber() >= firstLineOnScreen + linesOnScreen)
		scrollBy (caretPos.getLineNumber() - (firstLineOnScreen + linesOnScreen - 1));

	const int column = indexToColumn (caretPos.getLineNumber(), caretPos.getIndexInLine());
	if (column >= xOffset + columnsOnScreen - 1)
		scrollToColumn (column + 1 - columnsOnScreen);
	else if (column < xOffset)
		scrollToColumn (column);
}

const Rectangle<int> CodeEditorComponent::getCharacterBounds (const CodeDocument::Position& pos) const
{
	return Rectangle<int> (roundToInt ((gutter - xOffset * charWidth) + indexToColumn (pos.getLineNumber(), pos.getIndexInLine()) * charWidth),
						   (pos.getLineNumber() - firstLineOnScreen) * lineHeight,
						   roundToInt (charWidth),
						   lineHeight);
}

const CodeDocument::Position CodeEditorComponent::getPositionAt (int x, int y)
{
	const int line = y / lineHeight + firstLineOnScreen;
	const int column = roundToInt ((x - (gutter - xOffset * charWidth)) / charWidth);
	const int index = columnToIndex (line, column);

	return CodeDocument::Position (&document, line, index);
}

void CodeEditorComponent::insertTextAtCaret (const String& newText)
{
	document.deleteSection (selectionStart, selectionEnd);

	if (newText.isNotEmpty())
		document.insertText (caretPos, newText);

	scrollToKeepCaretOnScreen();
}

void CodeEditorComponent::insertTabAtCaret()
{
	if (CharacterFunctions::isWhitespace (caretPos.getCharacter())
		 && caretPos.getLineNumber() == caretPos.movedBy (1).getLineNumber())
	{
		moveCaretTo (document.findWordBreakAfter (caretPos), false);
	}

	if (useSpacesForTabs)
	{
		const int caretCol = indexToColumn (caretPos.getLineNumber(), caretPos.getIndexInLine());
		const int spacesNeeded = spacesPerTab - (caretCol % spacesPerTab);
		insertTextAtCaret (String::repeatedString (" ", spacesNeeded));
	}
	else
	{
		insertTextAtCaret ("\t");
	}
}

void CodeEditorComponent::cut()
{
	insertTextAtCaret (String::empty);
}

bool CodeEditorComponent::copyToClipboard()
{
	newTransaction();

	const String selection (document.getTextBetween (selectionStart, selectionEnd));

	if (selection.isNotEmpty())
		SystemClipboard::copyTextToClipboard (selection);

	return true;
}

bool CodeEditorComponent::cutToClipboard()
{
	copyToClipboard();
	cut();
	newTransaction();
	return true;
}

bool CodeEditorComponent::pasteFromClipboard()
{
	newTransaction();
	const String clip (SystemClipboard::getTextFromClipboard());

	if (clip.isNotEmpty())
		insertTextAtCaret (clip);

	newTransaction();
	return true;
}

bool CodeEditorComponent::moveCaretLeft (const bool moveInWholeWordSteps, const bool selecting)
{
	newTransaction();

	if (moveInWholeWordSteps)
		moveCaretTo (document.findWordBreakBefore (caretPos), selecting);
	else
		moveCaretTo (caretPos.movedBy (-1), selecting);

	return true;
}

bool CodeEditorComponent::moveCaretRight (const bool moveInWholeWordSteps, const bool selecting)
{
	newTransaction();

	if (moveInWholeWordSteps)
		moveCaretTo (document.findWordBreakAfter (caretPos), selecting);
	else
		moveCaretTo (caretPos.movedBy (1), selecting);

	return true;
}

void CodeEditorComponent::moveLineDelta (const int delta, const bool selecting)
{
	CodeDocument::Position pos (caretPos);
	const int newLineNum = pos.getLineNumber() + delta;

	if (columnToTryToMaintain < 0)
		columnToTryToMaintain = indexToColumn (pos.getLineNumber(), pos.getIndexInLine());

	pos.setLineAndIndex (newLineNum, columnToIndex (newLineNum, columnToTryToMaintain));

	const int colToMaintain = columnToTryToMaintain;
	moveCaretTo (pos, selecting);
	columnToTryToMaintain = colToMaintain;
}

bool CodeEditorComponent::moveCaretDown (const bool selecting)
{
	newTransaction();

	if (caretPos.getLineNumber() == document.getNumLines() - 1)
		moveCaretTo (CodeDocument::Position (&document, std::numeric_limits<int>::max(), std::numeric_limits<int>::max()), selecting);
	else
		moveLineDelta (1, selecting);

	return true;
}

bool CodeEditorComponent::moveCaretUp (const bool selecting)
{
	newTransaction();

	if (caretPos.getLineNumber() == 0)
		moveCaretTo (CodeDocument::Position (&document, 0, 0), selecting);
	else
		moveLineDelta (-1, selecting);

	return true;
}

bool CodeEditorComponent::pageDown (const bool selecting)
{
	newTransaction();
	scrollBy (jlimit (0, linesOnScreen, 1 + document.getNumLines() - firstLineOnScreen - linesOnScreen));
	moveLineDelta (linesOnScreen, selecting);
	return true;
}

bool CodeEditorComponent::pageUp (const bool selecting)
{
	newTransaction();
	scrollBy (-linesOnScreen);
	moveLineDelta (-linesOnScreen, selecting);
	return true;
}

bool CodeEditorComponent::scrollUp()
{
	newTransaction();
	scrollBy (1);

	if (caretPos.getLineNumber() < firstLineOnScreen)
		moveLineDelta (1, false);

	return true;
}

bool CodeEditorComponent::scrollDown()
{
	newTransaction();
	scrollBy (-1);

	if (caretPos.getLineNumber() >= firstLineOnScreen + linesOnScreen)
		moveLineDelta (-1, false);

	return true;
}

bool CodeEditorComponent::moveCaretToTop (const bool selecting)
{
	newTransaction();
	moveCaretTo (CodeDocument::Position (&document, 0, 0), selecting);
	return true;
}

namespace CodeEditorHelpers
{
	int findFirstNonWhitespaceChar (const String& line) noexcept
	{
		String::CharPointerType t (line.getCharPointer());
		int i = 0;

		while (! t.isEmpty())
		{
			if (! t.isWhitespace())
				return i;

			++t;
			++i;
		}

		return 0;
	}
}

bool CodeEditorComponent::moveCaretToStartOfLine (const bool selecting)
{
	newTransaction();

	int index = CodeEditorHelpers::findFirstNonWhitespaceChar (caretPos.getLineText());

	if (index >= caretPos.getIndexInLine() && caretPos.getIndexInLine() > 0)
		index = 0;

	moveCaretTo (CodeDocument::Position (&document, caretPos.getLineNumber(), index), selecting);
	return true;
}

bool CodeEditorComponent::moveCaretToEnd (const bool selecting)
{
	newTransaction();
	moveCaretTo (CodeDocument::Position (&document, std::numeric_limits<int>::max(), std::numeric_limits<int>::max()), selecting);
	return true;
}

bool CodeEditorComponent::moveCaretToEndOfLine (const bool selecting)
{
	newTransaction();
	moveCaretTo (CodeDocument::Position (&document, caretPos.getLineNumber(), std::numeric_limits<int>::max()), selecting);
	return true;
}

bool CodeEditorComponent::deleteBackwards (const bool moveInWholeWordSteps)
{
	if (moveInWholeWordSteps)
	{
		cut(); // in case something is already highlighted
		moveCaretTo (document.findWordBreakBefore (caretPos), true);
	}
	else
	{
		if (selectionStart == selectionEnd)
			selectionStart.moveBy (-1);
	}

	cut();
	return true;
}

bool CodeEditorComponent::deleteForwards (const bool moveInWholeWordSteps)
{
	if (moveInWholeWordSteps)
	{
		cut(); // in case something is already highlighted
		moveCaretTo (document.findWordBreakAfter (caretPos), true);
	}
	else
	{
		if (selectionStart == selectionEnd)
			selectionEnd.moveBy (1);
		else
			newTransaction();
	}

	cut();
	return true;
}

bool CodeEditorComponent::selectAll()
{
	newTransaction();
	moveCaretTo (CodeDocument::Position (&document, std::numeric_limits<int>::max(), std::numeric_limits<int>::max()), false);
	moveCaretTo (CodeDocument::Position (&document, 0, 0), true);
	return true;
}

bool CodeEditorComponent::undo()
{
	document.undo();
	scrollToKeepCaretOnScreen();
	return true;
}

bool CodeEditorComponent::redo()
{
	document.redo();
	scrollToKeepCaretOnScreen();
	return true;
}

void CodeEditorComponent::newTransaction()
{
	document.newTransaction();
	startTimer (600);
}

void CodeEditorComponent::timerCallback()
{
	newTransaction();
}

const Range<int> CodeEditorComponent::getHighlightedRegion() const
{
	return Range<int> (selectionStart.getPosition(), selectionEnd.getPosition());
}

void CodeEditorComponent::setHighlightedRegion (const Range<int>& newRange)
{
	moveCaretTo (CodeDocument::Position (&document, newRange.getStart()), false);
	moveCaretTo (CodeDocument::Position (&document, newRange.getEnd()), true);
}

const String CodeEditorComponent::getTextInRange (const Range<int>& range) const
{
	return document.getTextBetween (CodeDocument::Position (&document, range.getStart()),
									CodeDocument::Position (&document, range.getEnd()));
}

bool CodeEditorComponent::keyPressed (const KeyPress& key)
{
	if (! TextEditorKeyMapper<CodeEditorComponent>::invokeKeyFunction (*this, key))
	{
		if (key == KeyPress::tabKey || key.getTextCharacter() == '\t')
		{
			insertTabAtCaret();
		}
		else if (key == KeyPress::returnKey)
		{
			newTransaction();
			insertTextAtCaret (document.getNewLineCharacters());
		}
		else if (key.isKeyCode (KeyPress::escapeKey))
		{
			newTransaction();
		}
		else if (key.getTextCharacter() >= ' ')
		{
			insertTextAtCaret (String::charToString (key.getTextCharacter()));
		}
		else
		{
			return false;
		}
	}

	return true;
}

void CodeEditorComponent::mouseDown (const MouseEvent& e)
{
	newTransaction();
	dragType = notDragging;

	if (! e.mods.isPopupMenu())
	{
		beginDragAutoRepeat (100);
		moveCaretTo (getPositionAt (e.x, e.y), e.mods.isShiftDown());
	}
	else
	{
		/*PopupMenu m;
		addPopupMenuItems (m, &e);

		const int result = m.show();

		if (result != 0)
			performPopupMenuAction (result);
		*/
	}
}

void CodeEditorComponent::mouseDrag (const MouseEvent& e)
{
	if (! e.mods.isPopupMenu())
		moveCaretTo (getPositionAt (e.x, e.y), true);
}

void CodeEditorComponent::mouseUp (const MouseEvent&)
{
	newTransaction();
	beginDragAutoRepeat (0);
	dragType = notDragging;
}

void CodeEditorComponent::mouseDoubleClick (const MouseEvent& e)
{
	CodeDocument::Position tokenStart (getPositionAt (e.x, e.y));
	CodeDocument::Position tokenEnd (tokenStart);

	if (e.getNumberOfClicks() > 2)
	{
		tokenStart.setLineAndIndex (tokenStart.getLineNumber(), 0);
		tokenEnd.setLineAndIndex (tokenStart.getLineNumber() + 1, 0);
	}
	else
	{
		while (CharacterFunctions::isLetterOrDigit (tokenEnd.getCharacter()))
			tokenEnd.moveBy (1);

		tokenStart = tokenEnd;

		while (tokenStart.getIndexInLine() > 0
				&& CharacterFunctions::isLetterOrDigit (tokenStart.movedBy (-1).getCharacter()))
			tokenStart.moveBy (-1);
	}

	moveCaretTo (tokenEnd, false);
	moveCaretTo (tokenStart, true);
}

void CodeEditorComponent::mouseWheelMove (const MouseEvent& e, float wheelIncrementX, float wheelIncrementY)
{
	if ((verticalScrollBar.isVisible() && wheelIncrementY != 0)
		 || (horizontalScrollBar.isVisible() && wheelIncrementX != 0))
	{
		verticalScrollBar.mouseWheelMove (e, 0, wheelIncrementY);
		horizontalScrollBar.mouseWheelMove (e, wheelIncrementX, 0);
	}
	else
	{
		Component::mouseWheelMove (e, wheelIncrementX, wheelIncrementY);
	}
}

void CodeEditorComponent::scrollBarMoved (ScrollBar* scrollBarThatHasMoved, double newRangeStart)
{
	if (scrollBarThatHasMoved == &verticalScrollBar)
		scrollToLineInternal ((int) newRangeStart);
	else
		scrollToColumnInternal (newRangeStart);
}

void CodeEditorComponent::focusGained (FocusChangeType)
{
	updateCaretPosition();
}

void CodeEditorComponent::focusLost (FocusChangeType)
{
	updateCaretPosition();
}

void CodeEditorComponent::setTabSize (const int numSpaces, const bool insertSpaces)
{
	useSpacesForTabs = insertSpaces;

	if (spacesPerTab != numSpaces)
	{
		spacesPerTab = numSpaces;
		triggerAsyncUpdate();
	}
}

int CodeEditorComponent::indexToColumn (int lineNum, int index) const noexcept
{
	String::CharPointerType t (document.getLine (lineNum).getCharPointer());

	int col = 0;
	for (int i = 0; i < index; ++i)
	{
		if (t.isEmpty())
		{
			jassertfalse;
			break;
		}

		if (t.getAndAdvance() != '\t')
			++col;
		else
			col += getTabSize() - (col % getTabSize());
	}

	return col;
}

int CodeEditorComponent::columnToIndex (int lineNum, int column) const noexcept
{
	String::CharPointerType t (document.getLine (lineNum).getCharPointer());

	int i = 0, col = 0;

	while (! t.isEmpty())
	{
		if (t.getAndAdvance() != '\t')
			++col;
		else
			col += getTabSize() - (col % getTabSize());

		if (col > column)
			break;

		++i;
	}

	return i;
}

void CodeEditorComponent::setFont (const Font& newFont)
{
	font = newFont;
	charWidth = font.getStringWidthFloat ("0");
	lineHeight = roundToInt (font.getHeight());
	resized();
}

void CodeEditorComponent::resetToDefaultColours()
{
	coloursForTokenCategories.clear();

	if (codeTokeniser != nullptr)
	{
		for (int i = codeTokeniser->getTokenTypes().size(); --i >= 0;)
			setColourForTokenType (i, codeTokeniser->getDefaultColour (i));
	}
}

void CodeEditorComponent::setColourForTokenType (const int tokenType, const Colour& colour)
{
	jassert (tokenType < 256);

	while (coloursForTokenCategories.size() < tokenType)
		coloursForTokenCategories.add (Colours::black);

	coloursForTokenCategories.set (tokenType, colour);
	repaint();
}

const Colour CodeEditorComponent::getColourForTokenType (const int tokenType) const
{
	if (! isPositiveAndBelow (tokenType, coloursForTokenCategories.size()))
		return findColour (CodeEditorComponent::defaultTextColourId);

	return coloursForTokenCategories.getReference (tokenType);
}

void CodeEditorComponent::clearCachedIterators (const int firstLineToBeInvalid)
{
	int i;
	for (i = cachedIterators.size(); --i >= 0;)
		if (cachedIterators.getUnchecked (i)->getLine() < firstLineToBeInvalid)
			break;

	cachedIterators.removeRange (jmax (0, i - 1), cachedIterators.size());
}

void CodeEditorComponent::updateCachedIterators (int maxLineNum)
{
	const int maxNumCachedPositions = 5000;
	const int linesBetweenCachedSources = jmax (10, document.getNumLines() / maxNumCachedPositions);

	if (cachedIterators.size() == 0)
		cachedIterators.add (new CodeDocument::Iterator (&document));

	if (codeTokeniser == nullptr)
		return;

	for (;;)
	{
		CodeDocument::Iterator* last = cachedIterators.getLast();

		if (last->getLine() >= maxLineNum)
			break;

		CodeDocument::Iterator* t = new CodeDocument::Iterator (*last);
		cachedIterators.add (t);
		const int targetLine = last->getLine() + linesBetweenCachedSources;

		for (;;)
		{
			codeTokeniser->readNextToken (*t);

			if (t->getLine() >= targetLine)
				break;

			if (t->isEOF())
				return;
		}
	}
}

void CodeEditorComponent::getIteratorForPosition (int position, CodeDocument::Iterator& source)
{
	if (codeTokeniser == nullptr)
		return;

	for (int i = cachedIterators.size(); --i >= 0;)
	{
		CodeDocument::Iterator* t = cachedIterators.getUnchecked (i);
		if (t->getPosition() <= position)
		{
			source = *t;
			break;
		}
	}

	while (source.getPosition() < position)
	{
		const CodeDocument::Iterator original (source);
		codeTokeniser->readNextToken (source);

		if (source.getPosition() > position || source.isEOF())
		{
			source = original;
			break;
		}
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_CodeEditorComponent.cpp ***/


/*** Start of inlined file: juce_CPlusPlusCodeTokeniser.cpp ***/
BEGIN_JUCE_NAMESPACE

CPlusPlusCodeTokeniser::CPlusPlusCodeTokeniser()
{
}

CPlusPlusCodeTokeniser::~CPlusPlusCodeTokeniser()
{
}

namespace CppTokeniser
{
	bool isIdentifierStart (const juce_wchar c) noexcept
	{
		return CharacterFunctions::isLetter (c)
				|| c == '_' || c == '@';
	}

	bool isIdentifierBody (const juce_wchar c) noexcept
	{
		return CharacterFunctions::isLetterOrDigit (c)
				|| c == '_' || c == '@';
	}

	bool isReservedKeyword (String::CharPointerType token, const int tokenLength) noexcept
	{
		static const char* const keywords2Char[] =
			{ "if", "do", "or", "id", 0 };

		static const char* const keywords3Char[] =
			{ "for", "int", "new", "try", "xor", "and", "asm", "not", 0 };

		static const char* const keywords4Char[] =
			{ "bool", "void", "this", "true", "long", "else", "char",
			  "enum", "case", "goto", "auto", 0 };

		static const char* const keywords5Char[] =
			{  "while", "bitor", "break", "catch", "class", "compl", "const", "false",
				"float", "short", "throw", "union", "using", "or_eq", 0 };

		static const char* const keywords6Char[] =
			{ "return", "struct", "and_eq", "bitand", "delete", "double", "extern",
			  "friend", "inline", "not_eq", "public", "sizeof", "static", "signed",
			  "switch", "typeid", "wchar_t", "xor_eq", 0};

		static const char* const keywords7Char[] =
			{ "default", "mutable", "private", "typedef", "nullptr", "virtual", 0 };

		static const char* const keywordsOther[] =
			{ "noexcept", "const_cast", "continue", "explicit", "namespace",
			  "operator", "protected", "register", "reinterpret_cast", "static_cast",
			  "template", "typename", "unsigned", "volatile", "constexpr",
			  "@implementation", "@interface", "@end", "@synthesize", "@dynamic", "@public",
			  "@private", "@property", "@protected", "@class", 0 };

		const char* const* k;

		switch (tokenLength)
		{
			case 2:	 k = keywords2Char; break;
			case 3:	 k = keywords3Char; break;
			case 4:	 k = keywords4Char; break;
			case 5:	 k = keywords5Char; break;
			case 6:	 k = keywords6Char; break;
			case 7:	 k = keywords7Char; break;

			default:
				if (tokenLength < 2 || tokenLength > 16)
					return false;

				k = keywordsOther;
				break;
		}

		int i = 0;
		while (k[i] != 0)
		{
			if (token.compare (CharPointer_ASCII (k[i])) == 0)
				return true;

			++i;
		}

		return false;
	}

	int parseIdentifier (CodeDocument::Iterator& source) noexcept
	{
		int tokenLength = 0;
		String::CharPointerType::CharType possibleIdentifier [100];
		String::CharPointerType possible (possibleIdentifier);

		while (isIdentifierBody (source.peekNextChar()))
		{
			const juce_wchar c = source.nextChar();

			if (tokenLength < 20)
				possible.write (c);

			++tokenLength;
		}

		if (tokenLength > 1 && tokenLength <= 16)
		{
			possible.writeNull();

			if (isReservedKeyword (String::CharPointerType (possibleIdentifier), tokenLength))
				return CPlusPlusCodeTokeniser::tokenType_builtInKeyword;
		}

		return CPlusPlusCodeTokeniser::tokenType_identifier;
	}

	bool skipNumberSuffix (CodeDocument::Iterator& source)
	{
		const juce_wchar c = source.peekNextChar();
		if (c == 'l' || c == 'L' || c == 'u' || c == 'U')
			source.skip();

		if (CharacterFunctions::isLetterOrDigit (source.peekNextChar()))
			return false;

		return true;
	}

	bool isHexDigit (const juce_wchar c) noexcept
	{
		return (c >= '0' && c <= '9')
				|| (c >= 'a' && c <= 'f')
				|| (c >= 'A' && c <= 'F');
	}

	bool parseHexLiteral (CodeDocument::Iterator& source) noexcept
	{
		if (source.nextChar() != '0')
			return false;

		juce_wchar c = source.nextChar();
		if (c != 'x' && c != 'X')
			return false;

		int numDigits = 0;
		while (isHexDigit (source.peekNextChar()))
		{
			++numDigits;
			source.skip();
		}

		if (numDigits == 0)
			return false;

		return skipNumberSuffix (source);
	}

	bool isOctalDigit (const juce_wchar c) noexcept
	{
		return c >= '0' && c <= '7';
	}

	bool parseOctalLiteral (CodeDocument::Iterator& source) noexcept
	{
		if (source.nextChar() != '0')
			return false;

		if (! isOctalDigit (source.nextChar()))
			 return false;

		while (isOctalDigit (source.peekNextChar()))
			source.skip();

		return skipNumberSuffix (source);
	}

	bool isDecimalDigit (const juce_wchar c) noexcept
	{
		return c >= '0' && c <= '9';
	}

	bool parseDecimalLiteral (CodeDocument::Iterator& source) noexcept
	{
		int numChars = 0;
		while (isDecimalDigit (source.peekNextChar()))
		{
			++numChars;
			source.skip();
		}

		if (numChars == 0)
			return false;

		return skipNumberSuffix (source);
	}

	bool parseFloatLiteral (CodeDocument::Iterator& source) noexcept
	{
		int numDigits = 0;

		while (isDecimalDigit (source.peekNextChar()))
		{
			source.skip();
			++numDigits;
		}

		const bool hasPoint = (source.peekNextChar() == '.');

		if (hasPoint)
		{
			source.skip();

			while (isDecimalDigit (source.peekNextChar()))
			{
				source.skip();
				++numDigits;
			}
		}

		if (numDigits == 0)
			return false;

		juce_wchar c = source.peekNextChar();
		const bool hasExponent = (c == 'e' || c == 'E');

		if (hasExponent)
		{
			source.skip();

			c = source.peekNextChar();
			if (c == '+' || c == '-')
				source.skip();

			int numExpDigits = 0;
			while (isDecimalDigit (source.peekNextChar()))
			{
				source.skip();
				++numExpDigits;
			}

			if (numExpDigits == 0)
				return false;
		}

		c = source.peekNextChar();
		if (c == 'f' || c == 'F')
			source.skip();
		else if (! (hasExponent || hasPoint))
			return false;

		return true;
	}

	int parseNumber (CodeDocument::Iterator& source)
	{
		const CodeDocument::Iterator original (source);

		if (parseFloatLiteral (source))
			return CPlusPlusCodeTokeniser::tokenType_floatLiteral;

		source = original;

		if (parseHexLiteral (source))
			return CPlusPlusCodeTokeniser::tokenType_integerLiteral;

		source = original;

		if (parseOctalLiteral (source))
			return CPlusPlusCodeTokeniser::tokenType_integerLiteral;

		source = original;

		if (parseDecimalLiteral (source))
			return CPlusPlusCodeTokeniser::tokenType_integerLiteral;

		source = original;
		source.skip();

		return CPlusPlusCodeTokeniser::tokenType_error;
	}

	void skipQuotedString (CodeDocument::Iterator& source) noexcept
	{
		const juce_wchar quote = source.nextChar();

		for (;;)
		{
			const juce_wchar c = source.nextChar();

			if (c == quote || c == 0)
				break;

			if (c == '\\')
				source.skip();
		}
	}

	void skipComment (CodeDocument::Iterator& source) noexcept
	{
		bool lastWasStar = false;

		for (;;)
		{
			const juce_wchar c = source.nextChar();

			if (c == 0 || (c == '/' && lastWasStar))
				break;

			lastWasStar = (c == '*');
		}
	}
}

int CPlusPlusCodeTokeniser::readNextToken (CodeDocument::Iterator& source)
{
	int result = tokenType_error;
	source.skipWhitespace();

	juce_wchar firstChar = source.peekNextChar();

	switch (firstChar)
	{
	case 0:
		source.skip();
		break;

	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
		result = CppTokeniser::parseNumber (source);
		break;

	case '.':
		result = CppTokeniser::parseNumber (source);

		if (result == tokenType_error)
			result = tokenType_punctuation;

		break;

	case ',':
	case ';':
	case ':':
		source.skip();
		result = tokenType_punctuation;
		break;

	case '(':
	case ')':
	case '{':
	case '}':
	case '[':
	case ']':
		source.skip();
		result = tokenType_bracket;
		break;

	case '"':
	case '\'':
		CppTokeniser::skipQuotedString (source);
		result = tokenType_stringLiteral;
		break;

	case '+':
		result = tokenType_operator;
		source.skip();

		if (source.peekNextChar() == '+')
			source.skip();
		else if (source.peekNextChar() == '=')
			source.skip();

		break;

	case '-':
		source.skip();
		result = CppTokeniser::parseNumber (source);

		if (result == tokenType_error)
		{
			result = tokenType_operator;

			if (source.peekNextChar() == '-')
				source.skip();
			else if (source.peekNextChar() == '=')
				source.skip();
		}
		break;

	case '*':
	case '%':
	case '=':
	case '!':
		result = tokenType_operator;
		source.skip();

		if (source.peekNextChar() == '=')
			source.skip();

		break;

	case '/':
		result = tokenType_operator;
		source.skip();

		if (source.peekNextChar() == '=')
		{
			source.skip();
		}
		else if (source.peekNextChar() == '/')
		{
			result = tokenType_comment;
			source.skipToEndOfLine();
		}
		else if (source.peekNextChar() == '*')
		{
			source.skip();
			result = tokenType_comment;
			CppTokeniser::skipComment (source);
		}

		break;

	case '?':
	case '~':
		source.skip();
		result = tokenType_operator;
		break;

	case '<':
		source.skip();
		result = tokenType_operator;

		if (source.peekNextChar() == '=')
		{
			source.skip();
		}
		else if (source.peekNextChar() == '<')
		{
			source.skip();

			if (source.peekNextChar() == '=')
				source.skip();
		}

		break;

	case '>':
		source.skip();
		result = tokenType_operator;

		if (source.peekNextChar() == '=')
		{
			source.skip();
		}
		else if (source.peekNextChar() == '<')
		{
			source.skip();

			if (source.peekNextChar() == '=')
				source.skip();
		}

		break;

	case '|':
		source.skip();
		result = tokenType_operator;

		if (source.peekNextChar() == '=')
		{
			source.skip();
		}
		else if (source.peekNextChar() == '|')
		{
			source.skip();

			if (source.peekNextChar() == '=')
				source.skip();
		}

		break;

	case '&':
		source.skip();
		result = tokenType_operator;

		if (source.peekNextChar() == '=')
		{
			source.skip();
		}
		else if (source.peekNextChar() == '&')
		{
			source.skip();

			if (source.peekNextChar() == '=')
				source.skip();
		}

		break;

	case '^':
		source.skip();
		result = tokenType_operator;

		if (source.peekNextChar() == '=')
		{
			source.skip();
		}
		else if (source.peekNextChar() == '^')
		{
			source.skip();

			if (source.peekNextChar() == '=')
				source.skip();
		}

		break;

	case '#':
		result = tokenType_preprocessor;
		source.skipToEndOfLine();
		break;

	default:
		if (CppTokeniser::isIdentifierStart (firstChar))
			result = CppTokeniser::parseIdentifier (source);
		else
			source.skip();

		break;
	}

	return result;
}

StringArray CPlusPlusCodeTokeniser::getTokenTypes()
{
	const char* const types[] =
	{
		"Error",
		"Comment",
		"C++ keyword",
		"Identifier",
		"Integer literal",
		"Float literal",
		"String literal",
		"Operator",
		"Bracket",
		"Punctuation",
		"Preprocessor line",
		0
	};

	return StringArray (types);
}

const Colour CPlusPlusCodeTokeniser::getDefaultColour (const int tokenType)
{
	const uint32 colours[] =
	{
		0xffcc0000,  // error
		0xff00aa00,  // comment
		0xff0000cc,  // keyword
		0xff000000,  // identifier
		0xff880000,  // int literal
		0xff885500,  // float literal
		0xff990099,  // string literal
		0xff225500,  // operator
		0xff000055,  // bracket
		0xff004400,  // punctuation
		0xff660000   // preprocessor
	};

	if (tokenType >= 0 && tokenType < numElementsInArray (colours))
		return Colour (colours [tokenType]);

	return Colours::black;
}

bool CPlusPlusCodeTokeniser::isReservedKeyword (const String& token) noexcept
{
	return CppTokeniser::isReservedKeyword (token.getCharPointer(), token.length());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_CPlusPlusCodeTokeniser.cpp ***/


/*** Start of inlined file: juce_ComboBox.cpp ***/
BEGIN_JUCE_NAMESPACE

ComboBox::ItemInfo::ItemInfo (const String& name_, int itemId_, bool isEnabled_, bool isHeading_)
	: name (name_), itemId (itemId_), isEnabled (isEnabled_), isHeading (isHeading_)
{
}

bool ComboBox::ItemInfo::isSeparator() const noexcept
{
	return name.isEmpty();
}

bool ComboBox::ItemInfo::isRealItem() const noexcept
{
	return ! (isHeading || name.isEmpty());
}

ComboBox::ComboBox (const String& name)
	: Component (name),
	  lastCurrentId (0),
	  isButtonDown (false),
	  separatorPending (false),
	  menuActive (false),
	  noChoicesMessage (TRANS("(no choices)"))
{
	setRepaintsOnMouseActivity (true);
	ComboBox::lookAndFeelChanged();
	currentId.addListener (this);
}

ComboBox::~ComboBox()
{
	currentId.removeListener (this);

	if (menuActive)
		PopupMenu::dismissAllActiveMenus();

	label = nullptr;
}

void ComboBox::setEditableText (const bool isEditable)
{
	if (label->isEditableOnSingleClick() != isEditable || label->isEditableOnDoubleClick() != isEditable)
	{
		label->setEditable (isEditable, isEditable, false);
		setWantsKeyboardFocus (! isEditable);
		resized();
	}
}

bool ComboBox::isTextEditable() const noexcept
{
	return label->isEditable();
}

void ComboBox::setJustificationType (const Justification& justification)
{
	label->setJustificationType (justification);
}

const Justification ComboBox::getJustificationType() const noexcept
{
	return label->getJustificationType();
}

void ComboBox::setTooltip (const String& newTooltip)
{
	SettableTooltipClient::setTooltip (newTooltip);
	label->setTooltip (newTooltip);
}

void ComboBox::addItem (const String& newItemText, const int newItemId)
{
	// you can't add empty strings to the list..
	jassert (newItemText.isNotEmpty());

	// IDs must be non-zero, as zero is used to indicate a lack of selecion.
	jassert (newItemId != 0);

	// you shouldn't use duplicate item IDs!
	jassert (getItemForId (newItemId) == nullptr);

	if (newItemText.isNotEmpty() && newItemId != 0)
	{
		if (separatorPending)
		{
			separatorPending = false;
			items.add (new ItemInfo (String::empty, 0, false, false));
		}

		items.add (new ItemInfo (newItemText, newItemId, true, false));
	}
}

void ComboBox::addSeparator()
{
	separatorPending = (items.size() > 0);
}

void ComboBox::addSectionHeading (const String& headingName)
{
	// you can't add empty strings to the list..
	jassert (headingName.isNotEmpty());

	if (headingName.isNotEmpty())
	{
		if (separatorPending)
		{
			separatorPending = false;
			items.add (new ItemInfo (String::empty, 0, false, false));
		}

		items.add (new ItemInfo (headingName, 0, true, true));
	}
}

void ComboBox::setItemEnabled (const int itemId, const bool shouldBeEnabled)
{
	ItemInfo* const item = getItemForId (itemId);

	if (item != nullptr)
		item->isEnabled = shouldBeEnabled;
}

bool ComboBox::isItemEnabled (int itemId) const noexcept
{
	const ItemInfo* const item = getItemForId (itemId);
	return item != nullptr && item->isEnabled;
}

void ComboBox::changeItemText (const int itemId, const String& newText)
{
	ItemInfo* const item = getItemForId (itemId);

	jassert (item != nullptr);

	if (item != nullptr)
		item->name = newText;
}

void ComboBox::clear (const bool dontSendChangeMessage)
{
	items.clear();
	separatorPending = false;

	if (! label->isEditable())
		setSelectedItemIndex (-1, dontSendChangeMessage);
}

ComboBox::ItemInfo* ComboBox::getItemForId (const int itemId) const noexcept
{
	if (itemId != 0)
	{
		for (int i = items.size(); --i >= 0;)
			if (items.getUnchecked(i)->itemId == itemId)
				return items.getUnchecked(i);
	}

	return nullptr;
}

ComboBox::ItemInfo* ComboBox::getItemForIndex (const int index) const noexcept
{
	for (int n = 0, i = 0; i < items.size(); ++i)
	{
		ItemInfo* const item = items.getUnchecked(i);

		if (item->isRealItem())
			if (n++ == index)
				return item;
	}

	return nullptr;
}

int ComboBox::getNumItems() const noexcept
{
	int n = 0;

	for (int i = items.size(); --i >= 0;)
		if (items.getUnchecked(i)->isRealItem())
			++n;

	return n;
}

String ComboBox::getItemText (const int index) const
{
	const ItemInfo* const item = getItemForIndex (index);

	return item != nullptr ? item->name : String::empty;
}

int ComboBox::getItemId (const int index) const noexcept
{
	const ItemInfo* const item = getItemForIndex (index);

	return item != nullptr ? item->itemId : 0;
}

int ComboBox::indexOfItemId (const int itemId) const noexcept
{
	for (int n = 0, i = 0; i < items.size(); ++i)
	{
		const ItemInfo* const item = items.getUnchecked(i);

		if (item->isRealItem())
		{
			if (item->itemId == itemId)
				return n;

			++n;
		}
	}

	return -1;
}

int ComboBox::getSelectedItemIndex() const
{
	int index = indexOfItemId (currentId.getValue());

	if (getText() != getItemText (index))
		index = -1;

	return index;
}

void ComboBox::setSelectedItemIndex (const int index, const bool dontSendChangeMessage)
{
	setSelectedId (getItemId (index), dontSendChangeMessage);
}

int ComboBox::getSelectedId() const noexcept
{
	const ItemInfo* const item = getItemForId (currentId.getValue());

	return (item != nullptr && getText() == item->name) ? item->itemId : 0;
}

void ComboBox::setSelectedId (const int newItemId, const bool dontSendChangeMessage)
{
	const ItemInfo* const item = getItemForId (newItemId);
	const String newItemText (item != nullptr ? item->name : String::empty);

	if (lastCurrentId != newItemId || label->getText() != newItemText)
	{
		if (! dontSendChangeMessage)
			triggerAsyncUpdate();

		label->setText (newItemText, false);
		lastCurrentId = newItemId;
		currentId = newItemId;

		repaint();  // for the benefit of the 'none selected' text
	}
}

bool ComboBox::selectIfEnabled (const int index)
{
	const ItemInfo* const item = getItemForIndex (index);

	if (item != nullptr && item->isEnabled)
	{
		setSelectedItemIndex (index);
		return true;
	}

	return false;
}

void ComboBox::valueChanged (Value&)
{
	if (lastCurrentId != (int) currentId.getValue())
		setSelectedId (currentId.getValue(), false);
}

String ComboBox::getText() const
{
	return label->getText();
}

void ComboBox::setText (const String& newText, const bool dontSendChangeMessage)
{
	for (int i = items.size(); --i >= 0;)
	{
		const ItemInfo* const item = items.getUnchecked(i);

		if (item->isRealItem()
			 && item->name == newText)
		{
			setSelectedId (item->itemId, dontSendChangeMessage);
			return;
		}
	}

	lastCurrentId = 0;
	currentId = 0;

	if (label->getText() != newText)
	{
		label->setText (newText, false);

		if (! dontSendChangeMessage)
			triggerAsyncUpdate();
	}

	repaint();
}

void ComboBox::showEditor()
{
	jassert (isTextEditable()); // you probably shouldn't do this to a non-editable combo box?

	label->showEditor();
}

void ComboBox::setTextWhenNothingSelected (const String& newMessage)
{
	if (textWhenNothingSelected != newMessage)
	{
		textWhenNothingSelected = newMessage;
		repaint();
	}
}

String ComboBox::getTextWhenNothingSelected() const
{
	return textWhenNothingSelected;
}

void ComboBox::setTextWhenNoChoicesAvailable (const String& newMessage)
{
	noChoicesMessage = newMessage;
}

String ComboBox::getTextWhenNoChoicesAvailable() const
{
	return noChoicesMessage;
}

void ComboBox::paint (Graphics& g)
{
	getLookAndFeel().drawComboBox (g, getWidth(), getHeight(), isButtonDown,
								   label->getRight(), 0, getWidth() - label->getRight(), getHeight(),
								   *this);

	if (textWhenNothingSelected.isNotEmpty()
		 && label->getText().isEmpty()
		 && ! label->isBeingEdited())
	{
		g.setColour (findColour (textColourId).withMultipliedAlpha (0.5f));
		g.setFont (label->getFont());
		g.drawFittedText (textWhenNothingSelected,
						  label->getX() + 2, label->getY() + 1,
						  label->getWidth() - 4, label->getHeight() - 2,
						  label->getJustificationType(),
						  jmax (1, (int) (label->getHeight() / label->getFont().getHeight())));
	}
}

void ComboBox::resized()
{
	if (getHeight() > 0 && getWidth() > 0)
		getLookAndFeel().positionComboBoxText (*this, *label);
}

void ComboBox::enablementChanged()
{
	repaint();
}

void ComboBox::lookAndFeelChanged()
{
	repaint();

	{
		ScopedPointer <Label> newLabel (getLookAndFeel().createComboBoxTextBox (*this));
		jassert (newLabel != nullptr);

		if (label != nullptr)
		{
			newLabel->setEditable (label->isEditable());
			newLabel->setJustificationType (label->getJustificationType());
			newLabel->setTooltip (label->getTooltip());
			newLabel->setText (label->getText(), false);
		}

		label = newLabel;
	}

	addAndMakeVisible (label);
	setWantsKeyboardFocus (! label->isEditable());

	label->addListener (this);
	label->addMouseListener (this, false);

	label->setColour (Label::backgroundColourId, Colours::transparentBlack);
	label->setColour (Label::textColourId, findColour (ComboBox::textColourId));

	label->setColour (TextEditor::textColourId, findColour (ComboBox::textColourId));
	label->setColour (TextEditor::backgroundColourId, Colours::transparentBlack);
	label->setColour (TextEditor::highlightColourId, findColour (TextEditor::highlightColourId));
	label->setColour (TextEditor::outlineColourId, Colours::transparentBlack);

	resized();
}

void ComboBox::colourChanged()
{
	lookAndFeelChanged();
}

bool ComboBox::keyPressed (const KeyPress& key)
{
	if (key.isKeyCode (KeyPress::upKey) || key.isKeyCode (KeyPress::leftKey))
	{
		int index = getSelectedItemIndex() - 1;

		while (index >= 0 && ! selectIfEnabled (index))
			--index;

		return true;
	}
	else if (key.isKeyCode (KeyPress::downKey) || key.isKeyCode (KeyPress::rightKey))
	{
		int index = getSelectedItemIndex() + 1;

		while (index < getNumItems() && ! selectIfEnabled (index))
			++index;

		return true;
	}
	else if (key.isKeyCode (KeyPress::returnKey))
	{
		showPopup();
		return true;
	}

	return false;
}

bool ComboBox::keyStateChanged (const bool isKeyDown)
{
	// only forward key events that aren't used by this component
	return isKeyDown
			&& (KeyPress::isKeyCurrentlyDown (KeyPress::upKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::leftKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::downKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::rightKey));
}

void ComboBox::focusGained (FocusChangeType)	{ repaint(); }
void ComboBox::focusLost (FocusChangeType)	  { repaint(); }

void ComboBox::labelTextChanged (Label*)
{
	triggerAsyncUpdate();
}

void ComboBox::popupMenuFinishedCallback (int result, ComboBox* box)
{
	if (box != nullptr)
	{
		box->menuActive = false;

		if (result != 0)
			box->setSelectedId (result);
	}
}

void ComboBox::showPopup()
{
	if (! menuActive)
	{
		const int selectedId = getSelectedId();

		PopupMenu menu;
		menu.setLookAndFeel (&getLookAndFeel());

		for (int i = 0; i < items.size(); ++i)
		{
			const ItemInfo* const item = items.getUnchecked(i);

			if (item->isSeparator())
				menu.addSeparator();
			else if (item->isHeading)
				menu.addSectionHeader (item->name);
			else
				menu.addItem (item->itemId, item->name,
							  item->isEnabled, item->itemId == selectedId);
		}

		if (items.size() == 0)
			menu.addItem (1, noChoicesMessage, false);

		menuActive = true;

		menu.showMenuAsync (PopupMenu::Options().withTargetComponent (this)
												.withItemThatMustBeVisible (selectedId)
												.withMinimumWidth (getWidth())
												.withMaximumNumColumns (1)
												.withStandardItemHeight (jlimit (12, 24, getHeight())),
							ModalCallbackFunction::forComponent (popupMenuFinishedCallback, this));
	}
}

void ComboBox::mouseDown (const MouseEvent& e)
{
	beginDragAutoRepeat (300);

	isButtonDown = isEnabled() && ! e.mods.isPopupMenu();

	if (isButtonDown && (e.eventComponent == this || ! label->isEditable()))
		showPopup();
}

void ComboBox::mouseDrag (const MouseEvent& e)
{
	beginDragAutoRepeat (50);

	if (isButtonDown && ! e.mouseWasClicked())
		showPopup();
}

void ComboBox::mouseUp (const MouseEvent& e2)
{
	if (isButtonDown)
	{
		isButtonDown = false;
		repaint();

		const MouseEvent e (e2.getEventRelativeTo (this));

		if (reallyContains (e.getPosition(), true)
			 && (e2.eventComponent == this || ! label->isEditable()))
		{
			showPopup();
		}
	}
}

void ComboBox::addListener (ComboBoxListener* const listener)
{
	listeners.add (listener);
}

void ComboBox::removeListener (ComboBoxListener* const listener)
{
	listeners.remove (listener);
}

void ComboBox::handleAsyncUpdate()
{
	Component::BailOutChecker checker (this);
	listeners.callChecked (checker, &ComboBoxListener::comboBoxChanged, this);  // (can't use ComboBox::Listener due to idiotic VC2005 bug)
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComboBox.cpp ***/


/*** Start of inlined file: juce_ImageComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

ImageComponent::ImageComponent (const String& name)
	: Component (name),
	  placement (RectanglePlacement::centred)
{
}

ImageComponent::~ImageComponent()
{
}

void ImageComponent::setImage (const Image& newImage)
{
	if (image != newImage)
	{
		image = newImage;
		repaint();
	}
}

void ImageComponent::setImage (const Image& newImage, const RectanglePlacement& placementToUse)
{
	if (image != newImage || placement != placementToUse)
	{
		image = newImage;
		placement = placementToUse;
		repaint();
	}
}

void ImageComponent::setImagePlacement (const RectanglePlacement& newPlacement)
{
	if (placement != newPlacement)
	{
		placement = newPlacement;
		repaint();
	}
}

const Image& ImageComponent::getImage() const
{
	return image;
}

const RectanglePlacement ImageComponent::getImagePlacement() const
{
	return placement;
}

void ImageComponent::paint (Graphics& g)
{
	g.setOpacity (1.0f);
	g.drawImageWithin (image, 0, 0, getWidth(), getHeight(), placement, false);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ImageComponent.cpp ***/


/*** Start of inlined file: juce_Label.cpp ***/
BEGIN_JUCE_NAMESPACE

Label::Label (const String& name,
			  const String& labelText)
	: Component (name),
	  textValue (labelText),
	  lastTextValue (labelText),
	  font (15.0f),
	  justification (Justification::centredLeft),
	  horizontalBorderSize (5),
	  verticalBorderSize (1),
	  minimumHorizontalScale (0.7f),
	  editSingleClick (false),
	  editDoubleClick (false),
	  lossOfFocusDiscardsChanges (false)
{
	setColour (TextEditor::textColourId, Colours::black);
	setColour (TextEditor::backgroundColourId, Colours::transparentBlack);
	setColour (TextEditor::outlineColourId, Colours::transparentBlack);

	textValue.addListener (this);
}

Label::~Label()
{
	textValue.removeListener (this);

	if (ownerComponent != nullptr)
		ownerComponent->removeComponentListener (this);

	editor = nullptr;
}

void Label::setText (const String& newText,
					 const bool broadcastChangeMessage)
{
	hideEditor (true);

	if (lastTextValue != newText)
	{
		lastTextValue = newText;
		textValue = newText;
		repaint();

		textWasChanged();

		if (ownerComponent != nullptr)
			componentMovedOrResized (*ownerComponent, true, true);

		if (broadcastChangeMessage)
			callChangeListeners();
	}
}

String Label::getText (const bool returnActiveEditorContents) const
{
	return (returnActiveEditorContents && isBeingEdited())
				? editor->getText()
				: textValue.toString();
}

void Label::valueChanged (Value&)
{
	if (lastTextValue != textValue.toString())
		setText (textValue.toString(), true);
}

void Label::setFont (const Font& newFont)
{
	if (font != newFont)
	{
		font = newFont;
		repaint();
	}
}

const Font& Label::getFont() const noexcept
{
	return font;
}

void Label::setEditable (const bool editOnSingleClick,
						 const bool editOnDoubleClick,
						 const bool lossOfFocusDiscardsChanges_)
{
	editSingleClick = editOnSingleClick;
	editDoubleClick = editOnDoubleClick;
	lossOfFocusDiscardsChanges = lossOfFocusDiscardsChanges_;

	setWantsKeyboardFocus (editOnSingleClick || editOnDoubleClick);
	setFocusContainer (editOnSingleClick || editOnDoubleClick);
}

void Label::setJustificationType (const Justification& newJustification)
{
	if (justification != newJustification)
	{
		justification = newJustification;
		repaint();
	}
}

void Label::setBorderSize (int h, int v)
{
	if (horizontalBorderSize != h || verticalBorderSize != v)
	{
		horizontalBorderSize = h;
		verticalBorderSize = v;
		repaint();
	}
}

Component* Label::getAttachedComponent() const
{
	return static_cast<Component*> (ownerComponent);
}

void Label::attachToComponent (Component* owner,
							   const bool onLeft)
{
	if (ownerComponent != nullptr)
		ownerComponent->removeComponentListener (this);

	ownerComponent = owner;

	leftOfOwnerComp = onLeft;

	if (ownerComponent != nullptr)
	{
		setVisible (owner->isVisible());
		ownerComponent->addComponentListener (this);
		componentParentHierarchyChanged (*ownerComponent);
		componentMovedOrResized (*ownerComponent, true, true);
	}
}

void Label::componentMovedOrResized (Component& component, bool /*wasMoved*/, bool /*wasResized*/)
{
	if (leftOfOwnerComp)
	{
		setSize (jmin (getFont().getStringWidth (textValue.toString()) + 8, component.getX()),
				 component.getHeight());

		setTopRightPosition (component.getX(), component.getY());
	}
	else
	{
		setSize (component.getWidth(),
				 8 + roundToInt (getFont().getHeight()));

		setTopLeftPosition (component.getX(), component.getY() - getHeight());
	}
}

void Label::componentParentHierarchyChanged (Component& component)
{
	if (component.getParentComponent() != nullptr)
		component.getParentComponent()->addChildComponent (this);
}

void Label::componentVisibilityChanged (Component& component)
{
	setVisible (component.isVisible());
}

void Label::textWasEdited()
{
}

void Label::textWasChanged()
{
}

void Label::showEditor()
{
	if (editor == nullptr)
	{
		addAndMakeVisible (editor = createEditorComponent());
		editor->setText (getText(), false);
		editor->addListener (this);
		editor->grabKeyboardFocus();
		editor->setHighlightedRegion (Range<int> (0, textValue.toString().length()));
		editor->addListener (this);

		resized();
		repaint();

		editorShown (editor);

		enterModalState (false);
		editor->grabKeyboardFocus();
	}
}

void Label::editorShown (TextEditor*)
{
}

void Label::editorAboutToBeHidden (TextEditor*)
{
}

bool Label::updateFromTextEditorContents (TextEditor& ed)
{
	const String newText (ed.getText());

	if (textValue.toString() != newText)
	{
		lastTextValue = newText;
		textValue = newText;
		repaint();

		textWasChanged();

		if (ownerComponent != nullptr)
			componentMovedOrResized (*ownerComponent, true, true);

		return true;
	}

	return false;
}

void Label::hideEditor (const bool discardCurrentEditorContents)
{
	if (editor != nullptr)
	{
		WeakReference<Component> deletionChecker (this);

		ScopedPointer<TextEditor> outgoingEditor (editor);

		editorAboutToBeHidden (outgoingEditor);

		const bool changed = (! discardCurrentEditorContents)
							   && updateFromTextEditorContents (*outgoingEditor);
		outgoingEditor = nullptr;
		repaint();

		if (changed)
			textWasEdited();

		if (deletionChecker != nullptr)
			exitModalState (0);

		if (changed && deletionChecker != nullptr)
			callChangeListeners();
	}
}

void Label::inputAttemptWhenModal()
{
	if (editor != nullptr)
	{
		if (lossOfFocusDiscardsChanges)
			textEditorEscapeKeyPressed (*editor);
		else
			textEditorReturnKeyPressed (*editor);
	}
}

bool Label::isBeingEdited() const noexcept
{
	return editor != nullptr;
}

TextEditor* Label::createEditorComponent()
{
	TextEditor* const ed = new TextEditor (getName());
	ed->setFont (font);

	// copy these colours from our own settings..
	const int cols[] = { TextEditor::backgroundColourId,
						 TextEditor::textColourId,
						 TextEditor::highlightColourId,
						 TextEditor::highlightedTextColourId,
						 TextEditor::outlineColourId,
						 TextEditor::focusedOutlineColourId,
						 TextEditor::shadowColourId,
						 CaretComponent::caretColourId };

	for (int i = 0; i < numElementsInArray (cols); ++i)
		ed->setColour (cols[i], findColour (cols[i]));

	return ed;
}

void Label::paint (Graphics& g)
{
	getLookAndFeel().drawLabel (g, *this);
}

void Label::mouseUp (const MouseEvent& e)
{
	if (editSingleClick
		 && e.mouseWasClicked()
		 && contains (e.getPosition())
		 && ! e.mods.isPopupMenu())
	{
		showEditor();
	}
}

void Label::mouseDoubleClick (const MouseEvent& e)
{
	if (editDoubleClick && ! e.mods.isPopupMenu())
		showEditor();
}

void Label::resized()
{
	if (editor != nullptr)
		editor->setBoundsInset (BorderSize<int> (0));
}

void Label::focusGained (FocusChangeType cause)
{
	if (editSingleClick && cause == focusChangedByTabKey)
		showEditor();
}

void Label::enablementChanged()
{
	repaint();
}

void Label::colourChanged()
{
	repaint();
}

void Label::setMinimumHorizontalScale (const float newScale)
{
	if (minimumHorizontalScale != newScale)
	{
		minimumHorizontalScale = newScale;
		repaint();
	}
}

// We'll use a custom focus traverser here to make sure focus goes from the
// text editor to another component rather than back to the label itself.
class LabelKeyboardFocusTraverser   : public KeyboardFocusTraverser
{
public:
	LabelKeyboardFocusTraverser() {}

	Component* getNextComponent (Component* current)
	{
		return KeyboardFocusTraverser::getNextComponent (dynamic_cast <TextEditor*> (current) != nullptr
															? current->getParentComponent() : current);
	}

	Component* getPreviousComponent (Component* current)
	{
		return KeyboardFocusTraverser::getPreviousComponent (dynamic_cast <TextEditor*> (current) != nullptr
																? current->getParentComponent() : current);
	}
};

KeyboardFocusTraverser* Label::createFocusTraverser()
{
	return new LabelKeyboardFocusTraverser();
}

void Label::addListener (LabelListener* const listener)
{
	listeners.add (listener);
}

void Label::removeListener (LabelListener* const listener)
{
	listeners.remove (listener);
}

void Label::callChangeListeners()
{
	Component::BailOutChecker checker (this);
	listeners.callChecked (checker, &LabelListener::labelTextChanged, this);  // (can't use Label::Listener due to idiotic VC2005 bug)
}

void Label::textEditorTextChanged (TextEditor& ed)
{
	if (editor != nullptr)
	{
		jassert (&ed == editor);

		if (! (hasKeyboardFocus (true) || isCurrentlyBlockedByAnotherModalComponent()))
		{
			if (lossOfFocusDiscardsChanges)
				textEditorEscapeKeyPressed (ed);
			else
				textEditorReturnKeyPressed (ed);
		}
	}
}

void Label::textEditorReturnKeyPressed (TextEditor& ed)
{
	if (editor != nullptr)
	{
		jassert (&ed == editor);

		const bool changed = updateFromTextEditorContents (ed);
		hideEditor (true);

		if (changed)
		{
			WeakReference<Component> deletionChecker (this);
			textWasEdited();

			if (deletionChecker != nullptr)
				callChangeListeners();
		}
	}
}

void Label::textEditorEscapeKeyPressed (TextEditor& ed)
{
	if (editor != nullptr)
	{
		jassert (&ed == editor);
		(void) ed;

		editor->setText (textValue.toString(), false);
		hideEditor (true);
	}
}

void Label::textEditorFocusLost (TextEditor& ed)
{
	textEditorTextChanged (ed);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Label.cpp ***/


/*** Start of inlined file: juce_ListBox.cpp ***/
BEGIN_JUCE_NAMESPACE

class ListBoxRowComponent  : public Component,
							 public TooltipClient
{
public:
	ListBoxRowComponent (ListBox& owner_)
		: owner (owner_), row (-1),
		  selected (false), isDragging (false), selectRowOnMouseUp (false)
	{
	}

	void paint (Graphics& g)
	{
		if (owner.getModel() != nullptr)
			owner.getModel()->paintListBoxItem (row, g, getWidth(), getHeight(), selected);
	}

	void update (const int row_, const bool selected_)
	{
		if (row != row_ || selected != selected_)
		{
			repaint();
			row = row_;
			selected = selected_;
		}

		if (owner.getModel() != nullptr)
		{
			customComponent = owner.getModel()->refreshComponentForRow (row_, selected_, customComponent.release());

			if (customComponent != nullptr)
			{
				addAndMakeVisible (customComponent);
				customComponent->setBounds (getLocalBounds());
			}
		}
	}

	void mouseDown (const MouseEvent& e)
	{
		isDragging = false;
		selectRowOnMouseUp = false;

		if (isEnabled())
		{
			if (! selected)
			{
				owner.selectRowsBasedOnModifierKeys (row, e.mods, false);

				if (owner.getModel() != nullptr)
					owner.getModel()->listBoxItemClicked (row, e);
			}
			else
			{
				selectRowOnMouseUp = true;
			}
		}
	}

	void mouseUp (const MouseEvent& e)
	{
		if (isEnabled() && selectRowOnMouseUp && ! isDragging)
		{
			owner.selectRowsBasedOnModifierKeys (row, e.mods, true);

			if (owner.getModel() != nullptr)
				owner.getModel()->listBoxItemClicked (row, e);
		}
	}

	void mouseDoubleClick (const MouseEvent& e)
	{
		if (owner.getModel() != nullptr && isEnabled())
			owner.getModel()->listBoxItemDoubleClicked (row, e);
	}

	void mouseDrag (const MouseEvent& e)
	{
		if (isEnabled() && owner.getModel() != nullptr && ! (e.mouseWasClicked() || isDragging))
		{
			const SparseSet<int> selectedRows (owner.getSelectedRows());

			if (selectedRows.size() > 0)
			{
				const var dragDescription (owner.getModel()->getDragSourceDescription (selectedRows));

				if (! (dragDescription.isVoid() || (dragDescription.isString() && dragDescription.toString().isEmpty())))
				{
					isDragging = true;
					owner.startDragAndDrop (e, dragDescription);
				}
			}
		}
	}

	void resized()
	{
		if (customComponent != nullptr)
			customComponent->setBounds (getLocalBounds());
	}

	const String getTooltip()
	{
		if (owner.getModel() != nullptr)
			return owner.getModel()->getTooltipForRow (row);

		return String::empty;
	}

	ScopedPointer<Component> customComponent;

private:
	ListBox& owner;
	int row;
	bool selected, isDragging, selectRowOnMouseUp;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ListBoxRowComponent);
};

class ListViewport  : public Viewport
{
public:

	ListViewport (ListBox& owner_)
		: owner (owner_)
	{
		setWantsKeyboardFocus (false);

		Component* const content = new Component();
		setViewedComponent (content);
		content->addMouseListener (this, false);
		content->setWantsKeyboardFocus (false);
	}

	ListBoxRowComponent* getComponentForRow (const int row) const noexcept
	{
		return rows [row % jmax (1, rows.size())];
	}

	ListBoxRowComponent* getComponentForRowIfOnscreen (const int row) const noexcept
	{
		return (row >= firstIndex && row < firstIndex + rows.size())
				 ? getComponentForRow (row) : nullptr;
	}

	int getRowNumberOfComponent (Component* const rowComponent) const noexcept
	{
		const int index = getIndexOfChildComponent (rowComponent);
		const int num = rows.size();

		for (int i = num; --i >= 0;)
			if (((firstIndex + i) % jmax (1, num)) == index)
				return firstIndex + i;

		return -1;
	}

	void visibleAreaChanged (const Rectangle<int>&)
	{
		updateVisibleArea (true);

		if (owner.getModel() != nullptr)
			owner.getModel()->listWasScrolled();
	}

	void updateVisibleArea (const bool makeSureItUpdatesContent)
	{
		hasUpdated = false;

		const int newX = getViewedComponent()->getX();
		int newY = getViewedComponent()->getY();
		const int newW = jmax (owner.minimumRowWidth, getMaximumVisibleWidth());
		const int newH = owner.totalItems * owner.getRowHeight();

		if (newY + newH < getMaximumVisibleHeight() && newH > getMaximumVisibleHeight())
			newY = getMaximumVisibleHeight() - newH;

		getViewedComponent()->setBounds (newX, newY, newW, newH);

		if (makeSureItUpdatesContent && ! hasUpdated)
			updateContents();
	}

	void updateContents()
	{
		hasUpdated = true;
		const int rowHeight = owner.getRowHeight();

		if (rowHeight > 0)
		{
			const int y = getViewPositionY();
			const int w = getViewedComponent()->getWidth();

			const int numNeeded = 2 + getMaximumVisibleHeight() / rowHeight;
			rows.removeRange (numNeeded, rows.size());

			while (numNeeded > rows.size())
			{
				ListBoxRowComponent* newRow = new ListBoxRowComponent (owner);
				rows.add (newRow);
				getViewedComponent()->addAndMakeVisible (newRow);
			}

			firstIndex = y / rowHeight;
			firstWholeIndex = (y + rowHeight - 1) / rowHeight;
			lastWholeIndex = (y + getMaximumVisibleHeight() - 1) / rowHeight;

			for (int i = 0; i < numNeeded; ++i)
			{
				const int row = i + firstIndex;
				ListBoxRowComponent* const rowComp = getComponentForRow (row);

				if (rowComp != nullptr)
				{
					rowComp->setBounds (0, row * rowHeight, w, rowHeight);
					rowComp->update (row, owner.isRowSelected (row));
				}
			}
		}

		if (owner.headerComponent != nullptr)
			owner.headerComponent->setBounds (owner.outlineThickness + getViewedComponent()->getX(),
											  owner.outlineThickness,
											  jmax (owner.getWidth() - owner.outlineThickness * 2,
													getViewedComponent()->getWidth()),
											  owner.headerComponent->getHeight());
	}

	void selectRow (const int row, const int rowHeight, const bool dontScroll,
					const int lastRowSelected, const int totalItems, const bool isMouseClick)
	{
		hasUpdated = false;

		if (row < firstWholeIndex && ! dontScroll)
		{
			setViewPosition (getViewPositionX(), row * rowHeight);
		}
		else if (row >= lastWholeIndex && ! dontScroll)
		{
			const int rowsOnScreen = lastWholeIndex - firstWholeIndex;

			if (row >= lastRowSelected + rowsOnScreen
				 && rowsOnScreen < totalItems - 1
				 && ! isMouseClick)
			{
				setViewPosition (getViewPositionX(),
								 jlimit (0, jmax (0, totalItems - rowsOnScreen), row) * rowHeight);
			}
			else
			{
				setViewPosition (getViewPositionX(),
								 jmax (0, (row  + 1) * rowHeight - getMaximumVisibleHeight()));
			}
		}

		if (! hasUpdated)
			updateContents();
	}

	void scrollToEnsureRowIsOnscreen (const int row, const int rowHeight)
	{
		if (row < firstWholeIndex)
		{
			setViewPosition (getViewPositionX(), row * rowHeight);
		}
		else if (row >= lastWholeIndex)
		{
			setViewPosition (getViewPositionX(),
							 jmax (0, (row  + 1) * rowHeight - getMaximumVisibleHeight()));
		}
	}

	void paint (Graphics& g)
	{
		if (isOpaque())
			g.fillAll (owner.findColour (ListBox::backgroundColourId));
	}

	bool keyPressed (const KeyPress& key)
	{
		if (key.isKeyCode (KeyPress::upKey)
			|| key.isKeyCode (KeyPress::downKey)
			|| key.isKeyCode (KeyPress::pageUpKey)
			|| key.isKeyCode (KeyPress::pageDownKey)
			|| key.isKeyCode (KeyPress::homeKey)
			|| key.isKeyCode (KeyPress::endKey))
		{
			// we want to avoid these keypresses going to the viewport, and instead allow
			// them to pass up to our listbox..
			return false;
		}

		return Viewport::keyPressed (key);
	}

private:
	ListBox& owner;
	OwnedArray<ListBoxRowComponent> rows;
	int firstIndex, firstWholeIndex, lastWholeIndex;
	bool hasUpdated;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ListViewport);
};

ListBox::ListBox (const String& name, ListBoxModel* const model_)
	: Component (name),
	  model (model_),
	  totalItems (0),
	  rowHeight (22),
	  minimumRowWidth (0),
	  outlineThickness (0),
	  lastRowSelected (-1),
	  mouseMoveSelects (false),
	  multipleSelection (false),
	  hasDoneInitialUpdate (false)
{
	addAndMakeVisible (viewport = new ListViewport (*this));

	ListBox::setWantsKeyboardFocus (true);
	ListBox::colourChanged();
}

ListBox::~ListBox()
{
	headerComponent = nullptr;
	viewport = nullptr;
}

void ListBox::setModel (ListBoxModel* const newModel)
{
	if (model != newModel)
	{
		model = newModel;
		repaint();
		updateContent();
	}
}

void ListBox::setMultipleSelectionEnabled (bool b)
{
	multipleSelection = b;
}

void ListBox::setMouseMoveSelectsRows (bool b)
{
	mouseMoveSelects = b;

	if (b)
		addMouseListener (this, true);
}

void ListBox::paint (Graphics& g)
{
	if (! hasDoneInitialUpdate)
		updateContent();

	g.fillAll (findColour (backgroundColourId));
}

void ListBox::paintOverChildren (Graphics& g)
{
	if (outlineThickness > 0)
	{
		g.setColour (findColour (outlineColourId));
		g.drawRect (0, 0, getWidth(), getHeight(), outlineThickness);
	}
}

void ListBox::resized()
{
	viewport->setBoundsInset (BorderSize<int> (outlineThickness + ((headerComponent != nullptr) ? headerComponent->getHeight() : 0),
											   outlineThickness, outlineThickness, outlineThickness));

	viewport->setSingleStepSizes (20, getRowHeight());

	viewport->updateVisibleArea (false);
}

void ListBox::visibilityChanged()
{
	viewport->updateVisibleArea (true);
}

Viewport* ListBox::getViewport() const noexcept
{
	return viewport;
}

void ListBox::updateContent()
{
	hasDoneInitialUpdate = true;
	totalItems = (model != nullptr) ? model->getNumRows() : 0;

	bool selectionChanged = false;

	if (selected.size() > 0 && selected [selected.size() - 1] >= totalItems)
	{
		selected.removeRange (Range <int> (totalItems, std::numeric_limits<int>::max()));
		lastRowSelected = getSelectedRow (0);
		selectionChanged = true;
	}

	viewport->updateVisibleArea (isVisible());
	viewport->resized();

	if (selectionChanged && model != nullptr)
		model->selectedRowsChanged (lastRowSelected);
}

void ListBox::selectRow (const int row,
						 bool dontScroll,
						 bool deselectOthersFirst)
{
	selectRowInternal (row, dontScroll, deselectOthersFirst, false);
}

void ListBox::selectRowInternal (const int row,
								 bool dontScroll,
								 bool deselectOthersFirst,
								 bool isMouseClick)
{
	if (! multipleSelection)
		deselectOthersFirst = true;

	if ((! isRowSelected (row))
		 || (deselectOthersFirst && getNumSelectedRows() > 1))
	{
		if (isPositiveAndBelow (row, totalItems))
		{
			if (deselectOthersFirst)
				selected.clear();

			selected.addRange (Range<int> (row, row + 1));

			if (getHeight() == 0 || getWidth() == 0)
				dontScroll = true;

			viewport->selectRow (row, getRowHeight(), dontScroll,
								 lastRowSelected, totalItems, isMouseClick);

			lastRowSelected = row;
			model->selectedRowsChanged (row);
		}
		else
		{
			if (deselectOthersFirst)
				deselectAllRows();
		}
	}
}

void ListBox::deselectRow (const int row)
{
	if (selected.contains (row))
	{
		selected.removeRange (Range <int> (row, row + 1));

		if (row == lastRowSelected)
			lastRowSelected = getSelectedRow (0);

		viewport->updateContents();
		model->selectedRowsChanged (lastRowSelected);
	}
}

void ListBox::setSelectedRows (const SparseSet<int>& setOfRowsToBeSelected,
							   const bool sendNotificationEventToModel)
{
	selected = setOfRowsToBeSelected;
	selected.removeRange (Range <int> (totalItems, std::numeric_limits<int>::max()));

	if (! isRowSelected (lastRowSelected))
		lastRowSelected = getSelectedRow (0);

	viewport->updateContents();

	if ((model != nullptr) && sendNotificationEventToModel)
		model->selectedRowsChanged (lastRowSelected);
}

const SparseSet<int> ListBox::getSelectedRows() const
{
	return selected;
}

void ListBox::selectRangeOfRows (int firstRow, int lastRow)
{
	if (multipleSelection && (firstRow != lastRow))
	{
		const int numRows = totalItems - 1;
		firstRow = jlimit (0, jmax (0, numRows), firstRow);
		lastRow = jlimit (0, jmax (0, numRows), lastRow);

		selected.addRange (Range <int> (jmin (firstRow, lastRow),
										jmax (firstRow, lastRow) + 1));

		selected.removeRange (Range <int> (lastRow, lastRow + 1));
	}

	selectRowInternal (lastRow, false, false, true);
}

void ListBox::flipRowSelection (const int row)
{
	if (isRowSelected (row))
		deselectRow (row);
	else
		selectRowInternal (row, false, false, true);
}

void ListBox::deselectAllRows()
{
	if (! selected.isEmpty())
	{
		selected.clear();
		lastRowSelected = -1;

		viewport->updateContents();

		if (model != nullptr)
			model->selectedRowsChanged (lastRowSelected);
	}
}

void ListBox::selectRowsBasedOnModifierKeys (const int row,
											 const ModifierKeys& mods,
											 const bool isMouseUpEvent)
{
	if (multipleSelection && mods.isCommandDown())
	{
		flipRowSelection (row);
	}
	else if (multipleSelection && mods.isShiftDown() && lastRowSelected >= 0)
	{
		selectRangeOfRows (lastRowSelected, row);
	}
	else if ((! mods.isPopupMenu()) || ! isRowSelected (row))
	{
		selectRowInternal (row, false, ! (multipleSelection && (! isMouseUpEvent) && isRowSelected (row)), true);
	}
}

int ListBox::getNumSelectedRows() const
{
	return selected.size();
}

int ListBox::getSelectedRow (const int index) const
{
	return (isPositiveAndBelow (index, selected.size()))
				? selected [index] : -1;
}

bool ListBox::isRowSelected (const int row) const
{
	return selected.contains (row);
}

int ListBox::getLastRowSelected() const
{
	return (isRowSelected (lastRowSelected)) ? lastRowSelected : -1;
}

int ListBox::getRowContainingPosition (const int x, const int y) const noexcept
{
	if (isPositiveAndBelow (x, getWidth()))
	{
		const int row = (viewport->getViewPositionY() + y - viewport->getY()) / rowHeight;

		if (isPositiveAndBelow (row, totalItems))
			return row;
	}

	return -1;
}

int ListBox::getInsertionIndexForPosition (const int x, const int y) const noexcept
{
	if (isPositiveAndBelow (x, getWidth()))
	{
		const int row = (viewport->getViewPositionY() + y + rowHeight / 2 - viewport->getY()) / rowHeight;
		return jlimit (0, totalItems, row);
	}

	return -1;
}

Component* ListBox::getComponentForRowNumber (const int row) const noexcept
{
	ListBoxRowComponent* const listRowComp = viewport->getComponentForRowIfOnscreen (row);
	return listRowComp != nullptr ? static_cast <Component*> (listRowComp->customComponent) : nullptr;
}

int ListBox::getRowNumberOfComponent (Component* const rowComponent) const noexcept
{
	return viewport->getRowNumberOfComponent (rowComponent);
}

const Rectangle<int> ListBox::getRowPosition (const int rowNumber,
											  const bool relativeToComponentTopLeft) const noexcept
{
	int y = viewport->getY() + rowHeight * rowNumber;

	if (relativeToComponentTopLeft)
		y -= viewport->getViewPositionY();

	return Rectangle<int> (viewport->getX(), y,
						   viewport->getViewedComponent()->getWidth(), rowHeight);
}

void ListBox::setVerticalPosition (const double proportion)
{
	const int offscreen = viewport->getViewedComponent()->getHeight() - viewport->getHeight();

	viewport->setViewPosition (viewport->getViewPositionX(),
							   jmax (0, roundToInt (proportion * offscreen)));
}

double ListBox::getVerticalPosition() const
{
	const int offscreen = viewport->getViewedComponent()->getHeight() - viewport->getHeight();

	return (offscreen > 0) ? viewport->getViewPositionY() / (double) offscreen
						   : 0;
}

int ListBox::getVisibleRowWidth() const noexcept
{
	return viewport->getViewWidth();
}

void ListBox::scrollToEnsureRowIsOnscreen (const int row)
{
	viewport->scrollToEnsureRowIsOnscreen (row, getRowHeight());
}

bool ListBox::keyPressed (const KeyPress& key)
{
	const int numVisibleRows = viewport->getHeight() / getRowHeight();

	const bool multiple = multipleSelection
							&& (lastRowSelected >= 0)
							&& (key.getModifiers().isShiftDown()
								 || key.getModifiers().isCtrlDown()
								 || key.getModifiers().isCommandDown());

	if (key.isKeyCode (KeyPress::upKey))
	{
		if (multiple)
			selectRangeOfRows (lastRowSelected, lastRowSelected - 1);
		else
			selectRow (jmax (0, lastRowSelected - 1));
	}
	else if (key.isKeyCode (KeyPress::returnKey)
			  && isRowSelected (lastRowSelected))
	{
		if (model != nullptr)
			model->returnKeyPressed (lastRowSelected);
	}
	else if (key.isKeyCode (KeyPress::pageUpKey))
	{
		if (multiple)
			selectRangeOfRows (lastRowSelected, lastRowSelected - numVisibleRows);
		else
			selectRow (jmax (0, jmax (0, lastRowSelected) - numVisibleRows));
	}
	else if (key.isKeyCode (KeyPress::pageDownKey))
	{
		if (multiple)
			selectRangeOfRows (lastRowSelected, lastRowSelected + numVisibleRows);
		else
			selectRow (jmin (totalItems - 1, jmax (0, lastRowSelected) + numVisibleRows));
	}
	else if (key.isKeyCode (KeyPress::homeKey))
	{
		if (multiple && key.getModifiers().isShiftDown())
			selectRangeOfRows (lastRowSelected, 0);
		else
			selectRow (0);
	}
	else if (key.isKeyCode (KeyPress::endKey))
	{
		if (multiple && key.getModifiers().isShiftDown())
			selectRangeOfRows (lastRowSelected, totalItems - 1);
		else
			selectRow (totalItems - 1);
	}
	else if (key.isKeyCode (KeyPress::downKey))
	{
		if (multiple)
			selectRangeOfRows (lastRowSelected, lastRowSelected + 1);
		else
			selectRow (jmin (totalItems - 1, jmax (0, lastRowSelected) + 1));
	}
	else if ((key.isKeyCode (KeyPress::deleteKey) || key.isKeyCode (KeyPress::backspaceKey))
			   && isRowSelected (lastRowSelected))
	{
		if (model != nullptr)
			model->deleteKeyPressed (lastRowSelected);
	}
	else if (multiple && key == KeyPress ('a', ModifierKeys::commandModifier, 0))
	{
		selectRangeOfRows (0, std::numeric_limits<int>::max());
	}
	else
	{
		return false;
	}

	return true;
}

bool ListBox::keyStateChanged (const bool isKeyDown)
{
	return isKeyDown
			&& (KeyPress::isKeyCurrentlyDown (KeyPress::upKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::pageUpKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::downKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::pageDownKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::homeKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::endKey)
				|| KeyPress::isKeyCurrentlyDown (KeyPress::returnKey));
}

void ListBox::mouseWheelMove (const MouseEvent& e, float wheelIncrementX, float wheelIncrementY)
{
	bool eventWasUsed = false;

	if (viewport->getHorizontalScrollBar()->isVisible() && wheelIncrementX != 0)
	{
		eventWasUsed = true;
		viewport->getHorizontalScrollBar()->mouseWheelMove (e, wheelIncrementX, 0);
	}

	if (viewport->getVerticalScrollBar()->isVisible() && wheelIncrementY != 0)
	{
		eventWasUsed = true;
		viewport->getVerticalScrollBar()->mouseWheelMove (e, 0, wheelIncrementY);
	}

	if (! eventWasUsed)
		Component::mouseWheelMove (e, wheelIncrementX, wheelIncrementY);
}

void ListBox::mouseMove (const MouseEvent& e)
{
	if (mouseMoveSelects)
	{
		const MouseEvent e2 (e.getEventRelativeTo (this));
		selectRow (getRowContainingPosition (e2.x, e2.y), true);
	}
}

void ListBox::mouseExit (const MouseEvent& e)
{
	mouseMove (e);
}

void ListBox::mouseUp (const MouseEvent& e)
{
	if (e.mouseWasClicked() && model != nullptr)
		model->backgroundClicked();
}

void ListBox::setRowHeight (const int newHeight)
{
	rowHeight = jmax (1, newHeight);
	viewport->setSingleStepSizes (20, rowHeight);
	updateContent();
}

int ListBox::getNumRowsOnScreen() const noexcept
{
	return viewport->getMaximumVisibleHeight() / rowHeight;
}

void ListBox::setMinimumContentWidth (const int newMinimumWidth)
{
	minimumRowWidth = newMinimumWidth;
	updateContent();
}

int ListBox::getVisibleContentWidth() const noexcept
{
	return viewport->getMaximumVisibleWidth();
}

ScrollBar* ListBox::getVerticalScrollBar() const noexcept
{
	return viewport->getVerticalScrollBar();
}

ScrollBar* ListBox::getHorizontalScrollBar() const noexcept
{
	return viewport->getHorizontalScrollBar();
}

void ListBox::colourChanged()
{
	setOpaque (findColour (backgroundColourId).isOpaque());
	viewport->setOpaque (isOpaque());
	repaint();
}

void ListBox::setOutlineThickness (const int outlineThickness_)
{
	outlineThickness = outlineThickness_;
	resized();
}

void ListBox::setHeaderComponent (Component* const newHeaderComponent)
{
	if (headerComponent != newHeaderComponent)
	{
		headerComponent = newHeaderComponent;

		addAndMakeVisible (newHeaderComponent);
		ListBox::resized();
	}
}

void ListBox::repaintRow (const int rowNumber) noexcept
{
	repaint (getRowPosition (rowNumber, true));
}

const Image ListBox::createSnapshotOfSelectedRows (int& imageX, int& imageY)
{
	Rectangle<int> imageArea;
	const int firstRow = getRowContainingPosition (0, 0);

	int i;
	for (i = getNumRowsOnScreen() + 2; --i >= 0;)
	{
		Component* rowComp = viewport->getComponentForRowIfOnscreen (firstRow + i);

		if (rowComp != nullptr && isRowSelected (firstRow + i))
		{
			const Point<int> pos (getLocalPoint (rowComp, Point<int>()));
			const Rectangle<int> rowRect (pos.getX(), pos.getY(), rowComp->getWidth(), rowComp->getHeight());
			imageArea = imageArea.getUnion (rowRect);
		}
	}

	imageArea = imageArea.getIntersection (getLocalBounds());
	imageX = imageArea.getX();
	imageY = imageArea.getY();
	Image snapshot (Image::ARGB, imageArea.getWidth(), imageArea.getHeight(), true, Image::NativeImage);

	for (i = getNumRowsOnScreen() + 2; --i >= 0;)
	{
		Component* rowComp = viewport->getComponentForRowIfOnscreen (firstRow + i);

		if (rowComp != nullptr && isRowSelected (firstRow + i))
		{
			const Point<int> pos (getLocalPoint (rowComp, Point<int>()));

			Graphics g (snapshot);
			g.setOrigin (pos.getX() - imageX, pos.getY() - imageY);

			if (g.reduceClipRegion (rowComp->getLocalBounds()))
			{
				g.beginTransparencyLayer (0.6f);
				rowComp->paintEntireComponent (g, false);
				g.endTransparencyLayer();
			}
		}
	}

	return snapshot;
}

void ListBox::startDragAndDrop (const MouseEvent& e, const var& dragDescription, bool allowDraggingToOtherWindows)
{
	DragAndDropContainer* const dragContainer
		= DragAndDropContainer::findParentDragContainerFor (this);

	if (dragContainer != nullptr)
	{
		int x, y;
		Image dragImage (createSnapshotOfSelectedRows (x, y));

		MouseEvent e2 (e.getEventRelativeTo (this));
		const Point<int> p (x - e2.x, y - e2.y);
		dragContainer->startDragging (dragDescription, this, dragImage, allowDraggingToOtherWindows, &p);
	}
	else
	{
		// to be able to do a drag-and-drop operation, the listbox needs to
		// be inside a component which is also a DragAndDropContainer.
		jassertfalse;
	}
}

Component* ListBoxModel::refreshComponentForRow (int, bool, Component* existingComponentToUpdate)
{
	(void) existingComponentToUpdate;
	jassert (existingComponentToUpdate == nullptr); // indicates a failure in the code the recycles the components
	return nullptr;
}

void ListBoxModel::listBoxItemClicked (int, const MouseEvent&) {}
void ListBoxModel::listBoxItemDoubleClicked (int, const MouseEvent&) {}
void ListBoxModel::backgroundClicked() {}
void ListBoxModel::selectedRowsChanged (int) {}
void ListBoxModel::deleteKeyPressed (int) {}
void ListBoxModel::returnKeyPressed (int) {}
void ListBoxModel::listWasScrolled() {}
const var ListBoxModel::getDragSourceDescription (const SparseSet<int>&)	{ return var::null; }
const String ListBoxModel::getTooltipForRow (int)			   { return String::empty; }

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ListBox.cpp ***/


/*** Start of inlined file: juce_ProgressBar.cpp ***/
BEGIN_JUCE_NAMESPACE

ProgressBar::ProgressBar (double& progress_)
   : progress (progress_),
	 displayPercentage (true),
	 lastCallbackTime (0)
{
	currentValue = jlimit (0.0, 1.0, progress);
}

ProgressBar::~ProgressBar()
{
}

void ProgressBar::setPercentageDisplay (const bool shouldDisplayPercentage)
{
	displayPercentage = shouldDisplayPercentage;
	repaint();
}

void ProgressBar::setTextToDisplay (const String& text)
{
	displayPercentage = false;
	displayedMessage = text;
}

void ProgressBar::lookAndFeelChanged()
{
	setOpaque (findColour (backgroundColourId).isOpaque());
}

void ProgressBar::colourChanged()
{
	lookAndFeelChanged();
}

void ProgressBar::paint (Graphics& g)
{
	String text;

	if (displayPercentage)
	{
		if (currentValue >= 0 && currentValue <= 1.0)
			text << roundToInt (currentValue * 100.0) << '%';
	}
	else
	{
		text = displayedMessage;
	}

	getLookAndFeel().drawProgressBar (g, *this,
									  getWidth(), getHeight(),
									  currentValue, text);
}

void ProgressBar::visibilityChanged()
{
	if (isVisible())
		startTimer (30);
	else
		stopTimer();
}

void ProgressBar::timerCallback()
{
	double newProgress = progress;

	const uint32 now = Time::getMillisecondCounter();
	const int timeSinceLastCallback = (int) (now - lastCallbackTime);
	lastCallbackTime = now;

	if (currentValue != newProgress
		 || newProgress < 0 || newProgress >= 1.0
		 || currentMessage != displayedMessage)
	{
		if (currentValue < newProgress
			 && newProgress >= 0 && newProgress < 1.0
			 && currentValue >= 0 && currentValue < 1.0)
		{
			newProgress = jmin (currentValue + 0.0008 * timeSinceLastCallback,
								newProgress);
		}

		currentValue = newProgress;
		currentMessage = displayedMessage;
		repaint();
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ProgressBar.cpp ***/


/*** Start of inlined file: juce_Slider.cpp ***/
BEGIN_JUCE_NAMESPACE

class Slider::PopupDisplayComponent  : public BubbleComponent,
									   public Timer
{
public:

	PopupDisplayComponent (Slider& owner_)
		: owner (owner_),
		  font (15.0f, Font::bold)
	{
		setAlwaysOnTop (true);
	}

	void paintContent (Graphics& g, int w, int h)
	{
		g.setFont (font);
		g.setColour (findColour (TooltipWindow::textColourId, true));
		g.drawFittedText (text, 0, 0, w, h, Justification::centred, 1);
	}

	void getContentSize (int& w, int& h)
	{
		w = font.getStringWidth (text) + 18;
		h = (int) (font.getHeight() * 1.6f);
	}

	void updatePosition (const String& newText)
	{
		if (text != newText)
		{
			text = newText;
			repaint();
		}

		BubbleComponent::setPosition (&owner);
		repaint();
	}

	void timerCallback()
	{
		owner.popupDisplay = nullptr;
	}

private:
	Slider& owner;
	Font font;
	String text;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (PopupDisplayComponent);
};

Slider::Slider (const String& name)
  : Component (name),
	lastCurrentValue (0),
	lastValueMin (0),
	lastValueMax (0),
	minimum (0),
	maximum (10),
	interval (0),
	skewFactor (1.0),
	velocityModeSensitivity (1.0),
	velocityModeOffset (0.0),
	velocityModeThreshold (1),
	rotaryStart (float_Pi * 1.2f),
	rotaryEnd (float_Pi * 2.8f),
	numDecimalPlaces (7),
	sliderRegionStart (0),
	sliderRegionSize (1),
	sliderBeingDragged (-1),
	pixelsForFullDragExtent (250),
	style (LinearHorizontal),
	textBoxPos (TextBoxLeft),
	textBoxWidth (80),
	textBoxHeight (20),
	incDecButtonMode (incDecButtonsNotDraggable),
	editableText (true),
	doubleClickToValue (false),
	isVelocityBased (false),
	userKeyOverridesVelocity (true),
	rotaryStop (true),
	incDecButtonsSideBySide (false),
	sendChangeOnlyOnRelease (false),
	popupDisplayEnabled (false),
	menuEnabled (false),
	menuShown (false),
	scrollWheelEnabled (true),
	snapsToMousePos (true),
	parentForPopupDisplay (nullptr)
{
	setWantsKeyboardFocus (false);
	setRepaintsOnMouseActivity (true);

	Slider::lookAndFeelChanged();
	updateText();

	currentValue.addListener (this);
	valueMin.addListener (this);
	valueMax.addListener (this);
}

Slider::~Slider()
{
	currentValue.removeListener (this);
	valueMin.removeListener (this);
	valueMax.removeListener (this);
	popupDisplay = nullptr;
}

void Slider::handleAsyncUpdate()
{
	cancelPendingUpdate();

	Component::BailOutChecker checker (this);
	listeners.callChecked (checker, &SliderListener::sliderValueChanged, this);  // (can't use Slider::Listener due to idiotic VC2005 bug)
}

void Slider::sendDragStart()
{
	startedDragging();

	Component::BailOutChecker checker (this);
	listeners.callChecked (checker, &SliderListener::sliderDragStarted, this);
}

void Slider::sendDragEnd()
{
	stoppedDragging();

	sliderBeingDragged = -1;

	Component::BailOutChecker checker (this);
	listeners.callChecked (checker, &SliderListener::sliderDragEnded, this);
}

void Slider::addListener (SliderListener* const listener)
{
	listeners.add (listener);
}

void Slider::removeListener (SliderListener* const listener)
{
	listeners.remove (listener);
}

void Slider::setSliderStyle (const SliderStyle newStyle)
{
	if (style != newStyle)
	{
		style = newStyle;
		repaint();
		lookAndFeelChanged();
	}
}

void Slider::setRotaryParameters (const float startAngleRadians,
								  const float endAngleRadians,
								  const bool stopAtEnd)
{
	// make sure the values are sensible..
	jassert (rotaryStart >= 0 && rotaryEnd >= 0);
	jassert (rotaryStart < float_Pi * 4.0f && rotaryEnd < float_Pi * 4.0f);
	jassert (rotaryStart < rotaryEnd);

	rotaryStart = startAngleRadians;
	rotaryEnd = endAngleRadians;
	rotaryStop = stopAtEnd;
}

void Slider::setVelocityBasedMode (const bool velBased)
{
	isVelocityBased = velBased;
}

void Slider::setVelocityModeParameters (const double sensitivity,
										const int threshold,
										const double offset,
										const bool userCanPressKeyToSwapMode)
{
	jassert (threshold >= 0);
	jassert (sensitivity > 0);
	jassert (offset >= 0);

	velocityModeSensitivity = sensitivity;
	velocityModeOffset = offset;
	velocityModeThreshold = threshold;
	userKeyOverridesVelocity = userCanPressKeyToSwapMode;
}

void Slider::setSkewFactor (const double factor)
{
	skewFactor = factor;
}

void Slider::setSkewFactorFromMidPoint (const double sliderValueToShowAtMidPoint)
{
	if (maximum > minimum)
		skewFactor = log (0.5) / log ((sliderValueToShowAtMidPoint - minimum)
										/ (maximum - minimum));
}

void Slider::setMouseDragSensitivity (const int distanceForFullScaleDrag)
{
	jassert (distanceForFullScaleDrag > 0);

	pixelsForFullDragExtent = distanceForFullScaleDrag;
}

void Slider::setIncDecButtonsMode (const IncDecButtonMode mode)
{
	if (incDecButtonMode != mode)
	{
		incDecButtonMode = mode;
		lookAndFeelChanged();
	}
}

void Slider::setTextBoxStyle (const TextEntryBoxPosition newPosition,
							  const bool isReadOnly,
							  const int textEntryBoxWidth,
							  const int textEntryBoxHeight)
{
	if (textBoxPos != newPosition
		 || editableText != (! isReadOnly)
		 || textBoxWidth != textEntryBoxWidth
		 || textBoxHeight != textEntryBoxHeight)
	{
		textBoxPos = newPosition;
		editableText = ! isReadOnly;
		textBoxWidth = textEntryBoxWidth;
		textBoxHeight = textEntryBoxHeight;

		repaint();
		lookAndFeelChanged();
	}
}

void Slider::setTextBoxIsEditable (const bool shouldBeEditable)
{
	editableText = shouldBeEditable;

	if (valueBox != nullptr)
		valueBox->setEditable (shouldBeEditable && isEnabled());
}

void Slider::showTextBox()
{
	jassert (editableText); // this should probably be avoided in read-only sliders.

	if (valueBox != nullptr)
		valueBox->showEditor();
}

void Slider::hideTextBox (const bool discardCurrentEditorContents)
{
	if (valueBox != nullptr)
	{
		valueBox->hideEditor (discardCurrentEditorContents);

		if (discardCurrentEditorContents)
			updateText();
	}
}

void Slider::setChangeNotificationOnlyOnRelease (const bool onlyNotifyOnRelease)
{
	sendChangeOnlyOnRelease = onlyNotifyOnRelease;
}

void Slider::setSliderSnapsToMousePosition (const bool shouldSnapToMouse)
{
	snapsToMousePos = shouldSnapToMouse;
}

void Slider::setPopupDisplayEnabled (const bool enabled,
									 Component* const parentComponentToUse)
{
	popupDisplayEnabled = enabled;
	parentForPopupDisplay = parentComponentToUse;
}

void Slider::colourChanged()
{
	lookAndFeelChanged();
}

void Slider::lookAndFeelChanged()
{
	LookAndFeel& lf = getLookAndFeel();

	if (textBoxPos != NoTextBox)
	{
		const String previousTextBoxContent (valueBox != nullptr ? valueBox->getText()
																 : getTextFromValue (currentValue.getValue()));

		valueBox = nullptr;
		addAndMakeVisible (valueBox = getLookAndFeel().createSliderTextBox (*this));

		valueBox->setWantsKeyboardFocus (false);
		valueBox->setText (previousTextBoxContent, false);

		if (valueBox->isEditable() != editableText) // (avoid overriding the single/double click flags unless we have to)
			valueBox->setEditable (editableText && isEnabled());

		valueBox->addListener (this);

		if (style == LinearBar)
			valueBox->addMouseListener (this, false);

		valueBox->setTooltip (getTooltip());
	}
	else
	{
		valueBox = nullptr;
	}

	if (style == IncDecButtons)
	{
		addAndMakeVisible (incButton = lf.createSliderButton (true));
		incButton->addListener (this);

		addAndMakeVisible (decButton = lf.createSliderButton (false));
		decButton->addListener (this);

		if (incDecButtonMode != incDecButtonsNotDraggable)
		{
			incButton->addMouseListener (this, false);
			decButton->addMouseListener (this, false);
		}
		else
		{
			incButton->setRepeatSpeed (300, 100, 20);
			incButton->addMouseListener (decButton, false);

			decButton->setRepeatSpeed (300, 100, 20);
			decButton->addMouseListener (incButton, false);
		}

		incButton->setTooltip (getTooltip());
		decButton->setTooltip (getTooltip());
	}
	else
	{
		incButton = nullptr;
		decButton = nullptr;
	}

	setComponentEffect (lf.getSliderEffect());

	resized();
	repaint();
}

void Slider::setRange (const double newMin,
					   const double newMax,
					   const double newInt)
{
	if (minimum != newMin
		|| maximum != newMax
		|| interval != newInt)
	{
		minimum = newMin;
		maximum = newMax;
		interval = newInt;

		// figure out the number of DPs needed to display all values at this
		// interval setting.
		numDecimalPlaces = 7;

		if (newInt != 0)
		{
			int v = abs ((int) (newInt * 10000000));

			while ((v % 10) == 0)
			{
				--numDecimalPlaces;
				v /= 10;
			}
		}

		// keep the current values inside the new range..
		if (style != TwoValueHorizontal && style != TwoValueVertical)
		{
			setValue (getValue(), false, false);
		}
		else
		{
			setMinValue (getMinValue(), false, false);
			setMaxValue (getMaxValue(), false, false);
		}

		updateText();
	}
}

void Slider::triggerChangeMessage (const bool synchronous)
{
	if (synchronous)
		handleAsyncUpdate();
	else
		triggerAsyncUpdate();

	valueChanged();
}

void Slider::valueChanged (Value& value)
{
	if (value.refersToSameSourceAs (currentValue))
	{
		if (style != TwoValueHorizontal && style != TwoValueVertical)
			setValue (currentValue.getValue(), false, false);
	}
	else if (value.refersToSameSourceAs (valueMin))
		setMinValue (valueMin.getValue(), false, false, true);
	else if (value.refersToSameSourceAs (valueMax))
		setMaxValue (valueMax.getValue(), false, false, true);
}

double Slider::getValue() const
{
	// for a two-value style slider, you should use the getMinValue() and getMaxValue()
	// methods to get the two values.
	jassert (style != TwoValueHorizontal && style != TwoValueVertical);

	return currentValue.getValue();
}

void Slider::setValue (double newValue,
					   const bool sendUpdateMessage,
					   const bool sendMessageSynchronously)
{
	// for a two-value style slider, you should use the setMinValue() and setMaxValue()
	// methods to set the two values.
	jassert (style != TwoValueHorizontal && style != TwoValueVertical);

	newValue = constrainedValue (newValue);

	if (style == ThreeValueHorizontal || style == ThreeValueVertical)
	{
		jassert ((double) valueMin.getValue() <= (double) valueMax.getValue());

		newValue = jlimit ((double) valueMin.getValue(),
						   (double) valueMax.getValue(),
						   newValue);
	}

	if (newValue != lastCurrentValue)
	{
		if (valueBox != nullptr)
			valueBox->hideEditor (true);

		lastCurrentValue = newValue;

		// (need to do this comparison because the Value will use equalsWithSameType to compare
		// the new and old values, so will generate unwanted change events if the type changes)
		if (currentValue != newValue)
			currentValue = newValue;

		updateText();
		repaint();

		if (popupDisplay != nullptr)
			popupDisplay->updatePosition (getTextFromValue (newValue));

		if (sendUpdateMessage)
			triggerChangeMessage (sendMessageSynchronously);
	}
}

double Slider::getMinValue() const
{
	// The minimum value only applies to sliders that are in two- or three-value mode.
	jassert (style == TwoValueHorizontal || style == TwoValueVertical
			  || style == ThreeValueHorizontal || style == ThreeValueVertical);

	return valueMin.getValue();
}

double Slider::getMaxValue() const
{
	// The maximum value only applies to sliders that are in two- or three-value mode.
	jassert (style == TwoValueHorizontal || style == TwoValueVertical
			  || style == ThreeValueHorizontal || style == ThreeValueVertical);

	return valueMax.getValue();
}

void Slider::setMinValue (double newValue, const bool sendUpdateMessage, const bool sendMessageSynchronously, const bool allowNudgingOfOtherValues)
{
	// The minimum value only applies to sliders that are in two- or three-value mode.
	jassert (style == TwoValueHorizontal || style == TwoValueVertical
			  || style == ThreeValueHorizontal || style == ThreeValueVertical);

	newValue = constrainedValue (newValue);

	if (style == TwoValueHorizontal || style == TwoValueVertical)
	{
		if (allowNudgingOfOtherValues && newValue > (double) valueMax.getValue())
			setMaxValue (newValue, sendUpdateMessage, sendMessageSynchronously);

		newValue = jmin ((double) valueMax.getValue(), newValue);
	}
	else
	{
		if (allowNudgingOfOtherValues && newValue > lastCurrentValue)
			setValue (newValue, sendUpdateMessage, sendMessageSynchronously);

		newValue = jmin (lastCurrentValue, newValue);
	}

	if (lastValueMin != newValue)
	{
		lastValueMin = newValue;
		valueMin = newValue;
		repaint();

		if (popupDisplay != nullptr)
			popupDisplay->updatePosition (getTextFromValue (newValue));

		if (sendUpdateMessage)
			triggerChangeMessage (sendMessageSynchronously);
	}
}

void Slider::setMaxValue (double newValue, const bool sendUpdateMessage, const bool sendMessageSynchronously, const bool allowNudgingOfOtherValues)
{
	// The maximum value only applies to sliders that are in two- or three-value mode.
	jassert (style == TwoValueHorizontal || style == TwoValueVertical
			  || style == ThreeValueHorizontal || style == ThreeValueVertical);

	newValue = constrainedValue (newValue);

	if (style == TwoValueHorizontal || style == TwoValueVertical)
	{
		if (allowNudgingOfOtherValues && newValue < (double) valueMin.getValue())
			setMinValue (newValue, sendUpdateMessage, sendMessageSynchronously);

		newValue = jmax ((double) valueMin.getValue(), newValue);
	}
	else
	{
		if (allowNudgingOfOtherValues && newValue < lastCurrentValue)
			setValue (newValue, sendUpdateMessage, sendMessageSynchronously);

		newValue = jmax (lastCurrentValue, newValue);
	}

	if (lastValueMax != newValue)
	{
		lastValueMax = newValue;
		valueMax = newValue;
		repaint();

		if (popupDisplay != nullptr)
			popupDisplay->updatePosition (getTextFromValue (valueMax.getValue()));

		if (sendUpdateMessage)
			triggerChangeMessage (sendMessageSynchronously);
	}
}

void Slider::setMinAndMaxValues (double newMinValue, double newMaxValue, bool sendUpdateMessage, bool sendMessageSynchronously)
{
	// The maximum value only applies to sliders that are in two- or three-value mode.
	jassert (style == TwoValueHorizontal || style == TwoValueVertical
			  || style == ThreeValueHorizontal || style == ThreeValueVertical);

	if (newMaxValue < newMinValue)
		std::swap (newMaxValue, newMinValue);

	newMinValue = constrainedValue (newMinValue);
	newMaxValue = constrainedValue (newMaxValue);

	if (lastValueMax != newMaxValue || lastValueMin != newMinValue)
	{
		lastValueMax = newMaxValue;
		lastValueMin = newMinValue;
		valueMin = newMinValue;
		valueMax = newMaxValue;
		repaint();

		if (sendUpdateMessage)
			triggerChangeMessage (sendMessageSynchronously);
	}
}

void Slider::setDoubleClickReturnValue (const bool isDoubleClickEnabled,
										const double valueToSetOnDoubleClick)
{
	doubleClickToValue = isDoubleClickEnabled;
	doubleClickReturnValue = valueToSetOnDoubleClick;
}

double Slider::getDoubleClickReturnValue (bool& isEnabled_) const
{
	isEnabled_ = doubleClickToValue;
	return doubleClickReturnValue;
}

void Slider::updateText()
{
	if (valueBox != nullptr)
		valueBox->setText (getTextFromValue (currentValue.getValue()), false);
}

void Slider::setTextValueSuffix (const String& suffix)
{
	if (textSuffix != suffix)
	{
		textSuffix = suffix;
		updateText();
	}
}

String Slider::getTextValueSuffix() const
{
	return textSuffix;
}

const String Slider::getTextFromValue (double v)
{
	if (getNumDecimalPlacesToDisplay() > 0)
		return String (v, getNumDecimalPlacesToDisplay()) + getTextValueSuffix();
	else
		return String (roundToInt (v)) + getTextValueSuffix();
}

double Slider::getValueFromText (const String& text)
{
	String t (text.trimStart());

	if (t.endsWith (textSuffix))
		t = t.substring (0, t.length() - textSuffix.length());

	while (t.startsWithChar ('+'))
		t = t.substring (1).trimStart();

	return t.initialSectionContainingOnly ("0123456789.,-")
			.getDoubleValue();
}

double Slider::proportionOfLengthToValue (double proportion)
{
	if (skewFactor != 1.0 && proportion > 0.0)
		proportion = exp (log (proportion) / skewFactor);

	return minimum + (maximum - minimum) * proportion;
}

double Slider::valueToProportionOfLength (double value)
{
	const double n = (value - minimum) / (maximum - minimum);

	return skewFactor == 1.0 ? n : pow (n, skewFactor);
}

double Slider::snapValue (double attemptedValue, const bool)
{
	return attemptedValue;
}

void Slider::startedDragging()
{
}

void Slider::stoppedDragging()
{
}

void Slider::valueChanged()
{
}

void Slider::enablementChanged()
{
	repaint();
}

void Slider::setPopupMenuEnabled (const bool menuEnabled_)
{
	menuEnabled = menuEnabled_;
}

void Slider::setScrollWheelEnabled (const bool enabled)
{
	scrollWheelEnabled = enabled;
}

void Slider::labelTextChanged (Label* label)
{
	const double newValue = snapValue (getValueFromText (label->getText()), false);

	if (newValue != (double) currentValue.getValue())
	{
		sendDragStart();
		setValue (newValue, true, true);
		sendDragEnd();
	}

	updateText(); // force a clean-up of the text, needed in case setValue() hasn't done this.
}

void Slider::buttonClicked (Button* button)
{
	if (style == IncDecButtons)
	{
		sendDragStart();

		if (button == incButton)
			setValue (snapValue (getValue() + interval, false), true, true);
		else if (button == decButton)
			setValue (snapValue (getValue() - interval, false), true, true);

		sendDragEnd();
	}
}

double Slider::constrainedValue (double value) const
{
	if (interval > 0)
		value = minimum + interval * std::floor ((value - minimum) / interval + 0.5);

	if (value <= minimum || maximum <= minimum)
		value = minimum;
	else if (value >= maximum)
		value = maximum;

	return value;
}

float Slider::getLinearSliderPos (const double value)
{
	double sliderPosProportional;

	if (maximum > minimum)
	{
		if (value < minimum)
		{
			sliderPosProportional = 0.0;
		}
		else if (value > maximum)
		{
			sliderPosProportional = 1.0;
		}
		else
		{
			sliderPosProportional = valueToProportionOfLength (value);
			jassert (sliderPosProportional >= 0 && sliderPosProportional <= 1.0);
		}
	}
	else
	{
		sliderPosProportional = 0.5;
	}

	if (isVertical() || style == IncDecButtons)
		sliderPosProportional = 1.0 - sliderPosProportional;

	return (float) (sliderRegionStart + sliderPosProportional * sliderRegionSize);
}

bool Slider::isHorizontal() const
{
	return style == LinearHorizontal
		|| style == LinearBar
		|| style == TwoValueHorizontal
		|| style == ThreeValueHorizontal;
}

bool Slider::isVertical() const
{
	return style == LinearVertical
		|| style == TwoValueVertical
		|| style == ThreeValueVertical;
}

bool Slider::incDecDragDirectionIsHorizontal() const
{
	return incDecButtonMode == incDecButtonsDraggable_Horizontal
			|| (incDecButtonMode == incDecButtonsDraggable_AutoDirection && incDecButtonsSideBySide);
}

float Slider::getPositionOfValue (const double value)
{
	if (isHorizontal() || isVertical())
	{
		return getLinearSliderPos (value);
	}
	else
	{
		jassertfalse; // not a valid call on a slider that doesn't work linearly!
		return 0.0f;
	}
}

void Slider::paint (Graphics& g)
{
	if (style != IncDecButtons)
	{
		if (style == Rotary || style == RotaryHorizontalDrag || style == RotaryVerticalDrag)
		{
			const float sliderPos = (float) valueToProportionOfLength (lastCurrentValue);
			jassert (sliderPos >= 0 && sliderPos <= 1.0f);

			getLookAndFeel().drawRotarySlider (g,
											   sliderRect.getX(),
											   sliderRect.getY(),
											   sliderRect.getWidth(),
											   sliderRect.getHeight(),
											   sliderPos,
											   rotaryStart, rotaryEnd,
											   *this);
		}
		else
		{
			getLookAndFeel().drawLinearSlider (g,
											   sliderRect.getX(),
											   sliderRect.getY(),
											   sliderRect.getWidth(),
											   sliderRect.getHeight(),
											   getLinearSliderPos (lastCurrentValue),
											   getLinearSliderPos (lastValueMin),
											   getLinearSliderPos (lastValueMax),
											   style,
											   *this);
		}

		if (style == LinearBar && valueBox == nullptr)
		{
			g.setColour (findColour (Slider::textBoxOutlineColourId));
			g.drawRect (0, 0, getWidth(), getHeight(), 1);
		}
	}
}

void Slider::resized()
{
	int minXSpace = 0;
	int minYSpace = 0;

	if (textBoxPos == TextBoxLeft || textBoxPos == TextBoxRight)
		minXSpace = 30;
	else
		minYSpace = 15;

	const int tbw = jmax (0, jmin (textBoxWidth, getWidth() - minXSpace));
	const int tbh = jmax (0, jmin (textBoxHeight, getHeight() - minYSpace));

	if (style == LinearBar)
	{
		if (valueBox != nullptr)
			valueBox->setBounds (getLocalBounds());
	}
	else
	{
		if (textBoxPos == NoTextBox)
		{
			sliderRect = getLocalBounds();
		}
		else if (textBoxPos == TextBoxLeft)
		{
			valueBox->setBounds (0, (getHeight() - tbh) / 2, tbw, tbh);
			sliderRect.setBounds (tbw, 0, getWidth() - tbw, getHeight());
		}
		else if (textBoxPos == TextBoxRight)
		{
			valueBox->setBounds (getWidth() - tbw, (getHeight() - tbh) / 2, tbw, tbh);
			sliderRect.setBounds (0, 0, getWidth() - tbw, getHeight());
		}
		else if (textBoxPos == TextBoxAbove)
		{
			valueBox->setBounds ((getWidth() - tbw) / 2, 0, tbw, tbh);
			sliderRect.setBounds (0, tbh, getWidth(), getHeight() - tbh);
		}
		else if (textBoxPos == TextBoxBelow)
		{
			valueBox->setBounds ((getWidth() - tbw) / 2, getHeight() - tbh, tbw, tbh);
			sliderRect.setBounds (0, 0, getWidth(), getHeight() - tbh);
		}
	}

	const int indent = getLookAndFeel().getSliderThumbRadius (*this);

	if (style == LinearBar)
	{
		const int barIndent = 1;
		sliderRegionStart = barIndent;
		sliderRegionSize = getWidth() - barIndent * 2;

		sliderRect.setBounds (sliderRegionStart, barIndent,
							  sliderRegionSize, getHeight() - barIndent * 2);
	}
	else if (isHorizontal())
	{
		sliderRegionStart = sliderRect.getX() + indent;
		sliderRegionSize = jmax (1, sliderRect.getWidth() - indent * 2);

		sliderRect.setBounds (sliderRegionStart, sliderRect.getY(),
							  sliderRegionSize, sliderRect.getHeight());
	}
	else if (isVertical())
	{
		sliderRegionStart = sliderRect.getY() + indent;
		sliderRegionSize = jmax (1, sliderRect.getHeight() - indent * 2);

		sliderRect.setBounds (sliderRect.getX(), sliderRegionStart,
							  sliderRect.getWidth(), sliderRegionSize);
	}
	else
	{
		sliderRegionStart = 0;
		sliderRegionSize = 100;
	}

	if (style == IncDecButtons)
	{
		Rectangle<int> buttonRect (sliderRect);

		if (textBoxPos == TextBoxLeft || textBoxPos == TextBoxRight)
			buttonRect.expand (-2, 0);
		else
			buttonRect.expand (0, -2);

		incDecButtonsSideBySide = buttonRect.getWidth() > buttonRect.getHeight();

		if (incDecButtonsSideBySide)
		{
			decButton->setBounds (buttonRect.getX(),
								  buttonRect.getY(),
								  buttonRect.getWidth() / 2,
								  buttonRect.getHeight());

			decButton->setConnectedEdges (Button::ConnectedOnRight);

			incButton->setBounds (buttonRect.getCentreX(),
								  buttonRect.getY(),
								  buttonRect.getWidth() / 2,
								  buttonRect.getHeight());

			incButton->setConnectedEdges (Button::ConnectedOnLeft);
		}
		else
		{
			incButton->setBounds (buttonRect.getX(),
								  buttonRect.getY(),
								  buttonRect.getWidth(),
								  buttonRect.getHeight() / 2);

			incButton->setConnectedEdges (Button::ConnectedOnBottom);

			decButton->setBounds (buttonRect.getX(),
								  buttonRect.getCentreY(),
								  buttonRect.getWidth(),
								  buttonRect.getHeight() / 2);

			decButton->setConnectedEdges (Button::ConnectedOnTop);
		}
	}
}

void Slider::focusOfChildComponentChanged (FocusChangeType)
{
	repaint();
}

static void sliderMenuCallback (int result, Slider* slider)
{
	if (slider != nullptr)
	{
		switch (result)
		{
			case 1: slider->setVelocityBasedMode (! slider->getVelocityBasedMode()); break;
			case 2: slider->setSliderStyle (Slider::Rotary); break;
			case 3: slider->setSliderStyle (Slider::RotaryHorizontalDrag); break;
			case 4: slider->setSliderStyle (Slider::RotaryVerticalDrag); break;
			default: break;
		}
	}
}

void Slider::mouseDown (const MouseEvent& e)
{
	mouseWasHidden = false;
	incDecDragged = false;
	mousePosWhenLastDragged = e.getPosition();
	mouseDragStartX = e.getMouseDownX();
	mouseDragStartY = e.getMouseDownY();

	if (isEnabled())
	{
		if (e.mods.isPopupMenu() && menuEnabled)
		{
			menuShown = true;

			PopupMenu m;
			m.setLookAndFeel (&getLookAndFeel());
			m.addItem (1, TRANS ("velocity-sensitive mode"), true, isVelocityBased);
			m.addSeparator();

			if (style == Rotary || style == RotaryHorizontalDrag || style == RotaryVerticalDrag)
			{
				PopupMenu rotaryMenu;
				rotaryMenu.addItem (2, TRANS ("use circular dragging"), true, style == Rotary);
				rotaryMenu.addItem (3, TRANS ("use left-right dragging"), true, style == RotaryHorizontalDrag);
				rotaryMenu.addItem (4, TRANS ("use up-down dragging"), true, style == RotaryVerticalDrag);

				m.addSubMenu (TRANS ("rotary mode"), rotaryMenu);
			}

			m.showMenuAsync (PopupMenu::Options(),
							 ModalCallbackFunction::forComponent (sliderMenuCallback, this));
		}
		else if (maximum > minimum)
		{
			menuShown = false;

			if (valueBox != nullptr)
				valueBox->hideEditor (true);

			sliderBeingDragged = 0;

			if (style == TwoValueHorizontal
				 || style == TwoValueVertical
				 || style == ThreeValueHorizontal
				 || style == ThreeValueVertical)
			{
				const float mousePos = (float) (isVertical() ? e.y : e.x);

				const float normalPosDistance = std::abs (getLinearSliderPos (currentValue.getValue()) - mousePos);
				const float minPosDistance = std::abs (getLinearSliderPos (valueMin.getValue()) - 0.1f - mousePos);
				const float maxPosDistance = std::abs (getLinearSliderPos (valueMax.getValue()) + 0.1f - mousePos);

				if (style == TwoValueHorizontal || style == TwoValueVertical)
				{
					if (maxPosDistance <= minPosDistance)
						sliderBeingDragged = 2;
					else
						sliderBeingDragged = 1;
				}
				else if (style == ThreeValueHorizontal || style == ThreeValueVertical)
				{
					if (normalPosDistance >= minPosDistance && maxPosDistance >= minPosDistance)
						sliderBeingDragged = 1;
					else if (normalPosDistance >= maxPosDistance)
						sliderBeingDragged = 2;
				}
			}

			minMaxDiff = (double) valueMax.getValue() - (double) valueMin.getValue();

			lastAngle = rotaryStart + (rotaryEnd - rotaryStart)
										* valueToProportionOfLength (currentValue.getValue());

			valueWhenLastDragged = ((sliderBeingDragged == 2) ? valueMax
															  : ((sliderBeingDragged == 1) ? valueMin
																						   : currentValue)).getValue();

			valueOnMouseDown = valueWhenLastDragged;

			if (popupDisplayEnabled)
			{
				PopupDisplayComponent* const popup = new PopupDisplayComponent (*this);
				popupDisplay = popup;

				if (parentForPopupDisplay != nullptr)
					parentForPopupDisplay->addChildComponent (popup);
				else
					popup->addToDesktop (0);

				popup->setVisible (true);
			}

			sendDragStart();

			mouseDrag (e);
		}
	}
}

void Slider::mouseUp (const MouseEvent&)
{
	if (isEnabled()
		 && (! menuShown)
		 && (maximum > minimum)
		 && (style != IncDecButtons || incDecDragged))
	{
		restoreMouseIfHidden();

		if (sendChangeOnlyOnRelease && valueOnMouseDown != (double) currentValue.getValue())
			triggerChangeMessage (false);

		sendDragEnd();
		popupDisplay = nullptr;

		if (style == IncDecButtons)
		{
			incButton->setState (Button::buttonNormal);
			decButton->setState (Button::buttonNormal);
		}
	}
	else if (popupDisplay != nullptr)
	{
		popupDisplay->startTimer (2000);
	}
}

void Slider::restoreMouseIfHidden()
{
	if (mouseWasHidden)
	{
		mouseWasHidden = false;

		for (int i = Desktop::getInstance().getNumMouseSources(); --i >= 0;)
			Desktop::getInstance().getMouseSource(i)->enableUnboundedMouseMovement (false);

		const double pos = (sliderBeingDragged == 2) ? getMaxValue()
													 : ((sliderBeingDragged == 1) ? getMinValue()
																				  : (double) currentValue.getValue());

		Point<int> mousePos;

		if (style == RotaryHorizontalDrag || style == RotaryVerticalDrag)
		{
			mousePos = Desktop::getLastMouseDownPosition();

			if (style == RotaryHorizontalDrag)
			{
				const double posDiff = valueToProportionOfLength (pos) - valueToProportionOfLength (valueOnMouseDown);
				mousePos += Point<int> (roundToInt (pixelsForFullDragExtent * posDiff), 0);
			}
			else
			{
				const double posDiff = valueToProportionOfLength (valueOnMouseDown) - valueToProportionOfLength (pos);
				mousePos += Point<int> (0, roundToInt (pixelsForFullDragExtent * posDiff));
			}
		}
		else
		{
			const int pixelPos = (int) getLinearSliderPos (pos);

			mousePos = localPointToGlobal (Point<int> (isHorizontal() ? pixelPos : (getWidth() / 2),
													   isVertical()   ? pixelPos : (getHeight() / 2)));
		}

		Desktop::setMousePosition (mousePos);
	}
}

void Slider::modifierKeysChanged (const ModifierKeys& modifiers)
{
	if (isEnabled()
		 && style != IncDecButtons
		 && style != Rotary
		 && isVelocityBased == modifiers.isAnyModifierKeyDown())
	{
		restoreMouseIfHidden();
	}
}

namespace SliderHelpers
{
	double smallestAngleBetween (double a1, double a2) noexcept
	{
		return jmin (std::abs (a1 - a2),
					 std::abs (a1 + double_Pi * 2.0 - a2),
					 std::abs (a2 + double_Pi * 2.0 - a1));
	}
}

void Slider::mouseDrag (const MouseEvent& e)
{
	if (isEnabled()
		 && (! menuShown)
		 && (maximum > minimum))
	{
		if (style == Rotary)
		{
			const int dx = e.x - sliderRect.getCentreX();
			const int dy = e.y - sliderRect.getCentreY();

			if (dx * dx + dy * dy > 25)
			{
				double angle = std::atan2 ((double) dx, (double) -dy);
				while (angle < 0.0)
					angle += double_Pi * 2.0;

				if (rotaryStop && ! e.mouseWasClicked())
				{
					if (std::abs (angle - lastAngle) > double_Pi)
					{
						if (angle >= lastAngle)
							angle -= double_Pi * 2.0;
						else
							angle += double_Pi * 2.0;
					}

					if (angle >= lastAngle)
						angle = jmin (angle, (double) jmax (rotaryStart, rotaryEnd));
					else
						angle = jmax (angle, (double) jmin (rotaryStart, rotaryEnd));
				}
				else
				{
					while (angle < rotaryStart)
						angle += double_Pi * 2.0;

					if (angle > rotaryEnd)
					{
						if (SliderHelpers::smallestAngleBetween (angle, rotaryStart)
							 <= SliderHelpers::smallestAngleBetween (angle, rotaryEnd))
							angle = rotaryStart;
						else
							angle = rotaryEnd;
					}
				}

				const double proportion = (angle - rotaryStart) / (rotaryEnd - rotaryStart);

				valueWhenLastDragged = proportionOfLengthToValue (jlimit (0.0, 1.0, proportion));

				lastAngle = angle;
			}
		}
		else
		{
			if (style == LinearBar && e.mouseWasClicked()
				 && valueBox != nullptr && valueBox->isEditable())
				return;

			if (style == IncDecButtons && ! incDecDragged)
			{
				if (e.getDistanceFromDragStart() < 10 || e.mouseWasClicked())
					return;

				incDecDragged = true;
				mouseDragStartX = e.x;
				mouseDragStartY = e.y;
			}

			if ((isVelocityBased == (userKeyOverridesVelocity ? e.mods.testFlags (ModifierKeys::ctrlModifier | ModifierKeys::commandModifier | ModifierKeys::altModifier)
															  : false))
				|| ((maximum - minimum) / sliderRegionSize < interval))
			{
				const int mousePos = (isHorizontal() || style == RotaryHorizontalDrag) ? e.x : e.y;

				double scaledMousePos = (mousePos - sliderRegionStart) / (double) sliderRegionSize;

				if (style == RotaryHorizontalDrag
					|| style == RotaryVerticalDrag
					|| style == IncDecButtons
					|| ((style == LinearHorizontal || style == LinearVertical || style == LinearBar)
						&& ! snapsToMousePos))
				{
					const int mouseDiff = (style == RotaryHorizontalDrag
											 || style == LinearHorizontal
											 || style == LinearBar
											 || (style == IncDecButtons && incDecDragDirectionIsHorizontal()))
											? e.x - mouseDragStartX
											: mouseDragStartY - e.y;

					double newPos = valueToProportionOfLength (valueOnMouseDown)
									   + mouseDiff * (1.0 / pixelsForFullDragExtent);

					valueWhenLastDragged = proportionOfLengthToValue (jlimit (0.0, 1.0, newPos));

					if (style == IncDecButtons)
					{
						incButton->setState (mouseDiff < 0 ? Button::buttonNormal : Button::buttonDown);
						decButton->setState (mouseDiff > 0 ? Button::buttonNormal : Button::buttonDown);
					}
				}
				else
				{
					if (isVertical())
						scaledMousePos = 1.0 - scaledMousePos;

					valueWhenLastDragged = proportionOfLengthToValue (jlimit (0.0, 1.0, scaledMousePos));
				}
			}
			else
			{
				const int mouseDiff = (isHorizontal() || style == RotaryHorizontalDrag
										 || (style == IncDecButtons && incDecDragDirectionIsHorizontal()))
										? e.x - mousePosWhenLastDragged.getX()
										: e.y - mousePosWhenLastDragged.getY();

				const double maxSpeed = jmax (200, sliderRegionSize);
				double speed = jlimit (0.0, maxSpeed, (double) abs (mouseDiff));

				if (speed != 0)
				{
					speed = 0.2 * velocityModeSensitivity
							  * (1.0 + std::sin (double_Pi * (1.5 + jmin (0.5, velocityModeOffset
																			+ jmax (0.0, (double) (speed - velocityModeThreshold))
																				/ maxSpeed))));

					if (mouseDiff < 0)
						speed = -speed;

					if (isVertical() || style == RotaryVerticalDrag
						 || (style == IncDecButtons && ! incDecDragDirectionIsHorizontal()))
						speed = -speed;

					const double currentPos = valueToProportionOfLength (valueWhenLastDragged);

					valueWhenLastDragged = proportionOfLengthToValue (jlimit (0.0, 1.0, currentPos + speed));

					e.source.enableUnboundedMouseMovement (true, false);
					mouseWasHidden = true;
				}
			}
		}

		valueWhenLastDragged = jlimit (minimum, maximum, valueWhenLastDragged);

		if (sliderBeingDragged == 0)
		{
			setValue (snapValue (valueWhenLastDragged, true),
					  ! sendChangeOnlyOnRelease, true);
		}
		else if (sliderBeingDragged == 1)
		{
			setMinValue (snapValue (valueWhenLastDragged, true),
						 ! sendChangeOnlyOnRelease, false, true);

			if (e.mods.isShiftDown())
				setMaxValue (getMinValue() + minMaxDiff, false, false, true);
			else
				minMaxDiff = (double) valueMax.getValue() - (double) valueMin.getValue();
		}
		else
		{
			jassert (sliderBeingDragged == 2);

			setMaxValue (snapValue (valueWhenLastDragged, true),
						 ! sendChangeOnlyOnRelease, false, true);

			if (e.mods.isShiftDown())
				setMinValue (getMaxValue() - minMaxDiff, false, false, true);
			else
				minMaxDiff = (double) valueMax.getValue() - (double) valueMin.getValue();
		}

		mousePosWhenLastDragged = e.getPosition();
	}
}

void Slider::mouseDoubleClick (const MouseEvent&)
{
	if (doubleClickToValue
		 && isEnabled()
		 && style != IncDecButtons
		 && minimum <= doubleClickReturnValue
		 && maximum >= doubleClickReturnValue)
	{
		sendDragStart();
		setValue (doubleClickReturnValue, true, true);
		sendDragEnd();
	}
}

void Slider::mouseWheelMove (const MouseEvent& e, float wheelIncrementX, float wheelIncrementY)
{
	if (scrollWheelEnabled && isEnabled()
		 && style != TwoValueHorizontal
		 && style != TwoValueVertical)
	{
		if (maximum > minimum && ! e.mods.isAnyMouseButtonDown())
		{
			if (valueBox != nullptr)
				valueBox->hideEditor (false);

			const double value = (double) currentValue.getValue();
			const double proportionDelta = (wheelIncrementX != 0 ? -wheelIncrementX : wheelIncrementY) * 0.15f;
			const double currentPos = valueToProportionOfLength (value);
			const double newValue = proportionOfLengthToValue (jlimit (0.0, 1.0, currentPos + proportionDelta));

			double delta = (newValue != value)
							? jmax (std::abs (newValue - value), interval) : 0;

			if (value > newValue)
				delta = -delta;

			sendDragStart();
			setValue (snapValue (value + delta, false), true, true);
			sendDragEnd();
		}
	}
	else
	{
		Component::mouseWheelMove (e, wheelIncrementX, wheelIncrementY);
	}
}

void SliderListener::sliderDragStarted (Slider*)  // (can't write Slider::Listener due to idiotic VC2005 bug)
{
}

void SliderListener::sliderDragEnded (Slider*)
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Slider.cpp ***/


/*** Start of inlined file: juce_TableHeaderComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

class DragOverlayComp   : public Component
{
public:
	DragOverlayComp (const Image& image_)
		: image (image_)
	{
		image.duplicateIfShared();
		image.multiplyAllAlphas (0.8f);
		setAlwaysOnTop (true);
	}

	void paint (Graphics& g)
	{
		g.drawImageAt (image, 0, 0);
	}

private:
	Image image;

	JUCE_DECLARE_NON_COPYABLE (DragOverlayComp);
};

TableHeaderComponent::TableHeaderComponent()
	: columnsChanged (false),
	  columnsResized (false),
	  sortChanged (false),
	  menuActive (true),
	  stretchToFit (false),
	  columnIdBeingResized (0),
	  columnIdBeingDragged (0),
	  columnIdUnderMouse (0),
	  lastDeliberateWidth (0)
{
}

TableHeaderComponent::~TableHeaderComponent()
{
	dragOverlayComp = nullptr;
}

void TableHeaderComponent::setPopupMenuActive (const bool hasMenu)
{
	menuActive = hasMenu;
}

bool TableHeaderComponent::isPopupMenuActive() const			{ return menuActive; }

int TableHeaderComponent::getNumColumns (const bool onlyCountVisibleColumns) const
{
	if (onlyCountVisibleColumns)
	{
		int num = 0;

		for (int i = columns.size(); --i >= 0;)
			if (columns.getUnchecked(i)->isVisible())
				++num;

		return num;
	}
	else
	{
		return columns.size();
	}
}

String TableHeaderComponent::getColumnName (const int columnId) const
{
	const ColumnInfo* const ci = getInfoForId (columnId);
	return ci != nullptr ? ci->name : String::empty;
}

void TableHeaderComponent::setColumnName (const int columnId, const String& newName)
{
	ColumnInfo* const ci = getInfoForId (columnId);

	if (ci != nullptr && ci->name != newName)
	{
		ci->name = newName;
		sendColumnsChanged();
	}
}

void TableHeaderComponent::addColumn (const String& columnName,
									  const int columnId,
									  const int width,
									  const int minimumWidth,
									  const int maximumWidth,
									  const int propertyFlags,
									  const int insertIndex)
{
	// can't have a duplicate or null ID!
	jassert (columnId != 0 && getIndexOfColumnId (columnId, false) < 0);
	jassert (width > 0);

	ColumnInfo* const ci = new ColumnInfo();
	ci->name = columnName;
	ci->id = columnId;
	ci->width = width;
	ci->lastDeliberateWidth = width;
	ci->minimumWidth = minimumWidth;
	ci->maximumWidth = maximumWidth;
	if (ci->maximumWidth < 0)
		ci->maximumWidth = std::numeric_limits<int>::max();
	jassert (ci->maximumWidth >= ci->minimumWidth);
	ci->propertyFlags = propertyFlags;

	columns.insert (insertIndex, ci);
	sendColumnsChanged();
}

void TableHeaderComponent::removeColumn (const int columnIdToRemove)
{
	const int index = getIndexOfColumnId (columnIdToRemove, false);

	if (index >= 0)
	{
		columns.remove (index);
		sortChanged = true;
		sendColumnsChanged();
	}
}

void TableHeaderComponent::removeAllColumns()
{
	if (columns.size() > 0)
	{
		columns.clear();
		sendColumnsChanged();
	}
}

void TableHeaderComponent::moveColumn (const int columnId, int newIndex)
{
	const int currentIndex = getIndexOfColumnId (columnId, false);
	newIndex = visibleIndexToTotalIndex (newIndex);

	if (columns [currentIndex] != 0 && currentIndex != newIndex)
	{
		columns.move (currentIndex, newIndex);
		sendColumnsChanged();
	}
}

int TableHeaderComponent::getColumnWidth (const int columnId) const
{
	const ColumnInfo* const ci = getInfoForId (columnId);
	return ci != nullptr ? ci->width : 0;
}

void TableHeaderComponent::setColumnWidth (const int columnId, const int newWidth)
{
	ColumnInfo* const ci = getInfoForId (columnId);

	if (ci != nullptr && ci->width != newWidth)
	{
		const int numColumns = getNumColumns (true);

		ci->lastDeliberateWidth = ci->width
			= jlimit (ci->minimumWidth, ci->maximumWidth, newWidth);

		if (stretchToFit)
		{
			const int index = getIndexOfColumnId (columnId, true) + 1;

			if (isPositiveAndBelow (index, numColumns))
			{
				const int x = getColumnPosition (index).getX();

				if (lastDeliberateWidth == 0)
					lastDeliberateWidth = getTotalWidth();

				resizeColumnsToFit (visibleIndexToTotalIndex (index), lastDeliberateWidth - x);
			}
		}

		repaint();
		columnsResized = true;
		triggerAsyncUpdate();
	}
}

int TableHeaderComponent::getIndexOfColumnId (const int columnId, const bool onlyCountVisibleColumns) const
{
	int n = 0;

	for (int i = 0; i < columns.size(); ++i)
	{
		if ((! onlyCountVisibleColumns) || columns.getUnchecked(i)->isVisible())
		{
			if (columns.getUnchecked(i)->id == columnId)
				return n;

			++n;
		}
	}

	return -1;
}

int TableHeaderComponent::getColumnIdOfIndex (int index, const bool onlyCountVisibleColumns) const
{
	if (onlyCountVisibleColumns)
		index = visibleIndexToTotalIndex (index);

	const ColumnInfo* const ci = columns [index];
	return (ci != nullptr) ? ci->id : 0;
}

const Rectangle<int> TableHeaderComponent::getColumnPosition (const int index) const
{
	int x = 0, width = 0, n = 0;

	for (int i = 0; i < columns.size(); ++i)
	{
		x += width;

		if (columns.getUnchecked(i)->isVisible())
		{
			width = columns.getUnchecked(i)->width;

			if (n++ == index)
				break;
		}
		else
		{
			width = 0;
		}
	}

	return Rectangle<int> (x, 0, width, getHeight());
}

int TableHeaderComponent::getColumnIdAtX (const int xToFind) const
{
	if (xToFind >= 0)
	{
		int x = 0;

		for (int i = 0; i < columns.size(); ++i)
		{
			const ColumnInfo* const ci = columns.getUnchecked(i);

			if (ci->isVisible())
			{
				x += ci->width;

				if (xToFind < x)
					return ci->id;
			}
		}
	}

	return 0;
}

int TableHeaderComponent::getTotalWidth() const
{
	int w = 0;

	for (int i = columns.size(); --i >= 0;)
		if (columns.getUnchecked(i)->isVisible())
			w += columns.getUnchecked(i)->width;

	return w;
}

void TableHeaderComponent::setStretchToFitActive (const bool shouldStretchToFit)
{
	stretchToFit = shouldStretchToFit;
	lastDeliberateWidth = getTotalWidth();
	resized();
}

bool TableHeaderComponent::isStretchToFitActive() const
{
	return stretchToFit;
}

void TableHeaderComponent::resizeAllColumnsToFit (int targetTotalWidth)
{
	if (stretchToFit && getWidth() > 0
		 && columnIdBeingResized == 0 && columnIdBeingDragged == 0)
	{
		lastDeliberateWidth = targetTotalWidth;
		resizeColumnsToFit (0, targetTotalWidth);
	}
}

void TableHeaderComponent::resizeColumnsToFit (int firstColumnIndex, int targetTotalWidth)
{
	targetTotalWidth = jmax (targetTotalWidth, 0);

	StretchableObjectResizer sor;
	int i;
	for (i = firstColumnIndex; i < columns.size(); ++i)
	{
		ColumnInfo* const ci = columns.getUnchecked(i);

		if (ci->isVisible())
			sor.addItem (ci->lastDeliberateWidth, ci->minimumWidth, ci->maximumWidth);
	}

	sor.resizeToFit (targetTotalWidth);

	int visIndex = 0;
	for (i = firstColumnIndex; i < columns.size(); ++i)
	{
		ColumnInfo* const ci = columns.getUnchecked(i);

		if (ci->isVisible())
		{
			const int newWidth = jlimit (ci->minimumWidth, ci->maximumWidth,
										 (int) std::floor (sor.getItemSize (visIndex++)));

			if (newWidth != ci->width)
			{
				ci->width = newWidth;
				repaint();
				columnsResized = true;
				triggerAsyncUpdate();
			}
		}
	}
}

void TableHeaderComponent::setColumnVisible (const int columnId, const bool shouldBeVisible)
{
	ColumnInfo* const ci = getInfoForId (columnId);

	if (ci != nullptr && shouldBeVisible != ci->isVisible())
	{
		if (shouldBeVisible)
			ci->propertyFlags |= visible;
		else
			ci->propertyFlags &= ~visible;

		sendColumnsChanged();
		resized();
	}
}

bool TableHeaderComponent::isColumnVisible (const int columnId) const
{
	const ColumnInfo* const ci = getInfoForId (columnId);
	return ci != nullptr && ci->isVisible();
}

void TableHeaderComponent::setSortColumnId (const int columnId, const bool sortForwards)
{
	if (getSortColumnId() != columnId || isSortedForwards() != sortForwards)
	{
		for (int i = columns.size(); --i >= 0;)
			columns.getUnchecked(i)->propertyFlags &= ~(sortedForwards | sortedBackwards);

		ColumnInfo* const ci = getInfoForId (columnId);

		if (ci != nullptr)
			ci->propertyFlags |= (sortForwards ? sortedForwards : sortedBackwards);

		reSortTable();
	}
}

int TableHeaderComponent::getSortColumnId() const
{
	for (int i = columns.size(); --i >= 0;)
		if ((columns.getUnchecked(i)->propertyFlags & (sortedForwards | sortedBackwards)) != 0)
			return columns.getUnchecked(i)->id;

	return 0;
}

bool TableHeaderComponent::isSortedForwards() const
{
	for (int i = columns.size(); --i >= 0;)
		if ((columns.getUnchecked(i)->propertyFlags & (sortedForwards | sortedBackwards)) != 0)
			return (columns.getUnchecked(i)->propertyFlags & sortedForwards) != 0;

	return true;
}

void TableHeaderComponent::reSortTable()
{
	sortChanged = true;
	repaint();
	triggerAsyncUpdate();
}

String TableHeaderComponent::toString() const
{
	String s;

	XmlElement doc ("TABLELAYOUT");

	doc.setAttribute ("sortedCol", getSortColumnId());
	doc.setAttribute ("sortForwards", isSortedForwards());

	for (int i = 0; i < columns.size(); ++i)
	{
		const ColumnInfo* const ci = columns.getUnchecked (i);

		XmlElement* const e = doc.createNewChildElement ("COLUMN");
		e->setAttribute ("id", ci->id);
		e->setAttribute ("visible", ci->isVisible());
		e->setAttribute ("width", ci->width);
	}

	return doc.createDocument (String::empty, true, false);
}

void TableHeaderComponent::restoreFromString (const String& storedVersion)
{
	ScopedPointer <XmlElement> storedXml (XmlDocument::parse (storedVersion));
	int index = 0;

	if (storedXml != nullptr && storedXml->hasTagName ("TABLELAYOUT"))
	{
		forEachXmlChildElement (*storedXml, col)
		{
			const int tabId = col->getIntAttribute ("id");

			ColumnInfo* const ci = getInfoForId (tabId);

			if (ci != nullptr)
			{
				columns.move (columns.indexOf (ci), index);
				ci->width = col->getIntAttribute ("width");
				setColumnVisible (tabId, col->getBoolAttribute ("visible"));
			}

			++index;
		}

		columnsResized = true;
		sendColumnsChanged();

		setSortColumnId (storedXml->getIntAttribute ("sortedCol"),
						 storedXml->getBoolAttribute ("sortForwards", true));
	}
}

void TableHeaderComponent::addListener (Listener* const newListener)
{
	listeners.addIfNotAlreadyThere (newListener);
}

void TableHeaderComponent::removeListener (Listener* const listenerToRemove)
{
	listeners.removeValue (listenerToRemove);
}

void TableHeaderComponent::columnClicked (int columnId, const ModifierKeys& mods)
{
	const ColumnInfo* const ci = getInfoForId (columnId);

	if (ci != nullptr && (ci->propertyFlags & sortable) != 0 && ! mods.isPopupMenu())
		setSortColumnId (columnId, (ci->propertyFlags & sortedForwards) == 0);
}

void TableHeaderComponent::addMenuItems (PopupMenu& menu, const int /*columnIdClicked*/)
{
	for (int i = 0; i < columns.size(); ++i)
	{
		const ColumnInfo* const ci = columns.getUnchecked(i);

		if ((ci->propertyFlags & appearsOnColumnMenu) != 0)
			menu.addItem (ci->id, ci->name,
						  (ci->propertyFlags & (sortedForwards | sortedBackwards)) == 0,
						  isColumnVisible (ci->id));
	}
}

void TableHeaderComponent::reactToMenuItem (const int menuReturnId, const int /*columnIdClicked*/)
{
	if (getIndexOfColumnId (menuReturnId, false) >= 0)
		setColumnVisible (menuReturnId, ! isColumnVisible (menuReturnId));
}

void TableHeaderComponent::paint (Graphics& g)
{
	LookAndFeel& lf = getLookAndFeel();

	lf.drawTableHeaderBackground (g, *this);

	const Rectangle<int> clip (g.getClipBounds());

	int x = 0;
	for (int i = 0; i < columns.size(); ++i)
	{
		const ColumnInfo* const ci = columns.getUnchecked(i);

		if (ci->isVisible())
		{
			if (x + ci->width > clip.getX()
				 && (ci->id != columnIdBeingDragged
					  || dragOverlayComp == nullptr
					  || ! dragOverlayComp->isVisible()))
			{
				Graphics::ScopedSaveState ss (g);

				g.setOrigin (x, 0);
				g.reduceClipRegion (0, 0, ci->width, getHeight());

				lf.drawTableHeaderColumn (g, ci->name, ci->id, ci->width, getHeight(),
										  ci->id == columnIdUnderMouse,
										  ci->id == columnIdUnderMouse && isMouseButtonDown(),
										  ci->propertyFlags);
			}

			x += ci->width;

			if (x >= clip.getRight())
				break;
		}
	}
}

void TableHeaderComponent::resized()
{
}

void TableHeaderComponent::mouseMove (const MouseEvent& e)
{
	updateColumnUnderMouse (e.x, e.y);
}

void TableHeaderComponent::mouseEnter (const MouseEvent& e)
{
	updateColumnUnderMouse (e.x, e.y);
}

void TableHeaderComponent::mouseExit (const MouseEvent& e)
{
	updateColumnUnderMouse (e.x, e.y);
}

void TableHeaderComponent::mouseDown (const MouseEvent& e)
{
	repaint();
	columnIdBeingResized = 0;
	columnIdBeingDragged = 0;

	if (columnIdUnderMouse != 0)
	{
		draggingColumnOffset = e.x - getColumnPosition (getIndexOfColumnId (columnIdUnderMouse, true)).getX();

		if (e.mods.isPopupMenu())
			columnClicked (columnIdUnderMouse, e.mods);
	}

	if (menuActive && e.mods.isPopupMenu())
		showColumnChooserMenu (columnIdUnderMouse);
}

void TableHeaderComponent::mouseDrag (const MouseEvent& e)
{
	if (columnIdBeingResized == 0
		 && columnIdBeingDragged == 0
		 && ! (e.mouseWasClicked() || e.mods.isPopupMenu()))
	{
		dragOverlayComp = nullptr;

		columnIdBeingResized = getResizeDraggerAt (e.getMouseDownX());

		if (columnIdBeingResized != 0)
		{
			const ColumnInfo* const ci = getInfoForId (columnIdBeingResized);
			initialColumnWidth = ci->width;
		}
		else
		{
			beginDrag (e);
		}
	}

	if (columnIdBeingResized != 0)
	{
		const ColumnInfo* const ci = getInfoForId (columnIdBeingResized);

		if (ci != nullptr)
		{
			int w = jlimit (ci->minimumWidth, ci->maximumWidth,
							initialColumnWidth + e.getDistanceFromDragStartX());

			if (stretchToFit)
			{
				// prevent us dragging a column too far right if we're in stretch-to-fit mode
				int minWidthOnRight = 0;
				for (int i = getIndexOfColumnId (columnIdBeingResized, false) + 1; i < columns.size(); ++i)
					if (columns.getUnchecked (i)->isVisible())
						minWidthOnRight += columns.getUnchecked (i)->minimumWidth;

				const Rectangle<int> currentPos (getColumnPosition (getIndexOfColumnId (columnIdBeingResized, true)));
				w = jmax (ci->minimumWidth, jmin (w, getWidth() - minWidthOnRight - currentPos.getX()));
			}

			setColumnWidth (columnIdBeingResized, w);
		}
	}
	else if (columnIdBeingDragged != 0)
	{
		if (e.y >= -50 && e.y < getHeight() + 50)
		{
			if (dragOverlayComp != nullptr)
			{
				dragOverlayComp->setVisible (true);
				dragOverlayComp->setBounds (jlimit (0,
													jmax (0, getTotalWidth() - dragOverlayComp->getWidth()),
													e.x - draggingColumnOffset),
											0,
											dragOverlayComp->getWidth(),
											getHeight());

				for (int i = columns.size(); --i >= 0;)
				{
					const int currentIndex = getIndexOfColumnId (columnIdBeingDragged, true);
					int newIndex = currentIndex;

					if (newIndex > 0)
					{
						// if the previous column isn't draggable, we can't move our column
						// past it, because that'd change the undraggable column's position..
						const ColumnInfo* const previous = columns.getUnchecked (newIndex - 1);

						if ((previous->propertyFlags & draggable) != 0)
						{
							const int leftOfPrevious = getColumnPosition (newIndex - 1).getX();
							const int rightOfCurrent = getColumnPosition (newIndex).getRight();

							if (abs (dragOverlayComp->getX() - leftOfPrevious)
								< abs (dragOverlayComp->getRight() - rightOfCurrent))
							{
								--newIndex;
							}
						}
					}

					if (newIndex < columns.size() - 1)
					{
						// if the next column isn't draggable, we can't move our column
						// past it, because that'd change the undraggable column's position..
						const ColumnInfo* const nextCol = columns.getUnchecked (newIndex + 1);

						if ((nextCol->propertyFlags & draggable) != 0)
						{
							const int leftOfCurrent = getColumnPosition (newIndex).getX();
							const int rightOfNext = getColumnPosition (newIndex + 1).getRight();

							if (abs (dragOverlayComp->getX() - leftOfCurrent)
								> abs (dragOverlayComp->getRight() - rightOfNext))
							{
								++newIndex;
							}
						}
					}

					if (newIndex != currentIndex)
						moveColumn (columnIdBeingDragged, newIndex);
					else
						break;
				}
			}
		}
		else
		{
			endDrag (draggingColumnOriginalIndex);
		}
	}
}

void TableHeaderComponent::beginDrag (const MouseEvent& e)
{
	if (columnIdBeingDragged == 0)
	{
		columnIdBeingDragged = getColumnIdAtX (e.getMouseDownX());

		const ColumnInfo* const ci = getInfoForId (columnIdBeingDragged);

		if (ci == nullptr || (ci->propertyFlags & draggable) == 0)
		{
			columnIdBeingDragged = 0;
		}
		else
		{
			draggingColumnOriginalIndex = getIndexOfColumnId (columnIdBeingDragged, true);

			const Rectangle<int> columnRect (getColumnPosition (draggingColumnOriginalIndex));

			const int temp = columnIdBeingDragged;
			columnIdBeingDragged = 0;

			addAndMakeVisible (dragOverlayComp = new DragOverlayComp (createComponentSnapshot (columnRect, false)));
			columnIdBeingDragged = temp;

			dragOverlayComp->setBounds (columnRect);

			for (int i = listeners.size(); --i >= 0;)
			{
				listeners.getUnchecked(i)->tableColumnDraggingChanged (this, columnIdBeingDragged);
				i = jmin (i, listeners.size() - 1);
			}
		}
	}
}

void TableHeaderComponent::endDrag (const int finalIndex)
{
	if (columnIdBeingDragged != 0)
	{
		moveColumn (columnIdBeingDragged, finalIndex);

		columnIdBeingDragged = 0;
		repaint();

		for (int i = listeners.size(); --i >= 0;)
		{
			listeners.getUnchecked(i)->tableColumnDraggingChanged (this, 0);
			i = jmin (i, listeners.size() - 1);
		}
	}
}

void TableHeaderComponent::mouseUp (const MouseEvent& e)
{
	mouseDrag (e);

	for (int i = columns.size(); --i >= 0;)
		if (columns.getUnchecked (i)->isVisible())
			columns.getUnchecked (i)->lastDeliberateWidth = columns.getUnchecked (i)->width;

	columnIdBeingResized = 0;
	repaint();

	endDrag (getIndexOfColumnId (columnIdBeingDragged, true));

	updateColumnUnderMouse (e.x, e.y);

	if (columnIdUnderMouse != 0 && e.mouseWasClicked() && ! e.mods.isPopupMenu())
		columnClicked (columnIdUnderMouse, e.mods);

	dragOverlayComp = nullptr;
}

const MouseCursor TableHeaderComponent::getMouseCursor()
{
	if (columnIdBeingResized != 0 || (getResizeDraggerAt (getMouseXYRelative().getX()) != 0 && ! isMouseButtonDown()))
		return MouseCursor (MouseCursor::LeftRightResizeCursor);

	return Component::getMouseCursor();
}

bool TableHeaderComponent::ColumnInfo::isVisible() const
{
	return (propertyFlags & TableHeaderComponent::visible) != 0;
}

TableHeaderComponent::ColumnInfo* TableHeaderComponent::getInfoForId (const int id) const
{
	for (int i = columns.size(); --i >= 0;)
		if (columns.getUnchecked(i)->id == id)
			return columns.getUnchecked(i);

	return nullptr;
}

int TableHeaderComponent::visibleIndexToTotalIndex (const int visibleIndex) const
{
	int n = 0;

	for (int i = 0; i < columns.size(); ++i)
	{
		if (columns.getUnchecked(i)->isVisible())
		{
			if (n == visibleIndex)
				return i;

			++n;
		}
	}

	return -1;
}

void TableHeaderComponent::sendColumnsChanged()
{
	if (stretchToFit && lastDeliberateWidth > 0)
		resizeAllColumnsToFit (lastDeliberateWidth);

	repaint();
	columnsChanged = true;
	triggerAsyncUpdate();
}

void TableHeaderComponent::handleAsyncUpdate()
{
	const bool changed = columnsChanged || sortChanged;
	const bool sized = columnsResized || changed;
	const bool sorted = sortChanged;
	columnsChanged = false;
	columnsResized = false;
	sortChanged = false;

	if (sorted)
	{
		for (int i = listeners.size(); --i >= 0;)
		{
			listeners.getUnchecked(i)->tableSortOrderChanged (this);
			i = jmin (i, listeners.size() - 1);
		}
	}

	if (changed)
	{
		for (int i = listeners.size(); --i >= 0;)
		{
			listeners.getUnchecked(i)->tableColumnsChanged (this);
			i = jmin (i, listeners.size() - 1);
		}
	}

	if (sized)
	{
		for (int i = listeners.size(); --i >= 0;)
		{
			listeners.getUnchecked(i)->tableColumnsResized (this);
			i = jmin (i, listeners.size() - 1);
		}
	}
}

int TableHeaderComponent::getResizeDraggerAt (const int mouseX) const
{
	if (isPositiveAndBelow (mouseX, getWidth()))
	{
		const int draggableDistance = 3;
		int x = 0;

		for (int i = 0; i < columns.size(); ++i)
		{
			const ColumnInfo* const ci = columns.getUnchecked(i);

			if (ci->isVisible())
			{
				if (abs (mouseX - (x + ci->width)) <= draggableDistance
					 && (ci->propertyFlags & resizable) != 0)
					return ci->id;

				x += ci->width;
			}
		}
	}

	return 0;
}

void TableHeaderComponent::updateColumnUnderMouse (int x, int y)
{
	const int newCol = (reallyContains (Point<int> (x, y), true) && getResizeDraggerAt (x) == 0)
							? getColumnIdAtX (x) : 0;

	if (newCol != columnIdUnderMouse)
	{
		columnIdUnderMouse = newCol;
		repaint();
	}
}

static void tableHeaderMenuCallback (int result, TableHeaderComponent* tableHeader, int columnIdClicked)
{
	if (tableHeader != nullptr && result != 0)
		tableHeader->reactToMenuItem (result, columnIdClicked);
}

void TableHeaderComponent::showColumnChooserMenu (const int columnIdClicked)
{
	PopupMenu m;
	addMenuItems (m, columnIdClicked);

	if (m.getNumItems() > 0)
	{
		m.setLookAndFeel (&getLookAndFeel());

		m.showMenuAsync (PopupMenu::Options(),
						 ModalCallbackFunction::forComponent (tableHeaderMenuCallback, this, columnIdClicked));
	}
}

void TableHeaderComponent::Listener::tableColumnDraggingChanged (TableHeaderComponent*, int)
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TableHeaderComponent.cpp ***/


/*** Start of inlined file: juce_TableListBox.cpp ***/
BEGIN_JUCE_NAMESPACE

class TableListRowComp   : public Component,
						   public TooltipClient
{
public:
	TableListRowComp (TableListBox& owner_)
		: owner (owner_), row (-1), isSelected (false)
	{
	}

	void paint (Graphics& g)
	{
		TableListBoxModel* const model = owner.getModel();

		if (model != nullptr)
		{
			model->paintRowBackground (g, row, getWidth(), getHeight(), isSelected);

			const TableHeaderComponent& header = owner.getHeader();
			const int numColumns = header.getNumColumns (true);

			for (int i = 0; i < numColumns; ++i)
			{
				if (columnComponents[i] == nullptr)
				{
					const int columnId = header.getColumnIdOfIndex (i, true);
					const Rectangle<int> columnRect (header.getColumnPosition(i).withHeight (getHeight()));

					Graphics::ScopedSaveState ss (g);

					g.reduceClipRegion (columnRect);
					g.setOrigin (columnRect.getX(), 0);
					model->paintCell (g, row, columnId, columnRect.getWidth(), columnRect.getHeight(), isSelected);
				}
			}
		}
	}

	void update (const int newRow, const bool isNowSelected)
	{
		jassert (newRow >= 0);

		if (newRow != row || isNowSelected != isSelected)
		{
			row = newRow;
			isSelected = isNowSelected;
			repaint();
		}

		TableListBoxModel* const model = owner.getModel();

		if (model != nullptr && row < owner.getNumRows())
		{
			const Identifier columnProperty ("_tableColumnId");
			const int numColumns = owner.getHeader().getNumColumns (true);

			for (int i = 0; i < numColumns; ++i)
			{
				const int columnId = owner.getHeader().getColumnIdOfIndex (i, true);
				Component* comp = columnComponents[i];

				if (comp != nullptr && columnId != (int) comp->getProperties() [columnProperty])
				{
					columnComponents.set (i, nullptr);
					comp = nullptr;
				}

				comp = model->refreshComponentForCell (row, columnId, isSelected, comp);
				columnComponents.set (i, comp, false);

				if (comp != nullptr)
				{
					comp->getProperties().set (columnProperty, columnId);

					addAndMakeVisible (comp);
					resizeCustomComp (i);
				}
			}

			columnComponents.removeRange (numColumns, columnComponents.size());
		}
		else
		{
			columnComponents.clear();
		}
	}

	void resized()
	{
		for (int i = columnComponents.size(); --i >= 0;)
			resizeCustomComp (i);
	}

	void resizeCustomComp (const int index)
	{
		Component* const c = columnComponents.getUnchecked (index);

		if (c != nullptr)
			c->setBounds (owner.getHeader().getColumnPosition (index)
							.withY (0).withHeight (getHeight()));
	}

	void mouseDown (const MouseEvent& e)
	{
		isDragging = false;
		selectRowOnMouseUp = false;

		if (isEnabled())
		{
			if (! isSelected)
			{
				owner.selectRowsBasedOnModifierKeys (row, e.mods, false);

				const int columnId = owner.getHeader().getColumnIdAtX (e.x);

				if (columnId != 0 && owner.getModel() != nullptr)
					owner.getModel()->cellClicked (row, columnId, e);
			}
			else
			{
				selectRowOnMouseUp = true;
			}
		}
	}

	void mouseDrag (const MouseEvent& e)
	{
		if (isEnabled() && owner.getModel() != nullptr && ! (e.mouseWasClicked() || isDragging))
		{
			const SparseSet<int> selectedRows (owner.getSelectedRows());

			if (selectedRows.size() > 0)
			{
				const var dragDescription (owner.getModel()->getDragSourceDescription (selectedRows));

				if (! (dragDescription.isVoid() || (dragDescription.isString() && dragDescription.toString().isEmpty())))
				{
					isDragging = true;
					owner.startDragAndDrop (e, dragDescription);
				}
			}
		}
	}

	void mouseUp (const MouseEvent& e)
	{
		if (selectRowOnMouseUp && e.mouseWasClicked() && isEnabled())
		{
			owner.selectRowsBasedOnModifierKeys (row, e.mods, true);

			const int columnId = owner.getHeader().getColumnIdAtX (e.x);

			if (columnId != 0 && owner.getModel() != nullptr)
				owner.getModel()->cellClicked (row, columnId, e);
		}
	}

	void mouseDoubleClick (const MouseEvent& e)
	{
		const int columnId = owner.getHeader().getColumnIdAtX (e.x);

		if (columnId != 0 && owner.getModel() != nullptr)
			owner.getModel()->cellDoubleClicked (row, columnId, e);
	}

	const String getTooltip()
	{
		const int columnId = owner.getHeader().getColumnIdAtX (getMouseXYRelative().getX());

		if (columnId != 0 && owner.getModel() != nullptr)
			return owner.getModel()->getCellTooltip (row, columnId);

		return String::empty;
	}

	Component* findChildComponentForColumn (const int columnId) const
	{
		return columnComponents [owner.getHeader().getIndexOfColumnId (columnId, true)];
	}

private:
	TableListBox& owner;
	OwnedArray<Component> columnComponents;
	int row;
	bool isSelected, isDragging, selectRowOnMouseUp;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (TableListRowComp);
};

class TableListBoxHeader  : public TableHeaderComponent
{
public:
	TableListBoxHeader (TableListBox& owner_)
		: owner (owner_)
	{
	}

	void addMenuItems (PopupMenu& menu, int columnIdClicked)
	{
		if (owner.isAutoSizeMenuOptionShown())
		{
			menu.addItem (autoSizeColumnId, TRANS("Auto-size this column"), columnIdClicked != 0);
			menu.addItem (autoSizeAllId, TRANS("Auto-size all columns"), owner.getHeader().getNumColumns (true) > 0);
			menu.addSeparator();
		}

		TableHeaderComponent::addMenuItems (menu, columnIdClicked);
	}

	void reactToMenuItem (int menuReturnId, int columnIdClicked)
	{
		switch (menuReturnId)
		{
			case autoSizeColumnId:	  owner.autoSizeColumn (columnIdClicked); break;
			case autoSizeAllId:	 owner.autoSizeAllColumns(); break;
			default:			TableHeaderComponent::reactToMenuItem (menuReturnId, columnIdClicked); break;
		}
	}

private:
	TableListBox& owner;

	enum { autoSizeColumnId = 0xf836743, autoSizeAllId = 0xf836744 };

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (TableListBoxHeader);
};

TableListBox::TableListBox (const String& name, TableListBoxModel* const model_)
	: ListBox (name, nullptr),
	  header (nullptr),
	  model (model_),
	  autoSizeOptionsShown (true)
{
	ListBox::model = this;

	setHeader (new TableListBoxHeader (*this));
}

TableListBox::~TableListBox()
{
	header = nullptr;
}

void TableListBox::setModel (TableListBoxModel* const newModel)
{
	if (model != newModel)
	{
		model = newModel;
		updateContent();
	}
}

void TableListBox::setHeader (TableHeaderComponent* newHeader)
{
	jassert (newHeader != nullptr); // you need to supply a real header for a table!

	Rectangle<int> newBounds (0, 0, 100, 28);
	if (header != nullptr)
		newBounds = header->getBounds();

	header = newHeader;
	header->setBounds (newBounds);

	setHeaderComponent (header);

	header->addListener (this);
}

int TableListBox::getHeaderHeight() const
{
	return header->getHeight();
}

void TableListBox::setHeaderHeight (const int newHeight)
{
	header->setSize (header->getWidth(), newHeight);
	resized();
}

void TableListBox::autoSizeColumn (const int columnId)
{
	const int width = model != nullptr ? model->getColumnAutoSizeWidth (columnId) : 0;

	if (width > 0)
		header->setColumnWidth (columnId, width);
}

void TableListBox::autoSizeAllColumns()
{
	for (int i = 0; i < header->getNumColumns (true); ++i)
		autoSizeColumn (header->getColumnIdOfIndex (i, true));
}

void TableListBox::setAutoSizeMenuOptionShown (const bool shouldBeShown)
{
	autoSizeOptionsShown = shouldBeShown;
}

bool TableListBox::isAutoSizeMenuOptionShown() const
{
	return autoSizeOptionsShown;
}

const Rectangle<int> TableListBox::getCellPosition (const int columnId, const int rowNumber,
													const bool relativeToComponentTopLeft) const
{
	Rectangle<int> headerCell (header->getColumnPosition (header->getIndexOfColumnId (columnId, true)));

	if (relativeToComponentTopLeft)
		headerCell.translate (header->getX(), 0);

	return getRowPosition (rowNumber, relativeToComponentTopLeft)
			.withX (headerCell.getX())
			.withWidth (headerCell.getWidth());
}

Component* TableListBox::getCellComponent (int columnId, int rowNumber) const
{
	TableListRowComp* const rowComp = dynamic_cast <TableListRowComp*> (getComponentForRowNumber (rowNumber));
	return rowComp != nullptr ? rowComp->findChildComponentForColumn (columnId) : 0;
}

void TableListBox::scrollToEnsureColumnIsOnscreen (const int columnId)
{
	ScrollBar* const scrollbar = getHorizontalScrollBar();

	if (scrollbar != nullptr)
	{
		const Rectangle<int> pos (header->getColumnPosition (header->getIndexOfColumnId (columnId, true)));

		double x = scrollbar->getCurrentRangeStart();
		const double w = scrollbar->getCurrentRangeSize();

		if (pos.getX() < x)
			x = pos.getX();
		else if (pos.getRight() > x + w)
			x += jmax (0.0, pos.getRight() - (x + w));

		scrollbar->setCurrentRangeStart (x);
	}
}

int TableListBox::getNumRows()
{
	return model != nullptr ? model->getNumRows() : 0;
}

void TableListBox::paintListBoxItem (int, Graphics&, int, int, bool)
{
}

Component* TableListBox::refreshComponentForRow (int rowNumber, bool isRowSelected_, Component* existingComponentToUpdate)
{
	if (existingComponentToUpdate == nullptr)
		existingComponentToUpdate = new TableListRowComp (*this);

	static_cast <TableListRowComp*> (existingComponentToUpdate)->update (rowNumber, isRowSelected_);

	return existingComponentToUpdate;
}

void TableListBox::selectedRowsChanged (int row)
{
	if (model != nullptr)
		model->selectedRowsChanged (row);
}

void TableListBox::deleteKeyPressed (int row)
{
	if (model != nullptr)
		model->deleteKeyPressed (row);
}

void TableListBox::returnKeyPressed (int row)
{
	if (model != nullptr)
		model->returnKeyPressed (row);
}

void TableListBox::backgroundClicked()
{
	if (model != nullptr)
		model->backgroundClicked();
}

void TableListBox::listWasScrolled()
{
	if (model != nullptr)
		model->listWasScrolled();
}

void TableListBox::tableColumnsChanged (TableHeaderComponent*)
{
	setMinimumContentWidth (header->getTotalWidth());
	repaint();
	updateColumnComponents();
}

void TableListBox::tableColumnsResized (TableHeaderComponent*)
{
	setMinimumContentWidth (header->getTotalWidth());
	repaint();
	updateColumnComponents();
}

void TableListBox::tableSortOrderChanged (TableHeaderComponent*)
{
	if (model != nullptr)
		model->sortOrderChanged (header->getSortColumnId(),
								 header->isSortedForwards());
}

void TableListBox::tableColumnDraggingChanged (TableHeaderComponent*, int columnIdNowBeingDragged_)
{
	columnIdNowBeingDragged = columnIdNowBeingDragged_;
	repaint();
}

void TableListBox::resized()
{
	ListBox::resized();

	header->resizeAllColumnsToFit (getVisibleContentWidth());
	setMinimumContentWidth (header->getTotalWidth());
}

void TableListBox::updateColumnComponents() const
{
	const int firstRow = getRowContainingPosition (0, 0);

	for (int i = firstRow + getNumRowsOnScreen() + 2; --i >= firstRow;)
	{
		TableListRowComp* const rowComp = dynamic_cast <TableListRowComp*> (getComponentForRowNumber (i));

		if (rowComp != nullptr)
			rowComp->resized();
	}
}

void TableListBoxModel::cellClicked (int, int, const MouseEvent&)	   {}
void TableListBoxModel::cellDoubleClicked (int, int, const MouseEvent&) {}
void TableListBoxModel::backgroundClicked()				 {}
void TableListBoxModel::sortOrderChanged (int, const bool)		  {}
int TableListBoxModel::getColumnAutoSizeWidth (int)			 { return 0; }
void TableListBoxModel::selectedRowsChanged (int)			   {}
void TableListBoxModel::deleteKeyPressed (int)			  {}
void TableListBoxModel::returnKeyPressed (int)			  {}
void TableListBoxModel::listWasScrolled()				   {}

const String TableListBoxModel::getCellTooltip (int /*rowNumber*/, int /*columnId*/)	{ return String::empty; }
const var TableListBoxModel::getDragSourceDescription (const SparseSet<int>&)	   { return var::null; }

Component* TableListBoxModel::refreshComponentForCell (int, int, bool, Component* existingComponentToUpdate)
{
	(void) existingComponentToUpdate;
	jassert (existingComponentToUpdate == nullptr); // indicates a failure in the code the recycles the components
	return nullptr;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TableListBox.cpp ***/


/*** Start of inlined file: juce_TextEditor.cpp ***/
BEGIN_JUCE_NAMESPACE

// a word or space that can't be broken down any further
struct TextAtom
{

	String atomText;
	float width;
	int numChars;

	bool isWhitespace() const	   { return CharacterFunctions::isWhitespace (atomText[0]); }
	bool isNewLine() const	  { return atomText[0] == '\r' || atomText[0] == '\n'; }

	String getText (const juce_wchar passwordCharacter) const
	{
		if (passwordCharacter == 0)
			return atomText;
		else
			return String::repeatedString (String::charToString (passwordCharacter),
										   atomText.length());
	}

	String getTrimmedText (const juce_wchar passwordCharacter) const
	{
		if (passwordCharacter == 0)
			return atomText.substring (0, numChars);
		else if (isNewLine())
			return String::empty;
		else
			return String::repeatedString (String::charToString (passwordCharacter), numChars);
	}
};

// a run of text with a single font and colour
class TextEditor::UniformTextSection
{
public:

	UniformTextSection (const String& text,
						const Font& font_,
						const Colour& colour_,
						const juce_wchar passwordCharacter)
	  : font (font_),
		colour (colour_)
	{
		initialiseAtoms (text, passwordCharacter);
	}

	UniformTextSection (const UniformTextSection& other)
	  : font (other.font),
		colour (other.colour)
	{
		atoms.ensureStorageAllocated (other.atoms.size());

		for (int i = 0; i < other.atoms.size(); ++i)
			atoms.add (new TextAtom (*other.atoms.getUnchecked(i)));
	}

	~UniformTextSection()
	{
		// (no need to delete the atoms, as they're explicitly deleted by the caller)
	}

	void clear()
	{
		for (int i = atoms.size(); --i >= 0;)
			delete getAtom(i);

		atoms.clear();
	}

	int getNumAtoms() const
	{
		return atoms.size();
	}

	TextAtom* getAtom (const int index) const noexcept
	{
		return atoms.getUnchecked (index);
	}

	void append (const UniformTextSection& other, const juce_wchar passwordCharacter)
	{
		if (other.atoms.size() > 0)
		{
			TextAtom* const lastAtom = atoms.getLast();
			int i = 0;

			if (lastAtom != nullptr)
			{
				if (! CharacterFunctions::isWhitespace (lastAtom->atomText.getLastCharacter()))
				{
					TextAtom* const first = other.getAtom(0);

					if (! CharacterFunctions::isWhitespace (first->atomText[0]))
					{
						lastAtom->atomText += first->atomText;
						lastAtom->numChars = (uint16) (lastAtom->numChars + first->numChars);
						lastAtom->width = font.getStringWidthFloat (lastAtom->getText (passwordCharacter));
						delete first;
						++i;
					}
				}
			}

			atoms.ensureStorageAllocated (atoms.size() + other.atoms.size() - i);

			while (i < other.atoms.size())
			{
				atoms.add (other.getAtom(i));
				++i;
			}
		}
	}

	UniformTextSection* split (const int indexToBreakAt,
							   const juce_wchar passwordCharacter)
	{
		UniformTextSection* const section2 = new UniformTextSection (String::empty,
																	 font, colour,
																	 passwordCharacter);
		int index = 0;

		for (int i = 0; i < atoms.size(); ++i)
		{
			TextAtom* const atom = getAtom(i);

			const int nextIndex = index + atom->numChars;

			if (index == indexToBreakAt)
			{
				int j;
				for (j = i; j < atoms.size(); ++j)
					section2->atoms.add (getAtom (j));

				for (j = atoms.size(); --j >= i;)
					atoms.remove (j);

				break;
			}
			else if (indexToBreakAt >= index && indexToBreakAt < nextIndex)
			{
				TextAtom* const secondAtom = new TextAtom();

				secondAtom->atomText = atom->atomText.substring (indexToBreakAt - index);
				secondAtom->width = font.getStringWidthFloat (secondAtom->getText (passwordCharacter));
				secondAtom->numChars = (uint16) secondAtom->atomText.length();

				section2->atoms.add (secondAtom);

				atom->atomText = atom->atomText.substring (0, indexToBreakAt - index);
				atom->width = font.getStringWidthFloat (atom->getText (passwordCharacter));
				atom->numChars = (uint16) (indexToBreakAt - index);

				int j;
				for (j = i + 1; j < atoms.size(); ++j)
					section2->atoms.add (getAtom (j));

				for (j = atoms.size(); --j > i;)
					atoms.remove (j);

				break;
			}

			index = nextIndex;
		}

		return section2;
	}

	void appendAllText (MemoryOutputStream& mo) const
	{
		for (int i = 0; i < atoms.size(); ++i)
			mo << getAtom(i)->atomText;
	}

	void appendSubstring (MemoryOutputStream& mo, const Range<int>& range) const
	{
		int index = 0;
		for (int i = 0; i < atoms.size(); ++i)
		{
			const TextAtom* const atom = getAtom (i);
			const int nextIndex = index + atom->numChars;

			if (range.getStart() < nextIndex)
			{
				if (range.getEnd() <= index)
					break;

				const Range<int> r ((range - index).getIntersectionWith (Range<int> (0, (int) atom->numChars)));

				if (! r.isEmpty())
					mo << atom->atomText.substring (r.getStart(), r.getEnd());
			}

			index = nextIndex;
		}
	}

	int getTotalLength() const
	{
		int total = 0;

		for (int i = atoms.size(); --i >= 0;)
			total += getAtom(i)->numChars;

		return total;
	}

	void setFont (const Font& newFont,
				  const juce_wchar passwordCharacter)
	{
		if (font != newFont)
		{
			font = newFont;

			for (int i = atoms.size(); --i >= 0;)
			{
				TextAtom* const atom = atoms.getUnchecked(i);
				atom->width = newFont.getStringWidthFloat (atom->getText (passwordCharacter));
			}
		}
	}

	Font font;
	Colour colour;

private:
	Array <TextAtom*> atoms;

	void initialiseAtoms (const String& textToParse,
						  const juce_wchar passwordCharacter)
	{
		String::CharPointerType text (textToParse.getCharPointer());

		while (! text.isEmpty())
		{
			int numChars = 0;
			String::CharPointerType start (text);

			// create a whitespace atom unless it starts with non-ws
			if (text.isWhitespace() && *text != '\r' && *text != '\n')
			{
				do
				{
					++text;
					++numChars;
				}
				while (text.isWhitespace() && *text != '\r' && *text != '\n');
			}
			else
			{
				if (*text == '\r')
				{
					++text;
					++numChars;

					if (*text == '\n')
					{
						++start;
						++text;
					}
				}
				else if (*text == '\n')
				{
					++text;
					++numChars;
				}
				else
				{
					while (! (text.isEmpty() || text.isWhitespace()))
					{
						++text;
						++numChars;
					}
				}
			}

			TextAtom* const atom = new TextAtom();
			atom->atomText = String (start, numChars);

			atom->width = font.getStringWidthFloat (atom->getText (passwordCharacter));
			atom->numChars = (uint16) numChars;

			atoms.add (atom);
		}
	}

	UniformTextSection& operator= (const UniformTextSection& other);
	JUCE_LEAK_DETECTOR (UniformTextSection);
};

class TextEditor::Iterator
{
public:

	Iterator (const Array <UniformTextSection*>& sections_,
			  const float wordWrapWidth_,
			  const juce_wchar passwordCharacter_)
	  : indexInText (0),
		lineY (0),
		lineHeight (0),
		maxDescent (0),
		atomX (0),
		atomRight (0),
		atom (0),
		currentSection (nullptr),
		sections (sections_),
		sectionIndex (0),
		atomIndex (0),
		wordWrapWidth (wordWrapWidth_),
		passwordCharacter (passwordCharacter_)
	{
		jassert (wordWrapWidth_ > 0);

		if (sections.size() > 0)
		{
			currentSection = sections.getUnchecked (sectionIndex);

			if (currentSection != nullptr)
				beginNewLine();
		}
	}

	Iterator (const Iterator& other)
	  : indexInText (other.indexInText),
		lineY (other.lineY),
		lineHeight (other.lineHeight),
		maxDescent (other.maxDescent),
		atomX (other.atomX),
		atomRight (other.atomRight),
		atom (other.atom),
		currentSection (other.currentSection),
		sections (other.sections),
		sectionIndex (other.sectionIndex),
		atomIndex (other.atomIndex),
		wordWrapWidth (other.wordWrapWidth),
		passwordCharacter (other.passwordCharacter),
		tempAtom (other.tempAtom)
	{
	}

	bool next()
	{
		if (atom == &tempAtom)
		{
			const int numRemaining = tempAtom.atomText.length() - tempAtom.numChars;

			if (numRemaining > 0)
			{
				tempAtom.atomText = tempAtom.atomText.substring (tempAtom.numChars);

				atomX = 0;

				if (tempAtom.numChars > 0)
					lineY += lineHeight;

				indexInText += tempAtom.numChars;

				GlyphArrangement g;
				g.addLineOfText (currentSection->font, atom->getText (passwordCharacter), 0.0f, 0.0f);

				int split;
				for (split = 0; split < g.getNumGlyphs(); ++split)
					if (shouldWrap (g.getGlyph (split).getRight()))
						break;

				if (split > 0 && split <= numRemaining)
				{
					tempAtom.numChars = (uint16) split;
					tempAtom.width = g.getGlyph (split - 1).getRight();
					atomRight = atomX + tempAtom.width;
					return true;
				}
			}
		}

		bool forceNewLine = false;

		if (sectionIndex >= sections.size())
		{
			moveToEndOfLastAtom();
			return false;
		}
		else if (atomIndex >= currentSection->getNumAtoms() - 1)
		{
			if (atomIndex >= currentSection->getNumAtoms())
			{
				if (++sectionIndex >= sections.size())
				{
					moveToEndOfLastAtom();
					return false;
				}

				atomIndex = 0;
				currentSection = sections.getUnchecked (sectionIndex);
			}
			else
			{
				const TextAtom* const lastAtom = currentSection->getAtom (atomIndex);

				if (! lastAtom->isWhitespace())
				{
					// handle the case where the last atom in a section is actually part of the same
					// word as the first atom of the next section...
					float right = atomRight + lastAtom->width;
					float lineHeight2 = lineHeight;
					float maxDescent2 = maxDescent;

					for (int section = sectionIndex + 1; section < sections.size(); ++section)
					{
						const UniformTextSection* const s = sections.getUnchecked (section);

						if (s->getNumAtoms() == 0)
							break;

						const TextAtom* const nextAtom = s->getAtom (0);

						if (nextAtom->isWhitespace())
							break;

						right += nextAtom->width;

						lineHeight2 = jmax (lineHeight2, s->font.getHeight());
						maxDescent2 = jmax (maxDescent2, s->font.getDescent());

						if (shouldWrap (right))
						{
							lineHeight = lineHeight2;
							maxDescent = maxDescent2;

							forceNewLine = true;
							break;
						}

						if (s->getNumAtoms() > 1)
							break;
					}
				}
			}
		}

		if (atom != nullptr)
		{
			atomX = atomRight;
			indexInText += atom->numChars;

			if (atom->isNewLine())
				beginNewLine();
		}

		atom = currentSection->getAtom (atomIndex);
		atomRight = atomX + atom->width;
		++atomIndex;

		if (shouldWrap (atomRight) || forceNewLine)
		{
			if (atom->isWhitespace())
			{
				// leave whitespace at the end of a line, but truncate it to avoid scrolling
				atomRight = jmin (atomRight, wordWrapWidth);
			}
			else
			{
				atomRight = atom->width;

				if (shouldWrap (atomRight))  // atom too big to fit on a line, so break it up..
				{
					tempAtom = *atom;
					tempAtom.width = 0;
					tempAtom.numChars = 0;
					atom = &tempAtom;

					if (atomX > 0)
						beginNewLine();

					return next();
				}

				beginNewLine();
				return true;
			}
		}

		return true;
	}

	void beginNewLine()
	{
		atomX = 0;
		lineY += lineHeight;

		int tempSectionIndex = sectionIndex;
		int tempAtomIndex = atomIndex;
		const UniformTextSection* section = sections.getUnchecked (tempSectionIndex);

		lineHeight = section->font.getHeight();
		maxDescent = section->font.getDescent();

		float x = (atom != nullptr) ? atom->width : 0;

		while (! shouldWrap (x))
		{
			if (tempSectionIndex >= sections.size())
				break;

			bool checkSize = false;

			if (tempAtomIndex >= section->getNumAtoms())
			{
				if (++tempSectionIndex >= sections.size())
					break;

				tempAtomIndex = 0;
				section = sections.getUnchecked (tempSectionIndex);
				checkSize = true;
			}

			const TextAtom* const nextAtom = section->getAtom (tempAtomIndex);

			if (nextAtom == nullptr)
				break;

			x += nextAtom->width;

			if (shouldWrap (x) || nextAtom->isNewLine())
				break;

			if (checkSize)
			{
				lineHeight = jmax (lineHeight, section->font.getHeight());
				maxDescent = jmax (maxDescent, section->font.getDescent());
			}

			++tempAtomIndex;
		}
	}

	void draw (Graphics& g, const UniformTextSection*& lastSection) const
	{
		if (passwordCharacter != 0 || ! atom->isWhitespace())
		{
			if (lastSection != currentSection)
			{
				lastSection = currentSection;
				g.setColour (currentSection->colour);
				g.setFont (currentSection->font);
			}

			jassert (atom->getTrimmedText (passwordCharacter).isNotEmpty());

			GlyphArrangement ga;
			ga.addLineOfText (currentSection->font,
							  atom->getTrimmedText (passwordCharacter),
							  atomX,
							  (float) roundToInt (lineY + lineHeight - maxDescent));
			ga.draw (g);
		}
	}

	void drawSelection (Graphics& g, const Range<int>& selection) const
	{
		const int startX = roundToInt (indexToX (selection.getStart()));
		const int endX   = roundToInt (indexToX (selection.getEnd()));

		const int y = roundToInt (lineY);
		const int nextY = roundToInt (lineY + lineHeight);

		g.fillRect (startX, y, endX - startX, nextY - y);
	}

	void drawUnderline (Graphics& g, const Range<int>& underline, const Colour& colour) const
	{
		const int startX	= roundToInt (indexToX (underline.getStart()));
		const int endX	  = roundToInt (indexToX (underline.getEnd()));
		const int baselineY = roundToInt (lineY + currentSection->font.getAscent() + 0.5f);

		Graphics::ScopedSaveState state (g);
		g.reduceClipRegion (Rectangle<int> (startX, baselineY, endX - startX, 1));
		g.fillCheckerBoard (Rectangle<int> (0, 0, endX, baselineY + 1), 3, 1, colour, Colours::transparentBlack);
	}

	void drawSelectedText (Graphics& g,
						   const Range<int>& selection,
						   const Colour& selectedTextColour) const
	{
		if (passwordCharacter != 0 || ! atom->isWhitespace())
		{
			GlyphArrangement ga;
			ga.addLineOfText (currentSection->font,
							  atom->getTrimmedText (passwordCharacter),
							  atomX,
							  (float) roundToInt (lineY + lineHeight - maxDescent));

			if (selection.getEnd() < indexInText + atom->numChars)
			{
				GlyphArrangement ga2 (ga);
				ga2.removeRangeOfGlyphs (0, selection.getEnd() - indexInText);
				ga.removeRangeOfGlyphs (selection.getEnd() - indexInText, -1);

				g.setColour (currentSection->colour);
				ga2.draw (g);
			}

			if (selection.getStart() > indexInText)
			{
				GlyphArrangement ga2 (ga);
				ga2.removeRangeOfGlyphs (selection.getStart() - indexInText, -1);
				ga.removeRangeOfGlyphs (0, selection.getStart() - indexInText);

				g.setColour (currentSection->colour);
				ga2.draw (g);
			}

			g.setColour (selectedTextColour);
			ga.draw (g);
		}
	}

	float indexToX (const int indexToFind) const
	{
		if (indexToFind <= indexInText)
			return atomX;

		if (indexToFind >= indexInText + atom->numChars)
			return atomRight;

		GlyphArrangement g;
		g.addLineOfText (currentSection->font,
						 atom->getText (passwordCharacter),
						 atomX, 0.0f);

		if (indexToFind - indexInText >= g.getNumGlyphs())
			return atomRight;

		return jmin (atomRight, g.getGlyph (indexToFind - indexInText).getLeft());
	}

	int xToIndex (const float xToFind) const
	{
		if (xToFind <= atomX || atom->isNewLine())
			return indexInText;

		if (xToFind >= atomRight)
			return indexInText + atom->numChars;

		GlyphArrangement g;
		g.addLineOfText (currentSection->font,
						 atom->getText (passwordCharacter),
						 atomX, 0.0f);

		int j;
		for (j = 0; j < g.getNumGlyphs(); ++j)
			if ((g.getGlyph(j).getLeft() + g.getGlyph(j).getRight()) / 2 > xToFind)
				break;

		return indexInText + j;
	}

	bool getCharPosition (const int index, float& cx, float& cy, float& lineHeight_)
	{
		while (next())
		{
			if (indexInText + atom->numChars > index)
			{
				cx = indexToX (index);
				cy = lineY;
				lineHeight_ = lineHeight;
				return true;
			}
		}

		cx = atomX;
		cy = lineY;
		lineHeight_ = lineHeight;
		return false;
	}

	int indexInText;
	float lineY, lineHeight, maxDescent;
	float atomX, atomRight;
	const TextAtom* atom;
	const UniformTextSection* currentSection;

private:
	const Array <UniformTextSection*>& sections;
	int sectionIndex, atomIndex;
	const float wordWrapWidth;
	const juce_wchar passwordCharacter;
	TextAtom tempAtom;

	Iterator& operator= (const Iterator&);

	void moveToEndOfLastAtom()
	{
		if (atom != nullptr)
		{
			atomX = atomRight;

			if (atom->isNewLine())
			{
				atomX = 0.0f;
				lineY += lineHeight;
			}
		}
	}

	bool shouldWrap (const float x) const
	{
		return (x - 0.0001f) >= wordWrapWidth;
	}

	JUCE_LEAK_DETECTOR (Iterator);
};

class TextEditor::InsertAction  : public UndoableAction
{
public:
	InsertAction (TextEditor& owner_,
				  const String& text_,
				  const int insertIndex_,
				  const Font& font_,
				  const Colour& colour_,
				  const int oldCaretPos_,
				  const int newCaretPos_)
		: owner (owner_),
		  text (text_),
		  insertIndex (insertIndex_),
		  oldCaretPos (oldCaretPos_),
		  newCaretPos (newCaretPos_),
		  font (font_),
		  colour (colour_)
	{
	}

	bool perform()
	{
		owner.insert (text, insertIndex, font, colour, 0, newCaretPos);
		return true;
	}

	bool undo()
	{
		owner.remove (Range<int> (insertIndex, insertIndex + text.length()), 0, oldCaretPos);
		return true;
	}

	int getSizeInUnits()
	{
		return text.length() + 16;
	}

private:
	TextEditor& owner;
	const String text;
	const int insertIndex, oldCaretPos, newCaretPos;
	const Font font;
	const Colour colour;

	JUCE_DECLARE_NON_COPYABLE (InsertAction);
};

class TextEditor::RemoveAction  : public UndoableAction
{
public:
	RemoveAction (TextEditor& owner_,
				  const Range<int> range_,
				  const int oldCaretPos_,
				  const int newCaretPos_,
				  const Array <UniformTextSection*>& removedSections_)
		: owner (owner_),
		  range (range_),
		  oldCaretPos (oldCaretPos_),
		  newCaretPos (newCaretPos_),
		  removedSections (removedSections_)
	{
	}

	~RemoveAction()
	{
		for (int i = removedSections.size(); --i >= 0;)
		{
			UniformTextSection* const section = removedSections.getUnchecked (i);
			section->clear();
			delete section;
		}
	}

	bool perform()
	{
		owner.remove (range, 0, newCaretPos);
		return true;
	}

	bool undo()
	{
		owner.reinsert (range.getStart(), removedSections);
		owner.moveCaretTo (oldCaretPos, false);
		return true;
	}

	int getSizeInUnits()
	{
		int n = 0;

		for (int i = removedSections.size(); --i >= 0;)
			n += removedSections.getUnchecked (i)->getTotalLength();

		return n + 16;
	}

private:
	TextEditor& owner;
	const Range<int> range;
	const int oldCaretPos, newCaretPos;
	Array <UniformTextSection*> removedSections;

	JUCE_DECLARE_NON_COPYABLE (RemoveAction);
};

class TextEditor::TextHolderComponent  : public Component,
										 public Timer,
										 public ValueListener
{
public:
	TextHolderComponent (TextEditor& owner_)
		: owner (owner_)
	{
		setWantsKeyboardFocus (false);
		setInterceptsMouseClicks (false, true);

		owner.getTextValue().addListener (this);
	}

	~TextHolderComponent()
	{
		owner.getTextValue().removeListener (this);
	}

	void paint (Graphics& g)
	{
		owner.drawContent (g);
	}

	void restartTimer()
	{
		startTimer (350);
	}

	void timerCallback()
	{
		owner.timerCallbackInt();
	}

	const MouseCursor getMouseCursor()
	{
		return owner.getMouseCursor();
	}

	void valueChanged (Value&)
	{
		owner.textWasChangedByValue();
	}

private:
	TextEditor& owner;

	JUCE_DECLARE_NON_COPYABLE (TextHolderComponent);
};

class TextEditorViewport  : public Viewport
{
public:
	TextEditorViewport (TextEditor& owner_)
		: owner (owner_), lastWordWrapWidth (0), rentrant (false)
	{
	}

	void visibleAreaChanged (const Rectangle<int>&)
	{
		if (! rentrant) // it's rare, but possible to get into a feedback loop as the viewport's scrollbars
						// appear and disappear, causing the wrap width to change.
		{
			const float wordWrapWidth = owner.getWordWrapWidth();

			if (wordWrapWidth != lastWordWrapWidth)
			{
				lastWordWrapWidth = wordWrapWidth;

				rentrant = true;
				owner.updateTextHolderSize();
				rentrant = false;
			}
		}
	}

private:
	TextEditor& owner;
	float lastWordWrapWidth;
	bool rentrant;

	JUCE_DECLARE_NON_COPYABLE (TextEditorViewport);
};

namespace TextEditorDefs
{
	const int textChangeMessageId = 0x10003001;
	const int returnKeyMessageId  = 0x10003002;
	const int escapeKeyMessageId  = 0x10003003;
	const int focusLossMessageId  = 0x10003004;

	const int maxActionsPerTransaction = 100;

	int getCharacterCategory (const juce_wchar character)
	{
		return CharacterFunctions::isLetterOrDigit (character)
					? 2 : (CharacterFunctions::isWhitespace (character) ? 0 : 1);
	}
}

TextEditor::TextEditor (const String& name,
						const juce_wchar passwordCharacter_)
	: Component (name),
	  borderSize (1, 1, 1, 3),
	  readOnly (false),
	  multiline (false),
	  wordWrap (false),
	  returnKeyStartsNewLine (false),
	  popupMenuEnabled (true),
	  selectAllTextWhenFocused (false),
	  scrollbarVisible (true),
	  wasFocused (false),
	  keepCaretOnScreen (true),
	  tabKeyUsed (false),
	  menuActive (false),
	  valueTextNeedsUpdating (false),
	  maxTextLength (0),
	  leftIndent (4),
	  topIndent (4),
	  lastTransactionTime (0),
	  currentFont (14.0f),
	  totalNumChars (0),
	  caretPosition (0),
	  passwordCharacter (passwordCharacter_),
	  dragType (notDragging)
{
	setOpaque (true);

	addAndMakeVisible (viewport = new TextEditorViewport (*this));
	viewport->setViewedComponent (textHolder = new TextHolderComponent (*this));
	viewport->setWantsKeyboardFocus (false);
	viewport->setScrollBarsShown (false, false);

	setWantsKeyboardFocus (true);
	setCaretVisible (true);
}

TextEditor::~TextEditor()
{
	if (wasFocused)
	{
		ComponentPeer* const peer = getPeer();
		if (peer != nullptr)
			peer->dismissPendingTextInput();
	}

	textValue.referTo (Value());
	clearInternal (0);
	viewport = nullptr;
	textHolder = nullptr;
}

void TextEditor::newTransaction()
{
	lastTransactionTime = Time::getApproximateMillisecondCounter();
	undoManager.beginNewTransaction();
}

bool TextEditor::undoOrRedo (const bool shouldUndo)
{
	if (! isReadOnly())
	{
		newTransaction();

		if (shouldUndo ? undoManager.undo()
					   : undoManager.redo())
		{
			scrollToMakeSureCursorIsVisible();
			repaint();
			textChanged();
			return true;
		}
	}

	return false;
}

bool TextEditor::undo()	 { return undoOrRedo (true); }
bool TextEditor::redo()	 { return undoOrRedo (false); }

void TextEditor::setMultiLine (const bool shouldBeMultiLine,
							   const bool shouldWordWrap)
{
	if (multiline != shouldBeMultiLine
		 || wordWrap != (shouldWordWrap && shouldBeMultiLine))
	{
		multiline = shouldBeMultiLine;
		wordWrap = shouldWordWrap && shouldBeMultiLine;

		viewport->setScrollBarsShown (scrollbarVisible && multiline,
									  scrollbarVisible && multiline);
		viewport->setViewPosition (0, 0);
		resized();
		scrollToMakeSureCursorIsVisible();
	}
}

bool TextEditor::isMultiLine() const
{
	return multiline;
}

void TextEditor::setScrollbarsShown (bool shown)
{
	if (scrollbarVisible != shown)
	{
		scrollbarVisible = shown;
		shown = shown && isMultiLine();
		viewport->setScrollBarsShown (shown, shown);
	}
}

void TextEditor::setReadOnly (const bool shouldBeReadOnly)
{
	if (readOnly != shouldBeReadOnly)
	{
		readOnly = shouldBeReadOnly;
		enablementChanged();
	}
}

bool TextEditor::isReadOnly() const
{
	return readOnly || ! isEnabled();
}

bool TextEditor::isTextInputActive() const
{
	return ! isReadOnly();
}

void TextEditor::setReturnKeyStartsNewLine (const bool shouldStartNewLine)
{
	returnKeyStartsNewLine = shouldStartNewLine;
}

void TextEditor::setTabKeyUsedAsCharacter (const bool shouldTabKeyBeUsed)
{
	tabKeyUsed = shouldTabKeyBeUsed;
}

void TextEditor::setPopupMenuEnabled (const bool b)
{
	popupMenuEnabled = b;
}

void TextEditor::setSelectAllWhenFocused (const bool b)
{
	selectAllTextWhenFocused = b;
}

const Font& TextEditor::getFont() const
{
	return currentFont;
}

void TextEditor::setFont (const Font& newFont)
{
	currentFont = newFont;
	scrollToMakeSureCursorIsVisible();
}

void TextEditor::applyFontToAllText (const Font& newFont)
{
	currentFont = newFont;

	const Colour overallColour (findColour (textColourId));

	for (int i = sections.size(); --i >= 0;)
	{
		UniformTextSection* const uts = sections.getUnchecked (i);
		uts->setFont (newFont, passwordCharacter);
		uts->colour = overallColour;
	}

	coalesceSimilarSections();
	updateTextHolderSize();
	scrollToMakeSureCursorIsVisible();
	repaint();
}

void TextEditor::colourChanged()
{
	setOpaque (findColour (backgroundColourId).isOpaque());
	repaint();
}

void TextEditor::lookAndFeelChanged()
{
	if (isCaretVisible())
	{
		setCaretVisible (false);
		setCaretVisible (true);
	}
}

void TextEditor::setCaretVisible (const bool shouldCaretBeVisible)
{
	if (shouldCaretBeVisible && ! isReadOnly())
	{
		if (caret == nullptr)
			textHolder->addChildComponent (caret = getLookAndFeel().createCaretComponent (this));
	}
	else
	{
		caret = nullptr;
	}

	setMouseCursor (shouldCaretBeVisible ? MouseCursor::IBeamCursor
										 : MouseCursor::NormalCursor);
}

void TextEditor::updateCaretPosition()
{
	if (caret != nullptr)
		caret->setCaretPosition (getCaretRectangle().translated (leftIndent, topIndent));
}

void TextEditor::setInputRestrictions (const int maxLen,
									   const String& chars)
{
	maxTextLength = jmax (0, maxLen);
	allowedCharacters = chars;
}

void TextEditor::setTextToShowWhenEmpty (const String& text, const Colour& colourToUse)
{
	textToShowWhenEmpty = text;
	colourForTextWhenEmpty = colourToUse;
}

void TextEditor::setPasswordCharacter (const juce_wchar newPasswordCharacter)
{
	if (passwordCharacter != newPasswordCharacter)
	{
		passwordCharacter = newPasswordCharacter;
		resized();
		repaint();
	}
}

void TextEditor::setScrollBarThickness (const int newThicknessPixels)
{
	viewport->setScrollBarThickness (newThicknessPixels);
}

void TextEditor::setScrollBarButtonVisibility (const bool buttonsVisible)
{
	viewport->setScrollBarButtonVisibility (buttonsVisible);
}

void TextEditor::clear()
{
	clearInternal (0);
	updateTextHolderSize();
	undoManager.clearUndoHistory();
}

void TextEditor::setText (const String& newText,
						  const bool sendTextChangeMessage)
{
	const int newLength = newText.length();

	if (newLength != getTotalNumChars() || getText() != newText)
	{
		const int oldCursorPos = caretPosition;
		const bool cursorWasAtEnd = oldCursorPos >= getTotalNumChars();

		clearInternal (0);
		insert (newText, 0, currentFont, findColour (textColourId), 0, caretPosition);

		// if you're adding text with line-feeds to a single-line text editor, it
		// ain't gonna look right!
		jassert (multiline || ! newText.containsAnyOf ("\r\n"));

		if (cursorWasAtEnd && ! isMultiLine())
			moveCaretTo (getTotalNumChars(), false);
		else
			moveCaretTo (oldCursorPos, false);

		if (sendTextChangeMessage)
			textChanged();

		updateTextHolderSize();
		scrollToMakeSureCursorIsVisible();
		undoManager.clearUndoHistory();

		repaint();
	}
}

void TextEditor::updateValueFromText()
{
	if (valueTextNeedsUpdating)
	{
		valueTextNeedsUpdating = false;
		textValue = getText();
	}
}

Value& TextEditor::getTextValue()
{
	updateValueFromText();
	return textValue;
}

void TextEditor::textWasChangedByValue()
{
	if (textValue.getValueSource().getReferenceCount() > 1)
		setText (textValue.getValue());
}

void TextEditor::textChanged()
{
	updateTextHolderSize();
	postCommandMessage (TextEditorDefs::textChangeMessageId);

	if (textValue.getValueSource().getReferenceCount() > 1)
	{
		valueTextNeedsUpdating = false;
		textValue = getText();
	}
}

void TextEditor::returnPressed()
{
	postCommandMessage (TextEditorDefs::returnKeyMessageId);
}

void TextEditor::escapePressed()
{
	postCommandMessage (TextEditorDefs::escapeKeyMessageId);
}

void TextEditor::addListener (TextEditorListener* const newListener)
{
	listeners.add (newListener);
}

void TextEditor::removeListener (TextEditorListener* const listenerToRemove)
{
	listeners.remove (listenerToRemove);
}

void TextEditor::timerCallbackInt()
{
	if (hasKeyboardFocus (false) && ! isCurrentlyBlockedByAnotherModalComponent())
		wasFocused = true;

	const unsigned int now = Time::getApproximateMillisecondCounter();

	if (now > lastTransactionTime + 200)
		newTransaction();
}

void TextEditor::repaintText (const Range<int>& range)
{
	if (! range.isEmpty())
	{
		float x = 0, y = 0, lh = currentFont.getHeight();

		const float wordWrapWidth = getWordWrapWidth();

		if (wordWrapWidth > 0)
		{
			Iterator i (sections, wordWrapWidth, passwordCharacter);

			i.getCharPosition (range.getStart(), x, y, lh);

			const int y1 = (int) y;
			int y2;

			if (range.getEnd() >= getTotalNumChars())
			{
				y2 = textHolder->getHeight();
			}
			else
			{
				i.getCharPosition (range.getEnd(), x, y, lh);
				y2 = (int) (y + lh * 2.0f);
			}

			textHolder->repaint (0, y1, textHolder->getWidth(), y2 - y1);
		}
	}
}

void TextEditor::moveCaret (int newCaretPos)
{
	if (newCaretPos < 0)
		newCaretPos = 0;
	else if (newCaretPos > getTotalNumChars())
		newCaretPos = getTotalNumChars();

	if (newCaretPos != getCaretPosition())
	{
		caretPosition = newCaretPos;
		textHolder->restartTimer();
		scrollToMakeSureCursorIsVisible();
		updateCaretPosition();
	}
}

void TextEditor::setCaretPosition (const int newIndex)
{
	moveCaretTo (newIndex, false);
}

int TextEditor::getCaretPosition() const
{
	return caretPosition;
}

void TextEditor::scrollEditorToPositionCaret (const int desiredCaretX,
											  const int desiredCaretY)

{
	updateCaretPosition();

	const Rectangle<int> caretPos (getCaretRectangle());

	int vx = caretPos.getX() - desiredCaretX;
	int vy = caretPos.getY() - desiredCaretY;

	if (desiredCaretX < jmax (1, proportionOfWidth (0.05f)))
		vx += desiredCaretX - proportionOfWidth (0.2f);
	else if (desiredCaretX > jmax (0, viewport->getMaximumVisibleWidth() - (wordWrap ? 2 : 10)))
		vx += desiredCaretX + (isMultiLine() ? proportionOfWidth (0.2f) : 10) - viewport->getMaximumVisibleWidth();

	vx = jlimit (0, jmax (0, textHolder->getWidth() + 8 - viewport->getMaximumVisibleWidth()), vx);

	if (! isMultiLine())
	{
		vy = viewport->getViewPositionY();
	}
	else
	{
		vy = jlimit (0, jmax (0, textHolder->getHeight() - viewport->getMaximumVisibleHeight()), vy);

		if (desiredCaretY < 0)
			vy = jmax (0, desiredCaretY + vy);
		else if (desiredCaretY > jmax (0, viewport->getMaximumVisibleHeight() - topIndent - caretPos.getHeight()))
			vy += desiredCaretY + 2 + caretPos.getHeight() + topIndent - viewport->getMaximumVisibleHeight();
	}

	viewport->setViewPosition (vx, vy);
}

const Rectangle<int> TextEditor::getCaretRectangle()
{
	float cursorX, cursorY;
	float cursorHeight = currentFont.getHeight(); // (in case the text is empty and the call below doesn't set this value)
	getCharPosition (caretPosition, cursorX, cursorY, cursorHeight);

	return Rectangle<int> (roundToInt (cursorX), roundToInt (cursorY), 2, roundToInt (cursorHeight));
}

float TextEditor::getWordWrapWidth() const
{
	return wordWrap ? (float) (viewport->getMaximumVisibleWidth() - leftIndent - leftIndent / 2)
					: std::numeric_limits<float>::max();
}

void TextEditor::updateTextHolderSize()
{
	const float wordWrapWidth = getWordWrapWidth();

	if (wordWrapWidth > 0)
	{
		float maxWidth = 0.0f;

		Iterator i (sections, wordWrapWidth, passwordCharacter);

		while (i.next())
			maxWidth = jmax (maxWidth, i.atomRight);

		const int w = leftIndent + roundToInt (maxWidth);
		const int h = topIndent + roundToInt (jmax (i.lineY + i.lineHeight,
													currentFont.getHeight()));

		textHolder->setSize (w + 2, h + 1); // (the +2 allows a bit of space for the cursor to be at the right-hand-edge)
	}
}

int TextEditor::getTextWidth() const
{
	return textHolder->getWidth();
}

int TextEditor::getTextHeight() const
{
	return textHolder->getHeight();
}

void TextEditor::setIndents (const int newLeftIndent,
							 const int newTopIndent)
{
	leftIndent = newLeftIndent;
	topIndent = newTopIndent;
}

void TextEditor::setBorder (const BorderSize<int>& border)
{
	borderSize = border;
	resized();
}

const BorderSize<int> TextEditor::getBorder() const
{
	return borderSize;
}

void TextEditor::setScrollToShowCursor (const bool shouldScrollToShowCursor)
{
	keepCaretOnScreen = shouldScrollToShowCursor;
}

void TextEditor::scrollToMakeSureCursorIsVisible()
{
	updateCaretPosition();

	if (keepCaretOnScreen)
	{
		int x = viewport->getViewPositionX();
		int y = viewport->getViewPositionY();

		const Rectangle<int> caretPos (getCaretRectangle());

		const int relativeCursorX = caretPos.getX() - x;
		const int relativeCursorY = caretPos.getY() - y;

		if (relativeCursorX < jmax (1, proportionOfWidth (0.05f)))
		{
			x += relativeCursorX - proportionOfWidth (0.2f);
		}
		else if (relativeCursorX > jmax (0, viewport->getMaximumVisibleWidth() - (wordWrap ? 2 : 10)))
		{
			x += relativeCursorX + (isMultiLine() ? proportionOfWidth (0.2f) : 10) - viewport->getMaximumVisibleWidth();
		}

		x = jlimit (0, jmax (0, textHolder->getWidth() + 8 - viewport->getMaximumVisibleWidth()), x);

		if (! isMultiLine())
		{
			y = (getHeight() - textHolder->getHeight() - topIndent) / -2;
		}
		else
		{
			if (relativeCursorY < 0)
			{
				y = jmax (0, relativeCursorY + y);
			}
			else if (relativeCursorY > jmax (0, viewport->getMaximumVisibleHeight() - topIndent - caretPos.getHeight()))
			{
				y += relativeCursorY + 2 + caretPos.getHeight() + topIndent - viewport->getMaximumVisibleHeight();
			}
		}

		viewport->setViewPosition (x, y);
	}
}

void TextEditor::moveCaretTo (const int newPosition,
							  const bool isSelecting)
{
	if (isSelecting)
	{
		moveCaret (newPosition);

		const Range<int> oldSelection (selection);

		if (dragType == notDragging)
		{
			if (abs (getCaretPosition() - selection.getStart()) < abs (getCaretPosition() - selection.getEnd()))
				dragType = draggingSelectionStart;
			else
				dragType = draggingSelectionEnd;
		}

		if (dragType == draggingSelectionStart)
		{
			if (getCaretPosition() >= selection.getEnd())
				dragType = draggingSelectionEnd;

			selection = Range<int>::between (getCaretPosition(), selection.getEnd());
		}
		else
		{
			if (getCaretPosition() < selection.getStart())
				dragType = draggingSelectionStart;

			selection = Range<int>::between (getCaretPosition(), selection.getStart());
		}

		repaintText (selection.getUnionWith (oldSelection));
	}
	else
	{
		dragType = notDragging;

		repaintText (selection);

		moveCaret (newPosition);
		selection = Range<int>::emptyRange (getCaretPosition());
	}
}

int TextEditor::getTextIndexAt (const int x,
								const int y)
{
	return indexAtPosition ((float) (x + viewport->getViewPositionX() - leftIndent),
							(float) (y + viewport->getViewPositionY() - topIndent));
}

void TextEditor::insertTextAtCaret (const String& newText_)
{
	String newText (newText_);

	if (allowedCharacters.isNotEmpty())
		newText = newText.retainCharacters (allowedCharacters);

	if (! isMultiLine())
		newText = newText.replaceCharacters ("\r\n", "  ");
	else
		newText = newText.replace ("\r\n", "\n");

	const int newCaretPos = selection.getStart() + newText.length();
	const int insertIndex = selection.getStart();

	remove (selection, getUndoManager(),
			newText.isNotEmpty() ? newCaretPos - 1 : newCaretPos);

	if (maxTextLength > 0)
		newText = newText.substring (0, maxTextLength - getTotalNumChars());

	if (newText.isNotEmpty())
		insert (newText,
				insertIndex,
				currentFont,
				findColour (textColourId),
				getUndoManager(),
				newCaretPos);

	textChanged();
}

void TextEditor::setHighlightedRegion (const Range<int>& newSelection)
{
	moveCaretTo (newSelection.getStart(), false);
	moveCaretTo (newSelection.getEnd(), true);
}

void TextEditor::copy()
{
	if (passwordCharacter == 0)
	{
		const String selectedText (getHighlightedText());

		if (selectedText.isNotEmpty())
			SystemClipboard::copyTextToClipboard (selectedText);
	}
}

void TextEditor::paste()
{
	if (! isReadOnly())
	{
		const String clip (SystemClipboard::getTextFromClipboard());

		if (clip.isNotEmpty())
			insertTextAtCaret (clip);
	}
}

void TextEditor::cut()
{
	if (! isReadOnly())
	{
		moveCaret (selection.getEnd());
		insertTextAtCaret (String::empty);
	}
}

void TextEditor::drawContent (Graphics& g)
{
	const float wordWrapWidth = getWordWrapWidth();

	if (wordWrapWidth > 0)
	{
		g.setOrigin (leftIndent, topIndent);
		const Rectangle<int> clip (g.getClipBounds());
		Colour selectedTextColour;

		Iterator i (sections, wordWrapWidth, passwordCharacter);

		while (i.lineY + 200.0 < clip.getY() && i.next())
		{}

		if (! selection.isEmpty())
		{
			g.setColour (findColour (highlightColourId).withMultipliedAlpha (hasKeyboardFocus (true) ? 1.0f : 0.5f));

			selectedTextColour = findColour (highlightedTextColourId);

			Iterator i2 (i);

			while (i2.next() && i2.lineY < clip.getBottom())
			{
				if (i2.lineY + i2.lineHeight >= clip.getY()
					 && selection.intersects (Range<int> (i2.indexInText, i2.indexInText + i2.atom->numChars)))
				{
					i2.drawSelection (g, selection);
				}
			}
		}

		const UniformTextSection* lastSection = nullptr;

		while (i.next() && i.lineY < clip.getBottom())
		{
			if (i.lineY + i.lineHeight >= clip.getY())
			{
				if (selection.intersects (Range<int> (i.indexInText, i.indexInText + i.atom->numChars)))
				{
					i.drawSelectedText (g, selection, selectedTextColour);
					lastSection = nullptr;
				}
				else
				{
					i.draw (g, lastSection);
				}
			}
		}

		for (int j = underlinedSections.size(); --j >= 0;)
		{
			const Range<int>& underlinedSection = underlinedSections.getReference (j);

			Iterator i2 (sections, wordWrapWidth, passwordCharacter);

			while (i2.next() && i2.lineY < clip.getBottom())
			{
				if (i2.lineY + i2.lineHeight >= clip.getY()
					 && underlinedSection.intersects (Range<int> (i2.indexInText, i2.indexInText + i2.atom->numChars)))
				{
					i2.drawUnderline (g, underlinedSection, findColour (textColourId));
				}
			}
		}
	}
}

void TextEditor::paint (Graphics& g)
{
	getLookAndFeel().fillTextEditorBackground (g, getWidth(), getHeight(), *this);
}

void TextEditor::paintOverChildren (Graphics& g)
{
	if (textToShowWhenEmpty.isNotEmpty()
		 && (! hasKeyboardFocus (false))
		 && getTotalNumChars() == 0)
	{
		g.setColour (colourForTextWhenEmpty);
		g.setFont (getFont());

		if (isMultiLine())
		{
			g.drawText (textToShowWhenEmpty,
						0, 0, getWidth(), getHeight(),
						Justification::centred, true);
		}
		else
		{
			g.drawText (textToShowWhenEmpty,
						leftIndent, topIndent,
						viewport->getWidth() - leftIndent,
						viewport->getHeight() - topIndent,
						Justification::centredLeft, true);
		}
	}

	getLookAndFeel().drawTextEditorOutline (g, getWidth(), getHeight(), *this);
}

static void textEditorMenuCallback (int menuResult, TextEditor* editor)
{
	if (editor != nullptr && menuResult != 0)
		editor->performPopupMenuAction (menuResult);
}

void TextEditor::mouseDown (const MouseEvent& e)
{
	beginDragAutoRepeat (100);
	newTransaction();

	if (wasFocused || ! selectAllTextWhenFocused)
	{
		if (! (popupMenuEnabled && e.mods.isPopupMenu()))
		{
			moveCaretTo (getTextIndexAt (e.x, e.y),
						 e.mods.isShiftDown());
		}
		else
		{
			PopupMenu m;
			m.setLookAndFeel (&getLookAndFeel());
			addPopupMenuItems (m, &e);

			m.showMenuAsync (PopupMenu::Options(),
							 ModalCallbackFunction::forComponent (textEditorMenuCallback, this));
		}
	}
}

void TextEditor::mouseDrag (const MouseEvent& e)
{
	if (wasFocused || ! selectAllTextWhenFocused)
	{
		if (! (popupMenuEnabled && e.mods.isPopupMenu()))
		{
			moveCaretTo (getTextIndexAt (e.x, e.y), true);
		}
	}
}

void TextEditor::mouseUp (const MouseEvent& e)
{
	newTransaction();
	textHolder->restartTimer();

	if (wasFocused || ! selectAllTextWhenFocused)
	{
		if (e.mouseWasClicked() && ! (popupMenuEnabled && e.mods.isPopupMenu()))
		{
			moveCaret (getTextIndexAt (e.x, e.y));
		}
	}

	wasFocused = true;
}

void TextEditor::mouseDoubleClick (const MouseEvent& e)
{
	int tokenEnd = getTextIndexAt (e.x, e.y);
	int tokenStart = tokenEnd;

	if (e.getNumberOfClicks() > 3)
	{
		tokenStart = 0;
		tokenEnd = getTotalNumChars();
	}
	else
	{
		const String t (getText());
		const int totalLength = getTotalNumChars();

		while (tokenEnd < totalLength)
		{
			// (note the slight bodge here - it's because iswalnum only checks for alphabetic chars in the current locale)
			const juce_wchar c = t [tokenEnd];
			if (CharacterFunctions::isLetterOrDigit (c) || c > 128)
				++tokenEnd;
			else
				break;
		}

		tokenStart = tokenEnd;

		while (tokenStart > 0)
		{
			// (note the slight bodge here - it's because iswalnum only checks for alphabetic chars in the current locale)
			const juce_wchar c = t [tokenStart - 1];
			if (CharacterFunctions::isLetterOrDigit (c) || c > 128)
				--tokenStart;
			else
				break;
		}

		if (e.getNumberOfClicks() > 2)
		{
			while (tokenEnd < totalLength)
			{
				const juce_wchar c = t [tokenEnd];
				if (c != '\r' && c != '\n')
					++tokenEnd;
				else
					break;
			}

			while (tokenStart > 0)
			{
				const juce_wchar c = t [tokenStart - 1];
				if (c != '\r' && c != '\n')
					--tokenStart;
				else
					break;
			}
		}
	}

	moveCaretTo (tokenEnd, false);
	moveCaretTo (tokenStart, true);
}

void TextEditor::mouseWheelMove (const MouseEvent& e, float wheelIncrementX, float wheelIncrementY)
{
	if (! viewport->useMouseWheelMoveIfNeeded (e, wheelIncrementX, wheelIncrementY))
		Component::mouseWheelMove (e, wheelIncrementX, wheelIncrementY);
}

bool TextEditor::moveCaretWithTransation (const int newPos, const bool selecting)
{
	newTransaction();
	moveCaretTo (newPos, selecting);
	return true;
}

bool TextEditor::moveCaretLeft (bool moveInWholeWordSteps, bool selecting)
{
	int pos = getCaretPosition();

	if (moveInWholeWordSteps)
		pos = findWordBreakBefore (pos);
	else
		--pos;

	return moveCaretWithTransation (pos, selecting);
}

bool TextEditor::moveCaretRight (bool moveInWholeWordSteps, bool selecting)
{
	int pos = getCaretPosition();

	if (moveInWholeWordSteps)
		pos = findWordBreakAfter (pos);
	else
		++pos;

	return moveCaretWithTransation (pos, selecting);
}

bool TextEditor::moveCaretUp (bool selecting)
{
	if (! isMultiLine())
		return moveCaretToStartOfLine (selecting);

	const Rectangle<float> caretPos (getCaretRectangle().toFloat());
	return moveCaretWithTransation (indexAtPosition (caretPos.getX(), caretPos.getY() - 1.0f), selecting);
}

bool TextEditor::moveCaretDown (bool selecting)
{
	if (! isMultiLine())
		return moveCaretToEndOfLine (selecting);

	const Rectangle<float> caretPos (getCaretRectangle().toFloat());
	return moveCaretWithTransation (indexAtPosition (caretPos.getX(), caretPos.getBottom() + 1.0f), selecting);
}

bool TextEditor::pageUp (bool selecting)
{
	if (! isMultiLine())
		return moveCaretToStartOfLine (selecting);

	const Rectangle<float> caretPos (getCaretRectangle().toFloat());
	return moveCaretWithTransation (indexAtPosition (caretPos.getX(), caretPos.getY() - viewport->getViewHeight()), selecting);
}

bool TextEditor::pageDown (bool selecting)
{
	if (! isMultiLine())
		return moveCaretToEndOfLine (selecting);

	const Rectangle<float> caretPos (getCaretRectangle().toFloat());
	return moveCaretWithTransation (indexAtPosition (caretPos.getX(), caretPos.getBottom() + viewport->getViewHeight()), selecting);
}

void TextEditor::scrollByLines (int deltaLines)
{
	ScrollBar* scrollbar = viewport->getVerticalScrollBar();

	if (scrollbar != nullptr)
		scrollbar->moveScrollbarInSteps (deltaLines);
}

bool TextEditor::scrollDown()
{
	scrollByLines (-1);
	return true;
}

bool TextEditor::scrollUp()
{
	scrollByLines (1);
	return true;
}

bool TextEditor::moveCaretToTop (bool selecting)
{
	return moveCaretWithTransation (0, selecting);
}

bool TextEditor::moveCaretToStartOfLine (bool selecting)
{
	const Rectangle<float> caretPos (getCaretRectangle().toFloat());
	return moveCaretWithTransation (indexAtPosition (0.0f, caretPos.getY()), selecting);
}

bool TextEditor::moveCaretToEnd (bool selecting)
{
	return moveCaretWithTransation (getTotalNumChars(), selecting);
}

bool TextEditor::moveCaretToEndOfLine (bool selecting)
{
	const Rectangle<float> caretPos (getCaretRectangle().toFloat());
	return moveCaretWithTransation (indexAtPosition ((float) textHolder->getWidth(), caretPos.getY()), selecting);
}

bool TextEditor::deleteBackwards (bool moveInWholeWordSteps)
{
	if (moveInWholeWordSteps)
		moveCaretTo (findWordBreakBefore (getCaretPosition()), true);
	else if (selection.isEmpty() && selection.getStart() > 0)
		selection.setStart (selection.getEnd() - 1);

	cut();
	return true;
}

bool TextEditor::deleteForwards (bool /*moveInWholeWordSteps*/)
{
	if (selection.isEmpty() && selection.getStart() < getTotalNumChars())
		selection.setEnd (selection.getStart() + 1);

	cut();
	return true;
}

bool TextEditor::copyToClipboard()
{
	newTransaction();
	copy();
	return true;
}

bool TextEditor::cutToClipboard()
{
	newTransaction();
	copy();
	cut();
	return true;
}

bool TextEditor::pasteFromClipboard()
{
	newTransaction();
	paste();
	return true;
}

bool TextEditor::selectAll()
{
	newTransaction();
	moveCaretTo (getTotalNumChars(), false);
	moveCaretTo (0, true);
	return true;
}

bool TextEditor::keyPressed (const KeyPress& key)
{
	if (isReadOnly() && key != KeyPress ('c', ModifierKeys::commandModifier, 0))
		return false;

	if (! TextEditorKeyMapper<TextEditor>::invokeKeyFunction (*this, key))
	{
		if (key == KeyPress::returnKey)
		{
			newTransaction();

			if (returnKeyStartsNewLine)
				insertTextAtCaret ("\n");
			else
				returnPressed();
		}
		else if (key.isKeyCode (KeyPress::escapeKey))
		{
			newTransaction();
			moveCaretTo (getCaretPosition(), false);
			escapePressed();
		}
		else if (key.getTextCharacter() >= ' '
				  || (tabKeyUsed && (key.getTextCharacter() == '\t')))
		{
			insertTextAtCaret (String::charToString (key.getTextCharacter()));

			lastTransactionTime = Time::getApproximateMillisecondCounter();
		}
		else
		{
			return false;
		}
	}

	return true;
}

bool TextEditor::keyStateChanged (const bool isKeyDown)
{
	if (! isKeyDown)
		return false;

   #if JUCE_WINDOWS
	if (KeyPress (KeyPress::F4Key, ModifierKeys::altModifier, 0).isCurrentlyDown())
		return false;  // We need to explicitly allow alt-F4 to pass through on Windows
   #endif

	// (overridden to avoid forwarding key events to the parent)
	return ! ModifierKeys::getCurrentModifiers().isCommandDown();
}

const int baseMenuItemID = 0x7fff0000;

void TextEditor::addPopupMenuItems (PopupMenu& m, const MouseEvent*)
{
	const bool writable = ! isReadOnly();

	if (passwordCharacter == 0)
	{
		m.addItem (baseMenuItemID + 1, TRANS("cut"), writable);
		m.addItem (baseMenuItemID + 2, TRANS("copy"), ! selection.isEmpty());
		m.addItem (baseMenuItemID + 3, TRANS("paste"), writable);
	}

	m.addItem (baseMenuItemID + 4, TRANS("delete"), writable);
	m.addSeparator();
	m.addItem (baseMenuItemID + 5, TRANS("select all"));
	m.addSeparator();

	if (getUndoManager() != nullptr)
	{
		m.addItem (baseMenuItemID + 6, TRANS("undo"), undoManager.canUndo());
		m.addItem (baseMenuItemID + 7, TRANS("redo"), undoManager.canRedo());
	}
}

void TextEditor::performPopupMenuAction (const int menuItemID)
{
	switch (menuItemID)
	{
		case baseMenuItemID + 1:	cutToClipboard(); break;
		case baseMenuItemID + 2:	copyToClipboard(); break;
		case baseMenuItemID + 3:	pasteFromClipboard(); break;
		case baseMenuItemID + 4:	cut(); break;
		case baseMenuItemID + 5:	selectAll(); break;
		case baseMenuItemID + 6:	undo(); break;
		case baseMenuItemID + 7:	redo(); break;
		default: break;
	}
}

void TextEditor::focusGained (FocusChangeType)
{
	newTransaction();

	if (selectAllTextWhenFocused)
	{
		moveCaretTo (0, false);
		moveCaretTo (getTotalNumChars(), true);
	}

	repaint();
	updateCaretPosition();

	ComponentPeer* const peer = getPeer();
	if (peer != nullptr && ! isReadOnly())
		peer->textInputRequired (getScreenPosition() - peer->getScreenPosition());
}

void TextEditor::focusLost (FocusChangeType)
{
	newTransaction();

	wasFocused = false;
	textHolder->stopTimer();

	underlinedSections.clear();

	ComponentPeer* const peer = getPeer();
	if (peer != nullptr)
		peer->dismissPendingTextInput();

	updateCaretPosition();

	postCommandMessage (TextEditorDefs::focusLossMessageId);
	repaint();
}

void TextEditor::resized()
{
	viewport->setBoundsInset (borderSize);
	viewport->setSingleStepSizes (16, roundToInt (currentFont.getHeight()));

	updateTextHolderSize();

	if (! isMultiLine())
	{
		scrollToMakeSureCursorIsVisible();
	}
	else
	{
		updateCaretPosition();
	}
}

void TextEditor::handleCommandMessage (const int commandId)
{
	Component::BailOutChecker checker (this);

	switch (commandId)
	{
	case TextEditorDefs::textChangeMessageId:
		listeners.callChecked (checker, &TextEditorListener::textEditorTextChanged, (TextEditor&) *this);
		break;

	case TextEditorDefs::returnKeyMessageId:
		listeners.callChecked (checker, &TextEditorListener::textEditorReturnKeyPressed, (TextEditor&) *this);
		break;

	case TextEditorDefs::escapeKeyMessageId:
		listeners.callChecked (checker, &TextEditorListener::textEditorEscapeKeyPressed, (TextEditor&) *this);
		break;

	case TextEditorDefs::focusLossMessageId:
		updateValueFromText();
		listeners.callChecked (checker, &TextEditorListener::textEditorFocusLost, (TextEditor&) *this);
		break;

	default:
		jassertfalse;
		break;
	}
}

void TextEditor::enablementChanged()
{
	setMouseCursor (isReadOnly() ? MouseCursor::NormalCursor
								 : MouseCursor::IBeamCursor);
	repaint();
}

void TextEditor::setTemporaryUnderlining (const Array <Range<int> >& newUnderlinedSections)
{
	underlinedSections = newUnderlinedSections;
	repaint();
}

UndoManager* TextEditor::getUndoManager() noexcept
{
	return isReadOnly() ? 0 : &undoManager;
}

void TextEditor::clearInternal (UndoManager* const um)
{
	remove (Range<int> (0, getTotalNumChars()), um, caretPosition);
}

void TextEditor::insert (const String& text,
						 const int insertIndex,
						 const Font& font,
						 const Colour& colour,
						 UndoManager* const um,
						 const int caretPositionToMoveTo)
{
	if (text.isNotEmpty())
	{
		if (um != nullptr)
		{
			if (um->getNumActionsInCurrentTransaction() > TextEditorDefs::maxActionsPerTransaction)
				newTransaction();

			um->perform (new InsertAction (*this, text, insertIndex, font, colour,
										   caretPosition, caretPositionToMoveTo));
		}
		else
		{
			repaintText (Range<int> (insertIndex, getTotalNumChars())); // must do this before and after changing the data, in case
																		// a line gets moved due to word wrap

			int index = 0;
			int nextIndex = 0;

			for (int i = 0; i < sections.size(); ++i)
			{
				nextIndex = index + sections.getUnchecked (i)->getTotalLength();

				if (insertIndex == index)
				{
					sections.insert (i, new UniformTextSection (text,
																font, colour,
																passwordCharacter));
					break;
				}
				else if (insertIndex > index && insertIndex < nextIndex)
				{
					splitSection (i, insertIndex - index);
					sections.insert (i + 1, new UniformTextSection (text,
																	font, colour,
																	passwordCharacter));
					break;
				}

				index = nextIndex;
			}

			if (nextIndex == insertIndex)
				sections.add (new UniformTextSection (text,
													  font, colour,
													  passwordCharacter));

			coalesceSimilarSections();
			totalNumChars = -1;
			valueTextNeedsUpdating = true;

			updateTextHolderSize();
			moveCaretTo (caretPositionToMoveTo, false);

			repaintText (Range<int> (insertIndex, getTotalNumChars()));
		}
	}
}

void TextEditor::reinsert (const int insertIndex,
						   const Array <UniformTextSection*>& sectionsToInsert)
{
	int index = 0;
	int nextIndex = 0;

	for (int i = 0; i < sections.size(); ++i)
	{
		nextIndex = index + sections.getUnchecked (i)->getTotalLength();

		if (insertIndex == index)
		{
			for (int j = sectionsToInsert.size(); --j >= 0;)
				sections.insert (i, new UniformTextSection (*sectionsToInsert.getUnchecked(j)));

			break;
		}
		else if (insertIndex > index && insertIndex < nextIndex)
		{
			splitSection (i, insertIndex - index);

			for (int j = sectionsToInsert.size(); --j >= 0;)
				sections.insert (i + 1, new UniformTextSection (*sectionsToInsert.getUnchecked(j)));

			break;
		}

		index = nextIndex;
	}

	if (nextIndex == insertIndex)
	{
		for (int j = 0; j < sectionsToInsert.size(); ++j)
			sections.add (new UniformTextSection (*sectionsToInsert.getUnchecked(j)));
	}

	coalesceSimilarSections();
	totalNumChars = -1;
	valueTextNeedsUpdating = true;
}

void TextEditor::remove (const Range<int>& range,
						 UndoManager* const um,
						 const int caretPositionToMoveTo)
{
	if (! range.isEmpty())
	{
		int index = 0;

		for (int i = 0; i < sections.size(); ++i)
		{
			const int nextIndex = index + sections.getUnchecked(i)->getTotalLength();

			if (range.getStart() > index && range.getStart() < nextIndex)
			{
				splitSection (i, range.getStart() - index);
				--i;
			}
			else if (range.getEnd() > index && range.getEnd() < nextIndex)
			{
				splitSection (i, range.getEnd() - index);
				--i;
			}
			else
			{
				index = nextIndex;

				if (index > range.getEnd())
					break;
			}
		}

		index = 0;

		if (um != nullptr)
		{
			Array <UniformTextSection*> removedSections;

			for (int i = 0; i < sections.size(); ++i)
			{
				if (range.getEnd() <= range.getStart())
					break;

				UniformTextSection* const section = sections.getUnchecked (i);

				const int nextIndex = index + section->getTotalLength();

				if (range.getStart() <= index && range.getEnd() >= nextIndex)
					removedSections.add (new UniformTextSection (*section));

				index = nextIndex;
			}

			if (um->getNumActionsInCurrentTransaction() > TextEditorDefs::maxActionsPerTransaction)
				newTransaction();

			um->perform (new RemoveAction (*this, range, caretPosition,
										   caretPositionToMoveTo, removedSections));
		}
		else
		{
			Range<int> remainingRange (range);

			for (int i = 0; i < sections.size(); ++i)
			{
				UniformTextSection* const section = sections.getUnchecked (i);

				const int nextIndex = index + section->getTotalLength();

				if (remainingRange.getStart() <= index && remainingRange.getEnd() >= nextIndex)
				{
					sections.remove(i);
					section->clear();
					delete section;

					remainingRange.setEnd (remainingRange.getEnd() - (nextIndex - index));
					if (remainingRange.isEmpty())
						break;

					--i;
				}
				else
				{
					index = nextIndex;
				}
			}

			coalesceSimilarSections();
			totalNumChars = -1;
			valueTextNeedsUpdating = true;

			moveCaretTo (caretPositionToMoveTo, false);

			repaintText (Range<int> (range.getStart(), getTotalNumChars()));
		}
	}
}

String TextEditor::getText() const
{
	MemoryOutputStream mo;
	mo.preallocate (getTotalNumChars());

	for (int i = 0; i < sections.size(); ++i)
		sections.getUnchecked (i)->appendAllText (mo);

	return mo.toString();
}

const String TextEditor::getTextInRange (const Range<int>& range) const
{
	if (range.isEmpty())
		return String::empty;

	MemoryOutputStream mo;
	mo.preallocate (jmin (getTotalNumChars(), range.getLength()));

	int index = 0;

	for (int i = 0; i < sections.size(); ++i)
	{
		const UniformTextSection* const s = sections.getUnchecked (i);
		const int nextIndex = index + s->getTotalLength();

		if (range.getStart() < nextIndex)
		{
			if (range.getEnd() <= index)
				break;

			s->appendSubstring (mo, range - index);
		}

		index = nextIndex;
	}

	return mo.toString();
}

String TextEditor::getHighlightedText() const
{
	return getTextInRange (selection);
}

int TextEditor::getTotalNumChars() const
{
	if (totalNumChars < 0)
	{
		totalNumChars = 0;

		for (int i = sections.size(); --i >= 0;)
			totalNumChars += sections.getUnchecked (i)->getTotalLength();
	}

	return totalNumChars;
}

bool TextEditor::isEmpty() const
{
	return getTotalNumChars() == 0;
}

void TextEditor::getCharPosition (const int index, float& cx, float& cy, float& lineHeight) const
{
	const float wordWrapWidth = getWordWrapWidth();

	if (wordWrapWidth > 0 && sections.size() > 0)
	{
		Iterator i (sections, wordWrapWidth, passwordCharacter);

		i.getCharPosition (index, cx, cy, lineHeight);
	}
	else
	{
		cx = cy = 0;
		lineHeight = currentFont.getHeight();
	}
}

int TextEditor::indexAtPosition (const float x, const float y)
{
	const float wordWrapWidth = getWordWrapWidth();

	if (wordWrapWidth > 0)
	{
		Iterator i (sections, wordWrapWidth, passwordCharacter);

		while (i.next())
		{
			if (i.lineY + i.lineHeight > y)
			{
				if (i.lineY > y)
					return jmax (0, i.indexInText - 1);

				if (i.atomX >= x)
					return i.indexInText;

				if (x < i.atomRight)
					return i.xToIndex (x);
			}
		}
	}

	return getTotalNumChars();
}

int TextEditor::findWordBreakAfter (const int position) const
{
	const String t (getTextInRange (Range<int> (position, position + 512)));
	const int totalLength = t.length();
	int i = 0;

	while (i < totalLength && CharacterFunctions::isWhitespace (t[i]))
		++i;

	const int type = TextEditorDefs::getCharacterCategory (t[i]);

	while (i < totalLength && type == TextEditorDefs::getCharacterCategory (t[i]))
		++i;

	while (i < totalLength && CharacterFunctions::isWhitespace (t[i]))
		++i;

	return position + i;
}

int TextEditor::findWordBreakBefore (const int position) const
{
	if (position <= 0)
		return 0;

	const int startOfBuffer = jmax (0, position - 512);
	const String t (getTextInRange (Range<int> (startOfBuffer, position)));

	int i = position - startOfBuffer;

	while (i > 0 && CharacterFunctions::isWhitespace (t [i - 1]))
		--i;

	if (i > 0)
	{
		const int type = TextEditorDefs::getCharacterCategory (t [i - 1]);

		while (i > 0 && type == TextEditorDefs::getCharacterCategory (t [i - 1]))
			--i;
	}

	jassert (startOfBuffer + i >= 0);
	return startOfBuffer + i;
}

void TextEditor::splitSection (const int sectionIndex,
							   const int charToSplitAt)
{
	jassert (sections[sectionIndex] != nullptr);

	sections.insert (sectionIndex + 1,
					 sections.getUnchecked (sectionIndex)->split (charToSplitAt, passwordCharacter));
}

void TextEditor::coalesceSimilarSections()
{
	for (int i = 0; i < sections.size() - 1; ++i)
	{
		UniformTextSection* const s1 = sections.getUnchecked (i);
		UniformTextSection* const s2 = sections.getUnchecked (i + 1);

		if (s1->font == s2->font
			 && s1->colour == s2->colour)
		{
			s1->append (*s2, passwordCharacter);
			sections.remove (i + 1);
			delete s2;
			--i;
		}
	}
}

void TextEditor::Listener::textEditorTextChanged (TextEditor&) {}
void TextEditor::Listener::textEditorReturnKeyPressed (TextEditor&) {}
void TextEditor::Listener::textEditorEscapeKeyPressed (TextEditor&) {}
void TextEditor::Listener::textEditorFocusLost (TextEditor&) {}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TextEditor.cpp ***/


/*** Start of inlined file: juce_Toolbar.cpp ***/
BEGIN_JUCE_NAMESPACE

const char* const Toolbar::toolbarDragDescriptor = "_toolbarItem_";

class ToolbarSpacerComp  : public ToolbarItemComponent
{
public:
	ToolbarSpacerComp (const int itemId_, const float fixedSize_, const bool drawBar_)
		: ToolbarItemComponent (itemId_, String::empty, false),
		  fixedSize (fixedSize_),
		  drawBar (drawBar_)
	{
	}

	~ToolbarSpacerComp()
	{
	}

	bool getToolbarItemSizes (int toolbarThickness, bool /*isToolbarVertical*/,
							  int& preferredSize, int& minSize, int& maxSize)
	{
		if (fixedSize <= 0)
		{
			preferredSize = toolbarThickness * 2;
			minSize = 4;
			maxSize = 32768;
		}
		else
		{
			maxSize = roundToInt (toolbarThickness * fixedSize);
			minSize = drawBar ? maxSize : jmin (4, maxSize);
			preferredSize = maxSize;

			if (getEditingMode() == editableOnPalette)
				preferredSize = maxSize = toolbarThickness / (drawBar ? 3 : 2);
		}

		return true;
	}

	void paintButtonArea (Graphics&, int, int, bool, bool)
	{
	}

	void contentAreaChanged (const Rectangle<int>&)
	{
	}

	int getResizeOrder() const noexcept
	{
		return fixedSize <= 0 ? 0 : 1;
	}

	void paint (Graphics& g)
	{
		const int w = getWidth();
		const int h = getHeight();

		if (drawBar)
		{
			g.setColour (findColour (Toolbar::separatorColourId, true));

			const float thickness = 0.2f;

			if (isToolbarVertical())
				g.fillRect (w * 0.1f, h * (0.5f - thickness * 0.5f), w * 0.8f, h * thickness);
			else
				g.fillRect (w * (0.5f - thickness * 0.5f), h * 0.1f, w * thickness, h * 0.8f);
		}

		if (getEditingMode() != normalMode && ! drawBar)
		{
			g.setColour (findColour (Toolbar::separatorColourId, true));

			const int indentX = jmin (2, (w - 3) / 2);
			const int indentY = jmin (2, (h - 3) / 2);
			g.drawRect (indentX, indentY, w - indentX * 2, h - indentY * 2, 1);

			if (fixedSize <= 0)
			{
				float x1, y1, x2, y2, x3, y3, x4, y4, hw, hl;

				if (isToolbarVertical())
				{
					x1 = w * 0.5f;
					y1 = h * 0.4f;
					x2 = x1;
					y2 = indentX * 2.0f;

					x3 = x1;
					y3 = h * 0.6f;
					x4 = x1;
					y4 = h - y2;

					hw = w * 0.15f;
					hl = w * 0.2f;
				}
				else
				{
					x1 = w * 0.4f;
					y1 = h * 0.5f;
					x2 = indentX * 2.0f;
					y2 = y1;

					x3 = w * 0.6f;
					y3 = y1;
					x4 = w - x2;
					y4 = y1;

					hw = h * 0.15f;
					hl = h * 0.2f;
				}

				Path p;
				p.addArrow (Line<float> (x1, y1, x2, y2), 1.5f, hw, hl);
				p.addArrow (Line<float> (x3, y3, x4, y4), 1.5f, hw, hl);
				g.fillPath (p);
			}
		}
	}

private:
	const float fixedSize;
	const bool drawBar;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ToolbarSpacerComp);
};

class Toolbar::MissingItemsComponent  : public PopupMenu::CustomComponent
{
public:
	MissingItemsComponent (Toolbar& owner_, const int height_)
		: PopupMenu::CustomComponent (true),
		  owner (&owner_),
		  height (height_)
	{
		for (int i = owner_.items.size(); --i >= 0;)
		{
			ToolbarItemComponent* const tc = owner_.items.getUnchecked(i);

			if (dynamic_cast <ToolbarSpacerComp*> (tc) == nullptr && ! tc->isVisible())
			{
				oldIndexes.insert (0, i);
				addAndMakeVisible (tc, 0);
			}
		}

		layout (400);
	}

	~MissingItemsComponent()
	{
		if (owner != nullptr)
		{
			for (int i = 0; i < getNumChildComponents(); ++i)
			{
				ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (getChildComponent (i));

				if (tc != nullptr)
				{
					tc->setVisible (false);
					const int index = oldIndexes.remove (i);
					owner->addChildComponent (tc, index);
					--i;
				}
			}

			owner->resized();
		}
	}

	void layout (const int preferredWidth)
	{
		const int indent = 8;
		int x = indent;
		int y = indent;
		int maxX = 0;

		for (int i = 0; i < getNumChildComponents(); ++i)
		{
			ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (getChildComponent (i));

			if (tc != nullptr)
			{
				int preferredSize = 1, minSize = 1, maxSize = 1;

				if (tc->getToolbarItemSizes (height, false, preferredSize, minSize, maxSize))
				{
					if (x + preferredSize > preferredWidth && x > indent)
					{
						x = indent;
						y += height;
					}

					tc->setBounds (x, y, preferredSize, height);

					x += preferredSize;
					maxX = jmax (maxX, x);
				}
			}
		}

		setSize (maxX + 8, y + height + 8);
	}

	void getIdealSize (int& idealWidth, int& idealHeight)
	{
		idealWidth = getWidth();
		idealHeight = getHeight();
	}

private:
	Component::SafePointer<Toolbar> owner;
	const int height;
	Array <int> oldIndexes;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (MissingItemsComponent);
};

Toolbar::Toolbar()
	: vertical (false),
	  isEditingActive (false),
	  toolbarStyle (Toolbar::iconsOnly)
{
	addChildComponent (missingItemsButton = getLookAndFeel().createToolbarMissingItemsButton (*this));

	missingItemsButton->setAlwaysOnTop (true);
	missingItemsButton->addListener (this);
}

Toolbar::~Toolbar()
{
	items.clear();
}

void Toolbar::setVertical (const bool shouldBeVertical)
{
	if (vertical != shouldBeVertical)
	{
		vertical = shouldBeVertical;
		resized();
	}
}

void Toolbar::clear()
{
	items.clear();
	resized();
}

ToolbarItemComponent* Toolbar::createItem (ToolbarItemFactory& factory, const int itemId)
{
	if (itemId == ToolbarItemFactory::separatorBarId)
		return new ToolbarSpacerComp (itemId, 0.1f, true);
	else if (itemId == ToolbarItemFactory::spacerId)
		return new ToolbarSpacerComp (itemId, 0.5f, false);
	else if (itemId == ToolbarItemFactory::flexibleSpacerId)
		return new ToolbarSpacerComp (itemId, 0, false);

	return factory.createItem (itemId);
}

void Toolbar::addItemInternal (ToolbarItemFactory& factory,
							   const int itemId,
							   const int insertIndex)
{
	// An ID can't be zero - this might indicate a mistake somewhere?
	jassert (itemId != 0);

	ToolbarItemComponent* const tc = createItem (factory, itemId);

	if (tc != nullptr)
	{
#if JUCE_DEBUG
		Array <int> allowedIds;
		factory.getAllToolbarItemIds (allowedIds);

		// If your factory can create an item for a given ID, it must also return
		// that ID from its getAllToolbarItemIds() method!
		jassert (allowedIds.contains (itemId));
#endif

		items.insert (insertIndex, tc);
		addAndMakeVisible (tc, insertIndex);
	}
}

void Toolbar::addItem (ToolbarItemFactory& factory,
					   const int itemId,
					   const int insertIndex)
{
	addItemInternal (factory, itemId, insertIndex);
	resized();
}

void Toolbar::addDefaultItems (ToolbarItemFactory& factoryToUse)
{
	Array <int> ids;
	factoryToUse.getDefaultItemSet (ids);

	clear();

	for (int i = 0; i < ids.size(); ++i)
		addItemInternal (factoryToUse, ids.getUnchecked (i), -1);

	resized();
}

void Toolbar::removeToolbarItem (const int itemIndex)
{
	items.remove (itemIndex);
	resized();
}

int Toolbar::getNumItems() const noexcept
{
	return items.size();
}

int Toolbar::getItemId (const int itemIndex) const noexcept
{
	ToolbarItemComponent* const tc = getItemComponent (itemIndex);
	return tc != nullptr ? tc->getItemId() : 0;
}

ToolbarItemComponent* Toolbar::getItemComponent (const int itemIndex) const noexcept
{
	return items [itemIndex];
}

ToolbarItemComponent* Toolbar::getNextActiveComponent (int index, const int delta) const
{
	for (;;)
	{
		index += delta;
		ToolbarItemComponent* const tc = getItemComponent (index);

		if (tc == nullptr)
			break;

		if (tc->isActive)
			return tc;
	}

	return nullptr;
}

void Toolbar::setStyle (const ToolbarItemStyle& newStyle)
{
	if (toolbarStyle != newStyle)
	{
		toolbarStyle = newStyle;
		updateAllItemPositions (false);
	}
}

String Toolbar::toString() const
{
	String s ("TB:");

	for (int i = 0; i < getNumItems(); ++i)
		s << getItemId(i) << ' ';

	return s.trimEnd();
}

bool Toolbar::restoreFromString (ToolbarItemFactory& factoryToUse,
								 const String& savedVersion)
{
	if (! savedVersion.startsWith ("TB:"))
		return false;

	StringArray tokens;
	tokens.addTokens (savedVersion.substring (3), false);

	clear();

	for (int i = 0; i < tokens.size(); ++i)
		addItemInternal (factoryToUse, tokens[i].getIntValue(), -1);

	resized();
	return true;
}

void Toolbar::paint (Graphics& g)
{
	getLookAndFeel().paintToolbarBackground (g, getWidth(), getHeight(), *this);
}

int Toolbar::getThickness() const noexcept
{
	return vertical ? getWidth() : getHeight();
}

int Toolbar::getLength() const noexcept
{
	return vertical ? getHeight() : getWidth();
}

void Toolbar::setEditingActive (const bool active)
{
	if (isEditingActive != active)
	{
		isEditingActive = active;
		updateAllItemPositions (false);
	}
}

void Toolbar::resized()
{
	updateAllItemPositions (false);
}

void Toolbar::updateAllItemPositions (const bool animate)
{
	if (getWidth() > 0 && getHeight() > 0)
	{
		StretchableObjectResizer resizer;

		int i;
		for (i = 0; i < items.size(); ++i)
		{
			ToolbarItemComponent* const tc = items.getUnchecked(i);

			tc->setEditingMode (isEditingActive ? ToolbarItemComponent::editableOnToolbar
												: ToolbarItemComponent::normalMode);

			tc->setStyle (toolbarStyle);

			ToolbarSpacerComp* const spacer = dynamic_cast <ToolbarSpacerComp*> (tc);

			int preferredSize = 1, minSize = 1, maxSize = 1;

			if (tc->getToolbarItemSizes (getThickness(), isVertical(),
										 preferredSize, minSize, maxSize))
			{
				tc->isActive = true;
				resizer.addItem (preferredSize, minSize, maxSize,
								 spacer != nullptr ? spacer->getResizeOrder() : 2);
			}
			else
			{
				tc->isActive = false;
				tc->setVisible (false);
			}
		}

		resizer.resizeToFit (getLength());

		int totalLength = 0;

		for (i = 0; i < resizer.getNumItems(); ++i)
			totalLength += (int) resizer.getItemSize (i);

		const bool itemsOffTheEnd = totalLength > getLength();

		const int extrasButtonSize = getThickness() / 2;
		missingItemsButton->setSize (extrasButtonSize, extrasButtonSize);
		missingItemsButton->setVisible (itemsOffTheEnd);
		missingItemsButton->setEnabled (! isEditingActive);

		if (vertical)
			missingItemsButton->setCentrePosition (getWidth() / 2,
												   getHeight() - 4 - extrasButtonSize / 2);
		else
			missingItemsButton->setCentrePosition (getWidth() - 4 - extrasButtonSize / 2,
												   getHeight() / 2);

		const int maxLength = itemsOffTheEnd ? (vertical ? missingItemsButton->getY()
														 : missingItemsButton->getX()) - 4
											 : getLength();

		int pos = 0, activeIndex = 0;
		for (i = 0; i < items.size(); ++i)
		{
			ToolbarItemComponent* const tc = items.getUnchecked(i);

			if (tc->isActive)
			{
				const int size = (int) resizer.getItemSize (activeIndex++);

				Rectangle<int> newBounds;
				if (vertical)
					newBounds.setBounds (0, pos, getWidth(), size);
				else
					newBounds.setBounds (pos, 0, size, getHeight());

				if (animate)
				{
					Desktop::getInstance().getAnimator().animateComponent (tc, newBounds, 1.0f, 200, false, 3.0, 0.0);
				}
				else
				{
					Desktop::getInstance().getAnimator().cancelAnimation (tc, false);
					tc->setBounds (newBounds);
				}

				pos += size;
				tc->setVisible (pos <= maxLength
								 && ((! tc->isBeingDragged)
									  || tc->getEditingMode() == ToolbarItemComponent::editableOnPalette));
			}
		}
	}
}

void Toolbar::buttonClicked (Button*)
{
	jassert (missingItemsButton->isShowing());

	if (missingItemsButton->isShowing())
	{
		PopupMenu m;
		m.addCustomItem (1, new MissingItemsComponent (*this, getThickness()));
		m.showMenuAsync (PopupMenu::Options().withTargetComponent (missingItemsButton), nullptr);
	}
}

bool Toolbar::isInterestedInDragSource (const SourceDetails& dragSourceDetails)
{
	return dragSourceDetails.description == toolbarDragDescriptor && isEditingActive;
}

void Toolbar::itemDragMove (const SourceDetails& dragSourceDetails)
{
	ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (dragSourceDetails.sourceComponent.get());

	if (tc != nullptr)
	{
		if (! items.contains (tc))
		{
			if (tc->getEditingMode() == ToolbarItemComponent::editableOnPalette)
			{
				ToolbarItemPalette* const palette = tc->findParentComponentOfClass ((ToolbarItemPalette*) nullptr);

				if (palette != nullptr)
					palette->replaceComponent (tc);
			}
			else
			{
				jassert (tc->getEditingMode() == ToolbarItemComponent::editableOnToolbar);
			}

			items.add (tc);
			addChildComponent (tc);
			updateAllItemPositions (true);
		}

		for (int i = getNumItems(); --i >= 0;)
		{
			const int currentIndex = items.indexOf (tc);
			int newIndex = currentIndex;

			const int dragObjectLeft = vertical ? (dragSourceDetails.localPosition.getY() - tc->dragOffsetY)
												: (dragSourceDetails.localPosition.getX() - tc->dragOffsetX);
			const int dragObjectRight = dragObjectLeft + (vertical ? tc->getHeight() : tc->getWidth());

			const Rectangle<int> current (Desktop::getInstance().getAnimator()
											.getComponentDestination (getChildComponent (newIndex)));
			ToolbarItemComponent* const prev = getNextActiveComponent (newIndex, -1);

			if (prev != nullptr)
			{
				const Rectangle<int> previousPos (Desktop::getInstance().getAnimator().getComponentDestination (prev));

				if (abs (dragObjectLeft - (vertical ? previousPos.getY() : previousPos.getX())
					  < abs (dragObjectRight - (vertical ? current.getBottom() : current.getRight()))))
				{
					newIndex = getIndexOfChildComponent (prev);
				}
			}

			ToolbarItemComponent* const next = getNextActiveComponent (newIndex, 1);
			if (next != nullptr)
			{
				const Rectangle<int> nextPos (Desktop::getInstance().getAnimator().getComponentDestination (next));

				if (abs (dragObjectLeft - (vertical ? current.getY() : current.getX())
					 > abs (dragObjectRight - (vertical ? nextPos.getBottom() : nextPos.getRight()))))
				{
					newIndex = getIndexOfChildComponent (next) + 1;
				}
			}

			if (newIndex == currentIndex)
				break;

			items.removeObject (tc, false);
			removeChildComponent (tc);
			addChildComponent (tc, newIndex);
			items.insert (newIndex, tc);
			updateAllItemPositions (true);
		}
	}
}

void Toolbar::itemDragExit (const SourceDetails& dragSourceDetails)
{
	ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (dragSourceDetails.sourceComponent.get());

	if (tc != nullptr && isParentOf (tc))
	{
		items.removeObject (tc, false);
		removeChildComponent (tc);
		updateAllItemPositions (true);
	}
}

void Toolbar::itemDropped (const SourceDetails& dragSourceDetails)
{
	ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (dragSourceDetails.sourceComponent.get());

	if (tc != nullptr)
		tc->setState (Button::buttonNormal);
}

void Toolbar::mouseDown (const MouseEvent&)
{
}

class ToolbarCustomisationDialog   : public DialogWindow
{
public:
	ToolbarCustomisationDialog (ToolbarItemFactory& factory,
								Toolbar* const toolbar_,
								const int optionFlags)
		: DialogWindow (TRANS("Add/remove items from toolbar"), Colours::white, true, true),
		  toolbar (toolbar_)
	{
		setContentOwned (new CustomiserPanel (factory, toolbar, optionFlags), true);
		setResizable (true, true);
		setResizeLimits (400, 300, 1500, 1000);
		positionNearBar();
	}

	~ToolbarCustomisationDialog()
	{
		toolbar->setEditingActive (false);
	}

	void closeButtonPressed()
	{
		setVisible (false);
	}

	bool canModalEventBeSentToComponent (const Component* comp)
	{
		return toolbar->isParentOf (comp);
	}

	void positionNearBar()
	{
		const Rectangle<int> screenSize (toolbar->getParentMonitorArea());
		const int tbx = toolbar->getScreenX();
		const int tby = toolbar->getScreenY();
		const int gap = 8;

		int x, y;

		if (toolbar->isVertical())
		{
			y = tby;

			if (tbx > screenSize.getCentreX())
				x = tbx - getWidth() - gap;
			else
				x = tbx + toolbar->getWidth() + gap;
		}
		else
		{
			x = tbx + (toolbar->getWidth() - getWidth()) / 2;

			if (tby > screenSize.getCentreY())
				y = tby - getHeight() - gap;
			else
				y = tby + toolbar->getHeight() + gap;
		}

		setTopLeftPosition (x, y);
	}

private:
	Toolbar* const toolbar;

	class CustomiserPanel  : public Component,
							 private ComboBoxListener, // (can't use ComboBox::Listener due to idiotic VC2005 bug)
							 private ButtonListener
	{
	public:
		CustomiserPanel (ToolbarItemFactory& factory_,
						 Toolbar* const toolbar_,
						 const int optionFlags)
		  : factory (factory_),
			toolbar (toolbar_),
			palette (factory_, toolbar_),
			instructions (String::empty, TRANS ("You can drag the items above and drop them onto a toolbar to add them.\n\n"
												"Items on the toolbar can also be dragged around to change their order, or dragged off the edge to delete them.")),
			defaultButton (TRANS ("Restore to default set of items"))
		{
			addAndMakeVisible (&palette);

			if ((optionFlags & (Toolbar::allowIconsOnlyChoice
								 | Toolbar::allowIconsWithTextChoice
								 | Toolbar::allowTextOnlyChoice)) != 0)
			{
				addAndMakeVisible (&styleBox);
				styleBox.setEditableText (false);

				if ((optionFlags & Toolbar::allowIconsOnlyChoice) != 0)     styleBox.addItem (TRANS("Show icons only"), 1);
				if ((optionFlags & Toolbar::allowIconsWithTextChoice) != 0) styleBox.addItem (TRANS("Show icons and descriptions"), 2);
				if ((optionFlags & Toolbar::allowTextOnlyChoice) != 0)      styleBox.addItem (TRANS("Show descriptions only"), 3);

				int selectedStyle = 0;
				switch (toolbar_->getStyle())
				{
					case Toolbar::iconsOnly:	selectedStyle = 1; break;
					case Toolbar::iconsWithText:	selectedStyle = 2; break;
					case Toolbar::textOnly:	 selectedStyle = 3; break;
				}

				styleBox.setSelectedId (selectedStyle);

				styleBox.addListener (this);
			}

			if ((optionFlags & Toolbar::showResetToDefaultsButton) != 0)
			{
				addAndMakeVisible (&defaultButton);
				defaultButton.addListener (this);
			}

			addAndMakeVisible (&instructions);
			instructions.setFont (Font (13.0f));

			setSize (500, 300);
		}

		void comboBoxChanged (ComboBox*)
		{
			switch (styleBox.getSelectedId())
			{
				case 1:   toolbar->setStyle (Toolbar::iconsOnly); break;
				case 2:   toolbar->setStyle (Toolbar::iconsWithText); break;
				case 3:   toolbar->setStyle (Toolbar::textOnly); break;
			}

			palette.resized(); // to make it update the styles
		}

		void buttonClicked (Button*)
		{
			toolbar->addDefaultItems (factory);
		}

		void paint (Graphics& g)
		{
			Colour background;

			DialogWindow* const dw = findParentComponentOfClass ((DialogWindow*) nullptr);

			if (dw != nullptr)
				background = dw->getBackgroundColour();

			g.setColour (background.contrasting().withAlpha (0.3f));
			g.fillRect (palette.getX(), palette.getBottom() - 1, palette.getWidth(), 1);
		}

		void resized()
		{
			palette.setBounds (0, 0, getWidth(), getHeight() - 120);
			styleBox.setBounds (10, getHeight() - 110, 200, 22);

			defaultButton.changeWidthToFitText (22);
			defaultButton.setTopLeftPosition (240, getHeight() - 110);

			instructions.setBounds (10, getHeight() - 80, getWidth() - 20, 80);
		}

	private:
		ToolbarItemFactory& factory;
		Toolbar* const toolbar;

		ToolbarItemPalette palette;
		Label instructions;
		ComboBox styleBox;
		TextButton defaultButton;
	};
};

void Toolbar::showCustomisationDialog (ToolbarItemFactory& factory, const int optionFlags)
{
	setEditingActive (true);

	(new ToolbarCustomisationDialog (factory, this, optionFlags))
		->enterModalState (true, 0, true);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Toolbar.cpp ***/


/*** Start of inlined file: juce_ToolbarItemComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

ToolbarItemFactory::ToolbarItemFactory()
{
}

ToolbarItemFactory::~ToolbarItemFactory()
{
}

class ItemDragAndDropOverlayComponent	: public Component
{
public:
	ItemDragAndDropOverlayComponent()
		: isDragging (false)
	{
		setAlwaysOnTop (true);
		setRepaintsOnMouseActivity (true);
		setMouseCursor (MouseCursor::DraggingHandCursor);
	}

	void paint (Graphics& g)
	{
		ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (getParentComponent());

		if (isMouseOverOrDragging()
			  && tc != nullptr
			  && tc->getEditingMode() == ToolbarItemComponent::editableOnToolbar)
		{
			g.setColour (findColour (Toolbar::editingModeOutlineColourId, true));
			g.drawRect (0, 0, getWidth(), getHeight(),
						jmin (2, (getWidth() - 1) / 2, (getHeight() - 1) / 2));
		}
	}

	void mouseDown (const MouseEvent& e)
	{
		isDragging = false;
		ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (getParentComponent());

		if (tc != nullptr)
		{
			tc->dragOffsetX = e.x;
			tc->dragOffsetY = e.y;
		}
	}

	void mouseDrag (const MouseEvent& e)
	{
		if (! (isDragging || e.mouseWasClicked()))
		{
			isDragging = true;
			DragAndDropContainer* const dnd = DragAndDropContainer::findParentDragContainerFor (this);

			if (dnd != nullptr)
			{
				dnd->startDragging (Toolbar::toolbarDragDescriptor, getParentComponent(), Image::null, true);

				ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (getParentComponent());

				if (tc != nullptr)
				{
					tc->isBeingDragged = true;

					if (tc->getEditingMode() == ToolbarItemComponent::editableOnToolbar)
						tc->setVisible (false);
				}
			}
		}
	}

	void mouseUp (const MouseEvent&)
	{
		isDragging = false;
		ToolbarItemComponent* const tc = dynamic_cast <ToolbarItemComponent*> (getParentComponent());

		if (tc != nullptr)
		{
			tc->isBeingDragged = false;

			Toolbar* const tb = tc->getToolbar();

			if (tb != nullptr)
				tb->updateAllItemPositions (true);
			else if (tc->getEditingMode() == ToolbarItemComponent::editableOnToolbar)
				delete tc;
		}
	}

	void parentSizeChanged()
	{
		setBounds (0, 0, getParentWidth(), getParentHeight());
	}

private:

	bool isDragging;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ItemDragAndDropOverlayComponent);
};

ToolbarItemComponent::ToolbarItemComponent (const int itemId_,
											const String& labelText,
											const bool isBeingUsedAsAButton_)
	: Button (labelText),
	  itemId (itemId_),
	  mode (normalMode),
	  toolbarStyle (Toolbar::iconsOnly),
	  dragOffsetX (0),
	  dragOffsetY (0),
	  isActive (true),
	  isBeingDragged (false),
	  isBeingUsedAsAButton (isBeingUsedAsAButton_)
{
	// Your item ID can't be 0!
	jassert (itemId_ != 0);
}

ToolbarItemComponent::~ToolbarItemComponent()
{
	overlayComp = nullptr;
}

Toolbar* ToolbarItemComponent::getToolbar() const
{
	return dynamic_cast <Toolbar*> (getParentComponent());
}

bool ToolbarItemComponent::isToolbarVertical() const
{
	const Toolbar* const t = getToolbar();
	return t != nullptr && t->isVertical();
}

void ToolbarItemComponent::setStyle (const Toolbar::ToolbarItemStyle& newStyle)
{
	if (toolbarStyle != newStyle)
	{
		toolbarStyle = newStyle;
		repaint();
		resized();
	}
}

void ToolbarItemComponent::paintButton (Graphics& g, const bool over, const bool down)
{
	if (isBeingUsedAsAButton)
		getLookAndFeel().paintToolbarButtonBackground (g, getWidth(), getHeight(),
													   over, down, *this);

	if (toolbarStyle != Toolbar::iconsOnly)
	{
		const int indent = contentArea.getX();
		int y = indent;
		int h = getHeight() - indent * 2;

		if (toolbarStyle == Toolbar::iconsWithText)
		{
			y = contentArea.getBottom() + indent / 2;
			h -= contentArea.getHeight();
		}

		getLookAndFeel().paintToolbarButtonLabel (g, indent, y, getWidth() - indent * 2, h,
												  getButtonText(), *this);
	}

	if (! contentArea.isEmpty())
	{
		Graphics::ScopedSaveState ss (g);

		g.reduceClipRegion (contentArea);
		g.setOrigin (contentArea.getX(), contentArea.getY());

		paintButtonArea (g, contentArea.getWidth(), contentArea.getHeight(), over, down);
	}
}

void ToolbarItemComponent::resized()
{
	if (toolbarStyle != Toolbar::textOnly)
	{
		const int indent = jmin (proportionOfWidth (0.08f),
								 proportionOfHeight (0.08f));

		contentArea = Rectangle<int> (indent, indent,
									  getWidth() - indent * 2,
									  toolbarStyle == Toolbar::iconsWithText ? proportionOfHeight (0.55f)
																			 : (getHeight() - indent * 2));
	}
	else
	{
		contentArea = Rectangle<int>();
	}

	contentAreaChanged (contentArea);
}

void ToolbarItemComponent::setEditingMode (const ToolbarEditingMode newMode)
{
	if (mode != newMode)
	{
		mode = newMode;
		repaint();

		if (mode == normalMode)
		{
			overlayComp = nullptr;
		}
		else if (overlayComp == nullptr)
		{
			addAndMakeVisible (overlayComp = new ItemDragAndDropOverlayComponent());
			overlayComp->parentSizeChanged();
		}

		resized();
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ToolbarItemComponent.cpp ***/


/*** Start of inlined file: juce_ToolbarItemPalette.cpp ***/
BEGIN_JUCE_NAMESPACE

ToolbarItemPalette::ToolbarItemPalette (ToolbarItemFactory& factory_,
										Toolbar* const toolbar_)
	: factory (factory_),
	  toolbar (toolbar_)
{
	Component* const itemHolder = new Component();
	viewport.setViewedComponent (itemHolder);

	Array <int> allIds;
	factory.getAllToolbarItemIds (allIds);

	for (int i = 0; i < allIds.size(); ++i)
		addComponent (allIds.getUnchecked (i), -1);

	addAndMakeVisible (&viewport);
}

ToolbarItemPalette::~ToolbarItemPalette()
{
}

void ToolbarItemPalette::addComponent (const int itemId, const int index)
{
	ToolbarItemComponent* const tc = Toolbar::createItem (factory, itemId);
	jassert (tc != nullptr);

	if (tc != nullptr)
	{
		items.insert (index, tc);
		viewport.getViewedComponent()->addAndMakeVisible (tc, index);
		tc->setEditingMode (ToolbarItemComponent::editableOnPalette);
	}
}

void ToolbarItemPalette::replaceComponent (ToolbarItemComponent* const comp)
{
	const int index = items.indexOf (comp);
	jassert (index >= 0);
	items.removeObject (comp, false);

	addComponent (comp->getItemId(), index);
	resized();
}

void ToolbarItemPalette::resized()
{
	viewport.setBoundsInset (BorderSize<int> (1));

	Component* const itemHolder = viewport.getViewedComponent();

	const int indent = 8;
	const int preferredWidth = viewport.getWidth() - viewport.getScrollBarThickness() - indent;
	const int height = toolbar->getThickness();
	int x = indent;
	int y = indent;
	int maxX = 0;

	for (int i = 0; i < items.size(); ++i)
	{
		ToolbarItemComponent* const tc = items.getUnchecked(i);

		tc->setStyle (toolbar->getStyle());

		int preferredSize = 1, minSize = 1, maxSize = 1;

		if (tc->getToolbarItemSizes (height, false, preferredSize, minSize, maxSize))
		{
			if (x + preferredSize > preferredWidth && x > indent)
			{
				x = indent;
				y += height;
			}

			tc->setBounds (x, y, preferredSize, height);

			x += preferredSize + 8;
			maxX = jmax (maxX, x);
		}
	}

	itemHolder->setSize (maxX, y + height + 8);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ToolbarItemPalette.cpp ***/


/*** Start of inlined file: juce_TreeView.cpp ***/
BEGIN_JUCE_NAMESPACE

class TreeViewContentComponent  : public Component,
								  public TooltipClient
{
public:
	TreeViewContentComponent (TreeView& owner_)
		: owner (owner_),
		  buttonUnderMouse (nullptr),
		  isDragging (false)
	{
	}

	void mouseDown (const MouseEvent& e)
	{
		updateButtonUnderMouse (e);

		isDragging = false;
		needSelectionOnMouseUp = false;

		Rectangle<int> pos;
		TreeViewItem* const item = findItemAt (e.y, pos);

		if (item == nullptr)
			return;

		// (if the open/close buttons are hidden, we'll treat clicks to the left of the item
		// as selection clicks)
		if (e.x < pos.getX() && owner.openCloseButtonsVisible)
		{
			if (e.x >= pos.getX() - owner.getIndentSize())
				item->setOpen (! item->isOpen());

			// (clicks to the left of an open/close button are ignored)
		}
		else
		{
			// mouse-down inside the body of the item..
			if (! owner.isMultiSelectEnabled())
				item->setSelected (true, true);
			else if (item->isSelected())
				needSelectionOnMouseUp = ! e.mods.isPopupMenu();
			else
				selectBasedOnModifiers (item, e.mods);

			if (e.x >= pos.getX())
				item->itemClicked (e.withNewPosition (e.getPosition() - pos.getPosition()));
		}
	}

	void mouseUp (const MouseEvent& e)
	{
		updateButtonUnderMouse (e);

		if (needSelectionOnMouseUp && e.mouseWasClicked())
		{
			Rectangle<int> pos;
			TreeViewItem* const item = findItemAt (e.y, pos);

			if (item != nullptr)
				selectBasedOnModifiers (item, e.mods);
		}
	}

	void mouseDoubleClick (const MouseEvent& e)
	{
		if (e.getNumberOfClicks() != 3)  // ignore triple clicks
		{
			Rectangle<int> pos;
			TreeViewItem* const item = findItemAt (e.y, pos);

			if (item != nullptr && (e.x >= pos.getX() || ! owner.openCloseButtonsVisible))
				item->itemDoubleClicked (e.withNewPosition (e.getPosition() - pos.getPosition()));
		}
	}

	void mouseDrag (const MouseEvent& e)
	{
		if (isEnabled()
			 && ! (isDragging || e.mouseWasClicked()
					|| e.getDistanceFromDragStart() < 5
					|| e.mods.isPopupMenu()))
		{
			isDragging = true;

			Rectangle<int> pos;
			TreeViewItem* const item = findItemAt (e.getMouseDownY(), pos);

			if (item != nullptr && e.getMouseDownX() >= pos.getX())
			{
				const var dragDescription (item->getDragSourceDescription());

				if (! (dragDescription.isVoid() || (dragDescription.isString() && dragDescription.toString().isEmpty())))
				{
					DragAndDropContainer* const dragContainer
						= DragAndDropContainer::findParentDragContainerFor (this);

					if (dragContainer != nullptr)
					{
						pos.setSize (pos.getWidth(), item->itemHeight);
						Image dragImage (Component::createComponentSnapshot (pos, true));
						dragImage.multiplyAllAlphas (0.6f);

						Point<int> imageOffset (pos.getPosition() - e.getPosition());
						dragContainer->startDragging (dragDescription, &owner, dragImage, true, &imageOffset);
					}
					else
					{
						// to be able to do a drag-and-drop operation, the treeview needs to
						// be inside a component which is also a DragAndDropContainer.
						jassertfalse;
					}
				}
			}
		}
	}

	void mouseMove (const MouseEvent& e)
	{
		updateButtonUnderMouse (e);
	}

	void mouseExit (const MouseEvent& e)
	{
		updateButtonUnderMouse (e);
	}

	void paint (Graphics& g)
	{
		if (owner.rootItem != nullptr)
		{
			owner.handleAsyncUpdate();

			if (! owner.rootItemVisible)
				g.setOrigin (0, -owner.rootItem->itemHeight);

			owner.rootItem->paintRecursively (g, getWidth());
		}
	}

	TreeViewItem* findItemAt (int y, Rectangle<int>& itemPosition) const
	{
		if (owner.rootItem != nullptr)
		{
			owner.handleAsyncUpdate();

			if (! owner.rootItemVisible)
				y += owner.rootItem->itemHeight;

			TreeViewItem* const ti = owner.rootItem->findItemRecursively (y);

			if (ti != nullptr)
				itemPosition = ti->getItemPosition (false);

			return ti;
		}

		return nullptr;
	}

	void updateComponents()
	{
		const int visibleTop = -getY();
		const int visibleBottom = visibleTop + getParentHeight();

		{
			for (int i = items.size(); --i >= 0;)
				items.getUnchecked(i)->shouldKeep = false;
		}

		{
			TreeViewItem* item = owner.rootItem;
			int y = (item != nullptr && ! owner.rootItemVisible) ? -item->itemHeight : 0;

			while (item != nullptr && y < visibleBottom)
			{
				y += item->itemHeight;

				if (y >= visibleTop)
				{
					RowItem* const ri = findItem (item->uid);

					if (ri != nullptr)
					{
						ri->shouldKeep = true;
					}
					else
					{
						Component* const comp = item->createItemComponent();

						if (comp != nullptr)
						{
							items.add (new RowItem (item, comp, item->uid));
							addAndMakeVisible (comp);
						}
					}
				}

				item = item->getNextVisibleItem (true);
			}
		}

		for (int i = items.size(); --i >= 0;)
		{
			RowItem* const ri = items.getUnchecked(i);
			bool keep = false;

			if (isParentOf (ri->component))
			{
				if (ri->shouldKeep)
				{
					Rectangle<int> pos (ri->item->getItemPosition (false));
					pos.setSize (pos.getWidth(), ri->item->itemHeight);

					if (pos.getBottom() >= visibleTop && pos.getY() < visibleBottom)
					{
						keep = true;
						ri->component->setBounds (pos);
					}
				}

				if ((! keep) && isMouseDraggingInChildCompOf (ri->component))
				{
					keep = true;
					ri->component->setSize (0, 0);
				}
			}

			if (! keep)
				items.remove (i);
		}
	}

	void updateButtonUnderMouse (const MouseEvent& e)
	{
		TreeViewItem* newItem = nullptr;

		if (owner.openCloseButtonsVisible)
		{
			Rectangle<int> pos;
			TreeViewItem* item = findItemAt (e.y, pos);

			if (item != nullptr && e.x < pos.getX() && e.x >= pos.getX() - owner.getIndentSize())
			{
				newItem = item;

				if (! newItem->mightContainSubItems())
					newItem = nullptr;
			}
		}

		if (buttonUnderMouse != newItem)
		{
			if (buttonUnderMouse != nullptr && containsItem (buttonUnderMouse))
			{
				const Rectangle<int> r (buttonUnderMouse->getItemPosition (false));
				repaint (0, r.getY(), r.getX(), buttonUnderMouse->getItemHeight());
			}

			buttonUnderMouse = newItem;

			if (buttonUnderMouse != nullptr)
			{
				const Rectangle<int> r (buttonUnderMouse->getItemPosition (false));
				repaint (0, r.getY(), r.getX(), buttonUnderMouse->getItemHeight());
			}
		}
	}

	bool isMouseOverButton (TreeViewItem* const item) const noexcept
	{
		return item == buttonUnderMouse;
	}

	void resized()
	{
		owner.itemsChanged();
	}

	const String getTooltip()
	{
		Rectangle<int> pos;
		TreeViewItem* const item = findItemAt (getMouseXYRelative().getY(), pos);

		if (item != nullptr)
			return item->getTooltip();

		return owner.getTooltip();
	}

private:

	TreeView& owner;

	struct RowItem
	{
		RowItem (TreeViewItem* const item_, Component* const component_, const int itemUID)
			: component (component_), item (item_), uid (itemUID), shouldKeep (true)
		{
		}

		~RowItem()
		{
			delete component.get();
		}

		WeakReference<Component> component;
		TreeViewItem* item;
		int uid;
		bool shouldKeep;
	};

	OwnedArray <RowItem> items;

	TreeViewItem* buttonUnderMouse;
	bool isDragging, needSelectionOnMouseUp;

	void selectBasedOnModifiers (TreeViewItem* const item, const ModifierKeys& modifiers)
	{
		TreeViewItem* firstSelected = nullptr;

		if (modifiers.isShiftDown() && ((firstSelected = owner.getSelectedItem (0)) != nullptr))
		{
			TreeViewItem* const lastSelected = owner.getSelectedItem (owner.getNumSelectedItems() - 1);
			jassert (lastSelected != nullptr);

			int rowStart = firstSelected->getRowNumberInTree();
			int rowEnd = lastSelected->getRowNumberInTree();
			if (rowStart > rowEnd)
				std::swap (rowStart, rowEnd);

			int ourRow = item->getRowNumberInTree();
			int otherEnd = ourRow < rowEnd ? rowStart : rowEnd;

			if (ourRow > otherEnd)
				std::swap (ourRow, otherEnd);

			for (int i = ourRow; i <= otherEnd; ++i)
				owner.getItemOnRow (i)->setSelected (true, false);
		}
		else
		{
			const bool cmd = modifiers.isCommandDown();
			item->setSelected ((! cmd) || ! item->isSelected(), ! cmd);
		}
	}

	bool containsItem (TreeViewItem* const item) const noexcept
	{
		for (int i = items.size(); --i >= 0;)
			if (items.getUnchecked(i)->item == item)
				return true;

		return false;
	}

	RowItem* findItem (const int uid) const noexcept
	{
		for (int i = items.size(); --i >= 0;)
		{
			RowItem* const ri = items.getUnchecked(i);
			if (ri->uid == uid)
				return ri;
		}

		return nullptr;
	}

	static bool isMouseDraggingInChildCompOf (Component* const comp)
	{
		for (int i = Desktop::getInstance().getNumMouseSources(); --i >= 0;)
		{
			MouseInputSource* const source = Desktop::getInstance().getMouseSource(i);

			if (source->isDragging())
			{
				Component* const underMouse = source->getComponentUnderMouse();

				if (underMouse != nullptr && (comp == underMouse || comp->isParentOf (underMouse)))
					return true;
			}
		}

		return false;
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (TreeViewContentComponent);
};

class TreeView::TreeViewport  : public Viewport
{
public:
	TreeViewport() noexcept : lastX (-1)	{}

	void updateComponents (const bool triggerResize = false)
	{
		TreeViewContentComponent* const tvc = static_cast <TreeViewContentComponent*> (getViewedComponent());
		if (tvc != nullptr)
		{
			if (triggerResize)
				tvc->resized();
			else
				tvc->updateComponents();
		}

		repaint();
	}

	void visibleAreaChanged (const Rectangle<int>& newVisibleArea)
	{
		const bool hasScrolledSideways = (newVisibleArea.getX() != lastX);
		lastX = newVisibleArea.getX();
		updateComponents (hasScrolledSideways);
	}

private:

	int lastX;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (TreeViewport);
};

TreeView::TreeView (const String& name)
	: Component (name),
	  rootItem (nullptr),
	  indentSize (24),
	  defaultOpenness (false),
	  needsRecalculating (true),
	  rootItemVisible (true),
	  multiSelectEnabled (false),
	  openCloseButtonsVisible (true)
{
	addAndMakeVisible (viewport = new TreeViewport());
	viewport->setViewedComponent (new TreeViewContentComponent (*this));
	viewport->setWantsKeyboardFocus (false);
	setWantsKeyboardFocus (true);
}

TreeView::~TreeView()
{
	if (rootItem != nullptr)
		rootItem->setOwnerView (nullptr);
}

void TreeView::setRootItem (TreeViewItem* const newRootItem)
{
	if (rootItem != newRootItem)
	{
		if (newRootItem != nullptr)
		{
			jassert (newRootItem->ownerView == nullptr); // can't use a tree item in more than one tree at once..

			if (newRootItem->ownerView != nullptr)
				newRootItem->ownerView->setRootItem (nullptr);
		}

		if (rootItem != nullptr)
			rootItem->setOwnerView (nullptr);

		rootItem = newRootItem;

		if (newRootItem != nullptr)
			newRootItem->setOwnerView (this);

		needsRecalculating = true;
		handleAsyncUpdate();

		if (rootItem != nullptr && (defaultOpenness || ! rootItemVisible))
		{
			rootItem->setOpen (false); // force a re-open
			rootItem->setOpen (true);
		}
	}
}

void TreeView::deleteRootItem()
{
	const ScopedPointer <TreeViewItem> deleter (rootItem);
	setRootItem (nullptr);
}

void TreeView::setRootItemVisible (const bool shouldBeVisible)
{
	rootItemVisible = shouldBeVisible;

	if (rootItem != nullptr && (defaultOpenness || ! rootItemVisible))
	{
		rootItem->setOpen (false); // force a re-open
		rootItem->setOpen (true);
	}

	itemsChanged();
}

void TreeView::colourChanged()
{
	setOpaque (findColour (backgroundColourId).isOpaque());
	repaint();
}

void TreeView::setIndentSize (const int newIndentSize)
{
	if (indentSize != newIndentSize)
	{
		indentSize = newIndentSize;
		resized();
	}
}

void TreeView::setDefaultOpenness (const bool isOpenByDefault)
{
	if (defaultOpenness != isOpenByDefault)
	{
		defaultOpenness = isOpenByDefault;
		itemsChanged();
	}
}

void TreeView::setMultiSelectEnabled (const bool canMultiSelect)
{
	multiSelectEnabled = canMultiSelect;
}

void TreeView::setOpenCloseButtonsVisible (const bool shouldBeVisible)
{
	if (openCloseButtonsVisible != shouldBeVisible)
	{
		openCloseButtonsVisible = shouldBeVisible;
		itemsChanged();
	}
}

Viewport* TreeView::getViewport() const noexcept
{
	return viewport;
}

void TreeView::clearSelectedItems()
{
	if (rootItem != nullptr)
		rootItem->deselectAllRecursively();
}

int TreeView::getNumSelectedItems (int maximumDepthToSearchTo) const noexcept
{
	return rootItem != nullptr ? rootItem->countSelectedItemsRecursively (maximumDepthToSearchTo) : 0;
}

TreeViewItem* TreeView::getSelectedItem (const int index) const noexcept
{
	return rootItem != nullptr ? rootItem->getSelectedItemWithIndex (index) : 0;
}

int TreeView::getNumRowsInTree() const
{
	if (rootItem != nullptr)
		return rootItem->getNumRows() - (rootItemVisible ? 0 : 1);

	return 0;
}

TreeViewItem* TreeView::getItemOnRow (int index) const
{
	if (! rootItemVisible)
		++index;

	if (rootItem != nullptr && index >= 0)
		return rootItem->getItemOnRow (index);

	return nullptr;
}

TreeViewItem* TreeView::getItemAt (int y) const noexcept
{
	TreeViewContentComponent* const tc = static_cast <TreeViewContentComponent*> (viewport->getViewedComponent());
	Rectangle<int> pos;
	return tc->findItemAt (tc->getLocalPoint (this, Point<int> (0, y)).getY(), pos);
}

TreeViewItem* TreeView::findItemFromIdentifierString (const String& identifierString) const
{
	if (rootItem == nullptr)
		return nullptr;

	return rootItem->findItemFromIdentifierString (identifierString);
}

static void addAllSelectedItemIds (TreeViewItem* item, XmlElement& parent)
{
	if (item->isSelected())
		parent.createNewChildElement ("SELECTED")->setAttribute ("id", item->getItemIdentifierString());

	const int numSubItems = item->getNumSubItems();

	for (int i = 0; i < numSubItems; ++i)
		addAllSelectedItemIds (item->getSubItem(i), parent);
}

XmlElement* TreeView::getOpennessState (const bool alsoIncludeScrollPosition) const
{
	XmlElement* e = nullptr;

	if (rootItem != nullptr)
	{
		e = rootItem->getOpennessState();

		if (e != nullptr)
		{
			if (alsoIncludeScrollPosition)
				e->setAttribute ("scrollPos", viewport->getViewPositionY());

			addAllSelectedItemIds (rootItem, *e);
		}
	}

	return e;
}

void TreeView::restoreOpennessState (const XmlElement& newState, const bool restoreStoredSelection)
{
	if (rootItem != nullptr)
	{
		rootItem->restoreOpennessState (newState);

		if (newState.hasAttribute ("scrollPos"))
			viewport->setViewPosition (viewport->getViewPositionX(),
									   newState.getIntAttribute ("scrollPos"));

		if (restoreStoredSelection)
		{
			clearSelectedItems();

			forEachXmlChildElementWithTagName (newState, e, "SELECTED")
			{
				TreeViewItem* const item = rootItem->findItemFromIdentifierString (e->getStringAttribute ("id"));

				if (item != nullptr)
					item->setSelected (true, false);
			}
		}
	}
}

void TreeView::paint (Graphics& g)
{
	g.fillAll (findColour (backgroundColourId));
}

void TreeView::resized()
{
	viewport->setBounds (getLocalBounds());

	itemsChanged();
	handleAsyncUpdate();
}

void TreeView::enablementChanged()
{
	repaint();
}

void TreeView::moveSelectedRow (int delta)
{
	if (delta == 0)
		return;

	int rowSelected = 0;

	TreeViewItem* const firstSelected = getSelectedItem (0);
	if (firstSelected != nullptr)
		rowSelected = firstSelected->getRowNumberInTree();

	rowSelected = jlimit (0, getNumRowsInTree() - 1, rowSelected + delta);

	for (;;)
	{
		TreeViewItem* item = getItemOnRow (rowSelected);

		if (item != nullptr)
		{
			if (! item->canBeSelected())
			{
				// if the row we want to highlight doesn't allow it, try skipping
				// to the next item..
				const int nextRowToTry = jlimit (0, getNumRowsInTree() - 1,
												 rowSelected + (delta < 0 ? -1 : 1));

				if (rowSelected != nextRowToTry)
				{
					rowSelected = nextRowToTry;
					continue;
				}
				else
				{
					break;
				}
			}

			item->setSelected (true, true);

			scrollToKeepItemVisible (item);
		}

		break;
	}
}

void TreeView::scrollToKeepItemVisible (TreeViewItem* item)
{
	if (item != nullptr && item->ownerView == this)
	{
		handleAsyncUpdate();

		item = item->getDeepestOpenParentItem();

		int y = item->y;
		int viewTop = viewport->getViewPositionY();

		if (y < viewTop)
		{
			viewport->setViewPosition (viewport->getViewPositionX(), y);
		}
		else if (y + item->itemHeight > viewTop + viewport->getViewHeight())
		{
			viewport->setViewPosition (viewport->getViewPositionX(),
									   (y + item->itemHeight) - viewport->getViewHeight());
		}
	}
}

bool TreeView::keyPressed (const KeyPress& key)
{
	if (key.isKeyCode (KeyPress::upKey))
	{
		moveSelectedRow (-1);
	}
	else if (key.isKeyCode (KeyPress::downKey))
	{
		moveSelectedRow (1);
	}
	else if (key.isKeyCode (KeyPress::pageDownKey) || key.isKeyCode (KeyPress::pageUpKey))
	{
		if (rootItem != nullptr)
		{
			int rowsOnScreen = getHeight() / jmax (1, rootItem->itemHeight);

			if (key.isKeyCode (KeyPress::pageUpKey))
				rowsOnScreen = -rowsOnScreen;

			moveSelectedRow (rowsOnScreen);
		}
	}
	else if (key.isKeyCode (KeyPress::homeKey))
	{
		moveSelectedRow (-0x3fffffff);
	}
	else if (key.isKeyCode (KeyPress::endKey))
	{
		moveSelectedRow (0x3fffffff);
	}
	else if (key.isKeyCode (KeyPress::returnKey))
	{
		TreeViewItem* const firstSelected = getSelectedItem (0);
		if (firstSelected != nullptr)
			firstSelected->setOpen (! firstSelected->isOpen());
	}
	else if (key.isKeyCode (KeyPress::leftKey))
	{
		TreeViewItem* const firstSelected = getSelectedItem (0);

		if (firstSelected != nullptr)
		{
			if (firstSelected->isOpen())
			{
				firstSelected->setOpen (false);
			}
			else
			{
				TreeViewItem* parent = firstSelected->parentItem;

				if ((! rootItemVisible) && parent == rootItem)
					parent = nullptr;

				if (parent != nullptr)
				{
					parent->setSelected (true, true);
					scrollToKeepItemVisible (parent);
				}
			}
		}
	}
	else if (key.isKeyCode (KeyPress::rightKey))
	{
		TreeViewItem* const firstSelected = getSelectedItem (0);

		if (firstSelected != nullptr)
		{
			if (firstSelected->isOpen() || ! firstSelected->mightContainSubItems())
				moveSelectedRow (1);
			else
				firstSelected->setOpen (true);
		}
	}
	else
	{
		return false;
	}

	return true;
}

void TreeView::itemsChanged() noexcept
{
	needsRecalculating = true;
	repaint();
	triggerAsyncUpdate();
}

void TreeView::handleAsyncUpdate()
{
	if (needsRecalculating)
	{
		needsRecalculating = false;

		const ScopedLock sl (nodeAlterationLock);

		if (rootItem != nullptr)
			rootItem->updatePositions (rootItemVisible ? 0 : -rootItem->itemHeight);

		viewport->updateComponents();

		if (rootItem != nullptr)
		{
			viewport->getViewedComponent()
				->setSize (jmax (viewport->getMaximumVisibleWidth(), rootItem->totalWidth),
						   rootItem->totalHeight - (rootItemVisible ? 0 : rootItem->itemHeight));
		}
		else
		{
			viewport->getViewedComponent()->setSize (0, 0);
		}
	}
}

class TreeView::InsertPointHighlight   : public Component
{
public:
	InsertPointHighlight()
		: lastItem (nullptr)
	{
		setSize (100, 12);
		setAlwaysOnTop (true);
		setInterceptsMouseClicks (false, false);
	}

	void setTargetPosition (TreeViewItem* const item, int insertIndex, const int x, const int y, const int width) noexcept
	{
		lastItem = item;
		lastIndex = insertIndex;
		const int offset = getHeight() / 2;
		setBounds (x - offset, y - offset, width - (x - offset), getHeight());
	}

	void paint (Graphics& g)
	{
		Path p;
		const float h = (float) getHeight();
		p.addEllipse (2.0f, 2.0f, h - 4.0f, h - 4.0f);
		p.startNewSubPath (h - 2.0f, h / 2.0f);
		p.lineTo ((float) getWidth(), h / 2.0f);

		g.setColour (findColour (TreeView::dragAndDropIndicatorColourId, true));
		g.strokePath (p, PathStrokeType (2.0f));
	}

	TreeViewItem* lastItem;
	int lastIndex;

private:
	JUCE_DECLARE_NON_COPYABLE (InsertPointHighlight);
};

class TreeView::TargetGroupHighlight   : public Component
{
public:
	TargetGroupHighlight()
	{
		setAlwaysOnTop (true);
		setInterceptsMouseClicks (false, false);
	}

	void setTargetPosition (TreeViewItem* const item) noexcept
	{
		Rectangle<int> r (item->getItemPosition (true));
		r.setHeight (item->getItemHeight());
		setBounds (r);
	}

	void paint (Graphics& g)
	{
		g.setColour (findColour (TreeView::dragAndDropIndicatorColourId, true));
		g.drawRoundedRectangle (1.0f, 1.0f, getWidth() - 2.0f, getHeight() - 2.0f, 3.0f, 2.0f);
	}

private:
	JUCE_DECLARE_NON_COPYABLE (TargetGroupHighlight);
};

void TreeView::showDragHighlight (TreeViewItem* item, int insertIndex, int x, int y) noexcept
{
	beginDragAutoRepeat (100);

	if (dragInsertPointHighlight == nullptr)
	{
		addAndMakeVisible (dragInsertPointHighlight = new InsertPointHighlight());
		addAndMakeVisible (dragTargetGroupHighlight = new TargetGroupHighlight());
	}

	dragInsertPointHighlight->setTargetPosition (item, insertIndex, x, y, viewport->getViewWidth());
	dragTargetGroupHighlight->setTargetPosition (item);
}

void TreeView::hideDragHighlight() noexcept
{
	dragInsertPointHighlight = nullptr;
	dragTargetGroupHighlight = nullptr;
}

TreeViewItem* TreeView::getInsertPosition (int& x, int& y, int& insertIndex,
										   const StringArray& files, const SourceDetails& dragSourceDetails) const noexcept
{
	x = dragSourceDetails.localPosition.getX();
	y = dragSourceDetails.localPosition.getY();
	insertIndex = 0;
	TreeViewItem* item = getItemAt (y);

	if (item == nullptr)
		return nullptr;

	Rectangle<int> itemPos (item->getItemPosition (true));
	insertIndex = item->getIndexInParent();
	const int oldY = y;
	y = itemPos.getY();

	if (item->getNumSubItems() == 0 || ! item->isOpen())
	{
		if (files.size() > 0 ? item->isInterestedInFileDrag (files)
							 : item->isInterestedInDragSource (dragSourceDetails))
		{
			// Check if we're trying to drag into an empty group item..
			if (oldY > itemPos.getY() + itemPos.getHeight() / 4
				 && oldY < itemPos.getBottom() - itemPos.getHeight() / 4)
			{
				insertIndex = 0;
				x = itemPos.getX() + getIndentSize();
				y = itemPos.getBottom();
				return item;
			}
		}
	}

	if (oldY > itemPos.getCentreY())
	{
		y += item->getItemHeight();

		while (item->isLastOfSiblings() && item->parentItem != nullptr
				&& item->parentItem->parentItem != nullptr)
		{
			if (x > itemPos.getX())
				break;

			item = item->parentItem;
			itemPos = item->getItemPosition (true);
			insertIndex = item->getIndexInParent();
		}

		++insertIndex;
	}

	x = itemPos.getX();
	return item->parentItem;
}

void TreeView::handleDrag (const StringArray& files, const SourceDetails& dragSourceDetails)
{
	const bool scrolled = viewport->autoScroll (dragSourceDetails.localPosition.getX(),
												dragSourceDetails.localPosition.getY(), 20, 10);

	int insertIndex, x, y;
	TreeViewItem* const item = getInsertPosition (x, y, insertIndex, files, dragSourceDetails);

	if (item != nullptr)
	{
		if (scrolled || dragInsertPointHighlight == nullptr
			 || dragInsertPointHighlight->lastItem != item
			 || dragInsertPointHighlight->lastIndex != insertIndex)
		{
			if (files.size() > 0 ? item->isInterestedInFileDrag (files)
								 : item->isInterestedInDragSource (dragSourceDetails))
				showDragHighlight (item, insertIndex, x, y);
			else
				hideDragHighlight();
		}
	}
	else
	{
		hideDragHighlight();
	}
}

void TreeView::handleDrop (const StringArray& files, const SourceDetails& dragSourceDetails)
{
	hideDragHighlight();

	int insertIndex, x, y;
	TreeViewItem* item = getInsertPosition (x, y, insertIndex, files, dragSourceDetails);

	if (item == nullptr)
	{
		insertIndex = 0;
		item = rootItem;
	}

	if (item != nullptr)
	{
		if (files.size() > 0)
		{
			if (item->isInterestedInFileDrag (files))
				item->filesDropped (files, insertIndex);
		}
		else
		{
			if (item->isInterestedInDragSource (dragSourceDetails))
				item->itemDropped (dragSourceDetails, insertIndex);
		}
	}
}

bool TreeView::isInterestedInFileDrag (const StringArray&)
{
	return true;
}

void TreeView::fileDragEnter (const StringArray& files, int x, int y)
{
	fileDragMove (files, x, y);
}

void TreeView::fileDragMove (const StringArray& files, int x, int y)
{
	handleDrag (files, SourceDetails (String::empty, 0, Point<int> (x, y)));
}

void TreeView::fileDragExit (const StringArray&)
{
	hideDragHighlight();
}

void TreeView::filesDropped (const StringArray& files, int x, int y)
{
	handleDrop (files, SourceDetails (String::empty, 0, Point<int> (x, y)));
}

bool TreeView::isInterestedInDragSource (const SourceDetails& /*dragSourceDetails*/)
{
	return true;
}

void TreeView::itemDragEnter (const SourceDetails& dragSourceDetails)
{
	itemDragMove (dragSourceDetails);
}

void TreeView::itemDragMove (const SourceDetails& dragSourceDetails)
{
	handleDrag (StringArray(), dragSourceDetails);
}

void TreeView::itemDragExit (const SourceDetails& /*dragSourceDetails*/)
{
	hideDragHighlight();
}

void TreeView::itemDropped (const SourceDetails& dragSourceDetails)
{
	handleDrop (StringArray(), dragSourceDetails);
}

enum TreeViewOpenness
{
	opennessDefault = 0,
	opennessClosed = 1,
	opennessOpen = 2
};

TreeViewItem::TreeViewItem()
	: ownerView (nullptr),
	  parentItem (nullptr),
	  y (0),
	  itemHeight (0),
	  totalHeight (0),
	  selected (false),
	  redrawNeeded (true),
	  drawLinesInside (true),
	  drawsInLeftMargin (false),
	  openness (opennessDefault)
{
	static int nextUID = 0;
	uid = nextUID++;
}

TreeViewItem::~TreeViewItem()
{
}

const String TreeViewItem::getUniqueName() const
{
	return String::empty;
}

void TreeViewItem::itemOpennessChanged (bool)
{
}

int TreeViewItem::getNumSubItems() const noexcept
{
	return subItems.size();
}

TreeViewItem* TreeViewItem::getSubItem (const int index) const noexcept
{
	return subItems [index];
}

void TreeViewItem::clearSubItems()
{
	if (subItems.size() > 0)
	{
		if (ownerView != nullptr)
		{
			const ScopedLock sl (ownerView->nodeAlterationLock);
			subItems.clear();
			treeHasChanged();
		}
		else
		{
			subItems.clear();
		}
	}
}

void TreeViewItem::addSubItem (TreeViewItem* const newItem, const int insertPosition)
{
	if (newItem != nullptr)
	{
		newItem->parentItem = this;
		newItem->setOwnerView (ownerView);
		newItem->y = 0;
		newItem->itemHeight = newItem->getItemHeight();
		newItem->totalHeight = 0;
		newItem->itemWidth = newItem->getItemWidth();
		newItem->totalWidth = 0;

		if (ownerView != nullptr)
		{
			const ScopedLock sl (ownerView->nodeAlterationLock);
			subItems.insert (insertPosition, newItem);
			treeHasChanged();

			if (newItem->isOpen())
				newItem->itemOpennessChanged (true);
		}
		else
		{
			subItems.insert (insertPosition, newItem);

			if (newItem->isOpen())
				newItem->itemOpennessChanged (true);
		}
	}
}

void TreeViewItem::removeSubItem (const int index, const bool deleteItem)
{
	if (ownerView != nullptr)
	{
		const ScopedLock sl (ownerView->nodeAlterationLock);

		if (isPositiveAndBelow (index, subItems.size()))
		{
			subItems.remove (index, deleteItem);
			treeHasChanged();
		}
	}
	else
	{
		subItems.remove (index, deleteItem);
	}
}

bool TreeViewItem::isOpen() const noexcept
{
	if (openness == opennessDefault)
		return ownerView != nullptr && ownerView->defaultOpenness;
	else
		return openness == opennessOpen;
}

void TreeViewItem::setOpen (const bool shouldBeOpen)
{
	if (isOpen() != shouldBeOpen)
	{
		openness = shouldBeOpen ? opennessOpen
								: opennessClosed;
		treeHasChanged();

		itemOpennessChanged (isOpen());
	}
}

bool TreeViewItem::isSelected() const noexcept
{
	return selected;
}

void TreeViewItem::deselectAllRecursively()
{
	setSelected (false, false);

	for (int i = 0; i < subItems.size(); ++i)
		subItems.getUnchecked(i)->deselectAllRecursively();
}

void TreeViewItem::setSelected (const bool shouldBeSelected,
								const bool deselectOtherItemsFirst)
{
	if (shouldBeSelected && ! canBeSelected())
		return;

	if (deselectOtherItemsFirst)
		getTopLevelItem()->deselectAllRecursively();

	if (shouldBeSelected != selected)
	{
		selected = shouldBeSelected;
		if (ownerView != nullptr)
			ownerView->repaint();

		itemSelectionChanged (shouldBeSelected);
	}
}

void TreeViewItem::paintItem (Graphics&, int, int)
{
}

void TreeViewItem::paintOpenCloseButton (Graphics& g, int width, int height, bool isMouseOver)
{
	ownerView->getLookAndFeel()
	   .drawTreeviewPlusMinusBox (g, 0, 0, width, height, ! isOpen(), isMouseOver);
}

void TreeViewItem::itemClicked (const MouseEvent&)
{
}

void TreeViewItem::itemDoubleClicked (const MouseEvent&)
{
	if (mightContainSubItems())
		setOpen (! isOpen());
}

void TreeViewItem::itemSelectionChanged (bool)
{
}

const String TreeViewItem::getTooltip()
{
	return String::empty;
}

const var TreeViewItem::getDragSourceDescription()
{
	return var::null;
}

bool TreeViewItem::isInterestedInFileDrag (const StringArray&)
{
	return false;
}

void TreeViewItem::filesDropped (const StringArray& /*files*/, int /*insertIndex*/)
{
}

bool TreeViewItem::isInterestedInDragSource (const DragAndDropTarget::SourceDetails& /*dragSourceDetails*/)
{
	return false;
}

void TreeViewItem::itemDropped (const DragAndDropTarget::SourceDetails& /*dragSourceDetails*/, int /*insertIndex*/)
{
}

const Rectangle<int> TreeViewItem::getItemPosition (const bool relativeToTreeViewTopLeft) const noexcept
{
	const int indentX = getIndentX();
	int width = itemWidth;

	if (ownerView != nullptr && width < 0)
		width = ownerView->viewport->getViewWidth() - indentX;

	Rectangle<int> r (indentX, y, jmax (0, width), totalHeight);

	if (relativeToTreeViewTopLeft)
		r -= ownerView->viewport->getViewPosition();

	return r;
}

void TreeViewItem::treeHasChanged() const noexcept
{
	if (ownerView != nullptr)
		ownerView->itemsChanged();
}

void TreeViewItem::repaintItem() const
{
	if (ownerView != nullptr && areAllParentsOpen())
	{
		Rectangle<int> r (getItemPosition (true));
		r.setLeft (0);
		ownerView->viewport->repaint (r);
	}
}

bool TreeViewItem::areAllParentsOpen() const noexcept
{
	return parentItem == nullptr
			|| (parentItem->isOpen() && parentItem->areAllParentsOpen());
}

void TreeViewItem::updatePositions (int newY)
{
	y = newY;
	itemHeight = getItemHeight();
	totalHeight = itemHeight;
	itemWidth = getItemWidth();
	totalWidth = jmax (itemWidth, 0) + getIndentX();

	if (isOpen())
	{
		newY += totalHeight;

		for (int i = 0; i < subItems.size(); ++i)
		{
			TreeViewItem* const ti = subItems.getUnchecked(i);

			ti->updatePositions (newY);
			newY += ti->totalHeight;
			totalHeight += ti->totalHeight;
			totalWidth = jmax (totalWidth, ti->totalWidth);
		}
	}
}

TreeViewItem* TreeViewItem::getDeepestOpenParentItem() noexcept
{
	TreeViewItem* result = this;
	TreeViewItem* item = this;

	while (item->parentItem != nullptr)
	{
		item = item->parentItem;

		if (! item->isOpen())
			result = item;
	}

	return result;
}

void TreeViewItem::setOwnerView (TreeView* const newOwner) noexcept
{
	ownerView = newOwner;

	for (int i = subItems.size(); --i >= 0;)
		subItems.getUnchecked(i)->setOwnerView (newOwner);
}

int TreeViewItem::getIndentX() const noexcept
{
	const int indentWidth = ownerView->getIndentSize();
	int x = ownerView->rootItemVisible ? indentWidth : 0;

	if (! ownerView->openCloseButtonsVisible)
		x -= indentWidth;

	TreeViewItem* p = parentItem;

	while (p != nullptr)
	{
		x += indentWidth;
		p = p->parentItem;
	}

	return x;
}

void TreeViewItem::setDrawsInLeftMargin (bool canDrawInLeftMargin) noexcept
{
	drawsInLeftMargin = canDrawInLeftMargin;
}

void TreeViewItem::paintRecursively (Graphics& g, int width)
{
	jassert (ownerView != nullptr);
	if (ownerView == nullptr)
		return;

	const int indent = getIndentX();
	const int itemW = itemWidth < 0 ? width - indent : itemWidth;

	{
		Graphics::ScopedSaveState ss (g);
		g.setOrigin (indent, 0);

		if (g.reduceClipRegion (drawsInLeftMargin ? -indent : 0, 0,
								drawsInLeftMargin ? itemW + indent : itemW, itemHeight))
			paintItem (g, itemW, itemHeight);
	}

	g.setColour (ownerView->findColour (TreeView::linesColourId));

	const float halfH = itemHeight * 0.5f;
	int depth = 0;
	TreeViewItem* p = parentItem;

	while (p != nullptr)
	{
		++depth;
		p = p->parentItem;
	}

	if (! ownerView->rootItemVisible)
		--depth;

	const int indentWidth = ownerView->getIndentSize();

	if (depth >= 0 && ownerView->openCloseButtonsVisible)
	{
		float x = (depth + 0.5f) * indentWidth;

		if (depth >= 0)
		{
			if (parentItem != nullptr && parentItem->drawLinesInside)
				g.drawLine (x, 0, x, isLastOfSiblings() ? halfH : (float) itemHeight);

			if ((parentItem != nullptr && parentItem->drawLinesInside)
				 || (parentItem == nullptr && drawLinesInside))
				g.drawLine (x, halfH, x + indentWidth / 2, halfH);
		}

		p = parentItem;
		int d = depth;

		while (p != nullptr && --d >= 0)
		{
			x -= (float) indentWidth;

			if ((p->parentItem == nullptr || p->parentItem->drawLinesInside)
				 && ! p->isLastOfSiblings())
			{
				g.drawLine (x, 0, x, (float) itemHeight);
			}

			p = p->parentItem;
		}

		if (mightContainSubItems())
		{
			Graphics::ScopedSaveState ss (g);

			g.setOrigin (depth * indentWidth, 0);
			g.reduceClipRegion (0, 0, indentWidth, itemHeight);

			paintOpenCloseButton (g, indentWidth, itemHeight,
				static_cast <TreeViewContentComponent*> (ownerView->viewport->getViewedComponent())
					->isMouseOverButton (this));
		}
	}

	if (isOpen())
	{
		const Rectangle<int> clip (g.getClipBounds());

		for (int i = 0; i < subItems.size(); ++i)
		{
			TreeViewItem* const ti = subItems.getUnchecked(i);

			const int relY = ti->y - y;

			if (relY >= clip.getBottom())
				break;

			if (relY + ti->totalHeight >= clip.getY())
			{
				Graphics::ScopedSaveState ss (g);

				g.setOrigin (0, relY);

				if (g.reduceClipRegion (0, 0, width, ti->totalHeight))
					ti->paintRecursively (g, width);
			}
		}
	}
}

bool TreeViewItem::isLastOfSiblings() const noexcept
{
	return parentItem == nullptr
			|| parentItem->subItems.getLast() == this;
}

int TreeViewItem::getIndexInParent() const noexcept
{
	return parentItem == nullptr ? 0
								 : parentItem->subItems.indexOf (this);
}

TreeViewItem* TreeViewItem::getTopLevelItem() noexcept
{
	return parentItem == nullptr ? this
								 : parentItem->getTopLevelItem();
}

int TreeViewItem::getNumRows() const noexcept
{
	int num = 1;

	if (isOpen())
	{
		for (int i = subItems.size(); --i >= 0;)
			num += subItems.getUnchecked(i)->getNumRows();
	}

	return num;
}

TreeViewItem* TreeViewItem::getItemOnRow (int index) noexcept
{
	if (index == 0)
		return this;

	if (index > 0 && isOpen())
	{
		--index;

		for (int i = 0; i < subItems.size(); ++i)
		{
			TreeViewItem* const item = subItems.getUnchecked(i);

			if (index == 0)
				return item;

			const int numRows = item->getNumRows();

			if (numRows > index)
				return item->getItemOnRow (index);

			index -= numRows;
		}
	}

	return nullptr;
}

TreeViewItem* TreeViewItem::findItemRecursively (int targetY) noexcept
{
	if (isPositiveAndBelow (targetY, totalHeight))
	{
		const int h = itemHeight;

		if (targetY < h)
			return this;

		if (isOpen())
		{
			targetY -= h;

			for (int i = 0; i < subItems.size(); ++i)
			{
				TreeViewItem* const ti = subItems.getUnchecked(i);

				if (targetY < ti->totalHeight)
					return ti->findItemRecursively (targetY);

				targetY -= ti->totalHeight;
			}
		}
	}

	return nullptr;
}

int TreeViewItem::countSelectedItemsRecursively (int depth) const noexcept
{
	int total = isSelected() ? 1 : 0;

	if (depth != 0)
		for (int i = subItems.size(); --i >= 0;)
			total += subItems.getUnchecked(i)->countSelectedItemsRecursively (depth - 1);

	return total;
}

TreeViewItem* TreeViewItem::getSelectedItemWithIndex (int index) noexcept
{
	if (isSelected())
	{
		if (index == 0)
			return this;

		--index;
	}

	if (index >= 0)
	{
		for (int i = 0; i < subItems.size(); ++i)
		{
			TreeViewItem* const item = subItems.getUnchecked(i);

			TreeViewItem* const found = item->getSelectedItemWithIndex (index);

			if (found != nullptr)
				return found;

			index -= item->countSelectedItemsRecursively (-1);
		}
	}

	return nullptr;
}

int TreeViewItem::getRowNumberInTree() const noexcept
{
	if (parentItem != nullptr && ownerView != nullptr)
	{
		int n = 1 + parentItem->getRowNumberInTree();

		int ourIndex = parentItem->subItems.indexOf (this);
		jassert (ourIndex >= 0);

		while (--ourIndex >= 0)
			n += parentItem->subItems [ourIndex]->getNumRows();

		if (parentItem->parentItem == nullptr
			 && ! ownerView->rootItemVisible)
			--n;

		return n;
	}
	else
	{
		return 0;
	}
}

void TreeViewItem::setLinesDrawnForSubItems (const bool drawLines) noexcept
{
	drawLinesInside = drawLines;
}

TreeViewItem* TreeViewItem::getNextVisibleItem (const bool recurse) const noexcept
{
	if (recurse && isOpen() && subItems.size() > 0)
		return subItems [0];

	if (parentItem != nullptr)
	{
		const int nextIndex = parentItem->subItems.indexOf (this) + 1;

		if (nextIndex >= parentItem->subItems.size())
			return parentItem->getNextVisibleItem (false);

		return parentItem->subItems [nextIndex];
	}

	return nullptr;
}

String TreeViewItem::getItemIdentifierString() const
{
	String s;

	if (parentItem != nullptr)
		s = parentItem->getItemIdentifierString();

	return s + "/" + getUniqueName().replaceCharacter ('/', '\\');
}

TreeViewItem* TreeViewItem::findItemFromIdentifierString (const String& identifierString)
{
	const String thisId (getUniqueName());

	if (thisId == identifierString)
		return this;

	if (identifierString.startsWith (thisId + "/"))
	{
		const String remainingPath (identifierString.substring (thisId.length() + 1));

		const bool wasOpen = isOpen();
		setOpen (true);

		for (int i = subItems.size(); --i >= 0;)
		{
			TreeViewItem* item = subItems.getUnchecked(i)->findItemFromIdentifierString (remainingPath);

			if (item != nullptr)
				return item;
		}

		setOpen (wasOpen);
	}

	return nullptr;
}

void TreeViewItem::restoreOpennessState (const XmlElement& e) noexcept
{
	if (e.hasTagName ("CLOSED"))
	{
		setOpen (false);
	}
	else if (e.hasTagName ("OPEN"))
	{
		setOpen (true);

		forEachXmlChildElement (e, n)
		{
			const String id (n->getStringAttribute ("id"));

			for (int i = 0; i < subItems.size(); ++i)
			{
				TreeViewItem* const ti = subItems.getUnchecked(i);

				if (ti->getUniqueName() == id)
				{
					ti->restoreOpennessState (*n);
					break;
				}
			}
		}
	}
}

XmlElement* TreeViewItem::getOpennessState() const noexcept
{
	const String name (getUniqueName());

	if (name.isNotEmpty())
	{
		XmlElement* e;

		if (isOpen())
		{
			e = new XmlElement ("OPEN");

			for (int i = 0; i < subItems.size(); ++i)
				e->addChildElement (subItems.getUnchecked(i)->getOpennessState());
		}
		else
		{
			e = new XmlElement ("CLOSED");
		}

		e->setAttribute ("id", name);

		return e;
	}
	else
	{
		// trying to save the openness for an element that has no name - this won't
		// work because it needs the names to identify what to open.
		jassertfalse;
	}

	return nullptr;
}

TreeViewItem::OpennessRestorer::OpennessRestorer (TreeViewItem& treeViewItem_)
	: treeViewItem (treeViewItem_),
	  oldOpenness (treeViewItem_.getOpennessState())
{
}

TreeViewItem::OpennessRestorer::~OpennessRestorer()
{
	if (oldOpenness != nullptr)
		treeViewItem.restoreOpennessState (*oldOpenness);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TreeView.cpp ***/


/*** Start of inlined file: juce_DirectoryContentsDisplayComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

DirectoryContentsDisplayComponent::DirectoryContentsDisplayComponent (DirectoryContentsList& listToShow)
	: fileList (listToShow)
{
}

DirectoryContentsDisplayComponent::~DirectoryContentsDisplayComponent()
{
}

FileBrowserListener::~FileBrowserListener()
{
}

void DirectoryContentsDisplayComponent::addListener (FileBrowserListener* const listener)
{
	listeners.add (listener);
}

void DirectoryContentsDisplayComponent::removeListener (FileBrowserListener* const listener)
{
	listeners.remove (listener);
}

void DirectoryContentsDisplayComponent::sendSelectionChangeMessage()
{
	Component::BailOutChecker checker (dynamic_cast <Component*> (this));
	listeners.callChecked (checker, &FileBrowserListener::selectionChanged);
}

void DirectoryContentsDisplayComponent::sendMouseClickMessage (const File& file, const MouseEvent& e)
{
	if (fileList.getDirectory().exists())
	{
		Component::BailOutChecker checker (dynamic_cast <Component*> (this));
		listeners.callChecked (checker, &FileBrowserListener::fileClicked, file, e);
	}
}

void DirectoryContentsDisplayComponent::sendDoubleClickMessage (const File& file)
{
	if (fileList.getDirectory().exists())
	{
		Component::BailOutChecker checker (dynamic_cast <Component*> (this));
		listeners.callChecked (checker, &FileBrowserListener::fileDoubleClicked, file);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DirectoryContentsDisplayComponent.cpp ***/


/*** Start of inlined file: juce_DirectoryContentsList.cpp ***/
BEGIN_JUCE_NAMESPACE

DirectoryContentsList::DirectoryContentsList (const FileFilter* const fileFilter_,
											  TimeSliceThread& thread_)
   : fileFilter (fileFilter_),
	 thread (thread_),
	 fileTypeFlags (File::ignoreHiddenFiles | File::findFiles),
	 shouldStop (true)
{
}

DirectoryContentsList::~DirectoryContentsList()
{
	clear();
}

void DirectoryContentsList::setIgnoresHiddenFiles (const bool shouldIgnoreHiddenFiles)
{
	setTypeFlags (shouldIgnoreHiddenFiles ? (fileTypeFlags | File::ignoreHiddenFiles)
										  : (fileTypeFlags & ~File::ignoreHiddenFiles));
}

bool DirectoryContentsList::ignoresHiddenFiles() const
{
	return (fileTypeFlags & File::ignoreHiddenFiles) != 0;
}

const File& DirectoryContentsList::getDirectory() const
{
	return root;
}

void DirectoryContentsList::setDirectory (const File& directory,
										  const bool includeDirectories,
										  const bool includeFiles)
{
	jassert (includeDirectories || includeFiles); // you have to speciify at least one of these!

	if (directory != root)
	{
		clear();
		root = directory;

		// (this forces a refresh when setTypeFlags() is called, rather than triggering two refreshes)
		fileTypeFlags &= ~(File::findDirectories | File::findFiles);
	}

	int newFlags = fileTypeFlags;
	if (includeDirectories) newFlags |= File::findDirectories;  else newFlags &= ~File::findDirectories;
	if (includeFiles)	   newFlags |= File::findFiles;	else newFlags &= ~File::findFiles;

	setTypeFlags (newFlags);
}

void DirectoryContentsList::setTypeFlags (const int newFlags)
{
	if (fileTypeFlags != newFlags)
	{
		fileTypeFlags = newFlags;
		refresh();
	}
}

void DirectoryContentsList::clear()
{
	shouldStop = true;
	thread.removeTimeSliceClient (this);

	fileFindHandle = nullptr;

	if (files.size() > 0)
	{
		files.clear();
		changed();
	}
}

void DirectoryContentsList::refresh()
{
	clear();

	if (root.isDirectory())
	{
		fileFindHandle = new DirectoryIterator (root, false, "*", fileTypeFlags);
		shouldStop = false;
		thread.addTimeSliceClient (this);
	}
}

int DirectoryContentsList::getNumFiles() const
{
	return files.size();
}

bool DirectoryContentsList::getFileInfo (const int index,
										 FileInfo& result) const
{
	const ScopedLock sl (fileListLock);
	const FileInfo* const info = files [index];

	if (info != nullptr)
	{
		result = *info;
		return true;
	}

	return false;
}

File DirectoryContentsList::getFile (const int index) const
{
	const ScopedLock sl (fileListLock);
	const FileInfo* const info = files [index];

	if (info != nullptr)
		return root.getChildFile (info->filename);

	return File::nonexistent;
}

bool DirectoryContentsList::isStillLoading() const
{
	return fileFindHandle != nullptr;
}

void DirectoryContentsList::changed()
{
	sendChangeMessage();
}

int DirectoryContentsList::useTimeSlice()
{
	const uint32 startTime = Time::getApproximateMillisecondCounter();
	bool hasChanged = false;

	for (int i = 100; --i >= 0;)
	{
		if (! checkNextFile (hasChanged))
		{
			if (hasChanged)
				changed();

			return 500;
		}

		if (shouldStop || (Time::getApproximateMillisecondCounter() > startTime + 150))
			break;
	}

	if (hasChanged)
		changed();

	return 0;
}

bool DirectoryContentsList::checkNextFile (bool& hasChanged)
{
	if (fileFindHandle != nullptr)
	{
		bool fileFoundIsDir, isHidden, isReadOnly;
		int64 fileSize;
		Time modTime, creationTime;

		if (fileFindHandle->next (&fileFoundIsDir, &isHidden, &fileSize,
								  &modTime, &creationTime, &isReadOnly))
		{
			if (addFile (fileFindHandle->getFile(), fileFoundIsDir,
						 fileSize, modTime, creationTime, isReadOnly))
			{
				hasChanged = true;
			}

			return true;
		}
		else
		{
			fileFindHandle = nullptr;
		}
	}

	return false;
}

int DirectoryContentsList::compareElements (const DirectoryContentsList::FileInfo* const first,
											const DirectoryContentsList::FileInfo* const second)
{
   #if JUCE_WINDOWS
	if (first->isDirectory != second->isDirectory)
		return first->isDirectory ? -1 : 1;
   #endif

	return first->filename.compareIgnoreCase (second->filename);
}

bool DirectoryContentsList::addFile (const File& file,
									 const bool isDir,
									 const int64 fileSize,
									 const Time& modTime,
									 const Time& creationTime,
									 const bool isReadOnly)
{
	if (fileFilter == nullptr
		 || ((! isDir) && fileFilter->isFileSuitable (file))
		 || (isDir && fileFilter->isDirectorySuitable (file)))
	{
		ScopedPointer <FileInfo> info (new FileInfo());

		info->filename = file.getFileName();
		info->fileSize = fileSize;
		info->modificationTime = modTime;
		info->creationTime = creationTime;
		info->isDirectory = isDir;
		info->isReadOnly = isReadOnly;

		const ScopedLock sl (fileListLock);

		for (int i = files.size(); --i >= 0;)
			if (files.getUnchecked(i)->filename == info->filename)
				return false;

		files.addSorted (*this, info.release());
		return true;
	}

	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DirectoryContentsList.cpp ***/



/*** Start of inlined file: juce_FileBrowserComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

FileBrowserComponent::FileBrowserComponent (int flags_,
											const File& initialFileOrDirectory,
											const FileFilter* fileFilter_,
											FilePreviewComponent* previewComp_)
   : FileFilter (String::empty),
	 fileFilter (fileFilter_),
	 flags (flags_),
	 previewComp (previewComp_),
	 currentPathBox ("path"),
	 fileLabel ("f", TRANS ("file:")),
	 thread ("Juce FileBrowser")
{
	// You need to specify one or other of the open/save flags..
	jassert ((flags & (saveMode | openMode)) != 0);
	jassert ((flags & (saveMode | openMode)) != (saveMode | openMode));

	// You need to specify at least one of these flags..
	jassert ((flags & (canSelectFiles | canSelectDirectories)) != 0);

	String filename;

	if (initialFileOrDirectory == File::nonexistent)
	{
		currentRoot = File::getCurrentWorkingDirectory();
	}
	else if (initialFileOrDirectory.isDirectory())
	{
		currentRoot = initialFileOrDirectory;
	}
	else
	{
		chosenFiles.add (initialFileOrDirectory);
		currentRoot = initialFileOrDirectory.getParentDirectory();
		filename = initialFileOrDirectory.getFileName();
	}

	fileList = new DirectoryContentsList (this, thread);

	if ((flags & useTreeView) != 0)
	{
		FileTreeComponent* const tree = new FileTreeComponent (*fileList);
		fileListComponent = tree;

		if ((flags & canSelectMultipleItems) != 0)
			tree->setMultiSelectEnabled (true);

		addAndMakeVisible (tree);
	}
	else
	{
		FileListComponent* const list = new FileListComponent (*fileList);
		fileListComponent = list;
		list->setOutlineThickness (1);

		if ((flags & canSelectMultipleItems) != 0)
			list->setMultipleSelectionEnabled (true);

		addAndMakeVisible (list);
	}

	fileListComponent->addListener (this);

	addAndMakeVisible (&currentPathBox);
	currentPathBox.setEditableText (true);
	resetRecentPaths();
	currentPathBox.addListener (this);

	addAndMakeVisible (&filenameBox);
	filenameBox.setMultiLine (false);
	filenameBox.setSelectAllWhenFocused (true);
	filenameBox.setText (filename, false);
	filenameBox.addListener (this);
	filenameBox.setReadOnly ((flags & (filenameBoxIsReadOnly | canSelectMultipleItems)) != 0);

	addAndMakeVisible (&fileLabel);
	fileLabel.attachToComponent (&filenameBox, true);

	addAndMakeVisible (goUpButton = getLookAndFeel().createFileBrowserGoUpButton());
	goUpButton->addListener (this);
	goUpButton->setTooltip (TRANS ("go up to parent directory"));

	if (previewComp != nullptr)
		addAndMakeVisible (previewComp);

	setRoot (currentRoot);

	thread.startThread (4);
}

FileBrowserComponent::~FileBrowserComponent()
{
	fileListComponent = nullptr;
	fileList = nullptr;
	thread.stopThread (10000);
}

void FileBrowserComponent::addListener (FileBrowserListener* const newListener)
{
	listeners.add (newListener);
}

void FileBrowserComponent::removeListener (FileBrowserListener* const listener)
{
	listeners.remove (listener);
}

bool FileBrowserComponent::isSaveMode() const noexcept
{
	return (flags & saveMode) != 0;
}

int FileBrowserComponent::getNumSelectedFiles() const noexcept
{
	if (chosenFiles.size() == 0 && currentFileIsValid())
		return 1;

	return chosenFiles.size();
}

File FileBrowserComponent::getSelectedFile (int index) const noexcept
{
	if ((flags & canSelectDirectories) != 0 && filenameBox.getText().isEmpty())
		return currentRoot;

	if (! filenameBox.isReadOnly())
		return currentRoot.getChildFile (filenameBox.getText());

	return chosenFiles[index];
}

bool FileBrowserComponent::currentFileIsValid() const
{
	if (isSaveMode())
		return ! getSelectedFile (0).isDirectory();
	else
		return getSelectedFile (0).exists();
}

File FileBrowserComponent::getHighlightedFile() const noexcept
{
	return fileListComponent->getSelectedFile (0);
}

void FileBrowserComponent::deselectAllFiles()
{
	fileListComponent->deselectAllFiles();
}

bool FileBrowserComponent::isFileSuitable (const File& file) const
{
	return (flags & canSelectFiles) != 0 && (fileFilter == nullptr || fileFilter->isFileSuitable (file));
}

bool FileBrowserComponent::isDirectorySuitable (const File&) const
{
	return true;
}

bool FileBrowserComponent::isFileOrDirSuitable (const File& f) const
{
	if (f.isDirectory())
		return (flags & canSelectDirectories) != 0
				&& (fileFilter == nullptr || fileFilter->isDirectorySuitable (f));

	return (flags & canSelectFiles) != 0 && f.exists()
			&& (fileFilter == nullptr || fileFilter->isFileSuitable (f));
}

const File& FileBrowserComponent::getRoot() const
{
	return currentRoot;
}

void FileBrowserComponent::setRoot (const File& newRootDirectory)
{
	if (currentRoot != newRootDirectory)
	{
		fileListComponent->scrollToTop();

		String path (newRootDirectory.getFullPathName());

		if (path.isEmpty())
			path = File::separatorString;

		StringArray rootNames, rootPaths;
		getRoots (rootNames, rootPaths);

		if (! rootPaths.contains (path, true))
		{
			bool alreadyListed = false;

			for (int i = currentPathBox.getNumItems(); --i >= 0;)
			{
				if (currentPathBox.getItemText (i).equalsIgnoreCase (path))
				{
					alreadyListed = true;
					break;
				}
			}

			if (! alreadyListed)
				currentPathBox.addItem (path, currentPathBox.getNumItems() + 2);
		}
	}

	currentRoot = newRootDirectory;
	fileList->setDirectory (currentRoot, true, true);

	String currentRootName (currentRoot.getFullPathName());
	if (currentRootName.isEmpty())
		currentRootName = File::separatorString;

	currentPathBox.setText (currentRootName, true);

	goUpButton->setEnabled (currentRoot.getParentDirectory().isDirectory()
							 && currentRoot.getParentDirectory() != currentRoot);
}

void FileBrowserComponent::resetRecentPaths()
{
	currentPathBox.clear();

	StringArray rootNames, rootPaths;
	getRoots (rootNames, rootPaths);

	for (int i = 0; i < rootNames.size(); ++i)
	{
		if (rootNames[i].isEmpty())
			currentPathBox.addSeparator();
		else
			currentPathBox.addItem (rootNames[i], i + 1);
	}

	currentPathBox.addSeparator();
}

void FileBrowserComponent::goUp()
{
	setRoot (getRoot().getParentDirectory());
}

void FileBrowserComponent::refresh()
{
	fileList->refresh();
}

void FileBrowserComponent::setFileFilter (const FileFilter* const newFileFilter)
{
	if (fileFilter != newFileFilter)
	{
		fileFilter = newFileFilter;
		refresh();
	}
}

const String FileBrowserComponent::getActionVerb() const
{
	return isSaveMode() ? TRANS("Save") : TRANS("Open");
}

FilePreviewComponent* FileBrowserComponent::getPreviewComponent() const noexcept
{
	return previewComp;
}

void FileBrowserComponent::resized()
{
	getLookAndFeel()
		.layoutFileBrowserComponent (*this, fileListComponent, previewComp,
									 &currentPathBox, &filenameBox, goUpButton);
}

void FileBrowserComponent::sendListenerChangeMessage()
{
	Component::BailOutChecker checker (this);

	if (previewComp != nullptr)
		previewComp->selectedFileChanged (getSelectedFile (0));

	// You shouldn't delete the browser when the file gets changed!
	jassert (! checker.shouldBailOut());

	listeners.callChecked (checker, &FileBrowserListener::selectionChanged);
}

void FileBrowserComponent::selectionChanged()
{
	StringArray newFilenames;
	bool resetChosenFiles = true;

	for (int i = 0; i < fileListComponent->getNumSelectedFiles(); ++i)
	{
		const File f (fileListComponent->getSelectedFile (i));

		if (isFileOrDirSuitable (f))
		{
			if (resetChosenFiles)
			{
				chosenFiles.clear();
				resetChosenFiles = false;
			}

			chosenFiles.add (f);
			newFilenames.add (f.getRelativePathFrom (getRoot()));
		}
	}

	if (newFilenames.size() > 0)
		filenameBox.setText (newFilenames.joinIntoString (", "), false);

	sendListenerChangeMessage();
}

void FileBrowserComponent::fileClicked (const File& f, const MouseEvent& e)
{
	Component::BailOutChecker checker (this);
	listeners.callChecked (checker, &FileBrowserListener::fileClicked, f, e);
}

void FileBrowserComponent::fileDoubleClicked (const File& f)
{
	if (f.isDirectory())
	{
		setRoot (f);

		if ((flags & canSelectDirectories) != 0)
			filenameBox.setText (String::empty);
	}
	else
	{
		Component::BailOutChecker checker (this);
		listeners.callChecked (checker, &FileBrowserListener::fileDoubleClicked, f);
	}
}

bool FileBrowserComponent::keyPressed (const KeyPress& key)
{
	(void) key;

#if JUCE_LINUX || JUCE_WINDOWS
	if (key.getModifiers().isCommandDown()
		 && (key.getKeyCode() == 'H' || key.getKeyCode() == 'h'))
	{
		fileList->setIgnoresHiddenFiles (! fileList->ignoresHiddenFiles());
		fileList->refresh();
		return true;
	}
#endif

	return false;
}

void FileBrowserComponent::textEditorTextChanged (TextEditor&)
{
	sendListenerChangeMessage();
}

void FileBrowserComponent::textEditorReturnKeyPressed (TextEditor&)
{
	if (filenameBox.getText().containsChar (File::separator))
	{
		const File f (currentRoot.getChildFile (filenameBox.getText()));

		if (f.isDirectory())
		{
			setRoot (f);
			chosenFiles.clear();
			filenameBox.setText (String::empty);
		}
		else
		{
			setRoot (f.getParentDirectory());
			chosenFiles.clear();
			chosenFiles.add (f);
			filenameBox.setText (f.getFileName());
		}
	}
	else
	{
		fileDoubleClicked (getSelectedFile (0));
	}
}

void FileBrowserComponent::textEditorEscapeKeyPressed (TextEditor&)
{
}

void FileBrowserComponent::textEditorFocusLost (TextEditor&)
{
	if (! isSaveMode())
		selectionChanged();
}

void FileBrowserComponent::buttonClicked (Button*)
{
	goUp();
}

void FileBrowserComponent::comboBoxChanged (ComboBox*)
{
	const String newText (currentPathBox.getText().trim().unquoted());

	if (newText.isNotEmpty())
	{
		const int index = currentPathBox.getSelectedId() - 1;

		StringArray rootNames, rootPaths;
		getRoots (rootNames, rootPaths);

		if (rootPaths [index].isNotEmpty())
		{
			setRoot (File (rootPaths [index]));
		}
		else
		{
			File f (newText);

			for (;;)
			{
				if (f.isDirectory())
				{
					setRoot (f);
					break;
				}

				if (f.getParentDirectory() == f)
					break;

				f = f.getParentDirectory();
			}
		}
	}
}

void FileBrowserComponent::getRoots (StringArray& rootNames, StringArray& rootPaths)
{
   #if JUCE_WINDOWS
	Array<File> roots;
	File::findFileSystemRoots (roots);
	rootPaths.clear();

	for (int i = 0; i < roots.size(); ++i)
	{
		const File& drive = roots.getReference(i);

		String name (drive.getFullPathName());
		rootPaths.add (name);

		if (drive.isOnHardDisk())
		{
			String volume (drive.getVolumeLabel());

			if (volume.isEmpty())
				volume = TRANS("Hard Drive");

			name << " [" << volume << ']';
		}
		else if (drive.isOnCDRomDrive())
		{
			name << TRANS(" [CD/DVD drive]");
		}

		rootNames.add (name);
	}

	rootPaths.add (String::empty);
	rootNames.add (String::empty);

	rootPaths.add (File::getSpecialLocation (File::userDocumentsDirectory).getFullPathName());
	rootNames.add ("Documents");
	rootPaths.add (File::getSpecialLocation (File::userDesktopDirectory).getFullPathName());
	rootNames.add ("Desktop");

   #elif JUCE_MAC
	rootPaths.add (File::getSpecialLocation (File::userHomeDirectory).getFullPathName());
	rootNames.add ("Home folder");
	rootPaths.add (File::getSpecialLocation (File::userDocumentsDirectory).getFullPathName());
	rootNames.add ("Documents");
	rootPaths.add (File::getSpecialLocation (File::userDesktopDirectory).getFullPathName());
	rootNames.add ("Desktop");

	rootPaths.add (String::empty);
	rootNames.add (String::empty);

	Array <File> volumes;
	File vol ("/Volumes");
	vol.findChildFiles (volumes, File::findDirectories, false);

	for (int i = 0; i < volumes.size(); ++i)
	{
		const File& volume = volumes.getReference(i);

		if (volume.isDirectory() && ! volume.getFileName().startsWithChar ('.'))
		{
			rootPaths.add (volume.getFullPathName());
			rootNames.add (volume.getFileName());
		}
	}

   #else
	rootPaths.add ("/");
	rootNames.add ("/");
	rootPaths.add (File::getSpecialLocation (File::userHomeDirectory).getFullPathName());
	rootNames.add ("Home folder");
	rootPaths.add (File::getSpecialLocation (File::userDesktopDirectory).getFullPathName());
	rootNames.add ("Desktop");
   #endif
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileBrowserComponent.cpp ***/


/*** Start of inlined file: juce_FileChooser.cpp ***/
BEGIN_JUCE_NAMESPACE

FileChooser::FileChooser (const String& chooserBoxTitle,
						  const File& currentFileOrDirectory,
						  const String& fileFilters,
						  const bool useNativeDialogBox_)
	: title (chooserBoxTitle),
	  filters (fileFilters),
	  startingFile (currentFileOrDirectory),
	  useNativeDialogBox (useNativeDialogBox_)
{
#if JUCE_LINUX
	useNativeDialogBox = false;
#endif

	if (! fileFilters.containsNonWhitespaceChars())
		filters = "*";
}

FileChooser::~FileChooser()
{
}

#if JUCE_MODAL_LOOPS_PERMITTED
bool FileChooser::browseForFileToOpen (FilePreviewComponent* previewComponent)
{
	return showDialog (false, true, false, false, false, previewComponent);
}

bool FileChooser::browseForMultipleFilesToOpen (FilePreviewComponent* previewComponent)
{
	return showDialog (false, true, false, false, true, previewComponent);
}

bool FileChooser::browseForMultipleFilesOrDirectories (FilePreviewComponent* previewComponent)
{
	return showDialog (true, true, false, false, true, previewComponent);
}

bool FileChooser::browseForFileToSave (const bool warnAboutOverwritingExistingFiles)
{
	return showDialog (false, true, true, warnAboutOverwritingExistingFiles, false, nullptr);
}

bool FileChooser::browseForDirectory()
{
	return showDialog (true, false, false, false, false, nullptr);
}

bool FileChooser::showDialog (const bool selectsDirectories,
							  const bool selectsFiles,
							  const bool isSave,
							  const bool warnAboutOverwritingExistingFiles,
							  const bool selectMultipleFiles,
							  FilePreviewComponent* const previewComponent)
{
	WeakReference<Component> previouslyFocused (Component::getCurrentlyFocusedComponent());

	results.clear();

	// the preview component needs to be the right size before you pass it in here..
	jassert (previewComponent == nullptr || (previewComponent->getWidth() > 10
											   && previewComponent->getHeight() > 10));

   #if JUCE_WINDOWS
	if (useNativeDialogBox && ! (selectsFiles && selectsDirectories))
   #elif JUCE_MAC
	if (useNativeDialogBox && (previewComponent == nullptr))
   #else
	if (false)
   #endif
	{
		showPlatformDialog (results, title, startingFile, filters,
							selectsDirectories, selectsFiles, isSave,
							warnAboutOverwritingExistingFiles,
							selectMultipleFiles,
							previewComponent);
	}
	else
	{
		WildcardFileFilter wildcard (selectsFiles ? filters : String::empty,
									 selectsDirectories ? "*" : String::empty,
									 String::empty);

		int flags = isSave ? FileBrowserComponent::saveMode
						   : FileBrowserComponent::openMode;

		if (selectsFiles)
			flags |= FileBrowserComponent::canSelectFiles;

		if (selectsDirectories)
		{
			flags |= FileBrowserComponent::canSelectDirectories;

			if (! isSave)
				flags |= FileBrowserComponent::filenameBoxIsReadOnly;
		}

		if (selectMultipleFiles)
			flags |= FileBrowserComponent::canSelectMultipleItems;

		FileBrowserComponent browserComponent (flags, startingFile, &wildcard, previewComponent);

		FileChooserDialogBox box (title, String::empty,
								  browserComponent,
								  warnAboutOverwritingExistingFiles,
								  browserComponent.findColour (AlertWindow::backgroundColourId));

		if (box.show())
		{
			for (int i = 0; i < browserComponent.getNumSelectedFiles(); ++i)
				results.add (browserComponent.getSelectedFile (i));
		}
	}

	if (previouslyFocused != nullptr)
		previouslyFocused->grabKeyboardFocus();

	return results.size() > 0;
}
#endif

File FileChooser::getResult() const
{
	// if you've used a multiple-file select, you should use the getResults() method
	// to retrieve all the files that were chosen.
	jassert (results.size() <= 1);

	return results.getFirst();
}

const Array<File>& FileChooser::getResults() const
{
	return results;
}

FilePreviewComponent::FilePreviewComponent()
{
}

FilePreviewComponent::~FilePreviewComponent()
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileChooser.cpp ***/


/*** Start of inlined file: juce_FileChooserDialogBox.cpp ***/
BEGIN_JUCE_NAMESPACE

class FileChooserDialogBox::ContentComponent  : public Component
{
public:

	ContentComponent (const String& name, const String& instructions_, FileBrowserComponent& chooserComponent_)
		: Component (name),
		  chooserComponent (chooserComponent_),
		  okButton (chooserComponent_.getActionVerb()),
		  cancelButton (TRANS ("Cancel")),
		  newFolderButton (TRANS ("New Folder")),
		  instructions (instructions_)
	{
		addAndMakeVisible (&chooserComponent);

		addAndMakeVisible (&okButton);
		okButton.addShortcut (KeyPress::returnKey);

		addAndMakeVisible (&cancelButton);
		cancelButton.addShortcut (KeyPress::escapeKey);

		addChildComponent (&newFolderButton);

		setInterceptsMouseClicks (false, true);
	}

	void paint (Graphics& g)
	{
		g.setColour (getLookAndFeel().findColour (FileChooserDialogBox::titleTextColourId));
		text.draw (g);
	}

	void resized()
	{
		const int buttonHeight = 26;

		Rectangle<int> area (getLocalBounds());

		getLookAndFeel().createFileChooserHeaderText (getName(), instructions, text, getWidth());
		const Rectangle<float> bb (text.getBoundingBox (0, text.getNumGlyphs(), false));
		area.removeFromTop (roundToInt (bb.getBottom()) + 10);

		chooserComponent.setBounds (area.removeFromTop (area.getHeight() - buttonHeight - 20));
		Rectangle<int> buttonArea (area.reduced (16, 10));

		okButton.changeWidthToFitText (buttonHeight);
		okButton.setBounds (buttonArea.removeFromRight (okButton.getWidth() + 16));

		buttonArea.removeFromRight (16);

		cancelButton.changeWidthToFitText (buttonHeight);
		cancelButton.setBounds (buttonArea.removeFromRight (cancelButton.getWidth()));

		newFolderButton.changeWidthToFitText (buttonHeight);
		newFolderButton.setBounds (buttonArea.removeFromLeft (newFolderButton.getWidth()));
	}

	FileBrowserComponent& chooserComponent;
	TextButton okButton, cancelButton, newFolderButton;

private:
	String instructions;
	GlyphArrangement text;
};

FileChooserDialogBox::FileChooserDialogBox (const String& name,
											const String& instructions,
											FileBrowserComponent& chooserComponent,
											const bool warnAboutOverwritingExistingFiles_,
											const Colour& backgroundColour)
	: ResizableWindow (name, backgroundColour, true),
	  warnAboutOverwritingExistingFiles (warnAboutOverwritingExistingFiles_)
{
	content = new ContentComponent (name, instructions, chooserComponent);
	setContentOwned (content, false);

	setResizable (true, true);
	setResizeLimits (300, 300, 1200, 1000);

	content->okButton.addListener (this);
	content->cancelButton.addListener (this);
	content->newFolderButton.addListener (this);
	content->chooserComponent.addListener (this);

	FileChooserDialogBox::selectionChanged();
}

FileChooserDialogBox::~FileChooserDialogBox()
{
	content->chooserComponent.removeListener (this);
}

#if JUCE_MODAL_LOOPS_PERMITTED
bool FileChooserDialogBox::show (int w, int h)
{
	return showAt (-1, -1, w, h);
}

bool FileChooserDialogBox::showAt (int x, int y, int w, int h)
{
	if (w <= 0)
	{
		Component* const previewComp = content->chooserComponent.getPreviewComponent();

		if (previewComp != nullptr)
			w = 400 + previewComp->getWidth();
		else
			w = 600;
	}

	if (h <= 0)
		h = 500;

	if (x < 0 || y < 0)
		centreWithSize (w, h);
	else
		setBounds (x, y, w, h);

	const bool ok = (runModalLoop() != 0);
	setVisible (false);
	return ok;
}
#endif

void FileChooserDialogBox::centreWithDefaultSize (Component* componentToCentreAround)
{
	Component* const previewComp = content->chooserComponent.getPreviewComponent();

	centreAroundComponent (componentToCentreAround,
						   previewComp != nullptr ? 400 + previewComp->getWidth() : 600,
						   500);
}

void FileChooserDialogBox::buttonClicked (Button* button)
{
	if (button == &(content->okButton))
	{
		okButtonPressed();
	}
	else if (button == &(content->cancelButton))
	{
		closeButtonPressed();
	}
	else if (button == &(content->newFolderButton))
	{
		createNewFolder();
	}
}

void FileChooserDialogBox::closeButtonPressed()
{
	setVisible (false);
}

void FileChooserDialogBox::selectionChanged()
{
	content->okButton.setEnabled (content->chooserComponent.currentFileIsValid());

	content->newFolderButton.setVisible (content->chooserComponent.isSaveMode()
										  && content->chooserComponent.getRoot().isDirectory());
}

void FileChooserDialogBox::fileClicked (const File&, const MouseEvent&)
{
}

void FileChooserDialogBox::fileDoubleClicked (const File&)
{
	selectionChanged();
	content->okButton.triggerClick();
}

void FileChooserDialogBox::okToOverwriteFileCallback (int result, FileChooserDialogBox* box)
{
	if (result != 0 && box != nullptr)
		box->exitModalState (1);
}

void FileChooserDialogBox::okButtonPressed()
{
	if (warnAboutOverwritingExistingFiles
		 && content->chooserComponent.isSaveMode()
		 && content->chooserComponent.getSelectedFile(0).exists())
	{
		AlertWindow::showOkCancelBox (AlertWindow::WarningIcon,
										 TRANS("File already exists"),
										 TRANS("There's already a file called:")
										   + "\n\n" + content->chooserComponent.getSelectedFile(0).getFullPathName()
										   + "\n\n" + TRANS("Are you sure you want to overwrite it?"),
										 TRANS("overwrite"),
										 TRANS("cancel"),
									  this,
									  ModalCallbackFunction::forComponent (okToOverwriteFileCallback, this));
	}
	else
	{
		exitModalState (1);
	}
}

void FileChooserDialogBox::createNewFolderCallback (int result, FileChooserDialogBox* box,
													Component::SafePointer<AlertWindow> alert)
{
	if (result != 0 && alert != nullptr && box != nullptr)
	{
		alert->setVisible (false);
		box->createNewFolderConfirmed (alert->getTextEditorContents ("name"));
	}
}

void FileChooserDialogBox::createNewFolder()
{
	File parent (content->chooserComponent.getRoot());

	if (parent.isDirectory())
	{
		AlertWindow* aw = new AlertWindow (TRANS("New Folder"),
										   TRANS("Please enter the name for the folder"),
										   AlertWindow::NoIcon, this);

		aw->addTextEditor ("name", String::empty, String::empty, false);
		aw->addButton (TRANS("ok"), 1, KeyPress::returnKey);
		aw->addButton (TRANS("cancel"), KeyPress::escapeKey);

		aw->enterModalState (true,
							 ModalCallbackFunction::forComponent (createNewFolderCallback, this,
																  Component::SafePointer<AlertWindow> (aw)),
							 true);
	}
}

void FileChooserDialogBox::createNewFolderConfirmed (const String& nameFromDialog)
{
	const String name (File::createLegalFileName (nameFromDialog));

	if (! name.isEmpty())
	{
		const File parent (content->chooserComponent.getRoot());

		if (! parent.getChildFile (name).createDirectory())
		{
			AlertWindow::showMessageBoxAsync (AlertWindow::WarningIcon,
											  TRANS ("New Folder"),
											  TRANS ("Couldn't create the folder!"));
		}

		content->chooserComponent.refresh();
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileChooserDialogBox.cpp ***/


/*** Start of inlined file: juce_FileFilter.cpp ***/
BEGIN_JUCE_NAMESPACE

FileFilter::FileFilter (const String& filterDescription)
	: description (filterDescription)
{
}

FileFilter::~FileFilter()
{
}

const String& FileFilter::getDescription() const noexcept
{
	return description;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileFilter.cpp ***/


/*** Start of inlined file: juce_FileListComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

Image juce_createIconForFile (const File& file);

FileListComponent::FileListComponent (DirectoryContentsList& listToShow)
	: ListBox (String::empty, nullptr),
	  DirectoryContentsDisplayComponent (listToShow)
{
	setModel (this);
	fileList.addChangeListener (this);
}

FileListComponent::~FileListComponent()
{
	fileList.removeChangeListener (this);
}

int FileListComponent::getNumSelectedFiles() const
{
	return getNumSelectedRows();
}

const File FileListComponent::getSelectedFile (int index) const
{
	return fileList.getFile (getSelectedRow (index));
}

void FileListComponent::deselectAllFiles()
{
	deselectAllRows();
}

void FileListComponent::scrollToTop()
{
	getVerticalScrollBar()->setCurrentRangeStart (0);
}

void FileListComponent::changeListenerCallback (ChangeBroadcaster*)
{
	updateContent();

	if (lastDirectory != fileList.getDirectory())
	{
		lastDirectory = fileList.getDirectory();
		deselectAllRows();
	}
}

class FileListItemComponent  : public Component,
							   public TimeSliceClient,
							   public AsyncUpdater
{
public:

	FileListItemComponent (FileListComponent& owner_, TimeSliceThread& thread_)
		: owner (owner_), thread (thread_), index (0), highlighted (false)
	{
	}

	~FileListItemComponent()
	{
		thread.removeTimeSliceClient (this);
	}

	void paint (Graphics& g)
	{
		getLookAndFeel().drawFileBrowserRow (g, getWidth(), getHeight(),
											 file.getFileName(),
											 &icon, fileSize, modTime,
											 isDirectory, highlighted,
											 index, owner);
	}

	void mouseDown (const MouseEvent& e)
	{
		owner.selectRowsBasedOnModifierKeys (index, e.mods, false);
		owner.sendMouseClickMessage (file, e);
	}

	void mouseDoubleClick (const MouseEvent&)
	{
		owner.sendDoubleClickMessage (file);
	}

	void update (const File& root,
				 const DirectoryContentsList::FileInfo* const fileInfo,
				 const int index_,
				 const bool highlighted_)
	{
		thread.removeTimeSliceClient (this);

		if (highlighted_ != highlighted || index_ != index)
		{
			index = index_;
			highlighted = highlighted_;
			repaint();
		}

		File newFile;
		String newFileSize, newModTime;

		if (fileInfo != nullptr)
		{
			newFile = root.getChildFile (fileInfo->filename);
			newFileSize = File::descriptionOfSizeInBytes (fileInfo->fileSize);
			newModTime = fileInfo->modificationTime.formatted ("%d %b '%y %H:%M");
		}

		if (newFile != file
			 || fileSize != newFileSize
			 || modTime != newModTime)
		{
			file = newFile;
			fileSize = newFileSize;
			modTime = newModTime;
			icon = Image::null;
			isDirectory = fileInfo != nullptr && fileInfo->isDirectory;

			repaint();
		}

		if (file != File::nonexistent && icon.isNull() && ! isDirectory)
		{
			updateIcon (true);

			if (! icon.isValid())
				thread.addTimeSliceClient (this);
		}
	}

	int useTimeSlice()
	{
		updateIcon (false);
		return -1;
	}

	void handleAsyncUpdate()
	{
		repaint();
	}

private:

	FileListComponent& owner;
	TimeSliceThread& thread;
	File file;
	String fileSize, modTime;
	Image icon;
	int index;
	bool highlighted, isDirectory;

	void updateIcon (const bool onlyUpdateIfCached)
	{
		if (icon.isNull())
		{
			const int hashCode = (file.getFullPathName() + "_iconCacheSalt").hashCode();
			Image im (ImageCache::getFromHashCode (hashCode));

			if (im.isNull() && ! onlyUpdateIfCached)
			{
				im = juce_createIconForFile (file);

				if (im.isValid())
					ImageCache::addImageToCache (im, hashCode);
			}

			if (im.isValid())
			{
				icon = im;
				triggerAsyncUpdate();
			}
		}
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (FileListItemComponent);
};

int FileListComponent::getNumRows()
{
	return fileList.getNumFiles();
}

void FileListComponent::paintListBoxItem (int, Graphics&, int, int, bool)
{
}

Component* FileListComponent::refreshComponentForRow (int row, bool isSelected, Component* existingComponentToUpdate)
{
	FileListItemComponent* comp = dynamic_cast <FileListItemComponent*> (existingComponentToUpdate);

	if (comp == nullptr)
	{
		delete existingComponentToUpdate;
		comp = new FileListItemComponent (*this, fileList.getTimeSliceThread());
	}

	DirectoryContentsList::FileInfo fileInfo;

	if (fileList.getFileInfo (row, fileInfo))
		comp->update (fileList.getDirectory(), &fileInfo, row, isSelected);
	else
		comp->update (fileList.getDirectory(), nullptr, row, isSelected);

	return comp;
}

void FileListComponent::selectedRowsChanged (int /*lastRowSelected*/)
{
	sendSelectionChangeMessage();
}

void FileListComponent::deleteKeyPressed (int /*currentSelectedRow*/)
{
}

void FileListComponent::returnKeyPressed (int currentSelectedRow)
{
	sendDoubleClickMessage (fileList.getFile (currentSelectedRow));
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileListComponent.cpp ***/


/*** Start of inlined file: juce_FilenameComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

FilenameComponent::FilenameComponent (const String& name,
									  const File& currentFile,
									  const bool canEditFilename,
									  const bool isDirectory,
									  const bool isForSaving,
									  const String& fileBrowserWildcard,
									  const String& enforcedSuffix_,
									  const String& textWhenNothingSelected)
	: Component (name),
	  maxRecentFiles (30),
	  isDir (isDirectory),
	  isSaving (isForSaving),
	  isFileDragOver (false),
	  wildcard (fileBrowserWildcard),
	  enforcedSuffix (enforcedSuffix_)
{
	addAndMakeVisible (&filenameBox);
	filenameBox.setEditableText (canEditFilename);
	filenameBox.addListener (this);
	filenameBox.setTextWhenNothingSelected (textWhenNothingSelected);
	filenameBox.setTextWhenNoChoicesAvailable (TRANS("(no recently seleced files)"));

	setBrowseButtonText ("...");

	setCurrentFile (currentFile, true);
}

FilenameComponent::~FilenameComponent()
{
}

void FilenameComponent::paintOverChildren (Graphics& g)
{
	if (isFileDragOver)
	{
		g.setColour (Colours::red.withAlpha (0.2f));
		g.drawRect (0, 0, getWidth(), getHeight(), 3);
	}
}

void FilenameComponent::resized()
{
	getLookAndFeel().layoutFilenameComponent (*this, &filenameBox, browseButton);
}

void FilenameComponent::setBrowseButtonText (const String& newBrowseButtonText)
{
	browseButtonText = newBrowseButtonText;
	lookAndFeelChanged();
}

void FilenameComponent::lookAndFeelChanged()
{
	browseButton = nullptr;

	addAndMakeVisible (browseButton = getLookAndFeel().createFilenameComponentBrowseButton (browseButtonText));
	browseButton->setConnectedEdges (Button::ConnectedOnLeft);
	resized();

	browseButton->addListener (this);
}

void FilenameComponent::setTooltip (const String& newTooltip)
{
	SettableTooltipClient::setTooltip (newTooltip);
	filenameBox.setTooltip (newTooltip);
}

void FilenameComponent::setDefaultBrowseTarget (const File& newDefaultDirectory)
{
	defaultBrowseFile = newDefaultDirectory;
}

void FilenameComponent::buttonClicked (Button*)
{
   #if JUCE_MODAL_LOOPS_PERMITTED
	FileChooser fc (TRANS("Choose a new file"),
					getCurrentFile() == File::nonexistent ? defaultBrowseFile
														  : getCurrentFile(),
					wildcard);

	if (isDir ? fc.browseForDirectory()
			  : (isSaving ? fc.browseForFileToSave (false)
						  : fc.browseForFileToOpen()))
	{
		setCurrentFile (fc.getResult(), true);
	}
   #else
	jassertfalse; // needs rewriting to deal with non-modal environments
   #endif
}

void FilenameComponent::comboBoxChanged (ComboBox*)
{
	setCurrentFile (getCurrentFile(), true);
}

bool FilenameComponent::isInterestedInFileDrag (const StringArray&)
{
	return true;
}

void FilenameComponent::filesDropped (const StringArray& filenames, int, int)
{
	isFileDragOver = false;
	repaint();

	const File f (filenames[0]);

	if (f.exists() && (f.isDirectory() == isDir))
		setCurrentFile (f, true);
}

void FilenameComponent::fileDragEnter (const StringArray&, int, int)
{
	isFileDragOver = true;
	repaint();
}

void FilenameComponent::fileDragExit (const StringArray&)
{
	isFileDragOver = false;
	repaint();
}

File FilenameComponent::getCurrentFile() const
{
	File f (filenameBox.getText());

	if (enforcedSuffix.isNotEmpty())
		f = f.withFileExtension (enforcedSuffix);

	return f;
}

void FilenameComponent::setCurrentFile (File newFile,
										const bool addToRecentlyUsedList,
										const bool sendChangeNotification)
{
	if (enforcedSuffix.isNotEmpty())
		newFile = newFile.withFileExtension (enforcedSuffix);

	if (newFile.getFullPathName() != lastFilename)
	{
		lastFilename = newFile.getFullPathName();

		if (addToRecentlyUsedList)
			addRecentlyUsedFile (newFile);

		filenameBox.setText (lastFilename, true);

		if (sendChangeNotification)
			triggerAsyncUpdate();
	}
}

void FilenameComponent::setFilenameIsEditable (const bool shouldBeEditable)
{
	filenameBox.setEditableText (shouldBeEditable);
}

StringArray FilenameComponent::getRecentlyUsedFilenames() const
{
	StringArray names;

	for (int i = 0; i < filenameBox.getNumItems(); ++i)
		names.add (filenameBox.getItemText (i));

	return names;
}

void FilenameComponent::setRecentlyUsedFilenames (const StringArray& filenames)
{
	if (filenames != getRecentlyUsedFilenames())
	{
		filenameBox.clear();

		for (int i = 0; i < jmin (filenames.size(), maxRecentFiles); ++i)
			filenameBox.addItem (filenames[i], i + 1);
	}
}

void FilenameComponent::setMaxNumberOfRecentFiles (const int newMaximum)
{
	maxRecentFiles = jmax (1, newMaximum);

	setRecentlyUsedFilenames (getRecentlyUsedFilenames());
}

void FilenameComponent::addRecentlyUsedFile (const File& file)
{
	StringArray files (getRecentlyUsedFilenames());

	if (file.getFullPathName().isNotEmpty())
	{
		files.removeString (file.getFullPathName(), true);
		files.insert (0, file.getFullPathName());

		setRecentlyUsedFilenames (files);
	}
}

void FilenameComponent::addListener (FilenameComponentListener* const listener)
{
	listeners.add (listener);
}

void FilenameComponent::removeListener (FilenameComponentListener* const listener)
{
	listeners.remove (listener);
}

void FilenameComponent::handleAsyncUpdate()
{
	Component::BailOutChecker checker (this);
	listeners.callChecked (checker, &FilenameComponentListener::filenameComponentChanged, this);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FilenameComponent.cpp ***/


/*** Start of inlined file: juce_FileSearchPathListComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

FileSearchPathListComponent::FileSearchPathListComponent()
	: addButton ("+"),
	  removeButton ("-"),
	  changeButton (TRANS ("change...")),
	  upButton (String::empty, DrawableButton::ImageOnButtonBackground),
	  downButton (String::empty, DrawableButton::ImageOnButtonBackground)
{
	listBox.setModel (this);
	addAndMakeVisible (&listBox);
	listBox.setColour (ListBox::backgroundColourId, Colours::black.withAlpha (0.02f));
	listBox.setColour (ListBox::outlineColourId, Colours::black.withAlpha (0.1f));
	listBox.setOutlineThickness (1);

	addAndMakeVisible (&addButton);
	addButton.addListener (this);
	addButton.setConnectedEdges (Button::ConnectedOnLeft | Button::ConnectedOnRight | Button::ConnectedOnBottom | Button::ConnectedOnTop);

	addAndMakeVisible (&removeButton);
	removeButton.addListener (this);
	removeButton.setConnectedEdges (Button::ConnectedOnLeft | Button::ConnectedOnRight | Button::ConnectedOnBottom | Button::ConnectedOnTop);

	addAndMakeVisible (&changeButton);
	changeButton.addListener (this);

	addAndMakeVisible (&upButton);
	upButton.addListener (this);

	{
		Path arrowPath;
		arrowPath.addArrow (Line<float> (50.0f, 100.0f, 50.0f, 0.0f), 40.0f, 100.0f, 50.0f);
		DrawablePath arrowImage;
		arrowImage.setFill (Colours::black.withAlpha (0.4f));
		arrowImage.setPath (arrowPath);

		upButton.setImages (&arrowImage);
	}

	addAndMakeVisible (&downButton);
	downButton.addListener (this);

	{
		Path arrowPath;
		arrowPath.addArrow (Line<float> (50.0f, 0.0f, 50.0f, 100.0f), 40.0f, 100.0f, 50.0f);
		DrawablePath arrowImage;
		arrowImage.setFill (Colours::black.withAlpha (0.4f));
		arrowImage.setPath (arrowPath);

		downButton.setImages (&arrowImage);
	}

	updateButtons();
}

FileSearchPathListComponent::~FileSearchPathListComponent()
{
}

void FileSearchPathListComponent::updateButtons()
{
	const bool anythingSelected = listBox.getNumSelectedRows() > 0;

	removeButton.setEnabled (anythingSelected);
	changeButton.setEnabled (anythingSelected);
	upButton.setEnabled (anythingSelected);
	downButton.setEnabled (anythingSelected);
}

void FileSearchPathListComponent::changed()
{
	listBox.updateContent();
	listBox.repaint();
	updateButtons();
}

void FileSearchPathListComponent::setPath (const FileSearchPath& newPath)
{
	if (newPath.toString() != path.toString())
	{
		path = newPath;
		changed();
	}
}

void FileSearchPathListComponent::setDefaultBrowseTarget (const File& newDefaultDirectory)
{
	defaultBrowseTarget = newDefaultDirectory;
}

int FileSearchPathListComponent::getNumRows()
{
	return path.getNumPaths();
}

void FileSearchPathListComponent::paintListBoxItem (int rowNumber, Graphics& g, int width, int height, bool rowIsSelected)
{
	if (rowIsSelected)
		g.fillAll (findColour (TextEditor::highlightColourId));

	g.setColour (findColour (ListBox::textColourId));
	Font f (height * 0.7f);
	f.setHorizontalScale (0.9f);
	g.setFont (f);

	g.drawText (path [rowNumber].getFullPathName(),
				4, 0, width - 6, height,
				Justification::centredLeft, true);
}

void FileSearchPathListComponent::deleteKeyPressed (int row)
{
	if (isPositiveAndBelow (row, path.getNumPaths()))
	{
		path.remove (row);
		changed();
	}
}

void FileSearchPathListComponent::returnKeyPressed (int row)
{
   #if JUCE_MODAL_LOOPS_PERMITTED
	FileChooser chooser (TRANS("Change folder..."), path [row], "*");

	if (chooser.browseForDirectory())
	{
		path.remove (row);
		path.add (chooser.getResult(), row);
		changed();
	}
   #endif
}

void FileSearchPathListComponent::listBoxItemDoubleClicked (int row, const MouseEvent&)
{
	returnKeyPressed (row);
}

void FileSearchPathListComponent::selectedRowsChanged (int)
{
	updateButtons();
}

void FileSearchPathListComponent::paint (Graphics& g)
{
	g.fillAll (findColour (backgroundColourId));
}

void FileSearchPathListComponent::resized()
{
	const int buttonH = 22;
	const int buttonY = getHeight() - buttonH - 4;
	listBox.setBounds (2, 2, getWidth() - 4, buttonY - 5);

	addButton.setBounds (2, buttonY, buttonH, buttonH);
	removeButton.setBounds (addButton.getRight(), buttonY, buttonH, buttonH);

	changeButton.changeWidthToFitText (buttonH);
	downButton.setSize (buttonH * 2, buttonH);
	upButton.setSize (buttonH * 2, buttonH);

	downButton.setTopRightPosition (getWidth() - 2, buttonY);
	upButton.setTopRightPosition (downButton.getX() - 4, buttonY);
	changeButton.setTopRightPosition (upButton.getX() - 8, buttonY);
}

bool FileSearchPathListComponent::isInterestedInFileDrag (const StringArray&)
{
	return true;
}

void FileSearchPathListComponent::filesDropped (const StringArray& filenames, int, int mouseY)
{
	for (int i = filenames.size(); --i >= 0;)
	{
		const File f (filenames[i]);

		if (f.isDirectory())
		{
			const int row = listBox.getRowContainingPosition (0, mouseY - listBox.getY());
			path.add (f, row);
			changed();
		}
	}
}

void FileSearchPathListComponent::buttonClicked (Button* button)
{
	const int currentRow = listBox.getSelectedRow();

	if (button == &removeButton)
	{
		deleteKeyPressed (currentRow);
	}
	else if (button == &addButton)
	{
		File start (defaultBrowseTarget);

		if (start == File::nonexistent)
			start = path [0];

		if (start == File::nonexistent)
			start = File::getCurrentWorkingDirectory();

	   #if JUCE_MODAL_LOOPS_PERMITTED
		FileChooser chooser (TRANS("Add a folder..."), start, "*");

		if (chooser.browseForDirectory())
			path.add (chooser.getResult(), currentRow);
	   #else
		jassertfalse; // needs rewriting to deal with non-modal environments
	   #endif
	}
	else if (button == &changeButton)
	{
		returnKeyPressed (currentRow);
	}
	else if (button == &upButton)
	{
		if (currentRow > 0 && currentRow < path.getNumPaths())
		{
			const File f (path[currentRow]);
			path.remove (currentRow);
			path.add (f, currentRow - 1);
			listBox.selectRow (currentRow - 1);
		}
	}
	else if (button == &downButton)
	{
		if (currentRow >= 0 && currentRow < path.getNumPaths() - 1)
		{
			const File f (path[currentRow]);
			path.remove (currentRow);
			path.add (f, currentRow + 1);
			listBox.selectRow (currentRow + 1);
		}
	}

	changed();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileSearchPathListComponent.cpp ***/


/*** Start of inlined file: juce_FileTreeComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

Image juce_createIconForFile (const File& file);

class FileListTreeItem   : public TreeViewItem,
						   public TimeSliceClient,
						   public AsyncUpdater,
						   public ChangeListener
{
public:

	FileListTreeItem (FileTreeComponent& owner_,
					  DirectoryContentsList* const parentContentsList_,
					  const int indexInContentsList_,
					  const File& file_,
					  TimeSliceThread& thread_)
		: file (file_),
		  owner (owner_),
		  parentContentsList (parentContentsList_),
		  indexInContentsList (indexInContentsList_),
		  subContentsList (nullptr, false),
		  thread (thread_)
	{
		DirectoryContentsList::FileInfo fileInfo;

		if (parentContentsList_ != nullptr
			 && parentContentsList_->getFileInfo (indexInContentsList_, fileInfo))
		{
			fileSize = File::descriptionOfSizeInBytes (fileInfo.fileSize);
			modTime = fileInfo.modificationTime.formatted ("%d %b '%y %H:%M");
			isDirectory = fileInfo.isDirectory;
		}
		else
		{
			isDirectory = true;
		}
	}

	~FileListTreeItem()
	{
		thread.removeTimeSliceClient (this);
		clearSubItems();
	}

	bool mightContainSubItems()		 { return isDirectory; }
	const String getUniqueName() const	  { return file.getFullPathName(); }
	int getItemHeight() const		   { return 22; }

	const var getDragSourceDescription()	{ return owner.getDragAndDropDescription(); }

	void itemOpennessChanged (bool isNowOpen)
	{
		if (isNowOpen)
		{
			clearSubItems();

			isDirectory = file.isDirectory();

			if (isDirectory)
			{
				if (subContentsList == nullptr)
				{
					jassert (parentContentsList != nullptr);

					DirectoryContentsList* const l = new DirectoryContentsList (parentContentsList->getFilter(), thread);
					l->setDirectory (file, true, true);

					setSubContentsList (l, true);
				}

				changeListenerCallback (nullptr);
			}
		}
	}

	void setSubContentsList (DirectoryContentsList* newList, const bool canDeleteList)
	{
		OptionalScopedPointer<DirectoryContentsList> newPointer (newList, canDeleteList);
		subContentsList = newPointer;
		newList->addChangeListener (this);
	}

	void changeListenerCallback (ChangeBroadcaster*)
	{
		clearSubItems();

		if (isOpen() && subContentsList != nullptr)
		{
			for (int i = 0; i < subContentsList->getNumFiles(); ++i)
			{
				FileListTreeItem* const item
					= new FileListTreeItem (owner, subContentsList, i, subContentsList->getFile(i), thread);

				addSubItem (item);
			}
		}
	}

	void paintItem (Graphics& g, int width, int height)
	{
		if (file != File::nonexistent)
		{
			updateIcon (true);

			if (icon.isNull())
				thread.addTimeSliceClient (this);
		}

		owner.getLookAndFeel()
			.drawFileBrowserRow (g, width, height,
								 file.getFileName(),
								 &icon, fileSize, modTime,
								 isDirectory, isSelected(),
								 indexInContentsList, owner);
	}

	void itemClicked (const MouseEvent& e)
	{
		owner.sendMouseClickMessage (file, e);
	}

	void itemDoubleClicked (const MouseEvent& e)
	{
		TreeViewItem::itemDoubleClicked (e);

		owner.sendDoubleClickMessage (file);
	}

	void itemSelectionChanged (bool)
	{
		owner.sendSelectionChangeMessage();
	}

	int useTimeSlice()
	{
		updateIcon (false);
		return -1;
	}

	void handleAsyncUpdate()
	{
		owner.repaint();
	}

	const File file;

private:
	FileTreeComponent& owner;
	DirectoryContentsList* parentContentsList;
	int indexInContentsList;
	OptionalScopedPointer<DirectoryContentsList> subContentsList;
	bool isDirectory;
	TimeSliceThread& thread;
	Image icon;
	String fileSize;
	String modTime;

	void updateIcon (const bool onlyUpdateIfCached)
	{
		if (icon.isNull())
		{
			const int hashCode = (file.getFullPathName() + "_iconCacheSalt").hashCode();
			Image im (ImageCache::getFromHashCode (hashCode));

			if (im.isNull() && ! onlyUpdateIfCached)
			{
				im = juce_createIconForFile (file);

				if (im.isValid())
					ImageCache::addImageToCache (im, hashCode);
			}

			if (im.isValid())
			{
				icon = im;
				triggerAsyncUpdate();
			}
		}
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (FileListTreeItem);
};

FileTreeComponent::FileTreeComponent (DirectoryContentsList& listToShow)
	: DirectoryContentsDisplayComponent (listToShow)
{
	FileListTreeItem* const root
		= new FileListTreeItem (*this, 0, 0, listToShow.getDirectory(),
								listToShow.getTimeSliceThread());

	root->setSubContentsList (&listToShow, false);
	setRootItemVisible (false);
	setRootItem (root);
}

FileTreeComponent::~FileTreeComponent()
{
	deleteRootItem();
}

const File FileTreeComponent::getSelectedFile (const int index) const
{
	const FileListTreeItem* const item = dynamic_cast <const FileListTreeItem*> (getSelectedItem (index));

	return item != nullptr ? item->file
						   : File::nonexistent;
}

void FileTreeComponent::deselectAllFiles()
{
	clearSelectedItems();
}

void FileTreeComponent::scrollToTop()
{
	getViewport()->getVerticalScrollBar()->setCurrentRangeStart (0);
}

void FileTreeComponent::setDragAndDropDescription (const String& description)
{
	dragAndDropDescription = description;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FileTreeComponent.cpp ***/


/*** Start of inlined file: juce_ImagePreviewComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

ImagePreviewComponent::ImagePreviewComponent()
{
}

ImagePreviewComponent::~ImagePreviewComponent()
{
}

void ImagePreviewComponent::getThumbSize (int& w, int& h) const
{
	const int availableW = proportionOfWidth (0.97f);
	const int availableH = getHeight() - 13 * 4;

	const double scale = jmin (1.0,
							   availableW / (double) w,
							   availableH / (double) h);

	w = roundToInt (scale * w);
	h = roundToInt (scale * h);
}

void ImagePreviewComponent::selectedFileChanged (const File& file)
{
	if (fileToLoad != file)
	{
		fileToLoad = file;
		startTimer (100);
	}
}

void ImagePreviewComponent::timerCallback()
{
	stopTimer();

	currentThumbnail = Image::null;
	currentDetails = String::empty;
	repaint();

	ScopedPointer <FileInputStream> in (fileToLoad.createInputStream());

	if (in != nullptr)
	{
		ImageFileFormat* const format = ImageFileFormat::findImageFormatForStream (*in);

		if (format != nullptr)
		{
			currentThumbnail = format->decodeImage (*in);

			if (currentThumbnail.isValid())
			{
				int w = currentThumbnail.getWidth();
				int h = currentThumbnail.getHeight();

				currentDetails
					<< fileToLoad.getFileName() << "\n"
					<< format->getFormatName() << "\n"
					<< w << " x " << h << " pixels\n"
					<< File::descriptionOfSizeInBytes (fileToLoad.getSize());

				getThumbSize (w, h);

				currentThumbnail = currentThumbnail.rescaled (w, h);
			}
		}
	}
}

void ImagePreviewComponent::paint (Graphics& g)
{
	if (currentThumbnail.isValid())
	{
		g.setFont (13.0f);

		int w = currentThumbnail.getWidth();
		int h = currentThumbnail.getHeight();
		getThumbSize (w, h);

		const int numLines = 4;
		const int totalH = 13 * numLines + h + 4;
		const int y = (getHeight() - totalH) / 2;

		g.drawImageWithin (currentThumbnail,
						   (getWidth() - w) / 2, y, w, h,
						   RectanglePlacement::centred | RectanglePlacement::onlyReduceInSize,
						   false);

		g.drawFittedText (currentDetails,
						  0, y + h + 4, getWidth(), 100,
						  Justification::centredTop, numLines);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ImagePreviewComponent.cpp ***/


/*** Start of inlined file: juce_WildcardFileFilter.cpp ***/
BEGIN_JUCE_NAMESPACE

WildcardFileFilter::WildcardFileFilter (const String& fileWildcardPatterns,
										const String& directoryWildcardPatterns,
										const String& description_)
	: FileFilter (description_.isEmpty() ? fileWildcardPatterns
										 : (description_ + " (" + fileWildcardPatterns + ")"))
{
	parse (fileWildcardPatterns, fileWildcards);
	parse (directoryWildcardPatterns, directoryWildcards);
}

WildcardFileFilter::~WildcardFileFilter()
{
}

bool WildcardFileFilter::isFileSuitable (const File& file) const
{
	return match (file, fileWildcards);
}

bool WildcardFileFilter::isDirectorySuitable (const File& file) const
{
	return match (file, directoryWildcards);
}

void WildcardFileFilter::parse (const String& pattern, StringArray& result)
{
	result.addTokens (pattern.toLowerCase(), ";,", "\"'");

	result.trim();
	result.removeEmptyStrings();

	// special case for *.*, because people use it to mean "any file", but it
	// would actually ignore files with no extension.
	for (int i = result.size(); --i >= 0;)
		if (result[i] == "*.*")
			result.set (i, "*");
}

bool WildcardFileFilter::match (const File& file, const StringArray& wildcards)
{
	const String filename (file.getFileName());

	for (int i = wildcards.size(); --i >= 0;)
		if (filename.matchesWildcard (wildcards[i], true))
			return true;

	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_WildcardFileFilter.cpp ***/


/*** Start of inlined file: juce_KeyboardFocusTraverser.cpp ***/
BEGIN_JUCE_NAMESPACE

KeyboardFocusTraverser::KeyboardFocusTraverser()
{
}

KeyboardFocusTraverser::~KeyboardFocusTraverser()
{
}

namespace KeyboardFocusHelpers
{
	// This will sort a set of components, so that they are ordered in terms of
	// left-to-right and then top-to-bottom.
	class ScreenPositionComparator
	{
	public:
		ScreenPositionComparator() {}

		static int compareElements (const Component* const first, const Component* const second)
		{
			int explicitOrder1 = first->getExplicitFocusOrder();
			if (explicitOrder1 <= 0)
				explicitOrder1 = std::numeric_limits<int>::max() / 2;

			int explicitOrder2 = second->getExplicitFocusOrder();
			if (explicitOrder2 <= 0)
				explicitOrder2 = std::numeric_limits<int>::max() / 2;

			if (explicitOrder1 != explicitOrder2)
				return explicitOrder1 - explicitOrder2;

			const int diff = first->getY() - second->getY();

			return (diff == 0) ? first->getX() - second->getX()
							   : diff;
		}
	};

	static void findAllFocusableComponents (Component* const parent, Array <Component*>& comps)
	{
		if (parent->getNumChildComponents() > 0)
		{
			Array <Component*> localComps;
			ScreenPositionComparator comparator;

			int i;
			for (i = parent->getNumChildComponents(); --i >= 0;)
			{
				Component* const c = parent->getChildComponent (i);

				if (c->isVisible() && c->isEnabled())
					localComps.addSorted (comparator, c);
			}

			for (i = 0; i < localComps.size(); ++i)
			{
				Component* const c = localComps.getUnchecked (i);

				if (c->getWantsKeyboardFocus())
					comps.add (c);

				if (! c->isFocusContainer())
					findAllFocusableComponents (c, comps);
			}
		}
	}
}

namespace KeyboardFocusHelpers
{
	Component* getIncrementedComponent (Component* const current, const int delta)
	{
		Component* focusContainer = current->getParentComponent();

		if (focusContainer != nullptr)
		{
			while (focusContainer->getParentComponent() != nullptr && ! focusContainer->isFocusContainer())
				focusContainer = focusContainer->getParentComponent();

			if (focusContainer != nullptr)
			{
				Array <Component*> comps;
				KeyboardFocusHelpers::findAllFocusableComponents (focusContainer, comps);

				if (comps.size() > 0)
				{
					const int index = comps.indexOf (current);
					return comps [(index + comps.size() + delta) % comps.size()];
				}
			}
		}

		return nullptr;
	}
}

Component* KeyboardFocusTraverser::getNextComponent (Component* current)
{
	return KeyboardFocusHelpers::getIncrementedComponent (current, 1);
}

Component* KeyboardFocusTraverser::getPreviousComponent (Component* current)
{
	return KeyboardFocusHelpers::getIncrementedComponent (current, -1);
}

Component* KeyboardFocusTraverser::getDefaultComponent (Component* parentComponent)
{
	Array <Component*> comps;

	if (parentComponent != nullptr)
		KeyboardFocusHelpers::findAllFocusableComponents (parentComponent, comps);

	return comps.getFirst();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_KeyboardFocusTraverser.cpp ***/


/*** Start of inlined file: juce_KeyListener.cpp ***/
BEGIN_JUCE_NAMESPACE

bool KeyListener::keyStateChanged (const bool, Component*)
{
	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_KeyListener.cpp ***/


/*** Start of inlined file: juce_KeyMappingEditorComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

// N.B. these two includes are put here deliberately to avoid problems with
// old GCCs failing on long include paths

class KeyMappingEditorComponent::ChangeKeyButton  : public Button
{
public:
	ChangeKeyButton (KeyMappingEditorComponent& owner_,
					 const CommandID commandID_,
					 const String& keyName,
					 const int keyNum_)
		: Button (keyName),
		  owner (owner_),
		  commandID (commandID_),
		  keyNum (keyNum_)
	{
		setWantsKeyboardFocus (false);
		setTriggeredOnMouseDown (keyNum >= 0);

		setTooltip (keyNum_ < 0 ? TRANS("adds a new key-mapping")
								: TRANS("click to change this key-mapping"));
	}

	void paintButton (Graphics& g, bool /*isOver*/, bool /*isDown*/)
	{
		getLookAndFeel().drawKeymapChangeButton (g, getWidth(), getHeight(), *this,
												 keyNum >= 0 ? getName() : String::empty);
	}

	static void menuCallback (int result, ChangeKeyButton* button)
	{
		if (button != nullptr)
		{
			switch (result)
			{
				case 1: button->assignNewKey(); break;
				case 2: button->owner.getMappings().removeKeyPress (button->commandID, button->keyNum); break;
				default: break;
			}
		}
	}

	void clicked()
	{
		if (keyNum >= 0)
		{
			// existing key clicked..
			PopupMenu m;
			m.addItem (1, TRANS("change this key-mapping"));
			m.addSeparator();
			m.addItem (2, TRANS("remove this key-mapping"));

			m.showMenuAsync (PopupMenu::Options(),
							 ModalCallbackFunction::forComponent (menuCallback, this));
		}
		else
		{
			assignNewKey();  // + button pressed..
		}
	}

	void fitToContent (const int h) noexcept
	{
		if (keyNum < 0)
		{
			setSize (h, h);
		}
		else
		{
			Font f (h * 0.6f);
			setSize (jlimit (h * 4, h * 8, 6 + f.getStringWidth (getName())), h);
		}
	}

	class KeyEntryWindow  : public AlertWindow
	{
	public:
		KeyEntryWindow (KeyMappingEditorComponent& owner_)
			: AlertWindow (TRANS("New key-mapping"),
						   TRANS("Please press a key combination now..."),
						   AlertWindow::NoIcon),
			  owner (owner_)
		{
			addButton (TRANS("Ok"), 1);
			addButton (TRANS("Cancel"), 0);

			// (avoid return + escape keys getting processed by the buttons..)
			for (int i = getNumChildComponents(); --i >= 0;)
				getChildComponent (i)->setWantsKeyboardFocus (false);

			setWantsKeyboardFocus (true);
			grabKeyboardFocus();
		}

		bool keyPressed (const KeyPress& key)
		{
			lastPress = key;
			String message (TRANS("Key: ") + owner.getDescriptionForKeyPress (key));

			const CommandID previousCommand = owner.getMappings().findCommandForKeyPress (key);

			if (previousCommand != 0)
				message << "\n\n" << TRANS("(Currently assigned to \"")
						<< owner.getMappings().getCommandManager()->getNameOfCommand (previousCommand) << "\")";

			setMessage (message);
			return true;
		}

		bool keyStateChanged (bool)
		{
			return true;
		}

		KeyPress lastPress;

	private:
		KeyMappingEditorComponent& owner;

		JUCE_DECLARE_NON_COPYABLE (KeyEntryWindow);
	};

	static void assignNewKeyCallback (int result, ChangeKeyButton* button, KeyPress newKey)
	{
		if (result != 0 && button != nullptr)
			button->setNewKey (newKey, true);
	}

	void setNewKey (const KeyPress& newKey, bool dontAskUser)
	{
		if (newKey.isValid())
		{
			const CommandID previousCommand = owner.getMappings().findCommandForKeyPress (newKey);

			if (previousCommand == 0 || dontAskUser)
			{
				owner.getMappings().removeKeyPress (newKey);

				if (keyNum >= 0)
					owner.getMappings().removeKeyPress (commandID, keyNum);

				owner.getMappings().addKeyPress (commandID, newKey, keyNum);
			}
			else
			{
				AlertWindow::showOkCancelBox (AlertWindow::WarningIcon,
											  TRANS("Change key-mapping"),
											  TRANS("This key is already assigned to the command \"")
												+ owner.getMappings().getCommandManager()->getNameOfCommand (previousCommand)
												+ TRANS("\"\n\nDo you want to re-assign it to this new command instead?"),
											  TRANS("Re-assign"),
											  TRANS("Cancel"),
											  this,
											  ModalCallbackFunction::forComponent (assignNewKeyCallback,
																				   this, KeyPress (newKey)));
			}
		}
	}

	static void keyChosen (int result, ChangeKeyButton* button)
	{
		if (result != 0 && button != nullptr && button->currentKeyEntryWindow != nullptr)
		{
			button->currentKeyEntryWindow->setVisible (false);
			button->setNewKey (button->currentKeyEntryWindow->lastPress, false);
		}

		button->currentKeyEntryWindow = nullptr;
	}

	void assignNewKey()
	{
		currentKeyEntryWindow = new KeyEntryWindow (owner);
		currentKeyEntryWindow->enterModalState (true, ModalCallbackFunction::forComponent (keyChosen, this));
	}

private:
	KeyMappingEditorComponent& owner;
	const CommandID commandID;
	const int keyNum;
	ScopedPointer<KeyEntryWindow> currentKeyEntryWindow;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ChangeKeyButton);
};

class KeyMappingEditorComponent::ItemComponent  : public Component
{
public:
	ItemComponent (KeyMappingEditorComponent& owner_, const CommandID commandID_)
		: owner (owner_), commandID (commandID_)
	{
		setInterceptsMouseClicks (false, true);

		const bool isReadOnly = owner.isCommandReadOnly (commandID);

		const Array <KeyPress> keyPresses (owner.getMappings().getKeyPressesAssignedToCommand (commandID));

		for (int i = 0; i < jmin ((int) maxNumAssignments, keyPresses.size()); ++i)
			addKeyPressButton (owner.getDescriptionForKeyPress (keyPresses.getReference (i)), i, isReadOnly);

		addKeyPressButton (String::empty, -1, isReadOnly);
	}

	void addKeyPressButton (const String& desc, const int index, const bool isReadOnly)
	{
		ChangeKeyButton* const b = new ChangeKeyButton (owner, commandID, desc, index);
		keyChangeButtons.add (b);

		b->setEnabled (! isReadOnly);
		b->setVisible (keyChangeButtons.size() <= (int) maxNumAssignments);
		addChildComponent (b);
	}

	void paint (Graphics& g)
	{
		g.setFont (getHeight() * 0.7f);
		g.setColour (findColour (KeyMappingEditorComponent::textColourId));

		g.drawFittedText (owner.getMappings().getCommandManager()->getNameOfCommand (commandID),
						  4, 0, jmax (40, getChildComponent (0)->getX() - 5), getHeight(),
						  Justification::centredLeft, true);
	}

	void resized()
	{
		int x = getWidth() - 4;

		for (int i = keyChangeButtons.size(); --i >= 0;)
		{
			ChangeKeyButton* const b = keyChangeButtons.getUnchecked(i);

			b->fitToContent (getHeight() - 2);
			b->setTopRightPosition (x, 1);
			x = b->getX() - 5;
		}
	}

private:
	KeyMappingEditorComponent& owner;
	OwnedArray<ChangeKeyButton> keyChangeButtons;
	const CommandID commandID;

	enum { maxNumAssignments = 3 };

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ItemComponent);
};

class KeyMappingEditorComponent::MappingItem  : public TreeViewItem
{
public:
	MappingItem (KeyMappingEditorComponent& owner_, const CommandID commandID_)
		: owner (owner_), commandID (commandID_)
	{
	}

	const String getUniqueName() const          { return String ((int) commandID) + "_id"; }
	bool mightContainSubItems()		 { return false; }
	int getItemHeight() const		   { return 20; }

	Component* createItemComponent()
	{
		return new ItemComponent (owner, commandID);
	}

private:
	KeyMappingEditorComponent& owner;
	const CommandID commandID;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (MappingItem);
};

class KeyMappingEditorComponent::CategoryItem  : public TreeViewItem
{
public:
	CategoryItem (KeyMappingEditorComponent& owner_, const String& name)
		: owner (owner_), categoryName (name)
	{
	}

	const String getUniqueName() const          { return categoryName + "_cat"; }
	bool mightContainSubItems()		 { return true; }
	int getItemHeight() const		   { return 28; }

	void paintItem (Graphics& g, int width, int height)
	{
		g.setFont (height * 0.6f, Font::bold);
		g.setColour (owner.findColour (KeyMappingEditorComponent::textColourId));

		g.drawText (categoryName,
					2, 0, width - 2, height,
					Justification::centredLeft, true);
	}

	void itemOpennessChanged (bool isNowOpen)
	{
		if (isNowOpen)
		{
			if (getNumSubItems() == 0)
			{
				Array <CommandID> commands (owner.getMappings().getCommandManager()->getCommandsInCategory (categoryName));

				for (int i = 0; i < commands.size(); ++i)
				{
					if (owner.shouldCommandBeIncluded (commands[i]))
						addSubItem (new MappingItem (owner, commands[i]));
				}
			}
		}
		else
		{
			clearSubItems();
		}
	}

private:
	KeyMappingEditorComponent& owner;
	String categoryName;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (CategoryItem);
};

class KeyMappingEditorComponent::TopLevelItem   : public TreeViewItem,
												  public ChangeListener,
												  public ButtonListener
{
public:
	TopLevelItem (KeyMappingEditorComponent& owner_)
		: owner (owner_)
	{
		setLinesDrawnForSubItems (false);
		owner.getMappings().addChangeListener (this);
	}

	~TopLevelItem()
	{
		owner.getMappings().removeChangeListener (this);
	}

	bool mightContainSubItems()		 { return true; }
	const String getUniqueName() const      { return "keys"; }

	void changeListenerCallback (ChangeBroadcaster*)
	{
		const OpennessRestorer opennessRestorer (*this);
		clearSubItems();

		const StringArray categories (owner.getMappings().getCommandManager()->getCommandCategories());

		for (int i = 0; i < categories.size(); ++i)
		{
			const Array <CommandID> commands (owner.getMappings().getCommandManager()->getCommandsInCategory (categories[i]));
			int count = 0;

			for (int j = 0; j < commands.size(); ++j)
				if (owner.shouldCommandBeIncluded (commands[j]))
					++count;

			if (count > 0)
				addSubItem (new CategoryItem (owner, categories[i]));
		}
	}

	static void resetToDefaultsCallback (int result, KeyMappingEditorComponent* owner)
	{
		if (result != 0 && owner != nullptr)
			owner->getMappings().resetToDefaultMappings();
	}

	void buttonClicked (Button*)
	{
		AlertWindow::showOkCancelBox (AlertWindow::QuestionIcon,
									  TRANS("Reset to defaults"),
									  TRANS("Are you sure you want to reset all the key-mappings to their default state?"),
									  TRANS("Reset"),
									  String::empty,
									  &owner,
									  ModalCallbackFunction::forComponent (resetToDefaultsCallback, &owner));
	}

private:
	KeyMappingEditorComponent& owner;
};

KeyMappingEditorComponent::KeyMappingEditorComponent (KeyPressMappingSet& mappingManager,
													  const bool showResetToDefaultButton)
	: mappings (mappingManager),
	  resetButton (TRANS ("reset to defaults"))
{
	treeItem = new TopLevelItem (*this);

	if (showResetToDefaultButton)
	{
		addAndMakeVisible (&resetButton);
		resetButton.addListener (treeItem);
	}

	addAndMakeVisible (&tree);
	tree.setColour (TreeView::backgroundColourId, findColour (backgroundColourId));
	tree.setRootItemVisible (false);
	tree.setDefaultOpenness (true);
	tree.setRootItem (treeItem);
}

KeyMappingEditorComponent::~KeyMappingEditorComponent()
{
	tree.setRootItem (nullptr);
}

void KeyMappingEditorComponent::setColours (const Colour& mainBackground,
											const Colour& textColour)
{
	setColour (backgroundColourId, mainBackground);
	setColour (textColourId, textColour);
	tree.setColour (TreeView::backgroundColourId, mainBackground);
}

void KeyMappingEditorComponent::parentHierarchyChanged()
{
	treeItem->changeListenerCallback (nullptr);
}

void KeyMappingEditorComponent::resized()
{
	int h = getHeight();

	if (resetButton.isVisible())
	{
		const int buttonHeight = 20;
		h -= buttonHeight + 8;
		int x = getWidth() - 8;

		resetButton.changeWidthToFitText (buttonHeight);
		resetButton.setTopRightPosition (x, h + 6);
	}

	tree.setBounds (0, 0, getWidth(), h);
}

bool KeyMappingEditorComponent::shouldCommandBeIncluded (const CommandID commandID)
{
	const ApplicationCommandInfo* const ci = mappings.getCommandManager()->getCommandForID (commandID);

	return ci != nullptr && (ci->flags & ApplicationCommandInfo::hiddenFromKeyEditor) == 0;
}

bool KeyMappingEditorComponent::isCommandReadOnly (const CommandID commandID)
{
	const ApplicationCommandInfo* const ci = mappings.getCommandManager()->getCommandForID (commandID);

	return ci != nullptr && (ci->flags & ApplicationCommandInfo::readOnlyInKeyEditor) != 0;
}

const String KeyMappingEditorComponent::getDescriptionForKeyPress (const KeyPress& key)
{
	return key.getTextDescription();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_KeyMappingEditorComponent.cpp ***/


/*** Start of inlined file: juce_KeyPress.cpp ***/
BEGIN_JUCE_NAMESPACE

KeyPress::KeyPress() noexcept
	: keyCode (0),
	  mods (0),
	  textCharacter (0)
{
}

KeyPress::KeyPress (const int keyCode_,
					const ModifierKeys& mods_,
					const juce_wchar textCharacter_) noexcept
	: keyCode (keyCode_),
	  mods (mods_),
	  textCharacter (textCharacter_)
{
}

KeyPress::KeyPress (const int keyCode_) noexcept
	: keyCode (keyCode_),
	  textCharacter (0)
{
}

KeyPress::KeyPress (const KeyPress& other) noexcept
	: keyCode (other.keyCode),
	  mods (other.mods),
	  textCharacter (other.textCharacter)
{
}

KeyPress& KeyPress::operator= (const KeyPress& other) noexcept
{
	keyCode = other.keyCode;
	mods = other.mods;
	textCharacter = other.textCharacter;

	return *this;
}

bool KeyPress::operator== (const KeyPress& other) const noexcept
{
	return mods.getRawFlags() == other.mods.getRawFlags()
			&& (textCharacter == other.textCharacter
				 || textCharacter == 0
				 || other.textCharacter == 0)
			&& (keyCode == other.keyCode
				 || (keyCode < 256
					  && other.keyCode < 256
					  && CharacterFunctions::toLowerCase ((juce_wchar) keyCode)
						   == CharacterFunctions::toLowerCase ((juce_wchar) other.keyCode)));
}

bool KeyPress::operator!= (const KeyPress& other) const noexcept
{
	return ! operator== (other);
}

bool KeyPress::isCurrentlyDown() const
{
	return isKeyCurrentlyDown (keyCode)
			&& (ModifierKeys::getCurrentModifiers().getRawFlags() & ModifierKeys::allKeyboardModifiers)
				  == (mods.getRawFlags() & ModifierKeys::allKeyboardModifiers);
}

namespace KeyPressHelpers
{
	struct KeyNameAndCode
	{
		const char* name;
		int code;
	};

	const KeyNameAndCode translations[] =
	{
		{ "spacebar",       KeyPress::spaceKey },
		{ "return",         KeyPress::returnKey },
		{ "escape",         KeyPress::escapeKey },
		{ "backspace",      KeyPress::backspaceKey },
		{ "cursor left",    KeyPress::leftKey },
		{ "cursor right",   KeyPress::rightKey },
		{ "cursor up",      KeyPress::upKey },
		{ "cursor down",    KeyPress::downKey },
		{ "page up",        KeyPress::pageUpKey },
		{ "page down",      KeyPress::pageDownKey },
		{ "home",           KeyPress::homeKey },
		{ "end",            KeyPress::endKey },
		{ "delete",         KeyPress::deleteKey },
		{ "insert",         KeyPress::insertKey },
		{ "tab",            KeyPress::tabKey },
		{ "play",           KeyPress::playKey },
		{ "stop",           KeyPress::stopKey },
		{ "fast forward",   KeyPress::fastForwardKey },
		{ "rewind",         KeyPress::rewindKey }
	};

	struct ModifierDescription
	{
		const char* name;
		int flag;
	};

	static const ModifierDescription modifierNames[] =
	{
		{ "ctrl",      ModifierKeys::ctrlModifier },
		{ "control",   ModifierKeys::ctrlModifier },
		{ "ctl",       ModifierKeys::ctrlModifier },
		{ "shift",     ModifierKeys::shiftModifier },
		{ "shft",      ModifierKeys::shiftModifier },
		{ "alt",       ModifierKeys::altModifier },
		{ "option",    ModifierKeys::altModifier },
		{ "command",   ModifierKeys::commandModifier },
		{ "cmd",       ModifierKeys::commandModifier }
	};

	const char* numberPadPrefix() noexcept      { return "numpad "; }

	int getNumpadKeyCode (const String& desc)
	{
		if (desc.containsIgnoreCase (numberPadPrefix()))
		{
			const juce_wchar lastChar = desc.trimEnd().getLastCharacter();

			switch (lastChar)
			{
				case '0': case '1': case '2': case '3': case '4':
				case '5': case '6': case '7': case '8': case '9':
					return KeyPress::numberPad0 + lastChar - '0';

				case '+':   return KeyPress::numberPadAdd;
				case '-':   return KeyPress::numberPadSubtract;
				case '*':   return KeyPress::numberPadMultiply;
				case '/':   return KeyPress::numberPadDivide;
				case '.':   return KeyPress::numberPadDecimalPoint;
				case '=':   return KeyPress::numberPadEquals;

				default:	break;
			}

			if (desc.endsWith ("separator"))  return KeyPress::numberPadSeparator;
			if (desc.endsWith ("delete"))     return KeyPress::numberPadDelete;
		}

		return 0;
	}
}

const KeyPress KeyPress::createFromDescription (const String& desc)
{
	int modifiers = 0;

	for (int i = 0; i < numElementsInArray (KeyPressHelpers::modifierNames); ++i)
		if (desc.containsWholeWordIgnoreCase (KeyPressHelpers::modifierNames[i].name))
			modifiers |= KeyPressHelpers::modifierNames[i].flag;

	int key = 0;

	for (int i = 0; i < numElementsInArray (KeyPressHelpers::translations); ++i)
	{
		if (desc.containsWholeWordIgnoreCase (String (KeyPressHelpers::translations[i].name)))
		{
			key = KeyPressHelpers::translations[i].code;
			break;
		}
	}

	if (key == 0)
		key = KeyPressHelpers::getNumpadKeyCode (desc);

	if (key == 0)
	{
		// see if it's a function key..
		if (! desc.containsChar ('#')) // avoid mistaking hex-codes like "#f1"
			for (int i = 1; i <= 12; ++i)
				if (desc.containsWholeWordIgnoreCase ("f" + String (i)))
					key = F1Key + i - 1;

		if (key == 0)
		{
			// give up and use the hex code..
			const int hexCode = desc.fromFirstOccurrenceOf ("#", false, false)
									.toLowerCase()
									.retainCharacters ("0123456789abcdef")
									.getHexValue32();

			if (hexCode > 0)
				key = hexCode;
			else
				key = CharacterFunctions::toUpperCase (desc.getLastCharacter());
		}
	}

	return KeyPress (key, ModifierKeys (modifiers), 0);
}

String KeyPress::getTextDescription() const
{
	String desc;

	if (keyCode > 0)
	{
		// some keyboard layouts use a shift-key to get the slash, but in those cases, we
		// want to store it as being a slash, not shift+whatever.
		if (textCharacter == '/')
			return "/";

		if (mods.isCtrlDown())
			desc << "ctrl + ";

		if (mods.isShiftDown())
			desc << "shift + ";

	   #if JUCE_MAC
		if (mods.isAltDown())
			desc << "option + ";

		// only do this on the mac, because on Windows ctrl and command are the same,
		// and this would get confusing
		if (mods.isCommandDown())
			desc << "command + ";
	   #else
		if (mods.isAltDown())
			desc << "alt + ";
	   #endif

		for (int i = 0; i < numElementsInArray (KeyPressHelpers::translations); ++i)
			if (keyCode == KeyPressHelpers::translations[i].code)
				return desc + KeyPressHelpers::translations[i].name;

		if (keyCode >= F1Key && keyCode <= F16Key)		  desc << 'F' << (1 + keyCode - F1Key);
		else if (keyCode >= numberPad0 && keyCode <= numberPad9)	desc << KeyPressHelpers::numberPadPrefix() << (keyCode - numberPad0);
		else if (keyCode >= 33 && keyCode < 176)	desc += CharacterFunctions::toUpperCase ((juce_wchar) keyCode);
		else if (keyCode == numberPadAdd)		   desc << KeyPressHelpers::numberPadPrefix() << '+';
		else if (keyCode == numberPadSubtract)	  desc << KeyPressHelpers::numberPadPrefix() << '-';
		else if (keyCode == numberPadMultiply)	  desc << KeyPressHelpers::numberPadPrefix() << '*';
		else if (keyCode == numberPadDivide)		desc << KeyPressHelpers::numberPadPrefix() << '/';
		else if (keyCode == numberPadSeparator)         desc << KeyPressHelpers::numberPadPrefix() << "separator";
		else if (keyCode == numberPadDecimalPoint)	  desc << KeyPressHelpers::numberPadPrefix() << '.';
		else if (keyCode == numberPadDelete)            desc << KeyPressHelpers::numberPadPrefix() << "delete";
		else						desc << '#' << String::toHexString (keyCode);
	}

	return desc;
}

String KeyPress::getTextDescriptionWithIcons() const
{
   #if JUCE_MAC
	return getTextDescription().replace ("shift + ", String::charToString (0x21e7))
							   .replace ("command + ", String::charToString (0x2318))
							   .replace ("option + ", String::charToString (0x2325))
							   .replace ("ctrl + ", String::charToString (0x2303))
							   .replace ("return", String::charToString (0x23ce))
							   .replace ("cursor left", String::charToString (0x2190))
							   .replace ("cursor right", String::charToString (0x2192))
							   .replace ("cursor up", String::charToString (0x2191))
							   .replace ("cursor down", String::charToString (0x2193))
							   .replace ("backspace", String::charToString (0x232b))
							   .replace ("delete", String::charToString (0x2326));
   #else
	return getTextDescription();
   #endif
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_KeyPress.cpp ***/


/*** Start of inlined file: juce_KeyPressMappingSet.cpp ***/
BEGIN_JUCE_NAMESPACE

KeyPressMappingSet::KeyPressMappingSet (ApplicationCommandManager* const commandManager_)
	: commandManager (commandManager_)
{
	// A manager is needed to get the descriptions of commands, and will be called when
	// a command is invoked. So you can't leave this null..
	jassert (commandManager_ != nullptr);

	Desktop::getInstance().addFocusChangeListener (this);
}

KeyPressMappingSet::KeyPressMappingSet (const KeyPressMappingSet& other)
	: commandManager (other.commandManager)
{
	Desktop::getInstance().addFocusChangeListener (this);
}

KeyPressMappingSet::~KeyPressMappingSet()
{
	Desktop::getInstance().removeFocusChangeListener (this);
}

const Array <KeyPress> KeyPressMappingSet::getKeyPressesAssignedToCommand (const CommandID commandID) const
{
	for (int i = 0; i < mappings.size(); ++i)
		if (mappings.getUnchecked(i)->commandID == commandID)
			return mappings.getUnchecked (i)->keypresses;

	return Array <KeyPress> ();
}

void KeyPressMappingSet::addKeyPress (const CommandID commandID,
									  const KeyPress& newKeyPress,
									  int insertIndex)
{
	// If you specify an upper-case letter but no shift key, how is the user supposed to press it!?
	// Stick to lower-case letters when defining a keypress, to avoid ambiguity.
	jassert (! (CharacterFunctions::isUpperCase (newKeyPress.getTextCharacter())
				 && ! newKeyPress.getModifiers().isShiftDown()));

	if (findCommandForKeyPress (newKeyPress) != commandID)
	{
		removeKeyPress (newKeyPress);

		if (newKeyPress.isValid())
		{
			for (int i = mappings.size(); --i >= 0;)
			{
				if (mappings.getUnchecked(i)->commandID == commandID)
				{
					mappings.getUnchecked(i)->keypresses.insert (insertIndex, newKeyPress);

					sendChangeMessage();
					return;
				}
			}

			const ApplicationCommandInfo* const ci = commandManager->getCommandForID (commandID);

			if (ci != nullptr)
			{
				CommandMapping* const cm = new CommandMapping();
				cm->commandID = commandID;
				cm->keypresses.add (newKeyPress);
				cm->wantsKeyUpDownCallbacks = (ci->flags & ApplicationCommandInfo::wantsKeyUpDownCallbacks) != 0;

				mappings.add (cm);
				sendChangeMessage();
			}
		}
	}
}

void KeyPressMappingSet::resetToDefaultMappings()
{
	mappings.clear();

	for (int i = 0; i < commandManager->getNumCommands(); ++i)
	{
		const ApplicationCommandInfo* const ci = commandManager->getCommandForIndex (i);

		for (int j = 0; j < ci->defaultKeypresses.size(); ++j)
		{
			addKeyPress (ci->commandID,
						 ci->defaultKeypresses.getReference (j));
		}
	}

	sendChangeMessage();
}

void KeyPressMappingSet::resetToDefaultMapping (const CommandID commandID)
{
	clearAllKeyPresses (commandID);

	const ApplicationCommandInfo* const ci = commandManager->getCommandForID (commandID);

	for (int j = 0; j < ci->defaultKeypresses.size(); ++j)
	{
		addKeyPress (ci->commandID,
					 ci->defaultKeypresses.getReference (j));
	}
}

void KeyPressMappingSet::clearAllKeyPresses()
{
	if (mappings.size() > 0)
	{
		sendChangeMessage();
		mappings.clear();
	}
}

void KeyPressMappingSet::clearAllKeyPresses (const CommandID commandID)
{
	for (int i = mappings.size(); --i >= 0;)
	{
		if (mappings.getUnchecked(i)->commandID == commandID)
		{
			mappings.remove (i);
			sendChangeMessage();
		}
	}
}

void KeyPressMappingSet::removeKeyPress (const KeyPress& keypress)
{
	if (keypress.isValid())
	{
		for (int i = mappings.size(); --i >= 0;)
		{
			CommandMapping* const cm = mappings.getUnchecked(i);

			for (int j = cm->keypresses.size(); --j >= 0;)
			{
				if (keypress == cm->keypresses [j])
				{
					cm->keypresses.remove (j);
					sendChangeMessage();
				}
			}
		}
	}
}

void KeyPressMappingSet::removeKeyPress (const CommandID commandID, const int keyPressIndex)
{
	for (int i = mappings.size(); --i >= 0;)
	{
		if (mappings.getUnchecked(i)->commandID == commandID)
		{
			mappings.getUnchecked(i)->keypresses.remove (keyPressIndex);
			sendChangeMessage();
			break;
		}
	}
}

CommandID KeyPressMappingSet::findCommandForKeyPress (const KeyPress& keyPress) const noexcept
{
	for (int i = 0; i < mappings.size(); ++i)
		if (mappings.getUnchecked(i)->keypresses.contains (keyPress))
			return mappings.getUnchecked(i)->commandID;

	return 0;
}

bool KeyPressMappingSet::containsMapping (const CommandID commandID, const KeyPress& keyPress) const noexcept
{
	for (int i = mappings.size(); --i >= 0;)
		if (mappings.getUnchecked(i)->commandID == commandID)
			return mappings.getUnchecked(i)->keypresses.contains (keyPress);

	return false;
}

void KeyPressMappingSet::invokeCommand (const CommandID commandID,
										const KeyPress& key,
										const bool isKeyDown,
										const int millisecsSinceKeyPressed,
										Component* const originatingComponent) const
{
	ApplicationCommandTarget::InvocationInfo info (commandID);

	info.invocationMethod = ApplicationCommandTarget::InvocationInfo::fromKeyPress;
	info.isKeyDown = isKeyDown;
	info.keyPress = key;
	info.millisecsSinceKeyPressed = millisecsSinceKeyPressed;
	info.originatingComponent = originatingComponent;

	commandManager->invoke (info, false);
}

bool KeyPressMappingSet::restoreFromXml (const XmlElement& xmlVersion)
{
	if (xmlVersion.hasTagName ("KEYMAPPINGS"))
	{
		if (xmlVersion.getBoolAttribute ("basedOnDefaults", true))
		{
			// if the XML was created as a set of differences from the default mappings,
			// (i.e. by calling createXml (true)), then we need to first restore the defaults.
			resetToDefaultMappings();
		}
		else
		{
			// if the XML was created calling createXml (false), then we need to clear all
			// the keys and treat the xml as describing the entire set of mappings.
			clearAllKeyPresses();
		}

		forEachXmlChildElement (xmlVersion, map)
		{
			const CommandID commandId = map->getStringAttribute ("commandId").getHexValue32();

			if (commandId != 0)
			{
				const KeyPress key (KeyPress::createFromDescription (map->getStringAttribute ("key")));

				if (map->hasTagName ("MAPPING"))
				{
					addKeyPress (commandId, key);
				}
				else if (map->hasTagName ("UNMAPPING"))
				{
					if (containsMapping (commandId, key))
						removeKeyPress (key);
				}
			}
		}

		return true;
	}

	return false;
}

XmlElement* KeyPressMappingSet::createXml (const bool saveDifferencesFromDefaultSet) const
{
	ScopedPointer <KeyPressMappingSet> defaultSet;

	if (saveDifferencesFromDefaultSet)
	{
		defaultSet = new KeyPressMappingSet (commandManager);
		defaultSet->resetToDefaultMappings();
	}

	XmlElement* const doc = new XmlElement ("KEYMAPPINGS");

	doc->setAttribute ("basedOnDefaults", saveDifferencesFromDefaultSet);

	int i;
	for (i = 0; i < mappings.size(); ++i)
	{
		const CommandMapping* const cm = mappings.getUnchecked(i);

		for (int j = 0; j < cm->keypresses.size(); ++j)
		{
			if (defaultSet == nullptr
				 || ! defaultSet->containsMapping (cm->commandID, cm->keypresses.getReference (j)))
			{
				XmlElement* const map = doc->createNewChildElement ("MAPPING");

				map->setAttribute ("commandId", String::toHexString ((int) cm->commandID));
				map->setAttribute ("description", commandManager->getDescriptionOfCommand (cm->commandID));
				map->setAttribute ("key", cm->keypresses.getReference (j).getTextDescription());
			}
		}
	}

	if (defaultSet != nullptr)
	{
		for (i = 0; i < defaultSet->mappings.size(); ++i)
		{
			const CommandMapping* const cm = defaultSet->mappings.getUnchecked(i);

			for (int j = 0; j < cm->keypresses.size(); ++j)
			{
				if (! containsMapping (cm->commandID, cm->keypresses.getReference (j)))
				{
					XmlElement* const map = doc->createNewChildElement ("UNMAPPING");

					map->setAttribute ("commandId", String::toHexString ((int) cm->commandID));
					map->setAttribute ("description", commandManager->getDescriptionOfCommand (cm->commandID));
					map->setAttribute ("key", cm->keypresses.getReference (j).getTextDescription());
				}
			}
		}
	}

	return doc;
}

bool KeyPressMappingSet::keyPressed (const KeyPress& key,
									 Component* originatingComponent)
{
	bool used = false;

	const CommandID commandID = findCommandForKeyPress (key);

	const ApplicationCommandInfo* const ci = commandManager->getCommandForID (commandID);

	if (ci != nullptr
		 && (ci->flags & ApplicationCommandInfo::wantsKeyUpDownCallbacks) == 0)
	{
		ApplicationCommandInfo info (0);

		if (commandManager->getTargetForCommand (commandID, info) != 0
			 && (info.flags & ApplicationCommandInfo::isDisabled) == 0)
		{
			invokeCommand (commandID, key, true, 0, originatingComponent);
			used = true;
		}
		else
		{
			if (originatingComponent != nullptr)
				originatingComponent->getLookAndFeel().playAlertSound();
		}
	}

	return used;
}

bool KeyPressMappingSet::keyStateChanged (const bool /*isKeyDown*/, Component* originatingComponent)
{
	bool used = false;
	const uint32 now = Time::getMillisecondCounter();

	for (int i = mappings.size(); --i >= 0;)
	{
		CommandMapping* const cm =  mappings.getUnchecked(i);

		if (cm->wantsKeyUpDownCallbacks)
		{
			for (int j = cm->keypresses.size(); --j >= 0;)
			{
				const KeyPress key (cm->keypresses.getReference (j));
				const bool isDown = key.isCurrentlyDown();

				int keyPressEntryIndex = 0;
				bool wasDown = false;

				for (int k = keysDown.size(); --k >= 0;)
				{
					if (key == keysDown.getUnchecked(k)->key)
					{
						keyPressEntryIndex = k;
						wasDown = true;
						used = true;
						break;
					}
				}

				if (isDown != wasDown)
				{
					int millisecs = 0;

					if (isDown)
					{
						KeyPressTime* const k = new KeyPressTime();
						k->key = key;
						k->timeWhenPressed = now;

						keysDown.add (k);
					}
					else
					{
						const uint32 pressTime = keysDown.getUnchecked (keyPressEntryIndex)->timeWhenPressed;

						if (now > pressTime)
							millisecs = now - pressTime;

						keysDown.remove (keyPressEntryIndex);
					}

					invokeCommand (cm->commandID, key, isDown, millisecs, originatingComponent);
					used = true;
				}
			}
		}
	}

	return used;
}

void KeyPressMappingSet::globalFocusChanged (Component* focusedComponent)
{
	if (focusedComponent != nullptr)
		focusedComponent->keyStateChanged (false);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_KeyPressMappingSet.cpp ***/


/*** Start of inlined file: juce_ModifierKeys.cpp ***/
BEGIN_JUCE_NAMESPACE

ModifierKeys::ModifierKeys (const int flags_) noexcept
	: flags (flags_)
{
}

ModifierKeys::ModifierKeys (const ModifierKeys& other) noexcept
	: flags (other.flags)
{
}

ModifierKeys& ModifierKeys::operator= (const ModifierKeys& other) noexcept
{
	flags = other.flags;
	return *this;
}

ModifierKeys ModifierKeys::currentModifiers;

const ModifierKeys ModifierKeys::getCurrentModifiers() noexcept
{
	return currentModifiers;
}

int ModifierKeys::getNumMouseButtonsDown() const noexcept
{
	int num = 0;

	if (isLeftButtonDown())	 ++num;
	if (isRightButtonDown())	++num;
	if (isMiddleButtonDown())   ++num;

	return num;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ModifierKeys.cpp ***/


/*** Start of inlined file: juce_CaretComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

CaretComponent::CaretComponent (Component* const keyFocusOwner)
	: owner (keyFocusOwner)
{
	setAlwaysOnTop (true);
	setInterceptsMouseClicks (false, false);
}

CaretComponent::~CaretComponent()
{
}

void CaretComponent::paint (Graphics& g)
{
	g.fillAll (findColour (caretColourId, true));
}

void CaretComponent::timerCallback()
{
	setVisible (shouldBeShown() && ! isVisible());
}

void CaretComponent::setCaretPosition (const Rectangle<int>& characterArea)
{
	startTimer (380);
	setVisible (shouldBeShown());
	setBounds (characterArea.withWidth (2));
}

bool CaretComponent::shouldBeShown() const
{
	return owner == nullptr || (owner->hasKeyboardFocus (true)
								 && ! owner->isCurrentlyBlockedByAnotherModalComponent());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_CaretComponent.cpp ***/


/*** Start of inlined file: juce_ComponentAnimator.cpp ***/
BEGIN_JUCE_NAMESPACE

class ComponentAnimator::AnimationTask
{
public:
	AnimationTask (Component* const comp)
		: component (comp)
	{
	}

	void reset (const Rectangle<int>& finalBounds,
				float finalAlpha,
				int millisecondsToSpendMoving,
				bool useProxyComponent,
				double startSpeed_, double endSpeed_)
	{
		msElapsed = 0;
		msTotal = jmax (1, millisecondsToSpendMoving);
		lastProgress = 0;
		destination = finalBounds;
		destAlpha = finalAlpha;

		isMoving = (finalBounds != component->getBounds());
		isChangingAlpha = (finalAlpha != component->getAlpha());

		left = component->getX();
		top = component->getY();
		right = component->getRight();
		bottom = component->getBottom();
		alpha = component->getAlpha();

		const double invTotalDistance = 4.0 / (startSpeed_ + endSpeed_ + 2.0);
		startSpeed = jmax (0.0, startSpeed_ * invTotalDistance);
		midSpeed = invTotalDistance;
		endSpeed = jmax (0.0, endSpeed_ * invTotalDistance);

		if (useProxyComponent)
			proxy = new ProxyComponent (*component);
		else
			proxy = nullptr;

		component->setVisible (! useProxyComponent);
	}

	bool useTimeslice (const int elapsed)
	{
		Component* const c = proxy != nullptr ? static_cast <Component*> (proxy)
											  : static_cast <Component*> (component);

		if (c != nullptr)
		{
			msElapsed += elapsed;
			double newProgress = msElapsed / (double) msTotal;

			if (newProgress >= 0 && newProgress < 1.0)
			{
				newProgress = timeToDistance (newProgress);
				const double delta = (newProgress - lastProgress) / (1.0 - lastProgress);
				jassert (newProgress >= lastProgress);
				lastProgress = newProgress;

				if (delta < 1.0)
				{
					bool stillBusy = false;

					if (isMoving)
					{
						left += (destination.getX() - left) * delta;
						top += (destination.getY() - top) * delta;
						right += (destination.getRight() - right) * delta;
						bottom += (destination.getBottom() - bottom) * delta;

						const Rectangle<int> newBounds (roundToInt (left),
														roundToInt (top),
														roundToInt (right - left),
														roundToInt (bottom - top));

						if (newBounds != destination)
						{
							c->setBounds (newBounds);
							stillBusy = true;
						}
					}

					if (isChangingAlpha)
					{
						alpha += (destAlpha - alpha) * delta;
						c->setAlpha ((float) alpha);
						stillBusy = true;
					}

					if (stillBusy)
						return true;
				}
			}
		}

		moveToFinalDestination();
		return false;
	}

	void moveToFinalDestination()
	{
		if (component != nullptr)
		{
			component->setAlpha ((float) destAlpha);
			component->setBounds (destination);

			if (proxy != nullptr)
				component->setVisible (destAlpha > 0);
		}
	}

	class ProxyComponent  : public Component
	{
	public:
		ProxyComponent (Component& component)
			: image (component.createComponentSnapshot (component.getLocalBounds()))
		{
			setBounds (component.getBounds());
			setTransform (component.getTransform());
			setAlpha (component.getAlpha());
			setInterceptsMouseClicks (false, false);

			Component* const parent = component.getParentComponent();

			if (parent != nullptr)
				parent->addAndMakeVisible (this);
			else if (component.isOnDesktop() && component.getPeer() != nullptr)
				addToDesktop (component.getPeer()->getStyleFlags() | ComponentPeer::windowIgnoresKeyPresses);
			else
				jassertfalse; // seem to be trying to animate a component that's not visible..

			setVisible (true);
			toBehind (&component);
		}

		void paint (Graphics& g)
		{
			g.setOpacity (1.0f);
			g.drawImage (image, 0, 0, getWidth(), getHeight(),
						 0, 0, image.getWidth(), image.getHeight());
		}

	private:
		Image image;

		JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ProxyComponent);
	};

	WeakReference<Component> component;
	ScopedPointer<Component> proxy;

	Rectangle<int> destination;
	double destAlpha;

	int msElapsed, msTotal;
	double startSpeed, midSpeed, endSpeed, lastProgress;
	double left, top, right, bottom, alpha;
	bool isMoving, isChangingAlpha;

private:
	double timeToDistance (const double time) const noexcept
	{
		return (time < 0.5) ? time * (startSpeed + time * (midSpeed - startSpeed))
							: 0.5 * (startSpeed + 0.5 * (midSpeed - startSpeed))
								+ (time - 0.5) * (midSpeed + (time - 0.5) * (endSpeed - midSpeed));
	}
};

ComponentAnimator::ComponentAnimator()
	: lastTime (0)
{
}

ComponentAnimator::~ComponentAnimator()
{
}

ComponentAnimator::AnimationTask* ComponentAnimator::findTaskFor (Component* const component) const
{
	for (int i = tasks.size(); --i >= 0;)
		if (component == tasks.getUnchecked(i)->component.get())
			return tasks.getUnchecked(i);

	return nullptr;
}

void ComponentAnimator::animateComponent (Component* const component,
										  const Rectangle<int>& finalBounds,
										  const float finalAlpha,
										  const int millisecondsToSpendMoving,
										  const bool useProxyComponent,
										  const double startSpeed,
										  const double endSpeed)
{
	// the speeds must be 0 or greater!
	jassert (startSpeed >= 0 && endSpeed >= 0)

	if (component != nullptr)
	{
		AnimationTask* at = findTaskFor (component);

		if (at == nullptr)
		{
			at = new AnimationTask (component);
			tasks.add (at);
			sendChangeMessage();
		}

		at->reset (finalBounds, finalAlpha, millisecondsToSpendMoving,
				   useProxyComponent, startSpeed, endSpeed);

		if (! isTimerRunning())
		{
			lastTime = Time::getMillisecondCounter();
			startTimer (1000 / 50);
		}
	}
}

void ComponentAnimator::fadeOut (Component* component, int millisecondsToTake)
{
	if (component != nullptr)
	{
		if (component->isShowing() && millisecondsToTake > 0)
			animateComponent (component, component->getBounds(), 0.0f, millisecondsToTake, true, 1.0, 1.0);

		component->setVisible (false);
	}
}

void ComponentAnimator::fadeIn (Component* component, int millisecondsToTake)
{
	if (component != nullptr && ! (component->isVisible() && component->getAlpha() == 1.0f))
	{
		component->setAlpha (0.0f);
		component->setVisible (true);
		animateComponent (component, component->getBounds(), 1.0f, millisecondsToTake, false, 1.0, 1.0);
	}
}

void ComponentAnimator::cancelAllAnimations (const bool moveComponentsToTheirFinalPositions)
{
	if (tasks.size() > 0)
	{
		if (moveComponentsToTheirFinalPositions)
			for (int i = tasks.size(); --i >= 0;)
				tasks.getUnchecked(i)->moveToFinalDestination();

		tasks.clear();
		sendChangeMessage();
	}
}

void ComponentAnimator::cancelAnimation (Component* const component,
										 const bool moveComponentToItsFinalPosition)
{
	AnimationTask* const at = findTaskFor (component);

	if (at != nullptr)
	{
		if (moveComponentToItsFinalPosition)
			at->moveToFinalDestination();

		tasks.removeObject (at);
		sendChangeMessage();
	}
}

const Rectangle<int> ComponentAnimator::getComponentDestination (Component* const component)
{
	jassert (component != nullptr);
	AnimationTask* const at = findTaskFor (component);

	if (at != nullptr)
		return at->destination;

	return component->getBounds();
}

bool ComponentAnimator::isAnimating (Component* component) const
{
	return findTaskFor (component) != nullptr;
}

void ComponentAnimator::timerCallback()
{
	const uint32 timeNow = Time::getMillisecondCounter();

	if (lastTime == 0 || lastTime == timeNow)
		lastTime = timeNow;

	const int elapsed = timeNow - lastTime;

	for (int i = tasks.size(); --i >= 0;)
	{
		if (! tasks.getUnchecked(i)->useTimeslice (elapsed))
		{
			tasks.remove (i);
			sendChangeMessage();
		}
	}

	lastTime = timeNow;

	if (tasks.size() == 0)
		stopTimer();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComponentAnimator.cpp ***/


/*** Start of inlined file: juce_ComponentBuilder.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace ComponentBuilderHelpers
{
	String getStateId (const ValueTree& state)
	{
		return state [ComponentBuilder::idProperty].toString();
	}

	Component* findComponentWithID (OwnedArray<Component>& components, const String& compId)
	{
		jassert (compId.isNotEmpty());

		for (int i = components.size(); --i >= 0;)
		{
			Component* const c = components.getUnchecked (i);

			if (c->getComponentID() == compId)
				return components.removeAndReturn (i);
		}

		return nullptr;
	}

	Component* findComponentWithID (Component* const c, const String& compId)
	{
		jassert (compId.isNotEmpty());
		if (c->getComponentID() == compId)
			return c;

		for (int i = c->getNumChildComponents(); --i >= 0;)
		{
			Component* const child = findComponentWithID (c->getChildComponent (i), compId);

			if (child != nullptr)
				return child;
		}

		return nullptr;
	}

	Component* createNewComponent (ComponentBuilder::TypeHandler& type,
								   const ValueTree& state, Component* parent)
	{
		Component* const c = type.addNewComponentFromState (state, parent);
		jassert (c != nullptr && c->getParentComponent() == parent);
		c->setComponentID (getStateId (state));
		return c;
	}

	void updateComponent (ComponentBuilder& builder, const ValueTree& state)
	{
		Component* topLevelComp = builder.getManagedComponent();

		if (topLevelComp != nullptr)
		{
			ComponentBuilder::TypeHandler* const type = builder.getHandlerForState (state);
			const String uid (getStateId (state));

			if (type == nullptr || uid.isEmpty())
			{
				// ..handle the case where a child of the actual state node has changed.
				if (state.getParent().isValid())
					updateComponent (builder, state.getParent());
			}
			else
			{
				Component* const changedComp = findComponentWithID (topLevelComp, uid);

				if (changedComp != nullptr)
					type->updateComponentFromState (changedComp, state);
			}
		}
	}
}

const Identifier ComponentBuilder::idProperty ("id");

ComponentBuilder::ComponentBuilder (const ValueTree& state_)
	: state (state_), imageProvider (nullptr)
{
	state.addListener (this);
}

ComponentBuilder::~ComponentBuilder()
{
	state.removeListener (this);

   #if JUCE_DEBUG
	// Don't delete the managed component!! The builder owns that component, and will delete
	// it automatically when it gets deleted.
	jassert (componentRef.get() == static_cast <Component*> (component));
   #endif
}

Component* ComponentBuilder::getManagedComponent()
{
	if (component == nullptr)
	{
		component = createComponent();

	   #if JUCE_DEBUG
		componentRef = component;
	   #endif
	}

	return component;
}

Component* ComponentBuilder::createComponent()
{
	jassert (types.size() > 0);  // You need to register all the necessary types before you can load a component!

	TypeHandler* const type = getHandlerForState (state);
	jassert (type != nullptr); // trying to create a component from an unknown type of ValueTree

	return type != nullptr ? ComponentBuilderHelpers::createNewComponent (*type, state, nullptr) : nullptr;
}

void ComponentBuilder::registerTypeHandler (ComponentBuilder::TypeHandler* const type)
{
	jassert (type != nullptr);

	// Don't try to move your types around! Once a type has been added to a builder, the
	// builder owns it, and you should leave it alone!
	jassert (type->builder == nullptr);

	types.add (type);
	type->builder = this;
}

ComponentBuilder::TypeHandler* ComponentBuilder::getHandlerForState (const ValueTree& s) const
{
	const Identifier targetType (s.getType());

	for (int i = 0; i < types.size(); ++i)
	{
		TypeHandler* const t = types.getUnchecked(i);

		if (t->getType() == targetType)
			return t;
	}

	return nullptr;
}

int ComponentBuilder::getNumHandlers() const noexcept
{
	return types.size();
}

ComponentBuilder::TypeHandler* ComponentBuilder::getHandler (const int index) const noexcept
{
	return types [index];
}

void ComponentBuilder::setImageProvider (ImageProvider* newImageProvider) noexcept
{
	imageProvider = newImageProvider;
}

ComponentBuilder::ImageProvider* ComponentBuilder::getImageProvider() const noexcept
{
	return imageProvider;
}

void ComponentBuilder::valueTreePropertyChanged (ValueTree& tree, const Identifier&)
{
	ComponentBuilderHelpers::updateComponent (*this, tree);
}

void ComponentBuilder::valueTreeChildAdded (ValueTree& tree, ValueTree&)
{
	ComponentBuilderHelpers::updateComponent (*this, tree);
}

void ComponentBuilder::valueTreeChildRemoved (ValueTree& tree, ValueTree&)
{
	ComponentBuilderHelpers::updateComponent (*this, tree);
}

void ComponentBuilder::valueTreeChildOrderChanged (ValueTree& tree)
{
	ComponentBuilderHelpers::updateComponent (*this, tree);
}

void ComponentBuilder::valueTreeParentChanged (ValueTree& tree)
{
	ComponentBuilderHelpers::updateComponent (*this, tree);
}

ComponentBuilder::TypeHandler::TypeHandler (const Identifier& valueTreeType_)
   : builder (nullptr),
	 valueTreeType (valueTreeType_)
{
}

ComponentBuilder::TypeHandler::~TypeHandler()
{
}

ComponentBuilder* ComponentBuilder::TypeHandler::getBuilder() const noexcept
{
	// A type handler needs to be registered with a ComponentBuilder before using it!
	jassert (builder != nullptr);
	return builder;
}

void ComponentBuilder::updateChildComponents (Component& parent, const ValueTree& children)
{
	using namespace ComponentBuilderHelpers;

	const int numExistingChildComps = parent.getNumChildComponents();

	Array <Component*> componentsInOrder;
	componentsInOrder.ensureStorageAllocated (numExistingChildComps);

	{
		OwnedArray<Component> existingComponents;
		existingComponents.ensureStorageAllocated (numExistingChildComps);

		int i;
		for (i = 0; i < numExistingChildComps; ++i)
			existingComponents.add (parent.getChildComponent (i));

		const int newNumChildren = children.getNumChildren();
		for (i = 0; i < newNumChildren; ++i)
		{
			const ValueTree childState (children.getChild (i));

			ComponentBuilder::TypeHandler* const type = getHandlerForState (childState);
			jassert (type != nullptr);

			if (type != nullptr)
			{
				Component* c = findComponentWithID (existingComponents, getStateId (childState));

				if (c == nullptr)
					c = createNewComponent (*type, childState, &parent);

				componentsInOrder.add (c);
			}
		}

		// (remaining unused items in existingComponents get deleted here as it goes out of scope)
	}

	// Make sure the z-order is correct..
	if (componentsInOrder.size() > 0)
	{
		componentsInOrder.getLast()->toFront (false);

		for (int i = componentsInOrder.size() - 1; --i >= 0;)
			componentsInOrder.getUnchecked(i)->toBehind (componentsInOrder.getUnchecked (i + 1));
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComponentBuilder.cpp ***/


/*** Start of inlined file: juce_ComponentBoundsConstrainer.cpp ***/
BEGIN_JUCE_NAMESPACE

ComponentBoundsConstrainer::ComponentBoundsConstrainer() noexcept
	: minW (0),
	  maxW (0x3fffffff),
	  minH (0),
	  maxH (0x3fffffff),
	  minOffTop (0),
	  minOffLeft (0),
	  minOffBottom (0),
	  minOffRight (0),
	  aspectRatio (0.0)
{
}

ComponentBoundsConstrainer::~ComponentBoundsConstrainer()
{
}

void ComponentBoundsConstrainer::setMinimumWidth (const int minimumWidth) noexcept
{
	minW = minimumWidth;
}

void ComponentBoundsConstrainer::setMaximumWidth (const int maximumWidth) noexcept
{
	maxW = maximumWidth;
}

void ComponentBoundsConstrainer::setMinimumHeight (const int minimumHeight) noexcept
{
	minH = minimumHeight;
}

void ComponentBoundsConstrainer::setMaximumHeight (const int maximumHeight) noexcept
{
	maxH = maximumHeight;
}

void ComponentBoundsConstrainer::setMinimumSize (const int minimumWidth, const int minimumHeight) noexcept
{
	jassert (maxW >= minimumWidth);
	jassert (maxH >= minimumHeight);
	jassert (minimumWidth > 0 && minimumHeight > 0);

	minW = minimumWidth;
	minH = minimumHeight;

	if (minW > maxW)
		maxW = minW;

	if (minH > maxH)
		maxH = minH;
}

void ComponentBoundsConstrainer::setMaximumSize (const int maximumWidth, const int maximumHeight) noexcept
{
	jassert (maximumWidth >= minW);
	jassert (maximumHeight >= minH);
	jassert (maximumWidth > 0 && maximumHeight > 0);

	maxW = jmax (minW, maximumWidth);
	maxH = jmax (minH, maximumHeight);
}

void ComponentBoundsConstrainer::setSizeLimits (const int minimumWidth,
												const int minimumHeight,
												const int maximumWidth,
												const int maximumHeight) noexcept
{
	jassert (maximumWidth >= minimumWidth);
	jassert (maximumHeight >= minimumHeight);
	jassert (maximumWidth > 0 && maximumHeight > 0);
	jassert (minimumWidth > 0 && minimumHeight > 0);

	minW = jmax (0, minimumWidth);
	minH = jmax (0, minimumHeight);
	maxW = jmax (minW, maximumWidth);
	maxH = jmax (minH, maximumHeight);
}

void ComponentBoundsConstrainer::setMinimumOnscreenAmounts (const int minimumWhenOffTheTop,
															const int minimumWhenOffTheLeft,
															const int minimumWhenOffTheBottom,
															const int minimumWhenOffTheRight) noexcept
{
	minOffTop = minimumWhenOffTheTop;
	minOffLeft = minimumWhenOffTheLeft;
	minOffBottom = minimumWhenOffTheBottom;
	minOffRight = minimumWhenOffTheRight;
}

void ComponentBoundsConstrainer::setFixedAspectRatio (const double widthOverHeight) noexcept
{
	aspectRatio = jmax (0.0, widthOverHeight);
}

double ComponentBoundsConstrainer::getFixedAspectRatio() const noexcept
{
	return aspectRatio;
}

void ComponentBoundsConstrainer::setBoundsForComponent (Component* const component,
														const Rectangle<int>& targetBounds,
														const bool isStretchingTop,
														const bool isStretchingLeft,
														const bool isStretchingBottom,
														const bool isStretchingRight)
{
	jassert (component != nullptr);

	Rectangle<int> limits, bounds (targetBounds);
	BorderSize<int> border;

	Component* const parent = component->getParentComponent();

	if (parent == nullptr)
	{
		ComponentPeer* peer = component->getPeer();
		if (peer != nullptr)
			border = peer->getFrameSize();

		limits = Desktop::getInstance().getMonitorAreaContaining (bounds.getCentre());
	}
	else
	{
		limits.setSize (parent->getWidth(), parent->getHeight());
	}

	border.addTo (bounds);

	checkBounds (bounds,
				 border.addedTo (component->getBounds()), limits,
				 isStretchingTop, isStretchingLeft,
				 isStretchingBottom, isStretchingRight);

	border.subtractFrom (bounds);

	applyBoundsToComponent (component, bounds);
}

void ComponentBoundsConstrainer::checkComponentBounds (Component* component)
{
	setBoundsForComponent (component, component->getBounds(),
						   false, false, false, false);
}

void ComponentBoundsConstrainer::applyBoundsToComponent (Component* component,
														 const Rectangle<int>& bounds)
{
	Component::Positioner* const positioner = component->getPositioner();

	if (positioner != nullptr)
		positioner->applyNewBounds (bounds);
	else
		component->setBounds (bounds);
}

void ComponentBoundsConstrainer::resizeStart()
{
}

void ComponentBoundsConstrainer::resizeEnd()
{
}

void ComponentBoundsConstrainer::checkBounds (Rectangle<int>& bounds,
											  const Rectangle<int>& old,
											  const Rectangle<int>& limits,
											  const bool isStretchingTop,
											  const bool isStretchingLeft,
											  const bool isStretchingBottom,
											  const bool isStretchingRight)
{
	// constrain the size if it's being stretched..
	if (isStretchingLeft)
		bounds.setLeft (jlimit (old.getRight() - maxW, old.getRight() - minW, bounds.getX()));

	if (isStretchingRight)
		bounds.setWidth (jlimit (minW, maxW, bounds.getWidth()));

	if (isStretchingTop)
		bounds.setTop (jlimit (old.getBottom() - maxH, old.getBottom() - minH, bounds.getY()));

	if (isStretchingBottom)
		bounds.setHeight (jlimit (minH, maxH, bounds.getHeight()));

	if (bounds.isEmpty())
		return;

	if (minOffTop > 0)
	{
		const int limit = limits.getY() + jmin (minOffTop - bounds.getHeight(), 0);

		if (bounds.getY() < limit)
		{
			if (isStretchingTop)
				bounds.setTop (limits.getY());
			else
				bounds.setY (limit);
		}
	}

	if (minOffLeft > 0)
	{
		const int limit = limits.getX() + jmin (minOffLeft - bounds.getWidth(), 0);

		if (bounds.getX() < limit)
		{
			if (isStretchingLeft)
				bounds.setLeft (limits.getX());
			else
				bounds.setX (limit);
		}
	}

	if (minOffBottom > 0)
	{
		const int limit = limits.getBottom() - jmin (minOffBottom, bounds.getHeight());

		if (bounds.getY() > limit)
		{
			if (isStretchingBottom)
				bounds.setBottom (limits.getBottom());
			else
				bounds.setY (limit);
		}
	}

	if (minOffRight > 0)
	{
		const int limit = limits.getRight() - jmin (minOffRight, bounds.getWidth());

		if (bounds.getX() > limit)
		{
			if (isStretchingRight)
				bounds.setRight (limits.getRight());
			else
				bounds.setX (limit);
		}
	}

	// constrain the aspect ratio if one has been specified..
	if (aspectRatio > 0.0)
	{
		bool adjustWidth;

		if ((isStretchingTop || isStretchingBottom) && ! (isStretchingLeft || isStretchingRight))
		{
			adjustWidth = true;
		}
		else if ((isStretchingLeft || isStretchingRight) && ! (isStretchingTop || isStretchingBottom))
		{
			adjustWidth = false;
		}
		else
		{
			const double oldRatio = (old.getHeight() > 0) ? std::abs (old.getWidth() / (double) old.getHeight()) : 0.0;
			const double newRatio = std::abs (bounds.getWidth() / (double) bounds.getHeight());

			adjustWidth = (oldRatio > newRatio);
		}

		if (adjustWidth)
		{
			bounds.setWidth (roundToInt (bounds.getHeight() * aspectRatio));

			if (bounds.getWidth() > maxW || bounds.getWidth() < minW)
			{
				bounds.setWidth (jlimit (minW, maxW, bounds.getWidth()));
				bounds.setHeight (roundToInt (bounds.getWidth() / aspectRatio));
			}
		}
		else
		{
			bounds.setHeight (roundToInt (bounds.getWidth() / aspectRatio));

			if (bounds.getHeight() > maxH || bounds.getHeight() < minH)
			{
				bounds.setHeight (jlimit (minH, maxH, bounds.getHeight()));
				bounds.setWidth (roundToInt (bounds.getHeight() * aspectRatio));
			}
		}

		if ((isStretchingTop || isStretchingBottom) && ! (isStretchingLeft || isStretchingRight))
		{
			bounds.setX (old.getX() + (old.getWidth() - bounds.getWidth()) / 2);
		}
		else if ((isStretchingLeft || isStretchingRight) && ! (isStretchingTop || isStretchingBottom))
		{
			bounds.setY (old.getY() + (old.getHeight() - bounds.getHeight()) / 2);
		}
		else
		{
			if (isStretchingLeft)
				bounds.setX (old.getRight() - bounds.getWidth());

			if (isStretchingTop)
				bounds.setY (old.getBottom() - bounds.getHeight());
		}
	}

	jassert (! bounds.isEmpty());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComponentBoundsConstrainer.cpp ***/


/*** Start of inlined file: juce_ComponentMovementWatcher.cpp ***/
BEGIN_JUCE_NAMESPACE

ComponentMovementWatcher::ComponentMovementWatcher (Component* const component_)
	: component (component_),
	  lastPeerID (0),
	  reentrant (false),
	  wasShowing (component_->isShowing())
{
	jassert (component != nullptr); // can't use this with a null pointer..

	component->addComponentListener (this);

	registerWithParentComps();
}

ComponentMovementWatcher::~ComponentMovementWatcher()
{
	if (component != nullptr)
		component->removeComponentListener (this);

	unregister();
}

void ComponentMovementWatcher::componentParentHierarchyChanged (Component&)
{
	if (component != nullptr && ! reentrant)
	{
		const ScopedValueSetter<bool> setter (reentrant, true);

		ComponentPeer* const peer = component->getPeer();
		const uint32 peerID = peer != nullptr ? peer->getUniqueID() : 0;

		if (peerID != lastPeerID)
		{
			componentPeerChanged();

			if (component == nullptr)
				return;

			lastPeerID = peerID;
		}

		unregister();
		registerWithParentComps();

		componentMovedOrResized (*component, true, true);

		if (component != nullptr)
			componentVisibilityChanged (*component);
	}
}

void ComponentMovementWatcher::componentMovedOrResized (Component&, bool wasMoved, bool wasResized)
{
	if (component != nullptr)
	{
		if (wasMoved)
		{
			const Point<int> pos (component->getTopLevelComponent()->getLocalPoint (component, Point<int>()));

			wasMoved = lastBounds.getPosition() != pos;
			lastBounds.setPosition (pos);
		}

		wasResized = (lastBounds.getWidth() != component->getWidth() || lastBounds.getHeight() != component->getHeight());
		lastBounds.setSize (component->getWidth(), component->getHeight());

		if (wasMoved || wasResized)
			componentMovedOrResized (wasMoved, wasResized);
	}
}

void ComponentMovementWatcher::componentBeingDeleted (Component& comp)
{
	registeredParentComps.removeValue (&comp);

	if (component == &comp)
		unregister();
}

void ComponentMovementWatcher::componentVisibilityChanged (Component&)
{
	if (component != nullptr)
	{
		const bool isShowingNow = component->isShowing();

		if (wasShowing != isShowingNow)
		{
			wasShowing = isShowingNow;
			componentVisibilityChanged();
		}
	}
}

void ComponentMovementWatcher::registerWithParentComps()
{
	Component* p = component->getParentComponent();

	while (p != nullptr)
	{
		p->addComponentListener (this);
		registeredParentComps.add (p);
		p = p->getParentComponent();
	}
}

void ComponentMovementWatcher::unregister()
{
	for (int i = registeredParentComps.size(); --i >= 0;)
		registeredParentComps.getUnchecked(i)->removeComponentListener (this);

	registeredParentComps.clear();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComponentMovementWatcher.cpp ***/


/*** Start of inlined file: juce_GroupComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

GroupComponent::GroupComponent (const String& name,
								const String& labelText)
	: Component (name),
	  text (labelText),
	  justification (Justification::left)
{
	setInterceptsMouseClicks  (false, true);
}

GroupComponent::~GroupComponent()
{
}

void GroupComponent::setText (const String& newText)
{
	if (text != newText)
	{
		text = newText;
		repaint();
	}
}

String GroupComponent::getText() const
{
	return text;
}

void GroupComponent::setTextLabelPosition (const Justification& newJustification)
{
	if (justification != newJustification)
	{
		justification = newJustification;
		repaint();
	}
}

void GroupComponent::paint (Graphics& g)
{
	getLookAndFeel()
		.drawGroupComponentOutline (g, getWidth(), getHeight(),
									text, justification,
									*this);
}

void GroupComponent::enablementChanged()
{
	repaint();
}

void GroupComponent::colourChanged()
{
	repaint();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_GroupComponent.cpp ***/


/*** Start of inlined file: juce_MultiDocumentPanel.cpp ***/
BEGIN_JUCE_NAMESPACE

MultiDocumentPanelWindow::MultiDocumentPanelWindow (const Colour& backgroundColour)
	: DocumentWindow (String::empty, backgroundColour,
					  DocumentWindow::maximiseButton | DocumentWindow::closeButton, false)
{
}

MultiDocumentPanelWindow::~MultiDocumentPanelWindow()
{
}

void MultiDocumentPanelWindow::maximiseButtonPressed()
{
	MultiDocumentPanel* const owner = getOwner();

	jassert (owner != nullptr); // these windows are only designed to be used inside a MultiDocumentPanel!
	if (owner != nullptr)
		owner->setLayoutMode (MultiDocumentPanel::MaximisedWindowsWithTabs);
}

void MultiDocumentPanelWindow::closeButtonPressed()
{
	MultiDocumentPanel* const owner = getOwner();

	jassert (owner != nullptr); // these windows are only designed to be used inside a MultiDocumentPanel!
	if (owner != nullptr)
		owner->closeDocument (getContentComponent(), true);
}

void MultiDocumentPanelWindow::activeWindowStatusChanged()
{
	DocumentWindow::activeWindowStatusChanged();
	updateOrder();
}

void MultiDocumentPanelWindow::broughtToFront()
{
	DocumentWindow::broughtToFront();
	updateOrder();
}

void MultiDocumentPanelWindow::updateOrder()
{
	MultiDocumentPanel* const owner = getOwner();

	if (owner != nullptr)
		owner->updateOrder();
}

MultiDocumentPanel* MultiDocumentPanelWindow::getOwner() const noexcept
{
	// (unable to use the syntax findParentComponentOfClass <MultiDocumentPanel> () because of a VC6 compiler bug)
	return findParentComponentOfClass ((MultiDocumentPanel*) nullptr);
}

class MDITabbedComponentInternal   : public TabbedComponent
{
public:
	MDITabbedComponentInternal()
		: TabbedComponent (TabbedButtonBar::TabsAtTop)
	{
	}

	~MDITabbedComponentInternal()
	{
	}

	void currentTabChanged (int, const String&)
	{
		// (unable to use the syntax findParentComponentOfClass <MultiDocumentPanel> () because of a VC6 compiler bug)
		MultiDocumentPanel* const owner = findParentComponentOfClass ((MultiDocumentPanel*) nullptr);

		if (owner != nullptr)
			owner->updateOrder();
	}
};

MultiDocumentPanel::MultiDocumentPanel()
	: mode (MaximisedWindowsWithTabs),
	  backgroundColour (Colours::lightblue),
	  maximumNumDocuments (0),
	  numDocsBeforeTabsUsed (0)
{
	setOpaque (true);
}

MultiDocumentPanel::~MultiDocumentPanel()
{
	closeAllDocuments (false);
}

namespace MultiDocHelpers
{
	bool shouldDeleteComp (Component* const c)
	{
		return c->getProperties() ["mdiDocumentDelete_"];
	}
}

bool MultiDocumentPanel::closeAllDocuments (const bool checkItsOkToCloseFirst)
{
	while (components.size() > 0)
		if (! closeDocument (components.getLast(), checkItsOkToCloseFirst))
			return false;

	return true;
}

MultiDocumentPanelWindow* MultiDocumentPanel::createNewDocumentWindow()
{
	return new MultiDocumentPanelWindow (backgroundColour);
}

void MultiDocumentPanel::addWindow (Component* component)
{
	MultiDocumentPanelWindow* const dw = createNewDocumentWindow();

	dw->setResizable (true, false);
	dw->setContentNonOwned (component, true);
	dw->setName (component->getName());

	const var bkg (component->getProperties() ["mdiDocumentBkg_"]);
	dw->setBackgroundColour (bkg.isVoid() ? backgroundColour : Colour ((int) bkg));

	int x = 4;
	Component* const topComp = getChildComponent (getNumChildComponents() - 1);

	if (topComp != nullptr && topComp->getX() == x && topComp->getY() == x)
		x += 16;

	dw->setTopLeftPosition (x, x);

	const var pos (component->getProperties() ["mdiDocumentPos_"]);
	if (pos.toString().isNotEmpty())
		dw->restoreWindowStateFromString (pos.toString());

	addAndMakeVisible (dw);
	dw->toFront (true);
}

bool MultiDocumentPanel::addDocument (Component* const component,
									  const Colour& docColour,
									  const bool deleteWhenRemoved)
{
	// If you try passing a full DocumentWindow or ResizableWindow in here, you'll end up
	// with a frame-within-a-frame! Just pass in the bare content component.
	jassert (dynamic_cast <ResizableWindow*> (component) == nullptr);

	if (component == nullptr || (maximumNumDocuments > 0 && components.size() >= maximumNumDocuments))
		return false;

	components.add (component);
	component->getProperties().set ("mdiDocumentDelete_", deleteWhenRemoved);
	component->getProperties().set ("mdiDocumentBkg_", (int) docColour.getARGB());
	component->addComponentListener (this);

	if (mode == FloatingWindows)
	{
		if (isFullscreenWhenOneDocument())
		{
			if (components.size() == 1)
			{
				addAndMakeVisible (component);
			}
			else
			{
				if (components.size() == 2)
					addWindow (components.getFirst());

				addWindow (component);
			}
		}
		else
		{
		   addWindow (component);
		}
	}
	else
	{
		if (tabComponent == nullptr && components.size() > numDocsBeforeTabsUsed)
		{
			addAndMakeVisible (tabComponent = new MDITabbedComponentInternal());

			Array <Component*> temp (components);

			for (int i = 0; i < temp.size(); ++i)
				tabComponent->addTab (temp[i]->getName(), docColour, temp[i], false);

			resized();
		}
		else
		{
			if (tabComponent != nullptr)
				tabComponent->addTab (component->getName(), docColour, component, false);
			else
				addAndMakeVisible (component);
		}

		setActiveDocument (component);
	}

	resized();
	activeDocumentChanged();
	return true;
}

bool MultiDocumentPanel::closeDocument (Component* component,
										const bool checkItsOkToCloseFirst)
{
	if (components.contains (component))
	{
		if (checkItsOkToCloseFirst && ! tryToCloseDocument (component))
			return false;

		component->removeComponentListener (this);

		const bool shouldDelete = MultiDocHelpers::shouldDeleteComp (component);
		component->getProperties().remove ("mdiDocumentDelete_");
		component->getProperties().remove ("mdiDocumentBkg_");

		if (mode == FloatingWindows)
		{
			for (int i = getNumChildComponents(); --i >= 0;)
			{
				MultiDocumentPanelWindow* const dw = dynamic_cast <MultiDocumentPanelWindow*> (getChildComponent (i));

				if (dw != nullptr && dw->getContentComponent() == component)
				{
					ScopedPointer<MultiDocumentPanelWindow> (dw)->clearContentComponent();
					break;
				}
			}

			if (shouldDelete)
				delete component;

			components.removeValue (component);

			if (isFullscreenWhenOneDocument() && components.size() == 1)
			{
				for (int i = getNumChildComponents(); --i >= 0;)
				{
					ScopedPointer<MultiDocumentPanelWindow> dw (dynamic_cast <MultiDocumentPanelWindow*> (getChildComponent (i)));

					if (dw != nullptr)
						dw->clearContentComponent();
				}

				addAndMakeVisible (components.getFirst());
			}
		}
		else
		{
			jassert (components.indexOf (component) >= 0);

			if (tabComponent != nullptr)
			{
				for (int i = tabComponent->getNumTabs(); --i >= 0;)
					if (tabComponent->getTabContentComponent (i) == component)
						tabComponent->removeTab (i);
			}
			else
			{
				removeChildComponent (component);
			}

			if (shouldDelete)
				delete component;

			if (tabComponent != nullptr && tabComponent->getNumTabs() <= numDocsBeforeTabsUsed)
				tabComponent = nullptr;

			components.removeValue (component);

			if (components.size() > 0 && tabComponent == nullptr)
				addAndMakeVisible (components.getFirst());
		}

		resized();
		activeDocumentChanged();
	}
	else
	{
		jassertfalse;
	}

	return true;
}

int MultiDocumentPanel::getNumDocuments() const noexcept
{
	return components.size();
}

Component* MultiDocumentPanel::getDocument (const int index) const noexcept
{
	return components [index];
}

Component* MultiDocumentPanel::getActiveDocument() const noexcept
{
	if (mode == FloatingWindows)
	{
		for (int i = getNumChildComponents(); --i >= 0;)
		{
			MultiDocumentPanelWindow* const dw = dynamic_cast <MultiDocumentPanelWindow*> (getChildComponent (i));

			if (dw != nullptr && dw->isActiveWindow())
				return dw->getContentComponent();
		}
	}

	return components.getLast();
}

void MultiDocumentPanel::setActiveDocument (Component* component)
{
	if (mode == FloatingWindows)
	{
		component = getContainerComp (component);

		if (component != nullptr)
			component->toFront (true);
	}
	else if (tabComponent != nullptr)
	{
		jassert (components.indexOf (component) >= 0);

		for (int i = tabComponent->getNumTabs(); --i >= 0;)
		{
			if (tabComponent->getTabContentComponent (i) == component)
			{
				tabComponent->setCurrentTabIndex (i);
				break;
			}
		}
	}
	else
	{
		component->grabKeyboardFocus();
	}
}

void MultiDocumentPanel::activeDocumentChanged()
{
}

void MultiDocumentPanel::setMaximumNumDocuments (const int newNumber)
{
	maximumNumDocuments = newNumber;
}

void MultiDocumentPanel::useFullscreenWhenOneDocument (const bool shouldUseTabs)
{
	numDocsBeforeTabsUsed = shouldUseTabs ? 1 : 0;
}

bool MultiDocumentPanel::isFullscreenWhenOneDocument() const noexcept
{
	return numDocsBeforeTabsUsed != 0;
}

void MultiDocumentPanel::setLayoutMode (const LayoutMode newLayoutMode)
{
	if (mode != newLayoutMode)
	{
		mode = newLayoutMode;

		if (mode == FloatingWindows)
		{
			tabComponent = nullptr;
		}
		else
		{
			for (int i = getNumChildComponents(); --i >= 0;)
			{
				MultiDocumentPanelWindow* const dw = dynamic_cast <MultiDocumentPanelWindow*> (getChildComponent (i));

				if (dw != nullptr)
				{
					dw->getContentComponent()->getProperties().set ("mdiDocumentPos_", dw->getWindowStateAsString());
					dw->clearContentComponent();
					delete dw;
				}
			}
		}

		resized();

		const Array <Component*> tempComps (components);
		components.clear();

		for (int i = 0; i < tempComps.size(); ++i)
		{
			Component* const c = tempComps.getUnchecked(i);

			addDocument (c,
						 Colour ((int) c->getProperties().getWithDefault ("mdiDocumentBkg_", (int) Colours::white.getARGB())),
						 MultiDocHelpers::shouldDeleteComp (c));
		}
	}
}

void MultiDocumentPanel::setBackgroundColour (const Colour& newBackgroundColour)
{
	if (backgroundColour != newBackgroundColour)
	{
		backgroundColour = newBackgroundColour;
		setOpaque (newBackgroundColour.isOpaque());
		repaint();
	}
}

void MultiDocumentPanel::paint (Graphics& g)
{
	g.fillAll (backgroundColour);
}

void MultiDocumentPanel::resized()
{
	if (mode == MaximisedWindowsWithTabs || components.size() == numDocsBeforeTabsUsed)
	{
		for (int i = getNumChildComponents(); --i >= 0;)
			getChildComponent (i)->setBounds (getLocalBounds());
	}

	setWantsKeyboardFocus (components.size() == 0);
}

Component* MultiDocumentPanel::getContainerComp (Component* c) const
{
	if (mode == FloatingWindows)
	{
		for (int i = 0; i < getNumChildComponents(); ++i)
		{
			MultiDocumentPanelWindow* const dw = dynamic_cast <MultiDocumentPanelWindow*> (getChildComponent (i));

			if (dw != nullptr && dw->getContentComponent() == c)
			{
				c = dw;
				break;
			}
		}
	}

	return c;
}

void MultiDocumentPanel::componentNameChanged (Component&)
{
	if (mode == FloatingWindows)
	{
		for (int i = 0; i < getNumChildComponents(); ++i)
		{
			MultiDocumentPanelWindow* const dw = dynamic_cast <MultiDocumentPanelWindow*> (getChildComponent (i));

			if (dw != nullptr)
				dw->setName (dw->getContentComponent()->getName());
		}
	}
	else if (tabComponent != nullptr)
	{
		for (int i = tabComponent->getNumTabs(); --i >= 0;)
			tabComponent->setTabName (i, tabComponent->getTabContentComponent (i)->getName());
	}
}

void MultiDocumentPanel::updateOrder()
{
	const Array <Component*> oldList (components);

	if (mode == FloatingWindows)
	{
		components.clear();

		for (int i = 0; i < getNumChildComponents(); ++i)
		{
			MultiDocumentPanelWindow* const dw = dynamic_cast <MultiDocumentPanelWindow*> (getChildComponent (i));

			if (dw != nullptr)
				components.add (dw->getContentComponent());
		}
	}
	else
	{
		if (tabComponent != nullptr)
		{
			Component* const current = tabComponent->getCurrentContentComponent();

			if (current != nullptr)
			{
				components.removeValue (current);
				components.add (current);
			}
		}
	}

	if (components != oldList)
		activeDocumentChanged();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MultiDocumentPanel.cpp ***/


/*** Start of inlined file: juce_ResizableBorderComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

ResizableBorderComponent::Zone::Zone (const int zoneFlags) noexcept
	: zone (zoneFlags)
{}

ResizableBorderComponent::Zone::Zone (const ResizableBorderComponent::Zone& other) noexcept
	: zone (other.zone)
{}

ResizableBorderComponent::Zone& ResizableBorderComponent::Zone::operator= (const ResizableBorderComponent::Zone& other) noexcept
{
	zone = other.zone;
	return *this;
}

bool ResizableBorderComponent::Zone::operator== (const ResizableBorderComponent::Zone& other) const noexcept	  { return zone == other.zone; }
bool ResizableBorderComponent::Zone::operator!= (const ResizableBorderComponent::Zone& other) const noexcept	  { return zone != other.zone; }

const ResizableBorderComponent::Zone ResizableBorderComponent::Zone::fromPositionOnBorder (const Rectangle<int>& totalSize,
																						   const BorderSize<int>& border,
																						   const Point<int>& position)
{
	int z = 0;

	if (totalSize.contains (position)
		 && ! border.subtractedFrom (totalSize).contains (position))
	{
		const int minW = jmax (totalSize.getWidth() / 10, jmin (10, totalSize.getWidth() / 3));
		if (position.getX() < jmax (border.getLeft(), minW) && border.getLeft() > 0)
			z |= left;
		else if (position.getX() >= totalSize.getWidth() - jmax (border.getRight(), minW) && border.getRight() > 0)
			z |= right;

		const int minH = jmax (totalSize.getHeight() / 10, jmin (10, totalSize.getHeight() / 3));
		if (position.getY() < jmax (border.getTop(), minH) && border.getTop() > 0)
			z |= top;
		else if (position.getY() >= totalSize.getHeight() - jmax (border.getBottom(), minH) && border.getBottom() > 0)
			z |= bottom;
	}

	return Zone (z);
}

const MouseCursor ResizableBorderComponent::Zone::getMouseCursor() const noexcept
{
	MouseCursor::StandardCursorType mc = MouseCursor::NormalCursor;

	switch (zone)
	{
		case (left | top):	  mc = MouseCursor::TopLeftCornerResizeCursor; break;
		case top:		   mc = MouseCursor::TopEdgeResizeCursor; break;
		case (right | top):	 mc = MouseCursor::TopRightCornerResizeCursor; break;
		case left:		  mc = MouseCursor::LeftEdgeResizeCursor; break;
		case right:		 mc = MouseCursor::RightEdgeResizeCursor; break;
		case (left | bottom):   mc = MouseCursor::BottomLeftCornerResizeCursor; break;
		case bottom:		mc = MouseCursor::BottomEdgeResizeCursor; break;
		case (right | bottom):  mc = MouseCursor::BottomRightCornerResizeCursor; break;
		default:		break;
	}

	return mc;
}

ResizableBorderComponent::ResizableBorderComponent (Component* const componentToResize,
													ComponentBoundsConstrainer* const constrainer_)
   : component (componentToResize),
	 constrainer (constrainer_),
	 borderSize (5),
	 mouseZone (0)
{
}

ResizableBorderComponent::~ResizableBorderComponent()
{
}

void ResizableBorderComponent::paint (Graphics& g)
{
	getLookAndFeel().drawResizableFrame (g, getWidth(), getHeight(), borderSize);
}

void ResizableBorderComponent::mouseEnter (const MouseEvent& e)
{
	updateMouseZone (e);
}

void ResizableBorderComponent::mouseMove (const MouseEvent& e)
{
	updateMouseZone (e);
}

void ResizableBorderComponent::mouseDown (const MouseEvent& e)
{
	if (component == nullptr)
	{
		jassertfalse; // You've deleted the component that this resizer was supposed to be using!
		return;
	}

	updateMouseZone (e);

	originalBounds = component->getBounds();

	if (constrainer != nullptr)
		constrainer->resizeStart();
}

void ResizableBorderComponent::mouseDrag (const MouseEvent& e)
{
	if (component == nullptr)
	{
		jassertfalse; // You've deleted the component that this resizer was supposed to be using!
		return;
	}

	const Rectangle<int> newBounds (mouseZone.resizeRectangleBy (originalBounds, e.getOffsetFromDragStart()));

	if (constrainer != nullptr)
	{
		constrainer->setBoundsForComponent (component, newBounds,
											mouseZone.isDraggingTopEdge(),
											mouseZone.isDraggingLeftEdge(),
											mouseZone.isDraggingBottomEdge(),
											mouseZone.isDraggingRightEdge());
	}
	else
	{
		Component::Positioner* const pos = component->getPositioner();

		if (pos != nullptr)
			pos->applyNewBounds (newBounds);
		else
			component->setBounds (newBounds);
	}
}

void ResizableBorderComponent::mouseUp (const MouseEvent&)
{
	if (constrainer != nullptr)
		constrainer->resizeEnd();
}

bool ResizableBorderComponent::hitTest (int x, int y)
{
	return x < borderSize.getLeft()
			|| x >= getWidth() - borderSize.getRight()
			|| y < borderSize.getTop()
			|| y >= getHeight() - borderSize.getBottom();
}

void ResizableBorderComponent::setBorderThickness (const BorderSize<int>& newBorderSize)
{
	if (borderSize != newBorderSize)
	{
		borderSize = newBorderSize;
		repaint();
	}
}

const BorderSize<int> ResizableBorderComponent::getBorderThickness() const
{
	return borderSize;
}

void ResizableBorderComponent::updateMouseZone (const MouseEvent& e)
{
	Zone newZone (Zone::fromPositionOnBorder (getLocalBounds(), borderSize, e.getPosition()));

	if (mouseZone != newZone)
	{
		mouseZone = newZone;
		setMouseCursor (newZone.getMouseCursor());
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ResizableBorderComponent.cpp ***/


/*** Start of inlined file: juce_ResizableCornerComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

ResizableCornerComponent::ResizableCornerComponent (Component* const componentToResize,
													ComponentBoundsConstrainer* const constrainer_)
   : component (componentToResize),
	 constrainer (constrainer_)
{
	setRepaintsOnMouseActivity (true);
	setMouseCursor (MouseCursor::BottomRightCornerResizeCursor);
}

ResizableCornerComponent::~ResizableCornerComponent()
{
}

void ResizableCornerComponent::paint (Graphics& g)
{
	getLookAndFeel()
		.drawCornerResizer (g, getWidth(), getHeight(),
							isMouseOverOrDragging(),
							isMouseButtonDown());
}

void ResizableCornerComponent::mouseDown (const MouseEvent&)
{
	if (component == nullptr)
	{
		jassertfalse; // You've deleted the component that this resizer is supposed to be controlling!
		return;
	}

	originalBounds = component->getBounds();

	if (constrainer != nullptr)
		constrainer->resizeStart();
}

void ResizableCornerComponent::mouseDrag (const MouseEvent& e)
{
	if (component == nullptr)
	{
		jassertfalse; // You've deleted the component that this resizer is supposed to be controlling!
		return;
	}

	Rectangle<int> r (originalBounds.withSize (originalBounds.getWidth() + e.getDistanceFromDragStartX(),
											   originalBounds.getHeight() + e.getDistanceFromDragStartY()));

	if (constrainer != nullptr)
	{
		constrainer->setBoundsForComponent (component, r, false, false, true, true);
	}
	else
	{
		Component::Positioner* const pos = component->getPositioner();

		if (pos != nullptr)
			pos->applyNewBounds (r);
		else
			component->setBounds (r);
	}
}

void ResizableCornerComponent::mouseUp (const MouseEvent&)
{
	if (constrainer != nullptr)
		constrainer->resizeEnd();
}

bool ResizableCornerComponent::hitTest (int x, int y)
{
	if (getWidth() <= 0)
		return false;

	const int yAtX = getHeight() - (getHeight() * x / getWidth());

	return y >= yAtX - getHeight() / 4;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ResizableCornerComponent.cpp ***/


/*** Start of inlined file: juce_ResizableEdgeComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

ResizableEdgeComponent::ResizableEdgeComponent (Component* const componentToResize,
												ComponentBoundsConstrainer* const constrainer_,
												Edge edge_)
   : component (componentToResize),
	 constrainer (constrainer_),
	 edge (edge_)
{
	setRepaintsOnMouseActivity (true);
	setMouseCursor (MouseCursor (isVertical() ? MouseCursor::LeftRightResizeCursor
											  : MouseCursor::UpDownResizeCursor));
}

ResizableEdgeComponent::~ResizableEdgeComponent()
{
}

bool ResizableEdgeComponent::isVertical() const noexcept
{
	return edge == leftEdge || edge == rightEdge;
}

void ResizableEdgeComponent::paint (Graphics& g)
{
	getLookAndFeel().drawStretchableLayoutResizerBar (g, getWidth(), getHeight(), isVertical(),
													  isMouseOver(), isMouseButtonDown());
}

void ResizableEdgeComponent::mouseDown (const MouseEvent&)
{
	if (component == nullptr)
	{
		jassertfalse; // You've deleted the component that this resizer was supposed to be using!
		return;
	}

	originalBounds = component->getBounds();

	if (constrainer != nullptr)
		constrainer->resizeStart();
}

void ResizableEdgeComponent::mouseDrag (const MouseEvent& e)
{
	if (component == nullptr)
	{
		jassertfalse; // You've deleted the component that this resizer was supposed to be using!
		return;
	}

	Rectangle<int> newBounds (originalBounds);

	switch (edge)
	{
		case leftEdge:	  newBounds.setLeft (jmin (newBounds.getRight(), newBounds.getX() + e.getDistanceFromDragStartX())); break;
		case rightEdge:	 newBounds.setWidth (jmax (0, newBounds.getWidth() + e.getDistanceFromDragStartX())); break;
		case topEdge:	   newBounds.setTop (jmin (newBounds.getBottom(), newBounds.getY() + e.getDistanceFromDragStartY())); break;
		case bottomEdge:	newBounds.setHeight (jmax (0, newBounds.getHeight() + e.getDistanceFromDragStartY())); break;
		default:		jassertfalse; break;
	}

	if (constrainer != nullptr)
	{
		constrainer->setBoundsForComponent (component, newBounds,
											edge == topEdge,
											edge == leftEdge,
											edge == bottomEdge,
											edge == rightEdge);
	}
	else
	{
		Component::Positioner* const pos = component->getPositioner();

		if (pos != nullptr)
			pos->applyNewBounds (newBounds);
		else
			component->setBounds (newBounds);
	}
}

void ResizableEdgeComponent::mouseUp (const MouseEvent&)
{
	if (constrainer != nullptr)
		constrainer->resizeEnd();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ResizableEdgeComponent.cpp ***/


/*** Start of inlined file: juce_ScrollBar.cpp ***/
BEGIN_JUCE_NAMESPACE

class ScrollBar::ScrollbarButton  : public Button
{
public:
	ScrollbarButton (const int direction_, ScrollBar& owner_)
		: Button (String::empty),
		  direction (direction_),
		  owner (owner_)
	{
		setWantsKeyboardFocus (false);
	}

	void paintButton (Graphics& g, bool over, bool down)
	{
		getLookAndFeel()
			.drawScrollbarButton (g, owner,
								  getWidth(), getHeight(),
								  direction,
								  owner.isVertical(),
								  over, down);
	}

	void clicked()
	{
		owner.moveScrollbarInSteps ((direction == 1 || direction == 2) ? 1 : -1);
	}

	int direction;

private:
	ScrollBar& owner;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ScrollbarButton);
};

ScrollBar::ScrollBar (const bool vertical_,
					  const bool buttonsAreVisible)
	: totalRange (0.0, 1.0),
	  visibleRange (0.0, 0.1),
	  singleStepSize (0.1),
	  thumbAreaStart (0),
	  thumbAreaSize (0),
	  thumbStart (0),
	  thumbSize (0),
	  initialDelayInMillisecs (100),
	  repeatDelayInMillisecs (50),
	  minimumDelayInMillisecs (10),
	  vertical (vertical_),
	  isDraggingThumb (false),
	  autohides (true)
{
	setButtonVisibility (buttonsAreVisible);

	setRepaintsOnMouseActivity (true);
	setFocusContainer (true);
}

ScrollBar::~ScrollBar()
{
	upButton = nullptr;
	downButton = nullptr;
}

void ScrollBar::setRangeLimits (const Range<double>& newRangeLimit)
{
	if (totalRange != newRangeLimit)
	{
		totalRange = newRangeLimit;
		setCurrentRange (visibleRange);
		updateThumbPosition();
	}
}

void ScrollBar::setRangeLimits (const double newMinimum, const double newMaximum)
{
	jassert (newMaximum >= newMinimum); // these can't be the wrong way round!
	setRangeLimits (Range<double> (newMinimum, newMaximum));
}

void ScrollBar::setCurrentRange (const Range<double>& newRange)
{
	const Range<double> constrainedRange (totalRange.constrainRange (newRange));

	if (visibleRange != constrainedRange)
	{
		visibleRange = constrainedRange;

		updateThumbPosition();
		triggerAsyncUpdate();
	}
}

void ScrollBar::setCurrentRange (const double newStart, const double newSize)
{
	setCurrentRange (Range<double> (newStart, newStart + newSize));
}

void ScrollBar::setCurrentRangeStart (const double newStart)
{
	setCurrentRange (visibleRange.movedToStartAt (newStart));
}

void ScrollBar::setSingleStepSize (const double newSingleStepSize)
{
	singleStepSize = newSingleStepSize;
}

void ScrollBar::moveScrollbarInSteps (const int howManySteps)
{
	setCurrentRange (visibleRange + howManySteps * singleStepSize);
}

void ScrollBar::moveScrollbarInPages (const int howManyPages)
{
	setCurrentRange (visibleRange + howManyPages * visibleRange.getLength());
}

void ScrollBar::scrollToTop()
{
	setCurrentRange (visibleRange.movedToStartAt (getMinimumRangeLimit()));
}

void ScrollBar::scrollToBottom()
{
	setCurrentRange (visibleRange.movedToEndAt (getMaximumRangeLimit()));
}

void ScrollBar::setButtonRepeatSpeed (const int initialDelayInMillisecs_,
									  const int repeatDelayInMillisecs_,
									  const int minimumDelayInMillisecs_)
{
	initialDelayInMillisecs = initialDelayInMillisecs_;
	repeatDelayInMillisecs = repeatDelayInMillisecs_;
	minimumDelayInMillisecs = minimumDelayInMillisecs_;

	if (upButton != nullptr)
	{
		upButton->setRepeatSpeed (initialDelayInMillisecs,  repeatDelayInMillisecs,  minimumDelayInMillisecs);
		downButton->setRepeatSpeed (initialDelayInMillisecs,  repeatDelayInMillisecs,  minimumDelayInMillisecs);
	}
}

void ScrollBar::addListener (Listener* const listener)
{
	listeners.add (listener);
}

void ScrollBar::removeListener (Listener* const listener)
{
	listeners.remove (listener);
}

void ScrollBar::handleAsyncUpdate()
{
	double start = visibleRange.getStart(); // (need to use a temp variable for VC7 compatibility)
	listeners.call (&ScrollBar::Listener::scrollBarMoved, this, start);
}

void ScrollBar::updateThumbPosition()
{
	int newThumbSize = roundToInt (totalRange.getLength() > 0 ? (visibleRange.getLength() * thumbAreaSize) / totalRange.getLength()
															  : thumbAreaSize);

	if (newThumbSize < getLookAndFeel().getMinimumScrollbarThumbSize (*this))
		newThumbSize = jmin (getLookAndFeel().getMinimumScrollbarThumbSize (*this), thumbAreaSize - 1);

	if (newThumbSize > thumbAreaSize)
		newThumbSize = thumbAreaSize;

	int newThumbStart = thumbAreaStart;

	if (totalRange.getLength() > visibleRange.getLength())
		newThumbStart += roundToInt (((visibleRange.getStart() - totalRange.getStart()) * (thumbAreaSize - newThumbSize))
										 / (totalRange.getLength() - visibleRange.getLength()));

	setVisible ((! autohides) || (totalRange.getLength() > visibleRange.getLength() && visibleRange.getLength() > 0.0));

	if (thumbStart != newThumbStart  || thumbSize != newThumbSize)
	{
		const int repaintStart = jmin (thumbStart, newThumbStart) - 4;
		const int repaintSize = jmax (thumbStart + thumbSize, newThumbStart + newThumbSize) + 8 - repaintStart;

		if (vertical)
			repaint (0, repaintStart, getWidth(), repaintSize);
		else
			repaint (repaintStart, 0, repaintSize, getHeight());

		thumbStart = newThumbStart;
		thumbSize = newThumbSize;
	}
}

void ScrollBar::setOrientation (const bool shouldBeVertical)
{
	if (vertical != shouldBeVertical)
	{
		vertical = shouldBeVertical;

		if (upButton != nullptr)
		{
			upButton->direction	= vertical ? 0 : 3;
			downButton->direction  = vertical ? 2 : 1;
		}

		updateThumbPosition();
	}
}

void ScrollBar::setButtonVisibility (const bool buttonsAreVisible)
{
	upButton = nullptr;
	downButton = nullptr;

	if (buttonsAreVisible)
	{
		addAndMakeVisible (upButton   = new ScrollbarButton (vertical ? 0 : 3, *this));
		addAndMakeVisible (downButton = new ScrollbarButton (vertical ? 2 : 1, *this));

		setButtonRepeatSpeed (initialDelayInMillisecs, repeatDelayInMillisecs, minimumDelayInMillisecs);
	}

	updateThumbPosition();
}

void ScrollBar::setAutoHide (const bool shouldHideWhenFullRange)
{
	autohides = shouldHideWhenFullRange;
	updateThumbPosition();
}

bool ScrollBar::autoHides() const noexcept
{
	return autohides;
}

void ScrollBar::paint (Graphics& g)
{
	if (thumbAreaSize > 0)
	{
		LookAndFeel& lf = getLookAndFeel();

		const int thumb = (thumbAreaSize > lf.getMinimumScrollbarThumbSize (*this))
								? thumbSize : 0;

		if (vertical)
		{
			lf.drawScrollbar (g, *this,
							  0, thumbAreaStart,
							  getWidth(), thumbAreaSize,
							  vertical,
							  thumbStart, thumb,
							  isMouseOver(), isMouseButtonDown());
		}
		else
		{
			lf.drawScrollbar (g, *this,
							  thumbAreaStart, 0,
							  thumbAreaSize, getHeight(),
							  vertical,
							  thumbStart, thumb,
							  isMouseOver(), isMouseButtonDown());
		}
	}
}

void ScrollBar::lookAndFeelChanged()
{
	setComponentEffect (getLookAndFeel().getScrollbarEffect());
}

void ScrollBar::resized()
{
	const int length = vertical ? getHeight() : getWidth();

	const int buttonSize = upButton != nullptr ? jmin (getLookAndFeel().getScrollbarButtonSize (*this), length / 2)
											   : 0;

	if (length < 32 + getLookAndFeel().getMinimumScrollbarThumbSize (*this))
	{
		thumbAreaStart = length / 2;
		thumbAreaSize = 0;
	}
	else
	{
		thumbAreaStart = buttonSize;
		thumbAreaSize = length - (buttonSize << 1);
	}

	if (upButton != nullptr)
	{
		if (vertical)
		{
			upButton->setBounds (0, 0, getWidth(), buttonSize);
			downButton->setBounds (0, thumbAreaStart + thumbAreaSize, getWidth(), buttonSize);
		}
		else
		{
			upButton->setBounds (0, 0, buttonSize, getHeight());
			downButton->setBounds (thumbAreaStart + thumbAreaSize, 0, buttonSize, getHeight());
		}
	}

	updateThumbPosition();
}

void ScrollBar::mouseDown (const MouseEvent& e)
{
	isDraggingThumb = false;
	lastMousePos = vertical ? e.y : e.x;
	dragStartMousePos = lastMousePos;
	dragStartRange = visibleRange.getStart();

	if (dragStartMousePos < thumbStart)
	{
		moveScrollbarInPages (-1);
		startTimer (400);
	}
	else if (dragStartMousePos >= thumbStart + thumbSize)
	{
		moveScrollbarInPages (1);
		startTimer (400);
	}
	else
	{
		isDraggingThumb = (thumbAreaSize > getLookAndFeel().getMinimumScrollbarThumbSize (*this))
							&& (thumbAreaSize > thumbSize);
	}
}

void ScrollBar::mouseDrag (const MouseEvent& e)
{
	const int mousePos = vertical ? e.y : e.x;

	if (isDraggingThumb && lastMousePos != mousePos)
	{
		const int deltaPixels = mousePos - dragStartMousePos;

		setCurrentRangeStart (dragStartRange
								+ deltaPixels * (totalRange.getLength() - visibleRange.getLength())
									/ (thumbAreaSize - thumbSize));
	}

	lastMousePos = mousePos;
}

void ScrollBar::mouseUp (const MouseEvent&)
{
	isDraggingThumb = false;
	stopTimer();
	repaint();
}

void ScrollBar::mouseWheelMove (const MouseEvent&,
								float wheelIncrementX,
								float wheelIncrementY)
{
	float increment = vertical ? wheelIncrementY : wheelIncrementX;

	if (increment < 0)
		increment = jmin (increment * 10.0f, -1.0f);
	else if (increment > 0)
		increment = jmax (increment * 10.0f, 1.0f);

	setCurrentRange (visibleRange - singleStepSize * increment);
}

void ScrollBar::timerCallback()
{
	if (isMouseButtonDown())
	{
		startTimer (40);

		if (lastMousePos < thumbStart)
			setCurrentRange (visibleRange - visibleRange.getLength());
		else if (lastMousePos > thumbStart + thumbSize)
			setCurrentRangeStart (visibleRange.getEnd());
	}
	else
	{
		stopTimer();
	}
}

bool ScrollBar::keyPressed (const KeyPress& key)
{
	if (! isVisible())
		return false;

	if (key.isKeyCode (KeyPress::upKey) || key.isKeyCode (KeyPress::leftKey))
		moveScrollbarInSteps (-1);
	else if (key.isKeyCode (KeyPress::downKey) || key.isKeyCode (KeyPress::rightKey))
		moveScrollbarInSteps (1);
	else if (key.isKeyCode (KeyPress::pageUpKey))
		moveScrollbarInPages (-1);
	else if (key.isKeyCode (KeyPress::pageDownKey))
		moveScrollbarInPages (1);
	else if (key.isKeyCode (KeyPress::homeKey))
		scrollToTop();
	else if (key.isKeyCode (KeyPress::endKey))
		scrollToBottom();
	else
		return false;

	return true;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ScrollBar.cpp ***/


/*** Start of inlined file: juce_StretchableLayoutManager.cpp ***/
BEGIN_JUCE_NAMESPACE

StretchableLayoutManager::StretchableLayoutManager()
	: totalSize (0)
{
}

StretchableLayoutManager::~StretchableLayoutManager()
{
}

void StretchableLayoutManager::clearAllItems()
{
	items.clear();
	totalSize = 0;
}

void StretchableLayoutManager::setItemLayout (const int itemIndex,
											  const double minimumSize,
											  const double maximumSize,
											  const double preferredSize)
{
	ItemLayoutProperties* layout = getInfoFor (itemIndex);

	if (layout == nullptr)
	{
		layout = new ItemLayoutProperties();
		layout->itemIndex = itemIndex;

		int i;
		for (i = 0; i < items.size(); ++i)
			if (items.getUnchecked (i)->itemIndex > itemIndex)
				break;

		items.insert (i, layout);
	}

	layout->minSize = minimumSize;
	layout->maxSize = maximumSize;
	layout->preferredSize = preferredSize;
	layout->currentSize = 0;
}

bool StretchableLayoutManager::getItemLayout (const int itemIndex,
											  double& minimumSize,
											  double& maximumSize,
											  double& preferredSize) const
{
	const ItemLayoutProperties* const layout = getInfoFor (itemIndex);

	if (layout != nullptr)
	{
		minimumSize = layout->minSize;
		maximumSize = layout->maxSize;
		preferredSize = layout->preferredSize;
		return true;
	}

	return false;
}

void StretchableLayoutManager::setTotalSize (const int newTotalSize)
{
	totalSize = newTotalSize;

	fitComponentsIntoSpace (0, items.size(), totalSize, 0);
}

int StretchableLayoutManager::getItemCurrentPosition (const int itemIndex) const
{
	int pos = 0;

	for (int i = 0; i < itemIndex; ++i)
	{
		const ItemLayoutProperties* const layout = getInfoFor (i);

		if (layout != nullptr)
			pos += layout->currentSize;
	}

	return pos;
}

int StretchableLayoutManager::getItemCurrentAbsoluteSize (const int itemIndex) const
{
	const ItemLayoutProperties* const layout = getInfoFor (itemIndex);

	if (layout != nullptr)
		return layout->currentSize;

	return 0;
}

double StretchableLayoutManager::getItemCurrentRelativeSize (const int itemIndex) const
{
	const ItemLayoutProperties* const layout = getInfoFor (itemIndex);

	if (layout != nullptr)
		return -layout->currentSize / (double) totalSize;

	return 0;
}

void StretchableLayoutManager::setItemPosition (const int itemIndex,
												int newPosition)
{
	for (int i = items.size(); --i >= 0;)
	{
		const ItemLayoutProperties* const layout = items.getUnchecked(i);

		if (layout->itemIndex == itemIndex)
		{
			int realTotalSize = jmax (totalSize, getMinimumSizeOfItems (0, items.size()));
			const int minSizeAfterThisComp = getMinimumSizeOfItems (i, items.size());
			const int maxSizeAfterThisComp = getMaximumSizeOfItems (i + 1, items.size());

			newPosition = jmax (newPosition, totalSize - maxSizeAfterThisComp - layout->currentSize);
			newPosition = jmin (newPosition, realTotalSize - minSizeAfterThisComp);

			int endPos = fitComponentsIntoSpace (0, i, newPosition, 0);

			endPos += layout->currentSize;

			fitComponentsIntoSpace (i + 1, items.size(), totalSize - endPos, endPos);
			updatePrefSizesToMatchCurrentPositions();
			break;
		}
	}
}

void StretchableLayoutManager::layOutComponents (Component** const components,
												 int numComponents,
												 int x, int y, int w, int h,
												 const bool vertically,
												 const bool resizeOtherDimension)
{
	setTotalSize (vertically ? h : w);
	int pos = vertically ? y : x;

	for (int i = 0; i < numComponents; ++i)
	{
		const ItemLayoutProperties* const layout = getInfoFor (i);

		if (layout != nullptr)
		{
			Component* const c = components[i];

			if (c != nullptr)
			{
				if (i == numComponents - 1)
				{
					// if it's the last item, crop it to exactly fit the available space..
					if (resizeOtherDimension)
					{
						if (vertically)
							c->setBounds (x, pos, w, jmax (layout->currentSize, h - pos));
						else
							c->setBounds (pos, y, jmax (layout->currentSize, w - pos), h);
					}
					else
					{
						if (vertically)
							c->setBounds (c->getX(), pos, c->getWidth(), jmax (layout->currentSize, h - pos));
						else
							c->setBounds (pos, c->getY(), jmax (layout->currentSize, w - pos), c->getHeight());
					}
				}
				else
				{
					if (resizeOtherDimension)
					{
						if (vertically)
							c->setBounds (x, pos, w, layout->currentSize);
						else
							c->setBounds (pos, y, layout->currentSize, h);
					}
					else
					{
						if (vertically)
							c->setBounds (c->getX(), pos, c->getWidth(), layout->currentSize);
						else
							c->setBounds (pos, c->getY(), layout->currentSize, c->getHeight());
					}
				}
			}

			pos += layout->currentSize;
		}
	}
}

StretchableLayoutManager::ItemLayoutProperties* StretchableLayoutManager::getInfoFor (const int itemIndex) const
{
	for (int i = items.size(); --i >= 0;)
		if (items.getUnchecked(i)->itemIndex == itemIndex)
			return items.getUnchecked(i);

	return nullptr;
}

int StretchableLayoutManager::fitComponentsIntoSpace (const int startIndex,
													  const int endIndex,
													  const int availableSpace,
													  int startPos)
{
	// calculate the total sizes
	int i;
	double totalIdealSize = 0.0;
	int totalMinimums = 0;

	for (i = startIndex; i < endIndex; ++i)
	{
		ItemLayoutProperties* const layout = items.getUnchecked (i);

		layout->currentSize = sizeToRealSize (layout->minSize, totalSize);

		totalMinimums += layout->currentSize;
		totalIdealSize += sizeToRealSize (layout->preferredSize, totalSize);
   }

	if (totalIdealSize <= 0)
		totalIdealSize = 1.0;

	// now calc the best sizes..
	int extraSpace = availableSpace - totalMinimums;

	while (extraSpace > 0)
	{
		int numWantingMoreSpace = 0;
		int numHavingTakenExtraSpace = 0;

		// first figure out how many comps want a slice of the extra space..
		for (i = startIndex; i < endIndex; ++i)
		{
			ItemLayoutProperties* const layout = items.getUnchecked (i);

			double sizeWanted = sizeToRealSize (layout->preferredSize, totalSize);

			const int bestSize = jlimit (layout->currentSize,
										 jmax (layout->currentSize,
											   sizeToRealSize (layout->maxSize, totalSize)),
										 roundToInt (sizeWanted * availableSpace / totalIdealSize));

			if (bestSize > layout->currentSize)
				++numWantingMoreSpace;
		}

		// ..share out the extra space..
		for (i = startIndex; i < endIndex; ++i)
		{
			ItemLayoutProperties* const layout = items.getUnchecked (i);

			double sizeWanted = sizeToRealSize (layout->preferredSize, totalSize);

			int bestSize = jlimit (layout->currentSize,
								   jmax (layout->currentSize, sizeToRealSize (layout->maxSize, totalSize)),
								   roundToInt (sizeWanted * availableSpace / totalIdealSize));

			const int extraWanted = bestSize - layout->currentSize;

			if (extraWanted > 0)
			{
				const int extraAllowed = jmin (extraWanted,
												extraSpace / jmax (1, numWantingMoreSpace));

				if (extraAllowed > 0)
				{
					++numHavingTakenExtraSpace;
					--numWantingMoreSpace;

					layout->currentSize += extraAllowed;
					extraSpace -= extraAllowed;
				}
			}
		}

		if (numHavingTakenExtraSpace <= 0)
			break;
	}

	// ..and calculate the end position
	for (i = startIndex; i < endIndex; ++i)
	{
		ItemLayoutProperties* const layout = items.getUnchecked(i);
		startPos += layout->currentSize;
	}

	return startPos;
}

int StretchableLayoutManager::getMinimumSizeOfItems (const int startIndex,
													 const int endIndex) const
{
	int totalMinimums = 0;

	for (int i = startIndex; i < endIndex; ++i)
		totalMinimums += sizeToRealSize (items.getUnchecked (i)->minSize, totalSize);

	return totalMinimums;
}

int StretchableLayoutManager::getMaximumSizeOfItems (const int startIndex, const int endIndex) const
{
	int totalMaximums = 0;

	for (int i = startIndex; i < endIndex; ++i)
		totalMaximums += sizeToRealSize (items.getUnchecked (i)->maxSize, totalSize);

	return totalMaximums;
}

void StretchableLayoutManager::updatePrefSizesToMatchCurrentPositions()
{
	for (int i = 0; i < items.size(); ++i)
	{
		ItemLayoutProperties* const layout = items.getUnchecked (i);

		layout->preferredSize
			= (layout->preferredSize < 0) ? getItemCurrentRelativeSize (i)
										  : getItemCurrentAbsoluteSize (i);
	}
}

int StretchableLayoutManager::sizeToRealSize (double size, int totalSpace)
{
	if (size < 0)
		size *= -totalSpace;

	return roundToInt (size);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_StretchableLayoutManager.cpp ***/


/*** Start of inlined file: juce_StretchableLayoutResizerBar.cpp ***/
BEGIN_JUCE_NAMESPACE

StretchableLayoutResizerBar::StretchableLayoutResizerBar (StretchableLayoutManager* layout_,
														  const int itemIndex_,
														  const bool isVertical_)
	: layout (layout_),
	  itemIndex (itemIndex_),
	  isVertical (isVertical_)
{
	setRepaintsOnMouseActivity (true);
	setMouseCursor (MouseCursor (isVertical_ ? MouseCursor::LeftRightResizeCursor
											 : MouseCursor::UpDownResizeCursor));
}

StretchableLayoutResizerBar::~StretchableLayoutResizerBar()
{
}

void StretchableLayoutResizerBar::paint (Graphics& g)
{
	getLookAndFeel().drawStretchableLayoutResizerBar (g,
													  getWidth(), getHeight(),
													  isVertical,
													  isMouseOver(),
													  isMouseButtonDown());
}

void StretchableLayoutResizerBar::mouseDown (const MouseEvent&)
{
	mouseDownPos = layout->getItemCurrentPosition (itemIndex);
}

void StretchableLayoutResizerBar::mouseDrag (const MouseEvent& e)
{
	const int desiredPos = mouseDownPos + (isVertical ? e.getDistanceFromDragStartX()
													  : e.getDistanceFromDragStartY());

	if (layout->getItemCurrentPosition (itemIndex) != desiredPos)
	{
		layout->setItemPosition (itemIndex, desiredPos);
		hasBeenMoved();
	}
}

void StretchableLayoutResizerBar::hasBeenMoved()
{
	if (getParentComponent() != nullptr)
		getParentComponent()->resized();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_StretchableLayoutResizerBar.cpp ***/


/*** Start of inlined file: juce_StretchableObjectResizer.cpp ***/
BEGIN_JUCE_NAMESPACE

StretchableObjectResizer::StretchableObjectResizer()
{
}

StretchableObjectResizer::~StretchableObjectResizer()
{
}

void StretchableObjectResizer::addItem (const double size,
										const double minSize, const double maxSize,
										const int order)
{
	// the order must be >= 0 but less than the maximum integer value.
	jassert (order >= 0 && order < std::numeric_limits<int>::max());

	Item* const item = new Item();
	item->size = size;
	item->minSize = minSize;
	item->maxSize = maxSize;
	item->order = order;
	items.add (item);
}

double StretchableObjectResizer::getItemSize (const int index) const noexcept
{
	const Item* const it = items [index];
	return it != nullptr ? it->size : 0;
}

void StretchableObjectResizer::resizeToFit (const double targetSize)
{
	int order = 0;

	for (;;)
	{
		double currentSize = 0;
		double minSize = 0;
		double maxSize = 0;

		int nextHighestOrder = std::numeric_limits<int>::max();

		for (int i = 0; i < items.size(); ++i)
		{
			const Item* const it = items.getUnchecked(i);
			currentSize += it->size;

			if (it->order <= order)
			{
				minSize += it->minSize;
				maxSize += it->maxSize;
			}
			else
			{
				minSize += it->size;
				maxSize += it->size;
				nextHighestOrder = jmin (nextHighestOrder, it->order);
			}
		}

		const double thisIterationTarget = jlimit (minSize, maxSize, targetSize);

		if (thisIterationTarget >= currentSize)
		{
			const double availableExtraSpace = maxSize - currentSize;
			const double targetAmountOfExtraSpace = thisIterationTarget - currentSize;
			const double scale = targetAmountOfExtraSpace / availableExtraSpace;

			for (int i = 0; i < items.size(); ++i)
			{
				Item* const it = items.getUnchecked(i);

				if (it->order <= order)
					it->size = jmin (it->maxSize, it->size + (it->maxSize - it->size) * scale);
			}
		}
		else
		{
			const double amountOfSlack = currentSize - minSize;
			const double targetAmountOfSlack = thisIterationTarget - minSize;
			const double scale = targetAmountOfSlack / amountOfSlack;

			for (int i = 0; i < items.size(); ++i)
			{
				Item* const it = items.getUnchecked(i);

				if (it->order <= order)
					it->size = jmax (it->minSize, it->minSize + (it->size - it->minSize) * scale);
			}
		}

		if (nextHighestOrder < std::numeric_limits<int>::max())
			order = nextHighestOrder;
		else
			break;
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_StretchableObjectResizer.cpp ***/


/*** Start of inlined file: juce_TabbedButtonBar.cpp ***/
BEGIN_JUCE_NAMESPACE

TabBarButton::TabBarButton (const String& name, TabbedButtonBar& owner_)
	: Button (name),
	  owner (owner_),
	  overlapPixels (0)
{
	shadow.setShadowProperties (2.2f, 0.7f, 0, 0);
	setComponentEffect (&shadow);
	setWantsKeyboardFocus (false);
}

TabBarButton::~TabBarButton()
{
}

int TabBarButton::getIndex() const
{
	return owner.indexOfTabButton (this);
}

void TabBarButton::paintButton (Graphics& g,
								bool isMouseOverButton,
								bool isButtonDown)
{
	const Rectangle<int> area (getActiveArea());
	g.setOrigin (area.getX(), area.getY());

	getLookAndFeel()
		.drawTabButton (g, area.getWidth(), area.getHeight(),
						owner.getTabBackgroundColour (getIndex()),
						getIndex(), getButtonText(), *this,
						owner.getOrientation(),
						isMouseOverButton, isButtonDown,
						getToggleState());
}

void TabBarButton::clicked (const ModifierKeys& mods)
{
	if (mods.isPopupMenu())
		owner.popupMenuClickOnTab (getIndex(), getButtonText());
	else
		owner.setCurrentTabIndex (getIndex());
}

bool TabBarButton::hitTest (int mx, int my)
{
	const Rectangle<int> area (getActiveArea());

	if (owner.getOrientation() == TabbedButtonBar::TabsAtLeft
		 || owner.getOrientation() == TabbedButtonBar::TabsAtRight)
	{
		if (isPositiveAndBelow (mx, getWidth())
			 && my >= area.getY() + overlapPixels
			 && my < area.getBottom() - overlapPixels)
			return true;
	}
	else
	{
		if (mx >= area.getX() + overlapPixels && mx < area.getRight() - overlapPixels
			 && isPositiveAndBelow (my, getHeight()))
			return true;
	}

	Path p;
	getLookAndFeel()
		.createTabButtonShape (p, area.getWidth(), area.getHeight(), getIndex(), getButtonText(), *this,
							   owner.getOrientation(), false, false, getToggleState());

	return p.contains ((float) (mx - area.getX()),
					   (float) (my - area.getY()));
}

int TabBarButton::getBestTabLength (const int depth)
{
	return jlimit (depth * 2,
				   depth * 7,
				   getLookAndFeel().getTabButtonBestWidth (getIndex(), getButtonText(), depth, *this));
}

const Rectangle<int> TabBarButton::getActiveArea()
{
	Rectangle<int> r (getLocalBounds());
	const int spaceAroundImage = getLookAndFeel().getTabButtonSpaceAroundImage();

	if (owner.getOrientation() != TabbedButtonBar::TabsAtLeft)	  r.removeFromRight (spaceAroundImage);
	if (owner.getOrientation() != TabbedButtonBar::TabsAtRight)	 r.removeFromLeft (spaceAroundImage);
	if (owner.getOrientation() != TabbedButtonBar::TabsAtBottom)	r.removeFromTop (spaceAroundImage);
	if (owner.getOrientation() != TabbedButtonBar::TabsAtTop)	   r.removeFromBottom (spaceAroundImage);

	return r;
}

class TabbedButtonBar::BehindFrontTabComp  : public Component,
											 public ButtonListener // (can't use Button::Listener due to idiotic VC2005 bug)
{
public:
	BehindFrontTabComp (TabbedButtonBar& owner_)
		: owner (owner_)
	{
		setInterceptsMouseClicks (false, false);
	}

	void paint (Graphics& g)
	{
		getLookAndFeel().drawTabAreaBehindFrontButton (g, getWidth(), getHeight(),
													   owner, owner.getOrientation());
	}

	void enablementChanged()
	{
		repaint();
	}

	void buttonClicked (Button*)
	{
		owner.showExtraItemsMenu();
	}

private:
	TabbedButtonBar& owner;

	JUCE_DECLARE_NON_COPYABLE (BehindFrontTabComp);
};

TabbedButtonBar::TabbedButtonBar (const Orientation orientation_)
	: orientation (orientation_),
	  minimumScale (0.7),
	  currentTabIndex (-1)
{
	setInterceptsMouseClicks (false, true);
	addAndMakeVisible (behindFrontTab = new BehindFrontTabComp (*this));
	setFocusContainer (true);
}

TabbedButtonBar::~TabbedButtonBar()
{
	tabs.clear();
	extraTabsButton = nullptr;
}

void TabbedButtonBar::setOrientation (const Orientation newOrientation)
{
	orientation = newOrientation;

	for (int i = getNumChildComponents(); --i >= 0;)
		getChildComponent (i)->resized();

	resized();
}

TabBarButton* TabbedButtonBar::createTabButton (const String& name, const int /*index*/)
{
	return new TabBarButton (name, *this);
}

void TabbedButtonBar::setMinimumTabScaleFactor (double newMinimumScale)
{
	minimumScale = newMinimumScale;
	resized();
}

void TabbedButtonBar::clearTabs()
{
	tabs.clear();
	extraTabsButton = nullptr;
	setCurrentTabIndex (-1);
}

void TabbedButtonBar::addTab (const String& tabName,
							  const Colour& tabBackgroundColour,
							  int insertIndex)
{
	jassert (tabName.isNotEmpty()); // you have to give them all a name..

	if (tabName.isNotEmpty())
	{
		if (! isPositiveAndBelow (insertIndex, tabs.size()))
			insertIndex = tabs.size();

		TabInfo* newTab = new TabInfo();
		newTab->name = tabName;
		newTab->colour = tabBackgroundColour;
		newTab->component = createTabButton (tabName, insertIndex);

		jassert (newTab->component != nullptr);

		tabs.insert (insertIndex, newTab);
		addAndMakeVisible (newTab->component, insertIndex);

		resized();

		if (currentTabIndex < 0)
			setCurrentTabIndex (0);
	}
}

void TabbedButtonBar::setTabName (const int tabIndex, const String& newName)
{
	TabInfo* const tab = tabs [tabIndex];

	if (tab != nullptr && tab->name != newName)
	{
		tab->name = newName;
		tab->component->setButtonText (newName);
		resized();
	}
}

void TabbedButtonBar::removeTab (const int tabIndex)
{
	if (tabs [tabIndex] != nullptr)
	{
		const int oldTabIndex = currentTabIndex;
		if (currentTabIndex == tabIndex)
			currentTabIndex = -1;

		tabs.remove (tabIndex);
		resized();

		setCurrentTabIndex (jlimit (0, jmax (0, tabs.size() - 1), oldTabIndex));
	}
}

void TabbedButtonBar::moveTab (const int currentIndex, const int newIndex)
{
	tabs.move (currentIndex, newIndex);
	resized();
}

int TabbedButtonBar::getNumTabs() const
{
	return tabs.size();
}

String TabbedButtonBar::getCurrentTabName() const
{
	TabInfo* tab = tabs [currentTabIndex];
	return tab == nullptr ? String::empty : tab->name;
}

StringArray TabbedButtonBar::getTabNames() const
{
	StringArray names;

	for (int i = 0; i < tabs.size(); ++i)
		names.add (tabs.getUnchecked(i)->name);

	return names;
}

void TabbedButtonBar::setCurrentTabIndex (int newIndex, const bool sendChangeMessage_)
{
	if (currentTabIndex != newIndex)
	{
		if (! isPositiveAndBelow (newIndex, tabs.size()))
			newIndex = -1;

		currentTabIndex = newIndex;

		for (int i = 0; i < tabs.size(); ++i)
		{
			TabBarButton* tb = tabs.getUnchecked(i)->component;
			tb->setToggleState (i == newIndex, false);
		}

		resized();

		if (sendChangeMessage_)
			sendChangeMessage();

		currentTabChanged (newIndex, getCurrentTabName());
	}
}

TabBarButton* TabbedButtonBar::getTabButton (const int index) const
{
	TabInfo* const tab = tabs[index];
	return tab == nullptr ? nullptr : static_cast <TabBarButton*> (tab->component);
}

int TabbedButtonBar::indexOfTabButton (const TabBarButton* button) const
{
	for (int i = tabs.size(); --i >= 0;)
		if (tabs.getUnchecked(i)->component == button)
			return i;

	return -1;
}

void TabbedButtonBar::lookAndFeelChanged()
{
	extraTabsButton = nullptr;
	resized();
}

void TabbedButtonBar::resized()
{
	int depth = getWidth();
	int length = getHeight();

	if (orientation == TabsAtTop || orientation == TabsAtBottom)
		std::swap (depth, length);

	const int overlap = getLookAndFeel().getTabButtonOverlap (depth)
							+ getLookAndFeel().getTabButtonSpaceAroundImage() * 2;

	int i, totalLength = overlap;
	int numVisibleButtons = tabs.size();

	for (i = 0; i < tabs.size(); ++i)
	{
		TabBarButton* const tb = tabs.getUnchecked(i)->component;

		totalLength += tb->getBestTabLength (depth) - overlap;
		tb->overlapPixels = overlap / 2;
	}

	double scale = 1.0;

	if (totalLength > length)
		scale = jmax (minimumScale, length / (double) totalLength);

	const bool isTooBig = totalLength * scale > length;
	int tabsButtonPos = 0;

	if (isTooBig)
	{
		if (extraTabsButton == nullptr)
		{
			addAndMakeVisible (extraTabsButton = getLookAndFeel().createTabBarExtrasButton());
			extraTabsButton->addListener (behindFrontTab);
			extraTabsButton->setAlwaysOnTop (true);
			extraTabsButton->setTriggeredOnMouseDown (true);
		}

		const int buttonSize = jmin (proportionOfWidth (0.7f), proportionOfHeight (0.7f));
		extraTabsButton->setSize (buttonSize, buttonSize);

		if (orientation == TabsAtTop || orientation == TabsAtBottom)
		{
			tabsButtonPos = getWidth() - buttonSize / 2 - 1;
			extraTabsButton->setCentrePosition (tabsButtonPos, getHeight() / 2);
		}
		else
		{
			tabsButtonPos = getHeight() - buttonSize / 2 - 1;
			extraTabsButton->setCentrePosition (getWidth() / 2, tabsButtonPos);
		}

		totalLength = 0;

		for (i = 0; i < tabs.size(); ++i)
		{
			TabBarButton* const tb = tabs.getUnchecked(i)->component;

			const int newLength = totalLength + tb->getBestTabLength (depth);

			if (i > 0 && newLength * minimumScale > tabsButtonPos)
			{
				totalLength += overlap;
				break;
			}

			numVisibleButtons = i + 1;
			totalLength = newLength - overlap;
		}

		scale = jmax (minimumScale, tabsButtonPos / (double) totalLength);
	}
	else
	{
		extraTabsButton = nullptr;
	}

	int pos = 0;

	TabBarButton* frontTab = nullptr;

	for (i = 0; i < tabs.size(); ++i)
	{
		TabBarButton* const tb = getTabButton (i);

		if (tb != nullptr)
		{
			const int bestLength = roundToInt (scale * tb->getBestTabLength (depth));

			if (i < numVisibleButtons)
			{
				if (orientation == TabsAtTop || orientation == TabsAtBottom)
					tb->setBounds (pos, 0, bestLength, getHeight());
				else
					tb->setBounds (0, pos, getWidth(), bestLength);

				tb->toBack();

				if (i == currentTabIndex)
					frontTab = tb;

				tb->setVisible (true);
			}
			else
			{
				tb->setVisible (false);
			}

			pos += bestLength - overlap;
		}
	}

	behindFrontTab->setBounds (getLocalBounds());

	if (frontTab != nullptr)
	{
		frontTab->toFront (false);
		behindFrontTab->toBehind (frontTab);
	}
}

const Colour TabbedButtonBar::getTabBackgroundColour (const int tabIndex)
{
	TabInfo* const tab = tabs [tabIndex];
	return tab == nullptr ? Colours::white : tab->colour;
}

void TabbedButtonBar::setTabBackgroundColour (const int tabIndex, const Colour& newColour)
{
	TabInfo* const tab = tabs [tabIndex];

	if (tab != nullptr && tab->colour != newColour)
	{
		tab->colour = newColour;
		repaint();
	}
}

void TabbedButtonBar::extraItemsMenuCallback (int result, TabbedButtonBar* bar)
{
	if (bar != nullptr && result > 0)
		bar->setCurrentTabIndex (result - 1);
}

void TabbedButtonBar::showExtraItemsMenu()
{
	PopupMenu m;

	for (int i = 0; i < tabs.size(); ++i)
	{
		const TabInfo* const tab = tabs.getUnchecked(i);

		if (! tab->component->isVisible())
			m.addItem (i + 1, tab->name, true, i == currentTabIndex);
	}

	m.showMenuAsync (PopupMenu::Options().withTargetComponent (extraTabsButton),
					 ModalCallbackFunction::forComponent (extraItemsMenuCallback, this));
}

void TabbedButtonBar::currentTabChanged (const int, const String&)
{
}

void TabbedButtonBar::popupMenuClickOnTab (const int, const String&)
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TabbedButtonBar.cpp ***/


/*** Start of inlined file: juce_TabbedComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace TabbedComponentHelpers
{
	const Identifier deleteComponentId ("deleteByTabComp_");

	void deleteIfNecessary (Component* const comp)
	{
		if (comp != nullptr && (bool) comp->getProperties() [deleteComponentId])
			delete comp;
	}

	const Rectangle<int> getTabArea (Rectangle<int>& content, BorderSize<int>& outline,
									 const TabbedButtonBar::Orientation orientation, const int tabDepth)
	{
		switch (orientation)
		{
			case TabbedButtonBar::TabsAtTop:	outline.setTop (0);	 return content.removeFromTop (tabDepth);
			case TabbedButtonBar::TabsAtBottom: outline.setBottom (0);  return content.removeFromBottom (tabDepth);
			case TabbedButtonBar::TabsAtLeft:   outline.setLeft (0);	return content.removeFromLeft (tabDepth);
			case TabbedButtonBar::TabsAtRight:  outline.setRight (0);   return content.removeFromRight (tabDepth);
			default: jassertfalse; break;
		}

		return Rectangle<int>();
	}
}

class TabbedComponent::ButtonBar  : public TabbedButtonBar
{
public:
	ButtonBar (TabbedComponent& owner_, const TabbedButtonBar::Orientation orientation_)
		: TabbedButtonBar (orientation_),
		  owner (owner_)
	{
	}

	void currentTabChanged (int newCurrentTabIndex, const String& newTabName)
	{
		owner.changeCallback (newCurrentTabIndex, newTabName);
	}

	void popupMenuClickOnTab (int tabIndex, const String& tabName)
	{
		owner.popupMenuClickOnTab (tabIndex, tabName);
	}

	const Colour getTabBackgroundColour (const int tabIndex)
	{
		return owner.tabs->getTabBackgroundColour (tabIndex);
	}

	TabBarButton* createTabButton (const String& tabName, int tabIndex)
	{
		return owner.createTabButton (tabName, tabIndex);
	}

private:
	TabbedComponent& owner;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ButtonBar);
};

TabbedComponent::TabbedComponent (const TabbedButtonBar::Orientation orientation)
	: tabDepth (30),
	  outlineThickness (1),
	  edgeIndent (0)
{
	addAndMakeVisible (tabs = new ButtonBar (*this, orientation));
}

TabbedComponent::~TabbedComponent()
{
	clearTabs();
	tabs = nullptr;
}

void TabbedComponent::setOrientation (const TabbedButtonBar::Orientation orientation)
{
	tabs->setOrientation (orientation);
	resized();
}

TabbedButtonBar::Orientation TabbedComponent::getOrientation() const noexcept
{
	return tabs->getOrientation();
}

void TabbedComponent::setTabBarDepth (const int newDepth)
{
	if (tabDepth != newDepth)
	{
		tabDepth = newDepth;
		resized();
	}
}

TabBarButton* TabbedComponent::createTabButton (const String& tabName, const int /*tabIndex*/)
{
	return new TabBarButton (tabName, *tabs);
}

void TabbedComponent::clearTabs()
{
	if (panelComponent != nullptr)
	{
		panelComponent->setVisible (false);
		removeChildComponent (panelComponent);
		panelComponent = nullptr;
	}

	tabs->clearTabs();

	for (int i = contentComponents.size(); --i >= 0;)
		TabbedComponentHelpers::deleteIfNecessary (contentComponents.getReference (i));

	contentComponents.clear();
}

void TabbedComponent::addTab (const String& tabName,
							  const Colour& tabBackgroundColour,
							  Component* const contentComponent,
							  const bool deleteComponentWhenNotNeeded,
							  const int insertIndex)
{
	contentComponents.insert (insertIndex, WeakReference<Component> (contentComponent));

	if (deleteComponentWhenNotNeeded && contentComponent != nullptr)
		contentComponent->getProperties().set (TabbedComponentHelpers::deleteComponentId, true);

	tabs->addTab (tabName, tabBackgroundColour, insertIndex);
}

void TabbedComponent::setTabName (const int tabIndex, const String& newName)
{
	tabs->setTabName (tabIndex, newName);
}

void TabbedComponent::removeTab (const int tabIndex)
{
	if (isPositiveAndBelow (tabIndex, contentComponents.size()))
	{
		TabbedComponentHelpers::deleteIfNecessary (contentComponents.getReference (tabIndex));
		contentComponents.remove (tabIndex);
		tabs->removeTab (tabIndex);
	}
}

int TabbedComponent::getNumTabs() const
{
	return tabs->getNumTabs();
}

StringArray TabbedComponent::getTabNames() const
{
	return tabs->getTabNames();
}

Component* TabbedComponent::getTabContentComponent (const int tabIndex) const noexcept
{
	return contentComponents [tabIndex];
}

const Colour TabbedComponent::getTabBackgroundColour (const int tabIndex) const noexcept
{
	return tabs->getTabBackgroundColour (tabIndex);
}

void TabbedComponent::setTabBackgroundColour (const int tabIndex, const Colour& newColour)
{
	tabs->setTabBackgroundColour (tabIndex, newColour);

	if (getCurrentTabIndex() == tabIndex)
		repaint();
}

void TabbedComponent::setCurrentTabIndex (const int newTabIndex, const bool sendChangeMessage)
{
	tabs->setCurrentTabIndex (newTabIndex, sendChangeMessage);
}

int TabbedComponent::getCurrentTabIndex() const
{
	return tabs->getCurrentTabIndex();
}

String TabbedComponent::getCurrentTabName() const
{
	return tabs->getCurrentTabName();
}

void TabbedComponent::setOutline (const int thickness)
{
	outlineThickness = thickness;
	resized();
	repaint();
}

void TabbedComponent::setIndent (const int indentThickness)
{
	edgeIndent = indentThickness;
	resized();
	repaint();
}

void TabbedComponent::paint (Graphics& g)
{
	g.fillAll (findColour (backgroundColourId));

	Rectangle<int> content (getLocalBounds());
	BorderSize<int> outline (outlineThickness);
	TabbedComponentHelpers::getTabArea (content, outline, getOrientation(), tabDepth);

	g.reduceClipRegion (content);
	g.fillAll (tabs->getTabBackgroundColour (getCurrentTabIndex()));

	if (outlineThickness > 0)
	{
		RectangleList rl (content);
		rl.subtract (outline.subtractedFrom (content));

		g.reduceClipRegion (rl);
		g.fillAll (findColour (outlineColourId));
	}
}

void TabbedComponent::resized()
{
	Rectangle<int> content (getLocalBounds());
	BorderSize<int> outline (outlineThickness);

	tabs->setBounds (TabbedComponentHelpers::getTabArea (content, outline, getOrientation(), tabDepth));
	content = BorderSize<int> (edgeIndent).subtractedFrom (outline.subtractedFrom (content));

	for (int i = contentComponents.size(); --i >= 0;)
		if (contentComponents.getReference (i) != nullptr)
			contentComponents.getReference (i)->setBounds (content);
}

void TabbedComponent::lookAndFeelChanged()
{
	for (int i = contentComponents.size(); --i >= 0;)
		if (contentComponents.getReference (i) != nullptr)
			contentComponents.getReference (i)->lookAndFeelChanged();
}

void TabbedComponent::changeCallback (const int newCurrentTabIndex, const String& newTabName)
{
	if (panelComponent != nullptr)
	{
		panelComponent->setVisible (false);
		removeChildComponent (panelComponent);
		panelComponent = nullptr;
	}

	if (getCurrentTabIndex() >= 0)
	{
		panelComponent = getTabContentComponent (getCurrentTabIndex());

		if (panelComponent != nullptr)
		{
			// do these ops as two stages instead of addAndMakeVisible() so that the
			// component has always got a parent when it gets the visibilityChanged() callback
			addChildComponent (panelComponent);
			panelComponent->setVisible (true);
			panelComponent->toFront (true);
		}

		repaint();
	}

	resized();

	currentTabChanged (newCurrentTabIndex, newTabName);
}

void TabbedComponent::currentTabChanged (const int, const String&) {}
void TabbedComponent::popupMenuClickOnTab (const int, const String&) {}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TabbedComponent.cpp ***/


#endif // JUCE_BUILD_GUI
