/*
  ==============================================================================

   This file is part of the JUCE library - "Jules' Utility Class Extensions"
   Copyright 2004-9 by Raw Material Software Ltd.

  ------------------------------------------------------------------------------

   JUCE can be redistributed and/or modified under the terms of the GNU General
   Public License (Version 2), as published by the Free Software Foundation.
   A copy of the license is included in the JUCE distribution, or can be found
   online at www.gnu.org/licenses.

   JUCE is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  ------------------------------------------------------------------------------

   To release a closed-source product which uses JUCE, commercial licenses are
   available: visit www.rawmaterialsoftware.com/juce for more information.

  ==============================================================================
*/

#ifdef __JUCE_JUCEHEADER__
 /* When you add the amalgamated cpp file to your project, you mustn't include it in
	a file where you've already included juce.h - just put it inside a file on its own,
	possibly with your config flags preceding it, but don't include anything else. */
 #error
#endif


#include "juce_TargetPlatform.h"

 // FORCE_AMALGAMATOR_INCLUDE

#include "juce_Config.h"

  // FORCE_AMALGAMATOR_INCLUDE

#ifndef JUCE_BUILD_CORE
 #define JUCE_BUILD_CORE 1
#endif
#ifndef JUCE_BUILD_MISC
 #define JUCE_BUILD_MISC 1
#endif
#ifndef JUCE_BUILD_GUI
 #define JUCE_BUILD_GUI 1
#endif
#ifndef JUCE_BUILD_NATIVE
 #define JUCE_BUILD_NATIVE 1
#endif

#if JUCE_ONLY_BUILD_CORE_LIBRARY
 #undef JUCE_BUILD_MISC
 #undef JUCE_BUILD_GUI
#endif

#ifndef JUCE_NAMESPACE
 #define JUCE_NAMESPACE juce
#endif

//==============================================================================
#include "juce_NativeIncludes.h"
//==============================================================================
#define DONT_SET_USING_JUCE_NAMESPACE 1
#undef max
#undef min
#define NO_DUMMY_DECL

#define JUCE_AMALGAMATED_TEMPLATE 1

#if JUCE_BUILD_NATIVE
 #include "juce_amalgamated.h"  // FORCE_AMALGAMATOR_INCLUDE
#endif

#if (defined(_MSC_VER) && (_MSC_VER <= 1200))
 #pragma warning (disable: 4309 4305)
#endif


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////

#if JUCE_BUILD_GUI
/*** Start of inlined file: juce_Viewport.cpp ***/
BEGIN_JUCE_NAMESPACE

Viewport::Viewport (const String& name)
  : Component (name),
	scrollBarThickness (0),
	singleStepX (16),
	singleStepY (16),
	showHScrollbar (true),
	showVScrollbar (true),
	deleteContent (true),
	verticalScrollBar (true),
	horizontalScrollBar (false)
{
	// content holder is used to clip the contents so they don't overlap the scrollbars
	addAndMakeVisible (&contentHolder);
	contentHolder.setInterceptsMouseClicks (false, true);

	addChildComponent (&verticalScrollBar);
	addChildComponent (&horizontalScrollBar);

	verticalScrollBar.addListener (this);
	horizontalScrollBar.addListener (this);

	setInterceptsMouseClicks (false, true);
	setWantsKeyboardFocus (true);
}

Viewport::~Viewport()
{
	deleteContentComp();
}

void Viewport::visibleAreaChanged (const Rectangle<int>&)
{
}

void Viewport::deleteContentComp()
{
	if (deleteContent)
	{
		// This sets the content comp to a null pointer before deleting the old one, in case
		// anything tries to use the old one while it's in mid-deletion..
		ScopedPointer<Component> oldCompDeleter (contentComp);
		contentComp = nullptr;
	}
	else
	{
		contentComp = nullptr;
	}
}

void Viewport::setViewedComponent (Component* const newViewedComponent, const bool deleteComponentWhenNoLongerNeeded)
{
	if (contentComp.get() != newViewedComponent)
	{
		deleteContentComp();
		contentComp = newViewedComponent;
		deleteContent = deleteComponentWhenNoLongerNeeded;

		if (contentComp != nullptr)
		{
			contentHolder.addAndMakeVisible (contentComp);
			setViewPosition (0, 0);
			contentComp->addComponentListener (this);
		}

		updateVisibleArea();
	}
}

int Viewport::getMaximumVisibleWidth() const	{ return contentHolder.getWidth(); }
int Viewport::getMaximumVisibleHeight() const   { return contentHolder.getHeight(); }

void Viewport::setViewPosition (const int xPixelsOffset, const int yPixelsOffset)
{
	if (contentComp != nullptr)
		contentComp->setTopLeftPosition (jmax (jmin (0, contentHolder.getWidth() - contentComp->getWidth()), jmin (0, -xPixelsOffset)),
										 jmax (jmin (0, contentHolder.getHeight() - contentComp->getHeight()), jmin (0, -yPixelsOffset)));
}

void Viewport::setViewPosition (const Point<int>& newPosition)
{
	setViewPosition (newPosition.getX(), newPosition.getY());
}

void Viewport::setViewPositionProportionately (const double x, const double y)
{
	if (contentComp != nullptr)
		setViewPosition (jmax (0, roundToInt (x * (contentComp->getWidth() - getWidth()))),
						 jmax (0, roundToInt (y * (contentComp->getHeight() - getHeight()))));
}

bool Viewport::autoScroll (const int mouseX, const int mouseY, const int activeBorderThickness, const int maximumSpeed)
{
	if (contentComp != nullptr)
	{
		int dx = 0, dy = 0;

		if (horizontalScrollBar.isVisible() || contentComp->getX() < 0 || contentComp->getRight() > getWidth())
		{
			if (mouseX < activeBorderThickness)
				dx = activeBorderThickness - mouseX;
			else if (mouseX >= contentHolder.getWidth() - activeBorderThickness)
				dx = (contentHolder.getWidth() - activeBorderThickness) - mouseX;

			if (dx < 0)
				dx = jmax (dx, -maximumSpeed, contentHolder.getWidth() - contentComp->getRight());
			else
				dx = jmin (dx, maximumSpeed, -contentComp->getX());
		}

		if (verticalScrollBar.isVisible() || contentComp->getY() < 0 || contentComp->getBottom() > getHeight())
		{
			if (mouseY < activeBorderThickness)
				dy = activeBorderThickness - mouseY;
			else if (mouseY >= contentHolder.getHeight() - activeBorderThickness)
				dy = (contentHolder.getHeight() - activeBorderThickness) - mouseY;

			if (dy < 0)
				dy = jmax (dy, -maximumSpeed, contentHolder.getHeight() - contentComp->getBottom());
			else
				dy = jmin (dy, maximumSpeed, -contentComp->getY());
		}

		if (dx != 0 || dy != 0)
		{
			contentComp->setTopLeftPosition (contentComp->getX() + dx,
											 contentComp->getY() + dy);

			return true;
		}
	}

	return false;
}

void Viewport::componentMovedOrResized (Component&, bool, bool)
{
	updateVisibleArea();
}

void Viewport::resized()
{
	updateVisibleArea();
}

void Viewport::updateVisibleArea()
{
	const int scrollbarWidth = getScrollBarThickness();
	const bool canShowAnyBars = getWidth() > scrollbarWidth && getHeight() > scrollbarWidth;
	const bool canShowHBar = showHScrollbar && canShowAnyBars;
	const bool canShowVBar = showVScrollbar && canShowAnyBars;

	bool hBarVisible = canShowHBar && ! horizontalScrollBar.autoHides();
	bool vBarVisible = canShowVBar && ! verticalScrollBar.autoHides();

	Rectangle<int> contentArea (getLocalBounds());

	if (contentComp != nullptr && ! contentArea.contains (contentComp->getBounds()))
	{
		hBarVisible = canShowHBar && (hBarVisible || contentComp->getX() < 0 || contentComp->getRight() > contentArea.getWidth());
		vBarVisible = canShowVBar && (vBarVisible || contentComp->getY() < 0 || contentComp->getBottom() > contentArea.getHeight());

		if (vBarVisible)
			contentArea.setWidth (getWidth() - scrollbarWidth);

		if (hBarVisible)
			contentArea.setHeight (getHeight() - scrollbarWidth);

		if (! contentArea.contains (contentComp->getBounds()))
		{
			hBarVisible = canShowHBar && (hBarVisible || contentComp->getRight() > contentArea.getWidth());
			vBarVisible = canShowVBar && (vBarVisible || contentComp->getBottom() > contentArea.getHeight());
		}
	}

	if (vBarVisible)
		contentArea.setWidth (getWidth() - scrollbarWidth);

	if (hBarVisible)
		contentArea.setHeight (getHeight() - scrollbarWidth);

	contentHolder.setBounds (contentArea);

	Rectangle<int> contentBounds;
	if (contentComp != nullptr)
		contentBounds = contentHolder.getLocalArea (contentComp, contentComp->getLocalBounds());

	Point<int> visibleOrigin (-contentBounds.getPosition());

	if (hBarVisible)
	{
		horizontalScrollBar.setBounds (0, contentArea.getHeight(), contentArea.getWidth(), scrollbarWidth);
		horizontalScrollBar.setRangeLimits (0.0, contentBounds.getWidth());
		horizontalScrollBar.setCurrentRange (visibleOrigin.getX(), contentArea.getWidth());
		horizontalScrollBar.setSingleStepSize (singleStepX);
		horizontalScrollBar.cancelPendingUpdate();
	}
	else if (canShowHBar)
	{
		visibleOrigin.setX (0);
	}

	if (vBarVisible)
	{
		verticalScrollBar.setBounds (contentArea.getWidth(), 0, scrollbarWidth, contentArea.getHeight());
		verticalScrollBar.setRangeLimits (0.0, contentBounds.getHeight());
		verticalScrollBar.setCurrentRange (visibleOrigin.getY(), contentArea.getHeight());
		verticalScrollBar.setSingleStepSize (singleStepY);
		verticalScrollBar.cancelPendingUpdate();
	}
	else if (canShowVBar)
	{
		visibleOrigin.setY (0);
	}

	// Force the visibility *after* setting the ranges to avoid flicker caused by edge conditions in the numbers.
	horizontalScrollBar.setVisible (hBarVisible);
	verticalScrollBar.setVisible (vBarVisible);

	setViewPosition (visibleOrigin);

	const Rectangle<int> visibleArea (visibleOrigin.getX(), visibleOrigin.getY(),
									  jmin (contentBounds.getWidth() - visibleOrigin.getX(),  contentArea.getWidth()),
									  jmin (contentBounds.getHeight() - visibleOrigin.getY(), contentArea.getHeight()));

	if (lastVisibleArea != visibleArea)
	{
		lastVisibleArea = visibleArea;
		visibleAreaChanged (visibleArea);
	}

	horizontalScrollBar.handleUpdateNowIfNeeded();
	verticalScrollBar.handleUpdateNowIfNeeded();
}

void Viewport::setSingleStepSizes (const int stepX, const int stepY)
{
	if (singleStepX != stepX || singleStepY != stepY)
	{
		singleStepX = stepX;
		singleStepY = stepY;
		updateVisibleArea();
	}
}

void Viewport::setScrollBarsShown (const bool showVerticalScrollbarIfNeeded,
								   const bool showHorizontalScrollbarIfNeeded)
{
	if (showVScrollbar != showVerticalScrollbarIfNeeded
		 || showHScrollbar != showHorizontalScrollbarIfNeeded)
	{
		showVScrollbar = showVerticalScrollbarIfNeeded;
		showHScrollbar = showHorizontalScrollbarIfNeeded;
		updateVisibleArea();
	}
}

void Viewport::setScrollBarThickness (const int thickness)
{
	if (scrollBarThickness != thickness)
	{
		scrollBarThickness = thickness;
		updateVisibleArea();
	}
}

int Viewport::getScrollBarThickness() const
{
	return scrollBarThickness > 0 ? scrollBarThickness
								  : getLookAndFeel().getDefaultScrollbarWidth();
}

void Viewport::setScrollBarButtonVisibility (const bool buttonsVisible)
{
	verticalScrollBar.setButtonVisibility (buttonsVisible);
	horizontalScrollBar.setButtonVisibility (buttonsVisible);
}

void Viewport::scrollBarMoved (ScrollBar* scrollBarThatHasMoved, double newRangeStart)
{
	const int newRangeStartInt = roundToInt (newRangeStart);

	if (scrollBarThatHasMoved == &horizontalScrollBar)
	{
		setViewPosition (newRangeStartInt, getViewPositionY());
	}
	else if (scrollBarThatHasMoved == &verticalScrollBar)
	{
		setViewPosition (getViewPositionX(), newRangeStartInt);
	}
}

void Viewport::mouseWheelMove (const MouseEvent& e, const float wheelIncrementX, const float wheelIncrementY)
{
	if (! useMouseWheelMoveIfNeeded (e, wheelIncrementX, wheelIncrementY))
		Component::mouseWheelMove (e, wheelIncrementX, wheelIncrementY);
}

bool Viewport::useMouseWheelMoveIfNeeded (const MouseEvent& e, float wheelIncrementX, float wheelIncrementY)
{
	if (! (e.mods.isAltDown() || e.mods.isCtrlDown()))
	{
		const bool hasVertBar = verticalScrollBar.isVisible();
		const bool hasHorzBar = horizontalScrollBar.isVisible();

		if (hasHorzBar || hasVertBar)
		{
			if (wheelIncrementX != 0)
			{
				wheelIncrementX *= 14.0f * singleStepX;
				wheelIncrementX = (wheelIncrementX < 0) ? jmin (wheelIncrementX, -1.0f)
														: jmax (wheelIncrementX, 1.0f);
			}

			if (wheelIncrementY != 0)
			{
				wheelIncrementY *= 14.0f * singleStepY;
				wheelIncrementY = (wheelIncrementY < 0) ? jmin (wheelIncrementY, -1.0f)
														: jmax (wheelIncrementY, 1.0f);
			}

			Point<int> pos (getViewPosition());

			if (wheelIncrementX != 0 && wheelIncrementY != 0 && hasHorzBar && hasVertBar)
			{
				pos.setX (pos.getX() - roundToInt (wheelIncrementX));
				pos.setY (pos.getY() - roundToInt (wheelIncrementY));
			}
			else if (hasHorzBar && (wheelIncrementX != 0 || e.mods.isShiftDown() || ! hasVertBar))
			{
				if (wheelIncrementX == 0 && ! hasVertBar)
					wheelIncrementX = wheelIncrementY;

				pos.setX (pos.getX() - roundToInt (wheelIncrementX));
			}
			else if (hasVertBar && wheelIncrementY != 0)
			{
				pos.setY (pos.getY() - roundToInt (wheelIncrementY));
			}

			if (pos != getViewPosition())
			{
				setViewPosition (pos);
				return true;
			}
		}
	}

	return false;
}

bool Viewport::keyPressed (const KeyPress& key)
{
	const bool isUpDownKey = key.isKeyCode (KeyPress::upKey)
								|| key.isKeyCode (KeyPress::downKey)
								|| key.isKeyCode (KeyPress::pageUpKey)
								|| key.isKeyCode (KeyPress::pageDownKey)
								|| key.isKeyCode (KeyPress::homeKey)
								|| key.isKeyCode (KeyPress::endKey);

	if (verticalScrollBar.isVisible() && isUpDownKey)
		return verticalScrollBar.keyPressed (key);

	const bool isLeftRightKey = key.isKeyCode (KeyPress::leftKey)
								 || key.isKeyCode (KeyPress::rightKey);

	if (horizontalScrollBar.isVisible() && (isUpDownKey || isLeftRightKey))
		return horizontalScrollBar.keyPressed (key);

	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Viewport.cpp ***/


/*** Start of inlined file: juce_LookAndFeel.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace LookAndFeelHelpers
{
	void createRoundedPath (Path& p,
							const float x, const float y,
							const float w, const float h,
							const float cs,
							const bool curveTopLeft, const bool curveTopRight,
							const bool curveBottomLeft, const bool curveBottomRight) noexcept
	{
		const float cs2 = 2.0f * cs;

		if (curveTopLeft)
		{
			p.startNewSubPath (x, y + cs);
			p.addArc (x, y, cs2, cs2, float_Pi * 1.5f, float_Pi * 2.0f);
		}
		else
		{
			p.startNewSubPath (x, y);
		}

		if (curveTopRight)
		{
			p.lineTo (x + w - cs, y);
			p.addArc (x + w - cs2, y, cs2, cs2, 0.0f, float_Pi * 0.5f);
		}
		else
		{
			p.lineTo (x + w, y);
		}

		if (curveBottomRight)
		{
			p.lineTo (x + w, y + h - cs);
			p.addArc (x + w - cs2, y + h - cs2, cs2, cs2, float_Pi * 0.5f, float_Pi);
		}
		else
		{
			p.lineTo (x + w, y + h);
		}

		if (curveBottomLeft)
		{
			p.lineTo (x + cs, y + h);
			p.addArc (x, y + h - cs2, cs2, cs2, float_Pi, float_Pi * 1.5f);
		}
		else
		{
			p.lineTo (x, y + h);
		}

		p.closeSubPath();
	}

	const Colour createBaseColour (const Colour& buttonColour,
								   const bool hasKeyboardFocus,
								   const bool isMouseOverButton,
								   const bool isButtonDown) noexcept
	{
		const float sat = hasKeyboardFocus ? 1.3f : 0.9f;
		const Colour baseColour (buttonColour.withMultipliedSaturation (sat));

		if (isButtonDown)
			return baseColour.contrasting (0.2f);
		else if (isMouseOverButton)
			return baseColour.contrasting (0.1f);

		return baseColour;
	}

	const TextLayout layoutTooltipText (const String& text) noexcept
	{
		const float tooltipFontSize = 12.0f;
		const int maxToolTipWidth = 400;

		const Font f (tooltipFontSize, Font::bold);
		TextLayout tl (text, f);
		tl.layout (maxToolTipWidth, Justification::left, true);

		return tl;
	}
}

LookAndFeel::LookAndFeel()
	: useNativeAlertWindows (false)
{
	/* if this fails it means you're trying to create a LookAndFeel object before
	   the static Colours have been initialised. That ain't gonna work. It probably
	   means that you're using a static LookAndFeel object and that your compiler has
	   decided to intialise it before the Colours class.
	*/
	jassert (Colours::white == Colour (0xffffffff));

	// set up the standard set of colours..
	const unsigned int textButtonColour = 0xffbbbbff;
	const unsigned int textHighlightColour = 0x401111ee;
	const unsigned int standardOutlineColour = 0xb2808080;

	static const unsigned int standardColours[] =
	{
		TextButton::buttonColourId,		 textButtonColour,
		TextButton::buttonOnColourId,		   0xff4444ff,
		TextButton::textColourOnId,		 0xff000000,
		TextButton::textColourOffId,		0xff000000,

		ComboBox::buttonColourId,		   0xffbbbbff,
		ComboBox::outlineColourId,		  standardOutlineColour,

		ToggleButton::textColourId,		 0xff000000,

		TextEditor::backgroundColourId,		 0xffffffff,
		TextEditor::textColourId,		   0xff000000,
		TextEditor::highlightColourId,		  textHighlightColour,
		TextEditor::highlightedTextColourId,	0xff000000,
		TextEditor::outlineColourId,		0x00000000,
		TextEditor::focusedOutlineColourId,	 textButtonColour,
		TextEditor::shadowColourId,		 0x38000000,

		CaretComponent::caretColourId,		  0xff000000,

		Label::backgroundColourId,		  0x00000000,
		Label::textColourId,			0xff000000,
		Label::outlineColourId,			 0x00000000,

		ScrollBar::backgroundColourId,		  0x00000000,
		ScrollBar::thumbColourId,		   0xffffffff,

		TreeView::linesColourId,			0x4c000000,
		TreeView::backgroundColourId,		   0x00000000,
		TreeView::dragAndDropIndicatorColourId,	 0x80ff0000,

		PopupMenu::backgroundColourId,		  0xffffffff,
		PopupMenu::textColourId,			0xff000000,
		PopupMenu::headerTextColourId,		  0xff000000,
		PopupMenu::highlightedTextColourId,	 0xffffffff,
		PopupMenu::highlightedBackgroundColourId,   0x991111aa,

		ComboBox::textColourId,			 0xff000000,
		ComboBox::backgroundColourId,		   0xffffffff,
		ComboBox::arrowColourId,			0x99000000,

		ListBox::backgroundColourId,		0xffffffff,
		ListBox::outlineColourId,		   standardOutlineColour,
		ListBox::textColourId,			  0xff000000,

		Slider::backgroundColourId,		 0x00000000,
		Slider::thumbColourId,			  textButtonColour,
		Slider::trackColourId,			  0x7fffffff,
		Slider::rotarySliderFillColourId,	   0x7f0000ff,
		Slider::rotarySliderOutlineColourId,	0x66000000,
		Slider::textBoxTextColourId,		0xff000000,
		Slider::textBoxBackgroundColourId,	  0xffffffff,
		Slider::textBoxHighlightColourId,	   textHighlightColour,
		Slider::textBoxOutlineColourId,		 standardOutlineColour,

		ResizableWindow::backgroundColourId,	0xff777777,
		//DocumentWindow::textColourId,		   0xff000000, // (this is deliberately not set)

		AlertWindow::backgroundColourId,		0xffededed,
		AlertWindow::textColourId,		  0xff000000,
		AlertWindow::outlineColourId,		   0xff666666,

		ProgressBar::backgroundColourId,		0xffeeeeee,
		ProgressBar::foregroundColourId,		0xffaaaaee,

		TooltipWindow::backgroundColourId,	  0xffeeeebb,
		TooltipWindow::textColourId,		0xff000000,
		TooltipWindow::outlineColourId,		 0x4c000000,

		TabbedComponent::backgroundColourId,	0x00000000,
		TabbedComponent::outlineColourId,	   0xff777777,
		TabbedButtonBar::tabOutlineColourId,	0x80000000,
		TabbedButtonBar::frontOutlineColourId,	  0x90000000,

		Toolbar::backgroundColourId,		0xfff6f8f9,
		Toolbar::separatorColourId,		 0x4c000000,
		Toolbar::buttonMouseOverBackgroundColourId, 0x4c0000ff,
		Toolbar::buttonMouseDownBackgroundColourId, 0x800000ff,
		Toolbar::labelTextColourId,		 0xff000000,
		Toolbar::editingModeOutlineColourId,	0xffff0000,

		HyperlinkButton::textColourId,		  0xcc1111ee,

		GroupComponent::outlineColourId,		0x66000000,
		GroupComponent::textColourId,		   0xff000000,

		DirectoryContentsDisplayComponent::highlightColourId,   textHighlightColour,
		DirectoryContentsDisplayComponent::textColourId,	0xff000000,

		0x1000440, /*LassoComponent::lassoFillColourId*/	0x66dddddd,
		0x1000441, /*LassoComponent::lassoOutlineColourId*/	 0x99111111,

		CodeEditorComponent::backgroundColourId,		0xffffffff,
		CodeEditorComponent::highlightColourId,		 textHighlightColour,
		CodeEditorComponent::defaultTextColourId,		   0xff000000,

		ColourSelector::backgroundColourId,			 0xffe5e5e5,
		ColourSelector::labelTextColourId,			  0xff000000,

		KeyMappingEditorComponent::backgroundColourId,	  0x00000000,
		KeyMappingEditorComponent::textColourId,		0xff000000,

		FileSearchPathListComponent::backgroundColourId,	0xffffffff,

		FileChooserDialogBox::titleTextColourId,		0xff000000,

		DrawableButton::textColourId,			   0xff000000,
	};

	for (int i = 0; i < numElementsInArray (standardColours); i += 2)
		setColour (standardColours [i], Colour (standardColours [i + 1]));

	static String defaultSansName, defaultSerifName, defaultFixedName, defaultFallback;

	if (defaultSansName.isEmpty())
		Font::getPlatformDefaultFontNames (defaultSansName, defaultSerifName, defaultFixedName, defaultFallback);

	defaultSans = defaultSansName;
	defaultSerif = defaultSerifName;
	defaultFixed = defaultFixedName;
	Font::setFallbackFontName (defaultFallback);
}

LookAndFeel::~LookAndFeel()
{
	weakReferenceMaster.clear();
}

const WeakReference<LookAndFeel>::SharedRef& LookAndFeel::getWeakReference()
{
	return weakReferenceMaster (this);
}

const Colour LookAndFeel::findColour (const int colourId) const noexcept
{
	const int index = colourIds.indexOf (colourId);

	if (index >= 0)
		return colours [index];

	jassertfalse;
	return Colours::black;
}

void LookAndFeel::setColour (const int colourId, const Colour& colour) noexcept
{
	const int index = colourIds.indexOf (colourId);

	if (index >= 0)
	{
		colours.set (index, colour);
	}
	else
	{
		colourIds.add (colourId);
		colours.add (colour);
	}
}

bool LookAndFeel::isColourSpecified (const int colourId) const noexcept
{
	return colourIds.contains (colourId);
}

LookAndFeel& LookAndFeel::getDefaultLookAndFeel() noexcept
{
	return Desktop::getInstance().getDefaultLookAndFeel();
}

void LookAndFeel::setDefaultLookAndFeel (LookAndFeel* newDefaultLookAndFeel) noexcept
{
	Desktop::getInstance().setDefaultLookAndFeel (newDefaultLookAndFeel);
}

const Typeface::Ptr LookAndFeel::getTypefaceForFont (const Font& font)
{
	String faceName (font.getTypefaceName());

	if (faceName == Font::getDefaultSansSerifFontName())
		faceName = defaultSans;
	else if (faceName == Font::getDefaultSerifFontName())
		faceName = defaultSerif;
	else if (faceName == Font::getDefaultMonospacedFontName())
		faceName = defaultFixed;

	Font f (font);
	f.setTypefaceName (faceName);
	return Typeface::createSystemTypefaceFor (f);
}

void LookAndFeel::setDefaultSansSerifTypefaceName (const String& newName)
{
	defaultSans = newName;
}

const MouseCursor LookAndFeel::getMouseCursorFor (Component& component)
{
	return component.getMouseCursor();
}

void LookAndFeel::drawButtonBackground (Graphics& g,
										Button& button,
										const Colour& backgroundColour,
										bool isMouseOverButton,
										bool isButtonDown)
{
	const int width = button.getWidth();
	const int height = button.getHeight();

	const float outlineThickness = button.isEnabled() ? ((isButtonDown || isMouseOverButton) ? 1.2f : 0.7f) : 0.4f;
	const float halfThickness = outlineThickness * 0.5f;

	const float indentL = button.isConnectedOnLeft()   ? 0.1f : halfThickness;
	const float indentR = button.isConnectedOnRight()  ? 0.1f : halfThickness;
	const float indentT = button.isConnectedOnTop()	? 0.1f : halfThickness;
	const float indentB = button.isConnectedOnBottom() ? 0.1f : halfThickness;

	const Colour baseColour (LookAndFeelHelpers::createBaseColour (backgroundColour,
																   button.hasKeyboardFocus (true),
																   isMouseOverButton, isButtonDown)
							   .withMultipliedAlpha (button.isEnabled() ? 1.0f : 0.5f));

	drawGlassLozenge (g,
					  indentL,
					  indentT,
					  width - indentL - indentR,
					  height - indentT - indentB,
					  baseColour, outlineThickness, -1.0f,
					  button.isConnectedOnLeft(),
					  button.isConnectedOnRight(),
					  button.isConnectedOnTop(),
					  button.isConnectedOnBottom());
}

const Font LookAndFeel::getFontForTextButton (TextButton& button)
{
	return button.getFont();
}

void LookAndFeel::drawButtonText (Graphics& g, TextButton& button,
								  bool /*isMouseOverButton*/, bool /*isButtonDown*/)
{
	Font font (getFontForTextButton (button));
	g.setFont (font);
	g.setColour (button.findColour (button.getToggleState() ? TextButton::textColourOnId
															: TextButton::textColourOffId)
					   .withMultipliedAlpha (button.isEnabled() ? 1.0f : 0.5f));

	const int yIndent = jmin (4, button.proportionOfHeight (0.3f));
	const int cornerSize = jmin (button.getHeight(), button.getWidth()) / 2;

	const int fontHeight = roundToInt (font.getHeight() * 0.6f);
	const int leftIndent  = jmin (fontHeight, 2 + cornerSize / (button.isConnectedOnLeft() ? 4 : 2));
	const int rightIndent = jmin (fontHeight, 2 + cornerSize / (button.isConnectedOnRight() ? 4 : 2));

	g.drawFittedText (button.getButtonText(),
					  leftIndent,
					  yIndent,
					  button.getWidth() - leftIndent - rightIndent,
					  button.getHeight() - yIndent * 2,
					  Justification::centred, 2);
}

void LookAndFeel::drawTickBox (Graphics& g,
							   Component& component,
							   float x, float y, float w, float h,
							   const bool ticked,
							   const bool isEnabled,
							   const bool isMouseOverButton,
							   const bool isButtonDown)
{
	const float boxSize = w * 0.7f;

	drawGlassSphere (g, x, y + (h - boxSize) * 0.5f, boxSize,
					 LookAndFeelHelpers::createBaseColour (component.findColour (TextButton::buttonColourId)
																	.withMultipliedAlpha (isEnabled ? 1.0f : 0.5f),
														   true, isMouseOverButton, isButtonDown),
					 isEnabled ? ((isButtonDown || isMouseOverButton) ? 1.1f : 0.5f) : 0.3f);

	if (ticked)
	{
		Path tick;
		tick.startNewSubPath (1.5f, 3.0f);
		tick.lineTo (3.0f, 6.0f);
		tick.lineTo (6.0f, 0.0f);

		g.setColour (isEnabled ? Colours::black : Colours::grey);

		const AffineTransform trans (AffineTransform::scale (w / 9.0f, h / 9.0f)
										 .translated (x, y));

		g.strokePath (tick, PathStrokeType (2.5f), trans);
	}
}

void LookAndFeel::drawToggleButton (Graphics& g,
									ToggleButton& button,
									bool isMouseOverButton,
									bool isButtonDown)
{
	if (button.hasKeyboardFocus (true))
	{
		g.setColour (button.findColour (TextEditor::focusedOutlineColourId));
		g.drawRect (0, 0, button.getWidth(), button.getHeight());
	}

	float fontSize = jmin (15.0f, button.getHeight() * 0.75f);
	const float tickWidth = fontSize * 1.1f;

	drawTickBox (g, button, 4.0f, (button.getHeight() - tickWidth) * 0.5f,
				 tickWidth, tickWidth,
				 button.getToggleState(),
				 button.isEnabled(),
				 isMouseOverButton,
				 isButtonDown);

	g.setColour (button.findColour (ToggleButton::textColourId));
	g.setFont (fontSize);

	if (! button.isEnabled())
		g.setOpacity (0.5f);

	const int textX = (int) tickWidth + 5;

	g.drawFittedText (button.getButtonText(),
					  textX, 0,
					  button.getWidth() - textX - 2, button.getHeight(),
					  Justification::centredLeft, 10);
}

void LookAndFeel::changeToggleButtonWidthToFitText (ToggleButton& button)
{
	Font font (jmin (15.0f, button.getHeight() * 0.6f));

	const int tickWidth = jmin (24, button.getHeight());

	button.setSize (font.getStringWidth (button.getButtonText()) + tickWidth + 8,
					button.getHeight());
}

AlertWindow* LookAndFeel::createAlertWindow (const String& title,
											 const String& message,
											 const String& button1,
											 const String& button2,
											 const String& button3,
											 AlertWindow::AlertIconType iconType,
											 int numButtons,
											 Component* associatedComponent)
{
	AlertWindow* aw = new AlertWindow (title, message, iconType, associatedComponent);

	if (numButtons == 1)
	{
		aw->addButton (button1, 0,
					   KeyPress (KeyPress::escapeKey, 0, 0),
					   KeyPress (KeyPress::returnKey, 0, 0));
	}
	else
	{
		const KeyPress button1ShortCut (CharacterFunctions::toLowerCase (button1[0]), 0, 0);
		KeyPress button2ShortCut (CharacterFunctions::toLowerCase (button2[0]), 0, 0);
		if (button1ShortCut == button2ShortCut)
			button2ShortCut = KeyPress();

		if (numButtons == 2)
		{
			aw->addButton (button1, 1, KeyPress (KeyPress::returnKey, 0, 0), button1ShortCut);
			aw->addButton (button2, 0, KeyPress (KeyPress::escapeKey, 0, 0), button2ShortCut);
		}
		else if (numButtons == 3)
		{
			aw->addButton (button1, 1, button1ShortCut);
			aw->addButton (button2, 2, button2ShortCut);
			aw->addButton (button3, 0, KeyPress (KeyPress::escapeKey, 0, 0));
		}
	}

	return aw;
}

void LookAndFeel::drawAlertBox (Graphics& g,
								AlertWindow& alert,
								const Rectangle<int>& textArea,
								TextLayout& textLayout)
{
	g.fillAll (alert.findColour (AlertWindow::backgroundColourId));

	int iconSpaceUsed = 0;
	Justification alignment (Justification::horizontallyCentred);

	const int iconWidth = 80;
	int iconSize = jmin (iconWidth + 50, alert.getHeight() + 20);

	if (alert.containsAnyExtraComponents() || alert.getNumButtons() > 2)
		iconSize = jmin (iconSize, textArea.getHeight() + 50);

	const Rectangle<int> iconRect (iconSize / -10, iconSize / -10,
								   iconSize, iconSize);

	if (alert.getAlertType() != AlertWindow::NoIcon)
	{
		Path icon;
		uint32 colour;
		char character;

		if (alert.getAlertType() == AlertWindow::WarningIcon)
		{
			colour = 0x55ff5555;
			character = '!';

			icon.addTriangle (iconRect.getX() + iconRect.getWidth() * 0.5f, (float) iconRect.getY(),
							  (float) iconRect.getRight(), (float) iconRect.getBottom(),
							  (float) iconRect.getX(), (float) iconRect.getBottom());

			icon = icon.createPathWithRoundedCorners (5.0f);
		}
		else
		{
			colour = alert.getAlertType() == AlertWindow::InfoIcon ? 0x605555ff : 0x40b69900;
			character = alert.getAlertType() == AlertWindow::InfoIcon ? 'i' : '?';

			icon.addEllipse ((float) iconRect.getX(), (float) iconRect.getY(),
							 (float) iconRect.getWidth(), (float) iconRect.getHeight());
		}

		GlyphArrangement ga;
		ga.addFittedText (Font (iconRect.getHeight() * 0.9f, Font::bold),
						  String::charToString (character),
						  (float) iconRect.getX(), (float) iconRect.getY(),
						  (float) iconRect.getWidth(), (float) iconRect.getHeight(),
						  Justification::centred, false);
		ga.createPath (icon);

		icon.setUsingNonZeroWinding (false);
		g.setColour (Colour (colour));
		g.fillPath (icon);

		iconSpaceUsed = iconWidth;
		alignment = Justification::left;
	}

	g.setColour (alert.findColour (AlertWindow::textColourId));

	textLayout.drawWithin (g,
						   textArea.getX() + iconSpaceUsed, textArea.getY(),
						   textArea.getWidth() - iconSpaceUsed, textArea.getHeight(),
						   alignment.getFlags() | Justification::top);

	g.setColour (alert.findColour (AlertWindow::outlineColourId));
	g.drawRect (0, 0, alert.getWidth(), alert.getHeight());
}

int LookAndFeel::getAlertBoxWindowFlags()
{
	return ComponentPeer::windowAppearsOnTaskbar
			| ComponentPeer::windowHasDropShadow;
}

int LookAndFeel::getAlertWindowButtonHeight()
{
	return 28;
}

const Font LookAndFeel::getAlertWindowMessageFont()
{
	return Font (15.0f);
}

const Font LookAndFeel::getAlertWindowFont()
{
	return Font (12.0f);
}

void LookAndFeel::setUsingNativeAlertWindows (bool shouldUseNativeAlerts)
{
	useNativeAlertWindows = shouldUseNativeAlerts;
}

bool LookAndFeel::isUsingNativeAlertWindows()
{
   #if JUCE_LINUX
	return false; // not available currently..
   #else
	return useNativeAlertWindows;
   #endif
}

void LookAndFeel::drawProgressBar (Graphics& g, ProgressBar& progressBar,
								   int width, int height,
								   double progress, const String& textToShow)
{
	const Colour background (progressBar.findColour (ProgressBar::backgroundColourId));
	const Colour foreground (progressBar.findColour (ProgressBar::foregroundColourId));

	g.fillAll (background);

	if (progress >= 0.0f && progress < 1.0f)
	{
		drawGlassLozenge (g, 1.0f, 1.0f,
						  (float) jlimit (0.0, width - 2.0, progress * (width - 2.0)),
						  (float) (height - 2),
						  foreground,
						  0.5f, 0.0f,
						  true, true, true, true);
	}
	else
	{
		// spinning bar..
		g.setColour (foreground);

		const int stripeWidth = height * 2;
		const int position = (Time::getMillisecondCounter() / 15) % stripeWidth;

		Path p;

		for (float x = (float) (- position); x < width + stripeWidth; x += stripeWidth)
			p.addQuadrilateral (x, 0.0f,
								x + stripeWidth * 0.5f, 0.0f,
								x, (float) height,
								x - stripeWidth * 0.5f, (float) height);

		Image im (Image::ARGB, width, height, true);

		{
			Graphics g2 (im);
			drawGlassLozenge (g2, 1.0f, 1.0f,
							  (float) (width - 2),
							  (float) (height - 2),
							  foreground,
							  0.5f, 0.0f,
							  true, true, true, true);
		}

		g.setTiledImageFill (im, 0, 0, 0.85f);
		g.fillPath (p);
	}

	if (textToShow.isNotEmpty())
	{
		g.setColour (Colour::contrasting (background, foreground));
		g.setFont (height * 0.6f);

		g.drawText (textToShow, 0, 0, width, height, Justification::centred, false);
	}
}

void LookAndFeel::drawSpinningWaitAnimation (Graphics& g, const Colour& colour, int x, int y, int w, int h)
{
	const float radius = jmin (w, h) * 0.4f;
	const float thickness = radius * 0.15f;
	Path p;
	p.addRoundedRectangle (radius * 0.4f, thickness * -0.5f,
						   radius * 0.6f, thickness,
						   thickness * 0.5f);

	const float cx = x + w * 0.5f;
	const float cy = y + h * 0.5f;

	const uint32 animationIndex = (Time::getMillisecondCounter() / (1000 / 10)) % 12;

	for (int i = 0; i < 12; ++i)
	{
		const int n = (i + 12 - animationIndex) % 12;
		g.setColour (colour.withMultipliedAlpha ((n + 1) / 12.0f));

		g.fillPath (p, AffineTransform::rotation (i * (float_Pi / 6.0f))
									   .translated (cx, cy));
	}
}

void LookAndFeel::drawScrollbarButton (Graphics& g,
									   ScrollBar& scrollbar,
									   int width, int height,
									   int buttonDirection,
									   bool /*isScrollbarVertical*/,
									   bool /*isMouseOverButton*/,
									   bool isButtonDown)
{
	Path p;

	if (buttonDirection == 0)
		p.addTriangle (width * 0.5f, height * 0.2f,
					   width * 0.1f, height * 0.7f,
					   width * 0.9f, height * 0.7f);
	else if (buttonDirection == 1)
		p.addTriangle (width * 0.8f, height * 0.5f,
					   width * 0.3f, height * 0.1f,
					   width * 0.3f, height * 0.9f);
	else if (buttonDirection == 2)
		p.addTriangle (width * 0.5f, height * 0.8f,
					   width * 0.1f, height * 0.3f,
					   width * 0.9f, height * 0.3f);
	else if (buttonDirection == 3)
		p.addTriangle (width * 0.2f, height * 0.5f,
					   width * 0.7f, height * 0.1f,
					   width * 0.7f, height * 0.9f);

	if (isButtonDown)
		g.setColour (scrollbar.findColour (ScrollBar::thumbColourId).contrasting (0.2f));
	else
		g.setColour (scrollbar.findColour (ScrollBar::thumbColourId));

	g.fillPath (p);

	g.setColour (Colour (0x80000000));
	g.strokePath (p, PathStrokeType (0.5f));
}

void LookAndFeel::drawScrollbar (Graphics& g,
								 ScrollBar& scrollbar,
								 int x, int y,
								 int width, int height,
								 bool isScrollbarVertical,
								 int thumbStartPosition,
								 int thumbSize,
								 bool /*isMouseOver*/,
								 bool /*isMouseDown*/)
{
	g.fillAll (scrollbar.findColour (ScrollBar::backgroundColourId));

	Path slotPath, thumbPath;

	const float slotIndent = jmin (width, height) > 15 ? 1.0f : 0.0f;
	const float slotIndentx2 = slotIndent * 2.0f;
	const float thumbIndent = slotIndent + 1.0f;
	const float thumbIndentx2 = thumbIndent * 2.0f;

	float gx1 = 0.0f, gy1 = 0.0f, gx2 = 0.0f, gy2 = 0.0f;

	if (isScrollbarVertical)
	{
		slotPath.addRoundedRectangle (x + slotIndent,
									  y + slotIndent,
									  width - slotIndentx2,
									  height - slotIndentx2,
									  (width - slotIndentx2) * 0.5f);

		if (thumbSize > 0)
			thumbPath.addRoundedRectangle (x + thumbIndent,
										   thumbStartPosition + thumbIndent,
										   width - thumbIndentx2,
										   thumbSize - thumbIndentx2,
										   (width - thumbIndentx2) * 0.5f);
		gx1 = (float) x;
		gx2 = x + width * 0.7f;
	}
	else
	{
		slotPath.addRoundedRectangle (x + slotIndent,
									  y + slotIndent,
									  width - slotIndentx2,
									  height - slotIndentx2,
									  (height - slotIndentx2) * 0.5f);

		if (thumbSize > 0)
			thumbPath.addRoundedRectangle (thumbStartPosition + thumbIndent,
										   y + thumbIndent,
										   thumbSize - thumbIndentx2,
										   height - thumbIndentx2,
										   (height - thumbIndentx2) * 0.5f);
		gy1 = (float) y;
		gy2 = y + height * 0.7f;
	}

	const Colour thumbColour (scrollbar.findColour (ScrollBar::thumbColourId));
	Colour trackColour1, trackColour2;

	if (scrollbar.isColourSpecified (ScrollBar::trackColourId))
	{
		trackColour1 = trackColour2 = scrollbar.findColour (ScrollBar::trackColourId);
	}
	else
	{
		trackColour1 = thumbColour.overlaidWith (Colour (0x44000000));
		trackColour2 = thumbColour.overlaidWith (Colour (0x19000000));
	}

	g.setGradientFill (ColourGradient (trackColour1, gx1, gy1,
									   trackColour2, gx2, gy2, false));
	g.fillPath (slotPath);

	if (isScrollbarVertical)
	{
		gx1 = x + width * 0.6f;
		gx2 = (float) x + width;
	}
	else
	{
		gy1 = y + height * 0.6f;
		gy2 = (float) y + height;
	}

	g.setGradientFill (ColourGradient (Colours::transparentBlack,gx1, gy1,
					   Colour (0x19000000), gx2, gy2, false));
	g.fillPath (slotPath);

	g.setColour (thumbColour);
	g.fillPath (thumbPath);

	g.setGradientFill (ColourGradient (Colour (0x10000000), gx1, gy1,
					   Colours::transparentBlack, gx2, gy2, false));

	g.saveState();

	if (isScrollbarVertical)
		g.reduceClipRegion (x + width / 2, y, width, height);
	else
		g.reduceClipRegion (x, y + height / 2, width, height);

	g.fillPath (thumbPath);
	g.restoreState();

	g.setColour (Colour (0x4c000000));
	g.strokePath (thumbPath, PathStrokeType (0.4f));
}

ImageEffectFilter* LookAndFeel::getScrollbarEffect()
{
	return nullptr;
}

int LookAndFeel::getMinimumScrollbarThumbSize (ScrollBar& scrollbar)
{
	return jmin (scrollbar.getWidth(), scrollbar.getHeight()) * 2;
}

int LookAndFeel::getDefaultScrollbarWidth()
{
	return 18;
}

int LookAndFeel::getScrollbarButtonSize (ScrollBar& scrollbar)
{
	return 2 + (scrollbar.isVertical() ? scrollbar.getWidth()
									   : scrollbar.getHeight());
}

const Path LookAndFeel::getTickShape (const float height)
{
	static const unsigned char tickShapeData[] =
	{
		109,0,224,168,68,0,0,119,67,108,0,224,172,68,0,128,146,67,113,0,192,148,68,0,0,219,67,0,96,110,68,0,224,56,68,113,0,64,51,68,0,32,130,68,0,64,20,68,0,224,
		162,68,108,0,128,3,68,0,128,168,68,113,0,128,221,67,0,192,175,68,0,0,207,67,0,32,179,68,113,0,0,201,67,0,224,173,68,0,0,181,67,0,224,161,68,108,0,128,168,67,
		0,128,154,68,113,0,128,141,67,0,192,138,68,0,128,108,67,0,64,131,68,113,0,0,62,67,0,128,119,68,0,0,5,67,0,128,114,68,113,0,0,102,67,0,192,88,68,0,128,155,
		67,0,192,88,68,113,0,0,190,67,0,192,88,68,0,128,232,67,0,224,131,68,108,0,128,246,67,0,192,139,68,113,0,64,33,68,0,128,87,68,0,0,93,68,0,224,26,68,113,0,
		96,140,68,0,128,188,67,0,224,168,68,0,0,119,67,99,101
	};

	Path p;
	p.loadPathFromData (tickShapeData, sizeof (tickShapeData));
	p.scaleToFit (0, 0, height * 2.0f, height, true);
	return p;
}

const Path LookAndFeel::getCrossShape (const float height)
{
	static const unsigned char crossShapeData[] =
	{
		109,0,0,17,68,0,96,145,68,108,0,192,13,68,0,192,147,68,113,0,0,213,67,0,64,174,68,0,0,168,67,0,64,174,68,113,0,0,104,67,0,64,174,68,0,0,5,67,0,64,
		153,68,113,0,0,18,67,0,64,153,68,0,0,24,67,0,64,153,68,113,0,0,135,67,0,64,153,68,0,128,207,67,0,224,130,68,108,0,0,220,67,0,0,126,68,108,0,0,204,67,
		0,128,117,68,113,0,0,138,67,0,64,82,68,0,0,138,67,0,192,57,68,113,0,0,138,67,0,192,37,68,0,128,210,67,0,64,10,68,113,0,128,220,67,0,64,45,68,0,0,8,
		68,0,128,78,68,108,0,192,14,68,0,0,87,68,108,0,64,20,68,0,0,80,68,113,0,192,57,68,0,0,32,68,0,128,88,68,0,0,32,68,113,0,64,112,68,0,0,32,68,0,
		128,124,68,0,64,68,68,113,0,0,121,68,0,192,67,68,0,128,119,68,0,192,67,68,113,0,192,108,68,0,192,67,68,0,32,89,68,0,96,82,68,113,0,128,69,68,0,0,97,68,
		0,0,56,68,0,64,115,68,108,0,64,49,68,0,128,124,68,108,0,192,55,68,0,96,129,68,113,0,0,92,68,0,224,146,68,0,192,129,68,0,224,146,68,113,0,64,110,68,0,64,
		168,68,0,64,87,68,0,64,168,68,113,0,128,66,68,0,64,168,68,0,64,27,68,0,32,150,68,99,101
	};

	Path p;
	p.loadPathFromData (crossShapeData, sizeof (crossShapeData));
	p.scaleToFit (0, 0, height * 2.0f, height, true);
	return p;
}

void LookAndFeel::drawTreeviewPlusMinusBox (Graphics& g, int x, int y, int w, int h, bool isPlus, bool /*isMouseOver*/)
{
	const int boxSize = ((jmin (16, w, h) << 1) / 3) | 1;

	x += (w - boxSize) >> 1;
	y += (h - boxSize) >> 1;
	w = boxSize;
	h = boxSize;

	g.setColour (Colour (0xe5ffffff));
	g.fillRect (x, y, w, h);

	g.setColour (Colour (0x80000000));
	g.drawRect (x, y, w, h);

	const float size = boxSize / 2 + 1.0f;
	const float centre = (float) (boxSize / 2);

	g.fillRect (x + (w - size) * 0.5f, y + centre, size, 1.0f);

	if (isPlus)
		g.fillRect (x + centre, y + (h - size) * 0.5f, 1.0f, size);
}

void LookAndFeel::drawBubble (Graphics& g,
							  float tipX, float tipY,
							  float boxX, float boxY,
							  float boxW, float boxH)
{
	int side = 0;

	if (tipX < boxX)
		side = 1;
	else if (tipX > boxX + boxW)
		side = 3;
	else if (tipY > boxY + boxH)
		side = 2;

	const float indent = 2.0f;
	Path p;
	p.addBubble (boxX + indent,
				 boxY + indent,
				 boxW - indent * 2.0f,
				 boxH - indent * 2.0f,
				 5.0f,
				 tipX, tipY,
				 side,
				 0.5f,
				 jmin (15.0f, boxW * 0.3f, boxH * 0.3f));

	//xxx need to take comp as param for colour
	g.setColour (findColour (TooltipWindow::backgroundColourId).withAlpha (0.9f));
	g.fillPath (p);

	//xxx as above
	g.setColour (findColour (TooltipWindow::textColourId).withAlpha (0.4f));
	g.strokePath (p, PathStrokeType (1.33f));
}

const Font LookAndFeel::getPopupMenuFont()
{
	return Font (17.0f);
}

void LookAndFeel::getIdealPopupMenuItemSize (const String& text,
											 const bool isSeparator,
											 int standardMenuItemHeight,
											 int& idealWidth,
											 int& idealHeight)
{
	if (isSeparator)
	{
		idealWidth = 50;
		idealHeight = standardMenuItemHeight > 0 ? standardMenuItemHeight / 2 : 10;
	}
	else
	{
		Font font (getPopupMenuFont());

		if (standardMenuItemHeight > 0 && font.getHeight() > standardMenuItemHeight / 1.3f)
			font.setHeight (standardMenuItemHeight / 1.3f);

		idealHeight = standardMenuItemHeight > 0 ? standardMenuItemHeight : roundToInt (font.getHeight() * 1.3f);
		idealWidth = font.getStringWidth (text) + idealHeight * 2;
	}
}

void LookAndFeel::drawPopupMenuBackground (Graphics& g, int width, int height)
{
	const Colour background (findColour (PopupMenu::backgroundColourId));

	g.fillAll (background);
	g.setColour (background.overlaidWith (Colour (0x2badd8e6)));

	for (int i = 0; i < height; i += 3)
		g.fillRect (0, i, width, 1);

#if ! JUCE_MAC
	g.setColour (findColour (PopupMenu::textColourId).withAlpha (0.6f));
	g.drawRect (0, 0, width, height);
#endif
}

void LookAndFeel::drawPopupMenuUpDownArrow (Graphics& g,
											int width, int height,
											bool isScrollUpArrow)
{
	const Colour background (findColour (PopupMenu::backgroundColourId));

	g.setGradientFill (ColourGradient (background, 0.0f, height * 0.5f,
									   background.withAlpha (0.0f),
									   0.0f, isScrollUpArrow ? ((float) height) : 0.0f,
									   false));

	g.fillRect (1, 1, width - 2, height - 2);

	const float hw = width * 0.5f;
	const float arrowW = height * 0.3f;
	const float y1 = height * (isScrollUpArrow ? 0.6f : 0.3f);
	const float y2 = height * (isScrollUpArrow ? 0.3f : 0.6f);

	Path p;
	p.addTriangle (hw - arrowW, y1,
				   hw + arrowW, y1,
				   hw, y2);

	g.setColour (findColour (PopupMenu::textColourId).withAlpha (0.5f));
	g.fillPath (p);
}

void LookAndFeel::drawPopupMenuItem (Graphics& g,
									 int width, int height,
									 const bool isSeparator,
									 const bool isActive,
									 const bool isHighlighted,
									 const bool isTicked,
									 const bool hasSubMenu,
									 const String& text,
									 const String& shortcutKeyText,
									 Image* image,
									 const Colour* const textColourToUse)
{
	const float halfH = height * 0.5f;

	if (isSeparator)
	{
		const float separatorIndent = 5.5f;

		g.setColour (Colour (0x33000000));
		g.drawLine (separatorIndent, halfH, width - separatorIndent, halfH);

		g.setColour (Colour (0x66ffffff));
		g.drawLine (separatorIndent, halfH + 1.0f, width - separatorIndent, halfH + 1.0f);
	}
	else
	{
		Colour textColour (findColour (PopupMenu::textColourId));

		if (textColourToUse != nullptr)
			textColour = *textColourToUse;

		if (isHighlighted)
		{
			g.setColour (findColour (PopupMenu::highlightedBackgroundColourId));
			g.fillRect (1, 1, width - 2, height - 2);

			g.setColour (findColour (PopupMenu::highlightedTextColourId));
		}
		else
		{
			g.setColour (textColour);
		}

		if (! isActive)
			g.setOpacity (0.3f);

		Font font (getPopupMenuFont());

		if (font.getHeight() > height / 1.3f)
			font.setHeight (height / 1.3f);

		g.setFont (font);

		const int leftBorder = (height * 5) / 4;
		const int rightBorder = 4;

		if (image != nullptr)
		{
			g.drawImageWithin (*image,
							   2, 1, leftBorder - 4, height - 2,
							   RectanglePlacement::centred | RectanglePlacement::onlyReduceInSize, false);
		}
		else if (isTicked)
		{
			const Path tick (getTickShape (1.0f));
			const float th = font.getAscent();
			const float ty = halfH - th * 0.5f;

			g.fillPath (tick, tick.getTransformToScaleToFit (2.0f, ty, (float) (leftBorder - 4),
															 th, true));
		}

		g.drawFittedText (text,
						  leftBorder, 0,
						  width - (leftBorder + rightBorder), height,
						  Justification::centredLeft, 1);

		if (shortcutKeyText.isNotEmpty())
		{
			Font f2 (font);
			f2.setHeight (f2.getHeight() * 0.75f);
			f2.setHorizontalScale (0.95f);
			g.setFont (f2);

			g.drawText (shortcutKeyText,
						leftBorder,
						0,
						width - (leftBorder + rightBorder + 4),
						height,
						Justification::centredRight,
						true);
		}

		if (hasSubMenu)
		{
			const float arrowH = 0.6f * getPopupMenuFont().getAscent();
			const float x = width - height * 0.6f;

			Path p;
			p.addTriangle (x, halfH - arrowH * 0.5f,
						   x, halfH + arrowH * 0.5f,
						   x + arrowH * 0.6f, halfH);

			g.fillPath (p);
		}
	}
}

int LookAndFeel::getMenuWindowFlags()
{
	return ComponentPeer::windowHasDropShadow;
}

void LookAndFeel::drawMenuBarBackground (Graphics& g, int width, int height,
										 bool, MenuBarComponent& menuBar)
{
	const Colour baseColour (LookAndFeelHelpers::createBaseColour (menuBar.findColour (PopupMenu::backgroundColourId), false, false, false));

	if (menuBar.isEnabled())
	{
		drawShinyButtonShape (g,
							  -4.0f, 0.0f,
							  width + 8.0f, (float) height,
							  0.0f,
							  baseColour,
							  0.4f,
							  true, true, true, true);
	}
	else
	{
		g.fillAll (baseColour);
	}
}

const Font LookAndFeel::getMenuBarFont (MenuBarComponent& menuBar, int /*itemIndex*/, const String& /*itemText*/)
{
	return Font (menuBar.getHeight() * 0.7f);
}

int LookAndFeel::getMenuBarItemWidth (MenuBarComponent& menuBar, int itemIndex, const String& itemText)
{
	return getMenuBarFont (menuBar, itemIndex, itemText)
			.getStringWidth (itemText) + menuBar.getHeight();
}

void LookAndFeel::drawMenuBarItem (Graphics& g,
								   int width, int height,
								   int itemIndex,
								   const String& itemText,
								   bool isMouseOverItem,
								   bool isMenuOpen,
								   bool /*isMouseOverBar*/,
								   MenuBarComponent& menuBar)
{
	if (! menuBar.isEnabled())
	{
		g.setColour (menuBar.findColour (PopupMenu::textColourId)
							.withMultipliedAlpha (0.5f));
	}
	else if (isMenuOpen || isMouseOverItem)
	{
		g.fillAll (menuBar.findColour (PopupMenu::highlightedBackgroundColourId));
		g.setColour (menuBar.findColour (PopupMenu::highlightedTextColourId));
	}
	else
	{
		g.setColour (menuBar.findColour (PopupMenu::textColourId));
	}

	g.setFont (getMenuBarFont (menuBar, itemIndex, itemText));
	g.drawFittedText (itemText, 0, 0, width, height, Justification::centred, 1);
}

void LookAndFeel::fillTextEditorBackground (Graphics& g, int /*width*/, int /*height*/,
											TextEditor& textEditor)
{
	g.fillAll (textEditor.findColour (TextEditor::backgroundColourId));
}

void LookAndFeel::drawTextEditorOutline (Graphics& g, int width, int height, TextEditor& textEditor)
{
	if (textEditor.isEnabled())
	{
		if (textEditor.hasKeyboardFocus (true) && ! textEditor.isReadOnly())
		{
			const int border = 2;

			g.setColour (textEditor.findColour (TextEditor::focusedOutlineColourId));
			g.drawRect (0, 0, width, height, border);

			g.setOpacity (1.0f);
			const Colour shadowColour (textEditor.findColour (TextEditor::shadowColourId).withMultipliedAlpha (0.75f));
			g.drawBevel (0, 0, width, height + 2, border + 2, shadowColour, shadowColour);
		}
		else
		{
			g.setColour (textEditor.findColour (TextEditor::outlineColourId));
			g.drawRect (0, 0, width, height);

			g.setOpacity (1.0f);
			const Colour shadowColour (textEditor.findColour (TextEditor::shadowColourId));
			g.drawBevel (0, 0, width, height + 2, 3, shadowColour, shadowColour);
		}
	}
}

CaretComponent* LookAndFeel::createCaretComponent (Component* keyFocusOwner)
{
	return new CaretComponent (keyFocusOwner);
}

void LookAndFeel::drawComboBox (Graphics& g, int width, int height,
								const bool isButtonDown,
								int buttonX, int buttonY,
								int buttonW, int buttonH,
								ComboBox& box)
{
	g.fillAll (box.findColour (ComboBox::backgroundColourId));

	if (box.isEnabled() && box.hasKeyboardFocus (false))
	{
		g.setColour (box.findColour (TextButton::buttonColourId));
		g.drawRect (0, 0, width, height, 2);
	}
	else
	{
		g.setColour (box.findColour (ComboBox::outlineColourId));
		g.drawRect (0, 0, width, height);
	}

	const float outlineThickness = box.isEnabled() ? (isButtonDown ? 1.2f : 0.5f) : 0.3f;

	const Colour baseColour (LookAndFeelHelpers::createBaseColour (box.findColour (ComboBox::buttonColourId),
																   box.hasKeyboardFocus (true),
																   false, isButtonDown)
								.withMultipliedAlpha (box.isEnabled() ? 1.0f : 0.5f));

	drawGlassLozenge (g,
					  buttonX + outlineThickness, buttonY + outlineThickness,
					  buttonW - outlineThickness * 2.0f, buttonH - outlineThickness * 2.0f,
					  baseColour, outlineThickness, -1.0f,
					  true, true, true, true);

	if (box.isEnabled())
	{
		const float arrowX = 0.3f;
		const float arrowH = 0.2f;

		Path p;
		p.addTriangle (buttonX + buttonW * 0.5f,		buttonY + buttonH * (0.45f - arrowH),
					   buttonX + buttonW * (1.0f - arrowX), buttonY + buttonH * 0.45f,
					   buttonX + buttonW * arrowX,	  buttonY + buttonH * 0.45f);

		p.addTriangle (buttonX + buttonW * 0.5f,		buttonY + buttonH * (0.55f + arrowH),
					   buttonX + buttonW * (1.0f - arrowX), buttonY + buttonH * 0.55f,
					   buttonX + buttonW * arrowX,	  buttonY + buttonH * 0.55f);

		g.setColour (box.findColour (ComboBox::arrowColourId));
		g.fillPath (p);
	}
}

const Font LookAndFeel::getComboBoxFont (ComboBox& box)
{
	return Font (jmin (15.0f, box.getHeight() * 0.85f));
}

Label* LookAndFeel::createComboBoxTextBox (ComboBox&)
{
	return new Label (String::empty, String::empty);
}

void LookAndFeel::positionComboBoxText (ComboBox& box, Label& label)
{
	label.setBounds (1, 1,
					 box.getWidth() + 3 - box.getHeight(),
					 box.getHeight() - 2);

	label.setFont (getComboBoxFont (box));
}

void LookAndFeel::drawLabel (Graphics& g, Label& label)
{
	g.fillAll (label.findColour (Label::backgroundColourId));

	if (! label.isBeingEdited())
	{
		const float alpha = label.isEnabled() ? 1.0f : 0.5f;

		g.setColour (label.findColour (Label::textColourId).withMultipliedAlpha (alpha));
		g.setFont (label.getFont());
		g.drawFittedText (label.getText(),
						  label.getHorizontalBorderSize(),
						  label.getVerticalBorderSize(),
						  label.getWidth() - 2 * label.getHorizontalBorderSize(),
						  label.getHeight() - 2 * label.getVerticalBorderSize(),
						  label.getJustificationType(),
						  jmax (1, (int) (label.getHeight() / label.getFont().getHeight())),
						  label.getMinimumHorizontalScale());

		g.setColour (label.findColour (Label::outlineColourId).withMultipliedAlpha (alpha));
		g.drawRect (0, 0, label.getWidth(), label.getHeight());
	}
	else if (label.isEnabled())
	{
		g.setColour (label.findColour (Label::outlineColourId));
		g.drawRect (0, 0, label.getWidth(), label.getHeight());
	}
}

void LookAndFeel::drawLinearSliderBackground (Graphics& g,
											  int x, int y,
											  int width, int height,
											  float /*sliderPos*/,
											  float /*minSliderPos*/,
											  float /*maxSliderPos*/,
											  const Slider::SliderStyle /*style*/,
											  Slider& slider)
{
	const float sliderRadius = (float) (getSliderThumbRadius (slider) - 2);

	const Colour trackColour (slider.findColour (Slider::trackColourId));
	const Colour gradCol1 (trackColour.overlaidWith (Colours::black.withAlpha (slider.isEnabled() ? 0.25f : 0.13f)));
	const Colour gradCol2 (trackColour.overlaidWith (Colour (0x14000000)));
	Path indent;

	if (slider.isHorizontal())
	{
		const float iy = y + height * 0.5f - sliderRadius * 0.5f;
		const float ih = sliderRadius;

		g.setGradientFill (ColourGradient (gradCol1, 0.0f, iy,
										   gradCol2, 0.0f, iy + ih, false));

		indent.addRoundedRectangle (x - sliderRadius * 0.5f, iy,
									width + sliderRadius, ih,
									5.0f);
		g.fillPath (indent);
	}
	else
	{
		const float ix = x + width * 0.5f - sliderRadius * 0.5f;
		const float iw = sliderRadius;

		g.setGradientFill (ColourGradient (gradCol1, ix, 0.0f,
										   gradCol2, ix + iw, 0.0f, false));

		indent.addRoundedRectangle (ix, y - sliderRadius * 0.5f,
									iw, height + sliderRadius,
									5.0f);
		g.fillPath (indent);
	}

	g.setColour (Colour (0x4c000000));
	g.strokePath (indent, PathStrokeType (0.5f));
}

void LookAndFeel::drawLinearSliderThumb (Graphics& g,
										 int x, int y,
										 int width, int height,
										 float sliderPos,
										 float minSliderPos,
										 float maxSliderPos,
										 const Slider::SliderStyle style,
										 Slider& slider)
{
	const float sliderRadius = (float) (getSliderThumbRadius (slider) - 2);

	Colour knobColour (LookAndFeelHelpers::createBaseColour (slider.findColour (Slider::thumbColourId),
															 slider.hasKeyboardFocus (false) && slider.isEnabled(),
															 slider.isMouseOverOrDragging() && slider.isEnabled(),
															 slider.isMouseButtonDown() && slider.isEnabled()));

	const float outlineThickness = slider.isEnabled() ? 0.8f : 0.3f;

	if (style == Slider::LinearHorizontal || style == Slider::LinearVertical)
	{
		float kx, ky;

		if (style == Slider::LinearVertical)
		{
			kx = x + width * 0.5f;
			ky = sliderPos;
		}
		else
		{
			kx = sliderPos;
			ky = y + height * 0.5f;
		}

		drawGlassSphere (g,
						 kx - sliderRadius,
						 ky - sliderRadius,
						 sliderRadius * 2.0f,
						 knobColour, outlineThickness);
	}
	else
	{
		if (style == Slider::ThreeValueVertical)
		{
			drawGlassSphere (g, x + width * 0.5f - sliderRadius,
							 sliderPos - sliderRadius,
							 sliderRadius * 2.0f,
							 knobColour, outlineThickness);
		}
		else if (style == Slider::ThreeValueHorizontal)
		{
			drawGlassSphere (g,sliderPos - sliderRadius,
							 y + height * 0.5f - sliderRadius,
							 sliderRadius * 2.0f,
							 knobColour, outlineThickness);
		}

		if (style == Slider::TwoValueVertical || style == Slider::ThreeValueVertical)
		{
			const float sr = jmin (sliderRadius, width * 0.4f);

			drawGlassPointer (g, jmax (0.0f, x + width * 0.5f - sliderRadius * 2.0f),
							  minSliderPos - sliderRadius,
							  sliderRadius * 2.0f, knobColour, outlineThickness, 1);

			drawGlassPointer (g, jmin (x + width - sliderRadius * 2.0f, x + width * 0.5f), maxSliderPos - sr,
							  sliderRadius * 2.0f, knobColour, outlineThickness, 3);
		}
		else if (style == Slider::TwoValueHorizontal || style == Slider::ThreeValueHorizontal)
		{
			const float sr = jmin (sliderRadius, height * 0.4f);

			drawGlassPointer (g, minSliderPos - sr,
							  jmax (0.0f, y + height * 0.5f - sliderRadius * 2.0f),
							  sliderRadius * 2.0f, knobColour, outlineThickness, 2);

			drawGlassPointer (g, maxSliderPos - sliderRadius,
							  jmin (y + height - sliderRadius * 2.0f, y + height * 0.5f),
							  sliderRadius * 2.0f, knobColour, outlineThickness, 4);
		}
	}
}

void LookAndFeel::drawLinearSlider (Graphics& g,
									int x, int y,
									int width, int height,
									float sliderPos,
									float minSliderPos,
									float maxSliderPos,
									const Slider::SliderStyle style,
									Slider& slider)
{
	g.fillAll (slider.findColour (Slider::backgroundColourId));

	if (style == Slider::LinearBar)
	{
		const bool isMouseOver = slider.isMouseOverOrDragging() && slider.isEnabled();

		Colour baseColour (LookAndFeelHelpers::createBaseColour (slider.findColour (Slider::thumbColourId)
																	   .withMultipliedSaturation (slider.isEnabled() ? 1.0f : 0.5f),
																 false, isMouseOver,
																 isMouseOver || slider.isMouseButtonDown()));

		drawShinyButtonShape (g,
							  (float) x, (float) y, sliderPos - (float) x, (float) height, 0.0f,
							  baseColour,
							  slider.isEnabled() ? 0.9f : 0.3f,
							  true, true, true, true);
	}
	else
	{
		drawLinearSliderBackground (g, x, y, width, height, sliderPos, minSliderPos, maxSliderPos, style, slider);
		drawLinearSliderThumb (g, x, y, width, height, sliderPos, minSliderPos, maxSliderPos, style, slider);
	}
}

int LookAndFeel::getSliderThumbRadius (Slider& slider)
{
	return jmin (7,
				 slider.getHeight() / 2,
				 slider.getWidth() / 2) + 2;
}

void LookAndFeel::drawRotarySlider (Graphics& g,
									int x, int y,
									int width, int height,
									float sliderPos,
									const float rotaryStartAngle,
									const float rotaryEndAngle,
									Slider& slider)
{
	const float radius = jmin (width / 2, height / 2) - 2.0f;
	const float centreX = x + width * 0.5f;
	const float centreY = y + height * 0.5f;
	const float rx = centreX - radius;
	const float ry = centreY - radius;
	const float rw = radius * 2.0f;
	const float angle = rotaryStartAngle + sliderPos * (rotaryEndAngle - rotaryStartAngle);
	const bool isMouseOver = slider.isMouseOverOrDragging() && slider.isEnabled();

	if (radius > 12.0f)
	{
		if (slider.isEnabled())
			g.setColour (slider.findColour (Slider::rotarySliderFillColourId).withAlpha (isMouseOver ? 1.0f : 0.7f));
		else
			g.setColour (Colour (0x80808080));

		const float thickness = 0.7f;

		{
			Path filledArc;
			filledArc.addPieSegment (rx, ry, rw, rw,
									rotaryStartAngle,
									angle,
									thickness);

			g.fillPath (filledArc);
		}

		if (thickness > 0)
		{
			const float innerRadius = radius * 0.2f;
			Path p;
			p.addTriangle (-innerRadius, 0.0f,
						   0.0f, -radius * thickness * 1.1f,
						   innerRadius, 0.0f);

			p.addEllipse (-innerRadius, -innerRadius, innerRadius * 2.0f, innerRadius * 2.0f);

			g.fillPath (p, AffineTransform::rotation (angle).translated (centreX, centreY));
		}

		if (slider.isEnabled())
			g.setColour (slider.findColour (Slider::rotarySliderOutlineColourId));
		else
			g.setColour (Colour (0x80808080));

		Path outlineArc;
		outlineArc.addPieSegment (rx, ry, rw, rw, rotaryStartAngle, rotaryEndAngle, thickness);
		outlineArc.closeSubPath();

		g.strokePath (outlineArc, PathStrokeType (slider.isEnabled() ? (isMouseOver ? 2.0f : 1.2f) : 0.3f));
	}
	else
	{
		if (slider.isEnabled())
			g.setColour (slider.findColour (Slider::rotarySliderFillColourId).withAlpha (isMouseOver ? 1.0f : 0.7f));
		else
			g.setColour (Colour (0x80808080));

		Path p;
		p.addEllipse (-0.4f * rw, -0.4f * rw, rw * 0.8f, rw * 0.8f);
		PathStrokeType (rw * 0.1f).createStrokedPath (p, p);

		p.addLineSegment (Line<float> (0.0f, 0.0f, 0.0f, -radius), rw * 0.2f);

		g.fillPath (p, AffineTransform::rotation (angle).translated (centreX, centreY));
	}
}

Button* LookAndFeel::createSliderButton (const bool isIncrement)
{
	return new TextButton (isIncrement ? "+" : "-", String::empty);
}

class SliderLabelComp : public Label
{
public:
	SliderLabelComp() : Label (String::empty, String::empty) {}

	void mouseWheelMove (const MouseEvent&, float, float) {}
};

Label* LookAndFeel::createSliderTextBox (Slider& slider)
{
	Label* const l = new SliderLabelComp();

	l->setJustificationType (Justification::centred);

	l->setColour (Label::textColourId, slider.findColour (Slider::textBoxTextColourId));

	l->setColour (Label::backgroundColourId,
				  (slider.getSliderStyle() == Slider::LinearBar) ? Colours::transparentBlack
																 : slider.findColour (Slider::textBoxBackgroundColourId));
	l->setColour (Label::outlineColourId, slider.findColour (Slider::textBoxOutlineColourId));

	l->setColour (TextEditor::textColourId, slider.findColour (Slider::textBoxTextColourId));

	l->setColour (TextEditor::backgroundColourId,
				  slider.findColour (Slider::textBoxBackgroundColourId)
						.withAlpha (slider.getSliderStyle() == Slider::LinearBar ? 0.7f : 1.0f));

	l->setColour (TextEditor::outlineColourId, slider.findColour (Slider::textBoxOutlineColourId));

	return l;
}

ImageEffectFilter* LookAndFeel::getSliderEffect()
{
	return nullptr;
}

void LookAndFeel::getTooltipSize (const String& tipText, int& width, int& height)
{
	const TextLayout tl (LookAndFeelHelpers::layoutTooltipText (tipText));

	width = tl.getWidth() + 14;
	height = tl.getHeight() + 6;
}

void LookAndFeel::drawTooltip (Graphics& g, const String& text, int width, int height)
{
	g.fillAll (findColour (TooltipWindow::backgroundColourId));

   #if ! JUCE_MAC // The mac windows already have a non-optional 1 pix outline, so don't double it here..
	g.setColour (findColour (TooltipWindow::outlineColourId));
	g.drawRect (0, 0, width, height, 1);
   #endif

	const TextLayout tl (LookAndFeelHelpers::layoutTooltipText (text));

	g.setColour (findColour (TooltipWindow::textColourId));
	tl.drawWithin (g, 0, 0, width, height, Justification::centred);
}

Button* LookAndFeel::createFilenameComponentBrowseButton (const String& text)
{
	return new TextButton (text, TRANS("click to browse for a different file"));
}

void LookAndFeel::layoutFilenameComponent (FilenameComponent& filenameComp,
										   ComboBox* filenameBox,
										   Button* browseButton)
{
	browseButton->setSize (80, filenameComp.getHeight());

	TextButton* const tb = dynamic_cast <TextButton*> (browseButton);

	if (tb != nullptr)
		tb->changeWidthToFitText();

	browseButton->setTopRightPosition (filenameComp.getWidth(), 0);

	filenameBox->setBounds (0, 0, browseButton->getX(), filenameComp.getHeight());
}

void LookAndFeel::drawImageButton (Graphics& g, Image* image,
								   int imageX, int imageY, int imageW, int imageH,
								   const Colour& overlayColour,
								   float imageOpacity,
								   ImageButton& button)
{
	if (! button.isEnabled())
		imageOpacity *= 0.3f;

	if (! overlayColour.isOpaque())
	{
		g.setOpacity (imageOpacity);

		g.drawImage (*image, imageX, imageY, imageW, imageH,
					 0, 0, image->getWidth(), image->getHeight(), false);
	}

	if (! overlayColour.isTransparent())
	{
		g.setColour (overlayColour);

		g.drawImage (*image, imageX, imageY, imageW, imageH,
					 0, 0, image->getWidth(), image->getHeight(), true);
	}
}

void LookAndFeel::drawCornerResizer (Graphics& g,
									 int w, int h,
									 bool /*isMouseOver*/,
									 bool /*isMouseDragging*/)
{
	const float lineThickness = jmin (w, h) * 0.075f;

	for (float i = 0.0f; i < 1.0f; i += 0.3f)
	{
		g.setColour (Colours::lightgrey);

		g.drawLine (w * i,
					h + 1.0f,
					w + 1.0f,
					h * i,
					lineThickness);

		g.setColour (Colours::darkgrey);

		g.drawLine (w * i + lineThickness,
					h + 1.0f,
					w + 1.0f,
					h * i + lineThickness,
					lineThickness);
	}
}

void LookAndFeel::drawResizableFrame (Graphics& g, int w, int h, const BorderSize<int>& border)
{
	if (! border.isEmpty())
	{
		const Rectangle<int> fullSize (0, 0, w, h);
		const Rectangle<int> centreArea (border.subtractedFrom (fullSize));

		g.saveState();

		g.excludeClipRegion (centreArea);

		g.setColour (Colour (0x50000000));
		g.drawRect (fullSize);

		g.setColour (Colour (0x19000000));
		g.drawRect (centreArea.expanded (1, 1));

		g.restoreState();
	}
}

void LookAndFeel::fillResizableWindowBackground (Graphics& g, int /*w*/, int /*h*/,
												 const BorderSize<int>& /*border*/, ResizableWindow& window)
{
   g.fillAll (window.getBackgroundColour());
}

void LookAndFeel::drawResizableWindowBorder (Graphics&, int /*w*/, int /*h*/,
											 const BorderSize<int>& /*border*/, ResizableWindow&)
{
}

void LookAndFeel::drawDocumentWindowTitleBar (DocumentWindow& window,
											  Graphics& g, int w, int h,
											  int titleSpaceX, int titleSpaceW,
											  const Image* icon,
											  bool drawTitleTextOnLeft)
{
	const bool isActive = window.isActiveWindow();

	g.setGradientFill (ColourGradient (window.getBackgroundColour(),
									   0.0f, 0.0f,
									   window.getBackgroundColour().contrasting (isActive ? 0.15f : 0.05f),
									   0.0f, (float) h, false));
	g.fillAll();

	Font font (h * 0.65f, Font::bold);
	g.setFont (font);

	int textW = font.getStringWidth (window.getName());
	int iconW = 0;
	int iconH = 0;

	if (icon != nullptr)
	{
		iconH = (int) font.getHeight();
		iconW = icon->getWidth() * iconH / icon->getHeight() + 4;
	}

	textW = jmin (titleSpaceW, textW + iconW);
	int textX = drawTitleTextOnLeft ? titleSpaceX
									: jmax (titleSpaceX, (w - textW) / 2);

	if (textX + textW > titleSpaceX + titleSpaceW)
		textX = titleSpaceX + titleSpaceW - textW;

	if (icon != nullptr)
	{
		g.setOpacity (isActive ? 1.0f : 0.6f);
		g.drawImageWithin (*icon, textX, (h - iconH) / 2, iconW, iconH,
						   RectanglePlacement::centred, false);
		textX += iconW;
		textW -= iconW;
	}

	if (window.isColourSpecified (DocumentWindow::textColourId) || isColourSpecified (DocumentWindow::textColourId))
		g.setColour (findColour (DocumentWindow::textColourId));
	else
		g.setColour (window.getBackgroundColour().contrasting (isActive ? 0.7f : 0.4f));

	g.drawText (window.getName(), textX, 0, textW, h, Justification::centredLeft, true);
}

class GlassWindowButton   : public Button
{
public:

	GlassWindowButton (const String& name, const Colour& col,
					   const Path& normalShape_,
					   const Path& toggledShape_) noexcept
		: Button (name),
		  colour (col),
		  normalShape (normalShape_),
		  toggledShape (toggledShape_)
	{
	}

	~GlassWindowButton()
	{
	}

	void paintButton (Graphics& g, bool isMouseOverButton, bool isButtonDown)
	{
		float alpha = isMouseOverButton ? (isButtonDown ? 1.0f : 0.8f) : 0.55f;

		if (! isEnabled())
			alpha *= 0.5f;

		float x = 0, y = 0, diam;

		if (getWidth() < getHeight())
		{
			diam = (float) getWidth();
			y = (getHeight() - getWidth()) * 0.5f;
		}
		else
		{
			diam = (float) getHeight();
			y = (getWidth() - getHeight()) * 0.5f;
		}

		x += diam * 0.05f;
		y += diam * 0.05f;
		diam *= 0.9f;

		g.setGradientFill (ColourGradient (Colour::greyLevel (0.9f).withAlpha (alpha), 0, y + diam,
										   Colour::greyLevel (0.6f).withAlpha (alpha), 0, y, false));
		g.fillEllipse (x, y, diam, diam);

		x += 2.0f;
		y += 2.0f;
		diam -= 4.0f;

		LookAndFeel::drawGlassSphere (g, x, y, diam, colour.withAlpha (alpha), 1.0f);

		Path& p = getToggleState() ? toggledShape : normalShape;

		const AffineTransform t (p.getTransformToScaleToFit (x + diam * 0.3f, y + diam * 0.3f,
															 diam * 0.4f, diam * 0.4f, true));

		g.setColour (Colours::black.withAlpha (alpha * 0.6f));
		g.fillPath (p, t);
	}

private:
	Colour colour;
	Path normalShape, toggledShape;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (GlassWindowButton);
};

Button* LookAndFeel::createDocumentWindowButton (int buttonType)
{
	Path shape;
	const float crossThickness = 0.25f;

	if (buttonType == DocumentWindow::closeButton)
	{
		shape.addLineSegment (Line<float> (0.0f, 0.0f, 1.0f, 1.0f), crossThickness * 1.4f);
		shape.addLineSegment (Line<float> (1.0f, 0.0f, 0.0f, 1.0f), crossThickness * 1.4f);

		return new GlassWindowButton ("close", Colour (0xffdd1100), shape, shape);
	}
	else if (buttonType == DocumentWindow::minimiseButton)
	{
		shape.addLineSegment (Line<float> (0.0f, 0.5f, 1.0f, 0.5f), crossThickness);

		return new GlassWindowButton ("minimise", Colour (0xffaa8811), shape, shape);
	}
	else if (buttonType == DocumentWindow::maximiseButton)
	{
		shape.addLineSegment (Line<float> (0.5f, 0.0f, 0.5f, 1.0f), crossThickness);
		shape.addLineSegment (Line<float> (0.0f, 0.5f, 1.0f, 0.5f), crossThickness);

		Path fullscreenShape;
		fullscreenShape.startNewSubPath (45.0f, 100.0f);
		fullscreenShape.lineTo (0.0f, 100.0f);
		fullscreenShape.lineTo (0.0f, 0.0f);
		fullscreenShape.lineTo (100.0f, 0.0f);
		fullscreenShape.lineTo (100.0f, 45.0f);
		fullscreenShape.addRectangle (45.0f, 45.0f, 100.0f, 100.0f);
		PathStrokeType (30.0f).createStrokedPath (fullscreenShape, fullscreenShape);

		return new GlassWindowButton ("maximise", Colour (0xff119911), shape, fullscreenShape);
	}

	jassertfalse;
	return nullptr;
}

void LookAndFeel::positionDocumentWindowButtons (DocumentWindow&,
												 int titleBarX,
												 int titleBarY,
												 int titleBarW,
												 int titleBarH,
												 Button* minimiseButton,
												 Button* maximiseButton,
												 Button* closeButton,
												 bool positionTitleBarButtonsOnLeft)
{
	const int buttonW = titleBarH - titleBarH / 8;

	int x = positionTitleBarButtonsOnLeft ? titleBarX + 4
										  : titleBarX + titleBarW - buttonW - buttonW / 4;

	if (closeButton != nullptr)
	{
		closeButton->setBounds (x, titleBarY, buttonW, titleBarH);
		x += positionTitleBarButtonsOnLeft ? buttonW : -(buttonW + buttonW / 4);
	}

	if (positionTitleBarButtonsOnLeft)
		std::swap (minimiseButton, maximiseButton);

	if (maximiseButton != nullptr)
	{
		maximiseButton->setBounds (x, titleBarY, buttonW, titleBarH);
		x += positionTitleBarButtonsOnLeft ? buttonW : -buttonW;
	}

	if (minimiseButton != nullptr)
		minimiseButton->setBounds (x, titleBarY, buttonW, titleBarH);
}

int LookAndFeel::getDefaultMenuBarHeight()
{
	return 24;
}

DropShadower* LookAndFeel::createDropShadowerForComponent (Component*)
{
	return new DropShadower (0.4f, 1, 5, 10);
}

void LookAndFeel::drawStretchableLayoutResizerBar (Graphics& g,
												   int w, int h,
												   bool /*isVerticalBar*/,
												   bool isMouseOver,
												   bool isMouseDragging)
{
	float alpha = 0.5f;

	if (isMouseOver || isMouseDragging)
	{
		g.fillAll (Colour (0x190000ff));
		alpha = 1.0f;
	}

	const float cx = w * 0.5f;
	const float cy = h * 0.5f;
	const float cr = jmin (w, h) * 0.4f;

	g.setGradientFill (ColourGradient (Colours::white.withAlpha (alpha), cx + cr * 0.1f, cy + cr,
									   Colours::black.withAlpha (alpha), cx, cy - cr * 4.0f,
									   true));

	g.fillEllipse (cx - cr, cy - cr, cr * 2.0f, cr * 2.0f);
}

void LookAndFeel::drawGroupComponentOutline (Graphics& g, int width, int height,
											 const String& text,
											 const Justification& position,
											 GroupComponent& group)
{
	const float textH = 15.0f;
	const float indent = 3.0f;
	const float textEdgeGap = 4.0f;
	float cs = 5.0f;

	Font f (textH);

	Path p;
	float x = indent;
	float y = f.getAscent() - 3.0f;
	float w = jmax (0.0f, width - x * 2.0f);
	float h = jmax (0.0f, height - y  - indent);
	cs = jmin (cs, w * 0.5f, h * 0.5f);
	const float cs2 = 2.0f * cs;

	float textW = text.isEmpty() ? 0 : jlimit (0.0f, jmax (0.0f, w - cs2 - textEdgeGap * 2), f.getStringWidth (text) + textEdgeGap * 2.0f);
	float textX = cs + textEdgeGap;

	if (position.testFlags (Justification::horizontallyCentred))
		textX = cs + (w - cs2 - textW) * 0.5f;
	else if (position.testFlags (Justification::right))
		textX = w - cs - textW - textEdgeGap;

	p.startNewSubPath (x + textX + textW, y);
	p.lineTo (x + w - cs, y);

	p.addArc (x + w - cs2, y, cs2, cs2, 0, float_Pi * 0.5f);
	p.lineTo (x + w, y + h - cs);

	p.addArc (x + w - cs2, y + h - cs2, cs2, cs2, float_Pi * 0.5f, float_Pi);
	p.lineTo (x + cs, y + h);

	p.addArc (x, y + h - cs2, cs2, cs2, float_Pi, float_Pi * 1.5f);
	p.lineTo (x, y + cs);

	p.addArc (x, y, cs2, cs2, float_Pi * 1.5f, float_Pi * 2.0f);
	p.lineTo (x + textX, y);

	const float alpha = group.isEnabled() ? 1.0f : 0.5f;

	g.setColour (group.findColour (GroupComponent::outlineColourId)
					.withMultipliedAlpha (alpha));

	g.strokePath (p, PathStrokeType (2.0f));

	g.setColour (group.findColour (GroupComponent::textColourId)
					.withMultipliedAlpha (alpha));
	g.setFont (f);
	g.drawText (text,
				roundToInt (x + textX), 0,
				roundToInt (textW),
				roundToInt (textH),
				Justification::centred, true);
}

int LookAndFeel::getTabButtonOverlap (int tabDepth)
{
	return 1 + tabDepth / 3;
}

int LookAndFeel::getTabButtonSpaceAroundImage()
{
	return 4;
}

void LookAndFeel::createTabButtonShape (Path& p,
										int width, int height,
										int /*tabIndex*/,
										const String& /*text*/,
										Button& /*button*/,
										TabbedButtonBar::Orientation orientation,
										const bool /*isMouseOver*/,
										const bool /*isMouseDown*/,
										const bool /*isFrontTab*/)
{
	const float w = (float) width;
	const float h = (float) height;

	float length = w;
	float depth = h;

	if (orientation == TabbedButtonBar::TabsAtLeft
		 || orientation == TabbedButtonBar::TabsAtRight)
	{
		std::swap (length, depth);
	}

	const float indent = (float) getTabButtonOverlap ((int) depth);
	const float overhang = 4.0f;

	if (orientation == TabbedButtonBar::TabsAtLeft)
	{
		p.startNewSubPath (w, 0.0f);
		p.lineTo (0.0f, indent);
		p.lineTo (0.0f, h - indent);
		p.lineTo (w, h);
		p.lineTo (w + overhang, h + overhang);
		p.lineTo (w + overhang, -overhang);
	}
	else if (orientation == TabbedButtonBar::TabsAtRight)
	{
		p.startNewSubPath (0.0f, 0.0f);
		p.lineTo (w, indent);
		p.lineTo (w, h - indent);
		p.lineTo (0.0f, h);
		p.lineTo (-overhang, h + overhang);
		p.lineTo (-overhang, -overhang);
	}
	else if (orientation == TabbedButtonBar::TabsAtBottom)
	{
		p.startNewSubPath (0.0f, 0.0f);
		p.lineTo (indent, h);
		p.lineTo (w - indent, h);
		p.lineTo (w, 0.0f);
		p.lineTo (w + overhang, -overhang);
		p.lineTo (-overhang, -overhang);
	}
	else
	{
		p.startNewSubPath (0.0f, h);
		p.lineTo (indent, 0.0f);
		p.lineTo (w - indent, 0.0f);
		p.lineTo (w, h);
		p.lineTo (w + overhang, h + overhang);
		p.lineTo (-overhang, h + overhang);
	}

	p.closeSubPath();

	p = p.createPathWithRoundedCorners (3.0f);
}

void LookAndFeel::fillTabButtonShape (Graphics& g,
									  const Path& path,
									  const Colour& preferredColour,
									  int /*tabIndex*/,
									  const String& /*text*/,
									  Button& button,
									  TabbedButtonBar::Orientation /*orientation*/,
									  const bool /*isMouseOver*/,
									  const bool /*isMouseDown*/,
									  const bool isFrontTab)
{
	g.setColour (isFrontTab ? preferredColour
							: preferredColour.withMultipliedAlpha (0.9f));

	g.fillPath (path);

	g.setColour (button.findColour (isFrontTab ? TabbedButtonBar::frontOutlineColourId
											   : TabbedButtonBar::tabOutlineColourId, false)
					.withMultipliedAlpha (button.isEnabled() ? 1.0f : 0.5f));

	g.strokePath (path, PathStrokeType (isFrontTab ? 1.0f : 0.5f));
}

void LookAndFeel::drawTabButtonText (Graphics& g,
									 int x, int y, int w, int h,
									 const Colour& preferredBackgroundColour,
									 int /*tabIndex*/,
									 const String& text,
									 Button& button,
									 TabbedButtonBar::Orientation orientation,
									 const bool isMouseOver,
									 const bool isMouseDown,
									 const bool isFrontTab)
{
	int length = w;
	int depth = h;

	if (orientation == TabbedButtonBar::TabsAtLeft
		 || orientation == TabbedButtonBar::TabsAtRight)
	{
		std::swap (length, depth);
	}

	Font font (depth * 0.6f);
	font.setUnderline (button.hasKeyboardFocus (false));

	GlyphArrangement textLayout;
	textLayout.addFittedText (font, text.trim(),
							  0.0f, 0.0f, (float) length, (float) depth,
							  Justification::centred,
							  jmax (1, depth / 12));

	AffineTransform transform;

	if (orientation == TabbedButtonBar::TabsAtLeft)
	{
		transform = transform.rotated (float_Pi * -0.5f)
							 .translated ((float) x, (float) (y + h));
	}
	else if (orientation  == TabbedButtonBar::TabsAtRight)
	{
		transform = transform.rotated (float_Pi * 0.5f)
							 .translated ((float) (x + w), (float) y);
	}
	else
	{
		transform = transform.translated ((float) x, (float) y);
	}

	if (isFrontTab && (button.isColourSpecified (TabbedButtonBar::frontTextColourId) || isColourSpecified (TabbedButtonBar::frontTextColourId)))
		g.setColour (findColour (TabbedButtonBar::frontTextColourId));
	else if (button.isColourSpecified (TabbedButtonBar::tabTextColourId) || isColourSpecified (TabbedButtonBar::tabTextColourId))
		g.setColour (findColour (TabbedButtonBar::tabTextColourId));
	else
		g.setColour (preferredBackgroundColour.contrasting());

	if (! (isMouseOver || isMouseDown))
		g.setOpacity (0.8f);

	if (! button.isEnabled())
		g.setOpacity (0.3f);

	textLayout.draw (g, transform);
}

int LookAndFeel::getTabButtonBestWidth (int /*tabIndex*/,
										const String& text,
										int tabDepth,
										Button&)
{
	Font f (tabDepth * 0.6f);
	return f.getStringWidth (text.trim()) + getTabButtonOverlap (tabDepth) * 2;
}

void LookAndFeel::drawTabButton (Graphics& g,
								 int w, int h,
								 const Colour& preferredColour,
								 int tabIndex,
								 const String& text,
								 Button& button,
								 TabbedButtonBar::Orientation orientation,
								 const bool isMouseOver,
								 const bool isMouseDown,
								 const bool isFrontTab)
{
	int length = w;
	int depth = h;

	if (orientation == TabbedButtonBar::TabsAtLeft
			|| orientation == TabbedButtonBar::TabsAtRight)
	{
		std::swap (length, depth);
	}

	Path tabShape;

	createTabButtonShape (tabShape, w, h,
						  tabIndex, text, button, orientation,
						  isMouseOver, isMouseDown, isFrontTab);

	fillTabButtonShape (g, tabShape, preferredColour,
						tabIndex, text, button, orientation,
						isMouseOver, isMouseDown, isFrontTab);

	const int indent = getTabButtonOverlap (depth);
	int x = 0, y = 0;

	if (orientation == TabbedButtonBar::TabsAtLeft
		 || orientation == TabbedButtonBar::TabsAtRight)
	{
		y += indent;
		h -= indent * 2;
	}
	else
	{
		x += indent;
		w -= indent * 2;
	}

	drawTabButtonText (g, x, y, w, h, preferredColour,
					   tabIndex, text, button, orientation,
					   isMouseOver, isMouseDown, isFrontTab);
}

void LookAndFeel::drawTabAreaBehindFrontButton (Graphics& g,
												int w, int h,
												TabbedButtonBar& tabBar,
												TabbedButtonBar::Orientation orientation)
{
	const float shadowSize = 0.2f;

	float x1 = 0.0f, y1 = 0.0f, x2 = 0.0f, y2 = 0.0f;
	Rectangle<int> shadowRect;

	if (orientation == TabbedButtonBar::TabsAtLeft)
	{
		x1 = (float) w;
		x2 = w * (1.0f - shadowSize);
		shadowRect.setBounds ((int) x2, 0, w - (int) x2, h);
	}
	else if (orientation == TabbedButtonBar::TabsAtRight)
	{
		x2 = w * shadowSize;
		shadowRect.setBounds (0, 0, (int) x2, h);
	}
	else if (orientation == TabbedButtonBar::TabsAtBottom)
	{
		y2 = h * shadowSize;
		shadowRect.setBounds (0, 0, w, (int) y2);
	}
	else
	{
		y1 = (float) h;
		y2 = h * (1.0f - shadowSize);
		shadowRect.setBounds (0, (int) y2, w, h - (int) y2);
	}

	g.setGradientFill (ColourGradient (Colours::black.withAlpha (tabBar.isEnabled() ? 0.3f : 0.15f), x1, y1,
									   Colours::transparentBlack, x2, y2, false));

	shadowRect.expand (2, 2);
	g.fillRect (shadowRect);

	g.setColour (Colour (0x80000000));

	if (orientation == TabbedButtonBar::TabsAtLeft)
	{
		g.fillRect (w - 1, 0, 1, h);
	}
	else if (orientation == TabbedButtonBar::TabsAtRight)
	{
		g.fillRect (0, 0, 1, h);
	}
	else if (orientation == TabbedButtonBar::TabsAtBottom)
	{
		g.fillRect (0, 0, w, 1);
	}
	else
	{
		g.fillRect (0, h - 1, w, 1);
	}
}

Button* LookAndFeel::createTabBarExtrasButton()
{
	const float thickness = 7.0f;
	const float indent = 22.0f;

	Path p;
	p.addEllipse (-10.0f, -10.0f, 120.0f, 120.0f);

	DrawablePath ellipse;
	ellipse.setPath (p);
	ellipse.setFill (Colour (0x99ffffff));

	p.clear();
	p.addEllipse (0.0f, 0.0f, 100.0f, 100.0f);
	p.addRectangle (indent, 50.0f - thickness, 100.0f - indent * 2.0f, thickness * 2.0f);
	p.addRectangle (50.0f - thickness, indent, thickness * 2.0f, 50.0f - indent - thickness);
	p.addRectangle (50.0f - thickness, 50.0f + thickness, thickness * 2.0f, 50.0f - indent - thickness);
	p.setUsingNonZeroWinding (false);

	DrawablePath dp;
	dp.setPath (p);
	dp.setFill (Colour (0x59000000));

	DrawableComposite normalImage;
	normalImage.addAndMakeVisible (ellipse.createCopy());
	normalImage.addAndMakeVisible (dp.createCopy());

	dp.setFill (Colour (0xcc000000));

	DrawableComposite overImage;
	overImage.addAndMakeVisible (ellipse.createCopy());
	overImage.addAndMakeVisible (dp.createCopy());

	DrawableButton* db = new DrawableButton ("tabs", DrawableButton::ImageFitted);
	db->setImages (&normalImage, &overImage, nullptr);
	return db;
}

void LookAndFeel::drawTableHeaderBackground (Graphics& g, TableHeaderComponent& header)
{
	g.fillAll (Colours::white);

	const int w = header.getWidth();
	const int h = header.getHeight();

	g.setGradientFill (ColourGradient (Colour (0xffe8ebf9), 0.0f, h * 0.5f,
									   Colour (0xfff6f8f9), 0.0f, h - 1.0f,
									   false));
	g.fillRect (0, h / 2, w, h);

	g.setColour (Colour (0x33000000));
	g.fillRect (0, h - 1, w, 1);

	for (int i = header.getNumColumns (true); --i >= 0;)
		g.fillRect (header.getColumnPosition (i).getRight() - 1, 0, 1, h - 1);
}

void LookAndFeel::drawTableHeaderColumn (Graphics& g, const String& columnName, int /*columnId*/,
										 int width, int height,
										 bool isMouseOver, bool isMouseDown,
										 int columnFlags)
{
	if (isMouseDown)
		g.fillAll (Colour (0x8899aadd));
	else if (isMouseOver)
		g.fillAll (Colour (0x5599aadd));

	int rightOfText = width - 4;

	if ((columnFlags & (TableHeaderComponent::sortedForwards | TableHeaderComponent::sortedBackwards)) != 0)
	{
		const float top = height * ((columnFlags & TableHeaderComponent::sortedForwards) != 0 ? 0.35f : (1.0f - 0.35f));
		const float bottom = height - top;

		const float w = height * 0.5f;
		const float x = rightOfText - (w * 1.25f);
		rightOfText = (int) x;

		Path sortArrow;
		sortArrow.addTriangle (x, bottom, x + w * 0.5f, top, x + w, bottom);

		g.setColour (Colour (0x99000000));
		g.fillPath (sortArrow);
	}

	g.setColour (Colours::black);
	g.setFont (height * 0.5f, Font::bold);
	const int textX = 4;
	g.drawFittedText (columnName, textX, 0, rightOfText - textX, height, Justification::centredLeft, 1);
}

void LookAndFeel::paintToolbarBackground (Graphics& g, int w, int h, Toolbar& toolbar)
{
	const Colour background (toolbar.findColour (Toolbar::backgroundColourId));

	g.setGradientFill (ColourGradient (background, 0.0f, 0.0f,
									   background.darker (0.1f),
									   toolbar.isVertical() ? w - 1.0f : 0.0f,
									   toolbar.isVertical() ? 0.0f : h - 1.0f,
									   false));
	g.fillAll();
}

Button* LookAndFeel::createToolbarMissingItemsButton (Toolbar& /*toolbar*/)
{
	return createTabBarExtrasButton();
}

void LookAndFeel::paintToolbarButtonBackground (Graphics& g, int /*width*/, int /*height*/,
												bool isMouseOver, bool isMouseDown,
												ToolbarItemComponent& component)
{
	if (isMouseDown)
		g.fillAll (component.findColour (Toolbar::buttonMouseDownBackgroundColourId, true));
	else if (isMouseOver)
		g.fillAll (component.findColour (Toolbar::buttonMouseOverBackgroundColourId, true));
}

void LookAndFeel::paintToolbarButtonLabel (Graphics& g, int x, int y, int width, int height,
										   const String& text, ToolbarItemComponent& component)
{
	g.setColour (component.findColour (Toolbar::labelTextColourId, true)
					.withAlpha (component.isEnabled() ? 1.0f : 0.25f));

	const float fontHeight = jmin (14.0f, height * 0.85f);
	g.setFont (fontHeight);

	g.drawFittedText (text,
					  x, y, width, height,
					  Justification::centred,
					  jmax (1, height / (int) fontHeight));
}

void LookAndFeel::drawPropertyPanelSectionHeader (Graphics& g, const String& name,
												  bool isOpen, int width, int height)
{
	const int buttonSize = (height * 3) / 4;
	const int buttonIndent = (height - buttonSize) / 2;

	drawTreeviewPlusMinusBox (g, buttonIndent, buttonIndent, buttonSize, buttonSize, ! isOpen, false);

	const int textX = buttonIndent * 2 + buttonSize + 2;

	g.setColour (Colours::black);
	g.setFont (height * 0.7f, Font::bold);
	g.drawText (name, textX, 0, width - textX - 4, height, Justification::centredLeft, true);
}

void LookAndFeel::drawPropertyComponentBackground (Graphics& g, int width, int height,
												   PropertyComponent&)
{
	g.setColour (Colour (0x66ffffff));
	g.fillRect (0, 0, width, height - 1);
}

void LookAndFeel::drawPropertyComponentLabel (Graphics& g, int, int height,
											  PropertyComponent& component)
{
	g.setColour (Colours::black);

	if (! component.isEnabled())
		g.setOpacity (0.6f);

	g.setFont (jmin (height, 24) * 0.65f);

	const Rectangle<int> r (getPropertyComponentContentPosition (component));

	g.drawFittedText (component.getName(),
					  3, r.getY(), r.getX() - 5, r.getHeight(),
					  Justification::centredLeft, 2);
}

const Rectangle<int> LookAndFeel::getPropertyComponentContentPosition (PropertyComponent& component)
{
	return Rectangle<int> (component.getWidth() / 3, 1,
						   component.getWidth() - component.getWidth() / 3 - 1, component.getHeight() - 3);
}

void LookAndFeel::drawCallOutBoxBackground (CallOutBox& box, Graphics& g, const Path& path)
{
	Image content (Image::ARGB, box.getWidth(), box.getHeight(), true);

	{
		Graphics g2 (content);

		g2.setColour (Colour::greyLevel (0.23f).withAlpha (0.9f));
		g2.fillPath (path);

		g2.setColour (Colours::white.withAlpha (0.8f));
		g2.strokePath (path, PathStrokeType (2.0f));
	}

	DropShadowEffect shadow;
	shadow.setShadowProperties (5.0f, 0.4f, 0, 2);
	shadow.applyEffect (content, g, 1.0f);
}

void LookAndFeel::createFileChooserHeaderText (const String& title,
											   const String& instructions,
											   GlyphArrangement& text,
											   int width)
{
	text.clear();

	text.addJustifiedText (Font (17.0f, Font::bold), title,
						   8.0f, 22.0f, width - 16.0f,
						   Justification::centred);

	text.addJustifiedText (Font (14.0f), instructions,
						   8.0f, 24.0f + 16.0f, width - 16.0f,
						   Justification::centred);
}

void LookAndFeel::drawFileBrowserRow (Graphics& g, int width, int height,
									  const String& filename, Image* icon,
									  const String& fileSizeDescription,
									  const String& fileTimeDescription,
									  const bool isDirectory,
									  const bool isItemSelected,
									  const int /*itemIndex*/,
									  DirectoryContentsDisplayComponent&)
{
	if (isItemSelected)
		g.fillAll (findColour (DirectoryContentsDisplayComponent::highlightColourId));

	const int x = 32;
	g.setColour (Colours::black);

	if (icon != nullptr && icon->isValid())
	{
		g.drawImageWithin (*icon, 2, 2, x - 4, height - 4,
						   RectanglePlacement::centred | RectanglePlacement::onlyReduceInSize,
						   false);
	}
	else
	{
		const Drawable* d = isDirectory ? getDefaultFolderImage()
										: getDefaultDocumentFileImage();

		if (d != nullptr)
			d->drawWithin (g, Rectangle<float> (2.0f, 2.0f, x - 4.0f, height - 4.0f),
						   RectanglePlacement::centred | RectanglePlacement::onlyReduceInSize, 1.0f);
	}

	g.setColour (findColour (DirectoryContentsDisplayComponent::textColourId));
	g.setFont (height * 0.7f);

	if (width > 450 && ! isDirectory)
	{
		const int sizeX = roundToInt (width * 0.7f);
		const int dateX = roundToInt (width * 0.8f);

		g.drawFittedText (filename,
						  x, 0, sizeX - x, height,
						  Justification::centredLeft, 1);

		g.setFont (height * 0.5f);
		g.setColour (Colours::darkgrey);

		if (! isDirectory)
		{
			g.drawFittedText (fileSizeDescription,
							  sizeX, 0, dateX - sizeX - 8, height,
							  Justification::centredRight, 1);

			g.drawFittedText (fileTimeDescription,
							  dateX, 0, width - 8 - dateX, height,
							  Justification::centredRight, 1);
		}
	}
	else
	{
		g.drawFittedText (filename,
						  x, 0, width - x, height,
						  Justification::centredLeft, 1);

	}
}

Button* LookAndFeel::createFileBrowserGoUpButton()
{
	DrawableButton* goUpButton = new DrawableButton ("up", DrawableButton::ImageOnButtonBackground);

	Path arrowPath;
	arrowPath.addArrow (Line<float> (50.0f, 100.0f, 50.0f, 0.0f), 40.0f, 100.0f, 50.0f);

	DrawablePath arrowImage;
	arrowImage.setFill (Colours::black.withAlpha (0.4f));
	arrowImage.setPath (arrowPath);

	goUpButton->setImages (&arrowImage);

	return goUpButton;
}

void LookAndFeel::layoutFileBrowserComponent (FileBrowserComponent& browserComp,
											  DirectoryContentsDisplayComponent* fileListComponent,
											  FilePreviewComponent* previewComp,
											  ComboBox* currentPathBox,
											  TextEditor* filenameBox,
											  Button* goUpButton)
{
	const int x = 8;
	int w = browserComp.getWidth() - x - x;

	if (previewComp != nullptr)
	{
		const int previewWidth = w / 3;
		previewComp->setBounds (x + w - previewWidth, 0, previewWidth, browserComp.getHeight());

		w -= previewWidth + 4;
	}

	int y = 4;

	const int controlsHeight = 22;
	const int bottomSectionHeight = controlsHeight + 8;
	const int upButtonWidth = 50;

	currentPathBox->setBounds (x, y, w - upButtonWidth - 6, controlsHeight);
	goUpButton->setBounds (x + w - upButtonWidth, y, upButtonWidth, controlsHeight);

	y += controlsHeight + 4;

	Component* const listAsComp = dynamic_cast <Component*> (fileListComponent);
	listAsComp->setBounds (x, y, w, browserComp.getHeight() - y - bottomSectionHeight);

	y = listAsComp->getBottom() + 4;
	filenameBox->setBounds (x + 50, y, w - 50, controlsHeight);
}

// Pulls a drawable out of compressed valuetree data..
Drawable* LookAndFeel::loadDrawableFromData (const void* data, size_t numBytes)
{
	MemoryInputStream m (data, numBytes, false);
	GZIPDecompressorInputStream gz (m);
	ValueTree drawable (ValueTree::readFromStream (gz));
	return Drawable::createFromValueTree (drawable.getChild (0), nullptr);
}

const Drawable* LookAndFeel::getDefaultFolderImage()
{
	if (folderImage == nullptr)
	{
		static const unsigned char drawableData[] =
		{ 120,218,197,86,77,111,27,55,16,229,182,161,237,6,61,39,233,77,63,192,38,56,195,225,215,209,105,210,2,141,13,20,201,193,109,111,178,181,178,183,145,181,130,180,110,145,127,159,199,93,73,137,87,53,218,91,109,192,160,151,179,156,55,111,222,188,229,155,247,
		231,87,231,175,47,222,170,234,155,229,244,190,86,213,115,253,102,61,253,123,122,189,168,85,51,83,213,119,250,238,221,47,231,151,175,223,169,170,250,121,221,62,172,84,245,172,60,63,209,243,118,49,171,215,170,107,87,23,245,188,83,213,145,182,167,19,91,
		254,127,223,220,222,117,37,68,82,40,143,174,219,174,107,239,135,168,147,18,37,108,85,245,237,46,207,70,33,249,175,211,238,78,85,186,28,253,76,175,73,109,186,117,251,177,190,106,102,229,241,247,58,24,103,203,15,101,245,103,219,44,187,15,221,39,0,172,142,
		245,125,211,1,196,205,116,181,125,114,164,175,31,186,78,45,219,229,31,245,186,189,106,150,179,102,121,139,100,154,240,231,167,102,177,64,72,247,105,213,23,122,187,158,206,154,122,217,169,85,57,18,1,47,53,101,107,18,135,204,167,147,192,201,216,20,114,
		244,195,62,171,234,7,125,198,100,136,216,145,149,211,9,57,103,40,249,72,219,8,167,170,87,250,140,162,199,123,226,3,34,82,202,134,131,13,172,74,170,233,162,0,177,234,166,93,180,15,235,141,170,206,180,157,204,231,150,156,159,207,39,195,50,214,88,18,150,
		245,205,124,250,104,169,212,135,158,19,144,53,20,112,172,55,237,2,132,13,199,149,130,230,115,145,112,147,147,82,61,157,32,238,178,253,11,145,213,138,10,52,138,38,103,111,99,164,211,137,139,198,35,177,35,167,212,143,15,215,205,13,160,109,163,172,225,152,
		16,232,17,149,140,103,144,158,146,90,113,217,12,6,197,167,236,3,54,5,181,101,73,54,138,90,245,165,227,120,18,252,150,77,15,242,188,228,204,81,169,139,102,249,5,68,192,145,14,244,112,1,145,29,94,137,96,235,49,136,151,58,246,32,88,192,161,88,176,76,226,
		36,247,24,176,7,232,62,16,83,42,155,201,160,30,222,65,72,98,82,76,33,198,254,197,96,124,10,150,243,8,130,48,228,36,94,124,6,4,43,38,0,142,205,99,30,4,221,13,33,230,220,71,177,65,49,142,243,150,7,1,51,20,2,5,96,96,84,225,56,217,188,3,33,46,24,228,112,
		69,69,12,68,228,108,242,99,16,165,118,208,28,51,200,98,87,42,74,62,209,24,4,206,48,22,153,125,132,220,196,56,15,234,99,216,130,0,141,38,74,162,130,48,35,163,141,94,196,245,32,94,104,7,154,132,209,40,108,162,165,232,153,165,17,4,138,201,176,135,58,49,
		165,130,122,108,114,54,28,240,64,17,89,188,79,177,116,149,10,4,246,91,30,94,104,112,96,226,144,131,144,142,98,78,177,7,128,81,242,224,140,36,249,80,208,145,196,12,202,15,16,60,161,200,69,187,169,213,86,198,123,87,224,255,199,21,94,105,134,72,40,177,245,
		14,182,32,232,54,196,231,100,111,11,189,168,201,39,177,84,102,38,139,177,168,74,210,87,174,64,20,138,160,67,111,10,4,98,196,97,60,158,118,133,25,111,173,224,171,37,97,185,119,133,221,242,63,184,194,140,71,174,240,252,145,43,72,32,147,146,147,4,104,104,
		117,134,10,18,12,107,212,40,72,148,57,6,71,69,135,222,248,16,160,168,3,169,144,55,201,69,41,147,137,134,99,50,97,8,178,85,43,217,140,201,151,192,152,10,242,190,24,11,59,183,29,25,42,115,236,98,14,229,252,32,80,66,0,162,17,136,72,6,67,5,45,242,224,10,
		193,102,71,50,6,17,129,212,18,115,105,150,80,169,45,123,222,141,76,178,70,32,55,24,90,217,132,71,73,200,57,238,204,3,136,49,144,185,55,183,190,20,137,52,246,47,113,232,158,69,35,49,145,208,129,193,56,178,77,135,230,145,113,22,140,69,74,20,146,2,120,218,
		155,135,48,32,10,89,30,156,165,204,254,222,193,160,12,19,49,6,210,59,11,70,62,4,31,15,64,196,2,157,98,33,58,1,104,32,152,50,31,128,64,148,183,197,108,209,89,107,240,41,75,36,123,16,208,108,180,44,236,250,182,227,27,20,137,118,76,60,165,137,221,92,94,
		78,215,31,235,245,230,183,242,229,30,214,251,251,195,145,94,148,15,253,170,221,52,93,211,46,7,109,171,81,208,177,94,247,119,132,47,81,186,92,22,246,7,255,254,15,7,107,141,171,197,191,156,123,162,135,187,198,227,131,113,219,80,159,1,4,239,223,231,0,0 };

		folderImage = loadDrawableFromData (drawableData, sizeof (drawableData));
	}

	return folderImage;
}

const Drawable* LookAndFeel::getDefaultDocumentFileImage()
{
	if (documentImage == nullptr)
	{
		static const unsigned char drawableData[] =
		{ 120,218,213,88,77,115,219,54,16,37,147,208,246,228,214,75,155,246,164,123,29,12,176,216,197,199,49,105,218,94,156,153,78,114,72,219,155,108,75,137,26,89,212,200,116,59,233,175,239,3,105,201,164,68,50,158,166,233,76,196,11,69,60,173,128,197,123,139,183,
		124,241,234,217,155,103,207,207,126,204,242,7,171,233,213,44,203,31,23,47,54,211,191,166,231,203,89,182,184,204,242,147,226,195,165,219,252,125,150,229,249,207,155,242,102,157,229,143,210,227,199,197,101,121,113,115,53,91,85,89,85,174,207,102,243,42,
		203,143,10,125,58,209,233,251,171,197,219,119,85,250,173,97,151,30,157,151,85,85,94,53,168,147,132,50,226,179,252,225,246,143,174,179,44,63,254,101,90,189,203,242,34,5,127,84,172,77,118,93,109,202,247,179,55,139,203,244,248,97,161,179,63,202,197,170,
		122,93,125,192,196,242,227,226,106,81,205,54,217,197,116,125,251,228,168,56,191,169,170,108,85,174,126,159,109,202,55,139,213,229,98,245,182,249,97,254,240,167,197,114,137,5,86,31,214,245,111,175,203,37,254,230,162,92,150,55,155,180,148,249,237,39,203,
		94,215,127,58,10,213,245,39,203,234,249,102,249,87,47,203,63,129,204,49,227,252,73,225,149,145,104,131,245,254,116,34,202,82,164,16,153,179,236,108,177,234,7,49,41,237,130,144,167,17,144,15,42,104,239,93,12,35,32,99,68,9,187,24,125,7,244,77,23,36,164,
		40,56,226,61,12,107,229,130,215,100,105,24,227,89,17,246,211,105,55,140,49,218,43,207,100,245,72,28,195,70,17,230,201,118,8,243,164,139,233,95,88,23,52,152,162,54,104,48,217,237,105,15,111,91,107,253,131,160,118,34,239,69,128,54,232,135,101,121,61,203,
		110,169,181,147,2,253,159,82,48,180,229,247,167,74,193,41,141,188,35,93,241,116,18,148,113,214,120,207,113,47,19,109,16,51,182,153,193,5,59,2,10,90,69,114,218,135,48,2,50,198,43,171,189,152,81,144,88,108,85,136,78,246,64,54,42,163,35,69,30,3,121,82,38,
		98,81,98,70,64,70,139,34,111,163,167,49,144,13,202,138,179,58,220,23,52,180,186,54,104,48,79,109,208,96,198,219,19,31,220,187,118,10,6,65,237,100,222,139,5,109,80,191,30,236,151,162,135,147,142,30,68,105,182,58,6,22,84,43,229,124,148,116,97,145,55,231,
		139,11,76,228,16,37,14,48,205,145,77,134,34,176,55,152,182,200,57,99,93,204,144,145,253,65,97,229,132,72,104,63,62,71,21,140,54,186,41,226,59,84,19,63,130,15,222,235,224,185,59,104,27,226,68,101,153,241,227,177,248,29,20,136,26,8,252,178,183,241,219,
		131,137,160,209,107,109,92,79,124,16,211,184,104,93,77,130,110,124,2,65,172,67,201,60,157,88,163,2,91,99,92,216,198,55,78,69,75,190,150,119,84,98,200,71,150,109,124,36,204,227,52,8,33,229,223,68,167,173,167,131,248,137,212,226,141,19,233,160,154,248,
		144,142,195,140,137,185,59,104,15,247,119,40,126,23,69,81,200,242,110,254,123,20,49,94,112,110,245,199,111,241,167,87,36,252,101,138,132,149,22,22,38,65,134,29,182,139,24,230,192,31,144,184,133,130,72,44,131,210,142,111,147,216,30,76,123,30,113,206,242,
		150,196,157,65,129,130,76,180,194,61,34,225,160,5,228,233,160,118,34,137,26,202,115,212,29,108,72,134,243,223,90,114,226,199,226,119,80,6,245,152,197,122,217,146,184,53,24,140,210,30,21,59,80,79,124,182,202,71,207,218,112,159,72,80,53,140,109,68,2,191,
		227,217,210,78,36,94,137,88,231,82,157,8,176,61,0,122,191,19,137,3,255,13,39,183,228,20,193,151,144,119,166,79,36,40,253,156,138,72,11,181,19,137,14,46,176,217,27,180,135,251,219,31,255,235,61,148,165,96,72,122,118,23,229,81,52,135,24,250,163,183,216,
		211,43,17,217,151,136,253,116,137,28,53,188,127,92,188,221,76,47,23,169,59,90,167,144,141,239,197,86,104,141,189,60,157,80,84,142,140,4,31,154,241,122,105,132,41,107,13,201,39,86,120,24,82,114,206,198,6,96,27,227,172,36,232,168,201,36,219,24,113,62,163,
		154,101,233,143,166,203,102,26,141,206,174,179,252,89,161,39,243,249,197,121,186,38,233,246,146,211,53,1,123,56,194,231,122,143,103,179,217,60,204,167,19,147,110,41,93,173,219,123,72,89,248,35,173,16,220,50,179,111,60,181,24,88,103,156,235,7,78,248,14,
		4,119,78,162,93,60,112,35,109,16,124,126,12,17,71,67,24,1,165,142,1,181,215,248,56,6,66,235,193,137,167,61,22,30,5,3,27,101,71,64,169,25,112,216,2,63,22,169,110,43,18,200,140,129,208,160,88,44,220,208,125,65,67,171,107,131,6,243,212,6,13,102,188,61,241,
		225,189,107,165,96,16,212,78,230,189,88,208,6,245,235,214,237,235,150,62,167,110,155,106,170,53,133,192,117,193,20,84,78,74,174,98,39,92,156,8,112,21,46,80,106,12,209,207,225,228,16,113,59,225,126,87,60,133,25,209,34,36,2,99,242,52,197,48,30,75,244,247,
		212,238,246,182,173,221,185,78,215,127,167,221,162,163,221,250,152,217,146,196,222,145,100,223,235,105,108,28,250,149,212,74,224,86,2,213,118,110,119,204,224,144,208,38,214,131,200,14,214,223,120,189,230,53,1,193,70,133,154,131,56,223,16,229,48,188,14,
		201,205,213,121,71,233,68,89,15,124,103,37,53,26,11,118,176,127,169,88,166,158,219,178,117,173,83,108,75,95,55,68,186,193,53,246,146,206,127,6,63,53,78,58,228,204,155,224,113,74,91,232,221,195,240,105,215,34,29,138,64,128,183,8,130,233,71,173,56,54,101,
		99,75,186,111,65,58,28,229,145,82,19,152,12,99,180,81,130,131,75,234,229,220,247,53,231,154,79,205,185,185,155,199,249,172,38,85,253,204,76,68,95,92,204,207,255,221,75,178,227,14,187,224,224,97,202,172,173,219,12,167,130,133,9,54,135,245,92,176,29,134,
		165,110,139,141,18,16,223,29,188,183,65,207,144,106,144,151,143,128,224,176,168,110,140,32,62,56,110,219,195,54,235,20,68,209,216,34,232,21,6,41,234,157,39,211,201,107,160,230,66,225,56,153,9,101,21,37,237,150,204,14,115,208,22,221,54,216,230,33,116,
		14,65,14,44,19,8,236,73,71,246,182,110,125,224,75,132,195,214,247,163,36,51,252,84,76,124,37,212,100,88,62,183,179,76,67,217,218,242,244,229,116,243,126,182,185,254,21,105,126,208,220,239,94,229,30,21,203,244,202,117,93,94,47,170,69,185,106,246,60,219,
		3,29,23,155,250,109,237,29,170,72,175,109,119,129,127,235,9,92,20,85,185,254,72,220,147,162,121,235,219,13,44,144,225,63,241,244,165,51,0,0 };

		documentImage = loadDrawableFromData (drawableData, sizeof (drawableData));
	}

	return documentImage;
}

void LookAndFeel::drawLevelMeter (Graphics& g, int width, int height, float level)
{
	g.setColour (Colours::white.withAlpha (0.7f));
	g.fillRoundedRectangle (0.0f, 0.0f, (float) width, (float) height, 3.0f);
	g.setColour (Colours::black.withAlpha (0.2f));
	g.drawRoundedRectangle (1.0f, 1.0f, width - 2.0f, height - 2.0f, 3.0f, 1.0f);

	const int totalBlocks = 7;
	const int numBlocks = roundToInt (totalBlocks * level);
	const float w = (width - 6.0f) / (float) totalBlocks;

	for (int i = 0; i < totalBlocks; ++i)
	{
		if (i >= numBlocks)
			g.setColour (Colours::lightblue.withAlpha (0.6f));
		else
			g.setColour (i < totalBlocks - 1 ? Colours::blue.withAlpha (0.5f)
											 : Colours::red);

		g.fillRoundedRectangle (3.0f + i * w + w * 0.1f, 3.0f, w * 0.8f, height - 6.0f, w * 0.4f);
	}
}

void LookAndFeel::drawKeymapChangeButton (Graphics& g, int width, int height, Button& button, const String& keyDescription)
{
	const Colour textColour (button.findColour (KeyMappingEditorComponent::textColourId, true));

	if (keyDescription.isNotEmpty())
	{
		if (button.isEnabled())
		{
			const float alpha = button.isDown() ? 0.3f : (button.isOver() ? 0.15f : 0.08f);
			g.fillAll (textColour.withAlpha (alpha));

			g.setOpacity (0.3f);
			g.drawBevel (0, 0, width, height, 2);
		}

		g.setColour (textColour);
		g.setFont (height * 0.6f);
		g.drawFittedText (keyDescription,
						  3, 0, width - 6, height,
						  Justification::centred, 1);
	}
	else
	{
		const float thickness = 7.0f;
		const float indent = 22.0f;

		Path p;
		p.addEllipse (0.0f, 0.0f, 100.0f, 100.0f);
		p.addRectangle (indent, 50.0f - thickness, 100.0f - indent * 2.0f, thickness * 2.0f);
		p.addRectangle (50.0f - thickness, indent, thickness * 2.0f, 50.0f - indent - thickness);
		p.addRectangle (50.0f - thickness, 50.0f + thickness, thickness * 2.0f, 50.0f - indent - thickness);
		p.setUsingNonZeroWinding (false);

		g.setColour (textColour.withAlpha (button.isDown() ? 0.7f : (button.isOver() ? 0.5f : 0.3f)));
		g.fillPath (p, p.getTransformToScaleToFit (2.0f, 2.0f, width - 4.0f, height - 4.0f, true));
	}

	if (button.hasKeyboardFocus (false))
	{
		g.setColour (textColour.withAlpha (0.4f));
		g.drawRect (0, 0, width, height);
	}
}

void LookAndFeel::drawShinyButtonShape (Graphics& g,
										float x, float y, float w, float h,
										float maxCornerSize,
										const Colour& baseColour,
										const float strokeWidth,
										const bool flatOnLeft,
										const bool flatOnRight,
										const bool flatOnTop,
										const bool flatOnBottom) noexcept
{
	if (w <= strokeWidth * 1.1f || h <= strokeWidth * 1.1f)
		return;

	const float cs = jmin (maxCornerSize, w * 0.5f, h * 0.5f);

	Path outline;
	LookAndFeelHelpers::createRoundedPath (outline, x, y, w, h, cs,
											! (flatOnLeft || flatOnTop),
											! (flatOnRight || flatOnTop),
											! (flatOnLeft || flatOnBottom),
											! (flatOnRight || flatOnBottom));

	ColourGradient cg (baseColour, 0.0f, y,
					   baseColour.overlaidWith (Colour (0x070000ff)), 0.0f, y + h,
					   false);

	cg.addColour (0.5, baseColour.overlaidWith (Colour (0x33ffffff)));
	cg.addColour (0.51, baseColour.overlaidWith (Colour (0x110000ff)));

	g.setGradientFill (cg);
	g.fillPath (outline);

	g.setColour (Colour (0x80000000));
	g.strokePath (outline, PathStrokeType (strokeWidth));
}

void LookAndFeel::drawGlassSphere (Graphics& g,
								   const float x, const float y,
								   const float diameter,
								   const Colour& colour,
								   const float outlineThickness) noexcept
{
	if (diameter <= outlineThickness)
		return;

	Path p;
	p.addEllipse (x, y, diameter, diameter);

	{
		ColourGradient cg (Colours::white.overlaidWith (colour.withMultipliedAlpha (0.3f)), 0, y,
						   Colours::white.overlaidWith (colour.withMultipliedAlpha (0.3f)), 0, y + diameter, false);

		cg.addColour (0.4, Colours::white.overlaidWith (colour));

		g.setGradientFill (cg);
		g.fillPath (p);
	}

	g.setGradientFill (ColourGradient (Colours::white, 0, y + diameter * 0.06f,
									   Colours::transparentWhite, 0, y + diameter * 0.3f, false));
	g.fillEllipse (x + diameter * 0.2f, y + diameter * 0.05f, diameter * 0.6f, diameter * 0.4f);

	ColourGradient cg (Colours::transparentBlack,
					   x + diameter * 0.5f, y + diameter * 0.5f,
					   Colours::black.withAlpha (0.5f * outlineThickness * colour.getFloatAlpha()),
					   x, y + diameter * 0.5f, true);

	cg.addColour (0.7, Colours::transparentBlack);
	cg.addColour (0.8, Colours::black.withAlpha (0.1f * outlineThickness));

	g.setGradientFill (cg);
	g.fillPath (p);

	g.setColour (Colours::black.withAlpha (0.5f * colour.getFloatAlpha()));
	g.drawEllipse (x, y, diameter, diameter, outlineThickness);
}

void LookAndFeel::drawGlassPointer (Graphics& g,
									const float x, const float y,
									const float diameter,
									const Colour& colour, const float outlineThickness,
									const int direction) noexcept
{
	if (diameter <= outlineThickness)
		return;

	Path p;
	p.startNewSubPath (x + diameter * 0.5f, y);
	p.lineTo (x + diameter, y + diameter * 0.6f);
	p.lineTo (x + diameter, y + diameter);
	p.lineTo (x, y + diameter);
	p.lineTo (x, y + diameter * 0.6f);
	p.closeSubPath();

	p.applyTransform (AffineTransform::rotation (direction * (float_Pi * 0.5f), x + diameter * 0.5f, y + diameter * 0.5f));

	{
		ColourGradient cg (Colours::white.overlaidWith (colour.withMultipliedAlpha (0.3f)), 0, y,
						   Colours::white.overlaidWith (colour.withMultipliedAlpha (0.3f)), 0, y + diameter, false);

		cg.addColour (0.4, Colours::white.overlaidWith (colour));

		g.setGradientFill (cg);
		g.fillPath (p);
	}

	ColourGradient cg (Colours::transparentBlack,
					   x + diameter * 0.5f, y + diameter * 0.5f,
					   Colours::black.withAlpha (0.5f * outlineThickness * colour.getFloatAlpha()),
					   x - diameter * 0.2f, y + diameter * 0.5f, true);

	cg.addColour (0.5, Colours::transparentBlack);
	cg.addColour (0.7, Colours::black.withAlpha (0.07f * outlineThickness));

	g.setGradientFill (cg);
	g.fillPath (p);

	g.setColour (Colours::black.withAlpha (0.5f * colour.getFloatAlpha()));
	g.strokePath (p, PathStrokeType (outlineThickness));
}

void LookAndFeel::drawGlassLozenge (Graphics& g,
									const float x, const float y,
									const float width, const float height,
									const Colour& colour,
									const float outlineThickness,
									const float cornerSize,
									const bool flatOnLeft,
									const bool flatOnRight,
									const bool flatOnTop,
									const bool flatOnBottom) noexcept
{
	if (width <= outlineThickness || height <= outlineThickness)
		return;

	const int intX = (int) x;
	const int intY = (int) y;
	const int intW = (int) width;
	const int intH = (int) height;

	const float cs = cornerSize < 0 ? jmin (width * 0.5f, height * 0.5f) : cornerSize;
	const float edgeBlurRadius = height * 0.75f + (height - cs * 2.0f);
	const int intEdge = (int) edgeBlurRadius;

	Path outline;
	LookAndFeelHelpers::createRoundedPath (outline, x, y, width, height, cs,
											! (flatOnLeft || flatOnTop),
											! (flatOnRight || flatOnTop),
											! (flatOnLeft || flatOnBottom),
											! (flatOnRight || flatOnBottom));

	{
		ColourGradient cg (colour.darker (0.2f), 0, y,
						   colour.darker (0.2f), 0, y + height, false);

		cg.addColour (0.03, colour.withMultipliedAlpha (0.3f));
		cg.addColour (0.4, colour);
		cg.addColour (0.97, colour.withMultipliedAlpha (0.3f));

		g.setGradientFill (cg);
		g.fillPath (outline);
	}

	ColourGradient cg (Colours::transparentBlack, x + edgeBlurRadius, y + height * 0.5f,
					   colour.darker (0.2f), x, y + height * 0.5f, true);

	cg.addColour (jlimit (0.0, 1.0, 1.0 - (cs * 0.5f) / edgeBlurRadius), Colours::transparentBlack);
	cg.addColour (jlimit (0.0, 1.0, 1.0 - (cs * 0.25f) / edgeBlurRadius), colour.darker (0.2f).withMultipliedAlpha (0.3f));

	if (! (flatOnLeft || flatOnTop || flatOnBottom))
	{
		g.saveState();
		g.setGradientFill (cg);
		g.reduceClipRegion (intX, intY, intEdge, intH);
		g.fillPath (outline);
		g.restoreState();
	}

	if (! (flatOnRight || flatOnTop || flatOnBottom))
	{
		cg.point1.setX (x + width - edgeBlurRadius);
		cg.point2.setX (x + width);

		g.saveState();
		g.setGradientFill (cg);
		g.reduceClipRegion (intX + intW - intEdge, intY, 2 + intEdge, intH);
		g.fillPath (outline);
		g.restoreState();
	}

	{
		const float leftIndent = flatOnTop || flatOnLeft ? 0.0f : cs * 0.4f;
		const float rightIndent = flatOnTop || flatOnRight ? 0.0f : cs * 0.4f;

		Path highlight;
		LookAndFeelHelpers::createRoundedPath (highlight,
											   x + leftIndent,
											   y + cs * 0.1f,
											   width - (leftIndent + rightIndent),
											   height * 0.4f, cs * 0.4f,
											   ! (flatOnLeft || flatOnTop),
											   ! (flatOnRight || flatOnTop),
											   ! (flatOnLeft || flatOnBottom),
											   ! (flatOnRight || flatOnBottom));

		g.setGradientFill (ColourGradient (colour.brighter (10.0f), 0, y + height * 0.06f,
										   Colours::transparentWhite, 0, y + height * 0.4f, false));
		g.fillPath (highlight);
	}

	g.setColour (colour.darker().withMultipliedAlpha (1.5f));
	g.strokePath (outline, PathStrokeType (outlineThickness));
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_LookAndFeel.cpp ***/


/*** Start of inlined file: juce_OldSchoolLookAndFeel.cpp ***/
BEGIN_JUCE_NAMESPACE

OldSchoolLookAndFeel::OldSchoolLookAndFeel()
{
	setColour (TextButton::buttonColourId,	  Colour (0xffbbbbff));
	setColour (ListBox::outlineColourId,		findColour (ComboBox::outlineColourId));
	setColour (ScrollBar::thumbColourId,		Colour (0xffbbbbdd));
	setColour (ScrollBar::backgroundColourId,	   Colours::transparentBlack);
	setColour (Slider::thumbColourId,		   Colours::white);
	setColour (Slider::trackColourId,		   Colour (0x7f000000));
	setColour (Slider::textBoxOutlineColourId,	  Colours::grey);
	setColour (ProgressBar::backgroundColourId,	 Colours::white.withAlpha (0.6f));
	setColour (ProgressBar::foregroundColourId,	 Colours::green.withAlpha (0.7f));
	setColour (PopupMenu::backgroundColourId,		 Colour (0xffeef5f8));
	setColour (PopupMenu::highlightedBackgroundColourId,  Colour (0xbfa4c2ce));
	setColour (PopupMenu::highlightedTextColourId,	Colours::black);
	setColour (TextEditor::focusedOutlineColourId,  findColour (TextButton::buttonColourId));

	scrollbarShadow.setShadowProperties (2.2f, 0.5f, 0, 0);
}

OldSchoolLookAndFeel::~OldSchoolLookAndFeel()
{
}

void OldSchoolLookAndFeel::drawButtonBackground (Graphics& g,
												 Button& button,
												 const Colour& backgroundColour,
												 bool isMouseOverButton,
												 bool isButtonDown)
{
	const int width = button.getWidth();
	const int height = button.getHeight();

	const float indent = 2.0f;
	const int cornerSize = jmin (roundToInt (width * 0.4f),
								 roundToInt (height * 0.4f));

	Path p;
	p.addRoundedRectangle (indent, indent,
						   width - indent * 2.0f,
						   height - indent * 2.0f,
						   (float) cornerSize);

	Colour bc (backgroundColour.withMultipliedSaturation (0.3f));

	if (isMouseOverButton)
	{
		if (isButtonDown)
			bc = bc.brighter();
		else if (bc.getBrightness() > 0.5f)
			bc = bc.darker (0.1f);
		else
			bc = bc.brighter (0.1f);
	}

	g.setColour (bc);
	g.fillPath (p);

	g.setColour (bc.contrasting().withAlpha ((isMouseOverButton) ? 0.6f : 0.4f));
	g.strokePath (p, PathStrokeType ((isMouseOverButton) ? 2.0f : 1.4f));
}

void OldSchoolLookAndFeel::drawTickBox (Graphics& g,
										Component& /*component*/,
										float x, float y, float w, float h,
										const bool ticked,
										const bool isEnabled,
										const bool /*isMouseOverButton*/,
										const bool isButtonDown)
{
	Path box;
	box.addRoundedRectangle (0.0f, 2.0f, 6.0f, 6.0f, 1.0f);

	g.setColour (isEnabled ? Colours::blue.withAlpha (isButtonDown ? 0.3f : 0.1f)
						   : Colours::lightgrey.withAlpha (0.1f));

	AffineTransform trans (AffineTransform::scale (w / 9.0f, h / 9.0f).translated (x, y));

	g.fillPath (box, trans);

	g.setColour (Colours::black.withAlpha (0.6f));
	g.strokePath (box, PathStrokeType (0.9f), trans);

	if (ticked)
	{
		Path tick;
		tick.startNewSubPath (1.5f, 3.0f);
		tick.lineTo (3.0f, 6.0f);
		tick.lineTo (6.0f, 0.0f);

		g.setColour (isEnabled ? Colours::black : Colours::grey);
		g.strokePath (tick, PathStrokeType (2.5f), trans);
	}
}

void OldSchoolLookAndFeel::drawToggleButton (Graphics& g,
											 ToggleButton& button,
											 bool isMouseOverButton,
											 bool isButtonDown)
{
	if (button.hasKeyboardFocus (true))
	{
		g.setColour (button.findColour (TextEditor::focusedOutlineColourId));
		g.drawRect (0, 0, button.getWidth(), button.getHeight());
	}

	const int tickWidth = jmin (20, button.getHeight() - 4);

	drawTickBox (g, button, 4.0f, (button.getHeight() - tickWidth) * 0.5f,
				 (float) tickWidth, (float) tickWidth,
				 button.getToggleState(),
				 button.isEnabled(),
				 isMouseOverButton,
				 isButtonDown);

	g.setColour (button.findColour (ToggleButton::textColourId));
	g.setFont (jmin (15.0f, button.getHeight() * 0.6f));

	if (! button.isEnabled())
		g.setOpacity (0.5f);

	const int textX = tickWidth + 5;

	g.drawFittedText (button.getButtonText(),
					  textX, 4,
					  button.getWidth() - textX - 2, button.getHeight() - 8,
					  Justification::centredLeft, 10);
}

void OldSchoolLookAndFeel::drawProgressBar (Graphics& g, ProgressBar& progressBar,
											int width, int height,
											double progress, const String& textToShow)
{
	if (progress < 0 || progress >= 1.0)
	{
		LookAndFeel::drawProgressBar (g, progressBar, width, height, progress, textToShow);
	}
	else
	{
		const Colour background (progressBar.findColour (ProgressBar::backgroundColourId));
		const Colour foreground (progressBar.findColour (ProgressBar::foregroundColourId));

		g.fillAll (background);
		g.setColour (foreground);

		g.fillRect (1, 1,
					jlimit (0, width - 2, roundToInt (progress * (width - 2))),
					height - 2);

		if (textToShow.isNotEmpty())
		{
			g.setColour (Colour::contrasting (background, foreground));
			g.setFont (height * 0.6f);

			g.drawText (textToShow, 0, 0, width, height, Justification::centred, false);
		}
	}
}

void OldSchoolLookAndFeel::drawScrollbarButton (Graphics& g,
												ScrollBar& bar,
												int width, int height,
												int buttonDirection,
												bool isScrollbarVertical,
												bool isMouseOverButton,
												bool isButtonDown)
{
	if (isScrollbarVertical)
		width -= 2;
	else
		height -= 2;

	Path p;

	if (buttonDirection == 0)
		p.addTriangle (width * 0.5f, height * 0.2f,
					   width * 0.1f, height * 0.7f,
					   width * 0.9f, height * 0.7f);
	else if (buttonDirection == 1)
		p.addTriangle (width * 0.8f, height * 0.5f,
					   width * 0.3f, height * 0.1f,
					   width * 0.3f, height * 0.9f);
	else if (buttonDirection == 2)
		p.addTriangle (width * 0.5f, height * 0.8f,
					   width * 0.1f, height * 0.3f,
					   width * 0.9f, height * 0.3f);
	else if (buttonDirection == 3)
		p.addTriangle (width * 0.2f, height * 0.5f,
					   width * 0.7f, height * 0.1f,
					   width * 0.7f, height * 0.9f);

	if (isButtonDown)
		g.setColour (Colours::white);
	else if (isMouseOverButton)
		g.setColour (Colours::white.withAlpha (0.7f));
	else
		g.setColour (bar.findColour (ScrollBar::thumbColourId).withAlpha (0.5f));

	g.fillPath (p);

	g.setColour (Colours::black.withAlpha (0.5f));
	g.strokePath (p, PathStrokeType (0.5f));
}

void OldSchoolLookAndFeel::drawScrollbar (Graphics& g,
										  ScrollBar& bar,
										  int x, int y,
										  int width, int height,
										  bool isScrollbarVertical,
										  int thumbStartPosition,
										  int thumbSize,
										  bool isMouseOver,
										  bool isMouseDown)
{
	g.fillAll (bar.findColour (ScrollBar::backgroundColourId));

	g.setColour (bar.findColour (ScrollBar::thumbColourId)
					.withAlpha ((isMouseOver || isMouseDown) ? 0.4f : 0.15f));

	if (thumbSize > 0.0f)
	{
		Rectangle<int> thumb;

		if (isScrollbarVertical)
		{
			width -= 2;
			g.fillRect (x + roundToInt (width * 0.35f), y,
						roundToInt (width * 0.3f), height);

			thumb.setBounds (x + 1, thumbStartPosition,
							 width - 2, thumbSize);
		}
		else
		{
			height -= 2;
			g.fillRect (x, y + roundToInt (height * 0.35f),
						width, roundToInt (height * 0.3f));

			thumb.setBounds (thumbStartPosition, y + 1,
							 thumbSize, height - 2);
		}

		g.setColour (bar.findColour (ScrollBar::thumbColourId)
						.withAlpha ((isMouseOver || isMouseDown) ? 0.95f : 0.7f));

		g.fillRect (thumb);

		g.setColour (Colours::black.withAlpha ((isMouseOver || isMouseDown) ? 0.4f : 0.25f));
		g.drawRect (thumb.getX(), thumb.getY(), thumb.getWidth(), thumb.getHeight());

		if (thumbSize > 16)
		{
			for (int i = 3; --i >= 0;)
			{
				const float linePos = thumbStartPosition + thumbSize / 2 + (i - 1) * 4.0f;
				g.setColour (Colours::black.withAlpha (0.15f));

				if (isScrollbarVertical)
				{
					g.drawLine (x + width * 0.2f, linePos, width * 0.8f, linePos);
					g.setColour (Colours::white.withAlpha (0.15f));
					g.drawLine (width * 0.2f, linePos - 1, width * 0.8f, linePos - 1);
				}
				else
				{
					g.drawLine (linePos, height * 0.2f, linePos, height * 0.8f);
					g.setColour (Colours::white.withAlpha (0.15f));
					g.drawLine (linePos - 1, height * 0.2f, linePos - 1, height * 0.8f);
				}
			}
		}
	}
}

ImageEffectFilter* OldSchoolLookAndFeel::getScrollbarEffect()
{
	return &scrollbarShadow;
}

void OldSchoolLookAndFeel::drawPopupMenuBackground (Graphics& g, int width, int height)
{
	g.fillAll (findColour (PopupMenu::backgroundColourId));

	g.setColour (Colours::black.withAlpha (0.6f));
	g.drawRect (0, 0, width, height);
}

void OldSchoolLookAndFeel::drawMenuBarBackground (Graphics& g, int /*width*/, int /*height*/,
												  bool, MenuBarComponent& menuBar)
{
	g.fillAll (menuBar.findColour (PopupMenu::backgroundColourId));
}

void OldSchoolLookAndFeel::drawTextEditorOutline (Graphics& g, int width, int height, TextEditor& textEditor)
{
	if (textEditor.isEnabled())
	{
		g.setColour (textEditor.findColour (TextEditor::outlineColourId));
		g.drawRect (0, 0, width, height);
	}
}

void OldSchoolLookAndFeel::drawComboBox (Graphics& g, int width, int height,
										 const bool isButtonDown,
										 int buttonX, int buttonY,
										 int buttonW, int buttonH,
										 ComboBox& box)
{
	g.fillAll (box.findColour (ComboBox::backgroundColourId));

	g.setColour (box.findColour ((isButtonDown) ? ComboBox::buttonColourId
												: ComboBox::backgroundColourId));
	g.fillRect (buttonX, buttonY, buttonW, buttonH);

	g.setColour (box.findColour (ComboBox::outlineColourId));
	g.drawRect (0, 0, width, height);

	const float arrowX = 0.2f;
	const float arrowH = 0.3f;

	if (box.isEnabled())
	{
		Path p;
		p.addTriangle (buttonX + buttonW * 0.5f,		buttonY + buttonH * (0.45f - arrowH),
					   buttonX + buttonW * (1.0f - arrowX), buttonY + buttonH * 0.45f,
					   buttonX + buttonW * arrowX,	  buttonY + buttonH * 0.45f);

		p.addTriangle (buttonX + buttonW * 0.5f,		buttonY + buttonH * (0.55f + arrowH),
					   buttonX + buttonW * (1.0f - arrowX), buttonY + buttonH * 0.55f,
					   buttonX + buttonW * arrowX,	  buttonY + buttonH * 0.55f);

		g.setColour (box.findColour ((isButtonDown) ? ComboBox::backgroundColourId
													: ComboBox::buttonColourId));
		g.fillPath (p);
	}
}

const Font OldSchoolLookAndFeel::getComboBoxFont (ComboBox& box)
{
	Font f (jmin (15.0f, box.getHeight() * 0.85f));
	f.setHorizontalScale (0.9f);
	return f;
}

static void drawTriangle (Graphics& g, float x1, float y1, float x2, float y2, float x3, float y3, const Colour& fill, const Colour& outline) noexcept
{
	Path p;
	p.addTriangle (x1, y1, x2, y2, x3, y3);
	g.setColour (fill);
	g.fillPath (p);

	g.setColour (outline);
	g.strokePath (p, PathStrokeType (0.3f));
}

void OldSchoolLookAndFeel::drawLinearSlider (Graphics& g,
											 int x, int y,
											 int w, int h,
											 float sliderPos,
											 float minSliderPos,
											 float maxSliderPos,
											 const Slider::SliderStyle style,
											 Slider& slider)
{
	g.fillAll (slider.findColour (Slider::backgroundColourId));

	if (style == Slider::LinearBar)
	{
		g.setColour (slider.findColour (Slider::thumbColourId));
		g.fillRect (x, y, (int) sliderPos - x, h);

		g.setColour (slider.findColour (Slider::textBoxTextColourId).withMultipliedAlpha (0.5f));
		g.drawRect (x, y, (int) sliderPos - x, h);
	}
	else
	{
		g.setColour (slider.findColour (Slider::trackColourId)
						   .withMultipliedAlpha (slider.isEnabled() ? 1.0f : 0.3f));

		if (slider.isHorizontal())
		{
			g.fillRect (x, y + roundToInt (h * 0.6f),
						w, roundToInt (h * 0.2f));
		}
		else
		{
			g.fillRect (x + roundToInt (w * 0.5f - jmin (3.0f, w * 0.1f)), y,
						jmin (4, roundToInt (w * 0.2f)), h);
		}

		float alpha = 0.35f;

		if (slider.isEnabled())
			alpha = slider.isMouseOverOrDragging() ? 1.0f : 0.7f;

		const Colour fill (slider.findColour (Slider::thumbColourId).withAlpha (alpha));
		const Colour outline (Colours::black.withAlpha (slider.isEnabled() ? 0.7f : 0.35f));

		if (style == Slider::TwoValueVertical || style == Slider::ThreeValueVertical)
		{
			drawTriangle (g, x + w * 0.5f + jmin (4.0f, w * 0.3f), minSliderPos,
						  x + w * 0.5f - jmin (8.0f, w * 0.4f), minSliderPos - 7.0f,
						  x + w * 0.5f - jmin (8.0f, w * 0.4f), minSliderPos,
						  fill, outline);

			drawTriangle (g, x + w * 0.5f + jmin (4.0f, w * 0.3f), maxSliderPos,
						  x + w * 0.5f - jmin (8.0f, w * 0.4f), maxSliderPos,
						  x + w * 0.5f - jmin (8.0f, w * 0.4f), maxSliderPos + 7.0f,
						  fill, outline);
		}
		else if (style == Slider::TwoValueHorizontal || style == Slider::ThreeValueHorizontal)
		{
			drawTriangle (g, minSliderPos, y + h * 0.6f - jmin (4.0f, h * 0.3f),
						  minSliderPos - 7.0f, y + h * 0.9f ,
						  minSliderPos, y + h * 0.9f,
						  fill, outline);

			drawTriangle (g, maxSliderPos, y + h * 0.6f - jmin (4.0f, h * 0.3f),
						  maxSliderPos, y + h * 0.9f,
						  maxSliderPos + 7.0f, y + h * 0.9f,
						  fill, outline);
		}

		if (style == Slider::LinearHorizontal || style == Slider::ThreeValueHorizontal)
		{
			drawTriangle (g, sliderPos, y + h * 0.9f,
						  sliderPos - 7.0f, y + h * 0.2f,
						  sliderPos + 7.0f, y + h * 0.2f,
						  fill, outline);
		}
		else if (style == Slider::LinearVertical || style == Slider::ThreeValueVertical)
		{
			drawTriangle (g, x + w * 0.5f - jmin (4.0f, w * 0.3f), sliderPos,
						  x + w * 0.5f + jmin (8.0f, w * 0.4f), sliderPos - 7.0f,
						  x + w * 0.5f + jmin (8.0f, w * 0.4f), sliderPos + 7.0f,
						  fill, outline);
		}
	}
}

Button* OldSchoolLookAndFeel::createSliderButton (const bool isIncrement)
{
	if (isIncrement)
		return new ArrowButton ("u", 0.75f, Colours::white.withAlpha (0.8f));
	else
		return new ArrowButton ("d", 0.25f, Colours::white.withAlpha (0.8f));
}

ImageEffectFilter* OldSchoolLookAndFeel::getSliderEffect()
{
	return &scrollbarShadow;
}

int OldSchoolLookAndFeel::getSliderThumbRadius (Slider&)
{
	return 8;
}

void OldSchoolLookAndFeel::drawCornerResizer (Graphics& g,
											  int w, int h,
											  bool isMouseOver,
											  bool isMouseDragging)
{
	g.setColour ((isMouseOver || isMouseDragging) ? Colours::lightgrey
												  : Colours::darkgrey);

	const float lineThickness = jmin (w, h) * 0.1f;

	for (float i = 0.0f; i < 1.0f; i += 0.3f)
	{
		g.drawLine (w * i,
					h + 1.0f,
					w + 1.0f,
					h * i,
					lineThickness);
	}
}

Button* OldSchoolLookAndFeel::createDocumentWindowButton (int buttonType)
{
	Path shape;

	if (buttonType == DocumentWindow::closeButton)
	{
		shape.addLineSegment (Line<float> (0.0f, 0.0f, 1.0f, 1.0f), 0.35f);
		shape.addLineSegment (Line<float> (1.0f, 0.0f, 0.0f, 1.0f), 0.35f);

		ShapeButton* const b = new ShapeButton ("close",
												Colour (0x7fff3333),
												Colour (0xd7ff3333),
												Colour (0xf7ff3333));

		b->setShape (shape, true, true, true);
		return b;
	}
	else if (buttonType == DocumentWindow::minimiseButton)
	{
		shape.addLineSegment (Line<float> (0.0f, 0.5f, 1.0f, 0.5f), 0.25f);

		DrawableButton* b = new DrawableButton ("minimise", DrawableButton::ImageFitted);
		DrawablePath dp;
		dp.setPath (shape);
		dp.setFill (Colours::black.withAlpha (0.3f));
		b->setImages (&dp);
		return b;
	}
	else if (buttonType == DocumentWindow::maximiseButton)
	{
		shape.addLineSegment (Line<float> (0.5f, 0.0f, 0.5f, 1.0f), 0.25f);
		shape.addLineSegment (Line<float> (0.0f, 0.5f, 1.0f, 0.5f), 0.25f);

		DrawableButton* b = new DrawableButton ("maximise", DrawableButton::ImageFitted);
		DrawablePath dp;
		dp.setPath (shape);
		dp.setFill (Colours::black.withAlpha (0.3f));
		b->setImages (&dp);
		return b;
	}

	jassertfalse;
	return nullptr;
}

void OldSchoolLookAndFeel::positionDocumentWindowButtons (DocumentWindow&,
														  int titleBarX,
														  int titleBarY,
														  int titleBarW,
														  int titleBarH,
														  Button* minimiseButton,
														  Button* maximiseButton,
														  Button* closeButton,
														  bool positionTitleBarButtonsOnLeft)
{
	titleBarY += titleBarH / 8;
	titleBarH -= titleBarH / 4;

	const int buttonW = titleBarH;

	int x = positionTitleBarButtonsOnLeft ? titleBarX + 4
										  : titleBarX + titleBarW - buttonW - 4;

	if (closeButton != nullptr)
	{
		closeButton->setBounds (x, titleBarY, buttonW, titleBarH);
		x += positionTitleBarButtonsOnLeft ? buttonW + buttonW / 5
										   : -(buttonW + buttonW / 5);
	}

	if (positionTitleBarButtonsOnLeft)
		std::swap (minimiseButton, maximiseButton);

	if (maximiseButton != nullptr)
	{
		maximiseButton->setBounds (x, titleBarY - 2, buttonW, titleBarH);
		x += positionTitleBarButtonsOnLeft ? buttonW : -buttonW;
	}

	if (minimiseButton != nullptr)
		minimiseButton->setBounds (x, titleBarY - 2, buttonW, titleBarH);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_OldSchoolLookAndFeel.cpp ***/


/*** Start of inlined file: juce_MenuBarComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

MenuBarComponent::MenuBarComponent (MenuBarModel* model_)
	: model (nullptr),
	  itemUnderMouse (-1),
	  currentPopupIndex (-1),
	  topLevelIndexClicked (0),
	  lastMouseX (0),
	  lastMouseY (0)
{
	setRepaintsOnMouseActivity (true);
	setWantsKeyboardFocus (false);
	setMouseClickGrabsKeyboardFocus (false);

	setModel (model_);
}

MenuBarComponent::~MenuBarComponent()
{
	setModel (nullptr);
	Desktop::getInstance().removeGlobalMouseListener (this);
}

MenuBarModel* MenuBarComponent::getModel() const noexcept
{
	return model;
}

void MenuBarComponent::setModel (MenuBarModel* const newModel)
{
	if (model != newModel)
	{
		if (model != nullptr)
			model->removeListener (this);

		model = newModel;

		if (model != nullptr)
			model->addListener (this);

		repaint();
		menuBarItemsChanged (nullptr);
	}
}

void MenuBarComponent::paint (Graphics& g)
{
	const bool isMouseOverBar = currentPopupIndex >= 0 || itemUnderMouse >= 0 || isMouseOver();

	getLookAndFeel().drawMenuBarBackground (g,
											getWidth(),
											getHeight(),
											isMouseOverBar,
											*this);

	if (model != nullptr)
	{
		for (int i = 0; i < menuNames.size(); ++i)
		{
			Graphics::ScopedSaveState ss (g);

			g.setOrigin (xPositions [i], 0);
			g.reduceClipRegion (0, 0, xPositions[i + 1] - xPositions[i], getHeight());

			getLookAndFeel().drawMenuBarItem (g,
											  xPositions[i + 1] - xPositions[i],
											  getHeight(),
											  i,
											  menuNames[i],
											  i == itemUnderMouse,
											  i == currentPopupIndex,
											  isMouseOverBar,
											  *this);
		}
	}
}

void MenuBarComponent::resized()
{
	xPositions.clear();
	int x = 0;
	xPositions.add (x);

	for (int i = 0; i < menuNames.size(); ++i)
	{
		x += getLookAndFeel().getMenuBarItemWidth (*this, i, menuNames[i]);
		xPositions.add (x);
	}
}

int MenuBarComponent::getItemAt (const int x, const int y)
{
	for (int i = 0; i < xPositions.size(); ++i)
		if (x >= xPositions[i] && x < xPositions[i + 1])
			return reallyContains (Point<int> (x, y), true) ? i : -1;

	return -1;
}

void MenuBarComponent::repaintMenuItem (int index)
{
	if (isPositiveAndBelow (index, xPositions.size()))
	{
		const int x1 = xPositions [index];
		const int x2 = xPositions [index + 1];

		repaint (x1 - 2, 0, x2 - x1 + 4, getHeight());
	}
}

void MenuBarComponent::setItemUnderMouse (const int index)
{
	if (itemUnderMouse != index)
	{
		repaintMenuItem (itemUnderMouse);
		itemUnderMouse = index;
		repaintMenuItem (itemUnderMouse);
	}
}

void MenuBarComponent::setOpenItem (int index)
{
	if (currentPopupIndex != index)
	{
		repaintMenuItem (currentPopupIndex);
		currentPopupIndex = index;
		repaintMenuItem (currentPopupIndex);

		if (index >= 0)
			Desktop::getInstance().addGlobalMouseListener (this);
		else
			Desktop::getInstance().removeGlobalMouseListener (this);
	}
}

void MenuBarComponent::updateItemUnderMouse (int x, int y)
{
	setItemUnderMouse (getItemAt (x, y));
}

void MenuBarComponent::showMenu (int index)
{
	if (index != currentPopupIndex)
	{
		PopupMenu::dismissAllActiveMenus();
		menuBarItemsChanged (nullptr);

		setOpenItem (index);
		setItemUnderMouse (index);

		if (index >= 0)
		{
			PopupMenu m (model->getMenuForIndex (itemUnderMouse,
												 menuNames [itemUnderMouse]));

			if (m.lookAndFeel == nullptr)
				m.setLookAndFeel (&getLookAndFeel());

			const Rectangle<int> itemPos (xPositions [index], 0, xPositions [index + 1] - xPositions [index], getHeight());

			m.showMenuAsync (PopupMenu::Options().withTargetComponent (this)
												 .withTargetScreenArea (localAreaToGlobal (itemPos))
												 .withMinimumWidth (itemPos.getWidth()),
							 ModalCallbackFunction::forComponent (menuBarMenuDismissedCallback, this, index));
		}
	}
}

void MenuBarComponent::menuBarMenuDismissedCallback (int result, MenuBarComponent* bar, int topLevelIndex)
{
	if (bar != nullptr)
		bar->menuDismissed (topLevelIndex, result);
}

void MenuBarComponent::menuDismissed (int topLevelIndex, int itemId)
{
	topLevelIndexClicked = topLevelIndex;
	postCommandMessage (itemId);
}

void MenuBarComponent::handleCommandMessage (int commandId)
{
	const Point<int> mousePos (getMouseXYRelative());
	updateItemUnderMouse (mousePos.getX(), mousePos.getY());

	if (currentPopupIndex == topLevelIndexClicked)
		setOpenItem (-1);

	if (commandId != 0 && model != nullptr)
		model->menuItemSelected (commandId, topLevelIndexClicked);
}

void MenuBarComponent::mouseEnter (const MouseEvent& e)
{
	if (e.eventComponent == this)
		updateItemUnderMouse (e.x, e.y);
}

void MenuBarComponent::mouseExit (const MouseEvent& e)
{
	if (e.eventComponent == this)
		updateItemUnderMouse (e.x, e.y);
}

void MenuBarComponent::mouseDown (const MouseEvent& e)
{
	if (currentPopupIndex < 0)
	{
		const MouseEvent e2 (e.getEventRelativeTo (this));
		updateItemUnderMouse (e2.x, e2.y);

		currentPopupIndex = -2;
		showMenu (itemUnderMouse);
	}
}

void MenuBarComponent::mouseDrag (const MouseEvent& e)
{
	const MouseEvent e2 (e.getEventRelativeTo (this));
	const int item = getItemAt (e2.x, e2.y);

	if (item >= 0)
		showMenu (item);
}

void MenuBarComponent::mouseUp (const MouseEvent& e)
{
	const MouseEvent e2 (e.getEventRelativeTo (this));

	updateItemUnderMouse (e2.x, e2.y);

	if (itemUnderMouse < 0 && getLocalBounds().contains (e2.x, e2.y))
	{
		setOpenItem (-1);
		PopupMenu::dismissAllActiveMenus();
	}
}

void MenuBarComponent::mouseMove (const MouseEvent& e)
{
	const MouseEvent e2 (e.getEventRelativeTo (this));

	if (lastMouseX != e2.x || lastMouseY != e2.y)
	{
		if (currentPopupIndex >= 0)
		{
			const int item = getItemAt (e2.x, e2.y);

			if (item >= 0)
				showMenu (item);
		}
		else
		{
			updateItemUnderMouse (e2.x, e2.y);
		}

		lastMouseX = e2.x;
		lastMouseY = e2.y;
	}
}

bool MenuBarComponent::keyPressed (const KeyPress& key)
{
	bool used = false;
	const int numMenus = menuNames.size();
	const int currentIndex = jlimit (0, menuNames.size() - 1, currentPopupIndex);

	if (key.isKeyCode (KeyPress::leftKey))
	{
		showMenu ((currentIndex + numMenus - 1) % numMenus);
		used = true;
	}
	else if (key.isKeyCode (KeyPress::rightKey))
	{
		showMenu ((currentIndex + 1) % numMenus);
		used = true;
	}

	return used;
}

void MenuBarComponent::menuBarItemsChanged (MenuBarModel* /*menuBarModel*/)
{
	StringArray newNames;

	if (model != nullptr)
		newNames = model->getMenuBarNames();

	if (newNames != menuNames)
	{
		menuNames = newNames;
		repaint();
		resized();
	}
}

void MenuBarComponent::menuCommandInvoked (MenuBarModel* /*menuBarModel*/,
										   const ApplicationCommandTarget::InvocationInfo& info)
{
	if (model == nullptr || (info.commandFlags & ApplicationCommandInfo::dontTriggerVisualFeedback) != 0)
		return;

	for (int i = 0; i < menuNames.size(); ++i)
	{
		const PopupMenu menu (model->getMenuForIndex (i, menuNames [i]));

		if (menu.containsCommandItem (info.commandID))
		{
			setItemUnderMouse (i);
			startTimer (200);
			break;
		}
	}
}

void MenuBarComponent::timerCallback()
{
	stopTimer();

	const Point<int> mousePos (getMouseXYRelative());
	updateItemUnderMouse (mousePos.getX(), mousePos.getY());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MenuBarComponent.cpp ***/


/*** Start of inlined file: juce_MenuBarModel.cpp ***/
BEGIN_JUCE_NAMESPACE

MenuBarModel::MenuBarModel() noexcept
	: manager (nullptr)
{
}

MenuBarModel::~MenuBarModel()
{
	setApplicationCommandManagerToWatch (nullptr);
}

void MenuBarModel::menuItemsChanged()
{
	triggerAsyncUpdate();
}

void MenuBarModel::setApplicationCommandManagerToWatch (ApplicationCommandManager* const newManager) noexcept
{
	if (manager != newManager)
	{
		if (manager != nullptr)
			manager->removeListener (this);

		manager = newManager;

		if (manager != nullptr)
			manager->addListener (this);
	}
}

void MenuBarModel::addListener (Listener* const newListener) noexcept
{
	listeners.add (newListener);
}

void MenuBarModel::removeListener (Listener* const listenerToRemove) noexcept
{
	// Trying to remove a listener that isn't on the list!
	// If this assertion happens because this object is a dangling pointer, make sure you've not
	// deleted this menu model while it's still being used by something (e.g. by a MenuBarComponent)
	jassert (listeners.contains (listenerToRemove));

	listeners.remove (listenerToRemove);
}

void MenuBarModel::handleAsyncUpdate()
{
	listeners.call (&MenuBarModel::Listener::menuBarItemsChanged, this);
}

void MenuBarModel::applicationCommandInvoked (const ApplicationCommandTarget::InvocationInfo& info)
{
	listeners.call (&MenuBarModel::Listener::menuCommandInvoked, this, info);
}

void MenuBarModel::applicationCommandListChanged()
{
	menuItemsChanged();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MenuBarModel.cpp ***/


/*** Start of inlined file: juce_PopupMenu.cpp ***/
BEGIN_JUCE_NAMESPACE

class PopupMenu::Item
{
public:

	Item()
	  : itemId (0), active (true), isSeparator (true), isTicked (false),
		usesColour (false), commandManager (nullptr)
	{
	}

	Item (const int itemId_,
		  const String& text_,
		  const bool active_,
		  const bool isTicked_,
		  const Image& im,
		  const Colour& textColour_,
		  const bool usesColour_,
		  CustomComponent* const customComp_,
		  const PopupMenu* const subMenu_,
		  ApplicationCommandManager* const commandManager_)
	  : itemId (itemId_), text (text_), textColour (textColour_),
		active (active_), isSeparator (false), isTicked (isTicked_),
		usesColour (usesColour_), image (im), customComp (customComp_),
		commandManager (commandManager_)
	{
		if (subMenu_ != nullptr)
			subMenu = new PopupMenu (*subMenu_);

		if (commandManager_ != nullptr && itemId_ != 0)
		{
			String shortcutKey;

			Array <KeyPress> keyPresses (commandManager_->getKeyMappings()
											->getKeyPressesAssignedToCommand (itemId_));

			for (int i = 0; i < keyPresses.size(); ++i)
			{
				const String key (keyPresses.getReference(i).getTextDescriptionWithIcons());

				if (shortcutKey.isNotEmpty())
					shortcutKey << ", ";

				if (key.length() == 1 && key[0] < 128)
					shortcutKey << "shortcut: '" << key << '\'';
				else
					shortcutKey << key;
			}

			shortcutKey = shortcutKey.trim();

			if (shortcutKey.isNotEmpty())
				text << "<end>" << shortcutKey;
		}
	}

	Item (const Item& other)
		: itemId (other.itemId),
		  text (other.text),
		  textColour (other.textColour),
		  active (other.active),
		  isSeparator (other.isSeparator),
		  isTicked (other.isTicked),
		  usesColour (other.usesColour),
		  image (other.image),
		  customComp (other.customComp),
		  commandManager (other.commandManager)
	{
		if (other.subMenu != nullptr)
			subMenu = new PopupMenu (*(other.subMenu));
	}

	bool canBeTriggered() const noexcept	{ return active && ! (isSeparator || (subMenu != nullptr)); }
	bool hasActiveSubMenu() const noexcept  { return active && (subMenu != nullptr); }

	const int itemId;
	String text;
	const Colour textColour;
	const bool active, isSeparator, isTicked, usesColour;
	Image image;
	ReferenceCountedObjectPtr <CustomComponent> customComp;
	ScopedPointer <PopupMenu> subMenu;
	ApplicationCommandManager* const commandManager;

private:
	Item& operator= (const Item&);

	JUCE_LEAK_DETECTOR (Item);
};

class PopupMenu::ItemComponent  : public Component
{
public:

	ItemComponent (const PopupMenu::Item& itemInfo_, int standardItemHeight, Component* const parent)
	  : itemInfo (itemInfo_),
		isHighlighted (false)
	{
		if (itemInfo.customComp != nullptr)
			addAndMakeVisible (itemInfo.customComp);

		parent->addAndMakeVisible (this);

		int itemW = 80;
		int itemH = 16;
		getIdealSize (itemW, itemH, standardItemHeight);
		setSize (itemW, jlimit (2, 600, itemH));

		addMouseListener (parent, false);
	}

	~ItemComponent()
	{
		if (itemInfo.customComp != nullptr)
			removeChildComponent (itemInfo.customComp);
	}

	void getIdealSize (int& idealWidth, int& idealHeight, const int standardItemHeight)
	{
		if (itemInfo.customComp != nullptr)
			itemInfo.customComp->getIdealSize (idealWidth, idealHeight);
		else
			getLookAndFeel().getIdealPopupMenuItemSize (itemInfo.text,
														itemInfo.isSeparator,
														standardItemHeight,
														idealWidth, idealHeight);
	}

	void paint (Graphics& g)
	{
		if (itemInfo.customComp == nullptr)
		{
			String mainText (itemInfo.text);
			String endText;
			const int endIndex = mainText.indexOf ("<end>");

			if (endIndex >= 0)
			{
				endText = mainText.substring (endIndex + 5).trim();
				mainText = mainText.substring (0, endIndex);
			}

			getLookAndFeel()
				.drawPopupMenuItem (g, getWidth(), getHeight(),
									itemInfo.isSeparator,
									itemInfo.active,
									isHighlighted,
									itemInfo.isTicked,
									itemInfo.subMenu != 0,
									mainText, endText,
									itemInfo.image.isValid() ? &itemInfo.image : 0,
									itemInfo.usesColour ? &(itemInfo.textColour) : 0);
		}
	}

	void resized()
	{
		if (getNumChildComponents() > 0)
			getChildComponent(0)->setBounds (2, 0, getWidth() - 4, getHeight());
	}

	void setHighlighted (bool shouldBeHighlighted)
	{
		shouldBeHighlighted = shouldBeHighlighted && itemInfo.active;

		if (isHighlighted != shouldBeHighlighted)
		{
			isHighlighted = shouldBeHighlighted;

			if (itemInfo.customComp != nullptr)
				itemInfo.customComp->setHighlighted (shouldBeHighlighted);

			repaint();
		}
	}

	PopupMenu::Item itemInfo;

private:
	bool isHighlighted;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ItemComponent);
};

namespace PopupMenuSettings
{
	const int scrollZone = 24;
	const int borderSize = 2;
	const int timerInterval = 50;
	const int dismissCommandId = 0x6287345f;

	static bool menuWasHiddenBecauseOfAppChange = false;
}

class PopupMenu::Window  : public Component,
						   private Timer
{
public:

	Window (const PopupMenu& menu, Window* const owner_, const Rectangle<int>& target,
			const bool alignToRectangle, const int itemIdThatMustBeVisible,
			const int minimumWidth_, const int maximumNumColumns_,
			const int standardItemHeight_, const bool dismissOnMouseUp_,
			ApplicationCommandManager** const managerOfChosenCommand_,
			Component* const componentAttachedTo_)
	   : Component ("menu"),
		 owner (owner_),
		 activeSubMenu (nullptr),
		 managerOfChosenCommand (managerOfChosenCommand_),
		 componentAttachedTo (componentAttachedTo_),
		 componentAttachedToOriginal (componentAttachedTo_),
		 minimumWidth (minimumWidth_),
		 maximumNumColumns (maximumNumColumns_),
		 standardItemHeight (standardItemHeight_),
		 isOver (false),
		 hasBeenOver (false),
		 isDown (false),
		 needsToScroll (false),
		 dismissOnMouseUp (dismissOnMouseUp_),
		 hideOnExit (false),
		 disableMouseMoves (false),
		 hasAnyJuceCompHadFocus (false),
		 numColumns (0),
		 contentHeight (0),
		 childYOffset (0),
		 menuCreationTime (Time::getMillisecondCounter()),
		 lastMouseMoveTime (0),
		 timeEnteredCurrentChildComp (0),
		 scrollAcceleration (1.0)
	{
		lastFocused = lastScroll = menuCreationTime;
		setWantsKeyboardFocus (false);
		setMouseClickGrabsKeyboardFocus (false);
		setAlwaysOnTop (true);

		setLookAndFeel (menu.lookAndFeel);
		setOpaque (getLookAndFeel().findColour (PopupMenu::backgroundColourId).isOpaque() || ! Desktop::canUseSemiTransparentWindows());

		for (int i = 0; i < menu.items.size(); ++i)
			items.add (new PopupMenu::ItemComponent (*menu.items.getUnchecked(i), standardItemHeight, this));

		calculateWindowPos (target, alignToRectangle);
		setTopLeftPosition (windowPos.getX(), windowPos.getY());
		updateYPositions();

		if (itemIdThatMustBeVisible != 0)
		{
			const int y = target.getY() - windowPos.getY();
			ensureItemIsVisible (itemIdThatMustBeVisible,
								 isPositiveAndBelow (y, windowPos.getHeight()) ? y : -1);
		}

		resizeToBestWindowPos();
		addToDesktop (ComponentPeer::windowIsTemporary
					   | ComponentPeer::windowIgnoresKeyPresses
					   | getLookAndFeel().getMenuWindowFlags());

		getActiveWindows().add (this);
		Desktop::getInstance().addGlobalMouseListener (this);
	}

	~Window()
	{
		getActiveWindows().removeValue (this);
		Desktop::getInstance().removeGlobalMouseListener (this);
		activeSubMenu = nullptr;
		items.clear();
	}

	static Window* create (const PopupMenu& menu,
						   bool dismissOnMouseUp,
						   Window* const owner_,
						   const Rectangle<int>& target,
						   int minimumWidth,
						   int maximumNumColumns,
						   int standardItemHeight,
						   bool alignToRectangle,
						   int itemIdThatMustBeVisible,
						   ApplicationCommandManager** managerOfChosenCommand,
						   Component* componentAttachedTo)
	{
		if (menu.items.size() > 0)
			return new Window (menu, owner_, target, alignToRectangle, itemIdThatMustBeVisible,
							   minimumWidth, maximumNumColumns, standardItemHeight, dismissOnMouseUp,
							   managerOfChosenCommand, componentAttachedTo);

		return nullptr;
	}

	void paint (Graphics& g)
	{
		if (isOpaque())
			g.fillAll (Colours::white);

		getLookAndFeel().drawPopupMenuBackground (g, getWidth(), getHeight());
	}

	void paintOverChildren (Graphics& g)
	{
		if (isScrolling())
		{
			LookAndFeel& lf = getLookAndFeel();

			if (isScrollZoneActive (false))
				lf.drawPopupMenuUpDownArrow (g, getWidth(), PopupMenuSettings::scrollZone, true);

			if (isScrollZoneActive (true))
			{
				g.setOrigin (0, getHeight() - PopupMenuSettings::scrollZone);
				lf.drawPopupMenuUpDownArrow (g, getWidth(), PopupMenuSettings::scrollZone, false);
			}
		}
	}

	bool isScrollZoneActive (bool bottomOne) const
	{
		return isScrolling()
				&& (bottomOne ? childYOffset < contentHeight - windowPos.getHeight()
							  : childYOffset > 0);
	}

	// hide this and all sub-comps
	void hide (const PopupMenu::Item* const item, const bool makeInvisible)
	{
		if (isVisible())
		{
			WeakReference<Component> deletionChecker (this);

			activeSubMenu = nullptr;
			currentChild = nullptr;

			if (item != nullptr
				 && item->commandManager != nullptr
				 && item->itemId != 0)
			{
				*managerOfChosenCommand = item->commandManager;
			}

			exitModalState (item != nullptr ? item->itemId : 0);

			if (makeInvisible && (deletionChecker != nullptr))
				setVisible (false);
		}
	}

	void dismissMenu (const PopupMenu::Item* const item)
	{
		if (owner != nullptr)
		{
			owner->dismissMenu (item);
		}
		else
		{
			if (item != nullptr)
			{
				// need a copy of this on the stack as the one passed in will get deleted during this call
				const PopupMenu::Item mi (*item);
				hide (&mi, false);
			}
			else
			{
				hide (0, false);
			}
		}
	}

	void mouseMove (const MouseEvent&)	{ timerCallback(); }
	void mouseDown (const MouseEvent&)	{ timerCallback(); }
	void mouseDrag (const MouseEvent&)	{ timerCallback(); }
	void mouseUp   (const MouseEvent&)	{ timerCallback(); }

	void mouseWheelMove (const MouseEvent&, float /*amountX*/, float amountY)
	{
		alterChildYPos (roundToInt (-10.0f * amountY * PopupMenuSettings::scrollZone));
		lastMouse = Point<int> (-1, -1);
	}

	bool keyPressed (const KeyPress& key)
	{
		if (key.isKeyCode (KeyPress::downKey))
		{
			selectNextItem (1);
		}
		else if (key.isKeyCode (KeyPress::upKey))
		{
			selectNextItem (-1);
		}
		else if (key.isKeyCode (KeyPress::leftKey))
		{
			if (owner != nullptr)
			{
				Component::SafePointer<Window> parentWindow (owner);
				PopupMenu::ItemComponent* currentChildOfParent = parentWindow->currentChild;

				hide (0, true);

				if (parentWindow != nullptr)
					parentWindow->setCurrentlyHighlightedChild (currentChildOfParent);

				disableTimerUntilMouseMoves();
			}
			else if (componentAttachedTo != nullptr)
			{
				componentAttachedTo->keyPressed (key);
			}
		}
		else if (key.isKeyCode (KeyPress::rightKey))
		{
			disableTimerUntilMouseMoves();

			if (showSubMenuFor (currentChild))
			{
				if (activeSubMenu != nullptr && activeSubMenu->isVisible())
					activeSubMenu->selectNextItem (1);
			}
			else if (componentAttachedTo != nullptr)
			{
				componentAttachedTo->keyPressed (key);
			}
		}
		else if (key.isKeyCode (KeyPress::returnKey))
		{
			triggerCurrentlyHighlightedItem();
		}
		else if (key.isKeyCode (KeyPress::escapeKey))
		{
			dismissMenu (0);
		}
		else
		{
			return false;
		}

		return true;
	}

	void inputAttemptWhenModal()
	{
		WeakReference<Component> deletionChecker (this);

		timerCallback();

		if (deletionChecker != nullptr && ! isOverAnyMenu())
		{
			if (componentAttachedTo != nullptr)
			{
				// we want to dismiss the menu, but if we do it synchronously, then
				// the mouse-click will be allowed to pass through. That's good, except
				// when the user clicks on the button that orginally popped the menu up,
				// as they'll expect the menu to go away, and in fact it'll just
				// come back. So only dismiss synchronously if they're not on the original
				// comp that we're attached to.
				const Point<int> mousePos (componentAttachedTo->getMouseXYRelative());

				if (componentAttachedTo->reallyContains (mousePos, true))
				{
					postCommandMessage (PopupMenuSettings::dismissCommandId); // dismiss asynchrounously
					return;
				}
			}

			dismissMenu (0);
		}
	}

	void handleCommandMessage (int commandId)
	{
		Component::handleCommandMessage (commandId);

		if (commandId == PopupMenuSettings::dismissCommandId)
			dismissMenu (0);
	}

	void timerCallback()
	{
		if (! isVisible())
			return;

		if (componentAttachedTo != componentAttachedToOriginal)
		{
			dismissMenu (0);
			return;
		}

		Window* currentlyModalWindow = dynamic_cast <Window*> (Component::getCurrentlyModalComponent());

		if (currentlyModalWindow != nullptr && ! treeContains (currentlyModalWindow))
			return;

		startTimer (PopupMenuSettings::timerInterval);  // do this in case it was called from a mouse
														// move rather than a real timer callback

		const Point<int> globalMousePos (Desktop::getMousePosition());
		const Point<int> localMousePos (getLocalPoint (nullptr, globalMousePos));

		const uint32 now = Time::getMillisecondCounter();

		if (now > timeEnteredCurrentChildComp + 100
			 && reallyContains (localMousePos, true)
			 && currentChild != nullptr
			 && (! disableMouseMoves)
			 && ! (activeSubMenu != nullptr && activeSubMenu->isVisible()))
		{
			showSubMenuFor (currentChild);
		}

		if (globalMousePos != lastMouse || now > lastMouseMoveTime + 350)
		{
			highlightItemUnderMouse (globalMousePos, localMousePos);
		}

		bool overScrollArea = false;

		if (isScrolling()
			 && (isOver || (isDown && isPositiveAndBelow (localMousePos.getX(), getWidth())))
			 && ((isScrollZoneActive (false) && localMousePos.getY() < PopupMenuSettings::scrollZone)
				  || (isScrollZoneActive (true) && localMousePos.getY() > getHeight() - PopupMenuSettings::scrollZone)))
		{
			if (now > lastScroll + 20)
			{
				scrollAcceleration = jmin (4.0, scrollAcceleration * 1.04);
				int amount = 0;

				for (int i = 0; i < items.size() && amount == 0; ++i)
					amount = ((int) scrollAcceleration) * items.getUnchecked(i)->getHeight();

				alterChildYPos (localMousePos.getY() < PopupMenuSettings::scrollZone ? -amount : amount);

				lastScroll = now;
			}

			overScrollArea = true;
			lastMouse = Point<int> (-1, -1); // trigger a mouse-move
		}
		else
		{
			scrollAcceleration = 1.0;
		}

		const bool wasDown = isDown;
		bool isOverAny = isOverAnyMenu();

		if (hideOnExit && hasBeenOver && (! isOverAny) && activeSubMenu != nullptr)
		{
			activeSubMenu->updateMouseOverStatus (globalMousePos);
			isOverAny = isOverAnyMenu();
		}

		if (hideOnExit && hasBeenOver && ! isOverAny)
		{
			hide (0, true);
		}
		else
		{
			isDown = hasBeenOver
						&& (ModifierKeys::getCurrentModifiers().isAnyMouseButtonDown()
							|| ModifierKeys::getCurrentModifiersRealtime().isAnyMouseButtonDown());

			bool anyFocused = Process::isForegroundProcess();

			if (anyFocused && Component::getCurrentlyFocusedComponent() == nullptr)
			{
				// because no component at all may have focus, our test here will
				// only be triggered when something has focus and then loses it.
				anyFocused = ! hasAnyJuceCompHadFocus;

				for (int i = ComponentPeer::getNumPeers(); --i >= 0;)
				{
					if (ComponentPeer::getPeer (i)->isFocused())
					{
						anyFocused = true;
						hasAnyJuceCompHadFocus = true;
						break;
					}
				}
			}

			if (! anyFocused)
			{
				if (now > lastFocused + 10)
				{
					PopupMenuSettings::menuWasHiddenBecauseOfAppChange = true;
					dismissMenu (0);

					return;  // may have been deleted by the previous call..
				}
			}
			else if (wasDown && now > menuCreationTime + 250
					   && ! (isDown || overScrollArea))
			{
				isOver = reallyContains (localMousePos, true);

				if (isOver)
				{
					triggerCurrentlyHighlightedItem();
				}
				else if ((hasBeenOver || ! dismissOnMouseUp) && ! isOverAny)
				{
					dismissMenu (0);
				}

				return;  // may have been deleted by the previous calls..
			}
			else
			{
				lastFocused = now;
			}
		}
	}

	static Array<Window*>& getActiveWindows()
	{
		static Array<Window*> activeMenuWindows;
		return activeMenuWindows;
	}

private:
	Window* owner;
	OwnedArray <PopupMenu::ItemComponent> items;
	Component::SafePointer<PopupMenu::ItemComponent> currentChild;
	ScopedPointer <Window> activeSubMenu;
	ApplicationCommandManager** managerOfChosenCommand;
	WeakReference<Component> componentAttachedTo;
	Component* componentAttachedToOriginal;
	Rectangle<int> windowPos;
	Point<int> lastMouse;
	int minimumWidth, maximumNumColumns, standardItemHeight;
	bool isOver, hasBeenOver, isDown, needsToScroll;
	bool dismissOnMouseUp, hideOnExit, disableMouseMoves, hasAnyJuceCompHadFocus;
	int numColumns, contentHeight, childYOffset;
	Array <int> columnWidths;
	uint32 menuCreationTime, lastFocused, lastScroll, lastMouseMoveTime, timeEnteredCurrentChildComp;
	double scrollAcceleration;

	bool overlaps (const Rectangle<int>& r) const
	{
		return r.intersects (getBounds())
				|| (owner != nullptr && owner->overlaps (r));
	}

	bool isOverAnyMenu() const
	{
		return owner != nullptr ? owner->isOverAnyMenu()
								: isOverChildren();
	}

	bool isOverChildren() const
	{
		return isVisible()
				&& (isOver || (activeSubMenu != nullptr && activeSubMenu->isOverChildren()));
	}

	void updateMouseOverStatus (const Point<int>& globalMousePos)
	{
		const Point<int> relPos (getLocalPoint (nullptr, globalMousePos));
		isOver = reallyContains (relPos, true);

		if (activeSubMenu != nullptr)
			activeSubMenu->updateMouseOverStatus (globalMousePos);
	}

	bool treeContains (const Window* const window) const noexcept
	{
		const Window* mw = this;

		while (mw->owner != nullptr)
			mw = mw->owner;

		while (mw != nullptr)
		{
			if (mw == window)
				return true;

			mw = mw->activeSubMenu;
		}

		return false;
	}

	void calculateWindowPos (const Rectangle<int>& target, const bool alignToRectangle)
	{
		const Rectangle<int> mon (Desktop::getInstance()
									 .getMonitorAreaContaining (target.getCentre(),
															   #if JUCE_MAC
																true));
															   #else
																false)); // on windows, don't stop the menu overlapping the taskbar
															   #endif

		const int maxMenuHeight = mon.getHeight() - 24;

		int x, y, widthToUse, heightToUse;
		layoutMenuItems (mon.getWidth() - 24, maxMenuHeight, widthToUse, heightToUse);

		if (alignToRectangle)
		{
			x = target.getX();

			const int spaceUnder = mon.getHeight() - (target.getBottom() - mon.getY());
			const int spaceOver = target.getY() - mon.getY();

			if (heightToUse < spaceUnder - 30 || spaceUnder >= spaceOver)
				y = target.getBottom();
			else
				y = target.getY() - heightToUse;
		}
		else
		{
			bool tendTowardsRight = target.getCentreX() < mon.getCentreX();

			if (owner != nullptr)
			{
				if (owner->owner != nullptr)
				{
					const bool ownerGoingRight = (owner->getX() + owner->getWidth() / 2
													> owner->owner->getX() + owner->owner->getWidth() / 2);

					if (ownerGoingRight && target.getRight() + widthToUse < mon.getRight() - 4)
						tendTowardsRight = true;
					else if ((! ownerGoingRight) && target.getX() > widthToUse + 4)
						tendTowardsRight = false;
				}
				else if (target.getRight() + widthToUse < mon.getRight() - 32)
				{
					tendTowardsRight = true;
				}
			}

			const int biggestSpace = jmax (mon.getRight() - target.getRight(),
										   target.getX() - mon.getX()) - 32;

			if (biggestSpace < widthToUse)
			{
				layoutMenuItems (biggestSpace + target.getWidth() / 3, maxMenuHeight, widthToUse, heightToUse);

				if (numColumns > 1)
					layoutMenuItems (biggestSpace - 4, maxMenuHeight, widthToUse, heightToUse);

				tendTowardsRight = (mon.getRight() - target.getRight()) >= (target.getX() - mon.getX());
			}

			if (tendTowardsRight)
				x = jmin (mon.getRight() - widthToUse - 4, target.getRight());
			else
				x = jmax (mon.getX() + 4, target.getX() - widthToUse);

			y = target.getY();
			if (target.getCentreY() > mon.getCentreY())
				y = jmax (mon.getY(), target.getBottom() - heightToUse);
		}

		x = jmax (mon.getX() + 1, jmin (mon.getRight() - (widthToUse + 6), x));
		y = jmax (mon.getY() + 1, jmin (mon.getBottom() - (heightToUse + 6), y));

		windowPos.setBounds (x, y, widthToUse, heightToUse);

		// sets this flag if it's big enough to obscure any of its parent menus
		hideOnExit = owner != nullptr
					  && owner->windowPos.intersects (windowPos.expanded (-4, -4));
	}

	void layoutMenuItems (const int maxMenuW, const int maxMenuH, int& width, int& height)
	{
		numColumns = 0;
		contentHeight = 0;
		int totalW;

		do
		{
			++numColumns;
			totalW = workOutBestSize (maxMenuW);

			if (totalW > maxMenuW)
			{
				numColumns = jmax (1, numColumns - 1);
				totalW = workOutBestSize (maxMenuW); // to update col widths
				break;
			}
			else if (totalW > maxMenuW / 2 || contentHeight < maxMenuH)
			{
				break;
			}

		} while (numColumns < maximumNumColumns);

		const int actualH = jmin (contentHeight, maxMenuH);

		needsToScroll = contentHeight > actualH;

		width = updateYPositions();
		height = actualH + PopupMenuSettings::borderSize * 2;
	}

	int workOutBestSize (const int maxMenuW)
	{
		int totalW = 0;
		contentHeight = 0;
		int childNum = 0;

		for (int col = 0; col < numColumns; ++col)
		{
			int i, colW = standardItemHeight, colH = 0;

			const int numChildren = jmin (items.size() - childNum,
										  (items.size() + numColumns - 1) / numColumns);

			for (i = numChildren; --i >= 0;)
			{
				colW = jmax (colW, items.getUnchecked (childNum + i)->getWidth());
				colH += items.getUnchecked (childNum + i)->getHeight();
			}

			colW = jmin (maxMenuW / jmax (1, numColumns - 2), colW + PopupMenuSettings::borderSize * 2);

			columnWidths.set (col, colW);
			totalW += colW;
			contentHeight = jmax (contentHeight, colH);

			childNum += numChildren;
		}

		if (totalW < minimumWidth)
		{
			totalW = minimumWidth;

			for (int col = 0; col < numColumns; ++col)
				columnWidths.set (0, totalW / numColumns);
		}

		return totalW;
	}

	void ensureItemIsVisible (const int itemId, int wantedY)
	{
		jassert (itemId != 0)

		for (int i = items.size(); --i >= 0;)
		{
			PopupMenu::ItemComponent* const m = items.getUnchecked(i);

			if (m != nullptr
				&& m->itemInfo.itemId == itemId
				&& windowPos.getHeight() > PopupMenuSettings::scrollZone * 4)
			{
				const int currentY = m->getY();

				if (wantedY > 0 || currentY < 0 || m->getBottom() > windowPos.getHeight())
				{
					if (wantedY < 0)
						wantedY = jlimit (PopupMenuSettings::scrollZone,
										  jmax (PopupMenuSettings::scrollZone,
												windowPos.getHeight() - (PopupMenuSettings::scrollZone + m->getHeight())),
										  currentY);

					const Rectangle<int> mon (Desktop::getInstance().getMonitorAreaContaining (windowPos.getPosition(), true));

					int deltaY = wantedY - currentY;

					windowPos.setSize (jmin (windowPos.getWidth(), mon.getWidth()),
									   jmin (windowPos.getHeight(), mon.getHeight()));

					const int newY = jlimit (mon.getY(),
											 mon.getBottom() - windowPos.getHeight(),
											 windowPos.getY() + deltaY);

					deltaY -= newY - windowPos.getY();

					childYOffset -= deltaY;
					windowPos.setPosition (windowPos.getX(), newY);

					updateYPositions();
				}

				break;
			}
		}
	}

	void resizeToBestWindowPos()
	{
		Rectangle<int> r (windowPos);

		if (childYOffset < 0)
		{
			r.setBounds (r.getX(), r.getY() - childYOffset,
						 r.getWidth(), r.getHeight() + childYOffset);
		}
		else if (childYOffset > 0)
		{
			const int spaceAtBottom = r.getHeight() - (contentHeight - childYOffset);

			if (spaceAtBottom > 0)
				r.setSize (r.getWidth(), r.getHeight() - spaceAtBottom);
		}

		setBounds (r);
		updateYPositions();
	}

	void alterChildYPos (const int delta)
	{
		if (isScrolling())
		{
			childYOffset += delta;

			if (delta < 0)
			{
				childYOffset = jmax (childYOffset, 0);
			}
			else if (delta > 0)
			{
				childYOffset = jmin (childYOffset,
									 contentHeight - windowPos.getHeight() + PopupMenuSettings::borderSize);
			}

			updateYPositions();
		}
		else
		{
			childYOffset = 0;
		}

		resizeToBestWindowPos();
		repaint();
	}

	int updateYPositions()
	{
		int x = 0;
		int childNum = 0;

		for (int col = 0; col < numColumns; ++col)
		{
			const int numChildren = jmin (items.size() - childNum,
										  (items.size() + numColumns - 1) / numColumns);

			const int colW = columnWidths [col];

			int y = PopupMenuSettings::borderSize - (childYOffset + (getY() - windowPos.getY()));

			for (int i = 0; i < numChildren; ++i)
			{
				Component* const c = items.getUnchecked (childNum + i);
				c->setBounds (x, y, colW, c->getHeight());
				y += c->getHeight();
			}

			x += colW;
			childNum += numChildren;
		}

		return x;
	}

	bool isScrolling() const noexcept
	{
		return childYOffset != 0 || needsToScroll;
	}

	void setCurrentlyHighlightedChild (PopupMenu::ItemComponent* const child)
	{
		if (currentChild != nullptr)
			currentChild->setHighlighted (false);

		currentChild = child;

		if (currentChild != nullptr)
		{
			currentChild->setHighlighted (true);
			timeEnteredCurrentChildComp = Time::getApproximateMillisecondCounter();
		}
	}

	bool showSubMenuFor (PopupMenu::ItemComponent* const childComp)
	{
		activeSubMenu = nullptr;

		if (childComp != nullptr && childComp->itemInfo.hasActiveSubMenu())
		{
			activeSubMenu = Window::create (*(childComp->itemInfo.subMenu),
											dismissOnMouseUp,
											this,
											childComp->getScreenBounds(),
											0, maximumNumColumns,
											standardItemHeight,
											false, 0, managerOfChosenCommand,
											componentAttachedTo);

			if (activeSubMenu != nullptr)
			{
				activeSubMenu->setVisible (true);
				activeSubMenu->enterModalState (false);
				activeSubMenu->toFront (false);
				return true;
			}
		}

		return false;
	}

	void highlightItemUnderMouse (const Point<int>& globalMousePos, const Point<int>& localMousePos)
	{
		isOver = reallyContains (localMousePos, true);

		if (isOver)
			hasBeenOver = true;

		if (lastMouse.getDistanceFrom (globalMousePos) > 2)
		{
			lastMouseMoveTime = Time::getApproximateMillisecondCounter();

			if (disableMouseMoves && isOver)
				disableMouseMoves = false;
		}

		if (disableMouseMoves || (activeSubMenu != nullptr && activeSubMenu->isOverChildren()))
			return;

		bool isMovingTowardsMenu = false;

		if (isOver && (activeSubMenu != nullptr) && globalMousePos != lastMouse)
		{
			// try to intelligently guess whether the user is moving the mouse towards a currently-open
			// submenu. To do this, look at whether the mouse stays inside a triangular region that
			// extends from the last mouse pos to the submenu's rectangle..

			float subX = (float) activeSubMenu->getScreenX();

			if (activeSubMenu->getX() > getX())
			{
				lastMouse -= Point<int> (2, 0);  // to enlarge the triangle a bit, in case the mouse only moves a couple of pixels
			}
			else
			{
				lastMouse += Point<int> (2, 0);
				subX += activeSubMenu->getWidth();
			}

			Path areaTowardsSubMenu;
			areaTowardsSubMenu.addTriangle ((float) lastMouse.getX(), (float) lastMouse.getY(),
											subX, (float) activeSubMenu->getScreenY(),
											subX, (float) (activeSubMenu->getScreenY() + activeSubMenu->getHeight()));

			isMovingTowardsMenu = areaTowardsSubMenu.contains ((float) globalMousePos.getX(), (float) globalMousePos.getY());
		}

		lastMouse = globalMousePos;

		if (! isMovingTowardsMenu)
		{
			Component* c = getComponentAt (localMousePos.getX(), localMousePos.getY());
			if (c == this)
				c = nullptr;

			PopupMenu::ItemComponent* mic = dynamic_cast <PopupMenu::ItemComponent*> (c);

			if (mic == nullptr && c != nullptr)
				mic = c->findParentComponentOfClass ((PopupMenu::ItemComponent*) nullptr);

			if (mic != currentChild
				 && (isOver || (activeSubMenu == nullptr) || ! activeSubMenu->isVisible()))
			{
				if (isOver && (c != nullptr) && (activeSubMenu != nullptr))
					activeSubMenu->hide (0, true);

				if (! isOver)
					mic = nullptr;

				setCurrentlyHighlightedChild (mic);
			}
		}
	}

	void triggerCurrentlyHighlightedItem()
	{
		if (currentChild != nullptr
			 && currentChild->itemInfo.canBeTriggered()
			 && (currentChild->itemInfo.customComp == nullptr
				  || currentChild->itemInfo.customComp->isTriggeredAutomatically()))
		{
			dismissMenu (&currentChild->itemInfo);
		}
	}

	void selectNextItem (const int delta)
	{
		disableTimerUntilMouseMoves();
		PopupMenu::ItemComponent* mic = nullptr;
		bool wasLastOne = (currentChild == nullptr);
		const int numItems = items.size();

		for (int i = 0; i < numItems + 1; ++i)
		{
			int index = (delta > 0) ? i : (numItems - 1 - i);
			index = (index + numItems) % numItems;

			mic = items.getUnchecked (index);

			if (mic != nullptr && (mic->itemInfo.canBeTriggered() || mic->itemInfo.hasActiveSubMenu())
				 && wasLastOne)
				break;

			if (mic == currentChild)
				wasLastOne = true;
		}

		setCurrentlyHighlightedChild (mic);
	}

	void disableTimerUntilMouseMoves()
	{
		disableMouseMoves = true;

		if (owner != nullptr)
			owner->disableTimerUntilMouseMoves();
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (Window);
};

PopupMenu::PopupMenu()
	: lookAndFeel (nullptr),
	  separatorPending (false)
{
}

PopupMenu::PopupMenu (const PopupMenu& other)
	: lookAndFeel (other.lookAndFeel),
	  separatorPending (false)
{
	items.addCopiesOf (other.items);
}

PopupMenu& PopupMenu::operator= (const PopupMenu& other)
{
	if (this != &other)
	{
		lookAndFeel = other.lookAndFeel;

		clear();
		items.addCopiesOf (other.items);
	}

	return *this;
}

PopupMenu::~PopupMenu()
{
	clear();
}

void PopupMenu::clear()
{
	items.clear();
	separatorPending = false;
}

void PopupMenu::addSeparatorIfPending()
{
	if (separatorPending)
	{
		separatorPending = false;

		if (items.size() > 0)
			items.add (new Item());
	}
}

void PopupMenu::addItem (const int itemResultId, const String& itemText,
						 const bool isActive, const bool isTicked, const Image& iconToUse)
{
	jassert (itemResultId != 0);	// 0 is used as a return value to indicate that the user
									// didn't pick anything, so you shouldn't use it as the id
									// for an item..

	addSeparatorIfPending();

	items.add (new Item (itemResultId, itemText, isActive, isTicked, iconToUse,
						 Colours::black, false, 0, 0, 0));
}

void PopupMenu::addCommandItem (ApplicationCommandManager* commandManager,
								const int commandID,
								const String& displayName)
{
	jassert (commandManager != nullptr && commandID != 0);

	const ApplicationCommandInfo* const registeredInfo = commandManager->getCommandForID (commandID);

	if (registeredInfo != nullptr)
	{
		ApplicationCommandInfo info (*registeredInfo);
		ApplicationCommandTarget* const target = commandManager->getTargetForCommand (commandID, info);

		addSeparatorIfPending();

		items.add (new Item (commandID,
							 displayName.isNotEmpty() ? displayName
													  : info.shortName,
							 target != nullptr && (info.flags & ApplicationCommandInfo::isDisabled) == 0,
							 (info.flags & ApplicationCommandInfo::isTicked) != 0,
							 Image::null,
							 Colours::black,
							 false,
							 0, 0,
							 commandManager));
	}
}

void PopupMenu::addColouredItem (const int itemResultId,
								 const String& itemText,
								 const Colour& itemTextColour,
								 const bool isActive,
								 const bool isTicked,
								 const Image& iconToUse)
{
	jassert (itemResultId != 0);	// 0 is used as a return value to indicate that the user
									// didn't pick anything, so you shouldn't use it as the id
									// for an item..

	addSeparatorIfPending();

	items.add (new Item (itemResultId, itemText, isActive, isTicked, iconToUse,
						 itemTextColour, true, 0, 0, 0));
}

void PopupMenu::addCustomItem (const int itemResultId, CustomComponent* const customComponent)
{
	jassert (itemResultId != 0);	// 0 is used as a return value to indicate that the user
									// didn't pick anything, so you shouldn't use it as the id
									// for an item..

	addSeparatorIfPending();

	items.add (new Item (itemResultId, String::empty, true, false, Image::null,
						 Colours::black, false, customComponent, 0, 0));
}

class NormalComponentWrapper : public PopupMenu::CustomComponent
{
public:
	NormalComponentWrapper (Component* const comp, const int w, const int h,
							const bool triggerMenuItemAutomaticallyWhenClicked)
		: PopupMenu::CustomComponent (triggerMenuItemAutomaticallyWhenClicked),
		  width (w), height (h)
	{
		addAndMakeVisible (comp);
	}

	void getIdealSize (int& idealWidth, int& idealHeight)
	{
		idealWidth = width;
		idealHeight = height;
	}

	void resized()
	{
		if (getChildComponent(0) != nullptr)
			getChildComponent(0)->setBounds (getLocalBounds());
	}

private:
	const int width, height;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (NormalComponentWrapper);
};

void PopupMenu::addCustomItem (const int itemResultId,
							   Component* customComponent,
							   int idealWidth, int idealHeight,
							   const bool triggerMenuItemAutomaticallyWhenClicked)
{
	addCustomItem (itemResultId,
				   new NormalComponentWrapper (customComponent, idealWidth, idealHeight,
											   triggerMenuItemAutomaticallyWhenClicked));
}

void PopupMenu::addSubMenu (const String& subMenuName,
							const PopupMenu& subMenu,
							const bool isActive,
							const Image& iconToUse,
							const bool isTicked)
{
	addSeparatorIfPending();

	items.add (new Item (0, subMenuName, isActive && (subMenu.getNumItems() > 0), isTicked,
						 iconToUse, Colours::black, false, 0, &subMenu, 0));
}

void PopupMenu::addSeparator()
{
	separatorPending = true;
}

class HeaderItemComponent  : public PopupMenu::CustomComponent
{
public:
	HeaderItemComponent (const String& name)
		: PopupMenu::CustomComponent (false)
	{
		setName (name);
	}

	void paint (Graphics& g)
	{
		Font f (getLookAndFeel().getPopupMenuFont());
		f.setBold (true);
		g.setFont (f);
		g.setColour (findColour (PopupMenu::headerTextColourId));

		g.drawFittedText (getName(),
						  12, 0, getWidth() - 16, proportionOfHeight (0.8f),
						  Justification::bottomLeft, 1);
	}

	void getIdealSize (int& idealWidth, int& idealHeight)
	{
		getLookAndFeel().getIdealPopupMenuItemSize (getName(), false, -1, idealWidth, idealHeight);
		idealHeight += idealHeight / 2;
		idealWidth += idealWidth / 4;
	}

private:
	JUCE_LEAK_DETECTOR (HeaderItemComponent);
};

void PopupMenu::addSectionHeader (const String& title)
{
	addCustomItem (0X4734a34f, new HeaderItemComponent (title));
}

PopupMenu::Options::Options()
	: targetComponent (nullptr),
	  visibleItemID (0),
	  minWidth (0),
	  maxColumns (0),
	  standardHeight (0)
{
	targetArea.setPosition (Desktop::getMousePosition());
}

const PopupMenu::Options PopupMenu::Options::withTargetComponent (Component* comp) const
{
	Options o (*this);
	o.targetComponent = comp;

	if (comp != nullptr)
		o.targetArea = comp->getScreenBounds();

	return o;
}

const PopupMenu::Options PopupMenu::Options::withTargetScreenArea (const Rectangle<int>& area) const
{
	Options o (*this);
	o.targetArea = area;
	return o;
}

const PopupMenu::Options PopupMenu::Options::withMinimumWidth (int w) const
{
	Options o (*this);
	o.minWidth = w;
	return o;
}

const PopupMenu::Options PopupMenu::Options::withMaximumNumColumns (int cols) const
{
	Options o (*this);
	o.maxColumns = cols;
	return o;
}

const PopupMenu::Options PopupMenu::Options::withStandardItemHeight (int height) const
{
	Options o (*this);
	o.standardHeight = height;
	return o;
}

const PopupMenu::Options PopupMenu::Options::withItemThatMustBeVisible (int idOfItemToBeVisible) const
{
	Options o (*this);
	o.visibleItemID = idOfItemToBeVisible;
	return o;
}

Component* PopupMenu::createWindow (const Options& options,
									ApplicationCommandManager** managerOfChosenCommand) const
{
	return Window::create (*this, ModifierKeys::getCurrentModifiers().isAnyMouseButtonDown(),
						   0, options.targetArea, options.minWidth, options.maxColumns > 0 ? options.maxColumns : 7,
						   options.standardHeight, ! options.targetArea.isEmpty(), options.visibleItemID,
						   managerOfChosenCommand, options.targetComponent);
}

// This invokes any command manager commands and deletes the menu window when it is dismissed
class PopupMenuCompletionCallback  : public ModalComponentManager::Callback
{
public:
	PopupMenuCompletionCallback()
		: managerOfChosenCommand (nullptr),
		  prevFocused (Component::getCurrentlyFocusedComponent()),
		  prevTopLevel (prevFocused != nullptr ? prevFocused->getTopLevelComponent() : 0)
	{
		PopupMenuSettings::menuWasHiddenBecauseOfAppChange = false;
	}

	void modalStateFinished (int result)
	{
		if (managerOfChosenCommand != nullptr && result != 0)
		{
			ApplicationCommandTarget::InvocationInfo info (result);
			info.invocationMethod = ApplicationCommandTarget::InvocationInfo::fromMenu;

			managerOfChosenCommand->invoke (info, true);
		}

		// (this would be the place to fade out the component, if that's what's required)
		component = nullptr;

		if (! PopupMenuSettings::menuWasHiddenBecauseOfAppChange)
		{
			if (prevTopLevel != nullptr)
				prevTopLevel->toFront (true);

			if (prevFocused != nullptr)
				prevFocused->grabKeyboardFocus();
		}
	}

	ApplicationCommandManager* managerOfChosenCommand;
	ScopedPointer<Component> component;
	WeakReference<Component> prevFocused, prevTopLevel;

private:
	JUCE_DECLARE_NON_COPYABLE (PopupMenuCompletionCallback);
};

int PopupMenu::showWithOptionalCallback (const Options& options, ModalComponentManager::Callback* const userCallback,
										 const bool canBeModal)
{
	ScopedPointer<ModalComponentManager::Callback> userCallbackDeleter (userCallback);
	ScopedPointer<PopupMenuCompletionCallback> callback (new PopupMenuCompletionCallback());

	Component* window = createWindow (options, &(callback->managerOfChosenCommand));
	if (window == nullptr)
		return 0;

	callback->component = window;

	window->enterModalState (false, userCallbackDeleter.release());
	ModalComponentManager::getInstance()->attachCallback (window, callback.release());

	window->toFront (false);  // need to do this after making it modal, or it could
							  // be stuck behind other comps that are already modal..

   #if JUCE_MODAL_LOOPS_PERMITTED
	return (userCallback == nullptr && canBeModal) ? window->runModalLoop() : 0;
   #else
	jassert (userCallback != nullptr && canBeModal);
	return 0;
   #endif
}

#if JUCE_MODAL_LOOPS_PERMITTED
int PopupMenu::showMenu (const Options& options)
{
	return showWithOptionalCallback (options, 0, true);
}
#endif

void PopupMenu::showMenuAsync (const Options& options, ModalComponentManager::Callback* userCallback)
{
   #if ! JUCE_MODAL_LOOPS_PERMITTED
	jassert (userCallback != nullptr);
   #endif

	showWithOptionalCallback (options, userCallback, false);
}

#if JUCE_MODAL_LOOPS_PERMITTED
int PopupMenu::show (const int itemIdThatMustBeVisible,
					 const int minimumWidth, const int maximumNumColumns,
					 const int standardItemHeight,
					 ModalComponentManager::Callback* callback)
{
	return showWithOptionalCallback (Options().withItemThatMustBeVisible (itemIdThatMustBeVisible)
											  .withMinimumWidth (minimumWidth)
											  .withMaximumNumColumns (maximumNumColumns)
											  .withStandardItemHeight (standardItemHeight),
									 callback, true);
}

int PopupMenu::showAt (const Rectangle<int>& screenAreaToAttachTo,
					   const int itemIdThatMustBeVisible,
					   const int minimumWidth, const int maximumNumColumns,
					   const int standardItemHeight,
					   ModalComponentManager::Callback* callback)
{
	return showWithOptionalCallback (Options().withTargetScreenArea (screenAreaToAttachTo)
											  .withItemThatMustBeVisible (itemIdThatMustBeVisible)
											  .withMinimumWidth (minimumWidth)
											  .withMaximumNumColumns (maximumNumColumns)
											  .withStandardItemHeight (standardItemHeight),
									 callback, true);
}

int PopupMenu::showAt (Component* componentToAttachTo,
					   const int itemIdThatMustBeVisible,
					   const int minimumWidth, const int maximumNumColumns,
					   const int standardItemHeight,
					   ModalComponentManager::Callback* callback)
{
	Options options (Options().withItemThatMustBeVisible (itemIdThatMustBeVisible)
							  .withMinimumWidth (minimumWidth)
							  .withMaximumNumColumns (maximumNumColumns)
							  .withStandardItemHeight (standardItemHeight));

	if (componentToAttachTo != nullptr)
		options = options.withTargetComponent (componentToAttachTo);

	return showWithOptionalCallback (options, callback, true);
}
#endif

bool JUCE_CALLTYPE PopupMenu::dismissAllActiveMenus()
{
	const int numWindows = Window::getActiveWindows().size();
	for (int i = numWindows; --i >= 0;)
	{
		Window* const pmw = Window::getActiveWindows()[i];

		if (pmw != nullptr)
			pmw->dismissMenu (0);
	}

	return numWindows > 0;
}

int PopupMenu::getNumItems() const noexcept
{
	int num = 0;

	for (int i = items.size(); --i >= 0;)
		if (! items.getUnchecked(i)->isSeparator)
			++num;

	return num;
}

bool PopupMenu::containsCommandItem (const int commandID) const
{
	for (int i = items.size(); --i >= 0;)
	{
		const Item* mi = items.getUnchecked (i);

		if ((mi->itemId == commandID && mi->commandManager != nullptr)
			|| (mi->subMenu != nullptr && mi->subMenu->containsCommandItem (commandID)))
		{
			return true;
		}
	}

	return false;
}

bool PopupMenu::containsAnyActiveItems() const noexcept
{
	for (int i = items.size(); --i >= 0;)
	{
		const Item* const mi = items.getUnchecked (i);

		if (mi->subMenu != nullptr)
		{
			if (mi->subMenu->containsAnyActiveItems())
				return true;
		}
		else if (mi->active)
		{
			return true;
		}
	}

	return false;
}

void PopupMenu::setLookAndFeel (LookAndFeel* const newLookAndFeel)
{
	lookAndFeel = newLookAndFeel;
}

PopupMenu::CustomComponent::CustomComponent (const bool isTriggeredAutomatically_)
	: isHighlighted (false),
	  triggeredAutomatically (isTriggeredAutomatically_)
{
}

PopupMenu::CustomComponent::~CustomComponent()
{
}

void PopupMenu::CustomComponent::setHighlighted (bool shouldBeHighlighted)
{
	isHighlighted = shouldBeHighlighted;
	repaint();
}

void PopupMenu::CustomComponent::triggerMenuItem()
{
	PopupMenu::ItemComponent* const mic = dynamic_cast <PopupMenu::ItemComponent*> (getParentComponent());

	if (mic != nullptr)
	{
		PopupMenu::Window* const pmw = dynamic_cast <PopupMenu::Window*> (mic->getParentComponent());

		if (pmw != nullptr)
		{
			pmw->dismissMenu (&mic->itemInfo);
		}
		else
		{
			// something must have gone wrong with the component hierarchy if this happens..
			jassertfalse;
		}
	}
	else
	{
		// why isn't this component inside a menu? Not much point triggering the item if
		// there's no menu.
		jassertfalse;
	}
}

PopupMenu::MenuItemIterator::MenuItemIterator (const PopupMenu& menu_)
	: subMenu (nullptr),
	  itemId (0),
	  isSeparator (false),
	  isTicked (false),
	  isEnabled (false),
	  isCustomComponent (false),
	  isSectionHeader (false),
	  customColour (0),
	  menu (menu_),
	  index (0)
{
}

PopupMenu::MenuItemIterator::~MenuItemIterator()
{
}

bool PopupMenu::MenuItemIterator::next()
{
	if (index >= menu.items.size())
		return false;

	const Item* const item = menu.items.getUnchecked (index);
	++index;

	itemName = item->customComp != nullptr ? item->customComp->getName() : item->text;
	subMenu = item->subMenu;
	itemId = item->itemId;

	isSeparator = item->isSeparator;
	isTicked = item->isTicked;
	isEnabled = item->active;
	isSectionHeader = dynamic_cast <HeaderItemComponent*> (static_cast <CustomComponent*> (item->customComp)) != nullptr;
	isCustomComponent = (! isSectionHeader) && item->customComp != nullptr;
	customColour = item->usesColour ? &(item->textColour) : 0;
	customImage = item->image;
	commandManager = item->commandManager;

	return true;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_PopupMenu.cpp ***/


/*** Start of inlined file: juce_ComponentDragger.cpp ***/
BEGIN_JUCE_NAMESPACE

ComponentDragger::ComponentDragger()
{
}

ComponentDragger::~ComponentDragger()
{
}

void ComponentDragger::startDraggingComponent (Component* const componentToDrag, const MouseEvent& e)
{
	jassert (componentToDrag != nullptr);
	jassert (e.mods.isAnyMouseButtonDown()); // The event has to be a drag event!

	if (componentToDrag != nullptr)
		mouseDownWithinTarget = e.getEventRelativeTo (componentToDrag).getMouseDownPosition();
}

void ComponentDragger::dragComponent (Component* const componentToDrag, const MouseEvent& e,
									  ComponentBoundsConstrainer* const constrainer)
{
	jassert (componentToDrag != nullptr);
	jassert (e.mods.isAnyMouseButtonDown()); // The event has to be a drag event!

	if (componentToDrag != nullptr)
	{
		Rectangle<int> bounds (componentToDrag->getBounds());

		// If the component is a window, multiple mouse events can get queued while it's in the same position,
		// so their coordinates become wrong after the first one moves the window, so in that case, we'll use
		// the current mouse position instead of the one that the event contains...
		if (componentToDrag->isOnDesktop())
			bounds += componentToDrag->getMouseXYRelative() - mouseDownWithinTarget;
		else
			bounds += e.getEventRelativeTo (componentToDrag).getPosition() - mouseDownWithinTarget;

		if (constrainer != nullptr)
			constrainer->setBoundsForComponent (componentToDrag, bounds, false, false, false, false);
		else
			componentToDrag->setBounds (bounds);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComponentDragger.cpp ***/


/*** Start of inlined file: juce_DragAndDropContainer.cpp ***/
BEGIN_JUCE_NAMESPACE

bool juce_performDragDropFiles (const StringArray& files, const bool copyFiles, bool& shouldStop);
bool juce_performDragDropText (const String& text, bool& shouldStop);

class DragImageComponent  : public Component,
							public Timer
{
public:
	DragImageComponent (const Image& im,
						const var& desc,
						Component* const sourceComponent,
						Component* const mouseDragSource_,
						DragAndDropContainer* const o,
						const Point<int>& imageOffset_)
		: sourceDetails (desc, sourceComponent, Point<int>()),
		  image (im),
		  mouseDragSource (mouseDragSource_),
		  owner (o),
		  imageOffset (imageOffset_),
		  hasCheckedForExternalDrag (false),
		  drawImage (true)
	{
		setSize (im.getWidth(), im.getHeight());

		if (mouseDragSource == nullptr)
			mouseDragSource = sourceComponent;

		mouseDragSource->addMouseListener (this, false);

		startTimer (200);

		setInterceptsMouseClicks (false, false);
		setAlwaysOnTop (true);
	}

	~DragImageComponent()
	{
		if (owner->dragImageComponent == this)
			owner->dragImageComponent.release();

		if (mouseDragSource != nullptr)
		{
			mouseDragSource->removeMouseListener (this);

			DragAndDropTarget* const current = getCurrentlyOver();

			if (current != nullptr && current->isInterestedInDragSource (sourceDetails))
				current->itemDragExit (sourceDetails);
		}
	}

	void paint (Graphics& g)
	{
		if (isOpaque())
			g.fillAll (Colours::white);

		if (drawImage)
		{
			g.setOpacity (1.0f);
			g.drawImageAt (image, 0, 0);
		}
	}

	DragAndDropTarget* findTarget (const Point<int>& screenPos, Point<int>& relativePos)
	{
		Component* hit = getParentComponent();

		if (hit == nullptr)
		{
			hit = Desktop::getInstance().findComponentAt (screenPos);
		}
		else
		{
			const Point<int> relPos (hit->getLocalPoint (nullptr, screenPos));
			hit = hit->getComponentAt (relPos.getX(), relPos.getY());
		}

		// (note: use a local copy of this in case the callback runs
		// a modal loop and deletes this object before the method completes)
		const DragAndDropTarget::SourceDetails details (sourceDetails);

		while (hit != nullptr)
		{
			DragAndDropTarget* const ddt = dynamic_cast <DragAndDropTarget*> (hit);

			if (ddt != nullptr && ddt->isInterestedInDragSource (details))
			{
				relativePos = hit->getLocalPoint (nullptr, screenPos);
				return ddt;
			}

			hit = hit->getParentComponent();
		}

		return nullptr;
	}

	void mouseUp (const MouseEvent& e)
	{
		if (e.originalComponent != this)
		{
			if (mouseDragSource != nullptr)
				mouseDragSource->removeMouseListener (this);

			bool dropAccepted = false;
			DragAndDropTarget* ddt = nullptr;
			Point<int> relPos;

			if (isVisible())
			{
				setVisible (false);
				ddt = findTarget (e.getScreenPosition(), relPos);

				// fade this component and remove it - it'll be deleted later by the timer callback

				dropAccepted = ddt != nullptr;

				setVisible (true);

				if (dropAccepted || sourceDetails.sourceComponent == nullptr)
				{
					Desktop::getInstance().getAnimator().fadeOut (this, 120);
				}
				else
				{
					const Point<int> target (sourceDetails.sourceComponent->localPointToGlobal (sourceDetails.sourceComponent->getLocalBounds().getCentre()));
					const Point<int> ourCentre (localPointToGlobal (getLocalBounds().getCentre()));

					Desktop::getInstance().getAnimator().animateComponent (this,
																		   getBounds() + (target - ourCentre),
																		   0.0f, 120,
																		   true, 1.0, 1.0);
				}
			}

			if (getParentComponent() != nullptr)
				getParentComponent()->removeChildComponent (this);

			if (dropAccepted && ddt != nullptr)
			{
				// (note: use a local copy of this in case the callback runs
				// a modal loop and deletes this object before the method completes)
				DragAndDropTarget::SourceDetails details (sourceDetails);
				details.localPosition = relPos;

				currentlyOverComp = nullptr;

				ddt->itemDropped (details);
			}

			// careful - this object could now be deleted..
		}
	}

	void updateLocation (const bool canDoExternalDrag, const Point<int>& screenPos)
	{
		// (note: use a local copy of this in case a callback runs a modal loop and deletes
		// this object before it returns)
		DragAndDropTarget::SourceDetails details (sourceDetails);

		Point<int> newPos (screenPos + imageOffset);

		if (getParentComponent() != nullptr)
			newPos = getParentComponent()->getLocalPoint (nullptr, newPos);

		//if (newX != getX() || newY != getY())
		{
			setTopLeftPosition (newPos.getX(), newPos.getY());

			Point<int> relPos;
			DragAndDropTarget* const newTarget = findTarget (screenPos, relPos);
			Component* newTargetComp = dynamic_cast <Component*> (newTarget);

			drawImage = (newTarget == nullptr) || newTarget->shouldDrawDragImageWhenOver();

			if (newTargetComp != currentlyOverComp)
			{
				DragAndDropTarget* const lastTarget = getCurrentlyOver();

				if (lastTarget != nullptr && details.sourceComponent != nullptr
					  && lastTarget->isInterestedInDragSource (details))
					lastTarget->itemDragExit (details);

				currentlyOverComp = newTargetComp;

				if (newTarget != nullptr)
				{
					details.localPosition = relPos;

					if (newTarget->isInterestedInDragSource (details))
						newTarget->itemDragEnter (details);
				}
			}

			{
				DragAndDropTarget* const target = getCurrentlyOver();
				if (target != nullptr)
				{
					details.localPosition = relPos;

					if (target->isInterestedInDragSource (details))
						target->itemDragMove (details);
				}
			}

			if (getCurrentlyOver() == nullptr && canDoExternalDrag && ! hasCheckedForExternalDrag)
			{
				if (Desktop::getInstance().findComponentAt (screenPos) == nullptr)
				{
					hasCheckedForExternalDrag = true;
					StringArray files;
					bool canMoveFiles = false;

					if (owner->shouldDropFilesWhenDraggedExternally (details, files, canMoveFiles)
						 && files.size() > 0)
					{
						WeakReference<Component> thisWeakRef (this);
						setVisible (false);

						if (ModifierKeys::getCurrentModifiersRealtime().isAnyMouseButtonDown())
							DragAndDropContainer::performExternalDragDropOfFiles (files, canMoveFiles);

						delete thisWeakRef.get();
						return;
					}
				}
			}
		}
	}

	void mouseDrag (const MouseEvent& e)
	{
		if (e.originalComponent != this)
			updateLocation (true, e.getScreenPosition());
	}

	void timerCallback()
	{
		if (sourceDetails.sourceComponent == nullptr)
		{
			delete this;
		}
		else if (! isMouseButtonDownAnywhere())
		{
			if (mouseDragSource != nullptr)
				mouseDragSource->removeMouseListener (this);

			delete this;
		}
	}

private:
	DragAndDropTarget::SourceDetails sourceDetails;
	Image image;
	WeakReference<Component> mouseDragSource;
	DragAndDropContainer* const owner;

	WeakReference<Component> currentlyOverComp;
	DragAndDropTarget* getCurrentlyOver()
	{
		return dynamic_cast <DragAndDropTarget*> (currentlyOverComp.get());
	}

	const Point<int> imageOffset;
	bool hasCheckedForExternalDrag, drawImage;

	JUCE_DECLARE_NON_COPYABLE (DragImageComponent);
};

DragAndDropContainer::DragAndDropContainer()
{
}

DragAndDropContainer::~DragAndDropContainer()
{
	dragImageComponent = nullptr;
}

void DragAndDropContainer::startDragging (const var& sourceDescription,
										  Component* sourceComponent,
										  const Image& dragImage_,
										  const bool allowDraggingToExternalWindows,
										  const Point<int>* imageOffsetFromMouse)
{
	Image dragImage (dragImage_);

	if (dragImageComponent == nullptr)
	{
		Component* const thisComp = dynamic_cast <Component*> (this);

		if (thisComp == nullptr)
		{
			jassertfalse;   // Your DragAndDropContainer needs to be a Component!
			return;
		}

		MouseInputSource* draggingSource = Desktop::getInstance().getDraggingMouseSource (0);

		if (draggingSource == nullptr || ! draggingSource->isDragging())
		{
			jassertfalse;   // You must call startDragging() from within a mouseDown or mouseDrag callback!
			return;
		}

		const Point<int> lastMouseDown (Desktop::getLastMouseDownPosition());
		Point<int> imageOffset;

		if (dragImage.isNull())
		{
			dragImage = sourceComponent->createComponentSnapshot (sourceComponent->getLocalBounds())
							.convertedToFormat (Image::ARGB);

			dragImage.multiplyAllAlphas (0.6f);

			const int lo = 150;
			const int hi = 400;

			Point<int> relPos (sourceComponent->getLocalPoint (nullptr, lastMouseDown));
			Point<int> clipped (dragImage.getBounds().getConstrainedPoint (relPos));

			for (int y = dragImage.getHeight(); --y >= 0;)
			{
				const double dy = (y - clipped.getY()) * (y - clipped.getY());

				for (int x = dragImage.getWidth(); --x >= 0;)
				{
					const int dx = x - clipped.getX();
					const int distance = roundToInt (std::sqrt (dx * dx + dy));

					if (distance > lo)
					{
						const float alpha = (distance > hi) ? 0
															: (hi - distance) / (float) (hi - lo)
															   + Random::getSystemRandom().nextFloat() * 0.008f;

						dragImage.multiplyAlphaAt (x, y, alpha);
					}
				}
			}

			imageOffset = -clipped;
		}
		else
		{
			if (imageOffsetFromMouse == nullptr)
				imageOffset = -dragImage.getBounds().getCentre();
			else
				imageOffset = -(dragImage.getBounds().getConstrainedPoint (-*imageOffsetFromMouse));
		}

		dragImageComponent = new DragImageComponent (dragImage, sourceDescription, sourceComponent,
													 draggingSource->getComponentUnderMouse(), this, imageOffset);

		currentDragDesc = sourceDescription;

		if (allowDraggingToExternalWindows)
		{
			if (! Desktop::canUseSemiTransparentWindows())
				dragImageComponent->setOpaque (true);

			dragImageComponent->addToDesktop (ComponentPeer::windowIgnoresMouseClicks
											   | ComponentPeer::windowIsTemporary
											   | ComponentPeer::windowIgnoresKeyPresses);
		}
		else
			thisComp->addChildComponent (dragImageComponent);

		static_cast <DragImageComponent*> (static_cast <Component*> (dragImageComponent))->updateLocation (false, lastMouseDown);
		dragImageComponent->setVisible (true);

	   #if JUCE_WINDOWS
		// Under heavy load, the layered window's paint callback can often be lost by the OS,
		// so forcing a repaint at least once makes sure that the window becomes visible..
		ComponentPeer* const peer = dragImageComponent->getPeer();
		if (peer != nullptr)
			peer->performAnyPendingRepaintsNow();
	   #endif
	}
}

bool DragAndDropContainer::isDragAndDropActive() const
{
	return dragImageComponent != nullptr;
}

String DragAndDropContainer::getCurrentDragDescription() const
{
	return dragImageComponent != nullptr ? currentDragDesc
										 : String::empty;
}

DragAndDropContainer* DragAndDropContainer::findParentDragContainerFor (Component* c)
{
	return c == nullptr ? nullptr : c->findParentComponentOfClass ((DragAndDropContainer*) nullptr);
}

bool DragAndDropContainer::shouldDropFilesWhenDraggedExternally (const DragAndDropTarget::SourceDetails&, StringArray&, bool&)
{
	return false;
}

DragAndDropTarget::SourceDetails::SourceDetails (const var& description_, Component* sourceComponent_, const Point<int>& localPosition_) noexcept
	: description (description_),
	  sourceComponent (sourceComponent_),
	  localPosition (localPosition_)
{
}

void DragAndDropTarget::itemDragEnter (const SourceDetails&)  {}
void DragAndDropTarget::itemDragMove  (const SourceDetails&)  {}
void DragAndDropTarget::itemDragExit  (const SourceDetails&)  {}
bool DragAndDropTarget::shouldDrawDragImageWhenOver()	 { return true; }

void FileDragAndDropTarget::fileDragEnter (const StringArray&, int, int)  {}
void FileDragAndDropTarget::fileDragMove  (const StringArray&, int, int)  {}
void FileDragAndDropTarget::fileDragExit  (const StringArray&)		{}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DragAndDropContainer.cpp ***/


/*** Start of inlined file: juce_MouseCursor.cpp ***/
BEGIN_JUCE_NAMESPACE

class MouseCursor::SharedCursorHandle
{
public:
	explicit SharedCursorHandle (const MouseCursor::StandardCursorType type)
		: handle (createStandardMouseCursor (type)),
		  refCount (1),
		  standardType (type),
		  isStandard (true)
	{
	}

	SharedCursorHandle (const Image& image, const int hotSpotX, const int hotSpotY)
		: handle (createMouseCursorFromImage (image, hotSpotX, hotSpotY)),
		  refCount (1),
		  standardType (MouseCursor::NormalCursor),
		  isStandard (false)
	{
	}

	~SharedCursorHandle()
	{
		deleteMouseCursor (handle, isStandard);
	}

	static SharedCursorHandle* createStandard (const MouseCursor::StandardCursorType type)
	{
		const SpinLock::ScopedLockType sl (lock);

		for (int i = 0; i < getCursors().size(); ++i)
		{
			SharedCursorHandle* const sc = getCursors().getUnchecked(i);

			if (sc->standardType == type)
				return sc->retain();
		}

		SharedCursorHandle* const sc = new SharedCursorHandle (type);
		getCursors().add (sc);
		return sc;
	}

	SharedCursorHandle* retain() noexcept
	{
		++refCount;
		return this;
	}

	void release()
	{
		if (--refCount == 0)
		{
			if (isStandard)
			{
				const SpinLock::ScopedLockType sl (lock);
				getCursors().removeValue (this);
			}

			delete this;
		}
	}

	void* getHandle() const noexcept	{ return handle; }

private:

	void* const handle;
	Atomic <int> refCount;
	const MouseCursor::StandardCursorType standardType;
	const bool isStandard;
	static SpinLock lock;

	static Array <SharedCursorHandle*>& getCursors()
	{
		static Array <SharedCursorHandle*> cursors;
		return cursors;
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (SharedCursorHandle);
};

SpinLock MouseCursor::SharedCursorHandle::lock;

MouseCursor::MouseCursor()
	: cursorHandle (nullptr)
{
}

MouseCursor::MouseCursor (const StandardCursorType type)
	: cursorHandle (type != MouseCursor::NormalCursor ? SharedCursorHandle::createStandard (type) : 0)
{
}

MouseCursor::MouseCursor (const Image& image, const int hotSpotX, const int hotSpotY)
	: cursorHandle (new SharedCursorHandle (image, hotSpotX, hotSpotY))
{
}

MouseCursor::MouseCursor (const MouseCursor& other)
	: cursorHandle (other.cursorHandle == nullptr ? nullptr : other.cursorHandle->retain())
{
}

MouseCursor::~MouseCursor()
{
	if (cursorHandle != nullptr)
		cursorHandle->release();
}

MouseCursor& MouseCursor::operator= (const MouseCursor& other)
{
	if (other.cursorHandle != nullptr)
		other.cursorHandle->retain();

	if (cursorHandle != nullptr)
		cursorHandle->release();

	cursorHandle = other.cursorHandle;
	return *this;
}

bool MouseCursor::operator== (const MouseCursor& other) const noexcept
{
	return getHandle() == other.getHandle();
}

bool MouseCursor::operator!= (const MouseCursor& other) const noexcept
{
	return getHandle() != other.getHandle();
}

void* MouseCursor::getHandle() const noexcept
{
	return cursorHandle != nullptr ? cursorHandle->getHandle() : nullptr;
}

void MouseCursor::showWaitCursor()
{
	Desktop::getInstance().getMainMouseSource().showMouseCursor (MouseCursor::WaitCursor);
}

void MouseCursor::hideWaitCursor()
{
	Desktop::getInstance().getMainMouseSource().revealCursor();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MouseCursor.cpp ***/


/*** Start of inlined file: juce_MouseEvent.cpp ***/
BEGIN_JUCE_NAMESPACE

MouseEvent::MouseEvent (MouseInputSource& source_,
						const Point<int>& position,
						const ModifierKeys& mods_,
						Component* const eventComponent_,
						Component* const originator,
						const Time& eventTime_,
						const Point<int> mouseDownPos_,
						const Time& mouseDownTime_,
						const int numberOfClicks_,
						const bool mouseWasDragged) noexcept
	: x (position.getX()),
	  y (position.getY()),
	  mods (mods_),
	  eventComponent (eventComponent_),
	  originalComponent (originator),
	  eventTime (eventTime_),
	  source (source_),
	  mouseDownPos (mouseDownPos_),
	  mouseDownTime (mouseDownTime_),
	  numberOfClicks (numberOfClicks_),
	  wasMovedSinceMouseDown (mouseWasDragged)
{
}

MouseEvent::~MouseEvent() noexcept
{
}

MouseEvent MouseEvent::getEventRelativeTo (Component* const otherComponent) const noexcept
{
	if (otherComponent == nullptr)
	{
		jassertfalse;
		return *this;
	}

	return MouseEvent (source, otherComponent->getLocalPoint (eventComponent, getPosition()),
					   mods, otherComponent, originalComponent, eventTime,
					   otherComponent->getLocalPoint (eventComponent, mouseDownPos),
					   mouseDownTime, numberOfClicks, wasMovedSinceMouseDown);
}

MouseEvent MouseEvent::withNewPosition (const Point<int>& newPosition) const noexcept
{
	return MouseEvent (source, newPosition, mods, eventComponent, originalComponent,
					   eventTime, mouseDownPos, mouseDownTime,
					   numberOfClicks, wasMovedSinceMouseDown);
}

bool MouseEvent::mouseWasClicked() const noexcept
{
	return ! wasMovedSinceMouseDown;
}

int MouseEvent::getMouseDownX() const noexcept
{
	return mouseDownPos.getX();
}

int MouseEvent::getMouseDownY() const noexcept
{
	return mouseDownPos.getY();
}

const Point<int> MouseEvent::getMouseDownPosition() const noexcept
{
	return mouseDownPos;
}

int MouseEvent::getDistanceFromDragStartX() const noexcept
{
	return x - mouseDownPos.getX();
}

int MouseEvent::getDistanceFromDragStartY() const noexcept
{
	return y - mouseDownPos.getY();
}

int MouseEvent::getDistanceFromDragStart() const noexcept
{
	return mouseDownPos.getDistanceFrom (getPosition());
}

const Point<int> MouseEvent::getOffsetFromDragStart() const noexcept
{
	return getPosition() - mouseDownPos;
}

int MouseEvent::getLengthOfMousePress() const noexcept
{
	if (mouseDownTime.toMilliseconds() > 0)
		return jmax (0, (int) (eventTime - mouseDownTime).inMilliseconds());

	return 0;
}

const Point<int> MouseEvent::getPosition() const noexcept
{
	return Point<int> (x, y);
}

int MouseEvent::getScreenX() const
{
	return getScreenPosition().getX();
}

int MouseEvent::getScreenY() const
{
	return getScreenPosition().getY();
}

const Point<int> MouseEvent::getScreenPosition() const
{
	return eventComponent->localPointToGlobal (Point<int> (x, y));
}

int MouseEvent::getMouseDownScreenX() const
{
	return getMouseDownScreenPosition().getX();
}

int MouseEvent::getMouseDownScreenY() const
{
	return getMouseDownScreenPosition().getY();
}

const Point<int> MouseEvent::getMouseDownScreenPosition() const
{
	return eventComponent->localPointToGlobal (mouseDownPos);
}

int MouseEvent::doubleClickTimeOutMs = 400;

void MouseEvent::setDoubleClickTimeout (const int newTime) noexcept
{
	doubleClickTimeOutMs = newTime;
}

int MouseEvent::getDoubleClickTimeout() noexcept
{
	return doubleClickTimeOutMs;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MouseEvent.cpp ***/


/*** Start of inlined file: juce_MouseInputSource.cpp ***/
BEGIN_JUCE_NAMESPACE

class MouseInputSourceInternal   : public AsyncUpdater
{
public:

	MouseInputSourceInternal (MouseInputSource& source_, const int index_, const bool isMouseDevice_)
		: index (index_), isMouseDevice (isMouseDevice_), source (source_), lastPeer (nullptr),
		  isUnboundedMouseModeOn (false), isCursorVisibleUntilOffscreen (false), currentCursorHandle (nullptr),
		  mouseEventCounter (0)
	{
	}

	bool isDragging() const noexcept
	{
		return buttonState.isAnyMouseButtonDown();
	}

	Component* getComponentUnderMouse() const
	{
		return static_cast <Component*> (componentUnderMouse);
	}

	const ModifierKeys getCurrentModifiers() const
	{
		return ModifierKeys::getCurrentModifiers().withoutMouseButtons().withFlags (buttonState.getRawFlags());
	}

	ComponentPeer* getPeer()
	{
		if (! ComponentPeer::isValidPeer (lastPeer))
			lastPeer = nullptr;

		return lastPeer;
	}

	Component* findComponentAt (const Point<int>& screenPos)
	{
		ComponentPeer* const peer = getPeer();

		if (peer != nullptr)
		{
			Component* const comp = peer->getComponent();
			const Point<int> relativePos (comp->getLocalPoint (nullptr, screenPos));

			// (the contains() call is needed to test for overlapping desktop windows)
			if (comp->contains (relativePos))
				return comp->getComponentAt (relativePos);
		}

		return nullptr;
	}

	const Point<int> getScreenPosition() const
	{
		// This needs to return the live position if possible, but it mustn't update the lastScreenPos
		// value, because that can cause continuity problems.
		return unboundedMouseOffset + (isMouseDevice ? MouseInputSource::getCurrentMousePosition()
													 : lastScreenPos);
	}

	void sendMouseEnter (Component* const comp, const Point<int>& screenPos, const Time& time)
	{
		//DBG ("Mouse " + String (source.getIndex()) + " enter: " + comp->getLocalPoint (nullptr, screenPos).toString() + " - Comp: " + String::toHexString ((int) comp));
		comp->internalMouseEnter (source, comp->getLocalPoint (nullptr, screenPos), time);
	}

	void sendMouseExit (Component* const comp, const Point<int>& screenPos, const Time& time)
	{
		//DBG ("Mouse " + String (source.getIndex()) + " exit: " + comp->getLocalPoint (nullptr, screenPos).toString() + " - Comp: " + String::toHexString ((int) comp));
		comp->internalMouseExit (source, comp->getLocalPoint (nullptr, screenPos), time);
	}

	void sendMouseMove (Component* const comp, const Point<int>& screenPos, const Time& time)
	{
		//DBG ("Mouse " + String (source.getIndex()) + " move: " + comp->getLocalPoint (nullptr, screenPos).toString() + " - Comp: " + String::toHexString ((int) comp));
		comp->internalMouseMove (source, comp->getLocalPoint (nullptr, screenPos), time);
	}

	void sendMouseDown (Component* const comp, const Point<int>& screenPos, const Time& time)
	{
		//DBG ("Mouse " + String (source.getIndex()) + " down: " + comp->getLocalPoint (nullptr, screenPos).toString() + " - Comp: " + String::toHexString ((int) comp));
		comp->internalMouseDown (source, comp->getLocalPoint (nullptr, screenPos), time);
	}

	void sendMouseDrag (Component* const comp, const Point<int>& screenPos, const Time& time)
	{
		//DBG ("Mouse " + String (source.getIndex()) + " drag: " + comp->getLocalPoint (nullptr, screenPos).toString() + " - Comp: " + String::toHexString ((int) comp));
		comp->internalMouseDrag (source, comp->getLocalPoint (nullptr, screenPos), time);
	}

	void sendMouseUp (Component* const comp, const Point<int>& screenPos, const Time& time)
	{
		//DBG ("Mouse " + String (source.getIndex()) + " up: " + comp->getLocalPoint (nullptr, screenPos).toString() + " - Comp: " + String::toHexString ((int) comp));
		comp->internalMouseUp (source, comp->getLocalPoint (nullptr, screenPos), time, getCurrentModifiers());
	}

	void sendMouseWheel (Component* const comp, const Point<int>& screenPos, const Time& time, float x, float y)
	{
		//DBG ("Mouse " + String (source.getIndex()) + " wheel: " + comp->getLocalPoint (nullptr, screenPos).toString() + " - Comp: " + String::toHexString ((int) comp));
		comp->internalMouseWheel (source, comp->getLocalPoint (nullptr, screenPos), time, x, y);
	}

	// (returns true if the button change caused a modal event loop)
	bool setButtons (const Point<int>& screenPos, const Time& time, const ModifierKeys& newButtonState)
	{
		if (buttonState == newButtonState)
			return false;

		setScreenPos (screenPos, time, false);

		// (ignore secondary clicks when there's already a button down)
		if (buttonState.isAnyMouseButtonDown() == newButtonState.isAnyMouseButtonDown())
		{
			buttonState = newButtonState;
			return false;
		}

		const int lastCounter = mouseEventCounter;

		if (buttonState.isAnyMouseButtonDown())
		{
			Component* const current = getComponentUnderMouse();

			if (current != nullptr)
				sendMouseUp (current, screenPos + unboundedMouseOffset, time);

			enableUnboundedMouseMovement (false, false);
		}

		buttonState = newButtonState;

		if (buttonState.isAnyMouseButtonDown())
		{
			Desktop::getInstance().incrementMouseClickCounter();

			Component* const current = getComponentUnderMouse();

			if (current != nullptr)
			{
				registerMouseDown (screenPos, time, current, buttonState);
				sendMouseDown (current, screenPos, time);
			}
		}

		return lastCounter != mouseEventCounter;
	}

	void setComponentUnderMouse (Component* const newComponent, const Point<int>& screenPos, const Time& time)
	{
		Component* current = getComponentUnderMouse();

		if (newComponent != current)
		{
			WeakReference<Component> safeNewComp (newComponent);
			const ModifierKeys originalButtonState (buttonState);

			if (current != nullptr)
			{
				setButtons (screenPos, time, ModifierKeys());
				sendMouseExit (current, screenPos, time);
				buttonState = originalButtonState;
			}

			componentUnderMouse = safeNewComp;
			current = getComponentUnderMouse();

			if (current != nullptr)
				sendMouseEnter (current, screenPos, time);

			revealCursor (false);
			setButtons (screenPos, time, originalButtonState);
		}
	}

	void setPeer (ComponentPeer* const newPeer, const Point<int>& screenPos, const Time& time)
	{
		ModifierKeys::updateCurrentModifiers();

		if (newPeer != lastPeer)
		{
			setComponentUnderMouse (nullptr, screenPos, time);
			lastPeer = newPeer;
			setComponentUnderMouse (findComponentAt (screenPos), screenPos, time);
		}
	}

	void setScreenPos (const Point<int>& newScreenPos, const Time& time, const bool forceUpdate)
	{
		if (! isDragging())
			setComponentUnderMouse (findComponentAt (newScreenPos), newScreenPos, time);

		if (newScreenPos != lastScreenPos || forceUpdate)
		{
			cancelPendingUpdate();

			lastScreenPos = newScreenPos;
			Component* const current = getComponentUnderMouse();

			if (current != nullptr)
			{
				if (isDragging())
				{
					registerMouseDrag (newScreenPos);
					sendMouseDrag (current, newScreenPos + unboundedMouseOffset, time);

					if (isUnboundedMouseModeOn)
						handleUnboundedDrag (current);
				}
				else
				{
					sendMouseMove (current, newScreenPos, time);
				}
			}

			revealCursor (false);
		}
	}

	void handleEvent (ComponentPeer* const newPeer, const Point<int>& positionWithinPeer, const Time& time, const ModifierKeys& newMods)
	{
		jassert (newPeer != nullptr);
		lastTime = time;
		++mouseEventCounter;
		const Point<int> screenPos (newPeer->localToGlobal (positionWithinPeer));

		if (isDragging() && newMods.isAnyMouseButtonDown())
		{
			setScreenPos (screenPos, time, false);
		}
		else
		{
			setPeer (newPeer, screenPos, time);

			ComponentPeer* peer = getPeer();
			if (peer != nullptr)
			{
				if (setButtons (screenPos, time, newMods))
					return; // some modal events have been dispatched, so the current event is now out-of-date

				peer = getPeer();
				if (peer != nullptr)
					setScreenPos (screenPos, time, false);
			}
		}
	}

	void handleWheel (ComponentPeer* const peer, const Point<int>& positionWithinPeer, const Time& time, float x, float y)
	{
		jassert (peer != nullptr);
		lastTime = time;
		++mouseEventCounter;
		const Point<int> screenPos (peer->localToGlobal (positionWithinPeer));

		setPeer (peer, screenPos, time);
		setScreenPos (screenPos, time, false);
		triggerFakeMove();

		if (! isDragging())
		{
			Component* current = getComponentUnderMouse();
			if (current != nullptr)
				sendMouseWheel (current, screenPos, time, x, y);
		}
	}

	Time getLastMouseDownTime() const noexcept
	{
		return Time (mouseDowns[0].time);
	}

	Point<int> getLastMouseDownPosition() const noexcept
	{
		return mouseDowns[0].position;
	}

	int getNumberOfMultipleClicks() const noexcept
	{
		int numClicks = 0;

		if (mouseDowns[0].time != Time())
		{
			if (! mouseMovedSignificantlySincePressed)
				++numClicks;

			for (int i = 1; i < numElementsInArray (mouseDowns); ++i)
			{
				if (mouseDowns[0].canBePartOfMultipleClickWith (mouseDowns[i], (int) (MouseEvent::getDoubleClickTimeout() * (1.0 + 0.25 * (i - 1)))))
					++numClicks;
				else
					break;
			}
		}

		return numClicks;
	}

	bool hasMouseMovedSignificantlySincePressed() const noexcept
	{
		return mouseMovedSignificantlySincePressed
				|| lastTime > mouseDowns[0].time + RelativeTime::milliseconds (300);
	}

	void triggerFakeMove()
	{
		triggerAsyncUpdate();
	}

	void handleAsyncUpdate()
	{
		setScreenPos (lastScreenPos, jmax (lastTime, Time::getCurrentTime()), true);
	}

	void enableUnboundedMouseMovement (bool enable, bool keepCursorVisibleUntilOffscreen)
	{
		enable = enable && isDragging();
		isCursorVisibleUntilOffscreen = keepCursorVisibleUntilOffscreen;

		if (enable != isUnboundedMouseModeOn)
		{
			if ((! enable) && ((! isCursorVisibleUntilOffscreen) || ! unboundedMouseOffset.isOrigin()))
			{
				// when released, return the mouse to within the component's bounds
				Component* current = getComponentUnderMouse();
				if (current != nullptr)
					Desktop::setMousePosition (current->getScreenBounds()
												 .getConstrainedPoint (lastScreenPos));
			}

			isUnboundedMouseModeOn = enable;
			unboundedMouseOffset = Point<int>();

			revealCursor (true);
		}
	}

	void handleUnboundedDrag (Component* current)
	{
		const Rectangle<int> screenArea (current->getParentMonitorArea().expanded (-2, -2));

		if (! screenArea.contains (lastScreenPos))
		{
			const Point<int> componentCentre (current->getScreenBounds().getCentre());
			unboundedMouseOffset += (lastScreenPos - componentCentre);
			Desktop::setMousePosition (componentCentre);
		}
		else if (isCursorVisibleUntilOffscreen
				  && (! unboundedMouseOffset.isOrigin())
				  && screenArea.contains (lastScreenPos + unboundedMouseOffset))
		{
			Desktop::setMousePosition (lastScreenPos + unboundedMouseOffset);
			unboundedMouseOffset = Point<int>();
		}
	}

	void showMouseCursor (MouseCursor cursor, bool forcedUpdate)
	{
		if (isUnboundedMouseModeOn && ((! unboundedMouseOffset.isOrigin()) || ! isCursorVisibleUntilOffscreen))
		{
			cursor = MouseCursor::NoCursor;
			forcedUpdate = true;
		}

		if (forcedUpdate || cursor.getHandle() != currentCursorHandle)
		{
			currentCursorHandle = cursor.getHandle();
			cursor.showInWindow (getPeer());
		}
	}

	void hideCursor()
	{
		showMouseCursor (MouseCursor::NoCursor, true);
	}

	void revealCursor (bool forcedUpdate)
	{
		MouseCursor mc (MouseCursor::NormalCursor);

		Component* current = getComponentUnderMouse();
		if (current != nullptr)
			mc = current->getLookAndFeel().getMouseCursorFor (*current);

		showMouseCursor (mc, forcedUpdate);
	}

	const int index;
	const bool isMouseDevice;
	Point<int> lastScreenPos;
	ModifierKeys buttonState;

private:
	MouseInputSource& source;
	WeakReference<Component> componentUnderMouse;
	ComponentPeer* lastPeer;

	Point<int> unboundedMouseOffset;
	bool isUnboundedMouseModeOn, isCursorVisibleUntilOffscreen;
	void* currentCursorHandle;
	int mouseEventCounter;

	struct RecentMouseDown
	{
		RecentMouseDown()  : component (nullptr)
		{
		}

		Point<int> position;
		Time time;
		Component* component;
		ModifierKeys buttons;

		bool canBePartOfMultipleClickWith (const RecentMouseDown& other, const int maxTimeBetweenMs) const
		{
			return time - other.time < RelativeTime::milliseconds (maxTimeBetweenMs)
					&& abs (position.getX() - other.position.getX()) < 8
					&& abs (position.getY() - other.position.getY()) < 8
					&& buttons == other.buttons;;
		}
	};

	RecentMouseDown mouseDowns[4];
	bool mouseMovedSignificantlySincePressed;
	Time lastTime;

	void registerMouseDown (const Point<int>& screenPos, const Time& time,
							Component* const component, const ModifierKeys& modifiers) noexcept
	{
		for (int i = numElementsInArray (mouseDowns); --i > 0;)
			mouseDowns[i] = mouseDowns[i - 1];

		mouseDowns[0].position = screenPos;
		mouseDowns[0].time = time;
		mouseDowns[0].component = component;
		mouseDowns[0].buttons = modifiers.withOnlyMouseButtons();
		mouseMovedSignificantlySincePressed = false;
	}

	void registerMouseDrag (const Point<int>& screenPos) noexcept
	{
		mouseMovedSignificantlySincePressed = mouseMovedSignificantlySincePressed
			   || mouseDowns[0].position.getDistanceFrom (screenPos) >= 4;
	}

	JUCE_DECLARE_NON_COPYABLE (MouseInputSourceInternal);
};

MouseInputSource::MouseInputSource (const int index, const bool isMouseDevice)
{
	pimpl = new MouseInputSourceInternal (*this, index, isMouseDevice);
}

MouseInputSource::~MouseInputSource()
{
}

bool MouseInputSource::isMouse() const				  { return pimpl->isMouseDevice; }
bool MouseInputSource::isTouch() const				  { return ! isMouse(); }
bool MouseInputSource::canHover() const				 { return isMouse(); }
bool MouseInputSource::hasMouseWheel() const				{ return isMouse(); }
int MouseInputSource::getIndex() const				  { return pimpl->index; }
bool MouseInputSource::isDragging() const				   { return pimpl->isDragging(); }
const Point<int> MouseInputSource::getScreenPosition() const		{ return pimpl->getScreenPosition(); }
const ModifierKeys MouseInputSource::getCurrentModifiers() const	{ return pimpl->getCurrentModifiers(); }
Component* MouseInputSource::getComponentUnderMouse() const		 { return pimpl->getComponentUnderMouse(); }
void MouseInputSource::triggerFakeMove() const			  { pimpl->triggerFakeMove(); }
int MouseInputSource::getNumberOfMultipleClicks() const noexcept	{ return pimpl->getNumberOfMultipleClicks(); }
Time MouseInputSource::getLastMouseDownTime() const noexcept		{ return pimpl->getLastMouseDownTime(); }
Point<int> MouseInputSource::getLastMouseDownPosition() const noexcept  { return pimpl->getLastMouseDownPosition(); }
bool MouseInputSource::hasMouseMovedSignificantlySincePressed() const noexcept  { return pimpl->hasMouseMovedSignificantlySincePressed(); }
bool MouseInputSource::canDoUnboundedMovement() const noexcept	  { return isMouse(); }
void MouseInputSource::enableUnboundedMouseMovement (bool isEnabled, bool keepCursorVisibleUntilOffscreen)	{ pimpl->enableUnboundedMouseMovement (isEnabled, keepCursorVisibleUntilOffscreen); }
bool MouseInputSource::hasMouseCursor() const noexcept		  { return isMouse(); }
void MouseInputSource::showMouseCursor (const MouseCursor& cursor)	  { pimpl->showMouseCursor (cursor, false); }
void MouseInputSource::hideCursor()					 { pimpl->hideCursor(); }
void MouseInputSource::revealCursor()				   { pimpl->revealCursor (false); }
void MouseInputSource::forceMouseCursorUpdate()			 { pimpl->revealCursor (true); }

void MouseInputSource::handleEvent (ComponentPeer* peer, const Point<int>& positionWithinPeer, const int64 time, const ModifierKeys& mods)
{
	pimpl->handleEvent (peer, positionWithinPeer, Time (time), mods.withOnlyMouseButtons());
}

void MouseInputSource::handleWheel (ComponentPeer* const peer, const Point<int>& positionWithinPeer, const int64 time, const float x, const float y)
{
	pimpl->handleWheel (peer, positionWithinPeer, Time (time), x, y);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MouseInputSource.cpp ***/


/*** Start of inlined file: juce_MouseListener.cpp ***/
BEGIN_JUCE_NAMESPACE

void MouseListener::mouseEnter (const MouseEvent&)
{
}

void MouseListener::mouseExit (const MouseEvent&)
{
}

void MouseListener::mouseDown (const MouseEvent&)
{
}

void MouseListener::mouseUp (const MouseEvent&)
{
}

void MouseListener::mouseDrag (const MouseEvent&)
{
}

void MouseListener::mouseMove (const MouseEvent&)
{
}

void MouseListener::mouseDoubleClick (const MouseEvent&)
{
}

void MouseListener::mouseWheelMove (const MouseEvent&, float, float)
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MouseListener.cpp ***/


/*** Start of inlined file: juce_BooleanPropertyComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

BooleanPropertyComponent::BooleanPropertyComponent (const String& name,
													const String& buttonTextWhenTrue,
													const String& buttonTextWhenFalse)
	: PropertyComponent (name),
	  onText (buttonTextWhenTrue),
	  offText (buttonTextWhenFalse)
{
	addAndMakeVisible (&button);
	button.setClickingTogglesState (false);
	button.addListener (this);
}

BooleanPropertyComponent::BooleanPropertyComponent (const Value& valueToControl,
													const String& name,
													const String& buttonText)
	: PropertyComponent (name),
	  onText (buttonText),
	  offText (buttonText)
{
	addAndMakeVisible (&button);
	button.setClickingTogglesState (false);
	button.setButtonText (buttonText);
	button.getToggleStateValue().referTo (valueToControl);
	button.setClickingTogglesState (true);
}

BooleanPropertyComponent::~BooleanPropertyComponent()
{
}

void BooleanPropertyComponent::setState (const bool newState)
{
	button.setToggleState (newState, true);
}

bool BooleanPropertyComponent::getState() const
{
	return button.getToggleState();
}

void BooleanPropertyComponent::paint (Graphics& g)
{
	PropertyComponent::paint (g);

	g.setColour (Colours::white);
	g.fillRect (button.getBounds());

	g.setColour (findColour (ComboBox::outlineColourId));
	g.drawRect (button.getBounds());
}

void BooleanPropertyComponent::refresh()
{
	button.setToggleState (getState(), false);
	button.setButtonText (button.getToggleState() ? onText : offText);
}

void BooleanPropertyComponent::buttonClicked (Button*)
{
	setState (! getState());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_BooleanPropertyComponent.cpp ***/


/*** Start of inlined file: juce_ButtonPropertyComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

ButtonPropertyComponent::ButtonPropertyComponent (const String& name,
												  const bool triggerOnMouseDown)
	: PropertyComponent (name)
{
	addAndMakeVisible (&button);
	button.setTriggeredOnMouseDown (triggerOnMouseDown);
	button.addListener (this);
}

ButtonPropertyComponent::~ButtonPropertyComponent()
{
}

void ButtonPropertyComponent::refresh()
{
	button.setButtonText (getButtonText());
}

void ButtonPropertyComponent::buttonClicked (Button*)
{
	buttonClicked();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ButtonPropertyComponent.cpp ***/


/*** Start of inlined file: juce_ChoicePropertyComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

class ChoicePropertyComponent::RemapperValueSource	: public Value::ValueSource,
														public ValueListener
{
public:
	RemapperValueSource (const Value& sourceValue_, const Array<var>& mappings_)
	   : sourceValue (sourceValue_),
		 mappings (mappings_)
	{
		sourceValue.addListener (this);
	}

	~RemapperValueSource() {}

	var getValue() const
	{
		return mappings.indexOf (sourceValue.getValue()) + 1;
	}

	void setValue (const var& newValue)
	{
		const var remappedVal (mappings [(int) newValue - 1]);

		if (remappedVal != sourceValue)
			sourceValue = remappedVal;
	}

	void valueChanged (Value&)
	{
		sendChangeMessage (true);
	}

protected:

	Value sourceValue;
	Array<var> mappings;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (RemapperValueSource);
};

ChoicePropertyComponent::ChoicePropertyComponent (const String& name)
	: PropertyComponent (name),
	  isCustomClass (true)
{
}

ChoicePropertyComponent::ChoicePropertyComponent (const Value& valueToControl,
												  const String& name,
												  const StringArray& choices_,
												  const Array <var>& correspondingValues)
	: PropertyComponent (name),
	  choices (choices_),
	  isCustomClass (false)
{
	// The array of corresponding values must contain one value for each of the items in
	// the choices array!
	jassert (correspondingValues.size() == choices.size());

	createComboBox();

	comboBox.getSelectedIdAsValue().referTo (Value (new RemapperValueSource (valueToControl, correspondingValues)));
}

ChoicePropertyComponent::~ChoicePropertyComponent()
{
}

void ChoicePropertyComponent::createComboBox()
{
	addAndMakeVisible (&comboBox);

	for (int i = 0; i < choices.size(); ++i)
	{
		if (choices[i].isNotEmpty())
			comboBox.addItem (choices[i], i + 1);
		else
			comboBox.addSeparator();
	}

	comboBox.setEditableText (false);
}

void ChoicePropertyComponent::setIndex (const int /*newIndex*/)
{
	jassertfalse; // you need to override this method in your subclass!
}

int ChoicePropertyComponent::getIndex() const
{
	jassertfalse; // you need to override this method in your subclass!
	return -1;
}

const StringArray& ChoicePropertyComponent::getChoices() const
{
	return choices;
}

void ChoicePropertyComponent::refresh()
{
	if (isCustomClass)
	{
		if (! comboBox.isVisible())
		{
			createComboBox();
			comboBox.addListener (this);
		}

		comboBox.setSelectedId (getIndex() + 1, true);
	}
}

void ChoicePropertyComponent::comboBoxChanged (ComboBox*)
{
	if (isCustomClass)
	{
		const int newIndex = comboBox.getSelectedId() - 1;

		if (newIndex != getIndex())
			setIndex (newIndex);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ChoicePropertyComponent.cpp ***/


/*** Start of inlined file: juce_PropertyComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

PropertyComponent::PropertyComponent (const String& name,
									  const int preferredHeight_)
	: Component (name),
	  preferredHeight (preferredHeight_)
{
	jassert (name.isNotEmpty());
}

PropertyComponent::~PropertyComponent()
{
}

void PropertyComponent::paint (Graphics& g)
{
	getLookAndFeel().drawPropertyComponentBackground (g, getWidth(), getHeight(), *this);
	getLookAndFeel().drawPropertyComponentLabel (g, getWidth(), getHeight(), *this);
}

void PropertyComponent::resized()
{
	if (getNumChildComponents() > 0)
		getChildComponent (0)->setBounds (getLookAndFeel().getPropertyComponentContentPosition (*this));
}

void PropertyComponent::enablementChanged()
{
	repaint();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_PropertyComponent.cpp ***/


/*** Start of inlined file: juce_PropertyPanel.cpp ***/
BEGIN_JUCE_NAMESPACE

class PropertySectionComponent  : public Component
{
public:
	PropertySectionComponent (const String& sectionTitle,
							  const Array <PropertyComponent*>& newProperties,
							  const bool sectionIsOpen_)
		: Component (sectionTitle),
		  titleHeight (sectionTitle.isNotEmpty() ? 22 : 0),
		  sectionIsOpen (sectionIsOpen_)
	{
		propertyComps.addArray (newProperties);

		for (int i = propertyComps.size(); --i >= 0;)
		{
			addAndMakeVisible (propertyComps.getUnchecked(i));
			propertyComps.getUnchecked(i)->refresh();
		}
	}

	~PropertySectionComponent()
	{
		propertyComps.clear();
	}

	void paint (Graphics& g)
	{
		if (titleHeight > 0)
			getLookAndFeel().drawPropertyPanelSectionHeader (g, getName(), isOpen(), getWidth(), titleHeight);
	}

	void resized()
	{
		int y = titleHeight;

		for (int i = 0; i < propertyComps.size(); ++i)
		{
			PropertyComponent* const pec = propertyComps.getUnchecked (i);
			pec->setBounds (1, y, getWidth() - 2, pec->getPreferredHeight());
			y = pec->getBottom();
		}
	}

	int getPreferredHeight() const
	{
		int y = titleHeight;

		if (isOpen())
		{
			for (int i = propertyComps.size(); --i >= 0;)
				y += propertyComps.getUnchecked(i)->getPreferredHeight();
		}

		return y;
	}

	void setOpen (const bool open)
	{
		if (sectionIsOpen != open)
		{
			sectionIsOpen = open;

			for (int i = propertyComps.size(); --i >= 0;)
				propertyComps.getUnchecked(i)->setVisible (open);

			PropertyPanel* const pp = findParentComponentOfClass ((PropertyPanel*) nullptr);

			if (pp != nullptr)
				pp->resized();
		}
	}

	bool isOpen() const
	{
		return sectionIsOpen;
	}

	void refreshAll() const
	{
		for (int i = propertyComps.size(); --i >= 0;)
			propertyComps.getUnchecked (i)->refresh();
	}

	void mouseUp (const MouseEvent& e)
	{
		if (e.getMouseDownX() < titleHeight
			 && e.x < titleHeight
			 && e.y < titleHeight
			 && e.getNumberOfClicks() != 2)
		{
			setOpen (! isOpen());
		}
	}

	void mouseDoubleClick (const MouseEvent& e)
	{
		if (e.y < titleHeight)
			setOpen (! isOpen());
	}

private:
	OwnedArray <PropertyComponent> propertyComps;
	int titleHeight;
	bool sectionIsOpen;

	JUCE_DECLARE_NON_COPYABLE (PropertySectionComponent);
};

class PropertyPanel::PropertyHolderComponent  : public Component
{
public:
	PropertyHolderComponent() {}

	void paint (Graphics&) {}

	void updateLayout (int width)
	{
		int y = 0;

		for (int i = 0; i < sections.size(); ++i)
		{
			PropertySectionComponent* const section = sections.getUnchecked(i);

			section->setBounds (0, y, width, section->getPreferredHeight());
			y = section->getBottom();
		}

		setSize (width, y);
		repaint();
	}

	void refreshAll() const
	{
		for (int i = 0; i < sections.size(); ++i)
			sections.getUnchecked(i)->refreshAll();
	}

	void clear()
	{
		sections.clear();
	}

	void addSection (PropertySectionComponent* newSection)
	{
		sections.add (newSection);
		addAndMakeVisible (newSection, 0);
	}

	int getNumSections() const noexcept				 { return sections.size(); }
	PropertySectionComponent* getSection (const int index) const	{ return sections [index]; }

private:
	OwnedArray<PropertySectionComponent> sections;

	JUCE_DECLARE_NON_COPYABLE (PropertyHolderComponent);
};

PropertyPanel::PropertyPanel()
{
	messageWhenEmpty = TRANS("(nothing selected)");

	addAndMakeVisible (&viewport);
	viewport.setViewedComponent (propertyHolderComponent = new PropertyHolderComponent());
	viewport.setFocusContainer (true);
}

PropertyPanel::~PropertyPanel()
{
	clear();
}

void PropertyPanel::paint (Graphics& g)
{
	if (propertyHolderComponent->getNumSections() == 0)
	{
		g.setColour (Colours::black.withAlpha (0.5f));
		g.setFont (14.0f);
		g.drawText (messageWhenEmpty, 0, 0, getWidth(), 30,
					Justification::centred, true);
	}
}

void PropertyPanel::resized()
{
	viewport.setBounds (getLocalBounds());
	updatePropHolderLayout();
}

void PropertyPanel::clear()
{
	if (propertyHolderComponent->getNumSections() > 0)
	{
		propertyHolderComponent->clear();
		repaint();
	}
}

void PropertyPanel::addProperties (const Array <PropertyComponent*>& newProperties)
{
	if (propertyHolderComponent->getNumSections() == 0)
		repaint();

	propertyHolderComponent->addSection (new PropertySectionComponent (String::empty, newProperties, true));
	updatePropHolderLayout();
}

void PropertyPanel::addSection (const String& sectionTitle,
								const Array <PropertyComponent*>& newProperties,
								const bool shouldBeOpen)
{
	jassert (sectionTitle.isNotEmpty());

	if (propertyHolderComponent->getNumSections() == 0)
		repaint();

	propertyHolderComponent->addSection (new PropertySectionComponent (sectionTitle, newProperties, shouldBeOpen));
	updatePropHolderLayout();
}

void PropertyPanel::updatePropHolderLayout() const
{
	const int maxWidth = viewport.getMaximumVisibleWidth();
	propertyHolderComponent->updateLayout (maxWidth);

	const int newMaxWidth = viewport.getMaximumVisibleWidth();
	if (maxWidth != newMaxWidth)
	{
		// need to do this twice because of scrollbars changing the size, etc.
		propertyHolderComponent->updateLayout (newMaxWidth);
	}
}

void PropertyPanel::refreshAll() const
{
	propertyHolderComponent->refreshAll();
}

StringArray PropertyPanel::getSectionNames() const
{
	StringArray s;

	for (int i = 0; i < propertyHolderComponent->getNumSections(); ++i)
	{
		PropertySectionComponent* const section = propertyHolderComponent->getSection (i);

		if (section->getName().isNotEmpty())
			s.add (section->getName());
	}

	return s;
}

bool PropertyPanel::isSectionOpen (const int sectionIndex) const
{
	int index = 0;

	for (int i = 0; i < propertyHolderComponent->getNumSections(); ++i)
	{
		PropertySectionComponent* const section = propertyHolderComponent->getSection (i);

		if (section->getName().isNotEmpty())
		{
			if (index == sectionIndex)
				return section->isOpen();

			++index;
		}
	}

	return false;
}

void PropertyPanel::setSectionOpen (const int sectionIndex, const bool shouldBeOpen)
{
	int index = 0;

	for (int i = 0; i < propertyHolderComponent->getNumSections(); ++i)
	{
		PropertySectionComponent* const section = propertyHolderComponent->getSection (i);

		if (section->getName().isNotEmpty())
		{
			if (index == sectionIndex)
			{
				section->setOpen (shouldBeOpen);
				break;
			}

			++index;
		}
	}
}

void PropertyPanel::setSectionEnabled (const int sectionIndex, const bool shouldBeEnabled)
{
	int index = 0;

	for (int i = 0; i < propertyHolderComponent->getNumSections(); ++i)
	{
		PropertySectionComponent* const section = propertyHolderComponent->getSection (i);

		if (section->getName().isNotEmpty())
		{
			if (index == sectionIndex)
			{
				section->setEnabled (shouldBeEnabled);
				break;
			}

			++index;
		}
	}
}

XmlElement* PropertyPanel::getOpennessState() const
{
	XmlElement* const xml = new XmlElement ("PROPERTYPANELSTATE");

	xml->setAttribute ("scrollPos", viewport.getViewPositionY());

	const StringArray sections (getSectionNames());

	for (int i = 0; i < sections.size(); ++i)
	{
		if (sections[i].isNotEmpty())
		{
			XmlElement* const e = xml->createNewChildElement ("SECTION");
			e->setAttribute ("name", sections[i]);
			e->setAttribute ("open", isSectionOpen (i) ? 1 : 0);
		}
	}

	return xml;
}

void PropertyPanel::restoreOpennessState (const XmlElement& xml)
{
	if (xml.hasTagName ("PROPERTYPANELSTATE"))
	{
		const StringArray sections (getSectionNames());

		forEachXmlChildElementWithTagName (xml, e, "SECTION")
		{
			setSectionOpen (sections.indexOf (e->getStringAttribute ("name")),
							e->getBoolAttribute ("open"));
		}

		viewport.setViewPosition (viewport.getViewPositionX(),
								  xml.getIntAttribute ("scrollPos", viewport.getViewPositionY()));
	}
}

void PropertyPanel::setMessageWhenEmpty (const String& newMessage)
{
	if (messageWhenEmpty != newMessage)
	{
		messageWhenEmpty = newMessage;
		repaint();
	}
}

const String& PropertyPanel::getMessageWhenEmpty() const
{
	return messageWhenEmpty;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_PropertyPanel.cpp ***/


/*** Start of inlined file: juce_SliderPropertyComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

SliderPropertyComponent::SliderPropertyComponent (const String& name,
												  const double rangeMin,
												  const double rangeMax,
												  const double interval,
												  const double skewFactor)
	: PropertyComponent (name)
{
	addAndMakeVisible (&slider);

	slider.setRange (rangeMin, rangeMax, interval);
	slider.setSkewFactor (skewFactor);
	slider.setSliderStyle (Slider::LinearBar);

	slider.addListener (this);
}

SliderPropertyComponent::SliderPropertyComponent (const Value& valueToControl,
												  const String& name,
												  const double rangeMin,
												  const double rangeMax,
												  const double interval,
												  const double skewFactor)
	: PropertyComponent (name)
{
	addAndMakeVisible (&slider);

	slider.setRange (rangeMin, rangeMax, interval);
	slider.setSkewFactor (skewFactor);
	slider.setSliderStyle (Slider::LinearBar);

	slider.getValueObject().referTo (valueToControl);
}

SliderPropertyComponent::~SliderPropertyComponent()
{
}

void SliderPropertyComponent::setValue (const double /*newValue*/)
{
}

double SliderPropertyComponent::getValue() const
{
	return slider.getValue();
}

void SliderPropertyComponent::refresh()
{
	slider.setValue (getValue(), false);
}

void SliderPropertyComponent::sliderValueChanged (Slider*)
{
	if (getValue() != slider.getValue())
		setValue (slider.getValue());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_SliderPropertyComponent.cpp ***/


/*** Start of inlined file: juce_TextPropertyComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

class TextPropLabel  : public Label
{
public:
	TextPropLabel (TextPropertyComponent& owner_,
				   const int maxChars_, const bool isMultiline_)
		: Label (String::empty, String::empty),
		  owner (owner_),
		  maxChars (maxChars_),
		  isMultiline (isMultiline_)
	{
		setEditable (true, true, false);

		setColour (backgroundColourId, Colours::white);
		setColour (outlineColourId, findColour (ComboBox::outlineColourId));
	}

	TextEditor* createEditorComponent()
	{
		TextEditor* const textEditor = Label::createEditorComponent();
		textEditor->setInputRestrictions (maxChars);

		if (isMultiline)
		{
			textEditor->setMultiLine (true, true);
			textEditor->setReturnKeyStartsNewLine (true);
		}

		return textEditor;
	}

	void textWasEdited()
	{
		owner.textWasEdited();
	}

private:
	TextPropertyComponent& owner;
	int maxChars;
	bool isMultiline;
};

TextPropertyComponent::TextPropertyComponent (const String& name,
											  const int maxNumChars,
											  const bool isMultiLine)
	: PropertyComponent (name)
{
	createEditor (maxNumChars, isMultiLine);
}

TextPropertyComponent::TextPropertyComponent (const Value& valueToControl,
											  const String& name,
											  const int maxNumChars,
											  const bool isMultiLine)
	: PropertyComponent (name)
{
	createEditor (maxNumChars, isMultiLine);

	textEditor->getTextValue().referTo (valueToControl);
}

TextPropertyComponent::~TextPropertyComponent()
{
}

void TextPropertyComponent::setText (const String& newText)
{
	textEditor->setText (newText, true);
}

const String TextPropertyComponent::getText() const
{
	return textEditor->getText();
}

void TextPropertyComponent::createEditor (const int maxNumChars, const bool isMultiLine)
{
	addAndMakeVisible (textEditor = new TextPropLabel (*this, maxNumChars, isMultiLine));

	if (isMultiLine)
	{
		textEditor->setJustificationType (Justification::topLeft);
		preferredHeight = 120;
	}
}

void TextPropertyComponent::refresh()
{
	textEditor->setText (getText(), false);
}

void TextPropertyComponent::textWasEdited()
{
	const String newText (textEditor->getText());

	if (getText() != newText)
		setText (newText);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TextPropertyComponent.cpp ***/

/*** Start of inlined file: juce_BubbleComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

BubbleComponent::BubbleComponent()
  : side (0),
	allowablePlacements (above | below | left | right),
	arrowTipX (0.0f),
	arrowTipY (0.0f)
{
	setInterceptsMouseClicks (false, false);

	shadow.setShadowProperties (5.0f, 0.35f, 0, 0);
	setComponentEffect (&shadow);
}

BubbleComponent::~BubbleComponent()
{
}

void BubbleComponent::paint (Graphics& g)
{
	int x = content.getX();
	int y = content.getY();
	int w = content.getWidth();
	int h = content.getHeight();

	int cw, ch;
	getContentSize (cw, ch);

	if (side == 3)
		x += w - cw;
	else if (side != 1)
		x += (w - cw) / 2;

	w = cw;

	if (side == 2)
		y += h - ch;
	else if (side != 0)
		y += (h - ch) / 2;

	h = ch;

	getLookAndFeel().drawBubble (g, arrowTipX, arrowTipY,
								 (float) x, (float) y,
								 (float) w, (float) h);

	const int cx = x + (w - cw) / 2;
	const int cy = y + (h - ch) / 2;

	const int indent = 3;

	g.setOrigin (cx + indent, cy + indent);
	g.reduceClipRegion (0, 0, cw - indent * 2, ch - indent * 2);

	paintContent (g, cw - indent * 2, ch - indent * 2);
}

void BubbleComponent::setAllowedPlacement (const int newPlacement)
{
	allowablePlacements = newPlacement;
}

void BubbleComponent::setPosition (Component* componentToPointTo)
{
	jassert (componentToPointTo != nullptr);

	Point<int> pos;

	if (getParentComponent() != nullptr)
		pos = getParentComponent()->getLocalPoint (componentToPointTo, pos);
	else
		pos = componentToPointTo->localPointToGlobal (pos);

	setPosition (Rectangle<int> (pos.getX(), pos.getY(), componentToPointTo->getWidth(), componentToPointTo->getHeight()));
}

void BubbleComponent::setPosition (const int arrowTipX_,
								   const int arrowTipY_)
{
	setPosition (Rectangle<int> (arrowTipX_, arrowTipY_, 1, 1));
}

void BubbleComponent::setPosition (const Rectangle<int>& rectangleToPointTo)
{
	Rectangle<int> availableSpace (getParentComponent() != nullptr ? getParentComponent()->getLocalBounds()
																   : getParentMonitorArea());
	int x = 0;
	int y = 0;
	int w = 150;
	int h = 30;

	getContentSize (w, h);
	w += 30;
	h += 30;

	const float edgeIndent = 2.0f;
	const int arrowLength = jmin (10, h / 3, w / 3);

	int spaceAbove = ((allowablePlacements & above) != 0) ? jmax (0, rectangleToPointTo.getY() - availableSpace.getY()) : -1;
	int spaceBelow = ((allowablePlacements & below) != 0) ? jmax (0, availableSpace.getBottom() - rectangleToPointTo.getBottom()) : -1;
	int spaceLeft  = ((allowablePlacements & left)  != 0) ? jmax (0, rectangleToPointTo.getX() - availableSpace.getX()) : -1;
	int spaceRight = ((allowablePlacements & right) != 0) ? jmax (0, availableSpace.getRight() - rectangleToPointTo.getRight()) : -1;

	// look at whether the component is elongated, and if so, try to position next to its longer dimension.
	if (rectangleToPointTo.getWidth() > rectangleToPointTo.getHeight() * 2
		 && (spaceAbove > h + 20 || spaceBelow > h + 20))
	{
		spaceLeft = spaceRight = 0;
	}
	else if (rectangleToPointTo.getWidth() < rectangleToPointTo.getHeight() / 2
			  && (spaceLeft > w + 20 || spaceRight > w + 20))
	{
		spaceAbove = spaceBelow = 0;
	}

	if (jmax (spaceAbove, spaceBelow) >= jmax (spaceLeft, spaceRight))
	{
		x = rectangleToPointTo.getX() + (rectangleToPointTo.getWidth() - w) / 2;
		arrowTipX = w * 0.5f;
		content.setSize (w, h - arrowLength);

		if (spaceAbove >= spaceBelow)
		{
			// above
			y = rectangleToPointTo.getY() - h;
			content.setPosition (0, 0);
			arrowTipY = h - edgeIndent;
			side = 2;
		}
		else
		{
			// below
			y = rectangleToPointTo.getBottom();
			content.setPosition (0, arrowLength);
			arrowTipY = edgeIndent;
			side = 0;
		}
	}
	else
	{
		y = rectangleToPointTo.getY() + (rectangleToPointTo.getHeight() - h) / 2;
		arrowTipY = h * 0.5f;
		content.setSize (w - arrowLength, h);

		if (spaceLeft > spaceRight)
		{
			// on the left
			x = rectangleToPointTo.getX() - w;
			content.setPosition (0, 0);
			arrowTipX = w - edgeIndent;
			side = 3;
		}
		else
		{
			// on the right
			x = rectangleToPointTo.getRight();
			content.setPosition (arrowLength, 0);
			arrowTipX = edgeIndent;
			side = 1;
		}
	}

	setBounds (x, y, w, h);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_BubbleComponent.cpp ***/


/*** Start of inlined file: juce_BubbleMessageComponent.cpp ***/
BEGIN_JUCE_NAMESPACE

BubbleMessageComponent::BubbleMessageComponent (int fadeOutLengthMs)
	: fadeOutLength (fadeOutLengthMs),
	  deleteAfterUse (false)
{
}

BubbleMessageComponent::~BubbleMessageComponent()
{
	Desktop::getInstance().getAnimator().fadeOut (this, fadeOutLength);
}

void BubbleMessageComponent::showAt (int x, int y,
									 const String& text,
									 const int numMillisecondsBeforeRemoving,
									 const bool removeWhenMouseClicked,
									 const bool deleteSelfAfterUse)
{
	textLayout.clear();
	textLayout.setText (text, Font (14.0f));
	textLayout.layout (256, Justification::centredLeft, true);

	setPosition (x, y);

	init (numMillisecondsBeforeRemoving, removeWhenMouseClicked, deleteSelfAfterUse);
}

void BubbleMessageComponent::showAt (Component* const component,
									 const String& text,
									 const int numMillisecondsBeforeRemoving,
									 const bool removeWhenMouseClicked,
									 const bool deleteSelfAfterUse)
{
	textLayout.clear();
	textLayout.setText (text, Font (14.0f));
	textLayout.layout (256, Justification::centredLeft, true);

	setPosition (component);

	init (numMillisecondsBeforeRemoving, removeWhenMouseClicked, deleteSelfAfterUse);
}

void BubbleMessageComponent::init (const int numMillisecondsBeforeRemoving,
								   const bool removeWhenMouseClicked,
								   const bool deleteSelfAfterUse)
{
	setVisible (true);

	deleteAfterUse = deleteSelfAfterUse;

	if (numMillisecondsBeforeRemoving > 0)
		expiryTime = Time::getMillisecondCounter() + numMillisecondsBeforeRemoving;
	else
		expiryTime = 0;

	startTimer (77);

	mouseClickCounter = Desktop::getInstance().getMouseButtonClickCounter();

	if (! (removeWhenMouseClicked && isShowing()))
		mouseClickCounter += 0xfffff;

	repaint();
}

void BubbleMessageComponent::getContentSize (int& w, int& h)
{
	w = textLayout.getWidth() + 16;
	h = textLayout.getHeight() + 16;
}

void BubbleMessageComponent::paintContent (Graphics& g, int w, int h)
{
	g.setColour (findColour (TooltipWindow::textColourId));

	textLayout.drawWithin (g, 0, 0, w, h, Justification::centred);
}

void BubbleMessageComponent::timerCallback()
{
	if (Desktop::getInstance().getMouseButtonClickCounter() > mouseClickCounter)
	{
		stopTimer();
		setVisible (false);

		if (deleteAfterUse)
			delete this;
	}
	else if (expiryTime != 0 && Time::getMillisecondCounter() > expiryTime)
	{
		stopTimer();

		if (deleteAfterUse)
			delete this;
		else
			Desktop::getInstance().getAnimator().fadeOut (this, fadeOutLength);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_BubbleMessageComponent.cpp ***/


/*** Start of inlined file: juce_ColourSelector.cpp ***/
BEGIN_JUCE_NAMESPACE

class ColourComponentSlider  : public Slider
{
public:
	ColourComponentSlider (const String& name)
		: Slider (name)
	{
		setRange (0.0, 255.0, 1.0);
	}

	const String getTextFromValue (double value)
	{
		return String::toHexString ((int) value).toUpperCase().paddedLeft ('0', 2);
	}

	double getValueFromText (const String& text)
	{
		return (double) text.getHexValue32();
	}

private:
	JUCE_DECLARE_NON_COPYABLE (ColourComponentSlider);
};

class ColourSpaceMarker  : public Component
{
public:
	ColourSpaceMarker()
	{
		setInterceptsMouseClicks (false, false);
	}

	void paint (Graphics& g)
	{
		g.setColour (Colour::greyLevel (0.1f));
		g.drawEllipse (1.0f, 1.0f, getWidth() - 2.0f, getHeight() - 2.0f, 1.0f);
		g.setColour (Colour::greyLevel (0.9f));
		g.drawEllipse (2.0f, 2.0f, getWidth() - 4.0f, getHeight() - 4.0f, 1.0f);
	}

private:
	JUCE_DECLARE_NON_COPYABLE (ColourSpaceMarker);
};

class ColourSelector::ColourSpaceView  : public Component
{
public:
	ColourSpaceView (ColourSelector& owner_,
					 float& h_, float& s_, float& v_,
					 const int edgeSize)
		: owner (owner_),
		  h (h_), s (s_), v (v_),
		  lastHue (0.0f),
		  edge (edgeSize)
	{
		addAndMakeVisible (&marker);
		setMouseCursor (MouseCursor::CrosshairCursor);
	}

	void paint (Graphics& g)
	{
		if (colours.isNull())
		{
			const int width = getWidth() / 2;
			const int height = getHeight() / 2;
			colours = Image (Image::RGB, width, height, false);

			Image::BitmapData pixels (colours, Image::BitmapData::writeOnly);

			for (int y = 0; y < height; ++y)
			{
				const float val = 1.0f - y / (float) height;

				for (int x = 0; x < width; ++x)
				{
					const float sat = x / (float) width;

					pixels.setPixelColour (x, y, Colour (h, sat, val, 1.0f));
				}
			}
		}

		g.setOpacity (1.0f);
		g.drawImage (colours, edge, edge, getWidth() - edge * 2, getHeight() - edge * 2,
					 0, 0, colours.getWidth(), colours.getHeight());
	}

	void mouseDown (const MouseEvent& e)
	{
		mouseDrag (e);
	}

	void mouseDrag (const MouseEvent& e)
	{
		const float sat = (e.x - edge) / (float) (getWidth() - edge * 2);
		const float val = 1.0f - (e.y - edge) / (float) (getHeight() - edge * 2);

		owner.setSV (sat, val);
	}

	void updateIfNeeded()
	{
		if (lastHue != h)
		{
			lastHue = h;
			colours = Image::null;
			repaint();
		}

		updateMarker();
	}

	void resized()
	{
		colours = Image::null;
		updateMarker();
	}

private:
	ColourSelector& owner;
	float& h;
	float& s;
	float& v;
	float lastHue;
	ColourSpaceMarker marker;
	const int edge;
	Image colours;

	void updateMarker()
	{
		marker.setBounds (roundToInt ((getWidth() - edge * 2) * s),
						  roundToInt ((getHeight() - edge * 2) * (1.0f - v)),
						  edge * 2, edge * 2);
	}

	JUCE_DECLARE_NON_COPYABLE (ColourSpaceView);
};

class HueSelectorMarker  : public Component
{
public:
	HueSelectorMarker()
	{
		setInterceptsMouseClicks (false, false);
	}

	void paint (Graphics& g)
	{
		Path p;
		p.addTriangle (1.0f, 1.0f,
					   getWidth() * 0.3f, getHeight() * 0.5f,
					   1.0f, getHeight() - 1.0f);

		p.addTriangle (getWidth() - 1.0f, 1.0f,
					   getWidth() * 0.7f, getHeight() * 0.5f,
					   getWidth() - 1.0f, getHeight() - 1.0f);

		g.setColour (Colours::white.withAlpha (0.75f));
		g.fillPath (p);

		g.setColour (Colours::black.withAlpha (0.75f));
		g.strokePath (p, PathStrokeType (1.2f));
	}

private:
	JUCE_DECLARE_NON_COPYABLE (HueSelectorMarker);
};

class ColourSelector::HueSelectorComp  : public Component
{
public:
	HueSelectorComp (ColourSelector& owner_,
					 float& h_, float& s_, float& v_,
					 const int edgeSize)
		: owner (owner_),
		  h (h_), s (s_), v (v_),
		  lastHue (0.0f),
		  edge (edgeSize)
	{
		addAndMakeVisible (&marker);
	}

	void paint (Graphics& g)
	{
		const float yScale = 1.0f / (getHeight() - edge * 2);

		const Rectangle<int> clip (g.getClipBounds());

		for (int y = jmin (clip.getBottom(), getHeight() - edge); --y >= jmax (edge, clip.getY());)
		{
			g.setColour (Colour ((y - edge) * yScale, 1.0f, 1.0f, 1.0f));
			g.fillRect (edge, y, getWidth() - edge * 2, 1);
		}
	}

	void resized()
	{
		marker.setBounds (0, roundToInt ((getHeight() - edge * 2) * h),
						  getWidth(), edge * 2);
	}

	void mouseDown (const MouseEvent& e)
	{
		mouseDrag (e);
	}

	void mouseDrag (const MouseEvent& e)
	{
		owner.setHue ((e.y - edge) / (float) (getHeight() - edge * 2));
	}

	void updateIfNeeded()
	{
		resized();
	}

private:
	ColourSelector& owner;
	float& h;
	float& s;
	float& v;
	float lastHue;
	HueSelectorMarker marker;
	const int edge;

	JUCE_DECLARE_NON_COPYABLE (HueSelectorComp);
};

class ColourSelector::SwatchComponent   : public Component
{
public:
	SwatchComponent (ColourSelector& owner_, int index_)
		: owner (owner_), index (index_)
	{
	}

	void paint (Graphics& g)
	{
		const Colour colour (owner.getSwatchColour (index));

		g.fillCheckerBoard (getLocalBounds(), 6, 6,
							Colour (0xffdddddd).overlaidWith (colour),
							Colour (0xffffffff).overlaidWith (colour));
	}

	void mouseDown (const MouseEvent&)
	{
		PopupMenu m;
		m.addItem (1, TRANS("Use this swatch as the current colour"));
		m.addSeparator();
		m.addItem (2, TRANS("Set this swatch to the current colour"));

		m.showMenuAsync (PopupMenu::Options().withTargetComponent (this),
						 ModalCallbackFunction::forComponent (menuStaticCallback, this));
	}

private:
	ColourSelector& owner;
	const int index;

	static void menuStaticCallback (int result, SwatchComponent* comp)
	{
		if (comp != nullptr)
		{
			if (result == 1)
				comp->setColourFromSwatch();
			else if (result == 2)
				comp->setSwatchFromColour();
		}
	}

	void setColourFromSwatch()
	{
		owner.setCurrentColour (owner.getSwatchColour (index));
	}

	void setSwatchFromColour()
	{
		if (owner.getSwatchColour (index) != owner.getCurrentColour())
		{
			owner.setSwatchColour (index, owner.getCurrentColour());
			repaint();
		}
	}

	JUCE_DECLARE_NON_COPYABLE (SwatchComponent);
};

ColourSelector::ColourSelector (const int flags_,
								const int edgeGap_,
								const int gapAroundColourSpaceComponent)
	: colour (Colours::white),
	  flags (flags_),
	  edgeGap (edgeGap_)
{
	// not much point having a selector with no components in it!
	jassert ((flags_ & (showColourAtTop | showSliders | showColourspace)) != 0);

	updateHSV();

	if ((flags & showSliders) != 0)
	{
		addAndMakeVisible (sliders[0] = new ColourComponentSlider (TRANS ("red")));
		addAndMakeVisible (sliders[1] = new ColourComponentSlider (TRANS ("green")));
		addAndMakeVisible (sliders[2] = new ColourComponentSlider (TRANS ("blue")));
		addChildComponent (sliders[3] = new ColourComponentSlider (TRANS ("alpha")));

		sliders[3]->setVisible ((flags & showAlphaChannel) != 0);

		for (int i = 4; --i >= 0;)
			sliders[i]->addListener (this);
	}

	if ((flags & showColourspace) != 0)
	{
		addAndMakeVisible (colourSpace = new ColourSpaceView (*this, h, s, v, gapAroundColourSpaceComponent));
		addAndMakeVisible (hueSelector = new HueSelectorComp (*this, h, s, v, gapAroundColourSpaceComponent));
	}

	update();
}

ColourSelector::~ColourSelector()
{
	dispatchPendingMessages();
	swatchComponents.clear();
}

const Colour ColourSelector::getCurrentColour() const
{
	return ((flags & showAlphaChannel) != 0) ? colour
											 : colour.withAlpha ((uint8) 0xff);
}

void ColourSelector::setCurrentColour (const Colour& c)
{
	if (c != colour)
	{
		colour = ((flags & showAlphaChannel) != 0) ? c : c.withAlpha ((uint8) 0xff);

		updateHSV();
		update();
	}
}

void ColourSelector::setHue (float newH)
{
	newH = jlimit (0.0f, 1.0f, newH);

	if (h != newH)
	{
		h = newH;
		colour = Colour (h, s, v, colour.getFloatAlpha());
		update();
	}
}

void ColourSelector::setSV (float newS, float newV)
{
	newS = jlimit (0.0f, 1.0f, newS);
	newV = jlimit (0.0f, 1.0f, newV);

	if (s != newS || v != newV)
	{
		s = newS;
		v = newV;
		colour = Colour (h, s, v, colour.getFloatAlpha());
		update();
	}
}

void ColourSelector::updateHSV()
{
	colour.getHSB (h, s, v);
}

void ColourSelector::update()
{
	if (sliders[0] != nullptr)
	{
		sliders[0]->setValue ((int) colour.getRed());
		sliders[1]->setValue ((int) colour.getGreen());
		sliders[2]->setValue ((int) colour.getBlue());
		sliders[3]->setValue ((int) colour.getAlpha());
	}

	if (colourSpace != nullptr)
	{
		colourSpace->updateIfNeeded();
		hueSelector->updateIfNeeded();
	}

	if ((flags & showColourAtTop) != 0)
		repaint (previewArea);

	sendChangeMessage();
}

void ColourSelector::paint (Graphics& g)
{
	g.fillAll (findColour (backgroundColourId));

	if ((flags & showColourAtTop) != 0)
	{
		const Colour currentColour (getCurrentColour());

		g.fillCheckerBoard (previewArea, 10, 10,
							Colour (0xffdddddd).overlaidWith (currentColour),
							Colour (0xffffffff).overlaidWith (currentColour));

		g.setColour (Colours::white.overlaidWith (currentColour).contrasting());
		g.setFont (14.0f, true);
		g.drawText (currentColour.toDisplayString ((flags & showAlphaChannel) != 0),
					previewArea.getX(), previewArea.getY(), previewArea.getWidth(), previewArea.getHeight(),
					Justification::centred, false);
	}

	if ((flags & showSliders) != 0)
	{
		g.setColour (findColour (labelTextColourId));
		g.setFont (11.0f);

		for (int i = 4; --i >= 0;)
		{
			if (sliders[i]->isVisible())
				g.drawText (sliders[i]->getName() + ":",
							0, sliders[i]->getY(),
							sliders[i]->getX() - 8, sliders[i]->getHeight(),
							Justification::centredRight, false);
		}
	}
}

void ColourSelector::resized()
{
	const int swatchesPerRow = 8;
	const int swatchHeight = 22;

	const int numSliders = ((flags & showAlphaChannel) != 0) ? 4 : 3;
	const int numSwatches = getNumSwatches();

	const int swatchSpace = numSwatches > 0 ? edgeGap + swatchHeight * ((numSwatches + 7) / swatchesPerRow) : 0;
	const int sliderSpace = ((flags & showSliders) != 0)  ? jmin (22 * numSliders + edgeGap, proportionOfHeight (0.3f)) : 0;
	const int topSpace = ((flags & showColourAtTop) != 0) ? jmin (30 + edgeGap * 2, proportionOfHeight (0.2f)) : edgeGap;

	previewArea.setBounds (edgeGap, edgeGap, getWidth() - edgeGap * 2, topSpace - edgeGap * 2);

	int y = topSpace;

	if ((flags & showColourspace) != 0)
	{
		const int hueWidth = jmin (50, proportionOfWidth (0.15f));

		colourSpace->setBounds (edgeGap, y,
								getWidth() - hueWidth - edgeGap - 4,
								getHeight() - topSpace - sliderSpace - swatchSpace - edgeGap);

		hueSelector->setBounds (colourSpace->getRight() + 4, y,
								getWidth() - edgeGap - (colourSpace->getRight() + 4),
								colourSpace->getHeight());

		y = getHeight() - sliderSpace - swatchSpace - edgeGap;
	}

	if ((flags & showSliders) != 0)
	{
		const int sliderHeight = jmax (4, sliderSpace / numSliders);

		for (int i = 0; i < numSliders; ++i)
		{
			sliders[i]->setBounds (proportionOfWidth (0.2f), y,
								   proportionOfWidth (0.72f), sliderHeight - 2);

			y += sliderHeight;
		}
	}

	if (numSwatches > 0)
	{
		const int startX = 8;
		const int xGap = 4;
		const int yGap = 4;
		const int swatchWidth = (getWidth() - startX * 2) / swatchesPerRow;
		y += edgeGap;

		if (swatchComponents.size() != numSwatches)
		{
			swatchComponents.clear();

			for (int i = 0; i < numSwatches; ++i)
			{
				SwatchComponent* const sc = new SwatchComponent (*this, i);
				swatchComponents.add (sc);
				addAndMakeVisible (sc);
			}
		}

		int x = startX;

		for (int i = 0; i < swatchComponents.size(); ++i)
		{
			SwatchComponent* const sc = swatchComponents.getUnchecked(i);

			sc->setBounds (x + xGap / 2,
						   y + yGap / 2,
						   swatchWidth - xGap,
						   swatchHeight - yGap);

			if (((i + 1) % swatchesPerRow) == 0)
			{
				x = startX;
				y += swatchHeight;
			}
			else
			{
				x += swatchWidth;
			}
		}
	}
}

void ColourSelector::sliderValueChanged (Slider*)
{
	if (sliders[0] != nullptr)
		setCurrentColour (Colour ((uint8) sliders[0]->getValue(),
								  (uint8) sliders[1]->getValue(),
								  (uint8) sliders[2]->getValue(),
								  (uint8) sliders[3]->getValue()));
}

int ColourSelector::getNumSwatches() const
{
	return 0;
}

const Colour ColourSelector::getSwatchColour (const int) const
{
	jassertfalse; // if you've overridden getNumSwatches(), you also need to implement this method
	return Colours::black;
}

void ColourSelector::setSwatchColour (const int, const Colour&) const
{
	jassertfalse; // if you've overridden getNumSwatches(), you also need to implement this method
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ColourSelector.cpp ***/


/*** Start of inlined file: juce_DropShadower.cpp ***/
BEGIN_JUCE_NAMESPACE

class ShadowWindow  : public Component
{
public:
	ShadowWindow (Component& owner, const int type_, const Image shadowImageSections [12])
		: topLeft (shadowImageSections [type_ * 3]),
		  bottomRight (shadowImageSections [type_ * 3 + 1]),
		  filler (shadowImageSections [type_ * 3 + 2]),
		  type (type_)
	{
		setInterceptsMouseClicks (false, false);

		if (owner.isOnDesktop())
		{
			setSize (1, 1); // to keep the OS happy by not having zero-size windows
			addToDesktop (ComponentPeer::windowIgnoresMouseClicks
							| ComponentPeer::windowIsTemporary
							| ComponentPeer::windowIgnoresKeyPresses);
		}
		else if (owner.getParentComponent() != nullptr)
		{
			owner.getParentComponent()->addChildComponent (this);
		}
	}

	void paint (Graphics& g)
	{
		g.setOpacity (1.0f);

		if (type < 2)
		{
			int imH = jmin (topLeft.getHeight(), getHeight() / 2);
			g.drawImage (topLeft,
						 0, 0, topLeft.getWidth(), imH,
						 0, 0, topLeft.getWidth(), imH);

			imH = jmin (bottomRight.getHeight(), getHeight() - getHeight() / 2);
			g.drawImage (bottomRight,
						 0, getHeight() - imH, bottomRight.getWidth(), imH,
						 0, bottomRight.getHeight() - imH, bottomRight.getWidth(), imH);

			g.setTiledImageFill (filler, 0, 0, 1.0f);
			g.fillRect (0, topLeft.getHeight(), getWidth(), getHeight() - (topLeft.getHeight() + bottomRight.getHeight()));
		}
		else
		{
			int imW = jmin (topLeft.getWidth(), getWidth() / 2);
			g.drawImage (topLeft,
						 0, 0, imW, topLeft.getHeight(),
						 0, 0, imW, topLeft.getHeight());

			imW = jmin (bottomRight.getWidth(), getWidth() - getWidth() / 2);
			g.drawImage (bottomRight,
						 getWidth() - imW, 0, imW, bottomRight.getHeight(),
						 bottomRight.getWidth() - imW, 0, imW, bottomRight.getHeight());

			g.setTiledImageFill (filler, 0, 0, 1.0f);
			g.fillRect (topLeft.getWidth(), 0, getWidth() - (topLeft.getWidth() + bottomRight.getWidth()), getHeight());
		}
	}

	void resized()
	{
		repaint();  // (needed for correct repainting)
	}

private:
	const Image topLeft, bottomRight, filler;
	const int type;   // 0 = left, 1 = right, 2 = top, 3 = bottom. left + right are full-height

	JUCE_DECLARE_NON_COPYABLE (ShadowWindow);
};

DropShadower::DropShadower (const float alpha_,
							const int xOffset_,
							const int yOffset_,
							const float blurRadius_)
   : owner (nullptr),
	 xOffset (xOffset_),
	 yOffset (yOffset_),
	 alpha (alpha_),
	 blurRadius (blurRadius_),
	 reentrant (false)
{
}

DropShadower::~DropShadower()
{
	if (owner != nullptr)
		owner->removeComponentListener (this);

	reentrant = true;
	shadowWindows.clear();
}

void DropShadower::setOwner (Component* componentToFollow)
{
	if (componentToFollow != owner)
	{
		if (owner != nullptr)
			owner->removeComponentListener (this);

		// (the component can't be null)
		jassert (componentToFollow != nullptr);

		owner = componentToFollow;

		jassert (owner != nullptr);
		jassert (owner->isOpaque()); // doesn't work properly for semi-transparent comps!

		owner->addComponentListener (this);

		updateShadows();
	}
}

void DropShadower::componentMovedOrResized (Component&, bool /*wasMoved*/, bool /*wasResized*/)
{
	updateShadows();
}

void DropShadower::componentBroughtToFront (Component&)
{
	bringShadowWindowsToFront();
}

void DropShadower::componentParentHierarchyChanged (Component&)
{
	shadowWindows.clear();
	updateShadows();
}

void DropShadower::componentVisibilityChanged (Component&)
{
	updateShadows();
}

void DropShadower::updateShadows()
{
	if (reentrant || owner == nullptr)
		return;

	ComponentPeer* const peer = owner->getPeer();
	const bool isOwnerVisible = owner->isVisible() && (peer == nullptr || ! peer->isMinimised());

	const bool createShadowWindows  = shadowWindows.size() == 0
										 && owner->getWidth() > 0
										 && owner->getHeight() > 0
										 && isOwnerVisible
										 && (Desktop::canUseSemiTransparentWindows()
											  || owner->getParentComponent() != nullptr);

	{
		const ScopedValueSetter<bool> setter (reentrant, true, false);

		const int shadowEdge = jmax (xOffset, yOffset) + (int) blurRadius;

		if (createShadowWindows)
		{
			// keep a cached version of the image to save doing the gaussian too often
			String imageId;
			imageId << shadowEdge << ',' << xOffset << ',' << yOffset << ',' << alpha;

			const int hash = imageId.hashCode();

			Image bigIm (ImageCache::getFromHashCode (hash));

			if (bigIm.isNull())
			{
				bigIm = Image (Image::ARGB, shadowEdge * 5, shadowEdge * 5, true, Image::NativeImage);

				Graphics bigG (bigIm);
				bigG.setColour (Colours::black.withAlpha (alpha));
				bigG.fillRect (shadowEdge + xOffset,
							   shadowEdge + yOffset,
							   bigIm.getWidth() - (shadowEdge * 2),
							   bigIm.getHeight() - (shadowEdge * 2));

				ImageConvolutionKernel blurKernel (roundToInt (blurRadius * 2.0f));
				blurKernel.createGaussianBlur (blurRadius);

				blurKernel.applyToImage (bigIm, bigIm,
										 Rectangle<int> (xOffset, yOffset,
														 bigIm.getWidth(), bigIm.getHeight()));

				ImageCache::addImageToCache (bigIm, hash);
			}

			const int iw = bigIm.getWidth();
			const int ih = bigIm.getHeight();
			const int shadowEdge2 = shadowEdge * 2;

			setShadowImage (bigIm, 0, shadowEdge, shadowEdge2, 0, 0);
			setShadowImage (bigIm, 1, shadowEdge, shadowEdge2, 0, ih - shadowEdge2);
			setShadowImage (bigIm, 2, shadowEdge, shadowEdge, 0, shadowEdge2);
			setShadowImage (bigIm, 3, shadowEdge, shadowEdge2, iw - shadowEdge, 0);
			setShadowImage (bigIm, 4, shadowEdge, shadowEdge2, iw - shadowEdge, ih - shadowEdge2);
			setShadowImage (bigIm, 5, shadowEdge, shadowEdge, iw - shadowEdge, shadowEdge2);
			setShadowImage (bigIm, 6, shadowEdge, shadowEdge, shadowEdge, 0);
			setShadowImage (bigIm, 7, shadowEdge, shadowEdge, iw - shadowEdge2, 0);
			setShadowImage (bigIm, 8, shadowEdge, shadowEdge, shadowEdge2, 0);
			setShadowImage (bigIm, 9, shadowEdge, shadowEdge, shadowEdge, ih - shadowEdge);
			setShadowImage (bigIm, 10, shadowEdge, shadowEdge, iw - shadowEdge2, ih - shadowEdge);
			setShadowImage (bigIm, 11, shadowEdge, shadowEdge, shadowEdge2, ih - shadowEdge);

			for (int i = 0; i < 4; ++i)
				shadowWindows.add (new ShadowWindow (*owner, i, shadowImageSections));
		}

		if (shadowWindows.size() >= 4)
		{
			const int x = owner->getX();
			const int y = owner->getY() - shadowEdge;
			const int w = owner->getWidth();
			const int h = owner->getHeight() + shadowEdge + shadowEdge;

			for (int i = shadowWindows.size(); --i >= 0;)
			{
				// there seem to be rare situations where the dropshadower may be deleted by
				// callbacks during this loop, so use a weak ref to watch out for this..
				WeakReference<Component> sw (shadowWindows[i]);

				if (sw != nullptr)
					sw->setAlwaysOnTop (owner->isAlwaysOnTop());

				if (sw != nullptr)
					sw->setVisible (isOwnerVisible);

				if (sw != nullptr)
				{
					switch (i)
					{
						case 0: sw->setBounds (x - shadowEdge, y, shadowEdge, h); break;
						case 1: sw->setBounds (x + w, y, shadowEdge, h); break;
						case 2: sw->setBounds (x, y, w, shadowEdge); break;
						case 3: sw->setBounds (x, owner->getBottom(), w, shadowEdge); break;
						default: break;
					}
				}

				if (sw == nullptr)
					return;
			}
		}
	}

	if (createShadowWindows)
		bringShadowWindowsToFront();
}

void DropShadower::setShadowImage (const Image& src, const int num, const int w, const int h,
								   const int sx, const int sy)
{
	shadowImageSections[num] = Image (Image::ARGB, w, h, true, Image::NativeImage);

	Graphics g (shadowImageSections[num]);
	g.drawImage (src, 0, 0, w, h, sx, sy, w, h);
}

void DropShadower::bringShadowWindowsToFront()
{
	if (! reentrant)
	{
		updateShadows();

		const ScopedValueSetter<bool> setter (reentrant, true, false);

		for (int i = shadowWindows.size(); --i >= 0;)
			shadowWindows.getUnchecked(i)->toBehind (owner);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DropShadower.cpp ***/

/*** Start of inlined file: juce_PreferencesPanel.cpp ***/
BEGIN_JUCE_NAMESPACE

PreferencesPanel::PreferencesPanel()
	: buttonSize (70)
{
}

PreferencesPanel::~PreferencesPanel()
{
}

int PreferencesPanel::getButtonSize() const noexcept
{
	return buttonSize;
}

void PreferencesPanel::setButtonSize (int newSize)
{
	buttonSize = newSize;
	resized();
}

void PreferencesPanel::addSettingsPage (const String& title,
										const Drawable* icon,
										const Drawable* overIcon,
										const Drawable* downIcon)
{
	DrawableButton* const button = new DrawableButton (title, DrawableButton::ImageAboveTextLabel);
	buttons.add (button);

	button->setImages (icon, overIcon, downIcon);
	button->setRadioGroupId (1);
	button->addListener (this);
	button->setClickingTogglesState (true);
	button->setWantsKeyboardFocus (false);
	addAndMakeVisible (button);

	resized();

	if (currentPage == nullptr)
		setCurrentPage (title);
}

void PreferencesPanel::addSettingsPage (const String& title, const void* imageData, const int imageDataSize)
{
	DrawableImage icon, iconOver, iconDown;
	icon.setImage (ImageCache::getFromMemory (imageData, imageDataSize));

	iconOver.setImage (ImageCache::getFromMemory (imageData, imageDataSize));
	iconOver.setOverlayColour (Colours::black.withAlpha (0.12f));

	iconDown.setImage (ImageCache::getFromMemory (imageData, imageDataSize));
	iconDown.setOverlayColour (Colours::black.withAlpha (0.25f));

	addSettingsPage (title, &icon, &iconOver, &iconDown);
}

void PreferencesPanel::showInDialogBox (const String& dialogTitle, int dialogWidth, int dialogHeight, const Colour& backgroundColour)
{
	setSize (dialogWidth, dialogHeight);
	DialogWindow::showDialog (dialogTitle, this, 0, backgroundColour, false);
}

void PreferencesPanel::resized()
{
	for (int i = 0; i < buttons.size(); ++i)
		buttons.getUnchecked(i)->setBounds (i * buttonSize, 0, buttonSize, buttonSize);

	if (currentPage != nullptr)
		currentPage->setBounds (getLocalBounds().withTop (buttonSize + 5));
}

void PreferencesPanel::paint (Graphics& g)
{
	g.setColour (Colours::grey);
	g.fillRect (0, buttonSize + 2, getWidth(), 1);
}

void PreferencesPanel::setCurrentPage (const String& pageName)
{
	if (currentPageName != pageName)
	{
		currentPageName = pageName;

		currentPage = nullptr;
		currentPage = createComponentForPage (pageName);

		if (currentPage != nullptr)
		{
			addAndMakeVisible (currentPage);
			currentPage->toBack();
			resized();
		}

		for (int i = 0; i < buttons.size(); ++i)
		{
			if (buttons.getUnchecked(i)->getName() == pageName)
			{
				buttons.getUnchecked(i)->setToggleState (true, false);
				break;
			}
		}
	}
}

void PreferencesPanel::buttonClicked (Button*)
{
	for (int i = 0; i < buttons.size(); ++i)
	{
		if (buttons.getUnchecked(i)->getToggleState())
		{
			setCurrentPage (buttons.getUnchecked(i)->getName());
			break;
		}
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_PreferencesPanel.cpp ***/


/*** Start of inlined file: juce_SystemTrayIconComponent.cpp ***/
#if JUCE_WINDOWS || JUCE_LINUX

BEGIN_JUCE_NAMESPACE

SystemTrayIconComponent::SystemTrayIconComponent()
{
	addToDesktop (0);
}

SystemTrayIconComponent::~SystemTrayIconComponent()
{
}

END_JUCE_NAMESPACE

#endif

/*** End of inlined file: juce_SystemTrayIconComponent.cpp ***/


/*** Start of inlined file: juce_AlertWindow.cpp ***/
BEGIN_JUCE_NAMESPACE

class AlertWindowTextEditor  : public TextEditor
{
public:
	AlertWindowTextEditor (const String& name, const bool isPasswordBox)
		: TextEditor (name, isPasswordBox ? getDefaultPasswordChar() :  0)
	{
		setSelectAllWhenFocused (true);
	}

	void returnPressed()
	{
		// pass these up the component hierarchy to be trigger the buttons
		getParentComponent()->keyPressed (KeyPress (KeyPress::returnKey, 0, '\n'));
	}

	void escapePressed()
	{
		// pass these up the component hierarchy to be trigger the buttons
		getParentComponent()->keyPressed (KeyPress (KeyPress::escapeKey, 0, 0));
	}

private:
	JUCE_DECLARE_NON_COPYABLE (AlertWindowTextEditor);

	static juce_wchar getDefaultPasswordChar() noexcept
	{
	  #if JUCE_LINUX
		return 0x2022;
	  #else
		return 0x25cf;
	  #endif
	}
};

AlertWindow::AlertWindow (const String& title,
						  const String& message,
						  AlertIconType iconType,
						  Component* associatedComponent_)
   : TopLevelWindow (title, true),
	 alertIconType (iconType),
	 associatedComponent (associatedComponent_),
	 escapeKeyCancels (true)
{
	if (message.isEmpty())
		text = " "; // to force an update if the message is empty

	setMessage (message);

	for (int i = Desktop::getInstance().getNumComponents(); --i >= 0;)
	{
		Component* const c = Desktop::getInstance().getComponent (i);

		if (c != nullptr && c->isAlwaysOnTop() && c->isShowing())
		{
			setAlwaysOnTop (true);
			break;
		}
	}

	if (! JUCEApplication::isStandaloneApp())
		setAlwaysOnTop (true); // for a plugin, make it always-on-top because the host windows are often top-level

	AlertWindow::lookAndFeelChanged();

	constrainer.setMinimumOnscreenAmounts (0x10000, 0x10000, 0x10000, 0x10000);
}

AlertWindow::~AlertWindow()
{
	removeAllChildren();
}

void AlertWindow::userTriedToCloseWindow()
{
	if (escapeKeyCancels || buttons.size() > 0)
		exitModalState (0);
}

void AlertWindow::setMessage (const String& message)
{
	const String newMessage (message.substring (0, 2048));

	if (text != newMessage)
	{
		text = newMessage;

		font = getLookAndFeel().getAlertWindowMessageFont();

		Font titleFont (font.getHeight() * 1.1f, Font::bold);
		textLayout.setText (getName() + "\n\n", titleFont);

		textLayout.appendText (text, font);

		updateLayout (true);
		repaint();
	}
}

void AlertWindow::buttonClicked (Button* button)
{
	if (button->getParentComponent() != nullptr)
		button->getParentComponent()->exitModalState (button->getCommandID());
}

void AlertWindow::addButton (const String& name,
							 const int returnValue,
							 const KeyPress& shortcutKey1,
							 const KeyPress& shortcutKey2)
{
	TextButton* const b = new TextButton (name, String::empty);
	buttons.add (b);

	b->setWantsKeyboardFocus (true);
	b->setMouseClickGrabsKeyboardFocus (false);
	b->setCommandToTrigger (0, returnValue, false);
	b->addShortcut (shortcutKey1);
	b->addShortcut (shortcutKey2);
	b->addListener (this);
	b->changeWidthToFitText (getLookAndFeel().getAlertWindowButtonHeight());

	addAndMakeVisible (b, 0);

	updateLayout (false);
}

int AlertWindow::getNumButtons() const
{
	return buttons.size();
}

void AlertWindow::triggerButtonClick (const String& buttonName)
{
	for (int i = buttons.size(); --i >= 0;)
	{
		TextButton* const b = buttons.getUnchecked(i);

		if (buttonName == b->getName())
		{
			b->triggerClick();
			break;
		}
	}
}

void AlertWindow::setEscapeKeyCancels (bool shouldEscapeKeyCancel)
{
	escapeKeyCancels = shouldEscapeKeyCancel;
}

void AlertWindow::addTextEditor (const String& name,
								 const String& initialContents,
								 const String& onScreenLabel,
								 const bool isPasswordBox)
{
	AlertWindowTextEditor* const tc = new AlertWindowTextEditor (name, isPasswordBox);
	textBoxes.add (tc);
	allComps.add (tc);

	tc->setColour (TextEditor::outlineColourId, findColour (ComboBox::outlineColourId));
	tc->setFont (font);
	tc->setText (initialContents);
	tc->setCaretPosition (initialContents.length());
	addAndMakeVisible (tc);
	textboxNames.add (onScreenLabel);

	updateLayout (false);
}

TextEditor* AlertWindow::getTextEditor (const String& nameOfTextEditor) const
{
	for (int i = textBoxes.size(); --i >= 0;)
		if (textBoxes.getUnchecked(i)->getName() == nameOfTextEditor)
			return textBoxes.getUnchecked(i);

	return nullptr;
}

String AlertWindow::getTextEditorContents (const String& nameOfTextEditor) const
{
	TextEditor* const t = getTextEditor (nameOfTextEditor);
	return t != nullptr ? t->getText() : String::empty;
}

void AlertWindow::addComboBox (const String& name,
							   const StringArray& items,
							   const String& onScreenLabel)
{
	ComboBox* const cb = new ComboBox (name);
	comboBoxes.add (cb);
	allComps.add (cb);

	for (int i = 0; i < items.size(); ++i)
		cb->addItem (items[i], i + 1);

	addAndMakeVisible (cb);
	cb->setSelectedItemIndex (0);

	comboBoxNames.add (onScreenLabel);
	updateLayout (false);
}

ComboBox* AlertWindow::getComboBoxComponent (const String& nameOfList) const
{
	for (int i = comboBoxes.size(); --i >= 0;)
		if (comboBoxes.getUnchecked(i)->getName() == nameOfList)
			return comboBoxes.getUnchecked(i);

	return nullptr;
}

class AlertTextComp : public TextEditor
{
public:
	AlertTextComp (const String& message,
				   const Font& font)
	{
		setReadOnly (true);
		setMultiLine (true, true);
		setCaretVisible (false);
		setScrollbarsShown (true);
		lookAndFeelChanged();
		setWantsKeyboardFocus (false);

		setFont (font);
		setText (message, false);

		bestWidth = 2 * (int) std::sqrt (font.getHeight() * font.getStringWidth (message));

		setColour (TextEditor::backgroundColourId, Colours::transparentBlack);
		setColour (TextEditor::outlineColourId, Colours::transparentBlack);
		setColour (TextEditor::shadowColourId, Colours::transparentBlack);
	}

	int getPreferredWidth() const noexcept   { return bestWidth; }

	void updateLayout (const int width)
	{
		TextLayout text;
		text.appendText (getText(), getFont());
		text.layout (width - 8, Justification::topLeft, true);
		setSize (width, jmin (width, text.getHeight() + (int) getFont().getHeight()));
	}

private:
	int bestWidth;

	JUCE_DECLARE_NON_COPYABLE (AlertTextComp);
};

void AlertWindow::addTextBlock (const String& textBlock)
{
	AlertTextComp* const c = new AlertTextComp (textBlock, font);
	textBlocks.add (c);
	allComps.add (c);

	addAndMakeVisible (c);

	updateLayout (false);
}

void AlertWindow::addProgressBarComponent (double& progressValue)
{
	ProgressBar* const pb = new ProgressBar (progressValue);
	progressBars.add (pb);
	allComps.add (pb);

	addAndMakeVisible (pb);

	updateLayout (false);
}

void AlertWindow::addCustomComponent (Component* const component)
{
	customComps.add (component);
	allComps.add (component);

	addAndMakeVisible (component);

	updateLayout (false);
}

int AlertWindow::getNumCustomComponents() const
{
	return customComps.size();
}

Component* AlertWindow::getCustomComponent (const int index) const
{
	return customComps [index];
}

Component* AlertWindow::removeCustomComponent (const int index)
{
	Component* const c = getCustomComponent (index);

	if (c != nullptr)
	{
		customComps.removeValue (c);
		allComps.removeValue (c);
		removeChildComponent (c);

		updateLayout (false);
	}

	return c;
}

void AlertWindow::paint (Graphics& g)
{
	getLookAndFeel().drawAlertBox (g, *this, textArea, textLayout);

	g.setColour (findColour (textColourId));
	g.setFont (getLookAndFeel().getAlertWindowFont());

	int i;
	for (i = textBoxes.size(); --i >= 0;)
	{
		const TextEditor* const te = textBoxes.getUnchecked(i);

		g.drawFittedText (textboxNames[i],
						  te->getX(), te->getY() - 14,
						  te->getWidth(), 14,
						  Justification::centredLeft, 1);
	}

	for (i = comboBoxNames.size(); --i >= 0;)
	{
		const ComboBox* const cb = comboBoxes.getUnchecked(i);

		g.drawFittedText (comboBoxNames[i],
						  cb->getX(), cb->getY() - 14,
						  cb->getWidth(), 14,
						  Justification::centredLeft, 1);
	}

	for (i = customComps.size(); --i >= 0;)
	{
		const Component* const c = customComps.getUnchecked(i);

		g.drawFittedText (c->getName(),
						  c->getX(), c->getY() - 14,
						  c->getWidth(), 14,
						  Justification::centredLeft, 1);
	}
}

void AlertWindow::updateLayout (const bool onlyIncreaseSize)
{
	const int titleH = 24;
	const int iconWidth = 80;

	const int wid = jmax (font.getStringWidth (text),
						  font.getStringWidth (getName()));

	const int sw = (int) std::sqrt (font.getHeight() * wid);
	int w = jmin (300 + sw * 2, (int) (getParentWidth() * 0.7f));
	const int edgeGap = 10;
	const int labelHeight = 18;
	int iconSpace = 0;

	if (alertIconType == NoIcon)
	{
		textLayout.layout (w, Justification::horizontallyCentred, true);
	}
	else
	{
		textLayout.layout (w, Justification::left, true);
		iconSpace = iconWidth;
	}

	w = jmax (350, textLayout.getWidth() + iconSpace + edgeGap * 4);
	w = jmin (w, (int) (getParentWidth() * 0.7f));

	const int textLayoutH = textLayout.getHeight();
	const int textBottom = 16 + titleH + textLayoutH;
	int h = textBottom;

	int buttonW = 40;
	int i;
	for (i = 0; i < buttons.size(); ++i)
		buttonW += 16 + buttons.getUnchecked(i)->getWidth();

	w = jmax (buttonW, w);

	h += (textBoxes.size() + comboBoxes.size() + progressBars.size()) * 50;

	if (buttons.size() > 0)
		h += 20 + buttons.getUnchecked(0)->getHeight();

	for (i = customComps.size(); --i >= 0;)
	{
		Component* c = customComps.getUnchecked(i);
		w = jmax (w, (c->getWidth() * 100) / 80);
		h += 10 + c->getHeight();

		if (c->getName().isNotEmpty())
			h += labelHeight;
	}

	for (i = textBlocks.size(); --i >= 0;)
	{
		const AlertTextComp* const ac = static_cast <const AlertTextComp*> (textBlocks.getUnchecked(i));
		w = jmax (w, ac->getPreferredWidth());
	}

	w = jmin (w, (int) (getParentWidth() * 0.7f));

	for (i = textBlocks.size(); --i >= 0;)
	{
		AlertTextComp* const ac = static_cast <AlertTextComp*> (textBlocks.getUnchecked(i));
		ac->updateLayout ((int) (w * 0.8f));
		h += ac->getHeight() + 10;
	}

	h = jmin (getParentHeight() - 50, h);

	if (onlyIncreaseSize)
	{
		w = jmax (w, getWidth());
		h = jmax (h, getHeight());
	}

	if (! isVisible())
	{
		centreAroundComponent (associatedComponent, w, h);
	}
	else
	{
		const int cx = getX() + getWidth() / 2;
		const int cy = getY() + getHeight() / 2;

		setBounds (cx - w / 2,
				   cy - h / 2,
				   w, h);
	}

	textArea.setBounds (edgeGap, edgeGap, w - (edgeGap * 2), h - edgeGap);

	const int spacer = 16;
	int totalWidth = -spacer;

	for (i = buttons.size(); --i >= 0;)
		totalWidth += buttons.getUnchecked(i)->getWidth() + spacer;

	int x = (w - totalWidth) / 2;
	int y = (int) (getHeight() * 0.95f);

	for (i = 0; i < buttons.size(); ++i)
	{
		TextButton* const c = buttons.getUnchecked(i);
		int ny = proportionOfHeight (0.95f) - c->getHeight();
		c->setTopLeftPosition (x, ny);
		if (ny < y)
			y = ny;

		x += c->getWidth() + spacer;

		c->toFront (false);
	}

	y = textBottom;

	for (i = 0; i < allComps.size(); ++i)
	{
		Component* const c = allComps.getUnchecked(i);
		h = 22;

		const int comboIndex = comboBoxes.indexOf (dynamic_cast <ComboBox*> (c));
		if (comboIndex >= 0 && comboBoxNames [comboIndex].isNotEmpty())
			y += labelHeight;

		const int tbIndex = textBoxes.indexOf (dynamic_cast <TextEditor*> (c));
		if (tbIndex >= 0 && textboxNames[tbIndex].isNotEmpty())
			y += labelHeight;

		if (customComps.contains (c))
		{
			if (c->getName().isNotEmpty())
				y += labelHeight;

			c->setTopLeftPosition (proportionOfWidth (0.1f), y);
			h = c->getHeight();
		}
		else if (textBlocks.contains (c))
		{
			c->setTopLeftPosition ((getWidth() - c->getWidth()) / 2, y);
			h = c->getHeight();
		}
		else
		{
			c->setBounds (proportionOfWidth (0.1f), y, proportionOfWidth (0.8f), h);
		}

		y += h + 10;
	}

	setWantsKeyboardFocus (getNumChildComponents() == 0);
}

bool AlertWindow::containsAnyExtraComponents() const
{
	return allComps.size() > 0;
}

void AlertWindow::mouseDown (const MouseEvent& e)
{
	dragger.startDraggingComponent (this, e);
}

void AlertWindow::mouseDrag (const MouseEvent& e)
{
	dragger.dragComponent (this, e, &constrainer);
}

bool AlertWindow::keyPressed (const KeyPress& key)
{
	for (int i = buttons.size(); --i >= 0;)
	{
		TextButton* const b = buttons.getUnchecked(i);

		if (b->isRegisteredForShortcut (key))
		{
			b->triggerClick();
			return true;
		}
	}

	if (key.isKeyCode (KeyPress::escapeKey) && escapeKeyCancels && buttons.size() == 0)
	{
		exitModalState (0);
		return true;
	}
	else if (key.isKeyCode (KeyPress::returnKey) && buttons.size() == 1)
	{
		buttons.getUnchecked(0)->triggerClick();
		return true;
	}

	return false;
}

void AlertWindow::lookAndFeelChanged()
{
	const int newFlags = getLookAndFeel().getAlertBoxWindowFlags();

	setUsingNativeTitleBar ((newFlags & ComponentPeer::windowHasTitleBar) != 0);
	setDropShadowEnabled (isOpaque() && (newFlags & ComponentPeer::windowHasDropShadow) != 0);
}

int AlertWindow::getDesktopWindowStyleFlags() const
{
	return getLookAndFeel().getAlertBoxWindowFlags();
}

class AlertWindowInfo
{
public:
	AlertWindowInfo (const String& title_, const String& message_, Component* component,
					 AlertWindow::AlertIconType iconType_, int numButtons_,
					 ModalComponentManager::Callback* callback_, bool modal_)
		: title (title_), message (message_), iconType (iconType_),
		  numButtons (numButtons_), returnValue (0), associatedComponent (component),
		  callback (callback_), modal (modal_)
	{
	}

	String title, message, button1, button2, button3;

	int invoke() const
	{
		MessageManager::getInstance()->callFunctionOnMessageThread (showCallback, (void*) this);
		return returnValue;
	}

private:
	AlertWindow::AlertIconType iconType;
	int numButtons, returnValue;
	WeakReference<Component> associatedComponent;
	ModalComponentManager::Callback* callback;
	bool modal;

	void show()
	{
		LookAndFeel& lf = associatedComponent != nullptr ? associatedComponent->getLookAndFeel()
														 : LookAndFeel::getDefaultLookAndFeel();

		ScopedPointer <Component> alertBox (lf.createAlertWindow (title, message, button1, button2, button3,
																  iconType, numButtons, associatedComponent));

		jassert (alertBox != nullptr); // you have to return one of these!

	   #if JUCE_MODAL_LOOPS_PERMITTED
		if (modal)
		{
			returnValue = alertBox->runModalLoop();
		}
		else
	   #endif
		{
			alertBox->enterModalState (true, callback, true);
			alertBox.release();
		}
	}

	static void* showCallback (void* userData)
	{
		static_cast <AlertWindowInfo*> (userData)->show();
		return nullptr;
	}
};

#if JUCE_MODAL_LOOPS_PERMITTED
void AlertWindow::showMessageBox (AlertIconType iconType,
								  const String& title,
								  const String& message,
								  const String& buttonText,
								  Component* associatedComponent)
{
	if (LookAndFeel::getDefaultLookAndFeel().isUsingNativeAlertWindows())
	{
		NativeMessageBox::showMessageBox (iconType, title, message, associatedComponent);
	}
	else
	{
		AlertWindowInfo info (title, message, associatedComponent, iconType, 1, 0, true);
		info.button1 = buttonText.isEmpty() ? TRANS("ok") : buttonText;

		info.invoke();
	}
}
#endif

void AlertWindow::showMessageBoxAsync (AlertIconType iconType,
									   const String& title,
									   const String& message,
									   const String& buttonText,
									   Component* associatedComponent)
{
	if (LookAndFeel::getDefaultLookAndFeel().isUsingNativeAlertWindows())
	{
		return NativeMessageBox::showMessageBoxAsync (iconType, title, message, associatedComponent);
	}
	else
	{
		AlertWindowInfo info (title, message, associatedComponent, iconType, 1, 0, false);
		info.button1 = buttonText.isEmpty() ? TRANS("ok") : buttonText;

		info.invoke();
	}
}

bool AlertWindow::showOkCancelBox (AlertIconType iconType,
								   const String& title,
								   const String& message,
								   const String& button1Text,
								   const String& button2Text,
								   Component* associatedComponent,
								   ModalComponentManager::Callback* callback)
{
	if (LookAndFeel::getDefaultLookAndFeel().isUsingNativeAlertWindows())
	{
		return NativeMessageBox::showOkCancelBox (iconType, title, message, associatedComponent, callback);
	}
	else
	{
		AlertWindowInfo info (title, message, associatedComponent, iconType, 2, callback, callback == nullptr);
		info.button1 = button1Text.isEmpty() ? TRANS("ok")     : button1Text;
		info.button2 = button2Text.isEmpty() ? TRANS("cancel") : button2Text;

		return info.invoke() != 0;
	}
}

int AlertWindow::showYesNoCancelBox (AlertIconType iconType,
									 const String& title,
									 const String& message,
									 const String& button1Text,
									 const String& button2Text,
									 const String& button3Text,
									 Component* associatedComponent,
									 ModalComponentManager::Callback* callback)
{
	if (LookAndFeel::getDefaultLookAndFeel().isUsingNativeAlertWindows())
	{
		return NativeMessageBox::showYesNoCancelBox (iconType, title, message, associatedComponent, callback);
	}
	else
	{
		AlertWindowInfo info (title, message, associatedComponent, iconType, 3, callback, callback == nullptr);
		info.button1 = button1Text.isEmpty() ? TRANS("yes")     : button1Text;
		info.button2 = button2Text.isEmpty() ? TRANS("no")      : button2Text;
		info.button3 = button3Text.isEmpty() ? TRANS("cancel")  : button3Text;

		return info.invoke();
	}
}

#if JUCE_MODAL_LOOPS_PERMITTED
bool AlertWindow::showNativeDialogBox (const String& title,
									   const String& bodyText,
									   bool isOkCancel)
{
	if (isOkCancel)
	{
		return NativeMessageBox::showOkCancelBox (AlertWindow::NoIcon, title, bodyText);
	}
	else
	{
		NativeMessageBox::showMessageBox (AlertWindow::NoIcon, title, bodyText);
		return true;
	}
}
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_AlertWindow.cpp ***/


/*** Start of inlined file: juce_CallOutBox.cpp ***/
BEGIN_JUCE_NAMESPACE

CallOutBox::CallOutBox (Component& contentComponent,
						Component& componentToPointTo,
						Component* const parent)
	: borderSpace (20), arrowSize (16.0f), content (contentComponent)
{
	addAndMakeVisible (&content);

	if (parent != nullptr)
	{
		parent->addChildComponent (this);

		updatePosition (parent->getLocalArea (&componentToPointTo, componentToPointTo.getLocalBounds()),
						parent->getLocalBounds());

		setVisible (true);
	}
	else
	{
		if (! JUCEApplication::isStandaloneApp())
			setAlwaysOnTop (true); // for a plugin, make it always-on-top because the host windows are often top-level

		updatePosition (componentToPointTo.getScreenBounds(),
						componentToPointTo.getParentMonitorArea());

		addToDesktop (ComponentPeer::windowIsTemporary);
	}
}

CallOutBox::~CallOutBox()
{
}

void CallOutBox::setArrowSize (const float newSize)
{
	arrowSize = newSize;
	borderSpace = jmax (20, (int) arrowSize);
	refreshPath();
}

void CallOutBox::paint (Graphics& g)
{
	if (background.isNull())
	{
		background = Image (Image::ARGB, getWidth(), getHeight(), true);
		Graphics g2 (background);
		getLookAndFeel().drawCallOutBoxBackground (*this, g2, outline);
	}

	g.setColour (Colours::black);
	g.drawImageAt (background, 0, 0);
}

void CallOutBox::resized()
{
	content.setTopLeftPosition (borderSpace, borderSpace);
	refreshPath();
}

void CallOutBox::moved()
{
	refreshPath();
}

void CallOutBox::childBoundsChanged (Component*)
{
	updatePosition (targetArea, availableArea);
}

bool CallOutBox::hitTest (int x, int y)
{
	return outline.contains ((float) x, (float) y);
}

enum { callOutBoxDismissCommandId = 0x4f83a04b };

void CallOutBox::inputAttemptWhenModal()
{
	const Point<int> mousePos (getMouseXYRelative() + getBounds().getPosition());

	if (targetArea.contains (mousePos))
	{
		// if you click on the area that originally popped-up the callout, you expect it
		// to get rid of the box, but deleting the box here allows the click to pass through and
		// probably re-trigger it, so we need to dismiss the box asynchronously to consume the click..
		postCommandMessage (callOutBoxDismissCommandId);
	}
	else
	{
		exitModalState (0);
		setVisible (false);
	}
}

void CallOutBox::handleCommandMessage (int commandId)
{
	Component::handleCommandMessage (commandId);

	if (commandId == callOutBoxDismissCommandId)
	{
		exitModalState (0);
		setVisible (false);
	}
}

bool CallOutBox::keyPressed (const KeyPress& key)
{
	if (key.isKeyCode (KeyPress::escapeKey))
	{
		inputAttemptWhenModal();
		return true;
	}

	return false;
}

void CallOutBox::updatePosition (const Rectangle<int>& newAreaToPointTo, const Rectangle<int>& newAreaToFitIn)
{
	targetArea = newAreaToPointTo;
	availableArea = newAreaToFitIn;

	Rectangle<int> newBounds (0, 0,
							  content.getWidth() + borderSpace * 2,
							  content.getHeight() + borderSpace * 2);

	const int hw = newBounds.getWidth() / 2;
	const int hh = newBounds.getHeight() / 2;
	const float hwReduced = (float) (hw - borderSpace * 3);
	const float hhReduced = (float) (hh - borderSpace * 3);
	const float arrowIndent = borderSpace - arrowSize;

	Point<float> targets[4] = { Point<float> ((float) targetArea.getCentreX(), (float) targetArea.getBottom()),
								Point<float> ((float) targetArea.getRight(), (float) targetArea.getCentreY()),
								Point<float> ((float) targetArea.getX(), (float) targetArea.getCentreY()),
								Point<float> ((float) targetArea.getCentreX(), (float) targetArea.getY()) };

	Line<float> lines[4] = { Line<float> (targets[0].translated (-hwReduced, hh - arrowIndent), targets[0].translated (hwReduced, hh - arrowIndent)),
							 Line<float> (targets[1].translated (hw - arrowIndent, -hhReduced), targets[1].translated (hw - arrowIndent, hhReduced)),
							 Line<float> (targets[2].translated (-(hw - arrowIndent), -hhReduced), targets[2].translated (-(hw - arrowIndent), hhReduced)),
							 Line<float> (targets[3].translated (-hwReduced, -(hh - arrowIndent)), targets[3].translated (hwReduced, -(hh - arrowIndent))) };

	const Rectangle<float> centrePointArea (newAreaToFitIn.reduced (hw, hh).toFloat());

	float nearest = 1.0e9f;

	for (int i = 0; i < 4; ++i)
	{
		Line<float> constrainedLine (centrePointArea.getConstrainedPoint (lines[i].getStart()),
									 centrePointArea.getConstrainedPoint (lines[i].getEnd()));

		const Point<float> centre (constrainedLine.findNearestPointTo (centrePointArea.getCentre()));
		float distanceFromCentre = centre.getDistanceFrom (centrePointArea.getCentre());

		if (! (centrePointArea.contains (lines[i].getStart()) || centrePointArea.contains (lines[i].getEnd())))
			distanceFromCentre *= 2.0f;

		if (distanceFromCentre < nearest)
		{
			nearest = distanceFromCentre;

			targetPoint = targets[i];
			newBounds.setPosition ((int) (centre.getX() - hw),
								   (int) (centre.getY() - hh));
		}
	}

	setBounds (newBounds);
}

void CallOutBox::refreshPath()
{
	repaint();
	background = Image::null;
	outline.clear();

	const float gap = 4.5f;
	const float cornerSize = 9.0f;
	const float cornerSize2 = 2.0f * cornerSize;
	const float arrowBaseWidth = arrowSize * 0.7f;
	const float left = content.getX() - gap, top = content.getY() - gap, right = content.getRight() + gap, bottom = content.getBottom() + gap;
	const float targetX = targetPoint.getX() - getX(), targetY = targetPoint.getY() - getY();

	outline.startNewSubPath (left + cornerSize, top);

	if (targetY <= top)
	{
		outline.lineTo (targetX - arrowBaseWidth, top);
		outline.lineTo (targetX, targetY);
		outline.lineTo (targetX + arrowBaseWidth, top);
	}

	outline.lineTo (right - cornerSize, top);
	outline.addArc (right - cornerSize2, top, cornerSize2, cornerSize2, 0, float_Pi * 0.5f);

	if (targetX >= right)
	{
		outline.lineTo (right, targetY - arrowBaseWidth);
		outline.lineTo (targetX, targetY);
		outline.lineTo (right, targetY + arrowBaseWidth);
	}

	outline.lineTo (right, bottom - cornerSize);
	outline.addArc (right - cornerSize2, bottom - cornerSize2, cornerSize2, cornerSize2, float_Pi * 0.5f, float_Pi);

	if (targetY >= bottom)
	{
		outline.lineTo (targetX + arrowBaseWidth, bottom);
		outline.lineTo (targetX, targetY);
		outline.lineTo (targetX - arrowBaseWidth, bottom);
	}

	outline.lineTo (left + cornerSize, bottom);
	outline.addArc (left, bottom - cornerSize2, cornerSize2, cornerSize2, float_Pi, float_Pi * 1.5f);

	if (targetX <= left)
	{
		outline.lineTo (left, targetY + arrowBaseWidth);
		outline.lineTo (targetX, targetY);
		outline.lineTo (left, targetY - arrowBaseWidth);
	}

	outline.lineTo (left, top + cornerSize);
	outline.addArc (left, top, cornerSize2, cornerSize2, float_Pi * 1.5f, float_Pi * 2.0f - 0.05f);

	outline.closeSubPath();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_CallOutBox.cpp ***/


/*** Start of inlined file: juce_ComponentPeer.cpp ***/
BEGIN_JUCE_NAMESPACE

//#define JUCE_ENABLE_REPAINT_DEBUGGING 1

static Array <ComponentPeer*> heavyweightPeers;
static uint32 lastUniqueID = 1;

ComponentPeer::ComponentPeer (Component* const component_, const int styleFlags_)
	: component (component_),
	  styleFlags (styleFlags_),
	  lastPaintTime (0),
	  constrainer (nullptr),
	  lastDragAndDropCompUnderMouse (nullptr),
	  uniqueID (lastUniqueID += 2), // increment by 2 so that this can never hit 0
	  fakeMouseMessageSent (false),
	  isWindowMinimised (false)
{
	heavyweightPeers.add (this);
}

ComponentPeer::~ComponentPeer()
{
	heavyweightPeers.removeValue (this);
	Desktop::getInstance().triggerFocusCallback();
}

int ComponentPeer::getNumPeers() noexcept
{
	return heavyweightPeers.size();
}

ComponentPeer* ComponentPeer::getPeer (const int index) noexcept
{
	return heavyweightPeers [index];
}

ComponentPeer* ComponentPeer::getPeerFor (const Component* const component) noexcept
{
	for (int i = heavyweightPeers.size(); --i >= 0;)
	{
		ComponentPeer* const peer = heavyweightPeers.getUnchecked(i);

		if (peer->getComponent() == component)
			return peer;
	}

	return nullptr;
}

bool ComponentPeer::isValidPeer (const ComponentPeer* const peer) noexcept
{
	return heavyweightPeers.contains (const_cast <ComponentPeer*> (peer));
}

void ComponentPeer::updateCurrentModifiers() noexcept
{
	ModifierKeys::updateCurrentModifiers();
}

void ComponentPeer::handleMouseEvent (const int touchIndex, const Point<int>& positionWithinPeer, const ModifierKeys& newMods, const int64 time)
{
	MouseInputSource* const mouse = Desktop::getInstance().getMouseSource (touchIndex);
	jassert (mouse != nullptr); // not enough sources!

	mouse->handleEvent (this, positionWithinPeer, time, newMods);
}

void ComponentPeer::handleMouseWheel (const int touchIndex, const Point<int>& positionWithinPeer, const int64 time, const float x, const float y)
{
	MouseInputSource* const mouse = Desktop::getInstance().getMouseSource (touchIndex);
	jassert (mouse != nullptr); // not enough sources!

	mouse->handleWheel (this, positionWithinPeer, time, x, y);
}

void ComponentPeer::handlePaint (LowLevelGraphicsContext& contextToPaintTo)
{
	Graphics g (&contextToPaintTo);

   #if JUCE_ENABLE_REPAINT_DEBUGGING
	g.saveState();
   #endif

	JUCE_TRY
	{
		component->paintEntireComponent (g, true);
	}
	JUCE_CATCH_EXCEPTION

   #if JUCE_ENABLE_REPAINT_DEBUGGING
	// enabling this code will fill all areas that get repainted with a colour overlay, to show
	// clearly when things are being repainted.
	g.restoreState();

	g.fillAll (Colour ((uint8) Random::getSystemRandom().nextInt (255),
					   (uint8) Random::getSystemRandom().nextInt (255),
					   (uint8) Random::getSystemRandom().nextInt (255),
					   (uint8) 0x50));
   #endif

	/** If this fails, it's probably be because your CPU floating-point precision mode has
		been set to low.. This setting is sometimes changed by things like Direct3D, and can
		mess up a lot of the calculations that the library needs to do.
	*/
	jassert (roundToInt (10.1f) == 10);
}

bool ComponentPeer::handleKeyPress (const int keyCode,
									const juce_wchar textCharacter)
{
	updateCurrentModifiers();

	Component* target = Component::getCurrentlyFocusedComponent() != nullptr
							? Component::getCurrentlyFocusedComponent()
							: component;

	if (target->isCurrentlyBlockedByAnotherModalComponent())
	{
		Component* const currentModalComp = Component::getCurrentlyModalComponent();

		if (currentModalComp != nullptr)
			target = currentModalComp;
	}

	const KeyPress keyInfo (keyCode,
							ModifierKeys::getCurrentModifiers().getRawFlags()
							   & ModifierKeys::allKeyboardModifiers,
							textCharacter);

	bool keyWasUsed = false;

	while (target != nullptr)
	{
		const WeakReference<Component> deletionChecker (target);
		const Array <KeyListener*>* const keyListeners = target->keyListeners;

		if (keyListeners != nullptr)
		{
			for (int i = keyListeners->size(); --i >= 0;)
			{
				keyWasUsed = keyListeners->getUnchecked(i)->keyPressed (keyInfo, target);

				if (keyWasUsed || deletionChecker == nullptr)
					return keyWasUsed;

				i = jmin (i, keyListeners->size());
			}
		}

		keyWasUsed = target->keyPressed (keyInfo);

		if (keyWasUsed || deletionChecker == nullptr)
			break;

		if (keyInfo.isKeyCode (KeyPress::tabKey) && Component::getCurrentlyFocusedComponent() != nullptr)
		{
			Component* const currentlyFocused = Component::getCurrentlyFocusedComponent();
			currentlyFocused->moveKeyboardFocusToSibling (! keyInfo.getModifiers().isShiftDown());
			keyWasUsed = (currentlyFocused != Component::getCurrentlyFocusedComponent());
			break;
		}

		target = target->getParentComponent();
	}

	return keyWasUsed;
}

bool ComponentPeer::handleKeyUpOrDown (const bool isKeyDown)
{
	updateCurrentModifiers();

	Component* target = Component::getCurrentlyFocusedComponent() != nullptr
							? Component::getCurrentlyFocusedComponent()
							: component;

	if (target->isCurrentlyBlockedByAnotherModalComponent())
	{
		Component* const currentModalComp = Component::getCurrentlyModalComponent();

		if (currentModalComp != nullptr)
			target = currentModalComp;
	}

	bool keyWasUsed = false;

	while (target != nullptr)
	{
		const WeakReference<Component> deletionChecker (target);

		keyWasUsed = target->keyStateChanged (isKeyDown);

		if (keyWasUsed || deletionChecker == nullptr)
			break;

		const Array <KeyListener*>* const keyListeners = target->keyListeners;

		if (keyListeners != nullptr)
		{
			for (int i = keyListeners->size(); --i >= 0;)
			{
				keyWasUsed = keyListeners->getUnchecked(i)->keyStateChanged (isKeyDown, target);

				if (keyWasUsed || deletionChecker == nullptr)
					return keyWasUsed;

				i = jmin (i, keyListeners->size());
			}
		}

		target = target->getParentComponent();
	}

	return keyWasUsed;
}

void ComponentPeer::handleModifierKeysChange()
{
	updateCurrentModifiers();

	Component* target = Desktop::getInstance().getMainMouseSource().getComponentUnderMouse();

	if (target == nullptr)
		target = Component::getCurrentlyFocusedComponent();

	if (target == nullptr)
		target = component;

	if (target != nullptr)
		target->internalModifierKeysChanged();
}

TextInputTarget* ComponentPeer::findCurrentTextInputTarget()
{
	Component* const c = Component::getCurrentlyFocusedComponent();
	if (component->isParentOf (c))
	{
		TextInputTarget* const ti = dynamic_cast <TextInputTarget*> (c);
		if (ti != nullptr && ti->isTextInputActive())
			return ti;
	}

	return nullptr;
}

void ComponentPeer::dismissPendingTextInput()
{
}

void ComponentPeer::handleBroughtToFront()
{
	updateCurrentModifiers();

	if (component != nullptr)
		component->internalBroughtToFront();
}

void ComponentPeer::setConstrainer (ComponentBoundsConstrainer* const newConstrainer) noexcept
{
	constrainer = newConstrainer;
}

void ComponentPeer::handleMovedOrResized()
{
	updateCurrentModifiers();

	const bool nowMinimised = isMinimised();

	if (component->flags.hasHeavyweightPeerFlag && ! nowMinimised)
	{
		const WeakReference<Component> deletionChecker (component);

		const Rectangle<int> newBounds (getBounds());
		const bool wasMoved   = (component->getPosition() != newBounds.getPosition());
		const bool wasResized = (component->getWidth() != newBounds.getWidth() || component->getHeight() != newBounds.getHeight());

		if (wasMoved || wasResized)
		{
			component->bounds = newBounds;

			if (wasResized)
				component->repaint();

			component->sendMovedResizedMessages (wasMoved, wasResized);

			if (deletionChecker == nullptr)
				return;
		}
	}

	if (isWindowMinimised != nowMinimised)
	{
		isWindowMinimised = nowMinimised;
		component->minimisationStateChanged (nowMinimised);
		component->sendVisibilityChangeMessage();
	}

	if (! isFullScreen())
		lastNonFullscreenBounds = component->getBounds();
}

void ComponentPeer::handleFocusGain()
{
	updateCurrentModifiers();

	if (component->isParentOf (lastFocusedComponent))
	{
		Component::currentlyFocusedComponent = lastFocusedComponent;
		Desktop::getInstance().triggerFocusCallback();
		lastFocusedComponent->internalFocusGain (Component::focusChangedDirectly);
	}
	else
	{
		if (! component->isCurrentlyBlockedByAnotherModalComponent())
			component->grabKeyboardFocus();
		else
			ModalComponentManager::getInstance()->bringModalComponentsToFront();
	}
}

void ComponentPeer::handleFocusLoss()
{
	updateCurrentModifiers();

	if (component->hasKeyboardFocus (true))
	{
		lastFocusedComponent = Component::currentlyFocusedComponent;

		if (lastFocusedComponent != nullptr)
		{
			Component::currentlyFocusedComponent = nullptr;
			Desktop::getInstance().triggerFocusCallback();
			lastFocusedComponent->internalFocusLoss (Component::focusChangedByMouseClick);
		}
	}
}

Component* ComponentPeer::getLastFocusedSubcomponent() const noexcept
{
	return (component->isParentOf (lastFocusedComponent) && lastFocusedComponent->isShowing())
				? static_cast <Component*> (lastFocusedComponent)
				: component;
}

void ComponentPeer::handleScreenSizeChange()
{
	updateCurrentModifiers();

	component->parentSizeChanged();
	handleMovedOrResized();
}

void ComponentPeer::setNonFullScreenBounds (const Rectangle<int>& newBounds) noexcept
{
	lastNonFullscreenBounds = newBounds;
}

const Rectangle<int>& ComponentPeer::getNonFullScreenBounds() const noexcept
{
	return lastNonFullscreenBounds;
}

const Rectangle<int> ComponentPeer::localToGlobal (const Rectangle<int>& relativePosition)
{
	return relativePosition.withPosition (localToGlobal (relativePosition.getPosition()));
}

const Rectangle<int> ComponentPeer::globalToLocal (const Rectangle<int>& screenPosition)
{
	return screenPosition.withPosition (globalToLocal (screenPosition.getPosition()));
}

namespace ComponentPeerHelpers
{
	FileDragAndDropTarget* findDragAndDropTarget (Component* c,
												  const StringArray& files,
												  FileDragAndDropTarget* const lastOne)
	{
		while (c != nullptr)
		{
			FileDragAndDropTarget* const t = dynamic_cast <FileDragAndDropTarget*> (c);

			if (t != nullptr && (t == lastOne || t->isInterestedInFileDrag (files)))
				return t;

			c = c->getParentComponent();
		}

		return nullptr;
	}
}

bool ComponentPeer::handleFileDragMove (const StringArray& files, const Point<int>& position)
{
	updateCurrentModifiers();

	FileDragAndDropTarget* lastTarget
		= dynamic_cast<FileDragAndDropTarget*> (static_cast<Component*> (dragAndDropTargetComponent));

	FileDragAndDropTarget* newTarget = nullptr;

	Component* const compUnderMouse = component->getComponentAt (position);

	if (compUnderMouse != lastDragAndDropCompUnderMouse)
	{
		lastDragAndDropCompUnderMouse = compUnderMouse;
		newTarget = ComponentPeerHelpers::findDragAndDropTarget (compUnderMouse, files, lastTarget);

		if (newTarget != lastTarget)
		{
			if (lastTarget != nullptr)
				lastTarget->fileDragExit (files);

			dragAndDropTargetComponent = nullptr;

			if (newTarget != nullptr)
			{
				dragAndDropTargetComponent = dynamic_cast <Component*> (newTarget);
				const Point<int> pos (dragAndDropTargetComponent->getLocalPoint (component, position));
				newTarget->fileDragEnter (files, pos.getX(), pos.getY());
			}
		}
	}
	else
	{
		newTarget = lastTarget;
	}

	if (newTarget == nullptr)
		return false;

	Component* const targetComp = dynamic_cast <Component*> (newTarget);
	const Point<int> pos (targetComp->getLocalPoint (component, position));
	newTarget->fileDragMove (files, pos.getX(), pos.getY());
	return true;
}

bool ComponentPeer::handleFileDragExit (const StringArray& files)
{
	const bool used = handleFileDragMove (files, Point<int> (-1, -1));

	jassert (dragAndDropTargetComponent == nullptr);
	lastDragAndDropCompUnderMouse = nullptr;
	return used;
}

// We'll use an async message to deliver the drop, because if the target decides
// to run a modal loop, it can gum-up the operating system..
class AsyncFileDropMessage  : public CallbackMessage
{
public:
	AsyncFileDropMessage (Component* target_, FileDragAndDropTarget* dropTarget_,
						  const Point<int>& position_, const StringArray& files_)
		: target (target_), dropTarget (dropTarget_), position (position_), files (files_)
	{
	}

	void messageCallback()
	{
		if (target.get() != nullptr)
			dropTarget->filesDropped (files, position.getX(), position.getY());
	}

private:
	WeakReference<Component> target;
	FileDragAndDropTarget* const dropTarget;
	const Point<int> position;
	const StringArray files;

	JUCE_DECLARE_NON_COPYABLE (AsyncFileDropMessage);
};

bool ComponentPeer::handleFileDragDrop (const StringArray& files, const Point<int>& position)
{
	handleFileDragMove (files, position);

	if (dragAndDropTargetComponent != nullptr)
	{
		FileDragAndDropTarget* const target
			= dynamic_cast<FileDragAndDropTarget*> (static_cast<Component*> (dragAndDropTargetComponent));

		dragAndDropTargetComponent = nullptr;
		lastDragAndDropCompUnderMouse = nullptr;

		if (target != nullptr)
		{
			Component* const targetComp = dynamic_cast <Component*> (target);

			if (targetComp->isCurrentlyBlockedByAnotherModalComponent())
			{
				targetComp->internalModalInputAttempt();

				if (targetComp->isCurrentlyBlockedByAnotherModalComponent())
					return true;
			}

			(new AsyncFileDropMessage (targetComp, target, targetComp->getLocalPoint (component, position), files))->post();
			return true;
		}
	}

	return false;
}

void ComponentPeer::handleUserClosingWindow()
{
	updateCurrentModifiers();
	component->userTriedToCloseWindow();
}

void ComponentPeer::clearMaskedRegion()
{
	maskedRegion.clear();
}

void ComponentPeer::addMaskedRegion (int x, int y, int w, int h)
{
	maskedRegion.add (x, y, w, h);
}

StringArray ComponentPeer::getAvailableRenderingEngines()
{
	return StringArray ("Software Renderer");
}

int ComponentPeer::getCurrentRenderingEngine() const
{
	return 0;
}

void ComponentPeer::setCurrentRenderingEngine (int /*index*/)
{
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ComponentPeer.cpp ***/


/*** Start of inlined file: juce_DialogWindow.cpp ***/
BEGIN_JUCE_NAMESPACE

DialogWindow::DialogWindow (const String& name,
							const Colour& backgroundColour_,
							const bool escapeKeyTriggersCloseButton_,
							const bool addToDesktop_)
	: DocumentWindow (name, backgroundColour_, DocumentWindow::closeButton, addToDesktop_),
	  escapeKeyTriggersCloseButton (escapeKeyTriggersCloseButton_)
{
}

DialogWindow::~DialogWindow()
{
}

void DialogWindow::resized()
{
	DocumentWindow::resized();

	const KeyPress esc (KeyPress::escapeKey, 0, 0);

	if (escapeKeyTriggersCloseButton
		 && getCloseButton() != nullptr
		 && ! getCloseButton()->isRegisteredForShortcut (esc))
	{
		getCloseButton()->addShortcut (esc);
	}
}

class TempDialogWindow : public DialogWindow
{
public:
	TempDialogWindow (const String& title,
					  Component* contentComponent_,
					  Component* componentToCentreAround,
					  const Colour& colour,
					  const bool escapeKeyTriggersCloseButton_,
					  const bool shouldBeResizable,
					  const bool useBottomRightCornerResizer)
		: DialogWindow (title, colour, escapeKeyTriggersCloseButton_, true)
	{
		if (! JUCEApplication::isStandaloneApp())
			setAlwaysOnTop (true); // for a plugin, make it always-on-top because the host windows are often top-level

		setContentNonOwned (contentComponent_, true);
		centreAroundComponent (componentToCentreAround, getWidth(), getHeight());
		setResizable (shouldBeResizable, useBottomRightCornerResizer);
	}

	void closeButtonPressed()
	{
		setVisible (false);
	}

private:
	JUCE_DECLARE_NON_COPYABLE (TempDialogWindow);
};

void DialogWindow::showDialog (const String& dialogTitle,
							   Component* const contentComponent,
							   Component* const componentToCentreAround,
							   const Colour& backgroundColour,
							   const bool escapeKeyTriggersCloseButton,
							   const bool shouldBeResizable,
							   const bool useBottomRightCornerResizer)
{
	TempDialogWindow* dw = new TempDialogWindow (dialogTitle, contentComponent, componentToCentreAround,
												 backgroundColour, escapeKeyTriggersCloseButton,
												 shouldBeResizable, useBottomRightCornerResizer);

	dw->enterModalState (true, 0, true);
}

#if JUCE_MODAL_LOOPS_PERMITTED
int DialogWindow::showModalDialog (const String& dialogTitle,
								   Component* const contentComponent,
								   Component* const componentToCentreAround,
								   const Colour& backgroundColour,
								   const bool escapeKeyTriggersCloseButton,
								   const bool shouldBeResizable,
								   const bool useBottomRightCornerResizer)
{
	TempDialogWindow dw (dialogTitle, contentComponent, componentToCentreAround,
						 backgroundColour, escapeKeyTriggersCloseButton,
						 shouldBeResizable, useBottomRightCornerResizer);

	return dw.runModalLoop();
}
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DialogWindow.cpp ***/


/*** Start of inlined file: juce_DocumentWindow.cpp ***/
BEGIN_JUCE_NAMESPACE

class DocumentWindow::ButtonListenerProxy  : public ButtonListener // (can't use Button::Listener due to idiotic VC2005 bug)
{
public:
	ButtonListenerProxy (DocumentWindow& owner_)
		: owner (owner_)
	{
	}

	void buttonClicked (Button* button)
	{
		if	  (button == owner.getMinimiseButton())  owner.minimiseButtonPressed();
		else if (button == owner.getMaximiseButton())  owner.maximiseButtonPressed();
		else if (button == owner.getCloseButton())	 owner.closeButtonPressed();
	}

private:
	DocumentWindow& owner;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (ButtonListenerProxy);
};

DocumentWindow::DocumentWindow (const String& title,
								const Colour& backgroundColour,
								const int requiredButtons_,
								const bool addToDesktop_)
	: ResizableWindow (title, backgroundColour, addToDesktop_),
	  titleBarHeight (26),
	  menuBarHeight (24),
	  requiredButtons (requiredButtons_),
	#if JUCE_MAC
	  positionTitleBarButtonsOnLeft (true),
	#else
	  positionTitleBarButtonsOnLeft (false),
	#endif
	  drawTitleTextCentred (true),
	  menuBarModel (nullptr)
{
	setResizeLimits (128, 128, 32768, 32768);

	DocumentWindow::lookAndFeelChanged();
}

DocumentWindow::~DocumentWindow()
{
	// Don't delete or remove the resizer components yourself! They're managed by the
	// DocumentWindow, and you should leave them alone! You may have deleted them
	// accidentally by careless use of deleteAllChildren()..?
	jassert (menuBar == nullptr || getIndexOfChildComponent (menuBar) >= 0);
	jassert (titleBarButtons[0] == nullptr || getIndexOfChildComponent (titleBarButtons[0]) >= 0);
	jassert (titleBarButtons[1] == nullptr || getIndexOfChildComponent (titleBarButtons[1]) >= 0);
	jassert (titleBarButtons[2] == nullptr || getIndexOfChildComponent (titleBarButtons[2]) >= 0);

	for (int i = numElementsInArray (titleBarButtons); --i >= 0;)
		titleBarButtons[i] = nullptr;

	menuBar = nullptr;
}

void DocumentWindow::repaintTitleBar()
{
	repaint (getTitleBarArea());
}

void DocumentWindow::setName (const String& newName)
{
	if (newName != getName())
	{
		Component::setName (newName);
		repaintTitleBar();
	}
}

void DocumentWindow::setIcon (const Image& imageToUse)
{
	titleBarIcon = imageToUse;
	repaintTitleBar();
}

void DocumentWindow::setTitleBarHeight (const int newHeight)
{
	titleBarHeight = newHeight;
	resized();
	repaintTitleBar();
}

void DocumentWindow::setTitleBarButtonsRequired (const int requiredButtons_,
												 const bool positionTitleBarButtonsOnLeft_)
{
	requiredButtons = requiredButtons_;
	positionTitleBarButtonsOnLeft = positionTitleBarButtonsOnLeft_;
	lookAndFeelChanged();
}

void DocumentWindow::setTitleBarTextCentred (const bool textShouldBeCentred)
{
	drawTitleTextCentred = textShouldBeCentred;
	repaintTitleBar();
}

void DocumentWindow::setMenuBar (MenuBarModel* newMenuBarModel, const int newMenuBarHeight)
{
	if (menuBarModel != newMenuBarModel)
	{
		menuBar = nullptr;

		menuBarModel = newMenuBarModel;
		menuBarHeight = newMenuBarHeight > 0 ? newMenuBarHeight
											 : getLookAndFeel().getDefaultMenuBarHeight();

		if (menuBarModel != nullptr)
			setMenuBarComponent (new MenuBarComponent (menuBarModel));

		resized();
	}
}

Component* DocumentWindow::getMenuBarComponent() const noexcept
{
	return menuBar;
}

void DocumentWindow::setMenuBarComponent (Component* newMenuBarComponent)
{
	// (call the Component method directly to avoid the assertion in ResizableWindow)
	Component::addAndMakeVisible (menuBar = newMenuBarComponent);

	if (menuBar != nullptr)
		menuBar->setEnabled (isActiveWindow());

	resized();
}

void DocumentWindow::closeButtonPressed()
{
	/*  If you've got a close button, you have to override this method to get
		rid of your window!

		If the window is just a pop-up, you should override this method and make
		it delete the window in whatever way is appropriate for your app. E.g. you
		might just want to call "delete this".

		If your app is centred around this window such that the whole app should quit when
		the window is closed, then you will probably want to use this method as an opportunity
		to call JUCEApplication::quit(), and leave the window to be deleted later by your
		JUCEApplication::shutdown() method. (Doing it this way means that your window will
		still get cleaned-up if the app is quit by some other means (e.g. a cmd-Q on the mac
		or closing it via the taskbar icon on Windows).
	*/
	jassertfalse;
}

void DocumentWindow::minimiseButtonPressed()
{
	setMinimised (true);
}

void DocumentWindow::maximiseButtonPressed()
{
	setFullScreen (! isFullScreen());
}

void DocumentWindow::paint (Graphics& g)
{
	ResizableWindow::paint (g);

	if (resizableBorder == nullptr)
	{
		g.setColour (getBackgroundColour().overlaidWith (Colour (0x80000000)));

		const BorderSize<int> border (getBorderThickness());

		g.fillRect (0, 0, getWidth(), border.getTop());
		g.fillRect (0, border.getTop(), border.getLeft(), getHeight() - border.getTopAndBottom());
		g.fillRect (getWidth() - border.getRight(), border.getTop(), border.getRight(), getHeight() - border.getTopAndBottom());
		g.fillRect (0, getHeight() - border.getBottom(), getWidth(), border.getBottom());
	}

	const Rectangle<int> titleBarArea (getTitleBarArea());
	g.reduceClipRegion (titleBarArea);
	g.setOrigin (titleBarArea.getX(), titleBarArea.getY());

	int titleSpaceX1 = 6;
	int titleSpaceX2 = titleBarArea.getWidth() - 6;

	for (int i = 0; i < 3; ++i)
	{
		if (titleBarButtons[i] != nullptr)
		{
			if (positionTitleBarButtonsOnLeft)
				titleSpaceX1 = jmax (titleSpaceX1, titleBarButtons[i]->getRight() + (getWidth() - titleBarButtons[i]->getRight()) / 8);
			else
				titleSpaceX2 = jmin (titleSpaceX2, titleBarButtons[i]->getX() - (titleBarButtons[i]->getX() / 8));
		}
	}

	getLookAndFeel().drawDocumentWindowTitleBar (*this, g,
												 titleBarArea.getWidth(),
												 titleBarArea.getHeight(),
												 titleSpaceX1,
												 jmax (1, titleSpaceX2 - titleSpaceX1),
												 titleBarIcon.isValid() ? &titleBarIcon : 0,
												 ! drawTitleTextCentred);
}

void DocumentWindow::resized()
{
	ResizableWindow::resized();

	if (titleBarButtons[1] != nullptr)
		titleBarButtons[1]->setToggleState (isFullScreen(), false);

	const Rectangle<int> titleBarArea (getTitleBarArea());

	getLookAndFeel()
		.positionDocumentWindowButtons (*this,
										titleBarArea.getX(), titleBarArea.getY(),
										titleBarArea.getWidth(), titleBarArea.getHeight(),
										titleBarButtons[0],
										titleBarButtons[1],
										titleBarButtons[2],
										positionTitleBarButtonsOnLeft);

	if (menuBar != nullptr)
		menuBar->setBounds (titleBarArea.getX(), titleBarArea.getBottom(),
							titleBarArea.getWidth(), menuBarHeight);
}

const BorderSize<int> DocumentWindow::getBorderThickness()
{
	return BorderSize<int> ((isFullScreen() || isUsingNativeTitleBar())
								? 0 : (resizableBorder != nullptr ? 4 : 1));
}

const BorderSize<int> DocumentWindow::getContentComponentBorder()
{
	BorderSize<int> border (getBorderThickness());

	border.setTop (border.getTop()
					+ (isUsingNativeTitleBar() ? 0 : titleBarHeight)
					+ (menuBar != nullptr ? menuBarHeight : 0));

	return border;
}

int DocumentWindow::getTitleBarHeight() const
{
	return isUsingNativeTitleBar() ? 0 : jmin (titleBarHeight, getHeight() - 4);
}

const Rectangle<int> DocumentWindow::getTitleBarArea()
{
	const BorderSize<int> border (getBorderThickness());

	return Rectangle<int> (border.getLeft(), border.getTop(),
						   getWidth() - border.getLeftAndRight(),
						   getTitleBarHeight());
}

Button* DocumentWindow::getCloseButton() const noexcept	 { return titleBarButtons[2]; }
Button* DocumentWindow::getMinimiseButton() const noexcept  { return titleBarButtons[0]; }
Button* DocumentWindow::getMaximiseButton() const noexcept  { return titleBarButtons[1]; }

int DocumentWindow::getDesktopWindowStyleFlags() const
{
	int styleFlags = ResizableWindow::getDesktopWindowStyleFlags();

	if ((requiredButtons & minimiseButton) != 0)  styleFlags |= ComponentPeer::windowHasMinimiseButton;
	if ((requiredButtons & maximiseButton) != 0)  styleFlags |= ComponentPeer::windowHasMaximiseButton;
	if ((requiredButtons & closeButton) != 0)	 styleFlags |= ComponentPeer::windowHasCloseButton;

	return styleFlags;
}

void DocumentWindow::lookAndFeelChanged()
{
	int i;
	for (i = numElementsInArray (titleBarButtons); --i >= 0;)
		titleBarButtons[i] = nullptr;

	if (! isUsingNativeTitleBar())
	{
		LookAndFeel& lf = getLookAndFeel();

		if ((requiredButtons & minimiseButton) != 0)
			titleBarButtons[0] = lf.createDocumentWindowButton (minimiseButton);

		if ((requiredButtons & maximiseButton) != 0)
			titleBarButtons[1] = lf.createDocumentWindowButton (maximiseButton);

		if ((requiredButtons & closeButton) != 0)
			titleBarButtons[2] = lf.createDocumentWindowButton (closeButton);

		for (i = 0; i < 3; ++i)
		{
			if (titleBarButtons[i] != nullptr)
			{
				if (buttonListener == nullptr)
					buttonListener = new ButtonListenerProxy (*this);

				titleBarButtons[i]->addListener (buttonListener);
				titleBarButtons[i]->setWantsKeyboardFocus (false);

				// (call the Component method directly to avoid the assertion in ResizableWindow)
				Component::addAndMakeVisible (titleBarButtons[i]);
			}
		}

		if (getCloseButton() != nullptr)
		{
		   #if JUCE_MAC
			getCloseButton()->addShortcut (KeyPress ('w', ModifierKeys::commandModifier, 0));
		   #else
			getCloseButton()->addShortcut (KeyPress (KeyPress::F4Key, ModifierKeys::altModifier, 0));
		   #endif
		}
	}

	activeWindowStatusChanged();

	ResizableWindow::lookAndFeelChanged();
}

void DocumentWindow::parentHierarchyChanged()
{
	lookAndFeelChanged();
}

void DocumentWindow::activeWindowStatusChanged()
{
	ResizableWindow::activeWindowStatusChanged();

	for (int i = numElementsInArray (titleBarButtons); --i >= 0;)
		if (titleBarButtons[i] != nullptr)
			titleBarButtons[i]->setEnabled (isActiveWindow());

	if (menuBar != nullptr)
		menuBar->setEnabled (isActiveWindow());
}

void DocumentWindow::mouseDoubleClick (const MouseEvent& e)
{
	if (getTitleBarArea().contains (e.x, e.y)
		 && getMaximiseButton() != nullptr)
	{
		getMaximiseButton()->triggerClick();
	}
}

void DocumentWindow::userTriedToCloseWindow()
{
	closeButtonPressed();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DocumentWindow.cpp ***/


/*** Start of inlined file: juce_ResizableWindow.cpp ***/
BEGIN_JUCE_NAMESPACE

ResizableWindow::ResizableWindow (const String& name,
								  const bool addToDesktop_)
	: TopLevelWindow (name, addToDesktop_),
	  ownsContentComponent (false),
	  resizeToFitContent (false),
	  fullscreen (false),
	  constrainer (nullptr)
	 #if JUCE_DEBUG
	  , hasBeenResized (false)
	 #endif
{
	initialise (addToDesktop_);
}

ResizableWindow::ResizableWindow (const String& name,
								  const Colour& backgroundColour_,
								  const bool addToDesktop_)
	: TopLevelWindow (name, addToDesktop_),
	  ownsContentComponent (false),
	  resizeToFitContent (false),
	  fullscreen (false),
	  constrainer (nullptr)
	 #if JUCE_DEBUG
	  , hasBeenResized (false)
	 #endif
{
	setBackgroundColour (backgroundColour_);

	initialise (addToDesktop_);
}

ResizableWindow::~ResizableWindow()
{
	// Don't delete or remove the resizer components yourself! They're managed by the
	// ResizableWindow, and you should leave them alone! You may have deleted them
	// accidentally by careless use of deleteAllChildren()..?
	jassert (resizableCorner == nullptr || getIndexOfChildComponent (resizableCorner) >= 0);
	jassert (resizableBorder == nullptr || getIndexOfChildComponent (resizableBorder) >= 0);

	resizableCorner = nullptr;
	resizableBorder = nullptr;
	clearContentComponent();

	// have you been adding your own components directly to this window..? tut tut tut.
	// Read the instructions for using a ResizableWindow!
	jassert (getNumChildComponents() == 0);
}

void ResizableWindow::initialise (const bool shouldAddToDesktop)
{
	defaultConstrainer.setMinimumOnscreenAmounts (0x10000, 16, 24, 16);

	lastNonFullScreenPos.setBounds (50, 50, 256, 256);

	if (shouldAddToDesktop)
		addToDesktop();
}

int ResizableWindow::getDesktopWindowStyleFlags() const
{
	int styleFlags = TopLevelWindow::getDesktopWindowStyleFlags();

	if (isResizable() && (styleFlags & ComponentPeer::windowHasTitleBar) != 0)
		styleFlags |= ComponentPeer::windowIsResizable;

	return styleFlags;
}

void ResizableWindow::addToDesktop()
{
	Component::addToDesktop (ResizableWindow::getDesktopWindowStyleFlags());
	setDropShadowEnabled (isDropShadowEnabled()); // force an update to clear away any fake shadows if necessary.
}

void ResizableWindow::clearContentComponent()
{
	if (ownsContentComponent)
	{
		contentComponent.deleteAndZero();
	}
	else
	{
		removeChildComponent (contentComponent);
		contentComponent = nullptr;
	}
}

void ResizableWindow::setContent (Component* newContentComponent,
								  const bool takeOwnership,
								  const bool resizeToFitWhenContentChangesSize)
{
	if (newContentComponent != contentComponent)
	{
		clearContentComponent();

		contentComponent = newContentComponent;
		Component::addAndMakeVisible (contentComponent);
	}

	ownsContentComponent = takeOwnership;
	resizeToFitContent = resizeToFitWhenContentChangesSize;

	if (resizeToFitWhenContentChangesSize)
		childBoundsChanged (contentComponent);

	resized(); // must always be called to position the new content comp
}

void ResizableWindow::setContentOwned (Component* newContentComponent, const bool resizeToFitWhenContentChangesSize)
{
	setContent (newContentComponent, true, resizeToFitWhenContentChangesSize);
}

void ResizableWindow::setContentNonOwned (Component* newContentComponent, const bool resizeToFitWhenContentChangesSize)
{
	setContent (newContentComponent, false, resizeToFitWhenContentChangesSize);
}

void ResizableWindow::setContentComponent (Component* const newContentComponent,
										   const bool deleteOldOne,
										   const bool resizeToFitWhenContentChangesSize)
{
	if (newContentComponent != contentComponent)
	{
		if (deleteOldOne)
		{
			contentComponent.deleteAndZero();
		}
		else
		{
			removeChildComponent (contentComponent);
			contentComponent = nullptr;
		}
	}

	setContent (newContentComponent, true, resizeToFitWhenContentChangesSize);
}

void ResizableWindow::setContentComponentSize (int width, int height)
{
	jassert (width > 0 && height > 0); // not a great idea to give it a zero size..

	const BorderSize<int> border (getContentComponentBorder());

	setSize (width + border.getLeftAndRight(),
			 height + border.getTopAndBottom());
}

const BorderSize<int> ResizableWindow::getBorderThickness()
{
	return BorderSize<int> (isUsingNativeTitleBar() ? 0 : ((resizableBorder != nullptr && ! isFullScreen()) ? 5 : 3));
}

const BorderSize<int> ResizableWindow::getContentComponentBorder()
{
	return getBorderThickness();
}

void ResizableWindow::moved()
{
	updateLastPos();
}

void ResizableWindow::visibilityChanged()
{
	TopLevelWindow::visibilityChanged();

	updateLastPos();
}

void ResizableWindow::resized()
{
	if (resizableBorder != nullptr)
	{
	   #if JUCE_WINDOWS || JUCE_LINUX
		// hide the resizable border if the OS already provides one..
		resizableBorder->setVisible (! (isFullScreen() || isUsingNativeTitleBar()));
	   #else
		resizableBorder->setVisible (! isFullScreen());
	   #endif

		resizableBorder->setBorderThickness (getBorderThickness());
		resizableBorder->setSize (getWidth(), getHeight());
		resizableBorder->toBack();
	}

	if (resizableCorner != nullptr)
	{
	   #if JUCE_MAC
		// hide the resizable border if the OS already provides one..
		resizableCorner->setVisible (! (isFullScreen() || isUsingNativeTitleBar()));
	   #else
		resizableCorner->setVisible (! isFullScreen());
	   #endif

		const int resizerSize = 18;
		resizableCorner->setBounds (getWidth() - resizerSize,
									getHeight() - resizerSize,
									resizerSize, resizerSize);
	}

	if (contentComponent != nullptr)
		contentComponent->setBoundsInset (getContentComponentBorder());

	updateLastPos();

   #if JUCE_DEBUG
	hasBeenResized = true;
   #endif
}

void ResizableWindow::childBoundsChanged (Component* child)
{
	if ((child == contentComponent) && (child != nullptr) && resizeToFitContent)
	{
		// not going to look very good if this component has a zero size..
		jassert (child->getWidth() > 0);
		jassert (child->getHeight() > 0);

		const BorderSize<int> borders (getContentComponentBorder());

		setSize (child->getWidth() + borders.getLeftAndRight(),
				 child->getHeight() + borders.getTopAndBottom());
	}
}

void ResizableWindow::activeWindowStatusChanged()
{
	const BorderSize<int> border (getContentComponentBorder());

	Rectangle<int> area (getLocalBounds());
	repaint (area.removeFromTop (border.getTop()));
	repaint (area.removeFromLeft (border.getLeft()));
	repaint (area.removeFromRight (border.getRight()));
	repaint (area.removeFromBottom (border.getBottom()));
}

void ResizableWindow::setResizable (const bool shouldBeResizable,
									const bool useBottomRightCornerResizer)
{
	if (shouldBeResizable)
	{
		if (useBottomRightCornerResizer)
		{
			resizableBorder = nullptr;

			if (resizableCorner == nullptr)
			{
				Component::addChildComponent (resizableCorner = new ResizableCornerComponent (this, constrainer));
				resizableCorner->setAlwaysOnTop (true);
			}
		}
		else
		{
			resizableCorner = nullptr;

			if (resizableBorder == nullptr)
				Component::addChildComponent (resizableBorder = new ResizableBorderComponent (this, constrainer));
		}
	}
	else
	{
		resizableCorner = nullptr;
		resizableBorder = nullptr;
	}

	if (isUsingNativeTitleBar())
		recreateDesktopWindow();

	childBoundsChanged (contentComponent);
	resized();
}

bool ResizableWindow::isResizable() const noexcept
{
	return resizableCorner != nullptr
		|| resizableBorder != nullptr;
}

void ResizableWindow::setResizeLimits (const int newMinimumWidth,
									   const int newMinimumHeight,
									   const int newMaximumWidth,
									   const int newMaximumHeight) noexcept
{
	// if you've set up a custom constrainer then these settings won't have any effect..
	jassert (constrainer == &defaultConstrainer || constrainer == nullptr);

	if (constrainer == nullptr)
		setConstrainer (&defaultConstrainer);

	defaultConstrainer.setSizeLimits (newMinimumWidth, newMinimumHeight,
									  newMaximumWidth, newMaximumHeight);

	setBoundsConstrained (getBounds());
}

void ResizableWindow::setConstrainer (ComponentBoundsConstrainer* newConstrainer)
{
	if (constrainer != newConstrainer)
	{
		constrainer = newConstrainer;

		const bool useBottomRightCornerResizer = resizableCorner != nullptr;
		const bool shouldBeResizable = useBottomRightCornerResizer || resizableBorder != nullptr;

		resizableCorner = nullptr;
		resizableBorder = nullptr;

		setResizable (shouldBeResizable, useBottomRightCornerResizer);

		ComponentPeer* const peer = getPeer();
		if (peer != nullptr)
			peer->setConstrainer (newConstrainer);
	}
}

void ResizableWindow::setBoundsConstrained (const Rectangle<int>& newBounds)
{
	if (constrainer != nullptr)
		constrainer->setBoundsForComponent (this, newBounds, false, false, false, false);
	else
		setBounds (newBounds);
}

void ResizableWindow::paint (Graphics& g)
{
	getLookAndFeel().fillResizableWindowBackground (g, getWidth(), getHeight(),
													getBorderThickness(), *this);

	if (! isFullScreen())
	{
		getLookAndFeel().drawResizableWindowBorder (g, getWidth(), getHeight(),
													getBorderThickness(), *this);
	}

   #if JUCE_DEBUG
	/* If this fails, then you've probably written a subclass with a resized()
	   callback but forgotten to make it call its parent class's resized() method.

	   It's important when you override methods like resized(), moved(),
	   etc., that you make sure the base class methods also get called.

	   Of course you shouldn't really be overriding ResizableWindow::resized() anyway,
	   because your content should all be inside the content component - and it's the
	   content component's resized() method that you should be using to do your
	   layout.
	*/
	jassert (hasBeenResized || (getWidth() == 0 && getHeight() == 0));
   #endif
}

void ResizableWindow::lookAndFeelChanged()
{
	resized();

	if (isOnDesktop())
	{
		Component::addToDesktop (getDesktopWindowStyleFlags());

		ComponentPeer* const peer = getPeer();
		if (peer != nullptr)
			peer->setConstrainer (constrainer);
	}
}

const Colour ResizableWindow::getBackgroundColour() const noexcept
{
	return findColour (backgroundColourId, false);
}

void ResizableWindow::setBackgroundColour (const Colour& newColour)
{
	Colour backgroundColour (newColour);

	if (! Desktop::canUseSemiTransparentWindows())
		backgroundColour = newColour.withAlpha (1.0f);

	setColour (backgroundColourId, backgroundColour);

	setOpaque (backgroundColour.isOpaque());
	repaint();
}

bool ResizableWindow::isFullScreen() const
{
	if (isOnDesktop())
	{
		ComponentPeer* const peer = getPeer();
		return peer != nullptr && peer->isFullScreen();
	}

	return fullscreen;
}

void ResizableWindow::setFullScreen (const bool shouldBeFullScreen)
{
	if (shouldBeFullScreen != isFullScreen())
	{
		updateLastPos();
		fullscreen = shouldBeFullScreen;

		if (isOnDesktop())
		{
			ComponentPeer* const peer = getPeer();

			if (peer != nullptr)
			{
				// keep a copy of this intact in case the real one gets messed-up while we're un-maximising
				const Rectangle<int> lastPos (lastNonFullScreenPos);

				peer->setFullScreen (shouldBeFullScreen);

				if ((! shouldBeFullScreen) && ! lastPos.isEmpty())
					setBounds (lastPos);
			}
			else
			{
				jassertfalse;
			}
		}
		else
		{
			if (shouldBeFullScreen)
				setBounds (0, 0, getParentWidth(), getParentHeight());
			else
				setBounds (lastNonFullScreenPos);
		}

		resized();
	}
}

bool ResizableWindow::isMinimised() const
{
	ComponentPeer* const peer = getPeer();

	return (peer != nullptr) && peer->isMinimised();
}

void ResizableWindow::setMinimised (const bool shouldMinimise)
{
	if (shouldMinimise != isMinimised())
	{
		ComponentPeer* const peer = getPeer();

		if (peer != nullptr)
		{
			updateLastPos();
			peer->setMinimised (shouldMinimise);
		}
		else
		{
			jassertfalse;
		}
	}
}

void ResizableWindow::updateLastPos()
{
	if (isShowing() && ! (isFullScreen() || isMinimised()))
	{
		lastNonFullScreenPos = getBounds();
	}
}

void ResizableWindow::parentSizeChanged()
{
	if (isFullScreen() && getParentComponent() != nullptr)
	{
		setBounds (0, 0, getParentWidth(), getParentHeight());
	}
}

String ResizableWindow::getWindowStateAsString()
{
	updateLastPos();
	return (isFullScreen() ? "fs " : "") + lastNonFullScreenPos.toString();
}

bool ResizableWindow::restoreWindowStateFromString (const String& s)
{
	StringArray tokens;
	tokens.addTokens (s, false);
	tokens.removeEmptyStrings();
	tokens.trim();

	const bool fs = tokens[0].startsWithIgnoreCase ("fs");
	const int firstCoord = fs ? 1 : 0;

	if (tokens.size() != firstCoord + 4)
		return false;

	Rectangle<int> newPos (tokens[firstCoord].getIntValue(),
						   tokens[firstCoord + 1].getIntValue(),
						   tokens[firstCoord + 2].getIntValue(),
						   tokens[firstCoord + 3].getIntValue());

	if (newPos.isEmpty())
		return false;

	const Rectangle<int> screen (Desktop::getInstance().getMonitorAreaContaining (newPos.getCentre()));

	ComponentPeer* const peer = isOnDesktop() ? getPeer() : nullptr;
	if (peer != nullptr)
		peer->getFrameSize().addTo (newPos);

	if (! screen.contains (newPos))
	{
		newPos.setSize (jmin (newPos.getWidth(), screen.getWidth()),
						jmin (newPos.getHeight(), screen.getHeight()));

		newPos.setPosition (jlimit (screen.getX(), screen.getRight() - newPos.getWidth(), newPos.getX()),
							jlimit (screen.getY(), screen.getBottom() - newPos.getHeight(), newPos.getY()));
	}

	if (peer != nullptr)
	{
		peer->getFrameSize().subtractFrom (newPos);
		peer->setNonFullScreenBounds (newPos);
	}

	lastNonFullScreenPos = newPos;
	setFullScreen (fs);

	if (! fs)
		setBoundsConstrained (newPos);

	return true;
}

void ResizableWindow::mouseDown (const MouseEvent& e)
{
	if (! isFullScreen())
		dragger.startDraggingComponent (this, e);
}

void ResizableWindow::mouseDrag (const MouseEvent& e)
{
	if (! isFullScreen())
		dragger.dragComponent (this, e, constrainer);
}

#if JUCE_DEBUG
void ResizableWindow::addChildComponent (Component* const child, int zOrder)
{
	/* Agh! You shouldn't add components directly to a ResizableWindow - this class
	   manages its child components automatically, and if you add your own it'll cause
	   trouble. Instead, use setContentComponent() to give it a component which
	   will be automatically resized and kept in the right place - then you can add
	   subcomponents to the content comp. See the notes for the ResizableWindow class
	   for more info.

	   If you really know what you're doing and want to avoid this assertion, just call
	   Component::addChildComponent directly.
	*/
	jassertfalse;

	Component::addChildComponent (child, zOrder);
}

void ResizableWindow::addAndMakeVisible (Component* const child, int zOrder)
{
	/* Agh! You shouldn't add components directly to a ResizableWindow - this class
	   manages its child components automatically, and if you add your own it'll cause
	   trouble. Instead, use setContentComponent() to give it a component which
	   will be automatically resized and kept in the right place - then you can add
	   subcomponents to the content comp. See the notes for the ResizableWindow class
	   for more info.

	   If you really know what you're doing and want to avoid this assertion, just call
	   Component::addAndMakeVisible directly.
	*/
	jassertfalse;

	Component::addAndMakeVisible (child, zOrder);
}
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ResizableWindow.cpp ***/


/*** Start of inlined file: juce_ThreadWithProgressWindow.cpp ***/
BEGIN_JUCE_NAMESPACE

ThreadWithProgressWindow::ThreadWithProgressWindow (const String& title,
													const bool hasProgressBar,
													const bool hasCancelButton,
													const int timeOutMsWhenCancelling_,
													const String& cancelButtonText)
  : Thread ("Juce Progress Window"),
	progress (0.0),
	timeOutMsWhenCancelling (timeOutMsWhenCancelling_)
{
	alertWindow = LookAndFeel::getDefaultLookAndFeel()
					.createAlertWindow (title, String::empty, cancelButtonText,
										String::empty, String::empty,
										AlertWindow::NoIcon, hasCancelButton ? 1 : 0, 0);

	// if there are no buttons, we won't allow the user to interrupt the thread.
	alertWindow->setEscapeKeyCancels (false);

	if (hasProgressBar)
		alertWindow->addProgressBarComponent (progress);
}

ThreadWithProgressWindow::~ThreadWithProgressWindow()
{
	stopThread (timeOutMsWhenCancelling);
}

#if JUCE_MODAL_LOOPS_PERMITTED
bool ThreadWithProgressWindow::runThread (const int priority)
{
	startThread (priority);
	startTimer (100);

	{
		const ScopedLock sl (messageLock);
		alertWindow->setMessage (message);
	}

	const bool finishedNaturally = alertWindow->runModalLoop() != 0;

	stopThread (timeOutMsWhenCancelling);

	alertWindow->setVisible (false);

	return finishedNaturally;
}
#endif

void ThreadWithProgressWindow::setProgress (const double newProgress)
{
	progress = newProgress;
}

void ThreadWithProgressWindow::setStatusMessage (const String& newStatusMessage)
{
	const ScopedLock sl (messageLock);
	message = newStatusMessage;
}

void ThreadWithProgressWindow::timerCallback()
{
	if (! isThreadRunning())
	{
		// thread has finished normally..
		alertWindow->exitModalState (1);
		alertWindow->setVisible (false);
	}
	else
	{
		const ScopedLock sl (messageLock);
		alertWindow->setMessage (message);
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ThreadWithProgressWindow.cpp ***/


/*** Start of inlined file: juce_TooltipWindow.cpp ***/
BEGIN_JUCE_NAMESPACE

TooltipWindow::TooltipWindow (Component* const parent_,
							  const int millisecondsBeforeTipAppears_)
	: Component ("tooltip"),
	  millisecondsBeforeTipAppears (millisecondsBeforeTipAppears_),
	  mouseClicks (0),
	  lastHideTime (0),
	  lastComponentUnderMouse (nullptr),
	  changedCompsSinceShown (true)
{
	if (Desktop::getInstance().getMainMouseSource().canHover())
		startTimer (123);

	setAlwaysOnTop (true);
	setOpaque (true);

	if (parent_ != nullptr)
		parent_->addChildComponent (this);
}

TooltipWindow::~TooltipWindow()
{
	hide();
}

void TooltipWindow::setMillisecondsBeforeTipAppears (const int newTimeMs) noexcept
{
	millisecondsBeforeTipAppears = newTimeMs;
}

void TooltipWindow::paint (Graphics& g)
{
	getLookAndFeel().drawTooltip (g, tipShowing, getWidth(), getHeight());
}

void TooltipWindow::mouseEnter (const MouseEvent&)
{
	hide();
}

void TooltipWindow::showFor (const String& tip)
{
	jassert (tip.isNotEmpty());
	if (tipShowing != tip)
		repaint();

	tipShowing = tip;

	Point<int> mousePos (Desktop::getMousePosition());
	Rectangle<int> parentArea;

	if (getParentComponent() != nullptr)
	{
		mousePos = getParentComponent()->getLocalPoint (nullptr, mousePos);
		parentArea = getParentComponent()->getLocalBounds();
	}
	else
	{
		parentArea = Desktop::getInstance().getMonitorAreaContaining (mousePos);
	}

	int w, h;
	getLookAndFeel().getTooltipSize (tip, w, h);

	int x = mousePos.getX();
	if (x > parentArea.getCentreX())
		x -= (w + 12);
	else
		x += 24;

	int y = mousePos.getY();
	if (y > parentArea.getCentreY())
		y -= (h + 6);
	else
		y += 6;

	x = jlimit (parentArea.getX(), parentArea.getRight() - w, x);
	y = jlimit (parentArea.getY(), parentArea.getBottom() - h, y);

	setBounds (x, y, w, h);
	setVisible (true);

	if (getParentComponent() == nullptr)
	{
		addToDesktop (ComponentPeer::windowHasDropShadow
						| ComponentPeer::windowIsTemporary
						| ComponentPeer::windowIgnoresKeyPresses);
	}

	toFront (false);
}

String TooltipWindow::getTipFor (Component* const c)
{
	if (c != nullptr
		 && Process::isForegroundProcess()
		 && ! Component::isMouseButtonDownAnywhere())
	{
		TooltipClient* const ttc = dynamic_cast <TooltipClient*> (c);

		if (ttc != nullptr && ! c->isCurrentlyBlockedByAnotherModalComponent())
			return ttc->getTooltip();
	}

	return String::empty;
}

void TooltipWindow::hide()
{
	tipShowing = String::empty;
	removeFromDesktop();
	setVisible (false);
}

void TooltipWindow::timerCallback()
{
	const unsigned int now = Time::getApproximateMillisecondCounter();
	Component* const newComp = Desktop::getInstance().getMainMouseSource().getComponentUnderMouse();
	const String newTip (getTipFor (newComp));

	const bool tipChanged = (newTip != lastTipUnderMouse || newComp != lastComponentUnderMouse);
	lastComponentUnderMouse = newComp;
	lastTipUnderMouse = newTip;

	const int clickCount = Desktop::getInstance().getMouseButtonClickCounter();
	const bool mouseWasClicked = clickCount > mouseClicks;
	mouseClicks = clickCount;

	const Point<int> mousePos (Desktop::getMousePosition());
	const bool mouseMovedQuickly = mousePos.getDistanceFrom (lastMousePos) > 12;
	lastMousePos = mousePos;

	if (tipChanged || mouseWasClicked || mouseMovedQuickly)
		lastCompChangeTime = now;

	if (isVisible() || now < lastHideTime + 500)
	{
		// if a tip is currently visible (or has just disappeared), update to a new one
		// immediately if needed..
		if (newComp == nullptr || mouseWasClicked || newTip.isEmpty())
		{
			if (isVisible())
			{
				lastHideTime = now;
				hide();
			}
		}
		else if (tipChanged)
		{
			showFor (newTip);
		}
	}
	else
	{
		// if there isn't currently a tip, but one is needed, only let it
		// appear after a timeout..
		if (newTip.isNotEmpty()
			 && newTip != tipShowing
			 && now > lastCompChangeTime + millisecondsBeforeTipAppears)
		{
			showFor (newTip);
		}
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TooltipWindow.cpp ***/


/*** Start of inlined file: juce_TopLevelWindow.cpp ***/
BEGIN_JUCE_NAMESPACE

/** Keeps track of the active top level window.
*/
class TopLevelWindowManager  : public Timer,
							   public DeletedAtShutdown
{
public:

	TopLevelWindowManager()
		: currentActive (nullptr)
	{
	}

	~TopLevelWindowManager()
	{
		clearSingletonInstance();
	}

	juce_DeclareSingleton_SingleThreaded_Minimal (TopLevelWindowManager);

	void timerCallback()
	{
		startTimer (jmin (1731, getTimerInterval() * 2));

		TopLevelWindow* active = nullptr;

		if (Process::isForegroundProcess())
		{
			active = currentActive;

			Component* const c = Component::getCurrentlyFocusedComponent();
			TopLevelWindow* tlw = dynamic_cast <TopLevelWindow*> (c);

			if (tlw == nullptr && c != nullptr)
				// (unable to use the syntax findParentComponentOfClass <TopLevelWindow> () because of a VC6 compiler bug)
				tlw = c->findParentComponentOfClass ((TopLevelWindow*) nullptr);

			if (tlw != nullptr)
				active = tlw;
		}

		if (active != currentActive)
		{
			currentActive = active;

			for (int i = windows.size(); --i >= 0;)
			{
				TopLevelWindow* const tlw = windows.getUnchecked (i);
				tlw->setWindowActive (isWindowActive (tlw));

				i = jmin (i, windows.size() - 1);
			}

			Desktop::getInstance().triggerFocusCallback();
		}
	}

	bool addWindow (TopLevelWindow* const w)
	{
		windows.add (w);
		startTimer (10);

		return isWindowActive (w);
	}

	void removeWindow (TopLevelWindow* const w)
	{
		startTimer (10);

		if (currentActive == w)
			currentActive = nullptr;

		windows.removeValue (w);

		if (windows.size() == 0)
			deleteInstance();
	}

	Array <TopLevelWindow*> windows;

private:
	TopLevelWindow* currentActive;

	bool isWindowActive (TopLevelWindow* const tlw) const
	{
		return (tlw == currentActive
				 || tlw->isParentOf (currentActive)
				 || tlw->hasKeyboardFocus (true))
				&& tlw->isShowing();
	}

	JUCE_DECLARE_NON_COPYABLE (TopLevelWindowManager);
};

juce_ImplementSingleton_SingleThreaded (TopLevelWindowManager)

void juce_CheckCurrentlyFocusedTopLevelWindow()
{
	if (TopLevelWindowManager::getInstanceWithoutCreating() != nullptr)
		TopLevelWindowManager::getInstanceWithoutCreating()->startTimer (20);
}

TopLevelWindow::TopLevelWindow (const String& name,
								const bool addToDesktop_)
	: Component (name),
	  useDropShadow (true),
	  useNativeTitleBar (false),
	  windowIsActive_ (false)
{
	setOpaque (true);

	if (addToDesktop_)
		Component::addToDesktop (TopLevelWindow::getDesktopWindowStyleFlags());
	else
		setDropShadowEnabled (true);

	setWantsKeyboardFocus (true);
	setBroughtToFrontOnMouseClick (true);
	windowIsActive_ = TopLevelWindowManager::getInstance()->addWindow (this);
}

TopLevelWindow::~TopLevelWindow()
{
	shadower = nullptr;
	TopLevelWindowManager::getInstance()->removeWindow (this);
}

void TopLevelWindow::focusOfChildComponentChanged (FocusChangeType)
{
	if (hasKeyboardFocus (true))
		TopLevelWindowManager::getInstance()->timerCallback();
	else
		TopLevelWindowManager::getInstance()->startTimer (10);
}

void TopLevelWindow::setWindowActive (const bool isNowActive)
{
	if (windowIsActive_ != isNowActive)
	{
		windowIsActive_ = isNowActive;
		activeWindowStatusChanged();
	}
}

void TopLevelWindow::activeWindowStatusChanged()
{
}

void TopLevelWindow::visibilityChanged()
{
	if (isShowing()
		 && (getPeer()->getStyleFlags() & (ComponentPeer::windowIsTemporary
											| ComponentPeer::windowIgnoresKeyPresses)) == 0)
	{
		toFront (true);
	}
}

void TopLevelWindow::parentHierarchyChanged()
{
	setDropShadowEnabled (useDropShadow);
}

int TopLevelWindow::getDesktopWindowStyleFlags() const
{
	int styleFlags = ComponentPeer::windowAppearsOnTaskbar;

	if (useDropShadow)
		styleFlags |= ComponentPeer::windowHasDropShadow;

	if (useNativeTitleBar)
		styleFlags |= ComponentPeer::windowHasTitleBar;

	return styleFlags;
}

void TopLevelWindow::setDropShadowEnabled (const bool useShadow)
{
	useDropShadow = useShadow;

	if (isOnDesktop())
	{
		shadower = nullptr;
		Component::addToDesktop (getDesktopWindowStyleFlags());
	}
	else
	{
		if (useShadow && isOpaque())
		{
			if (shadower == nullptr)
			{
				shadower = getLookAndFeel().createDropShadowerForComponent (this);

				if (shadower != nullptr)
					shadower->setOwner (this);
			}
		}
		else
		{
			shadower = nullptr;
		}
	}
}

void TopLevelWindow::setUsingNativeTitleBar (const bool useNativeTitleBar_)
{
	if (useNativeTitleBar != useNativeTitleBar_)
	{
		useNativeTitleBar = useNativeTitleBar_;
		recreateDesktopWindow();
		sendLookAndFeelChange();
	}
}

void TopLevelWindow::recreateDesktopWindow()
{
	if (isOnDesktop())
	{
		Component::addToDesktop (getDesktopWindowStyleFlags());
		toFront (true);
	}
}

void TopLevelWindow::addToDesktop (int windowStyleFlags, void* nativeWindowToAttachTo)
{
	/* It's not recommended to change the desktop window flags directly for a TopLevelWindow,
	   because this class needs to make sure its layout corresponds with settings like whether
	   it's got a native title bar or not.

	   If you need custom flags for your window, you can override the getDesktopWindowStyleFlags()
	   method. If you do this, it's best to call the base class's getDesktopWindowStyleFlags()
	   method, then add or remove whatever flags are necessary from this value before returning it.
	*/

	jassert ((windowStyleFlags & ~ComponentPeer::windowIsSemiTransparent)
			   == (getDesktopWindowStyleFlags() & ~ComponentPeer::windowIsSemiTransparent));

	Component::addToDesktop (windowStyleFlags, nativeWindowToAttachTo);

	if (windowStyleFlags != getDesktopWindowStyleFlags())
		sendLookAndFeelChange();
}

void TopLevelWindow::centreAroundComponent (Component* c, const int width, const int height)
{
	if (c == nullptr)
		c = TopLevelWindow::getActiveTopLevelWindow();

	if (c == nullptr || c->getBounds().isEmpty())
	{
		centreWithSize (width, height);
	}
	else
	{
		Point<int> targetCentre (c->localPointToGlobal (c->getLocalBounds().getCentre()));
		Rectangle<int> parentArea (c->getParentMonitorArea());

		if (getParentComponent() != nullptr)
		{
			targetCentre = getParentComponent()->getLocalPoint (nullptr, targetCentre);
			parentArea = getParentComponent()->getLocalBounds();
		}

		parentArea.reduce (12, 12);

		setBounds (jlimit (parentArea.getX(), jmax (parentArea.getX(), parentArea.getRight() - width), targetCentre.getX() - width / 2),
				   jlimit (parentArea.getY(), jmax (parentArea.getY(), parentArea.getBottom() - height), targetCentre.getY() - height / 2),
				   width, height);
	}
}

int TopLevelWindow::getNumTopLevelWindows() noexcept
{
	return TopLevelWindowManager::getInstance()->windows.size();
}

TopLevelWindow* TopLevelWindow::getTopLevelWindow (const int index) noexcept
{
	return static_cast <TopLevelWindow*> (TopLevelWindowManager::getInstance()->windows [index]);
}

TopLevelWindow* TopLevelWindow::getActiveTopLevelWindow() noexcept
{
	TopLevelWindow* best = nullptr;
	int bestNumTWLParents = -1;

	for (int i = TopLevelWindow::getNumTopLevelWindows(); --i >= 0;)
	{
		TopLevelWindow* const tlw = TopLevelWindow::getTopLevelWindow (i);

		if (tlw->isActiveWindow())
		{
			int numTWLParents = 0;

			const Component* c = tlw->getParentComponent();

			while (c != nullptr)
			{
				if (dynamic_cast <const TopLevelWindow*> (c) != nullptr)
					++numTWLParents;

				c = c->getParentComponent();
			}

			if (bestNumTWLParents < numTWLParents)
			{
				best = tlw;
				bestNumTWLParents = numTWLParents;
			}
		}
	}

	return best;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TopLevelWindow.cpp ***/


/*** Start of inlined file: juce_MarkerList.cpp ***/
BEGIN_JUCE_NAMESPACE

MarkerList::MarkerList()
{
}

MarkerList::MarkerList (const MarkerList& other)
{
	operator= (other);
}

MarkerList& MarkerList::operator= (const MarkerList& other)
{
	if (other != *this)
	{
		markers.clear();
		markers.addCopiesOf (other.markers);
		markersHaveChanged();
	}

	return *this;
}

MarkerList::~MarkerList()
{
	listeners.call (&MarkerList::Listener::markerListBeingDeleted, this);
}

bool MarkerList::operator== (const MarkerList& other) const noexcept
{
	if (other.markers.size() != markers.size())
		return false;

	for (int i = markers.size(); --i >= 0;)
	{
		const Marker* const m1 = markers.getUnchecked(i);
		jassert (m1 != nullptr);

		const Marker* const m2 = other.getMarker (m1->name);

		if (m2 == nullptr || *m1 != *m2)
			return false;
	}

	return true;
}

bool MarkerList::operator!= (const MarkerList& other) const noexcept
{
	return ! operator== (other);
}

int MarkerList::getNumMarkers() const noexcept
{
	return markers.size();
}

const MarkerList::Marker* MarkerList::getMarker (const int index) const noexcept
{
	return markers [index];
}

const MarkerList::Marker* MarkerList::getMarker (const String& name) const noexcept
{
	for (int i = 0; i < markers.size(); ++i)
	{
		const Marker* const m = markers.getUnchecked(i);

		if (m->name == name)
			return m;
	}

	return nullptr;
}

void MarkerList::setMarker (const String& name, const RelativeCoordinate& position)
{
	Marker* const m = const_cast <Marker*> (getMarker (name));

	if (m != nullptr)
	{
		if (m->position != position)
		{
			m->position = position;
			markersHaveChanged();
		}

		return;
	}

	markers.add (new Marker (name, position));
	markersHaveChanged();
}

void MarkerList::removeMarker (const int index)
{
	if (isPositiveAndBelow (index, markers.size()))
	{
		markers.remove (index);
		markersHaveChanged();
	}
}

void MarkerList::removeMarker (const String& name)
{
	for (int i = 0; i < markers.size(); ++i)
	{
		const Marker* const m = markers.getUnchecked(i);

		if (m->name == name)
		{
			markers.remove (i);
			markersHaveChanged();
		}
	}
}

void MarkerList::markersHaveChanged()
{
	listeners.call (&MarkerList::Listener::markersChanged, this);
}

void MarkerList::Listener::markerListBeingDeleted (MarkerList*)
{
}

void MarkerList::addListener (Listener* listener)
{
	listeners.add (listener);
}

void MarkerList::removeListener (Listener* listener)
{
	listeners.remove (listener);
}

MarkerList::Marker::Marker (const Marker& other)
	: name (other.name), position (other.position)
{
}

MarkerList::Marker::Marker (const String& name_, const RelativeCoordinate& position_)
	: name (name_), position (position_)
{
}

bool MarkerList::Marker::operator== (const Marker& other) const noexcept
{
	return name == other.name && position == other.position;
}

bool MarkerList::Marker::operator!= (const Marker& other) const noexcept
{
	return ! operator== (other);
}

const Identifier MarkerList::ValueTreeWrapper::markerTag ("Marker");
const Identifier MarkerList::ValueTreeWrapper::nameProperty ("name");
const Identifier MarkerList::ValueTreeWrapper::posProperty ("position");

MarkerList::ValueTreeWrapper::ValueTreeWrapper (const ValueTree& state_)
	: state (state_)
{
}

int MarkerList::ValueTreeWrapper::getNumMarkers() const
{
	return state.getNumChildren();
}

ValueTree MarkerList::ValueTreeWrapper::getMarkerState (int index) const
{
	return state.getChild (index);
}

ValueTree MarkerList::ValueTreeWrapper::getMarkerState (const String& name) const
{
	return state.getChildWithProperty (nameProperty, name);
}

bool MarkerList::ValueTreeWrapper::containsMarker (const ValueTree& marker) const
{
	return marker.isAChildOf (state);
}

MarkerList::Marker MarkerList::ValueTreeWrapper::getMarker (const ValueTree& marker) const
{
	jassert (containsMarker (marker));

	return MarkerList::Marker (marker [nameProperty], RelativeCoordinate (marker [posProperty].toString()));
}

void MarkerList::ValueTreeWrapper::setMarker (const MarkerList::Marker& m, UndoManager* undoManager)
{
	ValueTree marker (state.getChildWithProperty (nameProperty, m.name));

	if (marker.isValid())
	{
		marker.setProperty (posProperty, m.position.toString(), undoManager);
	}
	else
	{
		marker = ValueTree (markerTag);
		marker.setProperty (nameProperty, m.name, nullptr);
		marker.setProperty (posProperty, m.position.toString(), nullptr);
		state.addChild (marker, -1, undoManager);
	}
}

void MarkerList::ValueTreeWrapper::removeMarker (const ValueTree& marker, UndoManager* undoManager)
{
	state.removeChild (marker, undoManager);
}

double MarkerList::getMarkerPosition (const Marker& marker, Component* parentComponent) const
{
	if (parentComponent != nullptr)
	{
		RelativeCoordinatePositionerBase::ComponentScope scope (*parentComponent);
		return marker.position.resolve (&scope);
	}
	else
	{
		return marker.position.resolve (nullptr);
	}
}

void MarkerList::ValueTreeWrapper::applyTo (MarkerList& markerList)
{
	const int numMarkers = getNumMarkers();

	StringArray updatedMarkers;

	int i;
	for (i = 0; i < numMarkers; ++i)
	{
		const ValueTree marker (state.getChild (i));
		const String name (marker [nameProperty].toString());
		markerList.setMarker (name, RelativeCoordinate (marker [posProperty].toString()));
		updatedMarkers.add (name);
	}

	for (i = markerList.getNumMarkers(); --i >= 0;)
		if (! updatedMarkers.contains (markerList.getMarker (i)->name))
			markerList.removeMarker (i);
}

void MarkerList::ValueTreeWrapper::readFrom (const MarkerList& markerList, UndoManager* undoManager)
{
	state.removeAllChildren (undoManager);

	for (int i = 0; i < markerList.getNumMarkers(); ++i)
		setMarker (*markerList.getMarker(i), undoManager);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_MarkerList.cpp ***/


/*** Start of inlined file: juce_RelativeCoordinate.cpp ***/
BEGIN_JUCE_NAMESPACE

const String RelativeCoordinate::Strings::parent ("parent");
const String RelativeCoordinate::Strings::left ("left");
const String RelativeCoordinate::Strings::right ("right");
const String RelativeCoordinate::Strings::top ("top");
const String RelativeCoordinate::Strings::bottom ("bottom");
const String RelativeCoordinate::Strings::x ("x");
const String RelativeCoordinate::Strings::y ("y");
const String RelativeCoordinate::Strings::width ("width");
const String RelativeCoordinate::Strings::height ("height");

RelativeCoordinate::StandardStrings::Type RelativeCoordinate::StandardStrings::getTypeOf (const String& s) noexcept
{
	if (s == Strings::left)	return left;
	if (s == Strings::right)   return right;
	if (s == Strings::top)	 return top;
	if (s == Strings::bottom)  return bottom;
	if (s == Strings::x)	   return x;
	if (s == Strings::y)	   return y;
	if (s == Strings::width)   return width;
	if (s == Strings::height)  return height;
	if (s == Strings::parent)  return parent;
	return unknown;
}

RelativeCoordinate::RelativeCoordinate()
{
}

RelativeCoordinate::RelativeCoordinate (const Expression& term_)
	: term (term_)
{
}

RelativeCoordinate::RelativeCoordinate (const RelativeCoordinate& other)
	: term (other.term)
{
}

RelativeCoordinate& RelativeCoordinate::operator= (const RelativeCoordinate& other)
{
	term = other.term;
	return *this;
}

RelativeCoordinate::RelativeCoordinate (const double absoluteDistanceFromOrigin)
	: term (absoluteDistanceFromOrigin)
{
}

RelativeCoordinate::RelativeCoordinate (const String& s)
{
	try
	{
		term = Expression (s);
	}
	catch (...)
	{}
}

RelativeCoordinate::~RelativeCoordinate()
{
}

bool RelativeCoordinate::operator== (const RelativeCoordinate& other) const noexcept
{
	return term.toString() == other.term.toString();
}

bool RelativeCoordinate::operator!= (const RelativeCoordinate& other) const noexcept
{
	return ! operator== (other);
}

double RelativeCoordinate::resolve (const Expression::Scope* scope) const
{
	try
	{
		if (scope != nullptr)
			return term.evaluate (*scope);
		else
			return term.evaluate();
	}
	catch (...)
	{}

	return 0.0;
}

bool RelativeCoordinate::isRecursive (const Expression::Scope* scope) const
{
	try
	{
		if (scope != nullptr)
			term.evaluate (*scope);
		else
			term.evaluate();
	}
	catch (...)
	{
		return true;
	}

	return false;
}

void RelativeCoordinate::moveToAbsolute (double newPos, const Expression::Scope* scope)
{
	try
	{
		if (scope != nullptr)
		{
			term = term.adjustedToGiveNewResult (newPos, *scope);
		}
		else
		{
			Expression::Scope defaultScope;
			term = term.adjustedToGiveNewResult (newPos, defaultScope);
		}
	}
	catch (...)
	{}
}

bool RelativeCoordinate::isDynamic() const
{
	return term.usesAnySymbols();
}

String RelativeCoordinate::toString() const
{
	return term.toString();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RelativeCoordinate.cpp ***/


/*** Start of inlined file: juce_RelativePoint.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace RelativePointHelpers
{
	inline void skipComma (String::CharPointerType& s)
	{
		s = s.findEndOfWhitespace();

		if (*s == ',')
			++s;
	}
}

RelativePoint::RelativePoint()
{
}

RelativePoint::RelativePoint (const Point<float>& absolutePoint)
	: x (absolutePoint.getX()), y (absolutePoint.getY())
{
}

RelativePoint::RelativePoint (const float x_, const float y_)
	: x (x_), y (y_)
{
}

RelativePoint::RelativePoint (const RelativeCoordinate& x_, const RelativeCoordinate& y_)
	: x (x_), y (y_)
{
}

RelativePoint::RelativePoint (const String& s)
{
	String::CharPointerType text (s.getCharPointer());
	x = RelativeCoordinate (Expression::parse (text));
	RelativePointHelpers::skipComma (text);
	y = RelativeCoordinate (Expression::parse (text));
}

bool RelativePoint::operator== (const RelativePoint& other) const noexcept
{
	return x == other.x && y == other.y;
}

bool RelativePoint::operator!= (const RelativePoint& other) const noexcept
{
	return ! operator== (other);
}

const Point<float> RelativePoint::resolve (const Expression::Scope* scope) const
{
	return Point<float> ((float) x.resolve (scope),
						 (float) y.resolve (scope));
}

void RelativePoint::moveToAbsolute (const Point<float>& newPos, const Expression::Scope* scope)
{
	x.moveToAbsolute (newPos.getX(), scope);
	y.moveToAbsolute (newPos.getY(), scope);
}

String RelativePoint::toString() const
{
	return x.toString() + ", " + y.toString();
}

bool RelativePoint::isDynamic() const
{
	return x.isDynamic() || y.isDynamic();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RelativePoint.cpp ***/


/*** Start of inlined file: juce_RelativeRectangle.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace RelativeRectangleHelpers
{
	inline void skipComma (String::CharPointerType& s)
	{
		s = s.findEndOfWhitespace();

		if (*s == ',')
			++s;
	}

	bool dependsOnSymbolsOtherThanThis (const Expression& e)
	{
		if (e.getType() == Expression::operatorType && e.getSymbolOrFunction() == ".")
			return true;

		if (e.getType() == Expression::symbolType)
		{
			switch (RelativeCoordinate::StandardStrings::getTypeOf (e.getSymbolOrFunction()))
			{
				case RelativeCoordinate::StandardStrings::x:
				case RelativeCoordinate::StandardStrings::y:
				case RelativeCoordinate::StandardStrings::left:
				case RelativeCoordinate::StandardStrings::right:
				case RelativeCoordinate::StandardStrings::top:
				case RelativeCoordinate::StandardStrings::bottom:   return false;
				default: break;
			}

			return true;
		}
		else
		{
			for (int i = e.getNumInputs(); --i >= 0;)
				if (dependsOnSymbolsOtherThanThis (e.getInput(i)))
					return true;
		}

		return false;
	}
}

RelativeRectangle::RelativeRectangle()
{
}

RelativeRectangle::RelativeRectangle (const RelativeCoordinate& left_, const RelativeCoordinate& right_,
									  const RelativeCoordinate& top_, const RelativeCoordinate& bottom_)
	: left (left_), right (right_), top (top_), bottom (bottom_)
{
}

RelativeRectangle::RelativeRectangle (const Rectangle<float>& rect)
	: left (rect.getX()),
	  right (Expression::symbol (RelativeCoordinate::Strings::left) + Expression ((double) rect.getWidth())),
	  top (rect.getY()),
	  bottom (Expression::symbol (RelativeCoordinate::Strings::top) + Expression ((double) rect.getHeight()))
{
}

RelativeRectangle::RelativeRectangle (const String& s)
{
	String::CharPointerType text (s.getCharPointer());
	left = RelativeCoordinate (Expression::parse (text));
	RelativeRectangleHelpers::skipComma (text);
	top = RelativeCoordinate (Expression::parse (text));
	RelativeRectangleHelpers::skipComma (text);
	right = RelativeCoordinate (Expression::parse (text));
	RelativeRectangleHelpers::skipComma (text);
	bottom = RelativeCoordinate (Expression::parse (text));
}

bool RelativeRectangle::operator== (const RelativeRectangle& other) const noexcept
{
	return left == other.left && top == other.top && right == other.right && bottom == other.bottom;
}

bool RelativeRectangle::operator!= (const RelativeRectangle& other) const noexcept
{
	return ! operator== (other);
}

// An expression context that can evaluate expressions using "this"
class RelativeRectangleLocalScope  : public Expression::Scope
{
public:
	RelativeRectangleLocalScope (const RelativeRectangle& rect_)  : rect (rect_) {}

	Expression getSymbolValue (const String& symbol) const
	{
		switch (RelativeCoordinate::StandardStrings::getTypeOf (symbol))
		{
			case RelativeCoordinate::StandardStrings::x:
			case RelativeCoordinate::StandardStrings::left:	 return rect.left.getExpression();
			case RelativeCoordinate::StandardStrings::y:
			case RelativeCoordinate::StandardStrings::top:	  return rect.top.getExpression();
			case RelativeCoordinate::StandardStrings::right:	return rect.right.getExpression();
			case RelativeCoordinate::StandardStrings::bottom:   return rect.bottom.getExpression();
			default: break;
		}

		return Expression::Scope::getSymbolValue (symbol);
	}

private:
	const RelativeRectangle& rect;

	JUCE_DECLARE_NON_COPYABLE (RelativeRectangleLocalScope);
};

const Rectangle<float> RelativeRectangle::resolve (const Expression::Scope* scope) const
{
	if (scope == nullptr)
	{
		RelativeRectangleLocalScope defaultScope (*this);
		return resolve (&defaultScope);
	}
	else
	{
		const double l = left.resolve (scope);
		const double r = right.resolve (scope);
		const double t = top.resolve (scope);
		const double b = bottom.resolve (scope);

		return Rectangle<float> ((float) l, (float) t, (float) jmax (0.0, r - l), (float) jmax (0.0, b - t));
	}
}

void RelativeRectangle::moveToAbsolute (const Rectangle<float>& newPos, const Expression::Scope* scope)
{
	left.moveToAbsolute (newPos.getX(), scope);
	right.moveToAbsolute (newPos.getRight(), scope);
	top.moveToAbsolute (newPos.getY(), scope);
	bottom.moveToAbsolute (newPos.getBottom(), scope);
}

bool RelativeRectangle::isDynamic() const
{
	using namespace RelativeRectangleHelpers;

	return dependsOnSymbolsOtherThanThis (left.getExpression())
			|| dependsOnSymbolsOtherThanThis (right.getExpression())
			|| dependsOnSymbolsOtherThanThis (top.getExpression())
			|| dependsOnSymbolsOtherThanThis (bottom.getExpression());
}

String RelativeRectangle::toString() const
{
	return left.toString() + ", " + top.toString() + ", " + right.toString() + ", " + bottom.toString();
}

void RelativeRectangle::renameSymbol (const Expression::Symbol& oldSymbol, const String& newName, const Expression::Scope& scope)
{
	left = left.getExpression().withRenamedSymbol (oldSymbol, newName, scope);
	right = right.getExpression().withRenamedSymbol (oldSymbol, newName, scope);
	top = top.getExpression().withRenamedSymbol (oldSymbol, newName, scope);
	bottom = bottom.getExpression().withRenamedSymbol (oldSymbol, newName, scope);
}

class RelativeRectangleComponentPositioner  : public RelativeCoordinatePositionerBase
{
public:
	RelativeRectangleComponentPositioner (Component& component_, const RelativeRectangle& rectangle_)
		: RelativeCoordinatePositionerBase (component_),
		  rectangle (rectangle_)
	{
	}

	bool registerCoordinates()
	{
		bool ok = addCoordinate (rectangle.left);
		ok = addCoordinate (rectangle.right) && ok;
		ok = addCoordinate (rectangle.top) && ok;
		ok = addCoordinate (rectangle.bottom) && ok;
		return ok;
	}

	bool isUsingRectangle (const RelativeRectangle& other) const noexcept
	{
		return rectangle == other;
	}

	void applyToComponentBounds()
	{
		for (int i = 4; --i >= 0;)
		{
			ComponentScope scope (getComponent());
			const Rectangle<int> newBounds (rectangle.resolve (&scope).getSmallestIntegerContainer());

			if (newBounds == getComponent().getBounds())
				return;

			getComponent().setBounds (newBounds);
		}

		jassertfalse; // must be a recursive reference!
	}

	void applyNewBounds (const Rectangle<int>& newBounds)
	{
		if (newBounds != getComponent().getBounds())
		{
			ComponentScope scope (getComponent());
			rectangle.moveToAbsolute (newBounds.toFloat(), &scope);

			applyToComponentBounds();
		}
	}

private:
	RelativeRectangle rectangle;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (RelativeRectangleComponentPositioner);
};

void RelativeRectangle::applyToComponent (Component& component) const
{
	if (isDynamic())
	{
		RelativeRectangleComponentPositioner* current = dynamic_cast <RelativeRectangleComponentPositioner*> (component.getPositioner());

		if (current == nullptr || ! current->isUsingRectangle (*this))
		{
			RelativeRectangleComponentPositioner* p = new RelativeRectangleComponentPositioner (component, *this);

			component.setPositioner (p);
			p->apply();
		}
	}
	else
	{
		component.setPositioner (nullptr);
		component.setBounds (resolve (nullptr).getSmallestIntegerContainer());
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RelativeRectangle.cpp ***/


/*** Start of inlined file: juce_RelativePointPath.cpp ***/
BEGIN_JUCE_NAMESPACE

RelativePointPath::RelativePointPath()
	: usesNonZeroWinding (true),
	  containsDynamicPoints (false)
{
}

RelativePointPath::RelativePointPath (const RelativePointPath& other)
	: usesNonZeroWinding (true),
	  containsDynamicPoints (false)
{
	for (int i = 0; i < other.elements.size(); ++i)
		elements.add (other.elements.getUnchecked(i)->clone());
}

RelativePointPath::RelativePointPath (const Path& path)
	: usesNonZeroWinding (path.isUsingNonZeroWinding()),
	  containsDynamicPoints (false)
{
	for (Path::Iterator i (path); i.next();)
	{
		switch (i.elementType)
		{
			case Path::Iterator::startNewSubPath:   elements.add (new StartSubPath (RelativePoint (i.x1, i.y1))); break;
			case Path::Iterator::lineTo:		elements.add (new LineTo (RelativePoint (i.x1, i.y1))); break;
			case Path::Iterator::quadraticTo:	   elements.add (new QuadraticTo (RelativePoint (i.x1, i.y1), RelativePoint (i.x2, i.y2))); break;
			case Path::Iterator::cubicTo:	   elements.add (new CubicTo (RelativePoint (i.x1, i.y1), RelativePoint (i.x2, i.y2), RelativePoint (i.x3, i.y3))); break;
			case Path::Iterator::closePath:	 elements.add (new CloseSubPath()); break;
			default:				jassertfalse; break;
		}
	}
}

RelativePointPath::~RelativePointPath()
{
}

bool RelativePointPath::operator== (const RelativePointPath& other) const noexcept
{
	if (elements.size() != other.elements.size()
		 || usesNonZeroWinding != other.usesNonZeroWinding
		 || containsDynamicPoints != other.containsDynamicPoints)
		return false;

	for (int i = 0; i < elements.size(); ++i)
	{
		ElementBase* const e1 = elements.getUnchecked(i);
		ElementBase* const e2 = other.elements.getUnchecked(i);

		if (e1->type != e2->type)
			return false;

		int numPoints1, numPoints2;
		const RelativePoint* const points1 = e1->getControlPoints (numPoints1);
		const RelativePoint* const points2 = e2->getControlPoints (numPoints2);

		jassert (numPoints1 == numPoints2);

		for (int j = numPoints1; --j >= 0;)
			if (points1[j] != points2[j])
				return false;
	}

	return true;
}

bool RelativePointPath::operator!= (const RelativePointPath& other) const noexcept
{
	return ! operator== (other);
}

void RelativePointPath::swapWith (RelativePointPath& other) noexcept
{
	elements.swapWithArray (other.elements);
	std::swap (usesNonZeroWinding, other.usesNonZeroWinding);
	std::swap (containsDynamicPoints, other.containsDynamicPoints);
}

void RelativePointPath::createPath (Path& path, Expression::Scope* scope) const
{
	for (int i = 0; i < elements.size(); ++i)
		elements.getUnchecked(i)->addToPath (path, scope);
}

bool RelativePointPath::containsAnyDynamicPoints() const
{
	return containsDynamicPoints;
}

void RelativePointPath::addElement (ElementBase* newElement)
{
	if (newElement != nullptr)
	{
		elements.add (newElement);
		containsDynamicPoints = containsDynamicPoints || newElement->isDynamic();
	}
}

RelativePointPath::ElementBase::ElementBase (const ElementType type_) : type (type_)
{
}

bool RelativePointPath::ElementBase::isDynamic()
{
	int numPoints;
	const RelativePoint* const points = getControlPoints (numPoints);

	for (int i = numPoints; --i >= 0;)
		if (points[i].isDynamic())
			return true;

	return false;
}

RelativePointPath::StartSubPath::StartSubPath (const RelativePoint& pos)
	: ElementBase (startSubPathElement), startPos (pos)
{
}

ValueTree RelativePointPath::StartSubPath::createTree() const
{
	ValueTree v (DrawablePath::ValueTreeWrapper::Element::startSubPathElement);
	v.setProperty (DrawablePath::ValueTreeWrapper::point1, startPos.toString(), nullptr);
	return v;
}

void RelativePointPath::StartSubPath::addToPath (Path& path, Expression::Scope* scope) const
{
	path.startNewSubPath (startPos.resolve (scope));
}

RelativePoint* RelativePointPath::StartSubPath::getControlPoints (int& numPoints)
{
	numPoints = 1;
	return &startPos;
}

RelativePointPath::ElementBase* RelativePointPath::StartSubPath::clone() const
{
	return new StartSubPath (startPos);
}

RelativePointPath::CloseSubPath::CloseSubPath()
	: ElementBase (closeSubPathElement)
{
}

ValueTree RelativePointPath::CloseSubPath::createTree() const
{
	return ValueTree (DrawablePath::ValueTreeWrapper::Element::closeSubPathElement);
}

void RelativePointPath::CloseSubPath::addToPath (Path& path, Expression::Scope*) const
{
	path.closeSubPath();
}

RelativePoint* RelativePointPath::CloseSubPath::getControlPoints (int& numPoints)
{
	numPoints = 0;
	return nullptr;
}

RelativePointPath::ElementBase* RelativePointPath::CloseSubPath::clone() const
{
	return new CloseSubPath();
}

RelativePointPath::LineTo::LineTo (const RelativePoint& endPoint_)
	: ElementBase (lineToElement), endPoint (endPoint_)
{
}

ValueTree RelativePointPath::LineTo::createTree() const
{
	ValueTree v (DrawablePath::ValueTreeWrapper::Element::lineToElement);
	v.setProperty (DrawablePath::ValueTreeWrapper::point1, endPoint.toString(), nullptr);
	return v;
}

void RelativePointPath::LineTo::addToPath (Path& path, Expression::Scope* scope) const
{
	path.lineTo (endPoint.resolve (scope));
}

RelativePoint* RelativePointPath::LineTo::getControlPoints (int& numPoints)
{
	numPoints = 1;
	return &endPoint;
}

RelativePointPath::ElementBase* RelativePointPath::LineTo::clone() const
{
	return new LineTo (endPoint);
}

RelativePointPath::QuadraticTo::QuadraticTo (const RelativePoint& controlPoint, const RelativePoint& endPoint)
	: ElementBase (quadraticToElement)
{
	controlPoints[0] = controlPoint;
	controlPoints[1] = endPoint;
}

ValueTree RelativePointPath::QuadraticTo::createTree() const
{
	ValueTree v (DrawablePath::ValueTreeWrapper::Element::quadraticToElement);
	v.setProperty (DrawablePath::ValueTreeWrapper::point1, controlPoints[0].toString(), nullptr);
	v.setProperty (DrawablePath::ValueTreeWrapper::point2, controlPoints[1].toString(), nullptr);
	return v;
}

void RelativePointPath::QuadraticTo::addToPath (Path& path, Expression::Scope* scope) const
{
	path.quadraticTo (controlPoints[0].resolve (scope),
					  controlPoints[1].resolve (scope));
}

RelativePoint* RelativePointPath::QuadraticTo::getControlPoints (int& numPoints)
{
	numPoints = 2;
	return controlPoints;
}

RelativePointPath::ElementBase* RelativePointPath::QuadraticTo::clone() const
{
	return new QuadraticTo (controlPoints[0], controlPoints[1]);
}

RelativePointPath::CubicTo::CubicTo (const RelativePoint& controlPoint1, const RelativePoint& controlPoint2, const RelativePoint& endPoint)
	: ElementBase (cubicToElement)
{
	controlPoints[0] = controlPoint1;
	controlPoints[1] = controlPoint2;
	controlPoints[2] = endPoint;
}

ValueTree RelativePointPath::CubicTo::createTree() const
{
	ValueTree v (DrawablePath::ValueTreeWrapper::Element::cubicToElement);
	v.setProperty (DrawablePath::ValueTreeWrapper::point1, controlPoints[0].toString(), nullptr);
	v.setProperty (DrawablePath::ValueTreeWrapper::point2, controlPoints[1].toString(), nullptr);
	v.setProperty (DrawablePath::ValueTreeWrapper::point3, controlPoints[2].toString(), nullptr);
	return v;
}

void RelativePointPath::CubicTo::addToPath (Path& path, Expression::Scope* scope) const
{
	path.cubicTo (controlPoints[0].resolve (scope),
				  controlPoints[1].resolve (scope),
				  controlPoints[2].resolve (scope));
}

RelativePoint* RelativePointPath::CubicTo::getControlPoints (int& numPoints)
{
	numPoints = 3;
	return controlPoints;
}

RelativePointPath::ElementBase* RelativePointPath::CubicTo::clone() const
{
	return new CubicTo (controlPoints[0], controlPoints[1], controlPoints[2]);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RelativePointPath.cpp ***/


/*** Start of inlined file: juce_RelativeParallelogram.cpp ***/
BEGIN_JUCE_NAMESPACE

RelativeParallelogram::RelativeParallelogram()
{
}

RelativeParallelogram::RelativeParallelogram (const Rectangle<float>& r)
	: topLeft (r.getTopLeft()), topRight (r.getTopRight()), bottomLeft (r.getBottomLeft())
{
}

RelativeParallelogram::RelativeParallelogram (const RelativePoint& topLeft_, const RelativePoint& topRight_, const RelativePoint& bottomLeft_)
	: topLeft (topLeft_), topRight (topRight_), bottomLeft (bottomLeft_)
{
}

RelativeParallelogram::RelativeParallelogram (const String& topLeft_, const String& topRight_, const String& bottomLeft_)
	: topLeft (topLeft_), topRight (topRight_), bottomLeft (bottomLeft_)
{
}

RelativeParallelogram::~RelativeParallelogram()
{
}

void RelativeParallelogram::resolveThreePoints (Point<float>* points, Expression::Scope* const scope) const
{
	points[0] = topLeft.resolve (scope);
	points[1] = topRight.resolve (scope);
	points[2] = bottomLeft.resolve (scope);
}

void RelativeParallelogram::resolveFourCorners (Point<float>* points, Expression::Scope* const scope) const
{
	resolveThreePoints (points, scope);
	points[3] = points[1] + (points[2] - points[0]);
}

const Rectangle<float> RelativeParallelogram::getBounds (Expression::Scope* const scope) const
{
	Point<float> points[4];
	resolveFourCorners (points, scope);
	return Rectangle<float>::findAreaContainingPoints (points, 4);
}

void RelativeParallelogram::getPath (Path& path, Expression::Scope* const scope) const
{
	Point<float> points[4];
	resolveFourCorners (points, scope);

	path.startNewSubPath (points[0]);
	path.lineTo (points[1]);
	path.lineTo (points[3]);
	path.lineTo (points[2]);
	path.closeSubPath();
}

const AffineTransform RelativeParallelogram::resetToPerpendicular (Expression::Scope* const scope)
{
	Point<float> corners[3];
	resolveThreePoints (corners, scope);

	const Line<float> top (corners[0], corners[1]);
	const Line<float> left (corners[0], corners[2]);
	const Point<float> newTopRight (corners[0] + Point<float> (top.getLength(), 0.0f));
	const Point<float> newBottomLeft (corners[0] + Point<float> (0.0f, left.getLength()));

	topRight.moveToAbsolute (newTopRight, scope);
	bottomLeft.moveToAbsolute (newBottomLeft, scope);

	return AffineTransform::fromTargetPoints (corners[0].getX(), corners[0].getY(), corners[0].getX(), corners[0].getY(),
											  corners[1].getX(), corners[1].getY(), newTopRight.getX(), newTopRight.getY(),
											  corners[2].getX(), corners[2].getY(), newBottomLeft.getX(), newBottomLeft.getY());
}

bool RelativeParallelogram::isDynamic() const
{
	return topLeft.isDynamic() || topRight.isDynamic() || bottomLeft.isDynamic();
}

bool RelativeParallelogram::operator== (const RelativeParallelogram& other) const noexcept
{
	return topLeft == other.topLeft && topRight == other.topRight && bottomLeft == other.bottomLeft;
}

bool RelativeParallelogram::operator!= (const RelativeParallelogram& other) const noexcept
{
	return ! operator== (other);
}

const Point<float> RelativeParallelogram::getInternalCoordForPoint (const Point<float>* const corners, Point<float> target) noexcept
{
	const Point<float> tr (corners[1] - corners[0]);
	const Point<float> bl (corners[2] - corners[0]);
	target -= corners[0];

	return Point<float> (Line<float> (Point<float>(), tr).getIntersection (Line<float> (target, target - bl)).getDistanceFromOrigin(),
						 Line<float> (Point<float>(), bl).getIntersection (Line<float> (target, target - tr)).getDistanceFromOrigin());
}

const Point<float> RelativeParallelogram::getPointForInternalCoord (const Point<float>* const corners, const Point<float>& point) noexcept
{
	return corners[0]
			+ Line<float> (Point<float>(), corners[1] - corners[0]).getPointAlongLine (point.getX())
			+ Line<float> (Point<float>(), corners[2] - corners[0]).getPointAlongLine (point.getY());
}

const Rectangle<float> RelativeParallelogram::getBoundingBox (const Point<float>* const p) noexcept
{
	const Point<float> points[] = { p[0], p[1], p[2], p[1] + (p[2] - p[0]) };
	return Rectangle<float>::findAreaContainingPoints (points, 4);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RelativeParallelogram.cpp ***/


/*** Start of inlined file: juce_RelativeCoordinatePositioner.cpp ***/
BEGIN_JUCE_NAMESPACE

RelativeCoordinatePositionerBase::ComponentScope::ComponentScope (Component& component_)
	: component (component_)
{
}

Expression RelativeCoordinatePositionerBase::ComponentScope::getSymbolValue (const String& symbol) const
{
	switch (RelativeCoordinate::StandardStrings::getTypeOf (symbol))
	{
		case RelativeCoordinate::StandardStrings::x:
		case RelativeCoordinate::StandardStrings::left:   return Expression ((double) component.getX());
		case RelativeCoordinate::StandardStrings::y:
		case RelativeCoordinate::StandardStrings::top:	return Expression ((double) component.getY());
		case RelativeCoordinate::StandardStrings::width:  return Expression ((double) component.getWidth());
		case RelativeCoordinate::StandardStrings::height: return Expression ((double) component.getHeight());
		case RelativeCoordinate::StandardStrings::right:  return Expression ((double) component.getRight());
		case RelativeCoordinate::StandardStrings::bottom: return Expression ((double) component.getBottom());
		default: break;
	}

	MarkerList* list;
	const MarkerList::Marker* const marker = findMarker (symbol, list);

	if (marker != nullptr)
		return marker->position.getExpression();

	return Expression::Scope::getSymbolValue (symbol);
}

void RelativeCoordinatePositionerBase::ComponentScope::visitRelativeScope (const String& scopeName, Visitor& visitor) const
{
	Component* targetComp = nullptr;

	if (scopeName == RelativeCoordinate::Strings::parent)
		targetComp = component.getParentComponent();
	else
		targetComp = findSiblingComponent (scopeName);

	if (targetComp != nullptr)
		visitor.visit (ComponentScope (*targetComp));
	else
		Expression::Scope::visitRelativeScope (scopeName, visitor);
}

String RelativeCoordinatePositionerBase::ComponentScope::getScopeUID() const
{
	return String::toHexString ((int) (pointer_sized_int) (void*) &component);
}

Component* RelativeCoordinatePositionerBase::ComponentScope::findSiblingComponent (const String& componentID) const
{
	Component* const parent = component.getParentComponent();

	if (parent != nullptr)
	{
		for (int i = parent->getNumChildComponents(); --i >= 0;)
		{
			Component* const c = parent->getChildComponent(i);

			if (c->getComponentID() == componentID)
				return c;
		}
	}

	return nullptr;
}

const MarkerList::Marker* RelativeCoordinatePositionerBase::ComponentScope::findMarker (const String& name, MarkerList*& list) const
{
	const MarkerList::Marker* marker = nullptr;

	Component* const parent = component.getParentComponent();

	if (parent != nullptr)
	{
		list = parent->getMarkers (true);
		if (list != nullptr)
			marker = list->getMarker (name);

		if (marker == nullptr)
		{
			list = parent->getMarkers (false);

			if (list != nullptr)
				marker = list->getMarker (name);
		}
	}

	return marker;
}

class RelativeCoordinatePositionerBase::DependencyFinderScope  : public ComponentScope
{
public:
	DependencyFinderScope (Component& component_, RelativeCoordinatePositionerBase& positioner_, bool& ok_)
		: ComponentScope (component_), positioner (positioner_), ok (ok_)
	{
	}

	Expression getSymbolValue (const String& symbol) const
	{
		if (symbol == RelativeCoordinate::Strings::left || symbol == RelativeCoordinate::Strings::x
			 || symbol == RelativeCoordinate::Strings::width || symbol == RelativeCoordinate::Strings::right
			 || symbol == RelativeCoordinate::Strings::top || symbol == RelativeCoordinate::Strings::y
			 || symbol == RelativeCoordinate::Strings::height || symbol == RelativeCoordinate::Strings::bottom)
		{
			positioner.registerComponentListener (component);
		}
		else
		{
			MarkerList* list;
			const MarkerList::Marker* const marker = findMarker (symbol, list);

			if (marker != nullptr)
			{
				positioner.registerMarkerListListener (list);
			}
			else
			{
				// The marker we want doesn't exist, so watch all lists in case they change and the marker appears later..
				positioner.registerMarkerListListener (component.getMarkers (true));
				positioner.registerMarkerListListener (component.getMarkers (false));
				ok = false;
			}
		}

		return ComponentScope::getSymbolValue (symbol);
	}

	void visitRelativeScope (const String& scopeName, Visitor& visitor) const
	{
		Component* targetComp = nullptr;

		if (scopeName == RelativeCoordinate::Strings::parent)
			targetComp = component.getParentComponent();
		else
			targetComp = findSiblingComponent (scopeName);

		if (targetComp != nullptr)
		{
			visitor.visit (DependencyFinderScope (*targetComp, positioner, ok));
		}
		else
		{
			// The named component doesn't exist, so we'll watch the parent for changes in case it appears later..
			Component* const parent = component.getParentComponent();
			if (parent != nullptr)
				positioner.registerComponentListener (*parent);

			positioner.registerComponentListener (component);
			ok = false;
		}
	}

private:
	RelativeCoordinatePositionerBase& positioner;
	bool& ok;

	JUCE_DECLARE_NON_COPYABLE (DependencyFinderScope);
};

RelativeCoordinatePositionerBase::RelativeCoordinatePositionerBase (Component& component_)
	: Component::Positioner (component_), registeredOk (false)
{
}

RelativeCoordinatePositionerBase::~RelativeCoordinatePositionerBase()
{
	unregisterListeners();
}

void RelativeCoordinatePositionerBase::componentMovedOrResized (Component&, bool /*wasMoved*/, bool /*wasResized*/)
{
	apply();
}

void RelativeCoordinatePositionerBase::componentParentHierarchyChanged (Component&)
{
	apply();
}

void RelativeCoordinatePositionerBase::componentChildrenChanged (Component& changed)
{
	if (getComponent().getParentComponent() == &changed && ! registeredOk)
		apply();
}

void RelativeCoordinatePositionerBase::componentBeingDeleted (Component& comp)
{
	jassert (sourceComponents.contains (&comp));
	sourceComponents.removeValue (&comp);
	registeredOk = false;
}

void RelativeCoordinatePositionerBase::markersChanged (MarkerList*)
{
	apply();
}

void RelativeCoordinatePositionerBase::markerListBeingDeleted (MarkerList* markerList)
{
	jassert (sourceMarkerLists.contains (markerList));
	sourceMarkerLists.removeValue (markerList);
}

void RelativeCoordinatePositionerBase::apply()
{
	if (! registeredOk)
	{
		unregisterListeners();
		registeredOk = registerCoordinates();
	}

	applyToComponentBounds();
}

bool RelativeCoordinatePositionerBase::addCoordinate (const RelativeCoordinate& coord)
{
	bool ok = true;
	DependencyFinderScope finderScope (getComponent(), *this, ok);
	coord.getExpression().evaluate (finderScope);
	return ok;
}

bool RelativeCoordinatePositionerBase::addPoint (const RelativePoint& point)
{
	const bool ok = addCoordinate (point.x);
	return addCoordinate (point.y) && ok;
}

void RelativeCoordinatePositionerBase::registerComponentListener (Component& comp)
{
	if (! sourceComponents.contains (&comp))
	{
		comp.addComponentListener (this);
		sourceComponents.add (&comp);
	}
}

void RelativeCoordinatePositionerBase::registerMarkerListListener (MarkerList* const list)
{
	if (list != nullptr && ! sourceMarkerLists.contains (list))
	{
		list->addListener (this);
		sourceMarkerLists.add (list);
	}
}

void RelativeCoordinatePositionerBase::unregisterListeners()
{
	int i;
	for (i = sourceComponents.size(); --i >= 0;)
		sourceComponents.getUnchecked(i)->removeComponentListener (this);

	for (i = sourceMarkerLists.size(); --i >= 0;)
		sourceMarkerLists.getUnchecked(i)->removeListener (this);

	sourceComponents.clear();
	sourceMarkerLists.clear();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RelativeCoordinatePositioner.cpp ***/

#endif // JUCE_BUILD_GUI


