/*
  ==============================================================================

   This file is part of the JUCE library - "Jules' Utility Class Extensions"
   Copyright 2004-9 by Raw Material Software Ltd.

  ------------------------------------------------------------------------------

   JUCE can be redistributed and/or modified under the terms of the GNU General
   Public License (Version 2), as published by the Free Software Foundation.
   A copy of the license is included in the JUCE distribution, or can be found
   online at www.gnu.org/licenses.

   JUCE is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  ------------------------------------------------------------------------------

   To release a closed-source product which uses JUCE, commercial licenses are
   available: visit www.rawmaterialsoftware.com/juce for more information.

  ==============================================================================
*/

#ifdef __JUCE_JUCEHEADER__
 /* When you add the amalgamated cpp file to your project, you mustn't include it in
	a file where you've already included juce.h - just put it inside a file on its own,
	possibly with your config flags preceding it, but don't include anything else. */
 #error
#endif


#include "juce_TargetPlatform.h"

 // FORCE_AMALGAMATOR_INCLUDE

#include "juce_Config.h"

  // FORCE_AMALGAMATOR_INCLUDE

#ifndef JUCE_BUILD_CORE
 #define JUCE_BUILD_CORE 1
#endif
#ifndef JUCE_BUILD_MISC
 #define JUCE_BUILD_MISC 1
#endif
#ifndef JUCE_BUILD_GUI
 #define JUCE_BUILD_GUI 1
#endif
#ifndef JUCE_BUILD_NATIVE
 #define JUCE_BUILD_NATIVE 1
#endif

#if JUCE_ONLY_BUILD_CORE_LIBRARY
 #undef JUCE_BUILD_MISC
 #undef JUCE_BUILD_GUI
#endif

#ifndef JUCE_NAMESPACE
 #define JUCE_NAMESPACE juce
#endif

//==============================================================================
#include "juce_NativeIncludes.h"
//==============================================================================
#define DONT_SET_USING_JUCE_NAMESPACE 1
#undef max
#undef min
#define NO_DUMMY_DECL

#define JUCE_AMALGAMATED_TEMPLATE 1

#if JUCE_BUILD_NATIVE
 #include "juce_amalgamated.h"  // FORCE_AMALGAMATOR_INCLUDE
#endif

#if (defined(_MSC_VER) && (_MSC_VER <= 1200))
 #pragma warning (disable: 4309 4305)
#endif


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


#if JUCE_BUILD_MISC  // (put these in misc to balance the file sizes and avoid problems in iphone build)

/*** Start of inlined file: juce_Colour.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace ColourHelpers
{
	uint8 floatAlphaToInt (const float alpha) noexcept
	{
		return (uint8) jlimit (0, 0xff, roundToInt (alpha * 255.0f));
	}

	void convertHSBtoRGB (float h, float s, float v,
						  uint8& r, uint8& g, uint8& b) noexcept
	{
		v = jlimit (0.0f, 1.0f, v);
		v *= 255.0f;
		const uint8 intV = (uint8) roundToInt (v);

		if (s <= 0)
		{
			r = intV;
			g = intV;
			b = intV;
		}
		else
		{
			s = jmin (1.0f, s);
			h = (h - std::floor (h)) * 6.0f + 0.00001f; // need a small adjustment to compensate for rounding errors
			const float f = h - std::floor (h);
			const uint8 x = (uint8) roundToInt (v * (1.0f - s));

			if (h < 1.0f)
			{
				r = intV;
				g = (uint8) roundToInt (v * (1.0f - (s * (1.0f - f))));
				b = x;
			}
			else if (h < 2.0f)
			{
				r = (uint8) roundToInt (v * (1.0f - s * f));
				g = intV;
				b = x;
			}
			else if (h < 3.0f)
			{
				r = x;
				g = intV;
				b = (uint8) roundToInt (v * (1.0f - (s * (1.0f - f))));
			}
			else if (h < 4.0f)
			{
				r = x;
				g = (uint8) roundToInt (v * (1.0f - s * f));
				b = intV;
			}
			else if (h < 5.0f)
			{
				r = (uint8) roundToInt (v * (1.0f - (s * (1.0f - f))));
				g = x;
				b = intV;
			}
			else
			{
				r = intV;
				g = x;
				b = (uint8) roundToInt (v * (1.0f - s * f));
			}
		}
	}
}

Colour::Colour() noexcept
	: argb (0)
{
}

Colour::Colour (const Colour& other) noexcept
	: argb (other.argb)
{
}

Colour& Colour::operator= (const Colour& other) noexcept
{
	argb = other.argb;
	return *this;
}

bool Colour::operator== (const Colour& other) const noexcept
{
	return argb.getARGB() == other.argb.getARGB();
}

bool Colour::operator!= (const Colour& other) const noexcept
{
	return argb.getARGB() != other.argb.getARGB();
}

Colour::Colour (const uint32 argb_) noexcept
	: argb (argb_)
{
}

Colour::Colour (const uint8 red,
				const uint8 green,
				const uint8 blue) noexcept
{
	argb.setARGB (0xff, red, green, blue);
}

Colour Colour::fromRGB (const uint8 red,
						const uint8 green,
						const uint8 blue) noexcept
{
	return Colour (red, green, blue);
}

Colour::Colour (const uint8 red,
				const uint8 green,
				const uint8 blue,
				const uint8 alpha) noexcept
{
	argb.setARGB (alpha, red, green, blue);
}

Colour Colour::fromRGBA (const uint8 red,
						 const uint8 green,
						 const uint8 blue,
						 const uint8 alpha) noexcept
{
	return Colour (red, green, blue, alpha);
}

Colour::Colour (const uint8 red,
				const uint8 green,
				const uint8 blue,
				const float alpha) noexcept
{
	argb.setARGB (ColourHelpers::floatAlphaToInt (alpha), red, green, blue);
}

Colour Colour::fromRGBAFloat (const uint8 red,
							  const uint8 green,
							  const uint8 blue,
							  const float alpha) noexcept
{
	return Colour (red, green, blue, alpha);
}

Colour::Colour (const float hue,
				const float saturation,
				const float brightness,
				const float alpha) noexcept
{
	uint8 r, g, b;
	ColourHelpers::convertHSBtoRGB (hue, saturation, brightness, r, g, b);

	argb.setARGB (ColourHelpers::floatAlphaToInt (alpha), r, g, b);
}

Colour Colour::fromHSV (const float hue,
						const float saturation,
						const float brightness,
						const float alpha) noexcept
{
	return Colour (hue, saturation, brightness, alpha);
}

Colour::Colour (const float hue,
				const float saturation,
				const float brightness,
				const uint8 alpha) noexcept
{
	uint8 r, g, b;
	ColourHelpers::convertHSBtoRGB (hue, saturation, brightness, r, g, b);

	argb.setARGB (alpha, r, g, b);
}

Colour::~Colour() noexcept
{
}

const PixelARGB Colour::getPixelARGB() const noexcept
{
	PixelARGB p (argb);
	p.premultiply();
	return p;
}

uint32 Colour::getARGB() const noexcept
{
	return argb.getARGB();
}

bool Colour::isTransparent() const noexcept
{
	return getAlpha() == 0;
}

bool Colour::isOpaque() const noexcept
{
	return getAlpha() == 0xff;
}

Colour Colour::withAlpha (const uint8 newAlpha) const noexcept
{
	PixelARGB newCol (argb);
	newCol.setAlpha (newAlpha);
	return Colour (newCol.getARGB());
}

Colour Colour::withAlpha (const float newAlpha) const noexcept
{
	jassert (newAlpha >= 0 && newAlpha <= 1.0f);

	PixelARGB newCol (argb);
	newCol.setAlpha (ColourHelpers::floatAlphaToInt (newAlpha));
	return Colour (newCol.getARGB());
}

Colour Colour::withMultipliedAlpha (const float alphaMultiplier) const noexcept
{
	jassert (alphaMultiplier >= 0);

	PixelARGB newCol (argb);
	newCol.setAlpha ((uint8) jmin (0xff, roundToInt (alphaMultiplier * newCol.getAlpha())));
	return Colour (newCol.getARGB());
}

Colour Colour::overlaidWith (const Colour& src) const noexcept
{
	const int destAlpha = getAlpha();

	if (destAlpha > 0)
	{
		const int invA = 0xff - (int) src.getAlpha();
		const int resA = 0xff - (((0xff - destAlpha) * invA) >> 8);

		if (resA > 0)
		{
			const int da = (invA * destAlpha) / resA;

			return Colour ((uint8) (src.getRed()   + ((((int) getRed() - src.getRed())	 * da) >> 8)),
						   (uint8) (src.getGreen() + ((((int) getGreen() - src.getGreen()) * da) >> 8)),
						   (uint8) (src.getBlue()  + ((((int) getBlue() - src.getBlue())   * da) >> 8)),
						   (uint8) resA);
		}

		return *this;
	}
	else
	{
		return src;
	}
}

Colour Colour::interpolatedWith (const Colour& other, float proportionOfOther) const noexcept
{
	if (proportionOfOther <= 0)
		return *this;

	if (proportionOfOther >= 1.0f)
		return other;

	PixelARGB c1 (getPixelARGB());
	const PixelARGB c2 (other.getPixelARGB());
	c1.tween (c2, roundToInt (proportionOfOther * 255.0f));
	c1.unpremultiply();

	return Colour (c1.getARGB());
}

float Colour::getFloatRed() const noexcept
{
	return getRed() / 255.0f;
}

float Colour::getFloatGreen() const noexcept
{
	return getGreen() / 255.0f;
}

float Colour::getFloatBlue() const noexcept
{
	return getBlue() / 255.0f;
}

float Colour::getFloatAlpha() const noexcept
{
	return getAlpha() / 255.0f;
}

void Colour::getHSB (float& h, float& s, float& v) const noexcept
{
	const int r = getRed();
	const int g = getGreen();
	const int b = getBlue();

	const int hi = jmax (r, g, b);
	const int lo = jmin (r, g, b);

	if (hi != 0)
	{
		s = (hi - lo) / (float) hi;

		if (s > 0)
		{
			const float invDiff = 1.0f / (hi - lo);

			const float red   = (hi - r) * invDiff;
			const float green = (hi - g) * invDiff;
			const float blue  = (hi - b) * invDiff;

			if (r == hi)
				h = blue - green;
			else if (g == hi)
				h = 2.0f + red - blue;
			else
				h = 4.0f + green - red;

			h *= 1.0f / 6.0f;

			if (h < 0)
				++h;
		}
		else
		{
			h = 0;
		}
	}
	else
	{
		s = 0;
		h = 0;
	}

	v = hi / 255.0f;
}

float Colour::getHue() const noexcept
{
	float h, s, b;
	getHSB (h, s, b);
	return h;
}

Colour Colour::withHue (const float hue) const noexcept
{
	float h, s, b;
	getHSB (h, s, b);

	return Colour (hue, s, b, getAlpha());
}

Colour Colour::withRotatedHue (const float amountToRotate) const noexcept
{
	float h, s, b;
	getHSB (h, s, b);

	return Colour (h + amountToRotate, s, b, getAlpha());
}

float Colour::getSaturation() const noexcept
{
	float h, s, b;
	getHSB (h, s, b);
	return s;
}

Colour Colour::withSaturation (const float saturation) const noexcept
{
	float h, s, b;
	getHSB (h, s, b);

	return Colour (h, saturation, b, getAlpha());
}

Colour Colour::withMultipliedSaturation (const float amount) const noexcept
{
	float h, s, b;
	getHSB (h, s, b);

	return Colour (h, jmin (1.0f, s * amount), b, getAlpha());
}

float Colour::getBrightness() const noexcept
{
	float h, s, b;
	getHSB (h, s, b);
	return b;
}

Colour Colour::withBrightness (const float brightness) const noexcept
{
	float h, s, b;
	getHSB (h, s, b);

	return Colour (h, s, brightness, getAlpha());
}

Colour Colour::withMultipliedBrightness (const float amount) const noexcept
{
	float h, s, b;
	getHSB (h, s, b);

	b *= amount;

	if (b > 1.0f)
		b = 1.0f;

	return Colour (h, s, b, getAlpha());
}

Colour Colour::brighter (float amount) const noexcept
{
	amount = 1.0f / (1.0f + amount);

	return Colour ((uint8) (255 - (amount * (255 - getRed()))),
				   (uint8) (255 - (amount * (255 - getGreen()))),
				   (uint8) (255 - (amount * (255 - getBlue()))),
				   getAlpha());
}

Colour Colour::darker (float amount) const noexcept
{
	amount = 1.0f / (1.0f + amount);

	return Colour ((uint8) (amount * getRed()),
				   (uint8) (amount * getGreen()),
				   (uint8) (amount * getBlue()),
				   getAlpha());
}

Colour Colour::greyLevel (const float brightness) noexcept
{
	const uint8 level
		= (uint8) jlimit (0x00, 0xff, roundToInt (brightness * 255.0f));

	return Colour (level, level, level);
}

Colour Colour::contrasting (const float amount) const noexcept
{
	return overlaidWith ((((int) getRed() + (int) getGreen() + (int) getBlue() >= 3 * 128)
							? Colours::black
							: Colours::white).withAlpha (amount));
}

Colour Colour::contrasting (const Colour& colour1,
							const Colour& colour2) noexcept
{
	const float b1 = colour1.getBrightness();
	const float b2 = colour2.getBrightness();
	float best = 0.0f;
	float bestDist = 0.0f;

	for (float i = 0.0f; i < 1.0f; i += 0.02f)
	{
		const float d1 = std::abs (i - b1);
		const float d2 = std::abs (i - b2);
		const float dist = jmin (d1, d2, 1.0f - d1, 1.0f - d2);

		if (dist > bestDist)
		{
			best = i;
			bestDist = dist;
		}
	}

	return colour1.overlaidWith (colour2.withMultipliedAlpha (0.5f))
				  .withBrightness (best);
}

String Colour::toString() const
{
	return String::toHexString ((int) argb.getARGB());
}

Colour Colour::fromString (const String& encodedColourString)
{
	return Colour ((uint32) encodedColourString.getHexValue32());
}

String Colour::toDisplayString (const bool includeAlphaValue) const
{
	return String::toHexString ((int) (argb.getARGB() & (includeAlphaValue ? 0xffffffff : 0xffffff)))
				  .paddedLeft ('0', includeAlphaValue ? 8 : 6)
				  .toUpperCase();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Colour.cpp ***/



/*** Start of inlined file: juce_ColourGradient.cpp ***/
BEGIN_JUCE_NAMESPACE

ColourGradient::ColourGradient() noexcept
{
   #if JUCE_DEBUG
	point1.setX (987654.0f);
   #endif
}

ColourGradient::ColourGradient (const Colour& colour1, const float x1_, const float y1_,
								const Colour& colour2, const float x2_, const float y2_,
								const bool isRadial_)
	: point1 (x1_, y1_),
	  point2 (x2_, y2_),
	  isRadial (isRadial_)
{
	colours.add (ColourPoint (0.0, colour1));
	colours.add (ColourPoint (1.0, colour2));
}

ColourGradient::~ColourGradient()
{
}

bool ColourGradient::operator== (const ColourGradient& other) const noexcept
{
	return point1 == other.point1 && point2 == other.point2
			&& isRadial == other.isRadial
			&& colours == other.colours;
}

bool ColourGradient::operator!= (const ColourGradient& other) const noexcept
{
	return ! operator== (other);
}

void ColourGradient::clearColours()
{
	colours.clear();
}

int ColourGradient::addColour (const double proportionAlongGradient, const Colour& colour)
{
	// must be within the two end-points
	jassert (proportionAlongGradient >= 0 && proportionAlongGradient <= 1.0);

	const double pos = jlimit (0.0, 1.0, proportionAlongGradient);

	int i;
	for (i = 0; i < colours.size(); ++i)
		if (colours.getReference(i).position > pos)
			break;

	colours.insert (i, ColourPoint (pos, colour));
	return i;
}

void ColourGradient::removeColour (int index)
{
	jassert (index > 0 && index < colours.size() - 1);
	colours.remove (index);
}

void ColourGradient::multiplyOpacity (const float multiplier) noexcept
{
	for (int i = 0; i < colours.size(); ++i)
	{
		Colour& c = colours.getReference(i).colour;
		c = c.withMultipliedAlpha (multiplier);
	}
}

int ColourGradient::getNumColours() const noexcept
{
	return colours.size();
}

double ColourGradient::getColourPosition (const int index) const noexcept
{
	if (isPositiveAndBelow (index, colours.size()))
		return colours.getReference (index).position;

	return 0;
 }

const Colour ColourGradient::getColour (const int index) const noexcept
{
	if (isPositiveAndBelow (index, colours.size()))
		return colours.getReference (index).colour;

	return Colour();
}

void ColourGradient::setColour (int index, const Colour& newColour) noexcept
{
	if (isPositiveAndBelow (index, colours.size()))
		colours.getReference (index).colour = newColour;
}

Colour ColourGradient::getColourAtPosition (const double position) const noexcept
{
	jassert (colours.getReference(0).position == 0); // the first colour specified has to go at position 0

	if (position <= 0 || colours.size() <= 1)
		return colours.getReference(0).colour;

	int i = colours.size() - 1;
	while (position < colours.getReference(i).position)
		--i;

	const ColourPoint& p1 = colours.getReference (i);

	if (i >= colours.size() - 1)
		return p1.colour;

	const ColourPoint& p2 = colours.getReference (i + 1);

	return p1.colour.interpolatedWith (p2.colour, (float) ((position - p1.position) / (p2.position - p1.position)));
}

int ColourGradient::createLookupTable (const AffineTransform& transform, HeapBlock <PixelARGB>& lookupTable) const
{
   #if JUCE_DEBUG
	// trying to use the object without setting its co-ordinates? Have a careful read of
	// the comments for the constructors.
	jassert (point1.getX() != 987654.0f);
   #endif

	const int numEntries = jlimit (1, jmax (1, (colours.size() - 1) << 8),
								   3 * (int) point1.transformedBy (transform)
												.getDistanceFrom (point2.transformedBy (transform)));
	lookupTable.malloc (numEntries);

	if (colours.size() >= 2)
	{
		jassert (colours.getReference(0).position == 0); // the first colour specified has to go at position 0

		PixelARGB pix1 (colours.getReference (0).colour.getPixelARGB());
		int index = 0;

		for (int j = 1; j < colours.size(); ++j)
		{
			const ColourPoint& p = colours.getReference (j);
			const int numToDo = roundToInt (p.position * (numEntries - 1)) - index;
			const PixelARGB pix2 (p.colour.getPixelARGB());

			for (int i = 0; i < numToDo; ++i)
			{
				jassert (index >= 0 && index < numEntries);

				lookupTable[index] = pix1;
				lookupTable[index].tween (pix2, (i << 8) / numToDo);
				++index;
			}

			pix1 = pix2;
		}

		while (index < numEntries)
			lookupTable [index++] = pix1;
	}
	else
	{
		jassertfalse; // no colours specified!
	}

	return numEntries;
}

bool ColourGradient::isOpaque() const noexcept
{
	for (int i = 0; i < colours.size(); ++i)
		if (! colours.getReference(i).colour.isOpaque())
			return false;

	return true;
}

bool ColourGradient::isInvisible() const noexcept
{
	for (int i = 0; i < colours.size(); ++i)
		if (! colours.getReference(i).colour.isTransparent())
			return false;

	return true;
}

bool ColourGradient::ColourPoint::operator== (const ColourPoint& other) const noexcept
{
	return position == other.position && colour == other.colour;
}

bool ColourGradient::ColourPoint::operator!= (const ColourPoint& other) const noexcept
{
	return position != other.position || colour != other.colour;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ColourGradient.cpp ***/


/*** Start of inlined file: juce_Colours.cpp ***/
BEGIN_JUCE_NAMESPACE

const Colour Colours::transparentBlack (0);
const Colour Colours::transparentWhite (0x00ffffff);

const Colour Colours::aliceblue (0xfff0f8ff);
const Colour Colours::antiquewhite (0xfffaebd7);
const Colour Colours::aqua (0xff00ffff);
const Colour Colours::aquamarine (0xff7fffd4);
const Colour Colours::azure (0xfff0ffff);
const Colour Colours::beige (0xfff5f5dc);
const Colour Colours::bisque (0xffffe4c4);
const Colour Colours::black (0xff000000);
const Colour Colours::blanchedalmond (0xffffebcd);
const Colour Colours::blue (0xff0000ff);
const Colour Colours::blueviolet (0xff8a2be2);
const Colour Colours::brown (0xffa52a2a);
const Colour Colours::burlywood (0xffdeb887);
const Colour Colours::cadetblue (0xff5f9ea0);
const Colour Colours::chartreuse (0xff7fff00);
const Colour Colours::chocolate (0xffd2691e);
const Colour Colours::coral (0xffff7f50);
const Colour Colours::cornflowerblue (0xff6495ed);
const Colour Colours::cornsilk (0xfffff8dc);
const Colour Colours::crimson (0xffdc143c);
const Colour Colours::cyan (0xff00ffff);
const Colour Colours::darkblue (0xff00008b);
const Colour Colours::darkcyan (0xff008b8b);
const Colour Colours::darkgoldenrod (0xffb8860b);
const Colour Colours::darkgrey (0xff555555);
const Colour Colours::darkgreen (0xff006400);
const Colour Colours::darkkhaki (0xffbdb76b);
const Colour Colours::darkmagenta (0xff8b008b);
const Colour Colours::darkolivegreen (0xff556b2f);
const Colour Colours::darkorange (0xffff8c00);
const Colour Colours::darkorchid (0xff9932cc);
const Colour Colours::darkred (0xff8b0000);
const Colour Colours::darksalmon (0xffe9967a);
const Colour Colours::darkseagreen (0xff8fbc8f);
const Colour Colours::darkslateblue (0xff483d8b);
const Colour Colours::darkslategrey (0xff2f4f4f);
const Colour Colours::darkturquoise (0xff00ced1);
const Colour Colours::darkviolet (0xff9400d3);
const Colour Colours::deeppink (0xffff1493);
const Colour Colours::deepskyblue (0xff00bfff);
const Colour Colours::dimgrey (0xff696969);
const Colour Colours::dodgerblue (0xff1e90ff);
const Colour Colours::firebrick (0xffb22222);
const Colour Colours::floralwhite (0xfffffaf0);
const Colour Colours::forestgreen (0xff228b22);
const Colour Colours::fuchsia (0xffff00ff);
const Colour Colours::gainsboro (0xffdcdcdc);
const Colour Colours::gold (0xffffd700);
const Colour Colours::goldenrod (0xffdaa520);
const Colour Colours::grey (0xff808080);
const Colour Colours::green (0xff008000);
const Colour Colours::greenyellow (0xffadff2f);
const Colour Colours::honeydew (0xfff0fff0);
const Colour Colours::hotpink (0xffff69b4);
const Colour Colours::indianred (0xffcd5c5c);
const Colour Colours::indigo (0xff4b0082);
const Colour Colours::ivory (0xfffffff0);
const Colour Colours::khaki (0xfff0e68c);
const Colour Colours::lavender (0xffe6e6fa);
const Colour Colours::lavenderblush (0xfffff0f5);
const Colour Colours::lemonchiffon (0xfffffacd);
const Colour Colours::lightblue (0xffadd8e6);
const Colour Colours::lightcoral (0xfff08080);
const Colour Colours::lightcyan (0xffe0ffff);
const Colour Colours::lightgoldenrodyellow (0xfffafad2);
const Colour Colours::lightgreen (0xff90ee90);
const Colour Colours::lightgrey (0xffd3d3d3);
const Colour Colours::lightpink (0xffffb6c1);
const Colour Colours::lightsalmon (0xffffa07a);
const Colour Colours::lightseagreen (0xff20b2aa);
const Colour Colours::lightskyblue (0xff87cefa);
const Colour Colours::lightslategrey (0xff778899);
const Colour Colours::lightsteelblue (0xffb0c4de);
const Colour Colours::lightyellow (0xffffffe0);
const Colour Colours::lime (0xff00ff00);
const Colour Colours::limegreen (0xff32cd32);
const Colour Colours::linen (0xfffaf0e6);
const Colour Colours::magenta (0xffff00ff);
const Colour Colours::maroon (0xff800000);
const Colour Colours::mediumaquamarine (0xff66cdaa);
const Colour Colours::mediumblue (0xff0000cd);
const Colour Colours::mediumorchid (0xffba55d3);
const Colour Colours::mediumpurple (0xff9370db);
const Colour Colours::mediumseagreen (0xff3cb371);
const Colour Colours::mediumslateblue (0xff7b68ee);
const Colour Colours::mediumspringgreen (0xff00fa9a);
const Colour Colours::mediumturquoise (0xff48d1cc);
const Colour Colours::mediumvioletred (0xffc71585);
const Colour Colours::midnightblue (0xff191970);
const Colour Colours::mintcream (0xfff5fffa);
const Colour Colours::mistyrose (0xffffe4e1);
const Colour Colours::navajowhite (0xffffdead);
const Colour Colours::navy (0xff000080);
const Colour Colours::oldlace (0xfffdf5e6);
const Colour Colours::olive (0xff808000);
const Colour Colours::olivedrab (0xff6b8e23);
const Colour Colours::orange (0xffffa500);
const Colour Colours::orangered (0xffff4500);
const Colour Colours::orchid (0xffda70d6);
const Colour Colours::palegoldenrod (0xffeee8aa);
const Colour Colours::palegreen (0xff98fb98);
const Colour Colours::paleturquoise (0xffafeeee);
const Colour Colours::palevioletred (0xffdb7093);
const Colour Colours::papayawhip (0xffffefd5);
const Colour Colours::peachpuff (0xffffdab9);
const Colour Colours::peru (0xffcd853f);
const Colour Colours::pink (0xffffc0cb);
const Colour Colours::plum (0xffdda0dd);
const Colour Colours::powderblue (0xffb0e0e6);
const Colour Colours::purple (0xff800080);
const Colour Colours::red (0xffff0000);
const Colour Colours::rosybrown (0xffbc8f8f);
const Colour Colours::royalblue (0xff4169e1);
const Colour Colours::saddlebrown (0xff8b4513);
const Colour Colours::salmon (0xfffa8072);
const Colour Colours::sandybrown (0xfff4a460);
const Colour Colours::seagreen (0xff2e8b57);
const Colour Colours::seashell (0xfffff5ee);
const Colour Colours::sienna (0xffa0522d);
const Colour Colours::silver (0xffc0c0c0);
const Colour Colours::skyblue (0xff87ceeb);
const Colour Colours::slateblue (0xff6a5acd);
const Colour Colours::slategrey (0xff708090);
const Colour Colours::snow (0xfffffafa);
const Colour Colours::springgreen (0xff00ff7f);
const Colour Colours::steelblue (0xff4682b4);
const Colour Colours::tan (0xffd2b48c);
const Colour Colours::teal (0xff008080);
const Colour Colours::thistle (0xffd8bfd8);
const Colour Colours::tomato (0xffff6347);
const Colour Colours::turquoise (0xff40e0d0);
const Colour Colours::violet (0xffee82ee);
const Colour Colours::wheat (0xfff5deb3);
const Colour Colours::white (0xffffffff);
const Colour Colours::whitesmoke (0xfff5f5f5);
const Colour Colours::yellow (0xffffff00);
const Colour Colours::yellowgreen (0xff9acd32);

const Colour Colours::findColourForName (const String& colourName,
										 const Colour& defaultColour)
{
	static const unsigned int presets[] =
	{
		// (first value is the string's hashcode, second is ARGB)

		0x05978fff, 0xff000000, /* black */
		0x06bdcc29, 0xffffffff, /* white */
		0x002e305a, 0xff0000ff, /* blue */
		0x00308adf, 0xff808080, /* grey */
		0x05e0cf03, 0xff008000, /* green */
		0x0001b891, 0xffff0000, /* red */
		0xd43c6474, 0xffffff00, /* yellow */
		0x620886da, 0xfff0f8ff, /* aliceblue */
		0x20a2676a, 0xfffaebd7, /* antiquewhite */
		0x002dcebc, 0xff00ffff, /* aqua */
		0x46bb5f7e, 0xff7fffd4, /* aquamarine */
		0x0590228f, 0xfff0ffff, /* azure */
		0x05947fe4, 0xfff5f5dc, /* beige */
		0xad388e35, 0xffffe4c4, /* bisque */
		0x00674f7e, 0xffffebcd, /* blanchedalmond */
		0x39129959, 0xff8a2be2, /* blueviolet */
		0x059a8136, 0xffa52a2a, /* brown */
		0x89cea8f9, 0xffdeb887, /* burlywood */
		0x0fa260cf, 0xff5f9ea0, /* cadetblue */
		0x6b748956, 0xff7fff00, /* chartreuse */
		0x2903623c, 0xffd2691e, /* chocolate */
		0x05a74431, 0xffff7f50, /* coral */
		0x618d42dd, 0xff6495ed, /* cornflowerblue */
		0xe4b479fd, 0xfffff8dc, /* cornsilk */
		0x3d8c4edf, 0xffdc143c, /* crimson */
		0x002ed323, 0xff00ffff, /* cyan */
		0x67cc74d0, 0xff00008b, /* darkblue */
		0x67cd1799, 0xff008b8b, /* darkcyan */
		0x31bbd168, 0xffb8860b, /* darkgoldenrod */
		0x67cecf55, 0xff555555, /* darkgrey */
		0x920b194d, 0xff006400, /* darkgreen */
		0x923edd4c, 0xffbdb76b, /* darkkhaki */
		0x5c293873, 0xff8b008b, /* darkmagenta */
		0x6b6671fe, 0xff556b2f, /* darkolivegreen */
		0xbcfd2524, 0xffff8c00, /* darkorange */
		0xbcfdf799, 0xff9932cc, /* darkorchid */
		0x55ee0d5b, 0xff8b0000, /* darkred */
		0xc2e5f564, 0xffe9967a, /* darksalmon */
		0x61be858a, 0xff8fbc8f, /* darkseagreen */
		0xc2b0f2bd, 0xff483d8b, /* darkslateblue */
		0xc2b34d42, 0xff2f4f4f, /* darkslategrey */
		0x7cf2b06b, 0xff00ced1, /* darkturquoise */
		0xc8769375, 0xff9400d3, /* darkviolet */
		0x25832862, 0xffff1493, /* deeppink */
		0xfcad568f, 0xff00bfff, /* deepskyblue */
		0x634c8b67, 0xff696969, /* dimgrey */
		0x45c1ce55, 0xff1e90ff, /* dodgerblue */
		0xef19e3cb, 0xffb22222, /* firebrick */
		0xb852b195, 0xfffffaf0, /* floralwhite */
		0xd086fd06, 0xff228b22, /* forestgreen */
		0xe106b6d7, 0xffff00ff, /* fuchsia */
		0x7880d61e, 0xffdcdcdc, /* gainsboro */
		0x00308060, 0xffffd700, /* gold */
		0xb3b3bc1e, 0xffdaa520, /* goldenrod */
		0xbab8a537, 0xffadff2f, /* greenyellow */
		0xe4cacafb, 0xfff0fff0, /* honeydew */
		0x41892743, 0xffff69b4, /* hotpink */
		0xd5796f1a, 0xffcd5c5c, /* indianred */
		0xb969fed2, 0xff4b0082, /* indigo */
		0x05fef6a9, 0xfffffff0, /* ivory */
		0x06149302, 0xfff0e68c, /* khaki */
		0xad5a05c7, 0xffe6e6fa, /* lavender */
		0x7c4d5b99, 0xfffff0f5, /* lavenderblush */
		0x195756f0, 0xfffffacd, /* lemonchiffon */
		0x28e4ea70, 0xffadd8e6, /* lightblue */
		0xf3c7ccdb, 0xfff08080, /* lightcoral */
		0x28e58d39, 0xffe0ffff, /* lightcyan */
		0x21234e3c, 0xfffafad2, /* lightgoldenrodyellow */
		0xf40157ad, 0xff90ee90, /* lightgreen */
		0x28e744f5, 0xffd3d3d3, /* lightgrey */
		0x28eb3b8c, 0xffffb6c1, /* lightpink */
		0x9fb78304, 0xffffa07a, /* lightsalmon */
		0x50632b2a, 0xff20b2aa, /* lightseagreen */
		0x68fb7b25, 0xff87cefa, /* lightskyblue */
		0xa8a35ba2, 0xff778899, /* lightslategrey */
		0xa20d484f, 0xffb0c4de, /* lightsteelblue */
		0xaa2cf10a, 0xffffffe0, /* lightyellow */
		0x0032afd5, 0xff00ff00, /* lime */
		0x607bbc4e, 0xff32cd32, /* limegreen */
		0x06234efa, 0xfffaf0e6, /* linen */
		0x316858a9, 0xffff00ff, /* magenta */
		0xbf8ca470, 0xff800000, /* maroon */
		0xbd58e0b3, 0xff66cdaa, /* mediumaquamarine */
		0x967dfd4f, 0xff0000cd, /* mediumblue */
		0x056f5c58, 0xffba55d3, /* mediumorchid */
		0x07556b71, 0xff9370db, /* mediumpurple */
		0x5369b689, 0xff3cb371, /* mediumseagreen */
		0x066be19e, 0xff7b68ee, /* mediumslateblue */
		0x3256b281, 0xff00fa9a, /* mediumspringgreen */
		0xc0ad9f4c, 0xff48d1cc, /* mediumturquoise */
		0x628e63dd, 0xffc71585, /* mediumvioletred */
		0x168eb32a, 0xff191970, /* midnightblue */
		0x4306b960, 0xfff5fffa, /* mintcream */
		0x4cbc0e6b, 0xffffe4e1, /* mistyrose */
		0xe97218a6, 0xffffdead, /* navajowhite */
		0x00337bb6, 0xff000080, /* navy */
		0xadd2d33e, 0xfffdf5e6, /* oldlace */
		0x064ee1db, 0xff808000, /* olive */
		0x9e33a98a, 0xff6b8e23, /* olivedrab */
		0xc3de262e, 0xffffa500, /* orange */
		0x58bebba3, 0xffff4500, /* orangered */
		0xc3def8a3, 0xffda70d6, /* orchid */
		0x28cb4834, 0xffeee8aa, /* palegoldenrod */
		0x3d9dd619, 0xff98fb98, /* palegreen */
		0x74022737, 0xffafeeee, /* paleturquoise */
		0x15e2ebc8, 0xffdb7093, /* palevioletred */
		0x5fd898e2, 0xffffefd5, /* papayawhip */
		0x93e1b776, 0xffffdab9, /* peachpuff */
		0x003472f8, 0xffcd853f, /* peru */
		0x00348176, 0xffffc0cb, /* pink */
		0x00348d94, 0xffdda0dd, /* plum */
		0xd036be93, 0xffb0e0e6, /* powderblue */
		0xc5c507bc, 0xff800080, /* purple */
		0xa89d65b3, 0xffbc8f8f, /* rosybrown */
		0xbd9413e1, 0xff4169e1, /* royalblue */
		0xf456044f, 0xff8b4513, /* saddlebrown */
		0xc9c6f66e, 0xfffa8072, /* salmon */
		0x0bb131e1, 0xfff4a460, /* sandybrown */
		0x34636c14, 0xff2e8b57, /* seagreen */
		0x3507fb41, 0xfffff5ee, /* seashell */
		0xca348772, 0xffa0522d, /* sienna */
		0xca37d30d, 0xffc0c0c0, /* silver */
		0x80da74fb, 0xff87ceeb, /* skyblue */
		0x44a8dd73, 0xff6a5acd, /* slateblue */
		0x44ab37f8, 0xff708090, /* slategrey */
		0x0035f183, 0xfffffafa, /* snow */
		0xd5440d16, 0xff00ff7f, /* springgreen */
		0x3e1524a5, 0xff4682b4, /* steelblue */
		0x0001bfa1, 0xffd2b48c, /* tan */
		0x0036425c, 0xff008080, /* teal */
		0xafc8858f, 0xffd8bfd8, /* thistle */
		0xcc41600a, 0xffff6347, /* tomato */
		0xfeea9b21, 0xff40e0d0, /* turquoise */
		0xcf57947f, 0xffee82ee, /* violet */
		0x06bdbae7, 0xfff5deb3, /* wheat */
		0x10802ee6, 0xfff5f5f5, /* whitesmoke */
		0xe1b5130f, 0xff9acd32  /* yellowgreen */
	};

	const int hash = colourName.trim().toLowerCase().hashCode();

	for (int i = 0; i < numElementsInArray (presets); i += 2)
		if (presets [i] == hash)
			return Colour (presets [i + 1]);

	return defaultColour;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Colours.cpp ***/


/*** Start of inlined file: juce_EdgeTable.cpp ***/
BEGIN_JUCE_NAMESPACE

const int juce_edgeTableDefaultEdgesPerLine = 32;

EdgeTable::EdgeTable (const Rectangle<int>& bounds_,
					  const Path& path, const AffineTransform& transform)
   : bounds (bounds_),
	 maxEdgesPerLine (juce_edgeTableDefaultEdgesPerLine),
	 lineStrideElements ((juce_edgeTableDefaultEdgesPerLine << 1) + 1),
	 needToCheckEmptinesss (true)
{
	table.malloc ((bounds.getHeight() + 1) * lineStrideElements);
	int* t = table;

	for (int i = bounds.getHeight(); --i >= 0;)
	{
		*t = 0;
		t += lineStrideElements;
	}

	const int topLimit = bounds.getY() << 8;
	const int heightLimit = bounds.getHeight() << 8;
	const int leftLimit = bounds.getX() << 8;
	const int rightLimit = bounds.getRight() << 8;

	PathFlatteningIterator iter (path, transform);

	while (iter.next())
	{
		int y1 = roundToInt (iter.y1 * 256.0f);
		int y2 = roundToInt (iter.y2 * 256.0f);

		if (y1 != y2)
		{
			y1 -= topLimit;
			y2 -= topLimit;

			const int startY = y1;
			int direction = -1;

			if (y1 > y2)
			{
				std::swap (y1, y2);
				direction = 1;
			}

			if (y1 < 0)
				y1 = 0;

			if (y2 > heightLimit)
				y2 = heightLimit;

			if (y1 < y2)
			{
				const double startX = 256.0f * iter.x1;
				const double multiplier = (iter.x2 - iter.x1) / (iter.y2 - iter.y1);
				const int stepSize = jlimit (1, 256, 256 / (1 + (int) std::abs (multiplier)));

				do
				{
					const int step = jmin (stepSize, y2 - y1, 256 - (y1 & 255));
					int x = roundToInt (startX + multiplier * ((y1 + (step >> 1)) - startY));

					if (x < leftLimit)
						x = leftLimit;
					else if (x >= rightLimit)
						x = rightLimit - 1;

					addEdgePoint (x, y1 >> 8, direction * step);
					y1 += step;
				}
				while (y1 < y2);
			}
		}
	}

	sanitiseLevels (path.isUsingNonZeroWinding());
}

EdgeTable::EdgeTable (const Rectangle<int>& rectangleToAdd)
   : bounds (rectangleToAdd),
	 maxEdgesPerLine (juce_edgeTableDefaultEdgesPerLine),
	 lineStrideElements ((juce_edgeTableDefaultEdgesPerLine << 1) + 1),
	 needToCheckEmptinesss (true)
{
	table.malloc (jmax (1, bounds.getHeight()) * lineStrideElements);
	table[0] = 0;

	const int x1 = rectangleToAdd.getX() << 8;
	const int x2 = rectangleToAdd.getRight() << 8;

	int* t = table;
	for (int i = rectangleToAdd.getHeight(); --i >= 0;)
	{
		t[0] = 2;
		t[1] = x1;
		t[2] = 255;
		t[3] = x2;
		t[4] = 0;
		t += lineStrideElements;
	}
}

EdgeTable::EdgeTable (const RectangleList& rectanglesToAdd)
   : bounds (rectanglesToAdd.getBounds()),
	 maxEdgesPerLine (juce_edgeTableDefaultEdgesPerLine),
	 lineStrideElements ((juce_edgeTableDefaultEdgesPerLine << 1) + 1),
	 needToCheckEmptinesss (true)
{
	table.malloc (jmax (1, bounds.getHeight()) * lineStrideElements);

	int* t = table;
	for (int i = bounds.getHeight(); --i >= 0;)
	{
		*t = 0;
		t += lineStrideElements;
	}

	for (RectangleList::Iterator iter (rectanglesToAdd); iter.next();)
	{
		const Rectangle<int>* const r = iter.getRectangle();

		const int x1 = r->getX() << 8;
		const int x2 = r->getRight() << 8;
		int y = r->getY() - bounds.getY();

		for (int j = r->getHeight(); --j >= 0;)
		{
			addEdgePoint (x1, y, 255);
			addEdgePoint (x2, y, -255);
			++y;
		}
	}

	sanitiseLevels (true);
}

EdgeTable::EdgeTable (const Rectangle<float>& rectangleToAdd)
   : bounds (Rectangle<int> ((int) std::floor (rectangleToAdd.getX()),
							 roundToInt (rectangleToAdd.getY() * 256.0f) >> 8,
							 2 + (int) rectangleToAdd.getWidth(),
							 2 + (int) rectangleToAdd.getHeight())),
	 maxEdgesPerLine (juce_edgeTableDefaultEdgesPerLine),
	 lineStrideElements ((juce_edgeTableDefaultEdgesPerLine << 1) + 1),
	 needToCheckEmptinesss (true)
{
	jassert (! rectangleToAdd.isEmpty());
	table.malloc (jmax (1, bounds.getHeight()) * lineStrideElements);
	table[0] = 0;

	const int x1 = roundToInt (rectangleToAdd.getX() * 256.0f);
	const int x2 = roundToInt (rectangleToAdd.getRight() * 256.0f);

	int y1 = roundToInt (rectangleToAdd.getY() * 256.0f) - (bounds.getY() << 8);
	jassert (y1 < 256);
	int y2 = roundToInt (rectangleToAdd.getBottom() * 256.0f) - (bounds.getY() << 8);

	if (x2 <= x1 || y2 <= y1)
	{
		bounds.setHeight (0);
		return;
	}

	int lineY = 0;
	int* t = table;

	if ((y1 >> 8) == (y2 >> 8))
	{
		t[0] = 2;
		t[1] = x1;
		t[2] = y2 - y1;
		t[3] = x2;
		t[4] = 0;
		++lineY;
		t += lineStrideElements;
	}
	else
	{
		t[0] = 2;
		t[1] = x1;
		t[2] = 255 - (y1 & 255);
		t[3] = x2;
		t[4] = 0;
		++lineY;
		t += lineStrideElements;

		while (lineY < (y2 >> 8))
		{
			t[0] = 2;
			t[1] = x1;
			t[2] = 255;
			t[3] = x2;
			t[4] = 0;
			++lineY;
			t += lineStrideElements;
		}

		jassert (lineY < bounds.getHeight());
		t[0] = 2;
		t[1] = x1;
		t[2] = y2 & 255;
		t[3] = x2;
		t[4] = 0;
		++lineY;
		t += lineStrideElements;
	}

	while (lineY < bounds.getHeight())
	{
		t[0] = 0;
		t += lineStrideElements;
		++lineY;
	}
}

EdgeTable::EdgeTable (const EdgeTable& other)
{
	operator= (other);
}

EdgeTable& EdgeTable::operator= (const EdgeTable& other)
{
	bounds = other.bounds;
	maxEdgesPerLine = other.maxEdgesPerLine;
	lineStrideElements = other.lineStrideElements;
	needToCheckEmptinesss = other.needToCheckEmptinesss;

	table.malloc (jmax (1, bounds.getHeight()) * lineStrideElements);
	copyEdgeTableData (table, lineStrideElements, other.table, lineStrideElements, bounds.getHeight());
	return *this;
}

EdgeTable::~EdgeTable()
{
}

void EdgeTable::copyEdgeTableData (int* dest, const int destLineStride, const int* src, const int srcLineStride, int numLines) noexcept
{
	while (--numLines >= 0)
	{
		memcpy (dest, src, (src[0] * 2 + 1) * sizeof (int));
		src += srcLineStride;
		dest += destLineStride;
	}
}

void EdgeTable::sanitiseLevels (const bool useNonZeroWinding) noexcept
{
	// Convert the table from relative windings to absolute levels..
	int* lineStart = table;

	for (int i = bounds.getHeight(); --i >= 0;)
	{
		int* line = lineStart;
		lineStart += lineStrideElements;

		int num = *line;
		if (num == 0)
			continue;

		int level = 0;

		if (useNonZeroWinding)
		{
			while (--num > 0)
			{
				line += 2;
				level += *line;
				int corrected = abs (level);
				if (corrected >> 8)
					corrected = 255;

				*line = corrected;
			}
		}
		else
		{
			while (--num > 0)
			{
				line += 2;
				level += *line;
				int corrected = abs (level);
				if (corrected >> 8)
				{
					corrected &= 511;
					if (corrected >> 8)
						corrected = 511 - corrected;
				}

				*line = corrected;
			}
		}

		line[2] = 0; // force the last level to 0, just in case something went wrong in creating the table
	}
}

void EdgeTable::remapTableForNumEdges (const int newNumEdgesPerLine)
{
	if (newNumEdgesPerLine != maxEdgesPerLine)
	{
		maxEdgesPerLine = newNumEdgesPerLine;

		jassert (bounds.getHeight() > 0);
		const int newLineStrideElements = maxEdgesPerLine * 2 + 1;

		HeapBlock <int> newTable (bounds.getHeight() * newLineStrideElements);

		copyEdgeTableData (newTable, newLineStrideElements, table, lineStrideElements, bounds.getHeight());

		table.swapWith (newTable);
		lineStrideElements = newLineStrideElements;
	}
}

void EdgeTable::optimiseTable()
{
	int maxLineElements = 0;

	for (int i = bounds.getHeight(); --i >= 0;)
		maxLineElements = jmax (maxLineElements, table [i * lineStrideElements]);

	remapTableForNumEdges (maxLineElements);
}

void EdgeTable::addEdgePoint (const int x, const int y, const int winding)
{
	jassert (y >= 0 && y < bounds.getHeight());

	int* line = table + lineStrideElements * y;
	const int numPoints = line[0];
	int n = numPoints << 1;

	if (n > 0)
	{
		while (n > 0)
		{
			const int cx = line [n - 1];

			if (cx <= x)
			{
				if (cx == x)
				{
					line [n] += winding;
					return;
				}

				break;
			}

			n -= 2;
		}

		if (numPoints >= maxEdgesPerLine)
		{
			remapTableForNumEdges (maxEdgesPerLine + juce_edgeTableDefaultEdgesPerLine);
			jassert (numPoints < maxEdgesPerLine);
			line = table + lineStrideElements * y;
		}

		memmove (line + (n + 3), line + (n + 1), sizeof (int) * ((numPoints << 1) - n));
	}

	line [n + 1] = x;
	line [n + 2] = winding;
	line[0]++;
}

void EdgeTable::translate (float dx, const int dy) noexcept
{
	bounds.translate ((int) std::floor (dx), dy);

	int* lineStart = table;
	const int intDx = (int) (dx * 256.0f);

	for (int i = bounds.getHeight(); --i >= 0;)
	{
		int* line = lineStart;
		lineStart += lineStrideElements;
		int num = *line++;

		while (--num >= 0)
		{
			*line += intDx;
			line += 2;
		}
	}
}

void EdgeTable::intersectWithEdgeTableLine (const int y, const int* otherLine)
{
	jassert (y >= 0 && y < bounds.getHeight());

	int* dest = table + lineStrideElements * y;
	if (dest[0] == 0)
		return;

	int otherNumPoints = *otherLine;
	if (otherNumPoints == 0)
	{
		*dest = 0;
		return;
	}

	const int right = bounds.getRight() << 8;

	// optimise for the common case where our line lies entirely within a
	// single pair of points, as happens when clipping to a simple rect.
	if (otherNumPoints == 2 && otherLine[2] >= 255)
	{
		clipEdgeTableLineToRange (dest, otherLine[1], jmin (right, otherLine[3]));
		return;
	}

	++otherLine;
	const size_t lineSizeBytes = (dest[0] * 2 + 1) * sizeof (int);
  std::vector<char> buffer(lineSizeBytes);
	int* temp = reinterpret_cast<int*> (&buffer[0]);
	memcpy (temp, dest, lineSizeBytes);

	const int* src1 = temp;
	int srcNum1 = *src1++;
	int x1 = *src1++;

	const int* src2 = otherLine;
	int srcNum2 = otherNumPoints;
	int x2 = *src2++;

	int destIndex = 0, destTotal = 0;
	int level1 = 0, level2 = 0;
	int lastX = std::numeric_limits<int>::min(), lastLevel = 0;

	while (srcNum1 > 0 && srcNum2 > 0)
	{
		int nextX;

		if (x1 < x2)
		{
			nextX = x1;
			level1 = *src1++;
			x1 = *src1++;
			--srcNum1;
		}
		else if (x1 == x2)
		{
			nextX = x1;
			level1 = *src1++;
			level2 = *src2++;
			x1 = *src1++;
			x2 = *src2++;
			--srcNum1;
			--srcNum2;
		}
		else
		{
			nextX = x2;
			level2 = *src2++;
			x2 = *src2++;
			--srcNum2;
		}

		if (nextX > lastX)
		{
			if (nextX >= right)
				break;

			lastX = nextX;

			const int nextLevel = (level1 * (level2 + 1)) >> 8;
			jassert (isPositiveAndBelow (nextLevel, (int) 256));

			if (nextLevel != lastLevel)
			{
				if (destTotal >= maxEdgesPerLine)
				{
					dest[0] = destTotal;
					remapTableForNumEdges (maxEdgesPerLine + juce_edgeTableDefaultEdgesPerLine);
					dest = table + lineStrideElements * y;
				}

				++destTotal;
				lastLevel = nextLevel;
				dest[++destIndex] = nextX;
				dest[++destIndex] = nextLevel;
			}
		}
	}

	if (lastLevel > 0)
	{
		if (destTotal >= maxEdgesPerLine)
		{
			dest[0] = destTotal;
			remapTableForNumEdges (maxEdgesPerLine + juce_edgeTableDefaultEdgesPerLine);
			dest = table + lineStrideElements * y;
		}

		++destTotal;
		dest[++destIndex] = right;
		dest[++destIndex] = 0;
	}

	dest[0] = destTotal;

#if JUCE_DEBUG
	int last = std::numeric_limits<int>::min();
	for (int i = 0; i < dest[0]; ++i)
	{
		jassert (dest[i * 2 + 1] > last);
		last = dest[i * 2 + 1];
	}

	jassert (dest [dest[0] * 2] == 0);
#endif
}

void EdgeTable::clipEdgeTableLineToRange (int* dest, const int x1, const int x2) noexcept
{
	int* lastItem = dest + (dest[0] * 2 - 1);

	if (x2 < lastItem[0])
	{
		if (x2 <= dest[1])
		{
			dest[0] = 0;
			return;
		}

		while (x2 < lastItem[-2])
		{
			--(dest[0]);
			lastItem -= 2;
		}

		lastItem[0] = x2;
		lastItem[1] = 0;
	}

	if (x1 > dest[1])
	{
		while (lastItem[0] > x1)
			lastItem -= 2;

		const int itemsRemoved = (int) (lastItem - (dest + 1)) / 2;

		if (itemsRemoved > 0)
		{
			dest[0] -= itemsRemoved;
			memmove (dest + 1, lastItem, dest[0] * (sizeof (int) * 2));
		}

		dest[1] = x1;
	}
}

void EdgeTable::clipToRectangle (const Rectangle<int>& r)
{
	const Rectangle<int> clipped (r.getIntersection (bounds));

	if (clipped.isEmpty())
	{
		needToCheckEmptinesss = false;
		bounds.setHeight (0);
	}
	else
	{
		const int top = clipped.getY() - bounds.getY();
		const int bottom = clipped.getBottom() - bounds.getY();

		if (bottom < bounds.getHeight())
			bounds.setHeight (bottom);

		if (clipped.getRight() < bounds.getRight())
			bounds.setRight (clipped.getRight());

		for (int i = top; --i >= 0;)
			table [lineStrideElements * i] = 0;

		if (clipped.getX() > bounds.getX())
		{
			const int x1 = clipped.getX() << 8;
			const int x2 = jmin (bounds.getRight(), clipped.getRight()) << 8;
			int* line = table + lineStrideElements * top;

			for (int i = bottom - top; --i >= 0;)
			{
				if (line[0] != 0)
					clipEdgeTableLineToRange (line, x1, x2);

				line += lineStrideElements;
			}
		}

		needToCheckEmptinesss = true;
	}
}

void EdgeTable::excludeRectangle (const Rectangle<int>& r)
{
	const Rectangle<int> clipped (r.getIntersection (bounds));

	if (! clipped.isEmpty())
	{
		const int top = clipped.getY() - bounds.getY();
		const int bottom = clipped.getBottom() - bounds.getY();

		//XXX optimise here by shortening the table if it fills top or bottom

		const int rectLine[] = { 4, std::numeric_limits<int>::min(), 255,
								 clipped.getX() << 8, 0,
								 clipped.getRight() << 8, 255,
								 std::numeric_limits<int>::max(), 0 };

		for (int i = top; i < bottom; ++i)
			intersectWithEdgeTableLine (i, rectLine);

		needToCheckEmptinesss = true;
	}
}

void EdgeTable::clipToEdgeTable (const EdgeTable& other)
{
	const Rectangle<int> clipped (other.bounds.getIntersection (bounds));

	if (clipped.isEmpty())
	{
		needToCheckEmptinesss = false;
		bounds.setHeight (0);
	}
	else
	{
		const int top = clipped.getY() - bounds.getY();
		const int bottom = clipped.getBottom() - bounds.getY();

		if (bottom < bounds.getHeight())
			bounds.setHeight (bottom);

		if (clipped.getRight() < bounds.getRight())
			bounds.setRight (clipped.getRight());

		int i = 0;
		for (i = top; --i >= 0;)
			table [lineStrideElements * i] = 0;

		const int* otherLine = other.table + other.lineStrideElements * (clipped.getY() - other.bounds.getY());

		for (i = top; i < bottom; ++i)
		{
			intersectWithEdgeTableLine (i, otherLine);
			otherLine += other.lineStrideElements;
		}

		needToCheckEmptinesss = true;
	}
}

void EdgeTable::clipLineToMask (int x, int y, const uint8* mask, int maskStride, int numPixels)
{
	y -= bounds.getY();

	if (y < 0 || y >= bounds.getHeight())
		return;

	needToCheckEmptinesss = true;

	if (numPixels <= 0)
	{
		table [lineStrideElements * y] = 0;
		return;
	}

  std::vector<int> buffer(numPixels * 2 + 4);
	int* tempLine = &buffer[0];
	int destIndex = 0, lastLevel = 0;

	while (--numPixels >= 0)
	{
		const int alpha = *mask;
		mask += maskStride;

		if (alpha != lastLevel)
		{
			tempLine[++destIndex] = (x << 8);
			tempLine[++destIndex] = alpha;
			lastLevel = alpha;
		}

		++x;
	}

	if (lastLevel > 0)
	{
		tempLine[++destIndex] = (x << 8);
		tempLine[++destIndex] = 0;
	}

	tempLine[0] = destIndex >> 1;

	intersectWithEdgeTableLine (y, tempLine);
}

bool EdgeTable::isEmpty() noexcept
{
	if (needToCheckEmptinesss)
	{
		needToCheckEmptinesss = false;
		int* t = table;

		for (int i = bounds.getHeight(); --i >= 0;)
		{
			if (t[0] > 1)
				return false;

			t += lineStrideElements;
		}

		bounds.setHeight (0);
	}

	return bounds.getHeight() == 0;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_EdgeTable.cpp ***/


/*** Start of inlined file: juce_FillType.cpp ***/
BEGIN_JUCE_NAMESPACE

FillType::FillType() noexcept
	: colour (0xff000000)
{
}

FillType::FillType (const Colour& colour_) noexcept
	: colour (colour_)
{
}

FillType::FillType (const ColourGradient& gradient_)
	: colour (0xff000000), gradient (new ColourGradient (gradient_))
{
}

FillType::FillType (const Image& image_, const AffineTransform& transform_) noexcept
	: colour (0xff000000), image (image_), transform (transform_)
{
}

FillType::FillType (const FillType& other)
	: colour (other.colour),
	  gradient (other.gradient != nullptr ? new ColourGradient (*other.gradient) : 0),
	  image (other.image), transform (other.transform)
{
}

FillType& FillType::operator= (const FillType& other)
{
	if (this != &other)
	{
		colour = other.colour;
		gradient = (other.gradient != nullptr ? new ColourGradient (*other.gradient) : 0);
		image = other.image;
		transform = other.transform;
	}

	return *this;
}

FillType::~FillType() noexcept
{
}

bool FillType::operator== (const FillType& other) const
{
	return colour == other.colour && image == other.image
			&& transform == other.transform
			&& (gradient == other.gradient
				 || (gradient != nullptr && other.gradient != nullptr && *gradient == *other.gradient));
}

bool FillType::operator!= (const FillType& other) const
{
	return ! operator== (other);
}

void FillType::setColour (const Colour& newColour) noexcept
{
	gradient = nullptr;
	image = Image::null;
	colour = newColour;
}

void FillType::setGradient (const ColourGradient& newGradient)
{
	if (gradient != nullptr)
	{
		*gradient = newGradient;
	}
	else
	{
		image = Image::null;
		gradient = new ColourGradient (newGradient);
		colour = Colours::black;
	}
}

void FillType::setTiledImage (const Image& image_, const AffineTransform& transform_) noexcept
{
	gradient = nullptr;
	image = image_;
	transform = transform_;
	colour = Colours::black;
}

void FillType::setOpacity (const float newOpacity) noexcept
{
	colour = colour.withAlpha (newOpacity);
}

bool FillType::isInvisible() const noexcept
{
	return colour.isTransparent() || (gradient != nullptr && gradient->isInvisible());
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_FillType.cpp ***/


/*** Start of inlined file: juce_GraphicsContext.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace
{
	template <typename Type>
	bool areCoordsSensibleNumbers (Type x, Type y, Type w, Type h)
	{
		const int maxVal = 0x3fffffff;

		return (int) x >= -maxVal && (int) x <= maxVal
			&& (int) y >= -maxVal && (int) y <= maxVal
			&& (int) w >= -maxVal && (int) w <= maxVal
			&& (int) h >= -maxVal && (int) h <= maxVal;
	}
}

LowLevelGraphicsContext::LowLevelGraphicsContext()
{
}

LowLevelGraphicsContext::~LowLevelGraphicsContext()
{
}

Graphics::Graphics (const Image& imageToDrawOnto)
	: context (imageToDrawOnto.createLowLevelContext()),
	  contextToDelete (context),
	  saveStatePending (false)
{
}

Graphics::Graphics (LowLevelGraphicsContext* const internalContext) noexcept
	: context (internalContext),
	  saveStatePending (false)
{
}

Graphics::~Graphics()
{
}

void Graphics::resetToDefaultState()
{
	saveStateIfPending();
	context->setFill (FillType());
	context->setFont (Font());
	context->setInterpolationQuality (Graphics::mediumResamplingQuality);
}

bool Graphics::isVectorDevice() const
{
	return context->isVectorDevice();
}

bool Graphics::reduceClipRegion (const Rectangle<int>& area)
{
	saveStateIfPending();
	return context->clipToRectangle (area);
}

bool Graphics::reduceClipRegion (const int x, const int y, const int w, const int h)
{
	return reduceClipRegion (Rectangle<int> (x, y, w, h));
}

bool Graphics::reduceClipRegion (const RectangleList& clipRegion)
{
	saveStateIfPending();
	return context->clipToRectangleList (clipRegion);
}

bool Graphics::reduceClipRegion (const Path& path, const AffineTransform& transform)
{
	saveStateIfPending();
	context->clipToPath (path, transform);
	return ! context->isClipEmpty();
}

bool Graphics::reduceClipRegion (const Image& image, const AffineTransform& transform)
{
	saveStateIfPending();
	context->clipToImageAlpha (image, transform);
	return ! context->isClipEmpty();
}

void Graphics::excludeClipRegion (const Rectangle<int>& rectangleToExclude)
{
	saveStateIfPending();
	context->excludeClipRectangle (rectangleToExclude);
}

bool Graphics::isClipEmpty() const
{
	return context->isClipEmpty();
}

const Rectangle<int> Graphics::getClipBounds() const
{
	return context->getClipBounds();
}

void Graphics::saveState()
{
	saveStateIfPending();
	saveStatePending = true;
}

void Graphics::restoreState()
{
	if (saveStatePending)
		saveStatePending = false;
	else
		context->restoreState();
}

void Graphics::saveStateIfPending()
{
	if (saveStatePending)
	{
		saveStatePending = false;
		context->saveState();
	}
}

void Graphics::setOrigin (const int newOriginX, const int newOriginY)
{
	saveStateIfPending();
	context->setOrigin (newOriginX, newOriginY);
}

void Graphics::addTransform (const AffineTransform& transform)
{
	saveStateIfPending();
	context->addTransform (transform);
}

bool Graphics::clipRegionIntersects (const Rectangle<int>& area) const
{
	return context->clipRegionIntersects (area);
}

void Graphics::beginTransparencyLayer (float layerOpacity)
{
	saveStateIfPending();
	context->beginTransparencyLayer (layerOpacity);
}

void Graphics::endTransparencyLayer()
{
	context->endTransparencyLayer();
}

void Graphics::setColour (const Colour& newColour)
{
	saveStateIfPending();
	context->setFill (newColour);
}

void Graphics::setOpacity (const float newOpacity)
{
	saveStateIfPending();
	context->setOpacity (newOpacity);
}

void Graphics::setGradientFill (const ColourGradient& gradient)
{
	setFillType (gradient);
}

void Graphics::setTiledImageFill (const Image& imageToUse, const int anchorX, const int anchorY, const float opacity)
{
	saveStateIfPending();
	context->setFill (FillType (imageToUse, AffineTransform::translation ((float) anchorX, (float) anchorY)));
	context->setOpacity (opacity);
}

void Graphics::setFillType (const FillType& newFill)
{
	saveStateIfPending();
	context->setFill (newFill);
}

void Graphics::setFont (const Font& newFont)
{
	saveStateIfPending();
	context->setFont (newFont);
}

void Graphics::setFont (const float newFontHeight, const int newFontStyleFlags)
{
	saveStateIfPending();
	Font f (context->getFont());
	f.setSizeAndStyle (newFontHeight, newFontStyleFlags, 1.0f, 0);
	context->setFont (f);
}

Font Graphics::getCurrentFont() const
{
	return context->getFont();
}

void Graphics::drawSingleLineText (const String& text, const int startX, const int baselineY) const
{
	if (text.isNotEmpty()
		 && startX < context->getClipBounds().getRight())
	{
		GlyphArrangement arr;
		arr.addLineOfText (context->getFont(), text, (float) startX, (float) baselineY);
		arr.draw (*this);
	}
}

void Graphics::drawTextAsPath (const String& text, const AffineTransform& transform) const
{
	if (text.isNotEmpty())
	{
		GlyphArrangement arr;
		arr.addLineOfText (context->getFont(), text, 0.0f, 0.0f);
		arr.draw (*this, transform);
	}
}

void Graphics::drawMultiLineText (const String& text, const int startX, const int baselineY, const int maximumLineWidth) const
{
	if (text.isNotEmpty()
		 && startX < context->getClipBounds().getRight())
	{
		GlyphArrangement arr;
		arr.addJustifiedText (context->getFont(), text,
							  (float) startX, (float) baselineY, (float) maximumLineWidth,
							  Justification::left);
		arr.draw (*this);
	}
}

void Graphics::drawText (const String& text,
						 const int x, const int y, const int width, const int height,
						 const Justification& justificationType,
						 const bool useEllipsesIfTooBig) const
{
	if (text.isNotEmpty() && context->clipRegionIntersects (Rectangle<int> (x, y, width, height)))
	{
		GlyphArrangement arr;

		arr.addCurtailedLineOfText (context->getFont(), text,
									0.0f, 0.0f, (float) width,
									useEllipsesIfTooBig);

		arr.justifyGlyphs (0, arr.getNumGlyphs(),
						   (float) x, (float) y, (float) width, (float) height,
						   justificationType);
		arr.draw (*this);
	}
}

void Graphics::drawFittedText (const String& text,
							   const int x, const int y, const int width, const int height,
							   const Justification& justification,
							   const int maximumNumberOfLines,
							   const float minimumHorizontalScale) const
{
	if (text.isNotEmpty()
		 && width > 0 && height > 0
		 && context->clipRegionIntersects (Rectangle<int> (x, y, width, height)))
	{
		GlyphArrangement arr;

		arr.addFittedText (context->getFont(), text,
						   (float) x, (float) y, (float) width, (float) height,
						   justification,
						   maximumNumberOfLines,
						   minimumHorizontalScale);

		arr.draw (*this);
	}
}

void Graphics::fillRect (int x, int y, int width, int height) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	context->fillRect (Rectangle<int> (x, y, width, height), false);
}

void Graphics::fillRect (const Rectangle<int>& r) const
{
	context->fillRect (r, false);
}

void Graphics::fillRect (const float x, const float y, const float width, const float height) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	Path p;
	p.addRectangle (x, y, width, height);
	fillPath (p);
}

void Graphics::setPixel (int x, int y) const
{
	context->fillRect (Rectangle<int> (x, y, 1, 1), false);
}

void Graphics::fillAll() const
{
	fillRect (context->getClipBounds());
}

void Graphics::fillAll (const Colour& colourToUse) const
{
	if (! colourToUse.isTransparent())
	{
		const Rectangle<int> clip (context->getClipBounds());

		context->saveState();
		context->setFill (colourToUse);
		context->fillRect (clip, false);
		context->restoreState();
	}
}

void Graphics::fillPath (const Path& path, const AffineTransform& transform) const
{
	if ((! context->isClipEmpty()) && ! path.isEmpty())
		context->fillPath (path, transform);
}

void Graphics::strokePath (const Path& path,
						   const PathStrokeType& strokeType,
						   const AffineTransform& transform) const
{
	Path stroke;
	strokeType.createStrokedPath (stroke, path, transform, context->getScaleFactor());
	fillPath (stroke);
}

void Graphics::drawRect (const int x, const int y, const int width, const int height,
						 const int lineThickness) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	context->fillRect (Rectangle<int> (x, y, width, lineThickness), false);
	context->fillRect (Rectangle<int> (x, y + lineThickness, lineThickness, height - lineThickness * 2), false);
	context->fillRect (Rectangle<int> (x + width - lineThickness, y + lineThickness, lineThickness, height - lineThickness * 2), false);
	context->fillRect (Rectangle<int> (x, y + height - lineThickness, width, lineThickness), false);
}

void Graphics::drawRect (const float x, const float y, const float width, const float height, const float lineThickness) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	Path p;
	p.addRectangle (x, y, width, lineThickness);
	p.addRectangle (x, y + lineThickness, lineThickness, height - lineThickness * 2.0f);
	p.addRectangle (x + width - lineThickness, y + lineThickness, lineThickness, height - lineThickness * 2.0f);
	p.addRectangle (x, y + height - lineThickness, width, lineThickness);
	fillPath (p);
}

void Graphics::drawRect (const Rectangle<int>& r, const int lineThickness) const
{
	drawRect (r.getX(), r.getY(), r.getWidth(), r.getHeight(), lineThickness);
}

void Graphics::drawBevel (const int x, const int y, const int width, const int height,
						  const int bevelThickness, const Colour& topLeftColour, const Colour& bottomRightColour,
						  const bool useGradient, const bool sharpEdgeOnOutside) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	if (clipRegionIntersects (Rectangle<int> (x, y, width, height)))
	{
		context->saveState();

		const float oldOpacity = 1.0f;//xxx state->colour.getFloatAlpha();
		const float ramp = oldOpacity / bevelThickness;

		for (int i = bevelThickness; --i >= 0;)
		{
			const float op = useGradient ? ramp * (sharpEdgeOnOutside ? bevelThickness - i : i)
										 : oldOpacity;

			context->setFill (topLeftColour.withMultipliedAlpha (op));
			context->fillRect (Rectangle<int> (x + i, y + i, width - i * 2, 1), false);
			context->setFill (topLeftColour.withMultipliedAlpha (op * 0.75f));
			context->fillRect (Rectangle<int> (x + i, y + i + 1, 1, height - i * 2 - 2), false);
			context->setFill (bottomRightColour.withMultipliedAlpha (op));
			context->fillRect (Rectangle<int> (x + i, y + height - i - 1, width - i * 2, 1), false);
			context->setFill (bottomRightColour.withMultipliedAlpha (op  * 0.75f));
			context->fillRect (Rectangle<int> (x + width - i - 1, y + i + 1, 1, height - i * 2 - 2), false);
		}

		context->restoreState();
	}
}

void Graphics::fillEllipse (const float x, const float y, const float width, const float height) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	Path p;
	p.addEllipse (x, y, width, height);
	fillPath (p);
}

void Graphics::drawEllipse (const float x, const float y, const float width, const float height,
							const float lineThickness) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	Path p;
	p.addEllipse (x, y, width, height);
	strokePath (p, PathStrokeType (lineThickness));
}

void Graphics::fillRoundedRectangle (const float x, const float y, const float width, const float height, const float cornerSize) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	Path p;
	p.addRoundedRectangle (x, y, width, height, cornerSize);
	fillPath (p);
}

void Graphics::fillRoundedRectangle (const Rectangle<float>& r, const float cornerSize) const
{
	fillRoundedRectangle (r.getX(), r.getY(), r.getWidth(), r.getHeight(), cornerSize);
}

void Graphics::drawRoundedRectangle (const float x, const float y, const float width, const float height,
									 const float cornerSize, const float lineThickness) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (x, y, width, height));

	Path p;
	p.addRoundedRectangle (x, y, width, height, cornerSize);
	strokePath (p, PathStrokeType (lineThickness));
}

void Graphics::drawRoundedRectangle (const Rectangle<float>& r, const float cornerSize, const float lineThickness) const
{
	drawRoundedRectangle (r.getX(), r.getY(), r.getWidth(), r.getHeight(), cornerSize, lineThickness);
}

void Graphics::drawArrow (const Line<float>& line, const float lineThickness, const float arrowheadWidth, const float arrowheadLength) const
{
	Path p;
	p.addArrow (line, lineThickness, arrowheadWidth, arrowheadLength);
	fillPath (p);
}

void Graphics::fillCheckerBoard (const Rectangle<int>& area,
								 const int checkWidth, const int checkHeight,
								 const Colour& colour1, const Colour& colour2) const
{
	jassert (checkWidth > 0 && checkHeight > 0); // can't be zero or less!

	if (checkWidth > 0 && checkHeight > 0)
	{
		context->saveState();

		if (colour1 == colour2)
		{
			context->setFill (colour1);
			context->fillRect (area, false);
		}
		else
		{
			const Rectangle<int> clipped (context->getClipBounds().getIntersection (area));

			if (! clipped.isEmpty())
			{
				context->clipToRectangle (clipped);

				const int checkNumX = (clipped.getX() - area.getX()) / checkWidth;
				const int checkNumY = (clipped.getY() - area.getY()) / checkHeight;
				const int startX = area.getX() + checkNumX * checkWidth;
				const int startY = area.getY() + checkNumY * checkHeight;
				const int right  = clipped.getRight();
				const int bottom = clipped.getBottom();

				for (int i = 0; i < 2; ++i)
				{
					context->setFill (i == ((checkNumX ^ checkNumY) & 1) ? colour1 : colour2);

					int cy = i;
					for (int y = startY; y < bottom; y += checkHeight)
						for (int x = startX + (cy++ & 1) * checkWidth; x < right; x += checkWidth * 2)
							context->fillRect (Rectangle<int> (x, y, checkWidth, checkHeight), false);
				}
			}
		}

		context->restoreState();
	}
}

void Graphics::drawVerticalLine (const int x, float top, float bottom) const
{
	context->drawVerticalLine (x, top, bottom);
}

void Graphics::drawHorizontalLine (const int y, float left, float right) const
{
	context->drawHorizontalLine (y, left, right);
}

void Graphics::drawLine (const float x1, const float y1, const float x2, const float y2) const
{
	context->drawLine (Line<float> (x1, y1, x2, y2));
}

void Graphics::drawLine (const Line<float>& line) const
{
	context->drawLine (line);
}

void Graphics::drawLine (const float x1, const float y1, const float x2, const float y2, const float lineThickness) const
{
	drawLine (Line<float> (x1, y1, x2, y2), lineThickness);
}

void Graphics::drawLine (const Line<float>& line, const float lineThickness) const
{
	Path p;
	p.addLineSegment (line, lineThickness);
	fillPath (p);
}

void Graphics::drawDashedLine (const Line<float>& line, const float* const dashLengths,
							   const int numDashLengths, const float lineThickness, int n) const
{
	jassert (n >= 0 && n < numDashLengths); // your start index must be valid!

	const Point<double> delta ((line.getEnd() - line.getStart()).toDouble());
	const double totalLen = delta.getDistanceFromOrigin();

	if (totalLen >= 0.1)
	{
		const double onePixAlpha = 1.0 / totalLen;

		for (double alpha = 0.0; alpha < 1.0;)
		{
			jassert (dashLengths[n] > 0); // can't have zero-length dashes!

			const double lastAlpha = alpha;
			alpha = jmin (1.0, alpha + dashLengths [n] * onePixAlpha);
			n = (n + 1) % numDashLengths;

			if ((n & 1) != 0)
			{
				const Line<float> segment (line.getStart() + (delta * lastAlpha).toFloat(),
										   line.getStart() + (delta * alpha).toFloat());

				if (lineThickness != 1.0f)
					drawLine (segment, lineThickness);
				else
					context->drawLine (segment);
			}
		}
	}
}

void Graphics::setImageResamplingQuality (const Graphics::ResamplingQuality newQuality)
{
	saveStateIfPending();
	context->setInterpolationQuality (newQuality);
}

void Graphics::drawImageAt (const Image& imageToDraw,
							const int topLeftX, const int topLeftY,
							const bool fillAlphaChannelWithCurrentBrush) const
{
	const int imageW = imageToDraw.getWidth();
	const int imageH = imageToDraw.getHeight();

	drawImage (imageToDraw,
			   topLeftX, topLeftY, imageW, imageH,
			   0, 0, imageW, imageH,
			   fillAlphaChannelWithCurrentBrush);
}

void Graphics::drawImageWithin (const Image& imageToDraw,
								const int destX, const int destY,
								const int destW, const int destH,
								const RectanglePlacement& placementWithinTarget,
								const bool fillAlphaChannelWithCurrentBrush) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (destX, destY, destW, destH));

	if (imageToDraw.isValid())
	{
		const int imageW = imageToDraw.getWidth();
		const int imageH = imageToDraw.getHeight();

		if (imageW > 0 && imageH > 0)
		{
			double newX = 0.0, newY = 0.0;
			double newW = imageW;
			double newH = imageH;

			placementWithinTarget.applyTo (newX, newY, newW, newH,
										   destX, destY, destW, destH);

			if (newW > 0 && newH > 0)
			{
				drawImage (imageToDraw,
						   roundToInt (newX), roundToInt (newY),
						   roundToInt (newW), roundToInt (newH),
						   0, 0, imageW, imageH,
						   fillAlphaChannelWithCurrentBrush);
			}
		}
	}
}

void Graphics::drawImage (const Image& imageToDraw,
						  int dx, int dy, int dw, int dh,
						  int sx, int sy, int sw, int sh,
						  const bool fillAlphaChannelWithCurrentBrush) const
{
	// passing in a silly number can cause maths problems in rendering!
	jassert (areCoordsSensibleNumbers (dx, dy, dw, dh));
	jassert (areCoordsSensibleNumbers (sx, sy, sw, sh));

	if (imageToDraw.isValid() && context->clipRegionIntersects  (Rectangle<int> (dx, dy, dw, dh)))
	{
		drawImageTransformed (imageToDraw.getClippedImage (Rectangle<int> (sx, sy, sw, sh)),
							  AffineTransform::scale (dw / (float) sw, dh / (float) sh)
											  .translated ((float) dx, (float) dy),
							  fillAlphaChannelWithCurrentBrush);
	}
}

void Graphics::drawImageTransformed (const Image& imageToDraw,
									 const AffineTransform& transform,
									 const bool fillAlphaChannelWithCurrentBrush) const
{
	if (imageToDraw.isValid() && ! context->isClipEmpty())
	{
		if (fillAlphaChannelWithCurrentBrush)
		{
			context->saveState();
			context->clipToImageAlpha (imageToDraw, transform);
			fillAll();
			context->restoreState();
		}
		else
		{
			context->drawImage (imageToDraw, transform, false);
		}
	}
}

Graphics::ScopedSaveState::ScopedSaveState (Graphics& g)
	: context (g)
{
	context.saveState();
}

Graphics::ScopedSaveState::~ScopedSaveState()
{
	context.restoreState();
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_GraphicsContext.cpp ***/


/*** Start of inlined file: juce_Justification.cpp ***/
BEGIN_JUCE_NAMESPACE

Justification::Justification (const Justification& other) noexcept
	: flags (other.flags)
{
}

Justification& Justification::operator= (const Justification& other) noexcept
{
	flags = other.flags;
	return *this;
}

int Justification::getOnlyVerticalFlags() const noexcept
{
	return flags & (top | bottom | verticallyCentred);
}

int Justification::getOnlyHorizontalFlags() const noexcept
{
	return flags & (left | right | horizontallyCentred | horizontallyJustified);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Justification.cpp ***/


/*** Start of inlined file: juce_LowLevelGraphicsSoftwareRenderer.cpp ***/
BEGIN_JUCE_NAMESPACE

#if JUCE_MSVC
 #pragma warning (push)
 #pragma warning (disable: 4127) // "expression is constant" warning

 #if JUCE_DEBUG
  #pragma optimize ("t", on)  // optimise just this file, to avoid sluggish graphics when debugging
  #pragma warning (disable: 4714) // warning about forcedinline methods not being inlined
 #endif
#endif

namespace SoftwareRendererClasses
{

template <class PixelType, bool replaceExisting = false>
class SolidColourEdgeTableRenderer
{
public:
	SolidColourEdgeTableRenderer (const Image::BitmapData& data_, const PixelARGB& colour)
		: data (data_),
		  sourceColour (colour)
	{
		if (sizeof (PixelType) == 3)
		{
			areRGBComponentsEqual = sourceColour.getRed() == sourceColour.getGreen()
										&& sourceColour.getGreen() == sourceColour.getBlue();
			filler[0].set (sourceColour);
			filler[1].set (sourceColour);
			filler[2].set (sourceColour);
			filler[3].set (sourceColour);
		}
	}

	forcedinline void setEdgeTableYPos (const int y) noexcept
	{
		linePixels = (PixelType*) data.getLinePointer (y);
	}

	forcedinline void handleEdgeTablePixel (const int x, const int alphaLevel) const noexcept
	{
		if (replaceExisting)
			linePixels[x].set (sourceColour);
		else
			linePixels[x].blend (sourceColour, alphaLevel);
	}

	forcedinline void handleEdgeTablePixelFull (const int x) const noexcept
	{
		if (replaceExisting)
			linePixels[x].set (sourceColour);
		else
			linePixels[x].blend (sourceColour);
	}

	forcedinline void handleEdgeTableLine (const int x, const int width, const int alphaLevel) const noexcept
	{
		PixelARGB p (sourceColour);
		p.multiplyAlpha (alphaLevel);

		PixelType* dest = linePixels + x;

		if (replaceExisting || p.getAlpha() >= 0xff)
			replaceLine (dest, p, width);
		else
			blendLine (dest, p, width);
	}

	forcedinline void handleEdgeTableLineFull (const int x, const int width) const noexcept
	{
		PixelType* dest = linePixels + x;

		if (replaceExisting || sourceColour.getAlpha() >= 0xff)
			replaceLine (dest, sourceColour, width);
		else
			blendLine (dest, sourceColour, width);
	}

private:
	const Image::BitmapData& data;
	PixelType* linePixels;
	PixelARGB sourceColour;
	PixelRGB filler [4];
	bool areRGBComponentsEqual;

	inline void blendLine (PixelType* dest, const PixelARGB& colour, int width) const noexcept
	{
		do
		{
			dest->blend (colour);
			++dest;
		} while (--width > 0);
	}

	forcedinline void replaceLine (PixelRGB* dest, const PixelARGB& colour, int width) const noexcept
	{
		if (areRGBComponentsEqual)  // if all the component values are the same, we can cheat..
		{
			memset (dest, colour.getRed(), width * 3);
		}
		else
		{
			if (width >> 5)
			{
				const int* const intFiller = reinterpret_cast<const int*> (filler);

				while (width > 8 && (((pointer_sized_int) dest) & 7) != 0)
				{
					dest->set (colour);
					++dest;
					--width;
				}

				while (width > 4)
				{
					int* d = reinterpret_cast<int*> (dest);
					*d++ = intFiller[0];
					*d++ = intFiller[1];
					*d++ = intFiller[2];
					dest = reinterpret_cast<PixelRGB*> (d);
					width -= 4;
				}
			}

			while (--width >= 0)
			{
				dest->set (colour);
				++dest;
			}
		}
	}

	forcedinline void replaceLine (PixelAlpha* const dest, const PixelARGB& colour, int const width) const noexcept
	{
		memset (dest, colour.getAlpha(), width);
	}

	forcedinline void replaceLine (PixelARGB* dest, const PixelARGB& colour, int width) const noexcept
	{
		do
		{
			dest->set (colour);
			++dest;

		} while (--width > 0);
	}

	JUCE_DECLARE_NON_COPYABLE (SolidColourEdgeTableRenderer);
};

class LinearGradientPixelGenerator
{
public:
	LinearGradientPixelGenerator (const ColourGradient& gradient, const AffineTransform& transform,
								  const PixelARGB* const lookupTable_, const int numEntries_)
		: lookupTable (lookupTable_),
		  numEntries (numEntries_)
	{
		jassert (numEntries_ >= 0);
		Point<float> p1 (gradient.point1);
		Point<float> p2 (gradient.point2);

		if (! transform.isIdentity())
		{
			const Line<float> l (p2, p1);
			Point<float> p3 = l.getPointAlongLine (0.0f, 100.0f);

			p1.applyTransform (transform);
			p2.applyTransform (transform);
			p3.applyTransform (transform);

			p2 = Line<float> (p2, p3).findNearestPointTo (p1);
		}

		vertical = std::abs (p1.getX() - p2.getX()) < 0.001f;
		horizontal = std::abs (p1.getY() - p2.getY()) < 0.001f;

		if (vertical)
		{
			scale = roundToInt ((numEntries << (int) numScaleBits) / (double) (p2.getY() - p1.getY()));
			start = roundToInt (p1.getY() * scale);
		}
		else if (horizontal)
		{
			scale = roundToInt ((numEntries << (int) numScaleBits) / (double) (p2.getX() - p1.getX()));
			start = roundToInt (p1.getX() * scale);
		}
		else
		{
			grad = (p2.getY() - p1.getY()) / (double) (p1.getX() - p2.getX());
			yTerm = p1.getY() - p1.getX() / grad;
			scale = roundToInt ((numEntries << (int) numScaleBits) / (yTerm * grad - (p2.getY() * grad - p2.getX())));
			grad *= scale;
		}
	}

	forcedinline void setY (const int y) noexcept
	{
		if (vertical)
			linePix = lookupTable [jlimit (0, numEntries, (y * scale - start) >> (int) numScaleBits)];
		else if (! horizontal)
			start = roundToInt ((y - yTerm) * grad);
	}

	inline const PixelARGB getPixel (const int x) const noexcept
	{
		return vertical ? linePix
						: lookupTable [jlimit (0, numEntries, (x * scale - start) >> (int) numScaleBits)];
	}

private:
	const PixelARGB* const lookupTable;
	const int numEntries;
	PixelARGB linePix;
	int start, scale;
	double grad, yTerm;
	bool vertical, horizontal;
	enum { numScaleBits = 12 };

	JUCE_DECLARE_NON_COPYABLE (LinearGradientPixelGenerator);
};

class RadialGradientPixelGenerator
{
public:
	RadialGradientPixelGenerator (const ColourGradient& gradient, const AffineTransform&,
								  const PixelARGB* const lookupTable_, const int numEntries_)
		: lookupTable (lookupTable_),
		  numEntries (numEntries_),
		  gx1 (gradient.point1.getX()),
		  gy1 (gradient.point1.getY())
	{
		jassert (numEntries_ >= 0);
		const Point<float> diff (gradient.point1 - gradient.point2);
		maxDist = diff.getX() * diff.getX() + diff.getY() * diff.getY();
		invScale = numEntries / std::sqrt (maxDist);
		jassert (roundToInt (std::sqrt (maxDist) * invScale) <= numEntries);
	}

	forcedinline void setY (const int y) noexcept
	{
		dy = y - gy1;
		dy *= dy;
	}

	inline const PixelARGB getPixel (const int px) const noexcept
	{
		double x = px - gx1;
		x *= x;
		x += dy;

		return lookupTable [x >= maxDist ? numEntries : roundToInt (std::sqrt (x) * invScale)];
	}

protected:
	const PixelARGB* const lookupTable;
	const int numEntries;
	const double gx1, gy1;
	double maxDist, invScale, dy;

	JUCE_DECLARE_NON_COPYABLE (RadialGradientPixelGenerator);
};

class TransformedRadialGradientPixelGenerator   : public RadialGradientPixelGenerator
{
public:
	TransformedRadialGradientPixelGenerator (const ColourGradient& gradient, const AffineTransform& transform,
											 const PixelARGB* const lookupTable_, const int numEntries_)
		: RadialGradientPixelGenerator (gradient, transform, lookupTable_, numEntries_),
		  inverseTransform (transform.inverted())
	{
		tM10 = inverseTransform.mat10;
		tM00 = inverseTransform.mat00;
	}

	forcedinline void setY (const int y) noexcept
	{
		lineYM01 = inverseTransform.mat01 * y + inverseTransform.mat02 - gx1;
		lineYM11 = inverseTransform.mat11 * y + inverseTransform.mat12 - gy1;
	}

	inline const PixelARGB getPixel (const int px) const noexcept
	{
		double x = px;
		const double y = tM10 * x + lineYM11;
		x = tM00 * x + lineYM01;
		x *= x;
		x += y * y;

		if (x >= maxDist)
			return lookupTable [numEntries];
		else
			return lookupTable [jmin (numEntries, roundToInt (std::sqrt (x) * invScale))];
	}

private:
	double tM10, tM00, lineYM01, lineYM11;
	const AffineTransform inverseTransform;

	JUCE_DECLARE_NON_COPYABLE (TransformedRadialGradientPixelGenerator);
};

template <class PixelType, class GradientType>
class GradientEdgeTableRenderer  : public GradientType
{
public:
	GradientEdgeTableRenderer (const Image::BitmapData& destData_, const ColourGradient& gradient, const AffineTransform& transform,
							   const PixelARGB* const lookupTable_, const int numEntries_)
		: GradientType (gradient, transform, lookupTable_, numEntries_ - 1),
		  destData (destData_)
	{
	}

	forcedinline void setEdgeTableYPos (const int y) noexcept
	{
		linePixels = (PixelType*) destData.getLinePointer (y);
		GradientType::setY (y);
	}

	forcedinline void handleEdgeTablePixel (const int x, const int alphaLevel) const noexcept
	{
		linePixels[x].blend (GradientType::getPixel (x), alphaLevel);
	}

	forcedinline void handleEdgeTablePixelFull (const int x) const noexcept
	{
		linePixels[x].blend (GradientType::getPixel (x));
	}

	void handleEdgeTableLine (int x, int width, const int alphaLevel) const noexcept
	{
		PixelType* dest = linePixels + x;

		if (alphaLevel < 0xff)
		{
			do
			{
				(dest++)->blend (GradientType::getPixel (x++), alphaLevel);
			} while (--width > 0);
		}
		else
		{
			do
			{
				(dest++)->blend (GradientType::getPixel (x++));
			} while (--width > 0);
		}
	}

	void handleEdgeTableLineFull (int x, int width) const noexcept
	{
		PixelType* dest = linePixels + x;

		do
		{
			(dest++)->blend (GradientType::getPixel (x++));
		} while (--width > 0);
	}

private:
	const Image::BitmapData& destData;
	PixelType* linePixels;

	JUCE_DECLARE_NON_COPYABLE (GradientEdgeTableRenderer);
};

namespace RenderingHelpers
{
	forcedinline int safeModulo (int n, const int divisor) noexcept
	{
		jassert (divisor > 0);
		n %= divisor;
		return (n < 0) ? (n + divisor) : n;
	}
}

template <class DestPixelType, class SrcPixelType, bool repeatPattern>
class ImageFillEdgeTableRenderer
{
public:
	ImageFillEdgeTableRenderer (const Image::BitmapData& destData_,
								const Image::BitmapData& srcData_,
								const int extraAlpha_,
								const int x, const int y)
		: destData (destData_),
		  srcData (srcData_),
		  extraAlpha (extraAlpha_ + 1),
		  xOffset (repeatPattern ? RenderingHelpers::safeModulo (x, srcData_.width) - srcData_.width : x),
		  yOffset (repeatPattern ? RenderingHelpers::safeModulo (y, srcData_.height) - srcData_.height : y)
	{
	}

	forcedinline void setEdgeTableYPos (int y) noexcept
	{
		linePixels = (DestPixelType*) destData.getLinePointer (y);

		y -= yOffset;
		if (repeatPattern)
		{
			jassert (y >= 0);
			y %= srcData.height;
		}

		sourceLineStart = (SrcPixelType*) srcData.getLinePointer (y);
	}

	forcedinline void handleEdgeTablePixel (const int x, int alphaLevel) const noexcept
	{
		alphaLevel = (alphaLevel * extraAlpha) >> 8;

		linePixels[x].blend (sourceLineStart [repeatPattern ? ((x - xOffset) % srcData.width) : (x - xOffset)], alphaLevel);
	}

	forcedinline void handleEdgeTablePixelFull (const int x) const noexcept
	{
		linePixels[x].blend (sourceLineStart [repeatPattern ? ((x - xOffset) % srcData.width) : (x - xOffset)], extraAlpha);
	}

	void handleEdgeTableLine (int x, int width, int alphaLevel) const noexcept
	{
		DestPixelType* dest = linePixels + x;
		alphaLevel = (alphaLevel * extraAlpha) >> 8;
		x -= xOffset;

		jassert (repeatPattern || (x >= 0 && x + width <= srcData.width));

		if (alphaLevel < 0xfe)
		{
			do
			{
				dest++ ->blend (sourceLineStart [repeatPattern ? (x++ % srcData.width) : x++], alphaLevel);
			} while (--width > 0);
		}
		else
		{
			if (repeatPattern)
			{
				do
				{
					dest++ ->blend (sourceLineStart [x++ % srcData.width]);
				} while (--width > 0);
			}
			else
			{
				copyRow (dest, sourceLineStart + x, width);
			}
		}
	}

	void handleEdgeTableLineFull (int x, int width) const noexcept
	{
		DestPixelType* dest = linePixels + x;
		x -= xOffset;

		jassert (repeatPattern || (x >= 0 && x + width <= srcData.width));

		if (extraAlpha < 0xfe)
		{
			do
			{
				dest++ ->blend (sourceLineStart [repeatPattern ? (x++ % srcData.width) : x++], extraAlpha);
			} while (--width > 0);
		}
		else
		{
			if (repeatPattern)
			{
				do
				{
					dest++ ->blend (sourceLineStart [x++ % srcData.width]);
				} while (--width > 0);
			}
			else
			{
				copyRow (dest, sourceLineStart + x, width);
			}
		}
	}

	void clipEdgeTableLine (EdgeTable& et, int x, int y, int width)
	{
		jassert (x - xOffset >= 0 && x + width - xOffset <= srcData.width);
		SrcPixelType* s = (SrcPixelType*) srcData.getLinePointer (y - yOffset);
		uint8* mask = (uint8*) (s + x - xOffset);

		if (sizeof (SrcPixelType) == sizeof (PixelARGB))
			mask += PixelARGB::indexA;

		et.clipLineToMask (x, y, mask, sizeof (SrcPixelType), width);
	}

private:
	const Image::BitmapData& destData;
	const Image::BitmapData& srcData;
	const int extraAlpha, xOffset, yOffset;
	DestPixelType* linePixels;
	SrcPixelType* sourceLineStart;

	template <class PixelType1, class PixelType2>
	static forcedinline void copyRow (PixelType1* dest, PixelType2* src, int width) noexcept
	{
		do
		{
			dest++ ->blend (*src++);
		} while (--width > 0);
	}

	static forcedinline void copyRow (PixelRGB* dest, PixelRGB* src, int width) noexcept
	{
		memcpy (dest, src, width * sizeof (PixelRGB));
	}

	JUCE_DECLARE_NON_COPYABLE (ImageFillEdgeTableRenderer);
};

template <class DestPixelType, class SrcPixelType, bool repeatPattern>
class TransformedImageFillEdgeTableRenderer
{
public:
	TransformedImageFillEdgeTableRenderer (const Image::BitmapData& destData_,
										   const Image::BitmapData& srcData_,
										   const AffineTransform& transform,
										   const int extraAlpha_,
										   const bool betterQuality_)
		: interpolator (transform,
						betterQuality_ ? 0.5f : 0.0f,
						betterQuality_ ? -128 : 0),
		  destData (destData_),
		  srcData (srcData_),
		  extraAlpha (extraAlpha_ + 1),
		  betterQuality (betterQuality_),
		  maxX (srcData_.width - 1),
		  maxY (srcData_.height - 1),
		  scratchSize (2048)
	{
		scratchBuffer.malloc (scratchSize);
	}

	forcedinline void setEdgeTableYPos (const int newY) noexcept
	{
		y = newY;
		linePixels = (DestPixelType*) destData.getLinePointer (newY);
	}

	forcedinline void handleEdgeTablePixel (const int x, const int alphaLevel) noexcept
	{
		SrcPixelType p;
		generate (&p, x, 1);

		linePixels[x].blend (p, (alphaLevel * extraAlpha) >> 8);
	}

	forcedinline void handleEdgeTablePixelFull (const int x) noexcept
	{
		SrcPixelType p;
		generate (&p, x, 1);

		linePixels[x].blend (p, extraAlpha);
	}

	void handleEdgeTableLine (const int x, int width, int alphaLevel) noexcept
	{
		if (width > scratchSize)
		{
			scratchSize = width;
			scratchBuffer.malloc (scratchSize);
		}

		SrcPixelType* span = scratchBuffer;
		generate (span, x, width);

		DestPixelType* dest = linePixels + x;
		alphaLevel *= extraAlpha;
		alphaLevel >>= 8;

		if (alphaLevel < 0xfe)
		{
			do
			{
				dest++ ->blend (*span++, alphaLevel);
			} while (--width > 0);
		}
		else
		{
			do
			{
				dest++ ->blend (*span++);
			} while (--width > 0);
		}
	}

	forcedinline void handleEdgeTableLineFull (const int x, int width) noexcept
	{
		handleEdgeTableLine (x, width, 255);
	}

	void clipEdgeTableLine (EdgeTable& et, int x, int y_, int width)
	{
		if (width > scratchSize)
		{
			scratchSize = width;
			scratchBuffer.malloc (scratchSize);
		}

		y = y_;
		generate (scratchBuffer.getData(), x, width);

		et.clipLineToMask (x, y_,
						   reinterpret_cast<uint8*> (scratchBuffer.getData()) + SrcPixelType::indexA,
						   sizeof (SrcPixelType), width);
	}

private:

	template <class PixelType>
	void generate (PixelType* dest, const int x, int numPixels) noexcept
	{
		this->interpolator.setStartOfLine ((float) x, (float) y, numPixels);

		do
		{
			int hiResX, hiResY;
			this->interpolator.next (hiResX, hiResY);

			int loResX = hiResX >> 8;
			int loResY = hiResY >> 8;

			if (repeatPattern)
			{
				loResX = RenderingHelpers::safeModulo (loResX, srcData.width);
				loResY = RenderingHelpers::safeModulo (loResY, srcData.height);
			}

			if (betterQuality)
			{
				if (isPositiveAndBelow (loResX, maxX))
				{
					if (isPositiveAndBelow (loResY, maxY))
					{
						// In the centre of the image..
						render4PixelAverage (dest, this->srcData.getPixelPointer (loResX, loResY),
											 hiResX & 255, hiResY & 255);
						++dest;
						continue;
					}
					else
					{
						// At a top or bottom edge..
						if (! repeatPattern)
						{
							if (loResY < 0)
								render2PixelAverageX (dest, this->srcData.getPixelPointer (loResX, 0), hiResX & 255);
							else
								render2PixelAverageX (dest, this->srcData.getPixelPointer (loResX, maxY), hiResX & 255);

							++dest;
							continue;
						}
					}
				}
				else
				{
					if (isPositiveAndBelow (loResY, maxY))
					{
						// At a left or right hand edge..
						if (! repeatPattern)
						{
							if (loResX < 0)
								render2PixelAverageY (dest, this->srcData.getPixelPointer (0, loResY), hiResY & 255);
							else
								render2PixelAverageY (dest, this->srcData.getPixelPointer (maxX, loResY), hiResY & 255);

							++dest;
							continue;
						}
					}
				}
			}

			if (! repeatPattern)
			{
				if (loResX < 0)	 loResX = 0;
				if (loResY < 0)	 loResY = 0;
				if (loResX > maxX)  loResX = maxX;
				if (loResY > maxY)  loResY = maxY;
			}

			dest->set (*(const PixelType*) this->srcData.getPixelPointer (loResX, loResY));
			++dest;

		} while (--numPixels > 0);
	}

	void render4PixelAverage (PixelARGB* const dest, const uint8* src, const int subPixelX, const int subPixelY) noexcept
	{
		uint32 c[4] = { 256 * 128, 256 * 128, 256 * 128, 256 * 128 };

		uint32 weight = (256 - subPixelX) * (256 - subPixelY);
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];
		c[3] += weight * src[3];

		weight = subPixelX * (256 - subPixelY);
		c[0] += weight * src[4];
		c[1] += weight * src[5];
		c[2] += weight * src[6];
		c[3] += weight * src[7];

		src += this->srcData.lineStride;

		weight = (256 - subPixelX) * subPixelY;
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];
		c[3] += weight * src[3];

		weight = subPixelX * subPixelY;
		c[0] += weight * src[4];
		c[1] += weight * src[5];
		c[2] += weight * src[6];
		c[3] += weight * src[7];

		dest->setARGB ((uint8) (c[PixelARGB::indexA] >> 16),
					   (uint8) (c[PixelARGB::indexR] >> 16),
					   (uint8) (c[PixelARGB::indexG] >> 16),
					   (uint8) (c[PixelARGB::indexB] >> 16));
	}

	void render2PixelAverageX (PixelARGB* const dest, const uint8* src, const int subPixelX) noexcept
	{
		uint32 c[4] = { 128, 128, 128, 128 };

		uint32 weight = 256 - subPixelX;
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];
		c[3] += weight * src[3];

		weight = subPixelX;
		c[0] += weight * src[4];
		c[1] += weight * src[5];
		c[2] += weight * src[6];
		c[3] += weight * src[7];

		dest->setARGB ((uint8) (c[PixelARGB::indexA] >> 8),
					   (uint8) (c[PixelARGB::indexR] >> 8),
					   (uint8) (c[PixelARGB::indexG] >> 8),
					   (uint8) (c[PixelARGB::indexB] >> 8));
	}

	void render2PixelAverageY (PixelARGB* const dest, const uint8* src, const int subPixelY) noexcept
	{
		uint32 c[4] = { 128, 128, 128, 128 };

		uint32 weight = 256 - subPixelY;
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];
		c[3] += weight * src[3];

		src += this->srcData.lineStride;

		weight = subPixelY;
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];
		c[3] += weight * src[3];

		dest->setARGB ((uint8) (c[PixelARGB::indexA] >> 8),
					   (uint8) (c[PixelARGB::indexR] >> 8),
					   (uint8) (c[PixelARGB::indexG] >> 8),
					   (uint8) (c[PixelARGB::indexB] >> 8));
	}

	void render4PixelAverage (PixelRGB* const dest, const uint8* src, const int subPixelX, const int subPixelY) noexcept
	{
		uint32 c[3] = { 256 * 128, 256 * 128, 256 * 128 };

		uint32 weight = (256 - subPixelX) * (256 - subPixelY);
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];

		weight = subPixelX * (256 - subPixelY);
		c[0] += weight * src[3];
		c[1] += weight * src[4];
		c[2] += weight * src[5];

		src += this->srcData.lineStride;

		weight = (256 - subPixelX) * subPixelY;
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];

		weight = subPixelX * subPixelY;
		c[0] += weight * src[3];
		c[1] += weight * src[4];
		c[2] += weight * src[5];

		dest->setARGB ((uint8) 255,
					   (uint8) (c[PixelRGB::indexR] >> 16),
					   (uint8) (c[PixelRGB::indexG] >> 16),
					   (uint8) (c[PixelRGB::indexB] >> 16));
	}

	void render2PixelAverageX (PixelRGB* const dest, const uint8* src, const int subPixelX) noexcept
	{
		uint32 c[3] = { 128, 128, 128 };

		const uint32 weight = 256 - subPixelX;
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];

		c[0] += subPixelX * src[3];
		c[1] += subPixelX * src[4];
		c[2] += subPixelX * src[5];

		dest->setARGB ((uint8) 255,
					   (uint8) (c[PixelRGB::indexR] >> 8),
					   (uint8) (c[PixelRGB::indexG] >> 8),
					   (uint8) (c[PixelRGB::indexB] >> 8));
	}

	void render2PixelAverageY (PixelRGB* const dest, const uint8* src, const int subPixelY) noexcept
	{
		uint32 c[3] = { 128, 128, 128 };

		const uint32 weight = 256 - subPixelY;
		c[0] += weight * src[0];
		c[1] += weight * src[1];
		c[2] += weight * src[2];

		src += this->srcData.lineStride;

		c[0] += subPixelY * src[0];
		c[1] += subPixelY * src[1];
		c[2] += subPixelY * src[2];

		dest->setARGB ((uint8) 255,
					   (uint8) (c[PixelRGB::indexR] >> 8),
					   (uint8) (c[PixelRGB::indexG] >> 8),
					   (uint8) (c[PixelRGB::indexB] >> 8));
	}

	void render4PixelAverage (PixelAlpha* const dest, const uint8* src, const int subPixelX, const int subPixelY) noexcept
	{
		uint32 c = 256 * 128;
		c += src[0] * ((256 - subPixelX) * (256 - subPixelY));
		c += src[1] * (subPixelX * (256 - subPixelY));
		src += this->srcData.lineStride;
		c += src[0] * ((256 - subPixelX) * subPixelY);
		c += src[1] * (subPixelX * subPixelY);

		*((uint8*) dest) = (uint8) (c >> 16);
	}

	void render2PixelAverageX (PixelAlpha* const dest, const uint8* src, const int subPixelX) noexcept
	{
		uint32 c = 128;
		c += src[0] * (256 - subPixelX);
		c += src[1] * subPixelX;
		*((uint8*) dest) = (uint8) (c >> 8);
	}

	void render2PixelAverageY (PixelAlpha* const dest, const uint8* src, const int subPixelY) noexcept
	{
		uint32 c = 128;
		c += src[0] * (256 - subPixelY);
		src += this->srcData.lineStride;
		c += src[0] * subPixelY;
		*((uint8*) dest) = (uint8) (c >> 8);
	}

	class TransformedImageSpanInterpolator
	{
	public:
		TransformedImageSpanInterpolator (const AffineTransform& transform, const float pixelOffset_, const int pixelOffsetInt_) noexcept
			: inverseTransform (transform.inverted()),
			  pixelOffset (pixelOffset_), pixelOffsetInt (pixelOffsetInt_)
		{}

		void setStartOfLine (float x, float y, const int numPixels) noexcept
		{
			jassert (numPixels > 0);

			x += pixelOffset;
			y += pixelOffset;
			float x1 = x, y1 = y;
			x += numPixels;
			inverseTransform.transformPoints (x1, y1, x, y);

			xBresenham.set ((int) (x1 * 256.0f), (int) (x * 256.0f), numPixels, pixelOffsetInt);
			yBresenham.set ((int) (y1 * 256.0f), (int) (y * 256.0f), numPixels, pixelOffsetInt);
		}

		void next (int& x, int& y) noexcept
		{
			x = xBresenham.n;
			xBresenham.stepToNext();
			y = yBresenham.n;
			yBresenham.stepToNext();
		}

	private:
		class BresenhamInterpolator
		{
		public:
			BresenhamInterpolator() noexcept {}

			void set (const int n1, const int n2, const int numSteps_, const int pixelOffsetInt) noexcept
			{
				numSteps = numSteps_;
				step = (n2 - n1) / numSteps;
				remainder = modulo = (n2 - n1) % numSteps;
				n = n1 + pixelOffsetInt;

				if (modulo <= 0)
				{
					modulo += numSteps;
					remainder += numSteps;
					--step;
				}

				modulo -= numSteps;
			}

			forcedinline void stepToNext() noexcept
			{
				modulo += remainder;
				n += step;

				if (modulo > 0)
				{
					modulo -= numSteps;
					++n;
				}
			}

			int n;

		private:
			int numSteps, step, modulo, remainder;
		};

		const AffineTransform inverseTransform;
		BresenhamInterpolator xBresenham, yBresenham;
		const float pixelOffset;
		const int pixelOffsetInt;

		JUCE_DECLARE_NON_COPYABLE (TransformedImageSpanInterpolator);
	};

	TransformedImageSpanInterpolator interpolator;
	const Image::BitmapData& destData;
	const Image::BitmapData& srcData;
	const int extraAlpha;
	const bool betterQuality;
	const int maxX, maxY;
	int y;
	DestPixelType* linePixels;
	HeapBlock <SrcPixelType> scratchBuffer;
	int scratchSize;

	JUCE_DECLARE_NON_COPYABLE (TransformedImageFillEdgeTableRenderer);
};

class ClipRegionBase  : public SingleThreadedReferenceCountedObject
{
public:
	ClipRegionBase() {}
	virtual ~ClipRegionBase() {}

	typedef ReferenceCountedObjectPtr<ClipRegionBase> Ptr;

	virtual Ptr clone() const = 0;
	virtual Ptr applyClipTo (const Ptr& target) const = 0;

	virtual Ptr clipToRectangle (const Rectangle<int>& r) = 0;
	virtual Ptr clipToRectangleList (const RectangleList& r) = 0;
	virtual Ptr excludeClipRectangle (const Rectangle<int>& r) = 0;
	virtual Ptr clipToPath (const Path& p, const AffineTransform& transform) = 0;
	virtual Ptr clipToEdgeTable (const EdgeTable& et) = 0;
	virtual Ptr clipToImageAlpha (const Image& image, const AffineTransform& t, const bool betterQuality) = 0;
	virtual Ptr translated (const Point<int>& delta) = 0;

	virtual bool clipRegionIntersects (const Rectangle<int>& r) const = 0;
	virtual Rectangle<int> getClipBounds() const = 0;

	virtual void fillRectWithColour (Image::BitmapData& destData, const Rectangle<int>& area, const PixelARGB& colour, bool replaceContents) const = 0;
	virtual void fillRectWithColour (Image::BitmapData& destData, const Rectangle<float>& area, const PixelARGB& colour) const = 0;
	virtual void fillAllWithColour (Image::BitmapData& destData, const PixelARGB& colour, bool replaceContents) const = 0;
	virtual void fillAllWithGradient (Image::BitmapData& destData, ColourGradient& gradient, const AffineTransform& transform, bool isIdentity) const = 0;
	virtual void renderImageTransformed (const Image::BitmapData& destData, const Image::BitmapData& srcData, const int alpha, const AffineTransform& t, bool betterQuality, bool tiledFill) const = 0;
	virtual void renderImageUntransformed (const Image::BitmapData& destData, const Image::BitmapData& srcData, const int alpha, int x, int y, bool tiledFill) const = 0;

protected:

	template <class Iterator>
	static void renderImageTransformedInternal (Iterator& iter, const Image::BitmapData& destData, const Image::BitmapData& srcData,
												const int alpha, const AffineTransform& transform, bool betterQuality, bool tiledFill)
	{
		switch (destData.pixelFormat)
		{
		case Image::ARGB:
			switch (srcData.pixelFormat)
			{
			case Image::ARGB:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelARGB, PixelARGB, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelARGB, PixelARGB, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			case Image::RGB:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelARGB, PixelRGB, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelARGB, PixelRGB, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			default:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelARGB, PixelAlpha, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelARGB, PixelAlpha, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			}
			break;

		case Image::RGB:
			switch (srcData.pixelFormat)
			{
			case Image::ARGB:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelRGB, PixelARGB, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelRGB, PixelARGB, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			case Image::RGB:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelRGB, PixelRGB, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelRGB, PixelRGB, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			default:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelRGB, PixelAlpha, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelRGB, PixelAlpha, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			}
			break;

		default:
			switch (srcData.pixelFormat)
			{
			case Image::ARGB:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelAlpha, PixelARGB, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelAlpha, PixelARGB, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			case Image::RGB:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelAlpha, PixelRGB, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelAlpha, PixelRGB, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			default:
				if (tiledFill)  { TransformedImageFillEdgeTableRenderer <PixelAlpha, PixelAlpha, true>  r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				else		{ TransformedImageFillEdgeTableRenderer <PixelAlpha, PixelAlpha, false> r (destData, srcData, transform, alpha, betterQuality); iter.iterate (r); }
				break;
			}
			break;
		}
	}

	template <class Iterator>
	static void renderImageUntransformedInternal (Iterator& iter, const Image::BitmapData& destData, const Image::BitmapData& srcData, const int alpha, int x, int y, bool tiledFill)
	{
		switch (destData.pixelFormat)
		{
		case Image::ARGB:
			switch (srcData.pixelFormat)
			{
			case Image::ARGB:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelARGB, PixelARGB, true>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelARGB, PixelARGB, false> r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			case Image::RGB:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelARGB, PixelRGB, true>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelARGB, PixelRGB, false> r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			default:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelARGB, PixelAlpha, true>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelARGB, PixelAlpha, false> r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			}
			break;

		case Image::RGB:
			switch (srcData.pixelFormat)
			{
			case Image::ARGB:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelRGB, PixelARGB, true>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelRGB, PixelARGB, false> r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			case Image::RGB:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelRGB, PixelRGB, true>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelRGB, PixelRGB, false> r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			default:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelRGB, PixelAlpha, true>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelRGB, PixelAlpha, false> r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			}
			break;

		default:
			switch (srcData.pixelFormat)
			{
			case Image::ARGB:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelAlpha, PixelARGB, true>   r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelAlpha, PixelARGB, false>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			case Image::RGB:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelAlpha, PixelRGB, true>	r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelAlpha, PixelRGB, false>   r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			default:
				if (tiledFill)  { ImageFillEdgeTableRenderer <PixelAlpha, PixelAlpha, true>  r (destData, srcData, alpha, x, y); iter.iterate (r); }
				else		{ ImageFillEdgeTableRenderer <PixelAlpha, PixelAlpha, false> r (destData, srcData, alpha, x, y); iter.iterate (r); }
				break;
			}
			break;
		}
	}

	template <class Iterator, class DestPixelType>
	static void renderSolidFill (Iterator& iter, const Image::BitmapData& destData, const PixelARGB& fillColour, const bool replaceContents, DestPixelType*)
	{
		jassert (destData.pixelStride == sizeof (DestPixelType));
		if (replaceContents)
		{
			SolidColourEdgeTableRenderer <DestPixelType, true> r (destData, fillColour);
			iter.iterate (r);
		}
		else
		{
			SolidColourEdgeTableRenderer <DestPixelType, false> r (destData, fillColour);
			iter.iterate (r);
		}
	}

	template <class Iterator, class DestPixelType>
	static void renderGradient (Iterator& iter, const Image::BitmapData& destData, const ColourGradient& g, const AffineTransform& transform,
								const PixelARGB* const lookupTable, const int numLookupEntries, const bool isIdentity, DestPixelType*)
	{
		jassert (destData.pixelStride == sizeof (DestPixelType));

		if (g.isRadial)
		{
			if (isIdentity)
			{
				GradientEdgeTableRenderer <DestPixelType, RadialGradientPixelGenerator> renderer (destData, g, transform, lookupTable, numLookupEntries);
				iter.iterate (renderer);
			}
			else
			{
				GradientEdgeTableRenderer <DestPixelType, TransformedRadialGradientPixelGenerator> renderer (destData, g, transform, lookupTable, numLookupEntries);
				iter.iterate (renderer);
			}
		}
		else
		{
			GradientEdgeTableRenderer <DestPixelType, LinearGradientPixelGenerator> renderer (destData, g, transform, lookupTable, numLookupEntries);
			iter.iterate (renderer);
		}
	}
};

class ClipRegion_EdgeTable  : public ClipRegionBase
{
public:
	ClipRegion_EdgeTable (const EdgeTable& e)	: edgeTable (e) {}
	ClipRegion_EdgeTable (const Rectangle<int>& r)   : edgeTable (r) {}
	ClipRegion_EdgeTable (const Rectangle<float>& r) : edgeTable (r) {}
	ClipRegion_EdgeTable (const RectangleList& r)	: edgeTable (r) {}
	ClipRegion_EdgeTable (const Rectangle<int>& bounds, const Path& p, const AffineTransform& t) : edgeTable (bounds, p, t) {}
	ClipRegion_EdgeTable (const ClipRegion_EdgeTable& other) : edgeTable (other.edgeTable) {}

	Ptr clone() const
	{
		return new ClipRegion_EdgeTable (*this);
	}

	Ptr applyClipTo (const Ptr& target) const
	{
		return target->clipToEdgeTable (edgeTable);
	}

	Ptr clipToRectangle (const Rectangle<int>& r)
	{
		edgeTable.clipToRectangle (r);
		return edgeTable.isEmpty() ? nullptr : this;
	}

	Ptr clipToRectangleList (const RectangleList& r)
	{
		RectangleList inverse (edgeTable.getMaximumBounds());

		if (inverse.subtract (r))
			for (RectangleList::Iterator iter (inverse); iter.next();)
				edgeTable.excludeRectangle (*iter.getRectangle());

		return edgeTable.isEmpty() ? nullptr : this;
	}

	Ptr excludeClipRectangle (const Rectangle<int>& r)
	{
		edgeTable.excludeRectangle (r);
		return edgeTable.isEmpty() ? nullptr : this;
	}

	Ptr clipToPath (const Path& p, const AffineTransform& transform)
	{
		EdgeTable et (edgeTable.getMaximumBounds(), p, transform);
		edgeTable.clipToEdgeTable (et);
		return edgeTable.isEmpty() ? nullptr : this;
	}

	Ptr clipToEdgeTable (const EdgeTable& et)
	{
		edgeTable.clipToEdgeTable (et);
		return edgeTable.isEmpty() ? nullptr : this;
	}

	Ptr clipToImageAlpha (const Image& image, const AffineTransform& transform, const bool betterQuality)
	{
		const Image::BitmapData srcData (image, Image::BitmapData::readOnly);

		if (transform.isOnlyTranslation())
		{
			// If our translation doesn't involve any distortion, just use a simple blit..
			const int tx = (int) (transform.getTranslationX() * 256.0f);
			const int ty = (int) (transform.getTranslationY() * 256.0f);

			if ((! betterQuality) || ((tx | ty) & 224) == 0)
			{
				const int imageX = ((tx + 128) >> 8);
				const int imageY = ((ty + 128) >> 8);

				if (image.getFormat() == Image::ARGB)
					straightClipImage (srcData, imageX, imageY, (PixelARGB*) 0);
				else
					straightClipImage (srcData, imageX, imageY, (PixelAlpha*) 0);

				return edgeTable.isEmpty() ? nullptr : this;
			}
		}

		if (transform.isSingularity())
			return nullptr;

		{
			Path p;
			p.addRectangle (0, 0, (float) srcData.width, (float) srcData.height);
			EdgeTable et2 (edgeTable.getMaximumBounds(), p, transform);
			edgeTable.clipToEdgeTable (et2);
		}

		if (! edgeTable.isEmpty())
		{
			if (image.getFormat() == Image::ARGB)
				transformedClipImage (srcData, transform, betterQuality, (PixelARGB*) 0);
			else
				transformedClipImage (srcData, transform, betterQuality, (PixelAlpha*) 0);
		}

		return edgeTable.isEmpty() ? nullptr : this;
	}

	Ptr translated (const Point<int>& delta)
	{
		edgeTable.translate ((float) delta.getX(), delta.getY());
		return edgeTable.isEmpty() ? nullptr : this;
	}

	bool clipRegionIntersects (const Rectangle<int>& r) const
	{
		return edgeTable.getMaximumBounds().intersects (r);
	}

	Rectangle<int> getClipBounds() const
	{
		return edgeTable.getMaximumBounds();
	}

	void fillRectWithColour (Image::BitmapData& destData, const Rectangle<int>& area, const PixelARGB& colour, bool replaceContents) const
	{
		const Rectangle<int> totalClip (edgeTable.getMaximumBounds());
		const Rectangle<int> clipped (totalClip.getIntersection (area));

		if (! clipped.isEmpty())
		{
			ClipRegion_EdgeTable et (clipped);
			et.edgeTable.clipToEdgeTable (edgeTable);
			et.fillAllWithColour (destData, colour, replaceContents);
		}
	}

	void fillRectWithColour (Image::BitmapData& destData, const Rectangle<float>& area, const PixelARGB& colour) const
	{
		const Rectangle<float> totalClip (edgeTable.getMaximumBounds().toFloat());
		const Rectangle<float> clipped (totalClip.getIntersection (area));

		if (! clipped.isEmpty())
		{
			ClipRegion_EdgeTable et (clipped);
			et.edgeTable.clipToEdgeTable (edgeTable);
			et.fillAllWithColour (destData, colour, false);
		}
	}

	void fillAllWithColour (Image::BitmapData& destData, const PixelARGB& colour, bool replaceContents) const
	{
		switch (destData.pixelFormat)
		{
			case Image::ARGB:   renderSolidFill (edgeTable, destData, colour, replaceContents, (PixelARGB*) 0); break;
			case Image::RGB:	renderSolidFill (edgeTable, destData, colour, replaceContents, (PixelRGB*) 0); break;
			default:		renderSolidFill (edgeTable, destData, colour, replaceContents, (PixelAlpha*) 0); break;
		}
	}

	void fillAllWithGradient (Image::BitmapData& destData, ColourGradient& gradient, const AffineTransform& transform, bool isIdentity) const
	{
		HeapBlock <PixelARGB> lookupTable;
		const int numLookupEntries = gradient.createLookupTable (transform, lookupTable);
		jassert (numLookupEntries > 0);

		switch (destData.pixelFormat)
		{
			case Image::ARGB:   renderGradient (edgeTable, destData, gradient, transform, lookupTable, numLookupEntries, isIdentity, (PixelARGB*) 0); break;
			case Image::RGB:	renderGradient (edgeTable, destData, gradient, transform, lookupTable, numLookupEntries, isIdentity, (PixelRGB*) 0); break;
			default:		renderGradient (edgeTable, destData, gradient, transform, lookupTable, numLookupEntries, isIdentity, (PixelAlpha*) 0); break;
		}
	}

	void renderImageTransformed (const Image::BitmapData& destData, const Image::BitmapData& srcData, const int alpha, const AffineTransform& transform, bool betterQuality, bool tiledFill) const
	{
		renderImageTransformedInternal (edgeTable, destData, srcData, alpha, transform, betterQuality, tiledFill);
	}

	void renderImageUntransformed (const Image::BitmapData& destData, const Image::BitmapData& srcData, const int alpha, int x, int y, bool tiledFill) const
	{
		renderImageUntransformedInternal (edgeTable, destData, srcData, alpha, x, y, tiledFill);
	}

	EdgeTable edgeTable;

private:

	template <class SrcPixelType>
	void transformedClipImage (const Image::BitmapData& srcData, const AffineTransform& transform, const bool betterQuality, const SrcPixelType*)
	{
		TransformedImageFillEdgeTableRenderer <SrcPixelType, SrcPixelType, false> renderer (srcData, srcData, transform, 255, betterQuality);

		for (int y = 0; y < edgeTable.getMaximumBounds().getHeight(); ++y)
			renderer.clipEdgeTableLine (edgeTable, edgeTable.getMaximumBounds().getX(), y + edgeTable.getMaximumBounds().getY(),
										edgeTable.getMaximumBounds().getWidth());
	}

	template <class SrcPixelType>
	void straightClipImage (const Image::BitmapData& srcData, int imageX, int imageY, const SrcPixelType*)
	{
		Rectangle<int> r (imageX, imageY, srcData.width, srcData.height);
		edgeTable.clipToRectangle (r);

		ImageFillEdgeTableRenderer <SrcPixelType, SrcPixelType, false> renderer (srcData, srcData, 255, imageX, imageY);

		for (int y = 0; y < r.getHeight(); ++y)
			renderer.clipEdgeTableLine (edgeTable, r.getX(), y + r.getY(), r.getWidth());
	}

	ClipRegion_EdgeTable& operator= (const ClipRegion_EdgeTable&);
};

class ClipRegion_RectangleList  : public ClipRegionBase
{
public:
	ClipRegion_RectangleList (const Rectangle<int>& r) : clip (r) {}
	ClipRegion_RectangleList (const RectangleList& r)  : clip (r) {}
	ClipRegion_RectangleList (const ClipRegion_RectangleList& other) : clip (other.clip) {}

	Ptr clone() const
	{
		return new ClipRegion_RectangleList (*this);
	}

	Ptr applyClipTo (const Ptr& target) const
	{
		return target->clipToRectangleList (clip);
	}

	Ptr clipToRectangle (const Rectangle<int>& r)
	{
		clip.clipTo (r);
		return clip.isEmpty() ? nullptr : this;
	}

	Ptr clipToRectangleList (const RectangleList& r)
	{
		clip.clipTo (r);
		return clip.isEmpty() ? nullptr : this;
	}

	Ptr excludeClipRectangle (const Rectangle<int>& r)
	{
		clip.subtract (r);
		return clip.isEmpty() ? nullptr : this;
	}

	Ptr clipToPath (const Path& p, const AffineTransform& transform)
	{
		return Ptr (new ClipRegion_EdgeTable (clip))->clipToPath (p, transform);
	}

	Ptr clipToEdgeTable (const EdgeTable& et)
	{
		return Ptr (new ClipRegion_EdgeTable (clip))->clipToEdgeTable (et);
	}

	Ptr clipToImageAlpha (const Image& image, const AffineTransform& transform, const bool betterQuality)
	{
		return Ptr (new ClipRegion_EdgeTable (clip))->clipToImageAlpha (image, transform, betterQuality);
	}

	Ptr translated (const Point<int>& delta)
	{
		clip.offsetAll (delta.getX(), delta.getY());
		return clip.isEmpty() ? nullptr : this;
	}

	bool clipRegionIntersects (const Rectangle<int>& r) const
	{
		return clip.intersects (r);
	}

	Rectangle<int> getClipBounds() const
	{
		return clip.getBounds();
	}

	void fillRectWithColour (Image::BitmapData& destData, const Rectangle<int>& area, const PixelARGB& colour, bool replaceContents) const
	{
		SubRectangleIterator iter (clip, area);

		switch (destData.pixelFormat)
		{
			case Image::ARGB:   renderSolidFill (iter, destData, colour, replaceContents, (PixelARGB*) 0); break;
			case Image::RGB:	renderSolidFill (iter, destData, colour, replaceContents, (PixelRGB*) 0); break;
			default:		renderSolidFill (iter, destData, colour, replaceContents, (PixelAlpha*) 0); break;
		}
	}

	void fillRectWithColour (Image::BitmapData& destData, const Rectangle<float>& area, const PixelARGB& colour) const
	{
		SubRectangleIteratorFloat iter (clip, area);

		switch (destData.pixelFormat)
		{
			case Image::ARGB:   renderSolidFill (iter, destData, colour, false, (PixelARGB*) 0); break;
			case Image::RGB:	renderSolidFill (iter, destData, colour, false, (PixelRGB*) 0); break;
			default:		renderSolidFill (iter, destData, colour, false, (PixelAlpha*) 0); break;
		}
	}

	void fillAllWithColour (Image::BitmapData& destData, const PixelARGB& colour, bool replaceContents) const
	{
		switch (destData.pixelFormat)
		{
			case Image::ARGB:   renderSolidFill (*this, destData, colour, replaceContents, (PixelARGB*) 0); break;
			case Image::RGB:	renderSolidFill (*this, destData, colour, replaceContents, (PixelRGB*) 0); break;
			default:		renderSolidFill (*this, destData, colour, replaceContents, (PixelAlpha*) 0); break;
		}
	}

	void fillAllWithGradient (Image::BitmapData& destData, ColourGradient& gradient, const AffineTransform& transform, bool isIdentity) const
	{
		HeapBlock <PixelARGB> lookupTable;
		const int numLookupEntries = gradient.createLookupTable (transform, lookupTable);
		jassert (numLookupEntries > 0);

		switch (destData.pixelFormat)
		{
			case Image::ARGB:   renderGradient (*this, destData, gradient, transform, lookupTable, numLookupEntries, isIdentity, (PixelARGB*) 0); break;
			case Image::RGB:	renderGradient (*this, destData, gradient, transform, lookupTable, numLookupEntries, isIdentity, (PixelRGB*) 0); break;
			default:		renderGradient (*this, destData, gradient, transform, lookupTable, numLookupEntries, isIdentity, (PixelAlpha*) 0); break;
		}
	}

	void renderImageTransformed (const Image::BitmapData& destData, const Image::BitmapData& srcData, const int alpha, const AffineTransform& transform, bool betterQuality, bool tiledFill) const
	{
		renderImageTransformedInternal (*this, destData, srcData, alpha, transform, betterQuality, tiledFill);
	}

	void renderImageUntransformed (const Image::BitmapData& destData, const Image::BitmapData& srcData, const int alpha, int x, int y, bool tiledFill) const
	{
		renderImageUntransformedInternal (*this, destData, srcData, alpha, x, y, tiledFill);
	}

	RectangleList clip;

	template <class Renderer>
	void iterate (Renderer& r) const noexcept
	{
		RectangleList::Iterator iter (clip);

		while (iter.next())
		{
			const Rectangle<int> rect (*iter.getRectangle());
			const int x = rect.getX();
			const int w = rect.getWidth();
			jassert (w > 0);
			const int bottom = rect.getBottom();

			for (int y = rect.getY(); y < bottom; ++y)
			{
				r.setEdgeTableYPos (y);
				r.handleEdgeTableLineFull (x, w);
			}
		}
	}

private:

	class SubRectangleIterator
	{
	public:
		SubRectangleIterator (const RectangleList& clip_, const Rectangle<int>& area_)
			: clip (clip_), area (area_)
		{
		}

		template <class Renderer>
		void iterate (Renderer& r) const noexcept
		{
			RectangleList::Iterator iter (clip);

			while (iter.next())
			{
				const Rectangle<int> rect (iter.getRectangle()->getIntersection (area));

				if (! rect.isEmpty())
				{
					const int x = rect.getX();
					const int w = rect.getWidth();
					const int bottom = rect.getBottom();

					for (int y = rect.getY(); y < bottom; ++y)
					{
						r.setEdgeTableYPos (y);
						r.handleEdgeTableLineFull (x, w);
					}
				}
			}
		}

	private:
		const RectangleList& clip;
		const Rectangle<int> area;

		JUCE_DECLARE_NON_COPYABLE (SubRectangleIterator);
	};

	class SubRectangleIteratorFloat
	{
	public:
		SubRectangleIteratorFloat (const RectangleList& clip_, const Rectangle<float>& area_)
			: clip (clip_), area (area_)
		{
		}

		template <class Renderer>
		void iterate (Renderer& r) const noexcept
		{
			int left	= roundToInt (area.getX() * 256.0f);
			int top	 = roundToInt (area.getY() * 256.0f);
			int right   = roundToInt (area.getRight() * 256.0f);
			int bottom  = roundToInt (area.getBottom() * 256.0f);

			int totalTop, totalLeft, totalBottom, totalRight;
			int topAlpha, leftAlpha, bottomAlpha, rightAlpha;

			if ((top >> 8) == (bottom >> 8))
			{
				topAlpha = bottom - top;
				bottomAlpha = 0;
				totalTop = top >> 8;
				totalBottom = bottom = top = totalTop + 1;
			}
			else
			{
				if ((top & 255) == 0)
				{
					topAlpha = 0;
					top = totalTop = (top >> 8);
				}
				else
				{
					topAlpha = 255 - (top & 255);
					totalTop = (top >> 8);
					top = totalTop + 1;
				}

				bottomAlpha = bottom & 255;
				bottom >>= 8;
				totalBottom = bottom + (bottomAlpha != 0 ? 1 : 0);
			}

			if ((left >> 8) == (right >> 8))
			{
				leftAlpha = right - left;
				rightAlpha = 0;
				totalLeft = (left >> 8);
				totalRight = right = left = totalLeft + 1;
			}
			else
			{
				if ((left & 255) == 0)
				{
					leftAlpha = 0;
					left = totalLeft = (left >> 8);
				}
				else
				{
					leftAlpha = 255 - (left & 255);
					totalLeft = (left >> 8);
					left = totalLeft + 1;
				}

				rightAlpha = right & 255;
				right >>= 8;
				totalRight = right + (rightAlpha != 0 ? 1 : 0);
			}

			RectangleList::Iterator iter (clip);

			while (iter.next())
			{
				const int clipLeft   = iter.getRectangle()->getX();
				const int clipRight  = iter.getRectangle()->getRight();
				const int clipTop	= iter.getRectangle()->getY();
				const int clipBottom = iter.getRectangle()->getBottom();

				if (totalBottom > clipTop && totalTop < clipBottom && totalRight > clipLeft && totalLeft < clipRight)
				{
					if (right - left == 1 && leftAlpha + rightAlpha == 0) // special case for 1-pix vertical lines
					{
						if (topAlpha != 0 && totalTop >= clipTop)
						{
							r.setEdgeTableYPos (totalTop);
							r.handleEdgeTablePixel (left, topAlpha);
						}

						const int endY = jmin (bottom, clipBottom);
						for (int y = jmax (clipTop, top); y < endY; ++y)
						{
							r.setEdgeTableYPos (y);
							r.handleEdgeTablePixelFull (left);
						}

						if (bottomAlpha != 0 && bottom < clipBottom)
						{
							r.setEdgeTableYPos (bottom);
							r.handleEdgeTablePixel (left, bottomAlpha);
						}
					}
					else
					{
						const int clippedLeft = jmax (left, clipLeft);
						const int clippedWidth = jmin (right, clipRight) - clippedLeft;
						const bool doLeftAlpha = leftAlpha != 0 && totalLeft >= clipLeft;
						const bool doRightAlpha = rightAlpha != 0 && right < clipRight;

						if (topAlpha != 0 && totalTop >= clipTop)
						{
							r.setEdgeTableYPos (totalTop);

							if (doLeftAlpha)	r.handleEdgeTablePixel (totalLeft, (leftAlpha * topAlpha) >> 8);
							if (clippedWidth > 0)   r.handleEdgeTableLine (clippedLeft, clippedWidth, topAlpha);
							if (doRightAlpha)	   r.handleEdgeTablePixel (right, (rightAlpha * topAlpha) >> 8);
						}

						const int endY = jmin (bottom, clipBottom);
						for (int y = jmax (clipTop, top); y < endY; ++y)
						{
							r.setEdgeTableYPos (y);

							if (doLeftAlpha)	r.handleEdgeTablePixel (totalLeft, leftAlpha);
							if (clippedWidth > 0)   r.handleEdgeTableLineFull (clippedLeft, clippedWidth);
							if (doRightAlpha)	   r.handleEdgeTablePixel (right, rightAlpha);
						}

						if (bottomAlpha != 0 && bottom < clipBottom)
						{
							r.setEdgeTableYPos (bottom);

							if (doLeftAlpha)	r.handleEdgeTablePixel (totalLeft, (leftAlpha * bottomAlpha) >> 8);
							if (clippedWidth > 0)   r.handleEdgeTableLine (clippedLeft, clippedWidth, bottomAlpha);
							if (doRightAlpha)	   r.handleEdgeTablePixel (right, (rightAlpha * bottomAlpha) >> 8);
						}
					}
				}
			}
		}

	private:
		const RectangleList& clip;
		const Rectangle<float>& area;

		JUCE_DECLARE_NON_COPYABLE (SubRectangleIteratorFloat);
	};

	ClipRegion_RectangleList& operator= (const ClipRegion_RectangleList&);
};

}

class LowLevelGraphicsSoftwareRenderer::SavedState
{
public:
	SavedState (const Image& image_, const Rectangle<int>& clip_, const int xOffset_, const int yOffset_)
		: image (image_), clip (new SoftwareRendererClasses::ClipRegion_RectangleList (clip_)),
		  xOffset (xOffset_), yOffset (yOffset_), compositionAlpha (1.0f),
		  isOnlyTranslated (true), interpolationQuality (Graphics::mediumResamplingQuality)
	{
	}

	SavedState (const Image& image_, const RectangleList& clip_, const int xOffset_, const int yOffset_)
		: image (image_), clip (new SoftwareRendererClasses::ClipRegion_RectangleList (clip_)),
		  xOffset (xOffset_), yOffset (yOffset_), compositionAlpha (1.0f),
		  isOnlyTranslated (true), interpolationQuality (Graphics::mediumResamplingQuality)
	{
	}

	SavedState (const SavedState& other)
		: image (other.image), clip (other.clip), complexTransform (other.complexTransform),
		  xOffset (other.xOffset), yOffset (other.yOffset), compositionAlpha (other.compositionAlpha),
		  isOnlyTranslated (other.isOnlyTranslated), font (other.font), fillType (other.fillType),
		  interpolationQuality (other.interpolationQuality)
	{
	}

	void setOrigin (const int x, const int y) noexcept
	{
		if (isOnlyTranslated)
		{
			xOffset += x;
			yOffset += y;
		}
		else
		{
			complexTransform = getTransformWith (AffineTransform::translation ((float) x, (float) y));
		}
	}

	void addTransform (const AffineTransform& t)
	{
		if ((! isOnlyTranslated)
			 || (! t.isOnlyTranslation())
			 || (int) (t.getTranslationX() * 256.0f) != 0
			 || (int) (t.getTranslationY() * 256.0f) != 0)
		{
			complexTransform = getTransformWith (t);
			isOnlyTranslated = false;
		}
		else
		{
			xOffset += (int) t.getTranslationX();
			yOffset += (int) t.getTranslationY();
		}
	}

	float getScaleFactor() const
	{
		return isOnlyTranslated ? 1.0f : complexTransform.getScaleFactor();
	}

	bool clipToRectangle (const Rectangle<int>& r)
	{
		if (clip != nullptr)
		{
			if (isOnlyTranslated)
			{
				cloneClipIfMultiplyReferenced();
				clip = clip->clipToRectangle (r.translated (xOffset, yOffset));
			}
			else
			{
				Path p;
				p.addRectangle (r);
				clipToPath (p, AffineTransform::identity);
			}
		}

		return clip != nullptr;
	}

	bool clipToRectangleList (const RectangleList& r)
	{
		if (clip != nullptr)
		{
			if (isOnlyTranslated)
			{
				cloneClipIfMultiplyReferenced();
				RectangleList offsetList (r);
				offsetList.offsetAll (xOffset, yOffset);
				clip = clip->clipToRectangleList (offsetList);
			}
			else
			{
				clipToPath (r.toPath(), AffineTransform::identity);
			}
		}

		return clip != nullptr;
	}

	bool excludeClipRectangle (const Rectangle<int>& r)
	{
		if (clip != nullptr)
		{
			cloneClipIfMultiplyReferenced();

			if (isOnlyTranslated)
			{
				clip = clip->excludeClipRectangle (r.translated (xOffset, yOffset));
			}
			else
			{
				Path p;
				p.addRectangle (r.toFloat());
				p.applyTransform (complexTransform);
				p.addRectangle (clip->getClipBounds().toFloat());
				p.setUsingNonZeroWinding (false);
				clip = clip->clipToPath (p, AffineTransform::identity);
			}
		}

		return clip != nullptr;
	}

	void clipToPath (const Path& p, const AffineTransform& transform)
	{
		if (clip != nullptr)
		{
			cloneClipIfMultiplyReferenced();
			clip = clip->clipToPath (p, getTransformWith (transform));
		}
	}

	void clipToImageAlpha (const Image& sourceImage, const AffineTransform& t)
	{
		if (clip != nullptr)
		{
			if (sourceImage.hasAlphaChannel())
			{
				cloneClipIfMultiplyReferenced();
				clip = clip->clipToImageAlpha (sourceImage, getTransformWith (t),
											   interpolationQuality != Graphics::lowResamplingQuality);
			}
			else
			{
				Path p;
				p.addRectangle (sourceImage.getBounds());
				clipToPath (p, t);
			}
		}
	}

	bool clipRegionIntersects (const Rectangle<int>& r) const
	{
		if (clip != nullptr)
		{
			if (isOnlyTranslated)
				return clip->clipRegionIntersects (r.translated (xOffset, yOffset));
			else
				return getClipBounds().intersects (r);
		}

		return false;
	}

	Rectangle<int> getUntransformedClipBounds() const
	{
		return clip != nullptr ? clip->getClipBounds() : Rectangle<int>();
	}

	Rectangle<int> getClipBounds() const
	{
		if (clip != nullptr)
		{
			if (isOnlyTranslated)
				return clip->getClipBounds().translated (-xOffset, -yOffset);
			else
				return clip->getClipBounds().toFloat().transformed (complexTransform.inverted()).getSmallestIntegerContainer();
		}

		return Rectangle<int>();
	}

	SavedState* beginTransparencyLayer (float opacity)
	{
		const Rectangle<int> layerBounds (getUntransformedClipBounds());

		SavedState* s = new SavedState (*this);
		s->image = Image (Image::ARGB, layerBounds.getWidth(), layerBounds.getHeight(), true);
		s->compositionAlpha = opacity;

		if (s->isOnlyTranslated)
		{
			s->xOffset -= layerBounds.getX();
			s->yOffset -= layerBounds.getY();
		}
		else
		{
			s->complexTransform = s->complexTransform.followedBy (AffineTransform::translation ((float) -layerBounds.getX(),
																								(float) -layerBounds.getY()));
		}

		s->cloneClipIfMultiplyReferenced();
		s->clip = s->clip->translated (-layerBounds.getPosition());
		return s;
	}

	void endTransparencyLayer (SavedState& layerState)
	{
		const Rectangle<int> layerBounds (getUntransformedClipBounds());

		const ScopedPointer<LowLevelGraphicsContext> g (image.createLowLevelContext());
		g->setOpacity (layerState.compositionAlpha);
		g->drawImage (layerState.image, AffineTransform::translation ((float) layerBounds.getX(),
																	  (float) layerBounds.getY()), false);
	}

	void fillRect (const Rectangle<int>& r, const bool replaceContents)
	{
		if (clip != nullptr)
		{
			if (isOnlyTranslated)
			{
				if (fillType.isColour())
				{
					Image::BitmapData destData (image, Image::BitmapData::readWrite);
					clip->fillRectWithColour (destData, r.translated (xOffset, yOffset), fillType.colour.getPixelARGB(), replaceContents);
				}
				else
				{
					const Rectangle<int> totalClip (clip->getClipBounds());
					const Rectangle<int> clipped (totalClip.getIntersection (r.translated (xOffset, yOffset)));

					if (! clipped.isEmpty())
						fillShape (new SoftwareRendererClasses::ClipRegion_RectangleList (clipped), false);
				}
			}
			else
			{
				Path p;
				p.addRectangle (r);
				fillPath (p, AffineTransform::identity);
			}
		}
	}

	void fillRect (const Rectangle<float>& r)
	{
		if (clip != nullptr)
		{
			if (isOnlyTranslated)
			{
				if (fillType.isColour())
				{
					Image::BitmapData destData (image, Image::BitmapData::readWrite);
					clip->fillRectWithColour (destData, r.translated ((float) xOffset, (float) yOffset), fillType.colour.getPixelARGB());
				}
				else
				{
					const Rectangle<float> totalClip (clip->getClipBounds().toFloat());
					const Rectangle<float> clipped (totalClip.getIntersection (r.translated ((float) xOffset, (float) yOffset)));

					if (! clipped.isEmpty())
						fillShape (new SoftwareRendererClasses::ClipRegion_EdgeTable (clipped), false);
				}
			}
			else
			{
				Path p;
				p.addRectangle (r);
				fillPath (p, AffineTransform::identity);
			}
		}
	}

	void fillPath (const Path& path, const AffineTransform& transform)
	{
		if (clip != nullptr)
			fillShape (new SoftwareRendererClasses::ClipRegion_EdgeTable (clip->getClipBounds(), path, getTransformWith (transform)), false);
	}

	void fillEdgeTable (const EdgeTable& edgeTable, const float x, const int y)
	{
		jassert (isOnlyTranslated);

		if (clip != nullptr)
		{
			SoftwareRendererClasses::ClipRegion_EdgeTable* edgeTableClip = new SoftwareRendererClasses::ClipRegion_EdgeTable (edgeTable);
			SoftwareRendererClasses::ClipRegionBase::Ptr shapeToFill (edgeTableClip);
			edgeTableClip->edgeTable.translate (x + xOffset, y + yOffset);
			fillShape (shapeToFill, false);
		}
	}

	void drawGlyph (const Font& f, int glyphNumber, const AffineTransform& transform)
	{
		const ScopedPointer<EdgeTable> et (f.getTypeface()->getEdgeTableForGlyph (glyphNumber, getTransformWith (transform)));

		if (et != nullptr && clip != nullptr)
		{
			SoftwareRendererClasses::ClipRegion_EdgeTable* edgeTableClip = new SoftwareRendererClasses::ClipRegion_EdgeTable (*et);
			SoftwareRendererClasses::ClipRegionBase::Ptr shapeToFill (edgeTableClip);
			fillShape (shapeToFill, false);
		}
	}

	void fillShape (SoftwareRendererClasses::ClipRegionBase::Ptr shapeToFill, const bool replaceContents)
	{
		jassert (clip != nullptr);

		shapeToFill = clip->applyClipTo (shapeToFill);

		if (shapeToFill != nullptr)
		{
			Image::BitmapData destData (image, Image::BitmapData::readWrite);

			if (fillType.isGradient())
			{
				jassert (! replaceContents); // that option is just for solid colours

				ColourGradient g2 (*(fillType.gradient));
				g2.multiplyOpacity (fillType.getOpacity());
				AffineTransform transform (getTransformWith (fillType.transform).translated (-0.5f, -0.5f));

				const bool isIdentity = transform.isOnlyTranslation();

				if (isIdentity && false) // disabled, seems to break the code generation
				{
					// If our translation doesn't involve any distortion, we can speed it up..
					g2.point1.applyTransform (transform);
					g2.point2.applyTransform (transform);
					transform = AffineTransform::identity;
				}

				shapeToFill->fillAllWithGradient (destData, g2, transform, isIdentity);
			}
			else if (fillType.isTiledImage())
			{
				renderImage (fillType.image, fillType.transform, shapeToFill);
			}
			else
			{
				shapeToFill->fillAllWithColour (destData, fillType.colour.getPixelARGB(), replaceContents);
			}
		}
	}

	void renderImage (const Image& sourceImage, const AffineTransform& t, const SoftwareRendererClasses::ClipRegionBase* const tiledFillClipRegion)
	{
		const AffineTransform transform (getTransformWith (t));

		const Image::BitmapData destData (image, Image::BitmapData::readWrite);
		const Image::BitmapData srcData (sourceImage, Image::BitmapData::readOnly);
		const int alpha = fillType.colour.getAlpha();
		const bool betterQuality = (interpolationQuality != Graphics::lowResamplingQuality);

		if (transform.isOnlyTranslation())
		{
			// If our translation doesn't involve any distortion, just use a simple blit..
			int tx = (int) (transform.getTranslationX() * 256.0f);
			int ty = (int) (transform.getTranslationY() * 256.0f);

			if ((! betterQuality) || ((tx | ty) & 224) == 0)
			{
				tx = ((tx + 128) >> 8);
				ty = ((ty + 128) >> 8);

				if (tiledFillClipRegion != nullptr)
				{
					tiledFillClipRegion->renderImageUntransformed (destData, srcData, alpha, tx, ty, true);
				}
				else
				{
					Rectangle<int> area (tx, ty, sourceImage.getWidth(), sourceImage.getHeight());
					area = area.getIntersection (image.getBounds());

					if (! area.isEmpty())
					{
						SoftwareRendererClasses::ClipRegionBase::Ptr c (new SoftwareRendererClasses::ClipRegion_EdgeTable (area));
						c = clip->applyClipTo (c);

						if (c != nullptr)
							c->renderImageUntransformed (destData, srcData, alpha, tx, ty, false);
					}
				}

				return;
			}
		}

		if (transform.isSingularity())
			return;

		if (tiledFillClipRegion != nullptr)
		{
			tiledFillClipRegion->renderImageTransformed (destData, srcData, alpha, transform, betterQuality, true);
		}
		else
		{
			Path p;
			p.addRectangle (sourceImage.getBounds());

			SoftwareRendererClasses::ClipRegionBase::Ptr c (clip->clone());
			c = c->clipToPath (p, transform);

			if (c != nullptr)
				c->renderImageTransformed (destData, srcData, alpha, transform, betterQuality, false);
		}
	}

	Image image;
	SoftwareRendererClasses::ClipRegionBase::Ptr clip;

private:
	AffineTransform complexTransform;
	int xOffset, yOffset;
	float compositionAlpha;

public:
	bool isOnlyTranslated;

	Font font;
	FillType fillType;
	Graphics::ResamplingQuality interpolationQuality;

private:
	void cloneClipIfMultiplyReferenced()
	{
		if (clip->getReferenceCount() > 1)
			clip = clip->clone();
	}

	const AffineTransform getTransform() const
	{
		if (isOnlyTranslated)
			return AffineTransform::translation ((float) xOffset, (float) yOffset);

		return complexTransform;
	}

	const AffineTransform getTransformWith (const AffineTransform& userTransform) const
	{
		if (isOnlyTranslated)
			return userTransform.translated ((float) xOffset, (float) yOffset);

		return userTransform.followedBy (complexTransform);
	}

	SavedState& operator= (const SavedState&);
};

LowLevelGraphicsSoftwareRenderer::LowLevelGraphicsSoftwareRenderer (const Image& image_)
	: image (image_),
	  currentState (new SavedState (image_, image_.getBounds(), 0, 0))
{
}

LowLevelGraphicsSoftwareRenderer::LowLevelGraphicsSoftwareRenderer (const Image& image_, const int xOffset, const int yOffset,
																	const RectangleList& initialClip)
	: image (image_),
	  currentState (new SavedState (image_, initialClip, xOffset, yOffset))
{
}

LowLevelGraphicsSoftwareRenderer::~LowLevelGraphicsSoftwareRenderer()
{
}

bool LowLevelGraphicsSoftwareRenderer::isVectorDevice() const
{
	return false;
}

void LowLevelGraphicsSoftwareRenderer::setOrigin (int x, int y)
{
	currentState->setOrigin (x, y);
}

void LowLevelGraphicsSoftwareRenderer::addTransform (const AffineTransform& transform)
{
	currentState->addTransform (transform);
}

float LowLevelGraphicsSoftwareRenderer::getScaleFactor()
{
	return currentState->getScaleFactor();
}

bool LowLevelGraphicsSoftwareRenderer::clipToRectangle (const Rectangle<int>& r)
{
	return currentState->clipToRectangle (r);
}

bool LowLevelGraphicsSoftwareRenderer::clipToRectangleList (const RectangleList& clipRegion)
{
	return currentState->clipToRectangleList (clipRegion);
}

void LowLevelGraphicsSoftwareRenderer::excludeClipRectangle (const Rectangle<int>& r)
{
	currentState->excludeClipRectangle (r);
}

void LowLevelGraphicsSoftwareRenderer::clipToPath (const Path& path, const AffineTransform& transform)
{
	currentState->clipToPath (path, transform);
}

void LowLevelGraphicsSoftwareRenderer::clipToImageAlpha (const Image& sourceImage, const AffineTransform& transform)
{
	currentState->clipToImageAlpha (sourceImage, transform);
}

bool LowLevelGraphicsSoftwareRenderer::clipRegionIntersects (const Rectangle<int>& r)
{
	return currentState->clipRegionIntersects (r);
}

const Rectangle<int> LowLevelGraphicsSoftwareRenderer::getClipBounds() const
{
	return currentState->getClipBounds();
}

bool LowLevelGraphicsSoftwareRenderer::isClipEmpty() const
{
	return currentState->clip == 0;
}

void LowLevelGraphicsSoftwareRenderer::saveState()
{
	stateStack.add (new SavedState (*currentState));
}

void LowLevelGraphicsSoftwareRenderer::restoreState()
{
	SavedState* const top = stateStack.getLast();

	if (top != nullptr)
	{
		currentState = top;
		stateStack.removeLast (1, false);
	}
	else
	{
		jassertfalse; // trying to pop with an empty stack!
	}
}

void LowLevelGraphicsSoftwareRenderer::beginTransparencyLayer (float opacity)
{
	saveState();
	currentState = currentState->beginTransparencyLayer (opacity);
}

void LowLevelGraphicsSoftwareRenderer::endTransparencyLayer()
{
	const ScopedPointer<SavedState> layer (currentState);
	restoreState();
	currentState->endTransparencyLayer (*layer);
}

void LowLevelGraphicsSoftwareRenderer::setFill (const FillType& fillType)
{
	currentState->fillType = fillType;
}

void LowLevelGraphicsSoftwareRenderer::setOpacity (float newOpacity)
{
	currentState->fillType.setOpacity (newOpacity);
}

void LowLevelGraphicsSoftwareRenderer::setInterpolationQuality (Graphics::ResamplingQuality quality)
{
	currentState->interpolationQuality = quality;
}

void LowLevelGraphicsSoftwareRenderer::fillRect (const Rectangle<int>& r, const bool replaceExistingContents)
{
	currentState->fillRect (r, replaceExistingContents);
}

void LowLevelGraphicsSoftwareRenderer::fillPath (const Path& path, const AffineTransform& transform)
{
	currentState->fillPath (path, transform);
}

void LowLevelGraphicsSoftwareRenderer::drawImage (const Image& sourceImage, const AffineTransform& transform, const bool fillEntireClipAsTiles)
{
	currentState->renderImage (sourceImage, transform, fillEntireClipAsTiles ? currentState->clip : 0);
}

void LowLevelGraphicsSoftwareRenderer::drawLine (const Line <float>& line)
{
	Path p;
	p.addLineSegment (line, 1.0f);
	fillPath (p, AffineTransform::identity);
}

void LowLevelGraphicsSoftwareRenderer::drawVerticalLine (const int x, float top, float bottom)
{
	if (bottom > top)
		currentState->fillRect (Rectangle<float> ((float) x, top, 1.0f, bottom - top));
}

void LowLevelGraphicsSoftwareRenderer::drawHorizontalLine (const int y, float left, float right)
{
	if (right > left)
		currentState->fillRect (Rectangle<float> (left, (float) y, right - left, 1.0f));
}

class LowLevelGraphicsSoftwareRenderer::CachedGlyph
{
public:
	CachedGlyph() : glyph (0), lastAccessCount (0) {}

	void draw (SavedState& state, float x, const float y) const
	{
		if (snapToIntegerCoordinate)
			x = std::floor (x + 0.5f);

		if (edgeTable != nullptr)
			state.fillEdgeTable (*edgeTable, x, roundToInt (y));
	}

	void generate (const Font& newFont, const int glyphNumber)
	{
		font = newFont;
		snapToIntegerCoordinate = newFont.getTypeface()->isHinted();
		glyph = glyphNumber;

		const float fontHeight = font.getHeight();
		edgeTable = font.getTypeface()->getEdgeTableForGlyph (glyphNumber,
															  AffineTransform::scale (fontHeight * font.getHorizontalScale(), fontHeight)
																			#if JUCE_MAC || JUCE_IOS
																			  .translated (0.0f, -0.5f)
																			#endif
															  );
	}

	Font font;
	int glyph, lastAccessCount;
	bool snapToIntegerCoordinate;

private:
	ScopedPointer <EdgeTable> edgeTable;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (CachedGlyph);
};

class LowLevelGraphicsSoftwareRenderer::GlyphCache  : private DeletedAtShutdown
{
public:
	GlyphCache()
		: accessCounter (0), hits (0), misses (0)
	{
		addNewGlyphSlots (120);
	}

	~GlyphCache()
	{
		clearSingletonInstance();
	}

	juce_DeclareSingleton_SingleThreaded_Minimal (GlyphCache);

	void drawGlyph (SavedState& state, const Font& font, const int glyphNumber, float x, float y)
	{
		++accessCounter;
		int oldestCounter = std::numeric_limits<int>::max();
		CachedGlyph* oldest = nullptr;

		for (int i = glyphs.size(); --i >= 0;)
		{
			CachedGlyph* const glyph = glyphs.getUnchecked (i);

			if (glyph->glyph == glyphNumber && glyph->font == font)
			{
				++hits;
				glyph->lastAccessCount = accessCounter;
				glyph->draw (state, x, y);
				return;
			}

			if (glyph->lastAccessCount <= oldestCounter)
			{
				oldestCounter = glyph->lastAccessCount;
				oldest = glyph;
			}
		}

		if (hits + ++misses > (glyphs.size() << 4))
		{
			if (misses * 2 > hits)
				addNewGlyphSlots (32);

			hits = misses = 0;
			oldest = glyphs.getLast();
		}

		jassert (oldest != nullptr);
		oldest->lastAccessCount = accessCounter;
		oldest->generate (font, glyphNumber);
		oldest->draw (state, x, y);
	}

private:
	friend class OwnedArray <CachedGlyph>;
	OwnedArray <CachedGlyph> glyphs;
	int accessCounter, hits, misses;

	void addNewGlyphSlots (int num)
	{
		while (--num >= 0)
			glyphs.add (new CachedGlyph());
	}

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (GlyphCache);
};

juce_ImplementSingleton_SingleThreaded (LowLevelGraphicsSoftwareRenderer::GlyphCache);

void LowLevelGraphicsSoftwareRenderer::setFont (const Font& newFont)
{
	currentState->font = newFont;
}

Font LowLevelGraphicsSoftwareRenderer::getFont()
{
	return currentState->font;
}

void LowLevelGraphicsSoftwareRenderer::drawGlyph (int glyphNumber, const AffineTransform& transform)
{
	Font& f = currentState->font;

	if (transform.isOnlyTranslation() && currentState->isOnlyTranslated)
	{
		GlyphCache::getInstance()->drawGlyph (*currentState, f, glyphNumber,
											  transform.getTranslationX(),
											  transform.getTranslationY());
	}
	else
	{
		const float fontHeight = f.getHeight();
		currentState->drawGlyph (f, glyphNumber, AffineTransform::scale (fontHeight * f.getHorizontalScale(), fontHeight)
																 .followedBy (transform));
	}
}

#if JUCE_MSVC
 #pragma warning (pop)

 #if JUCE_DEBUG
  #pragma optimize ("", on)  // resets optimisations to the project defaults
 #endif
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_LowLevelGraphicsSoftwareRenderer.cpp ***/


/*** Start of inlined file: juce_RectanglePlacement.cpp ***/
BEGIN_JUCE_NAMESPACE

RectanglePlacement::RectanglePlacement (const RectanglePlacement& other) noexcept
	: flags (other.flags)
{
}

RectanglePlacement& RectanglePlacement::operator= (const RectanglePlacement& other) noexcept
{
	flags = other.flags;
	return *this;
}

bool RectanglePlacement::operator== (const RectanglePlacement& other) const noexcept
{
	return flags == other.flags;
}

bool RectanglePlacement::operator!= (const RectanglePlacement& other) const noexcept
{
	return flags != other.flags;
}

void RectanglePlacement::applyTo (double& x, double& y, double& w, double& h,
								  const double dx, const double dy, const double dw, const double dh) const noexcept
{
	if (w == 0 || h == 0)
		return;

	if ((flags & stretchToFit) != 0)
	{
		x = dx;
		y = dy;
		w = dw;
		h = dh;
	}
	else
	{
		double scale = (flags & fillDestination) != 0 ? jmax (dw / w, dh / h)
													  : jmin (dw / w, dh / h);

		if ((flags & onlyReduceInSize) != 0)
			scale = jmin (scale, 1.0);

		if ((flags & onlyIncreaseInSize) != 0)
			scale = jmax (scale, 1.0);

		w *= scale;
		h *= scale;

		if ((flags & xLeft) != 0)
			x = dx;
		else if ((flags & xRight) != 0)
			x = dx + dw - w;
		else
			x = dx + (dw - w) * 0.5;

		if ((flags & yTop) != 0)
			y = dy;
		else if ((flags & yBottom) != 0)
			y = dy + dh - h;
		else
			y = dy + (dh - h) * 0.5;
	}
}

const AffineTransform RectanglePlacement::getTransformToFit (const Rectangle<float>& source, const Rectangle<float>& destination) const noexcept
{
	if (source.isEmpty())
		return AffineTransform::identity;

	float newX = destination.getX();
	float newY = destination.getY();

	float scaleX = destination.getWidth() / source.getWidth();
	float scaleY = destination.getHeight() / source.getHeight();

	if ((flags & stretchToFit) == 0)
	{
		scaleX = (flags & fillDestination) != 0 ? jmax (scaleX, scaleY)
												: jmin (scaleX, scaleY);

		if ((flags & onlyReduceInSize) != 0)
			scaleX = jmin (scaleX, 1.0f);

		if ((flags & onlyIncreaseInSize) != 0)
			scaleX = jmax (scaleX, 1.0f);

		scaleY = scaleX;

		if ((flags & xRight) != 0)
			newX += destination.getWidth() - source.getWidth() * scaleX;		 // right
		else if ((flags & xLeft) == 0)
			newX += (destination.getWidth() - source.getWidth() * scaleX) / 2.0f;	// centre

		if ((flags & yBottom) != 0)
			newY += destination.getHeight() - source.getHeight() * scaleX;		 // bottom
		else if ((flags & yTop) == 0)
			newY += (destination.getHeight() - source.getHeight() * scaleX) / 2.0f;	// centre
	}

	return AffineTransform::translation (-source.getX(), -source.getY())
				.scaled (scaleX, scaleY)
				.translated (newX, newY);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RectanglePlacement.cpp ***/


/*** Start of inlined file: juce_Drawable.cpp ***/
BEGIN_JUCE_NAMESPACE

Drawable::Drawable()
{
	setInterceptsMouseClicks (false, false);
	setPaintingIsUnclipped (true);
}

Drawable::~Drawable()
{
}

void Drawable::draw (Graphics& g, float opacity, const AffineTransform& transform) const
{
	const_cast <Drawable*> (this)->nonConstDraw (g, opacity, transform);
}

void Drawable::nonConstDraw (Graphics& g, float opacity, const AffineTransform& transform)
{
	Graphics::ScopedSaveState ss (g);

	const float oldOpacity = getAlpha();
	setAlpha (opacity);
	g.addTransform (AffineTransform::translation ((float) -originRelativeToComponent.getX(),
												  (float) -originRelativeToComponent.getY())
						.followedBy (getTransform())
						.followedBy (transform));

	if (! g.isClipEmpty())
		paintEntireComponent (g, false);

	setAlpha (oldOpacity);
}

void Drawable::drawAt (Graphics& g, float x, float y, float opacity) const
{
	draw (g, opacity, AffineTransform::translation (x, y));
}

void Drawable::drawWithin (Graphics& g, const Rectangle<float>& destArea, const RectanglePlacement& placement, float opacity) const
{
	draw (g, opacity, placement.getTransformToFit (getDrawableBounds(), destArea));
}

DrawableComposite* Drawable::getParent() const
{
	return dynamic_cast <DrawableComposite*> (getParentComponent());
}

void Drawable::transformContextToCorrectOrigin (Graphics& g)
{
	g.setOrigin (originRelativeToComponent.getX(),
				 originRelativeToComponent.getY());
}

void Drawable::parentHierarchyChanged()
{
	setBoundsToEnclose (getDrawableBounds());
}

void Drawable::setBoundsToEnclose (const Rectangle<float>& area)
{
	Drawable* const parent = getParent();
	Point<int> parentOrigin;
	if (parent != nullptr)
		parentOrigin = parent->originRelativeToComponent;

	const Rectangle<int> newBounds (area.getSmallestIntegerContainer() + parentOrigin);
	originRelativeToComponent = parentOrigin - newBounds.getPosition();
	setBounds (newBounds);
}

void Drawable::setOriginWithOriginalSize (const Point<float>& originWithinParent)
{
	setTransform (AffineTransform::translation (originWithinParent.getX(), originWithinParent.getY()));
}

void Drawable::setTransformToFit (const Rectangle<float>& area, const RectanglePlacement& placement)
{
	if (! area.isEmpty())
		setTransform (placement.getTransformToFit (getDrawableBounds(), area));
}

Drawable* Drawable::createFromImageData (const void* data, const size_t numBytes)
{
	Drawable* result = nullptr;

	Image image (ImageFileFormat::loadFrom (data, (int) numBytes));

	if (image.isValid())
	{
		DrawableImage* const di = new DrawableImage();
		di->setImage (image);
		result = di;
	}
	else
	{
		const String asString (String::createStringFromData (data, (int) numBytes));

		XmlDocument doc (asString);
		ScopedPointer <XmlElement> outer (doc.getDocumentElement (true));

		if (outer != nullptr && outer->hasTagName ("svg"))
		{
			ScopedPointer <XmlElement> svg (doc.getDocumentElement());

			if (svg != nullptr)
				result = Drawable::createFromSVG (*svg);
		}
	}

	return result;
}

Drawable* Drawable::createFromImageDataStream (InputStream& dataSource)
{
	MemoryOutputStream mo;
	mo.writeFromInputStream (dataSource, -1);

	return createFromImageData (mo.getData(), mo.getDataSize());
}

Drawable* Drawable::createFromImageFile (const File& file)
{
	const ScopedPointer <FileInputStream> fin (file.createInputStream());

	return fin != nullptr ? createFromImageDataStream (*fin) : nullptr;
}

template <class DrawableClass>
class DrawableTypeHandler  : public ComponentBuilder::TypeHandler
{
public:
	DrawableTypeHandler()
		: ComponentBuilder::TypeHandler (DrawableClass::valueTreeType)
	{
	}

	Component* addNewComponentFromState (const ValueTree& state, Component* parent)
	{
		DrawableClass* const d = new DrawableClass();

		if (parent != nullptr)
			parent->addAndMakeVisible (d);

		updateComponentFromState (d, state);
		return d;
	}

	void updateComponentFromState (Component* component, const ValueTree& state)
	{
		DrawableClass* const d = dynamic_cast <DrawableClass*> (component);
		jassert (d != nullptr);
		d->refreshFromValueTree (state, *this->getBuilder());
	}
};

void Drawable::registerDrawableTypeHandlers (ComponentBuilder& builder)
{
	builder.registerTypeHandler (new DrawableTypeHandler <DrawablePath>());
	builder.registerTypeHandler (new DrawableTypeHandler <DrawableComposite>());
	builder.registerTypeHandler (new DrawableTypeHandler <DrawableRectangle>());
	builder.registerTypeHandler (new DrawableTypeHandler <DrawableImage>());
	builder.registerTypeHandler (new DrawableTypeHandler <DrawableText>());
}

Drawable* Drawable::createFromValueTree (const ValueTree& tree, ComponentBuilder::ImageProvider* imageProvider)
{
	ComponentBuilder builder (tree);
	builder.setImageProvider (imageProvider);
	registerDrawableTypeHandlers (builder);

	ScopedPointer<Component> comp (builder.createComponent());
	Drawable* const d = dynamic_cast<Drawable*> (static_cast <Component*> (comp));

	if (d != nullptr)
		comp.release();

	return d;
}

Drawable::ValueTreeWrapperBase::ValueTreeWrapperBase (const ValueTree& state_)
	: state (state_)
{
}

String Drawable::ValueTreeWrapperBase::getID() const
{
	return state [ComponentBuilder::idProperty];
}

void Drawable::ValueTreeWrapperBase::setID (const String& newID)
{
	if (newID.isEmpty())
		state.removeProperty (ComponentBuilder::idProperty, nullptr);
	else
		state.setProperty (ComponentBuilder::idProperty, newID, nullptr);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Drawable.cpp ***/


/*** Start of inlined file: juce_DrawableShape.cpp ***/
BEGIN_JUCE_NAMESPACE

DrawableShape::DrawableShape()
	: strokeType (0.0f),
	  mainFill (Colours::black),
	  strokeFill (Colours::black)
{
}

DrawableShape::DrawableShape (const DrawableShape& other)
	: strokeType (other.strokeType),
	  mainFill (other.mainFill),
	  strokeFill (other.strokeFill)
{
}

DrawableShape::~DrawableShape()
{
}

class DrawableShape::RelativePositioner  : public RelativeCoordinatePositionerBase
{
public:
	RelativePositioner (DrawableShape& component_, const DrawableShape::RelativeFillType& fill_, bool isMainFill_)
		: RelativeCoordinatePositionerBase (component_),
		  owner (component_),
		  fill (fill_),
		  isMainFill (isMainFill_)
	{
	}

	bool registerCoordinates()
	{
		bool ok = addPoint (fill.gradientPoint1);
		ok = addPoint (fill.gradientPoint2) && ok;
		return addPoint (fill.gradientPoint3) && ok;
	}

	void applyToComponentBounds()
	{
		ComponentScope scope (owner);
		if (isMainFill ? owner.mainFill.recalculateCoords (&scope)
					   : owner.strokeFill.recalculateCoords (&scope))
			owner.repaint();
	}

	void applyNewBounds (const Rectangle<int>&)
	{
		jassertfalse; // drawables can't be resized directly!
	}

private:
	DrawableShape& owner;
	const DrawableShape::RelativeFillType fill;
	const bool isMainFill;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (RelativePositioner);
};

void DrawableShape::setFill (const FillType& newFill)
{
	setFill (RelativeFillType (newFill));
}

void DrawableShape::setStrokeFill (const FillType& newFill)
{
	setStrokeFill (RelativeFillType (newFill));
}

void DrawableShape::setFillInternal (RelativeFillType& fill, const RelativeFillType& newFill,
									 ScopedPointer<RelativeCoordinatePositionerBase>& pos)
{
	if (fill != newFill)
	{
		fill = newFill;
		pos = nullptr;

		if (fill.isDynamic())
		{
			pos = new RelativePositioner (*this, fill, true);
			pos->apply();
		}
		else
		{
			fill.recalculateCoords (nullptr);
		}

		repaint();
	}
}

void DrawableShape::setFill (const RelativeFillType& newFill)
{
	setFillInternal (mainFill, newFill, mainFillPositioner);
}

void DrawableShape::setStrokeFill (const RelativeFillType& newFill)
{
	setFillInternal (strokeFill, newFill, strokeFillPositioner);
}

void DrawableShape::setStrokeType (const PathStrokeType& newStrokeType)
{
	if (strokeType != newStrokeType)
	{
		strokeType = newStrokeType;
		strokeChanged();
	}
}

void DrawableShape::setStrokeThickness (const float newThickness)
{
	setStrokeType (PathStrokeType (newThickness, strokeType.getJointStyle(), strokeType.getEndStyle()));
}

bool DrawableShape::isStrokeVisible() const noexcept
{
	return strokeType.getStrokeThickness() > 0.0f && ! strokeFill.fill.isInvisible();
}

void DrawableShape::refreshFillTypes (const FillAndStrokeState& newState, ComponentBuilder::ImageProvider* imageProvider)
{
	setFill (newState.getFill (FillAndStrokeState::fill, imageProvider));
	setStrokeFill (newState.getFill (FillAndStrokeState::stroke, imageProvider));
}

void DrawableShape::writeTo (FillAndStrokeState& state, ComponentBuilder::ImageProvider* imageProvider, UndoManager* undoManager) const
{
	state.setFill (FillAndStrokeState::fill, mainFill, imageProvider, undoManager);
	state.setFill (FillAndStrokeState::stroke, strokeFill, imageProvider, undoManager);
	state.setStrokeType (strokeType, undoManager);
}

void DrawableShape::paint (Graphics& g)
{
	transformContextToCorrectOrigin (g);

	g.setFillType (mainFill.fill);
	g.fillPath (path);

	if (isStrokeVisible())
	{
		g.setFillType (strokeFill.fill);
		g.fillPath (strokePath);
	}
}

void DrawableShape::pathChanged()
{
	strokeChanged();
}

void DrawableShape::strokeChanged()
{
	strokePath.clear();
	strokeType.createStrokedPath (strokePath, path, AffineTransform::identity, 4.0f);

	setBoundsToEnclose (getDrawableBounds());
	repaint();
}

Rectangle<float> DrawableShape::getDrawableBounds() const
{
	if (isStrokeVisible())
		return strokePath.getBounds();
	else
		return path.getBounds();
}

bool DrawableShape::hitTest (int x, int y)
{
	bool allowsClicksOnThisComponent, allowsClicksOnChildComponents;
	getInterceptsMouseClicks (allowsClicksOnThisComponent, allowsClicksOnChildComponents);

	if (! allowsClicksOnThisComponent)
		return false;

	const float globalX = (float) (x - originRelativeToComponent.getX());
	const float globalY = (float) (y - originRelativeToComponent.getY());

	return path.contains (globalX, globalY)
			|| (isStrokeVisible() && strokePath.contains (globalX, globalY));
}

DrawableShape::RelativeFillType::RelativeFillType()
{
}

DrawableShape::RelativeFillType::RelativeFillType (const FillType& fill_)
	: fill (fill_)
{
	if (fill.isGradient())
	{
		const ColourGradient& g = *fill.gradient;

		gradientPoint1 = g.point1.transformedBy (fill.transform);
		gradientPoint2 = g.point2.transformedBy (fill.transform);
		gradientPoint3 = Point<float> (g.point1.getX() + g.point2.getY() - g.point1.getY(),
									   g.point1.getY() + g.point1.getX() - g.point2.getX())
							.transformedBy (fill.transform);
		fill.transform = AffineTransform::identity;
	}
}

DrawableShape::RelativeFillType::RelativeFillType (const RelativeFillType& other)
	: fill (other.fill),
	  gradientPoint1 (other.gradientPoint1),
	  gradientPoint2 (other.gradientPoint2),
	  gradientPoint3 (other.gradientPoint3)
{
}

DrawableShape::RelativeFillType& DrawableShape::RelativeFillType::operator= (const RelativeFillType& other)
{
	fill = other.fill;
	gradientPoint1 = other.gradientPoint1;
	gradientPoint2 = other.gradientPoint2;
	gradientPoint3 = other.gradientPoint3;
	return *this;
}

bool DrawableShape::RelativeFillType::operator== (const RelativeFillType& other) const
{
	return fill == other.fill
		&& ((! fill.isGradient())
			 || (gradientPoint1 == other.gradientPoint1
				 && gradientPoint2 == other.gradientPoint2
				 && gradientPoint3 == other.gradientPoint3));
}

bool DrawableShape::RelativeFillType::operator!= (const RelativeFillType& other) const
{
	return ! operator== (other);
}

bool DrawableShape::RelativeFillType::recalculateCoords (Expression::Scope* scope)
{
	if (fill.isGradient())
	{
		const Point<float> g1 (gradientPoint1.resolve (scope));
		const Point<float> g2 (gradientPoint2.resolve (scope));
		AffineTransform t;

		ColourGradient& g = *fill.gradient;

		if (g.isRadial)
		{
			const Point<float> g3 (gradientPoint3.resolve (scope));
			const Point<float> g3Source (g1.getX() + g2.getY() - g1.getY(),
										 g1.getY() + g1.getX() - g2.getX());

			t = AffineTransform::fromTargetPoints (g1.getX(), g1.getY(), g1.getX(), g1.getY(),
												   g2.getX(), g2.getY(), g2.getX(), g2.getY(),
												   g3Source.getX(), g3Source.getY(), g3.getX(), g3.getY());
		}

		if (g.point1 != g1 || g.point2 != g2 || fill.transform != t)
		{
			g.point1 = g1;
			g.point2 = g2;
			fill.transform = t;
			return true;
		}
	}

	return false;
}

bool DrawableShape::RelativeFillType::isDynamic() const
{
	return gradientPoint1.isDynamic() || gradientPoint2.isDynamic() || gradientPoint3.isDynamic();
}

void DrawableShape::RelativeFillType::writeTo (ValueTree& v, ComponentBuilder::ImageProvider* imageProvider, UndoManager* undoManager) const
{
	if (fill.isColour())
	{
		v.setProperty (FillAndStrokeState::type, "solid", undoManager);
		v.setProperty (FillAndStrokeState::colour, String::toHexString ((int) fill.colour.getARGB()), undoManager);
	}
	else if (fill.isGradient())
	{
		v.setProperty (FillAndStrokeState::type, "gradient", undoManager);
		v.setProperty (FillAndStrokeState::gradientPoint1, gradientPoint1.toString(), undoManager);
		v.setProperty (FillAndStrokeState::gradientPoint2, gradientPoint2.toString(), undoManager);
		v.setProperty (FillAndStrokeState::gradientPoint3, gradientPoint3.toString(), undoManager);

		const ColourGradient& cg = *fill.gradient;
		v.setProperty (FillAndStrokeState::radial, cg.isRadial, undoManager);

		String s;
		for (int i = 0; i < cg.getNumColours(); ++i)
			s << ' ' << cg.getColourPosition (i)
			  << ' ' << String::toHexString ((int) cg.getColour(i).getARGB());

		v.setProperty (FillAndStrokeState::colours, s.trimStart(), undoManager);
	}
	else if (fill.isTiledImage())
	{
		v.setProperty (FillAndStrokeState::type, "image", undoManager);

		if (imageProvider != nullptr)
			v.setProperty (FillAndStrokeState::imageId, imageProvider->getIdentifierForImage (fill.image), undoManager);

		if (fill.getOpacity() < 1.0f)
			v.setProperty (FillAndStrokeState::imageOpacity, fill.getOpacity(), undoManager);
		else
			v.removeProperty (FillAndStrokeState::imageOpacity, undoManager);
	}
	else
	{
		jassertfalse;
	}
}

bool DrawableShape::RelativeFillType::readFrom (const ValueTree& v, ComponentBuilder::ImageProvider* imageProvider)
{
	const String newType (v [FillAndStrokeState::type].toString());

	if (newType == "solid")
	{
		const String colourString (v [FillAndStrokeState::colour].toString());
		fill.setColour (Colour (colourString.isEmpty() ? (uint32) 0xff000000
													   : (uint32) colourString.getHexValue32()));
		return true;
	}
	else if (newType == "gradient")
	{
		ColourGradient g;
		g.isRadial = v [FillAndStrokeState::radial];

		StringArray colourSteps;
		colourSteps.addTokens (v [FillAndStrokeState::colours].toString(), false);

		for (int i = 0; i < colourSteps.size() / 2; ++i)
			g.addColour (colourSteps[i * 2].getDoubleValue(),
						 Colour ((uint32)  colourSteps[i * 2 + 1].getHexValue32()));

		fill.setGradient (g);

		gradientPoint1 = RelativePoint (v [FillAndStrokeState::gradientPoint1]);
		gradientPoint2 = RelativePoint (v [FillAndStrokeState::gradientPoint2]);
		gradientPoint3 = RelativePoint (v [FillAndStrokeState::gradientPoint3]);
		return true;
	}
	else if (newType == "image")
	{
		Image im;
		if (imageProvider != nullptr)
			im = imageProvider->getImageForIdentifier (v [FillAndStrokeState::imageId]);

		fill.setTiledImage (im, AffineTransform::identity);
		fill.setOpacity ((float) v.getProperty (FillAndStrokeState::imageOpacity, 1.0f));
		return true;
	}

	jassertfalse;
	return false;
}

const Identifier DrawableShape::FillAndStrokeState::type ("type");
const Identifier DrawableShape::FillAndStrokeState::colour ("colour");
const Identifier DrawableShape::FillAndStrokeState::colours ("colours");
const Identifier DrawableShape::FillAndStrokeState::fill ("Fill");
const Identifier DrawableShape::FillAndStrokeState::stroke ("Stroke");
const Identifier DrawableShape::FillAndStrokeState::path ("Path");
const Identifier DrawableShape::FillAndStrokeState::jointStyle ("jointStyle");
const Identifier DrawableShape::FillAndStrokeState::capStyle ("capStyle");
const Identifier DrawableShape::FillAndStrokeState::strokeWidth ("strokeWidth");
const Identifier DrawableShape::FillAndStrokeState::gradientPoint1 ("point1");
const Identifier DrawableShape::FillAndStrokeState::gradientPoint2 ("point2");
const Identifier DrawableShape::FillAndStrokeState::gradientPoint3 ("point3");
const Identifier DrawableShape::FillAndStrokeState::radial ("radial");
const Identifier DrawableShape::FillAndStrokeState::imageId ("imageId");
const Identifier DrawableShape::FillAndStrokeState::imageOpacity ("imageOpacity");

DrawableShape::FillAndStrokeState::FillAndStrokeState (const ValueTree& state_)
	: Drawable::ValueTreeWrapperBase (state_)
{
}

DrawableShape::RelativeFillType DrawableShape::FillAndStrokeState::getFill (const Identifier& fillOrStrokeType, ComponentBuilder::ImageProvider* imageProvider) const
{
	DrawableShape::RelativeFillType f;
	f.readFrom (state.getChildWithName (fillOrStrokeType), imageProvider);
	return f;
}

ValueTree DrawableShape::FillAndStrokeState::getFillState (const Identifier& fillOrStrokeType)
{
	ValueTree v (state.getChildWithName (fillOrStrokeType));
	if (v.isValid())
		return v;

	setFill (fillOrStrokeType, FillType (Colours::black), nullptr, nullptr);
	return getFillState (fillOrStrokeType);
}

void DrawableShape::FillAndStrokeState::setFill (const Identifier& fillOrStrokeType, const RelativeFillType& newFill,
												 ComponentBuilder::ImageProvider* imageProvider, UndoManager* undoManager)
{
	ValueTree v (state.getOrCreateChildWithName (fillOrStrokeType, undoManager));
	newFill.writeTo (v, imageProvider, undoManager);
}

PathStrokeType DrawableShape::FillAndStrokeState::getStrokeType() const
{
	const String jointStyleString (state [jointStyle].toString());
	const String capStyleString (state [capStyle].toString());

	return PathStrokeType (state [strokeWidth],
						   jointStyleString == "curved" ? PathStrokeType::curved
														: (jointStyleString == "bevel" ? PathStrokeType::beveled
																					   : PathStrokeType::mitered),
						   capStyleString == "square" ? PathStrokeType::square
													  : (capStyleString == "round" ? PathStrokeType::rounded
																				   : PathStrokeType::butt));
}

void DrawableShape::FillAndStrokeState::setStrokeType (const PathStrokeType& newStrokeType, UndoManager* undoManager)
{
	state.setProperty (strokeWidth, (double) newStrokeType.getStrokeThickness(), undoManager);
	state.setProperty (jointStyle, newStrokeType.getJointStyle() == PathStrokeType::mitered
									 ? "miter" : (newStrokeType.getJointStyle() == PathStrokeType::curved ? "curved" : "bevel"), undoManager);
	state.setProperty (capStyle, newStrokeType.getEndStyle() == PathStrokeType::butt
									 ? "butt" : (newStrokeType.getEndStyle() == PathStrokeType::square ? "square" : "round"), undoManager);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DrawableShape.cpp ***/


/*** Start of inlined file: juce_DrawableComposite.cpp ***/
BEGIN_JUCE_NAMESPACE

DrawableComposite::DrawableComposite()
	: bounds (Point<float>(), Point<float> (100.0f, 0.0f), Point<float> (0.0f, 100.0f)),
	  updateBoundsReentrant (false)
{
	setContentArea (RelativeRectangle (RelativeCoordinate (0.0),
									   RelativeCoordinate (100.0),
									   RelativeCoordinate (0.0),
									   RelativeCoordinate (100.0)));
}

DrawableComposite::DrawableComposite (const DrawableComposite& other)
	: bounds (other.bounds),
	  markersX (other.markersX),
	  markersY (other.markersY),
	  updateBoundsReentrant (false)
{
	for (int i = 0; i < other.getNumChildComponents(); ++i)
	{
		const Drawable* const d = dynamic_cast <const Drawable*> (other.getChildComponent(i));

		if (d != nullptr)
			addAndMakeVisible (d->createCopy());
	}
}

DrawableComposite::~DrawableComposite()
{
	deleteAllChildren();
}

Drawable* DrawableComposite::createCopy() const
{
	return new DrawableComposite (*this);
}

Rectangle<float> DrawableComposite::getDrawableBounds() const
{
	Rectangle<float> r;

	for (int i = getNumChildComponents(); --i >= 0;)
	{
		const Drawable* const d = dynamic_cast <const Drawable*> (getChildComponent(i));

		if (d != nullptr)
			r = r.getUnion (d->isTransformed() ? d->getDrawableBounds().transformed (d->getTransform())
											   : d->getDrawableBounds());
	}

	return r;
}

MarkerList* DrawableComposite::getMarkers (bool xAxis)
{
	return xAxis ? &markersX : &markersY;
}

RelativeRectangle DrawableComposite::getContentArea() const
{
	jassert (markersX.getNumMarkers() >= 2 && markersX.getMarker (0)->name == contentLeftMarkerName && markersX.getMarker (1)->name == contentRightMarkerName);
	jassert (markersY.getNumMarkers() >= 2 && markersY.getMarker (0)->name == contentTopMarkerName && markersY.getMarker (1)->name == contentBottomMarkerName);

	return RelativeRectangle (markersX.getMarker(0)->position, markersX.getMarker(1)->position,
							  markersY.getMarker(0)->position, markersY.getMarker(1)->position);
}

void DrawableComposite::setContentArea (const RelativeRectangle& newArea)
{
	markersX.setMarker (contentLeftMarkerName, newArea.left);
	markersX.setMarker (contentRightMarkerName, newArea.right);
	markersY.setMarker (contentTopMarkerName, newArea.top);
	markersY.setMarker (contentBottomMarkerName, newArea.bottom);
}

void DrawableComposite::setBoundingBox (const RelativeParallelogram& newBounds)
{
	if (bounds != newBounds)
	{
		bounds = newBounds;

		if (bounds.isDynamic())
		{
			Drawable::Positioner<DrawableComposite>* const p = new Drawable::Positioner<DrawableComposite> (*this);
			setPositioner (p);
			p->apply();
		}
		else
		{
			setPositioner (nullptr);
			recalculateCoordinates (nullptr);
		}
	}
}

void DrawableComposite::resetBoundingBoxToContentArea()
{
	const RelativeRectangle content (getContentArea());

	setBoundingBox (RelativeParallelogram (RelativePoint (content.left, content.top),
										   RelativePoint (content.right, content.top),
										   RelativePoint (content.left, content.bottom)));
}

void DrawableComposite::resetContentAreaAndBoundingBoxToFitChildren()
{
	const Rectangle<float> activeArea (getDrawableBounds());

	setContentArea (RelativeRectangle (RelativeCoordinate (activeArea.getX()),
									   RelativeCoordinate (activeArea.getRight()),
									   RelativeCoordinate (activeArea.getY()),
									   RelativeCoordinate (activeArea.getBottom())));
	resetBoundingBoxToContentArea();
}

bool DrawableComposite::registerCoordinates (RelativeCoordinatePositionerBase& pos)
{
	bool ok = pos.addPoint (bounds.topLeft);
	ok = pos.addPoint (bounds.topRight) && ok;
	return pos.addPoint (bounds.bottomLeft) && ok;
}

void DrawableComposite::recalculateCoordinates (Expression::Scope* scope)
{
	Point<float> resolved[3];
	bounds.resolveThreePoints (resolved, scope);

	const Rectangle<float> content (getContentArea().resolve (scope));

	AffineTransform t (AffineTransform::fromTargetPoints (content.getX(), content.getY(), resolved[0].getX(), resolved[0].getY(),
														  content.getRight(), content.getY(), resolved[1].getX(), resolved[1].getY(),
														  content.getX(), content.getBottom(), resolved[2].getX(), resolved[2].getY()));

	if (t.isSingularity())
		t = AffineTransform::identity;

	setTransform (t);
}

void DrawableComposite::parentHierarchyChanged()
{
	DrawableComposite* parent = getParent();
	if (parent != nullptr)
		originRelativeToComponent = parent->originRelativeToComponent - getPosition();
}

void DrawableComposite::childBoundsChanged (Component*)
{
	updateBoundsToFitChildren();
}

void DrawableComposite::childrenChanged()
{
	updateBoundsToFitChildren();
}

void DrawableComposite::updateBoundsToFitChildren()
{
	if (! updateBoundsReentrant)
	{
		const ScopedValueSetter<bool> setter (updateBoundsReentrant, true, false);

		Rectangle<int> childArea;

		for (int i = getNumChildComponents(); --i >= 0;)
			childArea = childArea.getUnion (getChildComponent(i)->getBoundsInParent());

		const Point<int> delta (childArea.getPosition());
		childArea += getPosition();

		if (childArea != getBounds())
		{
			if (! delta.isOrigin())
			{
				originRelativeToComponent -= delta;

				for (int i = getNumChildComponents(); --i >= 0;)
				{
					Component* const c = getChildComponent(i);

					if (c != nullptr)
						c->setBounds (c->getBounds() - delta);
				}
			}

			setBounds (childArea);
		}
	}
}

const char* const DrawableComposite::contentLeftMarkerName = "left";
const char* const DrawableComposite::contentRightMarkerName = "right";
const char* const DrawableComposite::contentTopMarkerName = "top";
const char* const DrawableComposite::contentBottomMarkerName = "bottom";

const Identifier DrawableComposite::valueTreeType ("Group");

const Identifier DrawableComposite::ValueTreeWrapper::topLeft ("topLeft");
const Identifier DrawableComposite::ValueTreeWrapper::topRight ("topRight");
const Identifier DrawableComposite::ValueTreeWrapper::bottomLeft ("bottomLeft");
const Identifier DrawableComposite::ValueTreeWrapper::childGroupTag ("Drawables");
const Identifier DrawableComposite::ValueTreeWrapper::markerGroupTagX ("MarkersX");
const Identifier DrawableComposite::ValueTreeWrapper::markerGroupTagY ("MarkersY");

DrawableComposite::ValueTreeWrapper::ValueTreeWrapper (const ValueTree& state_)
	: ValueTreeWrapperBase (state_)
{
	jassert (state.hasType (valueTreeType));
}

ValueTree DrawableComposite::ValueTreeWrapper::getChildList() const
{
	return state.getChildWithName (childGroupTag);
}

ValueTree DrawableComposite::ValueTreeWrapper::getChildListCreating (UndoManager* undoManager)
{
	return state.getOrCreateChildWithName (childGroupTag, undoManager);
}

RelativeParallelogram DrawableComposite::ValueTreeWrapper::getBoundingBox() const
{
	return RelativeParallelogram (state.getProperty (topLeft, "0, 0"),
								  state.getProperty (topRight, "100, 0"),
								  state.getProperty (bottomLeft, "0, 100"));
}

void DrawableComposite::ValueTreeWrapper::setBoundingBox (const RelativeParallelogram& newBounds, UndoManager* undoManager)
{
	state.setProperty (topLeft, newBounds.topLeft.toString(), undoManager);
	state.setProperty (topRight, newBounds.topRight.toString(), undoManager);
	state.setProperty (bottomLeft, newBounds.bottomLeft.toString(), undoManager);
}

void DrawableComposite::ValueTreeWrapper::resetBoundingBoxToContentArea (UndoManager* undoManager)
{
	const RelativeRectangle content (getContentArea());

	setBoundingBox (RelativeParallelogram (RelativePoint (content.left, content.top),
										   RelativePoint (content.right, content.top),
										   RelativePoint (content.left, content.bottom)), undoManager);
}

RelativeRectangle DrawableComposite::ValueTreeWrapper::getContentArea() const
{
	MarkerList::ValueTreeWrapper markersX (getMarkerList (true));
	MarkerList::ValueTreeWrapper markersY (getMarkerList (false));

	return RelativeRectangle (markersX.getMarker (markersX.getMarkerState (0)).position,
							  markersX.getMarker (markersX.getMarkerState (1)).position,
							  markersY.getMarker (markersY.getMarkerState (0)).position,
							  markersY.getMarker (markersY.getMarkerState (1)).position);
}

void DrawableComposite::ValueTreeWrapper::setContentArea (const RelativeRectangle& newArea, UndoManager* undoManager)
{
	MarkerList::ValueTreeWrapper markersX (getMarkerListCreating (true, nullptr));
	MarkerList::ValueTreeWrapper markersY (getMarkerListCreating (false, nullptr));

	markersX.setMarker (MarkerList::Marker (contentLeftMarkerName, newArea.left), undoManager);
	markersX.setMarker (MarkerList::Marker (contentRightMarkerName, newArea.right), undoManager);
	markersY.setMarker (MarkerList::Marker (contentTopMarkerName, newArea.top), undoManager);
	markersY.setMarker (MarkerList::Marker (contentBottomMarkerName, newArea.bottom), undoManager);
}

MarkerList::ValueTreeWrapper DrawableComposite::ValueTreeWrapper::getMarkerList (bool xAxis) const
{
	return state.getChildWithName (xAxis ? markerGroupTagX : markerGroupTagY);
}

MarkerList::ValueTreeWrapper DrawableComposite::ValueTreeWrapper::getMarkerListCreating (bool xAxis, UndoManager* undoManager)
{
	return state.getOrCreateChildWithName (xAxis ? markerGroupTagX : markerGroupTagY, undoManager);
}

void DrawableComposite::refreshFromValueTree (const ValueTree& tree, ComponentBuilder& builder)
{
	const ValueTreeWrapper wrapper (tree);
	setComponentID (wrapper.getID());

	wrapper.getMarkerList (true).applyTo (markersX);
	wrapper.getMarkerList (false).applyTo (markersY);

	setBoundingBox (wrapper.getBoundingBox());

	builder.updateChildComponents (*this, wrapper.getChildList());
}

ValueTree DrawableComposite::createValueTree (ComponentBuilder::ImageProvider* imageProvider) const
{
	ValueTree tree (valueTreeType);
	ValueTreeWrapper v (tree);

	v.setID (getComponentID());
	v.setBoundingBox (bounds, nullptr);

	ValueTree childList (v.getChildListCreating (nullptr));

	for (int i = 0; i < getNumChildComponents(); ++i)
	{
		const Drawable* const d = dynamic_cast <const Drawable*> (getChildComponent(i));
		jassert (d != nullptr); // You can't save a mix of Drawables and normal components!

		childList.addChild (d->createValueTree (imageProvider), -1, nullptr);
	}

	v.getMarkerListCreating (true, nullptr).readFrom (markersX, nullptr);
	v.getMarkerListCreating (false, nullptr).readFrom (markersY, nullptr);

	return tree;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DrawableComposite.cpp ***/


/*** Start of inlined file: juce_DrawableImage.cpp ***/
BEGIN_JUCE_NAMESPACE

DrawableImage::DrawableImage()
	: opacity (1.0f),
	  overlayColour (0x00000000)
{
	bounds.topRight = RelativePoint (Point<float> (1.0f, 0.0f));
	bounds.bottomLeft = RelativePoint (Point<float> (0.0f, 1.0f));
}

DrawableImage::DrawableImage (const DrawableImage& other)
	: image (other.image),
	  opacity (other.opacity),
	  overlayColour (other.overlayColour),
	  bounds (other.bounds)
{
}

DrawableImage::~DrawableImage()
{
}

void DrawableImage::setImage (const Image& imageToUse)
{
	image = imageToUse;
	setBounds (imageToUse.getBounds());

	bounds.topLeft = RelativePoint (Point<float> (0.0f, 0.0f));
	bounds.topRight = RelativePoint (Point<float> ((float) image.getWidth(), 0.0f));
	bounds.bottomLeft = RelativePoint (Point<float> (0.0f, (float) image.getHeight()));
	recalculateCoordinates (nullptr);

	repaint();
}

void DrawableImage::setOpacity (const float newOpacity)
{
	opacity = newOpacity;
}

void DrawableImage::setOverlayColour (const Colour& newOverlayColour)
{
	overlayColour = newOverlayColour;
}

void DrawableImage::setBoundingBox (const RelativeParallelogram& newBounds)
{
	if (bounds != newBounds)
	{
		bounds = newBounds;

		if (bounds.isDynamic())
		{
			Drawable::Positioner<DrawableImage>* const p = new Drawable::Positioner<DrawableImage> (*this);
			setPositioner (p);
			p->apply();
		}
		else
		{
			setPositioner (nullptr);
			recalculateCoordinates (nullptr);
		}
	}
}

bool DrawableImage::registerCoordinates (RelativeCoordinatePositionerBase& pos)
{
	bool ok = pos.addPoint (bounds.topLeft);
	ok = pos.addPoint (bounds.topRight) && ok;
	return pos.addPoint (bounds.bottomLeft) && ok;
}

void DrawableImage::recalculateCoordinates (Expression::Scope* scope)
{
	if (image.isValid())
	{
		Point<float> resolved[3];
		bounds.resolveThreePoints (resolved, scope);

		const Point<float> tr (resolved[0] + (resolved[1] - resolved[0]) / (float) image.getWidth());
		const Point<float> bl (resolved[0] + (resolved[2] - resolved[0]) / (float) image.getHeight());

		AffineTransform t (AffineTransform::fromTargetPoints (resolved[0].getX(), resolved[0].getY(),
															  tr.getX(), tr.getY(),
															  bl.getX(), bl.getY()));

		if (t.isSingularity())
			t = AffineTransform::identity;

		setTransform (t);
	}
}

void DrawableImage::paint (Graphics& g)
{
	if (image.isValid())
	{
		if (opacity > 0.0f && ! overlayColour.isOpaque())
		{
			g.setOpacity (opacity);
			g.drawImageAt (image, 0, 0, false);
		}

		if (! overlayColour.isTransparent())
		{
			g.setColour (overlayColour.withMultipliedAlpha (opacity));
			g.drawImageAt (image, 0, 0, true);
		}
	}
}

Rectangle<float> DrawableImage::getDrawableBounds() const
{
	return image.getBounds().toFloat();
}

bool DrawableImage::hitTest (int x, int y)
{
	return Drawable::hitTest (x, y) && image.isValid() && image.getPixelAt (x, y).getAlpha() >= 127;
}

Drawable* DrawableImage::createCopy() const
{
	return new DrawableImage (*this);
}

const Identifier DrawableImage::valueTreeType ("Image");

const Identifier DrawableImage::ValueTreeWrapper::opacity ("opacity");
const Identifier DrawableImage::ValueTreeWrapper::overlay ("overlay");
const Identifier DrawableImage::ValueTreeWrapper::image ("image");
const Identifier DrawableImage::ValueTreeWrapper::topLeft ("topLeft");
const Identifier DrawableImage::ValueTreeWrapper::topRight ("topRight");
const Identifier DrawableImage::ValueTreeWrapper::bottomLeft ("bottomLeft");

DrawableImage::ValueTreeWrapper::ValueTreeWrapper (const ValueTree& state_)
	: ValueTreeWrapperBase (state_)
{
	jassert (state.hasType (valueTreeType));
}

var DrawableImage::ValueTreeWrapper::getImageIdentifier() const
{
	return state [image];
}

Value DrawableImage::ValueTreeWrapper::getImageIdentifierValue (UndoManager* undoManager)
{
	return state.getPropertyAsValue (image, undoManager);
}

void DrawableImage::ValueTreeWrapper::setImageIdentifier (const var& newIdentifier, UndoManager* undoManager)
{
	state.setProperty (image, newIdentifier, undoManager);
}

float DrawableImage::ValueTreeWrapper::getOpacity() const
{
	return (float) state.getProperty (opacity, 1.0);
}

Value DrawableImage::ValueTreeWrapper::getOpacityValue (UndoManager* undoManager)
{
	if (! state.hasProperty (opacity))
		state.setProperty (opacity, 1.0, undoManager);

	return state.getPropertyAsValue (opacity, undoManager);
}

void DrawableImage::ValueTreeWrapper::setOpacity (float newOpacity, UndoManager* undoManager)
{
	state.setProperty (opacity, newOpacity, undoManager);
}

const Colour DrawableImage::ValueTreeWrapper::getOverlayColour() const
{
	return Colour (state [overlay].toString().getHexValue32());
}

void DrawableImage::ValueTreeWrapper::setOverlayColour (const Colour& newColour, UndoManager* undoManager)
{
	if (newColour.isTransparent())
		state.removeProperty (overlay, undoManager);
	else
		state.setProperty (overlay, String::toHexString ((int) newColour.getARGB()), undoManager);
}

Value DrawableImage::ValueTreeWrapper::getOverlayColourValue (UndoManager* undoManager)
{
	return state.getPropertyAsValue (overlay, undoManager);
}

RelativeParallelogram DrawableImage::ValueTreeWrapper::getBoundingBox() const
{
	return RelativeParallelogram (state.getProperty (topLeft, "0, 0"),
								  state.getProperty (topRight, "100, 0"),
								  state.getProperty (bottomLeft, "0, 100"));
}

void DrawableImage::ValueTreeWrapper::setBoundingBox (const RelativeParallelogram& newBounds, UndoManager* undoManager)
{
	state.setProperty (topLeft, newBounds.topLeft.toString(), undoManager);
	state.setProperty (topRight, newBounds.topRight.toString(), undoManager);
	state.setProperty (bottomLeft, newBounds.bottomLeft.toString(), undoManager);
}

void DrawableImage::refreshFromValueTree (const ValueTree& tree, ComponentBuilder& builder)
{
	const ValueTreeWrapper controller (tree);
	setComponentID (controller.getID());

	const float newOpacity = controller.getOpacity();
	const Colour newOverlayColour (controller.getOverlayColour());

	Image newImage;
	const var imageIdentifier (controller.getImageIdentifier());

	jassert (builder.getImageProvider() != 0 || imageIdentifier.isVoid()); // if you're using images, you need to provide something that can load and save them!

	if (builder.getImageProvider() != nullptr)
		newImage = builder.getImageProvider()->getImageForIdentifier (imageIdentifier);

	const RelativeParallelogram newBounds (controller.getBoundingBox());

	if (bounds != newBounds || newOpacity != opacity
		 || overlayColour != newOverlayColour || image != newImage)
	{
		repaint();
		opacity = newOpacity;
		overlayColour = newOverlayColour;

		if (image != newImage)
			setImage (newImage);

		setBoundingBox (newBounds);
	}
}

ValueTree DrawableImage::createValueTree (ComponentBuilder::ImageProvider* imageProvider) const
{
	ValueTree tree (valueTreeType);
	ValueTreeWrapper v (tree);

	v.setID (getComponentID());
	v.setOpacity (opacity, nullptr);
	v.setOverlayColour (overlayColour, nullptr);
	v.setBoundingBox (bounds, nullptr);

	if (image.isValid())
	{
		jassert (imageProvider != nullptr); // if you're using images, you need to provide something that can load and save them!

		if (imageProvider != nullptr)
			v.setImageIdentifier (imageProvider->getIdentifierForImage (image), nullptr);
	}

	return tree;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DrawableImage.cpp ***/


/*** Start of inlined file: juce_DrawablePath.cpp ***/
BEGIN_JUCE_NAMESPACE

DrawablePath::DrawablePath()
{
}

DrawablePath::DrawablePath (const DrawablePath& other)
	: DrawableShape (other)
{
	if (other.relativePath != nullptr)
		setPath (*other.relativePath);
	else
		setPath (other.path);
}

DrawablePath::~DrawablePath()
{
}

Drawable* DrawablePath::createCopy() const
{
	return new DrawablePath (*this);
}

void DrawablePath::setPath (const Path& newPath)
{
	path = newPath;
	pathChanged();
}

const Path& DrawablePath::getPath() const
{
	return path;
}

const Path& DrawablePath::getStrokePath() const
{
	return strokePath;
}

void DrawablePath::applyRelativePath (const RelativePointPath& newRelativePath, Expression::Scope* scope)
{
	Path newPath;
	newRelativePath.createPath (newPath, scope);

	if (path != newPath)
	{
		path.swapWithPath (newPath);
		pathChanged();
	}
}

class DrawablePath::RelativePositioner  : public RelativeCoordinatePositionerBase
{
public:
	RelativePositioner (DrawablePath& component_)
		: RelativeCoordinatePositionerBase (component_),
		  owner (component_)
	{
	}

	bool registerCoordinates()
	{
		bool ok = true;

		jassert (owner.relativePath != nullptr);
		const RelativePointPath& path = *owner.relativePath;

		for (int i = 0; i < path.elements.size(); ++i)
		{
			RelativePointPath::ElementBase* const e = path.elements.getUnchecked(i);

			int numPoints;
			RelativePoint* const points = e->getControlPoints (numPoints);

			for (int j = numPoints; --j >= 0;)
				ok = addPoint (points[j]) && ok;
		}

		return ok;
	}

	void applyToComponentBounds()
	{
		jassert (owner.relativePath != nullptr);

		ComponentScope scope (getComponent());
		owner.applyRelativePath (*owner.relativePath, &scope);
	}

	void applyNewBounds (const Rectangle<int>&)
	{
		jassertfalse; // drawables can't be resized directly!
	}

private:
	DrawablePath& owner;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (RelativePositioner);
};

void DrawablePath::setPath (const RelativePointPath& newRelativePath)
{
	if (newRelativePath.containsAnyDynamicPoints())
	{
		if (relativePath == nullptr || newRelativePath != *relativePath)
		{
			relativePath = new RelativePointPath (newRelativePath);

			RelativePositioner* const p = new RelativePositioner (*this);
			setPositioner (p);
			p->apply();
		}
	}
	else
	{
		relativePath = nullptr;
		applyRelativePath (newRelativePath, nullptr);
	}
}

const Identifier DrawablePath::valueTreeType ("Path");

const Identifier DrawablePath::ValueTreeWrapper::nonZeroWinding ("nonZeroWinding");
const Identifier DrawablePath::ValueTreeWrapper::point1 ("p1");
const Identifier DrawablePath::ValueTreeWrapper::point2 ("p2");
const Identifier DrawablePath::ValueTreeWrapper::point3 ("p3");

DrawablePath::ValueTreeWrapper::ValueTreeWrapper (const ValueTree& state_)
	: FillAndStrokeState (state_)
{
	jassert (state.hasType (valueTreeType));
}

ValueTree DrawablePath::ValueTreeWrapper::getPathState()
{
	return state.getOrCreateChildWithName (path, nullptr);
}

bool DrawablePath::ValueTreeWrapper::usesNonZeroWinding() const
{
	return state [nonZeroWinding];
}

void DrawablePath::ValueTreeWrapper::setUsesNonZeroWinding (bool b, UndoManager* undoManager)
{
	state.setProperty (nonZeroWinding, b, undoManager);
}

void DrawablePath::ValueTreeWrapper::readFrom (const RelativePointPath& relativePath, UndoManager* undoManager)
{
	setUsesNonZeroWinding (relativePath.usesNonZeroWinding, undoManager);

	ValueTree pathTree (getPathState());
	pathTree.removeAllChildren (undoManager);

	for (int i = 0; i < relativePath.elements.size(); ++i)
		pathTree.addChild (relativePath.elements.getUnchecked(i)->createTree(), -1, undoManager);
}

void DrawablePath::ValueTreeWrapper::writeTo (RelativePointPath& relativePath) const
{
	relativePath.usesNonZeroWinding = usesNonZeroWinding();
	RelativePoint points[3];

	const ValueTree pathTree (state.getChildWithName (path));
	const int num = pathTree.getNumChildren();
	for (int i = 0; i < num; ++i)
	{
		const Element e (pathTree.getChild(i));

		const int numCps = e.getNumControlPoints();
		for (int j = 0; j < numCps; ++j)
			points[j] = e.getControlPoint (j);

		RelativePointPath::ElementBase* newElement = 0;
		const Identifier t (e.getType());

		if	  (t == Element::startSubPathElement)  newElement = new RelativePointPath::StartSubPath (points[0]);
		else if (t == Element::closeSubPathElement)  newElement = new RelativePointPath::CloseSubPath();
		else if (t == Element::lineToElement)	newElement = new RelativePointPath::LineTo (points[0]);
		else if (t == Element::quadraticToElement)   newElement = new RelativePointPath::QuadraticTo (points[0], points[1]);
		else if (t == Element::cubicToElement)	   newElement = new RelativePointPath::CubicTo (points[0], points[1], points[2]);
		else					 jassertfalse;

		relativePath.addElement (newElement);
	}
}

const Identifier DrawablePath::ValueTreeWrapper::Element::mode ("mode");
const Identifier DrawablePath::ValueTreeWrapper::Element::startSubPathElement ("Move");
const Identifier DrawablePath::ValueTreeWrapper::Element::closeSubPathElement ("Close");
const Identifier DrawablePath::ValueTreeWrapper::Element::lineToElement ("Line");
const Identifier DrawablePath::ValueTreeWrapper::Element::quadraticToElement ("Quad");
const Identifier DrawablePath::ValueTreeWrapper::Element::cubicToElement ("Cubic");

const char* DrawablePath::ValueTreeWrapper::Element::cornerMode = "corner";
const char* DrawablePath::ValueTreeWrapper::Element::roundedMode = "round";
const char* DrawablePath::ValueTreeWrapper::Element::symmetricMode = "symm";

DrawablePath::ValueTreeWrapper::Element::Element (const ValueTree& state_)
	: state (state_)
{
}

DrawablePath::ValueTreeWrapper::Element::~Element()
{
}

DrawablePath::ValueTreeWrapper DrawablePath::ValueTreeWrapper::Element::getParent() const
{
	return ValueTreeWrapper (state.getParent().getParent());
}

DrawablePath::ValueTreeWrapper::Element DrawablePath::ValueTreeWrapper::Element::getPreviousElement() const
{
	return Element (state.getSibling (-1));
}

int DrawablePath::ValueTreeWrapper::Element::getNumControlPoints() const noexcept
{
	const Identifier i (state.getType());
	if (i == startSubPathElement || i == lineToElement) return 1;
	if (i == quadraticToElement) return 2;
	if (i == cubicToElement) return 3;
	return 0;
}

RelativePoint DrawablePath::ValueTreeWrapper::Element::getControlPoint (const int index) const
{
	jassert (index >= 0 && index < getNumControlPoints());
	return RelativePoint (state [index == 0 ? point1 : (index == 1 ? point2 : point3)].toString());
}

Value DrawablePath::ValueTreeWrapper::Element::getControlPointValue (int index, UndoManager* undoManager) const
{
	jassert (index >= 0 && index < getNumControlPoints());
	return state.getPropertyAsValue (index == 0 ? point1 : (index == 1 ? point2 : point3), undoManager);
}

void DrawablePath::ValueTreeWrapper::Element::setControlPoint (const int index, const RelativePoint& point, UndoManager* undoManager)
{
	jassert (index >= 0 && index < getNumControlPoints());
	state.setProperty (index == 0 ? point1 : (index == 1 ? point2 : point3), point.toString(), undoManager);
}

RelativePoint DrawablePath::ValueTreeWrapper::Element::getStartPoint() const
{
	const Identifier i (state.getType());

	if (i == startSubPathElement)
		return getControlPoint (0);

	jassert (i == lineToElement || i == quadraticToElement || i == cubicToElement || i == closeSubPathElement);

	return getPreviousElement().getEndPoint();
}

RelativePoint DrawablePath::ValueTreeWrapper::Element::getEndPoint() const
{
	const Identifier i (state.getType());
	if (i == startSubPathElement || i == lineToElement)  return getControlPoint (0);
	if (i == quadraticToElement)			 return getControlPoint (1);
	if (i == cubicToElement)				 return getControlPoint (2);

	jassert (i == closeSubPathElement);
	return RelativePoint();
}

float DrawablePath::ValueTreeWrapper::Element::getLength (Expression::Scope* scope) const
{
	const Identifier i (state.getType());

	if (i == lineToElement || i == closeSubPathElement)
		return getEndPoint().resolve (scope).getDistanceFrom (getStartPoint().resolve (scope));

	if (i == cubicToElement)
	{
		Path p;
		p.startNewSubPath (getStartPoint().resolve (scope));
		p.cubicTo (getControlPoint (0).resolve (scope), getControlPoint (1).resolve (scope), getControlPoint (2).resolve (scope));
		return p.getLength();
	}

	if (i == quadraticToElement)
	{
		Path p;
		p.startNewSubPath (getStartPoint().resolve (scope));
		p.quadraticTo (getControlPoint (0).resolve (scope), getControlPoint (1).resolve (scope));
		return p.getLength();
	}

	jassert (i == startSubPathElement);
	return 0;
}

String DrawablePath::ValueTreeWrapper::Element::getModeOfEndPoint() const
{
	return state [mode].toString();
}

void DrawablePath::ValueTreeWrapper::Element::setModeOfEndPoint (const String& newMode, UndoManager* undoManager)
{
	if (state.hasType (cubicToElement))
		state.setProperty (mode, newMode, undoManager);
}

void DrawablePath::ValueTreeWrapper::Element::convertToLine (UndoManager* undoManager)
{
	const Identifier i (state.getType());

	if (i == quadraticToElement || i == cubicToElement)
	{
		ValueTree newState (lineToElement);
		Element e (newState);
		e.setControlPoint (0, getEndPoint(), undoManager);
		state = newState;
	}
}

void DrawablePath::ValueTreeWrapper::Element::convertToCubic (Expression::Scope* scope, UndoManager* undoManager)
{
	const Identifier i (state.getType());

	if (i == lineToElement || i == quadraticToElement)
	{
		ValueTree newState (cubicToElement);
		Element e (newState);

		const RelativePoint start (getStartPoint());
		const RelativePoint end (getEndPoint());
		const Point<float> startResolved (start.resolve (scope));
		const Point<float> endResolved (end.resolve (scope));
		e.setControlPoint (0, startResolved + (endResolved - startResolved) * 0.3f, undoManager);
		e.setControlPoint (1, startResolved + (endResolved - startResolved) * 0.7f, undoManager);
		e.setControlPoint (2, end, undoManager);

		state = newState;
	}
}

void DrawablePath::ValueTreeWrapper::Element::convertToPathBreak (UndoManager* undoManager)
{
	const Identifier i (state.getType());

	if (i != startSubPathElement)
	{
		ValueTree newState (startSubPathElement);
		Element e (newState);
		e.setControlPoint (0, getEndPoint(), undoManager);
		state = newState;
	}
}

namespace DrawablePathHelpers
{
	Point<float> findCubicSubdivisionPoint (float proportion, const Point<float> points[4])
	{
		const Point<float> mid1 (points[0] + (points[1] - points[0]) * proportion),
						   mid2 (points[1] + (points[2] - points[1]) * proportion),
						   mid3 (points[2] + (points[3] - points[2]) * proportion);

		const Point<float> newCp1 (mid1 + (mid2 - mid1) * proportion),
						   newCp2 (mid2 + (mid3 - mid2) * proportion);

		return newCp1 + (newCp2 - newCp1) * proportion;
	}

	Point<float> findQuadraticSubdivisionPoint (float proportion, const Point<float> points[3])
	{
		const Point<float> mid1 (points[0] + (points[1] - points[0]) * proportion),
						   mid2 (points[1] + (points[2] - points[1]) * proportion);

		return mid1 + (mid2 - mid1) * proportion;
	}
}

float DrawablePath::ValueTreeWrapper::Element::findProportionAlongLine (const Point<float>& targetPoint, Expression::Scope* scope) const
{
	using namespace DrawablePathHelpers;
	const Identifier type (state.getType());
	float bestProp = 0;

	if (type == cubicToElement)
	{
		RelativePoint rp1 (getStartPoint()), rp2 (getControlPoint (0)), rp3 (getControlPoint (1)), rp4 (getEndPoint());

		const Point<float> points[] = { rp1.resolve (scope), rp2.resolve (scope), rp3.resolve (scope), rp4.resolve (scope) };

		float bestDistance = std::numeric_limits<float>::max();

		for (int i = 110; --i >= 0;)
		{
			float prop = i > 10 ? ((i - 10) / 100.0f) : (bestProp + ((i - 5) / 1000.0f));
			const Point<float> centre (findCubicSubdivisionPoint (prop, points));
			const float distance = centre.getDistanceFrom (targetPoint);

			if (distance < bestDistance)
			{
				bestProp = prop;
				bestDistance = distance;
			}
		}
	}
	else if (type == quadraticToElement)
	{
		RelativePoint rp1 (getStartPoint()), rp2 (getControlPoint (0)), rp3 (getEndPoint());
		const Point<float> points[] = { rp1.resolve (scope), rp2.resolve (scope), rp3.resolve (scope) };

		float bestDistance = std::numeric_limits<float>::max();

		for (int i = 110; --i >= 0;)
		{
			float prop = i > 10 ? ((i - 10) / 100.0f) : (bestProp + ((i - 5) / 1000.0f));
			const Point<float> centre (findQuadraticSubdivisionPoint ((float) prop, points));
			const float distance = centre.getDistanceFrom (targetPoint);

			if (distance < bestDistance)
			{
				bestProp = prop;
				bestDistance = distance;
			}
		}
	}
	else if (type == lineToElement)
	{
		RelativePoint rp1 (getStartPoint()), rp2 (getEndPoint());
		const Line<float> line (rp1.resolve (scope), rp2.resolve (scope));
		bestProp = line.findNearestProportionalPositionTo (targetPoint);
	}

	return bestProp;
}

ValueTree DrawablePath::ValueTreeWrapper::Element::insertPoint (const Point<float>& targetPoint, Expression::Scope* scope, UndoManager* undoManager)
{
	ValueTree newTree;
	const Identifier type (state.getType());

	if (type == cubicToElement)
	{
		float bestProp = findProportionAlongLine (targetPoint, scope);

		RelativePoint rp1 (getStartPoint()), rp2 (getControlPoint (0)), rp3 (getControlPoint (1)), rp4 (getEndPoint());
		const Point<float> points[] = { rp1.resolve (scope), rp2.resolve (scope), rp3.resolve (scope), rp4.resolve (scope) };

		const Point<float> mid1 (points[0] + (points[1] - points[0]) * bestProp),
						   mid2 (points[1] + (points[2] - points[1]) * bestProp),
						   mid3 (points[2] + (points[3] - points[2]) * bestProp);

		const Point<float> newCp1 (mid1 + (mid2 - mid1) * bestProp),
						   newCp2 (mid2 + (mid3 - mid2) * bestProp);

		const Point<float> newCentre (newCp1 + (newCp2 - newCp1) * bestProp);

		setControlPoint (0, mid1, undoManager);
		setControlPoint (1, newCp1, undoManager);
		setControlPoint (2, newCentre, undoManager);
		setModeOfEndPoint (roundedMode, undoManager);

		Element newElement (newTree = ValueTree (cubicToElement));
		newElement.setControlPoint (0, newCp2, nullptr);
		newElement.setControlPoint (1, mid3, nullptr);
		newElement.setControlPoint (2, rp4, nullptr);

		state.getParent().addChild (newTree, state.getParent().indexOf (state) + 1, undoManager);
	}
	else if (type == quadraticToElement)
	{
		float bestProp = findProportionAlongLine (targetPoint, scope);

		RelativePoint rp1 (getStartPoint()), rp2 (getControlPoint (0)), rp3 (getEndPoint());
		const Point<float> points[] = { rp1.resolve (scope), rp2.resolve (scope), rp3.resolve (scope) };

		const Point<float> mid1 (points[0] + (points[1] - points[0]) * bestProp),
						   mid2 (points[1] + (points[2] - points[1]) * bestProp);

		const Point<float> newCentre (mid1 + (mid2 - mid1) * bestProp);

		setControlPoint (0, mid1, undoManager);
		setControlPoint (1, newCentre, undoManager);
		setModeOfEndPoint (roundedMode, undoManager);

		Element newElement (newTree = ValueTree (quadraticToElement));
		newElement.setControlPoint (0, mid2, nullptr);
		newElement.setControlPoint (1, rp3, nullptr);

		state.getParent().addChild (newTree, state.getParent().indexOf (state) + 1, undoManager);
	}
	else if (type == lineToElement)
	{
		RelativePoint rp1 (getStartPoint()), rp2 (getEndPoint());
		const Line<float> line (rp1.resolve (scope), rp2.resolve (scope));
		const Point<float> newPoint (line.findNearestPointTo (targetPoint));

		setControlPoint (0, newPoint, undoManager);

		Element newElement (newTree = ValueTree (lineToElement));
		newElement.setControlPoint (0, rp2, nullptr);

		state.getParent().addChild (newTree, state.getParent().indexOf (state) + 1, undoManager);
	}
	else if (type == closeSubPathElement)
	{
	}

	return newTree;
}

void DrawablePath::ValueTreeWrapper::Element::removePoint (UndoManager* undoManager)
{
	state.getParent().removeChild (state, undoManager);
}

void DrawablePath::refreshFromValueTree (const ValueTree& tree, ComponentBuilder& builder)
{
	ValueTreeWrapper v (tree);
	setComponentID (v.getID());

	refreshFillTypes (v, builder.getImageProvider());
	setStrokeType (v.getStrokeType());

	RelativePointPath newRelativePath;
	v.writeTo (newRelativePath);
	setPath (newRelativePath);
}

ValueTree DrawablePath::createValueTree (ComponentBuilder::ImageProvider* imageProvider) const
{
	ValueTree tree (valueTreeType);
	ValueTreeWrapper v (tree);

	v.setID (getComponentID());
	writeTo (v, imageProvider, nullptr);

	if (relativePath != nullptr)
		v.readFrom (*relativePath, nullptr);
	else
		v.readFrom (RelativePointPath (path), nullptr);

	return tree;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DrawablePath.cpp ***/


/*** Start of inlined file: juce_DrawableRectangle.cpp ***/
BEGIN_JUCE_NAMESPACE

DrawableRectangle::DrawableRectangle()
{
}

DrawableRectangle::DrawableRectangle (const DrawableRectangle& other)
	: DrawableShape (other),
	  bounds (other.bounds),
	  cornerSize (other.cornerSize)
{
}

DrawableRectangle::~DrawableRectangle()
{
}

Drawable* DrawableRectangle::createCopy() const
{
	return new DrawableRectangle (*this);
}

void DrawableRectangle::setRectangle (const RelativeParallelogram& newBounds)
{
	if (bounds != newBounds)
	{
		bounds = newBounds;
		rebuildPath();
	}
}

void DrawableRectangle::setCornerSize (const RelativePoint& newSize)
{
	if (cornerSize != newSize)
	{
		cornerSize = newSize;
		rebuildPath();
	}
}

void DrawableRectangle::rebuildPath()
{
	if (bounds.isDynamic() || cornerSize.isDynamic())
	{
		Drawable::Positioner<DrawableRectangle>* const p = new Drawable::Positioner<DrawableRectangle> (*this);
		setPositioner (p);
		p->apply();
	}
	else
	{
		setPositioner (0);
		recalculateCoordinates (0);
	}
}

bool DrawableRectangle::registerCoordinates (RelativeCoordinatePositionerBase& pos)
{
	bool ok = pos.addPoint (bounds.topLeft);
	ok = pos.addPoint (bounds.topRight) && ok;
	ok = pos.addPoint (bounds.bottomLeft) && ok;
	return pos.addPoint (cornerSize) && ok;
}

void DrawableRectangle::recalculateCoordinates (Expression::Scope* scope)
{
	Point<float> points[3];
	bounds.resolveThreePoints (points, scope);

	const float cornerSizeX = (float) cornerSize.x.resolve (scope);
	const float cornerSizeY = (float) cornerSize.y.resolve (scope);

	const float w = Line<float> (points[0], points[1]).getLength();
	const float h = Line<float> (points[0], points[2]).getLength();

	Path newPath;

	if (cornerSizeX > 0 && cornerSizeY > 0)
		newPath.addRoundedRectangle (0, 0, w, h, cornerSizeX, cornerSizeY);
	else
		newPath.addRectangle (0, 0, w, h);

	newPath.applyTransform (AffineTransform::fromTargetPoints (0, 0, points[0].getX(), points[0].getY(),
															   w, 0, points[1].getX(), points[1].getY(),
															   0, h, points[2].getX(), points[2].getY()));

	if (path != newPath)
	{
		path.swapWithPath (newPath);
		pathChanged();
	}
}

const Identifier DrawableRectangle::valueTreeType ("Rectangle");
const Identifier DrawableRectangle::ValueTreeWrapper::topLeft ("topLeft");
const Identifier DrawableRectangle::ValueTreeWrapper::topRight ("topRight");
const Identifier DrawableRectangle::ValueTreeWrapper::bottomLeft ("bottomLeft");
const Identifier DrawableRectangle::ValueTreeWrapper::cornerSize ("cornerSize");

DrawableRectangle::ValueTreeWrapper::ValueTreeWrapper (const ValueTree& state_)
	: FillAndStrokeState (state_)
{
	jassert (state.hasType (valueTreeType));
}

RelativeParallelogram DrawableRectangle::ValueTreeWrapper::getRectangle() const
{
	return RelativeParallelogram (state.getProperty (topLeft, "0, 0"),
								  state.getProperty (topRight, "100, 0"),
								  state.getProperty (bottomLeft, "0, 100"));
}

void DrawableRectangle::ValueTreeWrapper::setRectangle (const RelativeParallelogram& newBounds, UndoManager* undoManager)
{
	state.setProperty (topLeft, newBounds.topLeft.toString(), undoManager);
	state.setProperty (topRight, newBounds.topRight.toString(), undoManager);
	state.setProperty (bottomLeft, newBounds.bottomLeft.toString(), undoManager);
}

void DrawableRectangle::ValueTreeWrapper::setCornerSize (const RelativePoint& newSize, UndoManager* undoManager)
{
	state.setProperty (cornerSize, newSize.toString(), undoManager);
}

RelativePoint DrawableRectangle::ValueTreeWrapper::getCornerSize() const
{
	return RelativePoint (state [cornerSize]);
}

Value DrawableRectangle::ValueTreeWrapper::getCornerSizeValue (UndoManager* undoManager) const
{
	return state.getPropertyAsValue (cornerSize, undoManager);
}

void DrawableRectangle::refreshFromValueTree (const ValueTree& tree, ComponentBuilder& builder)
{
	ValueTreeWrapper v (tree);
	setComponentID (v.getID());

	refreshFillTypes (v, builder.getImageProvider());
	setStrokeType (v.getStrokeType());
	setRectangle (v.getRectangle());
	setCornerSize (v.getCornerSize());
}

ValueTree DrawableRectangle::createValueTree (ComponentBuilder::ImageProvider* imageProvider) const
{
	ValueTree tree (valueTreeType);
	ValueTreeWrapper v (tree);

	v.setID (getComponentID());
	writeTo (v, imageProvider, nullptr);
	v.setRectangle (bounds, nullptr);
	v.setCornerSize (cornerSize, nullptr);

	return tree;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DrawableRectangle.cpp ***/


/*** Start of inlined file: juce_DrawableText.cpp ***/
BEGIN_JUCE_NAMESPACE

DrawableText::DrawableText()
	: colour (Colours::black),
	  justification (Justification::centredLeft)
{
	setBoundingBox (RelativeParallelogram (RelativePoint (0.0f, 0.0f),
										   RelativePoint (50.0f, 0.0f),
										   RelativePoint (0.0f, 20.0f)));
	setFont (Font (15.0f), true);
}

DrawableText::DrawableText (const DrawableText& other)
	: bounds (other.bounds),
	  fontSizeControlPoint (other.fontSizeControlPoint),
	  font (other.font),
	  text (other.text),
	  colour (other.colour),
	  justification (other.justification)
{
}

DrawableText::~DrawableText()
{
}

void DrawableText::setText (const String& newText)
{
	if (text != newText)
	{
		text = newText;
		refreshBounds();
	}
}

void DrawableText::setColour (const Colour& newColour)
{
	if (colour != newColour)
	{
		colour = newColour;
		repaint();
	}
}

void DrawableText::setFont (const Font& newFont, bool applySizeAndScale)
{
	if (font != newFont)
	{
		font = newFont;

		if (applySizeAndScale)
		{
			setFontSizeControlPoint (RelativePoint (RelativeParallelogram::getPointForInternalCoord (resolvedPoints,
									   Point<float> (font.getHorizontalScale() * font.getHeight(), font.getHeight()))));
		}

		refreshBounds();
	}
}

void DrawableText::setJustification (const Justification& newJustification)
{
	justification = newJustification;
	repaint();
}

void DrawableText::setBoundingBox (const RelativeParallelogram& newBounds)
{
	if (bounds != newBounds)
	{
		bounds = newBounds;
		refreshBounds();
	}
}

void DrawableText::setFontSizeControlPoint (const RelativePoint& newPoint)
{
	if (fontSizeControlPoint != newPoint)
	{
		fontSizeControlPoint = newPoint;
		refreshBounds();
	}
}

void DrawableText::refreshBounds()
{
	if (bounds.isDynamic() || fontSizeControlPoint.isDynamic())
	{
		Drawable::Positioner<DrawableText>* const p = new Drawable::Positioner<DrawableText> (*this);
		setPositioner (p);
		p->apply();
	}
	else
	{
		setPositioner (0);
		recalculateCoordinates (0);
	}
}

bool DrawableText::registerCoordinates (RelativeCoordinatePositionerBase& pos)
{
	bool ok = pos.addPoint (bounds.topLeft);
	ok = pos.addPoint (bounds.topRight) && ok;
	ok = pos.addPoint (bounds.bottomLeft) && ok;
	return pos.addPoint (fontSizeControlPoint) && ok;
}

void DrawableText::recalculateCoordinates (Expression::Scope* scope)
{
	bounds.resolveThreePoints (resolvedPoints, scope);

	const float w = Line<float> (resolvedPoints[0], resolvedPoints[1]).getLength();
	const float h = Line<float> (resolvedPoints[0], resolvedPoints[2]).getLength();

	const Point<float> fontCoords (RelativeParallelogram::getInternalCoordForPoint (resolvedPoints, fontSizeControlPoint.resolve (scope)));
	const float fontHeight = jlimit (0.01f, jmax (0.01f, h), fontCoords.getY());
	const float fontWidth = jlimit (0.01f, jmax (0.01f, w), fontCoords.getX());

	scaledFont = font;
	scaledFont.setHeight (fontHeight);
	scaledFont.setHorizontalScale (fontWidth / fontHeight);

	setBoundsToEnclose (getDrawableBounds());
	repaint();
}

const AffineTransform DrawableText::getArrangementAndTransform (GlyphArrangement& glyphs) const
{
	const float w = Line<float> (resolvedPoints[0], resolvedPoints[1]).getLength();
	const float h = Line<float> (resolvedPoints[0], resolvedPoints[2]).getLength();

	glyphs.addFittedText (scaledFont, text, 0, 0, w, h, justification, 0x100000);

	return AffineTransform::fromTargetPoints (0, 0, resolvedPoints[0].getX(), resolvedPoints[0].getY(),
											  w, 0, resolvedPoints[1].getX(), resolvedPoints[1].getY(),
											  0, h, resolvedPoints[2].getX(), resolvedPoints[2].getY());
}

void DrawableText::paint (Graphics& g)
{
	transformContextToCorrectOrigin (g);

	g.setColour (colour);

	GlyphArrangement ga;
	const AffineTransform transform (getArrangementAndTransform (ga));
	ga.draw (g, transform);
}

Rectangle<float> DrawableText::getDrawableBounds() const
{
	return RelativeParallelogram::getBoundingBox (resolvedPoints);
}

Drawable* DrawableText::createCopy() const
{
	return new DrawableText (*this);
}

const Identifier DrawableText::valueTreeType ("Text");

const Identifier DrawableText::ValueTreeWrapper::text ("text");
const Identifier DrawableText::ValueTreeWrapper::colour ("colour");
const Identifier DrawableText::ValueTreeWrapper::font ("font");
const Identifier DrawableText::ValueTreeWrapper::justification ("justification");
const Identifier DrawableText::ValueTreeWrapper::topLeft ("topLeft");
const Identifier DrawableText::ValueTreeWrapper::topRight ("topRight");
const Identifier DrawableText::ValueTreeWrapper::bottomLeft ("bottomLeft");
const Identifier DrawableText::ValueTreeWrapper::fontSizeAnchor ("fontSizeAnchor");

DrawableText::ValueTreeWrapper::ValueTreeWrapper (const ValueTree& state_)
	: ValueTreeWrapperBase (state_)
{
	jassert (state.hasType (valueTreeType));
}

String DrawableText::ValueTreeWrapper::getText() const
{
	return state [text].toString();
}

void DrawableText::ValueTreeWrapper::setText (const String& newText, UndoManager* undoManager)
{
	state.setProperty (text, newText, undoManager);
}

Value DrawableText::ValueTreeWrapper::getTextValue (UndoManager* undoManager)
{
	return state.getPropertyAsValue (text, undoManager);
}

Colour DrawableText::ValueTreeWrapper::getColour() const
{
	return Colour::fromString (state [colour].toString());
}

void DrawableText::ValueTreeWrapper::setColour (const Colour& newColour, UndoManager* undoManager)
{
	state.setProperty (colour, newColour.toString(), undoManager);
}

Justification DrawableText::ValueTreeWrapper::getJustification() const
{
	return Justification ((int) state [justification]);
}

void DrawableText::ValueTreeWrapper::setJustification (const Justification& newJustification, UndoManager* undoManager)
{
	state.setProperty (justification, newJustification.getFlags(), undoManager);
}

Font DrawableText::ValueTreeWrapper::getFont() const
{
	return Font::fromString (state [font]);
}

void DrawableText::ValueTreeWrapper::setFont (const Font& newFont, UndoManager* undoManager)
{
	state.setProperty (font, newFont.toString(), undoManager);
}

Value DrawableText::ValueTreeWrapper::getFontValue (UndoManager* undoManager)
{
	return state.getPropertyAsValue (font, undoManager);
}

RelativeParallelogram DrawableText::ValueTreeWrapper::getBoundingBox() const
{
	return RelativeParallelogram (state [topLeft].toString(), state [topRight].toString(), state [bottomLeft].toString());
}

void DrawableText::ValueTreeWrapper::setBoundingBox (const RelativeParallelogram& newBounds, UndoManager* undoManager)
{
	state.setProperty (topLeft, newBounds.topLeft.toString(), undoManager);
	state.setProperty (topRight, newBounds.topRight.toString(), undoManager);
	state.setProperty (bottomLeft, newBounds.bottomLeft.toString(), undoManager);
}

RelativePoint DrawableText::ValueTreeWrapper::getFontSizeControlPoint() const
{
	return state [fontSizeAnchor].toString();
}

void DrawableText::ValueTreeWrapper::setFontSizeControlPoint (const RelativePoint& p, UndoManager* undoManager)
{
	state.setProperty (fontSizeAnchor, p.toString(), undoManager);
}

void DrawableText::refreshFromValueTree (const ValueTree& tree, ComponentBuilder&)
{
	ValueTreeWrapper v (tree);
	setComponentID (v.getID());

	const RelativeParallelogram newBounds (v.getBoundingBox());
	const RelativePoint newFontPoint (v.getFontSizeControlPoint());
	const Colour newColour (v.getColour());
	const Justification newJustification (v.getJustification());
	const String newText (v.getText());
	const Font newFont (v.getFont());

	if (text != newText || font != newFont || justification != newJustification
		 || colour != newColour || bounds != newBounds || newFontPoint != fontSizeControlPoint)
	{
		setBoundingBox (newBounds);
		setFontSizeControlPoint (newFontPoint);
		setColour (newColour);
		setFont (newFont, false);
		setJustification (newJustification);
		setText (newText);
	}
}

ValueTree DrawableText::createValueTree (ComponentBuilder::ImageProvider*) const
{
	ValueTree tree (valueTreeType);
	ValueTreeWrapper v (tree);

	v.setID (getComponentID());
	v.setText (text, nullptr);
	v.setFont (font, nullptr);
	v.setJustification (justification, nullptr);
	v.setColour (colour, nullptr);
	v.setBoundingBox (bounds, nullptr);
	v.setFontSizeControlPoint (fontSizeControlPoint, nullptr);

	return tree;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DrawableText.cpp ***/


/*** Start of inlined file: juce_SVGParser.cpp ***/
BEGIN_JUCE_NAMESPACE

class SVGState
{
public:

	SVGState (const XmlElement* const topLevel)
		: topLevelXml (topLevel),
		  elementX (0), elementY (0),
		  width (512), height (512),
		  viewBoxW (0), viewBoxH (0)
	{
	}

	Drawable* parseSVGElement (const XmlElement& xml)
	{
		if (! xml.hasTagName ("svg"))
			return nullptr;

		DrawableComposite* const drawable = new DrawableComposite();

		drawable->setName (xml.getStringAttribute ("id"));

		SVGState newState (*this);

		if (xml.hasAttribute ("transform"))
			newState.addTransform (xml);

		newState.elementX = getCoordLength (xml.getStringAttribute ("x", String (newState.elementX)), viewBoxW);
		newState.elementY = getCoordLength (xml.getStringAttribute ("y", String (newState.elementY)), viewBoxH);
		newState.width    = getCoordLength (xml.getStringAttribute ("width", String (newState.width)), viewBoxW);
		newState.height   = getCoordLength (xml.getStringAttribute ("height", String (newState.height)), viewBoxH);

		if (xml.hasAttribute ("viewBox"))
		{
			const String viewBoxAtt (xml.getStringAttribute ("viewBox"));
			String::CharPointerType viewParams (viewBoxAtt.getCharPointer());
			float vx, vy, vw, vh;

			if (parseCoords (viewParams, vx, vy, true)
				 && parseCoords (viewParams, vw, vh, true)
				 && vw > 0
				 && vh > 0)
			{
				newState.viewBoxW = vw;
				newState.viewBoxH = vh;

				int placementFlags = 0;

				const String aspect (xml.getStringAttribute ("preserveAspectRatio"));

				if (aspect.containsIgnoreCase ("none"))
				{
					placementFlags = RectanglePlacement::stretchToFit;
				}
				else
				{
					if (aspect.containsIgnoreCase ("slice"))
						placementFlags |= RectanglePlacement::fillDestination;

					if (aspect.containsIgnoreCase ("xMin"))
						placementFlags |= RectanglePlacement::xLeft;
					else if (aspect.containsIgnoreCase ("xMax"))
						placementFlags |= RectanglePlacement::xRight;
					else
						placementFlags |= RectanglePlacement::xMid;

					if (aspect.containsIgnoreCase ("yMin"))
						placementFlags |= RectanglePlacement::yTop;
					else if (aspect.containsIgnoreCase ("yMax"))
						placementFlags |= RectanglePlacement::yBottom;
					else
						placementFlags |= RectanglePlacement::yMid;
				}

				const RectanglePlacement placement (placementFlags);

				newState.transform
					= placement.getTransformToFit (Rectangle<float> (vx, vy, vw, vh),
												   Rectangle<float> (0.0f, 0.0f, newState.width, newState.height))
							   .followedBy (newState.transform);
			}
		}
		else
		{
			if (viewBoxW == 0)
				newState.viewBoxW = newState.width;

			if (viewBoxH == 0)
				newState.viewBoxH = newState.height;
		}

		newState.parseSubElements (xml, drawable);

		drawable->resetContentAreaAndBoundingBoxToFitChildren();
		return drawable;
	}

private:

	const XmlElement* const topLevelXml;
	float elementX, elementY, width, height, viewBoxW, viewBoxH;
	AffineTransform transform;
	String cssStyleText;

	void parseSubElements (const XmlElement& xml, DrawableComposite* const parentDrawable)
	{
		forEachXmlChildElement (xml, e)
		{
			Drawable* d = nullptr;

			if (e->hasTagName ("g"))                d = parseGroupElement (*e);
			else if (e->hasTagName ("svg"))         d = parseSVGElement (*e);
			else if (e->hasTagName ("path"))        d = parsePath (*e);
			else if (e->hasTagName ("rect"))        d = parseRect (*e);
			else if (e->hasTagName ("circle"))      d = parseCircle (*e);
			else if (e->hasTagName ("ellipse"))     d = parseEllipse (*e);
			else if (e->hasTagName ("line"))        d = parseLine (*e);
			else if (e->hasTagName ("polyline"))    d = parsePolygon (*e, true);
			else if (e->hasTagName ("polygon"))     d = parsePolygon (*e, false);
			else if (e->hasTagName ("text"))        d = parseText (*e);
			else if (e->hasTagName ("switch"))      d = parseSwitch (*e);
			else if (e->hasTagName ("style"))       parseCSSStyle (*e);

			parentDrawable->addAndMakeVisible (d);
		}
	}

	DrawableComposite* parseSwitch (const XmlElement& xml)
	{
		const XmlElement* const group = xml.getChildByName ("g");

		if (group != nullptr)
			return parseGroupElement (*group);

		return nullptr;
	}

	DrawableComposite* parseGroupElement (const XmlElement& xml)
	{
		DrawableComposite* const drawable = new DrawableComposite();

		drawable->setName (xml.getStringAttribute ("id"));

		if (xml.hasAttribute ("transform"))
		{
			SVGState newState (*this);
			newState.addTransform (xml);

			newState.parseSubElements (xml, drawable);
		}
		else
		{
			parseSubElements (xml, drawable);
		}

		drawable->resetContentAreaAndBoundingBoxToFitChildren();
		return drawable;
	}

	Drawable* parsePath (const XmlElement& xml) const
	{
		const String dAttribute (xml.getStringAttribute ("d").trimStart());
		String::CharPointerType d (dAttribute.getCharPointer());
		Path path;

		if (getStyleAttribute (&xml, "fill-rule").trim().equalsIgnoreCase ("evenodd"))
			path.setUsingNonZeroWinding (false);

		float lastX = 0, lastY = 0;
		float lastX2 = 0, lastY2 = 0;
		juce_wchar lastCommandChar = 0;
		bool isRelative = true;
		bool carryOn = true;

		const CharPointer_ASCII validCommandChars ("MmLlHhVvCcSsQqTtAaZz");

		while (! d.isEmpty())
		{
			float x, y, x2, y2, x3, y3;

			if (validCommandChars.indexOf (*d) >= 0)
			{
				lastCommandChar = d.getAndAdvance();
				isRelative = (lastCommandChar >= 'a' && lastCommandChar <= 'z');
			}

			switch (lastCommandChar)
			{
			case 'M':
			case 'm':
			case 'L':
			case 'l':
				if (parseCoords (d, x, y, false))
				{
					if (isRelative)
					{
						x += lastX;
						y += lastY;
					}

					if (lastCommandChar == 'M' || lastCommandChar == 'm')
					{
						path.startNewSubPath (x, y);
						lastCommandChar = 'l';
					}
					else
						path.lineTo (x, y);

					lastX2 = lastX;
					lastY2 = lastY;
					lastX = x;
					lastY = y;
				}
				else
				{
					++d;
				}

				break;

			case 'H':
			case 'h':
				if (parseCoord (d, x, false, true))
				{
					if (isRelative)
						x += lastX;

					path.lineTo (x, lastY);

					lastX2 = lastX;
					lastX = x;
				}
				else
				{
					++d;
				}
				break;

			case 'V':
			case 'v':
				if (parseCoord (d, y, false, false))
				{
					if (isRelative)
						y += lastY;

					path.lineTo (lastX, y);

					lastY2 = lastY;
					lastY = y;
				}
				else
				{
					++d;
				}
				break;

			case 'C':
			case 'c':
				if (parseCoords (d, x, y, false)
					 && parseCoords (d, x2, y2, false)
					 && parseCoords (d, x3, y3, false))
				{
					if (isRelative)
					{
						x += lastX;
						y += lastY;
						x2 += lastX;
						y2 += lastY;
						x3 += lastX;
						y3 += lastY;
					}

					path.cubicTo (x, y, x2, y2, x3, y3);

					lastX2 = x2;
					lastY2 = y2;
					lastX = x3;
					lastY = y3;
				}
				else
				{
					++d;
				}
				break;

			case 'S':
			case 's':
				if (parseCoords (d, x, y, false)
					 && parseCoords (d, x3, y3, false))
				{
					if (isRelative)
					{
						x += lastX;
						y += lastY;
						x3 += lastX;
						y3 += lastY;
					}

					x2 = lastX + (lastX - lastX2);
					y2 = lastY + (lastY - lastY2);
					path.cubicTo (x2, y2, x, y, x3, y3);

					lastX2 = x;
					lastY2 = y;
					lastX = x3;
					lastY = y3;
				}
				else
				{
					++d;
				}
				break;

			case 'Q':
			case 'q':
				if (parseCoords (d, x, y, false)
					 && parseCoords (d, x2, y2, false))
				{
					if (isRelative)
					{
						x += lastX;
						y += lastY;
						x2 += lastX;
						y2 += lastY;
					}

					path.quadraticTo (x, y, x2, y2);

					lastX2 = x;
					lastY2 = y;
					lastX = x2;
					lastY = y2;
				}
				else
				{
					++d;
				}
				break;

			case 'T':
			case 't':
				if (parseCoords (d, x, y, false))
				{
					if (isRelative)
					{
						x += lastX;
						y += lastY;
					}

					x2 = lastX + (lastX - lastX2);
					y2 = lastY + (lastY - lastY2);
					path.quadraticTo (x2, y2, x, y);

					lastX2 = x2;
					lastY2 = y2;
					lastX = x;
					lastY = y;
				}
				else
				{
					++d;
				}
				break;

			case 'A':
			case 'a':
				if (parseCoords (d, x, y, false))
				{
					String num;

					if (parseNextNumber (d, num, false))
					{
						const float angle = num.getFloatValue() * (180.0f / float_Pi);

						if (parseNextNumber (d, num, false))
						{
							const bool largeArc = num.getIntValue() != 0;

							if (parseNextNumber (d, num, false))
							{
								const bool sweep = num.getIntValue() != 0;

								if (parseCoords (d, x2, y2, false))
								{
									if (isRelative)
									{
										x2 += lastX;
										y2 += lastY;
									}

									if (lastX != x2 || lastY != y2)
									{
										double centreX, centreY, startAngle, deltaAngle;
										double rx = x, ry = y;

										endpointToCentreParameters (lastX, lastY, x2, y2,
																	angle, largeArc, sweep,
																	rx, ry, centreX, centreY,
																	startAngle, deltaAngle);

										path.addCentredArc ((float) centreX, (float) centreY,
															(float) rx, (float) ry,
															angle, (float) startAngle, (float) (startAngle + deltaAngle),
															false);

										path.lineTo (x2, y2);
									}

									lastX2 = lastX;
									lastY2 = lastY;
									lastX = x2;
									lastY = y2;
								}
							}
						}
					}
				}
				else
				{
					++d;
				}

				break;

			case 'Z':
			case 'z':
				path.closeSubPath();
				d = d.findEndOfWhitespace();
				break;

			default:
				carryOn = false;
				break;
			}

			if (! carryOn)
				break;
		}

		return parseShape (xml, path);
	}

	Drawable* parseRect (const XmlElement& xml) const
	{
		Path rect;

		const bool hasRX = xml.hasAttribute ("rx");
		const bool hasRY = xml.hasAttribute ("ry");

		if (hasRX || hasRY)
		{
			float rx = getCoordLength (xml.getStringAttribute ("rx"), viewBoxW);
			float ry = getCoordLength (xml.getStringAttribute ("ry"), viewBoxH);

			if (! hasRX)
				rx = ry;
			else if (! hasRY)
				ry = rx;

			rect.addRoundedRectangle (getCoordLength (xml.getStringAttribute ("x"), viewBoxW),
									  getCoordLength (xml.getStringAttribute ("y"), viewBoxH),
									  getCoordLength (xml.getStringAttribute ("width"), viewBoxW),
									  getCoordLength (xml.getStringAttribute ("height"), viewBoxH),
									  rx, ry);
		}
		else
		{
			rect.addRectangle (getCoordLength (xml.getStringAttribute ("x"), viewBoxW),
							   getCoordLength (xml.getStringAttribute ("y"), viewBoxH),
							   getCoordLength (xml.getStringAttribute ("width"), viewBoxW),
							   getCoordLength (xml.getStringAttribute ("height"), viewBoxH));
		}

		return parseShape (xml, rect);
	}

	Drawable* parseCircle (const XmlElement& xml) const
	{
		Path circle;

		const float cx = getCoordLength (xml.getStringAttribute ("cx"), viewBoxW);
		const float cy = getCoordLength (xml.getStringAttribute ("cy"), viewBoxH);
		const float radius = getCoordLength (xml.getStringAttribute ("r"), viewBoxW);

		circle.addEllipse (cx - radius, cy - radius, radius * 2.0f, radius * 2.0f);

		return parseShape (xml, circle);
	}

	Drawable* parseEllipse (const XmlElement& xml) const
	{
		Path ellipse;

		const float cx      = getCoordLength (xml.getStringAttribute ("cx"), viewBoxW);
		const float cy      = getCoordLength (xml.getStringAttribute ("cy"), viewBoxH);
		const float radiusX = getCoordLength (xml.getStringAttribute ("rx"), viewBoxW);
		const float radiusY = getCoordLength (xml.getStringAttribute ("ry"), viewBoxH);

		ellipse.addEllipse (cx - radiusX, cy - radiusY, radiusX * 2.0f, radiusY * 2.0f);

		return parseShape (xml, ellipse);
	}

	Drawable* parseLine (const XmlElement& xml) const
	{
		Path line;

		const float x1 = getCoordLength (xml.getStringAttribute ("x1"), viewBoxW);
		const float y1 = getCoordLength (xml.getStringAttribute ("y1"), viewBoxH);
		const float x2 = getCoordLength (xml.getStringAttribute ("x2"), viewBoxW);
		const float y2 = getCoordLength (xml.getStringAttribute ("y2"), viewBoxH);

		line.startNewSubPath (x1, y1);
		line.lineTo (x2, y2);

		return parseShape (xml, line);
	}

	Drawable* parsePolygon (const XmlElement& xml, const bool isPolyline) const
	{
		const String pointsAtt (xml.getStringAttribute ("points"));
		String::CharPointerType points (pointsAtt.getCharPointer());
		Path path;
		float x, y;

		if (parseCoords (points, x, y, true))
		{
			float firstX = x;
			float firstY = y;
			float lastX = 0, lastY = 0;

			path.startNewSubPath (x, y);

			while (parseCoords (points, x, y, true))
			{
				lastX = x;
				lastY = y;
				path.lineTo (x, y);
			}

			if ((! isPolyline) || (firstX == lastX && firstY == lastY))
				path.closeSubPath();
		}

		return parseShape (xml, path);
	}

	Drawable* parseShape (const XmlElement& xml, Path& path,
						  const bool shouldParseTransform = true) const
	{
		if (shouldParseTransform && xml.hasAttribute ("transform"))
		{
			SVGState newState (*this);
			newState.addTransform (xml);

			return newState.parseShape (xml, path, false);
		}

		DrawablePath* dp = new DrawablePath();
		dp->setName (xml.getStringAttribute ("id"));
		dp->setFill (Colours::transparentBlack);

		path.applyTransform (transform);
		dp->setPath (path);

		Path::Iterator iter (path);

		bool containsClosedSubPath = false;
		while (iter.next())
		{
			if (iter.elementType == Path::Iterator::closePath)
			{
				containsClosedSubPath = true;
				break;
			}
		}

		dp->setFill (getPathFillType (path,
									  getStyleAttribute (&xml, "fill"),
									  getStyleAttribute (&xml, "fill-opacity"),
									  getStyleAttribute (&xml, "opacity"),
									  containsClosedSubPath ? Colours::black
															: Colours::transparentBlack));

		const String strokeType (getStyleAttribute (&xml, "stroke"));

		if (strokeType.isNotEmpty() && ! strokeType.equalsIgnoreCase ("none"))
		{
			dp->setStrokeFill (getPathFillType (path, strokeType,
												getStyleAttribute (&xml, "stroke-opacity"),
												getStyleAttribute (&xml, "opacity"),
												Colours::transparentBlack));

			dp->setStrokeType (getStrokeFor (&xml));
		}

		return dp;
	}

	const XmlElement* findLinkedElement (const XmlElement* e) const
	{
		const String id (e->getStringAttribute ("xlink:href"));

		if (! id.startsWithChar ('#'))
			return nullptr;

		return findElementForId (topLevelXml, id.substring (1));
	}

	void addGradientStopsIn (ColourGradient& cg, const XmlElement* const fillXml) const
	{
		if (fillXml == 0)
			return;

		forEachXmlChildElementWithTagName (*fillXml, e, "stop")
		{
			int index = 0;
			Colour col (parseColour (getStyleAttribute  (e, "stop-color"), index, Colours::black));

			const String opacity (getStyleAttribute (e, "stop-opacity", "1"));
			col = col.withMultipliedAlpha (jlimit (0.0f, 1.0f, opacity.getFloatValue()));

			double offset = e->getDoubleAttribute ("offset");

			if (e->getStringAttribute ("offset").containsChar ('%'))
				offset *= 0.01;

			cg.addColour (jlimit (0.0, 1.0, offset), col);
		}
	}

	FillType getPathFillType (const Path& path,
							  const String& fill,
							  const String& fillOpacity,
							  const String& overallOpacity,
							  const Colour& defaultColour) const
	{
		float opacity = 1.0f;

		if (overallOpacity.isNotEmpty())
			opacity = jlimit (0.0f, 1.0f, overallOpacity.getFloatValue());

		if (fillOpacity.isNotEmpty())
			opacity *= (jlimit (0.0f, 1.0f, fillOpacity.getFloatValue()));

		if (fill.startsWithIgnoreCase ("url"))
		{
			const String id (fill.fromFirstOccurrenceOf ("#", false, false)
								 .upToLastOccurrenceOf (")", false, false).trim());

			const XmlElement* const fillXml = findElementForId (topLevelXml, id);

			if (fillXml != nullptr
				 && (fillXml->hasTagName ("linearGradient")
					  || fillXml->hasTagName ("radialGradient")))
			{
				const XmlElement* inheritedFrom = findLinkedElement (fillXml);

				ColourGradient gradient;

				addGradientStopsIn (gradient, inheritedFrom);
				addGradientStopsIn (gradient, fillXml);

				if (gradient.getNumColours() > 0)
				{
					gradient.addColour (0.0, gradient.getColour (0));
					gradient.addColour (1.0, gradient.getColour (gradient.getNumColours() - 1));
				}
				else
				{
					gradient.addColour (0.0, Colours::black);
					gradient.addColour (1.0, Colours::black);
				}

				if (overallOpacity.isNotEmpty())
					gradient.multiplyOpacity (overallOpacity.getFloatValue());

				jassert (gradient.getNumColours() > 0);

				gradient.isRadial = fillXml->hasTagName ("radialGradient");

				float gradientWidth = viewBoxW;
				float gradientHeight = viewBoxH;
				float dx = 0.0f;
				float dy = 0.0f;

				const bool userSpace = fillXml->getStringAttribute ("gradientUnits").equalsIgnoreCase ("userSpaceOnUse");

				if (! userSpace)
				{
					const Rectangle<float> bounds (path.getBounds());
					dx = bounds.getX();
					dy = bounds.getY();
					gradientWidth = bounds.getWidth();
					gradientHeight = bounds.getHeight();
				}

				if (gradient.isRadial)
				{
					if (userSpace)
						gradient.point1.setXY (dx + getCoordLength (fillXml->getStringAttribute ("cx", "50%"), gradientWidth),
											   dy + getCoordLength (fillXml->getStringAttribute ("cy", "50%"), gradientHeight));
					else
						gradient.point1.setXY (dx + gradientWidth * getCoordLength (fillXml->getStringAttribute ("cx", "50%"), 1.0f),
											   dy + gradientHeight * getCoordLength (fillXml->getStringAttribute ("cy", "50%"), 1.0f));

					const float radius = getCoordLength (fillXml->getStringAttribute ("r", "50%"), gradientWidth);
					gradient.point2 = gradient.point1 + Point<float> (radius, 0.0f);

					//xxx (the fx, fy focal point isn't handled properly here..)
				}
				else
				{
					if (userSpace)
					{
						gradient.point1.setXY (dx + getCoordLength (fillXml->getStringAttribute ("x1", "0%"), gradientWidth),
											   dy + getCoordLength (fillXml->getStringAttribute ("y1", "0%"), gradientHeight));

						gradient.point2.setXY (dx + getCoordLength (fillXml->getStringAttribute ("x2", "100%"), gradientWidth),
											   dy + getCoordLength (fillXml->getStringAttribute ("y2", "0%"), gradientHeight));
					}
					else
					{
						gradient.point1.setXY (dx + gradientWidth * getCoordLength (fillXml->getStringAttribute ("x1", "0%"), 1.0f),
											   dy + gradientHeight * getCoordLength (fillXml->getStringAttribute ("y1", "0%"), 1.0f));

						gradient.point2.setXY (dx + gradientWidth * getCoordLength (fillXml->getStringAttribute ("x2", "100%"), 1.0f),
											   dy + gradientHeight * getCoordLength (fillXml->getStringAttribute ("y2", "0%"), 1.0f));
					}

					if (gradient.point1 == gradient.point2)
						return Colour (gradient.getColour (gradient.getNumColours() - 1));
				}

				FillType type (gradient);
				type.transform = parseTransform (fillXml->getStringAttribute ("gradientTransform"))
								   .followedBy (transform);
				return type;
			}
		}

		if (fill.equalsIgnoreCase ("none"))
			return Colours::transparentBlack;

		int i = 0;
		const Colour colour (parseColour (fill, i, defaultColour));
		return colour.withMultipliedAlpha (opacity);
	}

	PathStrokeType getStrokeFor (const XmlElement* const xml) const
	{
		const String strokeWidth (getStyleAttribute (xml, "stroke-width"));
		const String cap (getStyleAttribute (xml, "stroke-linecap"));
		const String join (getStyleAttribute (xml, "stroke-linejoin"));

		//const String mitreLimit (getStyleAttribute (xml, "stroke-miterlimit"));
		//const String dashArray (getStyleAttribute (xml, "stroke-dasharray"));
		//const String dashOffset (getStyleAttribute (xml, "stroke-dashoffset"));

		PathStrokeType::JointStyle joinStyle = PathStrokeType::mitered;
		PathStrokeType::EndCapStyle capStyle = PathStrokeType::butt;

		if (join.equalsIgnoreCase ("round"))
			joinStyle = PathStrokeType::curved;
		else if (join.equalsIgnoreCase ("bevel"))
			joinStyle = PathStrokeType::beveled;

		if (cap.equalsIgnoreCase ("round"))
			capStyle = PathStrokeType::rounded;
		else if (cap.equalsIgnoreCase ("square"))
			capStyle = PathStrokeType::square;

		float ox = 0.0f, oy = 0.0f;
		float x = getCoordLength (strokeWidth, viewBoxW), y = 0.0f;
		transform.transformPoints (ox, oy, x, y);

		return PathStrokeType (strokeWidth.isNotEmpty() ? juce_hypot (x - ox, y - oy) : 1.0f,
							   joinStyle, capStyle);
	}

	Drawable* parseText (const XmlElement& xml)
	{
		Array <float> xCoords, yCoords, dxCoords, dyCoords;

		getCoordList (xCoords, getInheritedAttribute (&xml, "x"), true, true);
		getCoordList (yCoords, getInheritedAttribute (&xml, "y"), true, false);
		getCoordList (dxCoords, getInheritedAttribute (&xml, "dx"), true, true);
		getCoordList (dyCoords, getInheritedAttribute (&xml, "dy"), true, false);

		//xxx not done text yet!

		forEachXmlChildElement (xml, e)
		{
			if (e->isTextElement())
			{
				const String text (e->getText());

				Path path;
				Drawable* s = parseShape (*e, path);
				delete s;  // xxx not finished!
			}
			else if (e->hasTagName ("tspan"))
			{
				Drawable* s = parseText (*e);
				delete s;  // xxx not finished!
			}
		}

		return nullptr;
	}

	void addTransform (const XmlElement& xml)
	{
		transform = parseTransform (xml.getStringAttribute ("transform"))
						.followedBy (transform);
	}

	bool parseCoord (String::CharPointerType& s, float& value, const bool allowUnits, const bool isX) const
	{
		String number;

		if (! parseNextNumber (s, number, allowUnits))
		{
			value = 0;
			return false;
		}

		value = getCoordLength (number, isX ? viewBoxW : viewBoxH);
		return true;
	}

	bool parseCoords (String::CharPointerType& s, float& x, float& y, const bool allowUnits) const
	{
		return parseCoord (s, x, allowUnits, true)
			&& parseCoord (s, y, allowUnits, false);
	}

	float getCoordLength (const String& s, const float sizeForProportions) const
	{
		float n = s.getFloatValue();
		const int len = s.length();

		if (len > 2)
		{
			const float dpi = 96.0f;

			const juce_wchar n1 = s [len - 2];
			const juce_wchar n2 = s [len - 1];

			if (n1 == 'i' && n2 == 'n')	 n *= dpi;
			else if (n1 == 'm' && n2 == 'm')	n *= dpi / 25.4f;
			else if (n1 == 'c' && n2 == 'm')	n *= dpi / 2.54f;
			else if (n1 == 'p' && n2 == 'c')	n *= 15.0f;
			else if (n2 == '%')		 n *= 0.01f * sizeForProportions;
		}

		return n;
	}

	void getCoordList (Array <float>& coords, const String& list,
					   const bool allowUnits, const bool isX) const
	{
		String::CharPointerType text (list.getCharPointer());
		float value;

		while (parseCoord (text, value, allowUnits, isX))
			coords.add (value);
	}

	void parseCSSStyle (const XmlElement& xml)
	{
		cssStyleText = xml.getAllSubText() + "\n" + cssStyleText;
	}

	String getStyleAttribute (const XmlElement* xml, const String& attributeName,
							  const String& defaultValue = String::empty) const
	{
		if (xml->hasAttribute (attributeName))
			return xml->getStringAttribute (attributeName, defaultValue);

		const String styleAtt (xml->getStringAttribute ("style"));

		if (styleAtt.isNotEmpty())
		{
			const String value (getAttributeFromStyleList (styleAtt, attributeName, String::empty));

			if (value.isNotEmpty())
				return value;
		}
		else if (xml->hasAttribute ("class"))
		{
			const String className ("." + xml->getStringAttribute ("class"));

			int index = cssStyleText.indexOfIgnoreCase (className + " ");

			if (index < 0)
				index = cssStyleText.indexOfIgnoreCase (className + "{");

			if (index >= 0)
			{
				const int openBracket = cssStyleText.indexOfChar (index, '{');

				if (openBracket > index)
				{
					const int closeBracket = cssStyleText.indexOfChar (openBracket, '}');

					if (closeBracket > openBracket)
					{
						const String value (getAttributeFromStyleList (cssStyleText.substring (openBracket + 1, closeBracket), attributeName, defaultValue));

						if (value.isNotEmpty())
							return value;
					}
				}
			}
		}

		xml = const_cast <XmlElement*> (topLevelXml)->findParentElementOf (xml);

		if (xml != nullptr)
			return getStyleAttribute (xml, attributeName, defaultValue);

		return defaultValue;
	}

	String getInheritedAttribute (const XmlElement* xml, const String& attributeName) const
	{
		if (xml->hasAttribute (attributeName))
			return xml->getStringAttribute (attributeName);

		xml = const_cast <XmlElement*> (topLevelXml)->findParentElementOf (xml);

		if (xml != nullptr)
			return getInheritedAttribute  (xml, attributeName);

		return String::empty;
	}

	static bool isIdentifierChar (const juce_wchar c)
	{
		return CharacterFunctions::isLetter (c) || c == '-';
	}

	static String getAttributeFromStyleList (const String& list, const String& attributeName, const String& defaultValue)
	{
		int i = 0;

		for (;;)
		{
			i = list.indexOf (i, attributeName);

			if (i < 0)
				break;

			if ((i == 0 || (i > 0 && ! isIdentifierChar (list [i - 1])))
				 && ! isIdentifierChar (list [i + attributeName.length()]))
			{
				i = list.indexOfChar (i, ':');

				if (i < 0)
					break;

				int end = list.indexOfChar (i, ';');

				if (end < 0)
					end = 0x7ffff;

				return list.substring (i + 1, end).trim();
			}

			++i;
		}

		return defaultValue;
	}

	static bool parseNextNumber (String::CharPointerType& s, String& value, const bool allowUnits)
	{
		while (s.isWhitespace() || *s == ',')
			++s;

		String::CharPointerType start (s);
		int numChars = 0;

		if (s.isDigit() || *s == '.' || *s == '-')
		{
			++numChars;
			++s;
		}

		while (s.isDigit() || *s == '.')
		{
			++numChars;
			++s;
		}

		if ((*s == 'e' || *s == 'E')
			 && ((s + 1).isDigit() || s[1] == '-' || s[1] == '+'))
		{
			s += 2;
			numChars += 2;

			while (s.isDigit())
			{
				++numChars;
				++s;
			}
		}

		if (allowUnits)
		{
			while (s.isLetter())
			{
				++numChars;
				++s;
			}
		}

		if (numChars == 0)
			return false;

		value = String (start, numChars);

		while (s.isWhitespace() || *s == ',')
			++s;

		return true;
	}

	static const Colour parseColour (const String& s, int& index, const Colour& defaultColour)
	{
		if (s [index] == '#')
		{
			uint32 hex[6] = { 0 };
			int numChars = 0;

			for (int i = 6; --i >= 0;)
			{
				const int hexValue = CharacterFunctions::getHexDigitValue (s [++index]);

				if (hexValue >= 0)
					hex [numChars++] = hexValue;
				else
					break;
			}

			if (numChars <= 3)
				return Colour ((uint8) (hex [0] * 0x11),
							   (uint8) (hex [1] * 0x11),
							   (uint8) (hex [2] * 0x11));
			else
				return Colour ((uint8) ((hex [0] << 4) + hex [1]),
							   (uint8) ((hex [2] << 4) + hex [3]),
							   (uint8) ((hex [4] << 4) + hex [5]));
		}
		else if (s [index] == 'r'
				  && s [index + 1] == 'g'
				  && s [index + 2] == 'b')
		{
			const int openBracket = s.indexOfChar (index, '(');
			const int closeBracket = s.indexOfChar (openBracket, ')');

			if (openBracket >= 3 && closeBracket > openBracket)
			{
				index = closeBracket;

				StringArray tokens;
				tokens.addTokens (s.substring (openBracket + 1, closeBracket), ",", "");
				tokens.trim();
				tokens.removeEmptyStrings();

				if (tokens[0].containsChar ('%'))
					return Colour ((uint8) roundToInt (2.55 * tokens[0].getDoubleValue()),
								   (uint8) roundToInt (2.55 * tokens[1].getDoubleValue()),
								   (uint8) roundToInt (2.55 * tokens[2].getDoubleValue()));
				else
					return Colour ((uint8) tokens[0].getIntValue(),
								   (uint8) tokens[1].getIntValue(),
								   (uint8) tokens[2].getIntValue());
			}
		}

		return Colours::findColourForName (s, defaultColour);
	}

	static const AffineTransform parseTransform (String t)
	{
		AffineTransform result;

		while (t.isNotEmpty())
		{
			StringArray tokens;
			tokens.addTokens (t.fromFirstOccurrenceOf ("(", false, false)
							   .upToFirstOccurrenceOf (")", false, false),
							  ", ", String::empty);

			tokens.removeEmptyStrings (true);

			float numbers [6];

			for (int i = 0; i < 6; ++i)
				numbers[i] = tokens[i].getFloatValue();

			AffineTransform trans;

			if (t.startsWithIgnoreCase ("matrix"))
			{
				trans = AffineTransform (numbers[0], numbers[2], numbers[4],
										 numbers[1], numbers[3], numbers[5]);
			}
			else if (t.startsWithIgnoreCase ("translate"))
			{
				jassert (tokens.size() == 2);
				trans = AffineTransform::translation (numbers[0], numbers[1]);
			}
			else if (t.startsWithIgnoreCase ("scale"))
			{
				if (tokens.size() == 1)
					trans = AffineTransform::scale (numbers[0], numbers[0]);
				else
					trans = AffineTransform::scale (numbers[0], numbers[1]);
			}
			else if (t.startsWithIgnoreCase ("rotate"))
			{
				if (tokens.size() != 3)
					trans = AffineTransform::rotation (numbers[0] / (180.0f / float_Pi));
				else
					trans = AffineTransform::rotation (numbers[0] / (180.0f / float_Pi),
													   numbers[1], numbers[2]);
			}
			else if (t.startsWithIgnoreCase ("skewX"))
			{
				trans = AffineTransform (1.0f, std::tan (numbers[0] * (float_Pi / 180.0f)), 0.0f,
										 0.0f, 1.0f, 0.0f);
			}
			else if (t.startsWithIgnoreCase ("skewY"))
			{
				trans = AffineTransform (1.0f, 0.0f, 0.0f,
										 std::tan (numbers[0] * (float_Pi / 180.0f)), 1.0f, 0.0f);
			}

			result = trans.followedBy (result);
			t = t.fromFirstOccurrenceOf (")", false, false).trimStart();
		}

		return result;
	}

	static void endpointToCentreParameters (const double x1, const double y1,
											const double x2, const double y2,
											const double angle,
											const bool largeArc, const bool sweep,
											double& rx, double& ry,
											double& centreX, double& centreY,
											double& startAngle, double& deltaAngle)
	{
		const double midX = (x1 - x2) * 0.5;
		const double midY = (y1 - y2) * 0.5;

		const double cosAngle = cos (angle);
		const double sinAngle = sin (angle);
		const double xp = cosAngle * midX + sinAngle * midY;
		const double yp = cosAngle * midY - sinAngle * midX;
		const double xp2 = xp * xp;
		const double yp2 = yp * yp;

		double rx2 = rx * rx;
		double ry2 = ry * ry;

		const double s = (xp2 / rx2) + (yp2 / ry2);
		double c;

		if (s <= 1.0)
		{
			c = std::sqrt (jmax (0.0, ((rx2 * ry2) - (rx2 * yp2) - (ry2 * xp2))
										 / (( rx2 * yp2) + (ry2 * xp2))));

			if (largeArc == sweep)
				c = -c;
		}
		else
		{
			const double s2 = std::sqrt (s);
			rx *= s2;
			ry *= s2;
			rx2 = rx * rx;
			ry2 = ry * ry;
			c = 0;
		}

		const double cpx = ((rx * yp) / ry) * c;
		const double cpy = ((-ry * xp) / rx) * c;

		centreX = ((x1 + x2) * 0.5) + (cosAngle * cpx) - (sinAngle * cpy);
		centreY = ((y1 + y2) * 0.5) + (sinAngle * cpx) + (cosAngle * cpy);

		const double ux = (xp - cpx) / rx;
		const double uy = (yp - cpy) / ry;
		const double vx = (-xp - cpx) / rx;
		const double vy = (-yp - cpy) / ry;

		const double length = juce_hypot (ux, uy);

		startAngle = acos (jlimit (-1.0, 1.0, ux / length));

		if (uy < 0)
			startAngle = -startAngle;

		startAngle += double_Pi * 0.5;

		deltaAngle = acos (jlimit (-1.0, 1.0, ((ux * vx) + (uy * vy))
												/ (length * juce_hypot (vx, vy))));

		if ((ux * vy) - (uy * vx) < 0)
			deltaAngle = -deltaAngle;

		if (sweep)
		{
			if (deltaAngle < 0)
				deltaAngle += double_Pi * 2.0;
		}
		else
		{
			if (deltaAngle > 0)
				deltaAngle -= double_Pi * 2.0;
		}

		deltaAngle = fmod (deltaAngle, double_Pi * 2.0);
	}

	static const XmlElement* findElementForId (const XmlElement* const parent, const String& id)
	{
		forEachXmlChildElement (*parent, e)
		{
			if (e->compareAttribute ("id", id))
				return e;

			const XmlElement* const found = findElementForId (e, id);

			if (found != nullptr)
				return found;
		}

		return nullptr;
	}

	SVGState& operator= (const SVGState&);
};

Drawable* Drawable::createFromSVG (const XmlElement& svgDocument)
{
	SVGState state (&svgDocument);
	return state.parseSVGElement (svgDocument);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_SVGParser.cpp ***/


/*** Start of inlined file: juce_DropShadowEffect.cpp ***/
BEGIN_JUCE_NAMESPACE

#if JUCE_MSVC && JUCE_DEBUG
  #pragma optimize ("t", on)
#endif

DropShadowEffect::DropShadowEffect()
  : offsetX (0),
	offsetY (0),
	radius (4),
	opacity (0.6f)
{
}

DropShadowEffect::~DropShadowEffect()
{
}

void DropShadowEffect::setShadowProperties (const float newRadius,
											const float newOpacity,
											const int newShadowOffsetX,
											const int newShadowOffsetY)
{
	radius = jmax (1.1f, newRadius);
	offsetX = newShadowOffsetX;
	offsetY = newShadowOffsetY;
	opacity = newOpacity;
}

void DropShadowEffect::applyEffect (Image& image, Graphics& g, float alpha)
{
	const int w = image.getWidth();
	const int h = image.getHeight();

	Image shadowImage (Image::SingleChannel, w, h, false);

	{
		const Image::BitmapData srcData (image, Image::BitmapData::readOnly);
		const Image::BitmapData destData (shadowImage, Image::BitmapData::readWrite);

		const int filter = roundToInt (63.0f / radius);
		const int radiusMinus1 = roundToInt ((radius - 1.0f) * 63.0f);

		for (int x = w; --x >= 0;)
		{
			int shadowAlpha = 0;

			const PixelARGB* src = ((const PixelARGB*) srcData.data) + x;
			uint8* shadowPix = destData.data + x;

			for (int y = h; --y >= 0;)
			{
				shadowAlpha = ((shadowAlpha * radiusMinus1 + (src->getAlpha() << 6)) * filter) >> 12;

				*shadowPix = (uint8) shadowAlpha;
				src = addBytesToPointer (src, srcData.lineStride);
				shadowPix += destData.lineStride;
			}
		}

		for (int y = h; --y >= 0;)
		{
			int shadowAlpha = 0;
			uint8* shadowPix = destData.getLinePointer (y);

			for (int x = w; --x >= 0;)
			{
				shadowAlpha = ((shadowAlpha * radiusMinus1 + (*shadowPix << 6)) * filter) >> 12;
				*shadowPix++ = (uint8) shadowAlpha;
			}
		}
	}

	g.setColour (Colours::black.withAlpha (opacity * alpha));
	g.drawImageAt (shadowImage, offsetX, offsetY, true);

	g.setOpacity (alpha);
	g.drawImageAt (image, 0, 0);
}

#if JUCE_MSVC && JUCE_DEBUG
  #pragma optimize ("", on)  // resets optimisations to the project defaults
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_DropShadowEffect.cpp ***/


/*** Start of inlined file: juce_GlowEffect.cpp ***/
BEGIN_JUCE_NAMESPACE

GlowEffect::GlowEffect()
  : radius (2.0f),
	colour (Colours::white)
{
}

GlowEffect::~GlowEffect()
{
}

void GlowEffect::setGlowProperties (const float newRadius,
									const Colour& newColour)
{
	radius = newRadius;
	colour = newColour;
}

void GlowEffect::applyEffect (Image& image, Graphics& g, float alpha)
{
	Image temp (image.getFormat(), image.getWidth(), image.getHeight(), true);

	ImageConvolutionKernel blurKernel (roundToInt (radius * 2.0f));

	blurKernel.createGaussianBlur (radius);
	blurKernel.rescaleAllValues (radius);

	blurKernel.applyToImage (temp, image, image.getBounds());

	g.setColour (colour.withMultipliedAlpha (alpha));
	g.drawImageAt (temp, 0, 0, true);

	g.setOpacity (alpha);
	g.drawImageAt (image, 0, 0, false);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_GlowEffect.cpp ***/


/*** Start of inlined file: juce_Font.cpp ***/
BEGIN_JUCE_NAMESPACE

namespace FontValues
{
	float limitFontHeight (const float height) noexcept
	{
		return jlimit (0.1f, 10000.0f, height);
	}

	const float defaultFontHeight = 14.0f;
	String fallbackFont;
}

class TypefaceCache  : public DeletedAtShutdown
{
public:
	TypefaceCache()
		: counter (0)
	{
		setSize (10);
	}

	~TypefaceCache()
	{
		clearSingletonInstance();
	}

	juce_DeclareSingleton_SingleThreaded_Minimal (TypefaceCache);

	void setSize (const int numToCache)
	{
		faces.clear();
		faces.insertMultiple (-1, CachedFace(), numToCache);
	}

	Typeface::Ptr findTypefaceFor (const Font& font)
	{
		const int flags = font.getStyleFlags() & (Font::bold | Font::italic);
		const String faceName (font.getTypefaceName());

		int i;
		for (i = faces.size(); --i >= 0;)
		{
			CachedFace& face = faces.getReference(i);

			if (face.flags == flags
				 && face.typefaceName == faceName
				 && face.typeface->isSuitableForFont (font))
			{
				face.lastUsageCount = ++counter;
				return face.typeface;
			}
		}

		int replaceIndex = 0;
		int bestLastUsageCount = std::numeric_limits<int>::max();

		for (i = faces.size(); --i >= 0;)
		{
			const int lu = faces.getReference(i).lastUsageCount;

			if (bestLastUsageCount > lu)
			{
				bestLastUsageCount = lu;
				replaceIndex = i;
			}
		}

		CachedFace& face = faces.getReference (replaceIndex);
		face.typefaceName = faceName;
		face.flags = flags;
		face.lastUsageCount = ++counter;
		face.typeface = LookAndFeel::getDefaultLookAndFeel().getTypefaceForFont (font);
		jassert (face.typeface != nullptr); // the look and feel must return a typeface!

		if (defaultFace == nullptr && font == Font())
			defaultFace = face.typeface;

		return face.typeface;
	}

	Typeface::Ptr getDefaultTypeface() const noexcept
	{
		return defaultFace;
	}

private:
	struct CachedFace
	{
		CachedFace() noexcept
			: lastUsageCount (0), flags (-1)
		{
		}

		// Although it seems a bit wacky to store the name here, it's because it may be a
		// placeholder rather than a real one, e.g. "<Sans-Serif>" vs the actual typeface name.
		// Since the typeface itself doesn't know that it may have this alias, the name under
		// which it was fetched needs to be stored separately.
		String typefaceName;
		int lastUsageCount, flags;
		Typeface::Ptr typeface;
	};

	Array <CachedFace> faces;
	Typeface::Ptr defaultFace;
	int counter;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (TypefaceCache);
};

juce_ImplementSingleton_SingleThreaded (TypefaceCache)

void Typeface::setTypefaceCacheSize (int numFontsToCache)
{
	TypefaceCache::getInstance()->setSize (numFontsToCache);
}

Font::SharedFontInternal::SharedFontInternal (const float height_, const int styleFlags_) noexcept
	: typefaceName (Font::getDefaultSansSerifFontName()),
	  height (height_),
	  horizontalScale (1.0f),
	  kerning (0),
	  ascent (0),
	  styleFlags (styleFlags_),
	  typeface ((styleFlags_ & (Font::bold | Font::italic)) == 0
					? TypefaceCache::getInstance()->getDefaultTypeface() : nullptr)
{
}

Font::SharedFontInternal::SharedFontInternal (const String& typefaceName_, const float height_, const int styleFlags_) noexcept
	: typefaceName (typefaceName_),
	  height (height_),
	  horizontalScale (1.0f),
	  kerning (0),
	  ascent (0),
	  styleFlags (styleFlags_),
	  typeface (nullptr)
{
}

Font::SharedFontInternal::SharedFontInternal (const Typeface::Ptr& typeface_) noexcept
	: typefaceName (typeface_->getName()),
	  height (FontValues::defaultFontHeight),
	  horizontalScale (1.0f),
	  kerning (0),
	  ascent (0),
	  styleFlags (Font::plain),
	  typeface (typeface_)
{
}

Font::SharedFontInternal::SharedFontInternal (const SharedFontInternal& other) noexcept
	: typefaceName (other.typefaceName),
	  height (other.height),
	  horizontalScale (other.horizontalScale),
	  kerning (other.kerning),
	  ascent (other.ascent),
	  styleFlags (other.styleFlags),
	  typeface (other.typeface)
{
}

bool Font::SharedFontInternal::operator== (const SharedFontInternal& other) const noexcept
{
	return height == other.height
			&& styleFlags == other.styleFlags
			&& horizontalScale == other.horizontalScale
			&& kerning == other.kerning
			&& typefaceName == other.typefaceName;
}

Font::Font()
	: font (new SharedFontInternal (FontValues::defaultFontHeight, Font::plain))
{
}

Font::Font (const float fontHeight, const int styleFlags_)
	: font (new SharedFontInternal (FontValues::limitFontHeight (fontHeight), styleFlags_))
{
}

Font::Font (const String& typefaceName_, const float fontHeight, const int styleFlags_)
	: font (new SharedFontInternal (typefaceName_, FontValues::limitFontHeight (fontHeight), styleFlags_))
{
}

Font::Font (const Typeface::Ptr& typeface)
	: font (new SharedFontInternal (typeface))
{
}

Font::Font (const Font& other) noexcept
	: font (other.font)
{
}

Font& Font::operator= (const Font& other) noexcept
{
	font = other.font;
	return *this;
}

Font::~Font() noexcept
{
}

bool Font::operator== (const Font& other) const noexcept
{
	return font == other.font
			|| *font == *other.font;
}

bool Font::operator!= (const Font& other) const noexcept
{
	return ! operator== (other);
}

void Font::dupeInternalIfShared()
{
	if (font->getReferenceCount() > 1)
		font = new SharedFontInternal (*font);
}

const String& Font::getDefaultSansSerifFontName()
{
	static const String name ("<Sans-Serif>");
	return name;
}

const String& Font::getDefaultSerifFontName()
{
	static const String name ("<Serif>");
	return name;
}

const String& Font::getDefaultMonospacedFontName()
{
	static const String name ("<Monospaced>");
	return name;
}

void Font::setTypefaceName (const String& faceName)
{
	if (faceName != font->typefaceName)
	{
		dupeInternalIfShared();
		font->typefaceName = faceName;
		font->typeface = nullptr;
		font->ascent = 0;
	}
}

const String& Font::getFallbackFontName()
{
	return FontValues::fallbackFont;
}

void Font::setFallbackFontName (const String& name)
{
	FontValues::fallbackFont = name;
}

void Font::setHeight (float newHeight)
{
	newHeight = FontValues::limitFontHeight (newHeight);

	if (font->height != newHeight)
	{
		dupeInternalIfShared();
		font->height = newHeight;
	}
}

void Font::setHeightWithoutChangingWidth (float newHeight)
{
	newHeight = FontValues::limitFontHeight (newHeight);

	if (font->height != newHeight)
	{
		dupeInternalIfShared();
		font->horizontalScale *= (font->height / newHeight);
		font->height = newHeight;
	}
}

void Font::setStyleFlags (const int newFlags)
{
	if (font->styleFlags != newFlags)
	{
		dupeInternalIfShared();
		font->styleFlags = newFlags;
		font->typeface = nullptr;
		font->ascent = 0;
	}
}

void Font::setSizeAndStyle (float newHeight,
							const int newStyleFlags,
							const float newHorizontalScale,
							const float newKerningAmount)
{
	newHeight = FontValues::limitFontHeight (newHeight);

	if (font->height != newHeight
		 || font->horizontalScale != newHorizontalScale
		 || font->kerning != newKerningAmount)
	{
		dupeInternalIfShared();
		font->height = newHeight;
		font->horizontalScale = newHorizontalScale;
		font->kerning = newKerningAmount;
	}

	setStyleFlags (newStyleFlags);
}

void Font::setHorizontalScale (const float scaleFactor)
{
	dupeInternalIfShared();
	font->horizontalScale = scaleFactor;
}

void Font::setExtraKerningFactor (const float extraKerning)
{
	dupeInternalIfShared();
	font->kerning = extraKerning;
}

void Font::setBold (const bool shouldBeBold)
{
	setStyleFlags (shouldBeBold ? (font->styleFlags | bold)
								: (font->styleFlags & ~bold));
}

Font Font::boldened() const
{
	Font f (*this);
	f.setBold (true);
	return f;
}

bool Font::isBold() const noexcept
{
	return (font->styleFlags & bold) != 0;
}

void Font::setItalic (const bool shouldBeItalic)
{
	setStyleFlags (shouldBeItalic ? (font->styleFlags | italic)
								  : (font->styleFlags & ~italic));
}

Font Font::italicised() const
{
	Font f (*this);
	f.setItalic (true);
	return f;
}

bool Font::isItalic() const noexcept
{
	return (font->styleFlags & italic) != 0;
}

void Font::setUnderline (const bool shouldBeUnderlined)
{
	setStyleFlags (shouldBeUnderlined ? (font->styleFlags | underlined)
									  : (font->styleFlags & ~underlined));
}

bool Font::isUnderlined() const noexcept
{
	return (font->styleFlags & underlined) != 0;
}

float Font::getAscent() const
{
	if (font->ascent == 0)
		font->ascent = getTypeface()->getAscent();

	return font->height * font->ascent;
}

float Font::getDescent() const
{
	return font->height - getAscent();
}

int Font::getStringWidth (const String& text) const
{
	return roundToInt (getStringWidthFloat (text));
}

float Font::getStringWidthFloat (const String& text) const
{
	float w = getTypeface()->getStringWidth (text);

	if (font->kerning != 0)
		w += font->kerning * text.length();

	return w * font->height * font->horizontalScale;
}

void Font::getGlyphPositions (const String& text, Array <int>& glyphs, Array <float>& xOffsets) const
{
	getTypeface()->getGlyphPositions (text, glyphs, xOffsets);

	const float scale = font->height * font->horizontalScale;
	const int num = xOffsets.size();

	if (num > 0)
	{
		float* const x = &(xOffsets.getReference(0));

		if (font->kerning != 0)
		{
			for (int i = 0; i < num; ++i)
				x[i] = (x[i] + i * font->kerning) * scale;
		}
		else
		{
			for (int i = 0; i < num; ++i)
				x[i] *= scale;
		}
	}
}

void Font::findFonts (Array<Font>& destArray)
{
	const StringArray names (findAllTypefaceNames());

	for (int i = 0; i < names.size(); ++i)
		destArray.add (Font (names[i], FontValues::defaultFontHeight, Font::plain));
}

String Font::toString() const
{
	String s (getTypefaceName());

	if (s == getDefaultSansSerifFontName())
		s = String::empty;
	else
		s += "; ";

	s += String (getHeight(), 1);

	if (isBold())
		s += " bold";

	if (isItalic())
		s += " italic";

	return s;
}

Font Font::fromString (const String& fontDescription)
{
	String name;

	const int separator = fontDescription.indexOfChar (';');

	if (separator > 0)
		name = fontDescription.substring (0, separator).trim();

	if (name.isEmpty())
		name = getDefaultSansSerifFontName();

	String sizeAndStyle (fontDescription.substring (separator + 1));

	float height = sizeAndStyle.getFloatValue();
	if (height <= 0)
		height = 10.0f;

	int flags = Font::plain;
	if (sizeAndStyle.containsIgnoreCase ("bold"))
		flags |= Font::bold;
	if (sizeAndStyle.containsIgnoreCase ("italic"))
		flags |= Font::italic;

	return Font (name, height, flags);
}

Typeface* Font::getTypeface() const
{
	if (font->typeface == 0)
		font->typeface = TypefaceCache::getInstance()->findTypefaceFor (*this);

	return font->typeface;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Font.cpp ***/


/*** Start of inlined file: juce_GlyphArrangement.cpp ***/
BEGIN_JUCE_NAMESPACE

PositionedGlyph::PositionedGlyph (const Font& font_, const juce_wchar character_, const int glyph_,
								  const float x_, const float y_, const float w_, const bool whitespace_)
	: font (font_), character (character_), glyph (glyph_),
	  x (x_), y (y_), w (w_), whitespace (whitespace_)
{
}

PositionedGlyph::PositionedGlyph (const PositionedGlyph& other)
	: font (other.font), character (other.character), glyph (other.glyph),
	  x (other.x), y (other.y), w (other.w), whitespace (other.whitespace)
{
}

PositionedGlyph::~PositionedGlyph() {}

PositionedGlyph& PositionedGlyph::operator= (const PositionedGlyph& other)
{
	font = other.font;
	character = other.character;
	glyph = other.glyph;
	x = other.x;
	y = other.y;
	w = other.w;
	whitespace = other.whitespace;
	return *this;
}

void PositionedGlyph::draw (const Graphics& g) const
{
	if (! isWhitespace())
	{
		LowLevelGraphicsContext* const context = g.getInternalContext();
		context->setFont (font);
		context->drawGlyph (glyph, AffineTransform::translation (x, y));
	}
}

void PositionedGlyph::draw (const Graphics& g,
							const AffineTransform& transform) const
{
	if (! isWhitespace())
	{
		LowLevelGraphicsContext* const context = g.getInternalContext();
		context->setFont (font);
		context->drawGlyph (glyph, AffineTransform::translation (x, y)
												   .followedBy (transform));
	}
}

void PositionedGlyph::createPath (Path& path) const
{
	if (! isWhitespace())
	{
		Typeface* const t = font.getTypeface();

		if (t != nullptr)
		{
			Path p;
			t->getOutlineForGlyph (glyph, p);

			path.addPath (p, AffineTransform::scale (font.getHeight() * font.getHorizontalScale(), font.getHeight())
											 .translated (x, y));
		}
	}
}

bool PositionedGlyph::hitTest (float px, float py) const
{
	if (getBounds().contains (px, py) && ! isWhitespace())
	{
		Typeface* const t = font.getTypeface();

		if (t != nullptr)
		{
			Path p;
			t->getOutlineForGlyph (glyph, p);

			AffineTransform::translation (-x, -y)
							.scaled (1.0f / (font.getHeight() * font.getHorizontalScale()), 1.0f / font.getHeight())
							.transformPoint (px, py);

			return p.contains (px, py);
		}
	}

	return false;
}

void PositionedGlyph::moveBy (const float deltaX,
							  const float deltaY)
{
	x += deltaX;
	y += deltaY;
}

GlyphArrangement::GlyphArrangement()
{
	glyphs.ensureStorageAllocated (128);
}

GlyphArrangement::GlyphArrangement (const GlyphArrangement& other)
{
	addGlyphArrangement (other);
}

GlyphArrangement& GlyphArrangement::operator= (const GlyphArrangement& other)
{
	if (this != &other)
	{
		clear();
		addGlyphArrangement (other);
	}

	return *this;
}

GlyphArrangement::~GlyphArrangement()
{
}

void GlyphArrangement::clear()
{
	glyphs.clear();
}

PositionedGlyph& GlyphArrangement::getGlyph (const int index) const
{
	jassert (isPositiveAndBelow (index, glyphs.size()));

	return *glyphs [index];
}

void GlyphArrangement::addGlyphArrangement (const GlyphArrangement& other)
{
	glyphs.ensureStorageAllocated (glyphs.size() + other.glyphs.size());
	glyphs.addCopiesOf (other.glyphs);
}

void GlyphArrangement::addGlyph (const PositionedGlyph& glyph)
{
	glyphs.add (new PositionedGlyph (glyph));
}

void GlyphArrangement::removeRangeOfGlyphs (int startIndex, const int num)
{
	glyphs.removeRange (startIndex, num < 0 ? glyphs.size() : num);
}

void GlyphArrangement::addLineOfText (const Font& font,
									  const String& text,
									  const float xOffset,
									  const float yOffset)
{
	addCurtailedLineOfText (font, text,
							xOffset, yOffset,
							1.0e10f, false);
}

void GlyphArrangement::addCurtailedLineOfText (const Font& font,
											   const String& text,
											   const float xOffset,
											   const float yOffset,
											   const float maxWidthPixels,
											   const bool useEllipsis)
{
	if (text.isNotEmpty())
	{
		Array <int> newGlyphs;
		Array <float> xOffsets;
		font.getGlyphPositions (text, newGlyphs, xOffsets);
		const int textLen = newGlyphs.size();
		glyphs.ensureStorageAllocated (glyphs.size() + textLen);

		String::CharPointerType t (text.getCharPointer());

		for (int i = 0; i < textLen; ++i)
		{
			const float thisX = xOffsets.getUnchecked (i);
			const float nextX = xOffsets.getUnchecked (i + 1);

			if (nextX > maxWidthPixels + 1.0f)
			{
				// curtail the string if it's too wide..
				if (useEllipsis && textLen > 3 && glyphs.size() >= 3)
					insertEllipsis (font, xOffset + maxWidthPixels, 0, glyphs.size());

				break;
			}
			else
			{
				const bool isWhitespace = t.isWhitespace();

				glyphs.add (new PositionedGlyph (font, t.getAndAdvance(),
												 newGlyphs.getUnchecked(i),
												 xOffset + thisX, yOffset,
												 nextX - thisX, isWhitespace));
			}
		}
	}
}

int GlyphArrangement::insertEllipsis (const Font& font, const float maxXPos,
									  const int startIndex, int endIndex)
{
	int numDeleted = 0;

	if (glyphs.size() > 0)
	{
		Array<int> dotGlyphs;
		Array<float> dotXs;
		font.getGlyphPositions ("..", dotGlyphs, dotXs);

		const float dx = dotXs[1];
		float xOffset = 0.0f, yOffset = 0.0f;

		while (endIndex > startIndex)
		{
			const PositionedGlyph* pg = glyphs.getUnchecked (--endIndex);
			xOffset = pg->x;
			yOffset = pg->y;

			glyphs.remove (endIndex);
			++numDeleted;

			if (xOffset + dx * 3 <= maxXPos)
				break;
		}

		for (int i = 3; --i >= 0;)
		{
			glyphs.insert (endIndex++, new PositionedGlyph (font, '.', dotGlyphs.getFirst(),
															xOffset, yOffset, dx, false));
			--numDeleted;
			xOffset += dx;

			if (xOffset > maxXPos)
				break;
		}
	}

	return numDeleted;
}

void GlyphArrangement::addJustifiedText (const Font& font,
										 const String& text,
										 float x, float y,
										 const float maxLineWidth,
										 const Justification& horizontalLayout)
{
	int lineStartIndex = glyphs.size();
	addLineOfText (font, text, x, y);

	const float originalY = y;

	while (lineStartIndex < glyphs.size())
	{
		int i = lineStartIndex;

		if (glyphs.getUnchecked(i)->getCharacter() != '\n'
			  && glyphs.getUnchecked(i)->getCharacter() != '\r')
			++i;

		const float lineMaxX = glyphs.getUnchecked (lineStartIndex)->getLeft() + maxLineWidth;
		int lastWordBreakIndex = -1;

		while (i < glyphs.size())
		{
			const PositionedGlyph* pg = glyphs.getUnchecked (i);
			const juce_wchar c = pg->getCharacter();

			if (c == '\r' || c == '\n')
			{
				++i;

				if (c == '\r' && i < glyphs.size()
					 && glyphs.getUnchecked(i)->getCharacter() == '\n')
					++i;

				break;
			}
			else if (pg->isWhitespace())
			{
				lastWordBreakIndex = i + 1;
			}
			else if (pg->getRight() - 0.0001f >= lineMaxX)
			{
				if (lastWordBreakIndex >= 0)
					i = lastWordBreakIndex;

				break;
			}

			++i;
		}

		const float currentLineStartX = glyphs.getUnchecked (lineStartIndex)->getLeft();
		float currentLineEndX = currentLineStartX;

		for (int j = i; --j >= lineStartIndex;)
		{
			if (! glyphs.getUnchecked (j)->isWhitespace())
			{
				currentLineEndX = glyphs.getUnchecked (j)->getRight();
				break;
			}
		}

		float deltaX = 0.0f;

		if (horizontalLayout.testFlags (Justification::horizontallyJustified))
			spreadOutLine (lineStartIndex, i - lineStartIndex, maxLineWidth);
		else if (horizontalLayout.testFlags (Justification::horizontallyCentred))
			deltaX = (maxLineWidth - (currentLineEndX - currentLineStartX)) * 0.5f;
		else if (horizontalLayout.testFlags (Justification::right))
			deltaX = maxLineWidth - (currentLineEndX - currentLineStartX);

		moveRangeOfGlyphs (lineStartIndex, i - lineStartIndex,
						   x + deltaX - currentLineStartX, y - originalY);

		lineStartIndex = i;

		y += font.getHeight();
	}
}

void GlyphArrangement::addFittedText (const Font& f,
									  const String& text,
									  const float x, const float y,
									  const float width, const float height,
									  const Justification& layout,
									  int maximumLines,
									  const float minimumHorizontalScale)
{
	// doesn't make much sense if this is outside a sensible range of 0.5 to 1.0
	jassert (minimumHorizontalScale > 0 && minimumHorizontalScale <= 1.0f);

	if (text.containsAnyOf ("\r\n"))
	{
		GlyphArrangement ga;
		ga.addJustifiedText (f, text, x, y, width, layout);

		const Rectangle<float> bb (ga.getBoundingBox (0, -1, false));

		float dy = y - bb.getY();

		if (layout.testFlags (Justification::verticallyCentred))
			dy += (height - bb.getHeight()) * 0.5f;
		else if (layout.testFlags (Justification::bottom))
			dy += height - bb.getHeight();

		ga.moveRangeOfGlyphs (0, -1, 0.0f, dy);

		glyphs.ensureStorageAllocated (glyphs.size() + ga.glyphs.size());

		for (int i = 0; i < ga.glyphs.size(); ++i)
			glyphs.add (ga.glyphs.getUnchecked (i));

		ga.glyphs.clear (false);
		return;
	}

	int startIndex = glyphs.size();
	addLineOfText (f, text.trim(), x, y);

	if (glyphs.size() > startIndex)
	{
		float lineWidth = glyphs.getUnchecked (glyphs.size() - 1)->getRight()
							- glyphs.getUnchecked (startIndex)->getLeft();

		if (lineWidth <= 0)
			return;

		if (lineWidth * minimumHorizontalScale < width)
		{
			if (lineWidth > width)
				stretchRangeOfGlyphs (startIndex, glyphs.size() - startIndex,
									  width / lineWidth);

			justifyGlyphs (startIndex, glyphs.size() - startIndex,
						   x, y, width, height, layout);
		}
		else if (maximumLines <= 1)
		{
			fitLineIntoSpace (startIndex, glyphs.size() - startIndex,
							  x, y, width, height, f, layout, minimumHorizontalScale);
		}
		else
		{
			Font font (f);
			String txt (text.trim());
			const int length = txt.length();
			const int originalStartIndex = startIndex;
			int numLines = 1;

			if (length <= 12 && ! txt.containsAnyOf (" -\t\r\n"))
				maximumLines = 1;

			maximumLines = jmin (maximumLines, length);

			while (numLines < maximumLines)
			{
				++numLines;

				const float newFontHeight = height / (float) numLines;

				if (newFontHeight < font.getHeight())
				{
					font.setHeight (jmax (8.0f, newFontHeight));

					removeRangeOfGlyphs (startIndex, -1);
					addLineOfText (font, txt, x, y);

					lineWidth = glyphs.getUnchecked (glyphs.size() - 1)->getRight()
									- glyphs.getUnchecked (startIndex)->getLeft();
				}

				if (numLines > lineWidth / width || newFontHeight < 8.0f)
					break;
			}

			if (numLines < 1)
				numLines = 1;

			float lineY = y;
			float widthPerLine = lineWidth / numLines;
			int lastLineStartIndex = 0;

			for (int line = 0; line < numLines; ++line)
			{
				int i = startIndex;
				lastLineStartIndex = i;
				float lineStartX = glyphs.getUnchecked (startIndex)->getLeft();

				if (line == numLines - 1)
				{
					widthPerLine = width;
					i = glyphs.size();
				}
				else
				{
					while (i < glyphs.size())
					{
						lineWidth = (glyphs.getUnchecked (i)->getRight() - lineStartX);

						if (lineWidth > widthPerLine)
						{
							// got to a point where the line's too long, so skip forward to find a
							// good place to break it..
							const int searchStartIndex = i;

							while (i < glyphs.size())
							{
								if ((glyphs.getUnchecked (i)->getRight() - lineStartX) * minimumHorizontalScale < width)
								{
									if (glyphs.getUnchecked (i)->isWhitespace()
										 || glyphs.getUnchecked (i)->getCharacter() == '-')
									{
										++i;
										break;
									}
								}
								else
								{
									// can't find a suitable break, so try looking backwards..
									i = searchStartIndex;

									for (int back = 1; back < jmin (5, i - startIndex - 1); ++back)
									{
										if (glyphs.getUnchecked (i - back)->isWhitespace()
											 || glyphs.getUnchecked (i - back)->getCharacter() == '-')
										{
											i -= back - 1;
											break;
										}
									}

									break;
								}

								++i;
							}

							break;
						}

						++i;
					}

					int wsStart = i;
					while (wsStart > 0 && glyphs.getUnchecked (wsStart - 1)->isWhitespace())
						--wsStart;

					int wsEnd = i;

					while (wsEnd < glyphs.size() && glyphs.getUnchecked (wsEnd)->isWhitespace())
						++wsEnd;

					removeRangeOfGlyphs (wsStart, wsEnd - wsStart);
					i = jmax (wsStart, startIndex + 1);
				}

				i -= fitLineIntoSpace (startIndex, i - startIndex,
									   x, lineY, width, font.getHeight(), font,
									   layout.getOnlyHorizontalFlags() | Justification::verticallyCentred,
									   minimumHorizontalScale);

				startIndex = i;
				lineY += font.getHeight();

				if (startIndex >= glyphs.size())
					break;
			}

			justifyGlyphs (originalStartIndex, glyphs.size() - originalStartIndex,
						   x, y, width, height, layout.getFlags() & ~Justification::horizontallyJustified);
		}
	}
}

void GlyphArrangement::moveRangeOfGlyphs (int startIndex, int num,
										  const float dx, const float dy)
{
	jassert (startIndex >= 0);

	if (dx != 0.0f || dy != 0.0f)
	{
		if (num < 0 || startIndex + num > glyphs.size())
			num = glyphs.size() - startIndex;

		while (--num >= 0)
			glyphs.getUnchecked (startIndex++)->moveBy (dx, dy);
	}
}

int GlyphArrangement::fitLineIntoSpace (int start, int numGlyphs, float x, float y, float w, float h, const Font& font,
										const Justification& justification, float minimumHorizontalScale)
{
	int numDeleted = 0;
	const float lineStartX = glyphs.getUnchecked (start)->getLeft();
	float lineWidth = glyphs.getUnchecked (start + numGlyphs - 1)->getRight() - lineStartX;

	if (lineWidth > w)
	{
		if (minimumHorizontalScale < 1.0f)
		{
			stretchRangeOfGlyphs (start, numGlyphs, jmax (minimumHorizontalScale, w / lineWidth));
			lineWidth = glyphs.getUnchecked (start + numGlyphs - 1)->getRight() - lineStartX - 0.5f;
		}

		if (lineWidth > w)
		{
			numDeleted = insertEllipsis (font, lineStartX + w, start, start + numGlyphs);
			numGlyphs -= numDeleted;
		}
	}

	justifyGlyphs (start, numGlyphs, x, y, w, h, justification);
	return numDeleted;
}

void GlyphArrangement::stretchRangeOfGlyphs (int startIndex, int num,
											 const float horizontalScaleFactor)
{
	jassert (startIndex >= 0);

	if (num < 0 || startIndex + num > glyphs.size())
		num = glyphs.size() - startIndex;

	if (num > 0)
	{
		const float xAnchor = glyphs.getUnchecked (startIndex)->getLeft();

		while (--num >= 0)
		{
			PositionedGlyph* const pg = glyphs.getUnchecked (startIndex++);

			pg->x = xAnchor + (pg->x - xAnchor) * horizontalScaleFactor;
			pg->font.setHorizontalScale (pg->font.getHorizontalScale() * horizontalScaleFactor);
			pg->w *= horizontalScaleFactor;
		}
	}
}

const Rectangle<float> GlyphArrangement::getBoundingBox (int startIndex, int num, const bool includeWhitespace) const
{
	jassert (startIndex >= 0);

	if (num < 0 || startIndex + num > glyphs.size())
		num = glyphs.size() - startIndex;

	Rectangle<float> result;

	while (--num >= 0)
	{
		const PositionedGlyph* const pg = glyphs.getUnchecked (startIndex++);

		if (includeWhitespace || ! pg->isWhitespace())
			result = result.getUnion (pg->getBounds());
	}

	return result;
}

void GlyphArrangement::justifyGlyphs (const int startIndex, const int num,
									  const float x, const float y, const float width, const float height,
									  const Justification& justification)
{
	jassert (num >= 0 && startIndex >= 0);

	if (glyphs.size() > 0 && num > 0)
	{
		const Rectangle<float> bb (getBoundingBox (startIndex, num, ! justification.testFlags (Justification::horizontallyJustified
																								| Justification::horizontallyCentred)));
		float deltaX = 0.0f;

		if (justification.testFlags (Justification::horizontallyJustified))
			deltaX = x - bb.getX();
		else if (justification.testFlags (Justification::horizontallyCentred))
			deltaX = x + (width - bb.getWidth()) * 0.5f - bb.getX();
		else if (justification.testFlags (Justification::right))
			deltaX = (x + width) - bb.getRight();
		else
			deltaX = x - bb.getX();

		float deltaY = 0.0f;

		if (justification.testFlags (Justification::top))
			deltaY = y - bb.getY();
		else if (justification.testFlags (Justification::bottom))
			deltaY = (y + height) - bb.getBottom();
		else
			deltaY = y + (height - bb.getHeight()) * 0.5f - bb.getY();

		moveRangeOfGlyphs (startIndex, num, deltaX, deltaY);

		if (justification.testFlags (Justification::horizontallyJustified))
		{
			int lineStart = 0;
			float baseY = glyphs.getUnchecked (startIndex)->getBaselineY();

			int i;
			for (i = 0; i < num; ++i)
			{
				const float glyphY = glyphs.getUnchecked (startIndex + i)->getBaselineY();

				if (glyphY != baseY)
				{
					spreadOutLine (startIndex + lineStart, i - lineStart, width);

					lineStart = i;
					baseY = glyphY;
				}
			}

			if (i > lineStart)
				spreadOutLine (startIndex + lineStart, i - lineStart, width);
		}
	}
}

void GlyphArrangement::spreadOutLine (const int start, const int num, const float targetWidth)
{
	if (start + num < glyphs.size()
		 && glyphs.getUnchecked (start + num - 1)->getCharacter() != '\r'
		 && glyphs.getUnchecked (start + num - 1)->getCharacter() != '\n')
	{
		int numSpaces = 0;
		int spacesAtEnd = 0;

		for (int i = 0; i < num; ++i)
		{
			if (glyphs.getUnchecked (start + i)->isWhitespace())
			{
				++spacesAtEnd;
				++numSpaces;
			}
			else
			{
				spacesAtEnd = 0;
			}
		}

		numSpaces -= spacesAtEnd;

		if (numSpaces > 0)
		{
			const float startX = glyphs.getUnchecked (start)->getLeft();
			const float endX = glyphs.getUnchecked (start + num - 1 - spacesAtEnd)->getRight();

			const float extraPaddingBetweenWords
				= (targetWidth - (endX - startX)) / (float) numSpaces;

			float deltaX = 0.0f;

			for (int i = 0; i < num; ++i)
			{
				glyphs.getUnchecked (start + i)->moveBy (deltaX, 0.0f);

				if (glyphs.getUnchecked (start + i)->isWhitespace())
					deltaX += extraPaddingBetweenWords;
			}
		}
	}
}

void GlyphArrangement::draw (const Graphics& g) const
{
	for (int i = 0; i < glyphs.size(); ++i)
	{
		const PositionedGlyph* const pg = glyphs.getUnchecked(i);

		if (pg->font.isUnderlined())
		{
			const float lineThickness = (pg->font.getDescent()) * 0.3f;

			float nextX = pg->x + pg->w;

			if (i < glyphs.size() - 1 && glyphs.getUnchecked (i + 1)->y == pg->y)
				nextX = glyphs.getUnchecked (i + 1)->x;

			g.fillRect (pg->x, pg->y + lineThickness * 2.0f,
						nextX - pg->x, lineThickness);
		}

		pg->draw (g);
	}
}

void GlyphArrangement::draw (const Graphics& g, const AffineTransform& transform) const
{
	for (int i = 0; i < glyphs.size(); ++i)
	{
		const PositionedGlyph* const pg = glyphs.getUnchecked(i);

		if (pg->font.isUnderlined())
		{
			const float lineThickness = (pg->font.getDescent()) * 0.3f;

			float nextX = pg->x + pg->w;

			if (i < glyphs.size() - 1 && glyphs.getUnchecked (i + 1)->y == pg->y)
				nextX = glyphs.getUnchecked (i + 1)->x;

			Path p;
			p.addLineSegment (Line<float> (pg->x, pg->y + lineThickness * 2.0f,
										   nextX, pg->y + lineThickness * 2.0f),
							  lineThickness);

			g.fillPath (p, transform);
		}

		pg->draw (g, transform);
	}
}

void GlyphArrangement::createPath (Path& path) const
{
	for (int i = 0; i < glyphs.size(); ++i)
		glyphs.getUnchecked (i)->createPath (path);
}

int GlyphArrangement::findGlyphIndexAt (float x, float y) const
{
	for (int i = 0; i < glyphs.size(); ++i)
		if (glyphs.getUnchecked (i)->hitTest (x, y))
			return i;

	return -1;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_GlyphArrangement.cpp ***/


/*** Start of inlined file: juce_TextLayout.cpp ***/
BEGIN_JUCE_NAMESPACE

class TextLayout::Token
{
public:
	Token (const String& t,
		   const Font& f,
		   const bool isWhitespace_)
		: text (t),
		  font (f),
		  x(0),
		  y(0),
		  isWhitespace (isWhitespace_)
	{
		w = font.getStringWidth (t);
		h = roundToInt (f.getHeight());
		isNewLine = t.containsChar ('\n') || t.containsChar ('\r');
	}

	Token (const Token& other)
		: text (other.text),
		  font (other.font),
		  x (other.x),
		  y (other.y),
		  w (other.w),
		  h (other.h),
		  line (other.line),
		  lineHeight (other.lineHeight),
		  isWhitespace (other.isWhitespace),
		  isNewLine (other.isNewLine)
	{
	}

	void draw (Graphics& g,
			   const int xOffset,
			   const int yOffset)
	{
		if (! isWhitespace)
		{
			g.setFont (font);
			g.drawSingleLineText (text.trimEnd(),
								  xOffset + x,
								  yOffset + y + (lineHeight - h)
									+ roundToInt (font.getAscent()));
		}
	}

	String text;
	Font font;
	int x, y, w, h;
	int line, lineHeight;
	bool isWhitespace, isNewLine;

private:
	JUCE_LEAK_DETECTOR (Token);
};

TextLayout::TextLayout()
	: totalLines (0)
{
	tokens.ensureStorageAllocated (64);
}

TextLayout::TextLayout (const String& text, const Font& font)
	: totalLines (0)
{
	tokens.ensureStorageAllocated (64);
	appendText (text, font);
}

TextLayout::TextLayout (const TextLayout& other)
	: totalLines (0)
{
	*this = other;
}

TextLayout& TextLayout::operator= (const TextLayout& other)
{
	if (this != &other)
	{
		clear();

		totalLines = other.totalLines;
		tokens.addCopiesOf (other.tokens);
	}

	return *this;
}

TextLayout::~TextLayout()
{
	clear();
}

void TextLayout::clear()
{
	tokens.clear();
	totalLines = 0;
}

bool TextLayout::isEmpty() const
{
	return tokens.size() == 0;
}

void TextLayout::appendText (const String& text, const Font& font)
{
	String::CharPointerType t (text.getCharPointer());
	String currentString;
	int lastCharType = 0;

	for (;;)
	{
		const juce_wchar c = t.getAndAdvance();
		if (c == 0)
			break;

		int charType;
		if (c == '\r' || c == '\n')
		{
			charType = 0;
		}
		else if (CharacterFunctions::isWhitespace (c))
		{
			charType = 2;
		}
		else
		{
			charType = 1;
		}

		if (charType == 0 || charType != lastCharType)
		{
			if (currentString.isNotEmpty())
			{
				tokens.add (new Token (currentString, font,
									   lastCharType == 2 || lastCharType == 0));
			}

			currentString = String::charToString (c);

			if (c == '\r' && *t == '\n')
				currentString += t.getAndAdvance();
		}
		else
		{
			currentString += c;
		}

		lastCharType = charType;
	}

	if (currentString.isNotEmpty())
		tokens.add (new Token (currentString, font, lastCharType == 2));
}

void TextLayout::setText (const String& text, const Font& font)
{
	clear();
	appendText (text, font);
}

void TextLayout::layout (int maxWidth,
						 const Justification& justification,
						 const bool attemptToBalanceLineLengths)
{
	if (attemptToBalanceLineLengths)
	{
		const int originalW = maxWidth;
		int bestWidth = maxWidth;
		float bestLineProportion = 0.0f;

		while (maxWidth > originalW / 2)
		{
			layout (maxWidth, justification, false);

			if (getNumLines() <= 1)
				return;

			const int lastLineW = getLineWidth (getNumLines() - 1);
			const int lastButOneLineW = getLineWidth (getNumLines() - 2);

			const float prop = lastLineW / (float) lastButOneLineW;

			if (prop > 0.9f)
				return;

			if (prop > bestLineProportion)
			{
				bestLineProportion = prop;
				bestWidth = maxWidth;
			}

			maxWidth -= 10;
		}

		layout (bestWidth, justification, false);
	}
	else
	{
		int x = 0;
		int y = 0;
		int h = 0;
		totalLines = 0;
		int i;

		for (i = 0; i < tokens.size(); ++i)
		{
			Token* const t = tokens.getUnchecked(i);
			t->x = x;
			t->y = y;
			t->line = totalLines;
			x += t->w;
			h = jmax (h, t->h);

			const Token* nextTok = tokens [i + 1];

			if (nextTok == 0)
				break;

			if (t->isNewLine || ((! nextTok->isWhitespace) && x + nextTok->w > maxWidth))
			{
				// finished a line, so go back and update the heights of the things on it
				for (int j = i; j >= 0; --j)
				{
					Token* const tok = tokens.getUnchecked(j);

					if (tok->line == totalLines)
						tok->lineHeight = h;
					else
						break;
				}

				x = 0;
				y += h;
				h = 0;
				++totalLines;
			}
		}

		// finished a line, so go back and update the heights of the things on it
		for (int j = jmin (i, tokens.size() - 1); j >= 0; --j)
		{
			Token* const t = tokens.getUnchecked(j);

			if (t->line == totalLines)
				t->lineHeight = h;
			else
				break;
		}

		++totalLines;

		if (! justification.testFlags (Justification::left))
		{
			int totalW = getWidth();

			for (i = totalLines; --i >= 0;)
			{
				const int lineW = getLineWidth (i);

				int dx = 0;
				if (justification.testFlags (Justification::horizontallyCentred))
					dx = (totalW - lineW) / 2;
				else if (justification.testFlags (Justification::right))
					dx = totalW - lineW;

				for (int j = tokens.size(); --j >= 0;)
				{
					Token* const t = tokens.getUnchecked(j);

					if (t->line == i)
						t->x += dx;
				}
			}
		}
	}
}

int TextLayout::getLineWidth (const int lineNumber) const
{
	int maxW = 0;

	for (int i = tokens.size(); --i >= 0;)
	{
		const Token* const t = tokens.getUnchecked(i);

		if (t->line == lineNumber && ! t->isWhitespace)
			maxW = jmax (maxW, t->x + t->w);
	}

	return maxW;
}

int TextLayout::getWidth() const
{
	int maxW = 0;

	for (int i = tokens.size(); --i >= 0;)
	{
		const Token* const t = tokens.getUnchecked(i);
		if (! t->isWhitespace)
			maxW = jmax (maxW, t->x + t->w);
	}

	return maxW;
}

int TextLayout::getHeight() const
{
	int maxH = 0;

	for (int i = tokens.size(); --i >= 0;)
	{
		const Token* const t = tokens.getUnchecked(i);

		if (! t->isWhitespace)
			maxH = jmax (maxH, t->y + t->h);
	}

	return maxH;
}

void TextLayout::draw (Graphics& g,
					   const int xOffset,
					   const int yOffset) const
{
	for (int i = tokens.size(); --i >= 0;)
		tokens.getUnchecked(i)->draw (g, xOffset, yOffset);
}

void TextLayout::drawWithin (Graphics& g,
							 int x, int y, int w, int h,
							 const Justification& justification) const
{
	justification.applyToRectangle (x, y, getWidth(), getHeight(),
									x, y, w, h);

	draw (g, x, y);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_TextLayout.cpp ***/


/*** Start of inlined file: juce_Typeface.cpp ***/
BEGIN_JUCE_NAMESPACE

Typeface::Typeface (const String& name_) noexcept
	: name (name_)
{
}

Typeface::~Typeface()
{
}

Typeface::Ptr Typeface::getFallbackTypeface()
{
	const Font fallbackFont (Font::getFallbackFontName(), 10, 0);
	return fallbackFont.getTypeface();
}

EdgeTable* Typeface::getEdgeTableForGlyph (int glyphNumber, const AffineTransform& transform)
{
	Path path;

	if (getOutlineForGlyph (glyphNumber, path) && ! path.isEmpty())
		return new EdgeTable (path.getBoundsTransformed (transform).getSmallestIntegerContainer().expanded (1, 0),
							  path, transform);

	return nullptr;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Typeface.cpp ***/


/*** Start of inlined file: juce_CustomTypeface.cpp ***/
BEGIN_JUCE_NAMESPACE

class CustomTypeface::GlyphInfo
{
public:
	GlyphInfo (const juce_wchar character_, const Path& path_, const float width_) noexcept
		: character (character_), path (path_), width (width_)
	{
	}

	struct KerningPair
	{
		juce_wchar character2;
		float kerningAmount;
	};

	void addKerningPair (const juce_wchar subsequentCharacter,
						 const float extraKerningAmount) noexcept
	{
		KerningPair kp;
		kp.character2 = subsequentCharacter;
		kp.kerningAmount = extraKerningAmount;
		kerningPairs.add (kp);
	}

	float getHorizontalSpacing (const juce_wchar subsequentCharacter) const noexcept
	{
		if (subsequentCharacter != 0)
		{
			for (int i = kerningPairs.size(); --i >= 0;)
				if (kerningPairs.getReference(i).character2 == subsequentCharacter)
					return width + kerningPairs.getReference(i).kerningAmount;
		}

		return width;
	}

	const juce_wchar character;
	const Path path;
	float width;
	Array <KerningPair> kerningPairs;

private:
	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (GlyphInfo);
};

namespace CustomTypefaceHelpers
{
	juce_wchar readChar (InputStream& in)
	{
		uint32 n = (uint32) (uint16) in.readShort();

		if (n >= 0xd800 && n <= 0xdfff)
		{
			const uint32 nextWord = (uint32) (uint16) in.readShort();
			jassert (nextWord >= 0xdc00); // illegal unicode character!

			n = 0x10000 + (((n - 0xd800) << 10) | (nextWord - 0xdc00));
		}

		return (juce_wchar) n;
	}

	void writeChar (OutputStream& out, juce_wchar charToWrite)
	{
		if (charToWrite >= 0x10000)
		{
			charToWrite -= 0x10000;
			out.writeShort ((short) (uint16) (0xd800 + (charToWrite >> 10)));
			out.writeShort ((short) (uint16) (0xdc00 + (charToWrite & 0x3ff)));
		}
		else
		{
			out.writeShort ((short) (uint16) charToWrite);
		}
	}
}

CustomTypeface::CustomTypeface()
	: Typeface (String::empty)
{
	clear();
}


CustomTypeface::~CustomTypeface()
{
}

void CustomTypeface::clear()
{
	defaultCharacter = 0;
	ascent = 1.0f;
	isBold = isItalic = false;
	zeromem (lookupTable, sizeof (lookupTable));
	glyphs.clear();
}

void CustomTypeface::setCharacteristics (const String& name_, const float ascent_, const bool isBold_,
										 const bool isItalic_, const juce_wchar defaultCharacter_) noexcept
{
	name = name_;
	defaultCharacter = defaultCharacter_;
	ascent = ascent_;
	isBold = isBold_;
	isItalic = isItalic_;
}

void CustomTypeface::addGlyph (const juce_wchar character, const Path& path, const float width) noexcept
{
	// Check that you're not trying to add the same character twice..
	jassert (findGlyph (character, false) == nullptr);

	if (isPositiveAndBelow ((int) character, (int) numElementsInArray (lookupTable)))
		lookupTable [character] = (short) glyphs.size();

	glyphs.add (new GlyphInfo (character, path, width));
}

void CustomTypeface::addKerningPair (const juce_wchar char1, const juce_wchar char2, const float extraAmount) noexcept
{
	if (extraAmount != 0)
	{
		GlyphInfo* const g = findGlyph (char1, true);
		jassert (g != nullptr); // can only add kerning pairs for characters that exist!

		if (g != nullptr)
			g->addKerningPair (char2, extraAmount);
	}
}

CustomTypeface::GlyphInfo* CustomTypeface::findGlyph (const juce_wchar character, const bool loadIfNeeded) noexcept
{
	if (isPositiveAndBelow ((int) character, (int) numElementsInArray (lookupTable)) && lookupTable [character] > 0)
		return glyphs [(int) lookupTable [(int) character]];

	for (int i = 0; i < glyphs.size(); ++i)
	{
		GlyphInfo* const g = glyphs.getUnchecked(i);
		if (g->character == character)
			return g;
	}

	if (loadIfNeeded && loadGlyphIfPossible (character))
		return findGlyph (character, false);

	return nullptr;
}

bool CustomTypeface::loadGlyphIfPossible (const juce_wchar /*characterNeeded*/)
{
	return false;
}

void CustomTypeface::addGlyphsFromOtherTypeface (Typeface& typefaceToCopy, juce_wchar characterStartIndex, int numCharacters) noexcept
{
	setCharacteristics (name, typefaceToCopy.getAscent(), isBold, isItalic, defaultCharacter);

	for (int i = 0; i < numCharacters; ++i)
	{
		const juce_wchar c = (juce_wchar) (characterStartIndex + i);

		Array <int> glyphIndexes;
		Array <float> offsets;
		typefaceToCopy.getGlyphPositions (String::charToString (c), glyphIndexes, offsets);

		const int glyphIndex = glyphIndexes.getFirst();

		if (glyphIndex >= 0 && glyphIndexes.size() > 0)
		{
			const float glyphWidth = offsets[1];

			Path p;
			typefaceToCopy.getOutlineForGlyph (glyphIndex, p);

			addGlyph (c, p, glyphWidth);

			for (int j = glyphs.size() - 1; --j >= 0;)
			{
				const juce_wchar char2 = glyphs.getUnchecked (j)->character;
				glyphIndexes.clearQuick();
				offsets.clearQuick();
				typefaceToCopy.getGlyphPositions (String::charToString (c) + String::charToString (char2), glyphIndexes, offsets);

				if (offsets.size() > 1)
					addKerningPair (c, char2, offsets[1] - glyphWidth);
			}
		}
	}
}

float CustomTypeface::getAscent() const
{
	return ascent;
}

float CustomTypeface::getDescent() const
{
	return 1.0f - ascent;
}

float CustomTypeface::getStringWidth (const String& text)
{
	float x = 0;
	String::CharPointerType t (text.getCharPointer());

	while (! t.isEmpty())
	{
		const juce_wchar c = t.getAndAdvance();
		const GlyphInfo* const glyph = findGlyph (c, true);

		if (glyph == nullptr)
		{
			const Typeface::Ptr fallbackTypeface (Typeface::getFallbackTypeface());

			if (fallbackTypeface != nullptr && fallbackTypeface != this)
				x += fallbackTypeface->getStringWidth (String::charToString (c));
		}

		if (glyph != nullptr)
			x += glyph->getHorizontalSpacing (*t);
	}

	return x;
}

void CustomTypeface::getGlyphPositions (const String& text, Array <int>& resultGlyphs, Array<float>& xOffsets)
{
	xOffsets.add (0);
	float x = 0;
	String::CharPointerType t (text.getCharPointer());

	while (! t.isEmpty())
	{
		const juce_wchar c = t.getAndAdvance();
		const GlyphInfo* const glyph = findGlyph (c, true);

		if (glyph == nullptr)
		{
			const Typeface::Ptr fallbackTypeface (Typeface::getFallbackTypeface());

			if (fallbackTypeface != nullptr && fallbackTypeface != this)
			{
				Array <int> subGlyphs;
				Array <float> subOffsets;
				fallbackTypeface->getGlyphPositions (String::charToString (c), subGlyphs, subOffsets);

				if (subGlyphs.size() > 0)
				{
					resultGlyphs.add (subGlyphs.getFirst());
					x += subOffsets[1];
					xOffsets.add (x);
				}
			}
		}

		if (glyph != nullptr)
		{
			x += glyph->getHorizontalSpacing (*t);
			resultGlyphs.add ((int) glyph->character);
			xOffsets.add (x);
		}
	}
}

bool CustomTypeface::getOutlineForGlyph (int glyphNumber, Path& path)
{
	const GlyphInfo* const glyph = findGlyph ((juce_wchar) glyphNumber, true);

	if (glyph == nullptr)
	{
		const Typeface::Ptr fallbackTypeface (Typeface::getFallbackTypeface());

		if (fallbackTypeface != nullptr && fallbackTypeface != this)
			fallbackTypeface->getOutlineForGlyph (glyphNumber, path);
	}

	if (glyph != nullptr)
	{
		path = glyph->path;
		return true;
	}

	return false;
}

EdgeTable* CustomTypeface::getEdgeTableForGlyph (int glyphNumber, const AffineTransform& transform)
{
	const GlyphInfo* const glyph = findGlyph ((juce_wchar) glyphNumber, true);

	if (glyph == nullptr)
	{
		const Typeface::Ptr fallbackTypeface (Typeface::getFallbackTypeface());

		if (fallbackTypeface != nullptr && fallbackTypeface != this)
			return fallbackTypeface->getEdgeTableForGlyph (glyphNumber, transform);
	}

	if (glyph != nullptr && ! glyph->path.isEmpty())
		return new EdgeTable (glyph->path.getBoundsTransformed (transform).getSmallestIntegerContainer().expanded (1, 0),
							  glyph->path, transform);

	return nullptr;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_CustomTypeface.cpp ***/


/*** Start of inlined file: juce_AffineTransform.cpp ***/
BEGIN_JUCE_NAMESPACE

AffineTransform::AffineTransform() noexcept
	: mat00 (1.0f), mat01 (0), mat02 (0),
	  mat10 (0), mat11 (1.0f), mat12 (0)
{
}

AffineTransform::AffineTransform (const AffineTransform& other) noexcept
  : mat00 (other.mat00), mat01 (other.mat01), mat02 (other.mat02),
	mat10 (other.mat10), mat11 (other.mat11), mat12 (other.mat12)
{
}

AffineTransform::AffineTransform (const float mat00_, const float mat01_, const float mat02_,
								  const float mat10_, const float mat11_, const float mat12_) noexcept
 :  mat00 (mat00_), mat01 (mat01_), mat02 (mat02_),
	mat10 (mat10_), mat11 (mat11_), mat12 (mat12_)
{
}

AffineTransform& AffineTransform::operator= (const AffineTransform& other) noexcept
{
	mat00 = other.mat00;
	mat01 = other.mat01;
	mat02 = other.mat02;
	mat10 = other.mat10;
	mat11 = other.mat11;
	mat12 = other.mat12;

	return *this;
}

bool AffineTransform::operator== (const AffineTransform& other) const noexcept
{
	return mat00 == other.mat00
		&& mat01 == other.mat01
		&& mat02 == other.mat02
		&& mat10 == other.mat10
		&& mat11 == other.mat11
		&& mat12 == other.mat12;
}

bool AffineTransform::operator!= (const AffineTransform& other) const noexcept
{
	return ! operator== (other);
}

bool AffineTransform::isIdentity() const noexcept
{
	return (mat01 == 0)
		&& (mat02 == 0)
		&& (mat10 == 0)
		&& (mat12 == 0)
		&& (mat00 == 1.0f)
		&& (mat11 == 1.0f);
}

const AffineTransform AffineTransform::identity;

AffineTransform AffineTransform::followedBy (const AffineTransform& other) const noexcept
{
	return AffineTransform (other.mat00 * mat00 + other.mat01 * mat10,
							other.mat00 * mat01 + other.mat01 * mat11,
							other.mat00 * mat02 + other.mat01 * mat12 + other.mat02,
							other.mat10 * mat00 + other.mat11 * mat10,
							other.mat10 * mat01 + other.mat11 * mat11,
							other.mat10 * mat02 + other.mat11 * mat12 + other.mat12);
}

AffineTransform AffineTransform::translated (const float dx, const float dy) const noexcept
{
	return AffineTransform (mat00, mat01, mat02 + dx,
							mat10, mat11, mat12 + dy);
}

AffineTransform AffineTransform::translation (const float dx, const float dy) noexcept
{
	return AffineTransform (1.0f, 0, dx,
							0, 1.0f, dy);
}

AffineTransform AffineTransform::rotated (const float rad) const noexcept
{
	const float cosRad = std::cos (rad);
	const float sinRad = std::sin (rad);

	return AffineTransform (cosRad * mat00 + -sinRad * mat10,
							cosRad * mat01 + -sinRad * mat11,
							cosRad * mat02 + -sinRad * mat12,
							sinRad * mat00 + cosRad * mat10,
							sinRad * mat01 + cosRad * mat11,
							sinRad * mat02 + cosRad * mat12);
}

AffineTransform AffineTransform::rotation (const float rad) noexcept
{
	const float cosRad = std::cos (rad);
	const float sinRad = std::sin (rad);

	return AffineTransform (cosRad, -sinRad, 0,
							sinRad, cosRad, 0);
}

AffineTransform AffineTransform::rotation (const float rad, const float pivotX, const float pivotY) noexcept
{
	const float cosRad = std::cos (rad);
	const float sinRad = std::sin (rad);

	return AffineTransform (cosRad, -sinRad, -cosRad * pivotX + sinRad * pivotY + pivotX,
							sinRad, cosRad, -sinRad * pivotX + -cosRad * pivotY + pivotY);
}

AffineTransform AffineTransform::rotated (const float angle, const float pivotX, const float pivotY) const noexcept
{
	return followedBy (rotation (angle, pivotX, pivotY));
}

AffineTransform AffineTransform::scaled (const float factorX, const float factorY) const noexcept
{
	return AffineTransform (factorX * mat00, factorX * mat01, factorX * mat02,
							factorY * mat10, factorY * mat11, factorY * mat12);
}

AffineTransform AffineTransform::scale (const float factorX, const float factorY) noexcept
{
	return AffineTransform (factorX, 0, 0,
							0, factorY, 0);
}

AffineTransform AffineTransform::scaled (const float factorX, const float factorY,
											   const float pivotX, const float pivotY) const noexcept
{
	return AffineTransform (factorX * mat00, factorX * mat01, factorX * mat02 + pivotX * (1.0f - factorX),
							factorY * mat10, factorY * mat11, factorY * mat12 + pivotY * (1.0f - factorY));
}

AffineTransform AffineTransform::scale (const float factorX, const float factorY,
											  const float pivotX, const float pivotY) noexcept
{
	return AffineTransform (factorX, 0, pivotX * (1.0f - factorX),
							0, factorY, pivotY * (1.0f - factorY));
}

AffineTransform AffineTransform::shear (float shearX, float shearY) noexcept
{
	return AffineTransform (1.0f, shearX, 0,
							shearY, 1.0f, 0);
}

AffineTransform AffineTransform::sheared (const float shearX, const float shearY) const noexcept
{
	return AffineTransform (mat00 + shearX * mat10,
							mat01 + shearX * mat11,
							mat02 + shearX * mat12,
							shearY * mat00 + mat10,
							shearY * mat01 + mat11,
							shearY * mat02 + mat12);
}

AffineTransform AffineTransform::inverted() const noexcept
{
	double determinant = (mat00 * mat11 - mat10 * mat01);

	if (determinant != 0.0)
	{
		determinant = 1.0 / determinant;

		const float dst00 = (float) (mat11 * determinant);
		const float dst10 = (float) (-mat10 * determinant);
		const float dst01 = (float) (-mat01 * determinant);
		const float dst11 = (float) (mat00 * determinant);

		return AffineTransform (dst00, dst01, -mat02 * dst00 - mat12 * dst01,
								dst10, dst11, -mat02 * dst10 - mat12 * dst11);
	}
	else
	{
		// singularity..
		return *this;
	}
}

bool AffineTransform::isSingularity() const noexcept
{
	return (mat00 * mat11 - mat10 * mat01) == 0;
}

AffineTransform AffineTransform::fromTargetPoints (const float x00, const float y00,
												   const float x10, const float y10,
												   const float x01, const float y01) noexcept
{
	return AffineTransform (x10 - x00, x01 - x00, x00,
							y10 - y00, y01 - y00, y00);
}

AffineTransform AffineTransform::fromTargetPoints (const float sx1, const float sy1, const float tx1, const float ty1,
												   const float sx2, const float sy2, const float tx2, const float ty2,
												   const float sx3, const float sy3, const float tx3, const float ty3) noexcept
{
	return fromTargetPoints (sx1, sy1, sx2, sy2, sx3, sy3)
			.inverted()
			.followedBy (fromTargetPoints (tx1, ty1, tx2, ty2, tx3, ty3));
}

bool AffineTransform::isOnlyTranslation() const noexcept
{
	return (mat01 == 0)
		&& (mat10 == 0)
		&& (mat00 == 1.0f)
		&& (mat11 == 1.0f);
}

float AffineTransform::getScaleFactor() const noexcept
{
	return juce_hypot (mat00 + mat01, mat10 + mat11);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_AffineTransform.cpp ***/


/*** Start of inlined file: juce_Path.cpp ***/
BEGIN_JUCE_NAMESPACE

// tests that some co-ords aren't NaNs
#define CHECK_COORDS_ARE_VALID(x, y) \
	jassert (x == x && y == y);

namespace PathHelpers
{
	const float ellipseAngularIncrement = 0.05f;

	String nextToken (String::CharPointerType& t)
	{
		t = t.findEndOfWhitespace();

		String::CharPointerType start (t);
		int numChars = 0;

		while (! (t.isEmpty() || t.isWhitespace()))
		{
			++t;
			++numChars;
		}

		return String (start, numChars);
	}

	inline double lengthOf (float x1, float y1, float x2, float y2) noexcept
	{
		return juce_hypot ((double) (x1 - x2), (double) (y1 - y2));
	}
}

const float Path::lineMarker	   = 100001.0f;
const float Path::moveMarker	   = 100002.0f;
const float Path::quadMarker	   = 100003.0f;
const float Path::cubicMarker	  = 100004.0f;
const float Path::closeSubPathMarker   = 100005.0f;

Path::Path()
	: numElements (0),
	  pathXMin (0),
	  pathXMax (0),
	  pathYMin (0),
	  pathYMax (0),
	  useNonZeroWinding (true)
{
}

Path::~Path()
{
}

Path::Path (const Path& other)
	: numElements (other.numElements),
	  pathXMin (other.pathXMin),
	  pathXMax (other.pathXMax),
	  pathYMin (other.pathYMin),
	  pathYMax (other.pathYMax),
	  useNonZeroWinding (other.useNonZeroWinding)
{
	if (numElements > 0)
	{
		data.setAllocatedSize ((int) numElements);
		memcpy (data.elements, other.data.elements, numElements * sizeof (float));
	}
}

Path& Path::operator= (const Path& other)
{
	if (this != &other)
	{
		data.ensureAllocatedSize ((int) other.numElements);

		numElements = other.numElements;
		pathXMin = other.pathXMin;
		pathXMax = other.pathXMax;
		pathYMin = other.pathYMin;
		pathYMax = other.pathYMax;
		useNonZeroWinding = other.useNonZeroWinding;

		if (numElements > 0)
			memcpy (data.elements, other.data.elements, numElements * sizeof (float));
	}

	return *this;
}

bool Path::operator== (const Path& other) const noexcept
{
	return ! operator!= (other);
}

bool Path::operator!= (const Path& other) const noexcept
{
	if (numElements != other.numElements || useNonZeroWinding != other.useNonZeroWinding)
		return true;

	for (size_t i = 0; i < numElements; ++i)
		if (data.elements[i] != other.data.elements[i])
			return true;

	return false;
}

void Path::clear() noexcept
{
	numElements = 0;
	pathXMin = 0;
	pathYMin = 0;
	pathYMax = 0;
	pathXMax = 0;
}

void Path::swapWithPath (Path& other) noexcept
{
	data.swapWith (other.data);
	swapVariables <size_t> (numElements, other.numElements);
	swapVariables <float> (pathXMin, other.pathXMin);
	swapVariables <float> (pathXMax, other.pathXMax);
	swapVariables <float> (pathYMin, other.pathYMin);
	swapVariables <float> (pathYMax, other.pathYMax);
	swapVariables <bool> (useNonZeroWinding, other.useNonZeroWinding);
}

void Path::setUsingNonZeroWinding (const bool isNonZero) noexcept
{
	useNonZeroWinding = isNonZero;
}

void Path::scaleToFit (const float x, const float y, const float w, const float h,
					   const bool preserveProportions) noexcept
{
	applyTransform (getTransformToScaleToFit (x, y, w, h, preserveProportions));
}

bool Path::isEmpty() const noexcept
{
	size_t i = 0;

	while (i < numElements)
	{
		const float type = data.elements [i++];

		if (type == moveMarker)
		{
			i += 2;
		}
		else if (type == lineMarker
				 || type == quadMarker
				 || type == cubicMarker)
		{
			return false;
		}
	}

	return true;
}

Rectangle<float> Path::getBounds() const noexcept
{
	return Rectangle<float> (pathXMin, pathYMin,
							 pathXMax - pathXMin,
							 pathYMax - pathYMin);
}

Rectangle<float> Path::getBoundsTransformed (const AffineTransform& transform) const noexcept
{
	return getBounds().transformed (transform);
}

void Path::startNewSubPath (const float x, const float y)
{
	CHECK_COORDS_ARE_VALID (x, y);

	if (numElements == 0)
	{
		pathXMin = pathXMax = x;
		pathYMin = pathYMax = y;
	}
	else
	{
		pathXMin = jmin (pathXMin, x);
		pathXMax = jmax (pathXMax, x);
		pathYMin = jmin (pathYMin, y);
		pathYMax = jmax (pathYMax, y);
	}

	data.ensureAllocatedSize ((int) numElements + 3);

	data.elements [numElements++] = moveMarker;
	data.elements [numElements++] = x;
	data.elements [numElements++] = y;
}

void Path::startNewSubPath (const Point<float>& start)
{
	startNewSubPath (start.getX(), start.getY());
}

void Path::lineTo (const float x, const float y)
{
	CHECK_COORDS_ARE_VALID (x, y);

	if (numElements == 0)
		startNewSubPath (0, 0);

	data.ensureAllocatedSize ((int) numElements + 3);

	data.elements [numElements++] = lineMarker;
	data.elements [numElements++] = x;
	data.elements [numElements++] = y;

	pathXMin = jmin (pathXMin, x);
	pathXMax = jmax (pathXMax, x);
	pathYMin = jmin (pathYMin, y);
	pathYMax = jmax (pathYMax, y);
}

void Path::lineTo (const Point<float>& end)
{
	lineTo (end.getX(), end.getY());
}

void Path::quadraticTo (const float x1, const float y1,
						const float x2, const float y2)
{
	CHECK_COORDS_ARE_VALID (x1, y1);
	CHECK_COORDS_ARE_VALID (x2, y2);

	if (numElements == 0)
		startNewSubPath (0, 0);

	data.ensureAllocatedSize ((int) numElements + 5);

	data.elements [numElements++] = quadMarker;
	data.elements [numElements++] = x1;
	data.elements [numElements++] = y1;
	data.elements [numElements++] = x2;
	data.elements [numElements++] = y2;

	pathXMin = jmin (pathXMin, x1, x2);
	pathXMax = jmax (pathXMax, x1, x2);
	pathYMin = jmin (pathYMin, y1, y2);
	pathYMax = jmax (pathYMax, y1, y2);
}

void Path::quadraticTo (const Point<float>& controlPoint,
						const Point<float>& endPoint)
{
	quadraticTo (controlPoint.getX(), controlPoint.getY(),
				 endPoint.getX(), endPoint.getY());
}

void Path::cubicTo (const float x1, const float y1,
					const float x2, const float y2,
					const float x3, const float y3)
{
	CHECK_COORDS_ARE_VALID (x1, y1);
	CHECK_COORDS_ARE_VALID (x2, y2);
	CHECK_COORDS_ARE_VALID (x3, y3);

	if (numElements == 0)
		startNewSubPath (0, 0);

	data.ensureAllocatedSize ((int) numElements + 7);

	data.elements [numElements++] = cubicMarker;
	data.elements [numElements++] = x1;
	data.elements [numElements++] = y1;
	data.elements [numElements++] = x2;
	data.elements [numElements++] = y2;
	data.elements [numElements++] = x3;
	data.elements [numElements++] = y3;

	pathXMin = jmin (pathXMin, x1, x2, x3);
	pathXMax = jmax (pathXMax, x1, x2, x3);
	pathYMin = jmin (pathYMin, y1, y2, y3);
	pathYMax = jmax (pathYMax, y1, y2, y3);
}

void Path::cubicTo (const Point<float>& controlPoint1,
					const Point<float>& controlPoint2,
					const Point<float>& endPoint)
{
	cubicTo (controlPoint1.getX(), controlPoint1.getY(),
			 controlPoint2.getX(), controlPoint2.getY(),
			 endPoint.getX(), endPoint.getY());
}

void Path::closeSubPath()
{
	if (numElements > 0
		 && data.elements [numElements - 1] != closeSubPathMarker)
	{
		data.ensureAllocatedSize ((int) numElements + 1);
		data.elements [numElements++] = closeSubPathMarker;
	}
}

Point<float> Path::getCurrentPosition() const
{
	int i = (int) numElements - 1;

	if (i > 0 && data.elements[i] == closeSubPathMarker)
	{
		while (i >= 0)
		{
			if (data.elements[i] == moveMarker)
			{
				i += 2;
				break;
			}

			--i;
		}
	}

	if (i > 0)
		return Point<float> (data.elements [i - 1], data.elements [i]);

	return Point<float>();
}

void Path::addRectangle (const float x, const float y,
						 const float w, const float h)
{
	float x1 = x, y1 = y, x2 = x + w, y2 = y + h;

	if (w < 0)
		swapVariables (x1, x2);

	if (h < 0)
		swapVariables (y1, y2);

	data.ensureAllocatedSize ((int) numElements + 13);

	if (numElements == 0)
	{
		pathXMin = x1;
		pathXMax = x2;
		pathYMin = y1;
		pathYMax = y2;
	}
	else
	{
		pathXMin = jmin (pathXMin, x1);
		pathXMax = jmax (pathXMax, x2);
		pathYMin = jmin (pathYMin, y1);
		pathYMax = jmax (pathYMax, y2);
	}

	data.elements [numElements++] = moveMarker;
	data.elements [numElements++] = x1;
	data.elements [numElements++] = y2;
	data.elements [numElements++] = lineMarker;
	data.elements [numElements++] = x1;
	data.elements [numElements++] = y1;
	data.elements [numElements++] = lineMarker;
	data.elements [numElements++] = x2;
	data.elements [numElements++] = y1;
	data.elements [numElements++] = lineMarker;
	data.elements [numElements++] = x2;
	data.elements [numElements++] = y2;
	data.elements [numElements++] = closeSubPathMarker;
}

void Path::addRoundedRectangle (const float x, const float y,
								const float w, const float h,
								float csx,
								float csy)
{
	csx = jmin (csx, w * 0.5f);
	csy = jmin (csy, h * 0.5f);
	const float cs45x = csx * 0.45f;
	const float cs45y = csy * 0.45f;
	const float x2 = x + w;
	const float y2 = y + h;

	startNewSubPath (x + csx, y);
	lineTo (x2 - csx, y);
	cubicTo (x2 - cs45x, y, x2, y + cs45y, x2, y + csy);
	lineTo (x2, y2 - csy);
	cubicTo (x2, y2 - cs45y, x2 - cs45x, y2, x2 - csx, y2);
	lineTo (x + csx, y2);
	cubicTo (x + cs45x, y2, x, y2 - cs45y, x, y2 - csy);
	lineTo (x, y + csy);
	cubicTo (x, y + cs45y, x + cs45x, y, x + csx, y);
	closeSubPath();
}

void Path::addRoundedRectangle (const float x, const float y,
								const float w, const float h,
								float cs)
{
	addRoundedRectangle (x, y, w, h, cs, cs);
}

void Path::addTriangle (const float x1, const float y1,
						const float x2, const float y2,
						const float x3, const float y3)
{
	startNewSubPath (x1, y1);
	lineTo (x2, y2);
	lineTo (x3, y3);
	closeSubPath();
}

void Path::addQuadrilateral (const float x1, const float y1,
							 const float x2, const float y2,
							 const float x3, const float y3,
							 const float x4, const float y4)
{
	startNewSubPath (x1, y1);
	lineTo (x2, y2);
	lineTo (x3, y3);
	lineTo (x4, y4);
	closeSubPath();
}

void Path::addEllipse (const float x, const float y,
					   const float w, const float h)
{
	const float hw = w * 0.5f;
	const float hw55 = hw * 0.55f;
	const float hh = h * 0.5f;
	const float hh55 = hh * 0.55f;
	const float cx = x + hw;
	const float cy = y + hh;

	startNewSubPath (cx, cy - hh);
	cubicTo (cx + hw55, cy - hh, cx + hw, cy - hh55, cx + hw, cy);
	cubicTo (cx + hw, cy + hh55, cx + hw55, cy + hh, cx, cy + hh);
	cubicTo (cx - hw55, cy + hh, cx - hw, cy + hh55, cx - hw, cy);
	cubicTo (cx - hw, cy - hh55, cx - hw55, cy - hh, cx, cy - hh);
	closeSubPath();
}

void Path::addArc (const float x, const float y,
				   const float w, const float h,
				   const float fromRadians,
				   const float toRadians,
				   const bool startAsNewSubPath)
{
	const float radiusX = w / 2.0f;
	const float radiusY = h / 2.0f;

	addCentredArc (x + radiusX,
				   y + radiusY,
				   radiusX, radiusY,
				   0.0f,
				   fromRadians, toRadians,
				   startAsNewSubPath);
}

void Path::addCentredArc (const float centreX, const float centreY,
						  const float radiusX, const float radiusY,
						  const float rotationOfEllipse,
						  const float fromRadians,
						  float toRadians,
						  const bool startAsNewSubPath)
{
	if (radiusX > 0.0f && radiusY > 0.0f)
	{
		const Point<float> centre (centreX, centreY);
		const AffineTransform rotation (AffineTransform::rotation (rotationOfEllipse, centreX, centreY));
		float angle = fromRadians;

		if (startAsNewSubPath)
			startNewSubPath (centre.getPointOnCircumference (radiusX, radiusY, angle).transformedBy (rotation));

		if (fromRadians < toRadians)
		{
			if (startAsNewSubPath)
				angle += PathHelpers::ellipseAngularIncrement;

			while (angle < toRadians)
			{
				lineTo (centre.getPointOnCircumference (radiusX, radiusY, angle).transformedBy (rotation));
				angle += PathHelpers::ellipseAngularIncrement;
			}
		}
		else
		{
			if (startAsNewSubPath)
				angle -= PathHelpers::ellipseAngularIncrement;

			while (angle > toRadians)
			{
				lineTo (centre.getPointOnCircumference (radiusX, radiusY, angle).transformedBy (rotation));
				angle -= PathHelpers::ellipseAngularIncrement;
			}
		}

		lineTo (centre.getPointOnCircumference (radiusX, radiusY, toRadians).transformedBy (rotation));
	}
}

void Path::addPieSegment (const float x, const float y,
						  const float width, const float height,
						  const float fromRadians,
						  const float toRadians,
						  const float innerCircleProportionalSize)
{
	float radiusX = width * 0.5f;
	float radiusY = height * 0.5f;
	const Point<float> centre (x + radiusX, y + radiusY);

	startNewSubPath (centre.getPointOnCircumference (radiusX, radiusY, fromRadians));
	addArc (x, y, width, height, fromRadians, toRadians);

	if (std::abs (fromRadians - toRadians) > float_Pi * 1.999f)
	{
		closeSubPath();

		if (innerCircleProportionalSize > 0)
		{
			radiusX *= innerCircleProportionalSize;
			radiusY *= innerCircleProportionalSize;

			startNewSubPath (centre.getPointOnCircumference (radiusX, radiusY, toRadians));
			addArc (centre.getX() - radiusX, centre.getY() - radiusY, radiusX * 2.0f, radiusY * 2.0f, toRadians, fromRadians);
		}
	}
	else
	{
		if (innerCircleProportionalSize > 0)
		{
			radiusX *= innerCircleProportionalSize;
			radiusY *= innerCircleProportionalSize;

			addArc (centre.getX() - radiusX, centre.getY() - radiusY, radiusX * 2.0f, radiusY * 2.0f, toRadians, fromRadians);
		}
		else
		{
			lineTo (centre);
		}
	}

	closeSubPath();
}

void Path::addLineSegment (const Line<float>& line, float lineThickness)
{
	const Line<float> reversed (line.reversed());
	lineThickness *= 0.5f;

	startNewSubPath (line.getPointAlongLine (0, lineThickness));
	lineTo (line.getPointAlongLine (0, -lineThickness));
	lineTo (reversed.getPointAlongLine (0, lineThickness));
	lineTo (reversed.getPointAlongLine (0, -lineThickness));
	closeSubPath();
}

void Path::addArrow (const Line<float>& line, float lineThickness,
					 float arrowheadWidth, float arrowheadLength)
{
	const Line<float> reversed (line.reversed());
	lineThickness *= 0.5f;
	arrowheadWidth *= 0.5f;
	arrowheadLength = jmin (arrowheadLength, 0.8f * line.getLength());

	startNewSubPath (line.getPointAlongLine (0, lineThickness));
	lineTo (line.getPointAlongLine (0, -lineThickness));
	lineTo (reversed.getPointAlongLine (arrowheadLength, lineThickness));
	lineTo (reversed.getPointAlongLine (arrowheadLength, arrowheadWidth));
	lineTo (line.getEnd());
	lineTo (reversed.getPointAlongLine (arrowheadLength, -arrowheadWidth));
	lineTo (reversed.getPointAlongLine (arrowheadLength, -lineThickness));
	closeSubPath();
}

void Path::addPolygon (const Point<float>& centre, const int numberOfSides,
					   const float radius, const float startAngle)
{
	jassert (numberOfSides > 1); // this would be silly.

	if (numberOfSides > 1)
	{
		const float angleBetweenPoints = float_Pi * 2.0f / numberOfSides;

		for (int i = 0; i < numberOfSides; ++i)
		{
			const float angle = startAngle + i * angleBetweenPoints;
			const Point<float> p (centre.getPointOnCircumference (radius, angle));

			if (i == 0)
				startNewSubPath (p);
			else
				lineTo (p);
		}

		closeSubPath();
	}
}

void Path::addStar (const Point<float>& centre, const int numberOfPoints,
					const float innerRadius, const float outerRadius, const float startAngle)
{
	jassert (numberOfPoints > 1); // this would be silly.

	if (numberOfPoints > 1)
	{
		const float angleBetweenPoints = float_Pi * 2.0f / numberOfPoints;

		for (int i = 0; i < numberOfPoints; ++i)
		{
			const float angle = startAngle + i * angleBetweenPoints;
			const Point<float> p (centre.getPointOnCircumference (outerRadius, angle));

			if (i == 0)
				startNewSubPath (p);
			else
				lineTo (p);

			lineTo (centre.getPointOnCircumference (innerRadius, angle + angleBetweenPoints * 0.5f));
		}

		closeSubPath();
	}
}

void Path::addBubble (float x, float y,
					  float w, float h,
					  float cs,
					  float tipX,
					  float tipY,
					  int whichSide,
					  float arrowPos,
					  float arrowWidth)
{
	if (w > 1.0f && h > 1.0f)
	{
		cs = jmin (cs, w * 0.5f, h * 0.5f);
		const float cs2 = 2.0f * cs;

		startNewSubPath (x + cs, y);

		if (whichSide == 0)
		{
			const float halfArrowW = jmin (arrowWidth, w - cs2) * 0.5f;
			const float arrowX1 = x + cs + jmax (0.0f, (w - cs2) * arrowPos - halfArrowW);
			lineTo (arrowX1, y);
			lineTo (tipX, tipY);
			lineTo (arrowX1 + halfArrowW * 2.0f, y);
		}

		lineTo (x + w - cs, y);

		if (cs > 0.0f)
			addArc (x + w - cs2, y, cs2, cs2, 0, float_Pi * 0.5f);

		if (whichSide == 3)
		{
			const float halfArrowH = jmin (arrowWidth, h - cs2) * 0.5f;
			const float arrowY1 = y + cs + jmax (0.0f, (h - cs2) * arrowPos - halfArrowH);
			lineTo (x + w, arrowY1);
			lineTo (tipX, tipY);
			lineTo (x + w, arrowY1 + halfArrowH * 2.0f);
		}

		lineTo (x + w, y + h - cs);

		if (cs > 0.0f)
			addArc (x + w - cs2, y + h - cs2, cs2, cs2, float_Pi * 0.5f, float_Pi);

		if (whichSide == 2)
		{
			const float halfArrowW = jmin (arrowWidth, w - cs2) * 0.5f;
			const float arrowX1 = x + cs + jmax (0.0f, (w - cs2) * arrowPos - halfArrowW);
			lineTo (arrowX1 + halfArrowW * 2.0f, y + h);
			lineTo (tipX, tipY);
			lineTo (arrowX1, y + h);
		}

		lineTo (x + cs, y + h);

		if (cs > 0.0f)
			addArc (x, y + h - cs2, cs2, cs2, float_Pi, float_Pi * 1.5f);

		if (whichSide == 1)
		{
			const float halfArrowH = jmin (arrowWidth, h - cs2) * 0.5f;
			const float arrowY1 = y + cs + jmax (0.0f, (h - cs2) * arrowPos - halfArrowH);
			lineTo (x, arrowY1 + halfArrowH * 2.0f);
			lineTo (tipX, tipY);
			lineTo (x, arrowY1);
		}

		lineTo (x, y + cs);

		if (cs > 0.0f)
			addArc (x, y, cs2, cs2, float_Pi * 1.5f, float_Pi * 2.0f - PathHelpers::ellipseAngularIncrement);

		closeSubPath();
	}
}

void Path::addPath (const Path& other)
{
	size_t i = 0;

	while (i < other.numElements)
	{
		const float type = other.data.elements [i++];

		if (type == moveMarker)
		{
			startNewSubPath (other.data.elements [i],
							 other.data.elements [i + 1]);

			i += 2;
		}
		else if (type == lineMarker)
		{
			lineTo (other.data.elements [i],
					other.data.elements [i + 1]);

			i += 2;
		}
		else if (type == quadMarker)
		{
			quadraticTo (other.data.elements [i],
						 other.data.elements [i + 1],
						 other.data.elements [i + 2],
						 other.data.elements [i + 3]);
			i += 4;
		}
		else if (type == cubicMarker)
		{
			cubicTo (other.data.elements [i],
					 other.data.elements [i + 1],
					 other.data.elements [i + 2],
					 other.data.elements [i + 3],
					 other.data.elements [i + 4],
					 other.data.elements [i + 5]);

			i += 6;
		}
		else if (type == closeSubPathMarker)
		{
			closeSubPath();
		}
		else
		{
			// something's gone wrong with the element list!
			jassertfalse;
		}
	}
}

void Path::addPath (const Path& other,
					const AffineTransform& transformToApply)
{
	size_t i = 0;

	while (i < other.numElements)
	{
		const float type = other.data.elements [i++];

		if (type == closeSubPathMarker)
		{
			closeSubPath();
		}
		else
		{
			float x = other.data.elements [i++];
			float y = other.data.elements [i++];
			transformToApply.transformPoint (x, y);

			if (type == moveMarker)
			{
				startNewSubPath (x, y);
			}
			else if (type == lineMarker)
			{
				lineTo (x, y);
			}
			else if (type == quadMarker)
			{
				float x2 = other.data.elements [i++];
				float y2 = other.data.elements [i++];
				transformToApply.transformPoint (x2, y2);

				quadraticTo (x, y, x2, y2);
			}
			else if (type == cubicMarker)
			{
				float x2 = other.data.elements [i++];
				float y2 = other.data.elements [i++];
				float x3 = other.data.elements [i++];
				float y3 = other.data.elements [i++];
				transformToApply.transformPoints (x2, y2, x3, y3);

				cubicTo (x, y, x2, y2, x3, y3);
			}
			else
			{
				// something's gone wrong with the element list!
				jassertfalse;
			}
		}
	}
}

void Path::applyTransform (const AffineTransform& transform) noexcept
{
	size_t i = 0;
	pathYMin = pathXMin = 0;
	pathYMax = pathXMax = 0;
	bool setMaxMin = false;

	while (i < numElements)
	{
		const float type = data.elements [i++];

		if (type == moveMarker)
		{
			transform.transformPoint (data.elements [i], data.elements [i + 1]);

			if (setMaxMin)
			{
				pathXMin = jmin (pathXMin, data.elements [i]);
				pathXMax = jmax (pathXMax, data.elements [i]);
				pathYMin = jmin (pathYMin, data.elements [i + 1]);
				pathYMax = jmax (pathYMax, data.elements [i + 1]);
			}
			else
			{
				pathXMin = pathXMax = data.elements [i];
				pathYMin = pathYMax = data.elements [i + 1];
				setMaxMin = true;
			}

			i += 2;
		}
		else if (type == lineMarker)
		{
			transform.transformPoint (data.elements [i], data.elements [i + 1]);

			pathXMin = jmin (pathXMin, data.elements [i]);
			pathXMax = jmax (pathXMax, data.elements [i]);
			pathYMin = jmin (pathYMin, data.elements [i + 1]);
			pathYMax = jmax (pathYMax, data.elements [i + 1]);

			i += 2;
		}
		else if (type == quadMarker)
		{
			transform.transformPoints (data.elements [i], data.elements [i + 1],
									   data.elements [i + 2], data.elements [i + 3]);

			pathXMin = jmin (pathXMin, data.elements [i], data.elements [i + 2]);
			pathXMax = jmax (pathXMax, data.elements [i], data.elements [i + 2]);
			pathYMin = jmin (pathYMin, data.elements [i + 1], data.elements [i + 3]);
			pathYMax = jmax (pathYMax, data.elements [i + 1], data.elements [i + 3]);

			i += 4;
		}
		else if (type == cubicMarker)
		{
			transform.transformPoints (data.elements [i], data.elements [i + 1],
									   data.elements [i + 2], data.elements [i + 3],
									   data.elements [i + 4], data.elements [i + 5]);

			pathXMin = jmin (pathXMin, data.elements [i], data.elements [i + 2], data.elements [i + 4]);
			pathXMax = jmax (pathXMax, data.elements [i], data.elements [i + 2], data.elements [i + 4]);
			pathYMin = jmin (pathYMin, data.elements [i + 1], data.elements [i + 3], data.elements [i + 5]);
			pathYMax = jmax (pathYMax, data.elements [i + 1], data.elements [i + 3], data.elements [i + 5]);

			i += 6;
		}
	}
}

AffineTransform Path::getTransformToScaleToFit (const float x, const float y,
												const float w, const float h,
												const bool preserveProportions,
												const Justification& justification) const
{
	Rectangle<float> bounds (getBounds());

	if (preserveProportions)
	{
		if (w <= 0 || h <= 0 || bounds.isEmpty())
			return AffineTransform::identity;

		float newW, newH;
		const float srcRatio = bounds.getHeight() / bounds.getWidth();

		if (srcRatio > h / w)
		{
			newW = h / srcRatio;
			newH = h;
		}
		else
		{
			newW = w;
			newH = w * srcRatio;
		}

		float newXCentre = x;
		float newYCentre = y;

		if (justification.testFlags (Justification::left))
			newXCentre += newW * 0.5f;
		else if (justification.testFlags (Justification::right))
			newXCentre += w - newW * 0.5f;
		else
			newXCentre += w * 0.5f;

		if (justification.testFlags (Justification::top))
			newYCentre += newH * 0.5f;
		else if (justification.testFlags (Justification::bottom))
			newYCentre += h - newH * 0.5f;
		else
			newYCentre += h * 0.5f;

		return AffineTransform::translation (bounds.getWidth() * -0.5f - bounds.getX(),
											 bounds.getHeight() * -0.5f - bounds.getY())
					.scaled (newW / bounds.getWidth(), newH / bounds.getHeight())
					.translated (newXCentre, newYCentre);
	}
	else
	{
		return AffineTransform::translation (-bounds.getX(), -bounds.getY())
					.scaled (w / bounds.getWidth(), h / bounds.getHeight())
					.translated (x, y);
	}
}

bool Path::contains (const float x, const float y, const float tolerance) const
{
	if (x <= pathXMin || x >= pathXMax
		 || y <= pathYMin || y >= pathYMax)
		return false;

	PathFlatteningIterator i (*this, AffineTransform::identity, tolerance);

	int positiveCrossings = 0;
	int negativeCrossings = 0;

	while (i.next())
	{
		if ((i.y1 <= y && i.y2 > y) || (i.y2 <= y && i.y1 > y))
		{
			const float intersectX = i.x1 + (i.x2 - i.x1) * (y - i.y1) / (i.y2 - i.y1);

			if (intersectX <= x)
			{
				if (i.y1 < i.y2)
					++positiveCrossings;
				else
					++negativeCrossings;
			}
		}
	}

	return useNonZeroWinding ? (negativeCrossings != positiveCrossings)
							 : ((negativeCrossings + positiveCrossings) & 1) != 0;
}

bool Path::contains (const Point<float>& point, const float tolerance) const
{
	return contains (point.getX(), point.getY(), tolerance);
}

bool Path::intersectsLine (const Line<float>& line, const float tolerance)
{
	PathFlatteningIterator i (*this, AffineTransform::identity, tolerance);
	Point<float> intersection;

	while (i.next())
		if (line.intersects (Line<float> (i.x1, i.y1, i.x2, i.y2), intersection))
			return true;

	return false;
}

Line<float> Path::getClippedLine (const Line<float>& line, const bool keepSectionOutsidePath) const
{
	Line<float> result (line);
	const bool startInside = contains (line.getStart());
	const bool endInside = contains (line.getEnd());

	if (startInside == endInside)
	{
		if (keepSectionOutsidePath == startInside)
			result = Line<float>();
	}
	else
	{
		PathFlatteningIterator i (*this, AffineTransform::identity);
		Point<float> intersection;

		while (i.next())
		{
			if (line.intersects (Line<float> (i.x1, i.y1, i.x2, i.y2), intersection))
			{
				if ((startInside && keepSectionOutsidePath) || (endInside && ! keepSectionOutsidePath))
					result.setStart (intersection);
				else
					result.setEnd (intersection);
			}
		}
	}

	return result;
}

float Path::getLength (const AffineTransform& transform) const
{
	float length = 0;
	PathFlatteningIterator i (*this, transform);

	while (i.next())
		length += Line<float> (i.x1, i.y1, i.x2, i.y2).getLength();

	return length;
}

Point<float> Path::getPointAlongPath (float distanceFromStart, const AffineTransform& transform) const
{
	PathFlatteningIterator i (*this, transform);

	while (i.next())
	{
		const Line<float> line (i.x1, i.y1, i.x2, i.y2);
		const float lineLength = line.getLength();

		if (distanceFromStart <= lineLength)
			return line.getPointAlongLine (distanceFromStart);

		distanceFromStart -= lineLength;
	}

	return Point<float> (i.x2, i.y2);
}

float Path::getNearestPoint (const Point<float>& targetPoint, Point<float>& pointOnPath,
							 const AffineTransform& transform) const
{
	PathFlatteningIterator i (*this, transform);
	float bestPosition = 0, bestDistance = std::numeric_limits<float>::max();
	float length = 0;
	Point<float> pointOnLine;

	while (i.next())
	{
		const Line<float> line (i.x1, i.y1, i.x2, i.y2);
		const float distance = line.getDistanceFromPoint (targetPoint, pointOnLine);

		if (distance < bestDistance)
		{
			bestDistance = distance;
			bestPosition = length + pointOnLine.getDistanceFrom (line.getStart());
			pointOnPath = pointOnLine;
		}

		length += line.getLength();
	}

	return bestPosition;
}

Path Path::createPathWithRoundedCorners (const float cornerRadius) const
{
	if (cornerRadius <= 0.01f)
		return *this;

	size_t indexOfPathStart = 0, indexOfPathStartThis = 0;
	size_t n = 0;
	bool lastWasLine = false, firstWasLine = false;
	Path p;

	while (n < numElements)
	{
		const float type = data.elements [n++];

		if (type == moveMarker)
		{
			indexOfPathStart = p.numElements;
			indexOfPathStartThis = n - 1;
			const float x = data.elements [n++];
			const float y = data.elements [n++];
			p.startNewSubPath (x, y);
			lastWasLine = false;
			firstWasLine = (data.elements [n] == lineMarker);
		}
		else if (type == lineMarker || type == closeSubPathMarker)
		{
			float startX = 0, startY = 0, joinX = 0, joinY = 0, endX, endY;

			if (type == lineMarker)
			{
				endX = data.elements [n++];
				endY = data.elements [n++];

				if (n > 8)
				{
					startX = data.elements [n - 8];
					startY = data.elements [n - 7];
					joinX = data.elements [n - 5];
					joinY = data.elements [n - 4];
				}
			}
			else
			{
				endX = data.elements [indexOfPathStartThis + 1];
				endY = data.elements [indexOfPathStartThis + 2];

				if (n > 6)
				{
					startX = data.elements [n - 6];
					startY = data.elements [n - 5];
					joinX = data.elements [n - 3];
					joinY = data.elements [n - 2];
				}
			}

			if (lastWasLine)
			{
				const double len1 = PathHelpers::lengthOf (startX, startY, joinX, joinY);

				if (len1 > 0)
				{
					const double propNeeded = jmin (0.5, cornerRadius / len1);

					p.data.elements [p.numElements - 2] = (float) (joinX - (joinX - startX) * propNeeded);
					p.data.elements [p.numElements - 1] = (float) (joinY - (joinY - startY) * propNeeded);
				}

				const double len2 = PathHelpers::lengthOf (endX, endY, joinX, joinY);

				if (len2 > 0)
				{
					const double propNeeded = jmin (0.5, cornerRadius / len2);

					p.quadraticTo (joinX, joinY,
								   (float) (joinX + (endX - joinX) * propNeeded),
								   (float) (joinY + (endY - joinY) * propNeeded));
				}

				p.lineTo (endX, endY);
			}
			else if (type == lineMarker)
			{
				p.lineTo (endX, endY);
				lastWasLine = true;
			}

			if (type == closeSubPathMarker)
			{
				if (firstWasLine)
				{
					startX = data.elements [n - 3];
					startY = data.elements [n - 2];
					joinX = endX;
					joinY = endY;
					endX = data.elements [indexOfPathStartThis + 4];
					endY = data.elements [indexOfPathStartThis + 5];

					const double len1 = PathHelpers::lengthOf (startX, startY, joinX, joinY);

					if (len1 > 0)
					{
						const double propNeeded = jmin (0.5, cornerRadius / len1);

						p.data.elements [p.numElements - 2] = (float) (joinX - (joinX - startX) * propNeeded);
						p.data.elements [p.numElements - 1] = (float) (joinY - (joinY - startY) * propNeeded);
					}

					const double len2 = PathHelpers::lengthOf (endX, endY, joinX, joinY);

					if (len2 > 0)
					{
						const double propNeeded = jmin (0.5, cornerRadius / len2);

						endX = (float) (joinX + (endX - joinX) * propNeeded);
						endY = (float) (joinY + (endY - joinY) * propNeeded);

						p.quadraticTo (joinX, joinY, endX, endY);

						p.data.elements [indexOfPathStart + 1] = endX;
						p.data.elements [indexOfPathStart + 2] = endY;
					}
				}

				p.closeSubPath();
			}
		}
		else if (type == quadMarker)
		{
			lastWasLine = false;
			const float x1 = data.elements [n++];
			const float y1 = data.elements [n++];
			const float x2 = data.elements [n++];
			const float y2 = data.elements [n++];
			p.quadraticTo (x1, y1, x2, y2);
		}
		else if (type == cubicMarker)
		{
			lastWasLine = false;
			const float x1 = data.elements [n++];
			const float y1 = data.elements [n++];
			const float x2 = data.elements [n++];
			const float y2 = data.elements [n++];
			const float x3 = data.elements [n++];
			const float y3 = data.elements [n++];
			p.cubicTo (x1, y1, x2, y2, x3, y3);
		}
	}

	return p;
}

void Path::loadPathFromStream (InputStream& source)
{
	while (! source.isExhausted())
	{
		switch (source.readByte())
		{
		case 'm':
		{
			const float x = source.readFloat();
			const float y = source.readFloat();
			startNewSubPath (x, y);
			break;
		}

		case 'l':
		{
			const float x = source.readFloat();
			const float y = source.readFloat();
			lineTo (x, y);
			break;
		}

		case 'q':
		{
			const float x1 = source.readFloat();
			const float y1 = source.readFloat();
			const float x2 = source.readFloat();
			const float y2 = source.readFloat();
			quadraticTo (x1, y1, x2, y2);
			break;
		}

		case 'b':
		{
			const float x1 = source.readFloat();
			const float y1 = source.readFloat();
			const float x2 = source.readFloat();
			const float y2 = source.readFloat();
			const float x3 = source.readFloat();
			const float y3 = source.readFloat();
			cubicTo (x1, y1, x2, y2, x3, y3);
			break;
		}

		case 'c':
			closeSubPath();
			break;

		case 'n':
			useNonZeroWinding = true;
			break;

		case 'z':
			useNonZeroWinding = false;
			break;

		case 'e':
			return; // end of path marker

		default:
			jassertfalse; // illegal char in the stream
			break;
		}
	}
}

void Path::loadPathFromData (const void* const pathData, const int numberOfBytes)
{
	MemoryInputStream in (pathData, numberOfBytes, false);
	loadPathFromStream (in);
}

void Path::writePathToStream (OutputStream& dest) const
{
	dest.writeByte (useNonZeroWinding ? 'n' : 'z');

	size_t i = 0;
	while (i < numElements)
	{
		const float type = data.elements [i++];

		if (type == moveMarker)
		{
			dest.writeByte ('m');
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
		}
		else if (type == lineMarker)
		{
			dest.writeByte ('l');
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
		}
		else if (type == quadMarker)
		{
			dest.writeByte ('q');
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
		}
		else if (type == cubicMarker)
		{
			dest.writeByte ('b');
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
			dest.writeFloat (data.elements [i++]);
		}
		else if (type == closeSubPathMarker)
		{
			dest.writeByte ('c');
		}
	}

	dest.writeByte ('e'); // marks the end-of-path
}

String Path::toString() const
{
	MemoryOutputStream s (2048);
	if (! useNonZeroWinding)
		s << 'a';

	size_t i = 0;
	float lastMarker = 0.0f;

	while (i < numElements)
	{
		const float marker = data.elements [i++];
		char markerChar = 0;
		int numCoords = 0;

		if (marker == moveMarker)
		{
			markerChar = 'm';
			numCoords = 2;
		}
		else if (marker == lineMarker)
		{
			markerChar = 'l';
			numCoords = 2;
		}
		else if (marker == quadMarker)
		{
			markerChar = 'q';
			numCoords = 4;
		}
		else if (marker == cubicMarker)
		{
			markerChar = 'c';
			numCoords = 6;
		}
		else
		{
			jassert (marker == closeSubPathMarker);
			markerChar = 'z';
		}

		if (marker != lastMarker)
		{
			if (s.getDataSize() != 0)
				s << ' ';

			s << markerChar;
			lastMarker = marker;
		}

		while (--numCoords >= 0 && i < numElements)
		{
			String coord (data.elements [i++], 3);

			while (coord.endsWithChar ('0') && coord != "0")
				coord = coord.dropLastCharacters (1);

			if (coord.endsWithChar ('.'))
				coord = coord.dropLastCharacters (1);

			if (s.getDataSize() != 0)
				s << ' ';

			s << coord;
		}
	}

	return s.toUTF8();
}

void Path::restoreFromString (const String& stringVersion)
{
	clear();
	setUsingNonZeroWinding (true);

	String::CharPointerType t (stringVersion.getCharPointer());
	juce_wchar marker = 'm';
	int numValues = 2;
	float values [6];

	for (;;)
	{
		const String token (PathHelpers::nextToken (t));
		const juce_wchar firstChar = token[0];
		int startNum = 0;

		if (firstChar == 0)
			break;

		if (firstChar == 'm' || firstChar == 'l')
		{
			marker = firstChar;
			numValues = 2;
		}
		else if (firstChar == 'q')
		{
			marker = firstChar;
			numValues = 4;
		}
		else if (firstChar == 'c')
		{
			marker = firstChar;
			numValues = 6;
		}
		else if (firstChar == 'z')
		{
			marker = firstChar;
			numValues = 0;
		}
		else if (firstChar == 'a')
		{
			setUsingNonZeroWinding (false);
			continue;
		}
		else
		{
			++startNum;
			values [0] = token.getFloatValue();
		}

		for (int i = startNum; i < numValues; ++i)
			values [i] = PathHelpers::nextToken (t).getFloatValue();

		switch (marker)
		{
			case 'm':   startNewSubPath (values[0], values[1]); break;
			case 'l':   lineTo (values[0], values[1]); break;
			case 'q':   quadraticTo (values[0], values[1], values[2], values[3]); break;
			case 'c':   cubicTo (values[0], values[1], values[2], values[3], values[4], values[5]); break;
			case 'z':   closeSubPath(); break;
			default:	jassertfalse; break; // illegal string format?
		}
	}
}

Path::Iterator::Iterator (const Path& path_)
	: path (path_),
	  index (0)
{
}

Path::Iterator::~Iterator()
{
}

bool Path::Iterator::next()
{
	const float* const elements = path.data.elements;

	if (index < path.numElements)
	{
		const float type = elements [index++];

		if (type == moveMarker)
		{
			elementType = startNewSubPath;
			x1 = elements [index++];
			y1 = elements [index++];
		}
		else if (type == lineMarker)
		{
			elementType = lineTo;
			x1 = elements [index++];
			y1 = elements [index++];
		}
		else if (type == quadMarker)
		{
			elementType = quadraticTo;
			x1 = elements [index++];
			y1 = elements [index++];
			x2 = elements [index++];
			y2 = elements [index++];
		}
		else if (type == cubicMarker)
		{
			elementType = cubicTo;
			x1 = elements [index++];
			y1 = elements [index++];
			x2 = elements [index++];
			y2 = elements [index++];
			x3 = elements [index++];
			y3 = elements [index++];
		}
		else if (type == closeSubPathMarker)
		{
			elementType = closePath;
		}

		return true;
	}

	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Path.cpp ***/


/*** Start of inlined file: juce_PathIterator.cpp ***/
BEGIN_JUCE_NAMESPACE

#if JUCE_MSVC && JUCE_DEBUG
  #pragma optimize ("t", on)
#endif

const float PathFlatteningIterator::defaultTolerance = 0.6f;

PathFlatteningIterator::PathFlatteningIterator (const Path& path_,
												const AffineTransform& transform_,
												const float tolerance)
	: x2 (0),
	  y2 (0),
	  closesSubPath (false),
	  subPathIndex (-1),
	  path (path_),
	  transform (transform_),
	  points (path_.data.elements),
	  toleranceSquared (tolerance * tolerance),
	  subPathCloseX (0),
	  subPathCloseY (0),
	  isIdentityTransform (transform_.isIdentity()),
	  stackBase (32),
	  index (0),
	  stackSize (32)
{
	stackPos = stackBase;
}

PathFlatteningIterator::~PathFlatteningIterator()
{
}

bool PathFlatteningIterator::isLastInSubpath() const noexcept
{
	return stackPos == stackBase.getData()
			 && (index >= path.numElements || points [index] == Path::moveMarker);
}

bool PathFlatteningIterator::next()
{
	x1 = x2;
	y1 = y2;

	float x3 = 0;
	float y3 = 0;
	float x4 = 0;
	float y4 = 0;
	float type;

	for (;;)
	{
		if (stackPos == stackBase)
		{
			if (index >= path.numElements)
			{
				return false;
			}
			else
			{
				type = points [index++];

				if (type != Path::closeSubPathMarker)
				{
					x2 = points [index++];
					y2 = points [index++];

					if (type == Path::quadMarker)
					{
						x3 = points [index++];
						y3 = points [index++];

						if (! isIdentityTransform)
							transform.transformPoints (x2, y2, x3, y3);
					}
					else if (type == Path::cubicMarker)
					{
						x3 = points [index++];
						y3 = points [index++];
						x4 = points [index++];
						y4 = points [index++];

						if (! isIdentityTransform)
							transform.transformPoints (x2, y2, x3, y3, x4, y4);
					}
					else
					{
						if (! isIdentityTransform)
							transform.transformPoint (x2, y2);
					}
				}
			}
		}
		else
		{
			type = *--stackPos;

			if (type != Path::closeSubPathMarker)
			{
				x2 = *--stackPos;
				y2 = *--stackPos;

				if (type == Path::quadMarker)
				{
					x3 = *--stackPos;
					y3 = *--stackPos;
				}
				else if (type == Path::cubicMarker)
				{
					x3 = *--stackPos;
					y3 = *--stackPos;
					x4 = *--stackPos;
					y4 = *--stackPos;
				}
			}
		}

		if (type == Path::lineMarker)
		{
			++subPathIndex;

			closesSubPath = (stackPos == stackBase)
							 && (index < path.numElements)
							 && (points [index] == Path::closeSubPathMarker)
							 && x2 == subPathCloseX
							 && y2 == subPathCloseY;

			return true;
		}
		else if (type == Path::quadMarker)
		{
			const size_t offset = (size_t) (stackPos - stackBase);

			if (offset >= stackSize - 10)
			{
				stackSize <<= 1;
				stackBase.realloc (stackSize);
				stackPos = stackBase + offset;
			}

			const float m1x = (x1 + x2) * 0.5f;
			const float m1y = (y1 + y2) * 0.5f;
			const float m2x = (x2 + x3) * 0.5f;
			const float m2y = (y2 + y3) * 0.5f;
			const float m3x = (m1x + m2x) * 0.5f;
			const float m3y = (m1y + m2y) * 0.5f;

			const float errorX = m3x - x2;
			const float errorY = m3y - y2;

			if (errorX * errorX + errorY * errorY > toleranceSquared)
			{
				*stackPos++ = y3;
				*stackPos++ = x3;
				*stackPos++ = m2y;
				*stackPos++ = m2x;
				*stackPos++ = Path::quadMarker;

				*stackPos++ = m3y;
				*stackPos++ = m3x;
				*stackPos++ = m1y;
				*stackPos++ = m1x;
				*stackPos++ = Path::quadMarker;
			}
			else
			{
				*stackPos++ = y3;
				*stackPos++ = x3;
				*stackPos++ = Path::lineMarker;

				*stackPos++ = m3y;
				*stackPos++ = m3x;
				*stackPos++ = Path::lineMarker;
			}

			jassert (stackPos < stackBase + stackSize);
		}
		else if (type == Path::cubicMarker)
		{
			const size_t offset = (size_t) (stackPos - stackBase);

			if (offset >= stackSize - 16)
			{
				stackSize <<= 1;
				stackBase.realloc (stackSize);
				stackPos = stackBase + offset;
			}

			const float m1x = (x1 + x2) * 0.5f;
			const float m1y = (y1 + y2) * 0.5f;
			const float m2x = (x3 + x2) * 0.5f;
			const float m2y = (y3 + y2) * 0.5f;
			const float m3x = (x3 + x4) * 0.5f;
			const float m3y = (y3 + y4) * 0.5f;
			const float m4x = (m1x + m2x) * 0.5f;
			const float m4y = (m1y + m2y) * 0.5f;
			const float m5x = (m3x + m2x) * 0.5f;
			const float m5y = (m3y + m2y) * 0.5f;

			const float error1X = m4x - x2;
			const float error1Y = m4y - y2;
			const float error2X = m5x - x3;
			const float error2Y = m5y - y3;

			if (error1X * error1X + error1Y * error1Y > toleranceSquared
				 || error2X * error2X + error2Y * error2Y > toleranceSquared)
			{
				*stackPos++ = y4;
				*stackPos++ = x4;
				*stackPos++ = m3y;
				*stackPos++ = m3x;
				*stackPos++ = m5y;
				*stackPos++ = m5x;
				*stackPos++ = Path::cubicMarker;

				*stackPos++ = (m4y + m5y) * 0.5f;
				*stackPos++ = (m4x + m5x) * 0.5f;
				*stackPos++ = m4y;
				*stackPos++ = m4x;
				*stackPos++ = m1y;
				*stackPos++ = m1x;
				*stackPos++ = Path::cubicMarker;
			}
			else
			{
				*stackPos++ = y4;
				*stackPos++ = x4;
				*stackPos++ = Path::lineMarker;

				*stackPos++ = m5y;
				*stackPos++ = m5x;
				*stackPos++ = Path::lineMarker;

				*stackPos++ = m4y;
				*stackPos++ = m4x;
				*stackPos++ = Path::lineMarker;
			}
		}
		else if (type == Path::closeSubPathMarker)
		{
			if (x2 != subPathCloseX || y2 != subPathCloseY)
			{
				x1 = x2;
				y1 = y2;
				x2 = subPathCloseX;
				y2 = subPathCloseY;
				closesSubPath = true;

				return true;
			}
		}
		else
		{
			jassert (type == Path::moveMarker);

			subPathIndex = -1;
			subPathCloseX = x1 = x2;
			subPathCloseY = y1 = y2;
		}
	}
}

#if JUCE_MSVC && JUCE_DEBUG
  #pragma optimize ("", on)  // resets optimisations to the project defaults
#endif

END_JUCE_NAMESPACE

/*** End of inlined file: juce_PathIterator.cpp ***/


/*** Start of inlined file: juce_PathStrokeType.cpp ***/
BEGIN_JUCE_NAMESPACE

PathStrokeType::PathStrokeType (const float strokeThickness,
								const JointStyle jointStyle_,
								const EndCapStyle endStyle_) noexcept
	: thickness (strokeThickness),
	  jointStyle (jointStyle_),
	  endStyle (endStyle_)
{
}

PathStrokeType::PathStrokeType (const PathStrokeType& other) noexcept
	: thickness (other.thickness),
	  jointStyle (other.jointStyle),
	  endStyle (other.endStyle)
{
}

PathStrokeType& PathStrokeType::operator= (const PathStrokeType& other) noexcept
{
	thickness = other.thickness;
	jointStyle = other.jointStyle;
	endStyle = other.endStyle;
	return *this;
}

PathStrokeType::~PathStrokeType() noexcept
{
}

bool PathStrokeType::operator== (const PathStrokeType& other) const noexcept
{
	return thickness == other.thickness
		&& jointStyle == other.jointStyle
		&& endStyle == other.endStyle;
}

bool PathStrokeType::operator!= (const PathStrokeType& other) const noexcept
{
	return ! operator== (other);
}

namespace PathStrokeHelpers
{
	bool lineIntersection (const float x1, const float y1,
						   const float x2, const float y2,
						   const float x3, const float y3,
						   const float x4, const float y4,
						   float& intersectionX,
						   float& intersectionY,
						   float& distanceBeyondLine1EndSquared) noexcept
	{
		if (x2 != x3 || y2 != y3)
		{
			const float dx1 = x2 - x1;
			const float dy1 = y2 - y1;
			const float dx2 = x4 - x3;
			const float dy2 = y4 - y3;
			const float divisor = dx1 * dy2 - dx2 * dy1;

			if (divisor == 0)
			{
				if (! ((dx1 == 0 && dy1 == 0) || (dx2 == 0 && dy2 == 0)))
				{
					if (dy1 == 0 && dy2 != 0)
					{
						const float along = (y1 - y3) / dy2;
						intersectionX = x3 + along * dx2;
						intersectionY = y1;

						distanceBeyondLine1EndSquared = intersectionX - x2;
						distanceBeyondLine1EndSquared *= distanceBeyondLine1EndSquared;
						if ((x2 > x1) == (intersectionX < x2))
							distanceBeyondLine1EndSquared = -distanceBeyondLine1EndSquared;

						return along >= 0 && along <= 1.0f;
					}
					else if (dy2 == 0 && dy1 != 0)
					{
						const float along = (y3 - y1) / dy1;
						intersectionX = x1 + along * dx1;
						intersectionY = y3;

						distanceBeyondLine1EndSquared = (along - 1.0f) * dx1;
						distanceBeyondLine1EndSquared *= distanceBeyondLine1EndSquared;
						if (along < 1.0f)
							distanceBeyondLine1EndSquared = -distanceBeyondLine1EndSquared;

						return along >= 0 && along <= 1.0f;
					}
					else if (dx1 == 0 && dx2 != 0)
					{
						const float along = (x1 - x3) / dx2;
						intersectionX = x1;
						intersectionY = y3 + along * dy2;

						distanceBeyondLine1EndSquared = intersectionY - y2;
						distanceBeyondLine1EndSquared *= distanceBeyondLine1EndSquared;

						if ((y2 > y1) == (intersectionY < y2))
							distanceBeyondLine1EndSquared = -distanceBeyondLine1EndSquared;

						return along >= 0 && along <= 1.0f;
					}
					else if (dx2 == 0 && dx1 != 0)
					{
						const float along = (x3 - x1) / dx1;
						intersectionX = x3;
						intersectionY = y1 + along * dy1;

						distanceBeyondLine1EndSquared = (along - 1.0f) * dy1;
						distanceBeyondLine1EndSquared *= distanceBeyondLine1EndSquared;
						if (along < 1.0f)
							distanceBeyondLine1EndSquared = -distanceBeyondLine1EndSquared;

						return along >= 0 && along <= 1.0f;
					}
				}

				intersectionX = 0.5f * (x2 + x3);
				intersectionY = 0.5f * (y2 + y3);

				distanceBeyondLine1EndSquared = 0.0f;
				return false;
			}
			else
			{
				const float along1 = ((y1 - y3) * dx2 - (x1 - x3) * dy2) / divisor;

				intersectionX = x1 + along1 * dx1;
				intersectionY = y1 + along1 * dy1;

				if (along1 >= 0 && along1 <= 1.0f)
				{
					const float along2 = ((y1 - y3) * dx1 - (x1 - x3) * dy1);

					if (along2 >= 0 && along2 <= divisor)
					{
						distanceBeyondLine1EndSquared = 0.0f;
						return true;
					}
				}

				distanceBeyondLine1EndSquared = along1 - 1.0f;
				distanceBeyondLine1EndSquared *= distanceBeyondLine1EndSquared;
				distanceBeyondLine1EndSquared *= (dx1 * dx1 + dy1 * dy1);

				if (along1 < 1.0f)
					distanceBeyondLine1EndSquared = -distanceBeyondLine1EndSquared;

				return false;
			}
		}

		intersectionX = x2;
		intersectionY = y2;

		distanceBeyondLine1EndSquared = 0.0f;
		return true;
	}

	void addEdgeAndJoint (Path& destPath,
						  const PathStrokeType::JointStyle style,
						  const float maxMiterExtensionSquared, const float width,
						  const float x1, const float y1,
						  const float x2, const float y2,
						  const float x3, const float y3,
						  const float x4, const float y4,
						  const float midX, const float midY)
	{
		if (style == PathStrokeType::beveled
			|| (x3 == x4 && y3 == y4)
			|| (x1 == x2 && y1 == y2))
		{
			destPath.lineTo (x2, y2);
			destPath.lineTo (x3, y3);
		}
		else
		{
			float jx, jy, distanceBeyondLine1EndSquared;

			// if they intersect, use this point..
			if (lineIntersection (x1, y1, x2, y2,
								  x3, y3, x4, y4,
								  jx, jy, distanceBeyondLine1EndSquared))
			{
				destPath.lineTo (jx, jy);
			}
			else
			{
				if (style == PathStrokeType::mitered)
				{
					if (distanceBeyondLine1EndSquared < maxMiterExtensionSquared
						&& distanceBeyondLine1EndSquared > 0.0f)
					{
						destPath.lineTo (jx, jy);
					}
					else
					{
						// the end sticks out too far, so just use a blunt joint
						destPath.lineTo (x2, y2);
						destPath.lineTo (x3, y3);
					}
				}
				else
				{
					// curved joints
					float angle1 = std::atan2 (x2 - midX, y2 - midY);
					float angle2 = std::atan2 (x3 - midX, y3 - midY);
					const float angleIncrement = 0.1f;

					destPath.lineTo (x2, y2);

					if (std::abs (angle1 - angle2) > angleIncrement)
					{
						if (angle2 > angle1 + float_Pi
							 || (angle2 < angle1 && angle2 >= angle1 - float_Pi))
						{
							if (angle2 > angle1)
								angle2 -= float_Pi * 2.0f;

							jassert (angle1 <= angle2 + float_Pi);

							angle1 -= angleIncrement;
							while (angle1 > angle2)
							{
								destPath.lineTo (midX + width * std::sin (angle1),
												 midY + width * std::cos (angle1));

								angle1 -= angleIncrement;
							}
						}
						else
						{
							if (angle1 > angle2)
								angle1 -= float_Pi * 2.0f;

							jassert (angle1 >= angle2 - float_Pi);

							angle1 += angleIncrement;
							while (angle1 < angle2)
							{
								destPath.lineTo (midX + width * std::sin (angle1),
												 midY + width * std::cos (angle1));

								angle1 += angleIncrement;
							}
						}
					}

					destPath.lineTo (x3, y3);
				}
			}
		}
	}

	void addLineEnd (Path& destPath,
					 const PathStrokeType::EndCapStyle style,
					 const float x1, const float y1,
					 const float x2, const float y2,
					 const float width)
	{
		if (style == PathStrokeType::butt)
		{
			destPath.lineTo (x2, y2);
		}
		else
		{
			float offx1, offy1, offx2, offy2;

			float dx = x2 - x1;
			float dy = y2 - y1;
			const float len = juce_hypot (dx, dy);

			if (len == 0)
			{
				offx1 = offx2 = x1;
				offy1 = offy2 = y1;
			}
			else
			{
				const float offset = width / len;
				dx *= offset;
				dy *= offset;

				offx1 = x1 + dy;
				offy1 = y1 - dx;
				offx2 = x2 + dy;
				offy2 = y2 - dx;
			}

			if (style == PathStrokeType::square)
			{
				// sqaure ends
				destPath.lineTo (offx1, offy1);
				destPath.lineTo (offx2, offy2);
				destPath.lineTo (x2, y2);
			}
			else
			{
				// rounded ends
				const float midx = (offx1 + offx2) * 0.5f;
				const float midy = (offy1 + offy2) * 0.5f;

				destPath.cubicTo (x1 + (offx1 - x1) * 0.55f, y1 + (offy1 - y1) * 0.55f,
								  offx1 + (midx - offx1) * 0.45f, offy1 + (midy - offy1) * 0.45f,
								  midx, midy);

				destPath.cubicTo (midx + (offx2 - midx) * 0.55f, midy + (offy2 - midy) * 0.55f,
								  offx2 + (x2 - offx2) * 0.45f, offy2 + (y2 - offy2) * 0.45f,
								  x2, y2);
			}
		}
	}

	struct Arrowhead
	{
		float startWidth, startLength;
		float endWidth, endLength;
	};

	void addArrowhead (Path& destPath,
					   const float x1, const float y1,
					   const float x2, const float y2,
					   const float tipX, const float tipY,
					   const float width,
					   const float arrowheadWidth)
	{
		Line<float> line (x1, y1, x2, y2);
		destPath.lineTo (line.getPointAlongLine (-(arrowheadWidth / 2.0f - width), 0));
		destPath.lineTo (tipX, tipY);
		destPath.lineTo (line.getPointAlongLine (arrowheadWidth - (arrowheadWidth / 2.0f - width), 0));
		destPath.lineTo (x2, y2);
	}

	struct LineSection
	{
		float x1, y1, x2, y2;	  // original line
		float lx1, ly1, lx2, ly2;  // the left-hand stroke
		float rx1, ry1, rx2, ry2;  // the right-hand stroke
	};

	void shortenSubPath (Array<LineSection>& subPath, float amountAtStart, float amountAtEnd)
	{
		while (amountAtEnd > 0 && subPath.size() > 0)
		{
			LineSection& l = subPath.getReference (subPath.size() - 1);
			float dx = l.rx2 - l.rx1;
			float dy = l.ry2 - l.ry1;
			const float len = juce_hypot (dx, dy);

			if (len <= amountAtEnd && subPath.size() > 1)
			{
				LineSection& prev = subPath.getReference (subPath.size() - 2);
				prev.x2 = l.x2;
				prev.y2 = l.y2;
				subPath.removeLast();
				amountAtEnd -= len;
			}
			else
			{
				const float prop = jmin (0.9999f, amountAtEnd / len);
				dx *= prop;
				dy *= prop;
				l.rx1 += dx;
				l.ry1 += dy;
				l.lx2 += dx;
				l.ly2 += dy;
				break;
			}
		}

		while (amountAtStart > 0 && subPath.size() > 0)
		{
			LineSection& l = subPath.getReference (0);
			float dx = l.rx2 - l.rx1;
			float dy = l.ry2 - l.ry1;
			const float len = juce_hypot (dx, dy);

			if (len <= amountAtStart && subPath.size() > 1)
			{
				LineSection& next = subPath.getReference (1);
				next.x1 = l.x1;
				next.y1 = l.y1;
				subPath.remove (0);
				amountAtStart -= len;
			}
			else
			{
				const float prop = jmin (0.9999f, amountAtStart / len);
				dx *= prop;
				dy *= prop;
				l.rx2 -= dx;
				l.ry2 -= dy;
				l.lx1 -= dx;
				l.ly1 -= dy;
				break;
			}
		}
	}

	void addSubPath (Path& destPath, Array<LineSection>& subPath,
					 const bool isClosed, const float width, const float maxMiterExtensionSquared,
					 const PathStrokeType::JointStyle jointStyle, const PathStrokeType::EndCapStyle endStyle,
					 const Arrowhead* const arrowhead)
	{
		jassert (subPath.size() > 0);

		if (arrowhead != nullptr)
			shortenSubPath (subPath, arrowhead->startLength, arrowhead->endLength);

		const LineSection& firstLine = subPath.getReference (0);

		float lastX1 = firstLine.lx1;
		float lastY1 = firstLine.ly1;
		float lastX2 = firstLine.lx2;
		float lastY2 = firstLine.ly2;

		if (isClosed)
		{
			destPath.startNewSubPath (lastX1, lastY1);
		}
		else
		{
			destPath.startNewSubPath (firstLine.rx2, firstLine.ry2);

			if (arrowhead != nullptr)
				addArrowhead (destPath, firstLine.rx2, firstLine.ry2, lastX1, lastY1, firstLine.x1, firstLine.y1,
							  width, arrowhead->startWidth);
			else
				addLineEnd (destPath, endStyle, firstLine.rx2, firstLine.ry2, lastX1, lastY1, width);
		}

		int i;
		for (i = 1; i < subPath.size(); ++i)
		{
			const LineSection& l = subPath.getReference (i);

			addEdgeAndJoint (destPath, jointStyle,
							 maxMiterExtensionSquared, width,
							 lastX1, lastY1, lastX2, lastY2,
							 l.lx1, l.ly1, l.lx2, l.ly2,
							 l.x1, l.y1);

			lastX1 = l.lx1;
			lastY1 = l.ly1;
			lastX2 = l.lx2;
			lastY2 = l.ly2;
		}

		const LineSection& lastLine = subPath.getReference (subPath.size() - 1);

		if (isClosed)
		{
			const LineSection& l = subPath.getReference (0);

			addEdgeAndJoint (destPath, jointStyle,
							 maxMiterExtensionSquared, width,
							 lastX1, lastY1, lastX2, lastY2,
							 l.lx1, l.ly1, l.lx2, l.ly2,
							 l.x1, l.y1);

			destPath.closeSubPath();
			destPath.startNewSubPath (lastLine.rx1, lastLine.ry1);
		}
		else
		{
			destPath.lineTo (lastX2, lastY2);

			if (arrowhead != nullptr)
				addArrowhead (destPath, lastX2, lastY2, lastLine.rx1, lastLine.ry1, lastLine.x2, lastLine.y2,
							  width, arrowhead->endWidth);
			else
				addLineEnd (destPath, endStyle, lastX2, lastY2, lastLine.rx1, lastLine.ry1, width);
		}

		lastX1 = lastLine.rx1;
		lastY1 = lastLine.ry1;
		lastX2 = lastLine.rx2;
		lastY2 = lastLine.ry2;

		for (i = subPath.size() - 1; --i >= 0;)
		{
			const LineSection& l = subPath.getReference (i);

			addEdgeAndJoint (destPath, jointStyle,
							 maxMiterExtensionSquared, width,
							 lastX1, lastY1, lastX2, lastY2,
							 l.rx1, l.ry1, l.rx2, l.ry2,
							 l.x2, l.y2);

			lastX1 = l.rx1;
			lastY1 = l.ry1;
			lastX2 = l.rx2;
			lastY2 = l.ry2;
		}

		if (isClosed)
		{
			addEdgeAndJoint (destPath, jointStyle,
							 maxMiterExtensionSquared, width,
							 lastX1, lastY1, lastX2, lastY2,
							 lastLine.rx1, lastLine.ry1, lastLine.rx2, lastLine.ry2,
							 lastLine.x2, lastLine.y2);
		}
		else
		{
			// do the last line
			destPath.lineTo (lastX2, lastY2);
		}

		destPath.closeSubPath();
	}

	void createStroke (const float thickness, const PathStrokeType::JointStyle jointStyle,
					   const PathStrokeType::EndCapStyle endStyle,
					   Path& destPath, const Path& source,
					   const AffineTransform& transform,
					   const float extraAccuracy, const Arrowhead* const arrowhead)
	{
		jassert (extraAccuracy > 0);

		if (thickness <= 0)
		{
			destPath.clear();
			return;
		}

		const Path* sourcePath = &source;
		Path temp;

		if (sourcePath == &destPath)
		{
			destPath.swapWithPath (temp);
			sourcePath = &temp;
		}
		else
		{
			destPath.clear();
		}

		destPath.setUsingNonZeroWinding (true);

		const float maxMiterExtensionSquared = 9.0f * thickness * thickness;
		const float width = 0.5f * thickness;

		// Iterate the path, creating a list of the
		// left/right-hand lines along either side of it...
		PathFlatteningIterator it (*sourcePath, transform, PathFlatteningIterator::defaultTolerance / extraAccuracy);

		Array <LineSection> subPath;
		subPath.ensureStorageAllocated (512);
		LineSection l;
		l.x1 = 0;
		l.y1 = 0;

		const float minSegmentLength = 0.0001f;

		while (it.next())
		{
			if (it.subPathIndex == 0)
			{
				if (subPath.size() > 0)
				{
					addSubPath (destPath, subPath, false, width, maxMiterExtensionSquared, jointStyle, endStyle, arrowhead);
					subPath.clearQuick();
				}

				l.x1 = it.x1;
				l.y1 = it.y1;
			}

			l.x2 = it.x2;
			l.y2 = it.y2;

			float dx = l.x2 - l.x1;
			float dy = l.y2 - l.y1;

			const float hypotSquared = dx*dx + dy*dy;

			if (it.closesSubPath || hypotSquared > minSegmentLength || it.isLastInSubpath())
			{
				const float len = std::sqrt (hypotSquared);

				if (len == 0)
				{
					l.rx1 = l.rx2 = l.lx1 = l.lx2 = l.x1;
					l.ry1 = l.ry2 = l.ly1 = l.ly2 = l.y1;
				}
				else
				{
					const float offset = width / len;
					dx *= offset;
					dy *= offset;

					l.rx2 = l.x1 - dy;
					l.ry2 = l.y1 + dx;
					l.lx1 = l.x1 + dy;
					l.ly1 = l.y1 - dx;

					l.lx2 = l.x2 + dy;
					l.ly2 = l.y2 - dx;
					l.rx1 = l.x2 - dy;
					l.ry1 = l.y2 + dx;
				}

				subPath.add (l);

				if (it.closesSubPath)
				{
					addSubPath (destPath, subPath, true, width, maxMiterExtensionSquared, jointStyle, endStyle, arrowhead);
					subPath.clearQuick();
				}
				else
				{
					l.x1 = it.x2;
					l.y1 = it.y2;
				}
			}
		}

		if (subPath.size() > 0)
			addSubPath (destPath, subPath, false, width, maxMiterExtensionSquared, jointStyle, endStyle, arrowhead);
	}
}

void PathStrokeType::createStrokedPath (Path& destPath, const Path& sourcePath,
										const AffineTransform& transform, const float extraAccuracy) const
{
	PathStrokeHelpers::createStroke (thickness, jointStyle, endStyle, destPath, sourcePath,
									 transform, extraAccuracy, 0);
}

void PathStrokeType::createDashedStroke (Path& destPath,
										 const Path& sourcePath,
										 const float* dashLengths,
										 int numDashLengths,
										 const AffineTransform& transform,
										 const float extraAccuracy) const
{
	jassert (extraAccuracy > 0);

	if (thickness <= 0)
		return;

	// this should really be an even number..
	jassert ((numDashLengths & 1) == 0);

	Path newDestPath;
	PathFlatteningIterator it (sourcePath, transform, PathFlatteningIterator::defaultTolerance / extraAccuracy);

	bool first = true;
	int dashNum = 0;
	float pos = 0.0f, lineLen = 0.0f, lineEndPos = 0.0f;
	float dx = 0.0f, dy = 0.0f;

	for (;;)
	{
		const bool isSolid = ((dashNum & 1) == 0);
		const float dashLen = dashLengths [dashNum++ % numDashLengths];

		jassert (dashLen > 0); // must be a positive increment!
		if (dashLen <= 0)
			break;

		pos += dashLen;

		while (pos > lineEndPos)
		{
			if (! it.next())
			{
				if (isSolid && ! first)
					newDestPath.lineTo (it.x2, it.y2);

				createStrokedPath (destPath, newDestPath, AffineTransform::identity, extraAccuracy);
				return;
			}

			if (isSolid && ! first)
				newDestPath.lineTo (it.x1, it.y1);
			else
				newDestPath.startNewSubPath (it.x1, it.y1);

			dx = it.x2 - it.x1;
			dy = it.y2 - it.y1;
			lineLen = juce_hypot (dx, dy);
			lineEndPos += lineLen;
			first = it.closesSubPath;
		}

		const float alpha = (pos - (lineEndPos - lineLen)) / lineLen;

		if (isSolid)
			newDestPath.lineTo (it.x1 + dx * alpha,
								it.y1 + dy * alpha);
		else
			newDestPath.startNewSubPath (it.x1 + dx * alpha,
										 it.y1 + dy * alpha);
	}
}

void PathStrokeType::createStrokeWithArrowheads (Path& destPath,
												 const Path& sourcePath,
												 const float arrowheadStartWidth, const float arrowheadStartLength,
												 const float arrowheadEndWidth, const float arrowheadEndLength,
												 const AffineTransform& transform,
												 const float extraAccuracy) const
{
	PathStrokeHelpers::Arrowhead head;
	head.startWidth = arrowheadStartWidth;
	head.startLength = arrowheadStartLength;
	head.endWidth = arrowheadEndWidth;
	head.endLength = arrowheadEndLength;

	PathStrokeHelpers::createStroke (thickness, jointStyle, endStyle,
									 destPath, sourcePath, transform, extraAccuracy, &head);
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_PathStrokeType.cpp ***/


/*** Start of inlined file: juce_RectangleList.cpp ***/
BEGIN_JUCE_NAMESPACE

RectangleList::RectangleList() noexcept
{
}

RectangleList::RectangleList (const Rectangle<int>& rect)
{
	if (! rect.isEmpty())
		rects.add (rect);
}

RectangleList::RectangleList (const RectangleList& other)
	: rects (other.rects)
{
}

RectangleList& RectangleList::operator= (const RectangleList& other)
{
	rects = other.rects;
	return *this;
}

RectangleList::~RectangleList()
{
}

void RectangleList::clear()
{
	rects.clearQuick();
}

Rectangle<int> RectangleList::getRectangle (const int index) const noexcept
{
	if (isPositiveAndBelow (index, rects.size()))
		return rects.getReference (index);

	return Rectangle<int>();
}

bool RectangleList::isEmpty() const noexcept
{
	return rects.size() == 0;
}

RectangleList::Iterator::Iterator (const RectangleList& list) noexcept
	: current (nullptr),
	  owner (list),
	  index (list.rects.size())
{
}

RectangleList::Iterator::~Iterator()
{
}

bool RectangleList::Iterator::next() noexcept
{
	if (--index >= 0)
	{
		current = & (owner.rects.getReference (index));
		return true;
	}

	return false;
}

void RectangleList::add (const Rectangle<int>& rect)
{
	if (! rect.isEmpty())
	{
		if (rects.size() == 0)
		{
			rects.add (rect);
		}
		else
		{
			bool anyOverlaps = false;

			int i;
			for (i = rects.size(); --i >= 0;)
			{
				Rectangle<int>& ourRect = rects.getReference (i);

				if (rect.intersects (ourRect))
				{
					if (rect.contains (ourRect))
						rects.remove (i);
					else if (! ourRect.reduceIfPartlyContainedIn (rect))
						anyOverlaps = true;
				}
			}

			if (anyOverlaps && rects.size() > 0)
			{
				RectangleList r (rect);

				for (i = rects.size(); --i >= 0;)
				{
					const Rectangle<int>& ourRect = rects.getReference (i);

					if (rect.intersects (ourRect))
					{
						r.subtract (ourRect);

						if (r.rects.size() == 0)
							return;
					}
				}

				for (i = r.getNumRectangles(); --i >= 0;)
					rects.add (r.rects.getReference (i));
			}
			else
			{
				rects.add (rect);
			}
		}
	}
}

void RectangleList::addWithoutMerging (const Rectangle<int>& rect)
{
	if (! rect.isEmpty())
		rects.add (rect);
}

void RectangleList::add (const int x, const int y, const int w, const int h)
{
	if (rects.size() == 0)
	{
		if (w > 0 && h > 0)
			rects.add (Rectangle<int> (x, y, w, h));
	}
	else
	{
		add (Rectangle<int> (x, y, w, h));
	}
}

void RectangleList::add (const RectangleList& other)
{
	for (int i = 0; i < other.rects.size(); ++i)
		add (other.rects.getReference (i));
}

void RectangleList::subtract (const Rectangle<int>& rect)
{
	const int originalNumRects = rects.size();

	if (originalNumRects > 0)
	{
		const int x1 = rect.x;
		const int y1 = rect.y;
		const int x2 = x1 + rect.w;
		const int y2 = y1 + rect.h;

		for (int i = getNumRectangles(); --i >= 0;)
		{
			Rectangle<int>& r = rects.getReference (i);

			const int rx1 = r.x;
			const int ry1 = r.y;
			const int rx2 = rx1 + r.w;
			const int ry2 = ry1 + r.h;

			if (! (x2 <= rx1 || x1 >= rx2 || y2 <= ry1 || y1 >= ry2))
			{
				if (x1 > rx1 && x1 < rx2)
				{
					if (y1 <= ry1 && y2 >= ry2 && x2 >= rx2)
					{
						r.w = x1 - rx1;
					}
					else
					{
						r.x = x1;
						r.w = rx2 - x1;

						rects.insert (i + 1, Rectangle<int> (rx1, ry1, x1 - rx1,  ry2 - ry1));
						i += 2;
					}
				}
				else if (x2 > rx1 && x2 < rx2)
				{
					r.x = x2;
					r.w = rx2 - x2;

					if (y1 > ry1 || y2 < ry2 || x1 > rx1)
					{
						rects.insert (i + 1, Rectangle<int> (rx1, ry1, x2 - rx1,  ry2 - ry1));
						i += 2;
					}
				}
				else if (y1 > ry1 && y1 < ry2)
				{
					if (x1 <= rx1 && x2 >= rx2 && y2 >= ry2)
					{
						r.h = y1 - ry1;
					}
					else
					{
						r.y = y1;
						r.h = ry2 - y1;

						rects.insert (i + 1, Rectangle<int> (rx1, ry1, rx2 - rx1, y1 - ry1));
						i += 2;
					}
				}
				else if (y2 > ry1 && y2 < ry2)
				{
					r.y = y2;
					r.h = ry2 - y2;

					if (x1 > rx1 || x2 < rx2 || y1 > ry1)
					{
						rects.insert (i + 1, Rectangle<int> (rx1, ry1, rx2 - rx1, y2 - ry1));
						i += 2;
					}
				}
				else
				{
					rects.remove (i);
				}
			}
		}
	}
}

bool RectangleList::subtract (const RectangleList& otherList)
{
	for (int i = otherList.rects.size(); --i >= 0 && rects.size() > 0;)
		subtract (otherList.rects.getReference (i));

	return rects.size() > 0;
}

bool RectangleList::clipTo (const Rectangle<int>& rect)
{
	bool notEmpty = false;

	if (rect.isEmpty())
	{
		clear();
	}
	else
	{
		for (int i = rects.size(); --i >= 0;)
		{
			Rectangle<int>& r = rects.getReference (i);

			if (! rect.intersectRectangle (r.x, r.y, r.w, r.h))
				rects.remove (i);
			else
				notEmpty = true;
		}
	}

	return notEmpty;
}

bool RectangleList::clipTo (const RectangleList& other)
{
	if (rects.size() == 0)
		return false;

	RectangleList result;

	for (int j = 0; j < rects.size(); ++j)
	{
		const Rectangle<int>& rect = rects.getReference (j);

		for (int i = other.rects.size(); --i >= 0;)
		{
			Rectangle<int> r (other.rects.getReference (i));

			if (rect.intersectRectangle (r.x, r.y, r.w, r.h))
				result.rects.add (r);
		}
	}

	swapWith (result);

	return ! isEmpty();
}

bool RectangleList::getIntersectionWith (const Rectangle<int>& rect, RectangleList& destRegion) const
{
	destRegion.clear();

	if (! rect.isEmpty())
	{
		for (int i = rects.size(); --i >= 0;)
		{
			Rectangle<int> r (rects.getReference (i));

			if (rect.intersectRectangle (r.x, r.y, r.w, r.h))
				destRegion.rects.add (r);
		}
	}

	return destRegion.rects.size() > 0;
}

void RectangleList::swapWith (RectangleList& otherList) noexcept
{
	rects.swapWithArray (otherList.rects);
}

void RectangleList::consolidate()
{
	int i;
	for (i = 0; i < getNumRectangles() - 1; ++i)
	{
		Rectangle<int>& r = rects.getReference (i);
		const int rx1 = r.x;
		const int ry1 = r.y;
		const int rx2 = rx1 + r.w;
		const int ry2 = ry1 + r.h;

		for (int j = rects.size(); --j > i;)
		{
			Rectangle<int>& r2 = rects.getReference (j);
			const int jrx1 = r2.x;
			const int jry1 = r2.y;
			const int jrx2 = jrx1 + r2.w;
			const int jry2 = jry1 + r2.h;

			// if the vertical edges of any blocks are touching and their horizontals don't
			// line up, split them horizontally..
			if (jrx1 == rx2 || jrx2 == rx1)
			{
				if (jry1 > ry1 && jry1 < ry2)
				{
					r.h = jry1 - ry1;
					rects.add (Rectangle<int> (rx1, jry1, rx2 - rx1, ry2 - jry1));
					i = -1;
					break;
				}

				if (jry2 > ry1 && jry2 < ry2)
				{
					r.h = jry2 - ry1;
					rects.add (Rectangle<int> (rx1, jry2, rx2 - rx1, ry2 - jry2));
					i = -1;
					break;
				}
				else if (ry1 > jry1 && ry1 < jry2)
				{
					r2.h = ry1 - jry1;
					rects.add (Rectangle<int> (jrx1, ry1, jrx2 - jrx1, jry2 - ry1));
					i = -1;
					break;
				}
				else if (ry2 > jry1 && ry2 < jry2)
				{
					r2.h = ry2 - jry1;
					rects.add (Rectangle<int> (jrx1, ry2, jrx2 - jrx1, jry2 - ry2));
					i = -1;
					break;
				}
			}
		}
	}

	for (i = 0; i < rects.size() - 1; ++i)
	{
		Rectangle<int>& r = rects.getReference (i);

		for (int j = rects.size(); --j > i;)
		{
			if (r.enlargeIfAdjacent (rects.getReference (j)))
			{
				rects.remove (j);
				i = -1;
				break;
			}
		}
	}
}

bool RectangleList::containsPoint (const int x, const int y) const noexcept
{
	for (int i = getNumRectangles(); --i >= 0;)
		if (rects.getReference (i).contains (x, y))
			return true;

	return false;
}

bool RectangleList::containsRectangle (const Rectangle<int>& rectangleToCheck) const
{
	if (rects.size() > 1)
	{
		RectangleList r (rectangleToCheck);

		for (int i = rects.size(); --i >= 0;)
		{
			r.subtract (rects.getReference (i));

			if (r.rects.size() == 0)
				return true;
		}
	}
	else if (rects.size() > 0)
	{
		return rects.getReference (0).contains (rectangleToCheck);
	}

	return false;
}

bool RectangleList::intersectsRectangle (const Rectangle<int>& rectangleToCheck) const noexcept
{
	for (int i = rects.size(); --i >= 0;)
		if (rects.getReference (i).intersects (rectangleToCheck))
			return true;

	return false;
}

bool RectangleList::intersects (const RectangleList& other) const noexcept
{
	for (int i = rects.size(); --i >= 0;)
		if (other.intersectsRectangle (rects.getReference (i)))
			return true;

	return false;
}

Rectangle<int> RectangleList::getBounds() const noexcept
{
	if (rects.size() <= 1)
	{
		if (rects.size() == 0)
			return Rectangle<int>();
		else
			return rects.getReference (0);
	}
	else
	{
		const Rectangle<int>& r = rects.getReference (0);

		int minX = r.x;
		int minY = r.y;
		int maxX = minX + r.w;
		int maxY = minY + r.h;

		for (int i = rects.size(); --i > 0;)
		{
			const Rectangle<int>& r2 = rects.getReference (i);

			minX = jmin (minX, r2.x);
			minY = jmin (minY, r2.y);
			maxX = jmax (maxX, r2.getRight());
			maxY = jmax (maxY, r2.getBottom());
		}

		return Rectangle<int> (minX, minY, maxX - minX, maxY - minY);
	}
}

void RectangleList::offsetAll (const int dx, const int dy) noexcept
{
	for (int i = rects.size(); --i >= 0;)
	{
		Rectangle<int>& r = rects.getReference (i);

		r.x += dx;
		r.y += dy;
	}
}

Path RectangleList::toPath() const
{
	Path p;

	for (int i = rects.size(); --i >= 0;)
	{
		const Rectangle<int>& r = rects.getReference (i);

		p.addRectangle ((float) r.x,
						(float) r.y,
						(float) r.w,
						(float) r.h);
	}

	return p;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_RectangleList.cpp ***/


/*** Start of inlined file: juce_Image.cpp ***/
BEGIN_JUCE_NAMESPACE

Image::SharedImage::SharedImage (const PixelFormat format_, const int width_, const int height_)
	: format (format_), width (width_), height (height_)
{
	jassert (format_ == RGB || format_ == ARGB || format_ == SingleChannel);
	jassert (width > 0 && height > 0); // It's illegal to create a zero-sized image!
}

Image::SharedImage::~SharedImage()
{
}

class SoftwareSharedImage  : public Image::SharedImage
{
public:
	SoftwareSharedImage (const Image::PixelFormat format_, const int width_, const int height_, const bool clearImage)
		: Image::SharedImage (format_, width_, height_),
		  pixelStride (format_ == Image::RGB ? 3 : ((format_ == Image::ARGB) ? 4 : 1)),
		  lineStride ((pixelStride * jmax (1, width_) + 3) & ~3)
	{
		imageData.allocate (lineStride * jmax (1, height_), clearImage);
	}

	Image::ImageType getType() const
	{
		return Image::SoftwareImage;
	}

	LowLevelGraphicsContext* createLowLevelContext()
	{
		return new LowLevelGraphicsSoftwareRenderer (Image (this));
	}

	void initialiseBitmapData (Image::BitmapData& bitmap, int x, int y, Image::BitmapData::ReadWriteMode /*mode*/)
	{
		bitmap.data = imageData + x * pixelStride + y * lineStride;
		bitmap.pixelFormat = format;
		bitmap.lineStride = lineStride;
		bitmap.pixelStride = pixelStride;
	}

	Image::SharedImage* clone()
	{
		SoftwareSharedImage* s = new SoftwareSharedImage (format, width, height, false);
		memcpy (s->imageData, imageData, lineStride * height);
		return s;
	}

private:
	HeapBlock<uint8> imageData;
	const int pixelStride, lineStride;

	JUCE_LEAK_DETECTOR (SoftwareSharedImage);
};

Image::SharedImage* Image::SharedImage::createSoftwareImage (Image::PixelFormat format, int width, int height, bool clearImage)
{
	return new SoftwareSharedImage (format, width, height, clearImage);
}

class SubsectionSharedImage  : public Image::SharedImage
{
public:
	SubsectionSharedImage (Image::SharedImage* const image_, const Rectangle<int>& area_)
		: Image::SharedImage (image_->getPixelFormat(), area_.getWidth(), area_.getHeight()),
		  image (image_), area (area_)
	{
	}

	Image::ImageType getType() const
	{
		return Image::SoftwareImage;
	}

	LowLevelGraphicsContext* createLowLevelContext()
	{
		LowLevelGraphicsContext* g = image->createLowLevelContext();
		g->clipToRectangle (area);
		g->setOrigin (area.getX(), area.getY());
		return g;
	}

	void initialiseBitmapData (Image::BitmapData& bitmap, int x, int y, Image::BitmapData::ReadWriteMode mode)
	{
		image->initialiseBitmapData (bitmap, x + area.getX(), y + area.getY(), mode);
	}

	Image::SharedImage* clone()
	{
		return new SubsectionSharedImage (image->clone(), area);
	}

private:
	const ReferenceCountedObjectPtr<Image::SharedImage> image;
	const Rectangle<int> area;

	JUCE_DECLARE_NON_COPYABLE_WITH_LEAK_DETECTOR (SubsectionSharedImage);
};

Image Image::getClippedImage (const Rectangle<int>& area) const
{
	if (area.contains (getBounds()))
		return *this;

	const Rectangle<int> validArea (area.getIntersection (getBounds()));
	if (validArea.isEmpty())
		return Image::null;

	return Image (new SubsectionSharedImage (image, validArea));
}

Image::Image()
{
}

Image::Image (SharedImage* const instance)
	: image (instance)
{
}

Image::Image (const PixelFormat format,
			  const int width, const int height,
			  const bool clearImage, const ImageType type)
	: image (type == Image::NativeImage ? SharedImage::createNativeImage (format, width, height, clearImage)
										: new SoftwareSharedImage (format, width, height, clearImage))
{
}

Image::Image (const Image& other)
   : image (other.image)
{
}

Image& Image::operator= (const Image& other)
{
	image = other.image;
	return *this;
}

Image::~Image()
{
}

const Image Image::null;

LowLevelGraphicsContext* Image::createLowLevelContext() const
{
	return image == nullptr ? nullptr : image->createLowLevelContext();
}

void Image::duplicateIfShared()
{
	if (image != nullptr && image->getReferenceCount() > 1)
		image = image->clone();
}

Image Image::rescaled (const int newWidth, const int newHeight, const Graphics::ResamplingQuality quality) const
{
	if (image == nullptr || (image->width == newWidth && image->height == newHeight))
		return *this;

	Image newImage (image->format, newWidth, newHeight, hasAlphaChannel(), image->getType());

	Graphics g (newImage);
	g.setImageResamplingQuality (quality);
	g.drawImage (*this, 0, 0, newWidth, newHeight, 0, 0, image->width, image->height, false);

	return newImage;
}

Image Image::convertedToFormat (PixelFormat newFormat) const
{
	if (image == nullptr || newFormat == image->format)
		return *this;

	const int w = image->width, h = image->height;
	Image newImage (newFormat, w, h, false, image->getType());

	if (newFormat == SingleChannel)
	{
		if (! hasAlphaChannel())
		{
			newImage.clear (getBounds(), Colours::black);
		}
		else
		{
			const BitmapData destData (newImage, 0, 0, w, h, BitmapData::writeOnly);
			const BitmapData srcData (*this, 0, 0, w, h);

			for (int y = 0; y < h; ++y)
			{
				const PixelARGB* src = (const PixelARGB*) srcData.getLinePointer(y);
				uint8* dst = destData.getLinePointer (y);

				for (int x = w; --x >= 0;)
				{
					*dst++ = src->getAlpha();
					++src;
				}
			}
		}
	}
	else
	{
		if (hasAlphaChannel())
			newImage.clear (getBounds());

		Graphics g (newImage);
		g.drawImageAt (*this, 0, 0);
	}

	return newImage;
}

NamedValueSet* Image::getProperties() const
{
	return image == nullptr ? nullptr : &(image->userData);
}

Image::BitmapData::BitmapData (Image& image, const int x, const int y, const int w, const int h, BitmapData::ReadWriteMode mode)
	: width (w),
	  height (h)
{
	// The BitmapData class must be given a valid image, and a valid rectangle within it!
	jassert (image.image != nullptr);
	jassert (x >= 0 && y >= 0 && w > 0 && h > 0 && x + w <= image.getWidth() && y + h <= image.getHeight());

	image.image->initialiseBitmapData (*this, x, y, mode);
	jassert (data != nullptr && pixelStride > 0 && lineStride != 0);
}

Image::BitmapData::BitmapData (const Image& image, const int x, const int y, const int w, const int h)
	: width (w),
	  height (h)
{
	// The BitmapData class must be given a valid image, and a valid rectangle within it!
	jassert (image.image != nullptr);
	jassert (x >= 0 && y >= 0 && w > 0 && h > 0 && x + w <= image.getWidth() && y + h <= image.getHeight());

	image.image->initialiseBitmapData (*this, x, y, readOnly);
	jassert (data != nullptr && pixelStride > 0 && lineStride != 0);
}

Image::BitmapData::BitmapData (const Image& image, BitmapData::ReadWriteMode mode)
	: width (image.getWidth()),
	  height (image.getHeight())
{
	// The BitmapData class must be given a valid image!
	jassert (image.image != nullptr);

	image.image->initialiseBitmapData (*this, 0, 0, mode);
	jassert (data != nullptr && pixelStride > 0 && lineStride != 0);
}

Image::BitmapData::~BitmapData()
{
}

const Colour Image::BitmapData::getPixelColour (const int x, const int y) const noexcept
{
	jassert (isPositiveAndBelow (x, width) && isPositiveAndBelow (y, height));

	const uint8* const pixel = getPixelPointer (x, y);

	switch (pixelFormat)
	{
		case Image::ARGB:	   return Colour (((const PixelARGB*) pixel)->getUnpremultipliedARGB());
		case Image::RGB:		return Colour (((const PixelRGB*) pixel)->getUnpremultipliedARGB());
		case Image::SingleChannel:  return Colour (((const PixelAlpha*) pixel)->getUnpremultipliedARGB());
		default:			jassertfalse; break;
	}

	return Colour();
}

void Image::BitmapData::setPixelColour (const int x, const int y, const Colour& colour) const noexcept
{
	jassert (isPositiveAndBelow (x, width) && isPositiveAndBelow (y, height));

	uint8* const pixel = getPixelPointer (x, y);
	const PixelARGB col (colour.getPixelARGB());

	switch (pixelFormat)
	{
		case Image::ARGB:	   ((PixelARGB*) pixel)->set (col); break;
		case Image::RGB:		((PixelRGB*) pixel)->set (col); break;
		case Image::SingleChannel:  *pixel = col.getAlpha(); break;
		default:			jassertfalse; break;
	}
}

void Image::setPixelData (int x, int y, int w, int h,
						  const uint8* const sourcePixelData, const int sourceLineStride)
{
	jassert (x >= 0 && y >= 0 && w > 0 && h > 0 && x + w <= getWidth() && y + h <= getHeight());

	if (Rectangle<int>::intersectRectangles (x, y, w, h, 0, 0, getWidth(), getHeight()))
	{
		const BitmapData dest (*this, x, y, w, h, BitmapData::writeOnly);

		for (int i = 0; i < h; ++i)
		{
			memcpy (dest.getLinePointer(i),
					sourcePixelData + sourceLineStride * i,
					w * dest.pixelStride);
		}
	}
}

void Image::clear (const Rectangle<int>& area, const Colour& colourToClearTo)
{
	const Rectangle<int> clipped (area.getIntersection (getBounds()));

	if (! clipped.isEmpty())
	{
		const PixelARGB col (colourToClearTo.getPixelARGB());

		const BitmapData destData (*this, clipped.getX(), clipped.getY(), clipped.getWidth(), clipped.getHeight(), BitmapData::writeOnly);
		uint8* dest = destData.data;
		int dh = clipped.getHeight();

		while (--dh >= 0)
		{
			uint8* line = dest;
			dest += destData.lineStride;

			if (isARGB())
			{
				for (int x = clipped.getWidth(); --x >= 0;)
				{
					((PixelARGB*) line)->set (col);
					line += destData.pixelStride;
				}
			}
			else if (isRGB())
			{
				for (int x = clipped.getWidth(); --x >= 0;)
				{
					((PixelRGB*) line)->set (col);
					line += destData.pixelStride;
				}
			}
			else
			{
				for (int x = clipped.getWidth(); --x >= 0;)
				{
					*line = col.getAlpha();
					line += destData.pixelStride;
				}
			}
		}
	}
}

const Colour Image::getPixelAt (const int x, const int y) const
{
	if (isPositiveAndBelow (x, getWidth()) && isPositiveAndBelow (y, getHeight()))
	{
		const BitmapData srcData (*this, x, y, 1, 1);
		return srcData.getPixelColour (0, 0);
	}

	return Colour();
}

void Image::setPixelAt (const int x, const int y, const Colour& colour)
{
	if (isPositiveAndBelow (x, getWidth()) && isPositiveAndBelow (y, getHeight()))
	{
		const BitmapData destData (*this, x, y, 1, 1, BitmapData::writeOnly);
		destData.setPixelColour (0, 0, colour);
	}
}

void Image::multiplyAlphaAt (const int x, const int y, const float multiplier)
{
	if (isPositiveAndBelow (x, getWidth()) && isPositiveAndBelow (y, getHeight())
		 && hasAlphaChannel())
	{
		const BitmapData destData (*this, x, y, 1, 1, BitmapData::readWrite);

		if (isARGB())
			((PixelARGB*) destData.data)->multiplyAlpha (multiplier);
		else
			*(destData.data) = (uint8) (*(destData.data) * multiplier);
	}
}

void Image::multiplyAllAlphas (const float amountToMultiplyBy)
{
	if (hasAlphaChannel())
	{
		const BitmapData destData (*this, 0, 0, getWidth(), getHeight(), BitmapData::readWrite);

		if (isARGB())
		{
			for (int y = 0; y < destData.height; ++y)
			{
				uint8* p = destData.getLinePointer (y);

				for (int x = 0; x < destData.width; ++x)
				{
					((PixelARGB*) p)->multiplyAlpha (amountToMultiplyBy);
					p += destData.pixelStride;
				}
			}
		}
		else
		{
			for (int y = 0; y < destData.height; ++y)
			{
				uint8* p = destData.getLinePointer (y);

				for (int x = 0; x < destData.width; ++x)
				{
					*p = (uint8) (*p * amountToMultiplyBy);
					p += destData.pixelStride;
				}
			}
		}
	}
	else
	{
		jassertfalse; // can't do this without an alpha-channel!
	}
}

void Image::desaturate()
{
	if (isARGB() || isRGB())
	{
		const BitmapData destData (*this, 0, 0, getWidth(), getHeight(), BitmapData::readWrite);

		if (isARGB())
		{
			for (int y = 0; y < destData.height; ++y)
			{
				uint8* p = destData.getLinePointer (y);

				for (int x = 0; x < destData.width; ++x)
				{
					((PixelARGB*) p)->desaturate();
					p += destData.pixelStride;
				}
			}
		}
		else
		{
			for (int y = 0; y < destData.height; ++y)
			{
				uint8* p = destData.getLinePointer (y);

				for (int x = 0; x < destData.width; ++x)
				{
					((PixelRGB*) p)->desaturate();
					p += destData.pixelStride;
				}
			}
		}
	}
}

void Image::createSolidAreaMask (RectangleList& result, const float alphaThreshold) const
{
	if (hasAlphaChannel())
	{
		const uint8 threshold = (uint8) jlimit (0, 255, roundToInt (alphaThreshold * 255.0f));
		SparseSet<int> pixelsOnRow;

		const BitmapData srcData (*this, 0, 0, getWidth(), getHeight());

		for (int y = 0; y < srcData.height; ++y)
		{
			pixelsOnRow.clear();
			const uint8* lineData = srcData.getLinePointer (y);

			if (isARGB())
			{
				for (int x = 0; x < srcData.width; ++x)
				{
					if (((const PixelARGB*) lineData)->getAlpha() >= threshold)
						pixelsOnRow.addRange (Range<int> (x, x + 1));

					lineData += srcData.pixelStride;
				}
			}
			else
			{
				for (int x = 0; x < srcData.width; ++x)
				{
					if (*lineData >= threshold)
						pixelsOnRow.addRange (Range<int> (x, x + 1));

					lineData += srcData.pixelStride;
				}
			}

			for (int i = 0; i < pixelsOnRow.getNumRanges(); ++i)
			{
				const Range<int> range (pixelsOnRow.getRange (i));
				result.add (Rectangle<int> (range.getStart(), y, range.getLength(), 1));
			}

			result.consolidate();
		}
	}
	else
	{
		result.add (0, 0, getWidth(), getHeight());
	}
}

void Image::moveImageSection (int dx, int dy,
							  int sx, int sy,
							  int w, int h)
{
	if (dx < 0)
	{
		w += dx;
		sx -= dx;
		dx = 0;
	}

	if (dy < 0)
	{
		h += dy;
		sy -= dy;
		dy = 0;
	}

	if (sx < 0)
	{
		w += sx;
		dx -= sx;
		sx = 0;
	}

	if (sy < 0)
	{
		h += sy;
		dy -= sy;
		sy = 0;
	}

	const int minX = jmin (dx, sx);
	const int minY = jmin (dy, sy);

	w = jmin (w, getWidth() - jmax (sx, dx));
	h = jmin (h, getHeight() - jmax (sy, dy));

	if (w > 0 && h > 0)
	{
		const int maxX = jmax (dx, sx) + w;
		const int maxY = jmax (dy, sy) + h;

		const BitmapData destData (*this, minX, minY, maxX - minX, maxY - minY, BitmapData::readWrite);

		uint8* dst	   = destData.getPixelPointer (dx - minX, dy - minY);
		const uint8* src = destData.getPixelPointer (sx - minX, sy - minY);

		const int lineSize = destData.pixelStride * w;

		if (dy > sy)
		{
			while (--h >= 0)
			{
				const int offset = h * destData.lineStride;
				memmove (dst + offset, src + offset, lineSize);
			}
		}
		else if (dst != src)
		{
			while (--h >= 0)
			{
				memmove (dst, src, lineSize);
				dst += destData.lineStride;
				src += destData.lineStride;
			}
		}
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_Image.cpp ***/


/*** Start of inlined file: juce_ImageCache.cpp ***/
BEGIN_JUCE_NAMESPACE

class ImageCache::Pimpl	 : public Timer,
							  public DeletedAtShutdown
{
public:
	Pimpl()
		: cacheTimeout (5000)
	{
	}

	~Pimpl()
	{
		clearSingletonInstance();
	}

	Image getFromHashCode (const int64 hashCode)
	{
		const ScopedLock sl (lock);

		for (int i = images.size(); --i >= 0;)
		{
			Item* const item = images.getUnchecked(i);

			if (item->hashCode == hashCode)
				return item->image;
		}

		return Image::null;
	}

	void addImageToCache (const Image& image, const int64 hashCode)
	{
		if (image.isValid())
		{
			if (! isTimerRunning())
				startTimer (2000);

			Item* const item = new Item();
			item->hashCode = hashCode;
			item->image = image;
			item->lastUseTime = Time::getApproximateMillisecondCounter();

			const ScopedLock sl (lock);
			images.add (item);
		}
	}

	void timerCallback()
	{
		const uint32 now = Time::getApproximateMillisecondCounter();

		const ScopedLock sl (lock);

		for (int i = images.size(); --i >= 0;)
		{
			Item* const item = images.getUnchecked(i);

			if (item->image.getReferenceCount() <= 1)
			{
				if (now > item->lastUseTime + cacheTimeout || now < item->lastUseTime - 1000)
					images.remove (i);
			}
			else
			{
				item->lastUseTime = now; // multiply-referenced, so this image is still in use.
			}
		}

		if (images.size() == 0)
			stopTimer();
	}

	struct Item
	{
		Image image;
		int64 hashCode;
		uint32 lastUseTime;
	};

	int cacheTimeout;

	juce_DeclareSingleton_SingleThreaded_Minimal (ImageCache::Pimpl);

private:
	OwnedArray<Item> images;
	CriticalSection lock;

	JUCE_DECLARE_NON_COPYABLE (Pimpl);
};

juce_ImplementSingleton_SingleThreaded (ImageCache::Pimpl);

Image ImageCache::getFromHashCode (const int64 hashCode)
{
	if (Pimpl::getInstanceWithoutCreating() != nullptr)
		return Pimpl::getInstanceWithoutCreating()->getFromHashCode (hashCode);

	return Image::null;
}

void ImageCache::addImageToCache (const Image& image, const int64 hashCode)
{
	Pimpl::getInstance()->addImageToCache (image, hashCode);
}

Image ImageCache::getFromFile (const File& file)
{
	const int64 hashCode = file.hashCode64();
	Image image (getFromHashCode (hashCode));

	if (image.isNull())
	{
		image = ImageFileFormat::loadFrom (file);
		addImageToCache (image, hashCode);
	}

	return image;
}

Image ImageCache::getFromMemory (const void* imageData, const int dataSize)
{
	const int64 hashCode = (int64) (pointer_sized_int) imageData;
	Image image (getFromHashCode (hashCode));

	if (image.isNull())
	{
		image = ImageFileFormat::loadFrom (imageData, dataSize);
		addImageToCache (image, hashCode);
	}

	return image;
}

void ImageCache::setCacheTimeout (const int millisecs)
{
	Pimpl::getInstance()->cacheTimeout = millisecs;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ImageCache.cpp ***/


/*** Start of inlined file: juce_ImageConvolutionKernel.cpp ***/
BEGIN_JUCE_NAMESPACE

ImageConvolutionKernel::ImageConvolutionKernel (const int size_)
	: values (size_ * size_),
	  size (size_)
{
	clear();
}

ImageConvolutionKernel::~ImageConvolutionKernel()
{
}

float ImageConvolutionKernel::getKernelValue (const int x, const int y) const noexcept
{
	if (isPositiveAndBelow (x, size) && isPositiveAndBelow (y, size))
		return values [x + y * size];

	jassertfalse;
	return 0;
}

void ImageConvolutionKernel::setKernelValue (const int x, const int y, const float value) noexcept
{
	if (isPositiveAndBelow (x, size) && isPositiveAndBelow (y, size))
	{
		values [x + y * size] = value;
	}
	else
	{
		jassertfalse;
	}
}

void ImageConvolutionKernel::clear()
{
	for (int i = size * size; --i >= 0;)
		values[i] = 0;
}

void ImageConvolutionKernel::setOverallSum (const float desiredTotalSum)
{
	double currentTotal = 0.0;

	for (int i = size * size; --i >= 0;)
		currentTotal += values[i];

	rescaleAllValues ((float) (desiredTotalSum / currentTotal));
}

void ImageConvolutionKernel::rescaleAllValues (const float multiplier)
{
	for (int i = size * size; --i >= 0;)
		values[i] *= multiplier;
}

void ImageConvolutionKernel::createGaussianBlur (const float radius)
{
	const double radiusFactor = -1.0 / (radius * radius * 2);
	const int centre = size >> 1;

	for (int y = size; --y >= 0;)
	{
		for (int x = size; --x >= 0;)
		{
			const int cx = x - centre;
			const int cy = y - centre;

			values [x + y * size] = (float) exp (radiusFactor * (cx * cx + cy * cy));
		}
	}

	setOverallSum (1.0f);
}

void ImageConvolutionKernel::applyToImage (Image& destImage,
										   const Image& sourceImage,
										   const Rectangle<int>& destinationArea) const
{
	if (sourceImage == destImage)
	{
		destImage.duplicateIfShared();
	}
	else
	{
		if (sourceImage.getWidth() != destImage.getWidth()
			 || sourceImage.getHeight() != destImage.getHeight()
			 || sourceImage.getFormat() != destImage.getFormat())
		{
			jassertfalse;
			return;
		}
	}

	const Rectangle<int> area (destinationArea.getIntersection (destImage.getBounds()));

	if (area.isEmpty())
		return;

	const int right = area.getRight();
	const int bottom = area.getBottom();

	const Image::BitmapData destData (destImage, area.getX(), area.getY(), area.getWidth(), area.getHeight(),
									  Image::BitmapData::writeOnly);
	uint8* line = destData.data;

	const Image::BitmapData srcData (sourceImage, Image::BitmapData::readOnly);

	if (destData.pixelStride == 4)
	{
		for (int y = area.getY(); y < bottom; ++y)
		{
			uint8* dest = line;
			line += destData.lineStride;

			for (int x = area.getX(); x < right; ++x)
			{
				float c1 = 0;
				float c2 = 0;
				float c3 = 0;
				float c4 = 0;

				for (int yy = 0; yy < size; ++yy)
				{
					const int sy = y + yy - (size >> 1);

					if (sy >= srcData.height)
						break;

					if (sy >= 0)
					{
						int sx = x - (size >> 1);
						const uint8* src = srcData.getPixelPointer (sx, sy);

						for (int xx = 0; xx < size; ++xx)
						{
							if (sx >= srcData.width)
								break;

							if (sx >= 0)
							{
								const float kernelMult = values [xx + yy * size];
								c1 += kernelMult * *src++;
								c2 += kernelMult * *src++;
								c3 += kernelMult * *src++;
								c4 += kernelMult * *src++;
							}
							else
							{
								src += 4;
							}

							++sx;
						}
					}
				}

				*dest++ = (uint8) jmin (0xff, roundToInt (c1));
				*dest++ = (uint8) jmin (0xff, roundToInt (c2));
				*dest++ = (uint8) jmin (0xff, roundToInt (c3));
				*dest++ = (uint8) jmin (0xff, roundToInt (c4));
			}
		}
	}
	else if (destData.pixelStride == 3)
	{
		for (int y = area.getY(); y < bottom; ++y)
		{
			uint8* dest = line;
			line += destData.lineStride;

			for (int x = area.getX(); x < right; ++x)
			{
				float c1 = 0;
				float c2 = 0;
				float c3 = 0;

				for (int yy = 0; yy < size; ++yy)
				{
					const int sy = y + yy - (size >> 1);

					if (sy >= srcData.height)
						break;

					if (sy >= 0)
					{
						int sx = x - (size >> 1);
						const uint8* src = srcData.getPixelPointer (sx, sy);

						for (int xx = 0; xx < size; ++xx)
						{
							if (sx >= srcData.width)
								break;

							if (sx >= 0)
							{
								const float kernelMult = values [xx + yy * size];
								c1 += kernelMult * *src++;
								c2 += kernelMult * *src++;
								c3 += kernelMult * *src++;
							}
							else
							{
								src += 3;
							}

							++sx;
						}
					}
				}

				*dest++ = (uint8) roundToInt (c1);
				*dest++ = (uint8) roundToInt (c2);
				*dest++ = (uint8) roundToInt (c3);
			}
		}
	}
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ImageConvolutionKernel.cpp ***/


/*** Start of inlined file: juce_ImageFileFormat.cpp ***/
BEGIN_JUCE_NAMESPACE

struct DefaultImageFormats
{
	PNGImageFormat  png;
	JPEGImageFormat jpg;
	GIFImageFormat  gif;
};

static DefaultImageFormats defaultImageFormats;

ImageFileFormat* ImageFileFormat::findImageFormatForStream (InputStream& input)
{
	ImageFileFormat* formats[] = { &defaultImageFormats.png,
								   &defaultImageFormats.jpg,
								   &defaultImageFormats.gif };

	const int64 streamPos = input.getPosition();

	for (int i = 0; i < numElementsInArray (formats); ++i)
	{
		const bool found = formats[i]->canUnderstand (input);
		input.setPosition (streamPos);

		if (found)
			return formats[i];
	}

	return nullptr;
}

Image ImageFileFormat::loadFrom (InputStream& input)
{
	ImageFileFormat* const format = findImageFormatForStream (input);

	if (format != nullptr)
		return format->decodeImage (input);

	return Image::null;
}

Image ImageFileFormat::loadFrom (const File& file)
{
	InputStream* const in = file.createInputStream();

	if (in != nullptr)
	{
		BufferedInputStream b (in, 8192, true);
		return loadFrom (b);
	}

	return Image::null;
}

Image ImageFileFormat::loadFrom (const void* rawData, const int numBytes)
{
	if (rawData != nullptr && numBytes > 4)
	{
		MemoryInputStream stream (rawData, numBytes, false);
		return loadFrom (stream);
	}

	return Image::null;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_ImageFileFormat.cpp ***/


/*** Start of inlined file: juce_GIFLoader.cpp ***/
BEGIN_JUCE_NAMESPACE

#if (JUCE_MAC || JUCE_IOS) && USE_COREGRAPHICS_RENDERING && ! DONT_USE_COREIMAGE_LOADER
 Image juce_loadWithCoreImage (InputStream& input);
#else

class GIFLoader
{
public:
	GIFLoader (InputStream& in)
		: input (in),
		  dataBlockIsZero (false), fresh (false), finished (false),
		  currentBit (0), lastBit (0), lastByteIndex (0),
		  codeSize (0), setCodeSize (0), maxCode (0), maxCodeSize (0),
		  firstcode (0), oldcode (0), clearCode (0), endCode (0)
	{
		int imageWidth, imageHeight;
		if (! getSizeFromHeader (imageWidth, imageHeight))
			return;

		uint8 buf [16];
		if (in.read (buf, 3) != 3)
			return;

		int numColours = 2 << (buf[0] & 7);
		int transparent = -1;

		if ((buf[0] & 0x80) != 0)
			readPalette (numColours);

		for (;;)
		{
			if (input.read (buf, 1) != 1 || buf[0] == ';')
				break;

			if (buf[0] == '!')
			{
				if (readExtension (transparent))
					continue;

				break;
			}

			if (buf[0] != ',')
				continue;

			if (input.read (buf, 9) == 9)
			{
				imageWidth  = (int) ByteOrder::littleEndianShort (buf + 4);
				imageHeight = (int) ByteOrder::littleEndianShort (buf + 6);

				numColours = 2 << (buf[8] & 7);

				if ((buf[8] & 0x80) != 0)
					if (! readPalette (numColours))
						break;

				image = Image (transparent >= 0 ? Image::ARGB : Image::RGB,
							   imageWidth, imageHeight, transparent >= 0);

				image.getProperties()->set ("originalImageHadAlpha", transparent >= 0);

				readImage ((buf[8] & 0x40) != 0, transparent);
			}

			break;
		}
	}

	Image image;

private:
	InputStream& input;
	uint8 buffer [260];
	PixelARGB palette [256];
	bool dataBlockIsZero, fresh, finished;
	int currentBit, lastBit, lastByteIndex;
	int codeSize, setCodeSize;
	int maxCode, maxCodeSize;
	int firstcode, oldcode;
	int clearCode, endCode;
	enum { maxGifCode = 1 << 12 };
	int table [2] [maxGifCode];
	int stack [2 * maxGifCode];
	int* sp;

	bool getSizeFromHeader (int& w, int& h)
	{
		char b[6];

		if (input.read (b, 6) == 6
			 && (strncmp ("GIF87a", b, 6) == 0
				  || strncmp ("GIF89a", b, 6) == 0))
		{
			if (input.read (b, 4) == 4)
			{
				w = (int) ByteOrder::littleEndianShort (b);
				h = (int) ByteOrder::littleEndianShort (b + 2);
				return w > 0 && h > 0;
			}
		}

		return false;
	}

	bool readPalette (const int numCols)
	{
		for (int i = 0; i < numCols; ++i)
		{
			uint8 rgb[4];
			input.read (rgb, 3);

			palette[i].setARGB (0xff, rgb[0], rgb[1], rgb[2]);
			palette[i].premultiply();
		}

		return true;
	}

	int readDataBlock (uint8* const dest)
	{
		uint8 n;
		if (input.read (&n, 1) == 1)
		{
			dataBlockIsZero = (n == 0);

			if (dataBlockIsZero || (input.read (dest, n) == n))
				return n;
		}

		return -1;
	}

	int readExtension (int& transparent)
	{
		uint8 type;
		if (input.read (&type, 1) != 1)
			return false;

		uint8 b [260];
		int n = 0;

		if (type == 0xf9)
		{
			n = readDataBlock (b);
			if (n < 0)
				return 1;

			if ((b[0] & 1) != 0)
				transparent = b[3];
		}

		do
		{
			n = readDataBlock (b);
		}
		while (n > 0);

		return n >= 0;
	}

	void clearTable()
	{
		int i;
		for (i = 0; i < clearCode; ++i)
		{
			table[0][i] = 0;
			table[1][i] = i;
		}

		for (; i < maxGifCode; ++i)
		{
			table[0][i] = 0;
			table[1][i] = 0;
		}
	}

	void initialise (const int inputCodeSize)
	{
		setCodeSize = inputCodeSize;
		codeSize = setCodeSize + 1;
		clearCode = 1 << setCodeSize;
		endCode = clearCode + 1;
		maxCodeSize = 2 * clearCode;
		maxCode = clearCode + 2;

		getCode (0, true);

		fresh = true;
		clearTable();
		sp = stack;
	}

	int readLZWByte()
	{
		if (fresh)
		{
			fresh = false;

			do
			{
				firstcode = oldcode
					= getCode (codeSize, false);
			}
			while (firstcode == clearCode);

			return firstcode;
		}

		if (sp > stack)
			return *--sp;

		int code;

		while ((code = getCode (codeSize, false)) >= 0)
		{
			if (code == clearCode)
			{
				clearTable();
				codeSize = setCodeSize + 1;
				maxCodeSize = 2 * clearCode;
				maxCode = clearCode + 2;
				sp = stack;
				firstcode = oldcode = getCode (codeSize, false);
				return firstcode;
			}
			else if (code == endCode)
			{
				if (dataBlockIsZero)
					return -2;

				uint8 buf [260];

				int n;
				while ((n = readDataBlock (buf)) > 0)
				{}

				if (n != 0)
					return -2;
			}

			const int incode = code;

			if (code >= maxCode)
			{
				*sp++ = firstcode;
				code = oldcode;
			}

			while (code >= clearCode)
			{
				*sp++ = table[1][code];
				if (code == table[0][code])
					return -2;

				code = table[0][code];
			}

			*sp++ = firstcode = table[1][code];

			if ((code = maxCode) < maxGifCode)
			{
				table[0][code] = oldcode;
				table[1][code] = firstcode;
				++maxCode;

				if ((maxCode >= maxCodeSize)
					&& (maxCodeSize < maxGifCode))
				{
					maxCodeSize <<= 1;
					++codeSize;
				}
			}

			oldcode = incode;

			if (sp > stack)
				return *--sp;
		}

		return code;
	}

	int getCode (const int codeSize_, const bool shouldInitialise)
	{
		if (shouldInitialise)
		{
			currentBit = 0;
			lastBit = 0;
			finished = false;
			return 0;
		}

		if ((currentBit + codeSize_) >= lastBit)
		{
			if (finished)
				return -1;

			buffer[0] = buffer [lastByteIndex - 2];
			buffer[1] = buffer [lastByteIndex - 1];

			const int n = readDataBlock (buffer + 2);

			if (n == 0)
				finished = true;

			lastByteIndex = 2 + n;
			currentBit = (currentBit - lastBit) + 16;
			lastBit = (2 + n) * 8 ;
		}

		int result = 0;
		int i = currentBit;

		for (int j = 0; j < codeSize_; ++j)
		{
			result |= ((buffer[i >> 3] & (1 << (i & 7))) != 0) << j;
			++i;
		}

		currentBit += codeSize_;

		return result;
	}

	bool readImage (const int interlace, const int transparent)
	{
		uint8 c;
		if (input.read (&c, 1) != 1)
			return false;

		initialise (c);

		if (transparent >= 0)
			palette [transparent].setARGB (0, 0, 0, 0);

		int xpos = 0, ypos = 0, yStep = 8, pass = 0;

		const Image::BitmapData destData (image, Image::BitmapData::writeOnly);
		uint8* p = destData.getPixelPointer (0, 0);
		const bool hasAlpha = image.hasAlphaChannel();

		for (;;)
		{
			const int index = readLZWByte();
			if (index < 0)
				break;

			if (hasAlpha)
				((PixelARGB*) p)->set (palette [index]);
			else
				((PixelRGB*)  p)->set (palette [index]);

			p += destData.pixelStride;

			if (++xpos == destData.width)
			{
				xpos = 0;

				if (interlace)
				{
					ypos += yStep;

					while (ypos >= destData.height)
					{
						++pass;

						switch (pass)
						{
							case 1:	 ypos = 4; yStep = 8; break;
							case 2:	 ypos = 2; yStep = 4; break;
							case 3:	 ypos = 1; yStep = 2; break;
							default:	return true;
						}
					}
				}
				else
				{
					if (++ypos >= destData.height)
						break;
				}

				p = destData.getPixelPointer (xpos, ypos);
			}
		}

		return true;
	}

	JUCE_DECLARE_NON_COPYABLE (GIFLoader);
};

#endif

GIFImageFormat::GIFImageFormat() {}
GIFImageFormat::~GIFImageFormat() {}

String GIFImageFormat::getFormatName()    { return "GIF"; }

bool GIFImageFormat::canUnderstand (InputStream& in)
{
	char header [4];

	return (in.read (header, sizeof (header)) == sizeof (header))
			 && header[0] == 'G'
			 && header[1] == 'I'
			 && header[2] == 'F';
}

Image GIFImageFormat::decodeImage (InputStream& in)
{
  #if (JUCE_MAC || JUCE_IOS) && USE_COREGRAPHICS_RENDERING && ! DONT_USE_COREIMAGE_LOADER
	return juce_loadWithCoreImage (in);
  #else
	const ScopedPointer <GIFLoader> loader (new GIFLoader (in));
	return loader->image;
  #endif
}

bool GIFImageFormat::writeImageToStream (const Image& /*sourceImage*/, OutputStream& /*destStream*/)
{
	jassertfalse; // writing isn't implemented for GIFs!
	return false;
}

END_JUCE_NAMESPACE

/*** End of inlined file: juce_GIFLoader.cpp ***/

#endif
